/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

//#ifdef HAVE_CONFIG_H
//#include "config.h"
//#endif

#include <getopt.h>

#include "ClicApplication.hpp"
#include "nmm/utils/net/NetReactor.hpp"
#include "nmm/utils/time/Rational.hpp"
#include "nmm/comm/ProxyObject.hpp"
#include "nmm/comm/messaging/TEDObject.hpp"
#include "nmm/comm/serialize/SerializeResult.hpp"
#include "nmm/multimedia/interfaces/ITrack.hpp"
#include "nmm/multimedia/interfaces/ITrackList.hpp"
#include "nmm/plugins/interfaces/IFileHandler.hpp"
#include "nmm/plugins/file/interfaces/ISinkFileHandler.hpp"
#include "nmm/plugins/file/interfaces/ISeekable.hpp"
#include "nmm/plugins/video/display/interfaces/IDisplayEvents.hpp"
#include "nmm/plugins/video/display/interfaces/IDisplayEventSender.hpp"
#include "nmm/plugins/audio/interfaces/IAudioDevice.hpp"
#include "nmm/plugins/video/display/interfaces/IVideoDisplay.hpp"
#include "nmm/services/utils/gdparse/GDParse.hpp"
#include "nmm/services/utils/gdparse/IClicValueStream.hpp"
#include "nmm/misc/graphurl/GraphURL.hpp"
#include "nmm/services/registry/ProxyApplication.hpp"
#include "nmm/services/registry/ISession.hpp"
#include "nmm/services/registry/IServerRegistry.hpp"
#include "nmm/services/graphbuilder/GraphBuilder.hpp"

#include "nmm/utils/directory_utils.hpp"

#include "clic.hpp"

#include "nmm/services/DefaultSinks.hpp"

extern unsigned int lineno;
extern char *yytext;

namespace NMM {
  ClicApplication::ClicApplication(int argc, char** argv)
    :  NamedObject("Clic-Application"), 
       __app(0), 
       __graph_url(), 
       __composite(0),
       __seekable(0),
       __playback(0),
       __playback_location(),
       __display(0),
       __display_location(),
       __audio_sink_name(),
       __video_sink_name(),
       __sink_synchronizer(0),
       __synchronizer(0),
       __curr_input_file(0),
       __sync(true),
       __prefetching(false),
       __eof_counter(0),
       __sink_nodes(0), 
       __fullscreen(false), 
       __delay(0),
       __infinite_loop(true),
       __loop(false),
       __timeout(0),
       __timer(*this), 
       __enable_registry(false), 
       __port(ClientRegistry::default_port),
       __paused(false),
       __x_listener(0)
  {
    /* Set message stream to cerr */
    setMessageStream(&cerr, NamedObject::ALL_LEVELS);
    setWarningStream(&cerr, NamedObject::ALL_LEVELS);
    setMessageColor(NamedObject::BLUE_COLOR, NamedObject::ALL_LEVELS);

    /* First parse the arguments to get all filenames and options */
    parseArguments(argc, argv);

    /* Begin> Create an NMMAplication */
    __app = ProxyApplication::getApplication(argc, argv);

    if (__enable_registry) {
      MESSAGE_STREAM("Starting server registry on port " << __port << endl);
      __app->enableServerRegistry(__port);
    }

    if( __graph_url.empty() ) {
      /* request the graph */
      __app->getRegistry().requestGraph(__gd);
      setupGraph();
    }
    else {
      buildGraph();
    }

    if (__timeout > 0) {
      __timer.start(__timeout);
    }
  }
  
  ClicApplication::~ClicApplication() {
    /* release the node inside the graph */
    try {
      if( __graph_url.empty() ) {
	/* If synchronization was enabled, disconnect the syncronizer */
	if (__sync) {
	  __gd.disconnectSinkSynchronizer();
        }
	
	/* release the graph and all nodes */
	__app->getRegistry().releaseGraph(__gd);
      }
      else {
	destroyGraph();
      }
    } catch (const Exception& e ) {
      cerr << e << endl;
    }
    delete __x_listener;
  }
  
  Result ClicApplication::keyReleased(string& key) {
    if(key == "KEY_f") { // 'f' to toggle fullscreen
      toggleFullscreen();
      return SUCCESS;
    } else if(key == "KEY_Right") {
      seekForward();
      return SUCCESS;
    } else if(key == "KEY_Left") {
      seekBackward();
      return SUCCESS;
    } else if(key == "KEY_q") {
      quit();
      return SUCCESS;
    } else if(key == "KEY_space") {
      togglePause();
      return SUCCESS;
    } else {
      cerr << "ClicApplication: Unsupported key pressed: " << key << endl;
      return FAILURE;
    }
  }

  void ClicApplication::seekForward() {
    if(__seekable) {
      if(__seekable->hasTimeSeek()) {
        __seekable->seekTime(Interval(10, 0));
      } else {
        __seekable->seekPercent(Rational(3,100));
      }
    } else {
      cerr << "Source not seekable" << endl;
    }
  }
  
  void ClicApplication::seekBackward() {
    if(__seekable) {
      if(__seekable->hasTimeSeek()) {
        __seekable->seekTime(Interval(-10, 0));
      } else {
        __seekable->seekPercent(Rational(-3,100));
      }
    } else {
      cerr << "Source not seekable" << endl;
    }
  }
  
  void ClicApplication::pause() {
    if(__synchronizer) {
      __synchronizer->pause();
    } else {
      if( __graph_url.empty() ) {
	__gd.stopGraph();
      }
      else {
	cerr << "No synchronizer to pause" << endl;
      }
    }
  }
  
  void ClicApplication::play() {
    if(__synchronizer) {
      __synchronizer->wakeup();
    } else {
      if( __graph_url.empty() ) {
	__gd.startGraph();
      }
      else {
	cerr << "No synchronizer to wakeup" << endl;
      }
    }
  }
  
  void ClicApplication::togglePause() {
    __paused = !__paused;
    if(__paused) {
      pause();
    } else {
      play();
    }
  }
  
  void ClicApplication::toggleFullscreen() {
    if(!__video_display.get()) {
      cerr << "ClicApplication: No VideoDisplay set!" << endl;
    }
    
    if(__video_display->hasFullscreen()) {
      __fullscreen = !__video_display->getFullscreen();
      __video_display->setFullscreen(__fullscreen);
    } else {
      cerr << "ClicApplication: Fullscreen not supported with current display!" << endl;
    }
  }
  
  void ClicApplication::quit() {
    /* interrupt console input */
    __console_input.interrupt();
  }
  
  void ClicApplication::parseArguments(int argc, char** argv) {
    int verbose = 0;

    const char *opt_string = "aDlLvV:rsPhi:WA:u:o:d:t:np:S"; 
    struct option options [] = {
      {"input-file",  1,NULL,'i'},
      {"url",         1,NULL,'u'},
      {"audio-output",1,NULL,'A'},
      {"video-output",1,NULL,'V'},
      {"audio-sink",  1,NULL,0x80},
      {"video-sink",  1,NULL,'V'},
      {"output-file", 1,NULL,'o'},
      {"verbose",     0,NULL,'v'},
      {"sync",        0,NULL,'s'},
      {"prefetching", 0,NULL,'P'},
      {"delay",       1,NULL,'d'},
      {"timeout",     1,NULL,'t'},
      {"loop",        0,NULL,'a'},
      {"Loop",        0,NULL,'L'},
      {"port",        1,NULL,'p'},
      {"non-interactive",   0,NULL,'n'},
      {"sever-registry",    0,NULL,'S'},
      {"license",     0,NULL,'l'},
      {"help",        0,NULL,'h'},
      {NULL,          0,NULL,0}
    };
    int ch,long_option_index;

    while((ch=getopt_long(argc,argv,opt_string,options,&long_option_index))!=EOF){
      // which option did we get?
      switch (ch) {
        case 'v':
          verbose++;
          break;
        case 'i':
	  //__input_files.push_back(NMM::getAbsolutePath(optarg));
	  __input_files.push_back(optarg);
	  break;
        case 'n':
          __console_input.setInteractive(false);
          break;
        case 'd':
          __delay = atoi(optarg);
          break;
        case 'o':
          __output_file = optarg;
          break;
        case 'h':
          usage(argv[0]);
          exit(0);
        case 's':
          WARNING_STREAM("Disable synchronizer" << endl);
          __sync = false;
          break;
        case 'P':
          __prefetching = true;
          break;
        case 'u':
          __graph_url = optarg;
          break;
        case 'A':
          __playback_location = optarg;
          break;
        case 'V':
          __display_location = optarg;
          break;
        case 'a':
          WARNING_STREAM("Disable loop" << endl);
          __infinite_loop = false;
          break;
        case 'L':
          WARNING_STREAM("Enable loop" << endl);
          __loop = true;
          break;	  
        case 't':
          __timeout = atoi(optarg);
          break;
        case 'S':
          __enable_registry = true;
          break;
        case 'l':        
           NamedObject::getGlobalInstance().setErrorStream(NULL, NamedObject::ALL_LEVELS);
           NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
           NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
           NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
           NMMApplication::printLicense("clic");
           exit(0);
        case 0x80:
          __audio_sink_name = optarg;
          break;
        case 0x81:
          __video_sink_name = optarg;
          break;
        case 'p' :
          __port = atoi(optarg);
          break;
        default:
          usage(argv[0]);
          throw Exception("Invalid option");
        }
    }

    /* handle *Stream output */
    if (verbose < 1) {
      NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
    }
    if (verbose < 2) {
      NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
    }
    if (verbose < 3) {
      NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
    }
    if (verbose < 4) {
      NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
    }
 
    if( __graph_url.empty() ) { // if no graph url was given 
      // check for filename
      if ( argc - optind < 1 ) {
	usage(argv[0]);
	string nmm_dir = NMM::getNMMDir();
	string error_string ;
	if (!nmm_dir.empty())
	  error_string = string( "Hint : Check directory ") + nmm_dir + string( "share/nmm/gd or directory ") + nmm_dir + string( "apps/clic/gd for graph descriptions.");
	
	throw Exception(string("No .gd file for a graph description set and no URL given\n") + error_string);
      }
      else {
	// first, check whether .gd file or other file was given ( = media file )
	string s = argv[optind];
	size_t dot_pos;
	if( ( (dot_pos = s.rfind(".")) != string::npos) && (s.substr(dot_pos, s.size()) != ".gd") ) {
	  // test, if valid url is given
	  GraphURL url(s);
	  if(url.isValid()) { // if yes, then use this url
	    __graph_url = s;
	    cerr << "No graph description file (.gd) given\n=> Running clic with URL : " << __graph_url << endl;
	  }
	  else { // otherwise, assume it is a file and add corresponding prefix
	    __graph_url = "file://" + s;
	    cerr << "No graph description file (.gd) given\n=> Running clic with file URL : " << __graph_url << endl;
	  }
	  if(__input_files.size()) {
	    cerr << "Ignoring additional input files" << endl;
	  }
	  if(__output_file.size()) {
	    cerr << "Ignoring additional output file" << endl;
	  }
	}
	else {
	  __gd_file = argv[optind];
	  /* create graph description from a file */
	  getGraphfromGD(__gd_file,__gd);
	}
      }
    }
    else { // some graph url was given -> do not use  .gd file
      if ( !(argc - optind < 1) ) {
	usage(argv[0]);
	throw Exception("Specifying url for graph builder and setting .gd file is not supported \n");
      }
    }
  }

  void ClicApplication::waitUntilKeyPressed(const char c) {
    MESSAGE_STREAM("Enter '" << c << "', to exit the application" << endl);
    MESSAGE_STREAM("Enter 'p', to pause current graph" << endl);
    MESSAGE_STREAM("Enter 's', to start a graph again" << endl);
    MESSAGE_STREAM("Enter 'n', to switch to next input file" << endl);
    
    while (1) {
      /* wait for input on console */
      char tmp = __console_input.waitUntilKeyPressed();
      
      /* if the waiting was interrupted, break out of the loop */
      if(tmp == 0) {
        break;
      }
      
      if (tmp == c) {
        cerr << "Exit... " << endl;
        // graph description is freed in desctructor
        break;
      }
      if (tmp == 'p' ) {
	pause();
        cerr << "Pause current graph. Enter 's' to start it again " << endl;
      }
      if (tmp == 's' ) {
	play();
        cerr << "Start current graph." << endl;
      }
      if (tmp == 'n' ) {
	if(__input_files.size() > 0) {
	  __curr_input_file++;
	  if(__curr_input_file >= __input_files.size() ) {
	    cerr << "reset input file to 0 " << endl;
	    __curr_input_file = 0;
	  }
	  cerr << "Switch to input file " << (__curr_input_file) << ": " << __input_files[__curr_input_file] << endl;
	  /* Begin> handle input files: */
	  SequenceValue<INode>& inode_list = __gd.getINodeList();
	  for (unsigned int i = 0; i< inode_list.size(); ++i) {
	    if(__gd.getInEdgeList(i).size() == 0) {
	      IFileHandler_var i_handler(inode_list[i]->getParentObject()->getInterface<IFileHandler>());
	      if (i_handler.get()) {
                cerr << "Change file at : " << __clock.getTime() << endl;	      
		inode_list[i]->reachInitialized();
		i_handler->setFilename(__input_files[__curr_input_file]);
		inode_list[i]->reachStarted();
		break; // currently only works for a single source
	      }
	    }
	  }
	  /* End< */
	}
      }
    }
  }

  void ClicApplication::nextInputFile() {
    if(__input_files.size() > 0) {
      __curr_input_file++;
      if(__curr_input_file >= __input_files.size()) {
        if(__loop) {
          cerr << "Loop: Start from first input file" << endl;
          __curr_input_file = 0;
        } else {
          cerr << "All files played" << endl;
          /* If we are not running in inifinte loop, set the exit flag to true and notify the main thread */
          if (!__infinite_loop) {
            /* quit application */
            quit();
          }
          /* End< */
          return;
        }
      }
      cerr << "Switch to input file " << (__curr_input_file) << ": " << __input_files[__curr_input_file] << endl;
      /* Begin> handle input files: */
      SequenceValue<INode>& inode_list = __gd.getINodeList();
      for (unsigned int i = 0; i< inode_list.size(); ++i) {
	if(__gd.getInEdgeList(i).size() == 0) { // if (inode_list[i]->getType() == INode::SOURCE) {
	  cerr << "found SOURCE" << endl;
          IFileHandler_var i_handler(inode_list[i]->getParentObject()->getInterface<IFileHandler>());
          if (i_handler.get()) {
            inode_list[i]->reachInitialized();
            i_handler->setFilename(__input_files[__curr_input_file]);
            inode_list[i]->reachStarted();
	    break; // currently only works for a single source
          }
        }
      }
      /* End< */
    }
  }

  void ClicApplication::setupGraph() {
    /* Begin> handle input files: */
    if ( __input_files.size() > 0 ) {
      cerr << "Switch to input file " << (__curr_input_file) << ": " << __input_files[__curr_input_file] << endl;
      SequenceValue<INode>& inode_list = __gd.getINodeList();
      for (unsigned int i = 0; i< inode_list.size(); ++i) {
	if(__gd.getInEdgeList(i).size() == 0) {
	  IFileHandler_var i_handler(inode_list[i]->getParentObject()->getInterface<IFileHandler>());
	  if (i_handler.get()) {
	    i_handler->setFilename(__input_files[__curr_input_file]);
	    break; // currently only works for a single source
	  }
	}
      }
    }
    /* End< */
    /* Begin> handle output files: */
    SequenceValue<INode>& inode_list = __gd.getINodeList();
    for (unsigned int i = 0; i< inode_list.size(); ++i) {
      if(__gd.getOutEdgeList(i).size() == 0) { 
      
	++__sink_nodes;
	inode_list[i]->getParentObject()->registerEventListener(ITrack::endTrack_event, 
								new TEDObject0<ClicApplication, Result (ClicApplication:: * )(), Result >
								(this, &ClicApplication::eofHandler));
	inode_list[i]->getParentObject()->registerEventListener(ITrackList::endTrackList_event, 
								new TEDObject0<ClicApplication, Result (ClicApplication:: * )(), Result >
								(this, &ClicApplication::finishedHandler));
      }
    }
    if (!__output_file.empty()) {
      for (unsigned int i = 0; i< inode_list.size(); ++i) {
	if(__gd.getOutEdgeList(i).size() == 0) { 
	  ISinkFileHandler_var o_handler(inode_list[i]->getParentObject()->getInterface<ISinkFileHandler>());
	  if (o_handler.get()) {
	    o_handler->setFilename(__output_file);
	    break; // currently only works for a single sink
	  }		    
	}
      }
    }
    /* End< */
    
    /* Begin> Configure the nodes according to its event-description and store the events in the NodeDescription */
    for (list<NodeConfiguration>::iterator it = GDParse::node_configuration.begin(); it != GDParse::node_configuration.end(); ++it) {
      if (!it->getEventList().empty()) {
	for (list<NCElement>::iterator it1 = it->getEventList().begin(); it1 != it->getEventList().end(); ++it1) {
	  /* create an ClicValueStream */
	  IClicValueStream iv(it1->event_key, it1->argument_list);

	  /* get the inode */
	  INode* inode = __gd.getINode(*it->getNodeDescription());
	  /* request the event from the node */
	  Event* event= inode->getParentObject()->matchEventPrototype( iv.getEventKey() );

	  if (!event) {
	    throw Exception ("Unknown event with key " + iv.getEventKey() + " in node " + it->getNodeDescription()->getNodeName() );
	  }

	  /* read the parameters from the string */
	  iv.readEvent(*event);

	  /* and add the event to the NodeDescription */
	  __gd.getNodeDescription(it->getNodeDescription()->getInternalID()).addNodeEvent(*event, it1->state);
	  /* delete the old event */
	  delete event;
	}
      }
    }
    /* End< */

    if (__sync) {
      /* connect the synchronizer if synchronization is enabled */
      __gd.connectSinkSynchronizer();      
      ISinkSynchronizer_var sink_synchronizer(__gd.getISession()->getSinkSynchronizer());
      __synchronizer = sink_synchronizer->getParentObject()->getInterface<IMultiAudioVideoSynchronizer>();
    } else {
      __gd.setSinkSynchronizationEnabled(false);
    }
    
    if (__prefetching) {
      /* connect the synchronizer if synchronization is enabled */
      __gd.setPrefetchingEnabled(true);      
    }

    __paused = false;

    /* realize the graph */
    __gd.realizeGraph();

    /* activate the graph */
    __gd.activateGraph();
    /* and finaly start it */
    __gd.startGraph();

    /* Begin> Find display node */
    for (unsigned int i = 0; i< inode_list.size(); ++i) {
      if (inode_list[i]->getType() == INode::SINK) {
	/* set video display */
	__video_display.reset(inode_list[i]->getParentObject()->getInterface<IVideoDisplay>());
	if (!__video_display.get()) {
	  continue;
	}
	
        /* enable display events*/
        enableDisplayEvents();
        
	/* found a display */
	break;
      }
    }
    /* End< */

  }
  
  void ClicApplication::buildGraph() {
    string demux_audio_jack_tag = "audio"; // e.g. set this to audio0 or audio1 to explicitly choose audio track for demux nodes
    string demux_video_jack_tag = "video";

    try {
      NodeDescription __playback_nd;

      // if a specific node name was set, use this one, otherwise
      // specify abstract and platform indepenendent description
      if(!__audio_sink_name.empty()) {
	__playback_nd.setNodeName(__audio_sink_name);
      }
      else {
	__playback_nd.setNodeType(INode::SINK);
	__playback_nd.addInterface(IAudioDevice::IAudioDevice_type);  
      }

      if( !__playback_location.empty() ) {
        __playback_nd.setLocation(__playback_location);
      }
      
      NodeDescription __display_nd;

      // if a specific node name was set, use this one, otherwise
      // specify abstract and platform indepenendent description
      if(!__video_sink_name.empty()) {
	__display_nd.setNodeName(__video_sink_name);
      }
      else {
	__display_nd.setNodeType(INode::SINK);
	__display_nd.addInterface(IVideoDisplay::IVideoDisplay_type);  
      }

      if( !__display_location.empty() ) {
	__display_nd.setLocation(__display_location);
      }
      
      // setup graph builder
      GraphBuilder gb;
      if(!gb.setURL(__graph_url)) {
	throw Exception("Invalid URL given");
      }
        
      if(__sync) {
	// create synchronizer
        __sink_synchronizer = __app->getRegistry().requestSinkSynchronizer();
        __synchronizer = __sink_synchronizer->getParentObject()->getCheckedInterface<IMultiAudioVideoSynchronizer>();
	gb.setMultiAudioVideoSynchronizer(__synchronizer);
        __paused = false;
      }
      gb.setAudioSink(__playback_nd);
      gb.setVideoSink(__display_nd);
      gb.setDemuxAudioJackTag(demux_audio_jack_tag);
      gb.setDemuxVideoJackTag(demux_video_jack_tag);
      if(__prefetching) {
	gb.setAudioVideoPrefetchingEnabled(__prefetching);
      }

      __composite = gb.createGraph(*__app);
      __composite->reachStarted();
      if(__display) {
        __display->reachStarted();
      }
      
      /* get interface for seeking */
      __seekable = __composite->getInterface<ISeekable>();
      
      /* get video display interface */
      __video_display.reset(__composite->getInterface<IVideoDisplay>());
      if(__video_display.get()) {
        enableDisplayEvents();
      }
    }
    catch (const Exception& e) {
      cerr << e << endl;
    }
    catch(...) {
      cerr << "something went wrong..." << endl;
    }
  } 

  void ClicApplication::destroyGraph() {
    if(__seekable) {
      delete __seekable;
      __seekable = 0;
    }
    
    disableDisplayEvents();
    if(__video_display.get()) {
      __video_display.reset(NULL);
    }    
    
    if(__composite) {
      if(__display && __display->isStarted()) {
        __display->stop();
      }
      __composite->reachActivated();
      __composite->flush();
      if(__display && __display->isActivated()) {
        __display->flush();
      }
      __composite->reachConstructed();
      if(__display) {
        __display->reachConstructed();
      }
      delete __composite;
      __composite = 0;

      if(__display) {
        __app->getRegistry().releaseNode(*__display);
        __display = NULL;
      }
    }
    
    if(__sync) {
      delete __synchronizer;
      __synchronizer = 0;
      __app->getRegistry().releaseNMMObject(*__sink_synchronizer);
      __sink_synchronizer = 0;
    }
  }
  
  void ClicApplication::enableDisplayEvents() {
    if(__video_display.get() && __composite) {
      /* enable display events */
      IDisplayEventSender_var display_event_sender(__video_display->getParentObject()->getInterface<IDisplayEventSender>());
      if (display_event_sender.get()) {
        DEBUG_STREAM( "ENABLE DISPLAY EVENTS!!!!" << endl);
        display_event_sender->enableDisplayEvents();
        
        __x_listener = new TEDObject1<ClicApplication, string>(this, &ClicApplication::keyReleased);
        /* register an event listener */
        __composite->
          registerEventListener(IDisplayEvents::keyReleased_event, __x_listener);
      }
    }
  }
  
  void ClicApplication::disableDisplayEvents() {
    if(__video_display.get()) {
      /* enable display events */
      IDisplayEventSender_var display_event_sender(__video_display->getParentObject()->getInterface<IDisplayEventSender>());
      if (display_event_sender.get()) {
        /* unregister event listener */
        __composite->removeEventListener(__x_listener);
        
        display_event_sender->disableDisplayEvents();        
      }
    }
  }
  
  void ClicApplication::usage(const string& name) {
    cerr << "Usage (for .gd files):" << endl
         << "  " << name << " [options] <.gd file with graph description>" << endl
         << endl
         << "Options:" << endl
         << "  -i : Try to set an input file to all source nodes which support IFilename" << endl
         << "  -o : Try to set an output file to all sink nodes support IFilename" << endl
         << "  -a : Exit clic after playing all files " << endl
         << "  -L : Loop after playing all files " << endl
         << endl;
    cerr << "Usage (for graph builder to play back given URL):" << endl
         << "  " << name << " [options] <url>" 
	 << endl << endl
	 << "  Supported URLs are:" << endl;

    GraphURL url;
    list<string> url_list = url.getRegisteredGraphURLs();
    for(list<string>::iterator i = url_list.begin(); i != url_list.end(); ++i) {
      cerr << "    " << (*i) << endl;
    }

    cerr << "  For your convenience, local file URLs can be specified directly, " << endl
	 << "  i.e. use '<path and filename>' instead of 'file://<path>/<filename>'" << endl
         << endl
         << "Options:" << endl
         << "  -A <hostname> : Sets host for possible audio output" << endl
         << "  -W <hostname> : Sets host for possible video output" << endl
         << "  --audio-sink <node> : Use plugin <node> for audio output" << endl
         << "  --video-sink <node> : Use plugin <node> for video output" << endl
         << endl
         << "General options for both modes:" << endl
         << "  -h : Show this help" << endl
         << "  -t <seconds> : Sets a timeout in seconds after that clic automatically exit" << endl
         << "  -S : Enable access to the local server registry from other applications " << endl
         << "  -p <port> : Sets a specific port for the server registry " << endl
         << "  -s : Disable synchronization " << endl
         << "  -P : Enable prefetching " << endl
         << "  -v : Increase verbositiy (may be given multiple times)" << endl
         << "  -l --license : print version and license information" << endl
         << endl;
  }

  Result ClicApplication::eofHandler() {
    ++__eof_counter;
    if (__eof_counter == __sink_nodes) {
      if( __input_files.size() > 0 ) {
        MESSAGE_STREAM( "Received \"end_of_file\" Event for input file " << __input_files[__curr_input_file] << endl);

        /* if a delay was defined and it was not the last input file, wait.*/
        if (__delay && !(__curr_input_file >= __input_files.size() - 1 ) )
		  suspendThread(Interval(__delay,0));

        nextInputFile();
      }
      __eof_counter = 0;
    }

    return SUCCESS;
  }

  Result ClicApplication::finishedHandler() {
    cerr << endl ;
    MESSAGE_STREAM( "Received \"finished\" Event for all input files. "<< endl);
    return SUCCESS;
  }
  
  void ClicApplication::timeout(Timer* ) {
    /* quit application */
    quit();
  }
}
