/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#ifndef _NMM_CLICAPPLICATION_HPP_
#define _NMM_CLICAPPLICATION_HPP_

#include "nmm/utils/NamedObject.hpp"

#include "nmm/comm/messaging/EDObject.hpp"

#include "nmm/plugins/video/display/interfaces/IVideoDisplay.hpp"

#include "nmm/services/graphbuilder/CompositeNode.hpp"
#include "nmm/services/registry/GraphDescription.hpp"
#include "nmm/services/registry/NMMApplication.hpp"

#include "nmm/utils/time/Clock.hpp"

#include "TimerListener.hpp"
#include "ClicConsoleInput.hpp"

namespace NMM {

class ISeekable;
class ISinkSynchronizer;
class IMultiAudioVideoSynchronizer;

  /** Main class of clic application.
   * Initializes NMMApplication, handles commandline options, keyboard events and
   * sets up NMM graphs corresponding to current input file.
   */

class ClicApplication : public NamedObject, private TimerListener {
public:
  /** Constructor
   *
   * @param argc number of arguments from commandline
   * @param argv array of arguments from commandline
   */
  ClicApplication(int argc, char** argv);

  /** Destructor
   */
  virtual ~ClicApplication();

  /** Sets up graph description
   * 
   * Sets up graph description using following steps:
   * (1) setup of out-of-band communication, 
   * (2,3) sets input and output files
   * (4) triggers calling of out-of-band events
   * (5) realizes graph (mainly connection setup)
   * (6) setup of synchronization (if enabled)
   * (7,8) activate and start graph 
   * (7) register listener (FIXME: should this be done before starting the graph?)
   */
  void setupGraph();

  /** Gets graph description.
   *
   *  @return graph description
   */
  GraphDescription& getGraphDescription()       { return __gd; }
  
  /** Gets graph description.
   *
   *  @return graph description
   */
  const GraphDescription& getGraphDescription() const { return __gd; }

  /** Gets NMM application.
   *
   *  @return NMM application
   */
  NMMApplication*   getNMMApplication()         { return __app; }
  
  /** Gets NMM application.
   *
   *  @return NMM application
   */
  const NMMApplication*   getNMMApplication()   const { return __app; }

  /** Waits until argument given key is pressed.
   *
   * @param c key to wait for
   */
  void waitUntilKeyPressed(const char c);

protected:
  /** Parses commandline arguments.
   *
   * @param argc number of arguments 
   * @param argv pointer to argument list.
   */
  void parseArguments(int argc, char**argv);

  /** Switches to next file if eof is reached.
   *
   * @return always SUCCESS
   */
  Result eofHandler();
  
  /** Outputs message if finished event has been received.
   *
   * @return always SUCCESS
   */
  Result finishedHandler();

  /** Outputs usage message.
   *
   * @param appname name of the application
   */
  void usage(const string& appname);
  
  /** Handler method for key released event.
   *
   * @param key key released
   * @return SUCCESS if known key has been pressed, FAILURE otherwise
   */
  Result keyReleased(string& key);

  /** Switches to next input file if known or quits.
   */
  void nextInputFile();
  
  void buildGraph();
  void destroyGraph();
  
  /** Seeks fix amount of time forward.
   */
  void seekForward();
  
  /** Seeks fix amount of time backward.
   */
  void seekBackward();

  /** Pauses playback.
   */
  void pause();

  /** (Re-)Starts playback.
   */
  void play();
  
  /** Switches between pause and play with check of last state.
   */
  void togglePause();
  
  /** Switches fullscreen on/off with check of last state.
   */
  void toggleFullscreen();

  /** Quits application.
   */
  void quit();
  
  void enableDisplayEvents();
  void disableDisplayEvents();
  
private:
  /** Gets called when internal timer has run out.
   */
  virtual void timeout(Timer* );
  
private:
  GraphDescription  __gd;                 ///< stores the GraphDescription
  NMMApplication*   __app;                ///< Pointer to the NMMApplication
  string            __gd_file;            ///< filename for the grahdescription file.
  string            __graph_url;          ///< string that holds url for GraphURL object and GraphBuilder
  CompositeNode*    __composite;          ///< composite node for graph builder
  ISeekable*        __seekable;           ///< Interface for seeking
  INode*            __playback;           ///< playback node for graph builder
  string            __playback_location;  ///< location for playback node
  INode*            __display;            ///< display node for graph builder
  string            __display_location;   ///< location for display node
  string            __audio_sink_name;    ///< audio sink node name
  string            __video_sink_name;    ///< video sink node name
  ISinkSynchronizer* __sink_synchronizer; ///< Sink synchronizer interface
  IMultiAudioVideoSynchronizer* __synchronizer; ///< synchronizer interface
  vector<string>    __input_files;        ///< input-filename which is set to source nodes with IFilename
  size_t            __curr_input_file;    ///< number of the current input file
  string            __output_file;        ///< output-filename which is set to sink nodes with IFilename
  bool              __sync;               ///< enables/disables synchronization
  bool              __prefetching;        ///< enables/disables prefetching
  int               __eof_counter;        ///< counts how many eof events occurs
  int               __sink_nodes;         ///< stores how many SinkeNodes are used

  IVideoDisplay_var __video_display;      ///< Video display
  bool              __fullscreen;         ///< Fullscreen mode
  unsigned int      __delay;              ///< specifies the delay in seconds between different input files (e.g. for slideshows)
  bool              __infinite_loop;      ///< is true if clic should terminate after playing all files
  bool              __loop;               ///< is true if clic should loop after termination of playing all files  
  
  int               __timeout;            ///< If timeout (in sec) is > 0, clic automatically exit after this time. 
  Timer             __timer;              ///< This timer is used to terminate clic. The attribute __timout must be > 0.

  bool              __enable_registry;    ///< if this flag is true, enable server registry
  int               __port;               ///< port used for server registry
  
  bool              __paused;             ///< Playback is paused
  EDObject*         __x_listener;         ///< event listener for x events
  
  Clock             __clock;              ///< Clock for performance measurements
  
  ClicConsoleInput  __console_input;      ///< Console input component
};
}

#endif 


