/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#ifndef NMM_CLIC_CONSOLE_INPUT_LINUX_HPP
#define NMM_CLIC_CONSOLE_INPUT_LINUX_HPP NMM_CLIC_CONSOLE_INPUT_LINUX_HPP

#include "nmm/utils/thread/ThreadMutex.hpp"
#include "nmm/utils/thread/ThreadCondition.hpp"

#include "nmm/utils/net/NetHandler.hpp"

namespace NMM {
  
  /** Class for handling console input.
   * 
   * A class that waits for keystrokes on the console (standard
   * input) and can be interrupted at any time.
   */
  class ClicConsoleInput : public NetHandler {
  public:
    /** Constructor.
     */
    ClicConsoleInput();
    
    /** Destructor.
     */
    ~ClicConsoleInput();
    
    /** Sets interactive mode.
     *
     * In interactive mode object waits for console input.
     *
     * @param interactive true for interactive mode, false otherwise
     */
    void setInteractive(bool interactive) {
      __interactive = interactive;
    }
    
    /** Waits until a key has been pressed.
     *
     * This method blocks and waits for a keystroke on the console. If
     * a key was pressed, it returns the pressed key. If the operation
     * was interrupted, 0 is returned.
     *
     * @return The key that was pressed or 0 if interrupted
     */
    char waitUntilKeyPressed();
    
    /** Interrupts the waiting operation.
     */
    void interrupt();
    
    /** @name %NetHandler Methods
     */
    //@{
    /** Receives event from the NetReactor.
     * \param socket The socket on which an event occurs.
     * \param event The events which have been registered.
     * \param revents The events which have been occured on the socket
     * \return VOID if nothing should be done with the socket, RECHECK if the socket should be checked for new events, or REMOVE if the Socket should be no longer handled
     */
    virtual NetResult receiveEvent(SocketHandle socket,
                                   short event, short revents);
    
    /** Checks if this handler requires a thread.
     *
     * @return always returns false
     */
    virtual bool requireThread() const {
      return false;
    }
    //@}
  private:
    bool              __interactive;        ///< true, if interactive mode is enabled; false otherwise
    bool              __interrupted;        ///< true, if interrupt was called
    
    ThreadMutex       __cond_mutex;         ///< thread mutex for sync. access to the __condition
    ThreadCondition   __condition;          ///< condition to wait for an input from stdin.
  };
  
}

#endif
