/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifndef VERSION
#define VERSION "1.0.0"
#endif

#include "nmm/services/registry/ServerRegistry.hpp"
#include "nmm/services/registry/ClientRegistry.hpp"
#include "nmm/services/registry/NMMApplication.hpp"
#include "nmm/services/registry/PluginManager.hpp"
#include "nmm/comm/netstrategy/IPGeneric.hpp"

#include "nmm/comm/ProcessUserInfo.hpp"
#include "nmm/comm/netstrategy/UDPBroadcaster.hpp"
#include "nmm/comm/netstrategy/TCPStrategy.hpp"
#include "nmm/comm/LocalStrategy.hpp"


#include "nmm/utils/SecurityCheck.hpp"
#include "nmm/utils/directory_utils.hpp"

#include <iostream>

#include "getopt.h"

using namespace NMM;

/** The server registry application manages NMM-Ressources available on a local system that can be
 * reserved by a ClientRegistry. It allows a networkwide access to use ressources available on this
 * system in a transparent manner.
 *
 * A PluginManager is used to extend the functionality at runtime. If you start
 * the server registry it loads all plugins available on the system.
 * The examples 'registryclient' and 'distregistryclient' implements to clients
 * and show how to access und use the RegistryServer.
 */

/*Typedefs*/
typedef void (*void_method)();
typedef void (*string_method)(const string&);

/* A Void command stores a method pointer with 0 arguments and a method description */
struct VoidCommand {
  VoidCommand(void_method method, const string& desc) : __method(method), __description(desc) {}
  void_method __method;
  string      __description;
};

/* A String command stores a method pointer with 1 argument and a method description */
struct StringCommand {
  StringCommand(string_method method, const string& desc) : __method(method), __description(desc) {}
  string_method __method;
  string        __description;
};

/** Global attributes */
bool                        is_running = true;           ///<While this flag is true, registry is running
IServerRegistry_var         i_server_registry;           ///<The ServerRegistry
map<string, VoidCommand>    void_command_map;            ///<command map for void methods
map<string, StringCommand>  string_command_map;          ///<command map for string methods

/** information of the server-registry */
int registry_port = ClientRegistry::default_port;        ///< stores the used port
static const string registry_version = VERSION;          ///< stores the Version from config.h



/** Prints the usage of this program  */
void usage( const char *name ) {
  cerr << endl
       << "Usage:" << endl
       << name << " [options] " << endl
       << endl
       << "    -p --port            sets the port used by the ServerRegistry to accept incoming connections" << endl
       << "    -b --broadcast       disables broadcasting of registry interfaces" << endl
       << "    -s --save            Stores information about plugins in ~/.nmm/plugins.<hostname>" << endl
       << "    -r --load            Load information about plugins from a specific file. " << endl
       << "    -f --force           force loading of all available plugins" << endl
       << "    -d --debug           Turn debug stream on" << endl
       << "    -e --error           Turn error stream on" << endl
       << "    -w --warning         Turn warning stream on" << endl
       << "    -m --message         Turn message stream on" << endl
       << "    -v --verbose:        Verbose mode (lots of messages!)" << endl

       << "    -Q                    set default QoS tags for out of band events (RMI_DATA), instream buffer(STREAMING_DATA) and instream events(INSTREAM_DATA)" << endl
       << "    -O <qos-tag>             enable QoS tagging for out of band communication. <qos-tag> can be STREAMING_DATA, INBAND_DATA or RMI_DATA " << endl
       << "    -E <qos-tag>             enable QoS tagging for instream events. <qos-tag> can be STREAMING_DATA, INBAND_DATA or RMI_DATA " << endl
       << "    -B <qos-tag>             enable QoS tagging for instream data. <qos-tag> can be STREAMING_DATA, INBAND_DATA or RMI_DATA" << endl

       << "    -l --license:        Print version and license information" << endl
       << "    -h --help:           Shows this help and exits." << endl;
  exit(0);
}

/** Disable all message streams, expect the error stream*/
void turnOffAllMessages() {
  NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  cerr << "All debug streams disabled" << endl;
}

/** Enable all message streams */
void turnOnAllMessages() {
  NamedObject::getGlobalInstance().setDebugStream(&cout, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setWarningStream(&cout, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setMessageStream(&cout, NamedObject::ALL_LEVELS);
  cerr << "All debug streams enabled" << endl;
}

/** Print the help info stored in the CommandObject */
void printHelpMessage(const string& command) {
  /* search the command in the VoidCommand-map */
  map<string, VoidCommand>::iterator void_it = void_command_map.find(command);
  if (void_it != void_command_map.end()) {
    /* print the description */
    cerr << void_it->second.__description << endl;
    return;
  }
  
  /* search the command in the StringCommand-map */
  map<string, StringCommand>::iterator string_it = string_command_map.find(command);
  if (string_it != string_command_map.end()) {
    /* print the description */
    cerr << string_it->second.__description << endl;
    return;
  }

  /* Handle unknown command */
  cerr << "Unknown command " << command << endl;
}

/** Split the command line and store it in a vector */
void split(const string& command, vector<string>& string_vector) {
  int start_pos = 0;
  int end_pos    = 0;

  /* While the end of the string is not reached */
  while( start_pos < static_cast<int>(command.size()) && start_pos >= 0  ) {
    /* find first char */
    start_pos = command.find_first_not_of(" ", start_pos);
    end_pos   = command.find(" ", start_pos);

    /* if end_pos > size no " " are available */
    if (end_pos > static_cast<int>(command.size()) || end_pos < 0)
      end_pos = command.size();

    string tmp_string;
    tmp_string.resize(end_pos - start_pos);
    /* store the substring in tmp_string */
    command.copy(&tmp_string[0], end_pos - start_pos , start_pos);
    /* store the command in the vector */
    string_vector.push_back(tmp_string);
    /* set the new position in the command string */
    start_pos = end_pos;
  }
}

void reset() {
  i_server_registry->resetInternalReservation();
  cerr << "ServerRegistry successfully reset" << endl;
}

void printCommands() {
  int default_space = 20;
  cerr << endl;
  cerr << "   commands         |               description " << endl;
  cerr << "----------------------------------------------------------------------------------------------------" << endl;

  for (map<string, VoidCommand>::iterator it = void_command_map.begin(); it != void_command_map.end(); ++it) {
    cerr << it->first;
    int space = default_space - it->first.size();
    for (int i = 0; i< space; ++i)
      cerr << " ";

    cerr << "| \t\t " << it->second.__description << endl;
  }

  for (map<string, StringCommand>::iterator it = string_command_map.begin(); it != string_command_map.end(); ++it)
    {
      cerr << it->first;
      int space = default_space - it->first.size();
      for (int i = 0; i< space; ++i)
        cerr << " ";

      cerr << "| \t\t " << it->second.__description << endl;
    }
}

void infoCommand() {
  /* Begin> Print some information about the ServerRegistry */
  cerr << "Server Registry Info: " << endl;
  cerr << "Version        : " << registry_version << endl;
  cerr << "Host           : " << SProcessUserInfo::getInstance().getHostname() << endl;
  cerr << "Used TCP-Port  : " << registry_port << endl;
  cerr << endl;
  /* End< */
}

void printStatus(const list<RegistryID>& reg_list ) ;
void statusCommand() {
  list<RegistryID> reg_list;
  reg_list.push_back( i_server_registry->getRegistryID() );

  printStatus(reg_list);
  cerr << endl;
}

void printStatus(const list<RegistryID>& reg_list ) {
  int default_space = 20;
  for (list<RegistryID>::const_iterator it = reg_list.begin(); it != reg_list.end(); ++it) {
    list<NMM::NodeInfo> info_list = i_server_registry->getNodeInfoList(*it);
	for (list<NodeInfo>::iterator it = info_list.begin(); it != info_list.end(); ++it) {
      /* print the node name */
      cerr << it->getNodeName();

      /* print some " "  */
      int space = default_space - it->getNodeName().size();
      for (int i = 0; i< space; ++i)
        cerr << " ";

      /* print the current instance count */
      cerr << " \t\t" << it->getInstanceCount() << "\t\t";

      /* Begin> Print the maximum instance count */
      if (it->getMaxInstanceCount() == -1 )
        cerr << "unlimited" << endl;
      else
        cerr << it->getMaxInstanceCount() << endl;
      /* End< */
    }
  }

  /* Print the status of the children */
  for (list<RegistryID>::const_iterator it = reg_list.begin(); it != reg_list.end(); ++it) {
    printStatus(i_server_registry->getChildList(*it));
  }
}


void exitRegistry() {
  is_running = false;
}

void setDebugOutput(const string& status) {
  /* Begin> Enable the debug stream, if status = on */
  if (status == "on") {
    NamedObject::getGlobalInstance().setDebugStream(&cout, NamedObject::ALL_LEVELS);
    cerr << "Debug stream enabled" << endl;
    return;
  }
  /* End< */

  /* Begin> Disable the debug stream, if status = off */
  if (status == "off"){
    NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
    cerr << "Debug stream disbaled" << endl;
    return;
  }
  /* End< */

  cerr << "Unknown parameter " << status << endl;
}

void setMessageOutput(const string& status) {
  /* Begin> Enable the debug stream, if status = on */
  if (status == "on") {
    NamedObject::getGlobalInstance().setMessageStream(&cout, NamedObject::ALL_LEVELS);
    cerr << "Message stream enabled" << endl;
    return;
  }
  /* End< */

  /* Begin> Disable the debug stream, if status = off */
  if (status == "off"){
    NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
    cerr << "Message stream disabled" << endl;
    return;
  }
  /* End< */

  cerr << "Unknown parameter " << status << endl;
}

void setWarningOutput(const string& status) {
  /* Begin> Enable the debug stream, if status = on */
  if (status == "on") {
    NamedObject::getGlobalInstance().setWarningStream(&cout, NamedObject::ALL_LEVELS);
    cerr << "Warning stream enabled" << endl;
    return;
  }
  /* End< */

  /* Begin> Disable the debug stream, if status = off */
  if (status == "off"){
    NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
    cerr << "Warning stream disabled" << endl;
    return;
  }
  /* End< */

  cerr << "Unknown parameter " << status << endl;
}

void setErrorOutput(const string& status) {
  /* Begin> Enable the debug stream, if status = on */
  if (status == "on") {
    NamedObject::getGlobalInstance().setErrorStream(&cout, NamedObject::ALL_LEVELS);
    cerr << "Error stream enabled" << endl;
    return;
  }
  /* End< */

  /* Begin> Disable the debug stream, if status = off */
  if (status == "off"){
    NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
    cerr << "Error stream diabled" << endl;
    return;
  }
  /* End< */

  cerr << "Unknown parameter " << status << endl;
}

void setAllMessages(const string& status) {
  if (status == "on") {
    turnOnAllMessages();
    return;
  }

  if (status == "off"){
    turnOffAllMessages();
    return;
  }

  cerr << "Unknown parameter " << status << endl;
}

/* Begin> Global objects for commands with 0 arguments */
VoidCommand reset_command(reset, "Resets the internal parameters of this registry and releases all reserved nodes");
VoidCommand exit_command(exitRegistry, "Exits the Registry");
VoidCommand print_command(printCommands, "Prints the commands available, with its description");
VoidCommand info_command(infoCommand, "Prints information about the Server-Registry");
VoidCommand status_command(statusCommand, "Shows the nodes managed by the ServerRegistry");
/* End< */

/* Begin> Global objects for commands with 1 argument */
StringCommand help_command(printHelpMessage, "Prints the description for a specific command. Expected parameter a <COMMAND>");
StringCommand message_output_command(setAllMessages, "Enables/disbales all output-message streams. Expected parameter 'on', 'off'.");
StringCommand error_stream_command(setErrorOutput, "Enables/disbales the error stream. Expected parameter 'on', 'off'.");
StringCommand warning_stream_command(setWarningOutput, "Enables/disbales the warning stream. Expected parameter 'on', 'off'.");
StringCommand debug_stream_command(setDebugOutput, "Enables/disbales the debug stream. Expected parameter 'on', 'off'.");
StringCommand message_stream_command(setMessageOutput, "Enables/disbales the message stream. Expected parameter 'on', 'off'.");
/* End< */
void parseCommand(const string& command) {
  vector<string> command_line;
  /* split the command and store each string in the vector */
  split(command, command_line);

  if (command_line.size() ==0 ) {
    return;
  }

  /* If the comand line has more than two arguments it is invalid */
  if (command_line.size() > 2) {
    cerr << "invalid command" << endl;
    return;
  }

  /* search the method that accepts no arguments */
  if (command_line.size() == 1) {
    if (void_command_map.find(command_line[0]) != void_command_map.end()) {
      return void_command_map.find(command_line[0])->second.__method();
    }
  }

  /* search the method that accepts one argument */
  if (command_line.size() == 2) {
    if (string_command_map.find(command_line[0]) != string_command_map.end()) {
      return string_command_map.find(command_line[0])->second.__method(command_line[1]);
    }
  }

  cerr << "Unknown command " << command_line[0] << endl;
  cerr << "Enter 'commands' to show available commands" << endl;
}


void initCommands() {
  void_command_map.insert(pair<string, VoidCommand>("reset", reset_command));
  void_command_map.insert(pair<string, VoidCommand>("exit", exit_command));
  void_command_map.insert(pair<string, VoidCommand>("q", exit_command));
  void_command_map.insert(pair<string, VoidCommand>("commands", print_command));
  void_command_map.insert(pair<string, VoidCommand>("info", info_command));
  void_command_map.insert(pair<string, VoidCommand>("status", status_command));

  string_command_map.insert(pair<string, StringCommand>("help", help_command));
  string_command_map.insert(pair<string, StringCommand>("streams", message_output_command));
  string_command_map.insert(pair<string, StringCommand>("error-stream", error_stream_command));
  string_command_map.insert(pair<string, StringCommand>("debug-stream", debug_stream_command));
  string_command_map.insert(pair<string, StringCommand>("message-stream", message_stream_command));
  string_command_map.insert(pair<string, StringCommand>("warning-stream", warning_stream_command));
}

void my_unexpected()
{
  cerr << "Unexpected exception!" << endl;
  exit(1);
}


int main(int argc, char **argv)
{
  LocalStrategy* ls = new LocalStrategy();
  TCPStrategy* tcp = new TCPStrategy();
  
  NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);

  try {
    /* disable error messages from getopt because getopt is called several times */
	::opterr = false;

    /* Begin> Initialize all attributes and handler */
    set_unexpected(my_unexpected);

    unsigned int verbose = 0;
    bool debugStream     = false;
    bool errorStream     = false;
    bool warningStream   = false;
    bool messageStream   = false;
    string read_config_file;
    string write_config_file;
    bool write_config = false;
    bool load_plugins = false;
    bool enable_broadcasting = true;
    UDPBroadcaster* broadcaster = NULL;

    /* Initialize the commands */
    initCommands();
    /* End< */

    /*Begin> Argument handling */
    const char *optstring = "QO:E:B:bs::lr:fdewmvhc:p:";
    struct option longopts[] = {
      {"broadcast",      0, NULL, 'b'},
      {"config",         1, NULL, 'c'},
      {"port",           1, NULL, 'p'},
      {"debug",          0, NULL, 'd'},
      {"error",          0, NULL, 'e'},
      {"warning",        0, NULL, 'w'},
      {"message",        0, NULL, 'm'},
      {"verbose",        0, NULL, 'v'},
      {"help",           0, NULL, 'h'},
      {"save",           2, NULL, 's'},
      {"load",           1, NULL, 'r'},
      {"force",          0, NULL, 'f'},
      {"license",          0, NULL, 'l'},
      {NULL,             0, NULL, 0  } /* last row must be zero */
    };
    string qos_type;
    int ch, optionindex;
    while(EOF != (ch=getopt_long(argc,argv,optstring,longopts,&optionindex))){
      // which option did we get?
      switch (ch) {
      case 'b':
         enable_broadcasting = false;
        break;
      case 'c':
        SSecurityCheck::getInstance().parseFile(string(optarg));
        break;
      case 'p':
        registry_port = atoi(optarg);
        cerr << "Set serverregistry port to " << registry_port << endl;
        break;
      case 's':
        NamedObject::getGlobalInstance().setErrorStream(NULL, NamedObject::ALL_LEVELS);
        NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
        NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
        NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
        NMMApplication::printLicense("serverregistry");

        if (optarg) {
          write_config_file = optarg;
        }
        write_config = true;
        break;
      case 'r':
        read_config_file = optarg;
        break;      
      case 'd':
        debugStream = true;
        break;
      case 'e':
        errorStream = true;
        break;
      case 'w':
        warningStream = true;
        break;
      case 'm':
        messageStream = true;
        break;
      case 'v':
        ++verbose;
        break;
      case 'f':
        load_plugins = true;
        break;

      case 'O' :
       qos_type = optarg;
       if (qos_type == "STREAMING_DATA") {
         IPGeneric::setGlobalOutOfBandPriority(Sock::STREAMING_DATA);
         break;
       }
       if (qos_type == "INBAND_DATA"){
         IPGeneric::setGlobalOutOfBandPriority(Sock::INBAND_DATA);
         break;
       }
       if (qos_type == "RMI_DATA"){
         IPGeneric::setGlobalOutOfBandPriority(Sock::RMI_DATA);
         break;
       }
       cerr << "Invalid QoS tag " << endl;
       usage(argv[0]);
       return 1;

      case 'B' :
       qos_type = optarg;
       if (qos_type == "STREAMING_DATA") {
         IPGeneric::setGlobalInstreamBufferPriority(Sock::STREAMING_DATA);
         break;
       }
       if (qos_type == "INBAND_DATA"){
         IPGeneric::setGlobalInstreamBufferPriority(Sock::INBAND_DATA);
         break;
       }
       if (qos_type == "RMI_DATA"){
         IPGeneric::setGlobalInstreamBufferPriority(Sock::RMI_DATA);
         break;
       }
       cerr << "Invalid QoS tag " << endl;
       usage(argv[0]);
       return 1;

      case 'E' :
       qos_type = optarg;
       if (qos_type == "STREAMING_DATA") {
         IPGeneric::setGlobalInstreamEventPriority(Sock::STREAMING_DATA);
         break;
       }
       if (qos_type == "INBAND_DATA"){
         IPGeneric::setGlobalInstreamEventPriority(Sock::INBAND_DATA);
         break;
       }
       if (qos_type == "RMI_DATA"){
         IPGeneric::setGlobalInstreamEventPriority(Sock::RMI_DATA);
         break;
       }
       cerr << "Invalid QoS tag " << endl;
       usage(argv[0]);
       return 1;
      case 'Q' :  
       // Set the default QoS tags
       IPGeneric::setGlobalInstreamBufferPriority(Sock::STREAMING_DATA);
       IPGeneric::setGlobalInstreamEventPriority(Sock::INBAND_DATA);
       IPGeneric::setGlobalOutOfBandPriority(Sock::RMI_DATA);
       break;

      case 'l' :
        NamedObject::getGlobalInstance().setErrorStream(NULL, NamedObject::ALL_LEVELS);
        NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
        NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
        NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
        NMMApplication::printLicense("serverregistry");
        return 1;
      case 'h':
        usage("server_registry");
        break;
      default: //includes 'h'
        break;
      }
    }

    // maybe turn off debug messages
    if (verbose > 0 ) {
      NamedObject::getGlobalInstance().setErrorStream(&cout, NamedObject::ALL_LEVELS);
    } else {
      NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
    }
    if (verbose > 1) {
      NamedObject::getGlobalInstance().setWarningStream(&cout, NamedObject::ALL_LEVELS);
    } else {
      NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
    }
    if (verbose > 2) {
      NamedObject::getGlobalInstance().setMessageStream(&cout, NamedObject::ALL_LEVELS);
    } else {
      NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
    }
    if (verbose > 3) {
      NamedObject::getGlobalInstance().setDebugStream(&cout, NamedObject::ALL_LEVELS);
    } else {
      NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
    }
    /*End< */

    /* Begin> if the config_file should be written, delete the old config file first */
    if ( write_config ) {
      if ( write_config_file.empty() ) {
	  string::size_type pos = 0;
	    string path = getNMMDir();

		while ( path[path.size()-1] == '\\' || path[path.size()-1] == '/') {
		  path = path.substr(0,path.size() - 1); 
		}

		for (unsigned int i = 0; i < path.size(); ++i) {
		  if (path[i] == '/' || path[i] == '\\' || path[i] == ':') {
		    path[i] = '_';
		  }
		}

	    write_config_file = 
			SProcessUserInfo::getInstance().getHomeDirectory() +
	        string("/.nmm/plugins.") +
	        string(VERSION) +
	        string(".") +
	        SProcessUserInfo::getInstance().getHostname() +
	        string(".") +
	        path;

        unlink(write_config_file.c_str());
		write_config_file = string();
      } else {
        unlink(write_config_file.c_str());
      }
    }
    /*End< */
        
    NMMApplication reg_app(argc, argv, true, registry_port, false);

	/* Check if the plugin configuration must be written */
    if (write_config) {
      if (!write_config_file.empty()) 
	  {
		  reg_app.writeRegistryConfig(write_config_file);
	  }
      return 0;;
	}
    
    if (load_plugins)
      reg_app.loadPlugins();

    if (!read_config_file.empty()) {
      reg_app.readRegistryConfig(read_config_file);
    }

    reg_app.enableServerRegistry(registry_port);
    i_server_registry.reset(reg_app.getServerRegistry()->getCheckedInterface<IServerRegistry>());

	/*Begin> Load all plugins available*/
    IPluginManager_var pm(SPluginManager::getInstance().getInterface<IPluginManager>());
    if (!pm.get()) {
      GLOBAL_ERROR_STREAM("Could not get PluginManager interface" << endl);
      return 0;
    }

    if (enable_broadcasting) {
      try {
        broadcaster = new UDPBroadcaster(registry_port+1, registry_port+2);
        reg_app.setUDPBroadcaster(broadcaster, registry_port);
      } catch ( Exception& e) {
        GLOBAL_ERROR_STREAM( e << endl << "Disable UDPBroadCaster" << endl);
      }
    }
    
    
    if (debugStream) {
      parseCommand(string("debug-stream on"));
    }
    if (errorStream) {
      parseCommand(string("error-stream on"));
    }
    if (warningStream) {
      parseCommand(string("warning-stream on"));
    }
    if (messageStream) {
      parseCommand(string("message-stream on"));
    }
    
    cerr << "                       ServerRegistry successfully started!                 " << endl;

    string command;
    while (is_running) {
      getline(cin,command);
      parseCommand(command);
    }
    delete broadcaster;
  } catch (const SerializedException& e) {
    GLOBAL_ERROR_STREAM("Exception occured!!" << endl);
    GLOBAL_ERROR_STREAM(e << endl);
  } catch (const Exception& e) {
    GLOBAL_ERROR_STREAM("Exception occured!!" << endl);
    GLOBAL_ERROR_STREAM(e << endl);
  }
}

