/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


// system includes
#include <getopt.h> 
#include <string.h>
#include <iostream>


// include NMM 
#include "nmm/multimedia/connect.hpp"
#include "nmm/comm/ProxyObject.hpp"

#include "nmm/services/registry/ProxyApplication.hpp"
#include "nmm/services/registry/GraphDescription.hpp"
#include "nmm/utils/time/Interval.hpp"

//interfaces
#include "nmm/plugins/audio/mux/IAudioMux.hpp"
#include "nmm/plugins/video/mux/IVideoMux.hpp"
#include "nmm/plugins/interfaces/IFileHandler.hpp"
#include "nmm/multimedia/IBiDirQueuedNode.hpp"

// import namespace
using namespace NMM;

int main(int argc, char ** argv)
{
  string filename1 = "";
  string filename2 = "";
  string filename3 = "";
  bool sync_enabled = true;

  const char* optstring = "a:b:c:sv";    
  int ch=0, optionindex=0;
  int verbose = 0;
    
  struct option longopts[] = {
    {"computer1"           , 1, NULL, 'a'},
    {"computer2"           , 1, NULL, 'b'},
    {"computer3"           , 1, NULL, 'c'},
    {"synchronize"         , 0, NULL, 's'},
    {"verbose"             , 0, NULL, 'v'},
    {NULL                  , 0, NULL, 0  } // last row must be zero
  };
    
  while(EOF != (ch=getopt_long(argc,argv,optstring,longopts,&optionindex))){ 
    switch (ch) {
	    
    case 'a': 
      if (optarg) 
	filename1 = optarg;
      cout << "filename1: " << filename1 << endl;
      break;
			    
    case 'b': 
      if (optarg) 
	filename2 = optarg;
      cout << "filename2: " << filename2 << endl;
      break;

    case 'c': 
      if (optarg) 
	filename3 = optarg;
      cout << "filename3: " << filename3 << endl;
      break;

    case 's': 
      sync_enabled = false;
      cout << "Synchronization disabled" << endl;
      break;

    case 'v': //verbose
      cout << "verbose" << endl;
      verbose++;
      break;

    default:
      GLOBAL_ERROR_STREAM("Unknown command line parameter" << endl);
    }
  }
  
  //NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  //NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  //NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  //NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
  // maybe turn off debug messages
  if (verbose < 1) {
    NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 2) {
    NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 3) {
    NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose < 4) {
    NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
  }
    
  // the graph description for the example
  GraphDescription graph;
    
  // the NMM application for the example
  NMMApplication* app = ProxyApplication::getApplication(argc, argv);
    
  Format* video_encoder_format = new Format(Format::video_mpeg4_type);
  video_encoder_format->addIntParamValue(Format::bitrate_param, 3840000);

  Format* video_scaler_format = new Format(Format::video_raw_type);
  video_scaler_format->addIntParamValue(Format::x_resolution_param, 720);
  video_scaler_format->addIntParamValue(Format::y_resolution_param, 540);
  video_scaler_format->addIntParamValue(Format::ratio_x_param, 4);
  video_scaler_format->addIntParamValue(Format::ratio_y_param, 3);

  Format* audio_encoder_format = new Format(Format::audio_mpeg_type);
  audio_encoder_format->addIntParamValue(Format::bitrate_param, 128000);

  NodeDescription read1("GenericReadNode");
  NodeDescription demux1("MPEGDemuxNode");
  NodeDescription decoder1("MPEGVideoDecodeNode");
  NodeDescription audio_decoder1("MPEGAudioDecodeNode");
  NodeDescription buffer_shaper1("BufferShapingNode");

  NodeDescription audio_encoder("MPEGAudioEncodeNode");

  NodeDescription read2("GenericReadNode");
  NodeDescription demux2("MPEGDemuxNode");
  NodeDescription decoder2("MPEGVideoDecodeNode");
  NodeDescription audio_decoder2("AC3DecodeNode");
  NodeDescription buffer_shaper2("BufferShapingNode");
  
  NodeDescription deinterlace1("YUVDeInterlaceNode");
  NodeDescription scaler1("VideoScalerNode");
  
  NodeDescription deinterlace2("YUVDeInterlaceNode");
  
  NodeDescription video_mux("VideoMuxNode");
  NodeDescription audio_mux("AudioMuxNode");

  NodeDescription xdisplay("XDisplayNode");
  NodeDescription playback("ALSAPlaybackNode");

  NodeDescription video_encoder("FFMpegEncodeNode");
  NodeDescription file_mux("AVMuxNode");
  NodeDescription file_writer("AVIWriteNode");

  // create the flow graph: 
  graph.addEdge(read1,demux1);
  graph.addEdge(demux1, "mpeg_video0", decoder1);
  graph.addEdge(decoder1,deinterlace1);
  graph.addEdge(deinterlace1, scaler1);

  graph.addEdge(demux1, "mpeg_audio0", audio_decoder1);
  graph.addEdge(audio_decoder1, buffer_shaper1);

  graph.addEdge(read2,demux2);
  graph.addEdge(demux2, "mpeg_video0", decoder2);
  graph.addEdge(decoder2,deinterlace2);
  graph.addEdge(demux2, "ac3_audio0", audio_decoder2);
  graph.addEdge(audio_decoder2, buffer_shaper2);

  graph.addEdge(buffer_shaper1, audio_mux, "input1");
  graph.addEdge(buffer_shaper2, audio_mux, "input2");
  graph.addEdge(audio_mux, playback);
  
  graph.addEdge(scaler1, video_mux, "input1", video_scaler_format);
  graph.addEdge(deinterlace2, video_mux, "input2");
  graph.addEdge(video_mux, xdisplay);
  
  if(!filename3.empty()) {
    graph.addEdge(video_mux, video_encoder);
    graph.addEdge(video_encoder, file_mux, "video", video_encoder_format);

    graph.addEdge(audio_mux, audio_encoder);
    graph.addEdge(audio_encoder, file_mux, "audio", audio_encoder_format);

    graph.addEdge(file_mux, file_writer);
  }
  
  // request complete graph
  cerr << "request GraphDescription from registry" << endl;
  ClientRegistry& registry = app->getRegistry();
  registry.requestGraph(graph);

   
  IBiDirQueuedNode_var decoder1_queued(graph.getINode(decoder1)->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
  decoder1_queued->setDownstreamMaxSize(50);
  IBiDirQueuedNode_var audio_decoder1_queued(graph.getINode(audio_decoder1)->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
  audio_decoder1_queued->setDownstreamMaxSize(50);

  IBiDirQueuedNode_var decoder2_queued(graph.getINode(decoder2)->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
  decoder2_queued->setDownstreamMaxSize(50);
  IBiDirQueuedNode_var audio_decoder2_queued(graph.getINode(audio_decoder2)->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
  audio_decoder2_queued->setDownstreamMaxSize(50);
   

  INode* readfile1 = graph.getINode(read1);
  IFileHandler_var filehandler1 (readfile1->getParentObject()->getCheckedInterface<IFileHandler>());
  filehandler1->setFilename(filename1);

  INode* readfile2 = graph.getINode(read2);
  IFileHandler_var filehandler2 (readfile2->getParentObject()->getCheckedInterface<IFileHandler>());
  filehandler2->setFilename(filename2);

  if(!filename3.empty()) {
    INode* writefile = graph.getINode(file_writer);
    IFileHandler_var filehandler3 (writefile->getParentObject()->getCheckedInterface<IFileHandler>());
    filehandler3->setFilename(filename3);
  }
  
  // realize graph
  if(sync_enabled) {
    graph.connectSinkSynchronizer();
  }
  cerr << "realizeGraph()" << endl;
  graph.realizeGraph();


  // setup video mux
  Format_var scaled_video_format(graph.getINode(scaler1)->getOutputStreamFormat());
  int full_width = scaled_video_format->getIntValue(Format::x_resolution_param);  // = 720;
  int full_height = scaled_video_format->getIntValue(Format::y_resolution_param); // = 540;
  int min_width = 176;
  int min_height = 132;
  int max_width = 528;
  int max_height = 396;

  IVideoMux_var video_mux_control(graph.getINode(video_mux)->getParentObject()->getCheckedInterface<IVideoMux>());

  video_mux_control->setPosition("input1", 0, 0);
  video_mux_control->setSize("input1", full_width, full_height);

  video_mux_control->setPosition("input2", full_width, full_height);
  video_mux_control->setSize("input2", 0, 0);

  // setup audio mux
  int max_volume1 = 30;
  int max_volume2 = 100;

  IAudioMux_var audio_mux_control(graph.getINode(audio_mux)->getParentObject()->getCheckedInterface<IAudioMux>());

  // enable volume of input1, disable on input2
  audio_mux_control->setVolume("input1", max_volume1);
  audio_mux_control->setVolume("input2", 0);

  // and finally start graph
  cerr << "startGraph()" << endl;
  graph.startGraph();

  cerr << "Test1: Enter '0 <enter>' (wait), then '1 <enter>' (wait), then '2 <enter>' (wait), then '3 <enter>' (wait) " << endl;
  cerr << "Or: restart application and " << endl;
  cerr << "Test2: Enter 'a <enter>' (wait), then 'b <enter>' (wait), then 'c <enter>' (wait) " << endl << endl;
  cerr << "Enter 'q <enter>' to exit." << endl;
  
  // control loop
  char s='0';
  while(s!='q') {
    cin >> s;

    Time current_audio_timestamp = audio_mux_control->getCurrentSyncTime();
    Time audio_start_time = current_audio_timestamp + Interval(0, 100000000); // + 100ms
    
    Time current_video_timestamp = video_mux_control->getCurrentSyncTime();
    Time video_start_time = current_video_timestamp + Interval(0, 100000000); // + 100ms

    if(s == 'a') {
      Time video_end_time   = video_start_time        + Interval(0, 250000000);         
      Time audio_end_time   = audio_start_time        + Interval(3, 500000000); 

      cerr << "Enabling a" << endl;

      // change volumes: input1: from current value to 0; input2: crom current value to maximum
      audio_mux_control->setVolumeControl("input1", audio_start_time, audio_end_time, audio_mux_control->getVolume("input1"), 0);
      audio_mux_control->setVolumeControl("input2", audio_start_time, audio_end_time, audio_mux_control->getVolume("input2"), max_volume2);

      // scale video frames
      video_mux_control->setWidthControl("input2", video_start_time, video_end_time, video_mux_control->getWidth("input2"), min_width, IVideoMux::BOTTOM_RIGHT);
      video_mux_control->setXPositionControl("input2", video_start_time, video_end_time, video_mux_control->getXPos("input2"), min_width);
      video_mux_control->setYPositionControl("input2", video_start_time, video_end_time, video_mux_control->getYPos("input2"), min_height);
    }

    else if(s == 'b') {
      Time audio_end_time   = audio_start_time        + Interval(1, 500000000); 
      Time video_end_time   = video_start_time        + Interval(0, 500000000);         

      cerr << "Enabling b" << endl;

      // change volumes: input1: from current value to 0; input2: crom current value to maximum
      audio_mux_control->setVolumeControl("input1", audio_start_time, audio_end_time, audio_mux_control->getVolume("input1"), 0);
      audio_mux_control->setVolumeControl("input2", audio_start_time, audio_end_time, audio_mux_control->getVolume("input2"), max_volume2);

      // scale video frames
      video_mux_control->setWidthControl("input2", video_start_time, video_end_time, video_mux_control->getWidth("input2"), full_width, IVideoMux::BOTTOM_RIGHT);
      video_mux_control->setXPositionControl("input2", video_start_time, video_end_time, video_mux_control->getXPos("input2"), 0);
      video_mux_control->setYPositionControl("input2", video_start_time, video_end_time, video_mux_control->getYPos("input2"), 0);
    }

    if(s == 'c') {
      Time audio_end_time   = audio_start_time        + Interval(1, 500000000); 
      Time video_end_time   = video_start_time        + Interval(0, 500000000);         
    
      cerr << "Enabling c" << endl;
      // change volumes: input1: from current value to maximum; input2: crom current value to 0

      audio_mux_control->setVolumeControl("input1", audio_start_time, audio_end_time, audio_mux_control->getVolume("input1"), max_volume1);
      audio_mux_control->setVolumeControl("input2", audio_start_time, audio_end_time, audio_mux_control->getVolume("input2"), 0);

      // scale video frames
      video_mux_control->setWidthControl("input1", video_start_time, video_end_time, video_mux_control->getWidth("input1"), full_width, IVideoMux::TOP_LEFT);
      video_mux_control->setWidthControl("input2", video_start_time, video_end_time, video_mux_control->getWidth("input2"), 0, IVideoMux::BOTTOM_RIGHT);
      video_mux_control->setXPositionControl("input2", video_start_time, video_end_time, video_mux_control->getXPos("input2"), full_width);
      video_mux_control->setYPositionControl("input2", video_start_time, video_end_time, video_mux_control->getYPos("input2"), full_height);
    }

    if(s == '0') {
      Time video_end_time   = video_start_time        + Interval(0, 250000000);         

      cerr << "Enabling 0" << endl;

      // scale video frames
      video_mux_control->setWidthControl("input1", video_start_time, video_end_time, video_mux_control->getWidth("input1"), max_width, IVideoMux::TOP_LEFT);
      video_mux_control->setWidthControl("input2", video_start_time, video_end_time, video_mux_control->getWidth("input2"), min_width, IVideoMux::BOTTOM_RIGHT);
      video_mux_control->setXPositionControl("input2", video_start_time, video_end_time, video_mux_control->getXPos("input2"), min_width);
      video_mux_control->setYPositionControl("input2", video_start_time, video_end_time, video_mux_control->getYPos("input2"), min_height);
    }

    else if(s == '1') {
      Time audio_end_time   = audio_start_time        + Interval(1, 500000000); 
      Time video_end_time   = video_start_time        + Interval(0, 500000000);         

      cerr << "Enabling 1" << endl;

      // change volumes: input1: from current value to 0; input2: crom current value to maximum
      audio_mux_control->setVolumeControl("input1", audio_start_time, audio_end_time, audio_mux_control->getVolume("input1"), 0);
      audio_mux_control->setVolumeControl("input2", audio_start_time, audio_end_time, audio_mux_control->getVolume("input2"), max_volume2);

      // scale video frames
      video_mux_control->setWidthControl("input1", video_start_time, video_end_time, video_mux_control->getWidth("input1"), min_width, IVideoMux::TOP_LEFT);
      video_mux_control->setWidthControl("input2", video_start_time, video_end_time, video_mux_control->getWidth("input2"), max_width, IVideoMux::BOTTOM_RIGHT);
      video_mux_control->setXPositionControl("input2", video_start_time, video_end_time, video_mux_control->getXPos("input2"), min_width);
      video_mux_control->setYPositionControl("input2", video_start_time, video_end_time, video_mux_control->getYPos("input2"), min_height);
    }

    if(s == '2') {
      Time audio_end_time   = audio_start_time        + Interval(1, 500000000); 
      Time video_end_time   = video_start_time        + Interval(0, 500000000);         
    
      cerr << "Enabling 2" << endl;
      // change volumes: input1: from current value to maximum; input2: crom current value to 0

      audio_mux_control->setVolumeControl("input1", audio_start_time, audio_end_time, audio_mux_control->getVolume("input1"), max_volume1);
      audio_mux_control->setVolumeControl("input2", audio_start_time, audio_end_time, audio_mux_control->getVolume("input2"), 0);

      // scale video frames
      video_mux_control->setWidthControl("input1", video_start_time, video_end_time, video_mux_control->getWidth("input1"), max_width, IVideoMux::TOP_LEFT);
      video_mux_control->setWidthControl("input2", video_start_time, video_end_time, video_mux_control->getWidth("input2"), min_width, IVideoMux::BOTTOM_RIGHT);
      video_mux_control->setXPositionControl("input2", video_start_time, video_end_time, video_mux_control->getXPos("input2"), max_width);
      video_mux_control->setYPositionControl("input2", video_start_time, video_end_time, video_mux_control->getYPos("input2"), max_height);
    }

    if(s == '3') {
      Time video_end_time   = video_start_time        + Interval(0, 250000000);         // + 500ms

      cerr << "Enabling 3" << endl;

      // scale video frames
      video_mux_control->setWidthControl("input1", video_start_time, video_end_time, video_mux_control->getWidth("input1"), full_width, IVideoMux::TOP_LEFT);
      video_mux_control->setWidthControl("input2", video_start_time, video_end_time, video_mux_control->getWidth("input2"), 0, IVideoMux::BOTTOM_RIGHT);
      video_mux_control->setXPositionControl("input2", video_start_time, video_end_time, video_mux_control->getXPos("input2"), full_width);
      video_mux_control->setYPositionControl("input2", video_start_time, video_end_time, video_mux_control->getYPos("input2"), full_height);
    }

  }
 
  // stop and release graph
  cerr << "stopGraph()" << endl;
  graph.stopGraph();

  cerr << "releaseGraph()" << endl;
  registry.releaseGraph(graph);
  
  return 0;
}
