/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
 
// hellographbuilder2 (modified version of hellographbuilder1): shows
// how to use the graph builder of NMM that automatically creates a
// distributed flow graph from a given URL; the audio and video sinks
// are located on the given hosts. All supported codecs and file
// formats supported by NMM can be played with this simple
// example. E.g. start with './hellographbuilder1
// file:///home/bob/video.mpg host1 host2'. Notice: you need a running
// serverregistry on host1 and host2. For more examples of possible
// URLs, see the corresponding section of the documentation for
// 'clic'.

#include <iostream>

// include NMM 
#include "nmm/comm/ProxyObject.hpp"
#include "nmm/services/registry/ProxyApplication.hpp"
#include "nmm/services/graphbuilder/GraphBuilder.hpp"
#include "nmm/multimedia/sync/MultiAudioVideoSynchronizer.hpp"

// include used interfaces
#include "nmm/plugins/interfaces/IFileHandler.hpp"
#include "nmm/plugins/audio/interfaces/IAudioDevice.hpp"
#include "nmm/plugins/video/display/interfaces/IVideoDisplay.hpp"

using namespace NMM;

int main(int argc, char * argv[]) 
{
  if(argc < 4) {
    cerr << "Usage:" << endl
         << "  " << argv[0] << " <url> <location of audio sink> <location of video sink>" << endl
         << endl;
    return 1;
  }
  cerr << argv[0] << " running ... " << endl << endl;

  // disable all debug output (error, warning, message, debug)
  // see documentation Debug, Error, Warning and Message Output in NMM
  NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);

  // the NMM application for the example
  NMMApplication* app = 0;
  // the composite node that holds the flow graph created by the graph builder
  CompositeNode* composite = 0;

  // NMM-operations may throw exception, in this simple example, all exceptions
  // are just caught at the end, and the application is terminated
  try {
    // create application object for example
    app = ProxyApplication::getApplication(argc, argv);

    // create node descriptions for possible sink nodes for audio and video
    // create the description for audio sink:
    // ... some sink node 
    // ... that supports the interface IAudioDevice
    // notice that this description will provide an audio sink
    // on all supported platforms
    NodeDescription audioplay_descr;
    audioplay_descr.setNodeType(INode::SINK);
    audioplay_descr.addInterface(IAudioDevice::IAudioDevice_type);  
    // create the description for video sink:
    // ... some sink node 
    // ... that supports the interface IVideoDisplay
    // notice that this description will provide a video sink
    // on all supported platforms
    NodeDescription display_descr;
    display_descr.setNodeType(INode::SINK);
    display_descr.addInterface(IVideoDisplay::IVideoDisplay_type);  
    
    // set location for audio sink to argv[2] and for video sink to argv[3]
    audioplay_descr.setLocation(argv[2]);
    display_descr.setLocation(argv[3]);

    // create synchronizer for audio/video
    MultiAudioVideoSynchronizer av_sync;
    IMultiAudioVideoSynchronizer_var sync(av_sync.getCheckedInterface<IMultiAudioVideoSynchronizer>());

    // setup graph builder
    GraphBuilder gb;
    if(!gb.setURL(argv[1])) {
      throw Exception("Invalid URL given");
    }
    gb.setMultiAudioVideoSynchronizer(sync.get());
    gb.setAudioSink(audioplay_descr);
    gb.setVideoSink(display_descr);

    // create flow graph for given URL
    composite = gb.createGraph(*app);

    // start all internal nodes of composite node
    composite->reachStarted();

    // wait ...
    cerr << "Playing ... enter 'q <enter>' to exit." << endl;
    string s;
    cin >> s;
  } 
  catch (const Exception& e) {
    cerr << e << endl;
  }
  catch(const SerializedException& e) { // catch serialized NMM-exception and print it
    cerr << e << endl;
  }
  catch(...) {
    cerr << "something went wrong..." << endl;
  }
  
  // release composite node
  if(composite) {
    composite->reachActivated();
    composite->flush();
    composite->reachConstructed();
    delete composite;
  }

  delete app;
  return 0;
}

