/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

// distributed wav player using GraphDescription with error handling
// and access to 'serverregistry' playback on remote host 
// notice: you need a running 'serverregistry' on the remote host

#include <iostream>
#include <getopt.h> //X

// include NMM 
#include "nmm/comm/ProxyObject.hpp"
#include "nmm/services/registry/ProxyApplication.hpp"

// include used interfaces
#include "nmm/plugins/interfaces/IFileHandler.hpp"
#include "nmm/plugins/audio/interfaces/IAudioDevice.hpp"
#include "nmm/plugins/interfaces/IBufferSize.hpp" //X
#include "nmm/comm/netstrategy/ITCPControl.hpp"//X
#include "nmm/multimedia/IBiDirQueuedNode.hpp"//X

// import namespace
using namespace NMM;

void usage(const string& name) {
  cerr << "Usage :" << endl
       << "  " << name << " [options] -H <Host> -i <wav file>" << endl
       << endl
       << "Options:" << endl
       << "  -i <wav file>    : Sets an input file " << endl
       << "  -H <Host>        : Sets remote host for audio playback " << endl
       << "  -p <queue size>  : Sets the queue size of the audio sink node for incoming buffers " << endl
       << "  -t <buffer size> : Sets the buffer size (in bytes) for incoming data of the TCP connection between source and sink (e.g. 8192)" << endl
       << "  -h               : Show this help" << endl
       << endl;
}

int main(int argc, char ** argv)
{
  //XX
  if(argc < 3) {
    usage(argv[0]);
    return 1;
  }

  const char *opt_string = "i:H:p:t:h";
  struct option options [] = {
    {"input-file",    1,NULL,'i'},
    {"host",          1,NULL,'H'},
    {"playback-queue",1,NULL,'p'},
    {"tcp-queue",     1,NULL,'t'},
    {"help",          0,NULL,'h'},
    {NULL,            0,NULL,0}
  };
  
  int ch,long_option_index;

  string input_file;
  string remote_host;
  int playback_queue_size = 0;
  int tcp_buffer_size     = 0;
        
  while((ch=getopt_long(argc,argv,opt_string,options,&long_option_index))!=EOF){
      // which option did we get?
    switch (ch) {
      case 'i':
        input_file = optarg;
        break;
      case 'H':
        remote_host = optarg;
        break;
      case 'p':
        playback_queue_size = atoi(optarg);
        break;
      case 't':
        tcp_buffer_size = atoi(optarg);
        break;
      case 'h':
        usage(argv[0]);
        exit(0);         
      default:
        usage(argv[0]);
        throw Exception("Invalid option");
    }
  } //XX

  cerr << argv[0] << " running ... " << endl << endl;
 
  // disable all debug output (error, warning, message, debug)
  // see documentation Debug, Error, Warning and Message Output in NMM
  NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);

  // the graph description for the example
  GraphDescription graph;

  // the NMM application for the example
  NMMApplication* app = 0;

  // NMM-operations may throw exception, in this simple example, all exceptions
  // are just caught at the end, and the application is terminated
  try {
    // create application object for example
    app = ProxyApplication::getApplication(argc, argv);

    // create the descriptions for the WavReadNode and platform specific PlaybackNode
    NodeDescription readfile_descr("WavReadNode");

    // create the description for audio sink:
    // ... some sink node 
    // ... that supports the interface IAudioDevice
    // notice that this description will provide an audio sink
    // on all supported platforms
    NodeDescription audioplay_descr;
    audioplay_descr.setNodeType(INode::SINK);
    audioplay_descr.addInterface(IAudioDevice::IAudioDevice_type);  

    // audio renderer will be requested from serverregistry
    // application running on given host
    audioplay_descr.setLocation(remote_host); //X

    // create a flow graph: readfile -> audioplay
    graph.addEdge(readfile_descr, audioplay_descr);

    // request complete graph
    cerr << "request GraphDescription from registry" << endl;
    ClientRegistry& registry = app->getRegistry();
    registry.requestGraph(graph);

    // set the filename by requesting appropriate interface
    cerr << "set filename to " << input_file << endl;
    INode* readfile = graph.getINode(readfile_descr);
    IFileHandler_var filehandler (readfile->getParentObject()->getCheckedInterface<IFileHandler>());
    filehandler->setFilename(input_file);
    
    //XX
    // set the queue size of the audio sink node for incoming NMM buffers in downstream direction, if it is specified
    if (playback_queue_size > 0) {
      cerr << "set queue size to " << playback_queue_size << endl;
      IBiDirQueuedNode_var audioplay_queue(graph.getINode(audioplay_descr)->getParentObject()->getInterface<IBiDirQueuedNode>());
      audioplay_queue->setDownstreamMaxSize(playback_queue_size, "default");
    }

    // set the buffer size of the tcp connection for incoming data, if it is specified
    if (tcp_buffer_size > 0) {
      cerr << "set tcp buffer size " << tcp_buffer_size << endl;

      // create event for setting tcp receive buffer size
      // this event will be forwarded to binding upon creation of the flow graph
      Event event(ITCPControl::setRecvBufferSize_event, new TInValue<nmm_int32>(tcp_buffer_size));

      // get binding description for the edge in the graph.
      ssize_t edge_id = graph.getEdgeID(readfile_descr, audioplay_descr);
      InstreamBindingDescription& binding = graph.getEdge(edge_id).getBindingDescription();

      // add a TCPStrategy to make sure we are using TCP.
      unsigned int ts_index = binding.addTransportStrategy("TCPStrategy", InstreamChannel::DOWNSTREAM_BUFFER);

      // add the event to the downstream configuration of the TCPStrategy.
      // This means it only increases the receive buffer for
      // downstream messages.
      binding.getTransportStrategy(ts_index).addDownstreamConfigEvent(event);
    }
    //XX

    // realize and start graph
    cerr << "realizeGraph()" << endl;
    graph.realizeGraph();
    cerr << "startGraph()" << endl;
    graph.startGraph();
    
    // wait ...
    cerr << "Playing audio ... enter 'q <enter>' to exit." << endl;
    string s;
    cin >> s;
    
    // stop and release graph
    cerr << "stopGraph()" << endl;
    graph.stopGraph();

    cerr << "releaseGraph()" << endl;
    registry.releaseGraph(graph);

    delete app;
    return 0;
  }
  catch(const Exception& e) { // catch NMM-exception and print it
    cerr << e << endl;

    // release graph
    cerr << "releaseGraph()" << endl;
    ClientRegistry& registry = app->getRegistry();
    registry.releaseGraph(graph);

    delete app;
    return 1;
  }
  catch(const SerializedException& e) { // catch serialized NMM-exception and print it
    cerr << e << endl;

    // release graph
    cerr << "releaseGraph()" << endl;
    ClientRegistry& registry = app->getRegistry();
    registry.releaseGraph(graph);

    delete app;
    return 1;
  }
  catch(...) { // any other exception
    cerr << "something went wrong!" << endl;

    // release graph
    cerr << "releaseGraph()" << endl;
    ClientRegistry& registry = app->getRegistry();
    registry.releaseGraph(graph);

    delete app;
    return 1;
  }
}
