/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

// rtp streaming server for mp3

#include <iostream>
#include <getopt.h>
#include "nmm/utils/directory_utils.hpp"
#include "nmm/comm/netstrategy/RTPAddress.hpp"
#include "nmm/comm/netstrategy/MulticastIPAddress.hpp"
// include NMM 
#include "nmm/multimedia/connect.hpp"
#include "nmm/comm/ProxyObject.hpp"
#include "nmm/services/registry/ProxyApplication.hpp"
#include "nmm/services/registry/GraphDescription.hpp"

// include used interfaces
#include "nmm/plugins/interfaces/IFileHandler.hpp"
#include "nmm/plugins/net/INet.hpp"

// import namespace
using namespace NMM;

void usage(const string& name) {
  cerr << "Usage :" << endl
       << "  " << name << " [options] -i <input file> -a <audio-port> " << endl
       << endl
       << "Options:" << endl
       << "  -i <input file> : Sets an input file (must be in mp3 format) " << endl
       << "  -a <audio-port> : Sets the multicast port to stream audio " << endl
       << "  -h : Show this help" << endl
       << endl;
}

// RTP and RTCP ports to send audio data
int rtp_port  = 44672;
int rtcp_port = 44673; 
  
string input_file;

int main(int argc, char ** argv)
{
  if(argc < 3) {
    usage( argv[0] );
    return 1;
  }
  cerr << argv[0] << " running ... " << endl << endl;

  
  const char *opt_string = "n:i:a:hv";
  struct option options [] = {
    {"input-file",    1,NULL,'i'},
    {"audio-port",    1,NULL,'a'},
    {"verbose",       0,NULL,'v'},    
    {"help",          0,NULL,'h'},    
    {NULL,            0,NULL,0}
  };
  
  int ch,long_option_index;
  int counter = 1;
  int verbose = 0;
  while((ch=getopt_long(argc,argv,opt_string,options,&long_option_index))!=EOF){
    // which option did we get?
    switch (ch) {
    case 'i':
      input_file = getAbsolutePath(optarg);
      break;
    case 'a':
      rtp_port  = atoi(optarg);
      rtcp_port = rtp_port+1;
      break;
    case 'n':
      counter = atoi(optarg);
      break;
    case 'v' :
      ++verbose;
      break;               
    case 'h':
      usage(argv[0]);
      exit(0);  
    default:
      usage(argv[0]);
      throw Exception("Invalid option");
    }
  }
  
  // disable all debug output (error, warning, message, debug)
  // see documentation Debug, Error, Warning and Message Output in NMM
  // maybe turn off debug messages
  if (verbose > 0 ) {
    NamedObject::getGlobalInstance().setErrorStream(&cout, NamedObject::ALL_LEVELS);
  } else {
    NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose > 1) {
    NamedObject::getGlobalInstance().setWarningStream(&cout, NamedObject::ALL_LEVELS);
  } else {
    NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose > 2) {
    NamedObject::getGlobalInstance().setMessageStream(&cout, NamedObject::ALL_LEVELS);
  } else {
    NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose > 3) {
    NamedObject::getGlobalInstance().setDebugStream(&cout, NamedObject::ALL_LEVELS);
  } else {
    NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
  }

  // the graph description for the example
  GraphDescription* graph = new GraphDescription();

  // the NMM application for the example
  NMMApplication* app = ProxyApplication::getApplication(argc, argv);

  // create the descriptions for the GenericReadNode and NetSinkNode
  NodeDescription readfile_descr("GenericReadNode");
  NodeDescription netsink_descr("NetSinkNode");

  // create a flow graph: readfile -> net_sink
  graph->addEdge(readfile_descr,  netsink_descr);

  // request complete graph
  cerr << "request GraphDescription from registry" << endl;
  ClientRegistry& registry = app->getRegistry();
  registry.requestGraph(*graph);

  // set the filename by requesting appropriate interface
  cerr << "set filename to " << input_file << endl;
  INode* readfile = graph->getINode(readfile_descr);
  IFileHandler_var filehandler (readfile->getParentObject()->getCheckedInterface<IFileHandler>());
  filehandler->setFilename(input_file);

  // set the port number:
  INode* netsink = graph->getINode(netsink_descr);
  
  // set multicast address and rtp/rtcp port to send audio data to
  INet_var i_net (netsink->getParentObject()->getCheckedInterface<INet>());
  MulticastIPAddress ip_addr("239.0.0.2");
  RTPAddress rtp_addr (ip_addr, rtp_port, rtcp_port);
  i_net->setTransportStrategy(&rtp_addr);
  cerr << "Using RTP port  : " << rtp_port << endl;
  cerr << "Using RTCP port : " << rtcp_port << endl; 
  
  // realize and start graph
  cerr << "realizeGraph()" << endl;
  graph->realizeGraph();
  cerr << "startGraph()" << endl;
  graph->startGraph();
  
  // wait ...
  cerr << "Playing audio ... enter 'q <enter>' to exit." << endl;
  string s;
  cin >> s;
    
  // stop and release graph
  cerr << "stopGraph()" << endl;
  graph->stopGraph();

  cerr << "releaseGraph()" << endl;
  app->getRegistry().releaseGraph(*graph);

  delete app;
  return 0;
}

