/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

// rtp streaming client for mp3

#include <iostream>
#include <getopt.h>
#include "nmm/utils/directory_utils.hpp"
#include "nmm/comm/netstrategy/RTPAddress.hpp"
#include "nmm/comm/netstrategy/MulticastIPAddress.hpp"
// include NMM 
#include "nmm/multimedia/connect.hpp"
#include "nmm/comm/ProxyObject.hpp"
#include "nmm/services/registry/ProxyApplication.hpp"
#include "nmm/services/registry/GraphDescription.hpp"

// include used interfaces
#include "nmm/plugins/interfaces/IFileHandler.hpp"
#include "nmm/plugins/net/INet.hpp"


// include used interfaces
#include "nmm/multimedia/sync/interfaces/ISynchronizedSink.hpp"
#include "nmm/multimedia/sync/MultiAudioVideoSynchronizer.hpp"
#include "nmm/utils/time/Clock.hpp" 
#include "nmm/comm/ReferenceManager.hpp"
#include "nmm/comm/ServiceProxy.hpp"
#include "nmm/comm/netstrategy/TCPAddress.hpp"
#include "nmm/comm/netstrategy/TCPAcceptor.hpp"
#include "nmm/multimedia/IBiDirQueuedNode.hpp"

// import namespace
using namespace NMM;

void usage(const string& name) {
  cerr << "Usage :" << endl
       << "  " << name << " [options] " << endl
       << endl
       << "Options:" << endl
       << "  -a <audio-port> : Sets the multicast port to stream audio " << endl
       << "  -s              : Disable synchronization " << endl
       << "  -q <millisecs>  : Set queue duration in milliseconds" << endl
       << "  -p <millisecs>  : Set prefetching duration in milliseconds" << endl
       << "  -H <hostname>   : Synchronize to specified host " << endl
       << "  -h : Show this help" << endl
       << endl;
}

int main(int argc, char ** argv)
{
  // RTP and RTCP ports to connect to
  int rtp_port  = 44672;
  int rtcp_port = 44673;

  MultiAudioVideoSynchronizer* synchronizer = 0;
  ISinkSynchronizer_var        i_sink_synchronizer;

  bool master = true;
  TCPAcceptor* acceptor = 0;
  bool is_synchronization_enabled = true;

  Interval downstream_max_duration(0,300000000); // 300ms
  Interval prefetch_queue_duration(0,200000000); // 200ms
  
  string host;
  
  cerr << argv[0] << " running ... " << endl << endl;

  const char *opt_string = "H:a:sp:q:vh";
  struct option options [] = {
    {"audio-port",    1,NULL,'a'},
    {"host",          1,NULL,'H'},
    {"nosync",        0,NULL,'s'},
    {"queue",         1,NULL,'q'},
    {"prefetch",      1,NULL,'p'},
    {"verbose",       0,NULL,'v'},
    {"help",          0,NULL,'h'},
    {NULL,            0,NULL,0}
  };
  
  int ch,long_option_index;
  int verbose = 0;
  while((ch=getopt_long(argc,argv,opt_string,options,&long_option_index))!=EOF){
    // which option did we get?
    switch (ch) {
    case 'a':
      rtp_port  = atoi(optarg);
      rtcp_port = rtp_port+1;
      break;
    case 'H':
      host = optarg;
      master = false;
      break;
    case 's':
      is_synchronization_enabled = false;
      break;
    case 'q':
      {
        unsigned int milliseconds = atoi(optarg);
        downstream_max_duration =  Interval(milliseconds / 1000, milliseconds * 1000000);
        break;
      }
    case 'p':
      {
        unsigned int milliseconds = atoi(optarg);
        prefetch_queue_duration =  Interval(milliseconds / 1000, milliseconds * 1000000);
        break;
      }
    case 'v':
      ++verbose;
      break;
    case 'h':
      usage(argv[0]);
      exit(0);         
    default:
      usage(argv[0]);
      throw Exception("Invalid option");
    }
  }

  // disable all debug output (error, warning, message, debug)
  // see documentation Debug, Error, Warning and Message Output in NMM
  // maybe turn off debug messages   
  if (verbose > 0 ) {
    NamedObject::getGlobalInstance().setErrorStream(&cout, NamedObject::ALL_LEVELS);
  } else {
    NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose > 1) {
    NamedObject::getGlobalInstance().setWarningStream(&cout, NamedObject::ALL_LEVELS);
  } else {
    NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose > 2) {
    NamedObject::getGlobalInstance().setMessageStream(&cout, NamedObject::ALL_LEVELS);
  } else {
    NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  }
  if (verbose > 3) {
    NamedObject::getGlobalInstance().setDebugStream(&cout, NamedObject::ALL_LEVELS);
  } else {
    NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
  }

  // the graph description for the example
  GraphDescription graph;

  // the NMM application for the example
  NMMApplication* app = ProxyApplication::getApplication(argc, argv);
  
  if (master)  {
    acceptor = new TCPAcceptor(9000);
  }
    
  // create the descriptions for the GenericReadNode, MPEGAudioDecodeNode, and PlaybackNode
  NodeDescription netsource_descr("NetSourceNode");
  NodeDescription mpegdec_descr("MPEGAudioDecodeNode");
  NodeDescription playback_descr("ALSAPlaybackNode");
      
  Format  in_format("audio/mpeg");

  // define the audio data format we want to receive  
  Format mp3_format (Format::audio_mpeg_type);
  mp3_format.addIntParamValue(Format::sample_rate_param, 44100) ; 
  mp3_format.addIntParamValue(Format::channels_param, 2);
  mp3_format.addIntParamValue(Format::layer_param, 3);
  mp3_format.addIntParamValue(Format::bitpersample_param, 16);
  
  // create a flow graph: readfile -> net_sink
  graph.addEdge(netsource_descr, mpegdec_descr, &mp3_format);
  graph.addEdge(mpegdec_descr, playback_descr);
      
  // request complete graph
  cerr << "request GraphDescription from registry" << endl;
  ClientRegistry& registry = app->getRegistry();
  registry.requestGraph(graph);

  // Set queue and prefetching durations
  INode* playback = graph.getINode(playback_descr);
  IBiDirQueuedNode_var iqueue(playback->getParentObject()->getCheckedInterface<IBiDirQueuedNode>());
  if(downstream_max_duration != Interval()) {
    cerr << downstream_max_duration << endl;
    iqueue->setDownstreamMaxDuration(downstream_max_duration);
  }
  if(prefetch_queue_duration != Interval()) {
    cerr << prefetch_queue_duration << endl;
    playback->setPrefetchQueueDuration(prefetch_queue_duration);
    graph.setPrefetchingEnabled(true);
  }
  
  // Try to allocate or to request a Synchronizer 
  try {  
    if ( master ) {
      // Create a synchronizer that can also be used by other applications 
      synchronizer = new MultiAudioVideoSynchronizer();
      i_sink_synchronizer.reset( synchronizer->getInterface<ISinkSynchronizer>() );      
      
      // Register the Synchronizer as service 
      SReferenceManager::getInstance().registerService( "SinkSynchronizer", synchronizer );
    } else {
      ServiceProxy service_proxy;
      
      try {
	// Get synchronizer of host given per commandline option. 
        TCPAddress tcp_address( host, 9000 );
        service_proxy.connectTo(tcp_address, "SinkSynchronizer");
        
	// Get the interface of synchronizer on given host
        i_sink_synchronizer.reset( service_proxy.getCheckedInterface<ISinkSynchronizer>() );

      } catch (const BindingException& e) {
	GLOBAL_ERROR_STREAM("Could not connect to location: " << host << ":" << 9000 << endl);
	cerr << e << endl;
	return 0;
      }      
    }
  } catch ( const Exception& e ) {
    cerr << e << endl;  
  }

  // Set RTP Payloadtype on NetSourceNode
  INode* netsource = graph.getINode(netsource_descr);
  INet_var           i_net (netsource->getParentObject()->getCheckedInterface<INet>());
  i_net->setPayloadType(14);
  cerr << "Using Port " << rtp_port << endl;
  cerr << "Realize Graph " << endl;  
  graph.realizeGraph();
  cerr << "Realize Graph " << endl;
  
  Clock clock;
  Time start_time;
  Time graph_started;
  
  if (! master ) {
    cerr << "Enter any key and press <Enter> to start 'listen in' " << endl;
    string s;
    cin >> s;
  }  
   
  start_time = clock.getTime();
    
  // set Multicast address to connect to
  
  MulticastIPAddress ip_addr("239.0.0.2");
  RTPAddress         rtp_addr (ip_addr, rtp_port, rtcp_port);
  
  i_net->setTransportStrategy(&rtp_addr);
  i_net->setPayloadType(14);
 
  try {
    if(is_synchronization_enabled) {
      // Use the allocated synchronizer 
      ISession* i_session = graph.getISession(); 
      i_session->setSinkSynchronizer( *i_sink_synchronizer.get() );

      graph.connectSinkSynchronizer();
    }
    else {
      graph.setSinkSynchronizationEnabled(false);
    }
    
    // start graph
    cerr << "startGraph()" << endl;
    graph.startGraph();
    graph_started = clock.getTime();
    if (! master ) {
      cerr << "Start Time : " << start_time << endl;
      cerr << "Graph Started : " << graph_started << endl; 
    }
  
  } catch (const Exception& e) { 
    cerr << e << endl;
  }  
  
  // wait ...
  
  cerr << "Playing audio ... enter 'q <enter>' to exit." << endl;
  string s;
  
  while (1) {
    cin >> s;
    if ( s == "q" ) {
      break;
    }
  }
  // stop and release graph
  cerr << "stopGraph()" << endl;
  graph.stopGraph();

  cerr << "releaseGraph()" << endl;
  registry.releaseGraph(graph);
  SReferenceManager::getInstance().removeService( "SinkSynchronizer"); 
  
  // do cleanup
  delete synchronizer;
  delete acceptor;
  delete app;
  return 0;
}
