/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

// rtp streaming client for video

#include <iostream>
#include <getopt.h>
#include "nmm/utils/directory_utils.hpp"
#include "nmm/comm/netstrategy/RTPAddress.hpp"
#include "nmm/comm/netstrategy/MulticastIPAddress.hpp"
// include NMM 
#include "nmm/multimedia/connect.hpp"
#include "nmm/comm/ProxyObject.hpp"
#include "nmm/services/registry/ProxyApplication.hpp"
#include "nmm/services/registry/GraphDescription.hpp"

// include used interfaces
#include "nmm/plugins/interfaces/IFileHandler.hpp"
#include "nmm/plugins/net/INet.hpp"


// include used interfaces
#include "nmm/multimedia/sync/interfaces/ISynchronizedSink.hpp"
#include "nmm/multimedia/sync/MultiAudioVideoSynchronizer.hpp" 
#include "nmm/utils/time/Clock.hpp" 
#include "nmm/comm/ReferenceManager.hpp"
#include "nmm/comm/ServiceProxy.hpp"
#include "nmm/comm/netstrategy/TCPAddress.hpp"
#include "nmm/comm/netstrategy/TCPAcceptor.hpp"

// import namespace
using namespace NMM;


void usage(const string& name) {
  cerr << "Usage :" << endl
       << "  " << name << " [options] -a <audio-port> -v <video-port>" << endl
       << endl
       << "Options:" << endl
       << "  -a <aduio-port> : Sets the multicast port to stream audio " << endl
       << "  -v <video-port> : Sets the multicast port to stream video " << endl
       << "  -V              : increase vervbose mode " << endl
       << "  -t              : receive transcoded video " << endl
       << "  -H <Host>       : connects to specified host to request a Synchronizer" << endl
       << "  -h : Show this help" << endl
       << endl;
}


int main(int argc, char ** argv)
{
  if(argc < 3) {
    usage(argv[0]);
    return 1;
  }
  
  cerr << argv[0] << " running ... " << endl << endl;
  
  int audio_rtp_port  = 0;
  int audio_rtcp_port = 0;  
  int video_rtp_port  = 0;
  int video_rtcp_port = 0;
  int transcoded_video_rtp_port  = 0;
  int transcoded_video_rtcp_port = 0;

  bool ac3_mpeg = false;
  
  string host;
  bool is_master = true;
  
  MultiAudioVideoSynchronizer* synchronizer = 0;
  ISinkSynchronizer_var        i_sink_synchronizer;
  TCPAcceptor* acceptor = 0;
  
  bool build_transcoding_graph = false;
  
  const char *opt_string = "H:a:v:dVth";
  struct option options [] = {
    {"audio-port",    1,NULL,'a'},
    {"video-port"    ,1,NULL,'v'},
    {"verbose",       0,NULL,'V'},
    {"transcoding",   0,NULL,'t'},
    {"dvd",           0,NULL,'d'},
    {"host",          1,NULL,'H'},
    {"help",          0,NULL,'h'},
    {NULL,            0,NULL,0}
  };
  
  int ch,long_option_index;
  int verbose = 0;
        
  while((ch=getopt_long(argc,argv,opt_string,options,&long_option_index))!=EOF){
      // which option did we get?
    switch (ch) {
      case 'a':
        audio_rtp_port  = atoi(optarg);
	audio_rtcp_port = audio_rtp_port+1;
        break;
      case 'v':
        video_rtp_port  = atoi(optarg);
	video_rtcp_port = video_rtp_port+1;
	transcoded_video_rtp_port  = video_rtp_port+2;
	transcoded_video_rtcp_port = video_rtp_port+3;
        break;
      case 'd':
        ac3_mpeg = true;
        break;
      case 'V': 
        ++verbose;
        break;
      case 'H':
        host = optarg;
	is_master = false;
	break;	
      case 't': 
        build_transcoding_graph = true;
        break;
      case 'h':
        usage(argv[0]);
        exit(0);         	
      default:
        usage(argv[0]);
        throw Exception("Invalid option");
    }
  }

  
  // disable all debug output (error, warning, message, debug)
  // see documentation Debug, Error, Warning and Message Output in NMM
  NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);
  if (verbose > 0) {
    NamedObject::getGlobalInstance().setErrorStream(&cout, NamedObject::ALL_LEVELS);
  }
  if (verbose > 1) {
    NamedObject::getGlobalInstance().setWarningStream(&cout, NamedObject::ALL_LEVELS);
  } 
  if (verbose > 2) {
    NamedObject::getGlobalInstance().setMessageStream(&cout, NamedObject::ALL_LEVELS);
  }
  if (verbose > 3) {
    NamedObject::getGlobalInstance().setDebugStream(&cout, NamedObject::ALL_LEVELS);
  }

  // the graph description for the example
  GraphDescription graph;

  // the NMM application for the example
  NMMApplication* app = ProxyApplication::getApplication(argc, argv);

  // create the descriptions 
  NodeDescription audio_netsource_descr("NetSourceNode");
  NodeDescription video_netsource_descr("NetSourceNode");
  NodeDescription mpeg_video_dec("MPEGVideoDecodeNode");
  NodeDescription ffmpeg_dec("FFMpegDecodeNode"); 
  NodeDescription display("XDisplayNode");
  NodeDescription mpeg_audio_dec("MPEGAudioDecodeNode");   
  NodeDescription playback_dec("ALSAPlaybackNode");


  // create a flow graph:  

  // define receiving audio format 
  Format mp3_format (Format::audio_mpeg_type);
  mp3_format.addIntParamValue(Format::sample_rate_param, 48000) ; 
  mp3_format.addIntParamValue(Format::channels_param, 2);
  mp3_format.addIntParamValue(Format::layer_param, 3);
  mp3_format.addIntParamValue(Format::bitpersample_param, 16);

  // build audio part of graph	
  graph.addEdge(audio_netsource_descr, mpeg_audio_dec, &mp3_format);
  graph.addEdge(mpeg_audio_dec,  playback_dec);

  // define receiving video format    
  Format video_transcode_format(Format::video_h263plus_type);
  video_transcode_format.addIntParamValue(Format::x_resolution_param, 320);
  video_transcode_format.addIntParamValue(Format::y_resolution_param, 240);
  video_transcode_format.addIntParamValue(Format::ratio_x_param, 320);
  video_transcode_format.addIntParamValue(Format::ratio_y_param, 240);
  video_transcode_format.addFloatParamValue(Format::framerate_param, 25.0);

  // define alternative receiving video format 
  Format video_format(Format::video_mpeg_type);
  video_format.addIntParamValue(Format::version_param, 2);
  video_format.addIntParamValue(Format::x_resolution_param, 720);
  video_format.addIntParamValue(Format::y_resolution_param, 540);
  video_format.addIntParamValue(Format::ratio_x_param, 4);
  video_format.addIntParamValue(Format::ratio_y_param, 3);
  video_format.addFloatParamValue(Format::framerate_param, 25.0);

  // build video part of graph      
  if (!build_transcoding_graph) {
    graph.addEdge(video_netsource_descr, mpeg_video_dec, &video_format);
    graph.addEdge(mpeg_video_dec, display);
  } else {
    graph.addEdge(video_netsource_descr, ffmpeg_dec, &video_transcode_format);
    graph.addEdge(ffmpeg_dec, display);
  } 
  
  // request complete graph
  cerr << "request GraphDescription from registry" << endl;
  ClientRegistry& registry = app->getRegistry();
  registry.requestGraph(graph);
  
  // Try to allocate or to request a Synchronizer 
  try {  
    if ( is_master ) {
      acceptor = new TCPAcceptor(9000);
      
      cerr << "master application " << endl;
      // Create a synchronizer that can also be used by other applications 
      synchronizer = new MultiAudioVideoSynchronizer();
      i_sink_synchronizer.reset( synchronizer->getInterface<ISinkSynchronizer>() );      
      
      // Register the Synchronizer as service 
      SReferenceManager::getInstance().registerService( "SinkSynchronizer", synchronizer );
    } else {
      cerr << "Connect to host " << host << endl;
      ServiceProxy service_proxy;
      
      try {
        TCPAddress tcp_address( host, 9000 );
        service_proxy.connectTo(tcp_address, "SinkSynchronizer");
        
        // Get the interface of the remote rtp client
        i_sink_synchronizer.reset( service_proxy.getCheckedInterface<ISinkSynchronizer>() );

      } catch (const BindingException& e) {
        GLOBAL_ERROR_STREAM("Could not connect to location: " << host << ":" << 9000 << endl);
        cerr << e << endl;
        return 0;
      }      
    }
  } catch ( const Exception& e ) {
    cerr << e << endl;  
  }
  
  Clock clock;
  Time start_time;
  Time graph_started;
    
  // set the audio rtp payload type
  INode* netsource1 =  graph.getINode(audio_netsource_descr); 
  INet_var i_net (netsource1->getParentObject()->getCheckedInterface<INet>());
  i_net->setPayloadType(14); //MP3 Audio
  
  // set the video rtp payload type
  INode* netsource2 = graph.getINode(video_netsource_descr);
  INet_var i_net2 (netsource2->getParentObject()->getCheckedInterface<INet>());
  if ( build_transcoding_graph ) {
      i_net2->setPayloadType(34);    
  }  else {
      i_net2->setPayloadType(32);
  }
      
  cerr << "realizeGraph()" << endl;
  graph.realizeGraph();
  
  if (! is_master ) {
    cerr << "Enter any key and press <Enter> to start 'listen in' " << endl;
    string s;
    cin >> s;
  }  

  start_time = clock.getTime();
      
  // define Multicast IP address for audio
  MulticastIPAddress ip_addr("239.0.0.2");
  RTPAddress rtp_addr (ip_addr, audio_rtp_port, audio_rtcp_port);
  i_net->setTransportStrategy(&rtp_addr);
  
  // define Multicast IP address for video
  MulticastIPAddress ip_addr2("239.0.0.2");
  if ( build_transcoding_graph ) {
    RTPAddress rtp_addr2 (ip_addr2, transcoded_video_rtp_port, transcoded_video_rtcp_port);
    i_net2->setTransportStrategy(&rtp_addr2);
  } else {
    RTPAddress rtp_addr2 (ip_addr2, video_rtp_port, video_rtcp_port);
    i_net2->setTransportStrategy(&rtp_addr2);
   }
  
  try {  
    // Use the allocated synchronizer 
    ISession* i_session = graph.getISession(); 
    i_session->setSinkSynchronizer( *i_sink_synchronizer.get() );
    graph.connectSinkSynchronizer();


    // start graph    
    cerr << "startGraph()" << endl;
    graph.startGraph();
    graph_started = clock.getTime();
    if (! is_master ) {
      cerr << "Start Time : " << start_time << endl;
      cerr << "Graph Started : " << graph_started << endl; 
    }
  } catch (const Exception& e) { 
    cerr << e << endl;
  }  
      
  // wait ...
  cerr << "Playing audio ... enter 'q <enter>' to exit." << endl;
  string s;
  cin >> s;
    
  // stop and release graph
  cerr << "stopGraph()" << endl;
  graph.stopGraph();

  cerr << "releaseGraph()" << endl;
  registry.releaseGraph(graph);
  
    
  // do cleanup
  delete synchronizer;
  delete acceptor;
  delete app;
  return 0;
}
