/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

// This example demonstrates how to discover running server registries on other hosts than localhost.
// In order to take full advantage of this feature, the names and
// ports printed out could be used to request nodes.
// notice: you need a running 'serverregistry' on the remote host(s)

#include <iostream>

// include NMM 
#include "nmm/comm/ProxyObject.hpp"
#include "nmm/services/registry/ProxyApplication.hpp"
#include "nmm/comm/ProcessUserInfo.hpp"

// include used interfaces
#include "nmm/plugins/interfaces/IFileHandler.hpp"
#include "nmm/plugins/audio/interfaces/IAudioDevice.hpp"

// import namespace
using namespace NMM;

int main(int argc, char ** argv)
{
  
  cerr << "Usage:" << endl
       << "  " << argv[0] << " [registry port]" << endl
       << endl << endl;

  int registry_port = ClientRegistry::default_port;

  if(argc >= 2) {
    registry_port = atoi(argv[1]);
  }

  cout << "Scanning for registries running on port " << registry_port << " ... " << endl << endl;

  // disable all debug output (error, warning, message, debug)
  // see documentation Debug, Error, Warning and Message Output in NMM
  NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);

  // the graph description for the example
  GraphDescription graph;

  // the NMM application for the example
  NMMApplication* app = 0;

  // NMM-operations may throw exception, in this simple example, all exceptions
  // are just caught at the end, and the application is terminated
  try {
    // create application object for example
    app = ProxyApplication::getApplication(argc, argv, SProcessUserInfo::getInstance().getHostname(), registry_port);

    //request client registry from application
    ClientRegistry& registry = app->getRegistry();

    // simple NMM discovery service: scan for running serverregistries that are using a certain registry port.
    // The first argument of this method specifies this registry por. The second argument specifies how long (in ms) this 
    // method waits to receive responses from running serverregistries before it returns. The default value is 50 ms. 
    list <RegistryName> result = registry.findServerRegistries(registry_port, 50);
  
    // print all available serverregistries 
    // in order to take full advantage of this feature, the names and
    // ports printed out could be used to request nodes 
    cerr << "Available registries " << endl;
    for ( list<RegistryName>::iterator it = result.begin(); it != result.end(); ++it)
    {
      cerr << "Found registry on " <<  it->first << " : " << it->second << endl;
    }

    // wait ...
    cerr << "Searching for running server registries ... enter 'q <enter>' to exit." << endl;
    string s;
    cin >> s;
    cerr << "Quit ... " << endl;

    delete app;
    return 0;
  }
  catch(const Exception& e) { // catch NMM-exception and print it
    cerr << e << endl;

    delete app;
    return 1;
  }
  catch(const SerializedException& e) { // catch serialized NMM-exception and print it
    cerr << e << endl;

    delete app;
    return 1;
  }
  catch(...) { // any other exception
    cerr << "something went wrong!" << endl;

    delete app;
    return 1;
  }
}
