/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

// simple wav player creating flow graph manually with error handling and access to 'serverregistry'

#include <iostream>

// include NMM
#include "nmm/comm/ProxyObject.hpp"
#include "nmm/services/registry/ProxyApplication.hpp"
#include "nmm/services/DefaultSinks.hpp"
#include "nmm/multimedia/connect.hpp"
#include "nmm/multimedia/NodeDescription.hpp"
#include "nmm/services/registry/RegistryLock.hpp"

// include used interfaces
#include "nmm/plugins/interfaces/IFileHandler.hpp"

// import namespace
using namespace NMM;

int main(int argc, char ** argv)
{
  if(argc != 2) {
    cerr << "Usage:" << endl
	 << "  " << argv[0] << " <wav file>" << endl
	 << endl;
    return 1;
  }
  cerr << argv[0] << " running ... " << endl << endl;

  // disable all debug output (error, warning, message, debug)
  // see documentation Debug, Error, Warning and Message Output in NMM
  NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);

  // the NMM application for the example
  NMMApplication* app = 0;

  // the _var variables for the INodes (C++ auto_ptr<>)
  INode_var readfile;
  INode_var audioplay;

  // NMM-operations may throw exception, in this simple example, all exceptions
  // are just caught at the end, and the application is terminated
  try {
    // create application object for example
    app = ProxyApplication::getApplication(argc, argv);

    // create the descriptions for the WavReadNode and platform specific PlaybackNode
    // using default name for used platform (different for Linux, MacOS, Windows, etc.)
    NodeDescription readfile_descr("WavReadNode");
    NodeDescription audioplay_descr(DEFAULT_AUDIO_SINK);

    // request Nodes: 
    // lock registry with guard, init request for each node, instantiate first result for every request 
    cerr << "request Nodes from registry" << endl;
    ClientRegistry& registry = app->getRegistry();
    {
      RegistryLock lock(registry);

      list<Response> readfile_response = registry.initRequest(readfile_descr);
      list<Response> audioplay_response = registry.initRequest(audioplay_descr);

      if (readfile_response.empty() || audioplay_response.empty() ) {
	cerr << "size of readfile_response  " << readfile_response.size() << endl;
	cerr << "size of audioplay_response " << audioplay_response.size() << endl;
	throw Exception("a required node is not available");
      }

      readfile.reset(registry.requestNode(readfile_response.front()));
      audioplay.reset(registry.requestNode(audioplay_response.front()));
    } // end of scope for lock -> will release lock

    // set the filename by requesting appropriate interface
    cerr << "set filename to " << argv[1] << endl;
    IFileHandler_var filehandler (readfile->getParentObject()->getCheckedInterface<IFileHandler>());
    filehandler->setFilename(argv[1]);

    // create a flow graph: readfile -> audioplay

    // initialize all three nodes ...
    readfile->init();
    audioplay->init();

    // initialize output of source node
    readfile->initOutput();

    // now, output of source node can be connected to initialized input of audio sink
    cerr << "connect(readfile, audioplay)" << endl;
    connect(readfile, audioplay);

    // activate source node so that audio sink can initialize its output
    readfile->activate();
    audioplay->initOutput();
    
    // audio sink can be activated now
    audioplay->activate();

    // ... and finally start them!
    cerr << "start() all Nodes" << endl;
    readfile->start();
    audioplay->start();
    
    // wait ...
    cerr << "Playing audio ... enter 'q <enter>' to exit." << endl;
    string s;
    cin >> s;
    
    // stop nodes, reset and destroy nodes
    cerr << "stop(), flush(), deactivate(), deinitOutput(), deinit() all Nodes" << endl;
    readfile->stop();
    audioplay->stop();

    readfile->flush();
    audioplay->flush();

    readfile->deactivate();
    audioplay->deactivate();

    readfile->deinitOutput();
    audioplay->deinitOutput();

    readfile->deinit();
    audioplay->deinit();

    // Release the reserved nodes
    cerr << "release all Nodes" << endl;
    registry.releaseNode(*readfile);
    registry.releaseNode(*audioplay);

    // need to first release INodes before deleting app
    readfile.reset();
    audioplay.reset();
    delete app;

    return 0;
  }
  catch(const Exception& e) { // catch NMM-exception and print it
    cerr << e << endl;

    // try to release the reserved nodes
    cerr << "release all Nodes" << endl;
    ClientRegistry& registry = app->getRegistry();
    RegistryLock lock(registry);
    if(readfile.get()) {
      registry.releaseNode(*readfile);
    }
    if(audioplay.get()) {
      registry.releaseNode(*audioplay);
    }

    // need to first release INodes before deleting app
    readfile.reset();
    audioplay.reset();
    delete app;

    return 1;
  }
  catch(const SerializedException& e) { // catch serialized NMM-exception and print it
    cerr << e << endl;

    // try to release the reserved nodes
    cerr << "release all Nodes" << endl;
    ClientRegistry& registry = app->getRegistry();
    RegistryLock lock(registry);
    if(readfile.get()) {
      registry.releaseNode(*readfile);
    }
    if(audioplay.get()) {
      registry.releaseNode(*audioplay);
    }

    // need to first release INodes before deleting app
    readfile.reset();
    audioplay.reset();
    delete app;

    return 1;
  }
  catch(...) { // any other exception
    cerr << "something went wrong!" << endl;

    // try to release the reserved nodes
    cerr << "release all Nodes" << endl;
    ClientRegistry& registry = app->getRegistry();
    RegistryLock lock(registry);
    if(readfile.get()) {
      registry.releaseNode(*readfile);
    }
    if(audioplay.get()) {
      registry.releaseNode(*audioplay);
    }
    
    // need to first release INodes before deleting app
    readfile.reset();
    audioplay.reset();
    delete app;

    return 1;
  }
}
