/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


// simple wav player using GraphDescription with error handling and access to 'serverregistry'
// same code as helloworld1, but added listener for endTrack event

#include <iostream>

// include NMM
#include "nmm/comm/ProxyObject.hpp"
#include "nmm/services/registry/ProxyApplication.hpp"
#include "nmm/services/DefaultSinks.hpp"
#include "nmm/comm/serialize/SerializeResult.hpp"

// include used interfaces
#include "nmm/plugins/interfaces/IFileHandler.hpp"
#include "nmm/multimedia/interfaces/ITrack.hpp"

// import namespace
using namespace NMM;

// a simple listener class
class TrackListener {
public:
  Result endTrack() 
  {
    cerr << endl << "endTrack" << endl << endl;
    return SUCCESS;
  }

  Result startTrack(const string& filename) 
  {
    cerr << endl << "startTrack " << filename << endl << endl;
    return SUCCESS;
  }
};

int main(int argc, char ** argv)
{
  if(argc != 2) {
    cerr << "Usage:" << endl
	 << "  " << argv[0] << " <wav file>" << endl
	 << endl;
    return 1;
  }
  cerr << argv[0] << " running ... " << endl << endl;

  // disable all debug output (error, warning, message, debug)
  // see documentation Debug, Error, Warning and Message Output in NMM
  NamedObject::getGlobalInstance().setErrorStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setWarningStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setMessageStream(0, NamedObject::ALL_LEVELS);
  NamedObject::getGlobalInstance().setDebugStream(0, NamedObject::ALL_LEVELS);

  // the graph description for the example
  GraphDescription graph;

  // the NMM application for the example
  NMMApplication* app = 0;

  // NMM-operations may throw exception, in this simple example, all exceptions
  // are just caught at the end, and the application is terminated
  try {
    // create application object for example
    app = ProxyApplication::getApplication(argc, argv);

    // create the descriptions for the WavReadNode and platform specific PlaybackNode
    // using default name for used platform (different for Linux, MacOS, Windows, etc.)
    NodeDescription readfile_descr("WavReadNode");
    NodeDescription audioplay_descr(DEFAULT_AUDIO_SINK);

    // create a flow graph: readfile -> audioplay
    graph.addEdge(readfile_descr, audioplay_descr);

    // request complete graph
    cerr << "request GraphDescription from registry" << endl;
    ClientRegistry& registry = app->getRegistry();
    registry.requestGraph(graph);

    // set the filename by requesting appropriate interface
    cerr << "set filename to " << argv[1] << endl;
    INode* readfile = graph.getINode(readfile_descr);
    IFileHandler_var filehandler (readfile->getParentObject()->getCheckedInterface<IFileHandler>());
    filehandler->setFilename(argv[1]);

    // register listener for events startTrack and endTrack of ITrack interface at audio sink node
    // -> the method TrackListener::startTrack will get triggered when the stream starts at audio sink
    // -> the method TrackListener::endTrack will get triggered when the stream ended at audio sink
    //
    // if you are unsure about the correct syntax for TEDObjects,
    // please refer to hpp/cpp sources generated from corresponding
    // .idl file
    INode* audioplay = graph.getINode(audioplay_descr);
    TrackListener end_track_listener;
    audioplay->getParentObject()->registerEventListener(ITrack::endTrack_event, 
							new TEDObject0<TrackListener>(&end_track_listener, &TrackListener::endTrack));
    audioplay->getParentObject()->registerEventListener(ITrack::startTrack_event, 
							new TEDObject1<TrackListener, string, Result(TrackListener:: * )(const string&), Result>
							(&end_track_listener, &TrackListener::startTrack));

    // realize and start graph
    cerr << "realizeGraph()" << endl;
    graph.realizeGraph();
    cerr << "startGraph()" << endl;
    graph.startGraph();
    
    // wait ...
    cerr << "Playing audio ... enter 'q <enter>' to exit." << endl;
    string s;
    cin >> s;
    
    // stop and release graph
    cerr << "stopGraph()" << endl;
    graph.stopGraph();

    cerr << "releaseGraph()" << endl;
    registry.releaseGraph(graph);

    delete app;
    return 0;
  }
  catch(const Exception& e) { // catch NMM-exception and print it
    cerr << e << endl;

    // release graph
    cerr << "releaseGraph()" << endl;
    ClientRegistry& registry = app->getRegistry();
    registry.releaseGraph(graph);

    delete app;
    return 1;
  }
  catch(const SerializedException& e) { // catch serialized NMM-exception and print it
    cerr << e << endl;

    // release graph
    cerr << "releaseGraph()" << endl;
    ClientRegistry& registry = app->getRegistry();
    registry.releaseGraph(graph);

    delete app;
    return 1;
  }
  catch(...) { // any other exception
    cerr << "something went wrong!" << endl;

    // release graph
    cerr << "releaseGraph()" << endl;
    ClientRegistry& registry = app->getRegistry();
    registry.releaseGraph(graph);

    delete app;
    return 1;
  }
}
