/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "CompositeStrategy.hpp"

#include "IConnectionManager.hpp"

#include "nmm/utils/thread/MutexGuard.hpp"

#include <cassert>

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  CompositeStrategy::CompositeStrategy(const string& name)
    : TransportStrategy(name),
      __partner_icm(0),
      __partner_reference(0)
  {}
  
  CompositeStrategy::~CompositeStrategy() {
    delete __partner_icm;
    removeAllChildStrategies();    
  }
  
  void CompositeStrategy::scheduleDestruction() {
    delete this;
  }
  
  string CompositeStrategy::getStrategyType() const {
    return "CompositeStrategy";
  }
  
  Result CompositeStrategy::sendReliableEvent(Event& event) const {
    // XXX: fw: sendReliableEvent
//     for(unsigned int i=0; i<getChildStrategyCount(); ++i) {
//       if(dynamic_cast<ITSRequestReplyEventReceiver*>(getOwnChildStrategy(i))) {
//         return getOwnChildStrategy(i)->sendReliableEvent(event);
//       }
//     }
    
    if(__partner_icm && __partner_reference) {
      /* send through connection manager */
      return __partner_icm->sendReliableEvent(__partner_reference, event);
    }
    
    GLOBAL_ERROR_STREAM1("CompositeStrategy> Failed to send reliable event" << endl);
    return FAILURE;
  }
  
  Result CompositeStrategy::sendReliableConfigEvent(Event& event) const {
    // XXX: fw: sendReliableEvent
//     for(unsigned int i=0; i<getChildStrategyCount(); ++i) {
//       if(dynamic_cast<ITSRequestReplyEventReceiver*>(getOwnChildStrategy(i))) {
//         return getOwnChildStrategy(i)->sendReliableConfigEvent(event);
//       }
//     }
    
    if(__partner_icm && __partner_reference) {
      /* send through connection manager */
      return __partner_icm->sendReliableConfigEvent(__partner_reference, event);
    }
    
    GLOBAL_ERROR_STREAM1("CompositeStrategy> Failed to send reliable config event" << endl);
    return FAILURE;
  }
  
  nmm_uint32 CompositeStrategy::getChildStrategyCount() const {
    MutexGuard guard(__child_strategy_mutex);
    return static_cast<nmm_uint32>(__child_strategies.size());
  }
  
  ITransportStrategy* CompositeStrategy::getChildStrategy(const nmm_uint32& index) const {
    MutexGuard guard(__child_strategy_mutex);
    return getOwnChildStrategy(index)->getInterface<ITransportStrategy>();
  }
  
  TransportStrategy* CompositeStrategy::getOwnChildStrategy(unsigned int i) const {
    MutexGuard guard(__child_strategy_mutex);
    assert(i < __child_strategies.size());
    return __child_strategies[i];
  }
  
  unsigned int CompositeStrategy::addChildStrategy(TransportStrategy* strategy) {
    MutexGuard guard(__child_strategy_mutex);
    
    /* add to strategy list */
    __child_strategies.push_back(strategy);
    unsigned int index = __child_strategies.size() - 1;
    
    /* set strategy name */
    // XXX: fw: TransportStrategy is not a NamedObject
//     std::ostringstream stream;
//     stream << getName() << "> " << index << "> " << strategy->getName();
//     strategy->setName(stream.str());
    
    /* set child strategy event receiver */
    strategy->setEventReceiver(this);
    
    /* return index */
    return index;
  }
  
  void CompositeStrategy::removeChildStrategy(unsigned int index) {
    MutexGuard guard(__child_strategy_mutex);
    
    if(index >= __child_strategies.size()) {
      return;
    }
    
    TransportStrategy* strategy = __child_strategies[index];
    __child_strategies[index] = __child_strategies[__child_strategies.size()-1];
    __child_strategies.pop_back();
    
    strategy->scheduleDestruction();
  }
  
  void CompositeStrategy::removeAllChildStrategies() {
    MutexGuard guard(__child_strategy_mutex);
    
    /* schedule destruction of all child strategies */
    for(unsigned int i=0; i<__child_strategies.size(); ++i) {
      __child_strategies[i]->scheduleDestruction();
      __child_strategies[i] = 0;
    }
    __child_strategies.clear();
  }
  
  void CompositeStrategy::setConnectionManager(IConnectionManager* partner_icm, const NMMReference& partner_reference) {
    __partner_icm = partner_icm;
    __partner_reference = partner_reference;
  }
  
  bool CompositeStrategy::isConnected() const {
    /* if the binding is not complete, the composite strategy is not
       connected. */
    if(!isBindingComplete()) {
      return false;
    }
    
    /* a cmoplete composite strategy is connected if all child strategies are
       connected. */
    for(unsigned int i=0; i<__child_strategies.size(); ++i) {
      if(!__child_strategies[i]->isConnected()) {
	return false;
      }
    }
    
    return true;
  }
  
}

#include "nmm/nmm_suffix.hpp"
