/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "ConnectionManager.hpp"
#include "ConnectorAcceptor.hpp"
#include "ConnectorAddress.hpp"

#include "OutOfBandChannel.hpp"
#include "ReferenceManager.hpp"
#include "FormatStrategyMap.hpp"
#include "TSFactory.hpp"
#include "ProxyObject.hpp"

#include "InstreamCompositeStrategy.hpp"

#include "nmm/comm/messaging/CEvent.hpp"
#include "nmm/comm/messaging/IMessageReceiver.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  ConnectionManager::ConnectionManager()
    : NamedObject("ConnectionManager"),
      __format_strategy_map(NULL)
  {
    /* Begin> Default OutOfBandChannel strategies */
    // XXX: This is a temporary solution until we have configuration files.
    setDefaultStrategy(TypeInfo("OutOfBandChannel", "NMM"), TSAddress::LOCAL,   "LocalStrategy");
    setDefaultStrategy(TypeInfo("OutOfBandChannel", "NMM"), TSAddress::SYSTEM,  "TCPStrategy");
    setDefaultStrategy(TypeInfo("OutOfBandChannel", "NMM"), TSAddress::NETWORK, "TCPStrategy");
    /* End< */
    
    /* Begin> Default InstreamChannel strategies */
    // XXX: This is a temporary solution until we have configuration files.
    setDefaultStrategy(TypeInfo("InstreamChannel", "NMM"), TSAddress::LOCAL,   "LocalStrategy");
    setDefaultStrategy(TypeInfo("InstreamChannel", "NMM"), TSAddress::SYSTEM,  "TCPStrategy");
    setDefaultStrategy(TypeInfo("InstreamChannel", "NMM"), TSAddress::NETWORK, "TCPStrategy");
    /* End< */
    
    /* Regster this object as a service */
    SReferenceManager::getInstance().registerService("ConnectionManager", this);
    
    __format_strategy_map = new FormatStrategyMap();
  }
  
  ConnectionManager::~ConnectionManager() {
    delete __format_strategy_map;
  }
  
  void ConnectionManager::handleConnect(const TSAddress* remote_addr, TSAddress*& local_addr) const
    throw (const BindingException&)
  {
    /* Get the strategy type */
    if (!remote_addr) {
      ERROR_STREAM("No remote address given!" << endl);
      throw BindingException("No remote address given");
    }
    string strategy_type = remote_addr->getStrategyType();
    
    /* Request a connector from the TSFactory */
    ConnectorAcceptor* acceptor = STSFactory::getInstance().requestConnectorAcceptor(strategy_type);
    if (!acceptor) {
      ERROR_STREAM("No acceoptor available for " << strategy_type << endl);
      throw BindingException("No acceptor available for " + strategy_type);
    }
    
    /* Forward request to acceptor */
    // XXX: need a dynamic cast here because we can't pass ConnectorAddress to the interface
    const ConnectorAddress* remote_connector_address = dynamic_cast<const ConnectorAddress*>(remote_addr);
    ConnectorAddress* local_connector_address;
    acceptor->handleConnect(remote_connector_address, local_connector_address);
    local_addr = local_connector_address;
  }
  
  Result ConnectionManager::sendReliableEvent(const NMMReference& receiver, Event& e) const
    throw(const BindingException&)
  {
    /* Get a pointer to the TransportStrategy from the ReferenceManager */
    TransportStrategy* ts = SReferenceManager::getInstance().getObject<TransportStrategy>(receiver);
    if (!ts) {
      ERROR_STREAM("Invalid reference to receiver TransportStrategy" << endl);
      throw BindingException("Invalid reference to receiver TransportStrategy");
    }
    
    /* Check if the TransportStrategy has an event receiver */
    if (!ts->getEventReceiver()) {
      ERROR_STREAM("Receiver TransportStrategy has no event receiver" << endl);
      throw BindingException("Receiver TransportStrategy has no event receiver");
    }
    
    /* Dispatch event to IEventReceiver of transport strategy */
    Result result = ts->getEventReceiver()->receiveEvent(e, ts);
    
    /* Return result */
    return result;
  }

  Result ConnectionManager::sendReliableConfigEvent(const NMMReference& receiver, Event& e) const
    throw(const BindingException&)
  {
    /* Get a pointer to the TransportStrategy from the ReferenceManager */
    TransportStrategy* ts = SReferenceManager::getInstance().getObject<TransportStrategy>(receiver);
    if (!ts) {
      ERROR_STREAM("Invalid reference to receiver TransportStrategy" << endl);
      throw BindingException("Invalid reference to receiver TransportStrategy");
    }
    
    /* Transport strategy dispatches event */
    Result result = ts->getEventDispatcher().dispatch(e, ts);
    
    /* Return result */
    return result;
  }

  Result ConnectionManager::sendReliableInstreamMessage(const NMMReference& receiver, CEvent& message) const
    throw(const BindingException&)
  {
    DEBUG_STREAM2("Received instream event" << endl);
    
    /* Get a pointer to the TransportStrategy from the ReferenceManager */
    InstreamCompositeStrategy* ts = SReferenceManager::getInstance().getObject<InstreamCompositeStrategy>(receiver);
    if (!ts) {
      ERROR_STREAM("Invalid reference to receiver TransportStrategy" << endl);
      throw BindingException("Invalid reference to receiver TransportStrategy");
    }
    
    /* Check if the TransportStrategy has a message receiver */
    if (!ts->getMessageReceiver()) {
      ERROR_STREAM("Receiver TransportStrategy has no message receiver" << endl);
      throw BindingException("Receiver TransportStrategy has no message receiver");
    }
    
    /* Dispatch event to IMessageReceiver of transport strategy */
    Result result = ts->getMessageReceiver()->receiveMessage(new CEvent(message), ts);
    
    /* Return result */
    return result;
  }
  
  list<TSDescription> ConnectionManager::getSupportedStrategies(const Format& format) const {
    return getFormatStrategyMap().getSupportedStrategies(format);
  }
  
  void ConnectionManager::connect(const ConnectorAddress* local_addr, ConnectorAddress*& remote_addr, IConnectionManager* remote_icm) const
    throw(const BindingException&) 
  {
    /* get the strategy type */
    if(!local_addr) {
      ERROR_STREAM2("No local address given!" << endl);
      throw BindingException("No local address given");
    }
    std::string strategy_type = local_addr->getStrategyType();
    
    /* request a ConnectorAcceptor from the TSFactory */
    DEBUG_STREAM2("Request connector for strategy type " << strategy_type << std::endl);
    ConnectorAcceptor* connector_acceptor = STSFactory::getInstance().requestConnectorAcceptor(strategy_type);
    if(!connector_acceptor) {
      ERROR_STREAM("No connector available for " << strategy_type << endl);
      throw BindingException("No connector available for " + strategy_type);
    }
    
    /* forward request to ConnectorAcceptor */
    connector_acceptor->connect(local_addr, remote_addr, remote_icm);
  }
  
  IConnectionManager* ConnectionManager::getConnectionManager(const Interface& iface) const
    throw(const BindingException&)
  {
    Interface* tmp_interface = getService(iface, "ConnectionManager");
    IConnectionManager* icm = tmp_interface->getParentObject()->getInterface<IConnectionManager>();
    if (!icm) {
      ERROR_STREAM("Found ConnectionManager service, but no interface to it" << endl);
      throw BindingException("Found ConnectionManager service, but no interface to it");
    }
    return icm;
  }
  
  Interface* ConnectionManager::getService(const TSAddress& address, const string& service) const
    throw(const BindingException&)
  {
    try {
      if (address.getBindingType() == TSAddress::LOCAL) {
        /* BEGIN> Handle local case */
        /* Get service from local reference manager */
        NMMObject* object = SReferenceManager::getInstance().getService(service);
        if (!object) {
          throw BindingException("No such service " + service);
        }
        
        /* Return an interface to it */
        return object->getInterface<Interface>();
        /* END< */
      } else {
        /* BEGIN> Handle distributed case */
        /* Create a ProxyObject for the service with automatic memory management */
        ProxyObject* new_service = new ProxyObject(true);
        
        /* set NMMReference to 0 */
        auto_ptr<TSAddress> tmp_address(address.copy());
        tmp_address->setNMMReference(0);
        
        /* and connect to the given address */
        new_service->connectToObject(*tmp_address.get());
        
        /* Get the IObjectAcceptor to access the services available on this Application. */
        IObjectAcceptor_var i_acceptor(new_service->getCheckedInterface<IObjectAcceptor>());
        
        /* Connect to the service */
        i_acceptor->connectToService(service);
        
        return i_acceptor->copy();
        /* END< */
      }
    } catch(const SerializedException& e) {
      throw BindingException(e.getComment());
    }
  }

  Interface* ConnectionManager::getService(const Interface& iface, const string& service) const
    throw(const BindingException&)
  {
    try {
      if (iface.isLocalTo(*getInterface<IConnectionManager>())) {
        /* BEGIN> Handle local case */
        /* Get service from local reference manager */
        NMMObject* object = SReferenceManager::getInstance().getService(service);
        if (!object) {
          throw BindingException("No such service " + service);
        }
        
        /* Return an interface to it */
        return object->getInterface<Interface>();
        /* END< */
      } else {
        /* BEGIN> Handle distributed case */
        /* BEGIN> Create a copy of the interface with NMM reference set to NULL */
        Interface tmp_interface(iface);
        const SequenceValue<TSAddress> &addresses = tmp_interface.getAddressList();
        for (SequenceValue<TSAddress>::const_iterator i(addresses.begin());
             i != addresses.end();
             ++i) {
          (*i)->setNMMReference(0);
        }
        /* END< */
        
        /* Create a ProxyObject for the service with automatic memory management */
        ProxyObject* new_service = new ProxyObject(tmp_interface, true);
        
        /* Get the default address for the first TransportStrategy from the Interface */
        const TSAddress* address = const_cast<Interface&>(iface).getDefaultAddress(new_service->getOutOfBandChannel().getStrategyList().front());
        
        /* If no address available we can not connect to the object and an exception is thrown */
        if (!address) {
          throw BindingException("No address available\n");
        }
        
        /* connect to the NMMObject */
        new_service->connectToObject(*address);
        
        /* Get the IObjectAcceptor to access the services available on this Application. */
        IObjectAcceptor_var i_acceptor(new_service->getCheckedInterface<IObjectAcceptor>());
        
        /* Connect to the service */
        i_acceptor->connectToService(service);
        
        return i_acceptor->copy();
        /* END< */
      }
    } catch(const SerializedException& e) {
      ERROR_STREAM("Failed to get service" << endl);
      ERROR_STREAM(e << endl);
      throw BindingException(e.getComment());
    }
  }
  
  TSAddress* ConnectionManager::getDefaultAddress(const string& strategy_type) const {
    TSAddress* address = STSFactory::getInstance().getDefaultAddress(strategy_type);
    
    /* Suppress shared addresses */
    if (dynamic_cast<ConnectorAddress*>(address)) {
      delete address;
      return 0;
    }
    
    if (address) {
      //TSAddress* tmp_address = address->copy();
      address->setNMMReference(getReference());
      return address;
    }
    
    return 0;
  }
  
  string ConnectionManager::getDefaultStrategy(const TypeInfo& channel_type, TSAddress::BindingType binding_type) const {
    StrategyConstIterator found(__strategy_map.find(ConnectionType(channel_type, binding_type)));
    if (found == __strategy_map.end()) {
      return "";
    } else {
      return found->second;
    }
  }
  
  void ConnectionManager::setDefaultStrategy(const TypeInfo& channel_type, TSAddress::BindingType binding_type,
                                             const string& strategy_type) 
  {
    __strategy_map[ConnectionType(channel_type, binding_type)] = strategy_type;
  }
}

DEFINE_SINGLETON(NMM::ConnectionManager);

#include "nmm/nmm_suffix.hpp"
