/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "ConnectorAcceptor.hpp"
#include "IConnectionManager.hpp"
#include "TSAddressOwner.hpp"

#include "TransportStrategy.hpp"
#include "ConnectorAddress.hpp"
#include "TSFactory.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  ConnectorAcceptor::ConnectorAcceptor(const std::string& strategy_type)
    : __strategy_type(strategy_type)
  {}
  
  ConnectorAcceptor* ConnectorAcceptor::copy() const throw() {
    return new ConnectorAcceptor(*this);
  }
  
  ConnectorAcceptor::~ConnectorAcceptor()
  {}
  
  void ConnectorAcceptor::handleConnect(const ConnectorAddress* remote_addr, ConnectorAddress*& local_addr)
    throw (const BindingException&)
  {
    TransportStrategy* strategy = NULL;
    
    try {
      GLOBAL_DEBUG_STREAM("ConnectorAcceptor: handle connection request" << std::endl);
      
      /* create a partner strategy */
      unsigned int partner_interfaces = TSDescription::getPartnerInterfaces(remote_addr->getStrategyInterfaces());
      GLOBAL_DEBUG_STREAM("ConnectorAcceptor: create partner strategy" << std::endl);
      strategy = STSFactory::getInstance().createTransportStrategy(getStrategyType(), partner_interfaces);
      if(!strategy) {
        GLOBAL_ERROR_STREAM("ConnectorAcceptor: Failed to create partner TransportStrategy" << endl);
        throw BindingException("Failed to create partner TransportStrategy");
      }
      
      /* get local reference */
      NMMReference local_ref = strategy->getReference();
      
      /* set event receiver so that other side can call connectToObject */
      strategy->setEventReceiver(this);
      TSAddressOwner* address_owner = dynamic_cast<TSAddressOwner*>(strategy);
      if(!address_owner) {
        GLOBAL_ERROR_STREAM("ConnectorAcceptor: Strategy has no TSAddressOwner interface" << std::endl);
        throw BindingException("Strategy has no TSAddressOwner interface");
      }
      
      /* connect local strategy to remote address */
      GLOBAL_DEBUG_STREAM("ConnectorAcceptor: Set remote address" << std::endl);
      strategy->setAddress(*remote_addr);
      
      /* get address of local strategy */
      GLOBAL_DEBUG_STREAM("ConnectorAcceptor: get local address" << std::endl);
      const TSAddress& ts_address = address_owner->getLocalAddress();
      local_addr = dynamic_cast<ConnectorAddress*>(ts_address.copy());
      local_addr->setTSReference(local_ref);
      
      GLOBAL_DEBUG_STREAM("ConnectorAcceptor: connection accepted" << std::endl);
    } catch(...) {
      /* clean up */
      if(strategy) {
        strategy->scheduleDestruction();
      }
      
      /* rethrow exception */
      throw;
    }
  }
  
  void ConnectorAcceptor::connect(const ConnectorAddress* local_addr, ConnectorAddress*& remote_addr, IConnectionManager* remote_icm)
    throw(const BindingException&)
  {
    GLOBAL_DEBUG_STREAM("ConnectorAcceptor: request connection" << std::endl);
    
    /* we need a remote connection manager */
    if(!remote_icm) {
      GLOBAL_ERROR_STREAM("ConnectorAcceptor: No remote connection manager interface given" << endl);
      throw BindingException("No remote connection manager interface given");
    }
    
    /* forward request to remote connection manager */
    // XXX: need a dynamic cast here because we can't pass ConnectorAddress to the interface
    TSAddress* remote_address = NULL;
    remote_icm->handleConnect(local_addr, remote_address);
    remote_addr = dynamic_cast<ConnectorAddress*>(remote_address);
    
    GLOBAL_DEBUG_STREAM("ConnectorAcceptor: connection successful" << std::endl);
  }
  
}

#include "nmm/nmm_suffix.hpp"
