/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "IConnectionManager.hpp"
#include "nmm/comm/serialize/SerializeResult.hpp"
#include "nmm/comm/serialize/InvalidTypeException.hpp"
#include "nmm/comm/serialize/TConstructor.hpp"

#include "nmm/nmm_prefix.hpp"

/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * See file LICENCE for terms and conditions of usage
 *
 */

#include "TSAddress.hpp"

#include "nmm/comm/format/Format.hpp"
#include "nmm/comm/messaging/CEvent.hpp"

#include "nmm/comm/serialize/StdValue.hpp"

namespace NMM
{
  void IConnectionManager::used() {} 
  
  
  void IConnectionManager::handleConnect(const TSAddress* remote_addr, TSAddress*& local_addr) const 
      throw(const NMM::SerializedException&, const BindingException&)
  {
    NMM::TInValue<TSAddress* > remote_addr_value (const_cast<TSAddress* >(remote_addr), false);
    NMM::TOutValue<TSAddress* > local_addr_value(local_addr);
    NMM::Event event("IConnectionManager::handleConnect", &remote_addr_value, &local_addr_value);
    if(const_cast<IConnectionManager*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        if(BindingException* ex = dynamic_cast<BindingException*>(event.getException())) {
          throw BindingException(*ex);
        }
        throw NMM::SerializedException(*event.getException());
      }
    }
    local_addr = local_addr_value.getRValue();
  }
  
  Result IConnectionManager::sendReliableEvent(const NMMReference& ts_reference, Event& e) const 
      throw(const NMM::SerializedException&, const BindingException&)
  {
    NMM::TOutValue<Result > return_value;
    NMM::TInValue<NMMReference > ts_reference_value(ts_reference, false);
    NMM::TInOutValue<Event > e_value;
    e_value.setValue(e);
    NMM::Event event("IConnectionManager::sendReliableEvent", &ts_reference_value, &e_value);
    event.setReturnValue(&return_value);
    if(const_cast<IConnectionManager*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        if(BindingException* ex = dynamic_cast<BindingException*>(event.getException())) {
          throw BindingException(*ex);
        }
        throw NMM::SerializedException(*event.getException());
      }
    }
    e = e_value.getRValue();
    return return_value.getRValue();
  }
  
  Result IConnectionManager::sendReliableConfigEvent(const NMMReference& receiver, Event& e) const 
      throw(const NMM::SerializedException&, const BindingException&)
  {
    NMM::TOutValue<Result > return_value;
    NMM::TInValue<NMMReference > receiver_value(receiver, false);
    NMM::TInOutValue<Event > e_value;
    e_value.setValue(e);
    NMM::Event event("IConnectionManager::sendReliableConfigEvent", &receiver_value, &e_value);
    event.setReturnValue(&return_value);
    if(const_cast<IConnectionManager*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        if(BindingException* ex = dynamic_cast<BindingException*>(event.getException())) {
          throw BindingException(*ex);
        }
        throw NMM::SerializedException(*event.getException());
      }
    }
    e = e_value.getRValue();
    return return_value.getRValue();
  }
  
  Result IConnectionManager::sendReliableInstreamMessage(const NMMReference& receiver, CEvent& message) const 
      throw(const NMM::SerializedException&, const BindingException&)
  {
    NMM::TOutValue<Result > return_value;
    NMM::TInValue<NMMReference > receiver_value(receiver, false);
    NMM::TInOutValue<CEvent > message_value;
    message_value.setValue(message);
    NMM::Event event("IConnectionManager::sendReliableInstreamMessage", &receiver_value, &message_value);
    event.setReturnValue(&return_value);
    if(const_cast<IConnectionManager*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        if(BindingException* ex = dynamic_cast<BindingException*>(event.getException())) {
          throw BindingException(*ex);
        }
        throw NMM::SerializedException(*event.getException());
      }
    }
    message = message_value.getRValue();
    return return_value.getRValue();
  }
  
  list<TSDescription >  IConnectionManager::getSupportedStrategies(const Format& format) const
      throw(const NMM::SerializedException&)
  {
    NMM::TOutValue<list<TSDescription >  > return_value;
    NMM::TInValue<Format > format_value(format, false);
    NMM::Event event("IConnectionManager::getSupportedStrategies", &format_value);
    event.setReturnValue(&return_value);
    if(const_cast<IConnectionManager*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
    return return_value.getRValue();
  }

  const char* IConnectionManager::IConnectionManager_type = "IConnectionManager";
  NMM::TConstructor<IConnectionManager > __IConnectionManager_prototype("IConnectionManager","NMM");


  const char* IConnectionManager::handleConnect_event = "IConnectionManager::handleConnect";
  const char* IConnectionManager::sendReliableEvent_event = "IConnectionManager::sendReliableEvent";
  const char* IConnectionManager::sendReliableConfigEvent_event = "IConnectionManager::sendReliableConfigEvent";
  const char* IConnectionManager::sendReliableInstreamMessage_event = "IConnectionManager::sendReliableInstreamMessage";
  const char* IConnectionManager::getSupportedStrategies_event = "IConnectionManager::getSupportedStrategies";


}


#include "nmm/nmm_suffix.hpp"

