/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#ifndef NMM_IOutOfBandCompositeStrategy_HPP
#define NMM_IOutOfBandCompositeStrategy_HPP

#include "nmm/nmm_prefix.hpp"

#include "nmm/comm/Interface.hpp"
#include "nmm/comm/messaging/Event.hpp"
#include "nmm/comm/messaging/IllegalStateException.hpp"
#include <memory>

#include "ICompositeStrategy.hpp"

#include "visibility.hpp"

namespace NMM
{
  /** Out-of-band composite transport strategy.
   * This is an interface for the %OutOfBandCompositeStrategy. It provides access to child transport strategies and
   * the request/reply unit. It also defines internal configuration events.
   */

  class DLLEXPORT IOutOfBandCompositeStrategy : virtual public ICompositeStrategy { 
  public:    /** Transport roles.
     */

  enum TransportRole { 
    RECEIVE_ONE_WAY_EVENT_REQUEST = 1 , ///< Receive one-way event request

    RECEIVE_ONE_WAY_EVENT_REPLY = 2 , ///< Receive one-way event reply

    SEND_ONE_WAY_EVENT_REQUEST = 4 , ///< Send one-way event request

    SEND_ONE_WAY_EVENT_REPLY = 8 , ///< Send one-way event reply

    RECEIVE_REQUEST_REPLY_EVENT = 16 , ///< Receive request-reply event

    SEND_REQUEST_REPLY_EVENT = 32 ///< Send request-reply event
 };

    IOutOfBandCompositeStrategy() : NMM::Interface(IOutOfBandCompositeStrategy::IOutOfBandCompositeStrategy_type, "NMM") { };

    IOutOfBandCompositeStrategy(NMM::NMMObject& parent) : NMM::Interface (parent, IOutOfBandCompositeStrategy::IOutOfBandCompositeStrategy_type, "NMM") { };

    virtual ~IOutOfBandCompositeStrategy() {};

    static const char* IOutOfBandCompositeStrategy_type;

    virtual  string getInterfaceType() const { return IOutOfBandCompositeStrategy_type; }

    NMM::Interface* copy() const { return new IOutOfBandCompositeStrategy(*this); }
    static void used(); 
    static const char* getOneWayEventRequestReceiverStrategy_event;
    static const char* getOneWayEventReplyReceiverStrategy_event;
    static const char* getOneWayEventRequestSenderStrategy_event;
    static const char* getOneWayEventReplySenderStrategy_event;
    static const char* getRequestReplyEventReceiverStrategy_event;
    static const char* getRequestReplyEventSenderStrategy_event;
    /** Returns child strategy for receiving one-way %Event requests.
     * If this method returns 0, then the binding is realized by a request-reply
     * leaf strategy, which is returned by getRequestReplyEventReceiverStrategy().
     *
     * @return Interface to child strategy for receiving one-way %Event requests.
     */

    virtual ITransportStrategy* getOneWayEventRequestReceiverStrategy() const
      throw(const NMM::SerializedException&);
    /** Returns child strategy for receiving one-way %Event replies.
     * If this method returns 0, then the binding is realized by a request-reply
     * leaf strategy, which is returned by getRequestReplyEventSenderStrategy().
     *
     * @return Interface to child strategy for receiving one-way %Event replies.
     */

    virtual ITransportStrategy* getOneWayEventReplyReceiverStrategy() const
      throw(const NMM::SerializedException&);
    /** Returns child strategy for sending one-way %Event requests.
     * If this method returns 0, then the binding is realized by a request-reply
     * leaf strategy, which is returned by getRequestReplyEventReceiverStrategy().
     *
     * @return Interface to child strategy for sending one-way %Event requests.
     */

    virtual ITransportStrategy* getOneWayEventRequestSenderStrategy() const
      throw(const NMM::SerializedException&);
    /** Returns child strategy for sending one-way %Event replies.
     * If this method returns 0, then the binding is realized by a request-reply
     * leaf strategy, which is returned by getRequestReplyEventSenderStrategy().
     *
     * @return Interface to child strategy for sending one-way %Event replies.
     */

    virtual ITransportStrategy* getOneWayEventReplySenderStrategy() const
      throw(const NMM::SerializedException&);
    /** Returns child strategy for receiving request-reply Events.
     * If this method returns 0, then the binding is realized by two one-way
     * leaf strategies, which are returned by getOneWayEventRequestReceiverStrategy()
     * and getOneWayEventReplySenderStrategy().
     *
     * @return Interface to child strategy for receiving request-reply Events
     */

    virtual ITransportStrategy* getRequestReplyEventReceiverStrategy() const
      throw(const NMM::SerializedException&);
    /** Returns child strategy for sending request-reply Events.
     * If this method returns 0, then the binding is realized by two one-way
     * leaf strategies, which are returned by getOneWayEventRequestSenderStrategy()
     * and getOneWayEventReplyReceiverStrategy().
     *
     * @return Interface to child strategy for sending request-reply Events
     */

    virtual ITransportStrategy* getRequestReplyEventSenderStrategy() const
      throw(const NMM::SerializedException&);


  };

  typedef auto_ptr<IOutOfBandCompositeStrategy> IOutOfBandCompositeStrategy_var;

  DLLEXPORT NMM::OValueStream& operator<<(NMM::OValueStream& ov, const IOutOfBandCompositeStrategy::TransportRole& t);
  DLLEXPORT NMM::IValueStream& operator>>(NMM::IValueStream& iv, IOutOfBandCompositeStrategy::TransportRole& t);



}

#include "nmm/nmm_suffix.hpp"

#endif
