/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "InstreamChannel.hpp"
#include "InstreamCompositeAddress.hpp"

#include "IInputJack.hpp"
#include "IOutputJack.hpp"
#include "IInstreamCompositeStrategy.hpp"
#include "ITransportStrategy.hpp"
#include "ISetValueStream.hpp"

#include "ProxyObject.hpp"
#include "ConnectionManager.hpp"

#include "nmm/comm/serialize/TConstructor.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  InstreamChannel::InstreamChannel()
    : CommunicationChannel(getStaticTypeInfo()),
      __input_jack(0),
      __output_jack(0),
      __input_jack_strategy(0),
      __output_jack_strategy(0),
      __in_strategy_valid(false),
      __out_strategy_valid(false),
      __strategy_list(),
      __strategy_list_valid(false),
      __strategies()
  {
    registerAttributes();
    
    /* init interface cache */
    for(unsigned int i=0; i<4; ++i) {
      __input_jack_child_strategy[i] = 0;
      __in_child_strategy_valid[i] = false;
      __output_jack_child_strategy[i] = 0;
      __out_child_strategy_valid[i] = false;
    }
  }
  
  InstreamChannel::InstreamChannel(IInputJack& input_jack, IOutputJack& output_jack)
    throw (const BindingException&)
    : CommunicationChannel(getStaticTypeInfo()),
      __input_jack(dynamic_cast<IInputJack*>(input_jack.copy())),
      __output_jack(dynamic_cast<IOutputJack*>(output_jack.copy())),
      __input_jack_strategy(0),
      __output_jack_strategy(0),
      __in_strategy_valid(false),
      __out_strategy_valid(false),
      __strategy_list(),
      __strategy_list_valid(false),
      __strategies()
  {
    registerAttributes();
    
    /* init interface cache */
    for(unsigned int i=0; i<4; ++i) {
      __input_jack_child_strategy[i] = 0;
      __in_child_strategy_valid[i] = false;
      __output_jack_child_strategy[i] = 0;
      __out_child_strategy_valid[i] = false;
    }
  }
  
  InstreamChannel::InstreamChannel(const InstreamChannel& channel)
    : CommunicationChannel(channel),
      __input_jack_strategy(0),
      __output_jack_strategy(0),
      __in_strategy_valid(false),
      __out_strategy_valid(false),
      __strategy_list(),
      __strategy_list_valid(false),
      __strategies()
  {
    registerAttributes();
    
    /* copy the IInputJack */
    if (channel.__input_jack.getValue()) {
      __input_jack.setValue(dynamic_cast<IInputJack*>(channel.__input_jack.getValue()->copy()));
    }
    
    /* copy the IOutputJack */
    if (channel.__output_jack.getValue()) {
      __output_jack.setValue(dynamic_cast<IOutputJack*>(channel.__output_jack.getValue()->copy()));
    }
    
    /* init interface cache */
    for(unsigned int i=0; i<4; ++i) {
      __input_jack_child_strategy[i] = 0;
      __in_child_strategy_valid[i] = false;
      __output_jack_child_strategy[i] = 0;
      __out_child_strategy_valid[i] = false;
    }
  }
  
  InstreamChannel* InstreamChannel::copy() const {
    return new InstreamChannel(*this);
  }
  
  InstreamChannel::~InstreamChannel() {
    for(unsigned int i=0; i<4; ++i) {
      delete __input_jack_child_strategy[i];
      delete __output_jack_child_strategy[i];
    }
    
    delete __input_jack_strategy;
    delete __output_jack_strategy;
    
    delete __input_jack.getValue();
    delete __output_jack.getValue();
  }
  
  void InstreamChannel::registerAttributes() {
    insertArgument(&__input_jack);
    insertArgument(&__output_jack);
  }
  
  void InstreamChannel::createStrategyList() const {
    /* clear strategy list */
    __strategy_list.clear();
    
    /* make sure jacks are valid */
    if(!__output_jack.getRValue()) {
      GLOBAL_ERROR_STREAM2("InstreamChannel: Invalid output jack" << std::endl);
      return;
    }
    if(!__input_jack.getRValue()) {
      GLOBAL_ERROR_STREAM2("InstreamChannel: Invalid input jack" << std::endl);
      return;
    }
    
    /* BEGIN> determine the binding type */
    TSAddress::BindingType binding_type;
    if (__output_jack.getRValue()->isLocalTo(*__input_jack.getRValue())) {
      binding_type = TSAddress::LOCAL;
    }
    
    if (__output_jack.getRValue()->isSystemTo(*__input_jack.getRValue())){
      binding_type = TSAddress::SYSTEM;
    }
    
    if (__output_jack.getRValue()->isNetworkTo(*__input_jack.getRValue())){
      binding_type = TSAddress::NETWORK;
    }
    /* END< */
    
    /* BEGIN> determine the strategies that can be used, based on binding type */
    switch (binding_type) {
    case TSAddress::LOCAL:
      /* store the LOCAL addresses that can be used. */
      if (__output_jack.getRValue()->isLocalTo(*__input_jack.getRValue())) {
        registerStrategyType(TSAddress::LOCAL);
      }
      /* Fall through */
    
    case TSAddress::SYSTEM:
      /* store the SYSTEM addresses that can be used. */
      if (__output_jack.getRValue()->isSystemTo(*__input_jack.getRValue())) {
        registerStrategyType(TSAddress::SYSTEM);
      }
      /* Fall through */
    
    case TSAddress::NETWORK:
      /* store the NETWORK addresses that can be used. */
      if (__output_jack.getRValue()->isNetworkTo(*__input_jack.getRValue())) {
        registerStrategyType(TSAddress::NETWORK);
      }
    }
    /* END< */
    
    /* make strategy list unique */
    __strategy_list.unique();
    
    /* move default strategy to front of list */
    string default_strategy = SConnectionManager::getInstance().getDefaultStrategy(getTypeInfo(), binding_type);
    if(!default_strategy.empty()) {
//       cerr << "Supported Strategies> " << endl;
      for(StrategyIterator i(__strategy_list.begin()); i != __strategy_list.end();) {
        /* BEGIN> Output strategyt config (debug) */
//         cerr << "  " << i->strategy_type << ">" << endl;
//         cerr << "    Upstream Interfaces>" << endl;
//         if(i->upstream_interfaces & TSDescription::ITS_ONE_WAY_BUFFER_RECEIVER) {
//           cerr << "      ITS_ONE_WAY_BUFFER_RECEIVER" << endl;
//         }
//         if(i->upstream_interfaces & TSDescription::ITS_ONE_WAY_BUFFER_SENDER) {
//           cerr << "      ITS_ONE_WAY_BUFFER_SENDER" << endl;
//         }
//         if(i->upstream_interfaces & TSDescription::ITS_ONE_WAY_CEVENT_RECEIVER) {
//           cerr << "      ITS_ONE_WAY_CEVENT_RECEIVER" << endl;
//         }
//         if(i->upstream_interfaces & TSDescription::ITS_ONE_WAY_CEVENT_SENDER) {
//           cerr << "      ITS_ONE_WAY_CEVENT_SENDER" << endl;
//         }
//         cerr << "    Downstream Interfaces>" << endl;
//         if(i->downstream_interfaces & TSDescription::ITS_ONE_WAY_BUFFER_RECEIVER) {
//           cerr << "      ITS_ONE_WAY_BUFFER_RECEIVER" << endl;
//         }
//         if(i->downstream_interfaces & TSDescription::ITS_ONE_WAY_BUFFER_SENDER) {
//           cerr << "      ITS_ONE_WAY_BUFFER_SENDER" << endl;
//         }
//         if(i->downstream_interfaces & TSDescription::ITS_ONE_WAY_CEVENT_RECEIVER) {
//           cerr << "      ITS_ONE_WAY_CEVENT_RECEIVER" << endl;
//         }
//         if(i->downstream_interfaces & TSDescription::ITS_ONE_WAY_CEVENT_SENDER) {
//           cerr << "      ITS_ONE_WAY_CEVENT_SENDER" << endl;
//         }
//         cerr << "    Capabilities>" << endl;
//         if(i->capabilities & TSDescription::TS_RELIABLE) {
//           cerr << "      TS_RELIABLE" << endl;
//         }
//         if(i->capabilities & TSDescription::TS_RELIABLE_CONNECT) {
//           cerr << "      TS_RELIABLE_CONNECT" << endl;
//         }
//         if(i->capabilities & TSDescription::TS_PRESERVE_MESSAGE_HEADERS) {
//           cerr << "      TS_PRESERVE_MESSAGE_HEADERS" << endl;
//         }
        /* END< */
        
        if(i->strategy_type == default_strategy) {
          __strategy_list.push_front(*i);
          __strategy_list.erase(i++);
        } else {
          ++i;
        }
      }
    }
    
    __strategy_list_valid = true;
  }
  
  void InstreamChannel::registerStrategyType(TSAddress::BindingType binding_type) const {
    /* get address lists */
    const SequenceValue<TSAddress>& input_jack_address_list  = __input_jack.getRValue()->getAddressList();
    const SequenceValue<TSAddress>& output_jack_address_list = __output_jack.getRValue()->getAddressList();

    /* get supported strategy lists */
    const list<TSDescription>& input_jack_list  = __input_jack.getRValue()->getSupportedStrategies();
    const list<TSDescription>& output_jack_list = __output_jack.getRValue()->getSupportedStrategies();
    
    /* register strategies */
    registerStrategyType(binding_type, output_jack_address_list, input_jack_list, Message::DOWNSTREAM);
    registerStrategyType(binding_type, input_jack_address_list, output_jack_list, Message::UPSTREAM);
  }
  
  void InstreamChannel::registerStrategy(const TSDescription& upstream_description, const TSDescription& downstream_description) const {
    StrategyListEntry entry;
    entry.strategy_type = upstream_description.getStrategyType();
    entry.capabilities = upstream_description.getCapabilities();
    entry.upstream_interfaces = 0;
    entry.downstream_interfaces = 0;
    entry.used_upstream_interfaces = 0;
    entry.used_downstream_interfaces = 0;
    
    unsigned int upstream_interfaces = upstream_description.getInterfaces();
    unsigned int downstream_interfaces = downstream_description.getInterfaces();
    
    if((upstream_interfaces   & TSDescription::ITS_ONE_WAY_BUFFER_SENDER) &&
       (downstream_interfaces & TSDescription::ITS_ONE_WAY_BUFFER_RECEIVER)) {
      entry.upstream_interfaces   |= TSDescription::ITS_ONE_WAY_BUFFER_SENDER;
      entry.downstream_interfaces |= TSDescription::ITS_ONE_WAY_BUFFER_RECEIVER;
    }
    if((upstream_interfaces   & TSDescription::ITS_ONE_WAY_CEVENT_SENDER) &&
       (downstream_interfaces & TSDescription::ITS_ONE_WAY_CEVENT_RECEIVER)) {
      entry.upstream_interfaces   |= TSDescription::ITS_ONE_WAY_CEVENT_SENDER;
      entry.downstream_interfaces |= TSDescription::ITS_ONE_WAY_CEVENT_RECEIVER;
    }
    if((upstream_interfaces   & TSDescription::ITS_ONE_WAY_BUFFER_RECEIVER) &&
       (downstream_interfaces & TSDescription::ITS_ONE_WAY_BUFFER_SENDER)) {
      entry.upstream_interfaces   |= TSDescription::ITS_ONE_WAY_BUFFER_RECEIVER;
      entry.downstream_interfaces |= TSDescription::ITS_ONE_WAY_BUFFER_SENDER;
    }
    if((upstream_interfaces   & TSDescription::ITS_ONE_WAY_CEVENT_RECEIVER) &&
       (downstream_interfaces & TSDescription::ITS_ONE_WAY_CEVENT_SENDER)) {
      entry.upstream_interfaces   |= TSDescription::ITS_ONE_WAY_CEVENT_RECEIVER;
      entry.downstream_interfaces |= TSDescription::ITS_ONE_WAY_CEVENT_SENDER;
    }
    
    /* check if any matching pairs have been found and, if so, register the strategy */
    if(entry.downstream_interfaces && entry.upstream_interfaces) {
      GLOBAL_DEBUG_STREAM2("InstreamChannel: Added available strategy " << entry.strategy_type << std::endl);
      __strategy_list.push_back(entry);
    }
  }
  
  void InstreamChannel::registerStrategyType(TSAddress::BindingType binding_type,
                                             const SequenceValue<TSAddress>& address_list,
                                             const list<TSDescription>& strategy_list,
                                             Message::Direction direction) const
  {
    for(unsigned int i=0; i<address_list.size(); ++i) {
      if(address_list[i] && address_list[i]->isBindingType(binding_type)) {
        string strategy_type = address_list[i]->getStrategyType();
        for(list<TSDescription>::const_iterator j=strategy_list.begin();
            j != strategy_list.end();
            ++j) {
          if(j->getStrategyType() == strategy_type) {
            if(direction == Message::DOWNSTREAM) {
              registerStrategy(address_list[i]->getTSDescription(), *j);
            } else {
              registerStrategy(*j, address_list[i]->getTSDescription());
            }
          }
        }
      }
    }
  }
  
  list<string> InstreamChannel::getStrategyList() const {
    if(!__strategy_list_valid) {
      createStrategyList();
    }
    
    list<string> temp;
    for(StrategyConstIterator i=__strategy_list.begin();
        i != __strategy_list.end();
        ++i) {
      temp.push_back(i->strategy_type);
    }
    return temp;
  }
  
  list<string> InstreamChannel::getStrategyList(unsigned int upstream_interfaces, unsigned int downstream_interfaces) const {
    if(!__strategy_list_valid) {
      createStrategyList();
    }
    
    list<string> temp;
    for(StrategyConstIterator i=__strategy_list.begin();
        i != __strategy_list.end();
        ++i) {
      if((upstream_interfaces & i->upstream_interfaces) == upstream_interfaces &&
         (downstream_interfaces & i->downstream_interfaces) == downstream_interfaces) {
        temp.push_back(i->strategy_type);
      }
    }
    return temp;
  }
  
  list<string> InstreamChannel::getStrategyList(unsigned int roles) const {
    unsigned int upstream_interfaces = 0;
    unsigned int downstream_interfaces = 0;
    getRequiredInterfaces(roles, upstream_interfaces, downstream_interfaces);
    
    return getStrategyList(upstream_interfaces, downstream_interfaces);
  }

  void InstreamChannel::addTransportStrategyInternal(const string& strategy_type, unsigned int upstream_interfaces,
                                                     unsigned int downstream_interfaces, bool is_explicit)
    throw(const BindingException&)
  {
    /* check if strategy list is valid */
    if(!__strategy_list_valid) {
      createStrategyList();
    }
    
    /* check if strategy is supported */
    bool strategy_found = false;
    StrategyConstIterator found;
    for(found=__strategy_list.begin(); found != __strategy_list.end(); ++found) {
      /* check if strategy has the right type */
      if(found->strategy_type == strategy_type) {
        strategy_found = true;
        
        /* check if strategy supports all required interfaces */
        if(upstream_interfaces == (upstream_interfaces & found->upstream_interfaces) &&
           downstream_interfaces == (downstream_interfaces & found->downstream_interfaces)) {
          break;
        }
      }
    }
    if(found == __strategy_list.end()) {
      if(!strategy_found) {
        throw BindingException(strategy_type + " can not be used to connect these jacks");
      } else {
        throw BindingException(strategy_type + " doesn't have all of the requested interfaces");
      }
    }
    
    /* check for collision */
    for(unsigned int i=0; i<__strategies.size(); ++i) {
      if(__strategies[i].used_upstream_interfaces & upstream_interfaces) {
        throw BindingException(__strategies[i].strategy_type + " has already been added with some of the requested upstream interfaces");
      }
      if(__strategies[i].used_downstream_interfaces & downstream_interfaces) {
        throw BindingException(__strategies[i].strategy_type + " has already been added with some of the requested downstream interfaces");
      }
    }
    
    /* add strategy to config */
    StrategyListEntry entry = *found;
    entry.used_upstream_interfaces = upstream_interfaces;
    entry.used_downstream_interfaces = downstream_interfaces;
    entry.is_explicit = is_explicit;
    __strategies.push_back(entry);
    unsigned int index = __strategies.size() - 1;
    if(is_explicit) {
      GLOBAL_DEBUG_STREAM2("InstreamChannel: Added explicit strategy to config " << entry.strategy_type << ":" << index << endl);
    } else {
      GLOBAL_DEBUG_STREAM2("InstreamChannel: Added strategy to config " << entry.strategy_type << ":" << index << endl);
    }
    GLOBAL_DEBUG_STREAM2("InstreamChannel: Upstream   interfaces: " << getInterfaceString(upstream_interfaces) << endl);
    GLOBAL_DEBUG_STREAM2("InstreamChannel: Downstream interfaces: " << getInterfaceString(downstream_interfaces) << endl);
  }
  
  void InstreamChannel::addTransportStrategy(const string& strategy_type, unsigned int upstream_interfaces, unsigned int downstream_interfaces)
    throw(const BindingException&)
  {
    addTransportStrategyInternal(strategy_type, upstream_interfaces, downstream_interfaces, false);
  }
  
  void InstreamChannel::addExplicitTransportStrategy(const string& strategy_type, unsigned int upstream_interfaces, unsigned int downstream_interfaces)
    throw(const BindingException&)
  {
    addTransportStrategyInternal(strategy_type, upstream_interfaces, downstream_interfaces, true);
  }
  
  void InstreamChannel::addTransportStrategy(const string& strategy_type, unsigned int roles)
    throw(const BindingException&)
  {
    /* first determine which interfaces we use */
    unsigned int upstream_interfaces = 0;
    unsigned int downstream_interfaces = 0;
    getRequiredInterfaces(roles, upstream_interfaces, downstream_interfaces);
    
    /* then add the strategy */
    return addTransportStrategy(strategy_type, upstream_interfaces, downstream_interfaces);
  }
  
  void InstreamChannel::addExplicitTransportStrategy(const string& strategy_type, unsigned int roles)
    throw(const BindingException&)
  {
    /* first determine which interfaces we use */
    unsigned int upstream_interfaces = 0;
    unsigned int downstream_interfaces = 0;
    getRequiredInterfaces(roles, upstream_interfaces, downstream_interfaces);
    
    /* then add the strategy */
    return addExplicitTransportStrategy(strategy_type, upstream_interfaces, downstream_interfaces);
  }
  
  void InstreamChannel::establishBinding() throw(const BindingException&) {
    /* create strategy list */
    if(!__strategy_list_valid) {
      createStrategyList();
    }
    
    /* BEGIN> check if for all roles a transport strategy has been set */
    unsigned int set_roles = 0;
    for(unsigned int i=0; i<__strategies.size(); ++i) {
      if((__strategies[i].used_upstream_interfaces & TSDescription::ITS_ONE_WAY_BUFFER_RECEIVER) &&
         (__strategies[i].used_downstream_interfaces & TSDescription::ITS_ONE_WAY_BUFFER_SENDER)) {
        set_roles |= DOWNSTREAM_BUFFER;
      }
      if((__strategies[i].used_upstream_interfaces & TSDescription::ITS_ONE_WAY_CEVENT_RECEIVER) &&
         (__strategies[i].used_downstream_interfaces & TSDescription::ITS_ONE_WAY_CEVENT_SENDER)) {
        set_roles |= DOWNSTREAM_CEVENT;
      }
      if((__strategies[i].used_upstream_interfaces & TSDescription::ITS_ONE_WAY_BUFFER_SENDER) &&
         (__strategies[i].used_downstream_interfaces & TSDescription::ITS_ONE_WAY_BUFFER_RECEIVER)) {
        set_roles |= UPSTREAM_BUFFER;
      }
      if((__strategies[i].used_upstream_interfaces & TSDescription::ITS_ONE_WAY_CEVENT_SENDER) &&
         (__strategies[i].used_downstream_interfaces & TSDescription::ITS_ONE_WAY_CEVENT_RECEIVER)) {
        set_roles |= UPSTREAM_CEVENT;
      }
    }
    /* END< */
    
    /* BEGIN> Automatically fill unset roles */
    unsigned int upstream_interfaces = 0;
    unsigned int downstream_interfaces = 0;
    
    /* BEGIN> Automatic binding configuration.
       Try to reuse already added strategies and add new strategies if needed.
       This is done in a loop, so the new added strategies can also be reused. */
    for(;;) {
      /* BEGIN> Try to reuse already added strategies */
      if(!(set_roles & DOWNSTREAM_BUFFER)) {
        getRequiredInterfaces(DOWNSTREAM_BUFFER, upstream_interfaces, downstream_interfaces);
        if(tryReuseStrategy(upstream_interfaces, downstream_interfaces)) {
          set_roles |= DOWNSTREAM_BUFFER;
        }
      }
      if(!(set_roles & DOWNSTREAM_CEVENT)) {
        getRequiredInterfaces(DOWNSTREAM_CEVENT, upstream_interfaces, downstream_interfaces);
        if(tryReuseStrategy(upstream_interfaces, downstream_interfaces)) {
          set_roles |= DOWNSTREAM_CEVENT;
        }
      }
      if(!(set_roles & UPSTREAM_BUFFER)) {
        getRequiredInterfaces(UPSTREAM_BUFFER, upstream_interfaces, downstream_interfaces);
        if(tryReuseStrategy(upstream_interfaces, downstream_interfaces)) {
          set_roles |= UPSTREAM_BUFFER;
        }
      }
      if(!(set_roles & UPSTREAM_CEVENT)) {
        getRequiredInterfaces(UPSTREAM_CEVENT, upstream_interfaces, downstream_interfaces);
        if(tryReuseStrategy(upstream_interfaces, downstream_interfaces)) {
          set_roles |= UPSTREAM_CEVENT;
        }
      }
      /* END< */
      
      /* BEGIN> Try to find suitable default strategies */
      if(!(set_roles & DOWNSTREAM_BUFFER)) {
        getRequiredInterfaces(DOWNSTREAM_BUFFER, upstream_interfaces, downstream_interfaces);
        if(findSuitableStrategy(upstream_interfaces, downstream_interfaces)) {
          set_roles |= DOWNSTREAM_BUFFER;
          continue;
        } else {
          throw BindingException("No suitable strategy for DOWNSTREAM_BUFFER available");
        }
      }
      if(!(set_roles & DOWNSTREAM_CEVENT)) {
        getRequiredInterfaces(DOWNSTREAM_CEVENT, upstream_interfaces, downstream_interfaces);
        if(findSuitableStrategy(upstream_interfaces, downstream_interfaces)) {
          set_roles |= DOWNSTREAM_CEVENT;
          continue;
        } else {
          throw BindingException("No suitable strategy for DOWNSTREAM_CEVENT available");
        }
      }
      if(!(set_roles & UPSTREAM_BUFFER)) {
        getRequiredInterfaces(UPSTREAM_BUFFER, upstream_interfaces, downstream_interfaces);
        if(findSuitableStrategy(upstream_interfaces, downstream_interfaces)) {
          set_roles |= UPSTREAM_BUFFER;
          continue;
        } else {
          throw BindingException("No suitable strategy for UPSTREAM_BUFFER available");
        }
      }
      if(!(set_roles & UPSTREAM_CEVENT)) {
        getRequiredInterfaces(UPSTREAM_CEVENT, upstream_interfaces, downstream_interfaces);
        if(findSuitableStrategy(upstream_interfaces, downstream_interfaces)) {
          set_roles |= UPSTREAM_CEVENT;
          continue;
        } else {
          throw BindingException("No suitable strategy for UPSTREAM_CEVENT available");
        }
      }
      /* END< */
      
      if(set_roles != INSTREAM) {
        /* give up */
        throw BindingException("Could not find suitable transport strategies to realize instream binding");
      } else {
        /* done */
        break;
      }
    }
    /* END< */
    
    /* BEGIN> Try to connect jacks in both directions */
    try {
      GLOBAL_DEBUG_STREAM2("InstreamChannel: Trying to connect input jack to output jack" << std::endl);
      tryConnectJacks(__input_jack.getValue(), __output_jack.getValue(), Message::UPSTREAM);
    } catch(const BindingException& e1) {
      GLOBAL_DEBUG_STREAM2("InstreamChannel: Connection from input jack to output jack failed:" << std::endl);
      GLOBAL_DEBUG_STREAM2(e1 << std::endl);
      try {
        GLOBAL_DEBUG_STREAM2("InstreamChannel: Trying to connect output jack to input jack" << std::endl);
        tryConnectJacks(__output_jack.getValue(), __input_jack.getValue(), Message::DOWNSTREAM);
      } catch(const BindingException& e2) {
        GLOBAL_ERROR_STREAM("InstreamChannel: Failed to connect jacks in both directions:" << std::endl);
        GLOBAL_ERROR_STREAM("InstreamChannel: Connection from input jack to output jack failed:" << std::endl);
        GLOBAL_ERROR_STREAM(e1 << std::endl);
        GLOBAL_ERROR_STREAM("InstreamChannel: Connection from output jack to input jack failed:" << std::endl);
        GLOBAL_ERROR_STREAM(e2 << std::endl);
        throw BindingException("Failed to connect jacks in both directions");
      }
    }
    /* END< */
  }
  
  list<Interface*>& InstreamChannel::getDownstreamInterfaceList() {
    IInstreamCompositeStrategy* ts = getInputJackStrategy();
    if(!ts) {
      /* return empty list */
      static list<Interface*> empty;
      return empty;
    }
    
    return ts->getParentObject()->getInterfaceList();
  }
  
  list<Interface*>& InstreamChannel::getUpstreamInterfaceList() {
    IInstreamCompositeStrategy* ts = getOutputJackStrategy();
    if(!ts) {
      /* return empty list */
      static list<Interface*> empty;
      return empty;
    }
    
    return ts->getParentObject()->getInterfaceList();
  }
  
  list<Interface*>& InstreamChannel::getDownstreamInterfaceList(CommunicationRole role) {
    ITransportStrategy* ts = getInputJackStrategy(role);
    if(!ts) {
      /* return empty list */
      static list<Interface*> empty;
      return empty;
    }
    
    return ts->getParentObject()->getInterfaceList();
  }
  
  list<Interface*>& InstreamChannel::getUpstreamInterfaceList(CommunicationRole role) {
    ITransportStrategy* ts = getOutputJackStrategy(role);
    if(!ts) {
      /* return empty list */
      static list<Interface*> empty;
      return empty;
    }
    
    return ts->getParentObject()->getInterfaceList();
  }
  
  void InstreamChannel::tryConnectJacks(IJack* jack1, IJack* jack2, Message::Direction direction)
    throw(const BindingException&)
  {
    /* BEGIN> Try to create a composite address */
    /* if one or more default addresses are not available, then this
       will fail.  however we don't need to clean up because we have not
       done anything yet, besides creating a TSAddress. */
    InstreamCompositeAddress composite_address;
    composite_address.setNMMReference(jack2->getParentReference());
    for(unsigned int i=0; i<__strategies.size(); ++i) {
      /* get default address */
      TSAddress* default_address = jack2->getDefaultAddress(__strategies[i].strategy_type);
      if(!default_address) {
        GLOBAL_ERROR_STREAM2("InstreamChannel: no default address for strategy " << __strategies[i].strategy_type << endl);
        throw BindingException("No default address for sub-binding");
      }
      
      /* determine used interfaces */
      unsigned int used_interfaces;
      if(direction == Message::DOWNSTREAM) {
        used_interfaces = __strategies[i].used_downstream_interfaces;
      } else {
        used_interfaces = __strategies[i].used_upstream_interfaces;
      }
      
      /* add address to composite address */
      unsigned int index = composite_address.addAddress(*default_address);
      
      /* set roles according to used interfaces */
      if(used_interfaces & TSDescription::ITS_ONE_WAY_BUFFER_SENDER) {
        composite_address.setOneWayBufferSenderAddress(index);
      }
      if(used_interfaces & TSDescription::ITS_ONE_WAY_CEVENT_SENDER) {
        composite_address.setOneWayCEventSenderAddress(index);
      }
      if(used_interfaces & TSDescription::ITS_ONE_WAY_BUFFER_RECEIVER) {
        composite_address.setOneWayBufferReceiverAddress(index);
      }
      if(used_interfaces & TSDescription::ITS_ONE_WAY_CEVENT_RECEIVER) {
        composite_address.setOneWayCEventReceiverAddress(index);
      }
    }
    /* END< */
    
    /* connect composite strategy */
    jack1->setTransportStrategy(&composite_address);
  }
  
  void InstreamChannel::unsetTransportStrategy() {        
    /* unset all roles */
    __strategies.clear();
  }
  
  IInputJack* InstreamChannel::getInputJack() const {
    return __input_jack.getRValue();
  }
  
  IOutputJack* InstreamChannel::getOutputJack() const {
    return __output_jack.getRValue();
  }
  
  IInstreamCompositeStrategy* InstreamChannel::getInputJackStrategy() const {
    if(!__in_strategy_valid) {
      if(__input_jack_strategy) {
        delete __input_jack_strategy;
        __input_jack_strategy = 0;
      }
      
      __input_jack_strategy = __input_jack.getRValue()->getTransportStrategy();
      __in_strategy_valid = true;
    }
    return __input_jack_strategy;
  }
  
  IInstreamCompositeStrategy* InstreamChannel::getOutputJackStrategy() const {
    if(!__out_strategy_valid) {
      if(__output_jack_strategy) {
        delete __output_jack_strategy;
        __output_jack_strategy = 0;
      }
      
      __output_jack_strategy = __output_jack.getRValue()->getTransportStrategy();
      __out_strategy_valid = true;
    }
    return __output_jack_strategy;
  }
  
  ITransportStrategy* InstreamChannel::getInputJackStrategy(CommunicationRole role) const {
    /* get index in interface cache */
    unsigned int index;
    for(index=0; index<4; ++index) {
      if((1 << index) == role) {
        break;
      }
    }
    
    /* check if we need to request this interface */
    if(!__in_child_strategy_valid[index]) {
      IInstreamCompositeStrategy* top_level_ts = getInputJackStrategy();
      if(!top_level_ts) {
        /* no top-level strategy exists */
        return 0;
      }
      
      /* get child strategy */
      ITransportStrategy* child_strategy = 0;
      switch(role) {
      case DOWNSTREAM_BUFFER:
        child_strategy = top_level_ts->getOneWayBufferSenderStrategy();
        break;
      case DOWNSTREAM_CEVENT:
        child_strategy = top_level_ts->getOneWayCEventSenderStrategy();
        break;
      case UPSTREAM_BUFFER:
        child_strategy = top_level_ts->getOneWayBufferReceiverStrategy();
        break;
      case UPSTREAM_CEVENT:
        child_strategy = top_level_ts->getOneWayCEventReceiverStrategy();
        break;
      default:
        GLOBAL_ERROR_STREAM("Invalid communication role specified for child strategy" << endl);
        return 0;
      }
      
      /* cache interface */
      __input_jack_child_strategy[index] = child_strategy;
      __in_child_strategy_valid[index] = true;
    }
    
    /* return cached interface */
    return __input_jack_child_strategy[index];
  }
  
  ITransportStrategy* InstreamChannel::getOutputJackStrategy(CommunicationRole role) const {
    /* get index in interface cache */
    unsigned int index;
    for(index=0; index<4; ++index) {
      if((1 << index) == role) {
        break;
      }
    }
    
    /* check if we need to request this interface */
    if(!__out_child_strategy_valid[index]) {
      IInstreamCompositeStrategy* top_level_ts = getOutputJackStrategy();
      if(!top_level_ts) {
        /* no top-level strategy exists */
        return 0;
      }
      
      /* get child strategy */
      ITransportStrategy* child_strategy = 0;
      switch(role) {
      case DOWNSTREAM_BUFFER:
        child_strategy = top_level_ts->getOneWayBufferReceiverStrategy();
        break;
      case DOWNSTREAM_CEVENT:
        child_strategy = top_level_ts->getOneWayCEventReceiverStrategy();
        break;
      case UPSTREAM_BUFFER:
        child_strategy = top_level_ts->getOneWayBufferSenderStrategy();
        break;
      case UPSTREAM_CEVENT:
        child_strategy = top_level_ts->getOneWayCEventSenderStrategy();
        break;
      default:
        GLOBAL_ERROR_STREAM("Invalid communication role specified for child strategy" << endl);
        return 0;
      }
      
      /* cache interface */
      __output_jack_child_strategy[index] = child_strategy;
      __out_child_strategy_valid[index] = true;
    }
    
    /* return cached interface */
    return __output_jack_child_strategy[index];
  }
  
  void InstreamChannel::updateAttributes() {
    createStrategyList();
  }
  
  void InstreamChannel::getRequiredInterfaces(unsigned int roles, unsigned int& upstream_interfaces, unsigned int& downstream_interfaces) {
    upstream_interfaces = 0;
    downstream_interfaces = 0;
    
    if(roles & DOWNSTREAM_BUFFER) {
      upstream_interfaces |= TSDescription::ITS_ONE_WAY_BUFFER_RECEIVER;
      downstream_interfaces |= TSDescription::ITS_ONE_WAY_BUFFER_SENDER;
    }
    if(roles & DOWNSTREAM_CEVENT) {
      upstream_interfaces |= TSDescription::ITS_ONE_WAY_CEVENT_RECEIVER;
      downstream_interfaces |= TSDescription::ITS_ONE_WAY_CEVENT_SENDER;
    }
    if(roles & UPSTREAM_BUFFER) {
      upstream_interfaces |= TSDescription::ITS_ONE_WAY_BUFFER_SENDER;
      downstream_interfaces |= TSDescription::ITS_ONE_WAY_BUFFER_RECEIVER;
    }
    if(roles & UPSTREAM_CEVENT) {
      upstream_interfaces |= TSDescription::ITS_ONE_WAY_CEVENT_SENDER;
      downstream_interfaces |= TSDescription::ITS_ONE_WAY_CEVENT_RECEIVER;
    }
  }

  bool InstreamChannel::tryReuseStrategy(unsigned int upstream_interfaces, unsigned int downstream_interfaces) {
    for(unsigned int i=0; i<__strategies.size(); ++i) {
      if(!__strategies[i].is_explicit &&
         (__strategies[i].upstream_interfaces & upstream_interfaces) == upstream_interfaces &&
         (__strategies[i].downstream_interfaces & downstream_interfaces) == downstream_interfaces) {
        __strategies[i].used_upstream_interfaces |= upstream_interfaces;
        __strategies[i].used_downstream_interfaces |= downstream_interfaces;
        GLOBAL_DEBUG_STREAM2("InstreamChannel: Automatically reused strategy " << __strategies[i].strategy_type << ":" << i << endl);
        GLOBAL_DEBUG_STREAM2("InstreamChannel: Upstream   interfaces: " << getInterfaceString(upstream_interfaces) << endl);
        GLOBAL_DEBUG_STREAM2("InstreamChannel: Downstream interfaces: " << getInterfaceString(downstream_interfaces) << endl);
        return true;
      }
    }
    
    return false;
  }
  
  bool InstreamChannel::findSuitableStrategy(unsigned int upstream_interfaces, unsigned int downstream_interfaces) {
    for(StrategyConstIterator i=__strategy_list.begin(); i != __strategy_list.end(); ++i) {
      if((i->upstream_interfaces & upstream_interfaces) == upstream_interfaces &&
         (i->downstream_interfaces & downstream_interfaces) == downstream_interfaces) {
        StrategyListEntry entry(*i);
        entry.used_upstream_interfaces = upstream_interfaces;
        entry.used_downstream_interfaces = downstream_interfaces;
        __strategies.push_back(entry);
        GLOBAL_DEBUG_STREAM2("InstreamChannel: Automatically added suitable strategy " << i->strategy_type << ":" << __strategies.size()-1 << endl);
        GLOBAL_DEBUG_STREAM2("InstreamChannel: Upstream   interfaces: " << getInterfaceString(upstream_interfaces) << endl);
        GLOBAL_DEBUG_STREAM2("InstreamChannel: Downstream interfaces: " << getInterfaceString(downstream_interfaces) << endl);
        return true;
      }
    }
    
    return false;
  }
  
  string InstreamChannel::getRoleName(CommunicationRole role) {
    switch(role) {
    case DOWNSTREAM_BUFFER:
      return "DOWNSTREAM_BUFFER";
    case DOWNSTREAM_CEVENT:
      return "DOWNSTREAM_CEVENT";
    case UPSTREAM_BUFFER:
      return "UPSTREAM_BUFFER";
    case UPSTREAM_CEVENT:
      return "UPSTREAM_CEVENT";
    default:
      return string();
    };
  }
  
  string InstreamChannel::getRoleString(int roles) {
    string str("(");
    bool first = true;
    if(roles & DOWNSTREAM_BUFFER) {
      if(!first) {
        str += ", ";
      }
      first = false;
      str += "DOWNSTREAM_BUFFER";
    }
    if(roles & DOWNSTREAM_CEVENT) {
      if(!first) {
        str += ", ";
      }
      first = false;
      str += "DOWNSTREAM_CEVENT";
    }
    if(roles & UPSTREAM_BUFFER) {
      if(!first) {
        str += ", ";
      }
      first = false;
      str += "UPSTREAM_BUFFER";
    }
    if(roles & UPSTREAM_CEVENT) {
      if(!first) {
        str += ", ";
      }
      first = false;
      str += "UPSTREAM_CEVENT";
    }
    return str + ")";
  };
  
  string InstreamChannel::getInterfaceName(TSDescription::Interfaces iface) {
    switch(iface) {
    case TSDescription::ITS_ONE_WAY_BUFFER_SENDER:
      return "ITS_ONE_WAY_BUFFER_SENDER";
    case TSDescription::ITS_ONE_WAY_CEVENT_SENDER:
      return "ITS_ONE_WAY_CEVENT_SENDER";
    case TSDescription::ITS_ONE_WAY_BUFFER_RECEIVER:
      return "ITS_ONE_WAY_BUFFER_RECEIVER";
    case TSDescription::ITS_ONE_WAY_CEVENT_RECEIVER:
      return "ITS_ONE_WAY_CEVENT_RECEIVER";
    default:
      return string();
    };
  }
  
  string InstreamChannel::getInterfaceString(int interfaces) {
    string str("(");
    bool first = true;
    if(interfaces & TSDescription::ITS_ONE_WAY_BUFFER_SENDER) {
      if(!first) {
        str += ", ";
      }
      first = false;
      str += "ITS_ONE_WAY_BUFFER_SENDER";
    }
    if(interfaces & TSDescription::ITS_ONE_WAY_CEVENT_SENDER) {
      if(!first) {
        str += ", ";
      }
      first = false;
      str += "ITS_ONE_WAY_CEVENT_SENDER";
    }
    if(interfaces & TSDescription::ITS_ONE_WAY_BUFFER_RECEIVER) {
      if(!first) {
        str += ", ";
      }
      first = false;
      str += "ITS_ONE_WAY_BUFFER_RECEIVER";
    }
    if(interfaces & TSDescription::ITS_ONE_WAY_CEVENT_RECEIVER) {
      if(!first) {
        str += ", ";
      }
      first = false;
      str += "ITS_ONE_WAY_CEVENT_RECEIVER";
    }
    return str + ")";
  };
  
  void InstreamChannel::setValueStream(unsigned int roles,
                                       const string& i_value_stream_type,
                                       const string& o_value_stream_type)
    throw(const BindingException&)
  {
    // XXX: fw: if one or more strategies are the same,
    // XXX: fw: don't set the value stream multiple times
    
    if(roles & DOWNSTREAM_BUFFER) {
      setValueStreamInternal(DOWNSTREAM_BUFFER, i_value_stream_type, o_value_stream_type);
    }
    if(roles & DOWNSTREAM_CEVENT) {
      setValueStreamInternal(DOWNSTREAM_CEVENT, i_value_stream_type, o_value_stream_type);
    }
    if(roles & UPSTREAM_BUFFER) {
      setValueStreamInternal(UPSTREAM_BUFFER, i_value_stream_type, o_value_stream_type);
    }
    if(roles & UPSTREAM_CEVENT) {
      setValueStreamInternal(UPSTREAM_CEVENT, i_value_stream_type, o_value_stream_type);
    }
  }
  
  void InstreamChannel::setValueStreamStack(unsigned int roles,
                                            const IValueStream* i_value_stream,
                                            const OValueStream* o_value_stream)
  {
    // XXX: fw: if one or more strategies are the same
    // XXX: fw: don't set the value stream multiple times
    
    if(roles & DOWNSTREAM_BUFFER) {
      setValueStreamStackInternal(DOWNSTREAM_BUFFER, i_value_stream, o_value_stream);
    }
    if(roles & DOWNSTREAM_CEVENT) {
      setValueStreamStackInternal(DOWNSTREAM_CEVENT, i_value_stream, o_value_stream);
    }
    if(roles & UPSTREAM_BUFFER) {
      setValueStreamStackInternal(UPSTREAM_BUFFER, i_value_stream, o_value_stream);
    }
    if(roles & UPSTREAM_CEVENT) {
      setValueStreamStackInternal(UPSTREAM_CEVENT, i_value_stream, o_value_stream);
    }
  }
  
  void InstreamChannel::setValueStreamInternal(CommunicationRole role,
                                               const string& i_value_stream_type,
                                               const string& o_value_stream_type)
    throw(const BindingException&)
  {
    /* get interfaces for setting the value stream */
    ISetValueStream_var upstream_value_stream(getUpstreamInterface<ISetValueStream>(role));
    ISetValueStream_var downstream_value_stream(getDownstreamInterface<ISetValueStream>(role));
    if(!upstream_value_stream.get() || !downstream_value_stream.get()) {
      GLOBAL_ERROR_STREAM("Can't set value stream for role " << getRoleName(role) << endl);
      return;
    }
    
    /* set the value stream */
    if(role & DOWNSTREAM) {
      upstream_value_stream->setOValueStream(o_value_stream_type);
      downstream_value_stream->setIValueStream(i_value_stream_type);
    } else {
      upstream_value_stream->setIValueStream(i_value_stream_type);
      downstream_value_stream->setOValueStream(o_value_stream_type);
    }
  }

  void InstreamChannel::setValueStreamStackInternal(CommunicationRole role,
                                                    const IValueStream* i_value_stream,
                                                    const OValueStream* o_value_stream)
  {
    /* get interfaces for setting the value stream */
    ISetValueStream_var upstream_value_stream(getUpstreamInterface<ISetValueStream>(role));
    ISetValueStream_var downstream_value_stream(getDownstreamInterface<ISetValueStream>(role));
    if(!upstream_value_stream.get() || !downstream_value_stream.get()) {
      GLOBAL_ERROR_STREAM("Can't set value stream for role " << getRoleName(role) << endl);
      return;
    }
    
    /* set the value stream */
    if(role & DOWNSTREAM) {
      upstream_value_stream->setOValueStreamStack(o_value_stream);
      downstream_value_stream->setIValueStreamStack(i_value_stream);
    } else {
      upstream_value_stream->setIValueStreamStack(i_value_stream);
      downstream_value_stream->setOValueStreamStack(o_value_stream);
    }
  }
  
  
  TConstructor<InstreamChannel> __nmm_instream_channel_prototype("InstreamChannel", "NMM");
  
  TypeInfo InstreamChannel::getStaticTypeInfo() {
    return TypeInfo("InstreamChannel", "NMM");
  }
  
}

#include "nmm/nmm_suffix.hpp"
