/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "InstreamCompositeStrategy.hpp"
#include "InstreamCompositeAddress.hpp"

#include "TSFactory.hpp"
#include "IObjectAcceptor.hpp"
#include "ConnectionManager.hpp"
#include "ConnectorAddress.hpp"

#include "ProxyObject.hpp"

#include "SimpleInstreamMultiplexer.hpp"
#include "ReliableInstreamMultiplexer.hpp"
#include "UnreliableInstreamMultiplexer.hpp"
#include "SimpleInstreamDemultiplexer.hpp"
#include "ReliableInstreamDemultiplexer.hpp"
#include "UnreliableInstreamDemultiplexer.hpp"

#include "nmm/comm/messaging/Message.hpp"
#include "nmm/comm/messaging/ICSSignalReceiver.hpp"
#include "nmm/comm/messaging/ITSOneWayBufferReceiver.hpp"
#include "nmm/comm/messaging/ITSOneWayBufferSender.hpp"
#include "nmm/comm/messaging/ITSOneWayCEventReceiver.hpp"
#include "nmm/comm/messaging/ITSOneWayCEventSender.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  void TSSignalAdapter::receiveSignal(Signal signal) {
    if(__receiver) {
      __receiver->receiveSignal(signal, __sender);
    } else {
      GLOBAL_WARNING_STREAM1("InstreamCompositeStrategy> TSSignalAdapter: No signal receiver");
    }
  }
  
  
  InstreamCompositeStrategy::InstreamCompositeStrategy(const string& name)
    : CompositeStrategy(name),
      __one_way_buffer_sender(0),
      __one_way_buffer_receiver(0),
      __one_way_cevent_sender(0),
      __one_way_cevent_receiver(0),
      __message_mux(0),
      __message_demux(0)
  {}
  
  InstreamCompositeStrategy::~InstreamCompositeStrategy() {

    // Cleanup: Flush the message_demux which also flushes the used sending TransportStrategies. 
    if ( __message_demux ) {
      __message_demux->flushITSReceiver();
    }
  
    // Disconnect section 
    /* BEGIN> Unset all roles */
    setOneWayBufferSenderStrategy(0);
    setOneWayCEventSenderStrategy(0);
    setOneWayBufferReceiverStrategy(0);
    setOneWayCEventReceiverStrategy(0);
    /* END< */
    
    
    //Delete section
    /* remove all child strategies (dangeling pointer...) */
    removeAllChildStrategies(); 
            
    /* delete message demultiplexer */
    if(__message_demux) {
      delete __message_demux;
    }
    
    /* delete message multiplexer */
    if(__message_mux) {
      delete __message_mux;
    }    
  }
  
  string InstreamCompositeStrategy::getStrategyType() const {
    return "InstreamCompositeStrategy";
  }
  
  bool InstreamCompositeStrategy::isBindingComplete() const throw() {
    if(!__one_way_buffer_sender) {
      return false;
    }
    if(!__one_way_buffer_receiver) {
      return false;
    }
    if(!__one_way_cevent_sender) {
      return false;
    }
    if(!__one_way_cevent_receiver) {
      return false;
    }
    
    return true;
  }
  
  void InstreamCompositeStrategy::setAddress(const TSAddress& address)
    throw(const BindingException&)
  {
    /* address must be an InstreamCompositeAddress */
    const InstreamCompositeAddress* composite_address = dynamic_cast<const InstreamCompositeAddress*>(&address);
    if(!composite_address) {
      throw BindingException("Address is not an InstreamCompositeAddress");
    }

    /* composite address must be complete */
    if(!composite_address->isComplete()) {
      throw BindingException("Address is incomplete");
    }
    
    /* check for null pointers */
    for(unsigned int i=0; i<composite_address->getNumberOfAddresses(); ++i) {
      const TSAddress* address = composite_address->getAddress(i);
      if(!address) {
        throw BindingException("Composite address contains NULL pointers");
      }
    }
    
    /* BEGIN> Create child strategies */
    TInOutValue<NMMReference> composite_reference(0);
    for(unsigned int i=0; i<composite_address->getNumberOfAddresses(); ++i) {
      int index = i;
      const TSAddress* address = composite_address->getAddress(i);
      
      /* invert interfaces */
      unsigned int partner_interfaces = TSDescription::getPartnerInterfaces(address->getStrategyInterfaces());
      
      /* create strategy */
      TransportStrategy* strategy = STSFactory::getInstance().createTransportStrategy(address->getStrategyType(), partner_interfaces);
      
      /* BEGIN> Set transport roles */
      /* determine which roles the strategy has on the remote side. this is defined directly by the TSAddress.
         also add the strategy locally for the inverse roles. */
      unsigned int roles = 0;
      if(index == composite_address->getOneWayBufferReceiverAddress()) {
        setOneWayBufferSenderStrategy(strategy);
        roles |= IInstreamCompositeStrategy::RECEIVE_ONE_WAY_BUFFER;
      }
      if(index == composite_address->getOneWayBufferSenderAddress()) {
        setOneWayBufferReceiverStrategy(strategy);
        roles |= IInstreamCompositeStrategy::SEND_ONE_WAY_BUFFER;
      }
      if(index == composite_address->getOneWayCEventReceiverAddress()) {
        setOneWayCEventSenderStrategy(strategy);
        roles |= IInstreamCompositeStrategy::RECEIVE_ONE_WAY_CEVENT;
      }
      if(index == composite_address->getOneWayCEventSenderAddress()) {
        setOneWayCEventReceiverStrategy(strategy);
        roles |= IInstreamCompositeStrategy::SEND_ONE_WAY_CEVENT;
      }
      /* END< */
      
      /* connect TransportStrategy */
      strategy->setAddress(*address);
      
      /* add TransportStrategy to this composite */
      addChildStrategy(strategy);
      
      /* connect TransportStrategy to partner strategy.
         for the first child strategy, this will also create the partner strategy. */
      Event connect_event(IObjectAcceptor::connectToInstreamCompositeStrategy_event, &composite_reference, new TInValue<nmm_uint32>(roles));
      if(FAILURE == strategy->sendReliableEvent(connect_event)) {
        throw BindingException("Failed to send connectToInstreamCompositeStrategy event");
      }
      if(connect_event.getException()) {
        if(dynamic_cast<const BindingException*>(connect_event.getException())) {
          throw *connect_event.getException();
        } else {
          throw BindingException(connect_event.getException()->getComment());
        }
      }
    }
    /* END< */
    
    /* BEGIN> set remote connection manager interface and reference */
    /* check if we have a connector address */
    const ConnectorAddress* any_connector_address = 0;
    for(unsigned int i=0; i<composite_address->getNumberOfAddresses(); ++i) {
      const TSAddress* address = composite_address->getAddress(i);
      any_connector_address = dynamic_cast<const ConnectorAddress*>(address);
      if(any_connector_address) {
        break;
      }
    }
    
    if(any_connector_address) {
      /* if we have a connector address, we have an IConnectionManager interface */
      const IConnectionManager* icm = any_connector_address->getConnectionManager();
      if(icm) {
        setConnectionManager(dynamic_cast<IConnectionManager*>(icm->copy()), composite_reference.getRValue());
      } else {
        GLOBAL_ERROR_STREAM1("InstreamCompositeStrategy> ConnectorAddress has no IConnectionManager interface" << endl);
        throw BindingException("ConnectorAddress has no IConnectionManager interface");
      }
    } else {
      /* otherwise we grab any address and connect to the remote connection manager service */
      // XXX: fw: this is a hack...
      const TSAddress* any_address = composite_address->getAddress(0);
      Interface_var tmp_interface ( SConnectionManager::getInstance().getService(*any_address, "ConnectionManager") );
      IConnectionManager* icm = tmp_interface->getParentObject()->getInterface<IConnectionManager>();
      if(!icm) {
        GLOBAL_ERROR_STREAM1("InstreamCompositeStrategy> Found ConnectionManager service, but no interface to it" << endl);
        throw BindingException("Found ConnectionManager service, but no interface to it");
      }
      setConnectionManager(icm, composite_reference.getRValue());
    }
    /* END< */
  }
  
  ITransportStrategy* InstreamCompositeStrategy::getOneWayBufferReceiverStrategy() const {
    TransportStrategy* strategy = getOwnOneWayBufferReceiverStrategy();
    if(!strategy) {
      return 0;
    }
    
    return strategy->getInterface<ITransportStrategy>();
  }
  
  ITransportStrategy* InstreamCompositeStrategy::getOneWayCEventReceiverStrategy() const {
    TransportStrategy* strategy = getOwnOneWayCEventReceiverStrategy();
    if(!strategy) {
      return 0;
    }
    
    return strategy->getInterface<ITransportStrategy>();
  }
  
  ITransportStrategy* InstreamCompositeStrategy::getOneWayBufferSenderStrategy() const {
    TransportStrategy* strategy = getOwnOneWayBufferSenderStrategy();
    if(!strategy) {
      return 0;
    }
    
    return strategy->getInterface<ITransportStrategy>();
  }
  
  ITransportStrategy* InstreamCompositeStrategy::getOneWayCEventSenderStrategy() const {
    TransportStrategy* strategy = getOwnOneWayCEventSenderStrategy();
    if(!strategy) {
      return 0;
    }
    
    return strategy->getInterface<ITransportStrategy>();
  }

  TransportStrategy* InstreamCompositeStrategy::getOwnOneWayBufferReceiverStrategy() const {
    if(!getOneWayBufferReceiver()) {
      return 0;
    }
    
    return dynamic_cast<TransportStrategy*>(getOneWayBufferReceiver());
  }
  
  TransportStrategy* InstreamCompositeStrategy::getOwnOneWayCEventReceiverStrategy() const {
    if(!getOneWayCEventReceiver()) {
      return 0;
    }
    
    return dynamic_cast<TransportStrategy*>(getOneWayCEventReceiver());
  }
  
  TransportStrategy* InstreamCompositeStrategy::getOwnOneWayBufferSenderStrategy() const {
    if(!getOneWayBufferSender()) {
      return 0;
    }
    
    return dynamic_cast<TransportStrategy*>(getOneWayBufferSender());
  }
  
  TransportStrategy* InstreamCompositeStrategy::getOwnOneWayCEventSenderStrategy() const {
    if(!getOneWayCEventSender()) {
      return 0;
    }
    
    return dynamic_cast<TransportStrategy*>(getOneWayCEventSender());
  }
  
  ITSOneWayBufferReceiver* InstreamCompositeStrategy::getOneWayBufferReceiver() const {
    return __one_way_buffer_receiver;
  }
  
  ITSOneWayCEventReceiver* InstreamCompositeStrategy::getOneWayCEventReceiver() const {
    return __one_way_cevent_receiver;
  }
  
  ITSOneWayBufferSender* InstreamCompositeStrategy::getOneWayBufferSender() const {
    return __one_way_buffer_sender;
  }
  
  ITSOneWayCEventSender* InstreamCompositeStrategy::getOneWayCEventSender() const {
    return __one_way_cevent_sender;
  }
  
  void InstreamCompositeStrategy::reconfigure() {
    /* check if both sender transport roles are set */
    if(__one_way_buffer_sender && __one_way_cevent_sender) {
      TransportStrategy* buffer_strategy = getOwnOneWayBufferSenderStrategy();
      TransportStrategy* cevent_strategy = getOwnOneWayCEventSenderStrategy();
      if(buffer_strategy == cevent_strategy) {
        /* simple binding */
        GLOBAL_DEBUG_STREAM1("InstreamCompositeStrategy> Using simple binding message mux" << endl);
        setInstreamMultiplexer(new SimpleInstreamMultiplexer());
      } else {
        if(cevent_strategy->getTSDescription().getCapabilities() & TSDescription::TS_RELIABLE) {
          if(buffer_strategy->getTSDescription().getCapabilities() & TSDescription::TS_RELIABLE) {
            /* reliable/reliable binding */
            GLOBAL_DEBUG_STREAM1("InstreamCompositeStrategy> Using reliable binding message mux" << endl);
            setInstreamMultiplexer(new ReliableInstreamMultiplexer());
          } else if(buffer_strategy->getTSDescription().getCapabilities() & TSDescription::TS_PRESERVE_MESSAGE_HEADERS) {
            /* reliable/unreliable binding */
            GLOBAL_DEBUG_STREAM1("InstreamCompositeStrategy> Using reliable/unreliable binding message mux" << endl);
            setInstreamMultiplexer(new UnreliableInstreamMultiplexer()); // XXX: fw: set correct strategy
          } else {
            /* reliable/multimedia binding */
            GLOBAL_DEBUG_STREAM1("InstreamCompositeStrategy> Using reliable/multimedia binding message mux" << endl);
            setInstreamMultiplexer(new UnreliableInstreamMultiplexer()); // XXX: set correct strategy
          }
        } else {
          /* unreliable/unreliable binding */
          GLOBAL_DEBUG_STREAM1("InstreamCompositeStrategy> Using unreliable/unreliable binding message mux" << endl);
          setInstreamMultiplexer(new UnreliableInstreamMultiplexer()); // XXX: set correct strategy
        }
      }
    } else {
      setInstreamMultiplexer(0);
    }
    
    /* check if both receiver transport roles are set */
    if(__one_way_buffer_receiver && __one_way_cevent_receiver) {
      TransportStrategy* buffer_strategy = getOwnOneWayBufferReceiverStrategy();
      TransportStrategy* cevent_strategy = getOwnOneWayCEventReceiverStrategy();
      if(buffer_strategy == cevent_strategy) {
        /* simple binding */
        GLOBAL_DEBUG_STREAM1("InstreamCompositeStrategy> Using simple binding message demux" << endl);
        setInstreamDemultiplexer(new SimpleInstreamDemultiplexer());
      } else {
        if(cevent_strategy->getTSDescription().getCapabilities() & TSDescription::TS_RELIABLE) {
          if(buffer_strategy->getTSDescription().getCapabilities() & TSDescription::TS_RELIABLE) {
            /* reliable/reliable binding */
            GLOBAL_DEBUG_STREAM1("InstreamCompositeStrategy> Using reliable binding message demux" << endl);
            setInstreamDemultiplexer(new ReliableInstreamDemultiplexer());
          } else if(buffer_strategy->getTSDescription().getCapabilities() & TSDescription::TS_PRESERVE_MESSAGE_HEADERS) {
            /* reliable/unreliable binding */
            GLOBAL_DEBUG_STREAM1("InstreamCompositeStrategy> Using reliable/unreliable binding message demux" << endl);
            setInstreamDemultiplexer(new UnreliableInstreamDemultiplexer()); // XXX: set correct strategy
          } else {
            /* reliable/multimedia binding */
            GLOBAL_DEBUG_STREAM1("InstreamCompositeStrategy> Using reliable/multimedia binding message demux" << endl);
            setInstreamDemultiplexer(new UnreliableInstreamDemultiplexer()); // XXX: set correct strategy
          }
        } else {
          /* unreliable/unreliable binding */
          GLOBAL_DEBUG_STREAM1("InstreamCompositeStrategy> Using unreliable/unreliable binding message demux" << endl);
          setInstreamDemultiplexer(new UnreliableInstreamDemultiplexer()); // XXX: set correct strategy
        }
      }
    } else {
      setInstreamDemultiplexer(0);
    }
  }
  
  void InstreamCompositeStrategy::setOneWayBufferSenderStrategy(TransportStrategy* strategy)
    throw(const BindingException&)
  {
    if(strategy) {
      // XXX: fw: TransportStrategy is not a NamedObject
//       GLOBAL_DEBUG_STREAM1("InstreamCompositeStrategy> Setting " << strategy->getName() << " (" << strategy << ") as one-way Buffer sender strategy" << endl);
      GLOBAL_DEBUG_STREAM1("InstreamCompositeStrategy> Setting " << strategy->getStrategyType() << " (" << strategy << ") as one-way Buffer sender strategy" << endl);
      __one_way_buffer_sender = dynamic_cast<ITSOneWayBufferSender*>(strategy);
      if(!__one_way_buffer_sender) {
        /* SHOULD NOT HAPPEN */
        throw BindingException("TransportStrategy does not support the required interfaces!");
      }
      
      __one_way_buffer_sender->setOneWayBufferReceiver(__message_mux);
    } else {
      if(__one_way_buffer_sender) {
        __one_way_buffer_sender->setOneWayBufferReceiver(0);
      }
      
      __one_way_buffer_sender = 0;
    }
    
    reconfigure();
  }
  
  void InstreamCompositeStrategy::setOneWayCEventSenderStrategy(TransportStrategy* strategy)
    throw(const BindingException&)
  {
    if(strategy) {
      // XXX: fw: TransportStrategy is not a NamedObject
//       GLOBAL_DEBUG_STREAM1("InstreamCompositeStrategy> Setting " << strategy->getName() << " (" << strategy << ") as one-way CEvent sender strategy" << endl);
      GLOBAL_DEBUG_STREAM1("InstreamCompositeStrategy> Setting " << strategy->getStrategyType() << " (" << strategy << ") as one-way CEvent sender strategy" << endl);
      __one_way_cevent_sender = dynamic_cast<ITSOneWayCEventSender*>(strategy);
      if(!__one_way_cevent_sender) {
        /* SHOULD NOT HAPPEN */
        throw BindingException("TransportStrategy does not support the required interfaces!");
      }
      
      __one_way_cevent_sender->setOneWayCEventReceiver(__message_mux);
    } else {
      if(__one_way_cevent_sender) {
        __one_way_cevent_sender->setOneWayCEventReceiver(0);
      }
      
      __one_way_cevent_sender = 0;
    }
    
    reconfigure();
  }
  
  void InstreamCompositeStrategy::setOneWayBufferReceiverStrategy(TransportStrategy* strategy)
    throw(const BindingException&)
  {
    if(strategy) {
      // XXX: fw: TransportStrategy is not a NamedObject
//       GLOBAL_DEBUG_STREAM1("InstreamCompositeStrategy> Setting " << strategy->getName() << " (" << strategy << ") as one-way Buffer receiver strategy" << endl);
      GLOBAL_DEBUG_STREAM1("InstreamCompositeStrategy> Setting " << strategy->getStrategyType() << " (" << strategy << ") as one-way Buffer receiver strategy" << endl);
      __one_way_buffer_receiver = dynamic_cast<ITSOneWayBufferReceiver*>(strategy);
      if(!__one_way_buffer_receiver) {
        /* SHOULD NOT HAPPEN */
        throw BindingException("TransportStrategy does not support the required interfaces!");
      }    
      
      /* signal multiplexing */
      __one_way_buffer_receiver->setSignalReceiver(&__buffer_adapter);
      __buffer_adapter.setSignalSender(__one_way_buffer_receiver);
    } else {
      if(__one_way_buffer_receiver) {
        __one_way_buffer_receiver->setSignalReceiver(0);
        __buffer_adapter.setSignalSender(0);
      }
      
      __one_way_buffer_receiver = 0;
    }
    
    reconfigure();
  }
  
  void InstreamCompositeStrategy::setOneWayCEventReceiverStrategy(TransportStrategy* strategy)
    throw(const BindingException&)
  {
    if(strategy) {
      // XXX: fw: TransportStrategy is not a NamedObject
//       GLOBAL_DEBUG_STREAM1("InstreamCompositeStrategy> Setting " << strategy->getName() << " (" << strategy << ") as one-way CEvent receiver strategy" << endl);
      GLOBAL_DEBUG_STREAM1("InstreamCompositeStrategy> Setting " << strategy->getStrategyType() << " (" << strategy << ") as one-way CEvent receiver strategy" << endl);
      __one_way_cevent_receiver = dynamic_cast<ITSOneWayCEventReceiver*>(strategy);
      if(!__one_way_cevent_receiver) {
        /* SHOULD NOT HAPPEN */
        throw BindingException("TransportStrategy does not support the required interfaces!");
      }
      
      /* signal multiplexing */
      __one_way_cevent_receiver->setSignalReceiver(&__cevent_adapter);
      __cevent_adapter.setSignalSender(__one_way_cevent_receiver);
    } else {
      if(__one_way_cevent_receiver) {
        __one_way_cevent_receiver->setSignalReceiver(0);
        __cevent_adapter.setSignalSender(0);
      }
      
      __one_way_cevent_receiver = 0;
    }
    
    reconfigure();
  }
  
  void InstreamCompositeStrategy::setInstreamMultiplexer(InstreamMultiplexer* message_mux) {
    if(__message_mux) {
      delete __message_mux;
    }
    
    __message_mux = message_mux;
    
    if(__message_mux) {
      __message_mux->setParent(this);
    }
    
    /* set receivers for sender interfaces to new message multiplexer */
    if(__one_way_buffer_sender) {
      __one_way_buffer_sender->setOneWayBufferReceiver(__message_mux);
    }
    if(__one_way_cevent_sender) {
      __one_way_cevent_sender->setOneWayCEventReceiver(__message_mux);
    }
  }
  
  void InstreamCompositeStrategy::setInstreamDemultiplexer(InstreamDemultiplexer* message_demux) {
    if(__message_demux) {
      delete __message_demux;
    }
    
    __message_demux = message_demux;
    
    if(__message_demux) {
      __message_demux->setParent(this);
    }
    
    /* connect message adapters to demultiplexer */
    __buffer_adapter.setCSSignalReceiver(__message_demux);
    __cevent_adapter.setCSSignalReceiver(__message_demux);
  }
  
  void InstreamCompositeStrategy::addTransportStrategy(TransportStrategy* strategy, unsigned int roles) {    
    /* add child strategy */
    addChildStrategy(strategy);
    
    // XXX: fw: TransportStrategy is not a NamedObject
//     GLOBAL_DEBUG_STREAM1("InstreamCompositeStrategy> Added incoming child strategy: " << strategy->getName() << endl);
    GLOBAL_DEBUG_STREAM1("InstreamCompositeStrategy> Added incoming child strategy: " << strategy->getStrategyType() << endl);
    
    /* BEGIN> set roles */
    if(roles & IInstreamCompositeStrategy::RECEIVE_ONE_WAY_BUFFER) {
      setOneWayBufferReceiverStrategy(strategy);
    }
    if(roles & IInstreamCompositeStrategy::SEND_ONE_WAY_BUFFER) {
      setOneWayBufferSenderStrategy(strategy);
    }
    if(roles & IInstreamCompositeStrategy::RECEIVE_ONE_WAY_CEVENT) {
      setOneWayCEventReceiverStrategy(strategy);
    }
    if(roles & IInstreamCompositeStrategy::SEND_ONE_WAY_CEVENT) {
      setOneWayCEventSenderStrategy(strategy);
    }
    /* END< */
  }
  
  Result InstreamCompositeStrategy::receiveCSMessage(Message* message) {
    if(!__message_demux) {
      GLOBAL_ERROR_STREAM1("InstreamCompositeStrategy> No message demultiplexer -> discarding message" << endl);
      message->release();
      return FAILURE;
    }
    
    /* forward message to demultiplexer */
    return __message_demux->receiveCSMessage(message);
  }
  
  void InstreamCompositeStrategy::receiveSignal(Signal signal) {
    if(!__message_mux) {
      GLOBAL_ERROR_STREAM1("InstreamCompositeStrategy> No message multiplexer -> discarding signal" << endl);
      return;
    }
    
    /* forward message to demultiplexer */
    return __message_mux->receiveSignal(signal);
  }
  
  void InstreamCompositeStrategy::setInstreamMuxStrategy(const IInstreamCompositeStrategy::MuxDemuxStrategy& strategy) {
    switch(strategy) {
    case IInstreamCompositeStrategy::SIMPLE_MUX_DEMUX:
      setInstreamMultiplexer(new SimpleInstreamMultiplexer());
      break;
    case IInstreamCompositeStrategy::RELIABLE_RELIABLE_MUX_DEMUX:
      setInstreamMultiplexer(new ReliableInstreamMultiplexer());
      break;
    case IInstreamCompositeStrategy::RELIABLE_UNRELIABLE_MUX_DEMUX:
      setInstreamMultiplexer(new UnreliableInstreamMultiplexer()); // XXX: fw: set correct strategy
      break;
    case IInstreamCompositeStrategy::RELIABLE_MULTIMEDIA_MUX_DEMUX:
      setInstreamMultiplexer(new UnreliableInstreamMultiplexer()); // XXX: fw: set correct strategy
      break;
    default:
      GLOBAL_ERROR_STREAM1("InstreamCompositeStrategy> Invalid mux strategy" << endl);
    }
  }
  
  void InstreamCompositeStrategy::setInstreamDemuxStrategy(const IInstreamCompositeStrategy::MuxDemuxStrategy& strategy) {
    switch(strategy) {
    case IInstreamCompositeStrategy::SIMPLE_MUX_DEMUX:
      setInstreamDemultiplexer(new SimpleInstreamDemultiplexer());
      break;
    case IInstreamCompositeStrategy::RELIABLE_RELIABLE_MUX_DEMUX:
      setInstreamDemultiplexer(new ReliableInstreamDemultiplexer());
      break;
    case IInstreamCompositeStrategy::RELIABLE_UNRELIABLE_MUX_DEMUX:
      setInstreamDemultiplexer(new UnreliableInstreamDemultiplexer()); // XXX: fw: set correct strategy
      break;
    case IInstreamCompositeStrategy::RELIABLE_MULTIMEDIA_MUX_DEMUX:
      setInstreamDemultiplexer(new UnreliableInstreamDemultiplexer()); // XXX: fw: set correct strategy
      break;
    default:
      GLOBAL_ERROR_STREAM1("InstreamCompositeStrategy> Invalid demux strategy" << endl);
    }
  }
  
  void InstreamCompositeStrategy::setBufferLossInterval(const nmm_uint32& /*loss_interval*/) {
    // XXX: fw: implement
  }
  
  void InstreamCompositeStrategy::notifyEventReceived(const nmm_int32& /*stream_counter*/) {
    // XXX: fw: implement
  }
  
}

#include "nmm/nmm_suffix.hpp"
