/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 * 
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#ifndef NMM_INSTREAM_MULTIPLEXER_HPP
#define NMM_INSTREAM_MULTIPLEXER_HPP NMM_INSTREAM_MULTIPLEXER_HPP

#include "nmm/nmm_prefix.hpp"

#include "nmm/comm/messaging/ITSOneWayBufferReceiver.hpp"
#include "nmm/comm/messaging/ITSOneWayCEventReceiver.hpp"
#include "nmm/comm/messaging/ISignalReceiver.hpp"
#include "nmm/comm/messaging/FormatNotSupportedException.hpp"

#include "nmm/utils/thread/ThreadMutex.hpp"

namespace NMM {  
  
  class InstreamCompositeStrategy;
  class Format;
  
  /** Instream message multiplexer. This class is an adapter between
   * an %InstreamCompositeStrategy and its child strategies.
   * Subclasses implement strategies for multiplexing instream
   * messages.
   */
  class InstreamMultiplexer
    : public ITSOneWayBufferReceiver,
      public ITSOneWayCEventReceiver,
      public ISignalReceiver
  {
  public:
    /** Default constructor.
     */
    InstreamMultiplexer() throw();
    
    /** Virtual destructor.
     */
    virtual ~InstreamMultiplexer() throw() {}
    
    /** @name Accesss parent strategy.
     */
    //@{
    /** Set the parent strategy.
     */
    inline void setParent(InstreamCompositeStrategy* parent) throw() { __parent = parent; }
    
    /** Set the parent strategy.
     */
    inline InstreamCompositeStrategy* getParent() const throw() { return __parent; }
    //@}
    
    /** @name Receiver Interface
     */
    //@{
    /** Receive %CEvent one-way.
     * This method passes a %CEvent to the %TransportStrategy without waiting for a reply.
     * @param cevent The buffer to be passed to the %TransportStrategy.
     */
    virtual Result receiveOneWayCEvent(CEvent* cevent) = 0;
    
    /** Receive %Buffer one-way.
     * This method passes a %Buffer to the %TransportStrategy without waiting for a reply.
     * @param buffer The buffer to be passed to the %TransportStrategy.
     */
    virtual Result receiveOneWayBuffer(Buffer* buffer) = 0;
    
    /** Receive a signal.
     */
    virtual void receiveSignal(Signal) = 0;
    
    /** This method does nothing.
     */
    virtual void configure(const Format&)
      throw(const FormatNotSupportedException&) {}
    
    /**
     *
     * This method flushes the buffer and event receiver. 
     *
     */
    virtual void flushITSReceiver() {};      
    //@}
    
  protected:
    /** Get signal mutex.
     */
    const ThreadMutex& getSignalMutex() const throw() {
      return __signal_mutex;
    }
    
    /** Get flag if partner is blocked.
     */
    bool isBlocked() const throw() {
      return __blocked;
    }
    
    /** Set flag if partner is blocked.
     */
    void setBlocked(bool blocked) throw() {
      __blocked = blocked;
    }
    
  private:
    InstreamCompositeStrategy* __parent;         ///< The parent strategy.
    ThreadMutex                __signal_mutex;   ///< Signal mutex.
    bool                       __blocked;        ///< Owner is blocked.
  };
  
}

#include "nmm/nmm_suffix.hpp"

#endif
