/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "Interface.hpp"
#include "NMMObject.hpp"
#include "ProxyObject.hpp"
#include "TSAddress.hpp"
#include "TSFactory.hpp"
#include "ReferenceManager.hpp"

#include "nmm/comm/serialize/TValueConstructor.hpp"
#include "nmm/comm/serialize/TConstructor.hpp"
#include "nmm/comm/serialize/SerializePtrList.hpp"

#include <algorithm>

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  Interface::Interface(NMMObject& receiver, const string& interface_type, const string& name_space)
    : SerializedValue(interface_type, name_space)
    , __parent_object(&receiver)
    , __receiver(0)
    , __parent_reference(receiver.getReference())
    , __parent_user_info(UserInfo())
    , __is_local(true)
  {
    // Get list of supported strategies from receiver object
    receiver.getSupportedStrategies(__supported_strategies.getRValue());

    __address_list.useTValue(false);
    registerAttributes();
  }

  Interface::Interface(const string& interface_type, const string& name_space)
    : SerializedValue(interface_type, name_space)
    , __parent_object(0)
    , __receiver(0)
    , __parent_reference(0)
    , __is_local(true)
  {
    __address_list.useTValue(false);
    registerAttributes();
  }

  Interface::Interface(const Interface& iface)
    : SerializedValue(iface)
    , IEventSender(iface)
    , __receiver(0)
    , __is_local(true)
  {
    __address_list.useTValue(false);
    registerAttributes();

    *this = iface;
  }

  Interface::~Interface() {
    if (__receiver)
      __receiver->decreaseInterfaceCounter();

    for (unsigned int i = 0; i < __address_list.size(); ++i)
      delete __address_list[i];
  }

  void Interface::registerAttributes() {
    /* Begin> Register the Attributes */
    insertArgument(&__parent_reference);
    insertArgument(&__parent_user_info);
    insertArgument(&__address_list);
    insertArgument(&__supported_strategies);
    /* End< */
  }

  bool Interface::isSubType(const string& type) {
	if (std::find(__interfaces.begin(), __interfaces.end(), type) == __interfaces.end())
      return false;
    return true;
  }

  void Interface::addInterfaceType(const string& iface) {
	if (std::find(__interfaces.begin(), __interfaces.end(), iface) == __interfaces.end())
      __interfaces.push_back(iface);
  }


  const SequenceValue<TSAddress>& Interface::getAddressList() const {
    if (__address_list.empty())
      const_cast<Interface*>(this)->prepareAttributes();

    return __address_list;
  }

  void Interface::prepareAttributes() {
    if (!__is_local)  //if the interface is not on its local system use the old strategy list
      return;
  
     for (unsigned int i = 0; i < __address_list.size(); ++i)
       delete __address_list[i];

    __address_list.clear();
  
    /* get the supported TransportStrategies */
    list<TSDescription> tmp_list;
    if(__parent_object) {
      /* if the parent object is known, get the transport strategies
         supported by the parent object */
      __parent_object->getSupportedStrategies(tmp_list);
    } else {
      /* otherwise get all locally supported transport strategies */
      tmp_list = STSFactory::getInstance().getSupportedStrategies();
    }
    
    /* Begin> Store the default address for each TransportStrategy */
    TSAddress* ts_address = 0;  
    for (list< TSDescription >::const_iterator it = tmp_list.begin(); it != tmp_list.end(); ++it) {
      ts_address = getDefaultAddress(it->getStrategyType());
      if(ts_address) {
        /* set the NMMReference of the NMMObject */
        ts_address->setNMMReference(__parent_reference.getRValue());
      
        /* set the low-level interfaces and capabilities to the values
           obtained from the parent object */
        ts_address->setStrategyInterfaces(it->getInterfaces());
        ts_address->setStrategyCapabilities(it->getCapabilities());
      
        __address_list.push_back(ts_address);
      }
    }
    /*End< */

  }

  Interface& Interface::operator=(const Interface& iface) {
    if (this == &iface) {
      return *this;
    }
    __interfaces = iface.__interfaces;
    __parent_object      = iface.__parent_object;


    /* Begin> decrease the reference counter if a receiver exist*/
    if (__receiver) {
      __receiver->decreaseInterfaceCounter();
    }
    /* End< */

    __receiver           = iface.__receiver;

    if (__receiver) {
      __receiver->increaseInterfaceCounter();
    }

    __parent_reference   = iface.__parent_reference;
    __parent_user_info   = iface.__parent_user_info;
    __address_list       = iface.__address_list;
    __is_local           = iface.__is_local;
    __supported_strategies = iface.__supported_strategies;
    return *this;                    
  }

  ProxyObject* Interface::getParentObject()          {
    if (!__receiver)
      requestParentObject();

    return __receiver;
  }

  const ProxyObject* Interface::getParentObject() const    {
    if (!__receiver)
      requestParentObject();

    return __receiver;
  }

  ProxyObject* Interface::getEventReceiver()   {
    if (!__receiver)
      requestParentObject();

    return __receiver;
  }

  const ProxyObject* Interface::getEventReceiver() const  {
    if (!__receiver)
      requestParentObject();

    return __receiver;
  }

  const NMMReference& Interface::getParentReference() const {
    return __parent_reference.getRValue();
  }

  void Interface::setEventReceiver(NMMObject& receiver) {
    // This interface now refers to the given object
    __parent_object     = &receiver;
    __parent_reference  = receiver.getReference();

    // This interface now refers to an object in the local address space
    __parent_user_info  = UserInfo();

    // This interface now supports exactly the transport strategies supported by the given object
    __supported_strategies.getRValue().clear();
    receiver.getSupportedStrategies(__supported_strategies.getRValue());
  }

  void Interface::requestParentObject() const {
    __receiver = SReferenceManager::getInstance().requestProxyObject(*this);
    __receiver->increaseInterfaceCounter();
  }

  TSAddress* Interface::getDefaultAddress(const string& strategy_type) {
    /* If the NMMObject is known, forward the method call*/
    if (__parent_object)
      return __parent_object->getDefaultAddress(strategy_type);

    /* Begin> Search the default address for a specific TransportStrategy */
    for (unsigned int i = 0; i < __address_list.size(); ++i) {
      if (__address_list[i])
        if (__address_list[i]->getStrategyType() == strategy_type)
          return __address_list[i]->copy();
    }
    /* End< */

    return 0;
  }

  void Interface::updateAttributes() {
    __is_local = false;
  }

  bool Interface::isLocalTo(const Interface& iface) const {
    return iface.getParentInfo() == getParentInfo();
  }

  bool Interface::isSystemTo(const Interface& iface) const {
    if (isLocalTo(iface))
      return false;

    if (getParentInfo().getHostname() != iface.getParentInfo().getHostname())
      return false;

    if (getParentInfo().getDomainname() != iface.getParentInfo().getDomainname())
      return false;

    return true;
  }

  bool Interface::isNetworkTo(const Interface& iface) const {
    if (isLocalTo(iface))
      return false;
    if (isSystemTo(iface))
      return false;

    return true;
  }

  void Interface::establishBinding()
    throw (const BindingException&)
  {
    getParentObject()->establishBinding();
  }


  Result Interface::sendEvent(Event& e)   {
    if (!__receiver)
      requestParentObject();

    return __receiver->receiveEvent(e, this);
  }
  
  
  OValueStream& operator<<(OValueStream& ov, const list<Interface*>& interfaceptr_list) {
    return writePtrList<Interface*>(ov, interfaceptr_list, "Interface");
  }
  IValueStream& operator>>(IValueStream& ov, list<Interface*>& interfaceptr_list) {
    return readPtrList<Interface*>(ov, interfaceptr_list, "Interface");
  }

  TValueConstructor<list<Interface*> >     __nmm_list_interfaceptr_value("listptr_Interface", "std");
  TConstructor<Interface>                  __nmm_interface_value("Interface", "NMM");
}

#include "nmm/nmm_suffix.hpp"
