/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 * 
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "LocalStrategy.hpp"
#include "LocalAcceptor.hpp"

#include "TTSFactoryRegistrator.hpp"
#include "TTSFactoryObject.hpp"

#include "nmm/comm/messaging/CEvent.hpp"
#include "nmm/comm/messaging/Buffer.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  LocalStrategy::LocalStrategy()
    : TransportStrategy("LocalStrategy"),
      __partner_strategy(0)
  {}
  
  LocalStrategy::~LocalStrategy() {
    if(__partner_strategy) {
      __partner_strategy->setPartnerStrategy(0);
    }
  }
  
  void LocalStrategy::scheduleDestruction() {
    delete this;
  }
  
  void LocalStrategy::setPartnerStrategy(LocalStrategy* partner_strategy)  {
    __partner_strategy = partner_strategy;
  }

  Result LocalStrategy::receiveOneWayCEvent(CEvent* cevent) {
    if (__partner_strategy) {
      return __partner_strategy->forwardOneWayCEvent(cevent);
    } else {
      GLOBAL_WARNING_STREAM1("LocalStrategy> No partner strategy -> Discarding one-way CEvent");
      cevent->release();
      return FAILURE;
    }
  }
  
  Result LocalStrategy::receiveOneWayBuffer(Buffer* buffer) {
    if (__partner_strategy) {
      return __partner_strategy->forwardOneWayBuffer(buffer);
    } else {
      GLOBAL_WARNING_STREAM1("LocalStrategy> No partner strategy -> Discarding one-way Buffer");
      buffer->release();
      return FAILURE;
    }
  }
  
  Result LocalStrategy::receiveRequestReplyEvent(Event& event) {
    if(__partner_strategy) {
      return __partner_strategy->forwardRequestReplyEvent(event);
    } else {
      GLOBAL_WARNING_STREAM1("LocalStrategy> No partner strategy -> Discarding request-reply Event");
      return FAILURE;
    }
  }
  
  Result LocalStrategy::sendReliableEvent(Event& event) const {
    if(__partner_strategy) {
      return __partner_strategy->forwardReliableEvent(event);
    } else {
      GLOBAL_WARNING_STREAM1("LocalStrategy> No partner strategy -> Discarding reliable Event");
      return FAILURE;
    }
  }
  
  Result LocalStrategy::sendReliableConfigEvent(Event& event) const {
    if(__partner_strategy) {
      return __partner_strategy->forwardReliableConfigEvent(event);
    } else {
      GLOBAL_WARNING_STREAM1("LocalStrategy> No partner strategy -> Discarding reliable config Event");
      return FAILURE;
    }
  }
  
  void LocalStrategy::receiveSignal(Signal signal) {
    if(signal == FREE_QUEUE_SLOT || signal == FREE_UPSTR_QUEUE_SLOT) {
      if(__partner_strategy) {
        __partner_strategy->forwardSignal(signal);
      } else {
        GLOBAL_WARNING_STREAM1("LocalStrategy> No partner strategy -> Discarding signal");
      }
    }
  }
  
  void LocalStrategy::forwardSignal(Signal signal) {
    if(signal == FREE_QUEUE_SLOT || signal == FREE_UPSTR_QUEUE_SLOT) {
      if(getSignalReceiver()) {
        getSignalReceiver()->receiveSignal(signal);
      } else {
        GLOBAL_WARNING_STREAM1("LocalStrategy> No signal receiver -> Discarding signal");
      }
    }
  }
  
  void LocalStrategy::setAddress(const TSAddress& addr)
    throw (const BindingException&)
  {
    const LocalAddress* tmp_address = dynamic_cast<const LocalAddress*>(&addr);
    if (!tmp_address) {
      throw BindingException("Invalid address type");
    }
    
    /* Begin> Create a partner strategy and set our object as partner strategy */
    __partner_strategy = new LocalStrategy();
    __partner_strategy->setPartnerStrategy(this);
    /* End< */
    
    /* Begin> Connect partner strategy to local acceptor */
    LocalAcceptor& local_acceptor = SLocalAcceptor::getInstance();
    __partner_strategy->setEventReceiver(&local_acceptor);
    /* End< */
  }
  
  void LocalStrategy::configure(const Format&)
    throw(const FormatNotSupportedException&)
  {}
  
  bool LocalStrategy::isConnected() const {
    return (__partner_strategy != 0);
  }
  
  Result LocalStrategy::forwardOneWayCEvent(CEvent* cevent) {
    if(getOneWayCEventReceiver()) {
      return getOneWayCEventReceiver()->receiveOneWayCEvent(cevent);
    } else {
      GLOBAL_WARNING_STREAM1("LocalStrategy> No one-way CEvent receiver -> Discarding CEvent");
      cevent->release();
      return FAILURE;
    }
  }
  
  Result LocalStrategy::forwardOneWayBuffer(Buffer* buffer) {
    if(getOneWayBufferReceiver()) {
      return getOneWayBufferReceiver()->receiveOneWayBuffer(buffer);
    } else {
      GLOBAL_WARNING_STREAM1("LocalStrategy> No one-way Buffer receiver -> Discarding Buffer");
      buffer->release();
      return FAILURE;
    }
  }
  
  Result LocalStrategy::forwardRequestReplyEvent(Event& event) {
    if(getRequestReplyEventReceiver()) {
      return getRequestReplyEventReceiver()->receiveRequestReplyEvent(event);
    } else {
      GLOBAL_WARNING_STREAM1("LocalStrategy> No request-reply Event receiver -> Discarding Event");
      return FAILURE;
    }
    
    return FAILURE;
  }
  
  Result LocalStrategy::forwardReliableEvent(Event& event) {
    if(getEventReceiver()) {
      return getEventReceiver()->receiveEvent(event, this);
    } else {
      GLOBAL_WARNING_STREAM1("LocalStrategy> No Event receiver -> Discarding reliable Event");
      return FAILURE;
    }
  }
  
  Result LocalStrategy::forwardReliableConfigEvent(Event& event) {
    return receiveEvent(event);
  }
  
  
  const char LocalStrategy::__strategy_type[] = "LocalStrategy";

  TTSFactoryRegistrator< TTSFactoryObject<LocalStrategy> >
  __local_strategy_prototype("LocalStrategy",
                             TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER |
                             TSDescription::ITS_ONE_WAY_CEVENT_RECEIVER |
                             TSDescription::ITS_ONE_WAY_BUFFER_RECEIVER |
                             TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER |
                             TSDescription::ITS_ONE_WAY_CEVENT_SENDER |
                             TSDescription::ITS_ONE_WAY_BUFFER_SENDER,
                             TSDescription::TS_RELIABLE |
                             TSDescription::TS_RELIABLE_CONNECT |
                             TSDescription::TS_PRESERVE_MESSAGE_HEADERS);
  
}

#include "nmm/nmm_suffix.hpp"
