/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "NMMObject.hpp"

#include "ReferenceManager.hpp"
#include "OutOfBandCompositeStrategy.hpp"
#include "TSFactory.hpp"
#include "TSAddress.hpp"

#include "nmm/comm/messaging/ICSEventReceiver.hpp"
#include "nmm/comm/messaging/TEDObject.hpp"

#include "nmm/comm/serialize/SerializeResult.hpp"
#include "nmm/comm/serialize/StdValue.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  NMMObject::NMMObject()
    : __current_sender(0), __internal_mutex(ThreadMutex::RECURSIVE), __i_registry(0), __ref_count(0), __reference_manager(0)
  {
    /* register the NMMObject and get the NMMReference*/
    __reference = SReferenceManager::getInstance().registerNMMObject(this);
    __reference_manager = &SReferenceManager::getInstance();

    /* Register the Events */
    registerEvents();
  }


  NMMObject::~NMMObject() {
    /* Unregister the NMMObject */
    if (__reference_manager) 
      SReferenceManager::getInstance().unregisterNMMObject(__reference);

    __internal_mutex.lock();

    /* Begin> Disconnect existing strategies */
    for(list<OutOfBandCompositeStrategy*>::iterator it = __strategy_list.begin(); it != __strategy_list.end(); ++it) {
      (*it)->scheduleDestruction();
    }
    __strategy_list.clear();
    /* End< */
    
    __internal_mutex.unlock();

    /* Delete the interfaces */
    for (list<Interface*>::iterator it = getInterfaceList().begin(); it != getInterfaceList().end(); ++it)
      delete *it;

    /* delete the IRegistry */
    delete __i_registry;
  }

  NMMObject::NMMObject(const NMMObject& nmm_object)
    : IEventReceiver(nmm_object)
  {
    /* Nothing must be done! */
  }

  Result NMMObject::receiveEvent(Event& event, const IEventSender* sender) {
    setCurrentSender(const_cast<IEventSender*>(sender));
    return __event_dispatcher.dispatch(event, sender);
  }
  
  Event* NMMObject::getEventPrototype(const string& event_key)    const {
    EDObject* edo = getEventDispatcher().getEDObject(event_key);
    if (edo) {
      return edo->createEvent(event_key);
    }

    GLOBAL_ERROR_STREAM("No prototype for event " << event_key << " available " << endl);
    return 0;
  }

  Event* NMMObject::matchEventPrototype(const string& event_key)    const {
    return getEventDispatcher().matchEventPrototype(event_key);
  }

  Result NMMObject::getReference(NMMReference& reference) const {
    reference = __reference;
    return SUCCESS;
  }

  Result NMMObject::getInterfaceList(list<string>& interface_list) const {
    for (list<Interface*>::const_iterator it = getInterfaceList().begin(); it != getInterfaceList().end(); ++it)
      interface_list.push_back((*it)->getInterfaceType());

    return SUCCESS;
  }

  void NMMObject::registerEvents() {
    getEventDispatcher().registerEvent("NMMObject::getReference",     new TEDObject1<NMMObject, NMMReference, Result(NMMObject::*)(NMMReference&) const>(this, &NMMObject::getReference));
    getEventDispatcher().registerEvent("NMMObject::getInternalInterface",     new TEDObject1<NMMObject, nmm_int32, Interface* (NMMObject::*)( nmm_int32 ) , Interface* >(this, &NMMObject::getInternalInterface));
    getEventDispatcher().registerEvent("NMMObject::getInterfaceSize", new TEDObject0<NMMObject, nmm_uint32 (NMMObject::*)( ) , nmm_uint32 >(this, &NMMObject::getInterfaceSize));
    getEventDispatcher().registerEvent("NMMObject::getInternalInterfaceList", new TEDObject0<NMMObject, SequenceValue<Interface> (NMMObject::*)( ) const, SequenceValue<Interface> >(this, &NMMObject::getInternalInterfaceList));
    getEventDispatcher().registerEvent("NMMObject::getUserInfo",      new TEDObject1<NMMObject, UserInfo, Result(NMMObject::*)(UserInfo& ) const >(this, &NMMObject::getUserInfo));
    getEventDispatcher().registerEvent("NMMObject::getStrategyList",  new TEDObject1<NMMObject, list<string>, Result(NMMObject::*)(list<string>& ) const >(this, &NMMObject::getStrategyList));
    getEventDispatcher().registerEvent("NMMObject::getDefaultAddress",  new TEDObject1<NMMObject, string, TSAddress* (NMMObject::*)(const string&) const, TSAddress* >(this, &NMMObject::getDefaultAddress));
    getEventDispatcher().registerEvent("NMMObject::registerEventListener", new TEDObject0<NMMObject, void (NMMObject::*)(), void >(this, &NMMObject::registerEventListener));
    getEventDispatcher().registerEvent("NMMObject::removeEventListener", new TEDObject0<NMMObject, void (NMMObject::*)(), void >(this, &NMMObject::removeEventListener));
    getEventDispatcher().registerEvent("NMMObject::registerSpecificEventListener", new TEDObject1<NMMObject, string, void (NMMObject::*)(const string&), void >(this, &NMMObject::registerSpecificEventListener));
    getEventDispatcher().registerEvent("NMMObject::removeSpecificEventListener", new TEDObject1<NMMObject, string, void (NMMObject::*)(const string&), void >(this, &NMMObject::removeSpecificEventListener));
    getEventDispatcher().registerEvent("NMMObject::disconnectFromObject", new TEDObject1<NMMObject, nmm_bool, void (NMMObject::*)(const nmm_bool&), void >(this, &NMMObject::disconnectFromObject));
    getEventDispatcher().registerEvent("NMMObject::getPartnerStrategy",new TEDObject0<NMMObject, ITransportStrategy* (NMMObject::*)(), ITransportStrategy* >(this, &NMMObject::getPartnerStrategy));
    getEventDispatcher().registerEvent("NMMObject::getEventList",new TEDObject0<NMMObject, list<string> (NMMObject::*)()const, list<string> >(this, &NMMObject::getEventList));
    getEventDispatcher().registerEvent("NMMObject::getEventPrototype",new TEDObject1<NMMObject, string, Event* (NMMObject::*)(const string&)const, Event* >(this, &NMMObject::getEventPrototype));
    getEventDispatcher().registerEvent("NMMObject::matchEventPrototype",new TEDObject1<NMMObject, string, Event* (NMMObject::*)(const string&)const, Event* >(this, &NMMObject::matchEventPrototype));
    getEventDispatcher().registerEvent("NMMObject::getRegistry",new TEDObject0<NMMObject, Interface* (NMMObject::*)(), Interface* >(this, &NMMObject::getRegistry));
    getEventDispatcher().registerEvent("NMMObject::setRegistry",new TEDObject1<NMMObject, Interface, void (NMMObject::*)(const Interface&), void >(this, &NMMObject::setRegistry));
  }

  Result NMMObject::getUserInfo(UserInfo& user_info) const {
    user_info = UserInfo();
    return SUCCESS;
  }

  Result NMMObject::getStrategyList(list<string>& strategy_list) const {
    strategy_list = STSFactory::getInstance().getStrategyList();
    return SUCCESS;
  }
  
  Result NMMObject::getSupportedStrategies(list<TSDescription>& strategy_list) const {
    // XXX: fw: disable out-of-band optimization
    strategy_list = STSFactory::getInstance().getSupportedStrategies(TSDescription::ITS_ONE_WAY_BUFFER_SENDER |
                                                                     TSDescription::ITS_ONE_WAY_CEVENT_SENDER |
                                                                     TSDescription::ITS_ONE_WAY_BUFFER_RECEIVER |
                                                                     TSDescription::ITS_ONE_WAY_CEVENT_RECEIVER |
                                                                     TSDescription::ITS_ONE_WAY_EVENT_SENDER |
                                                                     TSDescription::ITS_ONE_WAY_EVENT_RECEIVER |
                                                                     TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER |
                                                                     TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER);
    
    /* get a list of all strategies which support out-of-band interfaces.
       only out-of-band interface flags are set in the result. */
//     strategy_list = STSFactory::getInstance().getSupportedStrategies(TSDescription::ITS_ONE_WAY_EVENT_SENDER |
//                                                                      TSDescription::ITS_ONE_WAY_EVENT_RECEIVER |
//                                                                      TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER |
//                                                                      TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER);
    
    return SUCCESS;
  }
  
  TSAddress* NMMObject::getDefaultAddress(const string& strategy) const {
    TSAddress* address = STSFactory::getInstance().getDefaultAddress(strategy);
    if (address) {
      address->setNMMReference(getReference());
      return address;
    }

    return 0;
  }

  void NMMObject::registerEventListener(IEventReceiver* receiver) {
    if (receiver)
      __event_dispatcher.registerEventListener(receiver);
  }

  void NMMObject::registerEventListener(const string& event_key, IEventReceiver* receiver) {
    if (receiver)
      __event_dispatcher.registerEventListener(event_key, receiver);
  }

  void NMMObject::registerEventListener(const string& key, EDObject* edo) {
    __event_dispatcher.registerEventListener(key, edo);
  }

  void NMMObject::removeEventListener(IEventReceiver* receiver) {
    __event_dispatcher.removeEventListener(receiver);
  }

  void NMMObject::removeEventListener(const string& event_key, IEventReceiver* receiver) {
    if (receiver)
      __event_dispatcher.removeEventListener(event_key, receiver);
  }

  void NMMObject::removeEventListener(EDObject* ed) {
    __event_dispatcher.removeEventListener(ed);
  }
  
  void NMMObject::addTransportStrategy(OutOfBandCompositeStrategy* strategy) {
    /* Begin> Delete TransportStrategies that are no longer connected */
    for(list<OutOfBandCompositeStrategy*>::iterator it = __strategy_list.begin();
        it != __strategy_list.end();)
      {
        if(*it) { 
          if(!(*it)->isConnected()) {
            (*it)->scheduleDestruction();
            it = __strategy_list.erase(it);
            continue;
          }
        }
        ++it;
      }
    /* End< */
    
    if ( strategy ) {
      /* add strategy to list */
      __strategy_list.push_back(strategy);
    
      /* set event receiver of strategy */
      strategy->setEventReceiver(this);
    }
  }
  
  void NMMObject::removeTransportStrategy(OutOfBandCompositeStrategy* strategy) {
    /* unset event receiver of strategy */
    strategy->setEventReceiver(0);
    
    /* remove strategy from list */
    __strategy_list.remove(strategy);      
    
    /* remove listener and all its registrations at the event dispatcher */
    ListenerIterator found = __listener_map.find(strategy);
    if(found != __listener_map.end()) {
      removeEventListener(&found->second);
      __listener_map.erase(found);
    }
    
    /* destroy strategy */
    strategy->scheduleDestruction();
  }
  
  void NMMObject::disconnectFromObject(const nmm_bool& delete_strategy) {
    /* get strategy we received this event from */
    OutOfBandCompositeStrategy* strategy = dynamic_cast<OutOfBandCompositeStrategy*>(getCurrentSender());
    if(!strategy) {
      /* calling this method locally has no effect */
      return;
    }
    
    if(delete_strategy) {
      /* remove the strategy */
      removeTransportStrategy(strategy);
    }
  }
  
  bool NMMObject::isLocalTo(const Interface& iface)  {
    return iface.getParentInfo() == UserInfo();
  }
  
  bool NMMObject::isSystemTo(const Interface& iface) {
    if (isLocalTo(iface))
      return false;

    if (UserInfo().getHostname() != iface.getParentInfo().getHostname())
      return false;

    if (UserInfo().getDomainname() != iface.getParentInfo().getDomainname())
      return false;

    return true;
  }

  bool NMMObject::isNetworkTo(const Interface& iface) {
    if (isLocalTo(iface))
      return false;
    if (isSystemTo(iface))
      return false;

    return true;
  }

  SequenceValue<Interface> NMMObject::getInternalInterfaceList() const {
    SequenceValue<Interface> interface_list;
    interface_list.useTValue(false);
    for (list<Interface*>::const_iterator it = getInterfaceList().begin(); it != getInterfaceList().end(); ++it)
      interface_list.push_back((*it));

    return interface_list;
  }

  nmm_uint32 NMMObject::getInterfaceSize() {
    return static_cast<nmm_uint32>(getInterfaceList().size());
  }

  Interface* NMMObject::getInternalInterface( nmm_int32 i ) {
    if ( getInterfaceList().size() > (unsigned int)(i) ) {
      list<Interface*>::iterator it = getInterfaceList().begin();
      advance(it, i);
      return (*it)->copy();
    }
    return 0;
  }

  ITransportStrategy* NMMObject::getPartnerStrategy() {
    TransportStrategy* strategy = dynamic_cast<TransportStrategy*>(getCurrentSender());
    if (!strategy)
      return 0;

    return strategy->getInterface<ITransportStrategy>();    
  }

  list<string> NMMObject::getEventList() const {
    return getEventDispatcher().getEventList();
  }

  Interface* NMMObject::getRegistry() {
    if ( __i_registry )
      return dynamic_cast<Interface*> (__i_registry->copy());

    return 0;
  }

  void NMMObject::setRegistry(const Interface& registry) {
    delete __i_registry;
    __i_registry = dynamic_cast<Interface*>(registry.copy());
  }
  
  void NMMObject::registerEventListener() {
    OutOfBandCompositeStrategy* strategy = dynamic_cast<OutOfBandCompositeStrategy*>(getCurrentSender());
    if(!strategy) {
      GLOBAL_ERROR_STREAM2("Distributed event listeners may only be registered through a TransportStrategy" << endl);
      return;
    }
    
    ListenerIterator found = __listener_map.find(strategy);
    if(found == __listener_map.end()) {
      /* create a ProxyListener and add it */
      ProxyListener listener(strategy);
      __listener_map.insert(make_pair(strategy, listener));
      found = __listener_map.find(strategy);
    }
    registerEventListener(&found->second);
  }
  
  void NMMObject::removeEventListener() {
    OutOfBandCompositeStrategy* strategy = dynamic_cast<OutOfBandCompositeStrategy*>(getCurrentSender());
    if(!strategy) {
      GLOBAL_ERROR_STREAM2("Distributed event listeners may only be removed through a TransportStrategy" << endl);
      return;
    }
    
    ListenerIterator found = __listener_map.find(strategy);
    if(found != __listener_map.end()) {
      removeEventListener(&found->second);
    }
  }
  
  void NMMObject::registerSpecificEventListener(const string& key) {
    OutOfBandCompositeStrategy* strategy = dynamic_cast<OutOfBandCompositeStrategy*>(getCurrentSender());
    if(!strategy) {
      GLOBAL_ERROR_STREAM2("Distributed event listeners may only be registered through a TransportStrategy" << endl);
      return;
    }
    
    ListenerIterator found = __listener_map.find(strategy);
    if(found == __listener_map.end()) {
      /* create a ProxyListener and add it */
      ProxyListener listener(strategy);
      __listener_map.insert(make_pair(strategy, listener));
      found = __listener_map.find(strategy);
    }
    registerEventListener(key, &found->second);
  }
  
  void NMMObject::removeSpecificEventListener(const string& key) {
    OutOfBandCompositeStrategy* strategy = dynamic_cast<OutOfBandCompositeStrategy*>(getCurrentSender());
    if(!strategy) {
      GLOBAL_ERROR_STREAM2("Distributed event listeners may only be removed through a TransportStrategy" << endl);
      return;
    }
    
    ListenerIterator found = __listener_map.find(strategy);
    if(found != __listener_map.end()) {
      removeEventListener(key, &found->second);
    }
  }

  Result NMMObject::ProxyListener::receiveEvent(Event& event, const IEventSender* sender) {
    if ( sender )    {
      /* Begin> Check if the current sender is also registered as receiver */
      const ICSEventReceiver* tmp_receiver = dynamic_cast<const ICSEventReceiver*>( sender );
      if ( tmp_receiver == __receiver ) {
        /* If the sender is also regsitered as receiver we must not forward the event to avoid an infinite recursion!
           This happens if a ProxyObject sends an event to its NMMObject and is also registered as event listener.
           In this case the ProxyObject itself forwards the event to its listeners, so we must not send it back again. */
        return SUCCESS;
      }
    }
    
    /* Otherwise we forward the event to registered receiver */
    return __receiver->receiveCSEvent(event);
  }

  void NMMObject::setReferenceManager(ReferenceManager* object) {
    __reference_manager = object;
  }

}

#include "nmm/nmm_suffix.hpp"
