/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "ObjectAcceptor.hpp"
#include "ReferenceManager.hpp"
#include "OutOfBandCompositeStrategy.hpp"
#include "InstreamCompositeStrategy.hpp"
#include "InstreamObject.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  ObjectAcceptor::ObjectAcceptor()
  {}
  
  void ObjectAcceptor::connectToObject(const NMMReference& reference)
    throw(const BindingException&)
  {
    /* search the NMMObject with given reference */
    NMMObject* object = SReferenceManager::getInstance().getNMMObject(reference);
    if(!object) {
      GLOBAL_ERROR_STREAM("ObjectAcceptor: Unknown NMM object reference " << reference << endl);
      throw BindingException("Unknwon reference");
    }
    
    /* The current sender must be an OutOfBandCompositeStrategy */
    OutOfBandCompositeStrategy* strategy = dynamic_cast<OutOfBandCompositeStrategy*>(getCurrentSender());
    if(!strategy) {
      GLOBAL_ERROR_STREAM("ObjectAcceptor: Current sender is not an OutOfBandCompositeStrategy" << endl);
      throw BindingException("Unkown sender");
    }
    
    /* Adds the sender to the specific NMMObject */
    object->addTransportStrategy(strategy);
  }
  
  void ObjectAcceptor::connectToJack(const NMMReference& reference)
    throw(const BindingException&)
  {
    /* search the NMMObject with given reference */
    NMMObject* object = SReferenceManager::getInstance().getNMMObject(reference);
    if(!object) {
      GLOBAL_ERROR_STREAM("ObjectAcceptor: Unknown NMM object reference " << reference << endl);
      throw BindingException("Unknwon reference");
    }
    
    /* check if object is a Jack */
    InstreamObject* jack = dynamic_cast<InstreamObject*>(object);
    if(!jack) {
      GLOBAL_ERROR_STREAM("ObjectAcceptor: Referenced object is not a jack " << endl);
      throw BindingException("Invalid object");
    }
    
    /* The current sender must be an InstreamCompositeStrategy */
    InstreamCompositeStrategy* strategy = dynamic_cast<InstreamCompositeStrategy*>(getCurrentSender());
    if(!strategy) {
      GLOBAL_ERROR_STREAM("ObjectAcceptor: Current sender is not an InstreamCompositeStrategy" << endl);
      throw BindingException("Unkown sender");
    }
    
    /* Connect jack and TransportStrategy */
    jack->setTransportStrategy(strategy);
  }
  
  void ObjectAcceptor::connectToService(const string& service)
    throw(const BindingException&)
  {
    /* search the NMMObject with given service name */
    NMMObject* object = SReferenceManager::getInstance().getService(service);
    if(!object) {
      GLOBAL_ERROR_STREAM("ObjectAcceptor: Unknown service " << service << endl);
      throw BindingException("Unknwon service " + service);
    }
    
    /* The current sender must be an OutOfBandCompositeStrategy */
    OutOfBandCompositeStrategy* strategy = dynamic_cast<OutOfBandCompositeStrategy*>(getCurrentSender());
    if (!strategy) {
      GLOBAL_ERROR_STREAM("ObjectAcceptor: Current sender is not an OutOfBandCompositeStrategy" << endl);
      throw BindingException("Unkown sender");
    }
    
    /* Adds the sender to the specific NMMObject */
    object->addTransportStrategy(strategy);
  }
  
  void ObjectAcceptor:: connectToOutOfBandCompositeStrategy(NMMReference& reference, const nmm_uint32& roles)
    throw(const BindingException&)
  {
    /* BEGIN> get the child TransportStrategy which sent this event */
    TransportStrategy* strategy = dynamic_cast<TransportStrategy*>(getCurrentSender());
    if(!strategy) {
      GLOBAL_ERROR_STREAM("ObjectAcceptor: Current sender is not a TransportStrategy" << endl);
      throw BindingException("Unkown sender");
    }
    /* END< */
    
    /* BEGIN> get the parent CompositeStrategy */
    OutOfBandCompositeStrategy* cs;
    if(!reference) {
      /* create a new composite strategy */
      GLOBAL_DEBUG_STREAM2("Create new incoming OutOfbandCompositeStrategy" << std::endl);
      cs = new OutOfBandCompositeStrategy();
      reference = cs->getReference();
      
      /* connect new composite strategy to this acceptor (we could use any acceptor) */
      cs->setEventReceiver(this);
    } else {
      /* search the composite strategy with given reference */
      cs = SReferenceManager::getInstance().getObject<OutOfBandCompositeStrategy>(reference);
      if(!cs) {
        GLOBAL_ERROR_STREAM("ObjectAcceptor: Unknown reference to composite strategy " << reference << endl);
        throw BindingException("Invalid strategy");
      }
    }
    /* END< */
    
    /* add new strategy to composite */
    cs->addTransportStrategy(strategy, roles);
  }

  void ObjectAcceptor::connectToInstreamCompositeStrategy(NMMReference& reference, const nmm_uint32& roles) 
    throw(const BindingException&)
  {
    /* BEGIN> get the child TransportStrategy which sent this event */
    TransportStrategy* strategy = dynamic_cast<TransportStrategy*>(getCurrentSender());
    if(!strategy) {
      GLOBAL_ERROR_STREAM("ObjectAcceptor: Current sender is not a TransportStrategy" << endl);
      throw BindingException("Unkown sender");
    }
    /* END< */
    
    /* BEGIN> get the parent CompositeStrategy */
    InstreamCompositeStrategy* cs;
    if(!reference) {
      /* create a new composite strategy */
      GLOBAL_DEBUG_STREAM2("Create new incoming InstreamCompositeStrategy" << std::endl);
      cs = new InstreamCompositeStrategy();
      reference = cs->getReference();
      
      /* connect new composite strategy to this acceptor (we could use any acceptor) */
      cs->setEventReceiver(this);
    } else {
      /* search the composite strategy with given reference */
      cs = SReferenceManager::getInstance().getObject<InstreamCompositeStrategy>(reference);
      if(!cs) {
        GLOBAL_ERROR_STREAM("ObjectAcceptor: Unknown reference to composite strategy " << reference << endl);
        throw BindingException("Invalid strategy");
      }
    }
    /* END< */
    
    /* add new strategy to composite */
    cs->addTransportStrategy(strategy, roles);
  }
  
  list<string> ObjectAcceptor::getServiceList() const {
    return SReferenceManager::getInstance().getServiceList();
  }
  
}

#include "nmm/nmm_suffix.hpp"
