/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "ObjectDescription.hpp"
#include "Interface.hpp"
#include "serialize/TConstructor.hpp"
#include "NMMDefaultPort.hpp"


#include <algorithm>

#include "nmm/nmm_prefix.hpp"

namespace NMM {

ObjectDescription::ObjectDescription()
  : SerializedValue("ObjectDescription", "NMM"), __location(UserInfo().getHostname()),
    __location_flag(true), __port(NMM::default_port)
{
  registerAttributes();
}

ObjectDescription::ObjectDescription(const string& object_name)
  : SerializedValue("ObjectDescription", "NMM"), __object_name(object_name),
    __location(UserInfo().getHostname()),
    __location_flag(true), __port(NMM::default_port)
{
  registerAttributes();
}


ObjectDescription::~ObjectDescription() {

}


ObjectDescription::ObjectDescription(const ObjectDescription& nd)
  : SerializedValue(nd), __location(UserInfo().getHostname()), __location_flag(true)
{
  registerAttributes();
  *this = nd;
}

void ObjectDescription::addInterface(const Interface& iface) {
  addInterface(iface.getTypeName());
}

void ObjectDescription::addInterface(const string& iface) {
  __interface_list.getRValue().push_back(iface);
}

void ObjectDescription::setLocation(const string& location)
{ 
  if(location.empty() || location == "localhost") {
    __location.setValue(UserInfo().getHostname());
  }
  else {
    __location.setValue(location); 
  }
}
 
void ObjectDescription::disableLocation() { 
  __location_flag.setValue(false);
}

void ObjectDescription::enableLocation() { 
  __location_flag.setValue(true);
}

bool ObjectDescription::isSubset(const ObjectDescription& nd) const {
  
  /*Begin> Compare the ObjectName if it is set*/
  if (!nd.getObjectName().empty()) {
    if (getObjectName() != nd.getObjectName()) {
      return false;
    }
  }
  /*End< */


  /*Begin> Check if locations of objects are identical */
  if (nd.isLocationEnabled()){
    if((getLocation() != nd.getLocation())) {
      return false;
    }
  }
  /*End<*/


  /*Begin> Compare the library name if it is set*/
  if (!nd.getLibraryName().empty()) {
    if (getLibraryName() != nd.getLibraryName())
      return false;
  }
  /*End< */

  /* Begin> check the interfaces */
  if (!isInterfaceSubset(nd)){
    return false;
  }
  /* End< */

  /* If all parameters match return true */
  return true;
}

bool ObjectDescription::isInterfaceSubset(const ObjectDescription& np) const {
  for (list<string>::const_iterator it = np.getInterfaceList().begin(); it != np.getInterfaceList().end(); ++it)
  {
    if (find(__interface_list.getRValue().begin(), __interface_list.getRValue().end(), *it) == __interface_list.getRValue().end())
      return false;
  }
  return true;
}

void ObjectDescription::setPort(int port) {
  __port.setValue(port);
}

void ObjectDescription::registerAttributes() {
  insertArgument(&__object_name);
  insertArgument(&__library_name);
  insertArgument(&__interface_list);
  insertArgument(&__service_description);
  insertArgument(&__config_event);
  insertArgument(&__location);
  insertArgument(&__location_flag);
  insertArgument(&__port);
}

TConstructor<ObjectDescription>    __nmm__objectdescription_info("ObjectDescription", "NMM");

} //end namespace

#include "nmm/nmm_suffix.hpp"
