/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "OutOfBandChannel.hpp"
#include "OutOfBandCompositeAddress.hpp"
#include "OutOfBandCompositeStrategy.hpp"
#include "ProxyObject.hpp"

#include "ISetValueStream.hpp"

#include "TSFactory.hpp"
#include "TSAddress.hpp"
#include "TransportStrategy.hpp"
#include "ConnectionManager.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  OutOfBandChannel::OutOfBandChannel(ProxyObject& parent, const Interface& iface)
    : CommunicationChannel(getStaticTypeInfo()),
      __parent(parent),
      __strategy_list(),
      __strategy(0),
      __request_strategy(),
      __reply_strategy(),
      __request_strategy_valid(false),
      __reply_strategy_valid(false)
  {
    /* clear strategy interface cache */
    for(unsigned int i=0; i<4; ++i) {
      __request_child_strategy[i] = 0;
      __request_child_strategy_valid[i] = false;
      __reply_child_strategy[i] = 0;
      __reply_child_strategy_valid[i] = false;
    }
    
    /* create strategy list from interface */
    updateStrategyList(iface);
  }
  
  OutOfBandChannel::OutOfBandChannel(ProxyObject& parent, const TSAddress& address)
    : CommunicationChannel(getStaticTypeInfo()),
      __parent(parent),
      __strategy_list(),
      __strategy(0),
      __request_strategy(),
      __reply_strategy(),
      __request_strategy_valid(false),
      __reply_strategy_valid(false)
  {
    /* clear strategy interface cache */
    for(unsigned int i=0; i<4; ++i) {
      __request_child_strategy[i] = 0;
      __request_child_strategy_valid[i] = false;
      __reply_child_strategy[i] = 0;
      __reply_child_strategy_valid[i] = false;
    }
    
    /* get the interfaces supported by the strategy type of the given address. */
    unsigned int interfaces = STSFactory::getInstance().getInterfaces(address.getStrategyType());
    if(!interfaces) {
      GLOBAL_WARNING_STREAM("Requested OutOfBandChannel with address of locally unavailable strategy type " << address.getStrategyType() << endl);
      return;
    }
    
    /* get the capabilities of the strategy type of the given address.
       assume that they are the same on the remote side. */
    unsigned int capabilities = STSFactory::getInstance().getCapabilities(address.getStrategyType());    
    
    /* get the inverse interfaces. since we have no information about the remote side,
       we assume that it supports at least these interfaces. if it doesn't,
       then connection will fail. unfortunately we can't tell at this point.
    */
    unsigned int inverse_interfaces = 0;    
    if(interfaces & TSDescription::ITS_ONE_WAY_EVENT_SENDER) {
      inverse_interfaces |= TSDescription::ITS_ONE_WAY_EVENT_RECEIVER;
    }
    if(interfaces & TSDescription::ITS_ONE_WAY_EVENT_RECEIVER) {
      inverse_interfaces |= TSDescription::ITS_ONE_WAY_EVENT_SENDER;
    }
    if(interfaces & TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER) {
      inverse_interfaces |= TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER;
    }
    if(interfaces & TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER) {
      inverse_interfaces |= TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER;
    }
    
    /* create a new address with the inverse interfaces */
    TSAddress* new_address = address.copy();
    new_address->setStrategyInterfaces(inverse_interfaces);
    new_address->setStrategyCapabilities(capabilities);
    addAddress(*new_address);
  }
  
  OutOfBandChannel::OutOfBandChannel(NMMObject& parent)
    : CommunicationChannel(getStaticTypeInfo()),
      __parent(parent),
      __strategy_list(),
      __strategy(0),
      __request_strategy(),
      __reply_strategy(),
      __request_strategy_valid(false),
      __reply_strategy_valid(false)
  {
    /* clear strategy interface cache */
    for(unsigned int i=0; i<4; ++i) {
      __request_child_strategy[i] = 0;
      __request_child_strategy_valid[i] = false;
      __reply_child_strategy[i] = 0;
      __reply_child_strategy_valid[i] = false;
    }
  }
  
  OutOfBandChannel::OutOfBandChannel(const OutOfBandChannel& channel)
    : CommunicationChannel(channel),
      __parent(channel.__parent),
      __strategy_list(),
      __strategy(0),
      __request_strategy(),
      __reply_strategy(),
      __request_strategy_valid(false),
      __reply_strategy_valid(false)
  {
    /* clear strategy interface cache */
    for(unsigned int i=0; i<4; ++i) {
      __request_child_strategy[i] = 0;
      __request_child_strategy_valid[i] = false;
      __reply_child_strategy[i] = 0;
      __reply_child_strategy_valid[i] = false;
    }
  }
  
  OutOfBandChannel* OutOfBandChannel::copy() const {
    return new OutOfBandChannel(*this);
  }
  
  OutOfBandChannel::~OutOfBandChannel() {
    /* close binding */
    unsetTransportStrategy();
    
    /* clean up strategy list */
    for(StrategyIterator i = __strategy_list.begin(); i != __strategy_list.end(); ++i) {
      delete i->default_address;
    }
  }
  
  list<string> OutOfBandChannel::getStrategyList() const {
    list<string> temp;
    for(StrategyConstIterator i=__strategy_list.begin();
        i != __strategy_list.end();
        ++i) {
      temp.push_back(i->strategy_type);
    }
    return temp;
  }
  
  list<string> OutOfBandChannel::getStrategyList(unsigned int request_interfaces, unsigned int reply_interfaces) const {
    list<string> temp;
    for(StrategyConstIterator i=__strategy_list.begin();
        i != __strategy_list.end();
        ++i) {
      if((request_interfaces & i->request_interfaces) == request_interfaces &&
         (reply_interfaces & i->reply_interfaces) == reply_interfaces) {
        temp.push_back(i->strategy_type);
      }
    }
    return temp;
  }
  
  list<string> OutOfBandChannel::getStrategyList(unsigned int roles) const {
    /* the set of required interfaces is not unique. Since both method calls and listener notification
       are request-reply, we have the choice to use either two one-way leaf strategies or one request-reply
       leaf strategy for method calls and again for listener notification. Thus we get 4 different sets of
       required interfaces. For each set of interfaces, we get a set of strategies that support it. The union
       of these strategy sets is the final result.
       
       0 : request-reply method calls and request-reply listener notification
       1 : request-reply method calls and one-way       listener notification
       2 : one-way       method calls and request-reply listener notification
       3 : one-way       method calls and one-way       listener notification
       
    */
    unsigned int request_interfaces[4];
    unsigned int reply_interfaces[4];
    
    /* initialize interface sets */
    for(unsigned int i=0; i<4; ++i) {
      request_interfaces[i] = 0;
      reply_interfaces[i] = 0;
    }
    
    /* check whether we can use request-reply leaf strategies for method calls */
    if((roles & REQUEST) && (roles & REPLY)) {
      request_interfaces[0] |= TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER;
      request_interfaces[1] |= TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER;
      reply_interfaces[0] |= TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER;
      reply_interfaces[1] |= TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER;
    }
    
    /* check whether we can use request-reply leaf strategies for listener notificatin */
    if((roles & LISTENER_REQUEST) && (roles & LISTENER_REPLY)) {
      request_interfaces[0] |= TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER;
      request_interfaces[2] |= TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER;
      reply_interfaces[0] |= TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER;
      reply_interfaces[2] |= TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER;
    }
    
    /* check if we can use one-way leaf strategies for method calls */
    if(roles & REQUEST) {
      request_interfaces[2] |= TSDescription::ITS_ONE_WAY_EVENT_RECEIVER;
      request_interfaces[3] |= TSDescription::ITS_ONE_WAY_EVENT_RECEIVER;
      reply_interfaces[2] |= TSDescription::ITS_ONE_WAY_EVENT_SENDER;
      reply_interfaces[3] |= TSDescription::ITS_ONE_WAY_EVENT_SENDER;
    }
    if(roles & REPLY) {
      request_interfaces[2] |= TSDescription::ITS_ONE_WAY_EVENT_SENDER;
      request_interfaces[3] |= TSDescription::ITS_ONE_WAY_EVENT_SENDER;
      reply_interfaces[2] |= TSDescription::ITS_ONE_WAY_EVENT_RECEIVER;
      reply_interfaces[3] |= TSDescription::ITS_ONE_WAY_EVENT_RECEIVER;
    }
    
    /* check if we can use one-way leaf strategies for listener notification */
    if(roles & LISTENER_REQUEST) {
      request_interfaces[1] |= TSDescription::ITS_ONE_WAY_EVENT_SENDER;
      request_interfaces[3] |= TSDescription::ITS_ONE_WAY_EVENT_SENDER;
      reply_interfaces[1] |= TSDescription::ITS_ONE_WAY_EVENT_RECEIVER;
      reply_interfaces[3] |= TSDescription::ITS_ONE_WAY_EVENT_RECEIVER;
    }
    if(roles & LISTENER_REPLY) {
      request_interfaces[1] |= TSDescription::ITS_ONE_WAY_EVENT_RECEIVER;
      request_interfaces[3] |= TSDescription::ITS_ONE_WAY_EVENT_RECEIVER;
      reply_interfaces[1] |= TSDescription::ITS_ONE_WAY_EVENT_SENDER;
      reply_interfaces[3] |= TSDescription::ITS_ONE_WAY_EVENT_SENDER;
    }
    
    /* for each set of interfaces, find the set of strategies which supports it */
    list<string> temp;
    for(unsigned int i=0; i<4; ++i) {
      if(!request_interfaces[i] && !reply_interfaces[i]) {
        continue;
      }
      
      list<string> found = getStrategyList(request_interfaces[i], reply_interfaces[i]);
      temp.insert(temp.end(), found.begin(), found.end());
    }
    
    return temp;
  }

  void OutOfBandChannel::addTransportStrategyInternal(const string& strategy_type, unsigned int request_interfaces,
                                                      unsigned int reply_interfaces, bool is_explicit)
    throw(const BindingException&)
  {
    /* check for consistency */
    // XXX: fw: this is wrong! one interface can be used for listener notification and the other for request-reply
    if((request_interfaces & TSDescription::ITS_ONE_WAY_EVENT_RECEIVER) &&
       (request_interfaces & TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER)) {
      throw BindingException("Can't use a single strategy for both one-way and request-reply interaction");
    }
    if((reply_interfaces & TSDescription::ITS_ONE_WAY_EVENT_SENDER) &&
       (reply_interfaces & TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER)) {
      throw BindingException("Can't use a single strategy for both one-way and request-reply interaction");
    }

    /* check if strategy is supported */
    bool strategy_found = false;
    StrategyConstIterator found;
    for(found=__strategy_list.begin(); found != __strategy_list.end(); ++found) {
      /* check if strategy has the right type */
      if(found->strategy_type == strategy_type) {
        strategy_found = true;
        
        /* check if strategy supports all required interfaces */
        if(request_interfaces == (request_interfaces & found->request_interfaces) &&
           reply_interfaces == (reply_interfaces & found->reply_interfaces)) {
          break;
        }
      }
    }
    if(found == __strategy_list.end()) {
      if(!strategy_found) {
        throw BindingException(strategy_type + " can not be used to connect to this object");
      } else {
        throw BindingException(strategy_type + " doesn't have all of the requested interfaces");
      }
    }
    
    /* check for collision */
    for(unsigned int i=0; i<__strategies.size(); ++i) {
      /* same interface already used */
      if(__strategies[i].used_request_interfaces & request_interfaces) {
        throw BindingException(__strategies[i].strategy_type + " has already been added with some of the requested request interfaces");
      }
      if(__strategies[i].used_reply_interfaces & reply_interfaces) {
        throw BindingException(__strategies[i].strategy_type + " has already been added with some of the requested reply interfaces");
      }
      
      /* mutually exclusive interface already used */
      if((__strategies[i].used_request_interfaces & TSDescription::ITS_ONE_WAY_EVENT_RECEIVER) &&
         (request_interfaces & TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER)) {
        throw BindingException(__strategies[i].strategy_type + " has already been added for request-reply interaction");
      }
      if((__strategies[i].used_request_interfaces & TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER) &&
         (request_interfaces & TSDescription::ITS_ONE_WAY_EVENT_RECEIVER)) {
        throw BindingException(__strategies[i].strategy_type + " has already been added for one-way interaction");
      }
      if((__strategies[i].used_reply_interfaces & TSDescription::ITS_ONE_WAY_EVENT_SENDER) &&
         (reply_interfaces & TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER)) {
        throw BindingException(__strategies[i].strategy_type + " has already been added for request-reply interaction");
      }
      if((__strategies[i].used_reply_interfaces & TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER) &&
         (reply_interfaces & TSDescription::ITS_ONE_WAY_EVENT_SENDER)) {
        throw BindingException(__strategies[i].strategy_type + " has already been added for one-way interaction");
      }
    }
    
    /* add strategy to config. */
    StrategyListEntry entry = *found;
    entry.used_request_interfaces = request_interfaces;
    entry.used_reply_interfaces = reply_interfaces;
    entry.default_address = entry.default_address->copy();
    entry.default_address->setStrategyInterfaces(entry.used_request_interfaces);
    entry.is_explicit = is_explicit;
    __strategies.push_back(entry);
    if(is_explicit) {
      GLOBAL_DEBUG_STREAM2("InstreamChannel: Added explicit strategy to config " << entry.strategy_type << endl);
    } else {
      GLOBAL_DEBUG_STREAM2("InstreamChannel: Added strategy to config " << entry.strategy_type << endl);
    }
  }
  
  void OutOfBandChannel::addTransportStrategyInternal(const string& strategy_type, unsigned int roles, bool is_explicit)
    throw(const BindingException&) 
  {
    /* the set of required interfaces is not unique. Since both method calls and listener notification
       are request-reply, we have the choice to use either two one-way leaf strategies or one request-reply
       leaf strategy for method calls and again for listener notification. Thus we get 4 different sets of
       required interfaces. For each set of interfaces, we get a set of strategies that support it. The union
       of these strategy sets is the final result.
       
       0 : request-reply method calls and request-reply listener notification
       1 : request-reply method calls and one-way       listener notification
       2 : one-way       method calls and request-reply listener notification
       3 : one-way       method calls and one-way       listener notification
       
    */
    unsigned int request_interfaces[4];
    unsigned int reply_interfaces[4];
    
    /* initialize interface sets */
    for(unsigned int i=0; i<4; ++i) {
      request_interfaces[i] = 0;
      reply_interfaces[i] = 0;
    }
    
    /* check whether we can use request-reply leaf strategies for method calls */
    if((roles & REQUEST) && (roles & REPLY)) {
      request_interfaces[0] |= TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER;
      request_interfaces[1] |= TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER;
      reply_interfaces[0] |= TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER;
      reply_interfaces[1] |= TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER;
    }
    
    /* check whether we can use request-reply leaf strategies for listener notificatin */
    if((roles & LISTENER_REQUEST) && (roles & LISTENER_REPLY)) {
      request_interfaces[0] |= TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER;
      request_interfaces[2] |= TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER;
      reply_interfaces[0] |= TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER;
      reply_interfaces[2] |= TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER;
    }
    
    /* check if we can use one-way leaf strategies for method calls */
    if(roles & REQUEST) {
      request_interfaces[2] |= TSDescription::ITS_ONE_WAY_EVENT_RECEIVER;
      request_interfaces[3] |= TSDescription::ITS_ONE_WAY_EVENT_RECEIVER;
      reply_interfaces[2] |= TSDescription::ITS_ONE_WAY_EVENT_SENDER;
      reply_interfaces[3] |= TSDescription::ITS_ONE_WAY_EVENT_SENDER;
    }
    if(roles & REPLY) {
      request_interfaces[2] |= TSDescription::ITS_ONE_WAY_EVENT_SENDER;
      request_interfaces[3] |= TSDescription::ITS_ONE_WAY_EVENT_SENDER;
      reply_interfaces[2] |= TSDescription::ITS_ONE_WAY_EVENT_RECEIVER;
      reply_interfaces[3] |= TSDescription::ITS_ONE_WAY_EVENT_RECEIVER;
    }
    
    /* check if we can use one-way leaf strategies for listener notification */
    if(roles & LISTENER_REQUEST) {
      request_interfaces[1] |= TSDescription::ITS_ONE_WAY_EVENT_SENDER;
      request_interfaces[3] |= TSDescription::ITS_ONE_WAY_EVENT_SENDER;
      reply_interfaces[1] |= TSDescription::ITS_ONE_WAY_EVENT_RECEIVER;
      reply_interfaces[3] |= TSDescription::ITS_ONE_WAY_EVENT_RECEIVER;
    }
    if(roles & LISTENER_REPLY) {
      request_interfaces[1] |= TSDescription::ITS_ONE_WAY_EVENT_RECEIVER;
      request_interfaces[3] |= TSDescription::ITS_ONE_WAY_EVENT_RECEIVER;
      reply_interfaces[1] |= TSDescription::ITS_ONE_WAY_EVENT_SENDER;
      reply_interfaces[3] |= TSDescription::ITS_ONE_WAY_EVENT_SENDER;
    }
    
    /* for each set of interfaces, check if the given strategy can be used.
       Use the first set of interfaces that works (prefering request-reply over one-way). */
    if(request_interfaces[0] && reply_interfaces[0]) {
      try {
        GLOBAL_DEBUG_STREAM2("OutOfBandChannel: Trying to add strategy" << strategy_type << " with configuration 0" << std::endl);
        addTransportStrategyInternal(strategy_type, request_interfaces[0], reply_interfaces[0], is_explicit);
        return;
      } catch(const BindingException& e0) {
        GLOBAL_DEBUG_STREAM2("OutOfBandChannel: Failed to add strategy" << strategy_type << " with configuration 0" << std::endl);
        GLOBAL_DEBUG_STREAM2(e0 << endl);
        
        if(request_interfaces[1] && reply_interfaces[1]) {
          try {
            GLOBAL_DEBUG_STREAM2("OutOfBandChannel: Trying to add strategy" << strategy_type << " with configuration 1" << std::endl);
            addTransportStrategyInternal(strategy_type, request_interfaces[1], reply_interfaces[1], is_explicit);
            return;
          } catch(const BindingException& e1) {
            GLOBAL_DEBUG_STREAM2("OutOfBandChannel: Failed to add strategy" << strategy_type << " with configuration 1" << std::endl);
            GLOBAL_DEBUG_STREAM2(e1 << endl);
            
            if(request_interfaces[2] && reply_interfaces[2]) {
              try {
                GLOBAL_DEBUG_STREAM2("OutOfBandChannel: Trying to add strategy" << strategy_type << " with configuration 2" << std::endl);
                addTransportStrategyInternal(strategy_type, request_interfaces[2], reply_interfaces[2], is_explicit);
                return;
              } catch(const BindingException& e2) {
                GLOBAL_DEBUG_STREAM2("OutOfBandChannel: Failed to add strategy" << strategy_type << " with configuration 2" << std::endl);
                GLOBAL_DEBUG_STREAM2(e2 << endl);
                
                if(request_interfaces[3] && reply_interfaces[3]) {
                  try {
                    GLOBAL_DEBUG_STREAM2("OutOfBandChannel: Trying to add strategy" << strategy_type << " with configuration 3" << std::endl);
                    addTransportStrategyInternal(strategy_type, request_interfaces[3], reply_interfaces[3], is_explicit);
                    return;
                  } catch(const BindingException& e3) {
                    GLOBAL_ERROR_STREAM("OutOfBandChannel: Failed to add strategy" << strategy_type << " with configuration 0" << std::endl);
                    GLOBAL_ERROR_STREAM(e0 << endl);
                    GLOBAL_ERROR_STREAM("OutOfBandChannel: Failed to add strategy" << strategy_type << " with configuration 1" << std::endl);
                    GLOBAL_ERROR_STREAM(e1 << endl);
                    GLOBAL_ERROR_STREAM("OutOfBandChannel: Failed to add strategy" << strategy_type << " with configuration 2" << std::endl);
                    GLOBAL_ERROR_STREAM(e2 << endl);
                    GLOBAL_ERROR_STREAM("OutOfBandChannel: Failed to add strategy" << strategy_type << " with configuration 3" << std::endl);
                    GLOBAL_ERROR_STREAM(e3 << endl);
                    
                    throw BindingException("The strategy " + strategy_type + " can not be used for the given communication roles");
                  }
                }
              }
            }
          }
        }
      }
    }    
  }
  
  void OutOfBandChannel::addTransportStrategy(const string& strategy_type, unsigned int request_interfaces, unsigned int reply_interfaces)
    throw(const BindingException&) 
  {
    addTransportStrategyInternal(strategy_type, request_interfaces, reply_interfaces, false);
  }
  
  void OutOfBandChannel::addExplicitTransportStrategy(const string& strategy_type, unsigned int request_interfaces, unsigned int reply_interfaces)
    throw(const BindingException&) 
  {
    addTransportStrategyInternal(strategy_type, request_interfaces, reply_interfaces, true);
  }
  
  void OutOfBandChannel::addTransportStrategy(const string& strategy_type, unsigned int roles)
    throw(const BindingException&) 
  {
    addTransportStrategyInternal(strategy_type, roles, false);
  }
  
  void OutOfBandChannel::addExplicitTransportStrategy(const string& strategy_type, unsigned int roles)
    throw(const BindingException&) 
  {
    return addTransportStrategyInternal(strategy_type, roles, true);
  }
  
  void OutOfBandChannel::establishBinding() throw(const BindingException&) {
    /* check if binding is already established */
    if(getTransportStrategy()) {
      return;
    }
    
    /* BEGIN> check if for all roles a transport strategy has been set */
    unsigned int set_roles = 0;
    for(unsigned int i=0; i<__strategies.size(); ++i) {
      if((__strategies[i].used_request_interfaces & TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER) &&
         (__strategies[i].used_reply_interfaces & TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER)) {
        set_roles |= REQUEST_REPLY;
      }
      if((__strategies[i].used_request_interfaces & TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER) &&
         (__strategies[i].used_reply_interfaces & TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER)) {
        set_roles |= LISTENER_NOTIFICATION;
      }
      if((__strategies[i].used_request_interfaces & TSDescription::ITS_ONE_WAY_EVENT_RECEIVER) &&
         (__strategies[i].used_reply_interfaces & TSDescription::ITS_ONE_WAY_EVENT_SENDER)) {
        // XXX: fw: WRONG! store roles in strategies!
        set_roles |= REQUEST | LISTENER_REPLY;
      }
      if((__strategies[i].used_request_interfaces & TSDescription::ITS_ONE_WAY_EVENT_SENDER) &&
         (__strategies[i].used_reply_interfaces & TSDescription::ITS_ONE_WAY_EVENT_RECEIVER)) {
        // XXX: fw: WRONG! store roles in strategies!
        set_roles |= REPLY | LISTENER_REQUEST;
      }
    }
    /* END< */
    
    /* BEGIN> Automatically fill unset roles */
    /* we first try to reuse already added strategies, then we check for supported
       and suitable strategies. we do this in a loop, because any new strategy we might
       add for one unset role can be reused for another unset role.
       we prefer reqiest-reply leaf strategies over one-way leaf strategies, if we have
       to set both the request and the reply role. */
    while(set_roles != OUT_OF_BAND) {
      /* try to reuse strategies if possible */
      if(!(set_roles & REQUEST) && !(set_roles & REPLY)) {
        if(tryReuseStrategy(TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER, TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER)) {
          set_roles |= REQUEST | REPLY;
        }
      }
      if(!(set_roles & REQUEST)) {
        if(tryReuseStrategy(TSDescription::ITS_ONE_WAY_EVENT_RECEIVER, TSDescription::ITS_ONE_WAY_EVENT_SENDER)) {
          set_roles |= REQUEST;
        }
      }
      if(!(set_roles & REPLY)) {
        if(tryReuseStrategy(TSDescription::ITS_ONE_WAY_EVENT_SENDER, TSDescription::ITS_ONE_WAY_EVENT_RECEIVER)) {
          set_roles |= REPLY;
        }
      }
      if(!(set_roles & LISTENER_REQUEST) && !(set_roles & LISTENER_REPLY)) {
        if(tryReuseStrategy(TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER, TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER)) {
          set_roles |= LISTENER_REQUEST | LISTENER_REPLY;
        }
      }
      if(!(set_roles & LISTENER_REQUEST)) {
        if(tryReuseStrategy(TSDescription::ITS_ONE_WAY_EVENT_SENDER, TSDescription::ITS_ONE_WAY_EVENT_RECEIVER)) {
          set_roles |= LISTENER_REQUEST;
        }
      }
      if(!(set_roles & LISTENER_REPLY)) {
        if(tryReuseStrategy(TSDescription::ITS_ONE_WAY_EVENT_RECEIVER, TSDescription::ITS_ONE_WAY_EVENT_SENDER)) {
          set_roles |= LISTENER_REPLY;
        }
      }
      
      /* try to find suitable strategies */
      if(!(set_roles & REQUEST) && !(set_roles & REPLY)) {
        if(findSuitableStrategy(TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER, TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER)) {
          set_roles |= REQUEST | REPLY;
          continue;
        }
      }
      if(!(set_roles & REQUEST)) {
        if(findSuitableStrategy(TSDescription::ITS_ONE_WAY_EVENT_RECEIVER, TSDescription::ITS_ONE_WAY_EVENT_SENDER)) {
          set_roles |= REQUEST;
          continue;
        }
      }
      if(!(set_roles & REPLY)) {
        if(findSuitableStrategy(TSDescription::ITS_ONE_WAY_EVENT_SENDER, TSDescription::ITS_ONE_WAY_EVENT_RECEIVER)) {
          set_roles |= REPLY;
          continue;
        }
      }
      if(!(set_roles & LISTENER_REQUEST) && !(set_roles & LISTENER_REPLY)) {
        if(findSuitableStrategy(TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER, TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER)) {
          set_roles |= LISTENER_REQUEST | LISTENER_REPLY;
          continue;
        }
      }
      if(!(set_roles & LISTENER_REQUEST)) {
        if(findSuitableStrategy(TSDescription::ITS_ONE_WAY_EVENT_SENDER, TSDescription::ITS_ONE_WAY_EVENT_RECEIVER)) {
          set_roles |= LISTENER_REQUEST;
          continue;
        }
      }
      if(!(set_roles & LISTENER_REPLY)) {
        if(findSuitableStrategy(TSDescription::ITS_ONE_WAY_EVENT_RECEIVER, TSDescription::ITS_ONE_WAY_EVENT_SENDER)) {
          set_roles |= LISTENER_REPLY;
          continue;
        }
      }
      
      if(set_roles != OUT_OF_BAND) {
        /* give up */
        throw BindingException("Could not find suitable transport strategies to realize out-of-band binding");
      } else {
        /* done */
        break;
      }
    }
    /* END< */
    
    /* BEGIN> Create a composite address */
    OutOfBandCompositeAddress composite_address;
    for(unsigned int i=0; i<__strategies.size(); ++i) {
      /* add address to compoite address */
      unsigned int index = composite_address.addAddress(*__strategies[i].default_address);
      
      /* set transport roles */
      if(__strategies[i].used_request_interfaces & TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER) {
        composite_address.setRequestReplyEventReceiverAddress(index);
      }
      if(__strategies[i].used_request_interfaces & TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER) {
        composite_address.setRequestReplyEventSenderAddress(index);
      }
      if(__strategies[i].used_request_interfaces & TSDescription::ITS_ONE_WAY_EVENT_RECEIVER) {
        // XXX: fw: WRONG! store roles in strategies!
        composite_address.setOneWayEventRequestReceiverAddress(index);
        composite_address.setOneWayEventReplyReceiverAddress(index);
      }
      if(__strategies[i].used_request_interfaces & TSDescription::ITS_ONE_WAY_EVENT_SENDER) {
        // XXX: fw: WRONG! store roles in strategies!
        composite_address.setOneWayEventRequestSenderAddress(index);
        composite_address.setOneWayEventReplySenderAddress(index);
      }
    }
    /* END< */
    
    /* BEGIN> Create and connect a composite strategy */
    try {
      /* create strategy */
      OutOfBandCompositeStrategy* strategy = new OutOfBandCompositeStrategy();
      
      /* set strategy */
      setTransportStrategy(strategy);
      
      /* connect strategy */
      getTransportStrategy()->setAddress(composite_address);
    } catch(const BindingException&) {
      /* clean up and rethrow exception */
      unsetTransportStrategy();
      throw;
    }
    /* END< */
  }
  
  void OutOfBandChannel::unsetTransportStrategy() {
    /* unset transport strategy */
    setTransportStrategy(0);
    
    /* clear config */
    for(unsigned int i=0; i<__strategies.size(); ++i) {
      delete __strategies[i].default_address;
    }
    __strategies.clear();
  }
  
  void OutOfBandChannel::setTransportStrategy(OutOfBandCompositeStrategy* strategy) {
    /* delete interfaces */
    delete __request_strategy;
    delete __reply_strategy;
    __request_strategy = 0;
    __reply_strategy = 0;
    __request_strategy_valid = false;
    __reply_strategy_valid = false;
    
    for(unsigned int i=0; i<4; ++i) {
      delete __request_child_strategy[i];
      delete __reply_child_strategy[i];
      __request_child_strategy[i] = 0;
      __reply_child_strategy[i] = 0;
      __request_child_strategy_valid[i] = false;
      __reply_child_strategy_valid[i] = false;
    }
    
    /* destroy old strategy */
    if(__strategy) {
      OutOfBandCompositeStrategy* strategy = __strategy;
      __strategy = 0;
      strategy->scheduleDestruction();
    }
    
    /* set new strategy */
    __strategy = strategy;
    
    /* configure new strategy */
    if(__strategy) {
      /* set binding channel and event receiver */
      __strategy->setBindingChannel(this);
      __strategy->setEventReceiver(&__parent);
    }
  }
  
  void OutOfBandChannel::addAddress(const TSAddress& address) {
    /* check for local strategies that match the given address */
    list<TSDescription> local_strategies = STSFactory::getInstance().getSupportedStrategies();
    for(list<TSDescription>::const_iterator i = local_strategies.begin();
        i != local_strategies.end();
        ++i) {
      if(i->getStrategyType() == address.getStrategyType()) {
        /* check which interfaces are supported locally and remotely */
        unsigned int request_interfaces = i->getInterfaces();
        unsigned int reply_interfaces   = address.getStrategyInterfaces();
        
        /* prepare strategy list entry */
        StrategyListEntry entry;
        entry.strategy_type = i->getStrategyType();
        entry.capabilities = i->getCapabilities();
        entry.request_interfaces = 0;
        entry.reply_interfaces = 0;
        entry.used_request_interfaces = 0;
        entry.used_reply_interfaces = 0;
        
        /* check for matching pairs of interfaces */
        if((request_interfaces & TSDescription::ITS_ONE_WAY_EVENT_SENDER) &&
           (reply_interfaces   & TSDescription::ITS_ONE_WAY_EVENT_RECEIVER)) {
          entry.request_interfaces |= TSDescription::ITS_ONE_WAY_EVENT_SENDER;
          entry.reply_interfaces   |= TSDescription::ITS_ONE_WAY_EVENT_RECEIVER;
        }
        if((request_interfaces & TSDescription::ITS_ONE_WAY_EVENT_RECEIVER) &&
           (reply_interfaces   & TSDescription::ITS_ONE_WAY_EVENT_SENDER)) {
          entry.request_interfaces |= TSDescription::ITS_ONE_WAY_EVENT_RECEIVER;
          entry.reply_interfaces   |= TSDescription::ITS_ONE_WAY_EVENT_SENDER;
        }
        if((request_interfaces & TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER) &&
           (reply_interfaces   & TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER)) {
          entry.request_interfaces |= TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER;
          entry.reply_interfaces   |= TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER;
        }
        if((request_interfaces & TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER) &&
           (reply_interfaces   & TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER)) {
          entry.request_interfaces |= TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER;
          entry.reply_interfaces   |= TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER;
        }
        
        /* check if any matching pairs have been found and, if so, register the strategy */
        if(entry.request_interfaces && entry.reply_interfaces) {
          delete entry.default_address;
          entry.default_address = address.copy();
          GLOBAL_DEBUG_STREAM2("Added available strategy " << entry.strategy_type << std::endl);
          __strategy_list.push_back(entry);
        }
      }
    }
  }
  
  const TSAddress* OutOfBandChannel::getAddress(const string& strategy) const {
    for (StrategyConstIterator i = __strategy_list.begin(); i != __strategy_list.end(); ++i) {
      if (i->strategy_type == strategy) {
        return i->default_address;
      }
    }
    
    return 0;
  }
  
  IOutOfBandCompositeStrategy* OutOfBandChannel::getRequestStrategy() const {
    if(!__request_strategy_valid) {
      /* delete old interface */
      if(__request_strategy) {
        delete __request_strategy;
        __request_strategy = 0;
      }
      
      /* a local transport strategy must exist */
      if(!getTransportStrategy()) {
        GLOBAL_ERROR_STREAM("OutOfBandChannel: No transport strategy set" << endl);
      } else {
        /* get interface of local transport strategy */
        __request_strategy = getTransportStrategy()->getInterface<IOutOfBandCompositeStrategy>();
        if(!__request_strategy) {
          /* DOES NOT HAPPEN */
          GLOBAL_ERROR_STREAM("OutOfBandChannel: Can't get IOutOfBandCompositeStrategy interface of transport strategy" << endl);
        }
      }
      
      __request_strategy_valid = true;
    }
    
    return __request_strategy;
  }
  
  IOutOfBandCompositeStrategy* OutOfBandChannel::getReplyStrategy() const {
    if(!__reply_strategy_valid) {
      /* delete old interface */
      if(__reply_strategy) {
        delete __reply_strategy;
        __reply_strategy = 0;
      }
      
      /* a local transport strategy must exist */
      if(!getTransportStrategy()) {
        GLOBAL_ERROR_STREAM("OutOfBandChannel: No transport strategy set" << endl);
      } else {
        /* get partner transport strategy */
        ITransportStrategy* partner_strategy = getTransportStrategy()->getPartnerStrategy();
        if(!partner_strategy) {
          GLOBAL_ERROR_STREAM("OutOfBandChannel: Failed to get partner strategy" << endl);        
        } else {
          __reply_strategy = partner_strategy->getParentObject()->getInterface<IOutOfBandCompositeStrategy>();
          if(!__reply_strategy) {
            GLOBAL_ERROR_STREAM("OutOfBandChannel: Can't get IOutOfBandCompositeStrategy interface of partner transport strategy" << endl);
          }
        }
      }
      
      __reply_strategy_valid = true;
    }
    
    return __reply_strategy;
  }
  
  ITransportStrategy* OutOfBandChannel::getRequestStrategy(CommunicationRole role) const {
    /* get index in interface cache */
    unsigned int index;
    for(index=0; index<4; ++index) {
      if((1 << index) == role) {
        break;
      }
    }
    
    /* check if we need to request this interface */
    if(!__request_child_strategy_valid[index]) {
      IOutOfBandCompositeStrategy* top_level_ts = getRequestStrategy();
      if(!top_level_ts) {
        /* no top-level strategy exists */
        return 0;
      }
      
      /* get child strategy */
      ITransportStrategy* child_strategy = 0;
      switch(role) {
      case REQUEST:
        child_strategy = top_level_ts->getOneWayEventRequestReceiverStrategy();
        if(!child_strategy) {
          child_strategy = top_level_ts->getRequestReplyEventReceiverStrategy();
        }
        break;
      case REPLY:
        child_strategy = top_level_ts->getOneWayEventReplySenderStrategy();
        if(!child_strategy) {
          child_strategy = top_level_ts->getRequestReplyEventReceiverStrategy();
        }
        break;
      case LISTENER_REQUEST:
        child_strategy = top_level_ts->getOneWayEventRequestSenderStrategy();
        if(!child_strategy) {
          child_strategy = top_level_ts->getRequestReplyEventSenderStrategy();
        }
        break;
      case LISTENER_REPLY:
        child_strategy = top_level_ts->getOneWayEventReplyReceiverStrategy();
        if(!child_strategy) {
          child_strategy = top_level_ts->getRequestReplyEventSenderStrategy();
        }
        break;
      default:
        GLOBAL_ERROR_STREAM("Invalid communication role specified for child strategy" << endl);
        return 0;
      }
      
      /* cache interface */
      __request_child_strategy[index] = child_strategy;
      __request_child_strategy_valid[index] = true;
    }
    
    /* return cached interface */
    return __request_child_strategy[index];
  }

  ITransportStrategy* OutOfBandChannel::getReplyStrategy(CommunicationRole role) const {
    /* get index in interface cache */
    unsigned int index;
    for(index=0; index<4; ++index) {
      if((1 << index) == role) {
        break;
      }
    }
    
    /* check if we need to request this interface */
    if(!__reply_child_strategy_valid[index]) {
      IOutOfBandCompositeStrategy* top_level_ts = getReplyStrategy();
      if(!top_level_ts) {
        /* no top-level strategy exists */
        return 0;
      }
      
      /* get child strategy */
      ITransportStrategy* child_strategy = 0;
      switch(role) {
      case REQUEST:
        child_strategy = top_level_ts->getOneWayEventRequestSenderStrategy();
        if(!child_strategy) {
          child_strategy = top_level_ts->getRequestReplyEventSenderStrategy();
        }
        break;
      case REPLY:
        child_strategy = top_level_ts->getOneWayEventReplyReceiverStrategy();
        if(!child_strategy) {
          child_strategy = top_level_ts->getRequestReplyEventSenderStrategy();
        }
        break;
      case LISTENER_REQUEST:
        child_strategy = top_level_ts->getOneWayEventRequestReceiverStrategy();
        if(!child_strategy) {
          child_strategy = top_level_ts->getRequestReplyEventReceiverStrategy();
        }
        break;
      case LISTENER_REPLY:
        child_strategy = top_level_ts->getOneWayEventReplySenderStrategy();
        if(!child_strategy) {
          child_strategy = top_level_ts->getRequestReplyEventReceiverStrategy();
        }
        break;
      default:
        GLOBAL_ERROR_STREAM("Invalid communication role specified for child strategy" << endl);
        return 0;
      }
      
      /* cache interface */
      __reply_child_strategy[index] = child_strategy;
      __reply_child_strategy_valid[index] = true;
    }
    
    /* return cached interface */
    return __reply_child_strategy[index];
  }
  
  list<Interface*>& OutOfBandChannel::getReplyInterfaceList() {
    ITransportStrategy* ts = getReplyStrategy();
    if(!ts) {
      /* return empty list */
      static list<Interface*> empty;
      return empty;
    }
    
    return ts->getParentObject()->getInterfaceList();
  }
  
  list<Interface*>& OutOfBandChannel::getRequestInterfaceList() {
    ITransportStrategy* ts = getRequestStrategy();
    if(!ts) {
      /* return empty list */
      static list<Interface*> empty;
      return empty;
    }
    
    return ts->getParentObject()->getInterfaceList();
  }
  
  list<Interface*>& OutOfBandChannel::getReplyInterfaceList(CommunicationRole role) {
    ITransportStrategy* ts = getReplyStrategy(role);
    if(!ts) {
      /* return empty list */
      static list<Interface*> empty;
      return empty;
    }
    
    return ts->getParentObject()->getInterfaceList();
    
  }
  
  list<Interface*>& OutOfBandChannel::getRequestInterfaceList(CommunicationRole role) {
    ITransportStrategy* ts = getRequestStrategy(role);
    if(!ts) {
      /* return empty list */
      static list<Interface*> empty;
      return empty;
    }
    
    return ts->getParentObject()->getInterfaceList();
  }
  
  void OutOfBandChannel::updateStrategyList(const Interface& iface) {
    __strategy_list.clear();

    /*Begin> specify, wich binding type we can use. Use the method from base class
     *       to check the object position
     */
    TSAddress::BindingType binding_type;
    if (getParent().isLocalTo(iface)) {
      binding_type = TSAddress::LOCAL;
    }

    if (getParent().isSystemTo(iface)){
      binding_type = TSAddress::SYSTEM;
    }

    if (getParent().isNetworkTo(iface)){
      binding_type = TSAddress::NETWORK;
    }
    /* End< */

    /* Begin> Set the TransportStrategies we can use */
    switch ( binding_type ) {
    case TSAddress::LOCAL :
      /* Begin> Set the TransportStrategies we can use */
      for (unsigned int i = 0;i< iface.getAddressList().size(); ++i)
        if(iface.getAddressList()[i]->isBindingType(TSAddress::LOCAL)) {
          addAddress(*iface.getAddressList()[i]);
        }
      /* End< */
      /* Don't call break because System and Network strategies can be used too */
      
    case TSAddress::SYSTEM :
      /* Begin> Set the TransportStrategies we can use */
      for (unsigned int i = 0;i< iface.getAddressList().size(); ++i)
        if(iface.getAddressList()[i]->isBindingType(TSAddress::SYSTEM)) {
          addAddress(*iface.getAddressList()[i]);
        }
      /* End< */
      /* Don't call break because network strategies can be used too */
      
    case TSAddress::NETWORK :
      /* Begin> Set the TransportStrategies we can use */
      for (unsigned int i = 0;i< iface.getAddressList().size(); ++i)
        if (   iface.getAddressList()[i]->isBindingType(TSAddress::NETWORK)) {
          addAddress(*iface.getAddressList()[i]);
        }
      /* End< */
    }
    /* End< */
    
    /* make strategy list unique */
    __strategy_list.unique();
    
    /* move default strategy to front of list */
    string default_strategy = SConnectionManager::getInstance().getDefaultStrategy(getTypeInfo(), binding_type);
    if(!default_strategy.empty()) {
//       cerr << "Supported Strategies> " << endl;
      for(StrategyIterator i(__strategy_list.begin()); i != __strategy_list.end();) {
        
        /* BEGIN> Output strategyt config (debug) */
//         cerr << "  " << i->strategy_type << ">" << endl;
//         cerr << "    Request Interfaces>" << endl;
//         if(i->request_interfaces & TSDescription::ITS_ONE_WAY_EVENT_RECEIVER) {
//           cerr << "      ITS_ONE_WAY_EVENT_RECEIVER" << endl;
//         }
//         if(i->request_interfaces & TSDescription::ITS_ONE_WAY_EVENT_SENDER) {
//           cerr << "      ITS_ONE_WAY_EVENT_SENDER" << endl;
//         }
//         if(i->request_interfaces & TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER) {
//           cerr << "      ITS_REQUEST_REPLY_EVENT_RECEIVER" << endl;
//         }
//         if(i->request_interfaces & TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER) {
//           cerr << "      ITS_REQUEST_REPLY_EVENT_SENDER" << endl;
//         }
//         cerr << "    Reply Interfaces>" << endl;
//         if(i->reply_interfaces & TSDescription::ITS_ONE_WAY_EVENT_RECEIVER) {
//           cerr << "      ITS_ONE_WAY_EVENT_RECEIVER" << endl;
//         }
//         if(i->reply_interfaces & TSDescription::ITS_ONE_WAY_EVENT_SENDER) {
//           cerr << "      ITS_ONE_WAY_EVENT_SENDER" << endl;
//         }
//         if(i->reply_interfaces & TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER) {
//           cerr << "      ITS_REQUEST_REPLY_EVENT_RECEIVER" << endl;
//         }
//         if(i->reply_interfaces & TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER) {
//           cerr << "      ITS_REQUEST_REPLY_EVENT_SENDER" << endl;
//         }
        /* END< */
        
        if(i->strategy_type == default_strategy) {
          __strategy_list.push_front(*i);
          __strategy_list.erase(i++);
        } else {
          ++i;
        }
      }
    } else {
      GLOBAL_WARNING_STREAM2("No default strategy defined for " << getTypeName() << " and binding type " <<
                             ((binding_type == TSAddress::LOCAL) ? "LOCAL" :
                              (binding_type == TSAddress::SYSTEM) ? "SYSTEM" :
                              "NETWORK") << endl);
    }
  }
  
  bool OutOfBandChannel::tryReuseStrategy(unsigned int request_interfaces, unsigned int reply_interfaces) {
    for(unsigned int i=0; i<__strategies.size(); ++i) {
      if(!__strategies[i].is_explicit &&
         (__strategies[i].request_interfaces & request_interfaces) == request_interfaces &&
         (__strategies[i].reply_interfaces & reply_interfaces) == reply_interfaces) {
        __strategies[i].used_request_interfaces |= request_interfaces;
        __strategies[i].used_reply_interfaces |= reply_interfaces;
        GLOBAL_DEBUG_STREAM2("OutOfBandChannel: Automatically reused strategy " << __strategies[i].strategy_type << std::endl);
        return true;
      }
    }
    
    return false;
  }
  
  bool OutOfBandChannel::findSuitableStrategy(unsigned int request_interfaces, unsigned int reply_interfaces) {
    for(StrategyConstIterator i=__strategy_list.begin(); i != __strategy_list.end(); ++i) {
      if((i->request_interfaces & request_interfaces) == request_interfaces &&
         (i->reply_interfaces & reply_interfaces) == reply_interfaces) {
        StrategyListEntry entry(*i);
        entry.used_request_interfaces = request_interfaces;
        entry.used_reply_interfaces = reply_interfaces;
        entry.default_address = entry.default_address->copy();
        entry.default_address->setStrategyInterfaces(entry.used_request_interfaces);
        __strategies.push_back(entry);
        GLOBAL_DEBUG_STREAM2("OutOfBandChannel: Automatically added suitable strategy " << i->strategy_type << std::endl);
        return true;
      }
    }
    return false;
  }

  void OutOfBandChannel::setValueStream(unsigned int roles, const string& i_value_stream_type, const string& o_value_stream_type) {
    // XXX: fw: if one or more strategies are the same, don't set the communication channel multiple times
    
    if(roles & REQUEST) {
      setValueStreamInternal(REQUEST, i_value_stream_type, o_value_stream_type);
    }
    if(roles & REPLY) {
      setValueStreamInternal(REPLY, i_value_stream_type, o_value_stream_type);
    }
    if(roles & LISTENER_REQUEST) {
      setValueStreamInternal(LISTENER_REQUEST, i_value_stream_type, o_value_stream_type);
    }
    if(roles & LISTENER_REPLY) {
      setValueStreamInternal(LISTENER_REPLY, i_value_stream_type, o_value_stream_type);
    }
  }
  
  void OutOfBandChannel::setValueStreamStack(unsigned int roles, const IValueStream* i_value_stream, const OValueStream* o_value_stream) {
    // XXX: fw: if one or more strategies are the same, don't set the communication channel multiple times
    
    if(roles & REQUEST) {
      setValueStreamStackInternal(REQUEST, i_value_stream, o_value_stream);
    }
    if(roles & REPLY) {
      setValueStreamStackInternal(REPLY, i_value_stream, o_value_stream);
    }
    if(roles & LISTENER_REQUEST) {
      setValueStreamStackInternal(LISTENER_REQUEST, i_value_stream, o_value_stream);
    }
    if(roles & LISTENER_REPLY) {
      setValueStreamStackInternal(LISTENER_REPLY, i_value_stream, o_value_stream);
    }
  }
  
  void OutOfBandChannel::setValueStreamInternal(CommunicationRole role,
                                               const string& i_value_stream_type, const string& o_value_stream_type)
  {
    /* get interfaces for setting the value stream */
    ISetValueStream_var request_value_stream(getRequestInterface<ISetValueStream>(role));
    ISetValueStream_var reply_value_stream(getReplyInterface<ISetValueStream>(role));
    if(!request_value_stream.get() || !reply_value_stream.get()) {
      // XXX: fw: more detailed debug output, as in InstreamChannel
      //GLOBAL_ERROR_STREAM("Can't set value stream for role " << getRoleName(role) << endl);
      GLOBAL_ERROR_STREAM("Can't set value stream" << endl);
      return;
    }
    
    /* set the value stream */
    if(role == REQUEST || role == LISTENER_REPLY) {
      request_value_stream->setOValueStream(o_value_stream_type);
      reply_value_stream->setIValueStream(i_value_stream_type);
    } else {
      request_value_stream->setIValueStream(i_value_stream_type);
      reply_value_stream->setOValueStream(o_value_stream_type);
    }
  }
  
  TypeInfo OutOfBandChannel::getStaticTypeInfo() {
    return TypeInfo("OutOfBandChannel", "NMM");
  }
  
  void OutOfBandChannel::setValueStreamStackInternal(CommunicationRole role,
                                                     const IValueStream* i_value_stream, const OValueStream* o_value_stream)
  {
    /* get interfaces for setting the value stream */
    ISetValueStream_var request_value_stream(getRequestInterface<ISetValueStream>(role));
    ISetValueStream_var reply_value_stream(getReplyInterface<ISetValueStream>(role));
    if(!request_value_stream.get() || !reply_value_stream.get()) {
      // XXX: fw: more detailed debug output, as in InstreamChannel
      //GLOBAL_ERROR_STREAM("Can't set value stream for role " << getRoleName(role) << endl);
      GLOBAL_ERROR_STREAM("Can't set value stream" << endl);
      return;
    }
    
    /* set the value stream */
    if(role == REQUEST || role == LISTENER_REPLY) {
      request_value_stream->setOValueStreamStack(o_value_stream);
      reply_value_stream->setIValueStreamStack(i_value_stream);
    } else {
      request_value_stream->setIValueStreamStack(i_value_stream);
      reply_value_stream->setOValueStreamStack(o_value_stream);
    }
  }

}

#include "nmm/nmm_suffix.hpp"
