/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 * 
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "OutOfBandCompositeStrategy.hpp"
#include "OutOfBandCompositeAddress.hpp"

#include "TSFactory.hpp"
#include "IObjectAcceptor.hpp"
#include "TransportStrategy.hpp"

#include "nmm/comm/messaging/ITSOneWayEventReceiver.hpp"
#include "nmm/comm/messaging/ITSOneWayEventSender.hpp"
#include "nmm/comm/messaging/ITSRequestReplyEventReceiver.hpp"
#include "nmm/comm/messaging/ITSRequestReplyEventSender.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  OutOfBandCompositeStrategy::OutOfBandCompositeStrategy(const string& name)
    : CompositeStrategy(name),
      __one_way_event_request_receiver(0),
      __one_way_event_reply_receiver(0),
      __one_way_event_request_sender(0),
      __one_way_event_reply_sender(0),
      __request_reply_event_receiver(0),
      __request_reply_event_sender(0)
  {}
  
  OutOfBandCompositeStrategy::~OutOfBandCompositeStrategy() {
    /* BEGIN> unset all roles */
    setOneWayEventRequestReceiverStrategy(0);
    setOneWayEventRequestSenderStrategy(0);
    setOneWayEventReplyReceiverStrategy(0);
    setOneWayEventReplySenderStrategy(0);
    setRequestReplyEventReceiverStrategy(0);
    setRequestReplyEventSenderStrategy(0);
    /* END< */
    
    /* remove all child strategies (must be done here to avoid pure virtual method calls...) */
    removeAllChildStrategies();
  }
  
  string OutOfBandCompositeStrategy::getStrategyType() const {
    return "OutOfBandCompositeStrategy";
  }
  
  bool OutOfBandCompositeStrategy::isBindingComplete() const throw() {
    /* must have either a receiver for request-reply events or a receiver for
       one-way event requests and a sender for one-way event replies.
       (to handle event requests) */
    if((!__one_way_event_request_receiver ||
	!__one_way_event_reply_sender) &&
       !__request_reply_event_receiver) {
      return false;
    }

    /* must have either a receiver for one-way event replies or a sender for
       request-reply events and a sender for one-way event requests.
       (to handle event replies) */
    if((!__one_way_event_request_sender ||
	!__one_way_event_reply_receiver) &&
       !__request_reply_event_sender) {
      return false;
    }
    
    /* binding is complete */
    return true;
  }
  
  void OutOfBandCompositeStrategy::setAddress(const TSAddress& address)
    throw(const BindingException&)
  {
    /* address must be an OutOfBandCompositeAddress */
    const OutOfBandCompositeAddress* composite_address = dynamic_cast<const OutOfBandCompositeAddress*>(&address);
    if(!composite_address) {
      throw BindingException("Address is not an OutOfBandCompositeAddress");
    }
    
    /* BEGIN> Create child strategies */
    TInOutValue<NMMReference> composite_reference(0);
    for(unsigned int i=0; i<composite_address->getNumberOfAddresses(); ++i) {
      int index = i;
      const TSAddress* address = composite_address->getAddress(i);
      
      /* invert interfaces */
      unsigned int partner_interfaces = TSDescription::getPartnerInterfaces(address->getStrategyInterfaces());
      
      /* create strategy */
      TransportStrategy* strategy = STSFactory::getInstance().createTransportStrategy(address->getStrategyType(), partner_interfaces);
      
      /* BEGIN> Set transport roles */
      /* determine which roles the strategy has on the remote side. this is defined directly by the TSAddress.
         also add the strategy locally for the inverse roles. */
      unsigned int roles = 0;
      if(index == composite_address->getOneWayEventRequestReceiverAddress()) {
        setOneWayEventRequestSenderStrategy(strategy);
        roles |= IOutOfBandCompositeStrategy::RECEIVE_ONE_WAY_EVENT_REQUEST;
      }
      if(index == composite_address->getOneWayEventRequestSenderAddress()) {
        setOneWayEventRequestReceiverStrategy(strategy);
        roles |= IOutOfBandCompositeStrategy::SEND_ONE_WAY_EVENT_REQUEST;
      }
      if(index == composite_address->getOneWayEventReplyReceiverAddress()) {
        setOneWayEventReplySenderStrategy(strategy);
        roles |= IOutOfBandCompositeStrategy::RECEIVE_ONE_WAY_EVENT_REPLY;
      }
      if(index == composite_address->getOneWayEventReplySenderAddress()) {
        setOneWayEventReplyReceiverStrategy(strategy);
        roles |= IOutOfBandCompositeStrategy::SEND_ONE_WAY_EVENT_REPLY;
      }
      if(index == composite_address->getRequestReplyEventReceiverAddress()) {
        setRequestReplyEventSenderStrategy(strategy);
        roles |= IOutOfBandCompositeStrategy::RECEIVE_REQUEST_REPLY_EVENT;
      }
      if(index == composite_address->getRequestReplyEventSenderAddress()) {
        setRequestReplyEventReceiverStrategy(strategy);
        roles |= IOutOfBandCompositeStrategy::SEND_REQUEST_REPLY_EVENT;
      }
      /* END< */
      
      /* connect TransportStrategy */
      strategy->setAddress(*address);
      
      /* add TransportStrategy to this composite */
      addChildStrategy(strategy);
      
      /* connect TransportStrategy to partner strategy.
         for the first child strategy, this will also create the partner strategy. */
      Event connect_event(IObjectAcceptor::connectToOutOfBandCompositeStrategy_event, &composite_reference, new TInValue<nmm_uint32>(roles));
      if(FAILURE == strategy->sendReliableEvent(connect_event)) {
        throw BindingException("Failed to send connectToOutOfBandCompositeStrategy event");
      }
      if(connect_event.getException()) {
        if(dynamic_cast<const BindingException*>(connect_event.getException())) {
          throw *connect_event.getException();
        } else {
          throw BindingException(connect_event.getException()->getComment());
        }
      }
    }
    /* END< */
    
    /* BEGIN> set remote connection manager interface and reference */
    // XXX: fw: this is a bad idea (chicken & egg)
//     const TSAddress& any_address = composite_address->getAddress(0);
//     Interface* tmp_interface = SConnectionManager::getInstance().getService(any_address, "ConnectionManager");
//     IConnectionManager* icm = tmp_interface->getParentObject()->getInterface<IConnectionManager>();
//     if(!icm) {
//       GLOBAL_ERROR_STREAM1("OutOfBandCompositeStrategy> Found ConnectionManager service, but no interface to it" << endl);
//       throw BindingException("Found ConnectionManager service, but no interface to it");
//     }
//     setConnectionManager(icm, composite_reference.getRValue());
    /* END< */
  }
  
  ITransportStrategy* OutOfBandCompositeStrategy::getOneWayEventRequestReceiverStrategy() const {
    TransportStrategy* strategy = getOwnOneWayEventRequestReceiverStrategy();
    if(strategy) {
      return strategy->getInterface<ITransportStrategy>();
    } else {
      return 0;
    }
  }
  
  ITransportStrategy* OutOfBandCompositeStrategy::getOneWayEventReplyReceiverStrategy() const {
    TransportStrategy* strategy = getOwnOneWayEventReplyReceiverStrategy();
    if(strategy) {
      return strategy->getInterface<ITransportStrategy>();
    } else {
      return 0;
    }
  }
  
  ITransportStrategy* OutOfBandCompositeStrategy::getOneWayEventRequestSenderStrategy() const {
    TransportStrategy* strategy = getOwnOneWayEventRequestSenderStrategy();
    if(strategy) {
      return strategy->getInterface<ITransportStrategy>();
    } else {
      return 0;
    }
  }
  
  ITransportStrategy* OutOfBandCompositeStrategy::getOneWayEventReplySenderStrategy() const {
    TransportStrategy* strategy = getOwnOneWayEventReplySenderStrategy();
    if(strategy) {
      return strategy->getInterface<ITransportStrategy>();
    } else {
      return 0;
    }
  }
  
  ITransportStrategy* OutOfBandCompositeStrategy::getRequestReplyEventReceiverStrategy() const {
    TransportStrategy* strategy = getOwnRequestReplyEventReceiverStrategy();
    if(strategy) {
      return strategy->getInterface<ITransportStrategy>();
    } else {
      return 0;
    }
  }
  
  ITransportStrategy* OutOfBandCompositeStrategy::getRequestReplyEventSenderStrategy() const {
    TransportStrategy* strategy = getOwnRequestReplyEventSenderStrategy();
    if(strategy) {
      return strategy->getInterface<ITransportStrategy>();
    } else {
      return 0;
    }
  }
  
  TransportStrategy* OutOfBandCompositeStrategy::getOwnOneWayEventRequestReceiverStrategy() const {
    if(!getOneWayEventRequestReceiver()) {
      return 0;
    }
    
    TransportStrategy* strategy = dynamic_cast<TransportStrategy*>(getOneWayEventRequestReceiver());
    if(!strategy) {
      return 0;
    }
    
    return strategy;
  }
  
  TransportStrategy* OutOfBandCompositeStrategy::getOwnOneWayEventReplyReceiverStrategy() const {
    if(!getOneWayEventReplyReceiver()) {
      return 0;
    }
    
    TransportStrategy* strategy = dynamic_cast<TransportStrategy*>(getOneWayEventReplyReceiver());
    if(!strategy) {
      return 0;
    }
    
    return strategy;
  }
  
  TransportStrategy* OutOfBandCompositeStrategy::getOwnOneWayEventRequestSenderStrategy() const {
    if(!getOneWayEventRequestSender()) {
      return 0;
    }
    
    TransportStrategy* strategy = dynamic_cast<TransportStrategy*>(getOneWayEventRequestSender());
    if(!strategy) {
      return 0;
    }
    
    return strategy;
  }
  
  TransportStrategy* OutOfBandCompositeStrategy::getOwnOneWayEventReplySenderStrategy() const {
    if(!getOneWayEventReplySender()) {
      return 0;
    }
    
    TransportStrategy* strategy = dynamic_cast<TransportStrategy*>(getOneWayEventReplySender());
    if(!strategy) {
      return 0;
    }
    
    return strategy;
  }
  
  TransportStrategy* OutOfBandCompositeStrategy::getOwnRequestReplyEventReceiverStrategy() const {
    if(!getRequestReplyEventReceiver()) {
      return 0;
    }
    
    TransportStrategy* strategy = dynamic_cast<TransportStrategy*>(getRequestReplyEventReceiver());
    if(!strategy) {
      return 0;
    }
    
    return strategy;
  }
  
  TransportStrategy* OutOfBandCompositeStrategy::getOwnRequestReplyEventSenderStrategy() const {
    if(!getRequestReplyEventSender()) {
      return 0;
    }
    
    TransportStrategy* strategy = dynamic_cast<TransportStrategy*>(getRequestReplyEventSender());
    if(!strategy) {
      return 0;
    }
    
    return strategy;
  }


  
  ITSOneWayEventSender* OutOfBandCompositeStrategy::getOneWayEventRequestSender() const {
    return __one_way_event_request_sender;
  }
  
  ITSOneWayEventReceiver* OutOfBandCompositeStrategy::getOneWayEventReplyReceiver() const {
    return __one_way_event_reply_receiver;
  }
  
  ITSOneWayEventReceiver* OutOfBandCompositeStrategy::getOneWayEventRequestReceiver() const {
    return __one_way_event_request_receiver;
  }
  
  ITSOneWayEventSender* OutOfBandCompositeStrategy::getOneWayEventReplySender() const {
    return __one_way_event_reply_sender;
  }
  
  ITSRequestReplyEventSender* OutOfBandCompositeStrategy::getRequestReplyEventSender() const {
    return __request_reply_event_sender;
  }
  
  ITSRequestReplyEventReceiver* OutOfBandCompositeStrategy::getRequestReplyEventReceiver() const {
    return __request_reply_event_receiver;
  }
  
  void OutOfBandCompositeStrategy::setOneWayEventRequestReceiverStrategy(TransportStrategy* strategy)
    throw(const BindingException&)
  {
    if(strategy) {
      __one_way_event_request_receiver = dynamic_cast<ITSOneWayEventReceiver*>(strategy);
      if(!__one_way_event_request_receiver) {
        /* SHOULD NOT HAPPEN */
        throw BindingException("TransportStrategy does not support the required interfaces!");
      }
    } else {
      __one_way_event_request_receiver = 0;
    }
  }
  
  void OutOfBandCompositeStrategy::setOneWayEventRequestSenderStrategy(TransportStrategy* strategy)
    throw(const BindingException&)
  {
    if(strategy) {
      __one_way_event_request_sender = dynamic_cast<ITSOneWayEventSender*>(strategy);
      if(!__one_way_event_request_sender) {
        /* SHOULD NOT HAPPEN */
        throw BindingException("TransportStrategy does not support the required interfaces!");
      }
      
      // XXX: fw: connect to out-of-band multiplexer
      //__one_way_event_request_sender->setOneWayEventReceiver(...);
    } else {
      if(__one_way_event_request_sender) {
        /* disconnect from out-of-band multiplexer */
        __one_way_event_request_sender->setOneWayEventReceiver(0);
      }
      
      __one_way_event_request_sender = 0;      
    }
  }
  
  void OutOfBandCompositeStrategy::setOneWayEventReplyReceiverStrategy(TransportStrategy* strategy)
    throw(const BindingException&)
  {
    if(strategy) {
      __one_way_event_reply_receiver = dynamic_cast<ITSOneWayEventReceiver*>(strategy);
      if(!__one_way_event_reply_receiver) {
        /* SHOULD NOT HAPPEN */
        throw BindingException("TransportStrategy does not support the required interfaces!");
      }
    } else {
      __one_way_event_reply_receiver = 0;
    }
  }
  
  void OutOfBandCompositeStrategy::setOneWayEventReplySenderStrategy(TransportStrategy* strategy) 
    throw(const BindingException&)
  {
    if(strategy) {
      __one_way_event_reply_sender = dynamic_cast<ITSOneWayEventSender*>(strategy);
      if(!__one_way_event_reply_sender) {
        /* SHOULD NOT HAPPEN */
        throw BindingException("TransportStrategy does not support the required interfaces!");
      }
      
      // XXX: fw: connect to out-of-band multiplexer
      //__one_way_event_reply_sender->setOneWayEventReceiver(...);
    } else {
      if(__one_way_event_reply_sender) {
        /* disconnect from out-of-band multiplexer */
        __one_way_event_reply_sender->setOneWayEventReceiver(0);
      }
      
      __one_way_event_reply_sender = 0;
    }
  }
  
  void OutOfBandCompositeStrategy::setRequestReplyEventReceiverStrategy(TransportStrategy* strategy)
    throw(const BindingException&)
  {
    if(strategy) {
      __request_reply_event_receiver = dynamic_cast<ITSRequestReplyEventReceiver*>(strategy);
      if(!__request_reply_event_receiver) {
        /* SHOULD NOT HAPPEN */
        throw BindingException("TransportStrategy does not support the required interfaces!");
      }
      
    } else {
      __request_reply_event_receiver = 0;
    }
  }
  
  void OutOfBandCompositeStrategy::setRequestReplyEventSenderStrategy(TransportStrategy* strategy)
    throw(const BindingException&)
  {
    if(strategy) {
      __request_reply_event_sender = dynamic_cast<ITSRequestReplyEventSender*>(strategy);
      if(!__request_reply_event_sender) {
        /* SHOULD NOT HAPPEN */
        throw BindingException("TransportStrategy does not support the required interfaces!");
      }
      
      __request_reply_event_sender->setRequestReplyEventReceiver(this);
    } else {
      __request_reply_event_sender = 0;
    }
  }
  
  void OutOfBandCompositeStrategy::addTransportStrategy(TransportStrategy* strategy, unsigned int roles)
    throw(const BindingException&)
  {
    /* add child strategy */
    addChildStrategy(strategy);
    
    // XXX: fw: TransportStrategy is not a NamedObject
//     GLOBAL_DEBUG_STREAM1("OutOfBandCompositeStrategy> Added incoming child strategy: " << strategy->getName() << endl);
    GLOBAL_DEBUG_STREAM1("OutOfBandCompositeStrategy> Added incoming child strategy: " << strategy->getStrategyType() << endl);
    
    /* BEGIN> set roles */
    if(roles & IOutOfBandCompositeStrategy::RECEIVE_ONE_WAY_EVENT_REQUEST) {
      setOneWayEventRequestReceiverStrategy(strategy);
    }
    if(roles & IOutOfBandCompositeStrategy::SEND_ONE_WAY_EVENT_REQUEST) {
      setOneWayEventRequestSenderStrategy(strategy);
    }
    if(roles & IOutOfBandCompositeStrategy::RECEIVE_ONE_WAY_EVENT_REPLY) {
      setOneWayEventReplyReceiverStrategy(strategy);
    }
    if(roles & IOutOfBandCompositeStrategy::SEND_ONE_WAY_EVENT_REPLY) {
      setOneWayEventReplySenderStrategy(strategy);
    }
    if(roles & IOutOfBandCompositeStrategy::RECEIVE_REQUEST_REPLY_EVENT) {
      setRequestReplyEventReceiverStrategy(strategy);
    }
    if(roles & IOutOfBandCompositeStrategy::SEND_REQUEST_REPLY_EVENT) {
      setRequestReplyEventSenderStrategy(strategy);
    }
    /* END< */
  }
  
  Result OutOfBandCompositeStrategy::receiveCSEvent(Event& event) {
    if(__request_reply_event_receiver) {
      /* use request-reply leaf strategy if available */
      return __request_reply_event_receiver->receiveRequestReplyEvent(event);
    } else if(__one_way_event_request_receiver) {
      // XXX: TODO: assign a request ID to this event
      // XXX: TODO: pass the event to the one_way_event_request_receiver
      // XXX: TODO: block here until a reply has been received from the one_way_event_reply_sender
      return FAILURE;
    } else {
      GLOBAL_ERROR_STREAM1("OutOfBandCompositeStrategy> No child strategy set that can receive event requests" << endl);
      return FAILURE;
    }
  }
  
  Result OutOfBandCompositeStrategy::receiveRequestReplyEvent(Event& event) {
    /* forward to high-level event receiver */
    if(getEventReceiver()) {
      return getEventReceiver()->receiveEvent(event, this);
    } else {
      GLOBAL_ERROR_STREAM1("OutOfBandCompositeStrategy> No event receiver");
      return FAILURE;
    }
  }
  
  Result OutOfBandCompositeStrategy::sendReliableEvent(Event& event) const {
    // XXX: fw: reliable events
    // XXX: fw: we must check if the strategy is reliable (temporary hack)
    return const_cast<OutOfBandCompositeStrategy*>(this)->receiveCSEvent(event);
  }
  
  Result OutOfBandCompositeStrategy::sendReliableConfigEvent(Event& event) const {
    // XXX: fw: reliable events
    // XXX: fw: we must check if the strategy is reliable (temporary hack)

    if(__request_reply_event_receiver) {
      /* use request-reply leaf strategy if available */
      return getOwnRequestReplyEventReceiverStrategy()->sendReliableEvent(event);
    } else if(__one_way_event_request_receiver) {
      GLOBAL_ERROR_STREAM1("OutOfBandCompositeStrategy> Failed to send reliable config event" << endl);
      return FAILURE;
    } else {
      GLOBAL_ERROR_STREAM1("OutOfBandCompositeStrategy> No child strategy set that can receive event requests" << endl);
      return FAILURE;
    }
  }
  
};

#include "nmm/nmm_suffix.hpp"
