/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "ProxyObject.hpp"
#include "Interface.hpp"
#include "OutOfBandChannel.hpp"
#include "OutOfBandCompositeStrategy.hpp"

#include "IObjectAcceptor.hpp"

#include "TSAddress.hpp"
#include "ReferenceManager.hpp"

#include "nmm/comm/messaging/TEDObject.hpp"

#include "nmm/comm/serialize/StdValue.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  ProxyObject::ProxyObject(bool auto_delete)
    : NamedObject("Proxy"),
      __out_of_band_channel(0),
      __parent_reference(0),
      __is_connected(false),
      __interface_counter(0),
      __auto_delete(auto_delete),
      __local_optimisation_allowed(false),
      __is_local(false),
      __partner_object(0)
  {
    /* Register the Events */
    registerEvents();
  }

  ProxyObject::ProxyObject(const Interface& iface, bool auto_delete)
    : NamedObject("Proxy"),
      __out_of_band_channel(0),
      __parent_reference(0),
      __is_connected(false),
      __interface_counter(0),
      __auto_delete(auto_delete),
      __local_optimisation_allowed(false),
      __is_local(false),
      __partner_object(0)
  {
    /* Register the Events */
    registerEvents();
  
    /* Check if an address is available */
    __parent_reference = iface.getParentReference();
  
    /* Check if partner object is local */
    if(isLocalOptimisationAllowed() && isLocalTo(iface) && __parent_reference != 0) {
      __is_local = true;
      __partner_object = SReferenceManager::getInstance().getNMMObject(__parent_reference);
    }
  
    /* Create an OutOfBandChannel for the given interface */
    __out_of_band_channel = new OutOfBandChannel(*this, iface);  
  }


  ProxyObject::~ProxyObject() {
    /* Begin> Unregister the ProxyObejct by the ReferenceManager */
    if (NMMObject::getInterfaceList().empty())
      SReferenceManager::getInstance().unregisterProxyObject( *this, UserInfo() );
    else
      SReferenceManager::getInstance().unregisterProxyObject( *this, NMMObject::getInterfaceList().front()->getParentInfo() );
    /* End< */

    /* If the ProxyObject is still connected disconnect it */
    if (__is_connected) {
      disconnectFromObject();
    }
  
    /* Reset the parante pointer in our interfaces, so they dont dereference the internal interface counter */
    for (list<Interface*>::iterator it = NMMObject::getInterfaceList().begin(); it != NMMObject::getInterfaceList().end(); ++it)
      (*it)->setParentObject(0);
  }

  void ProxyObject::registerEvents() {
    getEventDispatcher().registerEvent("ProxyObject::updateInterfaces", new TEDObject0<ProxyObject, void (ProxyObject::*)( ) , void >(this, &ProxyObject::updateInterfaces));
  
    /* register methods to be informed if the connection to an object changes */
    NMMObject::registerEventListener(IObjectAcceptor::connectToObject_event, new TEDObject1<ProxyObject, NMMReference, void (ProxyObject::*)(const NMMReference&), void>(this, &ProxyObject::connectToObject));
    NMMObject::registerEventListener(IObjectAcceptor::connectToService_event, new TEDObject1<ProxyObject, string, void (ProxyObject::*)(const string&), void>(this, &ProxyObject::connectToService));
  }


  const list<Interface*>& ProxyObject::getInterfaceList()   const {
    /* If the ProxyObject is not connected, connect it */
    if (!__is_connected)
      const_cast<ProxyObject*>(this)->sendConnectToObject();

    /* return the interface list of the base class */
    return NMMObject::getInterfaceList();
  }

  list<Interface*>& ProxyObject::getInterfaceList() {
    /* If the ProxyObject is not connected, connect it */
    if (!__is_connected)
      sendConnectToObject();

    /* return the interface list of the base class */
    return NMMObject::getInterfaceList();
  }       

  Result ProxyObject::receiveEvent(Event& event, const IEventSender* sender)  {
    /* If not connected, connect with NMMObject */
    if (!__is_connected)
      sendConnectToObject();

    /* If the TransportStrategy or the partner object is the sender we dispatch the Event only*/
    // XXX: fw: we need to check whether the event comes from the partner object
    if (/*(__partner_object && sender == __partner_object) ||*/
        getOutOfBandChannel().getTransportStrategy() && sender == getOutOfBandChannel().getTransportStrategy()) {
      return NMMObject::receiveEvent(event, sender);
    }
  
    /* Local optimisation */
    if(__is_local) {
      return __partner_object->receiveEvent(event, sender);
    }
  
    /* Otherwise we send the Event over the TransportStrategy */
    if (getOutOfBandChannel().getTransportStrategy() ) {
      Result result = getOutOfBandChannel().getTransportStrategy()->receiveCSEvent(event);
      /* forward the event to all listener */
      getEventDispatcher().forwardEvent(event, sender);
      /* forward the event to all listener */
      getEventDispatcher().externalDispatch(event, sender);
      return result;
    }

    return FAILURE;
  }

  Result ProxyObject::receiveReliableEvent(Event& event, const IEventSender* sender) const {
    /* If not connected, connect with NMMObject */
    if (!__is_connected) {
      try {
        const_cast<ProxyObject*>(this)->sendConnectToObject();
      } catch(const BindingException&) {
        return FAILURE;
      }
    }

    /* If the TransportStrategy is the sender we dispatch the Event only */
    // XXX: fw: if the message comes from the partner object, dispatch it!!
    if (/*(__partner_object && sender == __partner_object) ||*/
        getOutOfBandChannel().getTransportStrategy() && sender == getOutOfBandChannel().getTransportStrategy()) {
      return const_cast<ProxyObject&>(*this).NMMObject::receiveEvent(event, sender);
    }
  
    /* Local optimisation */
    if(__is_local) {
      return __partner_object->receiveEvent(event, sender);
    }
  
    /* Otherwise we send the Event over the TransportStrategy */
    if (getOutOfBandChannel().getTransportStrategy() ) {
      Result result = getOutOfBandChannel().getTransportStrategy()->sendReliableEvent(event);
      /* forward the event to all listener */
      getEventDispatcher().forwardEvent(event, sender);
      /* forward the event to all listener */
      getEventDispatcher().externalDispatch(event, sender);
      return result;
    }
  
    return FAILURE;
  }


  void ProxyObject::connectToObject(const TSAddress& address) 
    throw(const BindingException&)
  {
    /* first disconnect from object */
    disconnectFromObject();
  
    /* copy the address to connect to */
    __out_of_band_channel = new OutOfBandChannel(*this, address);
  
    /* store the parent reference */
    __parent_reference  = address.getNMMReference();
  
    if(isLocalOptimisationAllowed()) {
      /* check if partner object is local. we can not tell this just from the address.
         so we have to be conservative and assume that the object is remote if the address suggests that it may be remote.
         This is the case if the address supports binding types other than LOCAL. For example if a TCPAddress is being used,
         we have to assume that the object is remote. Even if the hostname matches the local hostname, the object may belong to a different
         process, and since the only other information that we have is the port, we can't tell the difference. However, if a LocalAddress
         is used, we know for sure that the object is local, so we can allow the local optimisation. */
      if(address.getBindingType() == TSAddress::LOCAL) {
        __is_local = true;    
        if(__parent_reference != 0) {
          __partner_object = SReferenceManager::getInstance().getNMMObject(__parent_reference);
        }
      }
    }
  
    /* set the internal connection flag to false */
    __is_connected      = false;
  }

  void ProxyObject::disconnectFromObject() {
    /* delete the old TransportStrategy */
    if(__out_of_band_channel) {
      if(__is_connected) {
        /* unset TransportStrategy */
        getOutOfBandChannel().unsetTransportStrategy();
      
        /* we are not connected */
        __is_connected      = false;
      }
    
      /* delete channel */
      delete __out_of_band_channel;
      __out_of_band_channel = 0;
    }
  
    /* Unregister the ProxyObject by the ReferenceManager */
    if (NMMObject::getInterfaceList().empty()) {
      SReferenceManager::getInstance().unregisterProxyObject( *this, UserInfo() );
    } else {
      SReferenceManager::getInstance().unregisterProxyObject( *this, NMMObject::getInterfaceList().front()->getParentInfo() );
    }
  
    /* set the parent reference to  0 */
    __parent_reference  = 0;
  }


  void ProxyObject::establishBinding()
    throw (const BindingException&)
  {
    if(!__is_connected) {
      sendConnectToObject();
    }
  }

  void ProxyObject::sendConnectToObject()
    throw(const BindingException&)
  {
    try {
      /* BEGIN> Check if we have to disable local optimisation because we have no partner object
         (This shouldn't normally happen anymore) */
      if(__is_local && !__partner_object) {
        __is_local = false;
        WARNING_STREAM("No partner object available! Disabling local optimisation!" << endl);
      }
      /* END< */
    
      /* BEGIN> Establish binding */
      if(__is_local) {
        /* local optimisation works, we are done. */
        __is_connected = true;
      } else {
        /* connect the out-of-band channel */
        getOutOfBandChannel().establishBinding();
        __is_connected = true;
      
        /* connect to distributed partner object */
        if(!__is_local && __parent_reference != 0) {
          Event event(IObjectAcceptor::connectToObject_event, new TInValue<NMMReference>(__parent_reference));
          if(receiveReliableEvent(event) != SUCCESS) {
            ERROR_STREAM("Sending connectToObject event failed" << endl);
            throw BindingException("Sending connectToObject event failed");
          }
          if(event.getException()) {
            ERROR_STREAM("Exception in connectToObject: " << *event.getException() << endl);
            throw *event.getException();
          }
        } else {
          /* update the interface list */
          updateInterfaces();
        }
      }
      /* END< */    
    } catch (const SerializedException& e) {
      ERROR_STREAM("Exception during sendConnectToObject: " << e << std::endl);
      throw BindingException(e.getComment() );
    } catch ( ... ) {
      ERROR_STREAM("Unknown exception during sendConnectToObject" << std::endl);
      throw BindingException("Unknown exception in ProxyObject::sendConnectToObject");
    }
  }

  void ProxyObject::updateInterfaces()  {
    /* Begin> Delete the old interfaces and clear the list */
    for (list<Interface*>::iterator it = NMMObject::getInterfaceList().begin(); it != NMMObject::getInterfaceList().end(); ++it) {
      (*it)->setParentObject(0);
      delete *it;
    }

    NMMObject::getInterfaceList().clear();
    /* End< */
  
    /* Create the return value for method getInterfaceList */
    TOutValue<SequenceValue<Interface> > interface_list;
    interface_list.getRValue().useTValue(false);

    Event e("NMMObject::getInternalInterfaceList");
    e.setReturnValue(&interface_list);
    if(receiveReliableEvent(e) != SUCCESS) {
      ERROR_STREAM("Could not get InterfaceList ");
    } else {
      /* */
      if (!interface_list.getRValue().empty()) {
        for(unsigned int i = 0; i < interface_list.getRValue().size(); ++i) {
          /* If an interface pointer is NULL corresponding interface is unknown to the application */
          if ( interface_list.getRValue()[i] ) {
            interface_list.getRValue()[i]->setParentObject(this);
            addInterface(interface_list.getRValue()[i]);
            interface_list.getRValue()[i] = NULL;
          } else {
            WARNING_STREAM("Could not create Interface " << endl);
          }
        }
      } else {
        WARNING_STREAM("Distributed object without interfaces, so this object can only be accessed through its Proxy. " << endl);
      }
    }
    /*End< */
  }

  Result ProxyObject::getInterfaceList(SequenceValue<Interface>& interface_list) const {
    TOutValue<SequenceValue<Interface> > tmp_value(interface_list);
    Event e("NMMObject::getInternalInterfaceList");
    e.setReturnValue(&tmp_value);
    if(const_cast<ProxyObject&>(*this).receiveReliableEvent(e) == FAILURE) {
      WARNING_STREAM("Could not get interfaceList"<< endl);
      return FAILURE;
    }
    interface_list = tmp_value.getRValue();
    return SUCCESS;
  }

  Result ProxyObject::getUserInfo(UserInfo& user_info) const {
    TOutValue<UserInfo > value;
    Event event("NMMObject::getUserInfo", &value);
    if (const_cast<ProxyObject&>(*this).receiveReliableEvent(event) == FAILURE) {
      return FAILURE;
    }

    user_info = value.getRValue();
    return SUCCESS;
  }

  Result ProxyObject::getStrategyList(list<string>& strategy_list) const {
    TOutValue<list<string> > value;
    Event event("NMMObject::getStrategyList", &value);
    if (const_cast<ProxyObject&>(*this).receiveReliableEvent(event) == FAILURE) {
      return FAILURE;
    }

    strategy_list = value.getRValue();
    return SUCCESS;
  }

  TSAddress* ProxyObject::getDefaultAddress(const string& strategy) const {
    TOutValue<TSAddress*> address;
    Event e("NMMObject::getDefaultAddress", new TInValue<string>(strategy));
    e.setReturnValue(&address);
    if (const_cast<ProxyObject&>(*this).receiveReliableEvent(e) != SUCCESS) {
      if (e.getException())
        throw *e.getException();
    }
    return address.getRValue();
  }

  list<string> ProxyObject::getEventList() const {
    TOutValue<list<string> > event_list;
    Event e("NMMObject::getEventList");
    e.setReturnValue(&event_list);
    if (const_cast<ProxyObject&>(*this).receiveReliableEvent(e) != SUCCESS) {
      if (e.getException())
        throw *e.getException();
    }

    return event_list.getRValue();
  }

  Event* ProxyObject::getEventPrototype(const string& event_key)  const  {
    TOutValue<Event* > event(0);
    Event e("NMMObject::getEventPrototype", new TInValue<string>(event_key));
    e.setReturnValue(&event);
    if (const_cast<ProxyObject&>(*this).receiveReliableEvent(e) != SUCCESS) {
      if (e.getException())
        throw *e.getException();
    }

    return event.getRValue();
  }

  Event* ProxyObject::matchEventPrototype(const string& event_key)  const  {
    TOutValue<Event* > event(0);
    Event e("NMMObject::matchEventPrototype", new TInValue<string>(event_key));
    e.setReturnValue(&event);
    if (const_cast<ProxyObject&>(*this).receiveReliableEvent(e) != SUCCESS) {
      if (e.getException())
        throw *e.getException();
    }

    return event.getRValue();
  }


  void ProxyObject::registerEventListener(IEventReceiver* receiver) {
    bool register_listener = !getEventDispatcher().hasListener();
    NMMObject::registerEventListener(receiver);

    if (register_listener) {
      /* If we must register a generic listener, we test if specfifc listener are registerd */
      if(getEventDispatcher().hasSpecificListener()) {
        /* If specific listeners are registerd we send an "remove.." event to unregister Proxy for all specific listeners */
        Event event("NMMObject::removeEventListener");
        if (receiveReliableEvent(event) == FAILURE) {
          WARNING_STREAM("Could not unregister EventListener"<<endl);
        }
      }
      /* Then we register an eventhandler for all events*/
      Event event("NMMObject::registerEventListener");
      if(receiveReliableEvent(event) == FAILURE) {
        ERROR_STREAM("Could not register EventListener");
        return;
      }
    }
  }

  void ProxyObject::registerEventListener(const string& key, EDObject* ed) {
    /* Check if another EDObject or IEventReceiver is listening to the given event*/
    bool forward_registration = !(getEventDispatcher().hasSpecificListener(key));

    NMMObject::registerEventListener(key, ed);
    if (forward_registration && !getEventDispatcher().hasListener()) {
      Event event("NMMObject::registerSpecificEventListener", new TInValue<string>(key));
      if(receiveReliableEvent(event) == FAILURE) {
        ERROR_STREAM("Could not register EventListener");
      }
    }
  }

  void ProxyObject::registerEventListener(const string& key, IEventReceiver* handler) {
    /* Check if another EDObject or IEventReceiver is listening to the given event*/
    bool forward_registration = !(getEventDispatcher().hasSpecificListener(key));

    NMMObject::registerEventListener(key, handler);
    if (forward_registration &&  !getEventDispatcher().hasListener()) {
      Event event("NMMObject::registerSpecificEventListener", new TInValue<string>(key));
      if (receiveReliableEvent(event) == FAILURE) {
        ERROR_STREAM("Could not register EventListener");
      }
    }
  }

  void ProxyObject::removeEventListener(IEventReceiver* handler) {
    /* This bool is an additional test. If we have no listener and given listener is not in our map, we must not send the event*/
    bool had_listener_before_remove = getEventDispatcher().hasListener();
    NMMObject::removeEventListener(handler);

    if (had_listener_before_remove && !getEventDispatcher().hasListener()) {
      Event event("NMMObject::removeEventListener");
      if(receiveReliableEvent(event) == FAILURE) {
        WARNING_STREAM("Could not unregister EventListener"<< endl);
      }
      /* If we have specific listener the proxy object must send an registration event for these events*/
      if (getEventDispatcher().hasSpecificListener()) {
        list<string> specific_listener = getEventDispatcher().getSpecificListener();
        for (list<string>::iterator it = specific_listener.begin(); it != specific_listener.end(); ++it) {
          Event event("NMMObject::registerSpecificEventListener", new TInValue<string>(*it));
          if(receiveReliableEvent(event) == FAILURE) { 
            ERROR_STREAM("Could not register EventListener");
          }
        }
      }
    }
  }

  void ProxyObject::removeEventListener(const string& key, IEventReceiver* handler) {
    NMMObject::removeEventListener(key, handler);

    /* We send "removeEvent..." if no other listener is available and no generic listener is available*/
    if (!getEventDispatcher().hasSpecificListener(key) && !getEventDispatcher().hasListener()) {
      Event event("NMMObject::removeSpecificEventListener", new TInValue<string>(key));
      if (receiveReliableEvent(event) == FAILURE) {
        WARNING_STREAM("Could not unregister EventListener"<<endl);
      }
    }
  }

  void ProxyObject::removeEventListener(EDObject* edo) {
    list<string> unregistered_events = getEventDispatcher().removeEventListener(edo);

    if (!getEventDispatcher().hasListener()) {
      for (list<string>::iterator it = unregistered_events.begin(); it != unregistered_events.end(); ++it) {
        if (!getEventDispatcher().hasSpecificListener(*it)) {
          Event event("NMMObject::removeSpecificEventListener", new TInValue<string>(*it));
          if (receiveReliableEvent(event) == FAILURE) {
            WARNING_STREAM("Could not unregister EventListener"<<endl);
          }
        }
      }
    }
  }

  Interface* ProxyObject::getRegistry() {
    TOutValue<Interface*> r_value;
    Event e("NMMObject::getRegistry");
    e.setReturnValue(&r_value);
    if (receiveReliableEvent(e) != SUCCESS) {
      if (e.getException())
        throw *e.getException();
    }
    return r_value.getRValue();
  }

  void ProxyObject::setRegistry(const Interface& registry) {
    Event e("NMMObject::setRegistry", new TInValue<Interface>(registry));
    if (receiveReliableEvent(e) != SUCCESS) {
      if (e.getException())
        throw *e.getException();
    }
  }


  void ProxyObject::connectToObject(const NMMReference& reference) {
    /* set partner reference */
    __parent_reference = reference;
  
    /* update interfaces */
    updateInterfaces();
  }

  void ProxyObject::connectToService(const string& ) {
    /* get partner reference */
    TOutValue<NMMReference> reference;
    Event event("NMMObject::getReference", &reference);
    if (receiveReliableEvent(event) != SUCCESS) {
      WARNING_STREAM("Could not getNMMReference" << endl);
      __parent_reference = 0;
    } else {
      __parent_reference = reference.getRValue();
    }

    /* update interfaces */
    updateInterfaces();
  }

}

#include "nmm/nmm_suffix.hpp"
