/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 * 
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "ReliableInstreamDemultiplexer.hpp"
#include "InstreamCompositeStrategy.hpp"

#include "nmm/comm/messaging/Timestamp.hpp"
#include "nmm/comm/messaging/Buffer.hpp"
#include "nmm/comm/messaging/CEvent.hpp"
#include "nmm/comm/messaging/ITSOneWayBufferReceiver.hpp"
#include "nmm/comm/messaging/ITSOneWayCEventReceiver.hpp"

#include "nmm/utils/thread/MutexGuard.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  ReliableInstreamDemultiplexer::ReliableInstreamDemultiplexer()
    : __stream_counter(0)
  {}
  
  Result ReliableInstreamDemultiplexer::receiveCSMessage(Message* message) {
    MutexGuard lock(getSignalMutex());
    
    /* BEGIN> Message synchronization */
    /* force sequence numbers to be continuous and start with 0 */
    Timestamp timestamp = message->getTimestamp();
    timestamp.stream_counter = __stream_counter++;
    message->setTimestamp(timestamp);
    /* END< */
    
    /* BEGIN> Dispatch message to correct sub-binding */
    switch(message->getType()) {
    case Message::BUFFER:
      {
        Buffer* buffer = dynamic_cast<Buffer*>(message);
        Result result = getParent()->getOneWayBufferReceiver()->receiveOneWayBuffer(buffer);
        if(result == BLOCKED) {
          setBufferStrategyBlocked(true);
        }
        return result;
      }
    case Message::EVENT:
      {
        CEvent* cevent = dynamic_cast<CEvent*>(message);
        Result result = getParent()->getOneWayCEventReceiver()->receiveOneWayCEvent(cevent);
        if(result == BLOCKED) {
          setCEventStrategyBlocked(true);
        }
        return result;
      }
    default:
      /* DOES NOT HAPPEN */
      return FAILURE;
    }
    /* END< */
  }

  void ReliableInstreamDemultiplexer::receiveSignal(Signal signal, ISignalSender* sender) {
    if(signal == FREE_QUEUE_SLOT || signal == FREE_UPSTR_QUEUE_SLOT) {    
      MutexGuard lock(getSignalMutex());
      
      /* unblock child strategy */
      if(sender == getParent()->getOneWayBufferReceiver()) {
        setBufferStrategyBlocked(false);
      }
      if(sender == getParent()->getOneWayCEventReceiver()) {
        setCEventStrategyBlocked(false);
      }
      
      /* if both child strategies are unblocked, unblock composite strategy */
      if(!isBufferStrategyBlocked() && !isCEventStrategyBlocked()) {
        if(getParent()->getSignalReceiver()) {
          getParent()->getSignalReceiver()->receiveSignal(signal);
        } else {
          GLOBAL_WARNING_STREAM1("InstreamDemultiplexer> No signal receiver -> discarding signal" << endl);
        }
      }
    }
  }

  void ReliableInstreamDemultiplexer::flushITSReceiver() {
    MutexGuard lock(getSignalMutex());
  
    if ( getParent()->getOneWayCEventReceiver() ) {
      getParent()->getOneWayCEventReceiver()->flushITSReceiver();
    }

    if ( getParent()->getOneWayBufferReceiver() ) {
      getParent()->getOneWayBufferReceiver()->flushITSReceiver();
    }
  }
  
}

#include "nmm/nmm_suffix.hpp"
