/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 * 
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "ReliableInstreamMultiplexer.hpp"
#include "InstreamCompositeStrategy.hpp"

#include "nmm/comm/messaging/Buffer.hpp"
#include "nmm/comm/messaging/CEvent.hpp"
#include "nmm/comm/messaging/IMessageReceiver.hpp"
#include "nmm/comm/messaging/ITSOneWayBufferSender.hpp"
#include "nmm/comm/messaging/ITSOneWayCEventSender.hpp"

#include "nmm/utils/thread/MutexGuard.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  ReliableInstreamMultiplexer::ReliableInstreamMultiplexer() throw()
    : __pending_buffer(0),
      __pending_cevent(0),
      __stream_counter(0)
  {}
  
  ReliableInstreamMultiplexer::~ReliableInstreamMultiplexer() throw() {
    if(__pending_buffer) {
      /* release stored message, if any */
      __pending_buffer->release();
    }
    if(__pending_cevent) {
      /* release stored message, if any */
      __pending_cevent->release();
    }
  }
  
  Result ReliableInstreamMultiplexer::doForwardMessage(Message* message) {
    /* update the stream counter */
    __stream_counter = message->getTimestamp().stream_counter + 1;
    
    /* if we have no message receiver, discard nessage */
    if(!getParent()->getMessageReceiver()) {
      GLOBAL_WARNING_STREAM1("InstreamMultiplexer> No message receiver -> discarding message" << endl);
      message->release();
      return SUCCESS;
    }
    
    /* forward the message */
    GLOBAL_DEBUG_STREAM2("InstreamMultiplexer> Forwarding message with sequence number " << message->getTimestamp().stream_counter << " to partner" << endl);
    Result result = getParent()->getMessageReceiver()->receiveMessage(message, getParent());
    if(result == BLOCKED) {
      /* update the blocked flag */
      GLOBAL_DEBUG_STREAM2("InstreamMultiplexer> Partner returned BLOCKED" << endl);
      setBlocked(true);
    }
    
    return result;
  }
  
  Result ReliableInstreamMultiplexer::receiveOneWayCEvent(CEvent* cevent) {
    GLOBAL_DEBUG_STREAM2("InstreamMultiplexer> Received CEvent with sequence number " << cevent->getTimestamp().stream_counter << endl);
    MutexGuard lock(getSignalMutex());
    
    /* check if partner is already blocked.
       this may be the case if the partner returned BLOCKED on the other sub-binding. */
    if(isBlocked()) {
      /* we must not have a pending message */
      if(__pending_cevent) {
        GLOBAL_ERROR_STREAM1("InstreamMultiplexer> Message receiver previously returned BLOCKED! -> discarding event" << endl);
        cevent->release();
        return FAILURE;
      }
      
      /* store message and return BLOCKED */
      GLOBAL_DEBUG_STREAM2("InstreamMultiplexer> Partner receiver returned BLOCKED on other sub-binding " << endl);
      __pending_cevent = cevent;
      return BLOCKED;
    }
    
    /* check if sequence number is discontinuous */
    if(cevent->getTimestamp().stream_counter > __stream_counter) {
      /* we must not have a pending message */
      if(__pending_cevent) {
        GLOBAL_ERROR_STREAM1("InstreamMultiplexer> Unexpected discontinuous sequence number -> discarding event" << endl);
        cevent->release();
        return FAILURE;
      }
      
      /* store message and return BLOCKED */
      GLOBAL_DEBUG_STREAM2("InstreamMultiplexer> Sequence number is discontinuous" << endl);
      __pending_cevent = cevent;
      return BLOCKED;
    }
    
    /* forward this message */
    Result result = doForwardMessage(cevent);
    if(BLOCKED == result) {
      return BLOCKED;
    }
    
    /* handle pending messages */
    if(BLOCKED == handlePendingMessages()) {
      return BLOCKED;
    }
    
    /* return BLOCKED exactly if we have a pending message */
    if(__pending_cevent) {
      return BLOCKED;
    } else {
      return result;
    }
  }
  
  Result ReliableInstreamMultiplexer::receiveOneWayBuffer(Buffer* buffer) {
    GLOBAL_DEBUG_STREAM2("InstreamMultiplexer> Received Buffer with sequence number " << buffer->getTimestamp().stream_counter << endl);
    MutexGuard lock(getSignalMutex());
    
    /* check if partner is already blocked.
       this may be the case if the partner returned BLOCKED on the other sub-binding. */
    if(isBlocked()) {
      /* we must not have a pending message */
      if(__pending_buffer) {
        GLOBAL_ERROR_STREAM1("InstreamMultiplexer> Message receiver previously returned BLOCKED! -> discarding buffer" << endl);
        buffer->release();
        return FAILURE;
      }
      
      /* store message and return BLOCKED */
      GLOBAL_DEBUG_STREAM2("InstreamMultiplexer> Partner receiver returned BLOCKED on other sub-binding " << endl);
      __pending_buffer = buffer;
      return BLOCKED;
    }
    
    /* check if sequence number is discontinuous */
    if(buffer->getTimestamp().stream_counter > __stream_counter) {
      /* we must not have a pending message */
      if(__pending_buffer) {
        GLOBAL_ERROR_STREAM1("InstreamMultiplexer> Unexpected discontinuous sequence number -> discarding buffer" << endl);
        buffer->release();
        return FAILURE;
      }
      
      /* store message and return BLOCKED */
      GLOBAL_DEBUG_STREAM2("InstreamMultiplexer> Sequence number is discontinuous" << endl);
      __pending_buffer = buffer;
      return BLOCKED;
    }
    
    /* forward this message */
    Result result = doForwardMessage(buffer);
    if(BLOCKED == result) {
      return BLOCKED;
    }
    
    /* handle pending messages */
    if(BLOCKED == handlePendingMessages()) {
      return BLOCKED;
    }
    
    /* return BLOCKED exactly if we have a pending message */
    if(__pending_buffer) {
      return BLOCKED;
    } else {
      return result;
    }
  }
  
  Result ReliableInstreamMultiplexer::handlePendingMessages() {
    /* check for pending messages. if any pending message has a
     * continuous sequence number, forward it and check again if any
     * other pending message now has a continuous sequence number.  if
     * the partner returens BLOCKED after forwarding a pending
     * message, return.
     */
    do {
      if(__pending_cevent && __pending_cevent->getTimestamp().stream_counter <= __stream_counter) {
        /* forward the stored message */
        GLOBAL_DEBUG_STREAM2("InstreamMultiplexer> Forwarding pending message now" << endl);
        Result result = doForwardMessage(__pending_cevent);
        __pending_cevent = 0;
        if(result == BLOCKED) {
          /* both sub-bindings remain blocked */
          GLOBAL_DEBUG_STREAM2("InstreamMultiplexer> Partner returned BLOCKED, not unblocking sub-bindings" << endl);
          return BLOCKED;
        } else {
          /* unblock the sub-binding */
          GLOBAL_DEBUG_STREAM2("InstreamMultiplexer> Unblocking buffer sub-binding" << endl);
          getParent()->getOneWayCEventSender()->receiveSignal(FREE_QUEUE_SLOT); // XXX: fw: message direction !!!
        }
        continue;
      }
      
      if(__pending_buffer && __pending_buffer->getTimestamp().stream_counter <= __stream_counter) {
        /* forward the stored message */
        GLOBAL_DEBUG_STREAM2("InstreamMultiplexer> Forwarding pending message now" << endl);
        Result result = doForwardMessage(__pending_buffer);
        __pending_buffer = 0;
        if(result == BLOCKED) {
          /* both sub-bindings remain blocked */
          GLOBAL_DEBUG_STREAM2("InstreamMultiplexer> Partner returned BLOCKED, not unblocking sub-bindings" << endl);
          return BLOCKED;
        } else {
          /* unblock the sub-binding */
          GLOBAL_DEBUG_STREAM2("InstreamMultiplexer> Unblocking event sub-binding" << endl);
          getParent()->getOneWayBufferSender()->receiveSignal(FREE_QUEUE_SLOT); // XXX: fw: message direction !!!
        }
        continue;
      }
    } while(false);
    
    return SUCCESS;
  }
  
  void ReliableInstreamMultiplexer::receiveSignal(Signal signal) {
    if(signal == FREE_QUEUE_SLOT || signal == FREE_UPSTR_QUEUE_SLOT) {
      MutexGuard lock(getSignalMutex());
      GLOBAL_DEBUG_STREAM2("InstreamMultiplexer> Received FREE_QUEUE_SLOT signal" << endl);
      
      /* delete blocked flag */
      setBlocked(false);
      
      /* handle pending messages */
      if(BLOCKED == handlePendingMessages()) {
        /* partner is blocked again, so don't unblock any sub-bindings */
        return;
      }
      
      /* unblock sub-bindings. if we have a stored message, then the sub-binding from which it was received
         must remain blocked. */
      if(!__pending_buffer) {
        getParent()->getOneWayBufferSender()->receiveSignal(signal);
        GLOBAL_DEBUG_STREAM2("InstreamMultiplexer> Unblocking Buffer sub-binding" << endl);
      } else {
        GLOBAL_DEBUG_STREAM2("InstreamMultiplexer> Not unblocking Buffer sub-binding because a Buffer is still pending" << endl);
      }
      if(!__pending_cevent) {
        getParent()->getOneWayCEventSender()->receiveSignal(signal);
        GLOBAL_DEBUG_STREAM2("InstreamMultiplexer> Unblocking CEvent sub-binding" << endl);
      } else {
        GLOBAL_DEBUG_STREAM2("InstreamMultiplexer> Not unblocking CEvent sub-binding because a CEvent is still pending" << endl);
      }
    }
  }
  
}

#include "nmm/nmm_suffix.hpp"
