/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 * 
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#ifndef NMM_RELIABLE_INSTREAM_MULTIPLEXER_HPP
#define NMM_RELIABLE_INSTREAM_MULTIPLEXER_HPP NMM_RELIABLE_INSTREAM_MULTIPLEXER_HPP

#include "nmm/nmm_prefix.hpp"

#include "InstreamMultiplexer.hpp"

namespace NMM {
  
  class Message;
  
  /** Reliable instream message multiplexer. 
   *
   * This instream multiplexer is used for parallel bindings with exactly 
   * two reliable sub-bindings.  This class also performs message
   * synchronization. It orders messages for increasing sequence
   * numbers.
   */
  class ReliableInstreamMultiplexer
    : public InstreamMultiplexer
  {
  public:
    /** Default constructor.
     */
    ReliableInstreamMultiplexer() throw();
    
    /** Virtual destructor.
     */
    virtual ~ReliableInstreamMultiplexer() throw();
    
    /** @name Receiver Interface
     */
    //@{
    /** Receives %CEvent one-way.
     *
     * This method passes a %CEvent to the %TransportStrategy without waiting for a reply.
     *
     * @param cevent The buffer to be passed to the %TransportStrategy.
     * @return SUCCESS if call was successful, BLOCKED if blocked, FAILURE otherwise.
     */
    virtual Result receiveOneWayCEvent(CEvent* cevent);
    
    /** Receives %Buffer one-way.
     *
     * This method passes a %Buffer to the %TransportStrategy without waiting for a reply.
     *
     * @param buffer The buffer to be passed to the %TransportStrategy.
     * @return SUCCESS if call was successful, BLOCKED if blocked, FAILURE otherwise.
     */
    virtual Result receiveOneWayBuffer(Buffer* buffer);
    
    /** Receives a signal.
     *
     * @param signal Signal to receive
     */
    virtual void receiveSignal(Signal signal);
    //@}
    
  protected:

    /** Actually forwards a message. 
     *
     * This method forwards a message to the partner. 
     * It updates the blocked flag and the stream counter
     * and may perform additional tasks, such as benchmarks.
     *
     * @param message Message to forward
     * @return result of forwarding
     */
    virtual Result doForwardMessage(Message* message);
    
    /** This method forwards pending messages.
     *
     * This method forwards pending messages if they have continuous
     * sequence numbers. It may only be called if the partner is not
     * blocked.  It returns the queue status of the partner after
     * handling pending messages.
     */
    Result handlePendingMessages();
    
  private:
    Message* __pending_buffer;  ///< Pending Buffer.
    Message* __pending_cevent;  ///< Pending CEvent.
    
    nmm_int64 __stream_counter; ///< Sequence number of next message (stream counter).
  };
  
}

#include "nmm/nmm_suffix.hpp"

#endif
