/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 * 
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "SimpleInstreamMultiplexer.hpp"
#include "InstreamCompositeStrategy.hpp"

#include "nmm/comm/messaging/Buffer.hpp"
#include "nmm/comm/messaging/CEvent.hpp"
#include "nmm/comm/messaging/IMessageReceiver.hpp"
#include "nmm/comm/messaging/ITSOneWayBufferSender.hpp"

#include "nmm/utils/thread/MutexGuard.hpp"

#include <cassert>

#include "nmm/nmm_prefix.hpp"

// TEST: fw: for LatencyQueue
#include "nmm/utils/thread/ThreadedObject.hpp"
#include "nmm/utils/thread/ThreadCondition.hpp"
#include "nmm/utils/time/TimedElement.hpp"
#include "nmm/utils/time/Interval.hpp"

namespace NMM {

  // TODO: fw: This class is redundant! Include MessageDuration.hpp instead (but test it!)

  /** An accumulator function which computes the duration of a message in a stream.
   *
   * This class accepts a stream of messages and for each message computes the duration interval of the
   * previous message. It does this by computing timestamp differences. Only valid timestamps are used
   * for duration computation. Messages with invalid timestamps are assumed to have zero duration.
   *
   * Furthermore, this class detects syncReset events. A syncReset event is assumed to have zero duration,
   * and the duration of the last message with a valid timestamp before the syncReset also has zero duration.
   *
   * This class can not compute the duration of a single buffer. It can only determine the duration of a
   * buffer with a valid timestamp by being given the next subsequent buffer with a valid timestamp.
   */
  class MessageDuration
  {
    public:
      MessageDuration();

      /** Force a syncReset.
       */
      void syncReset();

      /** Insert a message.
       *
       * @param message a message in a stream
       * @return Duration of the last message with a valid timestamp not followed by a syncReset
       */
      Interval operator()(const Message* message);

    private:
      Timestamp last_valid_timestamp;
  };

  MessageDuration::MessageDuration()
    : last_valid_timestamp()
  {}

  void MessageDuration::syncReset()
  {
    // Invalidate last valid timestamp
    last_valid_timestamp = Timestamp();
  }

  Interval MessageDuration::operator()(const Message* message)
  {
    // Check for syncReset
    if(message->getType() == Message::EVENT || message->getType() == Message::FLUSH_EVENT)
    {
      const CEvent& cevent = static_cast<const CEvent&>(*message);
      for(size_t i=0; i<cevent.size(); ++i)
      {
        if(cevent[i].getKey() == "ISyncReset::syncReset")
        {
          // Found a syncReset
          syncReset();
          return Interval();
        }
      }
    }

    // Check if timestamp of incoming buffer is invalid
    if(!message->getTimestamp().is_valid) return Interval();

    // Store this timestamps as last valid timestamp
    Timestamp previous_timestamp = last_valid_timestamp;
    last_valid_timestamp = message->getTimestamp();

    // Check if a previous valid timestamp is known
    if(!previous_timestamp.is_valid) return Interval();

    // Compute difference of timestamps
    Interval duration = last_valid_timestamp.sync_time - previous_timestamp.sync_time;
    return duration;
  }


  class LatencyQueue : public ThreadedObject
  {
    public:
      LatencyQueue()
        : max_delay(2,0)
        , cv(mutex)
        , exit(false)
        , blocked(false)
        , blocking(false)
      {
        activateThread();
      }

      ~LatencyQueue()
      {
        {
          MutexGuard guard(mutex);

          // Flush the queue
          while(!latency_queue.empty())
          {
            latency_queue.rbegin()->first->release();
            latency_queue.pop_back();
          }

          // Tell the thread to stop
          exit = true;
          cv.notify();

          // Unlock the mutex so the thread can exit
        }

        // Join thread
        deactivateThread();
      }

      Result enqueue(Message* message)
      {
        MutexGuard guard(mutex);

        // If queue was empty, notify condition
        if(latency_queue.empty() && !blocked) cv.notify();

        // Compute duration of (previous) incoming message
        Interval duration = in_message_duration(message);

        // Add slot to queue
        LatencySlot slot;
        slot.first = message;
        slot.second = TimedElement::getTime();
        latency_queue.push_front(slot);
        total_delay += duration;

        // Blocking
        if(total_delay >= max_delay)
        {
          blocking = true;
          return BLOCKED;
        }
        else return SUCCESS;
      }

      void run()
      {
        MutexGuard guard(mutex);
        //cout << "latency queue running" << endl;

        while(isRunning())
        {
          // Wait for queue to be non-empty
          if(latency_queue.empty() || blocked) cv.wait();
          if(exit) break;

          assert(!blocked);
          assert(!latency_queue.empty());

          //cout << "latency queue not empty" << endl;

          // Compute remaining message delay
          Time now = TimedElement::getTime();
          Interval delay = now - latency_queue.rbegin()->second;
          Interval remaining_delay = max_delay - delay;

          // Wait for remaining message delay to expire
          Time absolute_wait_time = now + remaining_delay;
          cv.timedWait(absolute_wait_time.sec, absolute_wait_time.nsec);
          if(exit) break;

          // Compute duration of (previous) outgoing message
          Message* message = latency_queue.rbegin()->first;
          Interval duration = out_message_duration(message);

          // Dequeue message
          latency_queue.pop_back();
          total_delay -= duration;
          Result result = message_receiver->receiveMessage(message, message_sender);

          // Unblocking
          if(blocking)
          {
            signal_receiver->receiveSignal(FREE_QUEUE_SLOT);
            blocking = false;
          }

          // Blocked handling
          if(result == BLOCKED)
          {
            blocked = true;
          }
        }
      }

      void freeQueueSlot()
      {
        MutexGuard guard(mutex);
        if(blocked && !latency_queue.empty()) cv.notify();
        blocked = false;
      }

      void setMessageSender(IMessageSender* sender) { message_sender = sender; }
      void setMessageReceiver(IMessageReceiver* receiver) { message_receiver = receiver; }
      void setSignalReceiver(ISignalReceiver* receiver) { signal_receiver = receiver; }

    private:
      typedef std::pair<Message*, Time> LatencySlot;
      std::list<LatencySlot> latency_queue;
      MessageDuration in_message_duration;
      MessageDuration out_message_duration;
      Interval total_delay;
      Interval max_delay;

      IMessageReceiver* message_receiver;
      IMessageSender* message_sender;
      ISignalReceiver* signal_receiver;

      ThreadMutex mutex;
      ThreadCondition cv;
      bool exit;
      bool blocked;
      bool blocking;
  };


  SimpleInstreamMultiplexer::SimpleInstreamMultiplexer() throw()
    : __latency_queue(0)
  {
    if(UserInfo().getHostname() == "hannacv")
    {
      __latency_queue = new LatencyQueue();
    }
  }

  SimpleInstreamMultiplexer::~SimpleInstreamMultiplexer() throw()
  {
    if(__latency_queue)
    {
      delete __latency_queue;
      __latency_queue = 0;
    }
  }

  Result SimpleInstreamMultiplexer::receiveOneWayCEvent(CEvent* cevent) {
    if(!getParent()->getMessageReceiver()) {
      GLOBAL_WARNING_STREAM1("InstreamMultiplexer> No message receiver -> discarding event" << endl);
      cevent->release();
      return SUCCESS;
    }

    // TEST: fw: Latency hack
    if(__latency_queue)
    {
      // Set interfaces at latency queue
      __latency_queue->setMessageReceiver(getParent()->getMessageReceiver());
      __latency_queue->setMessageSender(getParent());
      __latency_queue->setSignalReceiver(dynamic_cast<ISignalReceiver*>(getParent()->getOneWayBufferSender()));

      // Enqueue message in latency queue
      return __latency_queue->enqueue(cevent);
    }
    else
    {
      Result result = getParent()->getMessageReceiver()->receiveMessage(cevent, getParent());
      return result;
    }
  }

  Result SimpleInstreamMultiplexer::receiveOneWayBuffer(Buffer* buffer) {
    if(!getParent()->getMessageReceiver()) {
      GLOBAL_WARNING_STREAM1("InstreamMultiplexer> No message receiver -> discarding buffer" << endl);
      buffer->release();
      return SUCCESS;
    }

    // TEST: fw: Latency hack
    if(__latency_queue)
    {
      // Set interfaces at latency queue
      __latency_queue->setMessageReceiver(getParent()->getMessageReceiver());
      __latency_queue->setMessageSender(getParent());
      __latency_queue->setSignalReceiver(dynamic_cast<ISignalReceiver*>(getParent()->getOneWayBufferSender()));

      // Enqueue message in latency queue
      return __latency_queue->enqueue(buffer);
    }
    else
    {
      Result result = getParent()->getMessageReceiver()->receiveMessage(buffer, getParent());
      return result;
    }
  }

  void SimpleInstreamMultiplexer::receiveSignal(Signal signal) {
    if(signal == FREE_QUEUE_SLOT || signal == FREE_UPSTR_QUEUE_SLOT) {
      // TEST: fw: Latency hack
      if(__latency_queue)
      {
        // Latency queue handles FREE_QUEUE_SLOT
        __latency_queue->freeQueueSlot();
      }
      else
      {
        /* BEGIN> forward this signal to both child strategies */
        assert(getParent()->getOneWayBufferSender());
        assert(getParent()->getOneWayCEventSender());
        getParent()->getOneWayBufferSender()->receiveSignal(signal);
        /* END< */
      }
    }
  }

}

#include "nmm/nmm_suffix.hpp"
