/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "TSFactory.hpp"
#include "TSFactoryObject.hpp"

#include "TransportStrategy.hpp"
#include "ConnectorAddress.hpp"
#include "ConnectionManager.hpp"
#include "FormatStrategyMap.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  TSFactory::TSFactory()
    : NamedObject("TSFactory")
  {}
  
  TSFactory::~TSFactory() {
    /* delete all registered default addresses */
    for(InternalIterator it = __factory_map.begin(); it != __factory_map.end(); ++it) {
      delete it->second.default_address;
    }
    
    /* clear factory map */
    __factory_map.clear();
  }
  
  void TSFactory::registerStrategy(const TSFactoryObject& factory, const string& strategy_type, unsigned int interfaces,
                                   const list<Format>& format_list, unsigned int capabilities) {
    /* add a new strategy to the map */
    InternalStruct is;
    is.factory_object  = &factory;
    is.default_address = 0;
    is.interfaces      = interfaces;
    is.capabilities    = capabilities;
    __factory_map.insert(pair<string, InternalStruct>(strategy_type, is));
    
    /* create a TSDescription object for this transport strategy */
    TSDescription tsd;
    tsd.setStrategyType(strategy_type);
    tsd.setInterfaces(interfaces);
    tsd.setCapabilities(capabilities);
    
    /* register formats */
    for(list<Format>::const_iterator i(format_list.begin());
        i != format_list.end();
        ++i) {
      SConnectionManager::getInstance().getFormatStrategyMap().addSupportedFormat(tsd, *i);
    }
  }
  
  void TSFactory::unregisterStrategy(const string& strategy_type, unsigned int interfaces, unsigned int capabilities) {
    /* find all registered strategies of this type */
    InternalIterator lower = __factory_map.lower_bound(strategy_type);
    InternalIterator upper = __factory_map.upper_bound(strategy_type);
    if(lower == upper) {
      /* no strategy of the given type was found */
      ERROR_STREAM("Unregister strategy: No TransportStrategy of type " << strategy_type << " registered" << endl);
      return;
    }
    
    /* unregister the transport strategy with exactly matching
       strategy type, interfaces and capabilities. */
    for(InternalIterator it=lower; it != upper; ++it) {
      if(it->second.interfaces == interfaces &&
         it->second.capabilities == capabilities) {
        /* delete default address */
        delete it->second.default_address;
        it->second.default_address = 0;
        
        /* create a TSDescription object for this transport strategy */
        TSDescription tsd;
        tsd.setStrategyType(strategy_type);
        tsd.setInterfaces(it->second.interfaces);
        tsd.setCapabilities(it->second.capabilities);
        
        /* unregister all formats */
        SConnectionManager::getInstance().getFormatStrategyMap().removeStrategyType(tsd);
        
        /* erase from list */
        __factory_map.erase(it);
        return;
      }
    }
    
    ERROR_STREAM("Unregister strategy: No TransportStrategy of type " << strategy_type << " with the given interfaces and capabilities registered" << endl);
  }  
  
  TransportStrategy* TSFactory::createTransportStrategy(const string& strategy_type, unsigned int interfaces) {
    /* if the strategy type is "NULL", this method succeeds and returns 0 */
    if(strategy_type == "NULL") {
      return 0;
    }
    
    /* first check if any strategies with the given type are available */
    InternalConstIterator lower = __factory_map.lower_bound(strategy_type);
    InternalConstIterator upper = __factory_map.upper_bound(strategy_type);
    if(lower == upper) {
      /* no strategy of the given type was found */
      ERROR_STREAM("No TransportStrategy of type " << strategy_type << " available" << endl);
      return 0;
    }
    
    /* check all registered strategies for supported interfaces */
    InternalConstIterator found;
    for(found=lower; found != upper; ++found) {
      if((found->second.interfaces & interfaces) == interfaces) {
        break;
      }
    }
    if(found == upper) {
      /* the requested strategy was found, but doesn't support all of the requested interfaces */
      ERROR_STREAM(strategy_type << " is available but does not support all of the requested interfaces" << endl);
      return 0;
    }
    
    /* create the instance */
    DEBUG_STREAM2("Create TransportStrategy " << strategy_type << endl);
    TransportStrategy* strategy = found->second.factory_object->createInstance();
    
    /* set the strategy description */
    TSDescription desc;
    desc.setStrategyType(strategy_type);
    desc.setInterfaces(found->second.interfaces);
    desc.setCapabilities(found->second.capabilities);
    strategy->setTSDescription(desc);
    
    return strategy;
  }
  
  void TSFactory::setDefaultAddress(const string& strategy_type, const TSAddress* address) {
    /* first check if any strategies with the given type are available */
    InternalIterator lower = __factory_map.lower_bound(strategy_type);
    InternalIterator upper = __factory_map.upper_bound(strategy_type);
    if(lower == upper) {
      ERROR_STREAM("No strategy " << strategy_type << " available" << endl);
      return;
    }
    
    /* replace all default addresses with the new default address */
    for(InternalIterator it=lower; it != upper; ++it) {
      delete it->second.default_address;
      if(!address) {
        it->second.default_address = 0;
      } else {
        /* store a copy of the address */
        it->second.default_address = address->copy();
        
        /* set the TSDescription of the address correctly */
        it->second.default_address->setStrategyInterfaces(it->second.interfaces);
        it->second.default_address->setStrategyCapabilities(it->second.capabilities);
      }
    }
  }
  
  TSAddress* TSFactory::getDefaultAddress(const string& strategy_type) const {
    /* first check if any strategies with the given type are available */
    InternalConstIterator lower = __factory_map.lower_bound(strategy_type);
    InternalConstIterator upper = __factory_map.upper_bound(strategy_type);
    if(lower == upper) {
      ERROR_STREAM("No strategy " << strategy_type << " available" << endl);
      return 0;
    }
    
    /* if a default address exist, return a copy of it. */
    if(lower->second.default_address) {
      return lower->second.default_address->copy();
    } else {
      DEBUG_STREAM("No default address for strategy " << strategy_type << " available" << endl);
      return 0;
    }
  }
  
  bool TSFactory::isStrategyAvailable(const string& strategy_type, unsigned int interfaces) const {
    /* first check if any strategies with the given type are available */
    InternalConstIterator lower = __factory_map.lower_bound(strategy_type);
    InternalConstIterator upper = __factory_map.upper_bound(strategy_type);
    
    /* check if any registered strategy suports the given interfaces */
    for(InternalConstIterator it=lower; it != upper; ++it) {
      if((it->second.interfaces & interfaces) == interfaces) {
        return true;
      }
    }
    return false;
  }
  
  unsigned int TSFactory::getInterfaces(const string& strategy_type) throw() {
    /* first check if any strategies with the given type are available */
    InternalConstIterator lower = __factory_map.lower_bound(strategy_type);
    InternalConstIterator upper = __factory_map.upper_bound(strategy_type);
    
    /* check if any registered strategy suports the given interfaces */
    unsigned int interfaces = 0;
    for(InternalConstIterator it=lower; it != upper; ++it) {
      interfaces |= it->second.interfaces;
    }
    return interfaces;
  }
  
  unsigned int TSFactory::getCapabilities(const string& strategy_type) throw() {
    /* first check if any strategies with the given type are available */
    InternalConstIterator lower = __factory_map.lower_bound(strategy_type);
    InternalConstIterator upper = __factory_map.upper_bound(strategy_type);
    
    /* check if any registered strategy suports the given interfaces */
    unsigned int capabilities = 0;
    for(InternalConstIterator it=lower; it != upper; ++it) {
      capabilities |= it->second.capabilities;
    }
    return capabilities;
  }
  
  list<string> TSFactory::getStrategyList() const {
    list<string> strategy_list;
    for(InternalConstIterator it = __factory_map.begin(); it != __factory_map.end(); ++it) {
      strategy_list.push_back(it->first);
    }
    
    return strategy_list;
  }
  
  list<TSDescription> TSFactory::getSupportedStrategies() const throw() {
    list<TSDescription> strategy_list;
    for(InternalConstIterator it = __factory_map.begin(); it != __factory_map.end(); ++it) {
      strategy_list.push_back(TSDescription(it->first, it->second.interfaces, it->second.capabilities));
    }
    
    return strategy_list;
  }

  list<TSDescription> TSFactory::getSupportedStrategies(unsigned int max_interfaces) const throw() {
    list<TSDescription> strategy_list;
    for(InternalConstIterator it = __factory_map.begin(); it != __factory_map.end(); ++it) {
      if((it->second.interfaces & max_interfaces) != 0) {
        strategy_list.push_back(TSDescription(it->first,it->second.interfaces & max_interfaces, it->second.capabilities));
      }
    }
    
    return strategy_list;
  }

  list<TSDescription> TSFactory::getSupportedStrategies(unsigned int max_interfaces, unsigned int min_capabilities) const throw() {
    list<TSDescription> strategy_list;
    for(InternalConstIterator it = __factory_map.begin(); it != __factory_map.end(); ++it) {
      if((it->second.interfaces & max_interfaces) != 0 &&
         (it->second.capabilities & min_capabilities) == min_capabilities) {
        strategy_list.push_back(TSDescription(it->first, it->second.interfaces & max_interfaces, it->second.capabilities));
      }
    }
    
    return strategy_list;
  }
  
  ConnectorAcceptor* TSFactory::requestConnectorAcceptor(const std::string& strategy) const {
    InternalConstIterator it = __factory_map.find(strategy);
    
    /* if the type name is unknown, return 0 */
    if (it == __factory_map.end()) {
      if (strategy == "NULL")
        return 0;

      ERROR_STREAM("No TransportStrategy of type " << strategy << " avaialable "<< endl);
      return 0;
    }

    /* create the instance*/
    ConnectorAcceptor* connector_acceptor = it->second.factory_object->getConnectorAcceptor();
    if(!connector_acceptor) {
      ERROR_STREAM("No ConnectorAcceptor for strategy " << strategy << " avaialable "<< endl);
    }
    
    return connector_acceptor;
  }

  void TSFactory::addConnectorAddress(const string& strategy_type) {
    /* Get IConnectionManager interface */
    IConnectionManager_var iface(SConnectionManager::getInstance().getInterface<IConnectionManager>());
    if (!iface.get()) {
      ERROR_STREAM("Could not get local connection manager interface" << endl);
      return;
    }
    ConnectorAddress address(strategy_type);
    address.setConnectionManager(*iface);
  
    /* Register default address for Strategy */
    setDefaultAddress(strategy_type, &address);
  }  
}

DEFINE_SINGLETON(NMM::TSFactory);

#include "nmm/nmm_suffix.hpp"
