/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 * 
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "UnreliableInstreamMultiplexer.hpp"
#include "InstreamCompositeStrategy.hpp"

#include "nmm/comm/messaging/Buffer.hpp"
#include "nmm/comm/messaging/CEvent.hpp"
#include "nmm/comm/messaging/IMessageReceiver.hpp"
#include "nmm/comm/messaging/ITSOneWayBufferSender.hpp"
#include "nmm/comm/messaging/ITSOneWayCEventSender.hpp"

#include "nmm/utils/thread/MutexGuard.hpp"

#include <cassert>

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  Result UnreliableInstreamMultiplexer::receiveOneWayCEvent(CEvent* cevent) {
    MutexGuard lock(getSignalMutex());
    if(!getParent()->getMessageReceiver()) {
      GLOBAL_WARNING_STREAM1("InstreamMultiplexer> No message receiver -> discarding event" << endl);
      cevent->release();
      return SUCCESS;
    }
    
    if(isBlocked()) {
      GLOBAL_ERROR_STREAM1("InstreamMultiplexer> Message receiver previously returned BLOCKED! -> discarding event" << endl);
      cevent->release();
      return FAILURE;
    }
    
    // XXX: fw: forward without synchronization.
    // XXX: fw: this is a hack, synchronization will be added later.
    
    Result result = getParent()->getMessageReceiver()->receiveMessage(cevent, getParent());
    if(result == BLOCKED) {
      setBlocked(true);
    }
    return result;
  }
  
  Result UnreliableInstreamMultiplexer::receiveOneWayBuffer(Buffer* buffer) {
    MutexGuard lock(getSignalMutex());
    
    if(!getParent()->getMessageReceiver()) {
      GLOBAL_WARNING_STREAM1("InstreamMultiplexer> No message receiver -> discarding buffer" << endl);
      buffer->release();
      return SUCCESS;
    }
    
    if(isBlocked()) {
      GLOBAL_ERROR_STREAM1("InstreamMultiplexer> Message receiver previously returned BLOCKED! -> discarding buffer" << endl);
      buffer->release();
      return FAILURE;
    }
    
    // XXX: fw: forward without synchronization.
    // XXX: fw: this is a hack, synchronization will be added later.
    
    Result result = getParent()->getMessageReceiver()->receiveMessage(buffer, getParent());
    if(result == BLOCKED) {
      setBlocked(true);
    }
    return result;
  }

  void UnreliableInstreamMultiplexer::receiveSignal(Signal signal) {
    MutexGuard lock(getSignalMutex());
    if(signal == FREE_QUEUE_SLOT || signal == FREE_UPSTR_QUEUE_SLOT) {
      
      // XXX: fw: forward without synchronization.
      // XXX: fw: this is a hack, synchronization will be added later.
      
      /* delete blocked flag */
      setBlocked(false);
      
      /* BEGIN> forward this signal to both child strategies */
      assert(getParent()->getOneWayBufferSender());
      assert(getParent()->getOneWayCEventSender());
      getParent()->getOneWayBufferSender()->receiveSignal(signal);
      getParent()->getOneWayCEventSender()->receiveSignal(signal);
      /* END< */
    }
  }
  
}

#include "nmm/nmm_suffix.hpp"
