/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "CParameter.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  CParameter::CParameter(const string & pName) : Parameter(pName) {
    m_parameter_type = Parameter::FPOINTER;
    m_counter        = m_value_list.begin();
  }
  
  CParameter::CParameter(const string & pName, Format * pFormatPtr) : Parameter(pName) {
    m_value_list.push_back(pFormatPtr);
    m_parameter_type = Parameter::FPOINTER;
    m_counter        = m_value_list.begin();
  }
  
  CParameter::~CParameter() {
    for (format_list::const_iterator i = m_value_list.begin(); i != m_value_list.end(); ++i) {
      delete *i;
    }
  }
  
  void CParameter::addValueInternal(Format* value) {
    m_value_list.push_back(value);
  }
  
  void CParameter::addValue(void * pValue) {
    addValue(*(NMM::Format **) pValue);
  }
  
  void CParameter::addValue(Format * value) {
    /*
     * Add unique value to list.
     * This method guarantees that there are no duplicate formats in the list.
     * We can't use STL sets because there is no meaningful less-operator for formats.
     */
    for (format_list::iterator i = m_value_list.begin(); i != m_value_list.end(); ++i) {
      if ((*i)->isFormatIdentical(value)) {
        m_value_list.erase(i);
        break;
      }
    }
    
    addValueInternal(value);
  }
  
  void CParameter::showParameter() const {
    GLOBAL_DEBUG_STREAM2( "Parameter-Name  : " << m_parameter_name << endl);
    GLOBAL_DEBUG_STREAM2( "Parameter-Type  : CFormat " << endl);
    GLOBAL_DEBUG_STREAM2( "Parameter-Weight: " << m_parameter_weight << endl);
    for (format_list::const_iterator i = m_value_list.begin(); i != m_value_list.end(); i++) {
      GLOBAL_DEBUG_STREAM2( m_parameter_name << " : " << (*i) << endl);
      GLOBAL_DEBUG_STREAM2( "****** BEGIN REKURSIVE FORMAT ******** " << endl);
      GLOBAL_DEBUG_STREAM2( **i);
      GLOBAL_DEBUG_STREAM2( "****** END REKURSIVE FORMAT ******** " << endl);
    }
  }
  
  ostream& CParameter::showParameter(ostream& os) const {
    os << m_parameter_name << "[weight:" << m_parameter_weight << "]=CFormat() = { ";

    char count = 'A';
    for (format_list::const_iterator i = m_value_list.begin();
         i != m_value_list.end();
         i++) {
      os << endl << count++ << ". " << **i;
    }

    os << "}";

    return os;
  }
  
  ostream& CParameter::showParameterVerbose(ostream& os) const {
    os << "Parameter-Name  : " << m_parameter_name << endl;
    os << "Parameter-Type  : CFormat " << endl;
    os << "Parameter-Weight: " << m_parameter_weight << endl;
    for (format_list::const_iterator i = m_value_list.begin(); i != m_value_list.end(); i++) {
      os << m_parameter_name << " : " << (*i) << endl;
      os << "****** BEGIN REKURSIVE FORMAT ******** " << endl;
      (*i)->showParameterVerbose(os);
      os << "****** END REKURSIVE FORMAT ******** " << endl;
    }
    return os;
  }
  
  bool CParameter::isValueIncluded(void *pValue) const {
    return isValueIncluded((NMM::Format **) pValue);
  }
  
  bool CParameter::isValueIncluded(const Format *value) const {
    for (format_list::const_iterator i = m_value_list.begin(); i != m_value_list.end(); ++i) {
      if ( (*i)->isFormatIdentical(value) ) {
        return true;
      }
    }
    return false;
  }
  
  void * CParameter::nextValue() const {
    format_list::const_iterator tempCounter;
    if (m_counter != m_value_list.end()) {
      tempCounter=m_counter;
      m_counter++;
      return *tempCounter;
    } else {
      m_counter=m_value_list.begin();
      return NULL;
    }
  }
  
  uint CParameter::getNumberOfValues() const {
    return m_value_list.size();
  }
  
  void * CParameter::getFirstValue() const {
    m_counter = (++m_value_list.begin());
    return *m_value_list.begin();
  }
  
  Parameter* CParameter::duplicate() const {
    CParameter * new_parameter = new CParameter(m_parameter_name);
    for (format_list::const_iterator i = m_value_list.begin(); i != m_value_list.end(); ++i) {
      new_parameter->addValueInternal((*i)->duplicateFormat());
    }
    return (Parameter *)new_parameter;
  }
  
  float CParameter::getWeight(void* /*value*/) const {
    return DEFAULT_WEIGHT;
  }
  
  Parameter * CParameter::getIntersection(const Parameter * reference) const {
    /*
     * Get the set of intersection formats between the first and second set.
     * We might get some intersection formats twice. This is handled in addValue.
     *
     * This method returns NULL if the set of intersections is empty.
     * Empty value lists ar not allowed in formats. If we are intersecting two
     * recursive ormats, and their subformats don't intersect, then the formats
     * themselves don't intersect.
     */
    const CParameter* cp_reference = dynamic_cast<const CParameter *>(reference);
    if (cp_reference == NULL) { return NULL; }
    CParameter* result = new CParameter(m_parameter_name);
    
    for (format_list::const_iterator i = m_value_list.begin(); i != m_value_list.end(); ++i) {
      for (format_list::const_iterator j = cp_reference->m_value_list.begin(); j != cp_reference->m_value_list.end(); ++j) {
        Format* intersec = (*i)->createIntersectionFormat(*j);
        if (intersec != NULL) {
          result->addValue(intersec);
        }
      }
    }
    
    if (result->m_value_list.empty()) {
      delete result;
      return NULL;
    }
    
    return result;
  }
  
  Format * CParameter::getFormat(const string & name) const {
    /*
     * Get format with a certain format type.
     * Only finds the first format of that type. Most methods in the old implementation assumed that
     * there is only one format of each type in the set. This has been changed, and this method
     * is no longer very useful.
     */
    for (format_list::const_iterator i = m_value_list.begin(); i != m_value_list.end(); ++i) {
      if ( (*i)->getFormatType() == name) {
        return *i;
      }
    }
    
    return NULL;
  }
  
  bool CParameter::isParameterIdentical(const Parameter * reference) const {
    /*
     * Check if both format sets contain exactly the same formats.
     * The order of the formats in the list is not relevant.
     * (This could be better implemented with a set instead of a list)
     */
    const CParameter* cp_reference = dynamic_cast<const CParameter*>(reference);
    if (cp_reference == NULL) { return false; }
    
    for (format_list::const_iterator i = m_value_list.begin(); i != m_value_list.end(); ++i) {
      if (!cp_reference->isValueIncluded(*i)) { return false; }
    }
    
    for (format_list::const_iterator i = cp_reference->m_value_list.begin(); i != cp_reference->m_value_list.end(); ++i) {
      if (!isValueIncluded(*i)) { return false; }
    }
    
    return true;
  }
  
  bool CParameter::isParameterMatching(const Parameter* reference) const {
    /*
     * Check if any format in the first set matches any format in the second set.
     */
    const CParameter* cp_reference = dynamic_cast<const CParameter*>(reference);
    if (cp_reference == NULL) { return false; }
    
    for (format_list::const_iterator i = m_value_list.begin(); i != m_value_list.end(); ++i) {
      for (format_list::const_iterator j = cp_reference->m_value_list.begin(); j != cp_reference->m_value_list.end(); ++j) {
        if ((*i)->isFormatMatching(*j)) { return true; }
      }
    }
    
    return false;
  }
  
  float CParameter::getStandardizedWeight() const {
    float local_weight = 0;
    int count = 0;
    for (format_list::const_iterator i = m_value_list.begin(); i != m_value_list.end(); ++i) {
      local_weight = local_weight + (*i)->getStandardizedWeight();
      ++count;
    }
    return ((local_weight / count) * m_parameter_weight);
  }
  
  const format_list & CParameter::getFPointerValueList() const {
    return m_value_list;
  }
  
  void CParameter::fromStream(IValueStream& iv) {
    if ( (iv.getCurrentType() != "CParameter") || (iv.getCurrentNamespace() != "NMM") ) {
      GLOBAL_ERROR_STREAM ("Invalid type or namespace:" << endl);
      GLOBAL_ERROR_STREAM ("Current type : " << iv.getCurrentType() << "                Expected type: CParameter" << endl);
      GLOBAL_ERROR_STREAM ("Current namespace : " << iv.getCurrentNamespace() << "      Expected namespace: NMM " << endl);
      iv.nextElement();
      return;
    }
    
    if (!iv.goDown()) {
      GLOBAL_MESSAGE_STREAM("CompositeFormat is empty" << endl);
      iv.nextElement();
      return;
    }
    
    string p_name;
    nmm_float weight = 0.0;
    iv >> p_name;
    iv >> weight;
    
    setParameterName(p_name);
    setParameterWeight(weight);
    
    Format* tmp_format = 0;
    while(!iv.isEmpty()) {
      tmp_format = new Format();
      iv >> *tmp_format;
      addValueInternal(tmp_format);
    }
    
    iv.goUp();
    return;
  }
  
  void CParameter::toStream(OValueStream& ov)  const {
    ov.beginType("CParameter", "NMM");
    
    string p_name = getParameterName();
    nmm_float weight = getParameterWeight();

    ov << p_name;
    ov << weight;
    
    for (list<Format*>::const_iterator i = m_value_list.begin(); i != m_value_list.end(); ++i) {
      ov << **i;
    }
    ov.endType();
  }
}

#include "nmm/nmm_suffix.hpp"
