/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "FParameter.hpp"
#include "nmm/comm/serialize/IValueStream.hpp"
#include "nmm/comm/serialize/OValueStream.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  FParameter::FParameter(const string & pName):Parameter(pName)
  {
    m_parameter_type = Parameter::FLOAT;
    m_value_counter  = m_value_list.begin();
    m_pair_counter   = m_value_list.begin();
  }

  FParameter::FParameter(const string & pName,float pValue, float pWeight) : Parameter(pName) 
  {
    m_parameter_type= Parameter::FLOAT;
    m_value_list.insert(make_pair(pValue, pWeight));
    m_value_counter = m_value_list.begin();
    m_pair_counter  = m_value_list.begin();
    m_default_value = pValue;
  }


  void FParameter::addValue(void *pValue) 
  {
    this->addValue(pValue, DEFAULT_WEIGHT);
  }

  void FParameter::addValue(void *pValue, float pWeight) 
  {
    if ( m_value_list.empty()) m_default_value = *(static_cast<float *>(pValue));
    m_value_list.insert(make_pair( *((float *)pValue), pWeight) );
    m_value_counter = m_value_list.begin();
    m_pair_counter  = m_value_list.begin();
  }

  void FParameter::addValue(float value, float weight)
  {
    if (m_value_list.empty()) m_default_value = value;
    m_value_list.insert(make_pair(value,weight));
    m_value_counter = m_value_list.begin();
    m_pair_counter  = m_value_list.begin();
  }

  void FParameter::showParameter() const
  {
    GLOBAL_DEBUG_STREAM2( "Parameter-Name  : " << m_parameter_name << endl);
    GLOBAL_DEBUG_STREAM2( "Parameter-Type  : FLOAT" << endl);
    GLOBAL_DEBUG_STREAM2( "Parameter-Weight: " << m_parameter_weight << endl);
    for (map<float,float>::const_iterator i = m_value_list.begin(); i != m_value_list.end(); i++)  
      {
	GLOBAL_DEBUG_STREAM2( "Value :" << (*i).first << " - Weight :" << (*i).second << endl);
      }
  }
  
  ostream& FParameter::showParameter(ostream& os) const
  {
    os << m_parameter_name << "[weight:" << m_parameter_weight << "]=float(";

    for (map<float,float>::const_iterator i = m_value_list.begin(); i != m_value_list.end(); i++)  
      {
        if (i != m_value_list.begin()) {
          os << ", ";
        }
	os << (*i).first << "[weight:" << (*i).second << "]";
      }
    os << ")";
    return os;
  }
  
  ostream& FParameter::showParameterVerbose(ostream& os) const
  {
    os << "Parameter-Name  : " << m_parameter_name << endl;
    os << "Parameter-Type  : FLOAT" << endl;
    os << "Parameter-Weight: " << m_parameter_weight << endl;
    for (map<float,float>::const_iterator i = m_value_list.begin(); i != m_value_list.end(); i++)  
      {
 	os << "Value :" << (*i).first << " - Weight :" << (*i).second << endl;
      }
    return os;
  }
  
  bool FParameter::isValueIncluded(void *pValue) const
  {
    map<float, float>::const_iterator i;
    for (i = m_value_list.begin(); i != m_value_list.end(); i++) 
      {
	if( (*i).first == *((float *)pValue) ) return true;
      }
    return false;
  }

  bool FParameter::isValueIncluded(float value) const
  {
    map<float , float>::const_iterator itr = m_value_list.find(value);
    if ( itr == m_value_list.end())
      return false;
    else
      return true;
  }
  
  void * FParameter::nextValue() const
  {
    map<float,float>::const_iterator tempitr;
    if (m_value_counter != m_value_list.end()) 
      {
	tempitr = m_value_counter;
	m_value_counter++;
	return static_cast<void*>( const_cast<float*>(&((*tempitr).first)));
      }
    else 
      {
	m_value_counter = m_value_list.begin();
	return NULL;
      }
  }
  
  map<float,float>::const_iterator FParameter::nextPair() const
  {
    map<float,float>::const_iterator tempitr;
    if (m_pair_counter != m_value_list.end()) 
      {
	tempitr = m_pair_counter;
	m_pair_counter ++;
	return tempitr;
      }
    else 
      {
	m_pair_counter = m_value_list.begin();
	return m_value_list.end();
      }
  }
  
  Parameter * FParameter::duplicate() const
  {
    FParameter *tempParameter = new FParameter(m_parameter_name);
    float tempfloat;
    for (map<float, float>::const_iterator itr = m_value_list.begin(); itr != m_value_list.end(); itr ++)
      {
	tempfloat = (*itr).first;
	tempParameter->addValue(tempfloat, (*itr).second );
      }
    return tempParameter;
  }

  float FParameter::getWeight(void *pValue) const
  {
    map<float,float>::const_iterator itr = m_value_list.find(*(float *)pValue);
    return (*itr).second;
  }

  uint FParameter::getNumberOfValues() const
  {
    return m_value_list.size();
  }

  bool FParameter::isParameterIdentical(const Parameter *reference) const
  {
    if (reference->getParameterType() != Parameter::FLOAT)
      return false;
    if (reference->getNumberOfValues() != m_value_list.size())
      return false;
    for (map<float,float>::const_iterator  itr = m_value_list.begin(); itr != m_value_list.end(); itr ++)
      {
	if (!reference->isValueIncluded((float*)&((*itr).first)))
	  return false;
      }
    return true;
  }

  bool FParameter::isParameterMatching(const Parameter *reference) const
  {
    for (map<float,float>::const_iterator itr = m_value_list.begin(); itr != m_value_list.end(); itr ++)
      {
	if (reference->isValueIncluded((*itr).first))
	  return true;
      }
    return false;
  }

  Parameter *FParameter::getIntersection(const Parameter *reference) const
  {
    FParameter *new_parameter = new FParameter(m_parameter_name);
    float tempfloat;
    for (map<float, float>::const_iterator itr = m_value_list.begin(); itr != m_value_list.end(); itr ++)
      {
	tempfloat = (*itr).first;
	if (reference->isValueIncluded(tempfloat))
	  {
	    new_parameter->addValue(tempfloat, 
				    ((*itr).second 
				     + reference->getWeight(&tempfloat))/2);
	  }
	else
	  {
// 	    cout << tempfloat << " nicht enthalten" << endl;
	  }
      }
    return new_parameter;
  }

  void * FParameter::getFirstValue() const
  {
    if(m_value_list.empty()){
      GLOBAL_ERROR_STREAM("FParameter::getFirstValue() > m_value_list is empty" << endl);
      return NULL;
    }

    m_value_counter = (++m_value_list.begin());
    return const_cast<float*>(&m_value_list.begin()->first);
  }

  float FParameter::getStandardizedWeight() const
  {
    float local_weight = 0;
    for (map<float,float>::const_iterator itr = m_value_list.begin(); itr != m_value_list.end(); itr ++)
      {
	local_weight += (*itr).second;
      }
    return ((local_weight/m_value_list.size()) * m_parameter_weight);
  }

  float FParameter::getDefaultValue() const
  {
    return m_default_value;
  }

  void FParameter::setDefaultValue(float param)
  {
    m_default_value = param;
  }

  float FParameter::getLowestValue() const
  {
    return (*m_value_list.begin()).first;
  }

  float FParameter::getHighestValue() const
  {
    return (*(--m_value_list.end())).first;
  }

  const float_list FParameter::getFloatValueList() const
  {
    float_list result;
    for (map<float, float>::const_iterator itr = m_value_list.begin(); itr != m_value_list.end(); itr ++)
      {
	result.push_back((*itr).first);
      }
    return result;
  }

  void FParameter::fromStream(IValueStream& iv) {
    if ((iv.getCurrentType() != "FParameter") || (iv.getCurrentNamespace() != "NMM")) {
      GLOBAL_ERROR_STREAM ("Invalid type or namespace:" << endl);
      GLOBAL_ERROR_STREAM ("Current type : " << iv.getCurrentType() << "                Expected type: FParameter" << endl);
      GLOBAL_ERROR_STREAM ("Current namespace : " << iv.getCurrentNamespace() << "      Expected namespace: NMM " << endl);
      iv.nextElement();
      return;
    }
    if (!iv.goDown()) {
      GLOBAL_ERROR_STREAM ("FParameter is empty" << endl);
      iv.nextElement();
      return;
    }

    string par_name;
    nmm_float weight = 0.0;

    iv >> par_name;
    iv >> weight;

    setParameterName(par_name);
    setParameterWeight(weight);

    pair<float, float> tmp_pair;
    while(!iv.isEmpty()) {
      iv >> tmp_pair.first;
      iv >> tmp_pair.second;
      addValue(tmp_pair.first, tmp_pair.second);
    }

    iv.goUp();
  }

  void FParameter::toStream(OValueStream& ov) const  {
    ov.beginType("FParameter", "NMM");

    string p_name = getParameterName();
    nmm_float weight = getParameterWeight();

    ov << p_name;
    ov << weight;

    for (map<float, float>::const_iterator it = m_value_list.begin(); it != m_value_list.end(); it++)
    {
      nmm_float v = (*it).first;
      nmm_float w = (*it).second;
      ov << v;
      ov << w;
    }

    ov.endType();
  }

}

#include "nmm/nmm_suffix.hpp"
