/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "Format.hpp"
#include "IParameter.hpp"
#include "FParameter.hpp"
#include "SParameter.hpp"
#include "CParameter.hpp"
#include "WParameter.hpp"
#include "R_IParameter.hpp"
#include "R_FParameter.hpp"
#include "R_INT.hpp"
#include "R_FLOAT.hpp"
#include "nmm/comm/serialize/TValueConstructor.hpp"

#include <iostream>

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  Format::Format() 
  {
    m_IOPartner     = NULL;
    m_jack_tag      = "default";
    m_value_counter = m_parameter_list.end();
  }

  Format::Format(const string & pFormatType, const string & param) 
  {
    m_format_type   = pFormatType;
    m_IOPartner     = NULL;
    m_jack_tag      = param;
    m_value_counter = m_parameter_list.end();
  }

  Format::~Format()
  {
    erase();
  }

  void Format::erase() {
    delete m_IOPartner;
    m_IOPartner = 0;
    for (parameter_iterator itr = m_parameter_list.begin(); itr != m_parameter_list.end(); ++itr )
    {
      delete *itr;
    }
    m_parameter_list.clear();
    m_value_counter = m_parameter_list.begin();
  }

  Format::Format(const Format &f) {
    m_IOPartner = 0;
    *this = f;
  }



  const string & Format::getFormatType() const 
  {
    return m_format_type;
  }

  void Format::setFormatType(const string & param) 
  {
    m_format_type = param;
  }

  void Format::addIntParamValue(const string & pParameterName , 
				int32_t pValue , 
				float weight) throw (ParameterException)
  {
    if ( (weight < 0) || (weight > 1))
      throw ParameterExceptionWeightRangeError(const_cast<char *>(pParameterName.c_str()),weight);

    if ( this->getParameter(pParameterName) != NULL)
      if (!dynamic_cast<IParameter *>(this->getParameter(pParameterName)))
	throw ParameterExceptionFalseType(const_cast<char *>(pParameterName.c_str()), 
					  this->getParameter(pParameterName)->getParameterType(), 
					  Parameter::INT32);

    if (this->getParameter(pParameterName))
      {
	this->getParameter(pParameterName)->addValue(pValue, weight);
      }
    else
      {
	m_parameter_list.push_back(new IParameter(pParameterName,pValue,weight));
	m_value_counter = m_parameter_list.begin();
      }
    return;
  }

  void Format::addFloatParamValue(const string & pParameterName,
				  float pValue, 
				  float weight)  throw (ParameterException)
  {
    if ( (weight < 0) || (weight > 1))
      throw ParameterExceptionWeightRangeError(const_cast<char *>(pParameterName.c_str()),weight);

    if (this->getParameter(pParameterName) != NULL)
      if (!dynamic_cast<FParameter *>(this->getParameter(pParameterName)))
	throw ParameterExceptionFalseType(const_cast<char *>(pParameterName.c_str()), 
					  this->getParameter(pParameterName)->getParameterType(), 
					  Parameter::FLOAT);

    if (this->getParameter(pParameterName))
      {
	static_cast<FParameter *>(this->getParameter(pParameterName))->addValue(pValue,weight);
      }
    else
      {
	m_parameter_list.push_back(new FParameter(pParameterName,pValue,weight));
	m_value_counter = m_parameter_list.begin();
      }
    return;
  }

  void Format::addStringParamValue(const string & pParameterName, const string & pValue, float weight)  throw (ParameterException)
  {
    if ( (weight < 0) || (weight > 1))
      throw ParameterExceptionWeightRangeError(const_cast<char *>(pParameterName.c_str()),weight);

    if (this->getParameter(pParameterName) != NULL)
      if (!dynamic_cast<SParameter *>(this->getParameter(pParameterName)))
	throw ParameterExceptionFalseType(const_cast<char *>(pParameterName.c_str()), 
					  this->getParameter(pParameterName)->getParameterType(), 
					  Parameter::STRING);

    if (this->getParameter(pParameterName))
      {
	this->getParameter(pParameterName)->addValue(pValue,weight);
      }
    else 
      {
	m_parameter_list.push_back(new SParameter(pParameterName, pValue, weight));
	m_value_counter = m_parameter_list.begin();
      }
    return;
  }

  void Format::addFPointerParamValue(const string & pParameterName, Format * pValue) throw (ParameterException)
  {
    if (this->getParameter(pParameterName))
      if (!dynamic_cast<CParameter *>(this->getParameter(pParameterName)))
	throw ParameterExceptionFalseType(const_cast<char *>(pParameterName.c_str()), 
					  this->getParameter(pParameterName)->getParameterType(), 
					  Parameter::FPOINTER);

    if (this->getParameter(pParameterName))
      {
	this->getParameter(pParameterName)->addValue(&pValue);
      }
    else 
      {
	m_parameter_list.push_back(new CParameter(pParameterName,pValue));
	m_value_counter = m_parameter_list.begin();
      }
    return;
  }

  void Format::addRIntParamValue(const string & pParameterName, 
				 int32_t lValue, 
				 int32_t hValue, 
				 float lweight, 
				 float hweight)  throw (ParameterException)
  {
    if ( (lweight < 0) || (lweight > 1))
      throw ParameterExceptionWeightRangeError(const_cast<char *>(pParameterName.c_str()),lweight);
    if ( (hweight < 0) || (hweight > 1))
      throw ParameterExceptionWeightRangeError(const_cast<char *>(pParameterName.c_str()),hweight);


    if (this->getParameter(pParameterName))
      if (!dynamic_cast<R_IParameter *>(this->getParameter(pParameterName)))
	throw ParameterExceptionFalseType(const_cast<char *>(pParameterName.c_str()), 
					  this->getParameter(pParameterName)->getParameterType(), 
					  Parameter::R_INT32);
    
    if (this->getParameter(pParameterName))
      {
	this->getParameter(pParameterName)->addValue(new R_INT(lValue,hValue,lweight,hweight));
      }
    else 
      {
	m_parameter_list.push_back(new R_IParameter(pParameterName,new R_INT(lValue,hValue,lweight,hweight)));
	m_value_counter = m_parameter_list.begin();
      }
    return;
  }
  
  void Format::addRFloatParamValue(const string & pParameterName,
				   float lValue, 
				   float hValue, 
				   float lweight, 
				   float hweight)  throw (ParameterException)
  {
    if ( (lweight < 0) || (lweight > 1))
      throw ParameterExceptionWeightRangeError(const_cast<char *>(pParameterName.c_str()),lweight);
    if ( (hweight < 0) || (hweight > 1))
      throw ParameterExceptionWeightRangeError(const_cast<char *>(pParameterName.c_str()),hweight);

    if (this->getParameter(pParameterName))
      if (!dynamic_cast<R_FParameter *>(this->getParameter(pParameterName)))
	throw ParameterExceptionFalseType(const_cast<char *>(pParameterName.c_str()), 
					  this->getParameter(pParameterName)->getParameterType(), 
					  Parameter::R_FLOAT);

    if ( this->getParameter(pParameterName) )
      {
	this->getParameter(pParameterName)->addValue(new R_FLOAT(lValue,hValue,lweight,hweight));
      }
    else
      {
	m_parameter_list.push_back(new R_FParameter(pParameterName,new R_FLOAT(lValue,hValue,lweight,hweight)));
	m_value_counter = m_parameter_list.begin();
      }
    return;
  }

  void Format::addWildcard(const string & pParameterName)  throw (ParameterException)
  {
    if (this->getParameter(pParameterName))
      if (!dynamic_cast<WParameter *>(this->getParameter(pParameterName)))
	throw ParameterExceptionFalseType(const_cast<char *>(pParameterName.c_str()), 
					  this->getParameter(pParameterName)->getParameterType(), 
					  Parameter::WILDCARD);
    
    if (this->getParameter(pParameterName))
      {
	return;
      }
    else
      {
	m_parameter_list.push_back(new WParameter(pParameterName));
	m_value_counter = m_parameter_list.begin();
      }
    return;
  }

  void Format::showParameter() const
  {
    GLOBAL_DEBUG_STREAM2( "Type/SubType       : " << m_format_type << endl  << "this               : " << this << endl);
    GLOBAL_DEBUG_STREAM2( "number of parameter: " << m_parameter_list.size() << endl);
    if (m_IOPartner != NULL)
      GLOBAL_DEBUG_STREAM2( "IO-Partner         : " << m_IOPartner << endl);
    if (!m_IOPartner_list.empty())
      {
 	format_list::const_iterator itr;
 	for (itr = m_IOPartner_list.begin(); itr != m_IOPartner_list.end(); itr ++)
 	  {
	    GLOBAL_DEBUG_STREAM2( "IO-Partner         : " << *itr << endl);
 	  }
      }
       
    GLOBAL_DEBUG_STREAM2("connected to ext.  : " << m_jack_tag << endl << endl);
     
    parameter_list::const_iterator i;
    for (i = m_parameter_list.begin(); i != m_parameter_list.end(); i++) 
      (*i)->showParameter();
   }
  
  ostream& Format::showParameter(ostream& os) const
  {
    os << "Format(\"" << m_format_type << "\", size(" << m_parameter_list.size() << "), jack_tag(" << m_jack_tag << "), partner(" << m_IOPartner << ")";

    if (!m_IOPartner_list.empty())
      {
        os << endl << "      partner_list(";
	format_list::const_iterator itr;
	for (itr = m_IOPartner_list.begin(); itr != m_IOPartner_list.end(); itr ++)
	  {
            if (itr != m_IOPartner_list.begin()) {
              os << ", ";
            }
	    os << *itr;
	  }
        os << ")";
      }
    os << ") = { ";

    unsigned int count = 1;
    for (parameter_list::const_iterator i = m_parameter_list.begin(); 
         i != m_parameter_list.end(); 
         i++) {
      os << endl << count++ << ". " << **i << " ";
    }

    os << "}";

    return os;
  }
  
  ostream& Format::showParameterVerbose(ostream& os) const
  {
    os << "Type/SubType       : " << m_format_type << endl  << "this               : " << this << endl;
    os << "number of parameter: " << m_parameter_list.size() << endl;
    if (m_IOPartner != NULL)
      os << "IO-Partner         : " << m_IOPartner << endl;
    if (!m_IOPartner_list.empty())
      {
  	format_list::const_iterator itr;
  	for (itr = m_IOPartner_list.begin(); itr != m_IOPartner_list.end(); itr ++)
  	  {
 	    os << "IO-Partner         : " << *itr << endl;
  	  }
      }
    
    os << "connected to ext.  : " << m_jack_tag << endl;
    
    parameter_list::const_iterator i;
    for (i = m_parameter_list.begin(); i != m_parameter_list.end(); i++) 
      (*i)->showParameterVerbose(os);
    
    return os;
  }

  Parameter::PType Format::getParameterType(const string & pParameterName) const throw (ParameterException)
  {
    if (this->getParameter(pParameterName))
      {
	return this->getParameter(pParameterName)->getParameterType();
      }
    throw ParameterExceptionNotExist(const_cast<char *>(pParameterName.c_str()));
  }

  bool Format::isParameterIncluded(const string & pParameterName) const
  {
    if (this->getParameter(pParameterName))
      return true;
    else
      return false;
  }
  
  bool Format::isParameterSpecified(const string & pParameterName) const
  {
    if (!isParameterIncluded(pParameterName)) { return false; }
    if (getParameter(pParameterName)->getParameterType() == Parameter::WILDCARD) { return false; }
    return true;
  }
  
  bool Format::isParamValueIncluded(const string & parameter_name, int32_t value) const 
  {
    return this->isParamValueIncludedInternal(parameter_name, &value);
  }

  bool Format::isParamValueIncluded(const string & parameter_name, float value) const 
  {
    return this->isParamValueIncludedInternal(parameter_name,&value);
  }

  bool Format::isParamValueIncluded(const string & parameter_name, const string & value) const
  {
    if (this->getParameter(parameter_name))
      return this->getParameter(parameter_name)->isValueIncluded(value);
    else
      return false;
  }

  bool Format::isParamValueIncluded(const string & parameter_name, Format * value) const
  {
    return this->isParamValueIncludedInternal(parameter_name, &value);
  }

  bool Format::isParamValueIncludedInternal(const string & parameter_name,void *value) const
  {
    if (this->getParameter(parameter_name))
      {
	return this->getParameter(parameter_name)->isValueIncluded(value);
      }
    else
      return false;
  }

  bool Format::isParamRangeValueIncluded(const string & parameter_name, int32_t l_value, int32_t h_value) const
  {
    if (this->getParameter(parameter_name))
      {
	return this->getParameter(parameter_name)->isValueIncluded(&l_value,&h_value);
      }
    else 
      return false;
  }

  bool Format::isParamRangeValueIncluded(const string & parameter_name, float l_value, float h_value) const
  {
    if (this->getParameter(parameter_name))
    {
      return this->getParameter(parameter_name)->isValueIncluded(&l_value,&h_value);
    }
    else
      return false;
  }

  int32_t Format::getIntValue(const string & parameter_name) const throw (ParameterException)
  {
    if (!this->getParameter(parameter_name))
      throw ParameterExceptionNotExist(const_cast<char *>(parameter_name.c_str()));
    
    if (dynamic_cast<IParameter *>(this->getParameter(parameter_name)))
      return (static_cast<IParameter *>(this->getParameter(parameter_name)))->getDefaultValue();
    if (dynamic_cast<R_IParameter *>(this->getParameter(parameter_name)))
      return (static_cast<R_IParameter *>(this->getParameter(parameter_name)))->getDefaultValue();
    
    throw ParameterExceptionFalseType(const_cast<char *>(parameter_name.c_str()),
				      this->getParameter(parameter_name)->getParameterType(),
				      Parameter::INT32);
  }

  float Format::getFloatValue(const string & parameter_name) const throw (ParameterException)
  {
    if (!this->getParameter(parameter_name))
      throw ParameterExceptionNotExist(const_cast<char *>(parameter_name.c_str()));

    if (dynamic_cast<FParameter *>(this->getParameter(parameter_name)))
      return (static_cast<FParameter *>(this->getParameter(parameter_name)))->getDefaultValue();
    if (dynamic_cast<R_FParameter *>(this->getParameter(parameter_name)))
      return (static_cast<R_FParameter *>(this->getParameter(parameter_name)))->getDefaultValue();

    throw ParameterExceptionFalseType(const_cast<char *>(parameter_name.c_str()),
				      this->getParameter(parameter_name)->getParameterType(),
				      Parameter::FLOAT);

  }

  string Format::getStringValue(const string & parameter_name)  const throw (ParameterException)
  {
    if (!this->getParameter(parameter_name))
      throw ParameterExceptionNotExist(const_cast<char *>(parameter_name.c_str()));

    if (dynamic_cast<SParameter *>(this->getParameter(parameter_name)))
      return (static_cast<SParameter *>(this->getParameter(parameter_name)))->getDefaultValue();

    throw ParameterExceptionFalseType(const_cast<char *>(parameter_name.c_str()),
				      this->getParameter(parameter_name)->getParameterType(),
				      Parameter::STRING);

  }

  R_INT * Format::getRIntValue(const string & parameter_name) const throw (ParameterException)
  {
    if (!this->getParameter(parameter_name))
      throw ParameterExceptionNotExist(const_cast<char *>(parameter_name.c_str()));
    
    if (dynamic_cast<R_IParameter *>(this->getParameter(parameter_name)))
      return (R_INT *)(static_cast<R_IParameter *>(this->getParameter(parameter_name)))->getFirstValue();
    
    throw ParameterExceptionFalseType(const_cast<char *>(parameter_name.c_str()),
				      this->getParameter(parameter_name)->getParameterType(),
				      Parameter::R_INT32);
    
  }

  NMM::R_FLOAT * Format::getRFloatValue(const string & parameter_name) const throw (ParameterException)
  {
    if (!this->getParameter(parameter_name))
      throw ParameterExceptionNotExist(const_cast<char *>(parameter_name.c_str()));

    if (dynamic_cast<R_FParameter *>(this->getParameter(parameter_name)))
      return (NMM::R_FLOAT *)(static_cast<R_FParameter *>(this->getParameter(parameter_name)))->getFirstValue();
    
    throw ParameterExceptionFalseType(const_cast<char *>(parameter_name.c_str()),
				      this->getParameter(parameter_name)->getParameterType(),
				      Parameter::R_FLOAT);

  }

  Format * Format::getFPointerValue(const string & parameter_name) const throw (ParameterException)
  {
    if (!this->getParameter(parameter_name))
      throw ParameterExceptionNotExist(const_cast<char *>(parameter_name.c_str()));
    
    if (dynamic_cast<CParameter *>(this->getParameter(parameter_name)))
      return (Format *)(static_cast<CParameter *>(this->getParameter(parameter_name)))->getFirstValue();

    throw ParameterExceptionFalseType(const_cast<char *>(parameter_name.c_str()),
				      this->getParameter(parameter_name)->getParameterType(),
				      Parameter::FPOINTER);

  }

  void Format::setDefaultValue(const string & parameter_name, int32_t value) const throw (ParameterException)
  {
    if (!this->getParameter(parameter_name))
      throw ParameterExceptionNotExist(const_cast<char *>(parameter_name.c_str()));

    if ( (dynamic_cast<IParameter *>(this->getParameter(parameter_name))))
      {
	(static_cast<IParameter *>(this->getParameter(parameter_name)))->setDefaultValue(value);
	return;
      }
    if (dynamic_cast<R_IParameter *>(this->getParameter(parameter_name)))
      {
	(static_cast<R_IParameter *>(this->getParameter(parameter_name)))->setDefaultValue(value);
	return;
      }
    throw ParameterExceptionFalseType(const_cast<char *>(parameter_name.c_str()),
				      this->getParameter(parameter_name)->getParameterType(),
				      Parameter::INT32);
  }

  void Format::setDefaultValue(const string & parameter_name, float value) const  throw (ParameterException)
  {
    if (!this->getParameter(parameter_name))
      throw ParameterExceptionNotExist(const_cast<char *>(parameter_name.c_str()));

    if (dynamic_cast<FParameter *>(this->getParameter(parameter_name)))
      {
	(static_cast<FParameter *>(this->getParameter(parameter_name)))->setDefaultValue(value);
	return;
      }
    if (dynamic_cast<R_FParameter *>(this->getParameter(parameter_name)))
      {
	(static_cast<R_FParameter *>(this->getParameter(parameter_name)))->setDefaultValue(value);
	return;
      }
    throw ParameterExceptionFalseType(const_cast<char *>(parameter_name.c_str()),
				      this->getParameter(parameter_name)->getParameterType(),
				      Parameter::FLOAT);
  }

  void Format::setDefaultValue(const string & parameter_name, string value) const throw (ParameterException)
  {
    if (!this->getParameter(parameter_name))
      throw ParameterExceptionNotExist(const_cast<char *>(parameter_name.c_str()));

    if (dynamic_cast<SParameter *>(this->getParameter(parameter_name)))
      (static_cast<SParameter *>(this->getParameter(parameter_name)))->setDefaultValue(value);
    throw ParameterExceptionFalseType(const_cast<char *>(parameter_name.c_str()),
				      this->getParameter(parameter_name)->getParameterType(),
				      Parameter::STRING);
  }

  const int32_t_list Format::getIntValueList(const string & parameter_name) const throw (ParameterException)
  {
    if (!this->getParameter(parameter_name))
      throw ParameterExceptionNotExist(const_cast<char *>(parameter_name.c_str()));

    if (dynamic_cast<IParameter *>(this->getParameter(parameter_name)))
      return (dynamic_cast<IParameter *>(this->getParameter(parameter_name)))->getIntValueList();
    throw ParameterExceptionFalseType(const_cast<char *>(parameter_name.c_str()),
				      this->getParameter(parameter_name)->getParameterType(),
				      Parameter::STRING);
  }

  const float_list Format::getFloatValueList(const string & parameter_name) const throw (ParameterException)
  {
    if (!this->getParameter(parameter_name))
      throw ParameterExceptionNotExist(const_cast<char *>(parameter_name.c_str()));

    if (dynamic_cast<FParameter *>(this->getParameter(parameter_name)))
      return (dynamic_cast<FParameter *>(this->getParameter(parameter_name)))->getFloatValueList();
    throw ParameterExceptionFalseType(const_cast<char *>(parameter_name.c_str()),
				      this->getParameter(parameter_name)->getParameterType(),
				      Parameter::FLOAT);
  }

  const string_list Format::getStringValueList(const string & parameter_name) const throw (ParameterException)
  {
    if (!this->getParameter(parameter_name))
      throw ParameterExceptionNotExist(const_cast<char *>(parameter_name.c_str()));

    if (dynamic_cast<SParameter *>(this->getParameter(parameter_name)))
      return (dynamic_cast<SParameter *>(this->getParameter(parameter_name)))->getStringValueList();
    throw ParameterExceptionFalseType(const_cast<char *>(parameter_name.c_str()),
				      this->getParameter(parameter_name)->getParameterType(),
				      Parameter::STRING);
  }

  const r_int32_t_list Format::getRIntValueList(const string & parameter_name) const throw (ParameterException)
  {
    if (!this->getParameter(parameter_name))
      throw ParameterExceptionNotExist(const_cast<char *>(parameter_name.c_str()));

    if (dynamic_cast<R_IParameter *>(this->getParameter(parameter_name)))
      return (dynamic_cast<R_IParameter *>(this->getParameter(parameter_name)))->getRIntValueList();
    throw ParameterExceptionFalseType(const_cast<char *>(parameter_name.c_str()),
				      this->getParameter(parameter_name)->getParameterType(),
				      Parameter::R_INT32);
  }

  const r_float_list Format::getRFloatValueList(const string & parameter_name) const throw (ParameterException)
  {
    if (!this->getParameter(parameter_name))
      throw ParameterExceptionNotExist(const_cast<char *>(parameter_name.c_str()));
    
    if (dynamic_cast<R_FParameter *>(this->getParameter(parameter_name)))
      return (dynamic_cast<R_FParameter *>(this->getParameter(parameter_name)))->getRFloatValueList();
    throw ParameterExceptionFalseType(const_cast<char *>(parameter_name.c_str()),
				      this->getParameter(parameter_name)->getParameterType(),
				      Parameter::R_FLOAT);
  }

  const format_list Format::getFPointerValueList(const string & parameter_name) const throw (ParameterException)
  {
    if (!this->getParameter(parameter_name))
      throw ParameterExceptionNotExist(const_cast<char *>(parameter_name.c_str()));

    if (dynamic_cast<CParameter *>(this->getParameter(parameter_name)))
      return (dynamic_cast<CParameter *>(this->getParameter(parameter_name)))->getFPointerValueList();
    throw ParameterExceptionFalseType(const_cast<char *>(parameter_name.c_str()),
				      this->getParameter(parameter_name)->getParameterType(),
				      Parameter::FPOINTER);
  }

  void Format::setParameterWeight(const string & parameter_name, float weight) const throw (ParameterException)
  {
    if ( (weight < 0) || (weight > 1) )
      throw ParameterExceptionWeightRangeError(const_cast<char *>(parameter_name.c_str()), weight);
    if (!this->getParameter(parameter_name))
      throw ParameterExceptionNotExist(const_cast<char *>(parameter_name.c_str()));
    else
      this->getParameter(parameter_name)->setParameterWeight(weight);
  }

  float Format::getParameterWeight(const string & parameter_name) const throw (ParameterException)
  {
    if (!this->getParameter(parameter_name))
      throw ParameterExceptionNotExist(const_cast<char *>(parameter_name.c_str()));
    else
      return this->getParameter(parameter_name)->getParameterWeight();
  }

  void Format::deleteParameter(const string & parameter_name) throw (ParameterException)
  {
    if (!this->getParameter(parameter_name))
      throw ParameterExceptionNotExist(const_cast<char *>(parameter_name.c_str()));
    
    Parameter * temp = this->getParameter(parameter_name);
    m_parameter_list.remove(this->getParameter(parameter_name));
    delete temp;
  }

  const parameter_list & Format::getParameterList() const
  {
    return m_parameter_list;
  }
  
  Parameter * Format::getParameter(const string & pParameterName) const
  {
    parameter_list::const_iterator itr;
    for (itr = m_parameter_list.begin(); itr != m_parameter_list.end(); itr++) 
      {
	if ((*itr)->getParameterName() == pParameterName) 
	  {
	    return (*itr);
	  }
      }
    return NULL;
  }

  void Format::addParameter(Parameter *param) 
  {
    m_parameter_list.push_back(param);
    m_value_counter = m_parameter_list.begin();
    return;
  }

  Parameter * Format::duplicateParameter(const string & parameter_name) const
  {
    if (this->getParameter(parameter_name) == NULL) 
      {
	return NULL;
      }
    else 
      {
	return this->getParameter(parameter_name)->duplicate();
      }
  }

  Format * Format::duplicateFormat() const
  {
    Format *new_format = new Format(m_format_type, m_jack_tag);
    parameter_list::const_iterator itr;
    for (itr = m_parameter_list.begin(); itr != m_parameter_list.end(); itr ++)
      {
	new_format->addParameter((*itr)->duplicate());
      }
    return new_format;
  }

  /*27.07.2002 Create a deep copy for the IOPartner*/
  void Format::setIOPartner(const Format *pFormat) 
  {
    delete m_IOPartner;
    if (pFormat) {
      m_IOPartner = new Format(*pFormat);
      m_IOPartner->setIOPartner(0); //avoid recursion
    }
    else
      m_IOPartner = 0;
  }

  const Format * Format::getIOPartner() const 
  {
    return m_IOPartner;
  }

  Format * Format::getIOPartner()
  {
    return m_IOPartner;
  }

  void Format::addIOPartnerInList(Format *pFormat) 
  {
    m_IOPartner_list.push_back(pFormat);
  }
  
  const format_list & Format::getIOPartnerList() const
  {
    return m_IOPartner_list;
  }

  bool Format::isParameterIdentical(const string & parameter_name) const
  {
    // if the IOPartner of this format is empty, the parameter is not mutable.
    if (m_IOPartner == NULL )
      return false;
    // if the parameter is not included in the list, the parameter is not mutable.
    if (!this->isParameterIncluded(parameter_name))
      return false;
    // if the parameter is not included in the list of the IOPartner, the parameter is not mutable.
    if (!m_IOPartner->isParameterIncluded(parameter_name))
      return false;
    if ( this ->getParameter(parameter_name)->getNumberOfValues() != m_IOPartner->getParameter(parameter_name)->getNumberOfValues() )
      return false;
    else
      {
	return this->getParameter(parameter_name)->isParameterIdentical(m_IOPartner->getParameter(parameter_name));
      }
  }

  bool Format::isFormatMatching(const Format * reference_format, bool compare_type_subtype) const
  {
    // if this format or reference_format has type */* then the formats are matching
    if((reference_format->getFormatType() == "*/*") ||
       (getFormatType() == "*/*")) {
      return true;
    }

    // alse check type/subtype ? (default is true)
    if (compare_type_subtype)
      if ( reference_format->getFormatType() != m_format_type)
	return false;
    
    // the number of values ( = number of parameters ) for the two formats has to be the same
    parameter_list::const_iterator itr;
    if (reference_format->getNumberOfValues() != this->getNumberOfValues())
      return false;

    // then, for each parameter ...
    for (itr = m_parameter_list.begin(); itr != m_parameter_list.end(); itr++)
      {
	// check if a parameter with the same name exists in the other format ...
	if ( reference_format->getParameter((*itr)->getParameterName()) != NULL)
	  {
	    // and if at least one value of the parameter exists in both formats ... 
	    // (if not, return false)
	    if (!(*itr)->isParameterMatching(reference_format->getParameter((*itr)->getParameterName())))
	      return false;
	  }
	else
	  return false;
      }
    return true;
  }

  bool Format::isFormatIncluded(const Format *reference_format) const
  {
    // if this format or reference_format has type */* then the formats are matching
    if((reference_format->getFormatType() == "*/*") ||
       (getFormatType() == "*/*")) {
      return true;
    }

    // check type/subtype 
    if (!reference_format->getFormatType().empty()) {
      if(reference_format->getFormatType() != m_format_type) {
        return false;
      }
    }

    /*Begin> all parameters of the reference_format have to 'match' the parameters in this format */
    for (parameter_list::const_iterator itr = reference_format->getParameterList().begin(); 
	 itr != reference_format->getParameterList().end(); 
	 ++itr)
    {
      if (getParameter((*itr)->getParameterName()) == NULL)
        return false;

      if (!(*itr)->isParameterMatching(getParameter((*itr)->getParameterName()))) {
	// GLOBAL_ERROR_STREAM("Parameter "<< (*itr)->getParameterName() <<  " mismatch" << endl);
        return false;
      }
    }

    return true;
  }

  bool Format::isFormatIdentical(const Format *reference_format) const
  {
    if (m_format_type != reference_format->getFormatType())
      return false;
    if (this->getNumberOfValues() != reference_format->getNumberOfValues() )
      return false;
    else
      {
	parameter_list::const_iterator itr;
	for (itr = m_parameter_list.begin(); itr != m_parameter_list.end(); itr ++)
	  {
	    if ( reference_format->getParameter((*itr)->getParameterName()) == NULL)
	      return false;
	    if (! (*itr)->isParameterIdentical(reference_format->getParameter((*itr)->getParameterName())))
	      return false;
	  }
      }
    return true;
  }
  
  const Parameter * Format::getNextParameter() const
  {
    parameter_iterator temp;
    if (m_value_counter != m_parameter_list.end())
      {
	temp = m_value_counter;
	m_value_counter++;
	if (!dynamic_cast<Parameter *>(*temp))
	  return NULL;
	else
	  return *temp;
      }
    else 
      {
	m_value_counter = m_parameter_list.begin();
	return NULL;
      }
  }
  
  int Format::getCrossProduct() const
  {
    int result = 1;
    parameter_list::const_iterator itr;
    for (itr = m_parameter_list.begin(); itr != m_parameter_list.end(); itr ++)
      {
	if ( this->isParameterIdentical((*itr)->getParameterName()) )
	  result = result * (*itr)->getNumberOfValues();
      }
    return result;
  }

  uint Format::getNumberOfValues() const
  {
    return m_parameter_list.size();
  }

  void Format::resetValueCounter() const
  {
    m_value_counter = m_parameter_list.begin();
  }

  const parameter_list Format::getLostParameter() const 
  {
    parameter_list temp_parameter_list;
    parameter_list io_list = m_IOPartner->getParameterList();
    
    for (parameter_iterator itr = io_list.begin(); itr != io_list.end(); itr ++)
      {
	if (this->getParameter( (*itr)->getParameterName()) == NULL)
	  {
	    temp_parameter_list.push_back( (*itr));
	  }
      }
    return temp_parameter_list;
  }

  Format* Format::createCheckedIntersectionFormat(const Format* reference_format) const
  {
    // returns copy of this Format object if 0 was given
    if(reference_format == 0) {
      return (new Format(*this));
    }
    else {
      return createIntersectionFormat(reference_format);
    }
  }

  Format* Format::createIntersectionFormat(const Format* reference_format) const
  {
    // added extra check
    if(reference_format == 0) {
      return 0;
    }

    // if reference format has type */* this format is copied and returned
    if(reference_format->getFormatType() == "*/*") {
      return (new Format(*this));
    }

    // if this format has type */* the reference format is copied and returned
    if(getFormatType() == "*/*") {
      return (new Format(*reference_format));
    }

    if ( (!this->isFormatMatching(reference_format)) && (!reference_format->isFormatMatching(this)) )
      return NULL;

    // hmm, this should not be necessary, ...
    if ( (!this->isFormatMatching(reference_format)) )
      return reference_format->createIntersectionFormat(this);
	 

    Format * new_format = new Format(m_format_type);
    parameter_list::const_iterator  itr;
    
    for (itr = m_parameter_list.begin(); itr != m_parameter_list.end() ; itr ++)
      {
	new_format->addParameter((*itr)->getIntersection(reference_format->getParameter((*itr)->getParameterName())));
      }

    if (this->isValidFormat(new_format))
      return new_format;
    else 
      {
	delete new_format;
	return NULL;
      }
  }

  // ****************
  Format* Format::createBestIntersectionFormat(const Format* subset_format, unsigned int& param_counter) const
  {
    // counts, how many parameter values of subset_format can be filled into intersection format
    param_counter = 0;

    // added extra check
    if(subset_format == 0) {
      return 0;
    }

    // if subset format has specified type and type is not */* -> check types
    if((subset_format->getFormatType() != "") && 
       (subset_format->getFormatType() != "*/*")) {
      // in this case types have to be equal
      if(subset_format->getFormatType() != getFormatType()) {
	return 0;
      }
    }

    // in this case, format type of this format is taken (whether or not it is equal to the subset_format)
    Format* new_format = new Format(getFormatType());

    // for all parameters in this format (since this format represents the format that is stored in the property)
    for(parameter_list::const_iterator itr = getParameterList().begin(); itr != getParameterList().end(); ++itr) {
      
      // try to get parameter from subset format -> try to create intersection
      Parameter* p = subset_format->getParameter((*itr)->getParameterName());
      if(p) {
	if(p->isParameterMatching(*itr)) {
	  new_format->addParameter(p->getIntersection(*itr));
	  ++param_counter; 
	}
	else { // same parameter, no intersection -> copy parameter from this format to new format 
	  new_format->addParameter((*itr)->duplicate());
	}
      }
      else { // parameter is not in subset format -> copy parameter from this format to new format
	new_format->addParameter((*itr)->duplicate());
      }

    }

    // valid format ( = contains values for all parameters) ?
    if (this->isValidFormat(new_format)) {
      return new_format;
    }
    else {
      param_counter = 0;
      delete new_format;
      return 0;
    }
  }

  Format* Format::createSubsetIntersectionFormat(const Format* subset_format) const
  {

    // added extra check
    if(subset_format == 0) {
      return 0;
    }

    // if subset format has specified type and type is not */* -> check types
    if((subset_format->getFormatType() != "") && 
       (subset_format->getFormatType() != "*/*")) {
      // in this case types have to be equal
      if(subset_format->getFormatType() != getFormatType()) {
	return 0;
      }
    }

    Format* new_format = new Format(getFormatType());
    unsigned param_counter = 0;

    // for all parameters in this format (since this format represents the superset)
    for(parameter_list::const_iterator itr = getParameterList().begin(); itr != getParameterList().end(); ++itr) {
      
      // try to get parameter from subset format -> try to create intersection
      Parameter* p = subset_format->getParameter((*itr)->getParameterName());
      if(p) {
	if(p->isParameterMatching(*itr)) {
	  new_format->addParameter(p->getIntersection(*itr));
	  ++param_counter; 
	}
	else { // same parameter, no intersection 
	  delete new_format;
	  return 0;
	}
      }
      else { // parameter is not in subset format -> copy parameter from this format to new format
	new_format->addParameter((*itr)->duplicate());
      }

    }

    // are all parameters of subset format in this format ?
    if(param_counter != subset_format->getParameterList().size()) {
      delete new_format;
      return 0;
    }

    // valid format ( = contains values for all parameters) ?
    if (this->isValidFormat(new_format)) {
      return new_format;
    }
    else {
      delete new_format;
      return 0;
    }
  }

  bool Format::isValidFormat(const Format * reference_format) const
  {
    parameter_list temp_parameter_list = reference_format->getParameterList();
    parameter_list::const_iterator itr;
    for (itr = temp_parameter_list.begin(); itr != temp_parameter_list.end(); itr ++)
      {
	if ( (*itr)->getNumberOfValues() == 0)
	  return false;
      }
    return true;
  }

  void Format::setJackTag(const string & param)
  {
    m_jack_tag = param;
  }

  const string & Format::getJackTag() const
  {
    return m_jack_tag;
  }

  bool Format::hasJackTag() const
  {
    return !m_jack_tag.empty();
//    if (m_jack_tag == "default")
//      return false;
//    else
//      return true;
  }

  float Format::getStandardizedWeight() const
  {
    float local_weight = 0;
    parameter_list::const_iterator itr;
    for (itr = m_parameter_list.begin(); itr != m_parameter_list.end(); itr ++)
      {
	local_weight += (*itr)->getStandardizedWeight();
      }
    return (local_weight/m_parameter_list.size());
  }

  void Format::buildOptimizedFormat()
  {
    for (parameter_list::const_iterator itr = m_parameter_list.begin(); itr != m_parameter_list.end(); itr++)
      {
	if ( (*itr)->getParameterType() == Parameter::R_INT32)
	  {
	    (dynamic_cast<R_IParameter *>(*itr))->setDefaultValue((dynamic_cast<R_IParameter *>(*itr))->getHighestValue());
	  }
	if ( (*itr)->getParameterType() == Parameter::R_FLOAT)
	  {
	    (dynamic_cast<R_IParameter *>(*itr))->setDefaultValue((dynamic_cast<R_IParameter *>(*itr))->getHighestValue());
	  }
      }
  }

  bool Format::hasFormatRanges() const
  {
    for (parameter_list::const_iterator itr = m_parameter_list.begin(); itr != m_parameter_list.end(); itr++)
      {
	if ( ((*itr)->getParameterType() == Parameter::R_INT32) || ((*itr)->getParameterType() == Parameter::R_FLOAT) ) 
	  return true;
      }
    return false;
  }

  bool Format::hasFormatWildcards() const
  {
    for (parameter_list::const_iterator itr = m_parameter_list.begin(); itr != m_parameter_list.end(); itr ++)
      {
	if ( (*itr)->getParameterType() == Parameter::WILDCARD)
	  return true;
      }
    return false;
  }

  void Format::actualizeFormat(const Format * format)
  {
    list <parameter_list::iterator> delete_list;
    
    for (parameter_list::iterator itr = m_parameter_list.begin(); itr != m_parameter_list.end(); itr ++)
      {
	if (format->getParameter((*itr)->getParameterName()) != NULL)
	  delete_list.push_back(itr);
      }
    for (list<parameter_list::iterator>::iterator itr = delete_list.begin(); itr != delete_list.end(); itr ++)
      {
	m_parameter_list.push_back( (**itr)->getIntersection(format->getParameter((**itr)->getParameterName())));
      }

    for (list<parameter_list::iterator>::iterator itr = delete_list.begin(); itr != delete_list.end(); itr++)
      {
	delete **itr;
	m_parameter_list.erase(*itr);
      }
    m_value_counter = m_parameter_list.begin();
  }

  void Format::resolveWildcards(const Format * format)
  {
    list <parameter_list::iterator> delete_list;

    // find all parameters with wildcards in own parameters, and put them to delete_list
    for (parameter_list::iterator itr = m_parameter_list.begin(); itr != m_parameter_list.end(); itr ++)
      {
	if ( (*itr)->getParameterType() == Parameter::WILDCARD)
	  {
	    delete_list.push_back(itr);
	  }
      }

    // for all these parameters in delete_list, copy the parameter of the given format to the end
    // of own parameter list (idea: the parameters of the given format are not wildcard, and match!)
    for (list<parameter_list::iterator>::iterator itr = delete_list.begin(); itr != delete_list.end(); itr++)
      {
	m_parameter_list.push_back( format->duplicateParameter((**itr)->getParameterName()));
      }
    
    // delete occurences of parameters with wildcard, this will leave the new entries (those without wildcard!)
    for (list<parameter_list::iterator>::iterator itr = delete_list.begin(); itr != delete_list.end(); itr++)
      {
	m_parameter_list.erase(*itr);
      }
//     cout << *this << endl;
  }

  void Format::resolveMatchingWildcards(const Format * format)
  {
    list <parameter_list::iterator> delete_list;

    // find all parameters with wildcards in own parameters that also exist in given format, and put them to delete_list
    for (parameter_list::iterator itr = m_parameter_list.begin(); itr != m_parameter_list.end(); itr ++)
      {
	if ( ((*itr)->getParameterType() == Parameter::WILDCARD) && (format->isParameterIncluded((*itr)->getParameterName())) )
	  {
	    delete_list.push_back(itr);
	  }
      }

    // for all these parameters in delete_list, copy the parameter of the given format to the end
    // of own parameter list (idea: the parameters of the given format are not wildcard, and match!)
    for (list<parameter_list::iterator>::iterator itr = delete_list.begin(); itr != delete_list.end(); itr++)
      {
	m_parameter_list.push_back( format->duplicateParameter((**itr)->getParameterName()));
      }
    
    // delete occurences of parameters with wildcard, this will leave the new entries (those without wildcard!)
    for (list<parameter_list::iterator>::iterator itr = delete_list.begin(); itr != delete_list.end(); itr++)
      {
	m_parameter_list.erase(*itr);
      }
//     cout << *this << endl;
  }

  parameter_list Format::getParameterList(Parameter::PType type) const
  {
    parameter_list result;
    for (parameter_iterator itr = m_parameter_list.begin(); itr != m_parameter_list.end(); itr ++)
      {
	if ( (*itr)->getParameterType() == type)
	  result.push_back(*itr);
      }
    return result;
  }

  void Format::setFormatID(const string & ID)
  {
    m_format_ID = ID;
  }

  const string & Format::getFormatID() const
  {
    return m_format_ID;
  }

  bool Format::fulfilConstraints(const Format * constraint) const
  {
    for (parameter_list::const_iterator itr = (constraint->getParameterList()).begin(); 
	 itr != (constraint->getParameterList()).end(); itr++)
      {
	if (dynamic_cast<IParameter *>(*itr))
	  if (!this->isParamValueIncluded((*itr)->getParameterName(), 
					  dynamic_cast<IParameter *>(*itr)->getDefaultValue()))
	    return false;
	if (dynamic_cast<R_IParameter *>(*itr))
	  if (!this->isParamRangeValueIncluded((*itr)->getParameterName(), 
					       dynamic_cast<R_IParameter *>(*itr)->getDefaultRange()->getLowValue(),
					       dynamic_cast<R_IParameter *>(*itr)->getDefaultRange()->getHighValue()))
	    return false;
	if (dynamic_cast<FParameter *>(*itr))
	  if (!this->isParamValueIncluded((*itr)->getParameterName(),
					  dynamic_cast<FParameter *>(*itr)->getDefaultValue()))
	    return false;
	if (dynamic_cast<R_FParameter *>(*itr))
	  if (!this->isParamRangeValueIncluded((*itr)->getParameterName(),
					       dynamic_cast<R_FParameter *>(*itr)->getDefaultRange()->getLowValue(),
					       dynamic_cast<R_FParameter *>(*itr)->getDefaultRange()->getHighValue()))
	    return false;
	if (dynamic_cast<SParameter *>(*itr))
	  if (!this->isParamValueIncluded((*itr)->getParameterName(),
					  dynamic_cast<SParameter *>(*itr)->getDefaultValue()))
	    return false;
      }
    return true;
  }

  format_list Format::getUniqueFormatListSingle() const
  {
    // result
    format_list result;

    //calculate cross product
    int cp = 1;
    parameter_list::const_iterator itr;
    for (itr = m_parameter_list.begin(); itr != m_parameter_list.end(); itr ++)
      {
	cp = cp * (*itr)->getNumberOfValues();
      }

    parameter_value_list        m_parameter_value_list;
    bool running              = true;
    Parameter *temp_parameter = NULL;
    void *tempvalue;

    // fill the list with the parameters from the reference_format and NULL as value.
    parameter_list parameterlist = this->getParameterList();
    for (parameter_iterator itr = parameterlist.begin(); itr != parameterlist.end(); itr ++)
      {
	Parameter_Value_Assignment *temp_pva = new Parameter_Value_Assignment;
	temp_pva->param = (*itr);
	temp_pva->value = NULL;
	temp_pva->move = true;
	m_parameter_value_list.push_back(temp_pva);
      }
    
    bool first = true;

    for (int i = 1; i <= cp; i++)
      {
	Format *new_format = new Format(this->getFormatType());
	new_format->setJackTag(this->getJackTag());

	while (running)
	  {
	    temp_parameter = const_cast<Parameter *>(this->getNextParameter());
	    if ( (temp_parameter == NULL)) // || ( !dynamic_cast<Parameter *>(const_cast<Parameter *>(this->getNextParameter()))))
	      running = false;
	    else
	      // iterate all the parameter once for one new format
	      // build the crossproduct with reference_format parameter-values
// 	      cout << *temp_parameter << endl;

	      {
		for (parameter_value_iterator pvi = m_parameter_value_list.begin(); pvi != m_parameter_value_list.end(); pvi++)
		  {
		    if ( (*pvi)->param == temp_parameter )
		      {
			if (first)
			  {
			    // 			    	GLOBAL_DEBUG_STREAM2( "ruecke immer 1. weiter" << endl);
			    (*pvi)->move = true;
			    first = false;
			  }
			if ( (*pvi)->move == true )
			  //if ( (*pvi)->value == NULL ) // at the end of an iterationstep
			  {
			    tempvalue = temp_parameter->nextValue();
			    (*pvi)->move = false;
			    if (tempvalue == NULL)
			      // start working on the following
			      {
				parameter_value_iterator temp = pvi;
				temp ++;
				if (temp != m_parameter_value_list.end())
				  {
				    (*temp)->move = true;
				  }
				tempvalue = temp_parameter->nextValue();
			      }
			    (*pvi)->value = tempvalue;
			  }
			else
			  {
			    tempvalue = (*pvi)->value;
			  }
			switch (temp_parameter->getParameterType())
			  {
			  case Parameter::INT32:
			    {
			      new_format->addParameter(new IParameter(temp_parameter->getParameterName(),
								      *(int32_t *)tempvalue,
								      temp_parameter->getWeight(tempvalue)));
			      break;
			    }
			  case Parameter::R_INT32:
			    {
			      new_format->addParameter(new R_IParameter(temp_parameter->getParameterName(),tempvalue));
			      break;
			    }
			  case Parameter::FLOAT:
			    {
			      new_format->addParameter(new FParameter(temp_parameter->getParameterName(),
								      *(float *)tempvalue, 
								      temp_parameter->getWeight(tempvalue)));
			      break;
			    }
			  case Parameter::R_FLOAT:
			    {
			      new_format->addParameter(new R_FParameter(temp_parameter->getParameterName(), tempvalue));
			      break;
			    }
			  case Parameter::STRING:
			    {
			      new_format->addParameter(new SParameter(temp_parameter->getParameterName(), 
								      *(string *)tempvalue, 
								      temp_parameter->getWeight(tempvalue)));
			      break;
			    }
			  case Parameter::WILDCARD:
			    {
			      new_format->addParameter(new WParameter(temp_parameter->getParameterName()));
			      break;
			    }
			  case Parameter::FPOINTER:
			    {
			      new_format->addParameter(temp_parameter->duplicate());
			    }
			  default:;
			  }
		      }
		  }
	      }
	    temp_parameter = NULL;
	  }
	result.push_back(new_format);
	first = true;
	running = true;
	
      }
    return result;
  }
  
  format_list Format::getUniqueFormatList(Format * reference_format) const
  {
    format_list result;
    parameter_value_list m_parameter_value_list;
    bool running         = true;
    bool include_io_format = false;
    bool include_format    = true;
    Parameter *temp_parameter = NULL;
    void *tempvalue;

    if (reference_format == NULL)
      {
	reference_format = const_cast<Format *>(this);
	include_io_format = true;
      }
    // fill the list with the parameters from the reference_format and NULL as value.
    parameter_list parameterlist = reference_format->getParameterList();
    for (parameter_iterator itr = parameterlist.begin(); itr != parameterlist.end(); itr ++)
      {
	Parameter_Value_Assignment *temp_pva = new Parameter_Value_Assignment;
	temp_pva->param = (*itr);
	temp_pva->value = NULL;
	temp_pva->move = true;
	//temp_pva->value = (*itr)->nextValue();
	m_parameter_value_list.push_back(temp_pva);
      }
    
    bool first = true;
    int limit = reference_format->getCrossProduct();

    for (int i = 1; i <= limit; i++)
      {
	Format *new_format = new Format(reference_format->getFormatType());
	new_format->setJackTag(reference_format->getJackTag());

	Format * new_io_format = NULL;
	new_io_format = new Format(reference_format->getIOPartner()->getFormatType());
	new_io_format->setJackTag(reference_format->getIOPartner()->getJackTag());

	while (running)
	  {
	    temp_parameter = const_cast<Parameter *>(reference_format->getNextParameter());
	    if (temp_parameter == NULL)
	      running = false;
	    else
	      // iterate all the parameter once for one new format
	      {
// 		temp_parameter->showParameter();
 		if (reference_format->isParameterIdentical(temp_parameter->getParameterName()))
		  // build the crossproduct with reference_format parameter-values
		  {
//   		    GLOBAL_DEBUG_STREAM2( "parameter " << temp_parameter->getParameterName() << " is identical" << endl);
		    for (parameter_value_iterator pvi = m_parameter_value_list.begin(); pvi != m_parameter_value_list.end(); pvi++)
		      {
			if ( (*pvi)->param == temp_parameter )
			  {
			    if (first)
			      {
// 			    	GLOBAL_DEBUG_STREAM2( "ruecke immer 1. weiter" << endl);
				(*pvi)->move = true;
			    	first = false;
			     }
			    if ( (*pvi)->move == true )
			      //if ( (*pvi)->value == NULL ) // at the end of an iterationstep
			      {
				tempvalue = temp_parameter->nextValue();
				(*pvi)->move = false;
				if (tempvalue == NULL)
				  // start working on the following
				  {
				    parameter_value_iterator temp = pvi;
				    temp ++;
				    if (temp != m_parameter_value_list.end())
				      {
					(*temp)->move = true;
				      }
				    tempvalue = temp_parameter->nextValue();
				  }
				(*pvi)->value = tempvalue;
			      }
			    else
			      {
				tempvalue = (*pvi)->value;
			      }
			    switch (temp_parameter->getParameterType())
			      {
			      case Parameter::INT32:
				{
				  new_format->addParameter(new IParameter(temp_parameter->getParameterName(),
									  *(int32_t *)tempvalue,
									  temp_parameter->getWeight(tempvalue)));
				  new_io_format->addParameter(new IParameter(temp_parameter->getParameterName(),
									     *(int32_t *)tempvalue,
									     reference_format->getParameter(temp_parameter->getParameterName())->getWeight(tempvalue)));
				  break;
				}
			      case Parameter::R_INT32:
				{
				  new_format->addParameter(new R_IParameter(temp_parameter->getParameterName(),tempvalue));
				  new_io_format->addParameter(new R_IParameter(temp_parameter->getParameterName(), tempvalue));
				  break;
				}
			      case Parameter::FLOAT:
				{
				  new_format->addParameter(new FParameter(temp_parameter->getParameterName(),
									  *(float *)tempvalue, 
									  temp_parameter->getWeight(tempvalue)));
				  new_io_format->addParameter(new FParameter(temp_parameter->getParameterName(),
									     *(float *)tempvalue, 
									     reference_format->getParameter(temp_parameter->getParameterName())->getWeight(tempvalue)));
				  break;
				}
			      case Parameter::R_FLOAT:
				{
				  new_format->addParameter(new R_FParameter(temp_parameter->getParameterName(), tempvalue));
				  new_io_format->addParameter(new R_FParameter(temp_parameter->getParameterName(), tempvalue));
				  break;
				}
			      case Parameter::STRING:
				{
				  new_format->addParameter(new SParameter(temp_parameter->getParameterName(), 
									  *(string *)tempvalue, 
									  temp_parameter->getWeight(tempvalue)));
				  new_io_format->addParameter(new SParameter(temp_parameter->getParameterName(), 
									     *(string *)tempvalue, 
									     reference_format->getParameter(temp_parameter->getParameterName())->getWeight(tempvalue)));
				  break;
				}
			      case Parameter::WILDCARD:
				{
				  new_format->addParameter(new WParameter(temp_parameter->getParameterName()));
				  new_io_format->addParameter(new WParameter(temp_parameter->getParameterName()));
				  break;
				}
			      case Parameter::FPOINTER:
				{
				  new_format->addParameter(temp_parameter->duplicate());
				  new_io_format->addParameter(temp_parameter->duplicate());
				}
			      default:;
			      }
// 			    break;
			  }
		      }
		  }
		else
		  // just duplicate the parameter and set the following entry of this parameter to move = true
		  {
		    new_format->addParameter(temp_parameter->duplicate());
		    if (reference_format->getIOPartner()->getParameter(temp_parameter->getParameterName()) != NULL)
		      new_io_format->addParameter(reference_format->getIOPartner()->getParameter(temp_parameter->getParameterName())->duplicate());
		    for (parameter_value_iterator pvi = m_parameter_value_list.begin(); pvi != m_parameter_value_list.end(); pvi ++)
		      {
			if ( ((*pvi)->param == temp_parameter) && (pvi != (--m_parameter_value_list.end()) ))
			  {
			    if ( (*pvi)->move == true )
			      {
				(*pvi)->move = false;
				parameter_value_iterator temp = pvi;
				temp ++;
				(*temp)->move = true;
			      }
			  }
		      }
		    
		  }
	      }
	    temp_parameter = NULL;
	  }
	// There is a problem with the parameter from the IOPartner which have no corresponding partner
	// in the IOPartner. Up till now, these parameters will be deleted. To solve the problem, we take
	// them back and add them like the "mutable" ones. That means they are not part of the cross product.
	parameter_list lost_parameter_list = reference_format->getLostParameter();
	if (!lost_parameter_list.empty())
	  {
	    for (parameter_iterator tempitr = lost_parameter_list.begin(); tempitr != lost_parameter_list.end(); tempitr ++)
	      {
		new_io_format->addParameter( (*tempitr)->duplicate());
	      }
	  }
	first = true;
	running = true;

	new_format->setIOPartner(new_io_format);
	new_io_format->setIOPartner(new_format);
	
	int _case = 1;
	
	format_list io_list;
	format_list io_partner_list;

	if (!new_format->isFormatUnique()) 
	  {
	    new_format->setIOPartner(new_format);
	    io_list = this->getUniqueFormatList(new_format);
	    _case = 2;
	  }
	if (!new_io_format->isFormatUnique()) 
	  {
	    new_io_format->setIOPartner(new_io_format);
	    io_partner_list = this->getUniqueFormatList(new_io_format);
	    if (_case ==2)
	      _case = 3; // all two possibilities
	    else
	      _case = 4; // only io_list;
	  }
	
	switch (_case)
	  {
	    case 1:
	      {
		break;
	      }
	      case 2:
	      {
		for (format_list::iterator itr = io_list.begin(); itr != io_list.end(); itr ++)
		  {
		    Format * temp = (*itr)->duplicateFormat();
		    Format * temp2= (new_io_format)->duplicateFormat();
		    temp2->setIOPartner(temp);
		    temp->setIOPartner(temp2);
		    result.push_back(temp2);
		    result.push_back(temp);
		    include_io_format = false;
		  }
		break;
	      }
	      case 3:
	      {
		for (format_list::iterator io_itr = io_list.begin(); io_itr != io_list.end(); io_itr ++)
		  {
		    for (format_list::iterator io_p_itr = io_partner_list.begin(); io_p_itr != io_partner_list.end(); io_p_itr++)
		      {
			Format * temp1 = (*io_itr)->duplicateFormat();
			Format * temp2 = (*io_p_itr)->duplicateFormat();
			temp1->setIOPartner(temp2);
			temp2->setIOPartner(temp1);
 			result.push_back(temp1);
 			result.push_back(temp2);
		      }
		  }
		break;
	      }
	  case 4:
	    {
	      for (format_list::iterator itr = io_partner_list.begin(); itr != io_partner_list.end() ; itr ++)
		{
		  Format * temp = (*itr)->duplicateFormat();
		  Format * temp2 = new_format->duplicateFormat();
		  temp->setIOPartner(temp2);
		  temp2->setIOPartner(temp);
		  result.push_back(temp2);
		  result.push_back(temp);
		  include_io_format = false;
		  include_format = false;
		}
	      break;
	    }
	  default:;
	  }

	if (include_format)
	  result.push_back(new_format);
	else
	  delete new_format;
	if (include_io_format)
	  result.push_back(new_io_format);
	else
	  delete new_io_format;
      }
    list < format_list::iterator > delete_list;
    for (format_list::iterator itr = result.begin(); itr != result.end(); itr ++)
      {
	if (!(*itr)->isFormatUnique())
	  (delete_list.push_back(itr));
      }
    for (list< format_list::iterator >::iterator itr = delete_list.begin(); itr != delete_list.end(); itr ++)
      {
	delete **itr;
	result.erase(*itr);
      }
    return result;
    
  }

  parameter_list Format::getSplittedParameterList() const
  {
    parameter_list result;
    for (parameter_list::const_iterator itr = m_parameter_list.begin(); itr != m_parameter_list.end(); itr ++)
      if ( (*itr)->getNumberOfValues() > 1)
	result.push_back(*itr);
    return result;
  }

  bool Format::isFormatUnique() const
  {
    for (parameter_list::const_iterator itr = m_parameter_list.begin(); itr != m_parameter_list.end(); itr ++)
      {
	if ( (*itr)->getNumberOfValues() > 1)
	  return false;
      }
    return true;
    
  }
  
  bool operator < (const Format& f1, const Format& f2)
  {
    if(f1.getStandardizedWeight() < f2.getStandardizedWeight()) {
      return true;
    }
    else {
      return false;
    }
  }

  bool operator == (const Format & lhs, const Format & rhs)
  {
    return lhs.isFormatIdentical(&rhs);
  }

  const Format & Format::operator = (const Format & ref)
  {
    // Assignment to itself
    if (this == &ref)
      return *this;

    erase(); //clears also the m_parameter list

    m_format_type = ref.getFormatType();

    m_format_ID   = ref.getFormatID();

    setIOPartner(ref.getIOPartner());

    for(format_list::const_iterator i = ref.getIOPartnerList().begin(); i != ref.getIOPartnerList().end(); ++i) {
      addIOPartnerInList(new Format(**i));
    }

    m_jack_tag    = ref.getJackTag();

    for (parameter_iterator itr = ref.m_parameter_list.begin(); itr != ref.m_parameter_list.end(); itr ++)
    {
      this->addParameter((*itr)->duplicate());
    }

    m_value_counter = m_parameter_list.begin(); //Is this ok?
    return *this;
  }

  ostream & operator << (ostream & os, const Format & format)
  {
    return format.showParameter(os);
  }


  void Format::createParameter(IValueStream& iv) {
    while(!iv.isEmpty()) {
      if (iv.getCurrentType() == "IParameter" && iv.getCurrentNamespace() == "NMM") {
	IParameter* param = new IParameter();
	param->fromStream(iv);
	addParameter(param);
	continue;
      }

      if (iv.getCurrentType() == "SParameter" && iv.getCurrentNamespace() == "NMM") {
	SParameter* param = new SParameter();
	param->fromStream(iv);
	addParameter(param);
	continue;
      }

      if (iv.getCurrentType() == "FParameter" && iv.getCurrentNamespace() == "NMM") {
	FParameter* param = new FParameter();
	param->fromStream(iv);
	addParameter(param);
	continue;
      }

      if (iv.getCurrentType() == "R_FParameter" && iv.getCurrentNamespace() == "NMM") {
	R_FParameter* param = new R_FParameter();
	param->fromStream(iv);
	addParameter(param);
	continue;
      }

      if (iv.getCurrentType() == "R_IParameter" && iv.getCurrentNamespace() == "NMM") {
	R_IParameter* param = new R_IParameter();
	param->fromStream(iv);
	addParameter(param);
	continue;
      }

      if (iv.getCurrentType() == "CParameter" && iv.getCurrentNamespace() == "NMM") {
	CParameter* param = new CParameter();
	param->fromStream(iv);
	addParameter(param);
	continue;
      }

      if (iv.getCurrentType() == "WParameter" && iv.getCurrentNamespace() == "NMM") {
	WParameter* param = new WParameter();
	param->fromStream(iv);
	addParameter(param);
	continue;
      }

      if (iv.getCurrentType() == "IOPartner") {
	if (!iv.goDown()) {
	  GLOBAL_WARNING_STREAM("No IOPartner available" );
	  iv.nextElement();
	  continue;
	}

	Format* format = new Format();
	iv >> *format;
	setIOPartner(format);
  delete format;
	iv.goUp();

	continue;
      }
      
      GLOBAL_ERROR_STREAM2("Format: Unkown Type " << iv.getCurrentType() << endl);
      iv.nextElement();
    }
  }

  IValueStream& operator>> (IValueStream& iv, Format& format) {
    if ( (iv.getCurrentType() != "Format") || (iv.getCurrentNamespace() != "NMM") ) {
      GLOBAL_ERROR_STREAM ("Invalid type or namespace:" << endl);
      GLOBAL_ERROR_STREAM ("Current type : " << iv.getCurrentType() << "                Expected type: Format" << endl);
      GLOBAL_ERROR_STREAM ("Current namespace : " << iv.getCurrentNamespace() << "      Expected namespace: NMM " << endl);
      iv.nextElement();
      return iv;
    }

    /* Access the internal attributes */
    if (!iv.goDown()) {
      iv.nextElement();
      return iv;
    }

    /* Cleanup format */
    format.erase();

    /* Begin> read the format type */
    string type;
    iv >> type;
    format.setFormatType(type);
    /* End< */

    /* Begin> read the jack tag */
    string jack_tag;
    iv >> jack_tag;
    format.setJackTag(jack_tag);
    /* End< */


    /* Begin> read the parameter */
    while(!iv.isEmpty()) {
      format.createParameter(iv);
    }
    /* End< */

    /* Leave the formats attributes */
    iv.goUp();

    return iv;
  }

  OValueStream& operator<< (OValueStream& ov, const Format& format) {
    ov.beginType("Format", "NMM");

    /* Write the format type */
    ov << format.getFormatType();
    /* Write the jack tag */
    ov << format.getJackTag();

    /* Write the parameter of this format */
    for (parameter_list::const_iterator it = format.m_parameter_list.begin(); it != format.m_parameter_list.end(); it++)
    {
      (*it)->toStream(ov);
    }

    /* Write the IO-Partner of this format */
    if (  format.getIOPartner() ) {
      /*Begin> set IOPartner of the IOpartner (which should be the given format itself) to 0 to avoid infinity recursion*/
      Format* io_partner = const_cast<Format*>(format.getIOPartner()->getIOPartner());
      const_cast<Format*>(format.getIOPartner())->setIOPartner(0);
      /*End< */

      /*Begin> Write the IOPartner */
      ov.beginType("IOPartner");
      ov << *format.getIOPartner();
      ov.endType();
      /*End< */

      /*Begin> Reset the changed IOPartner */
      const_cast<Format*>(format.getIOPartner())->setIOPartner(io_partner);//reset the old IOPartner
      /*End < */
    }

    ov.endType();
    return ov;
  }

  TValueConstructor<Format>               __nmm__format_value   ("Format", "NMM");
  
  const char* Format::wild_wild_type = "*/*";
  const char* Format::audio_wild_type = "audio/*_*";
  const char* Format::audio__type = "audio/";
  const char* Format::audio_ac3_type = "audio/ac3";
  const char* Format::audio_mp3_type = "audio/mp3";
  const char* Format::audio_amr_type = "audio/amr";
  const char* Format::audio_mpa_type = "audio/mpa";
  const char* Format::audio_mpeg_type = "audio/mpeg";
  const char* Format::audio_mpeg1_type = "audio/mpeg1";
  const char* Format::audio_x_ogg_type = "audio/x-ogg";
  const char* Format::audio_x_wav_type = "audio/x-wav";
  const char* Format::audio_ogg_type = "audio/ogg";
  const char* Format::audio_pcma_type = "audio/pcma";
  const char* Format::audio_pcmu_type = "audio/pcmu";
  const char* Format::audio_raw_type = "audio/raw";
  const char* Format::audio_wambi_type = "audio/wambi";
  const char* Format::audio_wma_type = "audio/wma";
  const char* Format::av_wild_type = "av/*_*";
  const char* Format::av_asf_type = "av/asf";
  const char* Format::av_avi_type = "av/avi";
  const char* Format::av_mpeg_type = "av/mpeg";
  const char* Format::av_mpegps_type = "av/mpegps";
  const char* Format::av_ogg_type = "av/ogg";
  const char* Format::av_ogm_type = "av/ogm";
  const char* Format::video_wild_type = "video/*_*";
  const char* Format::video_h263_type = "video/h263";
  const char* Format::video_h264_type = "video/h264";
  const char* Format::video_h263plus_type = "video/h263-1998";
  const char* Format::video_jpeg_type = "video/jpeg";
  const char* Format::video_mpeg_type = "video/mpeg";
  const char* Format::video_mpeg2_type = "video/mpeg2";
  const char* Format::video_mpeg2pes_type = "video/mpeg2pes";
  const char* Format::video_mpeg4_type = "video/mpeg4";
  const char* Format::video_mpv_type = "video/mpv";
  const char* Format::video_raw_type = "video/raw";
  const char* Format::video_rgb_type = "video/rgb";
  const char* Format::video_rtjpeg_type = "video/rtjpeg";
  const char* Format::video_spu_type = "video/spu";
  const char* Format::video_test_type = "video/test";
  const char* Format::video_yuv_type = "video/yuv";

  const char* Format::audio_codec_param = "audio_codec";
  const char* Format::audio_format_param = "audio_format";
  const char* Format::audio_format_string_param = "audio_format_string";
  const char* Format::bitperpixel_param = "bitperpixel";
  const char* Format::bitpersample_param = "bitpersample";
  const char* Format::bitrate_param = "bitrate";
  const char* Format::blockalign_param = "blockalign";
  const char* Format::channelorder_param = "channelorder";
  const char* Format::channels_param = "channels";
  const char* Format::colorspace_param = "colorspace";
  const char* Format::compressor_param = "compressor";
  const char* Format::encoding_param = "encoding";
  const char* Format::endian_param = "endian";
  const char* Format::extradata_param = "extradata";
  const char* Format::extradata_size_param = "extradata_size";
  const char* Format::format_param = "format";
  const char* Format::format_ac3_audio_streams_param = "format_ac3_audio_streams";
  const char* Format::format_mpeg_audio_streams_param = "format_mpeg_audio_streams";
  const char* Format::format_dts_audio_streams_param = "format_dts_audio_streams";
  const char* Format::format_lpcm_audio_streams_param = "format_lpcm_audio_streams";
  const char* Format::format_mpeg_video_streams_param = "format_mpeg_video_streams";
  const char* Format::format_spu_streams_param = "format_spu_streams";
  const char* Format::framerate_param = "framerate";
  const char* Format::layer_param = "layer";
  const char* Format::lfe_param = "lfe";
  const char* Format::map_ac3_audio_streams_param = "map_ac3_audio_streams";
  const char* Format::map_mpeg_audio_streams_param = "map_mpeg_audio_streams";
  const char* Format::map_dts_audio_streams_param = "map_dtc_audio_streams";
  const char* Format::map_lpcm_audio_streams_param = "map_lpcm_audio_streams";
  const char* Format::map_mpeg_video_streams_param = "map_mpeg_video_streams";
  const char* Format::map_spu_streams_param = "map_spu_streams";
  const char* Format::num_audio_streams_param = "num_video_streams";
  const char* Format::num_video_streams_param = "num_video_streams";
  const char* Format::quality_param = "quality";
  const char* Format::ratio_x_param = "ratio_x";
  const char* Format::ratio_y_param = "ratio_y";
  const char* Format::sample_rate_param = "sample_rate";
  const char* Format::sign_param = "sign";
  const char* Format::speaker_setup_param = "speaker_setup";
  const char* Format::version_param = "version";
  const char* Format::video_format_param = "video_format";
  const char* Format::x_resolution_param = "x_resolution";
  const char* Format::y_resolution_param = "y_resolution";

}

#include "nmm/nmm_suffix.hpp"
