/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "Property.hpp"
#include "Format.hpp"
#include "nmm/comm/serialize/TValueConstructor.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  Property::Property() {
  }

  /* Destructor deletes the format list now  (repplix, 28.06.2002)*/
  Property::~Property() {
    erase();
  }

  /* Added Copy Constructor   (repplix, 28.06.2002)*/
  Property::Property(const Property& property) {
    *this = property;
  }

  void Property::erase() {
    m_default_formats.clear();//clear the map

    for (format_list::iterator it = m_format_list.begin(); it != m_format_list.end(); ++it) {
      delete (*it);
    }

    m_format_list.clear(); //clear the old pointer list
  }

  const Property& Property::operator=(const Property& property) {
    // make sure, the property is empty aka clean
    erase();

    m_property_type = property.m_property_type;

    for (format_list::const_iterator it = property.getFormatList().begin(); it != property.getFormatList().end(); ++it) {
      // copy each format
      Format* added_fmt = new Format(**it);
      // and insert it in the format list
      m_format_list.push_back(added_fmt);

      //if the format has a jack-tag
      if (added_fmt->hasJackTag()) {
        map<const string, Format*>::const_iterator i_it = property.m_default_formats.find(added_fmt->getJackTag());
        if (i_it != property.m_default_formats.end())
          //check if it is the default-format for this jack-tag
          if (i_it->second == (*it)) {
            //set it as default format in this property too
            setDefaultFormat(added_fmt, added_fmt->getJackTag());
          }
      }
    }

    return *this;
  }

  Format* Property::addNewFormat(const string & pFormatType) {
    Format *newFormat = new Format(pFormatType);
    m_format_list.push_back(newFormat);
    return newFormat;
  }

  void Property::addNewFormat(Format* param) {
    m_format_list.push_back(param);
  }

  format_list Property::getFormat(const string & pFormatType) const {
    // check if format exists
    format_list tempList;
    format_list::const_iterator i;
    for (i = m_format_list.begin(); i != m_format_list.end(); i++) {
      if ((*i)->getFormatType() == pFormatType) {
	tempList.push_back(*i);
      }	
    }	
    return tempList;
  }

  format_list Property::getFormat(const Format * pFormat) const {
    format_list result;
    format_list::const_iterator itr;
    for (itr = m_format_list.begin(); itr != m_format_list.end(); itr ++) {
      if ( pFormat->isFormatMatching(*itr)) {
	result.push_back(*itr);
      }
    }
    return result;
  }
  
  const format_list& Property::getFormatList() const {
    return m_format_list;
  }

  const Format* Property::getIOPartner(const Format* pFormat) const {
    return (pFormat->getIOPartner());
  }

  void Property::showFormats() const {
    format_list::const_iterator i;
    for (i = m_format_list.begin(); i != m_format_list.end(); i++) {
      (*i)->showParameter();
    }
  }

  void Property::setDefaultFormat(Format* pFormat, const string jack_tag /* = "default" */) {
    // make sure, format has set the correct corresponding jack_tag
    pFormat->setJackTag(jack_tag);

    pair<map<const string, Format*>::iterator, bool> res = m_default_formats.insert(make_pair(jack_tag, pFormat));
    // if insert failed, remove the old key-value pair and add the new one
    if (!res.second) {
      map<const string, Format*>::iterator search = m_default_formats.find(jack_tag);
      if (m_default_formats.end() != search) {
        // remove the old default Format entry for this jack_tag ...
        // do not delete the Format itselve as the m_default_formats
        // map only stores pointers to the Formats stored in the
        // m_format_list.
	m_default_formats.erase(search);
        // ... and insert the new one
	m_default_formats.insert(make_pair(jack_tag, pFormat));
      }
    }
  }

  const Format* Property::getDefaultFormat(const string jack_tag ) const {
    // search for a entry with the key jack_tag in the
    // m_default_formats map
    map<const string, Format*>::const_iterator search = m_default_formats.find(jack_tag);
    if (m_default_formats.end() == search) {
      // no entry found for this jack_tag, returning NULL-Ptr.
      GLOBAL_WARNING_STREAM("No default format for jack with name '" << jack_tag << "'! Maybe no jack with this name exists?" << endl);
      return 0;
    }
    // return the Format-Ptr of the found entry
    return search->second;
  }

  void Property::deleteFormat(Format* pFormat) {
    // if the given Format has a jack_tag, we check, if it is the
    // jack_tag's default Format
    if (pFormat->hasJackTag()) {
      map<const string, Format*>::iterator search = m_default_formats.find(pFormat->getJackTag());
      if (m_default_formats.end() != search) {
        // we found a default Format for the given Format's jack_tag,
        // let's check if the default Format is equal to the given one
        // (== the given Format is the default Format for the jack_tag)
	if (search->second == pFormat) {
	  // the given Format is a default one, remove it from the map
	  // (Format will be deleted later)
	  m_default_formats.erase(search);
	}
      }
    }

    // search the given Format in the m_format_list
    format_iterator tempiter = m_format_list.end();
    for (format_iterator i = m_format_list.begin(); i != m_format_list.end(); i++) {
	if ((*i) == pFormat) {
          // found it, store the found iterator for later erase() (we
          // cannot erase the iterator used as for-iterator!)
	  tempiter = i;
          // delete the Format object
	  delete (*i);
	}
    }
    // remove the iterator from the map if found
    if (tempiter != m_format_list.end()) {
      m_format_list.erase(tempiter);
    }
  }

  bool Property::isFormatIncluded(const Format& format, const string& jack_tag) const {
    /*Begin> Check if one of the other formats include the given format */
    for (list<Format*>::const_iterator it = getFormatList().begin(); it != getFormatList().end(); ++it) {
      if ((*it)->getJackTag() == jack_tag && (*it)->isFormatIncluded(&format)) {
	return true;
      }
    }
    return false;
    /*End< */
  }
  
  const_format_list Property::createIntersectionFormats(const Format* format, 
							const string& property_jack_tag, 
							bool include_def_format /* = true */) const 
  {
    const_format_list l;
    Format* f = 0;
    
    const Format *def_format = getDefaultFormat(property_jack_tag);

    for (format_list::const_iterator it = getFormatList().begin(); it != getFormatList().end(); ++it) {
      
      // if default format should not be included, and current format equals def format -> continue
      if((!include_def_format) && (def_format) && (*def_format == *(*it))) {
	continue;
      }

      if ((*it)->getJackTag() == property_jack_tag) {
	f = (*it)->createIntersectionFormat(format);
	if(f) {
	  l.push_front(f);
	}
      }
    }
    return l;
  }

  const_format_list Property::createIntersectionFormats(const string& out_jack_tag, 
							const Property* property, const string& in_jack_tag, 
							bool include_def_format /* = true */) const 
  {
    const_format_list l;
    Format* f = 0;
    
    const Format *out_def_format = getDefaultFormat(out_jack_tag);
    const Format *in_def_format  = property->getDefaultFormat(in_jack_tag);

    for (format_list::const_iterator out_it = getFormatList().begin(); out_it != getFormatList().end(); ++out_it) {
      for (format_list::const_iterator in_it = property->getFormatList().begin(); in_it != property->getFormatList().end(); ++in_it) {
      
	if(((*in_it)->getJackTag() == in_jack_tag) && ((*out_it)->getJackTag() == out_jack_tag)) {
	  // if default format should not be included, and current format equals def format -> continue
	  if((!include_def_format) && (in_def_format) && (*in_def_format == *(*in_it))) {
	    continue;
	  }
	  if((!include_def_format) && (out_def_format) && (*out_def_format == *(*out_it))) {
	    continue;
	  }

	  f = (*out_it)->createIntersectionFormat(*in_it);
	  if(f) {
	    l.push_front(f);
	  }
	}
      }
    }
    return l;
  }

  bool Property::isFormatMatching(const Format& format, const string& jack_tag) const 
  {
    /*Begin> Check if one of the other formats match the given format and jack tag */
    for (list<Format*>::const_iterator it = getFormatList().begin(); it != getFormatList().end(); ++it) {
      if ((*it)->getJackTag() == jack_tag && (*it)->isFormatMatching(&format)) {
	return true;
      }
    }
    return false;
    /*End< */
  }
  
  IValueStream& operator>>(IValueStream& v, Property& property) {
    if ( (v.getCurrentType() != "Property") || (v.getCurrentNamespace() != "NMM")) {
      GLOBAL_ERROR_STREAM ("Invalid type or namespace:" << endl);
      GLOBAL_ERROR_STREAM ("Current type : " << v.getCurrentType() << "                Expected type: Property" << endl);
      GLOBAL_ERROR_STREAM ("Current namespace : " << v.getCurrentNamespace() << "      Expected namespace: NMM " << endl);

      v.nextElement();
      return v;
    }

    property.erase();

    if(!v.goDown()) {
      GLOBAL_DEBUG_STREAM("Property has no foramts" << endl);
      v.nextElement();
      return v;
    } else {
      // first, check if there are default Formats in the serialized
      // input
      if (v.getCurrentType() == "DefaultFormat") {
	if (!v.goDown()) {
	  GLOBAL_ERROR_STREAM("No default format available" << endl);
	  v.nextElement();
	} else {
          // we iterate over all default Formats
	  while (!v.isEmpty()) {
            // read in the Format, ...
	    Format* format = new Format();
	    v >> *format;
            // ... insert it in the m_format_list ...
	    property.m_format_list.push_back(format);
            // ... and set it as defualt Format for its own jack_tag
	    property.setDefaultFormat(format, format->getJackTag());
	  }
          // don't forget to go up again as we successful went down
          // (in the if-statement!)
	  v.goUp();
	}
      }
      
      // now we can iterate over the rest of Formats in this
      // serialized Property
      while(!v.isEmpty()) {
        // read in the Format ...
	Format* format = new Format();
	v >>  *format;
        // ... and insert it in the m_format_list (only!)
	property.m_format_list.push_back(format);
      }

      // don't forget to go up again as we successful went down
      // (in the if-statement!)
      v.goUp();
    }

    return v;
  }

  OValueStream& operator<<(OValueStream& v, const Property& property) {
    v.beginType("Property", "NMM");

    // first we list the default Formats
    v.beginType("DefaultFormat");
    for (map<const string, Format*>::const_iterator i = property.m_default_formats.begin(); i != property.m_default_formats.end(); i++) {
      // we iterate over the map-entries of the m_default_formats map
      // this assignment is only to make the code easier to read
      // (without thousands of derefencings)
      Format& tmp_fmt = *(*i).second;
      if (tmp_fmt.hasJackTag()) {
        // make sure, that the default format has a valid jack tag set!
        if (tmp_fmt.getJackTag() == (*i).first) {
          // OK, searialize the Format into the DefaultFormat-Type
          v << *(*i).second;
        } else {
          // strange: jack_tag of map-entry and Format do not match,
          // but they should as the default Formats are inserted like
          // this (see Property::setDefaultFormat()).
          GLOBAL_ERROR_STREAM("Invalid jack tag in default format" << endl);
        }
      } else {
        // strange: jack_tag of map-entry present, but Format does not
        // have one, but it should as the default Formats are inserted
        // like this (see Property::setDefaultFormat()).
        GLOBAL_ERROR_STREAM("Invalid jack tag in default format" << endl);
      }
    }
    v.endType();

    // now list the rest of the Formats in the Property
    for (format_list::const_iterator it = property.m_format_list.begin(); it != property.m_format_list.end(); ++it) {
      if ((*it)->hasJackTag()) {
        const Format* def_fmt = property.getDefaultFormat((*it)->getJackTag());
        if (def_fmt == (*it)) {
          continue; // format already listed in default formats
          // so we don't have to list it here, as default Formats are
          // also added to the m_format_list while deserializing
          // Properties (see operator<<() above!).
        }
      }
  
      v << **it;
    }

    v.endType();

    return v;
  }
  TValueConstructor<Property>             __nmm__property_value ("Property", "NMM");

}



#include "nmm/nmm_suffix.hpp"
