/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "R_IParameter.hpp"
#include "R_INT.hpp"
#include "nmm/comm/serialize/IValueStream.hpp"
#include "nmm/comm/serialize/OValueStream.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  R_IParameter::R_IParameter(const string & pName) : Parameter(pName) 
  {
    m_parameter_type = Parameter::R_INT32;
    m_value_counter  = m_value_list.begin();
    m_default_range  = 0;
  }

  R_IParameter::R_IParameter(const string & pName, void *pValue) : Parameter(pName) 
  {
    m_parameter_type = Parameter::R_INT32;
    m_value_list.push_back((NMM::R_INT*)pValue);
    m_value_counter       = m_value_list.begin();
    m_default_value = static_cast<R_INT *>(pValue)->getLowValue();
    m_default_range = m_value_list.front();
  }

  R_IParameter::~R_IParameter()
  {
    for (list<R_INT *>::const_iterator itr = m_value_list.begin(); itr != m_value_list.end(); itr ++)
      {
	delete *itr;
      }
  }

  void R_IParameter::addValue(void *pValue) 
  {
    if (m_value_list.empty()) m_default_value = static_cast<R_INT *>(pValue)->getLowValue();

    // add the new element at the correct position
    R_INT_Iterator temp_itr;
    bool add = true;
    for (R_INT_Iterator itr = m_value_list.begin(); itr != m_value_list.end(); itr ++)
      {
	if ( (*itr)->getLowValue() >= ((R_INT *)pValue)->getLowValue())
	  {
	    if (add)
	      {
		temp_itr = itr;
		add      = false;
	      }
	  }
      }
    if (!add)
	m_value_list.insert(temp_itr,(R_INT *)pValue);
    else
	m_value_list.push_back((R_INT *)pValue);
    m_value_counter       = m_value_list.begin();
    
    m_default_range = m_value_list.front();
  }

  void R_IParameter::addValue(R_INT * value)
  {
    if (m_value_list.empty()) m_default_value = value->getLowValue();

    // add the new element at the correct position
    R_INT_Iterator temp_itr;
    bool add = true;
    for (R_INT_Iterator itr = m_value_list.begin(); itr != m_value_list.end(); itr ++)
      {
	if ( (*itr)->getLowValue() >= value->getLowValue())
	  {
	    if (add)
	      {
		temp_itr = itr;
		add      = false;
	      }
	  }
      }
    if (!add)
	m_value_list.insert(temp_itr,value);
    else
	m_value_list.push_back(value);
    m_value_counter       = m_value_list.begin();
   
    m_default_range = m_value_list.front();
  }

  R_INT * R_IParameter::getDefaultRange() const
  {
    return m_default_range;
  }

  bool R_IParameter::isValueIncluded(void *pValue) const
  {
    r_int32_t_list::const_iterator i;
    for (i = m_value_list.begin(); i != m_value_list.end(); i++) 
      {
	if ( ((*i)->getLowValue()<=*((int32_t *)pValue)) && ((*i)->getHighValue()>=*((int32_t *)pValue)) ) 
	  {
	    return true;
	  }
      }
    return false;
  }

  bool R_IParameter::isValueIncluded(int32_t value) const
  {
    r_int32_t_list::const_iterator itr;
    for (itr = m_value_list.begin(); itr != m_value_list.end(); itr ++)
      {
	if ( (*itr)->isValueIncluded(value) )
	  return true;
      }
    return false;
  }

  bool R_IParameter::isRangeMatching(int32_t lValue, int32_t hValue) const
  {
    r_int32_t_list::const_iterator itr;
    for (itr = m_value_list.begin(); itr != m_value_list.end(); itr ++)
      {
	if ( (*itr)->isRangeMatching(lValue,hValue) )
	  return true;
      }
    return false;
  }

  bool R_IParameter::isRangeIdentical(int32_t lValue, int32_t hValue) const
  {
    for (r_int32_t_list::const_iterator itr = m_value_list.begin(); itr != m_value_list.end(); itr ++)
      {
	if ( (*itr)->isRangeIdentical(lValue,hValue))
	  return true;
      }
    return false;
  }

  bool R_IParameter::isValueIncluded(int32_t lValue, int32_t hValue) const
  {
    r_int32_t_list::const_iterator itr;
    for (itr = m_value_list.begin(); itr != m_value_list.end(); itr ++)
      {
	if ( (*itr)->isRangeIncluded(lValue,hValue) )
	  return true;
      }
    return false;
  }

  bool R_IParameter::isRangeIntersection(int32_t m_lValue, int32_t m_hValue, int32_t lValue, int32_t hValue) const
  {
    if (// [1 [2 2] 1] 
	( (m_lValue <= lValue)
	 && (m_hValue >= hValue))
	// [2 [1 1] 2]
	|| ((m_lValue >= lValue)
	    && (m_hValue <= hValue))
	// [2 [1 2] 1]
	|| ((m_lValue >= lValue)
	    && (m_lValue <= hValue))
	// [2 [1 2] 1]
	|| ((m_hValue >= lValue)
	    && (m_lValue <= hValue)))
      {
	return true;
      }
    return false;
  }

  bool R_IParameter::isValueIncluded(void *lValue,void *hValue) const
  {
    r_int32_t_list::const_iterator itr;
    for (itr = m_value_list.begin(); itr != m_value_list.end(); itr ++)
      {
	if (// [1 [2 2] 1] 
	    (((*itr)->getLowValue() <= *((int32_t *)lValue))
	      && ((*itr)->getHighValue() >= *((int32_t *)hValue)))
	    // [2 [1 1] 2]
	     || (((*itr)->getLowValue() >= *((int32_t *)lValue))
		 && ((*itr)->getHighValue() <= *((int32_t *)hValue)))
	    // [2 [1 2] 1]
	     || (((*itr)->getLowValue() >= *((int32_t *)lValue))
		 && ((*itr)->getLowValue() <= *((int32_t *)hValue)))
	    // [2 [1 2] 1]
	     || (((*itr)->getHighValue() >= *((int32_t *)lValue))
		 && ((*itr)->getHighValue() <= *((int32_t *)hValue))))
	     
	    //|| ( this->isValueIncluded(lValue) && this->isValueIncluded(hValue)))
	  {
	    return true;
	  }
      }
    return false;
  }

  void R_IParameter::showParameter() const
  {
    GLOBAL_DEBUG_STREAM2( "Parameter-Name  : " << m_parameter_name << endl);
    GLOBAL_DEBUG_STREAM2( "Parameter-Type  : R_INT32" << endl);
    GLOBAL_DEBUG_STREAM2( "Parameter-Weight: " << m_parameter_weight << endl);
    for (r_int32_t_list::const_iterator i = m_value_list.begin(); i != m_value_list.end(); i++) 
      {
	GLOBAL_DEBUG_STREAM2("Value: [ " << (*i)->getLowValue() 
			 << " , " << (*i)->getHighValue() << " ]" 
			 << " - Weight: [ " << (*i)->getLowWeight() << " , " << (*i)->getHighWeight() << " ] " 
			 << " Standardized weight : " << (*i)->getStandardizedWeight() << endl);
      }
  }

  ostream& R_IParameter::showParameter(ostream& os) const
  {
    os << m_parameter_name << "[weight:" << m_parameter_weight << "]=r_float(";

    for (r_int32_t_list::const_iterator i = m_value_list.begin(); i != m_value_list.end(); i++) 
      {
	os << "(" << (*i)->getLowValue() << "-" << (*i)->getHighValue() 
           << ")[low_weight:" << (*i)->getLowWeight() << ", std_weight:" << (*i)->getStandardizedWeight() << ", hight_weight:" << (*i)->getHighWeight() << "]";
      }
    os << ")";
    return os;
  }

  ostream& R_IParameter::showParameterVerbose(ostream& os) const
  {
    os << "Parameter-Name  : " << m_parameter_name << endl;
    os << "Parameter-Type  : R_INT32" << endl;
    os << "Parameter-Weight: " << m_parameter_weight << endl;
    for (r_int32_t_list::const_iterator i = m_value_list.begin(); i != m_value_list.end(); i++) 
      {
 	os <<"Value: [ " << (*i)->getLowValue() 
           << " , " << (*i)->getHighValue() << " ]" 
           << " - Weight: [ " << (*i)->getLowWeight() << " , " << (*i)->getHighWeight() << " ] " 
           << " Standardized weight : " << (*i)->getStandardizedWeight() << endl;
      }
    return os;
  }
  
  void * R_IParameter::nextValue() const
  {
    R_INT *temp = new R_INT();
    if ( m_value_counter != m_value_list.end() ) 
      {
	temp = *m_value_counter;
	m_value_counter++;
	return temp;
      }
    else 
      {
	m_value_counter = m_value_list.begin();
	return NULL;
      }
  }
  
  void * R_IParameter::getFirstValue() const
  {
    m_value_counter = (++m_value_list.begin());
    return *(m_value_list.begin());
  }

  Parameter * R_IParameter::duplicate() const
  {
    R_IParameter * tempParameter = new R_IParameter(m_parameter_name);
    for (r_int32_t_list::const_iterator itr = m_value_list.begin(); itr != m_value_list.end(); itr ++ )
      {
	R_INT *temp = new R_INT();
	temp->setLowValue((*itr)->getLowValue());
	temp->setHighValue((*itr)->getHighValue());
	temp->setLowWeight((*itr)->getLowWeight());
	temp->setHighWeight((*itr)->getHighWeight());
	tempParameter->addValue(temp);
      }
    return (Parameter *)tempParameter;
  }

  Parameter * R_IParameter::getIntersection(const Parameter * reference) const
  {
    // okay, this is a quick and dirty hack because someone else forgot
    // to handle this case and I am *not* going to implement it properly!
    // the idea is to use a new R_IParameter instead of the wildcard parameter
    if(reference->getParameterType() == Parameter::WILDCARD) {
      R_IParameter* tmp_reference = new R_IParameter(getParameterName());
      float l = this->getLowestValue(); 
      float h = this->getHighestValue(); 
      NMM::R_INT* f_range = new NMM::R_INT(this->getLowestValue(), 
					   this->getHighestValue(), 
					   this->getWeight((void *) &l),
					   this->getWeight((void *) &h));
      tmp_reference->addValue(f_range);
      Parameter* return_reference = this->getIntersection(tmp_reference);
      delete tmp_reference;
      return return_reference;
    }
    // quick and dirty hack END

    if (reference->getParameterType() == Parameter::INT32)
      {
	return reference->getIntersection(this);
      }
    R_IParameter * result_parameter = new R_IParameter(m_parameter_name);
    bool           running          = true;
    bool           increase_itr     = false;
    bool           increase_reference = false;
    int32_t        tempint;
//     R_INT_Iterator itr              = m_value_list.begin();
    list<R_INT *>::const_iterator itr = m_value_list.begin();
   
    R_INT * first_element     = *itr;
    R_INT * first_element_ref = static_cast<R_INT*>(reference->getFirstValue());
    
    if ( (first_element == NULL ) || (first_element_ref == NULL))
      running = false;

    while (running)
      {
	int32_t temp = first_element_ref->getLowValue();
	int32_t temp2= first_element_ref->getHighValue();
	increase_itr = false;
	increase_reference = false;

	if (this->isRangeIntersection( first_element->getLowValue(), first_element->getHighValue(),temp , temp2 ))
	  // there is somewhere a machting between the intervals
	  {
	    R_INT *new_element = new R_INT();
	    // [2 [1 1] 2]
	    if ( (first_element_ref->getLowValue() <= first_element->getLowValue()) 
		 && (first_element_ref->getHighValue() >= first_element->getHighValue()) )
	      {
		new_element->setLowValue(first_element->getLowValue());
		new_element->setHighValue(first_element->getHighValue());
		tempint = first_element->getLowValue();
		new_element->setLowWeight(((reference)->getWeight(&tempint) + first_element->getLowWeight())/2);
		tempint = first_element->getHighValue();
		new_element->setHighWeight(((reference)->getWeight(&tempint) + first_element->getHighWeight())/2);
		// next interval
		itr ++;
		increase_itr = true;
	      }
	    // [1 [2 2] 1]
	    if ( (first_element->getLowValue() <= first_element_ref->getLowValue()) 
		 && (first_element->getHighValue() >= first_element_ref->getHighValue()) )
	      {
		new_element->setLowValue(first_element_ref->getLowValue());
		new_element->setHighValue(first_element_ref->getHighValue());
		tempint = first_element_ref->getLowValue();
		new_element->setLowWeight((this->getWeight(&tempint) + first_element_ref->getLowWeight())/2);
		tempint = first_element_ref->getHighValue();
		new_element->setHighWeight((this->getWeight(&tempint) + first_element_ref->getHighWeight())/2);
		// next interval
		increase_reference = true;
	      }
	    // [1 [2 1] 2]
	    if ( (first_element->getLowValue() <= first_element_ref->getLowValue())
		 && (first_element->getHighValue() <= first_element_ref->getHighValue()) )
	      {
		new_element->setLowValue(first_element_ref->getLowValue());
		new_element->setHighValue(first_element->getHighValue());
		tempint = first_element_ref->getLowValue();
		new_element->setLowWeight((this->getWeight(&tempint) + first_element_ref->getLowWeight())/2);
		tempint = first_element->getHighValue();
		new_element->setHighWeight(((reference)->getWeight(&tempint) + first_element->getHighWeight())/2);
		// next interval
		itr ++;
		increase_itr = true;
	      }
	    // [2 [1 2] 1]
	    if ( (first_element->getLowValue() >= first_element_ref->getLowValue())
		 && (first_element->getHighValue() >= first_element_ref->getHighValue()) )
	      {
		new_element->setLowValue(first_element->getLowValue());
		new_element->setHighValue(first_element_ref->getHighValue());
		tempint = first_element->getLowValue();
		new_element->setLowWeight(((reference)->getWeight(&tempint) + first_element->getLowWeight())/2);
		tempint = first_element_ref->getHighValue();
		new_element->setHighWeight((this->getWeight(&tempint) + first_element_ref->getHighWeight())/2);
		// next interval
		increase_reference = true;
	      }
	    if (increase_reference)
	      first_element_ref = (R_INT *)reference->nextValue();
	    if (increase_itr)
	      first_element = *itr;
	    result_parameter->addValue(new_element);
	  }
	else
	  // there is no matching between the intervals
	  {
	    // [2 2] [1 1]
	    if (first_element->getLowValue() >= first_element_ref->getHighValue())
	      first_element_ref = (R_INT *)reference->nextValue();
	    // [1 1] [2 2]
	    else 
	      {
		itr ++;
		first_element = *itr;
	      }
	  }
	if ( (itr == m_value_list.end() ) || (first_element_ref == NULL) )
	  running = false;
      }
    return result_parameter;
  }

  float R_IParameter::getWeight(void *pValue) const
  {
    for (r_int32_t_list::const_iterator itr = m_value_list.begin(); itr != m_value_list.end(); itr ++)
      {
	if ( ((*itr)->getLowValue() <= *((int32_t *)pValue))
	     && ((*itr)->getHighValue() >= *((int32_t *)pValue)) )
	  {
	    return ( (*itr)->getLowWeight() + ( (*((int32_t *)pValue) - (*itr)->getLowValue()) * this->computeFactor(*itr)) );
	  }
      }
    return 0;
  }

  float R_IParameter::computeFactor(R_INT * param) const
  {
    if ((param->getHighWeight() - param->getLowWeight()) == 0 ) return 0;
    return ( (param->getHighWeight() - param->getLowWeight()) / ( param->getHighValue() - param->getLowValue() ));
  }

  uint R_IParameter::getNumberOfValues() const
  {
    return m_value_list.size();
  }

  bool R_IParameter::isParameterIdentical(const Parameter * reference) const
  {
    if (reference->getParameterType() != Parameter::R_INT32)
      return false;
    if (m_value_list.size() != reference->getNumberOfValues() )
      return false;
    
    for (r_int32_t_list::const_iterator itr = m_value_list.begin(); itr != m_value_list.end(); itr++)
      {
	if (!dynamic_cast<R_IParameter *>(const_cast<Parameter *>(reference))->isRangeIdentical((*itr)->getLowValue(),
												(*itr)->getHighValue()))
	  return false;
      }
    return true;
  }

  bool R_IParameter::isParameterMatching(const Parameter * reference) const
  {
    if (reference->getParameterType() == Parameter::INT32)
      return reference->isParameterMatching(this);

    for (r_int32_t_list::const_iterator itr = m_value_list.begin(); itr != m_value_list.end(); itr ++)
      {
	if ( reference->isRangeMatching( (*itr)->getLowValue(), (*itr)->getHighValue()) )
	  return true;
      }
    return false;
  }

  float R_IParameter::getStandardizedWeight() const
  {
    float local_weight = 0;
    for (r_int32_t_list::const_iterator itr = m_value_list.begin(); itr != m_value_list.end(); itr ++)
      {
	local_weight += (*itr)->getStandardizedWeight();
      }
    return ((local_weight/m_value_list.size()) * m_parameter_weight);
  }

  void R_IParameter::setDefaultValue(int32_t value)
  {
    m_default_value = value;
  }

  int32_t R_IParameter::getDefaultValue() const
  {
    return m_default_value;
  }

  int32_t R_IParameter::getLowestValue() const
  {
    int32_t tempint = (*m_value_list.begin())->getLowValue();
    for (r_int32_t_list::const_iterator itr = m_value_list.begin(); itr != m_value_list.end(); itr ++)
      {
	if ( (*itr)->getLowValue() < tempint)
	  tempint = (*itr)->getLowValue();
      }
    return tempint;
  }

  int32_t R_IParameter::getHighestValue() const
  {
    int32_t tempint = (*m_value_list.begin())->getHighValue();
    for (r_int32_t_list::const_iterator itr = m_value_list.begin(); itr != m_value_list.end(); itr ++)
      {
	if ( (*itr)->getHighValue() > tempint)
	  tempint = (*itr)->getHighValue();
      }
    return tempint;
  }

  const r_int32_t_list & R_IParameter::getRIntValueList() const
  {
    return m_value_list;
  }

  void R_IParameter::fromStream(IValueStream& iv) {
    if (iv.getCurrentType() != "R_IParameter" || iv.getCurrentNamespace() != "NMM" ) {
      GLOBAL_ERROR_STREAM ("Invalid type or namespace:" << endl);
      GLOBAL_ERROR_STREAM ("Current type : " << iv.getCurrentType() << "                Expected type: R_IParameter" << endl);
      GLOBAL_ERROR_STREAM ("Current namespace : " << iv.getCurrentNamespace() << "      Expected namespace: NMM " << endl);

      iv.nextElement();
      return;
    }
    if (!iv.goDown()) {
      GLOBAL_ERROR_STREAM ("R_IParameter is empty" << endl);
      iv.nextElement();
      return;
    }

    string par_name;
    nmm_float weight;

    iv >> par_name;
    iv >> weight;

    setParameterName(par_name);
    setParameterWeight(weight);

    float w1, w2;
    int32_t v1, v2;

    while(!iv.isEmpty()) {
      iv >> v1;
      iv >> w1;
      iv >> v2;
      iv >> w2;

      addValue(new R_INT(v1, v2, w1, w2));
    }
    iv.goUp();
  }

  void R_IParameter::toStream(OValueStream& ov) const  {
    ov.beginType("R_IParameter", "NMM");

    string p_name = getParameterName();
    nmm_float weight = getParameterWeight();

    ov << p_name;
    ov << weight;

    for (list<R_INT* >::const_iterator it = m_value_list.begin(); it != m_value_list.end(); it++)
    {
      nmm_float w1, w2;
      nmm_int32 v1, v2;
      v1 = (*it)->getLowValue();
      v2 = (*it)->getHighValue();
      w1 = (*it)->getLowWeight();
      w2 = (*it)->getHighWeight();

      ov << v1;
      ov << w1;
      ov << v2;
      ov << w2;
    }
    ov.endType();
  }

}

#include "nmm/nmm_suffix.hpp"
