/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include "B_Allocator.hpp"
#include "Buffer.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  B_Allocator::B_Allocator(size_t size, size_t start_num, size_t max_num, const char* name) : NamedObject(name)
  {
    // set useful default of parameter
    if(max_num == 0) {
      setMaxBufferCount(10);
    }
    else {
      setMaxBufferCount(max_num);
    }

    // create #start_num many buffers
    if ((size > 0) && (start_num > 0)) {
      for (size_t i=0; i < start_num; i++) {
        bufs.insert(bufset_t::value_type(size, new Buffer(static_cast<nmm_uint32>(size), this)));
      }
    }

    num_of_buffers = 0;
    erase_flag = false;
  }

  B_Allocator::~B_Allocator() 
  {
    // free Buffers...
    DEBUG_STREAM2("B_Allocator Destructor" << endl);
    
    mutex.lock();
    for(bufset_t::iterator i= bufs.begin(); i!=bufs.end(); ++i) {
      delete i->second;
      DEBUG_STREAM2("Deleted Buffer in B_Allocator" << endl);
    }
    mutex.unlock();
  }
  
  Buffer* B_Allocator::requestBuffer (const size_t size) const {
    bufset_t::iterator i;
    Buffer *b = 0;
    int new_size = size;
    mutex.lock();

    if (size <= 0) {
      new_size = 1;
    }

    if ((i = bufs.upper_bound(new_size-1)) != bufs.end()) {
      b = (*i).second;
      bufs.erase(i);
      b->reset();
      b->increaseCount(1);
      DEBUG_STREAM2(" ++++++++ FOUND BUFFER " << size << ", " << bufs.size() << " buffers left +++++++" << endl);
    } 
    else {
      b = new Buffer(static_cast<nmm_uint32>(size), this);
      if(bufs.size() == 0) {
	DEBUG_STREAM2(" ++++++++++++++++++++++ NO BUFFERS LEFT AT ALL, requested size " << size << " +++++++++++" << endl);
      }
      else {
	DEBUG_STREAM2(" ++++++++++++++++++++++ NO MATCHING BUFFER " 
                      << bufs.size() << " buffers left +++++++++" << endl);
      }
      b->increaseCount(1);
    }

    num_of_buffers++;
    mutex.unlock();

    return b;
  }

  void B_Allocator::releaseBuffer (Buffer* buffer) const 
  {
    mutex.lock();
    if (bufs.size() > max_buffer_count) {
      delete buffer;
      DEBUG_STREAM2(" ----------------- TOO MANY BUFFERS " << bufs.size() 
                    << " > " << max_buffer_count << " released buffer ----------------" << endl);
    }
    else {
      bufs.insert(bufset_t::value_type(buffer->getSize(), buffer));
    }

    num_of_buffers--;

    // The object will delete itself only if all buffers given away have returned
    if(erase_flag && (num_of_buffers <= 0)) {
      mutex.unlock();
      delete this;
      return;
    }

    mutex.unlock();
  }

  void B_Allocator::erase() const 
  {
    mutex.lock();
    // set flag, telling the object to try to delete itself
    erase_flag = true;

    // The object will delete itself only if all buffers given away have returned
    if (num_of_buffers <= 0) {
      mutex.unlock();
      DEBUG_STREAM("deleting myself" << endl);
      // todo: do we have to wait until all requestBuffer methods are
      // finished or can we simple delete ourself, potentially having
      // other threads waiting in our requestBuffer method? Maybe a
      // freebuffers_cond.notifyAll();
      // would even be enough!
      // see also: ShmBufferManager::erase()
      delete this;
    } else {
      mutex.unlock();
    }
  }

  Result B_Allocator::setMaxBufferCount(const size_t max)
  {
    max_buffer_count = max;
    return SUCCESS;
  }
  
  const size_t B_Allocator::getMaxBufferCount() const
  {
    return max_buffer_count;
  }

  
  
}

#include "nmm/nmm_suffix.hpp"
