/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "CEvent.hpp"
#include "Event.hpp"

#include "nmm/comm/serialize/TConstructor.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  CEvent::CEvent(bool b)
    : Message(Message::EVENT), b_mgr(b)
  {
    setMandatoryBit(true); // event should not be discarded during flush
    event_array.useTValue(false);
    insertArgument(&event_array);
    setTypeName(type_name);
    setNamespace("NMM");
    increaseCount();
  }


  CEvent::CEvent(size_t size, bool b)
    : Message(Message::EVENT), event_array(size), b_mgr(b)
  {
    setMandatoryBit(true); // event should not be discarded during flush
    event_array.useTValue(false);
    insertArgument(&event_array);
    for (size_t i = 0; i< size; ++i)
      event_array[i] = 0;

    setTypeName(type_name);
    setNamespace("NMM");

    increaseCount();
  }

  CEvent::CEvent(const Event& event, bool b)
    : Message(Message::EVENT), b_mgr(b)
  {
    setMandatoryBit(true); // event should not be discarded during flush
    event_array.useTValue(false);
    insertArgument(&event_array);

    setTypeName(type_name);
    setNamespace("NMM");

    event_array.insert(event_array.end(), new Event(event));
    increaseCount();
  }

  CEvent::CEvent(Event* event, bool b)
    : Message(Message::EVENT), b_mgr(b)
  {
    setMandatoryBit(true); // event should not be discarded during flush
    event_array.useTValue(false);
    insertArgument(&event_array);

    setTypeName(type_name);
    setNamespace("NMM");

    event_array.insert(event_array.end(), event);
    increaseCount();
  }

  CEvent::CEvent(const Event& e1, const Event& e2, bool b)
    : Message(Message::EVENT), b_mgr(b)
  {
    setMandatoryBit(true); // event should not be discarded during flush
    event_array.useTValue(false);
    insertArgument(&event_array);

    setTypeName(type_name);
    setNamespace("NMM");

    event_array.insert(event_array.end(), new Event(e1));
    event_array.insert(event_array.end(), new Event(e2));
    increaseCount();
  }

  void CEvent::erase() const {
    if (b_mgr)
      delete this;
  }

  CEvent::CEvent(const CEvent& ce) 
    : Message(ce), event_array(ce.size()) 
  {
    setMandatoryBit(true); // event should not be discarded during flush
    event_array.useTValue(false);
    insertArgument(&event_array);

    for (size_t i = 0; i< event_array.size(); ++i )
      event_array[i] = new Event(ce[i]);

    setTypeName(type_name);
    setNamespace("NMM");
    setDirection(ce.getDirection());

    increaseCount();
  }

  CEvent::~CEvent() {
    for (size_t i = 0; i< event_array.size(); ++i)
      delete event_array[i];
  }

  Event& CEvent::operator[](size_t i) {
    return *event_array[i];
  }

  const Event& CEvent::operator[](size_t i) const {
    return *event_array[i];
  }
  
  void CEvent::setEvent(size_t i, const Event& e) {
    delete event_array[i];
    event_array[i] = new Event(e);
  }
  
  void CEvent::setEvent(size_t i, Event* e) {
    delete event_array[i]; event_array[i] = e;
  }
  
  void CEvent::insertEvent(size_t p, const Event& event) {
    if (p >= event_array.size()) {
      event_array.insert(event_array.end(), new Event(event));
      return;
    }

    vector<Event*>::iterator it = event_array.begin();
    advance(it, p);
    event_array.insert(it, new Event(event));
  }
  
  void CEvent::insertEvent(const Event& e) {
    event_array.insert(event_array.end(), new Event(e));
  };
  
  void CEvent::insertEvent(Event* e) {
    event_array.insert(event_array.end(), e);
  };
  
  void CEvent::insertEvent(size_t p, Event* event) {
    if (p >= event_array.size()) {
      event_array.insert(event_array.end(), event);
      return;
    }

    vector<Event*>::iterator it = event_array.begin();
    advance(it, p);
    event_array.insert(it, event);
  }

  void CEvent::removeEvent(size_t p) {
    if (p >= event_array.size()) {
      GLOBAL_ERROR_STREAM2("CEvent > Invalid position. Could not remove Event" << endl);
      return;
    }
    vector<Event*>::iterator it = event_array.begin();
    for (size_t i = 0; i < p; ++i)
      ++it;
    delete *it;
    event_array.erase(it);
  }


  CEvent& CEvent::operator=(const CEvent& ce) {
    Message::operator=(ce); //call the assignmenet operator from base class

    for (size_t i = 0; i< event_array.size(); ++i) {
      delete event_array[i];
    }
    event_array.resize(ce.event_array.size());

    for (size_t i = 0; i< ce.event_array.size(); ++i) {
      if (ce.event_array[i])
        event_array[i] = ce.event_array[i]->copy();
      else
        event_array[i] = 0;
    }

    return *this;
  }
  const string CEvent::type_name = "CEvent";
  
  TConstructor<CEvent> __nmm__cevent_value("CEvent", "NMM");
  
}

#include "nmm/nmm_suffix.hpp"
