/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "EventDispatcher.hpp"

#include "EDObject.hpp"
#include "Event.hpp"
#include "CEvent.hpp"
#include "IEventReceiver.hpp"

#include "nmm/utils/thread/MutexGuard.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  EventDispatcher::EventDispatcher(const char* name, int state):
    NamedObject(name), mutex(ThreadMutex::RECURSIVE), external_mutex(ThreadMutex::RECURSIVE),
    __internal_process_mutex(ThreadMutex::RECURSIVE)
  {
    this->state = state;
  }


  EventDispatcher::EventDispatcher(const EventDispatcher& dispatcher)
    : IState(dispatcher), NamedObject(dispatcher),  mutex(ThreadMutex::RECURSIVE), external_mutex(ThreadMutex::RECURSIVE)
  {

  }

  EventDispatcher::~EventDispatcher() {
    for (HASH_MAP_STL::hash_map<string, EDObject*, string_hash >::iterator it = __int_event_map.begin(); it != __int_event_map.end(); ++it)
      {
	delete it->second;
      }
    __int_event_map.clear();
  }


  void EventDispatcher::registerEvent(const string& s, EDObject* ed) {
    event_it it = __int_event_map.find(s);
    if ( it != __int_event_map.end() )
      if ( it->second != ed ){
	GLOBAL_WARNING_STREAM("Another TEDObject is alreadey registered with SET event-key " << s << endl);	
      }
      else
	return;
    __int_event_map.insert(pair<string, EDObject*>(s, ed));
  }

  EDObject* EventDispatcher::removeEvent(const string& s) {
    event_it it = __int_event_map.find(s);
    if ( it == __int_event_map.end() )
      return 0;
    EDObject* tmp_object = it->second;
    __int_event_map.erase(it);
    return tmp_object;
  }


  Result EventDispatcher::dispatch(Event& event, const IEventSender* sender)
    throw(const IllegalStateException&)
  {
    // Removed mutex to avoid deadlock in distributed case 
    // NMMObject N1 calls method on NMMObject N2. N2 itself calls a method on N1. 
    //    MutexGuard mg(mutex);

    /*Begin> Check if a method is registered with given event-key */
    event_it it = __int_event_map.find(event.getKey());
    if (it == __int_event_map.end() ) {
      /*If no such key is registered forward it to all listener*/
      MutexGuard emg(external_mutex);
      externalDispatch(event, sender);
      forwardEvent(event, sender);

      if (__ext_recv_forwarding_list.empty()) {
         if ( __ext_event_map.find(event.getKey()) == __ext_event_map.end() ) {
           WARNING_STREAM2( "No event with key \""<< event.getKey() << "\" registerd" << endl);
         }
      } else {
        DEBUG_STREAM2( "No event with key \""<< event.getKey() << "\" registerd, b ut external listener for this event exist" << endl);
      }
      event.setException(UnknownEventException("Unkown event-key : " + event.getKey()));

      return FAILURE;
    }
    /*End< */

    /*Begin> Check if the method can be executed in current state */
    if (!it->second->isExecutionAllowed( state ) ) {
      ERROR_STREAM2("The method " << string(event.getKey()) << " cannot be executed in current state. My state : "<< state << endl);
      event.setException(IllegalStateException("The method " + event.getKey() + " cannot be executed in current state."));
      return FAILURE;
    }
    /*End< */

    /* Execute the method */
    Result r = it->second->dispatch(event);

    /*Begin> If no error occurs, mark the method as executed in the current state */
    if ( r != FAILURE ) {
      if(r != SUCCESS) GLOBAL_DEBUG_STREAM2("Ups, some event handling method did not return SUCCESS or FAILURE!" << endl);
      it->second->setExecuted(state);
      MutexGuard emg(external_mutex);
      externalDispatch(event, sender);
      forwardEvent(event, sender);
      return r;
    }
    /*End< */

    /*Begin> If an error occurs, mark the method as not executed in current state */
    it->second->reset(state); //Event could not be executed
    return FAILURE;
    /*End< */
  }


  Result EventDispatcher::internalProcessCEvent(CEvent* cevent)
    throw(const IllegalStateException&)
  {
    MutexGuard mg(__internal_process_mutex);

    if (cevent == 0) {
      return FAILURE;
    }

    // Try to dispatch all events of cevent
    // -> If one dispatched method returns FAILURE, remember this and return FAILURE in the end
    // -> If one dispatched method returns DELETE, remove it from CEvent.
    Result end_result = SUCCESS;
    unsigned int increment = 1;
    for(unsigned int cevent_working_index = 0; cevent_working_index < cevent->size(); cevent_working_index += increment) {
      increment = 1;
      Event& event = (*cevent)[cevent_working_index];
      // is handler registered for current event key ?
      event_it it = __int_event_map.find(event.getKey());
      if (it != __int_event_map.end() ) {
	/*Begin> Check if the method can be executed in current state */
	if(it->second->isExecutionAllowed(state)) {

	  /* Execute the method */
	  Result r = it->second->dispatch(event);

	  if(r == SUCCESS) { /* If no error occurs, mark the method as executed in the current state */
	    it->second->setExecuted(state);
	  }
	  else if(r == DELETE) { /* or delete the method */
	    DEBUG_STREAM2("Remove Event : "<< cevent_working_index << ": " << event.getKey() << endl);
	    it->second->setExecuted(state); // is this ok ?
	    cevent->removeEvent(cevent_working_index);
	    increment = 0; // now increment this time, since an event was removed
	  }
	  else if(r == FAILURE) { /* If an error occurs, mark the method as not executed in current state */
	    it->second->reset(state); 
	    end_result = FAILURE;
	  }
	  else {
	    WARNING_STREAM2("Undefined return " << r << " value for handling event " << event.getKey() << endl);
	  }
	}
	else { /* Not allowed in current state */
	  ERROR_STREAM2("The method " << string(event.getKey()) << " cannot be executed in current state. My state : "<< state << endl);
	  event.setException(IllegalStateException("The method " + event.getKey() + " cannot be executed in current state."));
	  end_result = FAILURE;
	}
	/*End< */
      }
    }
    return end_result;
  }

  Result EventDispatcher::externalProcessCEvent(CEvent* cevent) 
  {
    MutexGuard emg(external_mutex);

    if (cevent == 0) {
      return FAILURE;
    }

    // Try to dispatch all events of cevent
    Result end_result = SUCCESS;
    for(unsigned int cevent_working_index = 0; cevent_working_index < cevent->size(); ++cevent_working_index) {
      Event& event = (*cevent)[cevent_working_index];
      externalDispatch(event, 0);
      forwardEvent(event, 0);
    }
    return end_result;
  }

  void EventDispatcher::reset(int state) {
    for (HASH_MAP_STL::hash_map<string, EDObject*, string_hash >::iterator it = __int_event_map.begin(); it != __int_event_map.end(); ++it)
      {
	(*it).second->reset(state); //reset mode
      }
  }

  void EventDispatcher::reset(const string& key, int state) {
    event_it it = __int_event_map.find(key);
    if (it == __int_event_map.end())
      return;

    (*it).second->reset(state);                              //reset the state
  }


  void EventDispatcher::changeState(int s)
    throw (const UnsuccessfulRequiredCallException&)
  {
    checkMandatoryEvents(s);
    state = s;
    for (HASH_MAP_STL::hash_map<string, EDObject*, string_hash >::iterator it = __int_event_map.begin(); it != __int_event_map.end(); ++it)
      {
	/*Resets the execution list*/
	it->second->reset();
      }
  }

  list<string> EventDispatcher::getEventList() const {
    list<string> tmp_list;
    for (HASH_MAP_STL::hash_map<string, EDObject*, string_hash>::const_iterator it = __int_event_map.begin(); it != __int_event_map.end(); it++)
      {
	tmp_list.insert(tmp_list.end(), (*it).first);
      }
    return tmp_list;
  }


  EDObject* EventDispatcher::getEDObject(const string& event_key) {
    event_it it = __int_event_map.find(event_key);
    if (it != __int_event_map.end())
      return it->second;

    return 0;
  }

  const EDObject* EventDispatcher::getEDObject(const string& event_key) const{
    cevent_it it = __int_event_map.find(event_key);
    if (it != __int_event_map.end())
      return it->second;

    return 0;
  }

  Event* EventDispatcher::matchEventPrototype(const string& substring) {
    for (event_it it = __int_event_map.begin(); it != __int_event_map.end(); ++it) {
      size_t pos = it->first.rfind("::");
      if(it->first.substr(pos+2) == substring) {
	return it->second->createEvent(it->first);
      }
    }

    return 0;
  }



  /********************************** Register EventListener methods ***************************************/

  void EventDispatcher::registerEventListener(IEventReceiver* receiver) {
    MutexGuard mg(mutex);
    MutexGuard emg(external_mutex);
    MutexGuard intg(__internal_process_mutex);
    __ext_recv_forwarding_list.push_back(receiver);
  }


  void EventDispatcher::registerEventListener(const string& s, EDObject* edo) {
    MutexGuard mg(mutex);
    MutexGuard emg(external_mutex);
    MutexGuard intg(__internal_process_mutex);
    EDOMapIt it = __ext_event_map.find(s);
    if ( it == __ext_event_map.end()) {
      it = __ext_event_map.insert(pair<string, list<EDObject*> >(s, list<EDObject*>())).first;
    }
    it->second.push_back(edo);
  }

  void EventDispatcher::registerEventListener(const string& s, IEventReceiver* receiver) {
    MutexGuard mg(mutex);
    MutexGuard emg(external_mutex);
    MutexGuard intg(__internal_process_mutex);
    ReceiverMapIt it = __ext_event_recv_map.find(s);
    if ( it == __ext_event_recv_map.end()) {
      it = __ext_event_recv_map.insert(pair<string, list<IEventReceiver*> >(s, list<IEventReceiver*>())).first;
    }
    it->second.push_back(receiver);
  }


  /********************************** Remove EventListener methods ***************************************/

  list<string> EventDispatcher::removeEventListener(EDObject* edo) {
    list<string> event_list;
    MutexGuard mg(mutex);
    MutexGuard emg(external_mutex);
    MutexGuard intg(__internal_process_mutex);
    for (EDOMapIt it = __ext_event_map.begin(); it != __ext_event_map.end(); ++it) {
      if ( find(it->second.begin(), it->second.end(), edo) != it->second.end() )
        {
          it->second.remove(edo);
          event_list.push_back(it->first);
        }
    }
    // check EDOMap for entries with empty EDO-lists and remove these entries
    // otherwise hasEventListener(const string&) will report true
    // although there are non as it only checks if a EDOMap entry for
    // the given key exists, not if there are any handlers in the
    // entry's list
    for (list<string>::const_iterator event_key = event_list.begin();
         event_key != event_list.end();
         event_key++) {
      EDOMapIt it = __ext_event_map.find(*event_key);
      if (it != __ext_event_map.end()) {
        // found the entry
        if (it->second.empty()) {
          // no handlers for this entry, no need to keep the entry any more
          __ext_event_map.erase(it);
        }
      }
    }
    return event_list;
  }

  void EventDispatcher::removeEventListener(const string& event_key, IEventReceiver* receiver) {
    MutexGuard mg(mutex);
    MutexGuard emg(external_mutex);
    MutexGuard intg(__internal_process_mutex);
    ReceiverMapIt it = __ext_event_recv_map.find(event_key);
    if (it != __ext_event_recv_map.end()) {
      it->second.remove(receiver);
      if (it->second.empty()) {
        // no handlers for this entry, no need to keep the entry any more
        __ext_event_recv_map.erase(it);
        // otherwise hasEventListener(const string&) will report true
        // although there are non as it only checks if a ReceiverMap entry for
        // the given key exists, not if there are any receivers in the
        // entry's list
      }
    }
  }

  void EventDispatcher::removeEventListener(IEventReceiver* receiver) {
    MutexGuard mg(mutex);
    MutexGuard emg(external_mutex);
    MutexGuard intg(__internal_process_mutex);
    /* Begin> remove all specific event receiver */
    for (ReceiverMapIt it = __ext_event_recv_map.begin(); it != __ext_event_recv_map.end(); ++it)
      if (it != __ext_event_recv_map.end())
        it->second.remove(receiver);
    /* End< */

    /* Remove the receiver form the list*/
    __ext_recv_forwarding_list.remove(receiver);
  }

  void EventDispatcher::forwardEvent(Event& event, const IEventSender* sender) {

    if (!__ext_recv_forwarding_list.empty())
      for (list<IEventReceiver*>::iterator it = __ext_recv_forwarding_list.begin(); it != __ext_recv_forwarding_list.end(); ++it)
	(*it)->receiveEvent(event, sender);
  }

  void EventDispatcher::externalDispatch(Event& event, const IEventSender* sender) {
    EDOMapIt it = __ext_event_map.find(event.getKey());
    if (it != __ext_event_map.end() ) {
      for (list<EDObject*>::iterator l_it = it->second.begin(); l_it != it->second.end(); ++l_it) {
	if ( (*l_it)->isExecutionAllowed( state ) ) {
	  if ((*l_it)->dispatch(event) == SUCCESS ) {
	    (*l_it)->setExecuted(state);
	  } 
	  else {
	    (*l_it)->reset(state); //Event could not be executed
	  }
	}
      }
    }

    ReceiverMapIt r_it = __ext_event_recv_map.find(event.getKey());
    if (r_it != __ext_event_recv_map.end()) {
      for (list<IEventReceiver*>::iterator l_it = r_it->second.begin(); l_it != r_it->second.end(); ++l_it)
	(*l_it)->receiveEvent(event, sender);
    }
  }

  bool EventDispatcher::hasEvent(const string& key) const {
    if (__int_event_map.find(key) != __int_event_map.end())
      return true;
    return false;
  }

  void EventDispatcher::checkMandatoryEvents(int new_state)     const
    throw (const UnsuccessfulRequiredCallException&)
  {
    for (cevent_it it = __int_event_map.begin(); it != __int_event_map.end(); ++it) {
      if ( (*it).second->isRequired(state, new_state) &&  (!(*it).second->isExecuted(state)) ) {//this method must be executed
	GLOBAL_ERROR_STREAM2("No Event " << (*it).first << " received which is need to change state" <<endl);
	throw UnsuccessfulRequiredCallException("No Event " + (*it).first + " received which is need to change state");
      }
    }
  }

  list<Event*> EventDispatcher::getUnhandledMandatoryEvents(int new_state) const {
    list<Event*> tmp_list;
    for (cevent_it it = __int_event_map.begin(); it != __int_event_map.end(); ++it) {
      if ( (*it).second->isRequired(state, new_state) &&  (!(*it).second->isExecuted(state)) ) {//this method must be executed
	tmp_list.push_back( it->second->createEvent(it->first) );
      }
    }
    return tmp_list;
  }

  void EventDispatcher::addRequired(const string& event_key, int from_state, int to_state) {
    EDObject* ed_object = getEDObject(event_key);
    if (!ed_object) {
      ERROR_STREAM("No event " << event_key << " available which can be set to required "<< endl);
    } else {
      ed_object->addRequired(from_state, to_state);
    }
  }

  void EventDispatcher::addLimited(const string& event_key, int state) {
    EDObject* ed_object = getEDObject(event_key);
    if (!ed_object) {
      ERROR_STREAM("No event " << event_key << " available which can be limited to states " << endl);
    } else {
      ed_object->addLimited(state);
    }
  }

  bool EventDispatcher::hasSpecificListener(const string& event_key) const {
    if ( (__ext_event_map.find(event_key) == __ext_event_map.end()) && 
	 (__ext_event_recv_map.find(event_key) == __ext_event_recv_map.end())) {
      return false;
    }
    else {
      return true;
    }
  }

  bool EventDispatcher::hasSpecificListener() const {
    if ( __ext_event_map.empty() && 
	 __ext_event_recv_map.empty()) {
      return false;
    }
    else {
      return true;
    }
  }

  list<string> EventDispatcher::getSpecificListener() const {
    list<string> tmp_list;
    for (CEDOMapIt it = __ext_event_map.begin(); it != __ext_event_map.end(); ++it)
      tmp_list.push_back(it->first);

    for (CReceiverMapIt it = __ext_event_recv_map.begin(); it != __ext_event_recv_map.end(); ++it)
      if (find(tmp_list.begin(), tmp_list.end(), it->first) == tmp_list.end())
        tmp_list.push_back(it->first);

    return tmp_list;
  }
  
}

#include "nmm/nmm_suffix.hpp"
