/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "StreamQueue.hpp"
#include "ISignalReceiver.hpp"
#include "Message.hpp"
#include "Buffer.hpp"

#include "nmm/utils/thread/MutexGuard.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  StreamQueue::StreamQueue(const Mode mode,
			   const unsigned int size,
			   const char* name,
			   const Signal free_queue_slot_sig,
			   const Signal queue_not_empty_sig) 
    : NamedObject(name), 
      queue_mode(mode),
      use_duration(false),
      max_queue_size(size), 
      sig_free_queue_slot(free_queue_slot_sig), 
      sig_queue_not_empty(queue_not_empty_sig),
      limited_size(size),
      pre_fetch_queue_size(0),
      pre_fetch_flag(false),
      initial_pre_fetch_flag(false),
      pre_fetch_condition(pre_fetch_mutex)
  {
    setName(getName()+"> Queue");
    abort_receive_exception_count = 0;

    parent_of_jack = 0;
    parent_of_partner_jack = 0;
  };

  StreamQueue::~StreamQueue()
  { 
    flush(); // I guess, we should not call this method... BufferManagers should delete their Buffers
  };

  bool StreamQueue::isQueueFull() const
  {
    if(use_duration) {
      return message_queue_duration >= max_queue_duration;
    }
    else {
      return message_queue.size() >= max_queue_size;
    }
  }

  bool StreamQueue::isQueueMoreThanFull() const
  {
    if(use_duration) {
      return message_queue_duration > max_queue_duration;
    }
    else {
      return message_queue.size() > max_queue_size;
    }
  }

  bool StreamQueue::isPrefetchingFinished() const
  {
    if(use_duration) {
      return message_queue_duration >= pre_fetch_queue_duration;
    }
    else {
      return message_queue.size() == pre_fetch_queue_size;
    }
  }

  Result StreamQueue::setParentSignalReceiver(ISignalReceiver* parent)
  {
    MutexGuard mg(queue_mutex);
    parent_of_jack = parent;
    return SUCCESS;
  }

  Result StreamQueue::setPartnerSignalReceiver(ISignalReceiver* partner)
  {
    MutexGuard mg(queue_mutex);
    parent_of_partner_jack = partner;
    
    // if a new parent of partner jack is set, we have to check if he can send us messages or if he has to wait
    if ( parent_of_partner_jack && ( abort_receive_exception_count > 0) ) {
      if ( isQueueMoreThanFull() ) {    
        return BLOCKED;
      } else {
	// reset the internal counter
        abort_receive_exception_count = 0;
	return SUCCESS;
      }
    }
    
    return SUCCESS;
  }

  Result StreamQueue::setParentSignalReceiver(ISignalReceiver* parent, const Signal queue_not_empty_sig)
  {
    MutexGuard mg(queue_mutex);
    parent_of_jack = parent;
    sig_queue_not_empty = queue_not_empty_sig;
    return SUCCESS;
  }

  Result StreamQueue::setPartnerSignalReceiver(ISignalReceiver* partner, const Signal free_queue_slot_sig)
  {
    MutexGuard mg(queue_mutex);
    parent_of_partner_jack = partner;
    sig_free_queue_slot = free_queue_slot_sig;

    // if a new parent of partner jack is set, we have to check if he can send us messages or if he has to wait
    if ( parent_of_partner_jack && ( abort_receive_exception_count > 0) ) {
      if ( isQueueFull() ) {
        return BLOCKED;
      } else {
        // reset the internal counter
        abort_receive_exception_count = 0;
        return SUCCESS;
      }
    }
                                                           
    return SUCCESS;
  }

  Result StreamQueue::flush(const Timestamp& flush_ts)
  {
    DEBUG_STREAM2("Clear Queue." << endl);
    MutexGuard mg(queue_mutex);

    return internalFlush(flush_ts);
  }

  Result StreamQueue::internalFlush(const Timestamp& flush_ts)
  {
    queue<Message*> tmp_queue;
    Message* msg = 0;

    while (!message_queue.empty()) {
      msg = message_queue.front();
      message_queue.pop();

      if(msg->getMandatoryBit()) { // keep message
	tmp_queue.push(msg);
      }
      else { // message should be discarded
	// if timestamp for flushing was given -> flush all messages with 'older' timestamp only
	if(flush_ts.is_set && flush_ts.is_valid) {
	  Timestamp msg_ts = msg->getTimestamp();
	  if(msg_ts.is_set && msg_ts.is_valid) {
	    if(flush_ts.sync_time < msg_ts.sync_time) { // if flush_ts is 'younger' than msg_ts -> keep message
	      tmp_queue.push(msg);
	      continue;
	    }
	  }
	}

	// code taken from StreamQueue::dequeueMessage() :
	// signal if at least one buffer can be inserted, if we had signalled that
	// the queue was full.
	if (abort_receive_exception_count > 0) {
	  abort_receive_exception_count--;

	  // if partner_of_parent_jack is available, send internal event
	  if(parent_of_partner_jack) {
	    parent_of_partner_jack->receiveSignal(sig_free_queue_slot);
	    DEBUG_STREAM2("Signal sig_FREE_QUEUE_SLOT" << endl);
	  }
	  else {
	    WARNING_STREAM2("No subclass of ISignalReceiver, cannot send FREE_QUEUE_SLOT" << endl);
	  }
	}
	DEBUG_STREAM2("Releasing message of type " << msg->getType() << endl);
	msg->release();
	DEBUG_STREAM2("Releasing message done." << endl);
      }
    }
    DEBUG_STREAM2("Queue cleared." << endl);

    // Set total queue duration to zero
    message_queue_duration = Interval();
    in_message_duration.syncReset();

    while(!tmp_queue.empty()) { // re-insert mandatory message from tmp_queue to message_queue
      msg = tmp_queue.front();
      tmp_queue.pop();

      // Update message queue duration
      if(use_duration)
      {
        Interval duration = in_message_duration(msg);
        message_queue_duration += duration;
      }

      // Re-insert message into queue
      message_queue.push(msg);
      DEBUG_STREAM("Re-insert mandatory message " << msg->getType() << endl);
    }

    // reset the prefetch flag to its initial value
    pre_fetch_flag = initial_pre_fetch_flag;

    return SUCCESS;
  };

  Result StreamQueue::receiveMessage(Message* message)
  {
    MutexGuard mg(queue_mutex);

    if (message == 0) {
      ERROR_STREAM2("BUG: Received 0-Message" << endl);
      return FAILURE;
    }

    // check for FlushEvent
    if ( message->getType() == Message::FLUSH_EVENT) {
      DEBUG_STREAM2("received FlushEvent -> clearing queue" << endl);
      internalFlush(message->getTimestamp());
    }

    Message* oldest = 0;
    int insert_this_buffer = 1;

    if(isQueueFull()) {
      // debug test
      if(use_duration)
      {
        Interval more_than_max_queue_duration = max_queue_duration;
        more_than_max_queue_duration *= 10;
        if(message_queue_duration > more_than_max_queue_duration) {
          ERROR_STREAM2("ARGH. QUEUE MORE THAN FULL, THIS SHOULD NOT HAPPEN!" << endl);
        }
      }
      else
      {
        if(isQueueMoreThanFull()) {
          ERROR_STREAM2("ARGH. QUEUE MORE THAN FULL, THIS SHOULD NOT HAPPEN!" << endl);
        }
      }

      switch(queue_mode) {

      case MODE_SUSPEND:
	// Queue will suspend, but message will be inserted
	insert_this_buffer = true;
	break;

      case MODE_KEEP_NEWEST:
	// new message is newer than all enqueued messages, so
	// it has to be inserted because stream queue is in
	// mode MODE_KEEP_NEWEST
	insert_this_buffer = true;

	// get oldest message in queue
	oldest = message_queue.front();

        // update duration
        if(use_duration)
        {
          Interval duration = out_message_duration(oldest);
          message_queue_duration -= duration;
        }

        // remove message
	oldest->release();
	message_queue.pop();

	break;

      default:
	// this is MODE_KEEP_OLDEST (see header file), or anything else

	// release message, since older messages in queue should
	// be kept in MODE_KEEP_OLDEST
	message->release();
	insert_this_buffer = false;

	break;
      }
    }

    if(insert_this_buffer) {
      // if queue was empty before, signal that a new Buffer arrived and can be dequeued
      bool e = message_queue.empty();

      // flush events with valid timestamp will 'outrun' other messages in stream queue
      // -> this is necessary since the timestamp of the flush event should also be used for dropping all messages
      // in the successor of this node if 'older' (e.g. smaller) than timestamp of flush event
      // (just imagine: the flush event has to be inserted at the 'right' position within the queue!)
      if ( message->getType() == Message::FLUSH_EVENT) {
	// re-order entries in queue: put timed flush event to the beginning of the queue
	Timestamp ts = message->getTimestamp();
	if(ts.is_valid && ts.is_set) {
	  queue<Message*> tmp_queue;
	  Message* msg = 0;
	  while (!message_queue.empty()) {
	    msg = message_queue.front();
	    message_queue.pop();
	    tmp_queue.push(msg);
	  }
	  // re-insert messages to queue, put in 'flush_event' first!
	  message_queue.push(message);
	  while(!tmp_queue.empty()) { 
	    msg = tmp_queue.front();
	    tmp_queue.pop();
	    message_queue.push(msg);
	  }
	  DEBUG_STREAM2("reordered FlushEvent with timestamp " << message->getTimestamp() << endl);
	}
	else { // no timestamp set for flush event -> simply insert into queue
          if(use_duration)
          {
            // Update message queue duration
            Interval duration = in_message_duration(message);
            message_queue_duration += duration;
          }

	  message_queue.push(message);
	}
      }
      else { // no flush event -> simply insert into queue
        if(use_duration)
        {
          // Update message queue duration
          Interval duration = in_message_duration(message);
          message_queue_duration += duration;
        }
	message_queue.push(message);
      }

      if(e || pre_fetch_flag) {
        if(pre_fetch_flag) {

          // if partner_of_jack is available, send internal event
          if(isPrefetchingFinished()) {
            if(parent_of_jack) {
              MutexGuard mg(pre_fetch_mutex);
              pre_fetch_flag = false;
              parent_of_jack->receiveSignal(sig_queue_not_empty);
              pre_fetch_condition.notify();
            }
            else {
              DEBUG_STREAM2("No subclass of ISignalReceiver, cannot send QUEUE_NOT_EMPTY");
            }

            // TEST: fw: Debug output for prefetching
//            if(!use_duration) {
//              cout << "prefetching finished, size=" << message_queue.size() << "/" << pre_fetch_queue_size << endl;
//            }
//            else {
//              cout << "prefetching finished, duration=" << message_queue_duration << " / " << pre_fetch_queue_duration << endl;
//            }
          }
          else {
            // TEST: fw: Debug output for prefetching
//            if(!use_duration) {
//              cout << "prefetching, size=" << message_queue.size() << "/" << pre_fetch_queue_size << endl;
//            }
//            else {
//              cout << "prefetching, duration=" << message_queue_duration << " / " << pre_fetch_queue_duration << endl;
//            }
          }

        }
        else {
          // if partner_of_jack is available, send internal event
          if(parent_of_jack) {
            parent_of_jack->receiveSignal(sig_queue_not_empty);
          }
          else {
            DEBUG_STREAM2("No subclass of ISignalReceiver, cannot send QUEUE_NOT_EMPTY");
          }

        }
      }

      // if queue is "more than" full now, return BLOCKED
      if((isQueueFull()) &&   // changed from > to >= (mlohse 13-3-2002)
	 (queue_mode == MODE_SUSPEND)) {
	abort_receive_exception_count++;
	return BLOCKED;
      }
    }

    return SUCCESS;
  };

  Message* StreamQueue::dequeueMessage()
  {
    MutexGuard mg(queue_mutex);

    // empty queue -> return 0
    if (message_queue.empty()) {
      return 0; 
    }

    Message* msg = message_queue.front();

    // update duration
    if(use_duration)
    {
      Interval duration = out_message_duration(msg);
      message_queue_duration -= duration;
    }

    message_queue.pop();

    // signal if at least one buffer can be inserted, if we had signalled that
    // the queue was full.
    // (code also in StreamQueue::clear() )
    if ((abort_receive_exception_count > 0)) {

      // if partner_of_parent_jack is available, send internal event only if has empty slot
      if(parent_of_partner_jack && !isQueueFull()) {
	while(abort_receive_exception_count > 0) {
	  abort_receive_exception_count--;
	  parent_of_partner_jack->receiveSignal(sig_free_queue_slot);
	  DEBUG_STREAM2("Signal sig_FREE_QUEUE_SLOT" << endl);
	}
      }
      else {
	DEBUG_STREAM2("No subclass of ISignalReceiver, or queue more than full" << std::endl);
      }
    }

    return msg;
  };

  unsigned int StreamQueue::getSize() const 
  {
    MutexGuard mg(queue_mutex);
    return message_queue.size();
  }

  unsigned int StreamQueue::getMaxSize() const 
  {
    MutexGuard mg(queue_mutex);
    return max_queue_size;
  }

  Result StreamQueue::setMaxSize(unsigned int size)
  {
    MutexGuard mg(queue_mutex);
    max_queue_size = size;
    use_duration = false;

    return SUCCESS;
  }

  Result StreamQueue::setMaxDuration(Interval duration)
  {
    MutexGuard mg(queue_mutex);
    max_queue_duration = duration;
    use_duration = true;

    return SUCCESS;
  }

  void StreamQueue::unlimitSize()
  {
    MutexGuard mg(queue_mutex);
    limited_size = max_queue_size;
    DEBUG_STREAM2("unlimitSize() current max_queue_size " << max_queue_size << ", new 20000 " << endl);
    max_queue_size = 20000;
  }

  void StreamQueue::limitSize()
  {
    MutexGuard mg(queue_mutex);
    DEBUG_STREAM2("limitSize() current max_queue_size " << max_queue_size << ", new " << max_queue_size << endl);
    max_queue_size = limited_size;
  }

  StreamQueue::Mode StreamQueue::getMode() const
  {
    MutexGuard mg(queue_mutex);
    return queue_mode;
  }

  Result StreamQueue::setMode(StreamQueue::Mode mode)
  {
    MutexGuard mg(queue_mutex);
    queue_mode = mode;
    return SUCCESS;
  }

  void StreamQueue::lock() const {
    queue_mutex.lock();
  }

  void StreamQueue::unlock() const {
    queue_mutex.unlock();
  }

  void StreamQueue::setPrefetchQueueSize(unsigned int pre_fetch_queue_size) 
  {
    if(pre_fetch_queue_size > 0) {
      if(pre_fetch_queue_size > max_queue_size) {
	WARNING_STREAM2("Tried to set too large prefetch queue size " << pre_fetch_queue_size << ", using max. queue size of " << max_queue_size << " instead." << endl);
	this->pre_fetch_queue_size = max_queue_size;
      }
      else {
	this->pre_fetch_queue_size = pre_fetch_queue_size;
      }
      pre_fetch_flag = true;
    }
    else {
      this->pre_fetch_queue_size = 0;
      pre_fetch_flag = false;
    }

    initial_pre_fetch_flag = pre_fetch_flag;
    use_duration = false;
  }

  void StreamQueue::setPrefetchQueueDuration(const Interval& pre_fetch_queue_duration)
  {
    if(pre_fetch_queue_duration > Interval()) {
      if(pre_fetch_queue_duration > max_queue_duration) {
        WARNING_STREAM2("Tried to set too large prefetch queue duration " << pre_fetch_queue_duration << ", using max. queue duration of " << max_queue_duration << " instead." << endl);
        this->pre_fetch_queue_duration = max_queue_duration;
      }
      else {
        this->pre_fetch_queue_duration = pre_fetch_queue_duration;
      }
      pre_fetch_flag = true;
    }
    else {
      this->pre_fetch_queue_duration = Interval();
      pre_fetch_flag = false;
    }

    initial_pre_fetch_flag = pre_fetch_flag;
    use_duration = true;
  }

  void StreamQueue::waitPrefetchFinished()
  {
    MutexGuard mg(pre_fetch_mutex);
    while( pre_fetch_flag == true ) {
      pre_fetch_condition.wait();
    }
    DEBUG_STREAM2("waitPrefetchFinished() successful with " << message_queue.size() << " messages in queue" << endl);
  }

}

#include "nmm/nmm_suffix.hpp"
