/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "BufferedINetStream.hpp"

#include "nmm/comm/serialize/TConstructor.hpp"

#include "nmm/utils/net/Sock.hpp"
#include <cerrno>
#include <cassert>
#include <errno.h>

#include "nmm/nmm_prefix.hpp"

namespace NMM {  
  
  BufferedINetStream::BufferedINetStream()
    : INetStream(TypeInfo("BufferedINetStream", "NMM")),
      __buffer_size(2048),
      __recv_buffer(),
      __recv_buffer_begin(0),
      __recv_buffer_end(0)
  {
    insertArgument(&__buffer_size);
  }
  
  BufferedINetStream::BufferedINetStream(const BufferedINetStream& stream)
    : INetStream(stream),
      __buffer_size(stream.__buffer_size.getRValue()),
      __recv_buffer(),
      __recv_buffer_begin(0),
      __recv_buffer_end(0)
  {
    insertArgument(&__buffer_size);
  }
  
  BufferedINetStream::BufferedINetStream(const TypeInfo& ti)
    : INetStream(ti),
      __buffer_size(2048),
      __recv_buffer(),
      __recv_buffer_begin(0),
      __recv_buffer_end(0)
  {
    insertArgument(&__buffer_size);
  }
  
  BufferedINetStream* BufferedINetStream::copy() const {
    return new BufferedINetStream(*this);
  }
  
  BufferedINetStream::~BufferedINetStream()
  {}
  
  void BufferedINetStream::readValue(void* value, size_t size) {
    /* allocate or resize buffer on demand */
    allocateBuffer();
    
    while (size > 0) {
      /* get number of bytes in the buffer */
      unsigned used = __recv_buffer_end - __recv_buffer_begin;
      
      if (used >= size) {
        /* copy bytes from the buffer to the value */
        memcpy(value, &__recv_buffer[__recv_buffer_begin], size);
        __recv_buffer_begin += size;
        break;
      }
      
      if (used > 0) {
        /* copy all bytes from the buffer to the value */
        memcpy(value, &__recv_buffer[__recv_buffer_begin], used);
        size -= used;
        value = static_cast<char*>(value) + used;
      }
      
      /* buffer now empty */
      __recv_buffer_begin = 0;
      __recv_buffer_end = 0;
      
      if (size > __recv_buffer.size()) {
        /* read data directly to avoid unnecessary copying */
        size_t retval = internalReadValue(value, size);
        value = static_cast<char*>(value) + retval;
        size -= retval;
      } else {
        /* read new data into the buffer */
        __recv_buffer_end = internalReadValue(&__recv_buffer[__recv_buffer_begin], __recv_buffer.size());
      }
    }
  }
  
  size_t BufferedINetStream::internalReadValue(void* value, size_t size) {
    /* read bytes from the socket */
    assert(getSock());
    size_t retval = getSock()->recv(static_cast<char*>(value), size);
    if(retval == 0) {
      throw ConnectionClosed(EPIPE);
    }
    
    /* count bytes */
    countReadBytes(retval);
    
    return retval;
  }
  
  bool BufferedINetStream::isEmpty() const {
    if ( (__recv_buffer_end - __recv_buffer_begin) == 0 )
      return true;

    return false;
  }
  
  void BufferedINetStream::allocateBuffer() {
    if(__recv_buffer.size() != __buffer_size.getRValue()) {
      /* check if we can resize the buffer. If the buffer stores more data than the requested buffer size,
         we can't resize it yet. So this method must be called again later to resize the buffer. */
      if ( ((__recv_buffer_begin != 0) || (__recv_buffer_end != 0))
           && __recv_buffer.size() < __buffer_size.getRValue() ) {
        GLOBAL_ERROR_STREAM("BufferedINetStream: Could not resize receiving buffer, because it still stores data" << endl);
        return;
      }
      
      /* resize buffer */
      GLOBAL_DEBUG_STREAM("BufferedINetStream: Allocate buffer of size " << __buffer_size.getRValue() << endl);
      __recv_buffer.resize(__buffer_size.getRValue());
    }
  }
  
  
  template<>
  class TConstructor<BufferedINetStream> : public ValueConstructor {
  public:
    TConstructor(const TypeInfo& t_info)
      : ValueConstructor(t_info)
    { }
    virtual BufferedINetStream*          createValue()          const {return new BufferedINetStream();}
    virtual TValue<BufferedINetStream>*  createTValue()         const {return new TValue<BufferedINetStream>();}
  
    virtual TValue<INetStream*>*  createTPointerValue()  const {return new TValue<INetStream*>();}
  
    virtual TConstructor<BufferedINetStream>* copy()  const {return new TConstructor(*this); }
  private:
    string  __type_name;
    string  __namespace;
  };

  
  TConstructor<BufferedINetStream> __nmm__buffered_i_net_stream(TypeInfo("BufferedINetStream", "NMM"));
  
}

#include "nmm/nmm_suffix.hpp"
