/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "BufferedONetStream.hpp"

#include "nmm/comm/serialize/TConstructor.hpp"

#include "nmm/utils/net/Sock.hpp"

#include <cerrno>
#include <cassert>
#include <errno.h>

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  BufferedONetStream::BufferedONetStream()
    : ONetStream(TypeInfo("BufferedONetStream", "NMM")),
      __buffer_size(2048),
      __send_buffer(),
      __send_buffer_used(0)
  {
    insertArgument(&__buffer_size);
  }
  
  BufferedONetStream::BufferedONetStream(const BufferedONetStream& stream)
    : ONetStream(stream),
      __buffer_size(stream.__buffer_size.getRValue()),
      __send_buffer(),
      __send_buffer_used(0)
  {
    insertArgument(&__buffer_size);
  }
  
  BufferedONetStream::BufferedONetStream(const TypeInfo& ti)
    : ONetStream(ti),
      __buffer_size(2048),
      __send_buffer(),
      __send_buffer_used(0)
  {
    insertArgument(&__buffer_size);
  }
  
  BufferedONetStream* BufferedONetStream::copy() const {
    return new BufferedONetStream(*this);
  }
  
  BufferedONetStream::~BufferedONetStream()
  {}
  
  void BufferedONetStream::flush() {
    if(__send_buffer_used > 0) {
      /* write the temp array */
      internalWriteValue(&__send_buffer[0], __send_buffer_used);
      __send_buffer_used = 0;
    }
  }

  void BufferedONetStream::writeData() {
    flush();
  }

  void BufferedONetStream::writeValue(const void* value, size_t size) {
    /* allocate or resize buffer on demand */
    allocateBuffer();
    
    if(size >= __send_buffer.size()) {
      flush();
      internalWriteValue(value, size);
    } else {
      /* if not enough space is left flush the queue */
      if((__send_buffer.size() - __send_buffer_used) < size) {
        flush();
      }
      memcpy(&__send_buffer[__send_buffer_used], value, size);
      __send_buffer_used += size;
    }
  }

  void BufferedONetStream::internalWriteValue(const void* value, size_t size) {
    /* write bytes to the socket */
    assert(getSock());
    size_t retval = getSock()->send(static_cast<const char*>(value), size);
    if(retval == 0) {
      throw ConnectionClosed(EPIPE);
    }

    /* count bytes */
    countWriteBytes(retval);

    /* if not all values are written, write the rest */
    if(static_cast<size_t>(retval) != size) {
      const char* new_offset = static_cast<const char*>(value) + retval;
      internalWriteValue(new_offset, size - retval);
    }
  }
  
  void BufferedONetStream::allocateBuffer() {
    if(__send_buffer.size() != __buffer_size.getRValue()) {
      /* If more data is cached then the new buffer_size allows, we flush the NetStream */
      if (__send_buffer_used > __buffer_size.getRValue()) {
        flush();
      }
      
      /* resize the buffer */
      GLOBAL_DEBUG_STREAM("BufferedONetStream: Allocate buffer of size " << __buffer_size.getRValue() << endl);
      __send_buffer.resize(__buffer_size.getRValue());
    }
  }
  
  
  template<>
  class TConstructor<BufferedONetStream> : public ValueConstructor {
  public:
    TConstructor(const TypeInfo& t_info)
      : ValueConstructor(t_info)
    { }
    virtual BufferedONetStream*          createValue()          const {return new BufferedONetStream();}
    virtual TValue<BufferedONetStream>*  createTValue()         const {return new TValue<BufferedONetStream>();}
    
    virtual TValue<ONetStream*>*  createTPointerValue()  const {return new TValue<ONetStream*>();}
    
    virtual TConstructor<BufferedONetStream>* copy()  const {return new TConstructor(*this); }
  private:
    string  __type_name;
    string  __namespace;
  };
  
  TConstructor<BufferedONetStream> __nmm__buffered_o_net_stream(TypeInfo("BufferedONetStream", "NMM"));
  
}

#include "nmm/nmm_suffix.hpp"
