/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "BufferedSink.hpp"
#include "RTPReceiverStrategy.hpp"

#include "nmm/comm/messaging/Buffer.hpp"
#include "nmm/comm/messaging/IExternalBufferManager.hpp"
#include "nmm/comm/messaging/ITSOneWayBufferReceiver.hpp"

#include "nmm/utils/thread/MutexGuard.hpp"

#include "nmm/utils/time/timeutils.hpp"

#include <live/RTPSource.hh>
#include <live/AMRAudioSource.hh>
#include <live/RTPSource.hh>

#include <iostream>
#include <iomanip>

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  BufferedSink::BufferedSink(UsageEnvironment& env)
    : MediaSink(env),
      NamedObject("BufferedSink"),
      __current_buffer(0),
      __buffer_size(65000),
      __frame_size(8192),
      __blocked(false),
      __flush(false),
      __internal_cond(__internal_mutex),
      __buffer_manager(0),
      __receiver(0),
      __sender(0),
      __first_rtp_frame(true),
      __first_rtp_seqno(0)
  { 
    /* set time offset */
    gettimeofday(&__start_time, NULL);
    __has_first_sync_frame = false;
    __received_bytes = 0;
  }
  
  BufferedSink::~BufferedSink() {
    /* delete pending buffer */      
    if (__current_buffer) {
      __current_buffer->release();
      __current_buffer = NULL;
    }
  }
  
  BufferedSink* BufferedSink::createNew(UsageEnvironment& env) {
    return new BufferedSink(env);
  }
  
  void BufferedSink::setOneWayBufferReceiver(ITSOneWayBufferReceiver* receiver) throw() {
    __receiver = receiver;
  }
  
  void BufferedSink::setOneWayBufferSender(ITSOneWayBufferSender* sender) throw() {
    __sender = sender;
  }
  
  void BufferedSink::setBufferManager(IExternalBufferManager* buffer_manager) throw() {
    __buffer_manager = buffer_manager;
  }
  
  void BufferedSink::setBufferSize(unsigned int buffer_size) throw() {
    __buffer_size = buffer_size;
  }
  
  void BufferedSink::setMaxFrameSize(unsigned int max_frame_size) throw() {
    __frame_size = max_frame_size;
  }
  
  Boolean BufferedSink::continuePlaying() {
    /* check for exit condition */
    if (!fSource || __flush) {
      return False;
    }
    
    /* allocate buffer if necessary */
    if (!__current_buffer) {
      /* we must have a buffer manager */
      if(!__buffer_manager) {
        GLOBAL_ERROR_STREAM("No buffer manager" << std::endl);
        return False;
      }
      
      /* allocate buffer */
      __current_buffer = __buffer_manager->getNewBuffer(__buffer_size);
      if (!__current_buffer) {
        return False;
      }
      
      /* the buffer is initially empty */
      __current_buffer->setUsed(0);
    }
    
    /* get next frame */
    unsigned used = __current_buffer->getUsed();
    fSource->getNextFrame(reinterpret_cast<unsigned char*>(__current_buffer->getData()) + used, __buffer_size - used,
                          &afterGettingFrame, this, &onSourceClosure, this);
    
    return True;
  }
  
  void BufferedSink::afterGettingFrame(void* clientData, unsigned frameSize, unsigned /*numTruncatedBytes*/, struct timeval presentationTime, unsigned /*durationInMicroseconds*/ ) {
    /* notify the sink */
    BufferedSink* sink = static_cast<BufferedSink*>(clientData);
    sink->handleFrame(frameSize, presentationTime);
    
    /* try getting the next frame */
    sink->continuePlaying();
  }

  void BufferedSink::handleFrame(unsigned frameSize, struct timeval presentationTime) {
    if (__flush) {
      /* discard frame */
      return;
    }

    /* count bytes of frame */
    DEBUG_STREAM2("Received frame of " << frameSize << " bytes" << endl);
    unsigned used = __current_buffer->getUsed();

    RTPSource * rtp_source = 0;
    bool used_for_amr = false;
    AMRAudioSource* amr_source = dynamic_cast<AMRAudioSource*>(fSource);
    if(amr_source) {
      used_for_amr = true;

      // restore 1 byte AMR header that was removed by BufferedSource because liblive doesn't like it
      unsigned char* data_ptr = reinterpret_cast<unsigned char*>(__current_buffer->getData()) + used;
      memmove(data_ptr + 1, data_ptr, frameSize);
      frameSize++;
      *data_ptr = amr_source->lastFrameHeader();

      // TEST: fw: for testing of amr mode change
//      static int last_amr_mode = -1;
//      int amr_mode = -1;
//
//      switch (amr_source->lastFrameHeader()) {
//        case 0x04: { amr_mode = 0; break; }
//        case 0x0C: { amr_mode = 1; break; }
//        case 0x14: { amr_mode = 2; break; }
//        case 0x1C: { amr_mode = 3; break; }
//        case 0x24: { amr_mode = 4; break; }
//        case 0x2C: { amr_mode = 5; break; }
//        case 0x34: { amr_mode = 6; break; }
//        case 0x3C: { amr_mode = 7; break; }
//        case 0x44:
//        {
//          if (amr_source->isWideband()) {
//            amr_mode = 8;
//          }
//          else {
//            cerr << "Unrecognized Mode" << endl; amr_mode = -1;
//          }
//          break;
//        }
//        default:   { cerr << "Unrecognized Mode" << endl; amr_mode = -1; }
//      }
//
//      if ((amr_mode != -1) && (last_amr_mode != amr_mode)) {
//        cerr << "AMR-Mode: " << amr_mode << "; Header: " << static_cast<int>(amr_source->lastFrameHeader()) << endl;
//        last_amr_mode = amr_mode;
//      }

    }

    used += frameSize;
    __current_buffer->setUsed(used);

    /* check if the source is an RTP source */
    Timestamp t;

    if (used_for_amr) {
      RTPReceiverStrategy* receiver_strategy = dynamic_cast<RTPReceiverStrategy*>(__sender);
      if(receiver_strategy) {
        rtp_source = receiver_strategy->getRTPSource();
      }
      else {
        GLOBAL_ERROR_STREAM1("BufferedSink: Can't access RTPReceiverStrategy" << endl);
      }
    }

    if (!rtp_source) {
      // Get rtp_source from parent class if it is not yet set. For AMR it is set by receiver_strategy->getRTPSource().
      rtp_source = dynamic_cast<RTPSource*>(fSource);
    }

    if(rtp_source) {
      /* set the valid flag based on the marker bit */
      if (rtp_source->hasBeenSynchronizedUsingRTCP()) {
        t.is_valid = true;//rtp_source->curPacketMarkerBit();
      } else {
        t.is_valid = false;
      }

      /* get sequence number to reproduce stream counter */
      uint32_t seqno = rtp_source->curPacketRTPSeqNum();
      if(__first_rtp_frame) {
        /* this is the first frame. RTP sequence numbers have random offset */
        __first_rtp_seqno = seqno;
        __first_rtp_frame = false;
      }
      __last_rtp_seqno = seqno;
      t.stream_counter = seqno - __first_rtp_seqno;
    } else {
      /* this does not happen since we always read from RTP */
      GLOBAL_ERROR_STREAM2("NO RTP SOURCE!!!" << std::endl);
      return;
    }
    
    /* subtract time offset */
    // XXX: fw: uncomment for testing
//     presentationTime.tv_sec -= __start_time.tv_sec;
//     if (presentationTime.tv_usec < __start_time.tv_usec) {
//       presentationTime.tv_sec--;
//       presentationTime.tv_usec += 1000000;
//     }
//     presentationTime.tv_usec -= __start_time.tv_usec;
    
    /* convert to NMM timestamp */
    t.sync_time.sec = presentationTime.tv_sec;
    t.sync_time.nsec = presentationTime.tv_usec * 1000;
    __current_buffer->setTimestamp(t);
    
    /* enqueue buffer */
    // XXX: fw: uncomment for testing
//     if(used > 0 > __buffer_size - __frame_size) {
    if(used > 0) {
      MutexGuard mg(__internal_mutex);
      if (__blocked && !__flush) {
        DEBUG_STREAM2("Waiting for free queue slot" << endl);
        __internal_cond.wait();
      }
      if (__flush) {
        return;
      }

      __received_bytes += __current_buffer->getUsed();

      if (!__has_first_sync_frame && __current_buffer->getTimestamp().is_valid) {
        __has_first_sync_frame = true;
        __received_bytes = 0;
        __start_time.tv_sec  = presentationTime.tv_sec;
        __start_time.tv_usec = presentationTime.tv_usec;
      }

      // TEST: fw: Display different in timestamp and received bytes
//      int64_t tsdifference = (presentationTime.tv_sec - __start_time.tv_sec) * 1000000 + (presentationTime.tv_usec - __start_time.tv_usec);
//      int64_t datadifference =  ((__received_bytes * 2500) / (441)); // = (__received_bytes * 1000000) / (44100 * 2 * 2);
//      cerr << setprecision(12)
//          << tsdifference
//          << " "
//          << datadifference
//          << " "
//          << ((double)(tsdifference)/(double)(datadifference))
//          << " "
//          << (tsdifference - datadifference)
//          << endl;

      DEBUG_STREAM2("Passing buffer of " << __current_buffer->getUsed() << " bytes to receiver" << endl);
      if(BLOCKED == __receiver->receiveOneWayBuffer(__current_buffer)) {
        DEBUG_STREAM2("Receiver returned BLOCKED" << endl);
        __blocked = true;
      }
      __current_buffer = NULL;
    }
    /* End< */
  }
  
  void BufferedSink::receiveSignal(Signal signal) {
    switch (signal) {
    case FREE_QUEUE_SLOT:
      {
        DEBUG_STREAM2("Received signal FREE_QUEUE_SLOT from receiver" << endl);
        
        /* wake up the liveMedia thread */
        MutexGuard mg(__internal_mutex);
        __blocked = false;
        __internal_cond.notify();
      }
      break;
    default:
      break;
    }
  }
  
  void BufferedSink::flush() {
    /* wake up the thread */
    MutexGuard mg(__internal_mutex);
    __flush = true;
    __internal_cond.notify();
  }
  
}

#include "nmm/nmm_suffix.hpp"
