/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#ifndef NMM_BUFFERED_SINK_HPP
#define NMM_BUFFERED_SINK_HPP NMM_BUFFERED_SINK_HPP

#include "nmm/nmm_prefix.hpp"

#include "nmm/comm/messaging/ISignalReceiver.hpp"

#include "nmm/utils/thread/ThreadMutex.hpp"
#include "nmm/utils/thread/ThreadCondition.hpp"
#include "nmm/utils/NamedObject.hpp"

#include <live/MediaSink.hh>

namespace NMM {
  
  class IExternalBufferManager;
  class ITSOneWayBufferReceiver;
  class ITSOneWayBufferSender;
  class Buffer;
  
  /** Buffered media sink.
   *
   * This class is a liveMedia sink which produces NMM buffers. It
   * also reproduces the NMM timestamps from the RTP timestamps, the
   * RTP sequence numbers and the RTP marker bit.
   *
   * The buffered sink has no queue. It forwards messages directly to
   * the message receiver, which is usually a Jack. This avoids an
   * unnecessary context switch.
   *
   * @see BufferedSource RTPStrategy
   */
  class BufferedSink
    : public MediaSink,
      public NamedObject,
      public ISignalReceiver
  { 
  public:
    /** Create new BufferedSink.
     *
     * The frame size is the minimum count of bytes that must be free
     * in the buffer before a new one is allocated (the internal
     * buffer may contain more than one frame)
     *
     * @param env The liveMedia UsageEnvironment
     */
    static BufferedSink* createNew(UsageEnvironment& env);
    
    /** Set the one-way buffer receiver.
     *
     * This is the object that will receive the buffers produced by
     * this sink.  If the receiver is 0, buffers will be dropped
     * immediately.
     *
     * @param receiver Buffer receiver
     */
    void setOneWayBufferReceiver(ITSOneWayBufferReceiver* receiver) throw();
    
    /** Set the one-way buffer sender interface.
     *
     * This is the pointer to the object which is the actual buffer
     * sender. It will be passed to the buffer receiver with each
     * buffer. Typically, this is the %RTPStrategy. It can also be 0.
     *
     * @param sender Buffer sender interface
     */
    void setOneWayBufferSender(ITSOneWayBufferSender* sender) throw();
    
    /** Set the buffer manager from which NMM buffers are allocated.
     *
     * @param buffer_manager The buffer manager
     */
    void setBufferManager(IExternalBufferManager* buffer_manager) throw();
    
    /** Set the size of the buffers to allocate.
     *
     * This is the size of NMM buffers that will be allocated.
     *
     * @param Size of a buffer in bytes
     */
    void setBufferSize(unsigned int buffer_size) throw();
    
    /** Set the maximum size of a frame.
     *
     * This is the number of bytes which is requested from the source
     * at a time. It must be less than or equal to the buffer size and
     * at least as large as the largest frame which is expected to be
     * received. Normally buffer size and frame size are equal.
     *
     * @param frame_size Maximum size of a frame in bytes
     */
    void setMaxFrameSize(unsigned int max_frame_size) throw();
    
    /** Flush the stream.
     *
     * This method sets a flag which tells liveMedia to stop getting
     * frames. It must be called to unblock the thread before stopping
     * the liveMedia flow graph.
     */
    void flush();
    
    /** Signal handler.
     *
     * If @c FREE_QUEUE_SLOT is received from the message receiver,
     * then this method notifies the livemedia thread to unblock and
     * continue getting frames.
     */
    virtual void receiveSignal(Signal);
    
  protected:
    /** Constructor.
     *
     * This constructor is called by createNew.
     */
    BufferedSink(UsageEnvironment&);
    
    /** Destructor.
     */
    virtual ~BufferedSink();
    
    /** liveMedia after getting frame handler.
     */
    static void afterGettingFrame(void*, unsigned, unsigned,
                                  struct timeval, unsigned);
    
  private:
    /** @name MediaSink functions.
     */
    //@{
    virtual Boolean continuePlaying();
    //@}
    
    /** Called by afterGettingFrame.
     *
     * @param frameSize Frame size in bytes
     * @param presentationTime Presentation time
     */
    virtual void handleFrame(unsigned frameSize,
                             struct timeval presentationTime);
    
  private:
    int64_t  __received_bytes;
    bool __has_first_sync_frame;

    /** Current buffer to write to */
    Buffer* __current_buffer;
    
    /** The buffer size. */
    size_t __buffer_size;
    
    /** The RTP frame size. */
    size_t __frame_size;
    
    /** Initial RTP timestamp value (to compute offset). */
    struct timeval __start_time;
    
    /** Receiver is currently blocked. */
    bool __blocked;
    
    /** Flush method was called. */
    bool __flush;
    
    /** Internal thread mutex. */
    ThreadMutex __internal_mutex;

    /** Internal thread condition. */
    ThreadCondition __internal_cond;
    
    /** Buffer manager. */
    IExternalBufferManager* __buffer_manager;
    
    /** Message receiver. */
    ITSOneWayBufferReceiver* __receiver;

    /** Message sender. */
    ITSOneWayBufferSender* __sender;
    
    /** No RTP frames received yet. */
    bool __first_rtp_frame;
    
    /** RTP sequence number of first RTP frame. */
    uint32_t __first_rtp_seqno;
    
    /** RTP sequence number of last RTP frame. */
    uint32_t __last_rtp_seqno;
  };
  
}

#include "nmm/nmm_suffix.hpp"

#endif
