/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "BufferedSource.hpp"

#include "nmm/comm/messaging/Buffer.hpp"

#include "nmm/utils/thread/MutexGuard.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  BufferedSource::BufferedSource(UsageEnvironment& env, bool amr_is_wideband)
    : AMRAudioSource(env, amr_is_wideband, 1),
      NamedObject("BufferedSource"),
      __stream_queue(StreamQueue::MODE_SUSPEND, 4, "BufferedSource"),
      __current_buffer(0),
      __buffer_offset(0),
      __samples_per_second(0),
      __sample_size(0),
      __internal_cond(__internal_mutex),
      __flush(false),
      __waiting(false),
      __no_merge(true),
      __used_for_amr(false)
  {
    /* this object reads from the queue, so it is the parent of the queue */
    __stream_queue.setParentSignalReceiver(this);
  }
  
  void BufferedSource::setPartnerSignalReceiver(ISignalReceiver* parent) throw() {
    /* the parent writes to the queue, so it is the partner of the queue */
    __stream_queue.setPartnerSignalReceiver(parent);
  }
  
  void BufferedSource::setSamplesPerSecond(float samples_per_second) throw() {
    __samples_per_second = samples_per_second;
  }
  
  void BufferedSource::setSampleSize(unsigned int sample_size) throw() {
    __sample_size = sample_size;
  }
  
  void BufferedSource::setBufferMergingEnabled(bool enabled) throw() {
    __no_merge = !enabled;
  }

  void BufferedSource::setUsedForAMR(bool b) throw() {
    __used_for_amr = b;
  }

  BufferedSource::~BufferedSource()
  {}

  BufferedSource* BufferedSource::createNew(UsageEnvironment& env, bool amr_is_wideband) {
    return new BufferedSource(env, amr_is_wideband);
  }
  
  void BufferedSource::doGetNextFrame() {
    /* read data until we have a complete frame */
    unsigned char* to = fTo;
        
    fFrameSize = 0;

    /* maximum size of a frame */
    size_t max_size = fMaxSize;
    
    /* merge NMM buffers to frames */
    bool is_first_buffer = true;
    for(;;) {
      /* if we have no current buffer, fetch one from the queue */
      fetchData();
      
      /* check if flush was called */
      if (__flush) {
        /* if we still have a buffer, release it now */
        if(__current_buffer) {
          __current_buffer->release();
          __current_buffer = NULL;
        }
        
        /* exit */
        break;
      }
      
      /* get buffer size */
      unsigned used = __current_buffer->getUsed();

      if (__used_for_amr) {
        // remove 1 byte AMR header because liblive doesn't like it
        uint8_t* data_ptr = reinterpret_cast<uint8_t*>(__current_buffer->getData());
        fLastFrameHeader = *data_ptr;

        // TEST: fw: for testing of amr mode change
//        int amr_mode = -1;
//
//        switch (fLastFrameHeader) {
//          case 0x04: { amr_mode = 0; break; }
//          case 0x0C: { amr_mode = 1; break; }
//          case 0x14: { amr_mode = 2; break; }
//          case 0x1C: { amr_mode = 3; break; }
//          case 0x24: { amr_mode = 4; break; }
//          case 0x2C: { amr_mode = 5; break; }
//          case 0x34: { amr_mode = 6; break; }
//          case 0x3C: { amr_mode = 7; break; }
//          case 0x44:
//          {
//            if (isWideband()) {
//              amr_mode = 8;
//            }
//            else {
//              cerr << "Unrecognized Mode" << endl; amr_mode = -1;
//            }
//            break;
//          }
//          default:   { cerr << "Unrecognized Mode" << endl; amr_mode = -1; }
//        }
//
//        if ((amr_mode != -1) && (last_amr_mode != amr_mode)) {
//          cerr << "AMR-Mode: " << amr_mode << "; Header: " << static_cast<int>(fLastFrameHeader) << endl;
//          last_amr_mode = amr_mode;
//        }

        memmove(data_ptr, data_ptr + 1, used - 1);
        __current_buffer->setUsed(used - 1);
        used--;
      }

      /* convert NMM timestamp to RTP timestamp */
      if (!fFrameSize && __buffer_offset == 0) {
        /* If there are too many artefacts in the video stream, try gettimeofday-timestamps */
        //gettimeofday(&fPresentationTime, 0);
        fPresentationTime.tv_sec = __current_buffer->getTimestamp().sync_time.sec;
        fPresentationTime.tv_usec = __current_buffer->getTimestamp().sync_time.nsec / 1000;
      }

      /* check if maximum size has been reached */
      if ((fFrameSize + used - __buffer_offset) > max_size) {
        /* number of bytes to copy*/
        size_t size = max_size - fFrameSize;

        /* copy data */
        memcpy(to, __current_buffer->getData() + __buffer_offset, size);
        __buffer_offset += size;

        /* advance pointers */
        to += size;
        fFrameSize += size;

        /* maximum size of a frame reached */
        break;
      } else {
        /* number of bytes to copy*/
        size_t size = used - __buffer_offset;

        /* copy data */        
        memcpy(to, __current_buffer->getData() + __buffer_offset, size);
        __current_buffer->release();
        __current_buffer = NULL;

        /* advance pointers */
        to += size;
        fFrameSize += size;

        /* if buffers should not be merged, then this is the end of the frame */
        if(__no_merge) {
          break;
        }
      }

      /* stop if maximum frame size has been reached */
      if(fFrameSize == max_size) {
        break;
      }

      is_first_buffer = false;
    }

    DEBUG_STREAM2("Passing frame of " << fFrameSize << " bytes to RTP sink" << endl);

    /* set duration in microseconds */
    if(__sample_size > 0) {
      const float MICROSECONDS_PER_SECOND = 1000000;
      fDurationInMicroseconds = static_cast<unsigned>(MICROSECONDS_PER_SECOND * float(fFrameSize) / (__sample_size * float(__samples_per_second)));
    } else {
      fDurationInMicroseconds = 0;
    }

    DEBUG_STREAM2("Duration in microseconds " << fDurationInMicroseconds << endl);

    /* no bytes were truncated */
    fNumTruncatedBytes = 0;

    /* check if frame has zero size */
    if(!fFrameSize) {
      /* we are closed */
      DEBUG_STREAM2("BufferedSource: source closed!" << endl);
      handleClosure(this);
      return;
    }

    /* switch to another task, and inform the reader that he has data */
    nextTask() = envir().taskScheduler().scheduleDelayedTask(0, reinterpret_cast<TaskFunc*>(afterGetting), this);
  }

  void BufferedSource::doStopGettingFrames() {
    handleClosure(this);
  }
  
  Result BufferedSource::writeBuffer(const Buffer* buffer) {
    /* enqueue buffer */
    // XXX: fw: const_cast
    MutexGuard mg(__internal_mutex);
    DEBUG_STREAM2("Placing buffer of " << buffer->getUsed() << " bytes in send queue" << endl);
    Result result = __stream_queue.receiveMessage(const_cast<Buffer*>(buffer));
    if (result == BLOCKED) {
      DEBUG_STREAM2("Send queue returned BLOCKED" << endl);
    }
    return result;
  }
  
  void BufferedSource::receiveSignal(Signal signal) {
    switch (signal) {
    case QUEUE_NOT_EMPTY:
      {
        DEBUG_STREAM2("Received signal QUEUE_NOT_EMPTY from send queue" << endl);
        
        /* wake up the thread */
	__internal_cond.notify();
      }
      break;
    default:
      break;
    }
  }
  
  void BufferedSource::flush() {
    /* wake up the thread */
    __flush = true;
    
    MutexGuard mg(__internal_mutex);
    __internal_cond.notify();
    
    /* flush the send queue */
    __stream_queue.flush();
  }
  
  void BufferedSource::fetchData() {
    /* wait for a buffer or a flush */
    while(!__flush && !__current_buffer) {
      Message* message = NULL;
      {
	MutexGuard mg(__internal_mutex);
        
	/* dequeue message from queue */
	message = __stream_queue.dequeueMessage();
	
	/* check if the queue is empty */
	if(!message) {
	  __internal_cond.wait();
	}
      }
      
      /* check if we have a buffer */
      if(message) {
        Buffer* buffer = dynamic_cast<Buffer*>(message);
        if(buffer) {
          __current_buffer = buffer;
          __buffer_offset = 0;
        } else {
          /* should not happen */
          message->release();
          ERROR_STREAM2("Message is not a buffer" << std::endl);
          break;
        }
      }
    }
  }
  
}

#include "nmm/nmm_suffix.hpp"
