/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#ifndef NMM_BUFFERED_SOURCE_HPP
#define NMM_BUFFERED_SOURCE_HPP NMM_BUFFERED_SOURCE_HPP

#include "nmm/nmm_prefix.hpp"

#include "nmm/comm/messaging/ISignalReceiver.hpp"
#include "nmm/comm/messaging/StreamQueue.hpp"

#include "nmm/utils/NamedObject.hpp"

#include <live/AMRAudioSource.hh>

namespace NMM {
  
  class Buffer;
  
  /** Buffered media source.
   *
   * This class is a liveMedia FramedSource which reads data from an
   * NMM buffer. It has an internal StreamQueue to which buffers can
   * be passed.
   */
  class BufferedSource :
    public AMRAudioSource,
    public ISignalReceiver,
    public NamedObject
  {
  public:
    /** Create a new BufferedSource.
     *
     * @param env liveMedia UsageEnvironment
     */
    static BufferedSource* createNew(UsageEnvironment& env, bool amr_is_wideband = false);
    
    /** Set the signal receiver interface of the partner.
     *
     * This method sets the signal receiver which will receive
     * @c FREE_QUEUE_SLOT events when the input queue is no longer
     * full.
     */
    void setPartnerSignalReceiver(ISignalReceiver* parent) throw();
    
    /** Set the number of samples per second.
     *
     * @param samples_per_second Number of samples per second
     */
    void setSamplesPerSecond(float samples_per_second) throw();
    
    /** Set the size of a sample.
     *
     * If the sample size is set to 0, then this source will always
     * pass 0 as the play duration of a buffer to the sink.
     *
     * @param sample_size The sample size in bytes
     */
    void setSampleSize(unsigned int sample_size) throw();
    
    /** Set whether merging of incoming buffers is enabled.
     *
     * If enabled, incoming buffers may be merged by the
     * %BufferedSource and data is passed to the RTP library in larger
     * chunks.
     *
     * @note This option was used for testing and should not be used
     */
    void setBufferMergingEnabled(bool enabled) throw();

    void setUsedForAMR(bool b) throw();
    
    /** Write buffer.
     *
     * This method puts the buffer into the queue. If the previous
     * call to this method returned @c BLOCKED, then it may not be
     * called again until the parent has received @c FREE_QUEUE_SLOT.
     *
     * @return @c BLOCKED if the internal queue is full, otherwise @c SUCCESS.
     */
    Result writeBuffer(const Buffer* buffer);
    
    /** Flush the stream.
     *
     * This method flushes the queue and unblocks the thread.
     */
    void flush();
    
    /** Get stream queue.
     */
    StreamQueue& getStreamQueue() throw();
    
  protected:
    /** Protected constructor.
     *
     * This constructor is called by createNew.
     */
    BufferedSource(UsageEnvironment&, bool amr_is_wideband);
    
    /** Destructor.
     */
    virtual ~BufferedSource();
    
    /** Handler called upon stopPlaying.
     */
    virtual void doStopGettingFrames();
    
    /** Receive a signal.
     */
    virtual void receiveSignal(Signal signal);
    
  private:
    /** @name FramedSource functions
     */
    //@{
    /** Get next frame.
     *
     * This method copies the next frame to the destination buffer
     */
    virtual void doGetNextFrame();
    //@}
    
    /** Fetch new data.
     *
     * This method fetches new data from the send queue. It ensures
     * that a buffer is available for reading data from. If there is
     * no buffer, the method tries to dequeue a buffer from the send
     * queue. Messsages which are not buffers are discarded
     * immediately. If the queue is empty, then this method waits
     * until the queue becomes non-empty or BufferedSource::flush is
     * called.
     */
    void fetchData();
    
  private:
    /** The internal queue. */
    StreamQueue __stream_queue;
    
    /** Current buffer to read from. */
    Buffer* __current_buffer;
    
    /** Current byte offset within the buffer. */
    size_t __buffer_offset;
    
    /** Samples per second. */
    float __samples_per_second;
    
    /** Sample size in bytes. */
    unsigned int __sample_size;
    
    /** Internal thread mutex. */
    ThreadMutex __internal_mutex;
    
    /** Internal thread condition. */
    ThreadCondition __internal_cond;

    /** Flush method was called. */
    bool __flush;
    
    /** We are waiting for buffers. */
    bool __waiting;
    
    /** Don't merge buffers to frames. */
    bool __no_merge;

    /** BufferedSource is used for streaming AMR */
    bool __used_for_amr;
  };
  
  
  inline StreamQueue& BufferedSource::getStreamQueue() throw() {
    return __stream_queue;
  }
  
}

#include "nmm/nmm_suffix.hpp"

#endif
