/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#ifndef NMM_IP_ADDRESS_HPP
#define NMM_IP_ADDRESS_HPP NMM_IP_ADDRESS_HPP

#include "nmm/nmm_prefix.hpp"

#include "nmm/comm/serialize/SerializedValue.hpp"
#include "nmm/comm/BindingException.hpp"

#include "nmm/utils/net/nmm_sockaddr_in.hpp"

#include "visibility.hpp"

namespace NMM {

  /** A generic IP address.
   * An IP address which is a part of a transport strategy address.
   * This class can not be instantiated directly. Create a
   * UnicastIPAddress or a MulticastIPAddress instead.
   *
   * @see UnicastIPAddress, MulticastIPAddress
   */
  class DLLEXPORT IPAddress : public SerializedValue {
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Virtual copy constructor.
     *
     * @return A copy of this object which has the same type
     */
    inline virtual IPAddress* copy() const throw() {
      return new IPAddress(*this);
    }
    //@}

    /** @name Access Methods
     */
    //@{
    /** Get the IP address.
     * This method returns the IP address as a 32-bit integer.
     *
     * @return The IP address
     */
    inline uint32_t getIPAddress() const throw() {
      return __ip_address.getRValue();
    }

    /** Set the IP address.
     * This method sets the IP address as a 32-bit integer.
     *
     * @param ip The IP address
     */
    void setIPAddress(const uint32_t& ip) throw();

    /** Set the IP address by host name.
     * This method sets the IP address by a host name. The host is
     * looked up, which may block for a short time and possibly fail.
     * The host name may also be an IP address represented in dotted
     * form.
     *
     * @param hostname Host name or IP address
     * @exception BindingException if the host was not found
     */
    void setHostname(const std::string& hostname)
      throw(const BindingException&);
    //@}

    /** @name Access Internal Structures
     */
    //@{
    /** Get the internal representation.
     *
     * @return A reference to the internal representation
     */
    inline const in_addr& getINetAddr() const throw() {
      return __net_addr;
    }

    /** Get the internal representation.
     *
     * @return A reference to the internal representation
     */
    inline in_addr& getINetAddr() throw() {
      return __net_addr;
    }
    //@}

    /** @name Debugging
     */
    //@{
    /** Get a string representation of this IP address.
     * This method returns a string representation in dotted form of
     * this IP address.
     *
     * @return String representation of this IP address
     */
    string toString() const throw();
    //@}

  protected:
    /** Update members after serialization.
     * This implementation updates the internal representation.
     */
    virtual void updateAttributes();

  protected:
    /** @name Private Constructor & Destructor
     */
    //@{
    /** Create a new IP address.
     *
     * @param typeinfo The type info of the class of this object
     */
    IPAddress(const TypeInfo& typeinfo) throw();

    /** Create a new IP address.
     *
     * @param typeinfo The type info of the class of this object
     * @param hostname The hostname
     * @exception BindingException if the host was not found
     */
    IPAddress(const TypeInfo& typeinfo, const string& hostname)
      throw (const BindingException&);

    /** Copy constructor.
     *
     * @param addr Object to copy from
     */
    IPAddress(const IPAddress& addr) throw();

  private:
    in_addr __net_addr;               ///< Internal representation
    TValue<nmm_uint32> __ip_address;    ///< IP address
  };
  
}

#include "nmm/nmm_suffix.hpp"

#endif
