/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#ifndef NMM_IRTPControl_HPP
#define NMM_IRTPControl_HPP

#include "nmm/nmm_prefix.hpp"

#include "nmm/comm/Interface.hpp"
#include "nmm/comm/messaging/Event.hpp"
#include "nmm/comm/messaging/IllegalStateException.hpp"
#include <memory>

#include "IIPControl.hpp"

#include "visibility.hpp"

namespace NMM
{
  /** RTP control interface.
   * This interface provides method for getting and setting options
   * for the Real-time Transport Protocol (RTP).
   */

  class DLLEXPORT IRTPControl : virtual public IIPControl { 
  public:
    IRTPControl() : NMM::Interface(IRTPControl::IRTPControl_type, "NMM") { };

    IRTPControl(NMM::NMMObject& parent) : NMM::Interface (parent, IRTPControl::IRTPControl_type, "NMM") { };

    virtual ~IRTPControl() {};

    static const char* IRTPControl_type;

    virtual  string getInterfaceType() const { return IRTPControl_type; }

    NMM::Interface* copy() const { return new IRTPControl(*this); }
    static void used(); 
    static const char* setCNAME_event;
    static const char* getCNAME_event;
    static const char* setTTL_event;
    static const char* getTTL_event;
    static const char* setBufferSize_event;
    static const char* getBufferSize_event;
    static const char* setMaxFrameSize_event;
    static const char* getMaxFrameSize_event;
    static const char* setSessionBandwidth_event;
    static const char* getSessionBandwidth_event;
    static const char* sendRTCPReport_event;
    /** @name Get/Set RTP Options
     */
//@{
    /** Set canonical name.
     * This method sets the canonical name of the RTP session. This
     * is a free-form string which identifies an RTP session.
     * The default is "NMM-RTP".
     *
     * @param cname The canonical name
     */

    virtual void setCNAME(const string& cname)
      throw(const NMM::SerializedException&);
    /** Get canonical name.
     * This method gets the canonical name of the RTP session. This
     * is a free-form string which identifies an RTP session.
     * The default is "NMM-RTP".
     *
     * @return The canonical name
     */

    virtual string getCNAME() const
      throw(const NMM::SerializedException&);
//@}
    /** @name Connection Parameters
     */
//@{
    /** Set multicast TTL.
     * This method sets the time-to-live (TTL) for multicast packets.
     * For unicast sessions, this parameter has no effect.
     * The TTL is the maximum number of hops a packet travels until it is deleted.
     *
     * @param ttl The multicast TTL
     */

    virtual void setTTL(const nmm_uint32& ttl)
      throw(const NMM::SerializedException&);
    /** Get multicast TTL.
     * This method gets the time-to-live (TTL) for multicast packets.
     * For unicast sessions, this parameter has no effect.
     * The TTL is the maximum number of hops a packet travels until it is deleted.
     *
     * @return The multicast TTL
     */

    virtual nmm_uint32 getTTL() const
      throw(const NMM::SerializedException&);
    /** Set buffer size.
     * This method sets the size of NMM buffers allocated by the RTP receiver.
     *
     * @param buffer_size The buffer size
     */

    virtual void setBufferSize(const nmm_uint32& buffer_size)
      throw(const NMM::SerializedException&);
    /** Get buffer size.
     * This method sets the size of NMM buffers allocated by the RTP receiver.
     *
     * @return The buffer size (in bytes)
     */

    virtual nmm_uint32 getBufferSize() const
      throw(const NMM::SerializedException&);
    /** Set maximum frame size.
     * This method sets the expected maximum size of a received RTP packet (frame).
     * If an incoming frame is larger than this size, it is truncated.
     *
     * @param max_frame_size The maximum frame size (in bytes)
     */

    virtual void setMaxFrameSize(const nmm_uint32& max_frame_size)
      throw(const NMM::SerializedException&);
    /** Get maximum frame size.
     * This method gets the expected maximum size of a received RTP packet (frame).
     * If an incoming frame is larger than this size, it is truncated.
     *
     * @return The maximum frame size (in bytes)
     */

    virtual nmm_uint32 getMaxFrameSize() const
      throw(const NMM::SerializedException&);
    /** Set RTCP session bandwidth.
     * This method sets the session bandwidth of the RTCP channel.
     *
     * @param bandwidth The session bandwidth (in kb/s)
     */

    virtual void setSessionBandwidth(const nmm_uint32& bandwidth)
      throw(const NMM::SerializedException&);
    /** Get RTCP session bandwidth.
     * This method sets the session bandwidth of the RTCP channel.
     *
     * @return The session bandwidth (in kb/s)
     */

    virtual nmm_uint32 getSessionBandwidth() const
      throw(const NMM::SerializedException&);
//@}
    /** @name RTCP Control
     */
//@{
    /** Explicitly send an RTCP report.
     *
     * This method causes the transport strategy to send an RTCP sender or receiver report.
     * Normally, these reports are sent automatically at regular intervals. This method can
     * be used to enforce sending additional reports or sending reports at shorter intervals.
     */

    virtual void sendRTCPReport()
      throw(const NMM::SerializedException&);

//@}

  };

  typedef auto_ptr<IRTPControl> IRTPControl_var;




}

#include "nmm/nmm_suffix.hpp"

#endif
