/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#ifndef NMM_IRTPTransmissionStats_HPP
#define NMM_IRTPTransmissionStats_HPP

#include "nmm/nmm_prefix.hpp"

#include "nmm/comm/Interface.hpp"
#include "nmm/comm/messaging/Event.hpp"
#include "nmm/comm/messaging/IllegalStateException.hpp"
#include <memory>

/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * See file LICENCE for terms and conditions of usage
 *
 */

#include "nmm/utils/time/Interval.hpp"
#include "nmm/comm/serialize/SerializeUtilsTime.hpp" 

#include "visibility.hpp"

namespace NMM
{
  /** Interface for querying RTP transmission statistics.
   *
   * This interface can be used for querying the transmission
   * statistics of the RTP sink which are obtained by RTCP receiver
   * reports.
   *
   * @todo Find out what all methods do, remove undocumented methods
   *
   * @see IRTPControl IRTPReceptionStats RTPSenderStrategy
   */

  class DLLEXPORT IRTPTransmissionStats : virtual public NMM::Interface { 
  public:
    IRTPTransmissionStats() : NMM::Interface(IRTPTransmissionStats::IRTPTransmissionStats_type, "NMM") { };

    IRTPTransmissionStats(NMM::NMMObject& parent) : NMM::Interface (parent, IRTPTransmissionStats::IRTPTransmissionStats_type, "NMM") { };

    virtual ~IRTPTransmissionStats() {};

    static const char* IRTPTransmissionStats_type;

    virtual  string getInterfaceType() const { return IRTPTransmissionStats_type; }

    NMM::Interface* copy() const { return new IRTPTransmissionStats(*this); }
    static void used(); 
    static const char* getLastReceivedPacketNumber_event;
    static const char* getFirstReportetPacketNumber_event;
    static const char* getTotalNumberOfPackets_event;
    static const char* getTotalNumberOfBytes_event;
    static const char* getNumberOfLostPackets_event;
    static const char* getJitter_event;
    static const char* getPacketLossRatio_event;
    static const char* getRoundTripDelay_event;
    static const char* getPacketsLostBetweenRR_event;
    /** Get the sequence number of the last received packet.
     *
     * @todo Find out what this method REALLY returns 
     */

    virtual nmm_uint32 getLastReceivedPacketNumber() const
      throw(const NMM::SerializedException&);
    /** Get the sequence number of the first received packet.
     *
     * @todo Find out what this method REALLY returns
     */

    virtual nmm_uint32 getFirstReportetPacketNumber() const
      throw(const NMM::SerializedException&);
    /** Get the total number of packets transmitted.
     */

    virtual nmm_uint64 getTotalNumberOfPackets() const
      throw(const NMM::SerializedException&);
    /** Get the total number of bytes transmitted.
     */

    virtual nmm_uint64 getTotalNumberOfBytes() const
      throw(const NMM::SerializedException&);
    /** Get the number of lost packets.
     *
     * @todo Find out what this method REALLY returns
     */

    virtual nmm_uint32 getNumberOfLostPackets() const
      throw(const NMM::SerializedException&);
    /** Get the inter-arrival jitter.
     *
     * This method gets the packet inter-arrival jitter based on the
     * receiver reports. It is defined by RFC3550 as follows:
     *
     * "An estimate of the statistical variance of the RTP data packet
     * interarrival time, measured in timestamp units and expressed as
     * an unsigned integer. The interarrival jitter @a J is defined to
     * be the mean deviation (smoothed absolute value) of the
     * difference @a D in packet spacing at the receiver compared to
     * the sender for a pair of packets."
     *
     * @return The inter-arrival jitter in timestamp units
     */

    virtual Interval getJitter() const
      throw(const NMM::SerializedException&);

    virtual nmm_float getPacketLossRatio() const
      throw(const NMM::SerializedException&);

    virtual Interval getRoundTripDelay() const
      throw(const NMM::SerializedException&);

    virtual nmm_int32 getPacketsLostBetweenRR() const
      throw(const NMM::SerializedException&);


  };

  typedef auto_ptr<IRTPTransmissionStats> IRTPTransmissionStats_var;




}

#include "nmm/nmm_suffix.hpp"

#endif
