/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "PacketINetStream.hpp"
#include "PacketHeader.hpp"

#include "nmm/comm/serialize/TConstructor.hpp"

// for checking whether socket is SockStream
#include "nmm/utils/net/SockStream.hpp"

#include "nmm/utils/thread/MutexGuard.hpp"

#include <cerrno>

#include "nmm/nmm_prefix.hpp"

namespace NMM {  
  
  PacketINetStream::PacketINetStream()
    : INetStream(TypeInfo("PacketINetStream", "NMM")),
      __max_packet_size(8192),
      __max_partial(10),
      __discard_old_packets(false),
      __i_packet(__max_packet_size.getRValue()),
      __stream_socket(false),
      __i_message_pos(0),
      __i_message_id(0),
      __total_received_bytes(0),
      __total_received_packets(0),
      __total_discarded_packets(0),
      __total_lost_packets(0)
  {
    registerAttributes();
    
    if(dynamic_cast<SockStream*>(getSock())) {
      __stream_socket = true;
    }
  }
  
  PacketINetStream::PacketINetStream(const TypeInfo& ti)
    : INetStream(ti),
      __max_packet_size(8192),
      __max_partial(10),
      __discard_old_packets(false),
      __i_packet(__max_packet_size.getRValue()),
      __stream_socket(false),
      __i_message_pos(0),
      __i_message_id(0),
      __total_received_bytes(0),
      __total_received_packets(0),
      __total_discarded_packets(0),
      __total_lost_packets(0)
  {
    registerAttributes();
    
    if(dynamic_cast<SockStream*>(getSock())) {
      __stream_socket = true;
    }
  }
  
  PacketINetStream::PacketINetStream(const PacketINetStream& other)
    : INetStream(other),
      __max_packet_size(other.__max_packet_size.getRValue()),
      __max_partial(other.__max_partial.getRValue()),
      __discard_old_packets(other.__discard_old_packets.getRValue()),
      __i_packet(__max_packet_size.getRValue()),
      __stream_socket(false),
      __i_message_pos(0),
      __i_message_id(0),
      __total_received_bytes(0),
      __total_received_packets(0),
      __total_discarded_packets(0),
      __total_lost_packets(0)
  {
    registerAttributes();
    
    if (dynamic_cast<SockStream*>(getSock())) {
      __stream_socket = true;
    } 
  }
  
  void PacketINetStream::registerAttributes() {
    insertArgument(&__max_packet_size);
    insertArgument(&__max_partial);
    insertArgument(&__discard_old_packets);	
  }
  
  void PacketINetStream::updateAttributes() {
    INetStream::updateAttributes();
    
    /* resize internal buffer */
    MutexGuard g(__packet_mutex);
    if(static_cast<nmm_uint32>(__i_packet.size()) != __max_packet_size.getRValue()) {
      __i_packet.resize(__max_packet_size.getRValue());
    }
  }
  
  PacketINetStream::~PacketINetStream() {}
  
  PacketINetStream* PacketINetStream::copy() const {
    return new PacketINetStream(*this);
  }
  
  void PacketINetStream::readValue(void* value, size_t size) {
    /* Check if size exceeds buffer size */
    if (__i_message_pos + size > __i_message.size()) {
      cerr << "PacketINetStream: Read buffer overflow!" << endl;
      throw Exception("Read buffer overflow!");
    }
    
    /* Read data from internal buffer */
    memcpy(value, &__i_message[__i_message_pos], size);
    __i_message_pos += size;
  }
  
  bool PacketINetStream::receivePacket() {
    MutexGuard g(__packet_mutex);
    
    /* Read packet from stream */
    unsigned int size = readPacket(&__i_packet[0]);
    if (!size) {
      throw ConnectionClosed(EPIPE);
    }
    
    /* Check if packet is big enough to contain a complete header */
    if (size < static_cast<unsigned int>(sizeof(PacketHeader))) {
      GLOBAL_ERROR_STREAM("PacketINetStream: Received packet is truncated (no valid header)" << endl);
      return false;
    }
    
    /* Analyze new packet header */
    PacketHeader& new_header = *reinterpret_cast<PacketHeader*>(&__i_packet[0]);
    int new_message_id = new_header.getMessageID();
    int new_packet_no = new_header.getPacketNumber();
    unsigned int new_packet_size = new_header.getPacketSize();
    
    /* Verify packet size versus header data */
    unsigned int correct_size = sizeof(PacketHeader) + new_packet_size;
    if (size != correct_size) {
      GLOBAL_ERROR_STREAM("PacketINetStream: Received packet is corrupt (size=" << size
                          << ", should be " << correct_size << ")" << endl);
      return false;
    }
    
    /* From this point on, assume packet is consistent */
    ++__total_received_packets;
    GLOBAL_DEBUG_STREAM2("PacketINetStream: receivePacket"
                         << " message_id=" << new_message_id
                         << " packet_no=" << new_packet_no
                         << " packet_count=" << new_header.getPacketCount()
                         << " size=" << new_header.getPacketSize()
                         << endl);
    
    /* Insert the packet into the queue */
    /* The list is sorted by message ID and packet number */
    list<Packet>::iterator i;
    for (i=__i_packets.begin(); i != __i_packets.end(); ++i) {
      /* Analyze existing packet header */
      Packet& packet = *i;
      PacketHeader& header = *reinterpret_cast<PacketHeader*>(&packet[0]);
      int message_id = header.getMessageID();
      int packet_no = header.getPacketNumber();
      
      if (message_id == new_message_id) {
        if (packet_no == new_packet_no) {
          /* Packet received multipl times. ignore it */
          continue;
        } else if (packet_no < new_packet_no) {
          continue;
        }
      } else if (message_id < new_message_id) {
        continue;
      }
      
      break;
    }
    
    /* Insert before this packet */
    __i_packets.insert(i, __i_packet);
    
    if ( __discard_old_packets.getRValue() ) {
      /* Discard old messages */
      discardOldPackets();
    } 

    /* Check if we have a complete message now */
    return checkForCompleteMessage();
  }
  
  void PacketINetStream::readData() {
    int read_pos = 0;
    int packet_no = 0;
    int packet_count;

    GLOBAL_DEBUG_STREAM2("PacketINetStream: readData" << endl);
    
    if (__i_message_id < 0) {
      // XXX: fw: should throw an exception
//       throw Exception("Don't have a complete message!");
      GLOBAL_ERROR_STREAM("PacketINetStream::writeData() called, but no complete message available!");
      return;
    }
    
    /* Put all packets of the current input message together */
    /* Assume that the packets of this message are stored in the right order */
    for (list<Packet>::iterator i=__i_packets.begin(); i != __i_packets.end(); ) {
      Packet& packet = *i;
      
      /* Read packet header and check message ID */
      PacketHeader& header = *reinterpret_cast<PacketHeader*>(&packet[0]);
      if ((int)(header.getMessageID()) != __i_message_id) {
        ++i;
        continue;
      }
      unsigned int packet_bytes = header.getPacketSize();
      packet_count = header.getPacketCount();
      if (packet_no != (int)(header.getPacketNumber())) {
        GLOBAL_ERROR_STREAM2("PacketINetStream: readData: WRONG PACKET_NO!!!" << endl);
        // XXX: fw: should throw an exception
//         throw Exception("Packet not stored in correct order! SHOULD NOT HAPPEN!");
      }
      
      GLOBAL_DEBUG_STREAM2("PacketINetStream: add packet " << packet_no << endl);
      
      /* Enlarge message input buffer if necessary */
      if (static_cast<unsigned int>(read_pos + packet_bytes) > static_cast<unsigned int>(__i_message.size())) {
        unsigned int new_size = __i_message.size();
        if (!new_size) {
          new_size = 8;
        }
        while (new_size < static_cast<unsigned int>(read_pos + packet_bytes)) {
          new_size += new_size;
        }
        __i_message.resize(new_size);
        
        GLOBAL_DEBUG_STREAM2("PacketINetStream: resize input message buffer. Size is now " << __i_message.size() << endl);
      }
      
      /* Read packet data */
      memcpy(&__i_message[read_pos], &packet[sizeof(PacketHeader)], packet_bytes);
      
      read_pos += packet_bytes;
      ++packet_no;
      
      /* Remove packet from list */
      i = __i_packets.erase(i);
      
      /* Check if done */
      if (packet_no >= packet_count) {
        break;
      }
    }
    
    __i_message_pos = 0;
    ++__i_message_id;
  }
  
  unsigned int PacketINetStream::readPacket(void* packet) {
    unsigned int size;
    if (__stream_socket) {
      SockStream* stream_sock = dynamic_cast<SockStream*>(getSock());
      stream_sock->readStream(static_cast<char*>(packet), sizeof(PacketHeader), sizeof(PacketHeader));
      size = static_cast<PacketHeader*>(packet)->getPacketSize() + sizeof(PacketHeader);
      stream_sock->readStream(static_cast<char*>(packet) + sizeof(PacketHeader), size - sizeof(PacketHeader), size - sizeof(PacketHeader));
    } else {
      size = getSock()->recv(static_cast<char*>(packet), __max_packet_size.getRValue());
      if (!size) {
        return 0;
      }
    }
    
    __total_received_bytes += size;
    
    GLOBAL_DEBUG_STREAM2("PacketINetStream: readPacket raw_size=" << size << endl);
    return size;
  }
  
  bool PacketINetStream::checkForCompleteMessage() {
    int curr_message_id = -1;
    int curr_packet_no = -1;

    GLOBAL_DEBUG_STREAM2("PacketINetStream: check for complete message" << endl);
    
    for (list<Packet>::iterator i=__i_packets.begin(); i != __i_packets.end(); ++i) {
      /* Analyze existing packet header */
      Packet& packet = *i;
      PacketHeader& header = *reinterpret_cast<PacketHeader*>(&packet[0]);
      int message_id = header.getMessageID();
      int packet_count = header.getPacketCount();
      
      if (message_id > curr_message_id) {
        /* New message. Restart the packet counter */
        curr_message_id = message_id;
        curr_packet_no = 0;
      }
      
      if (curr_packet_no == packet_count-1) {
        /* We have all packets of this message */
        __i_message_id = message_id;
        return true;
      }
      
      ++curr_packet_no;
    }
    
    /* No complete message found */
    __i_message_id = -1;
    return false;
  }
  
  void PacketINetStream::discardOldPackets() {
    /* This method discards all packets with a message ID smaller than
     * __i_message_id - 10.  This eventually limits the size of the
     * incoming packet queue.
     *
     * It also counts the number of lost packets. A packet is
     * considered lost, if all packets with the same message ID have
     * been discarded by this method and the packet in question has
     * not been received.  In other words, this method discards all
     * packets in the queue whose message ID is too old and counts the
     * number of discarded packets. The difference between the number
     * of discarded packets and the number of packets that make up the
     * discarded messages is the estimated packet lost.  It is an
     * estimation because the packets considered lost might still
     * arrive, though this is rather unlikely. If we store "waterline"
     * of discarded message fragments, we can also count these
     * packets.
     */
    
    /* outer loop - iterates over messages */
    while(!__i_packets.empty()) {
      /* get first packet information */
      Packet& first_packet = *__i_packets.begin();
      PacketHeader& first_header = *reinterpret_cast<PacketHeader*>(&first_packet[0]);
      int message_id = first_header.getMessageID();
      int packet_count = first_header.getPacketCount();
      
      /* check if message is old enough */
      if((int)message_id >= (int)(__i_message_id - __max_partial.getRValue())) {
        /* no more old packets */
        break;
      }
      
      /* inner loop - iterates over packets*/
      int packets_found = 0;
      int prev_packet_number = -1;
      do {
        /* get packet information */
        Packet& packet = *__i_packets.begin();
        PacketHeader& header = *reinterpret_cast<PacketHeader*>(&packet[0]);
        int this_message_id = header.getMessageID();
        int packet_number = header.getPacketNumber();
        
        /* display lost packets */
        for(int i=prev_packet_number+1; i<packet_number; ++i) {
          cout << "*";
        }
        prev_packet_number = packet_number;
        
        if(message_id != this_message_id) {
          /* packet belongs to a new message */
          break;
        }
        
        /* discard packet */
        ++packets_found;
        __i_packets.pop_front();
        cerr << "Discard old messages with " << header.getMessageID() << endl;
        cerr << "Discard old packet with " << header.getPacketNumber() << endl;      
     } while(!__i_packets.empty());

      /* display lost packets at end of message */
      for(int i=prev_packet_number+1; i<packet_count; ++i) {
        cout << "*";
      }
      
      if(packets_found < packet_count) {
        cout << endl;
      }
      
      /* update statistics */
      __total_lost_packets += packet_count - packets_found;
      __total_discarded_packets += packets_found;
    }
  }
  
  void PacketINetStream::setPacketSize(const unsigned int& size) {
    MutexGuard g(__packet_mutex);
    if (size == __max_packet_size.getRValue()) {
      return;
    }
    
    __max_packet_size = size;
    __i_packet.resize(size);
  }
  
  unsigned int PacketINetStream::getPacketSize() const {
    MutexGuard g(__packet_mutex);
    return __max_packet_size.getRValue();
  }
  
  void PacketINetStream::setMaxPartialMessages(const unsigned int& max_partial) {
    MutexGuard g(__packet_queue_mutex);
    __max_partial = max_partial;
  }
  
  unsigned int PacketINetStream::getMaxPartialMessages() const {
    MutexGuard g(__packet_queue_mutex);
    return __max_partial.getRValue();
  }
  
  void PacketINetStream::resetStatistics() {
    __total_received_bytes = 0;
    __total_received_packets = 0;
    __total_discarded_packets = 0;
    __total_lost_packets = 0;
  }
  
  unsigned int PacketINetStream::getTotalReceivedBytes() const {
    return __total_received_bytes;
  }
  
  unsigned int PacketINetStream::getTotalReceivedPackets() const {
    return __total_received_packets;
  }
  
  unsigned int PacketINetStream::getTotalDiscardedPackets() const {
    return __total_discarded_packets;
  }
  
  unsigned int PacketINetStream::getTotalLostPackets() const {
    return __total_lost_packets;
  }

  template<>
  class TConstructor<PacketINetStream> : public ValueConstructor {
  public:
    TConstructor(const TypeInfo& t_info)
      : ValueConstructor(t_info)
    { }
    virtual PacketINetStream*          createValue()          const {return new PacketINetStream();}
    virtual TValue<PacketINetStream>*  createTValue()         const {return new TValue<PacketINetStream>();}
  
    virtual TValue<INetStream*>*  createTPointerValue()  const {return new TValue<INetStream*>();}
  
    virtual TConstructor<PacketINetStream>* copy()  const {return new TConstructor(*this); }
  private:
    string  __type_name;
    string  __namespace;
  };
  
  TConstructor<PacketINetStream> __nmm__packet_i_net_stream(TypeInfo("PacketINetStream", "NMM"));
  
}

#include "nmm/nmm_suffix.hpp"
