/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "PacketONetStream.hpp"
#include "PacketHeader.hpp"

#include "nmm/comm/serialize/TConstructor.hpp"

// for checking whether socket is SockStream
#include "nmm/utils/net/SockStream.hpp"

#include "nmm/utils/thread/MutexGuard.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  PacketONetStream::PacketONetStream()
    : ONetStream(TypeInfo("PacketONetStream", "NMM")),
      __max_packet_size(8192),
      __o_packet(__max_packet_size.getRValue()),
      __stream_socket(false),
      __o_message_pos(0),
      __o_message_id(0),
      __total_sent_bytes(0),
      __total_sent_packets(0)
  {
    registerAttributes();
    
    if(dynamic_cast<SockStream*>(getSock())) {
      __stream_socket = true;
    }
  }

  PacketONetStream::PacketONetStream(const TypeInfo& ti)
    : ONetStream(ti),
      __max_packet_size(8192),
      __o_packet(__max_packet_size.getRValue()),
      __stream_socket(false),
      __o_message_pos(0),
      __o_message_id(0),
      __total_sent_bytes(0),
      __total_sent_packets(0)
  {
    registerAttributes();
    
    if(dynamic_cast<SockStream*>(getSock())) {
      __stream_socket = true;
    }
  }
  
  PacketONetStream::PacketONetStream(const PacketONetStream& other)
    : ONetStream(other),
      __max_packet_size(other.__max_packet_size.getRValue()),
      __o_packet(__max_packet_size.getRValue()),
      __stream_socket(false),
      __o_message_pos(0),
      __o_message_id(0),
      __total_sent_bytes(0),
      __total_sent_packets(0)
  {
    registerAttributes();
    
    if (dynamic_cast<SockStream*>(getSock())) {
      __stream_socket = true;
    } 
  }
  
  void PacketONetStream::registerAttributes() {
    insertArgument(&__max_packet_size);
  }
  
  void PacketONetStream::updateAttributes() {
    ONetStream::updateAttributes();
    
    /* resize internal buffer */
    MutexGuard g(__packet_mutex);
    if(static_cast<int>(__o_packet.size()) != (int)__max_packet_size.getRValue()) {
      __o_packet.resize(__max_packet_size.getRValue());
    }
  }
  
  PacketONetStream::~PacketONetStream() {}
  
  PacketONetStream* PacketONetStream::copy() const {
    return new PacketONetStream(*this);
  }
  
  void PacketONetStream::writeValue(const void* value, size_t size) {
    /* Check if buffer is large enough, enlarge it if needed */
    if (__o_message_pos + size > __o_message.size()) {
      unsigned int new_size = __o_message.size();
      if (!new_size) {
        new_size = 8;
      }
      while (new_size < static_cast<unsigned int>(__o_message_pos + size)) {
        new_size += new_size;
      }
      __o_message.resize(new_size);
      
      GLOBAL_DEBUG_STREAM2("PacketONetStream: resize output message buffer. Size is now " << __o_message.size() << endl);
    }
    
    /* Write data to internal buffer */
    memcpy(&__o_message[__o_message_pos], value, size);
    __o_message_pos += size;
  }
  
  void PacketONetStream::writeData() {
    GLOBAL_DEBUG_STREAM2("PacketONetStream: writeData" << endl);
    
    unsigned int size = __o_message_pos;
    int packet_no = 0;
    int write_pos = 0;
    
    __o_message_pos = 0;
    
    /* Calculate minimum number of packets needed */
    unsigned int max_bytes_per_packet = __max_packet_size.getRValue() - sizeof(PacketHeader);
    int packet_count = (size + max_bytes_per_packet-1) / max_bytes_per_packet;
    
    /* Split the message into packets and send them */
    while (size > 0) {
      /* Calculate number of bytes to put in this packet */
      unsigned int packet_bytes = max_bytes_per_packet;
      if (packet_bytes > size) {
        packet_bytes = size;
      }
      
      /* Write packet header */
      MutexGuard g(__packet_mutex);
      PacketHeader* header = reinterpret_cast<PacketHeader*>(&__o_packet[0]);
      *header = PacketHeader(__o_message_id, packet_no, packet_count, packet_bytes);
      
      GLOBAL_DEBUG_STREAM2("PacketONetStream: sendPacket"
                           << " message_id=" << __o_message_id
                           << " packet_no=" << packet_no
                           << " packet_count=" << packet_count
                           << " size=" << packet_bytes
                           << endl);
      
      /* Write packet data */
      memcpy(&__o_packet[sizeof(PacketHeader)], &__o_message[write_pos], packet_bytes);
      
      write_pos += packet_bytes;
      size -= packet_bytes;
      ++packet_no;
      
      /* Send packet */
      writePacket(&__o_packet[0]);
    }
    
    ++__o_message_id;
  }
  
  void PacketONetStream::writePacket(const void* packet) {
    const char* data = static_cast<const char*>(packet);
    unsigned int size = static_cast<const PacketHeader*>(packet)->getPacketSize() + sizeof(PacketHeader);
    size = getSock()->send(data, size);
    
    __total_sent_bytes += size;
    ++__total_sent_packets;
    
    GLOBAL_DEBUG_STREAM2("PacketONetStream: writePacket raw_size=" << size << endl);
  }
  
  void PacketONetStream::setPacketSize(const unsigned int& size) {
    MutexGuard g(__packet_mutex);
    if (size == __max_packet_size.getRValue()) {
      return;
    }
    
    __max_packet_size.setValue(size);
    __o_packet.resize(size);
  }
  
  unsigned int PacketONetStream::getPacketSize() const {
    MutexGuard g(__packet_mutex);
    return __max_packet_size.getRValue();
  }
  
  void PacketONetStream::resetStatistics() {
    __total_sent_bytes = 0;
    __total_sent_packets = 0;
  }
  
  unsigned int PacketONetStream::getTotalSentBytes() const {
    return __total_sent_bytes;
  }
  
  unsigned int PacketONetStream::getTotalSentPackets() const {
    return __total_sent_packets;
  }
  
  
  template<>
  class TConstructor<PacketONetStream> : public ValueConstructor {
  public:
    TConstructor(const TypeInfo& t_info)
      : ValueConstructor(t_info)
    { }
    virtual PacketONetStream*          createValue()          const {return new PacketONetStream();}
    virtual TValue<PacketONetStream>*  createTValue()         const {return new TValue<PacketONetStream>();}
  
    virtual TValue<ONetStream*>*  createTPointerValue()  const {return new TValue<ONetStream*>();}
  
    virtual TConstructor<PacketONetStream>* copy()  const {return new TConstructor(*this); }
  private:
    string  __type_name;
    string  __namespace;
  };
  
  TConstructor<PacketONetStream> __nmm__packet_o_net_stream(TypeInfo("PacketONetStream", "NMM"));
  
}

#include "nmm/nmm_suffix.hpp"
