/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#ifndef NMM_RTP_ADDRESS_HPP
#define NMM_RTP_ADDRESS_HPP NMM_RTP_ADDRESS_HPP

#include "nmm/nmm_prefix.hpp"

#include "nmm/comm/ConnectorAddress.hpp"

#include "visibility.hpp"

namespace NMM {

  class IPAddress;

  /** Transport strategy address for RTPStrategy.
   * This is the type of address which is used by the RTPStrategy to connect to a partner strategy.
   * It can be used for indirect connection setup.
   *
   * An RTP address consists of an IP address and two port numbers for RTP and RTCP.
   * The IP address is represented by an IPAddress object, which can be a UnicastIPAddress or a
   * MulticastIPAddress.
   */
  class DLLEXPORT RTPAddress : public ConnectorAddress {
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Default constructor.
     *
     * @param nmm_reference NMM reference of partner object
     */
    RTPAddress(const NMMReference& nmm_reference = 0) throw();

    /** Constructor.
     *
     * @param ip_address The IP address
     * @param rtp_port The port used for RTP communication
     * @param rtcp_port the port used for RTCP communication
     * @param nmm_reference NMM reference of partner object
     */
    RTPAddress(const IPAddress& ip_address,
               unsigned int rtp_port, unsigned int rtcp_port,
               const NMMReference& nmm_reference = 0) throw();

    /** Copy constructor.
     *
     * @param address Object to copy from
     */
    RTPAddress(const RTPAddress& address) throw();

    /** Virtual copy constructor.
     *
     * @return A copy of this object which has the same type
     */
    RTPAddress* copy() const throw() {
      return new RTPAddress(*this);
    }
    //@}

    /** @name Attributes
     */
    //@{
    /** Set the port used for RTP communication.
     *
     * @param port The port used for RTP communication
     */
    inline void setRTPPort(unsigned int port) throw() {
      __rtp_port = port;
    }

    /** Set the port used for RTCP communication.
     *
     * @param port The port used for RTCP communication
     */
    inline void setRTCPPort(unsigned int port) throw() {
      __rtcp_port = port;
    }

    /** Set the IP address.
     *
     * @param ip_address The IP address
     */
    void setIPAddress(const IPAddress& ip_address) throw();

    /** Get the port used for RTP communication.
     *
     * @return The port used for RTP communication
     */
    inline unsigned int getRTPPort() const throw() {
      return __rtp_port.getRValue(); 
    }

    /** Get the port used for RTCP communication.
     *
     * @return The port used for RTCP communication
     */
    inline unsigned int getRTCPPort() const throw() {
      return __rtcp_port.getRValue();
    }

    /** Get the IP address.
     *
     * @return The IP address
     */
    inline const IPAddress& getIPAddress() const throw() {
      return *__ip_address.getValue();
    }
    //@}

  private:
    TValue<nmm_uint32>  __rtp_port;         ///< The port used for RTP communication
    TValue<nmm_uint32>  __rtcp_port;        ///< The port used for RTCP communication
    TValue<IPAddress*>    __ip_address;       ///< The IP address

  public:
    /** Get static type information.
     * This method returns the type information for this class.
     */
    static TypeInfo getStaticTypeInfo() throw();
  };

}

#include "nmm/nmm_suffix.hpp"

#endif
