/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "RTPReceiverStrategy.hpp"
#include "BufferedSink.hpp"

#include "nmm/comm/TTSFactoryObject.hpp"
#include "nmm/comm/TTSFactoryRegistrator.hpp"
#include "nmm/comm/ConnectorAcceptor.hpp"

#include <live/RTCP.hh>

// for MPEG 1/2 video
#include <live/MPEG1or2VideoRTPSource.hh>

// for H263plus video
#include <live/H263plusVideoRTPSource.hh>

// for JPEG video
#include <live/JPEGVideoRTPSource.hh>

// for MPEG 1/2 audio
#include <live/MPEG1or2AudioRTPSource.hh>

#include <live/AC3AudioRTPSource.hh>
// for some raw formats
#include <live/SimpleRTPSource.hh>

// for AMR audio
#include <live/AMRAudioRTPSource.hh>
#include <live/AMRAudioSource.hh>

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  RTPReceiverStrategy::RTPReceiverStrategy(const string& name)
    : RTPStrategy(name),
      __rtp_source(0),
      __amr_audio_source(0),
      __buffered_sink(0)
  {}
  
  RTPReceiverStrategy::~RTPReceiverStrategy() {
    /* stop strategy */
    stop();

    if(__amr_audio_source) {
      GLOBAL_DEBUG_STREAM2("Closing __amr_audio_source" << endl);
      Medium::close(__amr_audio_source);
      __amr_audio_source = NULL;
    }

    /* close sink */
    if(__buffered_sink) {
      GLOBAL_DEBUG_STREAM2("Closing buffered sink" << endl);
      Medium::close(__buffered_sink);
      __buffered_sink = NULL;
    }
  }
  
  void RTPReceiverStrategy::configure(const Format& format)
    throw(const FormatNotSupportedException&)
  {
    UsageEnvironment& env(getUsageEnvironment());
    Groupsock* rtp_sock(&getRTPSocket());
    
    string type = format.getFormatType();
    
    /* Begin> Create RTP source depending on format*/
    if(type == Format::video_mpeg_type) {
      /* Create an RTP source for MPEG video */
      __rtp_source = MPEG1or2VideoRTPSource::createNew(env, rtp_sock);
      
    } else if (type == Format::video_h263plus_type) {
      /* Create an RTP source for h263plus video */
      __rtp_source = H263plusVideoRTPSource::createNew(env, rtp_sock, 34);
      
    } else if (type == Format::video_jpeg_type) {
      /* Create an RTP source for motion-JPEG video */
      __rtp_source = JPEGVideoRTPSource::createNew(env, rtp_sock);
      
    } else if(type == Format::audio_mpeg_type) {
      /* Create an RTP source for MPEG audio */
      __rtp_source = MPEG1or2AudioRTPSource::createNew(env, rtp_sock);

      // fw: audio/ac3 doesn't seem to work at the moment

    } else if(type == Format::audio_amr_type) {
      string encoding = format.getStringValue(Format::encoding_param);

      if (encoding == "amr_nb") {
         // Create an RTP source for AMR_NB audio
         __amr_audio_source = AMRAudioRTPSource::createNew(env, rtp_sock, __rtp_source, 97, false);

      } else if (encoding == "amr_wb") {
        // Create an RTP source for AMR_WB audio
         __amr_audio_source = AMRAudioRTPSource::createNew(env, rtp_sock, __rtp_source, 98, true);

      } else {
        GLOBAL_ERROR_STREAM1("RTPReceiverStrategy> Unsupported audio/amr format" << endl);
        throw BindingException("Unsupported payload format");
      }

    } else if (type == Format::audio_ac3_type) {
      /* Create an RTP source for AC3 audio */
     __rtp_source = AC3AudioRTPSource::createNew(env, rtp_sock, 100, 48000);
      
    } else if(type == Format::audio_raw_type) {
      int channels = format.getIntValue(Format::channels_param);
      int bitpersample = format.getIntValue(Format::bitpersample_param);
      int sample_rate = format.getIntValue(Format::sample_rate_param);
      string encoding = format.getStringValue(Format::encoding_param);

      if(encoding == "linear") {
        if (bitpersample == 24 && (sample_rate == 8000 || sample_rate == 44100 || sample_rate == 48000)) {
          __rtp_source = SimpleRTPSource::createNew(env, rtp_sock, 100, sample_rate, "audio/raw");
        }
        else if (bitpersample == 16 && sample_rate == 44100) {
          if (channels == 2) {
            __rtp_source = SimpleRTPSource::createNew(env, rtp_sock, PT_AUDIO_L16_STEREO, 44100, "audio/raw");
          }
          else if (channels == 1) {
            __rtp_source = SimpleRTPSource::createNew(env, rtp_sock, PT_AUDIO_L16_MONO, 44100, "audio/raw");
          }
        }
        else {
          GLOBAL_ERROR_STREAM1("RTPStrategy> Unsupported payload format" << endl);
          throw BindingException("Unsupported payload format");
        }
      }
      else if (encoding == "mu_law" && bitpersample == 8) {
        __rtp_source = SimpleRTPSource::createNew(env, rtp_sock, PT_AUDIO_PCMU, sample_rate, "audio/pcmu");
      }
      else if (encoding == "a_law" && bitpersample == 8) {
        __rtp_source = SimpleRTPSource::createNew(env, rtp_sock, PT_AUDIO_PCMA, sample_rate, "audio/pcma");
      }
      else {
        cout << " The unsupported payload type: " << type << endl;
        GLOBAL_ERROR_STREAM1("RTPStrategy> Unsupported encoding for audio/raw format" << endl);
      }
    }
    else {
      GLOBAL_ERROR_STREAM1("RTPStrategy> Unsupported payload format" << endl);
      throw BindingException("Unsupported payload format");
    }
    /* End< */
    
    /* create and configure the buffered sink */
    __buffered_sink = BufferedSink::createNew(getUsageEnvironment());
    __buffered_sink->setOneWayBufferReceiver(getOneWayBufferReceiver());
    __buffered_sink->setOneWayBufferSender(this);
    __buffered_sink->setBufferManager(getBufferManager());
    __buffered_sink->setBufferSize(getBufferSize());
    __buffered_sink->setMaxFrameSize(getMaxFrameSize());
    
    /* create RTCP instance */
    setRTCPInstance(RTCPInstance::createNew(getUsageEnvironment(), &getRTCPSocket(), getSessionBandwidth(),
                                            reinterpret_cast<const unsigned char*>(getCNAME().c_str()), NULL, __rtp_source));
    
    /* activate thread */
    start();
  }
  
  void RTPReceiverStrategy::receiveSignal(Signal signal) {
    switch(signal) {
    case FREE_QUEUE_SLOT:
    case FREE_UPSTR_QUEUE_SLOT:
      {
        /* Our IMessageReceiver sends a FREE_QUEUE_SLOT.
           Forward it to the BufferedSink so it can send new buffers to it. */
        __buffered_sink->receiveSignal(signal);
      }
      break;
      
    default:
      break;
    }
  }

  RTPSource* RTPReceiverStrategy::getRTPSource() const throw() {
    return __rtp_source;
  }
  
  MediaSource* RTPReceiverStrategy::getMediaSource() const throw() {
    if (__amr_audio_source) {
      return __amr_audio_source;
    }
    return __rtp_source;
  }
  
  MediaSink* RTPReceiverStrategy::getMediaSink() const throw() {
    return __buffered_sink;
  }
  
  void RTPReceiverStrategy::flush() throw() {
    if(__buffered_sink) {
      __buffered_sink->flush();
    }
  }
  
  RTPReceptionStats* RTPReceiverStrategy::getRTPReceptionStats() const {
    /* must have an RTP source */
    if(!__rtp_source) {
      GLOBAL_ERROR_STREAM("No RTP source, can't get reception stats" << endl);
      return 0;
    }
    
    /* get RTPReceptionStatsDB of RTP source */
    RTPReceptionStatsDB& db = __rtp_source->receptionStatsDB();
    
    /* get transmission stats for this synchronized source */
    RTPReceptionStatsDB::Iterator iter(db);
    RTPReceptionStats* stats = iter.next();
    if(!stats) {
      GLOBAL_ERROR_STREAM("Failed to lookup RTP reception stats" << endl);
      return 0;
    }
    return stats;
  }
  
  nmm_uint32 RTPReceiverStrategy::getNumberOfReceivedPackets() const {
    RTPReceptionStats* stats = getRTPReceptionStats();
    if(!stats) {
      return 0;
    }
    
    return stats->totNumPacketsReceived();
  }
  
  nmm_uint32 RTPReceiverStrategy::getNumberOfExpectedPackets() const {
    RTPReceptionStats* stats = getRTPReceptionStats();
    if(!stats) {
      return 0;
    }
    
    return stats->totNumPacketsExpected();
  }
  
  nmm_uint64 RTPReceiverStrategy::getNumberOfBytesReceived() const {
    RTPReceptionStats* stats = getRTPReceptionStats();
    if(!stats) {
      return 0;
    }
    
    return uint64_t(stats->totNumKBytesReceived()) * 1024;
  }
  
  Interval RTPReceiverStrategy::getJitter() const {
    RTPReceptionStats* stats = getRTPReceptionStats();
    if(!stats || !__rtp_source) {
      return 0;
    }

    // convert jitter from timestamp units to nanoseconds
    // FIXME: fw: This conversion only works if jitter is less than 1 second! Needs to be fixed!
    float jitter = static_cast<float>(1000 * 1000 * 1000 * stats->jitter()) / static_cast<float>(__rtp_source->timestampFrequency());
    return Interval(static_cast<unsigned int>(jitter));
  }
  
  
  TTSFactoryRegistrator< TTSFactoryObject<RTPReceiverStrategy> >
  __rtp_recv_strategy_prototype("RTPStrategy",
                                new ConnectorAcceptor("RTPStrategy"),
                                TSDescription::ITS_ONE_WAY_BUFFER_SENDER,
                                0,
                                RTPStrategy::FormatList());
  
}

#include "nmm/nmm_suffix.hpp"
