/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#ifndef NMM_RTP_RECEIVER_STRATEGY_HPP
#define NMM_RTP_RECEIVER_STRATEGY_HPP NMM_RTP_RECEIVER_STRATEGY_HPP

#include "nmm/nmm_prefix.hpp"

#include "RTPStrategy.hpp"
#include "IRTPReceptionStatsImpl.hpp"

#include "nmm/comm/messaging/ITSOneWayBufferSender.hpp"

class RTPSource;
class RTPReceptionStats;
class AMRAudioSource;

#include "visibility.hpp"

namespace NMM {
  
  class BufferedSink;
  
  /** RTP receiver transport strategy.
   *
   * This transport strategy receives instream messages by the
   * Realtime TransportProtocol (RTP).
   */
  class DLLEXPORT RTPReceiverStrategy
    : public RTPStrategy,
      public ITSOneWayBufferSender,
      public IRTPReceptionStatsImpl
  {
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Default constructor.
     */
    RTPReceiverStrategy(const std::string& name = "RTPReceiverStrategy");
    
    /** Virtual destructor.
     */
    virtual ~RTPReceiverStrategy();
    //@}
    
    /** @name %InstreamReceiverTransportStrategy methods
     */
    //@{
    /** Configure instream strategy.
     *
     * This method configures the transport strategy with the given
     * format.  The method fails, if the format is not supported by
     * RTP (i.e. no payload type is available or implemented).
     *
     * @param format Stream format
     */
    virtual void configure(const Format& format)
      throw(const FormatNotSupportedException&);
    //@}
    
    /** @name %ISignalReceiver methods
     */
    //@{
    /** Receive a signal.
     *
     * This method handles signals from the parent of this transport
     * strategy.
     *
     * If a @c FREE_QUEUE_SLOT or @c FREE_UPSTR_QUEUE_SLOT event is
     * received, then this transport strategy will continue to forward
     * messages to its parent.
     */
    virtual void receiveSignal(Signal);
    //@}
    
    /** @name %IRTPSoufceImpl interface
     */
    //@{
    /** Get the total number of RTP packets received.
     */
    nmm_uint32 getNumberOfReceivedPackets() const;
    
    /** Get the total number of RTP packets expected.
     */
    nmm_uint32 getNumberOfExpectedPackets() const;
    
    /** Get the number of received bytes.
     */
    nmm_uint64 getNumberOfBytesReceived() const;
    
    /** Get the inter-packet reception jitter.
     */
    Interval getJitter() const;
    //@}

    unsigned int getNumberOfLostPackets() const;

    /// explicitly return the RTPSource (needed for AMR)
    RTPSource* getRTPSource() const throw();

  protected:
    /** Get media source.
     *
     * @return The RTP source or AMR RTP source
     */
    virtual MediaSource* getMediaSource() const throw();
    
    /** Get media sink.
     *
     * @return The buffered sink
     */
    virtual MediaSink* getMediaSink() const throw();
    
    /** Flush queues and cancel blocking calls.
     *
     * This method flushes all queues and cancels all blocking calls
     * in the LIVE.COM thread.
     */
    virtual void flush() throw();
    
    /** Get reception statistics.
     */
    RTPReceptionStats* getRTPReceptionStats() const;
    
  private:
    /** RTP network source */
    RTPSource*      __rtp_source;

    /** RTP AMRAudioSource */
    AMRAudioSource* __amr_audio_source;

    /** Buffered sink. */
    BufferedSink*   __buffered_sink;
  };
  
}

#include "nmm/nmm_suffix.hpp"

#endif
