/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "RTPSenderStrategy.hpp"
#include "BufferedSource.hpp"

#include "nmm/comm/TTSFactoryObject.hpp"
#include "nmm/comm/TTSFactoryRegistrator.hpp"
#include "nmm/comm/ConnectorAcceptor.hpp"

#include <live/RTCP.hh>

// for AMR audio
#include <live/AMRAudioRTPSink.hh>

// for MPEG 1/2 video
#include <live/MPEG1or2VideoRTPSink.hh>
#include <live/MPEG1or2VideoStreamFramer.hh>

// for MPEG 1/2 audio
#include <live/MPEG1or2AudioRTPSink.hh>
#include <live/MPEG1or2AudioStreamFramer.hh>

// for H263plus video
#include <live/H263plusVideoRTPSink.hh>

// for JPEG video
#include <live/JPEGVideoRTPSink.hh>

// for MPEG4 video
#include <live/MPEG4ESVideoRTPSink.hh>
#include <live/MPEG4VideoStreamFramer.hh>

#include <live/AC3AudioRTPSink.hh>
#include <live/AC3AudioStreamFramer.hh>
// for some raw formats
#include <live/SimpleRTPSink.hh>

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  RTPSenderStrategy::RTPSenderStrategy(const string& name)
    : RTPStrategy(name),
      __buffered_source(0),
      __framer(0),
      __rtp_sink(0),
      __send_queue_size(1),
      __send_queue_mode(IQueued::MODE_SUSPEND)
  {}
  
  RTPSenderStrategy::~RTPSenderStrategy() {
    /* stop strategy */
    stop();
    
    /* close sink */
    if(__rtp_sink) {
      GLOBAL_DEBUG_STREAM2("Closing RTP sink" << endl);
      Medium::close(__rtp_sink);
      __rtp_sink = NULL;
    }
  }
  
  void RTPSenderStrategy::configure(const Format& format)
    throw(const FormatNotSupportedException&)
  {
    /* set up sockets */
    setupSenderSockets();
    
    UsageEnvironment& env(getUsageEnvironment());
    Groupsock* rtp_sock(&getRTPSocket());
    
    string type = format.getFormatType();
    
    /* signal receiver must have been set before calling this method */
    if(!getSignalReceiver()) {
      GLOBAL_ERROR_STREAM1("RTPStrategy> configureForOutputJack called but signal receiver not set yet (SHOULD NOT HAPPEN!)" << std::endl);
      throw BindingException("Signal receiver not set");
    }
    
    /* Begin> Create RTP sink and framer depending on format*/
    if (type == Format::video_mpeg_type) {
      /* Create an RTP sink for MPEG video */
      __rtp_sink = MPEG1or2VideoRTPSink::createNew(env, rtp_sock);
      
      /* Create a buffered source */
      __buffered_source = BufferedSource::createNew(env);
      __buffered_source->setBufferMergingEnabled(true);
      
      /* Create a framer for MPEG video */
      __framer = MPEG1or2VideoStreamFramer::createNew(env, __buffered_source);
      
    } else if (type == Format::video_h263plus_type) {
      /* Create an RTP sink for h263plus video */
      __rtp_sink = H263plusVideoRTPSink::createNew(env, rtp_sock, 34);
      
      /* Create a buffered source */
      __buffered_source = BufferedSource::createNew(env);
      
      /* The source is the framer */
      __framer = __buffered_source;
      
    } else if (type == Format::video_jpeg_type) {
      /* Create an RTP sink for motion-JPEG video */
      __rtp_sink = JPEGVideoRTPSink::createNew(env, rtp_sock);
      
      /* Create a buffered source */
      __buffered_source = BufferedSource::createNew(env);
      
      /* The source is the framer */
      __framer = __buffered_source;
    
    } else if (type == Format::video_mpeg4_type) {
      /* Create an RTP sink for MPEG4 video */
      __rtp_sink = MPEG4ESVideoRTPSink::createNew(env, rtp_sock, 96);
      
      /* Create a buffered source */
      __buffered_source = BufferedSource::createNew(env);
      
      /* Create a framer for MPEG video */
      __framer = MPEG4VideoStreamFramer::createNew(env, __buffered_source);
      
    } else if (type == Format::audio_mpeg_type) {
      /* Create an RTP sink for MPEG audio */
      __rtp_sink = MPEG1or2AudioRTPSink::createNew(env, rtp_sock);
      
      /* Create a buffered source */
      __buffered_source = BufferedSource::createNew(env);
      
      /* Create a framer for MPEG audio */
      __framer = MPEG1or2AudioStreamFramer::createNew(env, __buffered_source);

     } else if (type == Format::audio_amr_type) {
       string encoding = format.getStringValue(Format::encoding_param);

       if (encoding == "amr_nb") {
         /* Create an RTP sink for AMR_NB audio */
         __rtp_sink = AMRAudioRTPSink::createNew(env, rtp_sock, 97, false);

         /* Create a buffered source */                     //wideband = false
         __buffered_source = BufferedSource::createNew(env, false);
         __buffered_source->setUsedForAMR(true);

         /* The source is the framer */
         __framer = __buffered_source;

       } else if (encoding == "amr_wb") {
         /* Create an RTP sink for AMR_WB audio */
         __rtp_sink = AMRAudioRTPSink::createNew(env, rtp_sock, 98, true);

         /* Create a buffered source */                     // wideband = true
         __buffered_source = BufferedSource::createNew(env, true);
         __buffered_source->setUsedForAMR(true);

         /* The source is the framer */
         __framer = __buffered_source;

       } else {
         GLOBAL_ERROR_STREAM1("RTPSenderStrategy> Unsupported audio/amr format" << endl);
         throw BindingException("Unsupported payload format");
       }

     } else if (type == Format::audio_ac3_type) {
       // XXX: fw: audio/ac3 doesn't seem to work at the moment

       cerr << "Using AC3 Sink " << endl;
       /* Create an RTP sink for AC3 audio */
       __rtp_sink = AC3AudioRTPSink::createNew(env, rtp_sock, 100, 48000);
       if (__rtp_sink) dynamic_cast<MultiFramedRTPSink*>(__rtp_sink)->setPacketSizes(2 * 1024, 16 * 1024);

       /* Create a buffered source */
       __buffered_source = BufferedSource::createNew(env);

       /* Create a framer for AC3 audio */
       __framer = AC3AudioStreamFramer::createNew(env, __buffered_source,0x80);
       //__rtp_sink->setRTPTimestampFrequency(dynamic_cast<AC3AudioStreamFramer*>(__framer)->samplingRate());
       //cerr << "Frequenz: " << dynamic_cast<AC3AudioStreamFramer*>(__framer)->samplingRate() << endl;
     } else if (type == Format::audio_raw_type) {
      int channels = format.getIntValue(Format::channels_param);
      int sample_rate = format.getIntValue(Format::sample_rate_param);
      int bitpersample = format.getIntValue(Format::bitpersample_param);
      string encoding = format.getStringValue(Format::encoding_param);
      if(encoding == "linear") {
        if (bitpersample == 24 && (sample_rate == 8000 || sample_rate == 44100 || sample_rate == 48000 )) {
          /* Create a simple RTP sink */
          __rtp_sink = SimpleRTPSink::createNew(env, rtp_sock, 100, sample_rate, "audio/raw", "L24");

          // comment in to adapt PacketSizes in liblive
          dynamic_cast<SimpleRTPSink*>(__rtp_sink)->setPacketSizes(16392, 16392);

          if (channels == 2) {
            /* Create a buffered source */
            __buffered_source = BufferedSource::createNew(env);
          }
          else if (channels == 1) {
            /* Create a buffered source */
            __buffered_source = BufferedSource::createNew(env);
          }
          else {
            GLOBAL_ERROR_STREAM1("RTPStrategy> Unsupported encoding for audio/raw format" << endl);
            throw BindingException("Unsupported payload type");
          }

          /* The source is the framer */
          __framer = __buffered_source;
        }
        else if (bitpersample == 16 && (sample_rate == 8000 || sample_rate == 44100 || sample_rate == 48000 )) {
          if (channels == 2) {
            /* Create a simple RTP sink */
            __rtp_sink = SimpleRTPSink::createNew(env, rtp_sock, PT_AUDIO_L16_STEREO, sample_rate, "audio/raw", "L16");

            // comment in to adapt PacketSizes in liblive
            dynamic_cast<SimpleRTPSink*>(__rtp_sink)->setPacketSizes(16384, 16384); // 8192, 8192);

            /* Create a buffered source */
            __buffered_source = BufferedSource::createNew(env);

            /* The source is the framer */
            __framer = __buffered_source;
          }
          else if (channels == 1) {
            /* Create a simple RTP sink */
            __rtp_sink = SimpleRTPSink::createNew(env, rtp_sock, PT_AUDIO_L16_MONO, sample_rate, "audio/raw", "L16");

            /* Create a buffered source */
            __buffered_source = BufferedSource::createNew(env);

            /* The source is the framer */
            __framer = __buffered_source;
          }
          else {
            GLOBAL_ERROR_STREAM1("RTPStrategy> Unsupported encoding for audio/raw format" << endl);
            throw BindingException("Unsupported payload type");
          }
        }
      }
      else if(encoding == "mu_law" && bitpersample == 8) {
        /* Create a simple RTP sink */
        __rtp_sink = SimpleRTPSink::createNew(env, rtp_sock, PT_AUDIO_PCMU, sample_rate, "audio/pcmu", "PCMU");
        /* set packet size in Live555 */
        dynamic_cast<SimpleRTPSink*>(__rtp_sink)->setPacketSizes(128, 724);

        /* Create a buffered source */
        __buffered_source = BufferedSource::createNew(env);

        /* The source is the framer */
        __framer = __buffered_source;
      }
      else if (encoding == "a_law" && bitpersample == 8) {
        /* Create a simple RTP sink */
        __rtp_sink = SimpleRTPSink::createNew(env, rtp_sock, PT_AUDIO_PCMA, sample_rate, "audio/pcma", "PCMA");
        /* set packet size in Live555 */
        dynamic_cast<SimpleRTPSink*>(__rtp_sink)->setPacketSizes(128, 724);

        /* Create a buffered source */
        __buffered_source = BufferedSource::createNew(env);

        /* The source is the framer */
        __framer = __buffered_source;
      }
      else {
        GLOBAL_ERROR_STREAM1("RTPStrategy> Unsupported encoding for audio/raw format" << endl);
        throw BindingException("Unsupported payload type");
      }

      /* set sample rate and sample size for audio/raw*/
      __buffered_source->setSamplesPerSecond(sample_rate);
      __buffered_source->setSampleSize(channels * bitpersample/8);
    }
    else {
      GLOBAL_ERROR_STREAM1("RTPStrategy> Unsupported payload format" << endl);
      throw BindingException("Unsupported payload format");
    }
    /* End< */

    /* set common attributes of BufferedSource */
    __buffered_source->setPartnerSignalReceiver(getSignalReceiver());
    __buffered_source->getStreamQueue().setMode((StreamQueue::Mode)(__send_queue_mode));
    __buffered_source->getStreamQueue().setMaxSize(__send_queue_size);

    /* Create RTCP instance */
    setRTCPInstance(RTCPInstance::createNew(getUsageEnvironment(), &getRTCPSocket(), getSessionBandwidth(),
                                            reinterpret_cast<const unsigned char*>(getCNAME().c_str()), __rtp_sink, NULL));
    /* Activate thread */
    start();
  }
  
  Result RTPSenderStrategy::receiveOneWayBuffer(Buffer* buffer) {
    return internalSendBuffer(buffer);
  }
  
  nmm_uint32 RTPSenderStrategy::getSize(const string& /*= "default"*/) const {
    if(__buffered_source) {
      /* return the current size of the BufferedSource's queue */
      return __buffered_source->getStreamQueue().getSize();
    } else {
      /* there is no queue */
      return 0;
    }
  }
  
  nmm_uint32 RTPSenderStrategy::getMaxSize(const string& /*= "default"*/) const {
    if(__buffered_source) {
      /* return the maximum size of the BufferedSource's queue */
      return __buffered_source->getStreamQueue().getMaxSize();
    } else {
      /* return the configured queue size */
      return __send_queue_size;
    }
  }
  
  Result RTPSenderStrategy::setMaxSize(const nmm_uint32& size, const string& /*= "default"*/) {
    if(__buffered_source) {
      /* set the maximum size of the BufferedSource's queue */
      __buffered_source->getStreamQueue().setMaxSize(size);
    } else {
      /* configure the queue size */
      __send_queue_size = size;
    }
    
    return SUCCESS;
  }
  
  IQueued::Mode RTPSenderStrategy::getMode(const string& /*= "default"*/) const {
    if(__buffered_source) {
      /* set the mode of the BufferedSource's queue */
      return (IQueued::Mode) __buffered_source->getStreamQueue().getMode();
    } else {
      /* return the configured queue mode */
      return __send_queue_mode;
    }
  }
  
  Result RTPSenderStrategy::setMode(const IQueued::Mode& mode, const string& /*= "default"*/) {
    if(__buffered_source) {
      /* set the mode of the BufferedSource's queue */
      return __buffered_source->getStreamQueue().setMode((StreamQueue::Mode) mode);
    } else {
      /* configure the queue mode */
      __send_queue_mode = mode;
      return SUCCESS;
    }
  }
  
  RTPTransmissionStats* RTPSenderStrategy::getRTPTransmissionStats() const {
    /* must have an RTP sink */
    if(!__rtp_sink) {
      GLOBAL_ERROR_STREAM("No RTP sink, can't get transmission stats" << endl);
      return 0;
    }
    
    /* get RTPTransmissionStatsDB of RTP sink */
    RTPTransmissionStatsDB& db = __rtp_sink->transmissionStatsDB();
    
    /* get transmission stats for this synchronized source */
    RTPTransmissionStatsDB::Iterator iter(db);
    RTPTransmissionStats* stats = iter.next();
    if(!stats) {
      GLOBAL_ERROR_STREAM("Failed to lookup RTP transmission stats" << endl);
      return 0;
    }
    return stats;
  }
  
  nmm_uint32 RTPSenderStrategy::getLastReceivedPacketNumber() const {
    RTPTransmissionStats* stats = getRTPTransmissionStats();
    if(!stats) {
      return 0;
    }
    
    return stats->lastPacketNumReceived();
  }
  
  nmm_uint32 RTPSenderStrategy::getFirstReportetPacketNumber() const {
    RTPTransmissionStats* stats = getRTPTransmissionStats();
    if(!stats) {
      return 0;
    }
    
    return stats->firstPacketNumReported();
  }
  
  nmm_uint64 RTPSenderStrategy::getTotalNumberOfPackets() const {
    RTPTransmissionStats* stats = getRTPTransmissionStats();
    if(!stats) {
      return 0;
    }
    
    u_int32_t hi=0, lo=0;
    stats->getTotalPacketCount(hi, lo);
    return (uint64_t(hi) << 32) + lo;
  }
  
  nmm_uint64 RTPSenderStrategy::getTotalNumberOfBytes() const {
    RTPTransmissionStats* stats = getRTPTransmissionStats();
    if(!stats) {
      return 0;
    }
    
    u_int32_t hi=0, lo=0;
    stats->getTotalOctetCount(hi, lo);
    return (uint64_t(hi) << 32) + lo;
  }
  
  nmm_uint32 RTPSenderStrategy::getNumberOfLostPackets() const {
    RTPTransmissionStats* stats = getRTPTransmissionStats();
    if(!stats) {
      return 0;
    }
    
    return stats->totNumPacketsLost();
  }
  
  Interval RTPSenderStrategy::getJitter() const {
    RTPTransmissionStats* stats = getRTPTransmissionStats();
    if(!stats || !__rtp_sink) {
      return 0;
    }

    // convert jitter from timestamp units to nanoseconds
    // FIXME: fw: This conversion only works if jitter is less than 1 second! Needs to be fixed!
    float jitter = static_cast<float>(stats->jitter()) / static_cast<float>(__rtp_sink->rtpTimestampFrequency());
    jitter = jitter * 1000 * 1000 * 1000;

    return Interval(static_cast<unsigned int>(jitter));
  }

  nmm_float RTPSenderStrategy::getPacketLossRatio() const {
    RTPTransmissionStats* stats = getRTPTransmissionStats();

    if(!stats) {
      // convert 8 bit fixed point number (see RFC 3550) to float
      return static_cast<nmm_float>( static_cast<int>( stats->packetLossRatio() )) / 256 ;
    }
    else {
      return 0;
    }
  }

  nmm_int32 RTPSenderStrategy::getPacketsLostBetweenRR() const {
    RTPTransmissionStats* stats = getRTPTransmissionStats();

    if(!stats) {
      return 0;
    }
    return stats->packetsLostBetweenRR();
  }

  Interval RTPSenderStrategy::getRoundTripDelay() const {
    RTPTransmissionStats* stats = getRTPTransmissionStats();
    if(!stats) {
      return 0;
    }

    // stats->roundTripDelay() returns delay (in units of 1/65536 seconds) computed from the most recently-received RTCP RR packet.
    double delay = static_cast<double> (stats->roundTripDelay());

    // extract seconds
    long int delay_s = static_cast<long int>(delay/65536);

    // and nanoseconds
    long int delay_ns = static_cast<long int>((delay/65536.0 - delay_s)*1000*1000*1000);

    return Interval(delay_s, delay_ns);
  }


  string RTPSenderStrategy::getMediaType() const{
    if (__rtp_sink && __rtp_sink->sdpMediaType()) {
      return (string(__rtp_sink->sdpMediaType()));
    }
    return string("");
  }
  
  unsigned char RTPSenderStrategy::getPayloadType() const{
    if (__rtp_sink) {
      return (__rtp_sink->rtpPayloadType());
    }
    return 0;
  }
  
  MediaSource* RTPSenderStrategy::getMediaSource() const throw() {
    if(__framer) {
      return __framer;
    } else {
      return __buffered_source;
    }
  }
  
  MediaSink* RTPSenderStrategy::getMediaSink() const throw() {
    return __rtp_sink;
  }
  
  void RTPSenderStrategy::flush() throw() {
    if(__buffered_source) {
      __buffered_source->flush();
    }
  }
  
  Result RTPSenderStrategy::internalSendBuffer(const Buffer* buffer) {
    /* If we don't have a buffered source, we can't send buffers */
    if (!__buffered_source) {
      if ( getRTCPInstance() ) {
        //cerr << "Send RTCP report " << endl;
        /* !!!!!!!!! HACK !!!!!!! 
	   Sending a sender report after every buffer allows to synchronize remote hosts immediately 
	*/
	getRTCPInstance()->sendReport();
      }
      GLOBAL_ERROR_STREAM1("RTPStrategy> internalSendBuffer failed because there is no BufferedSource" << endl);
      return FAILURE;
    }
    
    /* Write buffer to source. */
    return __buffered_source->writeBuffer(buffer);
  }

  string RTPSenderStrategy::getMapLine() const{
    if (__rtp_sink && __rtp_sink->rtpmapLine()) {
      return string(__rtp_sink->rtpmapLine());
    }
    return string("");
  }
  
  string RTPSenderStrategy::getAuxSDPLine() const{
    if (__rtp_sink && __rtp_sink->auxSDPLine()) {
      return string(__rtp_sink->auxSDPLine());
    }
    return string("");
  }
  
  unsigned short RTPSenderStrategy::getRTPSeqNum() const {
    if (__rtp_sink && __rtp_sink->currentSeqNo()) {
      return __rtp_sink->currentSeqNo();
    }
    return 0;
  }

  unsigned RTPSenderStrategy::getRTPTimestamp() const {
    if (__rtp_sink && __rtp_sink->currentTimestamp()) {
      return __rtp_sink->currentTimestamp();
    }
    return 0;
  }

  unsigned RTPSenderStrategy::getRTPPort() const {
    return getLocalAddress().getRTPPort();
  }

  unsigned RTPSenderStrategy::getRTCPPort() const {
    return getLocalAddress().getRTCPPort();
  }

  Format* RTPSenderStrategy::getFormatFromPayloadType(int payload_type) {
    Format* format = 0;
    
    switch (payload_type) {
      
    case 0:
      format = new Format(Format::audio_raw_type);
      format->addIntParamValue(Format::channels_param, 1);
      format->addIntParamValue(Format::bitpersample_param, 8);
      format->addIntParamValue(Format::sample_rate_param, 8000);
      format->addStringParamValue(Format::encoding_param, "mu_law");
      format->addStringParamValue(Format::sign_param, "signed");
      break;
      
      // XXX: gerstner: Not supported yet
      // XXX: gerstner: Perhaps you have to add some more params to the format
//     case 3:
//       format = new Format(Format::audio_gsm_type);
//       format->addIntParamValue(Format::channels_param, 1);
//       format->addIntParamValue(Format::sample_rate_param, 8000);
//       break;
      
    case 8:
      format = new Format(Format::audio_raw_type);
      format->addIntParamValue(Format::channels_param, 1);
      format->addIntParamValue(Format::bitpersample_param, 8);
      format->addIntParamValue(Format::sample_rate_param, 8000);
      format->addStringParamValue(Format::encoding_param, "a_law");
      format->addStringParamValue(Format::sign_param, "signed");
      break;    
    
    case 10:
      format = new Format(Format::audio_raw_type);
      format->addIntParamValue(Format::channels_param, 2);
      format->addIntParamValue(Format::bitpersample_param, 16);
      format->addIntParamValue(Format::sample_rate_param, 44100);
      format->addStringParamValue(Format::encoding_param, "linear");
      format->addStringParamValue(Format::sign_param, "signed");
      break;
      
    case 11:
      format = new Format(Format::audio_raw_type);
      format->addIntParamValue(Format::channels_param, 1);
      format->addIntParamValue(Format::bitpersample_param, 16);
      format->addIntParamValue(Format::sample_rate_param, 44100);
      format->addStringParamValue(Format::encoding_param, "linear");
      break;
      
    case 14:
      format = new Format(Format::audio_mpeg_type);
      format->addIntParamValue(Format::layer_param, 2);
      format->addIntParamValue(Format::layer_param, 3);
      format->addIntParamValue(Format::audio_codec_param, 0x55);
      format->addWildcard(Format::sample_rate_param);
      format->addWildcard(Format::channels_param);
      format->addWildcard(Format::bitpersample_param);
      format->addWildcard(Format::bitrate_param);
      break;

    case 26:
      format = new Format(Format::video_jpeg_type);
      format->addWildcard(Format::x_resolution_param);
      format->addWildcard(Format::y_resolution_param);
      format->addWildcard(Format::ratio_x_param);
      format->addWildcard(Format::ratio_y_param);
      format->addWildcard(Format::framerate_param);
      format->addWildcard(Format::bitrate_param);
      break;
      
    case 32:
      format = new Format(Format::video_mpeg_type);
      format->addIntParamValue(Format::version_param, 1);
      format->addIntParamValue(Format::version_param, 2);
      // XXX: fw: epeters: comment in for testing
//       format->addWildcard(Format::x_resolution_param);
//       format->addWildcard(Format::y_resolution_param);
//       format->addWildcard(Format::ratio_x_param);
//       format->addWildcard(Format::ratio_y_param);
//       format->addWildcard(Format::framerate_param);
      break;
      
    case 34:
      format = new Format(Format::video_h263plus_type);
      format->addWildcard(Format::x_resolution_param);
      format->addWildcard(Format::y_resolution_param);
      format->addWildcard(Format::ratio_x_param);
      format->addWildcard(Format::ratio_y_param);
      format->addWildcard(Format::framerate_param);
      format->addWildcard(Format::bitrate_param);
      break;
      
    case 96:
      format = new Format(Format::video_mpeg4_type);
      format->addWildcard(Format::x_resolution_param);
      format->addWildcard(Format::y_resolution_param);
      format->addWildcard(Format::ratio_x_param);
      format->addWildcard(Format::ratio_y_param);
      format->addWildcard(Format::framerate_param);
      format->addWildcard(Format::bitrate_param);
      format->addStringParamValue(Format::compressor_param, "DIVX");
      format->addStringParamValue(Format::compressor_param, "DX50");

    case 97:
      format = new Format(Format::audio_amr_type);
      format->addIntParamValue(Format::channels_param, 1);
      format->addIntParamValue(Format::sample_rate_param, 8000);
      format->addIntParamValue(Format::bitpersample_param, 16);
      format->addIntParamValue(Format::bitrate_param, 4750);
      format->addIntParamValue(Format::bitrate_param, 5150);
      format->addIntParamValue(Format::bitrate_param, 5900);
      format->addIntParamValue(Format::bitrate_param, 6700);
      format->addIntParamValue(Format::bitrate_param, 7400);
      format->addIntParamValue(Format::bitrate_param, 7950);
      format->addIntParamValue(Format::bitrate_param, 10200);
      format->addIntParamValue(Format::bitrate_param, 12200);
      format->addStringParamValue(Format::encoding_param, "amr_nb");

    case 98:
      format = new Format(Format::audio_amr_type);
      format->addIntParamValue(Format::channels_param, 1);
      format->addIntParamValue(Format::sample_rate_param, 16000);
      format->addIntParamValue(Format::bitpersample_param, 16);
      //format->addWildcard(Format::bitrate_param);
      format->addIntParamValue(Format::bitrate_param, 6600);
      format->addIntParamValue(Format::bitrate_param, 8850);
      format->addIntParamValue(Format::bitrate_param, 12650);
      format->addIntParamValue(Format::bitrate_param, 14250);
      format->addIntParamValue(Format::bitrate_param, 15850);
      format->addIntParamValue(Format::bitrate_param, 18250);
      format->addIntParamValue(Format::bitrate_param, 19850);
      format->addIntParamValue(Format::bitrate_param, 23050);
      format->addIntParamValue(Format::bitrate_param, 23850);
      format->addStringParamValue(Format::encoding_param, "amr_wb");

      // XXX: fw: epeters: comment in for testing
//       format = new Format(Format::video_mpeg4_type);
//       format->addWildcard(Format::x_resolution_param);
//       format->addWildcard(Format::y_resolution_param);
//       format->addWildcard(Format::ratio_x_param);
//       format->addWildcard(Format::ratio_y_param);
//       format->addWildcard(Format::framerate_param);
//       format->addWildcard(Format::bitrate_param);
//       format->addStringParamValue(Format::compressor_param, "DIVX");
      break;
      
    default:
      format = new Format(Format::wild_wild_type);
      
    }
    
    return format;
  }
  
  TTSFactoryRegistrator< TTSFactoryObject<RTPSenderStrategy> >
  __rtp_send_strategy_prototype("RTPStrategy",
                                new ConnectorAcceptor("RTPStrategy"),
                                TSDescription::ITS_ONE_WAY_BUFFER_RECEIVER,
                                0,
                                RTPStrategy::FormatList());
  
}

#include "nmm/nmm_suffix.hpp"
