/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#ifndef NMM_RTP_SENDER_STRATEGY_HPP
#define NMM_RTP_SENDER_STRATEGY_HPP NMM_RTP_SENDER_STRATEGY_HPP

#include "nmm/nmm_prefix.hpp"

#include "RTPStrategy.hpp"

#include "nmm/comm/messaging/ITSOneWayBufferReceiver.hpp"

#include "nmm/comm/IQueuedImpl.hpp"
#include "IRTPTransmissionStatsImpl.hpp"

class RTPSink;
class FramedSource;
class RTPTransmissionStats;

#include "visibility.hpp"

namespace NMM {
  
  class BufferedSource;
  
  /** RTP sender transport strategy.
   *
   * This transport strategy sends instream messages by the Realtime
   * TransportProtocol (RTP).
   */
  class DLLEXPORT RTPSenderStrategy
    : public RTPStrategy,
      public ITSOneWayBufferReceiver,
      public IQueuedImpl,
      public IRTPTransmissionStatsImpl
  {
  public:
    /** Constructor & destructor.
     */
    //@{
    /** Default constructor.
     */
    RTPSenderStrategy(const std::string& name = "RTPSenderStrategy");
    
    /** Virtual destructor.
     */
    virtual ~RTPSenderStrategy();    
    //@}
    
    /** @name %InstreamSenderTransportStrategy methods
     */
    //@{
    /** Configure instream strategy.
     *
     * This method configures the transport strategy with the given
     * format.  The method fails, if the format is not supported by
     * RTP (i.e. no payload type is available or implemented).
     *
     * @param format Stream format
     */
    virtual void configure(const Format& format)
      throw(const FormatNotSupportedException&);
    //@}
    
    /** @name Receiver Interface
     */
    //@{
    /** Receive a one-way %Buffer.
     *
     * This method sends a one-way %Buffer over RTP
     *
     * @return Stream queue status
     */
    virtual Result receiveOneWayBuffer(Buffer*);
    //@}
    
    /** @name %IQueued methods
     */
    //@{
    /** Get current send queue size.
     *
     * @return The current size of the send queue
     */
    nmm_uint32 getSize(const std::string& = "default") const;
    
    /** Get maximum send queue size.
     *
     * @return The maximum size of the send queue
     */
    nmm_uint32 getMaxSize(const std::string& = "default") const;
    
    /** Set maximum send queue size.
     *
     * @param size The maximum send queue size
     */
    Result setMaxSize(const nmm_uint32& size, const std::string& = "default");
    
    /** Get send queue mode.
     *
     * @return The send queue mode
     */
    IQueued::Mode getMode(const std::string& = "default") const;
    
    /** Set send queue mode.
     *
     * @param mode The send queue mode
     */
    Result setMode(const IQueued::Mode& mode, const std::string& = "default");
    //@}
    
    /** @name %IRTPTransmissionStats methods
     */
    //@{
    /** Get the sequence number of the last received packet.
     *
     * @todo Find out what this method REALLY returns 
     */
    virtual nmm_uint32 getLastReceivedPacketNumber() const;
    
    /** Get the sequence number of the first received packet.
     *
     * @todo Find out what this method REALLY returns
     */
    virtual nmm_uint32 getFirstReportetPacketNumber() const;
    
    /** Get the total number of packets transmitted.
     */
    virtual nmm_uint64 getTotalNumberOfPackets() const;
    
    /** Get the total number of bytes transmitted.
     */
    virtual nmm_uint64 getTotalNumberOfBytes() const;
    
    /** Get the number of lost packets.
     *
     * @todo Find out what this method REALLY returns
     */
    virtual nmm_uint32 getNumberOfLostPackets() const;
    
    /** Get the inter-arrival jitter.
     *
     * This method gets the packet inter-arrival jitter based on the
     * receiver reports. It is defined by RFC3550 as follows:
     *
     * "An estimate of the statistical variance of the RTP data packet
     * interarrival time, measured in timestamp units and expressed as
     * an unsigned integer. The interarrival jitter @a J is defined to
     * be the mean deviation (smoothed absolute value) of the
     * difference @a D in packet spacing at the receiver compared to
     * the sender for a pair of packets."
     *
     * @return The inter-arrival jitter in timestamp units
     *
     * @todo Return in seconds as an %Interval
     */
    virtual Interval getJitter() const;

    nmm_float RTPSenderStrategy::getPacketLossRatio() const;
    nmm_int32 RTPSenderStrategy::getPacketsLostBetweenRR() const;
    Interval RTPSenderStrategy::getRoundTripDelay() const;
    //@}

    void flushITSReceiver() {}
    
    /** @name Special Functions for RTP.
     */
    //@{
    std::string getMediaType() const;
    
    unsigned char getPayloadType() const;
    
    std::string getMapLine() const;
    
    std::string getAuxSDPLine() const;
    
    unsigned short getRTPSeqNum() const;
    
    unsigned getRTPTimestamp() const;
    
    unsigned getRTPPort() const;
    
    unsigned getRTCPPort() const;
    
    static Format* getFormatFromPayloadType(int payload_type);
    //@}
    
  protected:
    /** Get media source.
     *
     * @return The buffered source
     */
    virtual MediaSource* getMediaSource() const throw();
    
    /** Get media sink.
     *
     * @return The RTP sink
     */
    virtual MediaSink* getMediaSink() const throw();
    
    /** Flush queues and cancel blocking calls.
     *
     * This method flushes all queues and cancels all blocking calls
     * in the LIVE.COM thread.
     */
    virtual void flush() throw();
    
  private:
    /** @name Static type information.
     */
    //@{
    /** Get transport strategy type name.
     *
     * @return "RTPStrategy"
     */
    static std::string getStaticStrategyType();
    //@}
    
  protected:
    /** Send a buffer via RTP.
     *
     * This method passes the data contained in @c buffer to the RTP
     * stack.
     *
     * @param buffer Buffer to send
     */
    Result internalSendBuffer(const Buffer* buffer);
    
    /** Get the transmission stats for this RTP sink.
     */
    RTPTransmissionStats* getRTPTransmissionStats() const;
    
  private:
    /** Buffered source */
    BufferedSource* __buffered_source;
    
    /** Stream framer. */
    FramedSource* __framer;
    
    /** RTP network sink. */
    RTPSink* __rtp_sink;
    
    /** Size of send queue. */
    unsigned __send_queue_size;
    
    /** Queue mode of send queue. */
    IQueued::Mode __send_queue_mode;
  };

}

#include "nmm/nmm_suffix.hpp"

#endif
