/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "RTPStrategy.hpp"
#include "UnicastIPAddress.hpp"
#include "MulticastIPAddress.hpp"
#include "LiveThread.hpp"

#include "nmm/comm/ConnectionManager.hpp"

#include <live/RTCP.hh>
#include <live/GroupsockHelper.hh> // for increaseReceiveBufferTo

#include <live/BasicUsageEnvironment.hh>

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  RTPStrategy::RTPStrategy(const std::string& name)
    : TransportStrategy(name),
      __cname("NMM"),
      __ttl(30),
      __buffer_size(65536),
      __max_frame_size(8192),
      __bandwidth(160),
      __scheduler(NULL),
      __env(NULL),
      __rtp_sock(NULL),
      __rtcp_sock(NULL),
      __rtcp_instance(0),
      __live_thread(0),
      __local_address(0),
      __remote_address(0),
      __icm(0),
      __remote_ts_ref(0), 
      __out_of_band_event(Sock::RMI_DATA), 
      __instream_event(Sock::INBAND_DATA),
      __instream_buffer(Sock::STREAMING_DATA)
  {
    /* create the task scheduler */
    __scheduler = BasicTaskScheduler::createNew();
    
    /* create the usage environment */
    __env = BasicUsageEnvironment::createNew(*__scheduler);
  }
  
  RTPStrategy::~RTPStrategy() {
    /* stop strategy */
    stop();
    
    /* close RTCP instance */
    if(__rtcp_instance) {
      Medium::close(__rtcp_instance);
      __rtcp_instance = 0;
    }
    
    // XXX: fw: new version of libLive claims to do that automatically
//     /* destroy the usage environment */
//     if (__env) {
//       delete __env;
//       __env = NULL;
//     }
    
    /* destroy the task scheduler */
    if (__scheduler) {
      delete __scheduler;
      __scheduler = NULL;
    }
  }
  
  void RTPStrategy::setRecvBufferSize(const nmm_int32& value) {
    /* set receive buffer size to highest possible value below the
       given value. */
    ::increaseReceiveBufferTo(getUsageEnvironment(),
                              getRTPSocket().socketNum(),
                              value);
  }
  
  nmm_int32 RTPStrategy::getRecvBufferSize() const {
    /* get receive buffer size */
    return ::getReceiveBufferSize(getUsageEnvironment(),
                                  getRTPSocket().socketNum());
  }
  
  void RTPStrategy::setSendBufferSize(const nmm_int32& value) {
    /* set send buffer size to highest possible value below the
       given value. */
    ::increaseSendBufferTo(getUsageEnvironment(),
                           getRTPSocket().socketNum(),
                           value);
  }
  
  nmm_int32 RTPStrategy::getSendBufferSize() const {
    /* get send buffer size */
    return ::getSendBufferSize(getUsageEnvironment(),
                               getRTPSocket().socketNum());
  }
  
  void RTPStrategy::setPriority(const nmm_int32& /*value*/) {
    // XXX: fw: not implemented
  }
  
  nmm_int32 RTPStrategy::getPriority() const {
    // XXX: fw: not implemented
    return 42;
  }
  
  void RTPStrategy::setCNAME(const string& cname) {
    __cname = cname;
  }
  
  string RTPStrategy::getCNAME() const {
    return __cname;
  }
  
  void RTPStrategy::setTTL(const nmm_uint32& ttl) {
    __ttl = ttl;
  }
  
  nmm_uint32 RTPStrategy::getTTL() const {
    return __ttl;
  }
  
  void RTPStrategy::setBufferSize(const nmm_uint32& buffer_size) {
    __buffer_size = buffer_size;
  }
  
  nmm_uint32 RTPStrategy::getBufferSize() const {
    return __buffer_size;
  }
  
  void RTPStrategy::setMaxFrameSize(const nmm_uint32& max_frame_size) {
    __max_frame_size = max_frame_size;
  }
  
  nmm_uint32 RTPStrategy::getMaxFrameSize() const {
    return __max_frame_size;
  }
  
  void RTPStrategy::setSessionBandwidth(const nmm_uint32& bandwidth) {
    __bandwidth = bandwidth;
  }
  
  nmm_uint32 RTPStrategy::getSessionBandwidth() const {
    return __bandwidth;
  }

  void RTPStrategy::sendRTCPReport() {
    if(__rtcp_instance) {
      __rtcp_instance->sendReport();
    }
    else {
      GLOBAL_ERROR_STREAM1("RTPStrategy> No RTCPInstance! Can't send report" << endl);
    }
  }
  
  void RTPStrategy::setRTPSocket(Groupsock* sock) {
    if (__rtp_sock) {
      delete __rtp_sock;
    }
    
    __rtp_sock = sock;
  }
  
  void RTPStrategy::setRTCPSocket(Groupsock* sock) {
    if (__rtcp_sock) {
      delete __rtcp_sock;
    }
    
    __rtcp_sock = sock;
  }
  
  void RTPStrategy::scheduleDestruction() {
    delete this;
  }
  
  std::string RTPStrategy::getStrategyType() const {
    return getStaticStrategyType();
  }
  
  void RTPStrategy::setAddress(const TSAddress& address)
    throw (const BindingException&)
  {
    /* create a local socket */
    setupReceiverSockets(dynamic_cast<const RTPAddress*>(&address));
    
    /* check if we have a fully specified RTP address */
    const RTPAddress* rtp_addr = dynamic_cast<const RTPAddress*>(&address);
    if (rtp_addr) {
      
      /* set remote address */
      delete __remote_address;
      __remote_address = rtp_addr->copy();
      
      /* set remote connection manager */
      if(rtp_addr->getConnectionManager()) {
        __icm = dynamic_cast<IConnectionManager*>(rtp_addr->getConnectionManager()->copy());
        __remote_ts_ref = rtp_addr->getTSReference();
      }
      
      return;
    }
    
    /* check if we have the address of a connection manager */
    const ConnectorAddress* connector_address = dynamic_cast<const ConnectorAddress*>(&address);
    if (connector_address) {
      /* make sure the address is valid */
      if (!connector_address->getConnectionManager()) {
        GLOBAL_ERROR_STREAM1("RTPStrategy> Shared address has no IConnectionManager interface" << endl);
        throw BindingException("Shared address has no IConnectionManager interface");
      }
      
      /* get the local connection manager and request connection */
      ConnectorAddress* remote_addr = NULL;
      const ConnectionManager& local_icm = SConnectionManager::getInstance();
      local_icm.connect(__local_address, remote_addr, connector_address->getConnectionManager());
      
      /* the request was successful */
      /* check if the returned address is actually a RTP address */
      if (!remote_addr) {
        GLOBAL_ERROR_STREAM1("RTPStrategy> Could not get remote RTP address" << endl);
        throw BindingException("Could not get remote RTP address");
      }
      const RTPAddress* rtp_addr = dynamic_cast<RTPAddress*>(remote_addr);
      if (!rtp_addr) {
        GLOBAL_ERROR_STREAM1("RTPStrategy> Remote address is not a RTPAddress" << endl);
        throw BindingException("Remote address is not a RTPAddress");
      }
      __remote_address = rtp_addr->copy();
      
      /* check if the returned NMM reference is valid */
      if (!__remote_address->getTSReference()) {
        GLOBAL_WARNING_STREAM1("RTPStrategy> Could not get NMM reference of remote transport strategy" << endl);
      }
      
      /* store the IConnectionManager interface for later use */
      __icm = dynamic_cast<IConnectionManager*>(connector_address->getConnectionManager()->copy());
      __remote_ts_ref = remote_addr->getTSReference();
      
      return;
    }
    
    throw BindingException(getStrategyType() + " can't handle address of type " + address.getTypeName());
  }
  
  Result RTPStrategy::sendReliableEvent(Event& event) const {
    /* If we have no IConnectionManager, return FAILURE */
    if (!__icm) {
      GLOBAL_ERROR_STREAM1("RTPStrategy> sendReliableEvent failed (No IConnectionManager)" << endl);
      return FAILURE;
    }
    
    /* If we have no partner reference, return FAILURE */
    if (!__remote_ts_ref) {
      GLOBAL_ERROR_STREAM1("RTPStrategy> sendReliableEvent failed (No partner reference)" << endl);
      return FAILURE;
    }
    
    try {
      return __icm->sendReliableEvent(__remote_ts_ref, event);
    } catch(BindingException& e) {
      GLOBAL_ERROR_STREAM1("RTPStrategy> sendReliableEvent failed (" << e << ")" << endl);
      return FAILURE;
    }
  }
  
  Result RTPStrategy::sendReliableConfigEvent(Event& event) const {
    /* If we have no IConnectionManager, return FAILURE */
    if (!__icm) {
      GLOBAL_ERROR_STREAM1("RTPStrategy> sendReliableConfigEvent failed (No IConnectionManager)" << endl);
      return FAILURE;
    }
    
    /* If we have no partner reference, return FAILURE */
    if (!__remote_ts_ref) {
      GLOBAL_ERROR_STREAM1("RTPStrategy> sendReliableConfigEvent failed (No partner reference)" << endl);
      return FAILURE;
    }
    
    try {
      return __icm->sendReliableConfigEvent(__remote_ts_ref, event);
    } catch(BindingException& e) {
      GLOBAL_ERROR_STREAM1("RTPStrategy> sendReliableConfigEvent failed (" << e << ")" << endl);
      return FAILURE;
    }
  }
  
  void RTPStrategy::setupReceiverSockets(const RTPAddress* address)
    throw (const BindingException&)
  {
    try {
      /* delete old local address */
      delete __local_address;
      __local_address = NULL;
      
      /* delete old sockets */
      setRTPSocket(NULL);
      setRTCPSocket(NULL);
      
      /* IP address to use for creating the socket (default: INADDR_NONE) */
      const UnicastIPAddress null_ip("0.0.0.0");
      const MulticastIPAddress* mcast_ip = NULL;
      const IPAddress* ip = &null_ip;
      
      /* ports for creating the sockets (default: random) */
      unsigned rtp_port = 0;
      unsigned rtcp_port = 0;
      
      /* check what kind of address we have */
      if(address) {
        mcast_ip = dynamic_cast<const MulticastIPAddress*>(&address->getIPAddress());
        if(mcast_ip) {
          /* use multicast */
          ip = mcast_ip;
        }
        
        // set the ports for unicast as well as multicast
        
        /* use ports from address */
        rtp_port = address->getRTPPort();
        rtcp_port = address->getRTCPPort();
      }
      
      /* variables for getting a random port */
      int socket;
      struct sockaddr_in addr;
      socklen_t len_addr = sizeof(addr);
      addr.sin_addr = null_ip.getINetAddr();
      addr.sin_port = 0;
      
      /* Begin> Create RTP socket */
      if(rtp_port == 0) {
        /* create a socket with a random port */
        setRTPSocket(new Groupsock(getUsageEnvironment(), ip->getINetAddr(), 0, getTTL()));
        
        /* force the socket to be bound */
        socket = getRTPSocket().socketNum();
        sendto(socket, &socket, 1, 0, reinterpret_cast<struct sockaddr*>(&addr), len_addr);
        
        /* find out the local port */
        if (socket >= 0 && getsockname(socket, reinterpret_cast<struct sockaddr*>(&addr), &len_addr) == 0) {
          rtp_port = ntohs(addr.sin_port);
        }
        if (!rtp_port) {
          throw BindingException("Could not find a free port for RTP");
        }
      } else {
        /* create a socket with a fixed port */
        setRTPSocket(new Groupsock(getUsageEnvironment(), ip->getINetAddr(), rtp_port, getTTL()));
      }
      /* End< */

      /* Begin> Create RTCP socket */
      if(rtcp_port == 0) {
        /* create a socket with a random port */
        setRTCPSocket(new Groupsock(getUsageEnvironment(), null_ip.getINetAddr(), 0, getTTL()));
        
        /* force the socket to be bound */
        socket = getRTCPSocket().socketNum();
        sendto(socket, &socket, 1, 0, reinterpret_cast<struct sockaddr*>(&addr), len_addr);
        
        /* find out the local port */
        if (socket >= 0 && getsockname(socket, reinterpret_cast<struct sockaddr*>(&addr), &len_addr) == 0) {
          rtcp_port = ntohs(addr.sin_port);
        }
        if (!rtcp_port) {
          throw BindingException("Could not find a free port for RTCP");
        }
      }
      else {
        /* create a socket with a fixed port */
        setRTCPSocket(new Groupsock(getUsageEnvironment(), ip->getINetAddr(), rtcp_port, getTTL()));

        if(address) {
          const UnicastIPAddress* ucast_ip = 0;
          ucast_ip = dynamic_cast<const UnicastIPAddress*>(&address->getIPAddress());
          if (ucast_ip) {
            ip = ucast_ip;
            // we used ip = null_ip to create RTCPSocket. To receive RR reports on sender side we have to
            // use changeDestinationParameters(...) to set IP correctly.
            getRTCPSocket().changeDestinationParameters(ip->getINetAddr(), rtcp_port, 255);
            GLOBAL_DEBUG_STREAM2("remote address is : ip_addr=" << ip->toString() <<
                " rtp_port=" << rtp_port <<
                " rtcp_port=" << rtcp_port << endl);
          }
        }
      }
      /* End< */
      
      /* set receive buffer size. we use a vbery high number here to
         actually get the largest receive buffer possible (it is
         limited by /proc/sys/net/rmem_max */
      setRecvBufferSize(2000000);
      
      /* get local IP address */
      char buffer[256];
      ::gethostname(buffer, 256);
      std::string hostname(buffer);
      UnicastIPAddress local_ip(hostname);
      
      /* setup local address */
      if(!mcast_ip) {
        ip = &local_ip;
      }
      __local_address = new RTPAddress(*ip, rtp_port, rtcp_port, getReference());
      GLOBAL_DEBUG_STREAM2("local address is : ip_addr=" << ip->toString() <<
                           " rtp_port=" << rtp_port <<
                           " rtcp_port=" << rtcp_port << endl);
      
      /* set strategy description */
      __local_address->setTSDescription(getTSDescription());
        
      /* set local connection manager interface */
      __local_address->setConnectionManager(*SConnectionManager::getInstance().getInterface<IConnectionManager>());
      __local_address->setTSReference(getReference());
 
    } catch(...) {
      /* clean up created sockets */
      setRTPSocket(NULL);
      setRTCPSocket(NULL);
      
      /* delete local address */
      delete __local_address;
      __local_address = NULL;
      
      /* rethrow exception */
      throw;
    }
  }
  
  void RTPStrategy::setupSenderSockets() throw(const BindingException&) {
    /* check whether we have a remote address */
    if(!__remote_address) {
      throw BindingException("No remote address");
    }
    
    /* check whether remote address is a multicast address */
    if(dynamic_cast<const MulticastIPAddress*>(&__remote_address->getIPAddress())) {
      /* nothing more to do */
      return;
    }
    
    /* Begin> set the destination address */
    const IPAddress& dest_ip(__remote_address->getIPAddress());
    unsigned int rtp_port = __remote_address->getRTPPort();
    unsigned int rtcp_port = __remote_address->getRTCPPort();
    getRTPSocket().changeDestinationParameters(dest_ip.getINetAddr(), rtp_port, getTTL());
    getRTCPSocket().changeDestinationParameters(dest_ip.getINetAddr(), rtcp_port, getTTL());
    GLOBAL_DEBUG_STREAM2("remote address is: ip_addr=" << dest_ip.toString() <<
                         " rtp_port=" << rtp_port <<
                         " rtcp_port=" << rtcp_port << endl);
    /* End< */
  }
  
  void RTPStrategy::start() {
    /* stop thread first */
    stop();
    
    /* create thread */
    __live_thread = new LiveThread(getUsageEnvironment(), *getMediaSource(), *getMediaSink());
    
    /* start thread */
    __live_thread->start();
    GLOBAL_DEBUG_STREAM2("---- RTP THREAD STARTED ----" << endl);
  }
  
  void RTPStrategy::stop() {
    if (__live_thread) {
      /* cancel blocking calls and flush all queues */
      flush();
      
      /* stop LIVE thread */
      __live_thread->stop();
      GLOBAL_DEBUG_STREAM2("---- RTP THREAD STOPPED ----" << endl);
      
      /* delete thread */
      delete __live_thread;
      __live_thread = NULL;
    }
  }
  
  void RTPStrategy::setRTCPInstance(RTCPInstance* rtcp_instance) {
    __rtcp_instance = rtcp_instance;
  }
  
  const char RTPStrategy::__strategy_type[] = "RTPStrategy";
  
  RTPStrategy::FormatList::FormatList() {
    /* Begin> MPEG streams (RFC 2250) */
    /* Begin> MPEG video formats used in NMM */
    Format video_mpeg(Format::video_mpeg_type);
    video_mpeg.addIntParamValue(Format::version_param, 1);
    video_mpeg.addIntParamValue(Format::version_param, 2);
    addFormat(video_mpeg);
      
    Format video_mpeg_detail(Format::video_mpeg_type);
    video_mpeg_detail.addIntParamValue(Format::version_param, 1);
    video_mpeg_detail.addIntParamValue(Format::version_param, 2);
    video_mpeg_detail.addWildcard(Format::x_resolution_param);
    video_mpeg_detail.addWildcard(Format::y_resolution_param);
    video_mpeg_detail.addWildcard(Format::ratio_x_param);
    video_mpeg_detail.addWildcard(Format::ratio_y_param);
    video_mpeg_detail.addWildcard(Format::framerate_param);
    addFormat(video_mpeg_detail);
    /* End< MPEG video formats used in NMM */
    
    /* Begin> motion-JPEG streams (RFC 2435) */
    /* Begin> motion-JPEG video formats used in NMM */
    Format video_jpeg(Format::video_jpeg_type);
    video_jpeg.addWildcard(Format::x_resolution_param);
    video_jpeg.addWildcard(Format::y_resolution_param);
    video_jpeg.addWildcard(Format::ratio_x_param);
    video_jpeg.addWildcard(Format::ratio_y_param);
    video_jpeg.addWildcard(Format::framerate_param);
    addFormat(video_jpeg);
    /* End< motion-JPEG video formats used in NMM */
    
    /* Begin> h623plus streams (RFC 2429) */
    /* Begin> h623plus video formats used in NMM */
    Format video_h623plus(Format::video_h263plus_type);
    video_h623plus.addWildcard(Format::x_resolution_param);
    video_h623plus.addWildcard(Format::y_resolution_param);
    video_h623plus.addWildcard(Format::ratio_x_param);
    video_h623plus.addWildcard(Format::ratio_y_param);
    video_h623plus.addWildcard(Format::framerate_param);
    video_h623plus.addWildcard(Format::bitrate_param);
    addFormat(video_h623plus);
    /* End< motion-JPEG video formats used in NMM */
    
    /* Begin> mpeg4 streams */
    /* Begin> mpeg4 video format used in NMM */
    Format video_mpeg4(Format::video_mpeg4_type);
    video_mpeg4.addWildcard(Format::x_resolution_param);
    video_mpeg4.addWildcard(Format::y_resolution_param);
    video_mpeg4.addWildcard(Format::ratio_x_param);
    video_mpeg4.addWildcard(Format::ratio_y_param);
    video_mpeg4.addWildcard(Format::framerate_param);
    video_mpeg4.addWildcard(Format::bitrate_param);
    video_mpeg4.addStringParamValue(Format::compressor_param, "DIVX");
    video_mpeg4.addStringParamValue(Format::compressor_param, "DX50");
    addFormat(video_mpeg4);
    /* End< mpeg4 video format used in NMM */
    
    /* Begin> MPEG audio formats used in NMM */
    Format audio_mpeg_mime(Format::audio_mpeg_type);
    addFormat(audio_mpeg_mime);
      
    Format audio_mpeg(Format::audio_mpeg_type);
    audio_mpeg.addIntParamValue(Format::layer_param, 2);
    audio_mpeg.addIntParamValue(Format::layer_param, 3);
    audio_mpeg.addIntParamValue(Format::audio_codec_param, 0x55);
    addFormat(audio_mpeg);
      
    Format audio_mpeg_bitrate(Format::audio_mpeg_type);
    audio_mpeg_bitrate.addIntParamValue(Format::layer_param, 2);
    audio_mpeg_bitrate.addIntParamValue(Format::layer_param, 3);
    audio_mpeg_bitrate.addIntParamValue(Format::audio_codec_param, 0x55);
    audio_mpeg_bitrate.addWildcard(Format::bitrate_param);
    addFormat(audio_mpeg_bitrate);
      
    Format audio_mpeg_bitrate_detail(Format::audio_mpeg_type);
    audio_mpeg_bitrate_detail.addIntParamValue(Format::layer_param, 2);
    audio_mpeg_bitrate_detail.addIntParamValue(Format::layer_param, 3);
    audio_mpeg_bitrate_detail.addIntParamValue(Format::audio_codec_param, 0x55);
    audio_mpeg_bitrate_detail.addWildcard(Format::sample_rate_param);
    audio_mpeg_bitrate_detail.addWildcard(Format::channels_param);
    audio_mpeg_bitrate_detail.addWildcard(Format::bitpersample_param);
    audio_mpeg_bitrate_detail.addWildcard(Format::bitrate_param);
    addFormat(audio_mpeg_bitrate_detail);

    Format audio_mpeg_detai(Format::audio_mpeg_type);
    audio_mpeg_detai.addIntParamValue(Format::layer_param, 2);
    audio_mpeg_detai.addIntParamValue(Format::layer_param, 3);
    audio_mpeg_detai.addIntParamValue(Format::audio_codec_param, 0x55);
    audio_mpeg_detai.addWildcard(Format::sample_rate_param);
    audio_mpeg_detai.addWildcard(Format::channels_param);
    audio_mpeg_detai.addWildcard(Format::bitpersample_param);
    addFormat(audio_mpeg_detai);
    /* End< MPEG audio formats used in NMM */
    /* End< MPEG streams (RFC 2250) */

    // AMR-NB format used in NMM
    Format amr_nb(Format::audio_amr_type);
    amr_nb.addIntParamValue(Format::channels_param, 1);
    amr_nb.addIntParamValue(Format::sample_rate_param, 8000);
    amr_nb.addIntParamValue(Format::bitpersample_param, 16);
    amr_nb.addIntParamValue(Format::bitrate_param, 4750);
    amr_nb.addIntParamValue(Format::bitrate_param, 5150);
    amr_nb.addIntParamValue(Format::bitrate_param, 5900);
    amr_nb.addIntParamValue(Format::bitrate_param, 6700);
    amr_nb.addIntParamValue(Format::bitrate_param, 7400);
    amr_nb.addIntParamValue(Format::bitrate_param, 7950);
    amr_nb.addIntParamValue(Format::bitrate_param, 10200);
    amr_nb.addIntParamValue(Format::bitrate_param, 12200);
    amr_nb.addStringParamValue(Format::encoding_param, "amr_nb");
    addFormat(amr_nb);

    // AMR-WB format used in NMM
    Format amr_wb(Format::audio_amr_type);
    amr_wb.addIntParamValue(Format::channels_param, 1);
    amr_wb.addIntParamValue(Format::sample_rate_param, 16000);
    amr_wb.addIntParamValue(Format::bitpersample_param, 16);
    //amr_wb.addWildcard(Format::bitrate_param);
    amr_wb.addIntParamValue(Format::bitrate_param, 6600);
    amr_wb.addIntParamValue(Format::bitrate_param, 8850);
    amr_wb.addIntParamValue(Format::bitrate_param, 12650);
    amr_wb.addIntParamValue(Format::bitrate_param, 14250);
    amr_wb.addIntParamValue(Format::bitrate_param, 15850);
    amr_wb.addIntParamValue(Format::bitrate_param, 18250);
    amr_wb.addIntParamValue(Format::bitrate_param, 19850);
    amr_wb.addIntParamValue(Format::bitrate_param, 23050);
    amr_wb.addIntParamValue(Format::bitrate_param, 23850);

    amr_wb.addStringParamValue(Format::encoding_param, "amr_wb");
    addFormat(amr_wb);

    /* Begin> raw formats (RFC 1890) */
    Format audio_raw_l16_mono(Format::audio_raw_type);
    audio_raw_l16_mono.addStringParamValue(Format::endian_param, "little");
    audio_raw_l16_mono.addStringParamValue(Format::encoding_param, "linear");
    audio_raw_l16_mono.addStringParamValue(Format::sign_param, "signed");
    audio_raw_l16_mono.addIntParamValue(Format::sample_rate_param, 8000);
    audio_raw_l16_mono.addIntParamValue(Format::sample_rate_param, 44100);
    audio_raw_l16_mono.addIntParamValue(Format::sample_rate_param, 48000);
    audio_raw_l16_mono.addIntParamValue(Format::channels_param, 1);
    audio_raw_l16_mono.addIntParamValue(Format::bitpersample_param, 16);
    addFormat(audio_raw_l16_mono);
    
    Format audio_raw_l16_stereo(Format::audio_raw_type);
    audio_raw_l16_stereo.addStringParamValue(Format::endian_param, "little");
    audio_raw_l16_stereo.addStringParamValue(Format::encoding_param, "linear");
    audio_raw_l16_stereo.addStringParamValue(Format::sign_param, "signed");
    audio_raw_l16_stereo.addIntParamValue(Format::sample_rate_param, 8000);
    audio_raw_l16_stereo.addIntParamValue(Format::sample_rate_param, 44100);
    audio_raw_l16_stereo.addIntParamValue(Format::sample_rate_param, 48000);
    audio_raw_l16_stereo.addIntParamValue(Format::channels_param, 2);
    audio_raw_l16_stereo.addIntParamValue(Format::bitpersample_param, 16);
    addFormat(audio_raw_l16_stereo);
    /* End< raw formats (RFC 1890) */
      
    /* Begin> raw formats (RFC 3190) */
    Format audio_raw_l24(Format::audio_raw_type);
    audio_raw_l24.addStringParamValue(Format::endian_param, "little");
    audio_raw_l24.addStringParamValue(Format::encoding_param, "linear");
    audio_raw_l24.addStringParamValue(Format::sign_param, "signed");
    audio_raw_l24.addIntParamValue(Format::sample_rate_param, 8000);
    audio_raw_l24.addIntParamValue(Format::sample_rate_param, 44100);
    audio_raw_l24.addIntParamValue(Format::sample_rate_param, 48000);
    audio_raw_l24.addIntParamValue(Format::channels_param, 1);
    audio_raw_l24.addIntParamValue(Format::channels_param, 2);
    audio_raw_l24.addIntParamValue(Format::bitpersample_param, 24);
    addFormat(audio_raw_l24);
      /* End< raw formats (RFC 3190) */
          
    /* Begin> pcmu formats (RFC 1890) */
    Format audio_pcmu(Format::audio_raw_type);
    audio_pcmu.addStringParamValue(Format::endian_param, "little");
    audio_pcmu.addStringParamValue(Format::encoding_param, "mu_law");
    audio_pcmu.addStringParamValue(Format::sign_param, "signed");
    audio_pcmu.addIntParamValue(Format::sample_rate_param, 8000);
    audio_pcmu.addIntParamValue(Format::sample_rate_param, 44100);
    audio_pcmu.addIntParamValue(Format::sample_rate_param, 48000);
    audio_pcmu.addIntParamValue(Format::channels_param, 1);
    audio_pcmu.addIntParamValue(Format::channels_param, 2);
    audio_pcmu.addIntParamValue(Format::bitpersample_param,8);
    addFormat(audio_pcmu);
    /* End< pcmu formats (RFC 1890) */

    /* Begin> pcma formats (RFC 1890) */
    Format audio_pcma(Format::audio_raw_type);
    audio_pcma.addStringParamValue(Format::endian_param, "little");
    audio_pcma.addStringParamValue(Format::encoding_param, "a_law");
    audio_pcma.addStringParamValue(Format::sign_param, "signed");
    audio_pcma.addIntParamValue(Format::sample_rate_param, 8000);
    audio_pcma.addIntParamValue(Format::sample_rate_param, 44100);
    audio_pcma.addIntParamValue(Format::sample_rate_param, 48000);
    audio_pcma.addIntParamValue(Format::channels_param, 1);
    audio_pcma.addIntParamValue(Format::channels_param, 2);
    audio_pcma.addIntParamValue(Format::bitpersample_param, 8);
    addFormat(audio_pcma);
    /* End< pcma formats (RFC 1890) */
    
    Format ac3_format("audio/ac3"); 
    addFormat(ac3_format);
  }  
  
}

#include "nmm/nmm_suffix.hpp"
