/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#ifndef NMM_RTP_STRATEGY_HPP
#define NMM_RTP_STRATEGY_HPP NMM_RTP_STRATEGY_HPP

#include "nmm/nmm_prefix.hpp"


#include "IRTPControlImpl.hpp"
#include "RTPAddress.hpp" // for covariant return
#include "nmm/comm/TransportStrategy.hpp"
#include "nmm/comm/TSAddressOwner.hpp"
#include "nmm/comm/TSFormatList.hpp"

/* live forwards */
class TaskScheduler;
class UsageEnvironment;
class Groupsock;
class MediaSource;
class MediaSink;
class RTCPInstance;

#include "visibility.hpp"

namespace NMM {
  
  class IConnectionManager;
  class RTPAddress;  
  class LiveThread;  
  
  /** RTP Transport strategy base class.
   *
   * This class implements common functionality of a
   * %TransportStrategy using the Realtime Transport Protocol (RTP).
   * The subclasses %RTPSenderStrategy and %RTPReceiverStrategy
   * implement sending and receiving instream messages.
   *
   * @see RTPSenderStrategy TransportStrategy RTPAddress IRTPControl
   */
  class DLLEXPORT RTPStrategy :
    public TransportStrategy,
    public TSAddressOwner,
    virtual public IRTPControlImpl
  {
  public:
    /** Format list for the %RTPStrategy.
     * 
     * This list contains all formats supported by RTP.
     */
    class FormatList : public TSFormatList {
    public:
      FormatList();
    };
    
  public:
    /** @name Constructor & destructor
     */
    //@{
    /** Default constructor.
     */
    RTPStrategy(const string& name = "RTPStrategy");

    /** Virtual destructor.
     */
    virtual ~RTPStrategy();
    //@}

    /** @name IIPControl interface
     */
    //@{
    /** This method sets the buffer size for incoming data.
     */
    virtual void setRecvBufferSize(const nmm_int32& value);

    /** This method returns the buffer size for incoming data.
     */
    virtual nmm_int32 getRecvBufferSize() const;

    /** This method sets the buffer size for outgoing data.
     */
    virtual void setSendBufferSize(const nmm_int32& value);

    /** This method returns the buffer size for outgoing data.
     */
    virtual nmm_int32 getSendBufferSize() const;

    /** This method sets the priority of the data
     */
    virtual void setPriority(const nmm_int32& value);
    
    /** This method returns the used priority to send the data
     */
    virtual nmm_int32 getPriority() const;
    //@}
    
    /** @name IRTPControl Interface
     */
    //@{
    /** Set canonical name.
     *
     * This method sets the canonical name of this RTP session. This
     * is a free-form string which identifies an RTP session. The
     * default is "NMM-RTP".
     *
     * @param cname The canonical name
     */
    virtual void setCNAME(const string& cname);
    
    /** Get canonical name.
     *
     * @return The canonical name
     */
    virtual string getCNAME() const;
    
    /** Set multicast TTL.
     *
     * This method sets the time to live (TTL) for multicast
     * packets. For unicast sessions, this parameter has no
     * effect. The integer number is the maximum number of hops until
     * a packet is deleted.
     *
     * @param ttl The multicast TTL
     */
    virtual void setTTL(const nmm_uint32& ttl);
    
    /** Get multicast TTL.
     *
     * @return The multicast TTL
     */
    virtual nmm_uint32 getTTL() const;
    
    /** Set buffer size.
     *
     * This method sets the buffer size for receiving RTP packets.
     *
     * @param buffer_size The buffer size
     */
    virtual void setBufferSize(const nmm_uint32& buffer_size);
    
    /** Get buffer size.
     *
     * @return The buffer size
     */
    virtual nmm_uint32 getBufferSize() const;
    
    /** Set maximum frame size.
     *
     * This method sets the maximum size that a received RTP frame may
     * have. If an incoming frame is larger than this size, it may be
     * truncated.
     *
     * @param max_frame_size The maximum frame size (in bytes)
     */
    virtual void setMaxFrameSize(const nmm_uint32& max_frame_size);
    
    /** Get maximum frame size.
     *
     * @return The maximum frame size (in bytes)
     */
    virtual nmm_uint32 getMaxFrameSize() const;
    
    /** Set session bandwidth.
     *
     * This method sets the session bandwidth. Note that this is only
     * an estimate which serves as a hint to the RTP library. The
     * actual bandwidth is determined by the data which is transferred
     * (for example the video bitrate).
     *
     * @param bandwidth The session bandwidth (in kb/s)
     * @todo Check if this parameter is really in kb/s
     */
    virtual void setSessionBandwidth(const nmm_uint32& bandwidth);
    
    /** Get session bandwidth.
     *
     * @retun The session bandwidth (in kb/s)
     */
    virtual nmm_uint32 getSessionBandwidth() const;

    /** Explicitly send an RTCP report.
     */
    virtual void sendRTCPReport();
    //@}
    
    /** @name %TransportStrategy methods
     */
    //@{
    /** Schedule destruction of this object.
     */
    virtual void scheduleDestruction();
    
    /** Get transport strategy type.
     *
     * @return "RTPStrategy"
     */
    virtual string getStrategyType() const;
    
    /** Set address.
     *
     * This method sets the address of the partner transport
     * strategy. The address must either be a %ConnectorAddress (for
     * indirect connection setup) or an RTPAddress (if the connection
     * parameters are known).
     *
     * @param address Address to connect to
     */
    virtual void setAddress(const TSAddress& address)
      throw (const BindingException&);
    
    /** Reliable events
     */
    //@{
    /** This method sends an Event reliably.
     *
     * This transport strategy does not support out-of-band events.
     *
     * @return @c FAILURE
     */
    virtual Result sendReliableEvent(Event&) const;
    
    /** This method sends a configuration Event reliably.
     *
     * This transport strategy does not support out-of-band events.
     *
     * @return @c FAILURE
     */
    virtual Result sendReliableConfigEvent(Event&) const;
    
    /** Check if this strategy is connected.
     */
    virtual bool isConnected() const {
      return ( __live_thread != 0 );
    }
    //@}
    
    /** @name %TSAddressOwner methods
     */
    //@{
    /** Get local address.
     *
     * This method returns the local RTP address. It contains all the
     * connection parameters needed by a partner strategy to connect
     * to this strategy.
     *
     * @return The local RTP address
     */
    virtual const RTPAddress& getLocalAddress() const throw();
    //@}
    
    /** Get transport strategy type name.
     *
     * @return "RTPStrategy"
     */
    static string getStaticStrategyType();
    //@}

    /** This method sets the priority of instream messages. The default value is STREAMING_DATA
     *
     * @param priority The priority
     */
    void setInstreamBufferPriority(const Sock::QOS_TAG& msg) { __instream_buffer = msg; }

    /** This method sets the priority of instream events. The default value is INBAND_DATA
     *
     * @param priority The priority
     */
    void setInstreamEventPriority(const Sock::QOS_TAG& msg)  { __instream_event = msg; }

    /** This method sets the priority of out of band events. The default value is RMI_DATA
     *
     * @param priority The priority
     */
    void setOutOfBandEventPriority(const Sock::QOS_TAG& msg) { __out_of_band_event = msg; }

    /** This method sets the priority of instream messages. The default value is STREAMING_DATA
     *
     * @param return The priority
     */
    Sock::QOS_TAG getInstreamBufferPriority() { return __instream_buffer; }

    /** This method gets the priority of instream events. The default value is INBAND_DATA
     *
     * @param return The priority
     */
    Sock::QOS_TAG getInstreamEventPriority()  { return __instream_event; }


    /** This method gets the priority of out of band events. The default value is RMI_DATA
     *
     * @param return The priority
     */
    Sock::QOS_TAG getOutOfBandEventPriority() { return __out_of_band_event; }
    //@}
    
  protected:
    /** Get RTP socket.
     *
     * This method returns the RTP socket used by this session.
     *
     * @return The RTP socket
     */
    Groupsock& getRTPSocket() const;
    
    /** Get RTCP socket.
     *
     * This method returns the RTCP socket used by this session.
     *
     * @return The RTCP socket
     */
    Groupsock& getRTCPSocket() const;
    
    /** Get task scheduler.
     *
     * This method returns the liveMedia task scheduler used by this
     * session.
     *
     * @return The task scheduler
     */
    TaskScheduler& getTaskScheduler() const;
    
    /** Get usage environment.
     *
     * This method returns the liveMedia usage environment used by
     * this session.
     *
     * @return The usage environment
     */
    UsageEnvironment& getUsageEnvironment() const;
    
    /** Set RTP socket.
     *
     * This method sets the RTP socket. A session does not
     * automatically allocate sockets.  This must be done by a
     * subclass. If a socket was already set, then it is destroyed.
     * If this object is destroyed, then the sockets are destroyed
     * automatically.
     *
     * @param sock the new RTP socket
     */
    void setRTPSocket(Groupsock* sock);
    
    /** Set RTCP socket.
     *
     * This method sets the RTCP socket. A session does not
     * automatically allocate sockets.  This must be done by a
     * subclass. If a socket was already set, then it is destroyed.
     * If this object is destroyed, then the sockets are destroyed
     * automatically.
     *
     * @param sock the new RTCP socket
     */
    void setRTCPSocket(Groupsock* sock);
    
    /** Start liveMedia thread.
     *
     * This method starts the event processing by the liveMedia RTP
     * stack.
     */
    void start();
    
    /** Stop liveMedia thread.
     *
     * This method stops the event processing by the liveMedia RTP
     * stack.
     */
    void stop();
    
    /** Setup receiver sockets address The session address.
     *
     * This method configures the receiver sockets for RTP and
     * RTCP. The sockets may later be reconfigured to be used for
     * sending by setting a destination address.
     *
     * The @c address parameter is interpreted as follows:
     *
     * - A @c NULL pointer configures the sockets for unicast with random ports
     * - An address with a UnicastIPAddress configures the sockets for unicast with fixed ports. The IP address is ignored.
     * - An address with a MulticastIPAddress configures the sockets for multicast with fixed ports
     *
     * @exception BindingException if something goes wrong
     */
    void setupReceiverSockets(const RTPAddress* address)
      throw (const BindingException&);
    
    /** Setup sender sockets.
     *
     * This method configures the sender sockets for RTP and RTCP
     *
     * @exception BindingException if something goes wrong
     */
    void setupSenderSockets() throw (const BindingException&);
    
    /** Set RTCP instance.
     *
     * This method sets the RTCP instance of this session. The
     * instance is created by the subclass and managed by the base
     * class after it is passed to this method.
     *
     * @param rtcp_instance The RTCP instance
     */
    void setRTCPInstance(RTCPInstance* rtcp_instance);
    
    /** Get RTCP instance.
     *
     * This method returns the RTCP instance of this session. 
     *
     */
    RTCPInstance* getRTCPInstance()  { return __rtcp_instance; }
        
    /** Get media source.
     *
     * This method returns the LIVE.COM source medium of this
     * strategy.
     *
     * @return The LIVE.COM media source
     */
    virtual MediaSource* getMediaSource() const throw() = 0;
    
    /** Get media sink.
     *
     * This method returns the LIVE.COM sink medium of this strategy.
     *
     * @return The LIVE.COM media sink
     */
    virtual MediaSink* getMediaSink() const throw() = 0;
    
    /** Flush queues and cancel blocking calls.
     *
     * This method flushes all queues and cancels all blocking calls
     * in the LIVE.COM thread.
     */
    virtual void flush() throw() = 0;
    
  protected:
    /** Payload types according to RFC.
     */
    enum PayloadType {
      /** PCM U-law raw audio */
      PT_AUDIO_PCMU       = 0,

      /** GSM audio */
      PT_AUDIO_GSM        = 3,

      /** PCM A-law raw audio */
      PT_AUDIO_PCMA       = 8,

      /** 16-bit linear stereo raw audio */
      PT_AUDIO_L16_STEREO = 10,

      /** 16-bit linear mono raw audio */
      PT_AUDIO_L16_MONO   = 11,

      /** MPEG audio */
      PT_AUDIO_MPEG       = 14,

      /** AMR audio */
      PT_AUDIO_AMR_NB     = 97,

      /** AMR audio */
      PT_AUDIO_AMR_WB     = 98,

      /** JPEG video */
      PT_VIDEO_JPEG       = 26,

      /** MPEG video */
      PT_VIDEO_MPEG       = 32,
    };

  private:
    /** @name Session Parameters
     */
    //@{
    /** The canonical name (CNAME). */
    string __cname;

    /** The multicast TTL. */
    unsigned int __ttl;

    /** The buffer size. */
    size_t __buffer_size;

    /** The maximum frame size. */
    size_t __max_frame_size;

    /** The session bandwidth (in kb/s). */
    unsigned int __bandwidth;
    //@}

    /** @name Session Components
     */
    //@{
    /** The liveMedia task scheduler. */
    TaskScheduler* __scheduler;
    
    /** The liveMedia usage environment. */
    UsageEnvironment* __env;
    
    /** The RTP socket. */
    Groupsock* __rtp_sock;
    
    /** The RTCP socket. */
    Groupsock* __rtcp_sock;
    
    /** RTCP instance. */
    RTCPInstance* __rtcp_instance;
    
    /** Thread for liveMedia components. */
    LiveThread* __live_thread;
    //@}
    
    /** @name Connection Information
     */
    //@{
    /** Local RTP address. */
    RTPAddress* __local_address;
    
    /** Remote RTP address. */
    RTPAddress* __remote_address;
    
    /** Connection manager interface of partner strategy. */
    IConnectionManager* __icm;
    
    /** NMM reference of partner strategy. */
    NMMReference __remote_ts_ref;
    //@}
    
    Sock::QOS_TAG                   __out_of_band_event;
    Sock::QOS_TAG                   __instream_event;
    Sock::QOS_TAG                   __instream_buffer;

    /** Strategy type. */
    static const char __strategy_type[];
  };
  
  
  inline Groupsock& RTPStrategy::getRTPSocket() const {
    return *__rtp_sock;
  }
  
  inline Groupsock& RTPStrategy::getRTCPSocket() const {
    return *__rtcp_sock;
  }
  
  inline TaskScheduler& RTPStrategy::getTaskScheduler() const {
    return *__scheduler;
  }
  
  inline UsageEnvironment& RTPStrategy::getUsageEnvironment() const {
    return *__env;
  }
  
  inline const RTPAddress& RTPStrategy::getLocalAddress() const throw() {
    return *__local_address;
  }
  
  inline string RTPStrategy::getStaticStrategyType() {
    return __strategy_type;
  }
  
}

#include "nmm/nmm_suffix.hpp"

#endif
