/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "TCPStrategy.hpp"
#include "TCPAddress.hpp"

#include "BufferedINetStream.hpp"
#include "BufferedONetStream.hpp"

#include "nmm/comm/serialize/MagicNumberIValueStream.hpp"
#include "nmm/comm/serialize/MagicNumberOValueStream.hpp"
#include "nmm/comm/serialize/SerializeResult.hpp"

#include "nmm/comm/messaging/Buffer.hpp"
#include "nmm/comm/messaging/CEvent.hpp"
#include "nmm/comm/messaging/IExternalBufferManager.hpp"

#include "nmm/comm/TTSFactoryRegistrator.hpp"
#include "nmm/comm/TTSFactoryObject.hpp"

#include "nmm/utils/net/NetReactor.hpp"

#include "nmm/utils/thread/MutexGuard.hpp"

#include <cassert>

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  TCPStrategy::TCPStrategy(const string& name)
    :
    /* Begin> Initialize all the Attributes with sensfull values */
    TransportStrategy(name),
    __event_condition(__event_mutex), __event_sending_mutex(),
    __request_count(0),
    __request_map(),
    __o_stream(0), __i_stream(0),
    __o_net_stream(0), __i_net_stream(0),
    __is_running(true), __thread_condition(__condition_thread_mutex),
    __send_message(0), __blocked_message(0), __send_message_mutex(ThreadMutex::RECURSIVE),
    __current_address(0), 
    __blocked(false), __blocked_condition(__blocked_mutex)
    /* End <*/
  {
    /* Set the TCP net stream */
    BufferedINetStream i_net_stream;
    BufferedONetStream o_net_stream;
    i_net_stream.setReceiveBufferSize(2048);
    o_net_stream.setSendBufferSize(2048);
    setINetStream(i_net_stream);
    setONetStream(o_net_stream);

    /* Set the default stream */
    MagicNumberIValueStream i_stream;
    setIValueStreamStack(&i_stream);
    MagicNumberOValueStream o_stream;
    setOValueStreamStack(&o_stream);

    /* activate the thread */
    activateThread();
  }

  TCPStrategy::~TCPStrategy() {
    /* send FREE_QUEUE_SLOT signals if there is a stored __send_message. However, a TransportStrategy should be flushed before it is destroyed */
    if ( __send_message ) {
      GLOBAL_WARNING_STREAM ("TransportStrategy was not flushed before desctruction");
      flushITSReceiver();
    }
    
    setSignalReceiver(0);
    setEventReceiver(0);
    setRequestReplyEventReceiver(0);
    setOneWayCEventReceiver(0);
    setOneWayBufferReceiver(0);

    if(getStreamSocket().getSocket() != -1) {
      closeConnection(getStreamSocket().getSocket());
    }

    /* That the running flag to false */
    __condition_thread_mutex.lock();
    __is_running = false;
    __condition_thread_mutex.unlock();

    /* Notify the thread to leave the while loop */
    conditionNotify();

    __blocked_mutex.lock();
    __blocked_condition.notify();
    __blocked_mutex.unlock();

    /* Deactivate the thread */
    deactivateThread();

    /* Destroy value streams */
    if (__i_stream) {
      delete __i_stream;
      __i_stream = 0;
    }
    if (__o_stream) {
      delete __o_stream;
      __o_stream = 0;
    }

    /* Destroy net streams */
    delete __o_net_stream;
    __o_net_stream = 0;
    delete __i_net_stream;
    __i_net_stream = 0;
  }


  Result TCPStrategy::receiveRequestReplyEvent(Event& event) {
    return sendRequest(event, EVENT_REQUEST);
  }

  Result TCPStrategy::receiveOneWayCEvent(CEvent* message) {
    return sendOneWayMessage(message, CEVENT_ONE_WAY);
  }

  Result TCPStrategy::receiveOneWayBuffer(Buffer* message) {
    return sendOneWayMessage(message, BUFFER_ONE_WAY);
  }

  Result TCPStrategy::sendReliableEvent(Event& event) const {
    return sendRequest(event, RELIABLE_EVENT_REQUEST);
  }

  Result TCPStrategy::sendReliableConfigEvent(Event& event) const {
    return sendRequest(event, RELIABLE_CONFIG_EVENT_REQUEST);
  }

  void TCPStrategy::configure(const Format&)
    throw(const FormatNotSupportedException&)
  {
    /* set the send buffer size of the ONetStream */
    if(__o_net_stream) {
      BufferedONetStream* buffered_o =
        dynamic_cast<BufferedONetStream*>(__o_net_stream);
      if(buffered_o) {
        buffered_o->setSendBufferSize(143);
      }
    }
  }

  Result TCPStrategy::sendRequest(const Event& event, TCPMessage message_type) const {
    /* If the thread is no longer running return FAILURE */
    if (!__is_running)
      return FAILURE;

    /* The event_mutex (needed to the event_condition) must be locked before
       writing to ensure that wait() is called before dispatchEventResponse
       can call notify(). */
    MutexGuard event_guard(__event_mutex);
    const_cast<Sock&>(getSocket()).setQoSTag(const_cast<TCPStrategy*>(this)->getOutOfBandEventPriority());


    /* The request ID for this request */
    RequestID request_id = ++__request_count;
     
    try {
      /* Lock the event_sending_mutex to ensure that no two threads access
         the streams and the socket simultaneously. This mutex is unlocked
         when the request has been completely written. */
      MutexGuard event_sending_guard(__event_sending_mutex);

      /* Create a new request state */
      RequestState request_state;
      request_state.event = const_cast<Event*>(&event);
      request_state.received_reply = false;
      request_state.result = FAILURE;

      /* Add the request state to the request map */
      __request_map.insert(std::make_pair(request_id, request_state));

      /* Write the message type to the stream. The receiver needs this information to know
         which receiveEvent-method must be called; the receiveEvent(const Event&) or the
         receiveEvent(const CEvent&) */
      *__o_stream << static_cast<nmm_int32>(message_type);

      /* Write the request ID */
      *__o_stream << __request_count;

      /* Disable the out_values */
      request_state.event->disableMemoryType(Value::OUTVALUE);

      /* Write the value to the stream */
      *__o_stream << event;

      /* Ensure that all data is written (similar to a 'flush') */
      __o_stream->writeData();
    } catch (const TransmissionException& te) {
      const_cast<TCPStrategy*>(this)->closeConnection(getStreamSocket().getSocket());
      throw BindingException(te.getComment());
    } catch(const ConnectionClosed& ce){
      const_cast<TCPStrategy*>(this)->closeConnection(getStreamSocket().getSocket());
      throw BindingException(ce.getComment());
    }

    /* Wait until a reply has been received for this request */
    for(;;) {
      /* Find request state in requets map */
      RequestMap::const_iterator found = __request_map.find(request_id);
      if(found == __request_map.end()) {
        GLOBAL_ERROR_STREAM("TCPStrategy: Request state no longer in request map!! This should not happen!!" << endl);
        return FAILURE;
      }

      /* Check if a reply has been received for this event */
      if(!found->second.received_reply) {
        /* Wait until a reply has been received */
        __event_condition.wait();

        /* The request map may have changed and the iterator may no longer be valid,
           so we need to find the request state again in the hash map. */
        continue;
      }

      /* Copy the request state */
      RequestState request_state = found->second;

      /* Remove the request state from the hash */
      __request_map.erase(request_id);

      /* Return the result */
      return request_state.result;
    }
  }

  void TCPStrategy::sendResponse(Result result, const Event* event, RequestID request_id, TCPMessage message_type) {
    try {
      /* Lock the event_sending_mutex to ensure that no two threads access
         the streams and the socket simultaneously. This mutex is unlocked
         when the reply has been completely written. */
      MutexGuard event_sending_guard(__event_sending_mutex);

      getSocket().setQoSTag(getOutOfBandEventPriority());

      /* Write the message_type */
      (*__o_stream) << static_cast<nmm_int32>(message_type);

      /* Write the request ID */
      (*__o_stream) << request_id;

      /* Write the result */
      (*__o_stream) << static_cast<const Result&>(result);

      /* If given value != 0*/
      if (event)
        /* write the value */
        (*__o_stream) << *event;

      /* call writeData to finish the transmission */
      __o_stream->writeData();

      /* and return */
      return;
    } catch (const TransmissionException& te) {
      GLOBAL_ERROR_STREAM1(te << endl);
    } catch(const ConnectionClosed& ce){
      GLOBAL_ERROR_STREAM1(ce << endl);
    } catch (const Exception& e) {
      GLOBAL_ERROR_STREAM1(e << endl);
    }
    closeConnection(getStreamSocket().getSocket());
  }

  Result TCPStrategy::sendOneWayMessage(const Message* message, TCPMessage /*message_type*/) {
    if (!message) {
      return FAILURE;
    }

    /* Begin> Store the message pointer */
    __condition_thread_mutex.lock();
    __send_message_mutex.lock();

    if (__send_message) {
      GLOBAL_ERROR_STREAM1( "ARGH ERROR IN TCPSTARTEGY __send_message != 0 !!!!!!!!!!!!!" << endl);
      cerr << "ARGH ERROR IN TCPSTARTEGY __send_message != 0 !!!!!!!!!!!!! EVENT? " << (message->getType() == Message::EVENT) << endl;
      __blocked_message = message;
    } else {
      __send_message = message;
    }
  
    __send_message_mutex.unlock();
    __condition_thread_mutex.unlock();
    /* End< */

    /* Notify the Thread */
    conditionNotify();
  
    /* Return blocked*/
    return BLOCKED;
  }

  NetHandler::NetResult TCPStrategy::receiveData(SocketHandle /*socket*/) {
    /* If the transport strategy is not running return remove */
    if(!__is_running) {
      return REMOVE;
    }

    /* Dispatch the message */
    NetHandler::NetResult result = dispatchMessage();
    if(result != RECHECK) return result;

    /* -- Thread is still leader. This is only the case for instream communication -- */

    // XXX: fw: NOTE: The following code is kept here because of the strange handling of the blocked_mutex
    // XXX: fw: NOTE: If the blocked_mutex is no longer needed, then this code can be removed as well
    // XXX: fw: NOTE: (because it is already done in state CHECK_FOR_EVENTS after this method returns)
    // XXX: fw: NOTE: Also note that this code may only be executed by the leader thread!

    /* Check for additional data which has already been read from the socket */
    while(__is_running) {
      if ( !__i_net_stream->isEmpty() && !__blocked && __is_running )  {
        /* Dispatch more data */
        dispatchMessage();
      } else {
        if (__i_net_stream->isEmpty()  && !__blocked && __is_running ) {
          /* Recheck for events */
          return RECHECK;
        }
        __blocked_mutex.lock();
        if ( __blocked ) {
          __blocked_condition.wait();
        }
        __blocked_mutex.unlock();
      }
    }

    /* Remove socket event handler */
    return REMOVE;
  }



  NetHandler::NetResult TCPStrategy::dispatchMessage() {
    try {
      /* Read new data from the socket */
      __i_stream->readData();

      /* Read the message type */
      nmm_int32 tmp_message;
      *__i_stream >> tmp_message;
      TCPMessage message_type = static_cast<TCPMessage>(tmp_message);

      /* Dispatch the message depending on the message type */
      switch(message_type) {
      case EVENT_REQUEST:
        /* Dispatch out-of-band event request */
        dispatchEventRequest(*__i_stream, message_type);
        /* Return VOID because this thread is no longer the leader */
        return VOID;

      case EVENT_RESPONSE:
        /* Dispatch out-of-band event reply */
        dispatchEventResponse(*__i_stream);
        /* Return VOID because this thread is no longer the leader */
        return VOID;

      case CEVENT_ONE_WAY:
        /* Dispatch one-way instream composite event */
        dispatchCEventOneWay(*__i_stream);
        /* Return RECHECK because this thread is still the leader */
        return RECHECK;

      case BUFFER_ONE_WAY:
        /* Dispatch one-way instream buffer */
        dispatchBufferOneWay(*__i_stream);
        /* Return RECHECK because this thread is still the leader */
        return RECHECK;

      case RELIABLE_EVENT_REQUEST:
        /* Dispatch reliable out-of-band event request */
        dispatchEventRequest(*__i_stream, message_type);
        /* Return VOID because this thread is no longer the leader */
        return VOID;

      case RELIABLE_CONFIG_EVENT_REQUEST:
        /* Dispatch reliable config event request */
        dispatchEventRequest(*__i_stream, message_type);
        /* Return VOID because this thread is no longer the leader */
        return VOID;

      default:
        /* Nothing to read or process */
        GLOBAL_ERROR_STREAM1("TCPStrategy> Unknown message type" << endl);

        /* Return RECHECK because this thread is still the leader */
        return RECHECK;
      }
    } catch (const Exception& te) {
      GLOBAL_ERROR_STREAM1( "TCPStrategy dispatch Message " << te << endl);
      __is_running = false;
      closeConnection(getStreamSocket().getSocket());
      return REMOVE;
    }
  }

  void TCPStrategy::setONetStream(const ONetStream& stream) {
    delete __o_net_stream;
    __o_net_stream = stream.copy();
    __o_net_stream->setSock(&getStreamSocket());
  }

  void TCPStrategy::setINetStream(const INetStream& stream) {
    delete __i_net_stream;
    __i_net_stream = stream.copy();
    __i_net_stream->setSock(&getStreamSocket());
  }

  void TCPStrategy::setIValueStream(const string& value_stream_type)
    throw(const BindingException&)
  {
    GLOBAL_DEBUG_STREAM2("Set new IValueStream of type " << value_stream_type << endl);
  
    /* no events can be sent */
    MutexGuard event_guard(__event_sending_mutex);
  
    /* no messages can be sent */
    MutexGuard message_guard(__send_message_mutex);
  
    /* get prototype */
    IValueStream* new_prototype = SValueFactory::getInstance().createIValueStream(value_stream_type);
    if(!new_prototype) {
      throw BindingException("Could not change to new receive ValueStream " + value_stream_type);
    }
  
    /* set value stream */
    delete __i_stream;
    __i_stream = new_prototype;
  
    /* set partner stream */
    GLOBAL_DEBUG_STREAM2("Connect new IValueStream to current INetStream" << endl);
    __i_stream->setIPartnerStream(__i_net_stream);
  }

  void TCPStrategy::setOValueStream(const string& value_stream_type)
    throw(const BindingException&)
  {
    GLOBAL_DEBUG_STREAM2("Set new OValueStream of type " << value_stream_type << endl);
  
    /* no events can be sent */
    MutexGuard event_guard(__event_sending_mutex);
  
    /* no messages can be sent */
    MutexGuard message_guard(__send_message_mutex);
  
    /* get prototype */
    OValueStream* new_prototype = SValueFactory::getInstance().createOValueStream(value_stream_type);
    if(!new_prototype) {
      throw BindingException("Could not change to new send ValueStream " + value_stream_type);
    }
  
    /* set value stream */
    delete __o_stream;
    __o_stream = new_prototype;
  
    /* set partner stream */
    GLOBAL_DEBUG_STREAM2("Connect new OValueStream to current ONetStream" << endl);
    __o_stream->setOPartnerStream(__o_net_stream);
  }

  void TCPStrategy::setIValueStreamStack(const IValueStream* value_stream) {
    /* no events can be sent */
    MutexGuard event_guard(__event_sending_mutex);
  
    /* no messages can be sent */
    MutexGuard message_guard(__send_message_mutex);
  
    /* create a copy of the stream */
    IValueStream* new_prototype = value_stream->copy();
  
    /* set value stream */
    delete __i_stream;
    __i_stream = new_prototype;
  
    /* get the bottom partner stream in the value stream stack */
    IStream* parent = 0;
    IStream* bottom = __i_stream;
    while(bottom->getIPartnerStream()) {
      parent = bottom;
      bottom = bottom->getIPartnerStream();
    }
  
    /* if the bottom stream is a NetStream, set it as our net
       stream. otherwise set our NetStream as partner stream of the
       bottom stream. */
    INetStream* net_stream = dynamic_cast<INetStream*>(bottom);
    if(net_stream) {
      /* top-level stream can't be a NetStream */
      assert(parent);
    
      /* the TCPStrategy owns the reference to the net stream */
      setINetStream(*net_stream);
      parent->setIPartnerStream(0);
      parent->setIPartnerStream(__i_net_stream);
    } else {
      bottom->setIPartnerStream(__i_net_stream);
    }
  }

  void TCPStrategy::setOValueStreamStack(const OValueStream* value_stream) {
    /* no events can be sent */
    MutexGuard event_guard(__event_sending_mutex);
  
    /* no messages can be sent */
    MutexGuard message_guard(__send_message_mutex);
  
    /* create a copy of the stream */
    OValueStream* new_prototype = value_stream->copy();
  
    /* set value stream */
    delete __o_stream;
    __o_stream = new_prototype;
  
    /* get the bottom partner stream in the value stream stack */
    OStream* parent = 0;
    OStream* bottom = __o_stream;
    while(bottom->getOPartnerStream()) {
      parent = bottom;
      bottom = bottom->getOPartnerStream();
    }
  
    /* if the bottom stream is a NetStream, set it as our net
       stream. otherwise set our NetStream as partner stream of the
       bottom stream. */
    ONetStream* net_stream = dynamic_cast<ONetStream*>(bottom);
    if(net_stream) {
      /* top-level stream can't be a NetStream */
      assert(parent);
    
      /* the TCPStrategy owns the reference to the net stream */
      setONetStream(*net_stream);
      parent->setOPartnerStream(0);
      parent->setOPartnerStream(__o_net_stream);
    } else {
      bottom->setOPartnerStream(__o_net_stream);
    }
  }

  int TCPStrategy::getSendingCounter() const {
    return __o_net_stream->getWriteCounter();
  }

  int TCPStrategy::getReceivingCounter() const {
    return __i_net_stream->getReadCounter();
  }

  void TCPStrategy::resetSendingCounter() {
    __o_net_stream->resetWriteCounter();
  }

  void TCPStrategy::resetReceivingCounter() {
    __i_net_stream->resetReadCounter();
  }

  bool TCPStrategy::isDataAvailable() {
    /* Check for additional data which has already been read from the socket */
    if(__is_running) {
      if(!__i_net_stream->isEmpty())  {
        /* Data is already available for reading */
        return true;
      }
    }

    /* No data available */
    return false;
  }

  void TCPStrategy::dispatchEventResponse(IValueStream& iv) {
    /* Lock the event mutex (needed for notifying the event_condition).
       This is done here in order to prevent simultaneous access to __current_value by
       sendRequest and dispatchEventResponse. */
    MutexGuard event_guard(__event_mutex);

    /* Read the request ID */
    RequestID request_id;
    *__i_stream >> request_id;

    /* Find request state in requets map */
    RequestMap::iterator found = __request_map.find(request_id);
    if(found == __request_map.end()) {
      GLOBAL_ERROR_STREAM("TCPStrategy: Received reply to unknown request!! This should not happen!!" << endl);
      return;
    }

    /* Set the reply flag in the request state. */
    /* This must be done first. Even if reading the result fails after this point,
       the thread waiting for the reply must wake up. */
    found->second.received_reply = true;

    /* Read the result */
    try {
      /* Read the result value */
      iv >> static_cast<Result&>(found->second.result);

      /* If the event was processed successfully read the value*/
      if(found->second.result == SUCCESS) {
        found->second.event->resetMemoryType();
        found->second.event->disableMemoryType(Value::INVALUE);
        iv >> *found->second.event;
        found->second.event->resetMemoryType();
      } else {
        /* If the return value is FAILURE read the data if the stream is not empty */
        if (!iv.isEmpty()) {
          found->second.event->resetMemoryType();
          found->second.event->disableMemoryType(Value::INVALUE);
          iv >> *found->second.event;
          found->second.event->resetMemoryType();
        }
      }
    } catch (const ConnectionClosed&) {
      __is_running = false;
      closeConnection(getStreamSocket().getSocket());
    } catch (const TransmissionException&) {
      __is_running = false;
      closeConnection(getStreamSocket().getSocket());
    } catch (const Exception& e) {
      Event * tmp_event = dynamic_cast<Event*>(found->second.event);
      cerr << "Exception during deserialization of event response (method dispatchEventResponse())" << endl;
      if (tmp_event) {
        cerr << "Could not deserialize response for Event " << tmp_event->getKey() << endl;
      } else {
        cerr << "Could not deserialize response for other Value"  << endl;
      }
      GLOBAL_ERROR_STREAM1(e << endl);
    }

    /* -- The messages has been completely read from the socket -- */

    /* Request a new leader thread. */
    requestLeaderThread();

    /* Notify the condition variable */
    __event_condition.notifyAll();
  }


  void TCPStrategy::dispatchEventRequest(IValueStream& iv, TCPMessage message_type) {
    /* Read the request ID */
    RequestID request_id;
    *__i_stream >> request_id;

    /* Begin> If no receiver available  the Event can not handled and return UNDEF*/
    if(message_type == EVENT_REQUEST && !getRequestReplyEventReceiver()) {
      GLOBAL_WARNING_STREAM1("TCPStrategy> No request-reply event receiver set, sending UNDEF response" << std::endl);
      return sendResponse(UNDEF, 0, request_id, EVENT_RESPONSE);
    }
    if(message_type == RELIABLE_EVENT_REQUEST && !getEventReceiver()) {
      GLOBAL_WARNING_STREAM1("TCPStrategy> No event receiver for reliable events set, sending UNDEF response" << std::endl);
      return sendResponse(UNDEF, 0, request_id, EVENT_RESPONSE);
    }
    /* End< */

    /* Create the Event*/
    Event* event = SValueFactory::getInstance().createType<Event>(iv.getCurrentTypeInfo());

    /* If the ValueFactory can not create an instance of the object the event can not be
     *  handled. So return UNDEF
     */
    if(!event) {
      GLOBAL_WARNING_STREAM1("TCPStrategy> Failed to create event, sending UNDEF response" << endl);
      return sendResponse(UNDEF, 0, request_id, EVENT_RESPONSE);
    }

    try {
      /* Otherwise read the event from stream */
      event->createFromStream(iv);
    } catch (const Exception& e) {
      cerr << "Exception during deserialization of event request (method dispatchEventRequest())" << endl;
      GLOBAL_ERROR_STREAM1( "DispatchEventRequest for event " << event->getKey() << endl << e << endl);
    }

    /* -- The messages has been completely read from the socket -- */

    /* Request a new leader thread. */
    requestLeaderThread();

    /* Send the event to the receiver */
    Result result = FAILURE;
    if(message_type == EVENT_REQUEST) {
      result = getRequestReplyEventReceiver()->receiveRequestReplyEvent(*event);
    } else if(message_type == RELIABLE_EVENT_REQUEST) {
      result = getEventReceiver()->receiveEvent(*event, this);
    } else if(message_type == RELIABLE_CONFIG_EVENT_REQUEST) {
      result = getEventDispatcher().dispatch(*event, this);
    } else {
      /* never happens */
      result = FAILURE;
    }

    /* disable the in-values */
    event->disableMemoryType(Value::INVALUE);

    /* and send the response back to the sender */
    sendResponse(result, event, request_id, EVENT_RESPONSE);

    /* reset the Memory type befor destruction */
    event->resetMemoryType();

    /* delete the event */
    delete event;
    /*delete the out and inoutvalues and the return value*/
  }


  void TCPStrategy::dispatchConfigEventRequest(IValueStream& iv) {
    /* Read the request ID */
    RequestID request_id;
    *__i_stream >> request_id;
    /* Create the Event*/
    Event* event = SValueFactory::getInstance().createType<Event>(iv.getCurrentTypeInfo());

    /* If the ValueFactory can not create an instance of the object the event can not be
     *  handled. So return UNDEF
     */
    if (!event)
      return sendResponse(UNDEF, 0, request_id, EVENT_RESPONSE);

    try {
      /* Otherwise read the event from stream */
      event->createFromStream(iv);
    } catch (const Exception& e) {
      cerr << "Exception during deserialization of a internal config event (method dispatchConfigEvent()): " << event->getKey() << endl;
      GLOBAL_ERROR_STREAM1( "DispatchConfigEvent for event " << event->getKey() << endl << e << endl);
    }

    /* -- The messages has been completely read from the socket -- */

    /* Request a new leader thread. */
    requestLeaderThread();

    /* The event is dispatched in the own EventDispatcher */
    Result result = getEventDispatcher().dispatch(*event, this);

    /* and send the response back to the sender */
    sendResponse(result, event, request_id, EVENT_RESPONSE);

    /* delete the event */
    delete event;
  }

  void TCPStrategy::dispatchBufferOneWay(IValueStream& iv) {
    try {
      /* BEGIN> Read size from the stream and allocate the Buffer */
      Buffer* buffer = 0;
      unsigned int size;
      iv >> size;
      if(!getBufferManager()) {
        GLOBAL_ERROR_STREAM1("TCPStrategy> No Buffer Manager set! Discarding buffer!" << std::endl);
        throw Exception("No buffer manager set! No data is read from the network");
      }
      if(getBufferManager()) {
        buffer = getBufferManager()->getNewBuffer(size);
      }
      if(!buffer) {
        GLOBAL_ERROR_STREAM("Failed to allocate Buffer! Discarding buffer!" << endl);
        throw Exception("Could not allocate buffer! No data is read from the network");
      }    
      // XXX: repplix: this should be done in the BufferManager!!
      buffer->setUsed(size);
      /* END< */

      /* BEGIN> Deserialize the message */
      try {
        buffer->createFromStream(iv);
      } catch(const Exception& e) {
        cerr << "Exception during deserialization of Buffer (method dispatchBufferOneWay()): " << endl;
        cerr << e << endl;
      }    
      /* END< Deserialize the message */

      /* BEGIN> Print some debug messages */
      if(buffer->getDirection() == Message::DOWNSTREAM) {
        GLOBAL_DEBUG_STREAM1("TCPStrategy>  Received Downstream Buffer " << buffer->getTimestamp().stream_counter << endl);
      } else if(buffer->getDirection() == Message::UPSTREAM) {
        GLOBAL_DEBUG_STREAM1("TCPStrategy>  Received Upstream Buffer " << buffer->getTimestamp().stream_counter << endl);
      } else {
        GLOBAL_WARNING_STREAM1("TCPStrategy> Received Buffer with UNKNOWN DIRECTION !!!!!!!" << endl);
      }
      /* END< */

      /* BEGIN> Forward the mesage to the receiver */
      // XXX: fw: the handling of BLOCKED here assumes that buffer and event receiver are the same object
      if(getOneWayBufferReceiver()) {
        __blocked_mutex.lock();
        if(getOneWayBufferReceiver()->receiveOneWayBuffer(buffer) == BLOCKED) {
          __blocked = true;
        }
        __blocked_mutex.unlock();
      } else {
        buffer->release();
      }
      /* END< */
    } catch(const Exception& e) {
      GLOBAL_ERROR_STREAM1("TCPStrategy> Exception in TCPStrategy while dispatching one-way Buffer:" << endl);
      GLOBAL_ERROR_STREAM1(e << endl);
      closeConnection(getStreamSocket().getSocket());
    }
  }

  void TCPStrategy::dispatchCEventOneWay(IValueStream& iv) {
    try {
      /* BEGIN> Create the CEvent */
      CEvent* cevent = SValueFactory::getInstance().createType<CEvent>(iv.getCurrentTypeInfo());
      if(!cevent) {
        cerr << "Failed to allocate CEvent" << endl;
        throw Exception("Could not allocate CEvent! No data is read from the network");
      }
      /* END< */
    
      /* BEGIN> Deserialize the message */
      try {
        cevent->createFromStream(iv);
      } catch(const Exception& e) {
        cerr << "Exception during deserialization of instream CEvent (method dispatchCEventOneWay()): " << endl;
        cerr << e << endl;
      }    
      /* END< Deserialize the message */
    
      /* BEGIN> Print some debug messages */
      if(cevent->getDirection() == Message::DOWNSTREAM) {
        GLOBAL_DEBUG_STREAM1("TCPStrategy>  Received Downstream CEvent " << cevent->getTimestamp().stream_counter << endl);
      } else if(cevent->getDirection() == Message::UPSTREAM) {
        GLOBAL_DEBUG_STREAM1("TCPStrategy>  Received Upstream CEvent " << cevent->getTimestamp().stream_counter << endl);
      } else {
        GLOBAL_WARNING_STREAM1("TCPStrategy> Received CEvent with UNKNOWN DIRECTION !!!!!!!" << endl);
      }
      /* END< */
    
      /* BEGIN> Forward the mesage to the receiver */
      // XXX: fw: the handling of BLOCKED here assumes that buffer and event receiver are the same object
      if(getOneWayCEventReceiver()) {
        __blocked_mutex.lock();
        if(getOneWayCEventReceiver()->receiveOneWayCEvent(cevent) == BLOCKED) {
          __blocked = true;
        }
        __blocked_mutex.unlock();
      } else {
        cevent->release();
      }
      /* END< */
    } catch(const Exception& e) {
      GLOBAL_ERROR_STREAM1("TCPStrategy> Exception in TCPStrategy while dispatching one-way CEvent:" << endl);
      GLOBAL_ERROR_STREAM1(e << endl);
      closeConnection(getStreamSocket().getSocket());
    }
  }

  NetHandler::NetResult TCPStrategy::closeConnection(SocketHandle socket) {
    /* That the running flag to false */
    __condition_thread_mutex.lock();
    __is_running = false;
    __condition_thread_mutex.unlock();

    /* Notify the thread to leave the while loop */
    conditionNotify();
    return internalCloseConnection(socket);
  }


  NetHandler::NetResult TCPStrategy::internalCloseConnection(SocketHandle) {
    /* Begin> Reset the Receiver if connection is closed */
    __socket_mutex.lock();
    if (getStreamSocket().getSocket() == -1) {
      __socket_mutex.unlock();
      return REMOVE;
    }

    try {
      getStreamSocket().close();
    } catch (const Exception& e) {
      GLOBAL_ERROR_STREAM1( e << endl );
    }
    /* End< */
    __socket_mutex.unlock();

    /* Begin >Notify all waiting event senders */
    __event_mutex.lock();
    for(RequestMap::iterator i=__request_map.begin(); i != __request_map.end(); ++i) {
      i->second.result = FAILURE;
    }
    __event_condition.notifyAll();
    __event_mutex.unlock();

    /* End< */
    return REMOVE;
  }


  void TCPStrategy::connectTo()
    throw (const ConnectionException&)
  {
    /* Begin> Use the connector to connect to the given address*/
    try {
      __connector.connect(getStreamSocket(), __current_address->getINetAddr());
    
      /* Begin> Set the connection flag to true and register the socket by the NetReactor */
      SNetReactor::getInstance().registerSocket(getStreamSocket().getSocket(), this);
      /* End< */

      /* Begin> Enable TCP NoDelay */
      enableNoDelay(true);
      /* End< */
    } catch (const ConnectionException& ce) {
      GLOBAL_ERROR_STREAM1(ce << endl);
      throw ce;
    }
    /* End< */

  }


  void TCPStrategy::setAddress(const TSAddress& addr)
    throw (const BindingException&)
  {
    const TCPAddress* address = dynamic_cast<const TCPAddress*>(&addr);
    if (!address) {
      throw BindingException("Invalid address type. Expect a TCPAddress");
    }

    delete __current_address;
    __current_address = address->copy();

    try {
      connectTo();
    } catch (const Exception& e) {
      throw BindingException(e.getComment());
    }
  }


  void TCPStrategy::run() {
    while( __is_running ) {
      __blocked_mutex.lock();
      if ( (!__send_message && __is_running) ||
           (!__send_message && __is_running && __blocked) ||
           (!__blocked_message && __is_running) //HACK!!!!!!
           )
        {
          __blocked_mutex.unlock();
          /* If nothing to do wait for jobs */
          conditionWait();
        } else {
          __blocked_mutex.unlock();
        }

      /* Begin> HACK FOR INCORRECT BLOCKED HANDLING */
      if ( __blocked_message ) {
        __send_message_mutex.lock();
        cerr << "SEND BLOCKED MESSAGE " << endl;
        const Message* tmp_message = __blocked_message;
        /* Begin> reset the __send_message_flag */
        __blocked_message = 0;
        /* Notify the Parent about FREE_QUEUE_SLOT*/
        if (getSignalReceiver()) {
          if(tmp_message->getDirection() == Message::DOWNSTREAM) {
            getSignalReceiver()->receiveSignal(FREE_DOWNSTR_QUEUE_SLOT);
          }
          else if(tmp_message->getDirection() == Message::UPSTREAM) {
            getSignalReceiver()->receiveSignal(FREE_UPSTR_QUEUE_SLOT);
          } else {
            GLOBAL_ERROR_STREAM1("TCPStrategy> Unknown message direction" << endl);
          }
        } else {
          GLOBAL_ERROR_STREAM1("TCPStrategy> TCPStrategy has no SignalReceiver" << endl);
        }
        __send_message_mutex.unlock();
        /* End< */

        /* Send message */
        sendQueuedMessage(tmp_message);
      } 
      /* END< HACK !!! */

      __send_message_mutex.lock();
      /*Begin> Handle a message that should be send */
      if ( __send_message ) {
        const Message* tmp_message = __send_message;
        /* Begin> reset the __send_message_flag */
        __send_message = 0;
        /* Notify the Parent about FREE_QUEUE_SLOT*/
        if (getSignalReceiver()) {
          if(tmp_message->getDirection() == Message::DOWNSTREAM) {
            getSignalReceiver()->receiveSignal(FREE_DOWNSTR_QUEUE_SLOT);
          }
          else if(tmp_message->getDirection() == Message::UPSTREAM) {
            getSignalReceiver()->receiveSignal(FREE_UPSTR_QUEUE_SLOT);
          } else {
            GLOBAL_ERROR_STREAM1("TCPStrategy> Unknown message direction" << endl);
          }
        } else {
          GLOBAL_ERROR_STREAM1("TCPStrategy> TCPStrategy has no SignalReceiver" << endl);
        }
        __send_message_mutex.unlock();
        /* End< */

        /* Send message */
        sendQueuedMessage(tmp_message);
      } else {
        __send_message_mutex.unlock();
      }
      /* End< */
    }
  }

  void TCPStrategy::receiveSignal(Signal signal) {
    switch(signal) {
    case FREE_QUEUE_SLOT:
      __blocked_mutex.lock();
      __blocked = false;
      __blocked_condition.notify();
      __blocked_mutex.unlock();
      break;
    default:
      break;
    }
  }

  void TCPStrategy::sendQueuedMessage(const Message* message) {
    if (!message) {
      return;
    }
    if (getStreamSocket().getSocket() == -1) {
      GLOBAL_WARNING_STREAM1( "TCPStrategy is not connected " << endl );
      message->release();
      return;
    }

    /* Begin> If the buffer stores no data, release it */
    if (message->getType() == Message::BUFFER) {
      if (static_cast<const Buffer*>(message)->getUsed() == 0) {
        GLOBAL_WARNING_STREAM1( "Try to send Buffer without data " << endl);
        message->release();
        return;
      }
    }
    /* End< */
  
    try {
      // TEST: fw: Display buffer header size
//      // Get write counter before writing the buffer
//      size_t write_counter_before_message = __o_net_stream->getWriteCounter();

      if (message->getType() == Message::BUFFER) {
        getSocket().setQoSTag(getInstreamBufferPriority());
        *__o_stream << static_cast<nmm_int32>(BUFFER_ONE_WAY);
        *__o_stream << static_cast<const Buffer*>(message)->getUsed();
      } else {
        getSocket().setQoSTag(getInstreamEventPriority());
        *__o_stream << static_cast<nmm_int32>(CEVENT_ONE_WAY);
      }

      if(message->getDirection() == Message::DOWNSTREAM) {
        GLOBAL_DEBUG_STREAM1( "Send DownStream Message " << endl);
      } else if(message->getDirection() == Message::UPSTREAM) {
        GLOBAL_DEBUG_STREAM1( "Send UpStream Message " << endl );
      } else {
        GLOBAL_WARNING_STREAM1( "Send Message with UNKNOWN DIRECTION !!!!!!!" << endl);
      }

      *__o_stream << *message;

      __o_stream->writeData();

      // TEST: fw: Display buffer header size
//      // Get write counter after writing the buffer
//      size_t write_counter_after_message = __o_net_stream->getWriteCounter();

      // Compute buffer header size
      if (message->getType() == Message::BUFFER) {
        // TEST: fw: Display buffer header size
//        size_t message_size = write_counter_after_message - write_counter_before_message;
//        size_t header_size = message_size - static_cast<const Buffer*>(message)->getUsed();

//        std::cerr << "Buffer header size: " << header_size << std::endl;

        // TEST: fw: Warn if buffer size is too small and/or adapt to header size
        // TEST: fw: Disabled because it requires a dynamic_cast to BufferedONetStream
//        // Check if buffer header size matches hardcoded buffer header size
//        if(header_size > __o_net_stream->getSendBufferSize())
//        {
//          // Warn because this affects performance of first buffer
//          GLOBAL_WARNING_STREAM1("TCPStrategy: Buffer header is bigger than send buffer size of network stream!" << endl);
//
//          // Adapt to header size
//          __o_net_stream->setSendBufferSize(header_size);
//        }
      }
    }
    /* Begin> Error handling */
    catch (const Exception& te) {
      GLOBAL_WARNING_STREAM1(te << endl);
      internalCloseConnection(getStreamSocket().getSocket());
      __is_running = false;
    }
    /* End< */
    /* Release the message */
    message->release();
  }

  void TCPStrategy::conditionWait() {
    __condition_thread_mutex.lock();
    if ( !__send_message && __is_running) {
      __thread_condition.wait();
    }
    __condition_thread_mutex.unlock();
  }

  void TCPStrategy::conditionNotify() {
    __condition_thread_mutex.lock();
    __thread_condition.notify();
    __condition_thread_mutex.unlock();
  }

  void TCPStrategy::scheduleDestruction() {
    SocketHandle socket = getStreamSocket().getSocket();
    __is_running = false;
    __blocked_mutex.lock(); 
    if (__blocked) {
      __blocked = false;
      __blocked_condition.notify();
    }
    __blocked_mutex.unlock();
    disableNotification();
    if(socket != -1) {
      SNetReactor::getInstance().unregisterSocket(socket);
      closeConnection(socket);
    }
    delete this;
  }
  
  void TCPStrategy::enableNoDelay(const nmm_bool& value) {
    getStreamSocket().enableNoDelay(value);
  }

  nmm_bool TCPStrategy::isDelayEnabled() const {
    return getStreamSocket().isDelayEnabled();
  }
  
  void TCPStrategy::setMaxSegment(const nmm_int32& value) {
    getStreamSocket().setMaxSegment(value);
  }
  
  nmm_int32 TCPStrategy::getMaxSegment() const {
    return getStreamSocket().getMaxSegment();
  }
  
  void TCPStrategy::enablePartialFrames(const nmm_bool& value) {
    getStreamSocket().enablePartialFrames(value);
  }
  
  nmm_bool TCPStrategy::isPartialFramesEnabled() {
    return getStreamSocket().isPartialFramesEnabled();
  }
  
  nmm_int32 TCPStrategy::getUnreadQueueCount() const {
    return getStreamSocket().getUnreadQueueCount();
  }
  
  nmm_bool TCPStrategy::allUrgentDataSend() const {
    return getStreamSocket().allUrgentDataSend();
  }
  
  nmm_int32 TCPStrategy::getUnsendQueueCount() const {
    return getStreamSocket().getUnsendQueueCount();
  }
  
  void TCPStrategy::enableKeepAlive(const nmm_bool& value) {
    getStreamSocket().enableKeepAlive(value);
  }
  
  nmm_bool TCPStrategy::getKeepAlive() const {
    return getStreamSocket().getKeepAlive();
  }
  

  void TCPStrategy::flushITSReceiver() {
    /* send FREE_QUEUE_SLOT signals */
    __send_message_mutex.lock();

    if(getSignalReceiver() && __send_message) {
      if(__send_message->getDirection() == Message::DOWNSTREAM) {
        GLOBAL_DEBUG_STREAM1("TCPStrategy> getSignalReceiver()->receiveSignal(FREE_DOWNSTR_QUEUE_SLOT); in TCPStrategy::flush" << endl); 
	getSignalReceiver()->receiveSignal(FREE_DOWNSTR_QUEUE_SLOT);
      } else if(__send_message->getDirection() == Message::UPSTREAM) {
        GLOBAL_DEBUG_STREAM1("TCPStrategy> getSignalReceiver()->receiveSignal(FREE_UPSTR_QUEUE_SLOT); in TCPStrategy::flush" << endl);
        getSignalReceiver()->receiveSignal(FREE_UPSTR_QUEUE_SLOT);
      } else {
        GLOBAL_ERROR_STREAM1("TCPStrategy> Unknown message direction in ~TCPStrategy" << endl);
      }
    }

    if(__send_message) {
      __send_message->release();
      __send_message = 0;
    }
    __send_message_mutex.unlock();  
  }
  
  const char TCPStrategy::__strategy_type[] = "TCPStrategy";

  TTSFactoryRegistrator< TTSFactoryObject<TCPStrategy> >
  __tcp_strategy_prototype("TCPStrategy",
                           TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER |
                           TSDescription::ITS_ONE_WAY_CEVENT_RECEIVER |
                           TSDescription::ITS_ONE_WAY_BUFFER_RECEIVER |
                           TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER |
                           TSDescription::ITS_ONE_WAY_CEVENT_SENDER |
                           TSDescription::ITS_ONE_WAY_BUFFER_SENDER,
                           TSDescription::TS_RELIABLE |
                           TSDescription::TS_RELIABLE_CONNECT |
                           TSDescription::TS_PRESERVE_MESSAGE_HEADERS);
  
}

#include "nmm/nmm_suffix.hpp"
