/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "UDPBroadcaster.hpp"


#include "nmm/utils/thread/MutexGuard.hpp"
#include "nmm/comm/ProcessUserInfo.hpp"
#include "nmm/comm/messaging/IEventReceiver.hpp"
#include "nmm/utils/net/NetReactor.hpp"
#include "nmm/utils/NamedObject.hpp"

#include <cerrno>

#include "nmm/nmm_prefix.hpp"

namespace NMM {  
  UDPBroadcaster::UDPBroadcaster(int sending_port, int receiving_port) 
        : __sending_port (sending_port), __receiving_port(receiving_port), __is_running(true)
  {
    string protocol = "udp";
    __sending_addr.setAddress(INetAddr::getSendingBroadcastAddress()); //set to multicast
    __sending_addr.setTransportProtocol(protocol);
    __sending_addr.setPort(__sending_port);
    
    
    __receiving_addr.setAddress(INetAddr::getReceivingBroadcastAddress()); //read from local address
    __receiving_addr.setTransportProtocol(protocol);
    __receiving_addr.setPort(__receiving_port);

    //connect to this port
    __connector.connect(__sending_socket, __sending_addr, protocol.c_str());
    __pos.setSock(&__sending_socket);
    __movs.setOPartnerStream(&__pos);

    // accept incoming data
    __acceptor.open(__receiving_addr, protocol.c_str());
    __acceptor.accept(__receiving_socket);
    __pis.setDiscardOldPackets(false);
    __pis.setSock(&__receiving_socket);
    __mivs.setIPartnerStream(&__pis);    
    SNetReactor::getInstance().registerSocket( __receiving_socket.getSocket(), this, NetReactor::NR_POLLIN | NetReactor::NR_POLLPRI | NetReactor::NR_POLLERR  );        
  }

  UDPBroadcaster::~UDPBroadcaster() {
    MutexGuard mg_send(__internal_send_mutex);
    MutexGuard mg_recv(__internal_recv_mutex);
    __is_running = false;
    SNetReactor::getInstance().unregisterSocket( __receiving_socket.getSocket());
  }

  Result UDPBroadcaster::sendEvent(const Event& event) {
    MutexGuard mg_send(__internal_send_mutex);

    event.toStream(__movs);
    __movs.writeData();

    return SUCCESS;
  }

  NetHandler::NetResult UDPBroadcaster::receiveData(SocketHandle ) {
    MutexGuard guard(__internal_recv_mutex);

    if (!__is_running) {
      return REMOVE;
    }

    try {
      /* Try to read data from PacketStream */ 
      if ( __pis.receivePacket() ) {
        __mivs.readData();
        /* Try to read an event */
        Event e;
        e.createFromStream(__mivs);
        if (getEventReceiver()) {
          getEventReceiver()->receiveEvent(e, this);
        } else {
          GLOBAL_WARNING_STREAM( "UDPBroadcaster has no event receiver" << endl);
        }
      }
    } catch (const Exception & e ) {
      GLOBAL_ERROR_STREAM(e << endl );
    } catch (const SerializedException & e ) {
      GLOBAL_ERROR_STREAM(e << endl );
    }

    /* register socket again */ 
    if (__is_running) {
      return RECHECK;
    }
    
    return REMOVE;
  }
  
}

#include "nmm/nmm_suffix.hpp"
