/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "UDPStrategy.hpp"

#include "PacketINetStream.hpp"
#include "PacketONetStream.hpp"

#include "nmm/comm/serialize/MagicNumberIValueStream.hpp"
#include "nmm/comm/serialize/MagicNumberOValueStream.hpp"
#include "nmm/comm/serialize/SerializeResult.hpp"

#include "nmm/comm/messaging/CEvent.hpp"
#include "nmm/comm/messaging/Buffer.hpp"
#include "nmm/comm/messaging/IExternalBufferManager.hpp"

#include "nmm/comm/ConnectionManager.hpp"
#include "nmm/comm/ConnectorAcceptor.hpp"
#include "nmm/comm/TTSFactoryRegistrator.hpp"
#include "nmm/comm/TTSFactoryObject.hpp"

#include "nmm/utils/net/SockAcceptor.hpp"
#include "nmm/utils/net/SockConnector.hpp"
#include "nmm/utils/net/NetReactor.hpp"

#include "nmm/utils/thread/MutexGuard.hpp"

#include <cassert>

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  UDPStrategy::UDPStrategy(const string& name)
    : TransportStrategy(name),
      __event_condition(__event_mutex),
      __event_sending_mutex(ThreadMutex::RECURSIVE),
      __current_value(0),
      __current_result(FAILURE),
      __o_stream(0),
      __i_stream(0),
      __o_net_stream(0),
      __i_net_stream(0),
      __use_packet_stream(true),
      __is_running(true),
      __thread_condition(__condition_thread_mutex),
      __send_message(0),
      __send_message_mutex(ThreadMutex::RECURSIVE),
      __local_address(0),
      __remote_address(0),
      __icm(0),
      __remote_ts_ref(0),
      __blocked(false),
      __incoming_message_mutex(),
      __socket_mutex(),
      __net_stream_mutex(),
      __blocked_mutex(),
      __blocked_condition(__blocked_mutex),
      __udp_socket_events(0)
  {
    /* Set the default net stream */
    setINetStream(PacketINetStream());
    setONetStream(PacketONetStream());
    
    /* Set the default stream */
    MagicNumberIValueStream i_stream;
    setIValueStreamStack(&i_stream);
    MagicNumberOValueStream o_stream;
    setOValueStreamStack(&o_stream);
    
    /* activate the thread */
    activateThread();
  }

  UDPStrategy::~UDPStrategy() {
    /* send FREE_QUEUE_SLOT signals */
    if(getSignalReceiver() && __send_message) {
      if(__send_message->getDirection() == Message::DOWNSTREAM) {
        GLOBAL_DEBUG_STREAM1("UDPStrategy> getSignalReceiver()->receiveSignal(FREE_DOWNSTR_QUEUE_SLOT); in ~UDPStrategy" << endl);
        getSignalReceiver()->receiveSignal(FREE_DOWNSTR_QUEUE_SLOT);
      } else if(__send_message->getDirection() == Message::UPSTREAM) {
        GLOBAL_DEBUG_STREAM1("UDPStrategy> getSignalReceiver()->receiveSignal(FREE_UPSTR_QUEUE_SLOT); in ~UDPStrategy" << endl);
        getSignalReceiver()->receiveSignal(FREE_UPSTR_QUEUE_SLOT);
      } else {
        GLOBAL_ERROR_STREAM1("UDPStrategy> Unknown message direction in ~UDPStrategy" << endl);
      }
    }
    if(__send_message) {
      __send_message->release();
    }
    
    setSignalReceiver(0);
    setEventReceiver(0);
    setRequestReplyEventReceiver(0);
    setOneWayCEventReceiver(0);
    setOneWayBufferReceiver(0);
    
    closeConnection(getDatagramSocket().getSocket());

    /* Set the running flag to false */
    __condition_thread_mutex.lock();
    __is_running = false;
    __condition_thread_mutex.unlock();
    
    /* Notify the thread to leave the while loop */
    conditionNotify();

    /* unblock blocked condition */
    __blocked_mutex.lock();
    __blocked_condition.notify();
    __blocked_mutex.unlock();

    /* Deactivate the thread */
    deactivateThread();
    
    /* Destroy value streams */
    if (__i_stream) {
      delete __i_stream;
      __i_stream = 0;
    }
    if (__o_stream) {
      delete __o_stream;
      __o_stream = 0;
    }
  
    /* Destroy net streams */
    MutexGuard g(__net_stream_mutex);
    delete __o_net_stream;
    __o_net_stream = 0;
    delete __i_net_stream;
    __i_net_stream = 0;
  }
  
  Result UDPStrategy::receiveRequestReplyEvent(Event& event) {
    return sendRequest(event, EVENT_REQUEST);
  }
  
  Result UDPStrategy::receiveOneWayCEvent(CEvent* message) {
    return sendOneWayMessage(message, CEVENT_ONE_WAY);
  }
  
  Result UDPStrategy::receiveOneWayBuffer(Buffer* message) {
    return sendOneWayMessage(message, BUFFER_ONE_WAY);
  }
  
  Result UDPStrategy::sendReliableEvent(Event& event) const {
    /* If we have no IConnectionManager, return FAILURE */
    if (!__icm) {
      GLOBAL_ERROR_STREAM1("UDPStrategy> UDP sendReliableEvent failed (No IConnectionManager)" << endl);
      return FAILURE;
    }
    
    /* If we have no partner reference, return FAILURE */
    if (!__remote_ts_ref) {
      GLOBAL_ERROR_STREAM1("UDPStrategy> UDP sendReliableEvent failed (No partner reference)" << endl);
      return FAILURE;
    }
    
    try {
      return __icm->sendReliableEvent(__remote_ts_ref, event);
    } catch(BindingException& e) {
      GLOBAL_ERROR_STREAM1("UDPStrategy> UDP sendReliableEvent failed (" << e << ")" << endl);
      return FAILURE;
    }
  }
  
  Result UDPStrategy::sendReliableConfigEvent(Event& event) const {
    /* If we have no IConnectionManager, return FAILURE */
    if (!__icm) {
      GLOBAL_ERROR_STREAM1("UDPStrategy> UDP sendReliableEvent failed (No IConnectionManager)" << endl);
      return FAILURE;
    }
    
    /* If we have no partner reference, return FAILURE */
    if (!__remote_ts_ref) {
      GLOBAL_ERROR_STREAM1("UDPStrategy> UDP sendReliableEvent failed (No partner reference)" << endl);
      return FAILURE;
    }
    
    try {
      return __icm->sendReliableConfigEvent(__remote_ts_ref, event);
    } catch(BindingException& e) {
      GLOBAL_ERROR_STREAM1("UDPStrategy> UDP sendReliableEvent failed (" << e << ")" << endl);
      return FAILURE;
    }
  }
  
  void UDPStrategy::configure(const Format&)
    throw(const FormatNotSupportedException&)
  {}
  
  Result UDPStrategy::sendRequest(const Value& value, UDPMessage message_type) const {
    /* If the thread is no longer running return FAILURE */
    if (!__is_running) {
      GLOBAL_ERROR_STREAM1("UDPStrategy> UDP sendValue failed (UDPStrategy not running)" << endl);
      return FAILURE;
    }
    
    try {
      /* Lock the __event_mutex to ensure that no other event can be send. Use a MutexGuard to unlock
       * the mutex after method is returned and return value is stored by the sender.*/
      MutexGuard guard(__event_sending_mutex);
      
      /* Store a pointer to the given value. So the internal thread can wirte the response to it*/
      __current_value = const_cast<Value*>(&value);
      
      /* Begin> Serialize the event and send it */
      {
        /* Lock the net stream while we use it */
        MutexGuard g(__net_stream_mutex);
        if(!__o_net_stream) {
          GLOBAL_ERROR_STREAM1("UDPStrategy> No NetStream has been set. Failed to send value.");
          return FAILURE;
        }
        
        /* before writing, lock the condition to ensure that the dispatchResponse method can not
           access the event until we have called wait */
        MutexGuard event_guard(__event_mutex);
        
        /*write the message type to the stream. The receiver needs this information to know
         * which receiveEvent-method must be called; the receiveEvent(const Event&) or the
         * receiveEvent(const CEvent&)
         */
        (*__o_stream) << static_cast<nmm_int32>(message_type);
        
        /* Disable the out_values */
        __current_value->disableMemoryType(Value::OUTVALUE);
        
        /* write the value to the stream */
        (*__o_stream) << value;
        
        /* ensure that all data is written */
        __o_stream->writeData();
      }
      
      /* wait for the response */
      __event_condition.wait();
      
      /* reset the internal pointer */
      __current_value = 0;

      /* return the result */
      return __current_result;
    }
    /* End< */
    
    /* Begin> Error handling */
    catch (const TransmissionException& te) {
      GLOBAL_ERROR_STREAM1(te << endl);
      const_cast<UDPStrategy*>(this)->closeConnection(getDatagramSocket().getSocket());
      throw BindingException(te.getComment());
    } catch(const ConnectionClosed& ce){
      GLOBAL_ERROR_STREAM1(ce << endl);
      const_cast<UDPStrategy*>(this)->closeConnection(getDatagramSocket().getSocket());
      throw BindingException(ce.getComment());
    } catch(const Exception& e){
      GLOBAL_ERROR_STREAM1(e << endl);
      const_cast<UDPStrategy*>(this)->closeConnection(getDatagramSocket().getSocket());
      throw BindingException(e.getComment());
    }
    /* End< */
  }
  
  void UDPStrategy::sendResponse(Result result, const Value* ce, UDPMessage message_type) {
    try {
      /* Begin> Create and serialize the response */
      {
        /* lock the net stream while we use it */
        MutexGuard g(__net_stream_mutex);
        if (!__o_net_stream) {
          GLOBAL_ERROR_STREAM1("UDPStrategy> UDP strategy has no NetStream - sendResponse failed!");
          throw Exception("No NetStream");
        }
        
        /* write the message_type */
        (*__o_stream) << static_cast<nmm_int32>(message_type);
        
        /* write the result */
        (*__o_stream) << static_cast<const Result&>(result);
        
        /* If given value != 0, write the value*/
        if(ce) {
          (*__o_stream) << *ce;
        }
        
        /* call writeData to finish the transmission */
        __o_stream->writeData();
      }
      
      /* and return */
      return;
    }
    /* End< */
    
    /* Begin> Error handling */
    catch (const TransmissionException& te) {
      GLOBAL_ERROR_STREAM1(te << endl);
    } catch(const ConnectionClosed& ce){
      GLOBAL_ERROR_STREAM1(ce << endl);
    } catch (const Exception& e) {
      GLOBAL_ERROR_STREAM1(e << endl);
    }
    closeConnection(getDatagramSocket().getSocket());
    /* End */
  }
  
  Result UDPStrategy::sendOneWayMessage(const Message* message, UDPMessage /*message_type*/) {
    if (!message) {
      return FAILURE;
    }
  
    /* Begin> Store the message pointer */
    __condition_thread_mutex.lock();
    __send_message_mutex.lock();
    
    if(__send_message) {
      GLOBAL_ERROR_STREAM1( "ARGH ERROR IN TCPSTARTEGY __send_message != 0 !!!!!!!!!!!!!" << endl);
      cerr << "ARGH ERROR IN TCPSTARTEGY __send_message != 0 !!!!!!!!!!!!! EVENT? " << (message->getType() == Message::EVENT) << endl;
      return FAILURE;
    } else {
      __send_message = const_cast<Message*>(message);
    }
    
    __send_message_mutex.unlock();
    __condition_thread_mutex.unlock();
    /* End< */
    
    /* Notify the Thread */
    conditionNotify();
    
    /* Return blocked */
    return BLOCKED;
  }
  
  bool UDPStrategy::isDataAvailable() {
    /* Check for additional data which has already been read from the socket */
    if(__is_running) {
      if(!__i_net_stream->isEmpty())  {
        /* Data is already available for reading */
        return true;      }
    }

    /* No data available */
    return false;
  }


  NetHandler::NetResult UDPStrategy::receiveData(SocketHandle /*socket*/) {
    /* If the transport strategy is not running, return remove */
    if (!__is_running ) {
      return REMOVE;
    }
    
    /* Dispatch the message */
    if(__i_net_stream->receivePacket()) {
      dispatchMessage();
    }
    
    /* After this there may be other data*/
    while (__is_running) {
      if ( !__i_net_stream->isEmpty() && !__blocked && __is_running )  {
        if(__i_net_stream->receivePacket()) {
          dispatchMessage();
        }
      } else {
        if (__i_net_stream->isEmpty()  && !__blocked && __is_running ) {
          return RECHECK; //return recheck to look for new data 
        }
        __blocked_mutex.lock();
        if ( __blocked ) {
          __blocked_condition.wait();
        }
        __blocked_mutex.unlock();
      }
    }
    
    return REMOVE;
  }
  
  void UDPStrategy::dispatchMessage() {
    try {
      __i_stream->setIPartnerStream(__i_net_stream);
      __i_stream->readData();
      nmm_int32 tmp_message;
      
      *__i_stream >> tmp_message;
      UDPMessage message_type = static_cast<UDPMessage>(tmp_message);
      switch(message_type) {
      case EVENT_REQUEST:
        dispatchEventRequest(*__i_stream, message_type);
        break;
        
      case EVENT_RESPONSE:
        dispatchEventResponse(*__i_stream);
        break;
        
      case CEVENT_ONE_WAY:
        cerr << "Receive OneWayBuffer " << endl;
        dispatchCEventOneWay(*__i_stream);
        break;
        
      case BUFFER_ONE_WAY:
        cerr << "Receive OneWayBuffer " << endl;
        dispatchBufferOneWay(*__i_stream);
        break;
        
      default:
        GLOBAL_ERROR_STREAM1("UDPStrategy> Unknown Type" << endl); break;
      }
    } catch(const Exception& e) {
      GLOBAL_ERROR_STREAM1( "UDPStrategy dispatch Message " << e << endl);
      __is_running = false;
      closeConnection(getDatagramSocket().getSocket());
      return;
    }
  }
  
  void UDPStrategy::setONetStream(const ONetStream& stream) {
    delete __o_net_stream;
    __o_net_stream = stream.copy();
    __o_net_stream->setSock(&getDatagramSocket());
  }
  
  void UDPStrategy::setINetStream(const INetStream& stream) {
    delete __i_net_stream;
    __i_net_stream = stream.copy();
    __i_net_stream->setSock(&getDatagramSocket());
  }

  void UDPStrategy::setIValueStream(const string& value_stream_type)
    throw(const BindingException&)
  {
    GLOBAL_DEBUG_STREAM2("Set new IValueStream of type " << value_stream_type << endl);
    
    /* no events can be sent */
    MutexGuard event_guard(__event_sending_mutex);
  
    /* no messages can be sent */
    MutexGuard message_guard(__send_message_mutex);
  
    /* get prototype */
    IValueStream* new_prototype = SValueFactory::getInstance().createIValueStream(value_stream_type);
    if(!new_prototype) {
      throw BindingException("Could not change to new send ValueStream " + value_stream_type);
    }
    
    /* set value stream */
    delete __i_stream;
    __i_stream = new_prototype;
    
    /* set partner stream */
    __i_stream->setIPartnerStream(__i_net_stream);
  }
  
  void UDPStrategy::setOValueStream(const string& value_stream_type)
    throw(const BindingException&)
  {
    GLOBAL_DEBUG_STREAM2("Set new OValueStream of type " << value_stream_type << endl);
    
    /* no events can be sent */
    MutexGuard event_guard(__event_sending_mutex);
  
    /* no messages can be sent */
    MutexGuard message_guard(__send_message_mutex);
  
    /* get prototype */
    OValueStream* new_prototype = SValueFactory::getInstance().createOValueStream(value_stream_type);
    if(!new_prototype) {
      throw BindingException("Could not change to new send ValueStream " + value_stream_type);
    }
  
    /* set value stream */
    delete __o_stream;
    __o_stream = new_prototype;
    
    /* set partner stream */
    __o_stream->setOPartnerStream(__o_net_stream);
  }
  
  void UDPStrategy::setIValueStreamStack(const IValueStream* value_stream) {
    /* no events can be sent */
    MutexGuard event_guard(__event_sending_mutex);
  
    /* no messages can be sent */
    MutexGuard message_guard(__send_message_mutex);
  
    /* create a copy of the stream */
    IValueStream* new_prototype = value_stream->copy();
  
    /* set value stream */
    delete __i_stream;
    __i_stream = new_prototype;
    
    /* get the bottom partner stream in the value stream stack */
    IStream* parent = 0;
    IStream* bottom = __i_stream;
    while(bottom->getIPartnerStream()) {
      parent = __i_stream;
      bottom = bottom->getIPartnerStream();
    }
    
    /* if the bottom stream is a NetStream, set it as our net
       stream. otherwise set our NetStream as partner stream of the
       bottom stream. */
    INetStream* net_stream = dynamic_cast<INetStream*>(bottom);
    if(net_stream) {
      /* top-level stream can't be a NetStream */
      assert(parent);
      
      /* the TCPStrategy owns the reference to the net stream */
      setINetStream(*net_stream);
      parent->setIPartnerStream(0);
      parent->setIPartnerStream(__i_net_stream);
    } else {
      bottom->setIPartnerStream(__i_net_stream);
    }
  }

  void UDPStrategy::setOValueStreamStack(const OValueStream* value_stream) {
    /* no events can be sent */
    MutexGuard event_guard(__event_sending_mutex);
  
    /* no messages can be sent */
    MutexGuard message_guard(__send_message_mutex);
  
    /* create a copy of the stream */
    OValueStream* new_prototype = value_stream->copy();
  
    /* set value stream */
    delete __o_stream;
    __o_stream = new_prototype;
    
    /* get the bottom partner stream in the value stream stack */
    OStream* parent = 0;
    OStream* bottom = __o_stream;
    while(bottom->getOPartnerStream()) {
      bottom = bottom->getOPartnerStream();
      parent = __o_stream;
    }
    
    /* if the bottom stream is a NetStream, set it as our net
       stream. otherwise set our NetStream as partner stream of the
       bottom stream. */
    ONetStream* net_stream = dynamic_cast<ONetStream*>(bottom);
    if(net_stream) {
      /* top-level stream can't be a NetStream */
      assert(parent);
      
      /* the TCPStrategy owns the reference to the net stream */
      setONetStream(*net_stream);
      parent->setOPartnerStream(0);
      parent->setOPartnerStream(__o_net_stream);
    } else {
      bottom->setOPartnerStream(__o_net_stream);
    }
  }
  
  nmm_int32 UDPStrategy::getSendingCounter() const {
    return __o_net_stream->getWriteCounter();
  }
  
  nmm_int32 UDPStrategy::getReceivingCounter() const {
    return __i_net_stream->getReadCounter();
  }
  
  void UDPStrategy::resetSendingCounter() {
    __o_net_stream->resetWriteCounter();
  }
  
  void UDPStrategy::resetReceivingCounter() {
    __i_net_stream->resetReadCounter();
  }
  
  void UDPStrategy::dispatchEventResponse(IValueStream& iv) {
    /* Read the result */
    try {
      iv >> static_cast<Result&>(__current_result);

      /* If the event is processed successfully read the value*/
      if (__current_result == SUCCESS) {
        __current_value->resetMemoryType();
        __current_value->disableMemoryType(Value::INVALUE);
        iv >> *__current_value;
        __current_value->resetMemoryType();
      } else {
        /* If the return value is FAILURE read the data if the stream is not empty */
        if (!iv.isEmpty()) {
          __current_value->resetMemoryType();
          __current_value->disableMemoryType(Value::INVALUE);
          iv >> *__current_value;
          __current_value->resetMemoryType();
        }
      }
    }
    /* Begin> Error Handling */
    catch (const ConnectionClosed& ) {
      __is_running = false;
      closeConnection(getDatagramSocket().getSocket());
    }
    catch (const TransmissionException& ) {
      __is_running = false;
      closeConnection(getDatagramSocket().getSocket());
    }
    catch (const Exception& e) {
      Event * tmp_event = dynamic_cast<Event*>(__current_value);
      cerr << "Exception during deserialization of event response (method dsipatchEventResponse())" << endl;
      if (tmp_event) {
        cerr << "Could not deserialize response for Event " << tmp_event->getKey() << endl;
      } else {
        cerr << "Could not deserialize response for other Value"  << endl;
      }
      GLOBAL_ERROR_STREAM1(e << endl);
    }
    /* End< */

    /* Begin> Notify the condtion variable */
    __event_mutex.lock();
    __event_condition.notify();
    __event_mutex.unlock();
    /* End< */
  }


  void UDPStrategy::dispatchEventRequest(IValueStream& iv, UDPMessage message_type) {
    /* Begin> If no receiver available  the Event can not handled and return UNDEF*/
    if(message_type == EVENT_REQUEST && !getRequestReplyEventReceiver()) {
      GLOBAL_WARNING_STREAM1("UDPStrategy> No request-reply event receiver set, sending UNDEF response" << std::endl);
      return sendResponse(UNDEF, 0, EVENT_RESPONSE);
    }
    /* End< */
    
    /* Create the Event*/
    Event* event = SValueFactory::getInstance().createType<Event>(iv.getCurrentTypeInfo());

    /* If the ValueFactory can not create an instance of the object the event can not be
     *  handled. So return UNDEF
     */
    if(!event) {
      GLOBAL_WARNING_STREAM1("UDPStrategy> Failed to create event, sending UNDEF response" << endl);
      return sendResponse(UNDEF, 0, EVENT_RESPONSE);
    }
    
    try {
      /* Otherwise read the event from stream */
      event->createFromStream(iv);
    } catch (const Exception& e) {
      cerr << "Exception during deserialization of event request (method dsipatchEventRequest())" << endl;
      GLOBAL_ERROR_STREAM1( "DispatchEventRequest for event " << event->getKey() << endl << e << endl);
    }

    /* send it to the receiver */
    Result result = FAILURE;
    if(message_type == EVENT_REQUEST) {
      result = getRequestReplyEventReceiver()->receiveRequestReplyEvent(*event);
    } else {
      /* never happens */
      result = FAILURE;
    }
    
    /* disable the in-values */
    event->disableMemoryType(Value::INVALUE);
    
    /* and send the response back to the sender */
    sendResponse(result, event, EVENT_RESPONSE);
    
    /* reset the Memory type befor destruction */
    event->resetMemoryType();
    
    /* delete the event */
    delete event;
    /*delete the out and inoutvalues and the return value*/
  }

  void UDPStrategy::dispatchBufferOneWay(IValueStream& iv) {
    try {
      /* BEGIN> Read size from the stream and allocate the Buffer */
      Buffer* buffer = 0;
      nmm_uint32 size;
      iv >> size;
      if(!getBufferManager()) {
        GLOBAL_ERROR_STREAM1("UDPStrategy> No Buffer Manager set! Discarding buffer!" << std::endl);
        throw Exception("No buffer manager set! No data is read from the network");
      }
      if(getBufferManager()) {
        buffer = getBufferManager()->getNewBuffer(size);
      }
      if(!buffer) {
        GLOBAL_ERROR_STREAM("Failed to allocate Buffer! Discarding buffer!" << endl);
        throw Exception("Could not allocate buffer! No data is read from the network");
      }    
      // XXX: repplix: this should be done in the BufferManager!!
      buffer->setUsed(size);
      /* END< */
    
      /* BEGIN> Deserialize the message */
      try {
        buffer->createFromStream(iv);
      } catch(const Exception& e) {
        cerr << "Exception during deserialization of Buffer (method dispatchBufferOneWay()): " << endl;
        cerr << e << endl;
      }    
      /* END< Deserialize the message */
    
      /* BEGIN> Print some debug messages */
      if(buffer->getDirection() == Message::DOWNSTREAM) {
        GLOBAL_DEBUG_STREAM1("UDPStrategy>  Received Downstream Buffer " << buffer->getTimestamp().stream_counter << endl);
      } else if(buffer->getDirection() == Message::UPSTREAM) {
        GLOBAL_DEBUG_STREAM1("UDPStrategy>  Received Upstream Buffer " << buffer->getTimestamp().stream_counter << endl);
      } else {
        GLOBAL_WARNING_STREAM1("UDPStrategy> Received Buffer with UNKNOWN DIRECTION !!!!!!!" << endl);
      }
      /* END< */
    
      /* BEGIN> Forward the mesage to the receiver */
      // XXX: fw: the handling of BLOCKED here assumes that buffer and event receiver are the same object
      if(getOneWayBufferReceiver()) {
        MutexGuard mg(__blocked_mutex);
        if(getOneWayBufferReceiver()->receiveOneWayBuffer(buffer) == BLOCKED) {
          __blocked = true;
        }
      } else {
        buffer->release();
      }
      /* END< */
    } catch(const Exception& e) {
      GLOBAL_ERROR_STREAM1("UDPStrategy> Exception in UDPStrategy while dispatching one-way Buffer:" << endl);
      GLOBAL_ERROR_STREAM1(e << endl);
      closeConnection(getDatagramSocket().getSocket());
    }
  }

  void UDPStrategy::dispatchCEventOneWay(IValueStream& iv) {
    try {
      /* BEGIN> Create the CEvent */
      CEvent* cevent = SValueFactory::getInstance().createType<CEvent>(iv.getCurrentTypeInfo());
      if(!cevent) {
        cerr << "Failed to allocate CEvent" << endl;
        throw Exception("Could not allocate CEvent! No data is read from the network");
      }
      /* END< */
    
      /* BEGIN> Deserialize the message */
      try {
        cevent->createFromStream(iv);
      } catch(const Exception& e) {
        cerr << "Exception during deserialization of instream CEvent (method dispatchCEventOneWay()): " << endl;
        cerr << e << endl;
      }    
      /* END< Deserialize the message */
      
      /* BEGIN> Print some debug messages */
      if(cevent->getDirection() == Message::DOWNSTREAM) {
        GLOBAL_DEBUG_STREAM1("UDPStrategy>  Received Downstream CEvent " << cevent->getTimestamp().stream_counter << endl);
      } else if(cevent->getDirection() == Message::UPSTREAM) {
        GLOBAL_DEBUG_STREAM1("UDPStrategy>  Received Upstream CEvent " << cevent->getTimestamp().stream_counter << endl);
      } else {
        GLOBAL_WARNING_STREAM1("UDPStrategy> Received CEvent with UNKNOWN DIRECTION !!!!!!!" << endl);
      }
      /* END< */
      
      /* BEGIN> Forward the mesage to the receiver */
      // XXX: fw: the handling of BLOCKED here assumes that buffer and event receiver are the same object
      if(getOneWayCEventReceiver()) {
        MutexGuard mg(__blocked_mutex);
        if(getOneWayCEventReceiver()->receiveOneWayCEvent(cevent) == BLOCKED) {
          __blocked = true;
        }
      } else {
        cevent->release();
      }
      /* END< */
    } catch(const Exception& e) {
      GLOBAL_ERROR_STREAM1("UDPStrategy> Exception in UDPStrategy while dispatching one-way CEvent:" << endl);
      GLOBAL_ERROR_STREAM1(e << endl);
      closeConnection(getDatagramSocket().getSocket());
    }
  }
  
  NetHandler::NetResult UDPStrategy::closeConnection(SocketHandle socket) {
    /* That the running flag to false */
    __condition_thread_mutex.lock();
    __is_running = false;
    __condition_thread_mutex.unlock();

    /* Notify the thread to leave the while loop */
    conditionNotify();
    return internalCloseConnection(socket);
  }
  
  NetHandler::NetResult UDPStrategy::internalCloseConnection(SocketHandle) {
    /* Begin> Reset the Receiver if connection is closed */
    __socket_mutex.lock();
    if (getDatagramSocket().getSocket() == -1) {
      __socket_mutex.unlock();
      return REMOVE;
    }
    
    try {
      getDatagramSocket().close();
    } catch (const Exception& e) {
      GLOBAL_ERROR_STREAM1( e << endl );
    }
    /* End< */
    __socket_mutex.unlock();
    
    /* Begin >Notify a waiting event sender */
    __event_mutex.lock();
    __current_result = FAILURE;
    __event_condition.notify();
    __event_mutex.unlock();
    
    /* End< */
    return REMOVE;
  }
  
  void UDPStrategy::receiveSignal(Signal signal) {
    switch(signal) {
    case FREE_QUEUE_SLOT:
      __blocked_mutex.lock();
      __blocked = false;
      __blocked_condition.notify();
      __blocked_mutex.unlock();
      break;
    default:
      break;
    }
  }
  
  void UDPStrategy::sendQueuedMessage(const Message* message) {
    if(!message) {
      return;
    }
    if(getDatagramSocket().getSocket() == -1) {
      GLOBAL_WARNING_STREAM1("UDPStrategy> Not connected" << endl );
      message->release();
      return;
    }
    
    /* BEGIN> If the buffer stores no data, release it */
    if(message->getType() == Message::BUFFER) {
      if (static_cast<const Buffer*>(message)->getUsed() == 0) {
        GLOBAL_WARNING_STREAM1("UDPStrategy> Trying to send Buffer without data" << endl);
        message->release();
        return;
      }
    }
    /* END< */
    
    /* BEGIN> serialize message */
    __o_stream->setOPartnerStream(__o_net_stream);
    try {
      if (message->getType() == Message::BUFFER) {
        *__o_stream << static_cast<nmm_int32>(BUFFER_ONE_WAY);
        *__o_stream << static_cast<const Buffer*>(message)->getUsed();
      } else {
        *__o_stream << static_cast<nmm_int32>(CEVENT_ONE_WAY);
      }

      if(message->getDirection() == Message::DOWNSTREAM) {
        GLOBAL_DEBUG_STREAM1("UDPStrategy> Send DownStream Message " << endl);
      } else if(message->getDirection() == Message::UPSTREAM) {
        GLOBAL_DEBUG_STREAM1("UDPStrategy> Send UpStream Message " << endl );
      } else {
        GLOBAL_ERROR_STREAM1("UDPStrategy> Send Message with UNKNOWN DIRECTION !!!!!!!" << endl);
      }
      
      *__o_stream << *message;
      
      __o_stream->writeData();
    }
    /* END< */
    
    /* BEGIN> Error handling */
    catch (const Exception& te) {
      GLOBAL_WARNING_STREAM1(te << endl);
      closeConnection(getDatagramSocket().getSocket());
      __is_running = false;
    }
    /* END< */
    
    /* release the message */
    message->release();
  }
  
  void UDPStrategy::conditionWait() {
    __condition_thread_mutex.lock();
    if(!__send_message && __is_running) {
      __thread_condition.wait();
    }
    __condition_thread_mutex.unlock();
  }
  
  void UDPStrategy::conditionNotify() {
    __condition_thread_mutex.lock();
    __thread_condition.notify();
    __condition_thread_mutex.unlock();
  }
  
  void UDPStrategy::scheduleDestruction() {
    SocketHandle socket = getDatagramSocket().getSocket();
    __is_running = false;
    disableNotification();
    SNetReactor::getInstance().unregisterSocket(socket);
    closeConnection(socket);
    delete this;
  }
  
  void UDPStrategy::createLocalSocket() throw (const ConnectionException&)
  {
    try {
      /* Get local host name */
      char buffer[256];
      ::gethostname(buffer, 256);
      string hostname(buffer);
      
      /* Set up local address and find a free port*/
      INetAddr udp_addr;
      udp_addr.setTransportProtocol("udp");
      udp_addr.setPort(__udp_min_port);
      udp_addr.setAddress(hostname.c_str());
      SockAcceptor udp_acceptor;
      udp_acceptor.tryOpen(udp_addr, __udp_min_port, __udp_max_port, "udp");
      udp_acceptor.accept(getDatagramSocket());
      
      INetAddr temp_addr(getDatagramSocket().getLocalAddress());
      GLOBAL_DEBUG_STREAM2("Local address is " << temp_addr << endl);
      
      /* Register socket at NetReactor */
      SNetReactor::getInstance().registerSocket(getDatagramSocket().getSocket(), this);
      __udp_socket_events = NetReactor::NR_POLLIN | NetReactor::NR_POLLERR;
    } catch (const ConnectionException& ce) {
      GLOBAL_ERROR_STREAM1(ce << endl);
      throw ce;
    }
  }
  
  void UDPStrategy::connectToPeer() throw (const ConnectionException&) {
    try {
      /* Set up peer address and connect */
      INetAddr udp_addr;
      udp_addr = __remote_address->getINetAddr();
      udp_addr.setTransportProtocol("udp");
      SockConnector udp_connector;
      udp_connector.connect(getDatagramSocket(), udp_addr, "udp");
      
      GLOBAL_DEBUG_STREAM2("Peer address is " << udp_addr << endl);
    } catch (const ConnectionException& e) {
      GLOBAL_ERROR_STREAM1(e << endl);
      throw e;
    }
  }
  
  void UDPStrategy::setAddress(const TSAddress& addr)
    throw (const BindingException&)
  {
    GLOBAL_DEBUG_STREAM2("UDPStrategy: Set address type=" << addr.getTypeName() << std::endl);
    
    /* create a local socket */
    try {
      createLocalSocket();
    } catch (const Exception& e) {
      throw BindingException(e.getComment());
    }
    
    /* set up the local address */
    // XXX: fw: because we don't know the local object, we can't set the NMM reference properly!
    GLOBAL_DEBUG_STREAM1("UDPStrategy> UDPStrategy: Setup local address" << std::endl);
    delete __local_address;
    __local_address = new UDPAddress(0);
    __local_address->setINetAddr(getDatagramSocket().getLocalAddress());    
    __local_address->setConnectionManager(*SConnectionManager::getInstance().getInterface<IConnectionManager>());
    __local_address->setTSReference(getReference());
    
    /* set strategy description */
    __local_address->setTSDescription(getTSDescription());
    
    /* check if we have a UDPAddress */
    const UDPAddress* udp_address = dynamic_cast<const UDPAddress*>(&addr);
    if(udp_address) {
      /* store this address as remote address */
      delete __remote_address;
      __remote_address = udp_address->copy();
      
      /* connect socket to peer address */
      try {
        connectToPeer();
      } catch (const Exception& e) {
        throw BindingException(e.getComment());
      }
      
      return;
    }
    
    /* check if we have a ConnnectorAddress */
    const ConnectorAddress* connector_address = dynamic_cast<const ConnectorAddress*>(&addr);
    if(connector_address) {      
      /* store the IConnectionManager interface */
      if (!connector_address->getConnectionManager()) {
        /* SHOULD NOT HAPPEN */
        GLOBAL_ERROR_STREAM2("UDPStrategy: ConnectorAddress has no IConnectionManager interface" << endl);
        throw BindingException("Shared address has no IConnectionManager interface");
      }
      delete __icm;
      __icm = dynamic_cast<IConnectionManager*>(connector_address->getConnectionManager()->copy());
      
      /* request a connection */
      ConnectorAddress* remote_addr = NULL;
      SConnectionManager::getInstance().connect(__local_address, remote_addr, __icm);
      
      /* the remote address must be a UDPAddress */
      if(!remote_addr) {
        /* SHOULD NOT HAPPEN */
        GLOBAL_ERROR_STREAM2("UDPStrategy: Could not get remote address" << endl);
        throw BindingException("Could not get remote address");
      }
      const UDPAddress* udp_addr = dynamic_cast<UDPAddress*>(remote_addr);
      if (!udp_addr) {
        /* SHOULD NOT HAPPEN */
        GLOBAL_ERROR_STREAM2("UDPStrategy: Remote address is not a UDPAddress" << endl);
        throw BindingException("Remote address is not a UDPAddress");
      }
      
      /* store the remote address */
      __remote_address = udp_addr->copy();
      
      /* check if the returned NMM reference is valid */
      if (!__remote_address->getTSReference()) {
        GLOBAL_WARNING_STREAM1("UDPStrategy> UDPStrategy: Could not get NMM reference of remote transport strategy" << endl);
      }
      __remote_ts_ref = __remote_address->getTSReference();
      
      /* connect socket to peer address */
      try {
        connectToPeer();
      } catch (const Exception& e) {
        throw BindingException(e.getComment());
      }
      
      return;
    }
    
    throw BindingException(getStrategyType() + " can't handle address of type " + addr.getTypeName());
  }
  
  void UDPStrategy::run() {
    while(__is_running) {
      __blocked_mutex.lock();
      if ( (!__send_message && __is_running) ||
           (!__send_message && __is_running && __blocked)
           )
        {
          __blocked_mutex.unlock();
          /* If nothing to do wait for jobs */
          conditionWait();
        } else {
          __blocked_mutex.unlock();
        }      
      
      /*Begin> Handle a message that should be send */
      __send_message_mutex.lock();
      if ( __send_message ) {
        Message* tmp_message = __send_message;
        /* Begin> reset the __send_message_flag */
        __send_message = 0;
        /* Notify the Parent about FREE_QUEUE_SLOT */
        if (getSignalReceiver()) {
          if(tmp_message->getDirection() == Message::DOWNSTREAM) {
            getSignalReceiver()->receiveSignal(FREE_DOWNSTR_QUEUE_SLOT);
          }
          else if(tmp_message->getDirection() == Message::UPSTREAM) {
            getSignalReceiver()->receiveSignal(FREE_UPSTR_QUEUE_SLOT);
          } else {
            GLOBAL_ERROR_STREAM1("UDPStrategy> Unknown message direction" << endl);
          }
        } else {
          GLOBAL_ERROR_STREAM1("UDPStrategy> UDPStrategy has no SignalReceiver" << endl);
        }
        __send_message_mutex.unlock();
        /* End< */

        /* Send message */
        sendQueuedMessage(tmp_message);
      } else {
        __send_message_mutex.unlock();
      }
      /* End< */
    }
  }
  
  void UDPStrategy::resetStatistics() {
    MutexGuard g(__net_stream_mutex);
    if (!__use_packet_stream) {
      return;
    }
    dynamic_cast<PacketINetStream*>(__i_net_stream)->resetStatistics();
    dynamic_cast<PacketONetStream*>(__o_net_stream)->resetStatistics();
  }
  
  nmm_uint32 UDPStrategy::getTotalSentBytes() const {
    MutexGuard g(__net_stream_mutex);
    if (!__use_packet_stream) {
      /* This is the only information we can get from a net stream */
      return __o_net_stream->getWriteCounter();
    }
    return dynamic_cast<const PacketONetStream*>(__o_net_stream)->getTotalSentBytes();
  }
  
  nmm_uint32 UDPStrategy::getTotalReceivedBytes() const {
    MutexGuard g(__net_stream_mutex);
    if (!__use_packet_stream) {
      return 0;
    }
    return dynamic_cast<const PacketINetStream*>(__i_net_stream)->getTotalReceivedBytes();
  }
  
  nmm_uint32 UDPStrategy::getTotalSentPackets() const {
    MutexGuard g(__net_stream_mutex);
    if (!__use_packet_stream) {
      return 0;
    }
    return dynamic_cast<const PacketONetStream*>(__o_net_stream)->getTotalSentPackets();
  }
  
  nmm_uint32 UDPStrategy::getTotalReceivedPackets() const {
    MutexGuard g(__net_stream_mutex);
    if (!__use_packet_stream) {
      return 0;
    }
    return dynamic_cast<const PacketINetStream*>(__i_net_stream)->getTotalReceivedPackets();
  }
  
  nmm_uint32 UDPStrategy::getTotalDiscardedPackets() const {
    MutexGuard g(__net_stream_mutex);
    if (!__use_packet_stream) {
      return 0;
    }
    return dynamic_cast<const PacketINetStream*>(__i_net_stream)->getTotalDiscardedPackets();
  }
  
  nmm_uint32 UDPStrategy::getTotalLostPackets() const {
    MutexGuard g(__net_stream_mutex);
    if (!__use_packet_stream) {
      return 0;
    }
    return dynamic_cast<const PacketINetStream*>(__i_net_stream)->getTotalLostPackets();
  }
  
  
  const char UDPStrategy::__strategy_type[] = "UDPStrategy";
  
  int UDPStrategy::__udp_min_port = 20000;
  int UDPStrategy::__udp_max_port = 40000;
  
  TTSFactoryRegistrator< TTSFactoryObject<UDPStrategy> >
  __udp_strategy_prototype("UDPStrategy",
                           new ConnectorAcceptor("UDPStrategy"),
                           TSDescription::ITS_REQUEST_REPLY_EVENT_RECEIVER |
                           TSDescription::ITS_ONE_WAY_CEVENT_RECEIVER |
                           TSDescription::ITS_ONE_WAY_BUFFER_RECEIVER |
                           TSDescription::ITS_REQUEST_REPLY_EVENT_SENDER |
                           TSDescription::ITS_ONE_WAY_CEVENT_SENDER |
                           TSDescription::ITS_ONE_WAY_BUFFER_SENDER,
                           TSDescription::TS_PRESERVE_MESSAGE_HEADERS);
  
}

#include "nmm/nmm_suffix.hpp"
