/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "MagicNumberIValueStream.hpp"
#include "MagicNumberManager.hpp"

#include "nmm/Exception.hpp"
#include "nmm/utils/NamedObject.hpp"

#include "TConstructor.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  MagicNumberIValueStream::MagicNumberIValueStream()
    : IValueStream(TypeInfo("MagicNumberIValueStream", "NMM")),
      __current_type_info(0),
      __next_element(0),
      __empty(true),
      __stream_empty(true),
      __magic_number_manager(&SMagicNumberManager::getInstance())
  {}
  
  MagicNumberIValueStream::MagicNumberIValueStream(const MagicNumberIValueStream& stream)
    : IValueStream(stream),
      __current_type_info(0),
      __next_element(0),
      __empty(true),
      __stream_empty(true),
      __magic_number_manager(&SMagicNumberManager::getInstance())
  {}

  MagicNumberIValueStream::~MagicNumberIValueStream() {}

  const TypeInfo& MagicNumberIValueStream::getCurrentTypeInfo() const {
    /* If no TypeInfo is available return a string */
    if (!__current_type_info) {
      throw Exception("Stream is empty");
    }

    return *__current_type_info;
  }

  const string& MagicNumberIValueStream::getCurrentType() const {
    /* If no TypeInfo is available return a string */
    if (!__current_type_info) {
      return __empty_string;
    }

    /* return the current typename */
    return __current_type_info->t_name;
  }

  const string& MagicNumberIValueStream::getCurrentNamespace() const {
    if (!__current_type_info) {
      const_cast<MagicNumberIValueStream*>(this)->nextElement();
    }

    /* If no TypeInfo is available return a string */
    if (!__current_type_info) {
      GLOBAL_ERROR_STREAM("Stream is empty");
      return __empty_string;
    }

    /* return the current namespace */
    return __current_type_info->t_namespace;
  }

  bool MagicNumberIValueStream::isEmpty() const {
    if (__stream_empty)
      return true;

    return __empty;
  }


  void MagicNumberIValueStream::goUp() {
    if (__open_stack.empty() )
      return ;
    /* remove the first type info of latest complex type */
    __open_stack.pop_back();
    __empty = false;
    /* read the TypeInfo of the next element */
    nextElement();
  }


  bool MagicNumberIValueStream::goDown() {
    if ( (__next_element != magic_number[COMPLEX_BEGIN]) && (__next_element < 20)) {
      return false;
    }
    /* push the current TypeInfo to the internal vector */
    __open_stack.push_back(__current_type_info);
    /* read the TypeInfo of the next element */
    nextElement();
    return true;
  }

  bool MagicNumberIValueStream::nextElement() {
    /* If no Partner stream is set, no element can be read */
    if(!getIPartnerStream()) {
      return false;
    }

    /* If the stream is empty return false */
    if (__stream_empty) {
      return false;
    }

    /* Is the next element a NULL-Pointer? */
    bool skip_null = (__next_element == 32);

    /* read the magic number of the next element */
    (*getIPartnerStream()) >> __next_element;
    switch (__next_element) {
      /* For the next element no magic number was available. We must read it. */
    case GenericMagicNumber::COMPLEX_BEGIN:
      __empty = false;
      (*getIPartnerStream()) >> __complex_type_info.t_name;
      (*getIPartnerStream()) >> __complex_type_info.t_namespace;
      __current_type_info = &__complex_type_info;
      break;


      /* A complex struct was closed. */
    case GenericMagicNumber::COMPLEX_END:
      __empty = true;
      if (__current_type_info)
        if (__current_type_info->t_name == "NULL")
          skip_null = true;
      __current_type_info = 0;
      /* If the current Element is a null pointerm, read the next TypeInfo */
      if (skip_null)
        nextElement();
      break;

      /* The Stream is empty */
    case GenericMagicNumber::FINFLAG:
      __stream_empty = true;
      break;

      /* Map the read magic number to a TypeInfo */
    default:
      __empty = false;
      /* is it in our internal map ?*/
      if ( (__next_element >= 0) && ( __next_element <= 18) )
        __current_type_info = magic_number_map[__next_element];
      else
        /* Test if its in the MagicNumberManager */
        __current_type_info = __magic_number_manager->getTypeInfo(__next_element);


      if (__current_type_info == 0) {
        cerr << "Invalid Magic Number " << __next_element << endl; cerr.flush();
        throw Exception("Invalid Magic number");
      }

      break;
    }
    return true;
  }

  void MagicNumberIValueStream::readBuffer(char* buffer, nmm_uint32 size) {
    /* If stream is empty, return */
    if (isEmpty()) {
      cerr << "Stream is Empty for type id : " << "char_buffer" << endl;
      return;
    }

    /* Type check */
    if (__next_element != CHAR_PTR) {
      cerr << "Invalid type conversion from char_buffer to " << getCurrentType()  << endl;
    }

    /* read the buffer from the partner stream */
    if (getIPartnerStream()) {
      getIPartnerStream()->readBuffer(buffer, size);
    }

    /* read the TypeInfo of the next element */
    nextElement();
  }

  void MagicNumberIValueStream::readBuffer(unsigned char* buffer, nmm_uint32 size) {
    /* If stream is empty, return */
    if (isEmpty()) {
      cerr << "Stream is Empty for type id : " << "unsigned_char_buffer" << endl;
      return;
    }

    /* Type check */
    if (__next_element != UNSIGNED_CHAR_PTR) {
      cerr << "Invalid type conversion from char_buffer to " << getCurrentType()  << endl;
    }

    /* read the buffer from the partner stream */
    if (getIPartnerStream()) {
      getIPartnerStream()->readBuffer(buffer, size);
    }

    /* read the TypeInfo of the next element */
    nextElement();
  }

  void MagicNumberIValueStream::readData() {
    if (getIPartnerStream()) {
      getIPartnerStream()->readData();
    }
  
    __empty = false;
    __stream_empty = false;
    nextElement();
  }


  MagicNumberIValueStream& MagicNumberIValueStream::operator>>(nmm_int16&  i) {
    if (getIPartnerStream()) {
      /* Type check */
      if (__next_element != magic_number[INT16]) {
        throw Exception("Invalid Type assignment. Excpected type INT16"  + getCurrentType());
      } else {
        /* read the value */
        (*getIPartnerStream()) >> i;
      }
      nextElement();
    }

    return *this;
  }


  MagicNumberIValueStream& MagicNumberIValueStream::operator>>(nmm_uint16& i) {
    if (getIPartnerStream()) {
      /* Type check */
      if (__next_element != magic_number[U_INT16]) {
        throw Exception("Invalid Type assignment. Excpected type U_INT16"  + getCurrentType());
      } else {
        /* read the value */
        (*getIPartnerStream()) >> i;
      }
      nextElement();
    }

    return *this;
  }


  MagicNumberIValueStream& MagicNumberIValueStream::operator>>(nmm_int32& i) {
    if (getIPartnerStream()) {
      /* Type check */
      if (__next_element != magic_number[INT32] && __next_element != magic_number[INT32]) {
        throw Exception("Invalid Type assignment. Excpected type INT32, actual type "  + getCurrentType());
      } else {
        /* read the value */
        (*getIPartnerStream()) >> i;
      }
      nextElement();
    }
  
    return *this;
  }

  MagicNumberIValueStream& MagicNumberIValueStream::operator>>(nmm_uint32&  i) {
    if (getIPartnerStream()) {
      /* Type check */
      if (__next_element != magic_number[U_INT32] && __next_element != magic_number[U_INT32]) {
        throw Exception("Invalid Type assignment. Excpected type U_INT32, actual type "  + getCurrentType());
      } else {
        /* read the value */
        (*getIPartnerStream()) >> i;
      }
      nextElement();
    }

    return *this;
  }


  MagicNumberIValueStream& MagicNumberIValueStream::operator>>(nmm_int64& i) {
    if (getIPartnerStream()) {
      /* Type check */
      if (__next_element != magic_number[INT64]) {
        throw Exception("Invalid Type assignment. Excpected type INT64, actual type "  + getCurrentType());
      } else {
        /* read the value */
        (*getIPartnerStream()) >> i;
      }
      nextElement();
    }

    return *this;
  }


  MagicNumberIValueStream& MagicNumberIValueStream::operator>>(nmm_uint64& i) {
    if (getIPartnerStream()) {
      /* Type check */
      if (__next_element != magic_number[U_INT64]) {
        throw Exception("Invalid Type assignment. Excpected type u_INT64, actual type "  + getCurrentType());
      } else {
        /* read the value */
        (*getIPartnerStream()) >> i;
      }
      nextElement();
    }

    return *this;
  }



  MagicNumberIValueStream& MagicNumberIValueStream::operator>>(nmm_double& i) {
    if (getIPartnerStream()) {
      /* Type check */
      if (__next_element != magic_number[DOUBLE]) {
        throw Exception("Invalid Type assignment. Excpected type DOUBLE"  + getCurrentType());
      } else {
        /* read the value */
        (*getIPartnerStream()) >> i;
      }
      nextElement();
    }

    return *this;
  }

  MagicNumberIValueStream& MagicNumberIValueStream::operator>>(nmm_float& i) {
    if (getIPartnerStream()) {
      /* Type check */
      if (__next_element != magic_number[FLOAT]) {
        throw Exception("Invalid Type assignment. Excpected type FLOAT. Current type " + getCurrentType());
      } else {
        /* read the value */
        (*getIPartnerStream()) >> i;
      }
      nextElement();
    }

    return *this;
  }

  MagicNumberIValueStream& MagicNumberIValueStream::operator>>(nmm_char& i) {
    if (getIPartnerStream()) {
      /* Type check */
      if (__next_element != magic_number[CHAR]) {
        throw Exception("Invalid Type assignment Excpected type char. Current type"  + getCurrentType());
      } else {
        /* read the value */
        (*getIPartnerStream()) >> i;
      }
      nextElement();
    }

    return *this;
  }


  MagicNumberIValueStream& MagicNumberIValueStream::operator>>(nmm_uchar& i) {
    if (getIPartnerStream()) {
      /* Type check */
      if (__next_element != magic_number[UNSIGNED_CHAR]) {
        throw Exception("Invalid Type assignment Excpected type unsigne char. Current type"  + getCurrentType());
      } else {
        /* read the value */
        (*getIPartnerStream()) >> i;
      }
      nextElement();
    }

    return *this;
  }

  MagicNumberIValueStream& MagicNumberIValueStream::operator>>(string& i) {
    if (isEmpty()) {
      return *this;
    }


    if (getIPartnerStream()) {
      /* Type check */
      if (__next_element != magic_number[STRING]) {
        throw Exception("Invalid Type assignment Excpected type string. Current type"  + getCurrentType());
      } else {
        /* read the value */
        (*getIPartnerStream()) >> i;
      }
    }

    nextElement();
    return *this;
  }

  MagicNumberIValueStream& MagicNumberIValueStream::operator>>(nmm_bool& i) {
    if (getIPartnerStream()) {
      /* Type check */
      if (__next_element != magic_number[BOOL]) {
        throw Exception("Invalid Type assignment Excpected type bool. Current type"  + getCurrentType());
      } else {
        /* read the value */
        (*getIPartnerStream()) >> i;
      }
      nextElement();
    }

    return *this;
  }
  
  const char*  MagicNumberIValueStream::__i_value_stream_type = "MagicNumberIValueStream";
  const string MagicNumberIValueStream::__empty_string;
  
  template<>
  class TConstructor<MagicNumberIValueStream> : public ValueConstructor {
  public:
    TConstructor(const TypeInfo& t_info)
      : ValueConstructor(t_info)
    { }
    virtual MagicNumberIValueStream*          createValue()          const {return new MagicNumberIValueStream();}
    virtual TValue<MagicNumberIValueStream>*  createTValue()         const {return new TValue<MagicNumberIValueStream>();}
  
    virtual TValue<IValueStream*>*  createTPointerValue()  const {return new TValue<IValueStream*>();}
  
    virtual TConstructor<MagicNumberIValueStream>* copy()  const {return new TConstructor(*this); }
  private:
    string  __type_name;
    string  __namespace;
  };
  
  TConstructor<MagicNumberIValueStream> __nmm__magic_number_i_value_stream(TypeInfo("MagicNumberIValueStream", "NMM"));
  
}

#include "nmm/nmm_suffix.hpp"
