/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "MagicNumberOValueStream.hpp"
#include "MagicNumberManager.hpp"

#include "nmm/Exception.hpp"

#include "TConstructor.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  MagicNumberOValueStream::MagicNumberOValueStream()
    : OValueStream(TypeInfo("MagicNumberOValueStream", "NMM")),
      __magic_number_manager(&SMagicNumberManager::getInstance())
  {}

  MagicNumberOValueStream::MagicNumberOValueStream(const MagicNumberOValueStream& stream)
    : OValueStream(stream),
      __magic_number_manager(&SMagicNumberManager::getInstance())
  {}

  MagicNumberOValueStream::~MagicNumberOValueStream() {}

  void MagicNumberOValueStream::beginType(const string& type, const string& ns) {
    int m_number = __magic_number_manager->getMagicNumber(type, ns);
    if (m_number == -1) {
      if (getOPartnerStream()) {
        (*getOPartnerStream()) << magic_number[COMPLEX_BEGIN] << type << ns;
      }
    } else {
      if (getOPartnerStream()) {
        (*getOPartnerStream()) << m_number;
      }
    }
  }


  void MagicNumberOValueStream::beginType(int m_number) {
    if (getOPartnerStream()) {
      (*getOPartnerStream()) << m_number;
    }
  }

  void MagicNumberOValueStream::endType() {
    if (getOPartnerStream()) {
      (*getOPartnerStream()) << magic_number[COMPLEX_END];
    }
  }


  void MagicNumberOValueStream::writeBuffer(const char* buffer, nmm_uint32 size) {
    if (getOPartnerStream()) {
      (*getOPartnerStream()) << magic_number[CHAR_PTR];
      getOPartnerStream()->writeBuffer(buffer, size);
    }
  }

  void MagicNumberOValueStream::writeData()  {
    if (getOPartnerStream()) {
      (*getOPartnerStream()) << magic_number[FINFLAG];
      getOPartnerStream()->writeData();
    }
  }

  void MagicNumberOValueStream::writeBuffer(const unsigned char* buffer, nmm_uint32 size) {
    if (getOPartnerStream()) {
      (*getOPartnerStream()) << magic_number[UNSIGNED_CHAR_PTR];
      getOPartnerStream()->writeBuffer(buffer, size);
    }
  }


  MagicNumberOValueStream& MagicNumberOValueStream::operator<<(const nmm_uint16 &i) {
    if (getOPartnerStream()) {
      (*getOPartnerStream()) << magic_number[U_INT16] << i;
    } else {
      throw Exception("No partner stream available");
    }
    return *this;
  }

  MagicNumberOValueStream& MagicNumberOValueStream::operator<<(const nmm_int16&  i) {
    if (getOPartnerStream()) {
      (*getOPartnerStream()) << magic_number[INT16] << i;
    }
    return *this;
  }

  MagicNumberOValueStream& MagicNumberOValueStream::operator<<(const nmm_uint32& i) {
    if (getOPartnerStream()) {
      (*getOPartnerStream()) << magic_number[U_INT32] << i;
    }
    return *this;
  }

  MagicNumberOValueStream& MagicNumberOValueStream::operator<<(const nmm_int32&  i) {
    if (getOPartnerStream()) {
      (*getOPartnerStream()) << magic_number[INT32] << i;
    }
    return *this;
  }

  MagicNumberOValueStream& MagicNumberOValueStream::operator<<(const nmm_uint64& i) {
    if (getOPartnerStream()) {
      (*getOPartnerStream()) << magic_number[U_INT64] << i;
    }
    return *this;
  }

  MagicNumberOValueStream& MagicNumberOValueStream::operator<<(const nmm_int64& i) {
    if (getOPartnerStream()) {
      (*getOPartnerStream()) << magic_number[INT64] << i;
    }
    return *this;
  }

  MagicNumberOValueStream& MagicNumberOValueStream::operator<<(const nmm_double& i) {
    if (getOPartnerStream()) {
      (*getOPartnerStream()) << magic_number[DOUBLE] << i;
    }
    return *this;
  }

  MagicNumberOValueStream& MagicNumberOValueStream::operator<<(const nmm_float& i) {
    if (getOPartnerStream()) {
      (*getOPartnerStream()) << magic_number[FLOAT] << i;
    }
    return *this;
  }

  MagicNumberOValueStream& MagicNumberOValueStream::operator<<(const nmm_char& i) {
    if (getOPartnerStream()) {
      (*getOPartnerStream()) << magic_number[CHAR] << i;
    }
    return *this;
  }

  MagicNumberOValueStream& MagicNumberOValueStream::operator<<(const nmm_uchar& i) {
    if (getOPartnerStream()) {
      (*getOPartnerStream()) << magic_number[UNSIGNED_CHAR] << i;
    }
    return *this;
  }

  MagicNumberOValueStream& MagicNumberOValueStream::operator<<(const string& i) {
    if (getOPartnerStream()) {
      (*getOPartnerStream()) << magic_number[STRING] << i;
    }
    return *this;
  }

  MagicNumberOValueStream& MagicNumberOValueStream::operator<<(const nmm_bool& i) {
    if (getOPartnerStream()) {
      (*getOPartnerStream()) << magic_number[BOOL] << i;
    }
    return *this;
  }

  const char* MagicNumberOValueStream::__o_value_stream_type = "MagicNumberOValueStream";
  
  
  template<>
  class TConstructor<MagicNumberOValueStream> : public ValueConstructor {
  public:
    TConstructor(const TypeInfo& t_info)
      : ValueConstructor(t_info)
    { }
    virtual MagicNumberOValueStream*          createValue()          const {return new MagicNumberOValueStream();}
    virtual TValue<MagicNumberOValueStream>*  createTValue()         const {return new TValue<MagicNumberOValueStream>();}
  
    virtual TValue<OValueStream*>*  createTPointerValue()  const {return new TValue<OValueStream*>();}
  
    virtual TConstructor<MagicNumberOValueStream>* copy()  const {return new TConstructor(*this); }
  private:
    string  __type_name;
    string  __namespace;
  };
  
  TConstructor<MagicNumberOValueStream> __nmm__magic_number_o_value_stream(TypeInfo("MagicNumberOValueStream", "NMM"));
  
}

#include "nmm/nmm_suffix.hpp"
