/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "SerializedValue.hpp"

#include "IValueStream.hpp"
#include "OValueStream.hpp"
#include "ValueFactory.hpp"

#include "nmm/utils/NamedObject.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  SerializedValue::SerializedValue(bool reconstruction_mode, Value::MemoryMgm memory_mgm)
    : Value(Value::COMPOUND,memory_mgm), __reconstruction_mode(reconstruction_mode)
  {

  }

  SerializedValue::SerializedValue(const string& type_name, const string& ns, bool reconstruction_mode,  Value::MemoryMgm memory_mgm)
    : Value(Value::COMPOUND, memory_mgm), __type_info(type_name, ns), __reconstruction_mode(reconstruction_mode)
  {

  }


  SerializedValue::SerializedValue(const TypeInfo& ti,  bool reconstruction_mode, Value::MemoryMgm memory_mgm)
    : Value(Value::COMPOUND, memory_mgm), __type_info(ti), __reconstruction_mode(reconstruction_mode)
  {

  }

  SerializedValue::SerializedValue(const SerializedValue& sv)
    : Value(sv), __type_info(sv.getTypeInfo()), __reconstruction_mode(sv.__reconstruction_mode)
  {
    //don't copy the vector from sv!
  }


  const SerializedValue& SerializedValue::operator=(const SerializedValue& sv) {
    Value::operator=(sv);
    return assign(sv);
  }

  const SerializedValue& SerializedValue::assign(const SerializedValue& sv) {
    __type_info = sv.getTypeInfo();
    return *this;
  }

  void  SerializedValue::assignFromStream(IValueStream& ivs) {
    if ( ivs.getCurrentTypeInfo() != getTypeInfo() ) {
      GLOBAL_ERROR_STREAM (" SerializedValue > Invalid type or namespace:" << endl);
      GLOBAL_ERROR_STREAM (" SerializedValue > Current type : " << ivs.getCurrentType() << "                Expected type: " << getTypeName() << endl);
      GLOBAL_ERROR_STREAM (" SerializedValue > Current namespace : " << ivs.getCurrentNamespace() << "      Expected namespace: " << getNamespace() << endl);
      ivs.nextElement();
      return;
    }

    if (!ivs.goDown()) {
      GLOBAL_ERROR_STREAM ("SerializedValue seems to be empty");
      ivs.nextElement();
      return;
    }

    unsigned int i = 0;
    while(!ivs.isEmpty()) {
      /** If the value already exist, use this one.
       */
      if (i < __arg_array.size()) {
        if (getArgument(i)) {
          getArgument(i)->assignFromStream(ivs);
        }
      } else {
        if (__reconstruction_mode) {
          Value* value = SValueFactory::getInstance().createValue(ivs.getCurrentTypeInfo());
          value->createFromStream(ivs);
          __arg_array.push_back(value);
        } else {
          GLOBAL_ERROR_STREAM("Runtime error> Stream has unhandled Elements with type name " << ivs.getCurrentType() << endl);
          ivs.nextElement();
        }
      }
      ++i;
    }

    ivs.goUp();

    updateAttributes();
  }

  void  SerializedValue::createFromStream(IValueStream& ivs) {
    if ( ivs.getCurrentTypeInfo() != getTypeInfo() ) {
      GLOBAL_ERROR_STREAM ("SerializedValue > Invalid type or namespace:" << ivs.getCurrentType() << endl);
      GLOBAL_ERROR_STREAM ("SerializedValue > Current type : " << ivs.getCurrentType() << "      Expected type: " << getTypeName() << endl);
      GLOBAL_ERROR_STREAM ("SerializedValue > Current namespace : " << ivs.getCurrentNamespace() << "      Expected namespace: " << getNamespace() << endl);
      ivs.nextElement();
      return;
    }

    if (!ivs.goDown()) {
      GLOBAL_ERROR_STREAM ("SerializedValue seems to be empty");
      ivs.nextElement();
      return;
    }

    unsigned int i = 0;
    while(!ivs.isEmpty()) {
      /** If the value already exist, use this one.
       */
      if (i < __arg_array.size()) {
        if (getArgument(i)) {
          getArgument(i)->createFromStream(ivs);
        }
      } else {
        if (__reconstruction_mode) {
          Value* value = SValueFactory::getInstance().createValue(ivs.getCurrentTypeInfo());
          value->createFromStream(ivs);
          __arg_array.push_back(value);
        } else {
          GLOBAL_ERROR_STREAM("Runtime error> Stream has unhandled Elements with type name " << ivs.getCurrentType() << endl);
          ivs.nextElement();
        }
      }
      ++i;
    }
    ivs.goUp();

    updateAttributes(); 
  }


  void  SerializedValue::toStream(OValueStream& ovs) const {
    /* Call the prepare method */
    const_cast<SerializedValue*>(this)->prepareAttributes();

    /* Start a new type */
    ovs.beginType(getTypeName(), getNamespace());

    /* Begin> Write the registered values to the stream */
    for (size_t i = 0; i< getSize(); ++i)
      if (getArgument(i))
        getArgument(i)->toStream(ovs);
      else {
        GLOBAL_ERROR_STREAM2("Could not serialze 0ptr in SerializedValue " << getTypeName() << endl);
      }
    /* End< */

    /* End the new type */
    ovs.endType();
  }


  void  SerializedValue::releaseValue()           const {
    if (getMemoryStrategy() == INVALUE)
      delete this;
  }
  
}

#include "nmm/nmm_suffix.hpp"
