/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "ValueFactory.hpp"

#include "ValueConstructor.hpp"
#include "ValueStreamConstructor.hpp"
#include "SerializedValue.hpp"
#include "TValue.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  ValueFactory::ValueFactory()
    : NamedObject("ValueFactory")
  {}
  
  ValueFactory::~ValueFactory() {
    __value_map.clear();
  }

  void ValueFactory::registerValue(const ValueConstructor& vc, const TypeInfo& ti) {
    registerValue(vc, ti.t_name, ti.t_namespace);
  }

  void ValueFactory::registerValue(const ValueConstructor& vc, const string& type, const string& ns) {
    MapIt it = __value_map.find(ns);
    /*Begin> Create a new namespace entry if it not exist*/
    if (it == __value_map.end() ) {
      map<string, ValueConstructor*> tmp_map;
      tmp_map.insert(pair<string, ValueConstructor*> (type, const_cast<ValueConstructor*>(&vc)));
      __value_map.insert(pair<string, map<string, ValueConstructor*> >(ns, tmp_map));
      return;
    }
    /*End< */
    else {
      InternalIt i_it = it->second.find(type);
      /*Begin> Add the new type*/
      if (i_it == it->second.end()) {
        it->second.insert(pair<string, ValueConstructor*>(type, const_cast<ValueConstructor*>(&vc)));
      }
      /*End< */
      else {
        WARNING_STREAM2(" ValueFactory > A ValueConstructor for type \"" << type << "\" in namespace " << ns << " is already registered" << endl);
      }
    }
  }

  Value* ValueFactory::createValue(const TypeInfo& ti) {
    return createValue(ti.t_name, ti.t_namespace);
  }

  void ValueFactory::unregisterValue(const TypeInfo& ti) {
    /*Begin> Search for the given namespace */
    MapIt it = __value_map.find(ti.t_namespace);
    if (it == __value_map.end()) {
      return;
    }
    /*End< */

    /*Begin> Search for the constructor */
    InternalIt i_it = it->second.find(ti.t_name);
    if (i_it != it->second.end()) {
      it->second.erase(i_it);
      return;
    }
    /*End< */

  }

  Value* ValueFactory::createValue(const string& type, const string& ns ) {
    /*Begin> Search for a registered constructor */
    ValueConstructor* vc = findValueConstructor(type, ns);
    if (vc)
      return vc->createValue();
    /*End< */
    if (type == "NULL")
      return 0;

    ERROR_STREAM("No type " << type << " in namespace " << ns << " available " << endl);

    /*Begin> If no such Constructor exist create a SequenceValue which can be used as compound object*/
    SerializedValue* tmp_value = new SerializedValue(TypeInfo(type,ns), true);
    WARNING_STREAM("Reconstruct the data structure! A C++ type check will fail" << endl);
    return tmp_value;
    /*End< */
  }

  Value* ValueFactory::createTValue(const TypeInfo& ti) {
    return createTValue(ti.t_name, ti.t_namespace);
  }

  Value* ValueFactory::createTValue(const string& type, const string& ns ) {
    /*Begin> Search for a registered constructor */
    ValueConstructor* vc = findValueConstructor(type, ns);
    if (vc)
      return vc->createTValue();
    /*End< */
    if (type == "NULL")
      return 0;

    ERROR_STREAM("No type " << type << " in namespace " << ns << " available " << endl);
    /*Begin> If no such Constructor exist create a SequenceValue which can be used as compound object*/
    Value* tmp_value = new TValue<SerializedValue>(SerializedValue(TypeInfo(type,ns), true));
    WARNING_STREAM("Reconstruct the data structure! A C++ type check will fail" << endl);

    return tmp_value;
    /*End< */
  }

  Value* ValueFactory::createTPointerValue(const TypeInfo& ti) {
    return createTPointerValue(ti.t_name, ti.t_namespace);
  }


  Value* ValueFactory::createTPointerValue(const string& type, const string& ns ) {
    /*Begin> Search for a registered constructor */
    ValueConstructor* vc = findValueConstructor(type, ns);
    if (vc)
      return vc->createTPointerValue();
    /*End< */
    if (type == "NULL")
      return 0;

    ERROR_STREAM("No type " << type << " in namespace " << ns << " available " << endl);
    /*Begin> If no such Constructor exist create a SequenceValue which can be used as compound object*/
    Value* tmp_value = new TValue<SerializedValue*>(new SerializedValue(TypeInfo(type,ns), true));
    return tmp_value;
    /*End< */
  }


  ValueConstructor* ValueFactory::findValueConstructor(const string& type, const string& ns) {
    /*Begin> Search for the given namespace */
    MapIt it = __value_map.find(ns);
    if (it == __value_map.end()) {
      WARNING_STREAM2("Namespace " << ns << " not available" << endl);
      return 0;
    }
    /*End< */

    /*Begin> Search for the constructor */
    InternalIt i_it = it->second.find(type);
    if (i_it == it->second.end()) {
      return 0;
    }
    /*End< */

    return i_it->second;
  }

  bool ValueFactory::isRegisterd( const TypeInfo& ti ) {
    return isRegisterd(ti.t_name, ti.t_namespace);
  }

  bool ValueFactory::isRegisterd( const string& type, const string& ns ){
    /*Begin> Search for the given namespace */
    MapIt it = __value_map.find(ns);
    if (it == __value_map.end()) {
      return false;
    }
    /*End< */

    /*Begin> Return true if a constructor is registered else false*/
    InternalIt i_it = it->second.find(type);
    if (i_it == it->second.end()) 
      return false;

    return true;
    /*End< */
  }

  IValueStream* ValueFactory::createIValueStream(const string& i_stream) {
    map<string, ValueStreamConstructor*>::iterator it = __i_valuestream_map.find(i_stream);
    if ( it == __i_valuestream_map.end())
      return 0;

    return it->second->createIValueStream();
  }

  OValueStream* ValueFactory::createOValueStream(const string& o_stream) {
    map<string, ValueStreamConstructor*>::iterator it = __o_valuestream_map.find(o_stream);
    if ( it == __o_valuestream_map.end())
      return 0;

    return it->second->createOValueStream();
  }

  string ValueFactory::getIValueStreamPartner(const string& i_stream) const {
    CValueStreamIt it = __i_valuestream_map.find(i_stream);
    if (it == __i_valuestream_map.end())
      return string();

    /* Use the valuestreamconstructor pointer to find the partner */
    for (CValueStreamIt o_it = __o_valuestream_map.begin(); o_it != __o_valuestream_map.end(); ++it)
      if (o_it->second == it->second)
        return o_it->first;

    return string();
  }

  bool ValueFactory::isValueAvailable(const TypeInfo& ti)  const {
    if (const_cast<ValueFactory*>(this)->findValueConstructor(ti.t_name, ti.t_namespace))
      return true;

    return false;
  }

  string ValueFactory::getOValueStreamPartner(const string& i_stream) const {
    CValueStreamIt it = __o_valuestream_map.find(i_stream);
    if (it == __o_valuestream_map.end())
      return string();

    /* Use the valuestreamconstructor pointer to find the partner */
    for (CValueStreamIt o_it = __i_valuestream_map.begin(); o_it != __i_valuestream_map.end(); ++it)
      if (o_it->second == it->second)
        return o_it->first;

    return string();
  }
  
  void ValueFactory::registerValueStream(const string& i_stream, const string& o_stream, const ValueStreamConstructor&vc ) {
    ValueStreamConstructor* tmp = const_cast<ValueStreamConstructor*>(&vc);
    __i_valuestream_map.insert(pair<string,ValueStreamConstructor*>(i_stream, tmp));
    __o_valuestream_map.insert(pair<string,ValueStreamConstructor*>(o_stream, tmp));
  }  
}

DEFINE_SINGLETON(NMM::ValueFactory);

#include "nmm/nmm_suffix.hpp"
