/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "XMLIValueStream.hpp"

#include "nmm/comm/serialize/StdTypeInfo.hpp"
#include "nmm/comm/serialize/TConstructor.hpp"

#include "nmm/utils/xml/XMLNode.hpp"

#include "nmm/utils/NamedObject.hpp"

// XXX: fw: needed for UTF8toisolat1 (should use XMLEncoder instead)
// XXX: fw: whoever fixes this should also remove @XML_CFLAGS@ from
// XXX: fw: Makefile.am
#include <libxml/encoding.h>

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  XMLIValueStream::XMLIValueStream(XMLNode* node)
    : IValueStream(TypeInfo("XMLIValueStream", "NMM"))
  {
    tree.set_root(node);
  }
  
  XMLIValueStream::XMLIValueStream(const XMLIValueStream& stream)
    : IValueStream(stream)
  {
    tree.set_root(0);
  }

  XMLIValueStream::XMLIValueStream(const string& s)
    : IValueStream(TypeInfo("XMLIValueStream", "NMM"))
  {
    tree.read_buffer(s);
    if (tree.root()->name() != "XMLValueStream") {
      GLOBAL_ERROR_STREAM("Wrong StreamType" << endl);
    }

    parent_stack.push_front(tree.root());
    pos_stack.push_front( tree.root()->children().begin() );
  } 

  XMLIValueStream::~XMLIValueStream() {}

  const TypeInfo& XMLIValueStream::getCurrentTypeInfo() const {
    __current_type.t_name = getCurrentType();
    __current_type.t_namespace = getCurrentNamespace();

    return __current_type;
  }

  const string& XMLIValueStream::getCurrentType() const {
    /* If the Stream is empty return an empty string */
    if (isEmpty())
      return __empty_string;
  
    /* Search the property that stores the namespace */
    XMLProperty* property = (**pos_stack.begin())->property("type");
    if ( property )
      return property->value();
    else 
      cerr << "No type property availbale !!!" << endl;
  
    /* If no c_ns entry is available return an empty string */
    return __empty_string;
  }

  const string& XMLIValueStream::getCurrentNamespace() const {
    /* If the Stream is empty return an empty string */
    if (isEmpty())
      return __empty_string;

    /* Search the property that stores the namespace */
    XMLProperty* property = (**pos_stack.begin())->property("c_ns");
    if ( property )
      return property->value();

    /* If no c_ns entry is available return an empty string */
    return __empty_string;
  }

  bool XMLIValueStream::isEmpty() const {
    if ((pos_stack.size() == 0) || (parent_stack.size() == 0))
      return false;

    if (*pos_stack.begin() == (*parent_stack.begin())->children().end())
      return true;

    return false;
  }


  void XMLIValueStream::goUp() {
    /* If the stacks are empty return, because we can not go up */
    if (parent_stack.empty() || pos_stack.empty())
      return ;

    /* pop the first element of the parent stack */
    parent_stack.pop_front();
    /* pop the first element of the position stack */
    pos_stack.pop_front();

    /* read the type information of the next element */
    nextElement();
  }


  bool XMLIValueStream::goDown() {
    if (parent_stack.empty() || pos_stack.empty())
      return false;

    if ( (**pos_stack.begin())->children().empty())
      return false;

    /* Store the current parent in the parent_stack */
    parent_stack.push_front(**pos_stack.begin());

    /* Store the current position in the position_stack */
    pos_stack.push_front((*parent_stack.begin())->children().begin());
    return true;
  }

  bool XMLIValueStream::nextElement() {
    if (isEmpty())
      return false;
    XMLNodeIterator it = *pos_stack.begin();
    ++it;
    pos_stack.pop_front();
    pos_stack.push_front(it);
    return true;
  }

  void XMLIValueStream::readBuffer(char* buffer, nmm_uint32 size) {
    if (isEmpty()) {
      GLOBAL_ERROR_STREAM( "Stream is Empty for type id : " << "char_buffer" << endl);
      return;
    }

    /* Begin> Type checking */
    if (getCurrentType() != "char_buffer") {
      GLOBAL_ERROR_STREAM( "Invalid type conversion from char_buffer to " << getCurrentType()  << endl);
      nextElement();
      return;
    }
    /* End< */

    XMLProperty* prop =  (**pos_stack.begin())->property("Value");
    if (prop) {
      unsigned char* tmp_ptr= reinterpret_cast<unsigned char*>(&prop->value()[0]);
      int new_size = prop->value().size();
      /* convert the buffer from 7 bytes to 8 bytes */
      UTF8Toisolat1(reinterpret_cast<unsigned char*>(buffer), reinterpret_cast<int*>(&size), tmp_ptr, reinterpret_cast<int*>(&new_size));
    }

    nextElement();
  }

  void XMLIValueStream::readBuffer(unsigned char* buffer, nmm_uint32 size) {
    if (isEmpty()) {
      GLOBAL_ERROR_STREAM("Stream is Empty for type id : " << "unsigned_char_buffer" << endl);
      return;
    }

    /* Begin> Type checking */
    if (getCurrentType() != "unsigned_char_buffer") {
      GLOBAL_ERROR_STREAM("Invalid type conversion from char_buffer to " << getCurrentType()  << endl);
      nextElement();
      return;
    }
    /* End< */

    /* Get the property "Value" that stores the data */
    XMLProperty* prop =  (**pos_stack.begin())->property("Value");
    if (prop) {
      unsigned char* tmp_ptr= reinterpret_cast<unsigned char*>(&prop->value()[0]);
      int new_size = prop->value().size();
      /* convert the buffer from 7 bytes to 8 bytes */
      UTF8Toisolat1(buffer, reinterpret_cast<int*>(&size), tmp_ptr, reinterpret_cast<int*>(&new_size));
    }
    nextElement();
  }

  void XMLIValueStream::readData() {
    /* Begin> Read the XML string from the partner stream */
    if (getIPartnerStream()) {
      getIPartnerStream()->readData();
      nmm_int32 size = 0;
      (*getIPartnerStream()) >> size;
      if (size > 0) {
        string s;
        s.resize(size);
        char* buffer = &(s[0]);
        getIPartnerStream()->readBuffer(buffer, size);
        fromString(s);
      }
    }
    /* End< */
  }


  XMLIValueStream& XMLIValueStream::operator>>(nmm_uint16& i) {
    readType<nmm_uint16>(i, nmm_uint16_type_info.t_name);
    return *this;
  }

  XMLIValueStream& XMLIValueStream::operator>>(nmm_int16&  i) {
    readType<nmm_int16>(i, nmm_int16_type_info.t_name);
    return *this;
  }

  XMLIValueStream& XMLIValueStream::operator>>(nmm_uint32& i) {
    readType<nmm_uint32>(i, nmm_uint32_type_info.t_name);
    return *this;
  }

  XMLIValueStream& XMLIValueStream::operator>>(nmm_int32&  i) {
    readType<nmm_int32>(i, nmm_int32_type_info.t_name);
    return *this;
  }

  XMLIValueStream& XMLIValueStream::operator>>(nmm_uint64& i) {
    readType<nmm_uint64>(i, nmm_uint64_type_info.t_name);
    return *this;
  }

  XMLIValueStream& XMLIValueStream::operator>>(nmm_int64& i) {
    readType<nmm_int64>(i, nmm_int64_type_info.t_name);
    return *this;
  }


  XMLIValueStream& XMLIValueStream::operator>>(nmm_double& i) {
    readType<nmm_double>(i, nmm_double_type_info.t_name);
    return *this;
  }

  XMLIValueStream& XMLIValueStream::operator>>(nmm_float& i) {
    readType<nmm_float>(i, nmm_float_type_info.t_name);
    return *this;
  }

  XMLIValueStream& XMLIValueStream::operator>>(nmm_char& i) {
    readType<nmm_char>(i, nmm_char_type_info.t_name);
    return *this;
  }

  XMLIValueStream& XMLIValueStream::operator>>(nmm_uchar& i) {
    readType<nmm_uchar>(i, nmm_uchar_type_info.t_name);
    return *this;
  }

  XMLIValueStream& XMLIValueStream::operator>>(string& i) {
    if (isEmpty()) {
      GLOBAL_ERROR_STREAM("Stream is Empty" << endl);
      return *this;
    }

    if (getCurrentType() != string_type_info.t_name) {
      GLOBAL_ERROR_STREAM("Try type conversion from type " << getCurrentType() << " to " << "string" << endl);
    }

    XMLProperty* prop =  (**pos_stack.begin())->property("Value");
    if (prop) {
      i = prop->value();
    }
    nextElement();
    return *this;
  }

  XMLIValueStream& XMLIValueStream::operator>>(nmm_bool& i) {
    readType<nmm_bool>(i, nmm_bool_type_info.t_name);
    return *this;
  }

  void XMLIValueStream::fromString(const string& data) {
    parent_stack.clear();
    pos_stack.clear();
  
    tree.read_buffer(data);
    if (tree.root()->name() != "XMLValueStream")
      GLOBAL_ERROR_STREAM("Wrong StreamType" << endl);

    parent_stack.push_front(tree.root());
    pos_stack.push_front( tree.root()->children().begin() );
  }
  
  const char*  XMLIValueStream::__i_value_stream_type = "XMLIValueStream";
  const string XMLIValueStream::__empty_string;
  
  template<>
  class TConstructor<XMLIValueStream> : public ValueConstructor {
  public:
    TConstructor(const TypeInfo& t_info)
      : ValueConstructor(t_info)
    { }
    virtual XMLIValueStream*          createValue()          const {return new XMLIValueStream();}
    virtual TValue<XMLIValueStream>*  createTValue()         const {return new TValue<XMLIValueStream>();}
  
    virtual TValue<IValueStream*>*  createTPointerValue()  const {return new TValue<IValueStream*>();}
  
    virtual TConstructor<XMLIValueStream>* copy()  const {return new TConstructor(*this); }
  private:
    string  __type_name;
    string  __namespace;
  };
  
  TConstructor<XMLIValueStream> xml_i_value_stream(TypeInfo("XMLIValueStream", "NMM"));
  
}

#include "nmm/nmm_suffix.hpp"
