/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#ifndef NMM_XML_I_VALUE_STREAM_HPP
#define NMM_XML_I_VALUE_STREAM_HPP NMM_XML_I_VALUE_STREAM_HPP

#include "nmm/nmm_prefix.hpp"

#include <iostream>

#include "IValueStream.hpp"

#include "nmm/utils/xml/XMLTree.hpp"
#include "nmm/utils/xml/XMLNode.hpp"

#include "nmm/nmm_strstream.hpp"
#include <iostream>

#include "visibility.hpp"

namespace NMM {
  
  /** XML Inupt value stream.
   *
   * This class implements an IValueStream, which converts the data
   * from a string that is stored in a XML structure.
   */
  class DLLEXPORT XMLIValueStream
    : public IValueStream
  {
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Constructor
     */
    XMLIValueStream(XMLNode* = 0);

    /** Copy constructor
     */
    XMLIValueStream(const XMLIValueStream&);
  
    /** Constructor that reads the data from given string.
     */
    XMLIValueStream(const string&);

    /** Destructor
     */
    virtual ~XMLIValueStream();

    /** Virtual copy constructor
     */
    XMLIValueStream* copy() const { return new XMLIValueStream(*this); }
    //@}

    /** @name The operators
     */
    //@{
    /** This method reads an u_int16 i from the stream and returns a reference to the stream.
     */
    virtual XMLIValueStream&operator>>(nmm_uint16&)       ;

    /** This method reads an int16 i from the stream and returns a reference to the stream.
     */
    virtual XMLIValueStream&operator>>(nmm_int16&)         ;

    /** This method reads an u_int32 i from the stream and returns a reference to the stream.
     */
    virtual XMLIValueStream&operator>>(nmm_uint32&  )     ;

    /** This method reads an int32 i from the stream and returns a reference to the stream.
     */
    virtual XMLIValueStream&operator>>(nmm_int32&  )       ;

    /** This method reads an u_int64 i from the stream and returns a reference to the stream.
     */
    virtual XMLIValueStream&operator>>(nmm_uint64&  )     ;

    /** This method reads an int64 i from the stream and returns a reference to the stream.
     */
    virtual XMLIValueStream&operator>>(nmm_int64&  )       ;

    /** This method reads a bool i from the stream and returns a reference to the stream.
     */
    virtual XMLIValueStream&operator>>(nmm_bool&  )          ;

    /** This method reads a double i from the stream and returns a reference to the stream.
     */
    virtual XMLIValueStream&operator>>(nmm_double& )         ;

    /** This method reads a float i from the stream and returns a reference to the stream.
     */
    virtual XMLIValueStream&operator>>(nmm_float& )          ;

    /** This method reads a char i from the stream and returns a reference to the stream.
     */
    virtual XMLIValueStream&operator>>(nmm_char& )           ;

    /** This method reads an unsigned char i from the stream and returns a reference to the stream.
     */
    virtual XMLIValueStream&operator>>(nmm_uchar&)   ;

    /** This method reads a string i from the stream and returns a reference to the stream.
     */
    virtual XMLIValueStream&operator>>(string&)          ;
    //@}


    /** This method reads all data need to use this MagicNumber ValueStream
     */
    virtual void readData();
    //@}

    /** @name Read char-array
     */
    //@{
    /** This method reads size chars and store they in the given buffer.
     * \param buffer The buffer to store the read char's. The minimum size of the array must be size
     * \param size The size of char's that should be read
     */
    virtual void readBuffer(char* buffer, nmm_uint32 size);

    /** This method reads size unsigned chars and store they in the given buffer.
     * \param buffer The buffer to store the read unsigned char's. The minimum size of the array must be size
     * \param size The size of char's that should be read
     */
    virtual void readBuffer(unsigned char* buffer, nmm_uint32 size);
    //@}


    /** @name Navigate through the attributes
     */
    //@{
    /** This method accesses the internal attributes of complex structure.
     */
    virtual bool goDown();

    /** This method leaves the internal attributes of complex structure.
     */
    virtual void goUp();

    /** This method returns true if no values are in the ValueStream.
     */
    virtual bool isEmpty()          const;

    /** This method reads all information from the next element
     */
    virtual bool nextElement();
    //@}

    /** @name Navigate through the attributes
     */
    //@{
    /** This method returns the current type information in a TypeInfo struct
     */
    virtual  const TypeInfo& getCurrentTypeInfo() const;

    /** This method returns the type name of the current element. The Stream must not be empty!
     * \sa isEmpty();
     */
    virtual const string& getCurrentType() const;

    /** This method returns the namespace of the current element. The Stream must not be empty!
     * \sa isEmpty();
     */
    virtual const std::string& getCurrentNamespace() const;

    /** This method returns the type name of this IValueStream.
     */
    virtual const char* getIValueStreamType() const { return __i_value_stream_type; }
    //@}

    /** @name XML specific methods
     */
    //@{
    /** This method returns a pointer to the XML-Root node
     */
    const XMLNode* getRootNode()  const {return tree.root();}

    /** This method returns the XMLTree
     */
    inline const XMLTree& getXMLTree()   const {return tree;}

    /** This method returns the XMLTree
     */
    inline       XMLTree& getXMLTree()         {return tree;}

    /** This method reads the data from a given string
     */
    virtual void fromString(const string& data);
    //@}

  protected:
    void setRootNode(XMLNode* node) {tree.set_root(node);}
  private:
    /* Type information of the IValueStream*/
    list<XMLNode*> parent_stack;                        ///< list of nodes that are currently used
    list<XMLNodeIterator> pos_stack;                    ///< current position in the XMLTree

    XMLTree tree;                                       ///< The XML-tree structure

    template <typename D>
    void readType (D& value, const string& type_id);
    mutable TypeInfo __current_type;
  public:
    static const char* __i_value_stream_type;          ///< The type name of this class
    static const string __empty_string;
  };
  
  template <typename D>
  void XMLIValueStream::readType (D& value, const string& type_id) {
    /* Begin> Check if stream is empty */
    if (isEmpty()) {
      cerr << "Stream is Empty for type id : " << type_id << endl;
      return;
    }
    /* End< */

    /* Compate the types */
    if (getCurrentType() != type_id) {
      cerr << "Try type conversion from type " << getCurrentType() << " to " << type_id << endl;
    }

    /* stream to convert the string to a specific type */
    strstream stream;
    XMLProperty* prop =  (**pos_stack.begin())->property("Value");
    if (prop) {
      stream << prop->value();
      /* read the value */
      stream >> value;
    }
    /* read the type information of the next element */
    nextElement();
  }

}

#include "nmm/nmm_suffix.hpp"

#endif
