/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#ifndef NMM_XML_O_VALUE_STREAM_HPP
#define NMM_XML_O_VALUE_STREAM_HPP NMM_XML_O_VALUE_STREAM_HPP

#include "nmm/nmm_prefix.hpp"

#include "OValueStream.hpp"

#include "nmm/utils/xml/XMLTree.hpp"
#include "nmm/utils/xml/XMLNode.hpp"

#include "nmm/nmm_strstream.hpp"

#include "visibility.hpp"

namespace NMM {
  
  /** XML Output value stream.
   *
   * This class implements an OValueStream, which converts the data to
   * string and stores them in a XML structure.
   */
  class DLLEXPORT XMLOValueStream
    : public OValueStream
  {
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Constructor
     */
    XMLOValueStream();

    /** Copy Constructor
     */
    XMLOValueStream(const XMLOValueStream&);

    /** Destructor
     */
    virtual ~XMLOValueStream();

    /** Virtual copy constructor
     */
    XMLOValueStream* copy() const { return new XMLOValueStream(*this); }
    //@}

    /** @name The operators
     */
    //@{

    /** This method writes an u_int16_t i to the stream and returns a reference to the stream.
     */
    virtual XMLOValueStream& operator<<(const nmm_uint16&)     ;

    /** This method writes an int16_t i to the stream and returns a reference to the stream.
     */
    virtual XMLOValueStream& operator<<(const nmm_int16&   )    ;

    /** This method writes an u_int32_t i to the stream and returns a reference to the stream.
     */
    virtual XMLOValueStream& operator<<(const nmm_uint32&)     ;

    /** This method writes an int32_t i to the stream and returns a reference to the stream.
     */
    virtual XMLOValueStream& operator<<(const nmm_int32&   )    ;

    /** This method writes an u_int64_t i to the stream and returns a reference to the stream.
     */
    virtual XMLOValueStream& operator<<(const nmm_uint64 & )    ;

    /** This method writes an int64_t i to the stream and returns a reference to the stream.
     */
    virtual XMLOValueStream& operator<<(const nmm_int64& )      ;

    /** This method writes a bool i to the stream and returns a reference to the stream.
     */
    virtual XMLOValueStream& operator<<(const nmm_bool& )         ;

    /** This method writes a double i to the stream and returns a reference to the stream.
     */
    virtual XMLOValueStream& operator<<(const nmm_double& )       ;

    /** This method writes an float i to the stream and returns a reference to the stream.
     */
    virtual XMLOValueStream& operator<<(const nmm_float& )        ;

    /** This method writes an char i to the stream and returns a reference to the stream.
     */
    virtual XMLOValueStream& operator<<(const nmm_char& )         ;

    /** This method writes an unsigned char i to the stream and returns a reference to the stream.
     */
    virtual XMLOValueStream& operator<<(const nmm_uchar&) ;
    
    /** This method writes a string i to the stream and returns a reference to the stream.
     */
    
    virtual XMLOValueStream& operator<<(const string&)  ;
    /** This method must be called if no other values are written.
     */
    virtual void writeData() ;
    //@}

    /** @name Write char-array
     */
    //@{
    /** This method writes an char buffer of given size to the stream.
     * \param buffer Poitner to the char array.
     * \param size Size of the buffer.
     */
    virtual void writeBuffer(const char* buffer, nmm_uint32 size);

    /** This method writes an unsigned char buffer of given size to the stream.
     * \param buffer Poitner to the unsigned char array.
     * \param size Size of the buffer.
     */
    virtual void writeBuffer(const unsigned char* buffer, nmm_uint32 size);
    //@}

    /** @name Write TypeInfo
     */
    //@{
    /** This method starts a new compound type and tries to map given information to a magic number.
     * If no such magic number exist, the type information are written.
     * \param type The type name of the struct
     * \param ns namespace of the struct
     */
    virtual void beginType(const string& type, const string& ns = "std");

    /** This method set the end flag of a current type and must be called after all attributes of a
     * complex struct are written.
     */
    virtual void endType() ;

    /** This method returns the type name of this ValueStream
     */
    virtual const char* getOValueStreamType() const { return __o_value_stream_type; }
    //@}

    /** @name XML specific methods
     */
    //@{
    /** This method returns the Root Node
     */
    inline  XMLNode* getRootNode()            {return tree.root();}

    /** This method returns the XML tree
     */
    inline  const XMLTree& getXMLTree() const {return tree;}

    /** This method returns the XML tree
     */
    inline        XMLTree& getXMLTree()       {return tree;}

    /** This method converts the XML tree to a string.
     */
    virtual string toString() const;
    //@}

  protected:
    list<XMLNode*> parent_stack;                ///< stack of used xml nodes

    strstream stream;                           ///< stream to convert the data
    XMLTree tree;                               ///< The XML tree

    XMLNode* latest_node;                       ///< The current node
    XMLNodeIterator current_pos;                ///< The current position in the XML-Tree

    template<class T>
    void writeType(const T& value, const string& type);

    template<class T>
    void writeType(const T& value, const string& type, const string& ns);
    
    static string convertToString(const nmm_int32& i);
    inline static string convertToString(const nmm_uint32& i)   {return convertToString(reinterpret_cast<const nmm_int32&>(i));}
    static string convertToString(const nmm_int16& i);
    inline static string convertToString(const nmm_uint16& i)  {return convertToString(reinterpret_cast<const nmm_int16&>(i));}
    inline static string convertToString(const nmm_char* i)           {return string(i);}
    inline static string convertToString(const nmm_uchar* i)  {return string(reinterpret_cast<const nmm_char*>(i));}
    inline static string convertToString(const nmm_double& i)         {strstream is; is << i; string tmp; is >> tmp; return tmp;}
    inline static string convertToString(const nmm_float& i)          {strstream is; is << i; string tmp; is >> tmp; return tmp;}
    inline static string convertToString(const nmm_int64& i)      {strstream is; is << i; string tmp; is >> tmp; return tmp;}
    inline static string convertToString(const nmm_uint64& i)        {strstream is; is << i; string tmp; is >> tmp; return tmp;}
    inline static const string& convertToString(const string& i)  {return i;}
    virtual void writeConvertedType(const string& value, const string& type);
    virtual void writeConvertedType(const string& value, const string& type, const string& ns);
    
  public:
    static const char* __o_value_stream_type; ///< The type name of this class
  };

  template<class T>
  void XMLOValueStream::writeType(const T& value, const string& type) {
    writeConvertedType(convertToString(value),type);
  }
		
  template<class T>
  void XMLOValueStream::writeType(const T& value, const string& type, const string& ns) {
    writeConvertedType(convertToString(value),type,ns);
  }
}

#include "nmm/nmm_suffix.hpp"

#endif
