/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include "GraphURL.hpp"

#include "FileURLObject.hpp"  
#include "HTTPURLObject.hpp"  
#include "AudioCDURLObject.hpp" 
#include "DVDURLObject.hpp" 

#ifdef BUILD_TVINTERFACES
#include "TVURLObject.hpp"
#include "DVBTVURLObject.hpp"
#include "IVTVURLObject.hpp"
#include "MPEGTVURLObject.hpp"
#endif

#include "nmm/Exception.hpp"
#include "nmm/multimedia/NodeDescription.hpp"
#include "nmm/utils/string_utils.hpp"
#include "nmm/comm/NMMDefaultPort.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  GraphURL::GraphURL()
    : url(""),
      is_valid(false),
      protocol(UNKNOWN),
      protocol_name("unknown"),
      user(""),
      password(""),
      host(""),
      port(NMM::default_port),
      uri(""),
      uri_path(""),
      uri_options(""),
      uri_extension("")
  {
    registerGraphURLObjects();
  }

  GraphURL::GraphURL(const string _url)
    : url(_url),
      protocol(UNKNOWN),
      protocol_name("unknown"),
      user(""),
      password(""),
      host(""),
      port(NMM::default_port),
      uri(""),
      uri_path(""),
      uri_options(""),
      uri_extension("")
  {
    registerGraphURLObjects();
    is_valid = parse();
    GLOBAL_DEBUG_STREAM("GraphURL> parsed \"" << *this << "\" from \"" << url << "\", " << (!is_valid ? "in" : "") << "valid!" << endl);
  }

  GraphURL::~GraphURL() 
  {
  }

  void GraphURL::registerGraphURLObject(GraphURLObject* object)
  {
    pair<GraphURLObjectsIt, bool> res =
      url_objects.insert(make_pair(object->getID(), object));
    if (!res.second) {
      GLOBAL_ERROR_STREAM("Could not insert GraphURLObject " << object->getID() << endl);
    }
  }

  void GraphURL::registerGraphURLObjects() 
  {
    registerGraphURLObject(new FileURLObject());  
    registerGraphURLObject(new HTTPURLObject());  
    registerGraphURLObject(new AudioCDURLObject()); 
    registerGraphURLObject(new DVDURLObject()); 
    #ifdef BUILD_TVINTERFACES
    registerGraphURLObject(new TVURLObject());
    registerGraphURLObject(new DVBTVURLObject());
    registerGraphURLObject(new IVTVURLObject());
    registerGraphURLObject(new MPEGTVURLObject());
    #endif
  }

  list<string> GraphURL::getRegisteredGraphURLs() 
  {
    list<string> l;
    for(GraphURLObjectsIt i = url_objects.begin(); i != url_objects.end(); ++i) {
      l.push_back( (*i).first );
    }
    return l;
  }

  GraphURLObject* GraphURL::getGraphURLObject() const
  {
    GraphURLObjectsIt i = url_objects.find(protocol_name);
    if(i == url_objects.end()) {
      return 0;
    }
    else {
      i->second->setGraphURL(this);
      return i->second;
    }
  }

  bool GraphURL::parse() {
    // parse protocol
    const size_t first_colon = url.find(":",0);
    string proto = url.substr(0, first_colon);

    protocol_name = proto;
    if(!getGraphURLObject()) { // url object registered for protocol_name?
      return false;
    }

    if ("http" == proto) {
      protocol = HTTP;
    } else if ("ftp" == proto) {
      protocol = FTP;
    } else if ("smb" == proto) {
      protocol = SMB;
    } else if ("fish" == proto) {
      protocol = FISH;
    } else if ("file" == proto) {
      protocol = FILE;
    } else if ("dvd" == proto) {
      protocol = DVD;
    } else if ("audiocd" == proto) {
      protocol = AUDIOCD;
    } else if ("dvbtv" == proto) {
      protocol = DVBTV;
    } else if ("tv" == proto) {
      protocol = TV;
    } else if ("mpegtv" == proto) {
      protocol = MPEGTV;
    } else if ("ivtv" == proto) {
      protocol = IVTV;
    } else if ("vnc" == proto) {
      protocol = VNC;
    } else {
      protocol = UNKNOWN;
      return false;
    }

    // parse host, uri, uri_path and uri_options

    if (url.npos == first_colon) {
      return false;
    }

    // verify two slash signs after the colon, 'scheme://'
    if (string("//") != url.substr(first_colon + 1, 2)) {
      return false;
    }
    
    const size_t user_begin = first_colon + 3;
    const size_t first_slash = url.find("/", user_begin); // first slash after 'scheme://'
    const size_t first_atsign = url.find("@", user_begin);
    size_t host_begin = user_begin;
    if (first_atsign < first_slash) { // there is a '@' sign in 'scheme://.../', determine user [and password]
      const size_t first_colon = url.find(":", user_begin);
      if (first_colon < first_atsign) {
        user = urldecode(url.substr(user_begin, first_colon - user_begin));
        password = urldecode(url.substr(first_colon+1, first_atsign - (first_colon+1)));
      } else {
        user = urldecode(url.substr(user_begin, first_atsign - user_begin));
        password = string("");
      }
      host_begin = first_atsign+1;
    }
    const size_t first_qmark = url.find("?", host_begin);
    if (first_qmark < first_slash) { // no path given, only options
      host = urldecode(url.substr(host_begin, first_qmark - host_begin));
      uri = url.substr(first_qmark);
      uri_path = "";
      uri_options = uri.substr(1);
    } else if (url.npos != first_slash) { // path given
      host = urldecode(url.substr(host_begin, first_slash - host_begin));
      uri = url.substr(first_slash);
      if (url.npos != first_qmark) { // and options given
        uri_path = urldecode(url.substr(first_slash, first_qmark - first_slash));
        uri_options = url.substr(first_qmark + 1);
      } else { // no options
        uri_path = urldecode(url.substr(first_slash));
        uri_options = string("");
      }
    } else { // neither path, nor options, use rest of url as host, even if empty
      host = urldecode(url.substr(host_begin));
      uri = string("");
      uri_path = string("");
      uri_options = string("");
    }
    const size_t host_colon = host.find(":", 0);
    if (host_colon != string::npos) {
      port = atoi(host.substr(host_colon+1).c_str());
      host = host.substr(0, host_colon);
    }

    // parse options
    if (string("") != uri_options) {
      string parse = uri_options;
      string optstring = string("");
      size_t option_end = 0;
      size_t equalsign = 0;
      while (0 != parse.length()) {
        option_end = parse.find("&");
        optstring = parse.substr(0, option_end);
        if (parse.npos == option_end) {
          parse = "";
        } else {
          parse = parse.substr(option_end + 1);
        }

        equalsign = optstring.find("=");
        if (uri.npos == equalsign) {
          return false;
        }

        pair<URLOptionsIt, bool> res =
              options.insert(make_pair(urldecode(optstring.substr(0, equalsign)), 
              urldecode(optstring.substr(equalsign + 1))));
        if (!res.second) {
          GLOBAL_ERROR_STREAM("Could not insert URL option " << optstring.substr(0, equalsign) << "=" << optstring.substr(equalsign + 1) << endl);
        }
      }
    }

    string::size_type extension_pos = uri_path.rfind(".");
    if(extension_pos != string::npos && extension_pos != uri_path.length() - 1) {
      uri_extension = uri_path.substr(extension_pos+1);
      for (unsigned int i = 0; i < uri_extension.length(); i++) {
          uri_extension[i] = tolower(uri_extension[i]);
      }
    }
    
    return true;
  }

  void GraphURL::configureNodeDescription(NodeDescription &node){
    try {
      getGraphURLObject()->configureNodeDescription(node);
    }
    catch(Exception e) {
      GLOBAL_ERROR_STREAM("" << e << endl);
    }

  }

  void GraphURL::addInterfacesFromProtocol(NodeDescription& n) const
  {
    list<string> interfaces = getGraphURLObject()->getInterfaces();
    for(list<string>::const_iterator i = interfaces.begin(); i != interfaces.end(); ++i) {
      n.addInterface(*i);
    }
  }

  Format* GraphURL::getFormat() const
  {
    // try to get an output format
    Format* output_format = 0;
    output_format = getGraphURLObject()->getFormatFromProtocol();
    if(!output_format) {
      output_format = getGraphURLObject()->getFormatFromExtension();
      if(!output_format) {
        output_format = getGraphURLObject()->getFormatFromContent();
        if(!output_format) {
          output_format = new Format(Format::wild_wild_type);
        }
      }
    }

    return output_format;
  }

  NodeDescription GraphURL::getNodeDescription() const
  {
    string node_name = "";

    NodeDescription n;
    // nodes for reading from URLs are always source nodes
    n.setNodeType(INode::SOURCE);

    // protocol is set, so try to get node name from protocol first
    node_name = getGraphURLObject()->getNodeNameFromProtocol();
    if(!node_name.empty()) {
      n.setNodeName(node_name);
    } 
    else { // then from extension
      node_name = getGraphURLObject()->getNodeNameFromExtension();
      if(!node_name.empty()) {
	n.setNodeName(node_name);
      }
      else {
	node_name = getGraphURLObject()->getNodeNameFromContent();
	if(!node_name.empty()) {
	  n.setNodeName(node_name);
	}
      }
    }

    // get format 
    Format_var output_format(getFormat());
    if(output_format.get()) {
      n.addOutputFormat(*(output_format.get()));
    }

    // add interfaces from url
    addInterfacesFromProtocol(n);

    // add node-events
    getGraphURLObject()->configureNodeDescription(n);

    // finally, set the URL of the node description to the URL of this GraphURL
    n.setURL(url);

    // set the host specified in the URL as the node location
    if(getGraphURLObject()->hasNMMRegistryLocationAndPort()) {
      n.setLocation(getHost());
      n.setPort(getPort());
    }

    return n;
  }

  const set<string> GraphURL::getOptionList() const {
    set<string> retval;

    for (URLOptionsIt i = options.begin(); i != options.end(); i++) {
      retval.insert(i->first);
    }

    return retval;
  }

  const string GraphURL::getOption(const string key) const {
    URLOptionsIt f = options.find(key);
    if (options.end() != f) {
      return f->second;
    }

    return "";
  }
  
  void GraphURL::setOption(const string& key, const string& value) {
    options[key] = value;
  }
  
  void GraphURL::clearOptions() {
    options.clear();
  }

ostream& operator<<(ostream& os, const NMM::GraphURL& url) {
  os << endl << "START --------------------------------" << endl;

  os << "protocol     : " << url.getProtocolName() << endl;
  if (!url.getUser().empty()) {
    os << "user         : " << NMM::urlencode(url.getUser()) << endl;
  }
  if (!url.getPassword().empty()) {
    os << "password     : " << NMM::urlencode(url.getPassword()) << endl;
  }
  os << "host         : " << url.getHost() << endl;
  if (url.getPort()) {
    os << "port         : " << url.getPort() << endl;
  }
  os << "URI path     : " << url.getURIPath() << endl;
  os << "URI extension: " << url.getURIExtension() << endl;
  if (!url.getURIOptions().empty()) {
    os << "URI options  : " << url.getURIOptions() << endl;
  }

  os << endl << "TOGETHER -----------------------------" << endl;

  os << url.getProtocolName();
  os << "://";
  if (!url.getUser().empty()) {
    os << NMM::urlencode(url.getUser());
  }
  if (!url.getPassword().empty()) {
    os << ":";
    os << NMM::urlencode(url.getPassword());
  }
  if (!url.getUser().empty() || !url.getPassword().empty()) {
    os << "@";
  }
  os << url.getHost();
  if (url.getPort()) {
    os << ":";
    os << url.getPort();
  }
  os << url.getURIPath();
  if (!url.getURIOptions().empty()) {
    os << "?" << url.getURIOptions();
  }

 os << endl << "END ----------------------------------" << endl;
  return os;
}

}

#include "nmm/nmm_suffix.hpp"
