/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "TimerEvents.hpp"
#include <iostream>

#include "nmm/comm/messaging/IEventReceiver.hpp"
#include "nmm/utils/thread/ThreadCondition.hpp"
#include "nmm/utils/thread/MutexGuard.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  TimerProducer::TimerProducer(){
    cond = new ThreadCondition(mutex);
    
    new_time = 0;
  }
  
  TimerProducer::~TimerProducer(){
    delete cond;
  }
  
  void TimerProducer::start() {
    MutexGuard guard(mutex);
    isRunning = true;
    activateThread();
  }
  
  void TimerProducer::stop() {
    isRunning = false;
    mutex.lock();
    cond->notify();
    mutex.unlock();
    deactivateThread();
  }
  
  void TimerProducer::registerTimer( int sec, int min, int h, int day, int mon, int year, Event* e ) {
    struct tm abs_time;
    
    abs_time.tm_sec  = sec;
    abs_time.tm_min  = min;
    abs_time.tm_hour = h;
    abs_time.tm_mday = day;
    abs_time.tm_mon  = mon-1;
    abs_time.tm_year = year-1900;
    
    registerTimer( mktime( &abs_time ), e );
  }

  void TimerProducer::registerTimer( time_t t, Event* e ) {
    if ( queue.empty() ) {
      // if queue is empty, set new timer and enqueue event
      te.set( t, e );
      queue.push( te );
      setTimer( t );
    } else {
      // if queue is not empty...
      te = queue.top();
      
      //...check if new timer must be set, if event should be scheduled before the earliest one
      if ( t < te.getTime() ) setTimer( t );
      
      // enqueue event
      te.set( t, e );
      queue.push( te );
    }
  }
  
  void TimerProducer::setTimer( time_t time ) {
    // new_time is used in timerThread to wait
    new_time = time;

    // awake waiting timerThread
    mutex.lock();
    cond->notify();
    mutex.unlock();
  }
  
  void TimerProducer::run(){
    while( isRunning ) {
      mutex.lock();
      // if time to wait is set to zero, wait infinitely
      if ( new_time == 0 ) {
        cond->wait();
        
      }else {
        // wait until timer expired or new timer is set
        if ( cond->timedWait( new_time, 0 ) != SUCCESS ) {
          if ( !queue.empty() ) {
	    // process event in queue
            te = queue.top();
	    queue.pop();
            
            mutex.unlock();
            getEventReceiver()->receiveEvent(*te.getEvent(), this);
            mutex.lock();
            
	    // set new timer
            if ( !queue.empty() ) {
              te = queue.top();
              mutex.unlock();
	      setTimer( te.getTime() );
              mutex.lock();
	    } else {
              mutex.unlock();                          
              setTimer( 0 );
              mutex.lock();
            }
          }
        }
      }
      mutex.unlock();
    } // while running
  }
} // NMM


#include "nmm/nmm_suffix.hpp"
