/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include <iostream>
#include <sstream>
#include "GenericNode.hpp"
#include "QueuedOutputJack.hpp"
#include "nmm/comm/messaging/Buffer.hpp"
#include "nmm/comm/messaging/CEvent.hpp"
#include "nmm/comm/format/Property.hpp"
#include "nmm/comm/ReferenceManager.hpp"
// #include "nmm/base/format/Property.hpp"
#include "ISyncReset.hpp"
#include "nmm/multimedia/interfaces/IMute.hpp"
#include "nmm/utils/thread/MutexGuard.hpp"
#include "nmm/utils/time/TimedElement.hpp"
#include "FlushEvent.hpp"

#include <cassert>

// XXX: fw: windows is evil
#undef DELETE

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  GenericNode::GenericNode(const char* name,
				     const StreamQueue::Mode queue_mode /* = StreamQueue::DEFAULT_MODE */,
				     const int queue_size /* = StreamQueue::DEFAULT_SIZE */)
    : ThreadedNode(name), 
      signal_queue_cv(signal_queue_mutex), 
      internal_started_cv(internal_started_mutex),
      __internal_statechange_mutex(ThreadMutex::RECURSIVE),
      perform_switch(false),
      perform_flush(false),
      perform_flush_ts(),
      switch_enabled(false),
      internal_message_queue_mutex(ThreadMutex::RECURSIVE),
      analysing_cache(NULL),
      analysing_flag(false),
      cached_analysing_flag(false),
      max_analysis_iterations(10),
      analyse_producing_flag(false),
      analyse_processing_flag(false),
      analysis_data_forwarded(false),
      send_data_failed(false),
      current_analysis_out_stream(NULL),
      current_analysis_in_stream(NULL),
      next_analysis_in_stream(input_streams.end()),
      delay_mutex(),
      delay_logging_enabled(false),
      delay_logging_counter(0),
      delay_logging_interval(20),
      max_delay_samples(20),
      in_delay_time(),
      delay_start_enabled(0),
      delay_stop_enabled(0)
  {
    filter_node = this->getInterface<IFilterNode>();

    internal_started_state = false;

    current_event       = 0;

    setWorkingFlag(false);
    setProducingFlag(false);

    this->queue_size = queue_size;
    this->queue_mode = queue_mode;

    downstr_blocked_queue_count = 0;
    output_queue_not_empty = false;

    upstr_blocked_queue_count = 0;
    input_queue_not_empty = false;

    pre_fetch_queue_size = 0;
    pre_fetch_queue_duration = Interval(-1, 0);

    curr_send_out_stream = 0;
    curr_send_in_stream = 0;

    curr_recv_in_stream = input_streams.end();
    curr_recv_out_stream = output_streams.end();

    current_msg_direction = Message::DOWNSTREAM;

    flush_counter = 0;
  }

  GenericNode::~GenericNode()
  {
    // remove inputs
    for (InputStreamMap_t::iterator i = input_streams.begin(); i != input_streams.end(); i++) {
      delete [] i->first; // should be ok, created new char[] in addInputStream // Bug! //Why char* and not string??
      delete i->second;
    }

    // remove outputs
    for (OutputStreamMap_t::iterator i = output_streams.begin();
	 i != output_streams.end(); i++) {
      delete [] i->first;
      delete i->second;
    }
  }

  void GenericNode::setWorkingFlag(const bool w)
  {
    working_flag = w;
    if(working_flag) {
      receiveSignal(WORKING);
    }
  }

  bool GenericNode::getWorkingFlag() const
  {
    return working_flag;
  }

  void GenericNode::setProducingFlag(const bool p)
  {
    producing_flag = p;
    if(producing_flag) {
      receiveSignal(PRODUCING);
    }
  }

  bool GenericNode::getProducingFlag() const
  {
    return producing_flag;
  }

  void GenericNode::block()
  {
    DEBUG_STREAM2("Blocking! ..." << endl);

    /* This mutex is used to avoid deadlocks that are caused between start and block
     * If a (sink) node is started and a sync-reset already waits in the node's queue
     * the sync reset can stop the node, befor internal_started_state is checked from start().
     * The state_mutex must not be used to avoid deadlocks during seek operation.
     */
    MutexGuard internal_mg(__internal_statechange_mutex);

    if(isStarted()) {
      receiveSignal(STOP);
    }
    DEBUG_STREAM2("Blocked!" << endl);
  }

  void GenericNode::unblock()
  {
    DEBUG_STREAM2("Unblocking! ..." << endl);

    /* This mutex is used to avoid deadlocks that are caused between start and block
     * If a (sink) node is started and a sync-reset already waits in the node's queue
     * the sync reset can stop the node, befor internal_started_state is checked from start().
     * The state_mutex must not be used to avoid deadlocks during seek operation.
     */
    MutexGuard internal_mg(__internal_statechange_mutex);

    if(isStarted()) {
      receiveSignal(START);
    }
    DEBUG_STREAM2("Unblocked!" << endl);
  }

  void GenericNode::waitPrefetchFinished()
  {
    DEBUG_STREAM2("waitPrefetchFinished() WARNING: this is a blocking operation which requires all predecessors to be STARTED!" << endl);
    for(InputStreamMap_t::iterator i = input_streams.begin(); i != input_streams.end(); ++i) {
      DEBUG_STREAM2("waitPrefetchFinished() : getPrefetchQueueSize() : " << getPrefetchQueueSize() << endl);
      (*i).second->jack->waitPrefetchFinished();
    }
    DEBUG_STREAM2("waitPrefetchFinished() finished!" << endl);
  }

  void GenericNode::setPrefetchQueueSize(const unsigned int& pre_fetch_queue_size)
  {
    this->pre_fetch_queue_size = pre_fetch_queue_size;
  }

  void GenericNode::setPrefetchQueueDuration(const Interval& pre_fetch_queue_duration)
  {
    this->pre_fetch_queue_duration = pre_fetch_queue_duration;
  }

  nmm_uint32 GenericNode::getPrefetchQueueSize()
  {
    return pre_fetch_queue_size;
  }

  void GenericNode::init()
    throw (const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&)
  {
    MutexGuard mg(state_mutex);
    ThreadedNode::init();
    // remove all input streams with 'needed == false' (needed has been set to false,
    // when calling removeInputStream in doDeinit, and is still 'false' if stream
    // was not re-added during doInit (called be Node::init())
    removeUnneededInputStreams();

    // set pre_fetch_size for all streams
    if(pre_fetch_queue_size > 0) {
      for(InputStreamMap_t::iterator i = input_streams.begin(); i != input_streams.end(); ++i) {
	(*i).second->jack->setPrefetchQueueSize(pre_fetch_queue_size);
      }
    }

    // set pre_fetch_duration for all streams
    if(pre_fetch_queue_duration != Interval(-1, 0)) {
      for(InputStreamMap_t::iterator i = input_streams.begin(); i != input_streams.end(); ++i) {
        (*i).second->jack->setPrefetchQueueDuration(pre_fetch_queue_duration);
      }
    }
  }

  void GenericNode::initOutput()
    throw (const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&)
  {
    MutexGuard mg(state_mutex);

    // Reset delay logging counter
    delay_logging_counter = 0;

    // Reset input delay timestamp
    in_delay_time = Time();

    ThreadedNode::initOutput();
    // remove all output streams with 'needed == false' (needed has been set to false,
    // when calling removeOutputStream in doDeinitOutput, and is still 'false' if stream
    // was not re-added during doInitOutput (called be Node::initOutput())
    removeUnneededOutputStreams();
  }

  void GenericNode::stop()
    throw (const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&)
  {
    DEBUG_STREAM2("Stopping! ..." << endl);

    MutexGuard mg(state_mutex);
    ThreadedNode::stop();
    receiveSignal(STOP);

    internal_started_mutex.lock();
    while(internal_started_state) {
      internal_started_cv.wait();
    }
    internal_started_mutex.unlock();

    DEBUG_STREAM2("Stopped!" << endl);
  }

  void GenericNode::start()
    throw (const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&)
  {
    DEBUG_STREAM2("Starting! ..." << endl);

    // always prefetch
    waitPrefetchFinished();

    // the following two tests are only for debugging. They should
    // never occur normally. However, if one of the occurs, this must
    // be a bug and should be indicated!
    if (analyse_processing_flag) {
      cerr << "Warning: AnalyseProcessingFlag of " << getName() << " still to in start(), disabling!" << endl;
      setAnalyseProcessingFlag(false);
    }

    if (analyse_producing_flag) {
      cerr << "Warning: AnalyseProducingFlag of " << getName() << " still to in start(), disabling!" << endl;
      setAnalyseProducingFlag(false);
    }

    MutexGuard mg(state_mutex);

    /* This mutex is used to avoid deadlocks that are caused between start and block
     * If a (sink) node is started and a sync-reset already waits in the node's queue
     * the sync reset can stop the node, befor internal_started_state is checked from start().
     */
    MutexGuard internal_mg(__internal_statechange_mutex);

    ThreadedNode::start();
    receiveSignal(START);

    internal_started_mutex.lock();
    while(!internal_started_state) {
      internal_started_cv.wait();
    }
    internal_started_mutex.unlock();

    DEBUG_STREAM2("Started!" << endl);
  }

  
  void GenericNode::deinitOutput()
    throw (const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&)
  {
    DEBUG_STREAM2("Deinit Output! ..." << endl);

    MutexGuard mg(state_mutex);

    // attention! here the signal must be sent before the call of ThreadedNode::deactivate
    // since the pthread_join in ThreadedNode::deactivate will block (remember, the thread
    // will finish if the run()-method is finished and this is done in the handling of DEACTIVATE)
    receiveSignal(DEACTIVATE);
    ThreadedNode::deinitOutput();
                
    DEBUG_STREAM2("Deinit Output!" << endl);
  }

  void GenericNode::activate()
    throw (const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&)
  {
    DEBUG_STREAM2("Activating! ..." << endl);

    MutexGuard mg(state_mutex);

    // disable caching for all output jacks
    // must be done here, since in initOutput, streams are not connected
    // yet, hence jack cannot decide to release data
    if (analysing_cache) {
      list<list<AnalysingCacheEntry>::iterator> empty_cache_entries;
      // we iterate over all analysing_cache entries
      for (list<AnalysingCacheEntry>::iterator cache_entry = analysing_cache->begin();
           cache_entry != analysing_cache->end();
           cache_entry++) {
        list<list<JackGroup*>::iterator> not_connected_jacks;
        // we iterate over all output jack (groups) and remember all
        // list enties of JackGroups which are not connected
        for (list<JackGroup*>::iterator jack = cache_entry->jacks.begin();
             jack != cache_entry->jacks.end();
             jack++) {
          if (!(*jack)->isConnected()) {
            not_connected_jacks.push_back(jack);
          }
        }
        // now we can remove the previously remembered (not connected)
        // JackGroups
        for (list<list<JackGroup*>::iterator>::iterator jackit = not_connected_jacks.begin();
             jackit != not_connected_jacks.end();
             jackit++) {
          cache_entry->jacks.erase(*jackit);
        }
        // if there are no more JackGroups to send this cached message
        // through, we remember it for later remove
        if (cache_entry->jacks.empty()) {
          empty_cache_entries.push_back(cache_entry);
        }
      }
      // now we can remove the previously remembered analysing_cache
      // entries (whose Message should me send to an empty list of JackGroups)
      for (list<list<AnalysingCacheEntry>::iterator>::iterator entryit = empty_cache_entries.begin();
           entryit != empty_cache_entries.end();
           entryit++) {
        (*entryit)->message.release();
        analysing_cache->erase(*entryit);
      }
      // if the analysing_cache is all empty, we can remove the entire
      // object as well.
      if (analysing_cache->empty()) {
        delete analysing_cache;
        analysing_cache = NULL;
      }
    }

    ThreadedNode::activate();

    DEBUG_STREAM2("Activated!" << endl);
  }

  Result GenericNode::setFormat(const Format& format)
  {
    DEBUG_STREAM("GenericNode::setFormat" << endl);

    if(getState() < INode::ACTIVATED) {
      DEBUG_STREAM("GenericNode::setFormat : node state is < ACTIVATED => do not handle setFormat and delete it" << endl);
      return DELETE;
    }

    // put input format in working property
    if(getCurrentMsgDirection() == Message::DOWNSTREAM) {
      DEBUG_STREAM("GenericNode::setFormat DOWNSTREAM" << endl);
      string tag = getCurrentRecvInputStream();

      // getInputJack will automatically set the input format in the generic layer
      if(getInputJack(&format, tag.c_str()) != 0) {
	DEBUG_STREAM("GenericNode::setFormat gotInputJack" << endl);
	// create copy and set tag according to current input stream
	Format* i_format = new Format(format);
	i_format->setJackTag(tag);

	doStop();
	doFlush();
	doDeactivate();
	doDeinitOutput();

	// as in Node::deinitOutput : after calling doDeinitOutput, erase working properties
	// (maybe plug-in nodes do the same in doInitOutput(), but this won't hurt
	getOwnWorkingInputProperty()->erase();
	getOwnWorkingOutputProperty()->erase();

	doInitOutput();

	// remove all output streams with 'needed == false' (needed has been set to false,
	// when calling removeOutputStream in doDeinitOutput, and is still 'false' if stream
	// was not re-added during doInitOutput (called be Node::initOutput())
	removeUnneededOutputStreams();

	if(getType() != INode::SINK) {
	  list<string> out_tags = getOutputStreamTags();
	  for(list<string>::const_iterator i = out_tags.begin(); i != out_tags.end(); ++i) {

	    // new output format is 0 in beginning
	    Format* o_format = 0;

	    // remember old output format for current jack
	    const Format* old_o_format = 0;
	    if(getOutputFormat((*i).c_str())) {
	      old_o_format = getOutputFormat((*i).c_str());
	    }
	    if(old_o_format != 0) {

	      // first, try to get intersection of new default out format and old out format
	      const Format* def_o_format = getOwnWorkingOutputProperty()->getDefaultFormat(*i);
	      if(def_o_format) {
		o_format = def_o_format->createIntersectionFormat(old_o_format);
		if(o_format && !(o_format->hasFormatWildcards()) && !(o_format->hasFormatRanges())) {
		  DEBUG_STREAM("GenericNode::setFormat intersec with def out successful " << (*o_format) << endl);
		}
		else {
		  delete o_format;
		  o_format = 0;
		}
	      }

	      // then, if not successful, try to create intersection with all other out formats of new property
	      if(o_format == 0) {
		const_format_list possible_o_formats
		  = getOwnWorkingOutputProperty()->createIntersectionFormats(old_o_format, (*i), false);

		DEBUG_STREAM("GenericNode::setFormat possible_o_formats " << possible_o_formats.size() << endl);
		if(possible_o_formats.size() > 0) {
		  possible_o_formats.sort();
		  for(const_format_list::reverse_iterator piter = possible_o_formats.rbegin();
		      piter != possible_o_formats.rend();
		      --piter)
                  {
		    if( !((*piter)->hasFormatWildcards()) && !((*piter)->hasFormatRanges()) ) {
		      o_format = new Format(**piter);
		      DEBUG_STREAM("GenericNode::setFormat intersec with other out formats successful" << (*o_format) << endl);
		      break;
		    }
		  }
		  for(const_format_list::const_iterator diter = possible_o_formats.begin(); diter != possible_o_formats.end(); ++diter) {
		    delete (*diter);
		  }
		}
	      }

	      // if, still, no output format was found, try to find output format with the least number of different parameter values
	      if(o_format == 0) {
		unsigned int best_num_matches = 0;
		unsigned int num_matches  = 0;
		Property* prop = getOwnWorkingOutputProperty(); 
    		for (format_list::const_iterator it = prop->getFormatList().begin(); 
		     it != prop->getFormatList().end(); 
		     ++it)
                {
		  DEBUG_STREAM("GenericNode::setFormat best match??? " << (**it) << endl);
		  
		  if ((*it)->getJackTag() == (*i)) {
		    Format* tmp = (*it)->createBestIntersectionFormat(old_o_format, num_matches);
		    if(tmp && (num_matches > best_num_matches)) {
		      DEBUG_STREAM("GenericNode::setFormat best match -> matches: " << num_matches << ": " << (*tmp) << endl);
		      best_num_matches = num_matches;
		      o_format = tmp;
		    }
		    else {
		      delete tmp;
		    }
		  }
		}
		// no match -> take default format for jack tag
		if(best_num_matches == 0) {
		  o_format = new Format(*getOwnWorkingOutputProperty()->getDefaultFormat(*i));
		}
		DEBUG_STREAM("GenericNode::setFormat new out format as 'best match' " << (*o_format) << endl);
	      }
	    }
	    else { // (old_o_format == 0) -> no old out format (!) -> take new default out format
	      o_format = new Format(*getOwnWorkingOutputProperty()->getDefaultFormat(*i));
	      DEBUG_STREAM("GenericNode::setFormat OOPS no old out format -> taking new default out format" << (*o_format) << endl);
	    }

	    // automatically internally sets out format
	    getOutputJack(o_format, (*i).c_str());
	    // set output stream according to jack tag
	    setCurrentSendOutputStream((*i).c_str());
	    DEBUG_STREAM("GenericNode::setFormat send output format to jack " << (*i) << " is " << *o_format << endl);
	    sendMessage(  new CEvent(IFormatChange::create_setFormat(*o_format)) );
	  } // end: for all out_tags
	} // end: if(getType() != INode::SINK)
	doActivate();
	doSetFormat(format); 	// call do-method here ?
	doStart();

	// non-sinks nodes have generated new events, sink nodes not 
	if(getType() != INode::SINK) {
	  return DELETE;
	}
	else {
	  return SUCCESS;
	}
      } // end: if(getInputJack(&format, tag.c_str()) != 0) 
      else { 
	ERROR_STREAM("Unsupported input format for jack " << tag << " for setFormat!" << endl);
	return FAILURE;
      }
    }
    else {
      ERROR_STREAM("Upstream setFormat not implemented!" << endl);
      return FAILURE;
    }
  }

  void GenericNode::receiveSignal(Signal event)
  {
    signal_queue_mutex.lock();
    signal_queue.push(event);
    signal_queue_cv.notifyAll();
    signal_queue_mutex.unlock();
  }

  bool GenericNode::isSignalQueueEmpty()
  {
    signal_queue_mutex.lock();
    bool e = signal_queue.empty();
    signal_queue_mutex.unlock();
    return e;
  }

  int GenericNode::dequeueSignal()
  {
    signal_queue_mutex.lock();
    assert(signal_queue.empty() == false);
    int event = signal_queue.front();
    signal_queue.pop();
    signal_queue_mutex.unlock();
    return event;
  }

  Result GenericNode::waitForSignal()
  {
    signal_queue_mutex.lock();
    while(signal_queue.empty()) {
      signal_queue_cv.wait();
    }
    signal_queue_mutex.unlock();
    return SUCCESS;
  }

  void GenericNode::insertEvent(const Event& e) // FIXME: This implementation is broken since event_working_index is a dummy
  {
    if (current_event == 0) {
      ERROR_STREAM2("ARGH. THIS SHOULD NOT HAPPEN, CEVENT is 0" << endl);
      return;
    }
    current_event->insertEvent(event_working_index + 1, e);
    ++event_working_index;
  }

  void GenericNode::insertEvent(Event* e)  // FIXME: This implementation is broken since event_working_index is a dummy
  {
    if (current_event == 0) {
      ERROR_STREAM2("ARGH. THIS SHOULD NOT HAPPEN, CEVENT is 0" << endl);
      return;
    }
    current_event->insertEvent(event_working_index +1, e);
    ++event_working_index;
  }

  Result GenericNode::removeInputStream(const char *jack_tag)
  {
    InputStreamMap_t::iterator i = input_streams.find(jack_tag);
    if (i == input_streams.end()) {
      WARNING_STREAM2("No matching jack found to remove" << endl);
      return FAILURE;
    }
    if (i->second->needed == false) {
      WARNING_STREAM2("No matching jack found to remove" << endl);
      return FAILURE;
    }
    else {
      DEBUG_STREAM2("Set needed = false for input stream " << jack_tag << endl);
      i->second->needed = false;
      if(curr_send_in_stream ==  i->second->jack) {
	curr_send_in_stream = 0;
      }
      if(curr_recv_in_stream ==  i) {
	curr_recv_in_stream = input_streams.begin();
      }
      return SUCCESS;
    }
  }

  void GenericNode::removeUnneededInputStreams()
  {
    for(InputStreamMap_t::iterator i = input_streams.begin(); i != input_streams.end(); ++i) {
      if(i->second->needed == false) {
	DEBUG_STREAM2("Remove unneeded input stream named " << i->first << endl);
	delete [] i->first;
	delete i->second;
	input_streams.erase(i);
	if(curr_send_in_stream ==  i->second->jack) {
	  curr_send_in_stream = 0;
	}
	if(curr_recv_in_stream ==  i) {
	  curr_recv_in_stream = input_streams.begin();
	}
      }
    }
  }

  Result GenericNode::flush()
  {
    MutexGuard mg(state_mutex);

    // called out-of-band
    if(current_event == 0) { 
      InputStreamMap_t::iterator i;
      for (i = input_streams.begin(); i != input_streams.end(); i++) {
	i->second->jack->flush();
      }
      OutputStreamMap_t::iterator j;
      for (j = output_streams.begin(); j != output_streams.end(); j++) {
	j->second->jack_group->flush();
      }
      return ThreadedNode::flush();
    }
    else { // called instream
      Result r = ThreadedNode::flush();
      if(r == DELETE) {
	--flush_counter;
      }
      else {
	++flush_counter;
      }
      return r;
    }
  }

  Result GenericNode::addInputStream(const char *jack_tag)
  {
    InputStreamMap_t::iterator i = input_streams.find(jack_tag);
    if (i != input_streams.end()) {
      if(i->second->needed == false) {
	i->second->needed = true;
	DEBUG_STREAM2("Input stream named " << jack_tag << " reactivated." << endl);
	// update curr_recv_in_stream
	curr_recv_in_stream = input_streams.begin();
	curr_send_in_stream = 0;
	return SUCCESS;
      }
      else {
	DEBUG_STREAM2("Input stream named " << jack_tag << " was already added before." << endl);
	return FAILURE;
      }
    }
    else {
      string jack_name = jack_tag;
      jack_name = getName() + "> InputJack> " + jack_name;

      InputStream_t *in = new InputStream_t();
      // fill struct with values
      in->jack = new QueuedInputJack(this, jack_tag, jack_name, queue_mode, queue_size);

      // create a copy of the jack_tag
      char * my_jack_tag = new char[strlen(jack_tag) + 1];
      strcpy(my_jack_tag, jack_tag);

      pair<InputStreamMap_t::iterator, bool> res 
	= input_streams.insert(make_pair(my_jack_tag, in));
      // could not insert object, then delete it
      if (!res.second) {
	delete [] my_jack_tag;
	delete in;
	return FAILURE;
      }
      DEBUG_STREAM2("Input stream named " << jack_tag << " successfully added." << endl);
      // update curr_recv_in_stream
      curr_recv_in_stream = input_streams.begin();
      curr_send_in_stream = 0;
      return SUCCESS;
    }
  }

  list<string> GenericNode::getInputStreamTags() const
  {
    list<string> tmp_list;
    for (InputStreamMap_t::const_iterator i = input_streams.begin();
         i != input_streams.end();
         i++) {
      if(i->second->needed == true) {
	tmp_list.push_back(string(i->first));
      }
    }
    return tmp_list;
  }

  const char *GenericNode::getCurrentRecvInputStream() const
  {
    return curr_recv_in_stream->first;
  }

  const char *GenericNode::getCurrentRecvOutputStream() const
  {
    return curr_recv_out_stream->first;
  }

  Result GenericNode::setRecvInputStreamEnabled(const char *jack_tag, bool enable)
  {
    InputStreamMap_t::iterator i = input_streams.find(jack_tag);
    if((i != input_streams.end()) && (i->second->needed == true)) {
      i->second->enabled = enable;
      return SUCCESS;
    }
    else {
      WARNING_STREAM2("Could not enable or disable jack " << jack_tag << endl);
      return FAILURE;
    }
  }

  bool GenericNode::getRecvInputStreamEnabled(const char *jack_tag) const
  {
    InputStreamMap_t::iterator i = input_streams.find(jack_tag);
    if((i != input_streams.end()) && (i->second->needed == true)) {
      return i->second->enabled;
    }
    else {
      WARNING_STREAM2("Could not find input jack " << jack_tag << endl);
      return false;
    }
  }

  Result GenericNode::setRecvOutputStreamEnabled(const char *jack_tag, bool enable)
  {
    OutputStreamMap_t::iterator i = output_streams.find(jack_tag);
    if((i != output_streams.end()) && (i->second->needed == true)) {
      i->second->enabled = enable;
      return SUCCESS;
    }
    else {
      WARNING_STREAM2("Could not enable or disable jack " << jack_tag << endl);
      return FAILURE;
    }
  }

  bool GenericNode::getRecvOutputStreamEnabled(const char *jack_tag) const
  {
    OutputStreamMap_t::iterator i = output_streams.find(jack_tag);
    if((i != output_streams.end()) && (i->second->needed == true)) {
      return i->second->enabled;
    }
    else {
      WARNING_STREAM2("Could not find output jack " << jack_tag << endl);
      return false;
    }
  }

  Result GenericNode::setCurrentSendOutputStream(const char *jack_tag)
  {
    OutputStreamMap_t::iterator i = output_streams.find(jack_tag);
    if(i == output_streams.end()) {
      ERROR_STREAM2("No matching jack found to send message!" << endl);
      return FAILURE;
    }
    if (i->second->needed == false) {
      ERROR_STREAM2("No matching jack found to send message!" << endl);
      return FAILURE;
    }
    curr_send_out_stream = i->second->jack_group;
    return SUCCESS;
  }

  const char* GenericNode::getCurrentSendOutputStream() const
  {
    if(curr_send_out_stream != 0) {
      return curr_send_out_stream->getMaster()->getJackTag().c_str();
    }
    else {
      return "unavailable";
    }
  }

  Result GenericNode::setCurrentSendInputStream(const char *jack_tag)
  {
    InputStreamMap_t::iterator i = input_streams.find(jack_tag);
    if (i == input_streams.end()) {
      ERROR_STREAM2("No matching jack found to send message!" << endl);
      return FAILURE;
    }
    if (i->second->needed == false) {
      ERROR_STREAM2("No matching jack found to send message!" << endl);
      return FAILURE;
    }
    curr_send_in_stream = i->second->jack;
    return SUCCESS;
  }

  const char* GenericNode::getCurrentSendInputStream() const
  {
    if(curr_send_in_stream != 0) {
      return curr_send_in_stream->getJackTag().c_str();
    }
    else {
      return "unavailable";
    }
  }

  Message *GenericNode::getNextUpstrMessage()
  {
    Message *msg = 0;

    // starting from current output, check all output,
    // and return first available message from enabled output
    for(size_t i = 0; i < output_streams.size(); ++i) {

      // round robin: move to next output
      curr_recv_out_stream++;
      // if end reached, move to beginning
      if(curr_recv_out_stream == output_streams.end()) {
        curr_recv_out_stream = output_streams.begin();
      }
      // if output enabled, try to dequeue message
      if((curr_recv_out_stream->second->enabled == true) && (curr_recv_out_stream->second->needed == true)) {
        msg = curr_recv_out_stream->second->jack_group->dequeueMessage();
        if(msg)
        {
          // Stop delay measurement
          stopDelayMeasurement(IDelayMeasurement::STOP_WITH_DEQUEUE_MESSAGE, msg);

          // Start delay measurement
          startDelayMeasurement(IDelayMeasurement::START_WITH_DEQUEUE_MESSAGE, msg);

          // Save delay start time for message
          // HACK: fw/mlohse: Hand the delay timestamp from the in_buffer to the out_buffer regardless what processBuffer does
          // HACK: fw/mlohse: Does not work with MUX nodes!
          // HACK: fw/mlohse: Does not work with upstream messages!
          in_delay_time = msg->getDelayStartTime();

          DEBUG_STREAM2("GOT UPSTREAM MSG" << endl);
          return msg;
        }
      }
    }

    return msg;
  }

  Message *GenericNode::getNextDownstrMessage()
  {
    Message *msg = 0;

    // starting from current input, check all inputs,
    // and return first available message from enabled input
    for(size_t i = 0; i < input_streams.size(); ++i) {

      // round robin: move to next input
      curr_recv_in_stream++;
      // if end reached, move to beginning
      if(curr_recv_in_stream == input_streams.end()) {
        curr_recv_in_stream = input_streams.begin();
      }
      // if input enabled, try to dequeue message
      if((curr_recv_in_stream->second->enabled == true) && (curr_recv_in_stream->second->needed == true)) {
        msg = curr_recv_in_stream->second->jack->dequeueMessage();
        if(msg) {
          // Stop delay measurement
          stopDelayMeasurement(IDelayMeasurement::STOP_WITH_DEQUEUE_MESSAGE, msg);

          // Start delay measurement
          startDelayMeasurement(IDelayMeasurement::START_WITH_DEQUEUE_MESSAGE, msg);

          // Save delay start time for message
          // HACK: fw/mlohse: Hand the delay timestamp from the in_buffer to the out_buffer regardless what processBuffer does
          // HACK: fw/mlohse: Does not work with MUX nodes!
          // HACK: fw/mlohse: Does not work with upstream messages!
          in_delay_time = msg->getDelayStartTime();

          return msg;
        }
      }
    }

    // found no message in all enabled queues
    return msg;
  }

  void GenericNode::switchOutputs()
  {
    MutexGuard mg(internal_message_queue_mutex);
    if(switch_enabled) {
      CEvent* ce_slave = new CEvent(IMute::create_unmute());
      CEvent* ce_master = new CEvent(IMute::create_mute());

      internal_message_queue.push(InternalMessage_t(ce_slave, ce_master, true));

      DEBUG_STREAM2("HANDOVER: switchOutputs! set working-flag to TRUE" << endl);
    }
  }

  void GenericNode::flushOutputs(const Timestamp& ts)
  {
    MutexGuard mg(internal_message_queue_mutex);
    if(switch_enabled) {
      if(ts.is_valid && ts.is_set) { // timestamp given -> need to send extra syncReset since FlushEvent will 'outrun' other messages
	CEvent* ce1_slave = new FlushEvent();
	ce1_slave->setTimestamp(ts);
	ce1_slave->insertEvent(IBiDirQueuedNode::create_limitDownstreamSize());

	internal_message_queue.push(InternalMessage_t(ce1_slave, 0, false));

	CEvent* ce2_slave = new CEvent();
	ce2_slave->insertEvent(ISyncReset::create_syncReset());
	CEvent* ce_master = new CEvent();
	ce_master->insertEvent(ISyncReset::create_syncReset());

	internal_message_queue.push(InternalMessage_t(ce2_slave, ce_master, false));
      }
      else { // no timestamp given -> insert syncReset into FlushEvent
	CEvent* ce_slave = new FlushEvent();
	ce_slave->insertEvent(ISyncReset::create_syncReset());
	ce_slave->insertEvent(IBiDirQueuedNode::create_limitDownstreamSize());
	CEvent* ce_master = new CEvent();
	ce_master->insertEvent(ISyncReset::create_syncReset());

	internal_message_queue.push(InternalMessage_t(ce_slave, ce_master, false));
      }

      DEBUG_STREAM2("HANDOVER: flushOutputs! set working-flag to TRUE" << endl);
    }
  }

  void GenericNode::setSwitchOutputsEnabled(const nmm_bool& enable)
  {
    MutexGuard mg(internal_message_queue_mutex);
    switch_enabled = enable;
  }

  nmm_bool GenericNode::getSwitchOutputsEnabled()
  {
    return switch_enabled;
  }

  void GenericNode::run(int num_times)
  {
    bool running = true;

    Message *message = 0;
    while (running) { /* begin > running */
      
      // run num_times
      if(num_times > 0) {
        num_times--;
        if(num_times == 0) {
          running = false;
        }
      }

      // First, dequeue all async signals
      while( !isSignalQueueEmpty() ) {
        int e = dequeueSignal();

        if(e == FREE_DOWNSTR_QUEUE_SLOT) {
          downstr_blocked_queue_count--;
          DEBUG_STREAM2("Signal FREE_DOWNSTR_QUEUE_SLOT, still blocked " << downstr_blocked_queue_count << endl);
          if (downstr_blocked_queue_count <= 0) {
            downstr_blocked_queue_count = 0;
          }
          continue;
        }
        if(e == INPUT_QUEUE_NOT_EMPTY) {
          DEBUG_STREAM2("Signal INPUT_QUEUE_NOT_EMPTY" << endl);
          input_queue_not_empty = true;
          continue;
        }
        if(e == FREE_UPSTR_QUEUE_SLOT) {
          upstr_blocked_queue_count--;
          DEBUG_STREAM2("Signal FREE_UPSTR_QUEUE_SLOT, still blocked " << upstr_blocked_queue_count << endl);
          if (upstr_blocked_queue_count <= 0) {
            upstr_blocked_queue_count = 0;
          }
          continue;
        }
        if(e == OUTPUT_QUEUE_NOT_EMPTY) {
          DEBUG_STREAM2("Signal OUTPUT_QUEUE_NOT_EMPTY" << endl);
          output_queue_not_empty = true;
          continue;
        }
        if(e == WORKING) { // just for debugging
	  DEBUG_STREAM2("WORKING signal" << endl);
          continue;
        }
        if(e == STOP) { 
	  // here, a locked access to internal_started is needed, since there is a wait in stop()
	  internal_started_mutex.lock();
	  internal_started_state = false;
	  internal_started_cv.notifyAll();
	  internal_started_mutex.unlock();
          // Wait for new signals here, otherwise, the following state Lock() might lock,
          // because of concurrent pthread_join in ThreadedNode::deactivate(). (FIXME is this still needed?)
	  // waitForSignal();
	  continue;
        }
	if(e == START) { 
	  // maybe here no locked access is needed ? (since start/stop hold state mutex and are
	  // therfore exclusive!) but it does not hurt much to lock anyway...
	  internal_started_mutex.lock();
	  internal_started_state = true;
	  internal_started_cv.notifyAll();
	  internal_started_mutex.unlock();
	  break; // this forces the run method to call processBuffer at least once (if not blocked)
	}
        if(e == DEACTIVATE) {
          DEBUG_STREAM2("DEACTIVATE Signal!" << endl);
	  // FIXME
	  // leave run()-method only if ThreadedNode::deactivate did return SUCCESS 
	  // because pthread_join is waiting of this thread (otherwise not, and the
	  // application has to call deactivate() again, which will enqueue another
	  // DEACTIVATE signal, which is bad because this will be handled after the 
	  // next time the node was activated!)
          return; // running = false; // do not return here? (to empty sig-queue)
        }
      }

      // if running was set to false by DEACTIVATE signal, return
      //if(!running) {
      //continue;
      //}

      // then try to call processBuffer
      if(internal_started_state || analyse_producing_flag || analyse_processing_flag || getAnalysingFlag()) { /* begin > isStarted */
        // there are different conditions, which allow normal
        // processing here:
        //   * internal_started_state:
        //     The Node is started.
        //   * analyse_producing_flag:
        //     This (Source-)Node has received a "INode::senData"
        //     Event, which forces it to generate a Message for the
        //     requesting output jack
        //   * analyse_processing_flag:
        //     This (Non-Source-)Node has received a "INode::sendData"
        //     Event and has no data from previous analysises. It has
        //     send a "INode::sendData" Event upstream and is waiting
        //     now for the arriving of the requested data resp. the
        //     error message, denying its request.
        //   * analysing_flag:
        //     This Node is currently analysing data for its output
        //     initialisation (e.g. decoding first MPEGAudio frame).
        //     In this case, the run()-method was triggered by
        //     anaylseData() called from within the
        //     doInitOutput()-method of a concret Node. Note that this
        //     implies, that the run()-code currently does not run in
        //     the Node's thread, but in the Application- --
        //     resp. Network- -- Thread.

        if (analysing_cache && !(analyse_producing_flag || analyse_processing_flag || getAnalysingFlag())) {
          // if there is still something in the analysing_cache and we
          // are NOT in one of the analysis-cases, we should first
          // empty the analysing_cache.
          // There cannot be a useful Message for the requesting
          // output jack in analyse_producing- and
          // analyse_processing-cases as this is tested before we
          // eneable these flags (see GenericNode::sendData()).
	  if(!downstr_blocked_queue_count && !upstr_blocked_queue_count) {
            // we only can send a entry from the analysing_cache if
            // there is space in the according queues
            list<AnalysingCacheEntry>::iterator cache_entry(analysing_cache->begin());
            message = &cache_entry->message;
            list<JackGroup*>::iterator cache_jack(cache_entry->jacks.begin());
            if (cache_entry->jacks.size() > 1) {
              // if we have more than one output to sent to
              // make a copy of the message as we need it for the
              // other outputs too!
              message = message->copy();
            }

	    downstr_blocked_queue_count += (*cache_jack)->sendMessage(message);
            message = NULL;

            // after successfully sent the cached message, we can
            // remove the output jack from the cache-entries list.
            cache_entry->jacks.erase(cache_jack);
            if (cache_entry->jacks.empty()) {
              // if the list of to be sent through jacks is empty now,
              // we can remove the analysing_cache entry
              analysing_cache->erase(cache_entry);
              if (analysing_cache->empty()) {
                // ... and if the complete cache is empty now, we can
                // destroy it, to avoid running in this block again in
                // the next iteration.
                delete analysing_cache;
                analysing_cache = NULL;
              }
            }
	  }
	  else {
            // no space for a cached message, wait for a FREE_QUEUE_SLOT
	    waitForSignal(); // because anaylsing_cache not empty but blocked...
	  }
        }
	
	else if(switch_enabled && internal_message_queue.size() && (!downstr_blocked_queue_count)) { // begin > internal message 
	  MutexGuard mg(internal_message_queue_mutex);
	  // get front message(s)
	  InternalMessage_t m = internal_message_queue.front();
	  internal_message_queue.pop();
	  // and send to all outputs
	  for(OutputStreamMap_t::iterator i = output_streams.begin(); i != output_streams.end(); i++) {
	    DEBUG_STREAM2("HANDOVER: SEND SWITCH MESSAGE !!!!!!!!!!!!!!!!!!!!!!!!!!!" << endl);
	    
	    downstr_blocked_queue_count += (*i).second->jack_group->sendSwitchMessages(m.switch_to_msg, m.switch_from_msg); 
	    // check whether jack group needs to be reset
	    if(m.reset_switch_to) {
	      (*i).second->jack_group->resetSwitchTo();
	    }
	  }
	  continue;
	} // end < internal message 
        
	else { /* begin > either (getWorkingFlag) or (dequeue new message) */
	  if(switch_enabled && (!downstr_blocked_queue_count)) {  // begin > check new connections 
	    // ISwitch functionality: check all outgoing jack groups for new connections
	    bool found_new_connections = false;
	    for (OutputStreamMap_t::iterator i = output_streams.begin(); i != output_streams.end(); i++) {
	      // if new connections were found -> send syncReset to particular output
	      if((*i).second->jack_group->checkNewConnections()) {
		found_new_connections = true;
	      }
	    }
	    if(found_new_connections) {
	      MutexGuard mg(internal_message_queue_mutex);
	      DEBUG_STREAM2("HANDOVER: FOUND NEW CONNECTIONS !!!!!!!!!!!!!!!!!!!!!!!!!!!" << endl);
	      CEvent* ce_slave = new CEvent(ISyncReset::create_syncReset());
	      CEvent* ce_master = new CEvent(ISyncReset::create_syncReset());
	      
	      internal_message_queue.push(InternalMessage_t(ce_slave, ce_master, false));
	      continue;
	    }
	  } // end < check new connections 

	  if(getWorkingFlag()) { /* begin > working-flag */
	    if(!downstr_blocked_queue_count && !upstr_blocked_queue_count) {

	      sendMessage(processBuffer(0));
	      
	      if(analyse_producing_flag || analyse_processing_flag || getAnalysingFlag()) { // if (!internal_started_state) { 
		// if we are not started, we must be in analysing,
		// analyse-processing or analysis-producing ...
		DEBUG_STREAM2("WorkingFlag: checkAnalysis()" << endl);
		// ... so we have to check, if we have to continue analysing
		if (checkAnalysis()) {
		  WARNING_STREAM2("checkAnalysis() with WorkingFlag leads to 'break;' in run()!" << endl);
		  // if we are analysing (analysing_flag==true), we
		  // might have enough and want to break out here
		  // (run() was triggered by analyseData() from the
		  // doInitOutput() method which waits for termination
		  // of run()!)
		  break;
		}
		DEBUG_STREAM2("checkAnalysis() with WorkingFlag leads to 'continue;' in run()!" << endl);
	      }
	      
	    }
	    else {
	      waitForSignal(); // because working-flag is true but blocked...
	    }
	  } /* end < working-flag */
	  else { /* begin > not working */

	    // try to get new message
	    message = 0;
	    /*
	    if(switch_enabled && !downstr_blocked_queue_count) {
	      // ISwitch functionality: check all outgoing jack groups for new connections
	      for (OutputStreamMap_t::iterator i = output_streams.begin(); i != output_streams.end(); i++) {
		bool found_new_connections = false;
		// if new connections were found -> send syncReset to particular output
		if((*i).second->jack_group->checkNewConnections()) {
		  found_new_connections = true;
		  setCurrentSendOutputStream((*i).second->jack_group->getMaster()->getJackTag().c_str());
		  cerr << getName() << " blocked " << downstr_blocked_queue_count 
		       << " send syncReset to " << (*i).second->jack_group->getMaster()->getJackTag() << endl;
		  sendMessage(new CEvent(ISyncReset::create_syncReset()));
		}
		// if new connections were found for some jack group -> restart run-loop
		if(found_new_connections) {
		  continue;
		}
	      }
	    }
	    */

	    if(output_queue_not_empty && !upstr_blocked_queue_count) { // first upstream
	      message = getNextUpstrMessage();
	      if(!message) { // no more upstr messages
		output_queue_not_empty = false;
		continue;
	      }
	      else { // set direction status
		current_msg_direction = Message::UPSTREAM;
	      }
	    }
	    else if(input_queue_not_empty && !downstr_blocked_queue_count) { // then downstream
	      message = getNextDownstrMessage();
	      if(!message) { // no more downstr messages
		input_queue_not_empty = false;
		continue;
	      }
	      else { // set direction status
		current_msg_direction = Message::DOWNSTREAM;
	      }
	    }

	    if(message) { /* begin > message */
	      if(message->getType() == Message::BUFFER) {
		if ((filter_node == 0) || (filter_node->isEnabled())) { // no filter or enabled filter
		  sendMessage(processBuffer(static_cast<Buffer*>(message)));
		}
		else { // disabled filters just send buffer unmodified
		  sendMessage(message);
		}
	      } /* end < BUFFER */
	      else if(message->getType() == Message::EVENT) {
		current_event = static_cast<CEvent*>(message);
		
		// need to store the current_event since handling of setFormat
		// can receive new cevents which will overwrite the current_event
		CEvent* tmp = current_event;
		/* dispatch the event */
		getEventDispatcher().internalProcessCEvent(current_event);
		current_event = tmp;

		/* call the processCEvent method, so that the plugin developer can access insrteam events */
		processCEvent(current_event);
		
		// finished whole cevent (with or without success)
		if(current_event->size() == 0) { // cevent no longer holds any events
		  current_event->release();
		}
		else {
		  sendMessage(current_event);
		}
		current_event = 0;
	      } /* end < EVENT */
	      else if(message->getType() == Message::FLUSH_EVENT) {
		MutexGuard mg(state_mutex); // lock for correct handling of flush event (need to double-lock flush())
		current_event = static_cast<CEvent*>(message);
		flush_counter = 0;
		
		// need to store the current_event since handling of setFormat
		// can receive new cevents which will overwrite the current_event
		CEvent* tmp = current_event;
		getEventDispatcher().internalProcessCEvent(current_event);
		current_event = tmp;

		/* call the processCEvent method, so that the plugin developer can access insrteam events */
		processCEvent(current_event);

		// finished whole cevent (with or without success)
		if(current_event->size() == 0) { // cevent no longer holds any events
		  current_event->release();
		}
		else {
		  if(flush_counter < 0) { // counted more DELETE than non-DELETE return values -> no more FlushEvent
		    CEvent* tmp_event = new CEvent(*current_event);
		    current_event->release();
		    current_event = tmp_event;
		  }
		  sendMessage(current_event);
		}
		current_event = 0;
	      } /* end < FLUSH_EVENT */
	      else {
		WARNING_STREAM2("Unhandled Message Type in GenericProcessorNode, just sending it..." << endl);
		sendMessage(message);
	      }

	      if(analyse_producing_flag || analyse_processing_flag || getAnalysingFlag()) { // if (!internal_started_state) { 
		// if we are not started, we must be in analysing, analyse-processing or analysis-producing
		DEBUG_STREAM2("no WorkingFlag: checkAnalysis()" << endl);
		// ... so we have to check, if we have to continue analysing
		if (checkAnalysis()) {
		  WARNING_STREAM2("checkAnalysis() without WorkingFlag leads to 'break;' in run()!" << endl);
		  // if we are analysing (analysing_flag==true), we
		  // might have enough and want to break out here
		  // (run() was triggered by analyseData() from the
		  // doInitOutput() method which waits for termination
		  // of run()!)
		  break;
		}
		DEBUG_STREAM2("checkAnalysis() without WorkingFlag leads to 'continue;' in run()!" << endl);
	      }

	    } /* end < message */
	    else { /* begin > no message */
	      if(getProducingFlag() && !downstr_blocked_queue_count && !upstr_blocked_queue_count) {
		sendMessage(processBuffer(0));
	      }
	      else {
		waitForSignal(); // because everything failed...
	      }
	    } /* end < no message */
	  } /* end < not working and no current event */
	} /* end < else-part : either (getWorkingFlag) or (dequeue new message) */
      } /* end < isStarted */
      else if (isOutputInitialized() || isActivated()) { /* begin > isOutputInitialized || isActivate */
        // although we should not process incoming data, we must
        // check, if we got a upstream Message like a
        // "INode::sendData" Event.
        // (code partly taken from case above!)

        // try to get new message
        message = 0;

        if(output_queue_not_empty && !upstr_blocked_queue_count) { // first upstream
          // as described above, we only check for upstream Messages
          message = getNextUpstrMessage();
          if(!message) { // no more upstr messages
            output_queue_not_empty = false;
            continue;
          }
          else { // set direction status
            current_msg_direction = Message::UPSTREAM;
          }
        }

        if(message) { /* begin > message */

          if(message->getType() == Message::BUFFER) {
            WARNING_STREAM("Upstream Buffer in " << getState() << " received. Releasing!" << endl);
            // we release all Buffers (== data-Messages) received
            // upstream. This should not happen anyway!
            message->release();
          } /* end < BUFFER */

          else if(message->getType() == Message::EVENT) {
            current_event = static_cast<CEvent*>(message);

	    // need to store the current_event since handling of setFormat
	    // can receive new cevents which will overwrite the current_event
	    CEvent* tmp = current_event;
            getEventDispatcher().internalProcessCEvent(current_event);
	    current_event = tmp;

            /* call the processCEvent method, so that the plugin developer can access insrteam events */
            processCEvent(current_event);

            // finished whole cevent (with or without success)
            if(current_event->size() == 0) { // cevent no longer holds any events
              current_event->release();
            }
            else {
              sendMessage(current_event);
            }
            current_event = 0;
          } /* end < EVENT */
	  else if(message->getType() == Message::FLUSH_EVENT) {
	    MutexGuard mg(state_mutex); // lock for correct handling of flush event (need to double-lock flush())
	    current_event = static_cast<CEvent*>(message);
	    flush_counter = 0;

	    // need to store the current_event since handling of setFormat
	    // can receive new cevents which will overwrite the current_event
	    CEvent* tmp = current_event;
	    getEventDispatcher().internalProcessCEvent(current_event);
	    current_event = tmp;

            /* call the processCEvent method, so that the plugin developer can access insrteam events */
            processCEvent(current_event);

	    // finished whole cevent (with or without success)
	    if(current_event->size() == 0) { // cevent no longer holds any events
	      current_event->release();
	    }
	    else {
	      if(flush_counter < 0) { // counted more DELETE than non-DELETE return values -> no more FlushEvent
		CEvent* tmp_event = new CEvent(*current_event);
		current_event->release();
		current_event = tmp_event;
	      }
	      sendMessage(current_event);
	    }
	    current_event = 0;
	  } /* end < FLUSH_EVENT */
          else {
	    WARNING_STREAM2("Unhandled Message Type in GenericProcessorNode, just sending it..." << endl);
	    sendMessage(message);
          }

        } /* end < message */
        else { /* begin > no message */
          waitForSignal(); // because everything failed...
        } /* end < no message */

      } /* end < isOutputInitialized || isActivate */
      else {
	waitForSignal(); // not started, nor activated, nor output-initialized
      }
    } /* end < running */
  }


  IInputJack* GenericNode::getInputJack(const NMM::Format& format, const string & tag)
  {
    InputJack* jack =  const_cast<InputJack*>(const_cast<const GenericNode*>(this)->getInputJack(&format, tag.c_str()));
    if (!jack) {
      ERROR_STREAM2("Input jack with name " << tag << " does not exist!" << endl);
      return 0;
    }

    return jack->getInterface<IInputJack>();
  }

  IInputJack*   GenericNode::getInputStream(const string & tag) {
    InputStreamMap_t::iterator i = input_streams.find(tag.c_str());
    if (i == input_streams.end()) {
      return 0;
    }
    if (i->second->needed == false) {
      return 0;
    }

    if((i->second->jack) && (i->second->jack->isConnected())) {
      return i->second->jack->getInterface<IInputJack>();
    }
    else {
      return 0;
    }
  }

  IOutputJack* GenericNode::getOutputJack(const NMM::Format & format, const string & tag)
  {
    OutputJack* jack =  const_cast<OutputJack*>(const_cast<const GenericNode*>(this)->getOutputJack(&format, tag.c_str(), true));
    if (!jack) {
      ERROR_STREAM2("Output jack with name " << tag << " does not exist!" << endl);
      return 0;
    }

    return jack->getInterface<IOutputJack>();
  }

  IOutputJack* GenericNode::getUnintializedOutputJack(const NMM::Format & format, const string & tag)
  {
    OutputJack* jack =  const_cast<OutputJack*>(const_cast<const GenericNode*>(this)->getOutputJack(&format, tag.c_str(), false));
    if (!jack)
      return 0;
    
    return jack->getInterface<IOutputJack>();
  }

  IOutputJack*  GenericNode::getOutputStream(const string & tag) {
    OutputStreamMap_t::iterator i = output_streams.find(tag.c_str());
    if (i == output_streams.end()) {
      return 0;
    }
    if (i->second->needed == false) {
      return 0;  
    }
    if((i->second->jack_group) &&  (i->second->jack_group->isConnected())) {
      return i->second->jack_group->getMaster()->getInterface<IOutputJack>();
    }
    else {
      return 0;
    }
  }

  const InputJack* GenericNode::getInputJack(const Format *format,
						  const char *jack) const
  {
    InputStreamMap_t::iterator i;

    if (jack == 0) {
      // Need a stream name
      WARNING_STREAM2("Input stream name null in getInputJack, set to default" << endl);
      jack = "default";
    }

    if(format != 0) {
      if(!const_cast<GenericNode*>(this)->getOwnInputProperty()->isFormatMatching(*format, jack)) {
	ERROR_STREAM2("Given format does not match any in input property." << endl);
	return 0;
      }
    }
    else { // no format given -> take default input format
      WARNING_STREAM2("No format given in getInputJack, taking default input format" << endl);
      format = const_cast<GenericNode*>(this)->getOwnWorkingInputProperty()->getDefaultFormat();
      if(format == 0) {
	ERROR_STREAM2("No default format in WorkingInputProperty for getInputJack" << endl);
	return 0;
      }
    }

    i = input_streams.find(const_cast<char*>(jack));
    if (i == input_streams.end()) {
      // No such stream defined!
      ERROR_STREAM2("Stream named " << jack << " not defined" << endl);
      return 0;
    }
    if (i->second->needed == false) {
      // No such stream defined!
      ERROR_STREAM2("Stream named " << jack << " not defined" << endl);
      return 0;
    }

    /*Begin> delete the old format and create a copy of the new one (should not be NULL here) */
    delete i->second->format;
    i->second->format = new Format(*format);
    /*End */

    return i->second->jack;
  }

  Result GenericNode::removeOutputStream(const char *jack_tag)
  {
    OutputStreamMap_t::iterator i = output_streams.find(jack_tag);
    if (i == output_streams.end()) {
      WARNING_STREAM2("No matching jack found to remove" << endl);
      return FAILURE;
    }
    if (i->second->needed == false) {
      WARNING_STREAM2("No matching jack found to remove" << endl);
      return FAILURE;
    }
    else {
      DEBUG_STREAM2("Set needed = false for output stream " << jack_tag << endl);
      i->second->needed = false;
      if(curr_send_out_stream == i->second->jack_group) {
	curr_send_out_stream = 0;
      }
      if(curr_recv_out_stream == i) {
	curr_recv_out_stream = output_streams.begin();
      }
      return SUCCESS;
    }
  }

  void GenericNode::removeUnneededOutputStreams()
  {
    for(OutputStreamMap_t::iterator i = output_streams.begin(); i != output_streams.end(); ++i) {
      if(i->second->needed == false) {
	DEBUG_STREAM2("Remove unneeded output stream named " << i->first << endl);
	delete [] i->first;
	delete i->second;
	output_streams.erase(i);
	if(curr_send_out_stream == i->second->jack_group) {
	  curr_send_out_stream = 0;
	}
	if(curr_recv_out_stream == i) {
	  curr_recv_out_stream = output_streams.begin();
	}
      }
    }
  }

  Result GenericNode::addOutputStream(const char *jack_tag)
  {
    OutputStreamMap_t::iterator i = output_streams.find(jack_tag);
    if (i != output_streams.end()) {
      if(i->second->needed == false) {
	i->second->needed = true;
	DEBUG_STREAM2("Output stream named " << jack_tag << " reactivated." << endl);
	// update curr_recv_out_stream
	curr_recv_out_stream = output_streams.begin();
	curr_send_out_stream = 0;
	return SUCCESS;
      }
      else {
	DEBUG_STREAM2("Output stream named " << jack_tag << " was already added before." << endl);
	return FAILURE;
      }
    }
    else {
      string jack_name = jack_tag;
      jack_name = getName() + "> OutputJack> " + jack_name;

      OutputStream_t *out = new OutputStream_t();
      out->jack_group   = new JackGroup(this, new QueuedOutputJack(this, jack_tag, jack_name, queue_mode, queue_size));

      // create a copy of the jack_tag
      char * my_jack_tag = new char[strlen(jack_tag) + 1];
      strcpy(my_jack_tag, jack_tag);

      pair<OutputStreamMap_t::iterator, bool> res
	= output_streams.insert(make_pair(my_jack_tag, out));
      // could not insert object, then delete it
      if (!res.second) {
	delete [] my_jack_tag;
	delete out;
	return FAILURE;
      }
      DEBUG_STREAM2("Output stream named " << jack_tag << " successfully added." << endl);
      // update curr_recv_out_stream
      curr_recv_out_stream = output_streams.begin();
      curr_send_out_stream = 0;
      return SUCCESS;
    }
  }

  Result GenericNode::sendMessage(Message *out)
  {
    if(out != 0) {
      // Set the saved delay start time for message
      // HACK: fw/mlohse: Hand the delay timestamp from the in_buffer to the out_buffer regardless what processBuffer does
      // HACK: fw/mlohse: Does not work with MUX nodes!
      if(!out->isDelayStartTimeSet() && in_delay_time != Time())
      {
        out->setDelayStartTime(in_delay_time);
      }

      // Stop delay measurement
      stopDelayMeasurement(IDelayMeasurement::STOP_WITH_SEND_MESSAGE, out);

      // Start delay measurement
      startDelayMeasurement(IDelayMeasurement::START_WITH_SEND_MESSAGE, out);

      // send cevents to external listeners (UPSTREAM or DOWNSTREAM)
      if((out->getType() == Message::EVENT) || (out->getType() == Message::FLUSH_EVENT)) {
        if (!isStarted() && (out->getDirection() == Message::DOWNSTREAM)) {
          WARNING_STREAM("Downstream Message in " << getState() << "." << endl);
        }
        getEventDispatcher().externalProcessCEvent(static_cast<CEvent*>(out));
      }

      if(out->getDirection() == Message::DOWNSTREAM) { // DOWNSTREAM
        if (getAnalysingFlag() || analyse_processing_flag) { // if we are anaysing
          // we have a Message to send downstream, but we are in
          // analysing- oder analyse_processing-mode, so we have to
          // cache the Message
          if (!analysing_cache) {
            DEBUG_STREAM("creating new analysing cache" << endl);
            analysing_cache = new list<AnalysingCacheEntry>();
          }

          list<JackGroup*> entry_jacks;
          if (curr_send_out_stream == NULL) {
            if (analyse_processing_flag) {
              // if we are in analyse_processing and have to send a
              // Message to ALL outputs, we generate a list of all
              // outputs to be inserted in the analysing_cache entry
              for (OutputStreamMap_t::const_iterator os = output_streams.begin();
                   os != output_streams.end();
                   os++) {
		if(os->second->needed == true) {
		  entry_jacks.push_back(os->second->jack_group);
		}
              }
              // note: in analysing mode, we cannot generate the list
              // of all outputs here as the according list might be
              // incomplete (e.g. the MPEGDemuxNode adds its output
              // streams while analysis, so there might be some
              // streams, which are not found yet and so not known
              // yet!)
              // Instead, we insert a empty list here, to be replaced
              // with the complete list after analysis (see
              // GenericNode::analyseData()).
            }
          } else {
            // we only have to send this Message to the specified
            // output stream, so insert it in the list for the
            // anaylsing_cache entry.
            entry_jacks.push_back(curr_send_out_stream);
          }
          // insert a reference to the to-be-cached Message (this
          // ensures, that it is not NULL as a NULL-Ptr cannot be
          // stored as reference)
          // and the list of outputs this Message should be send to
          // (computed above)
          analysing_cache->push_back(AnalysingCacheEntry(*out, entry_jacks));

          if (analyse_processing_flag &&
              ((curr_send_out_stream == current_analysis_out_stream) ||
               (curr_send_out_stream == 0))) {
            // if we are analyse_processing and have cached a Message
            // for the output which requested data, we remember this
            // in a internal flag, evaluated in
            // GenericNode::checkAnalysis().
            DEBUG_STREAM2("message was forwarded to requesting output " << current_analysis_out_stream->getMaster()->getJackTag() << endl);
            analysis_data_forwarded = true;
          }

          // if more then one output (demux) then reset
          // curr_send_out_stream to 0 after each send message
          if(output_streams.size() > 1) {
            curr_send_out_stream = 0;
          }

          // we return here, to avoid sending the cached Message all
          // the same (to the maybe not connected outputs!)
          return SUCCESS;
        }

	if(curr_send_out_stream != 0) {
	  downstr_blocked_queue_count += curr_send_out_stream->sendMessage(out);
	  // if more then one output (demux) then reset curr_send_out_stream to 0 after each send message
	  if(output_streams.size() > 1) {
	    curr_send_out_stream = 0;
	  }
	}
	else {
	  // send to all out jacks
	  OutputStream_t *out_stream = 0;
	  for(OutputStreamMap_t::iterator i = output_streams.begin(); i != output_streams.end(); i++) {
	    out_stream = i->second;
	    if(out_stream) {
	      if(out_stream->jack_group) {
		if(out_stream->needed) {
		  downstr_blocked_queue_count += out_stream->jack_group->sendMessage(out->copy());
		  DEBUG_STREAM2("SENT EVENT, ref count " << out->getCount()
				<< " blocked " << downstr_blocked_queue_count << endl);
		}
		else {
		  DEBUG_STREAM2("not sending to unneded out_stream" << endl);
		}
	      }
	      else {
		ERROR_STREAM2("ARGH. out_stream->jack is 0 " << endl);
	      }
	    }
	    else {
	      ERROR_STREAM2("ARGH. out_stream is 0 " << endl);
	    }
	  }
	  // need to release since out->copy() was called for each jack
	  out->release();
        } // end (no curr_send_out_stream set)


        if (analyse_producing_flag) {  // if we are producing for Analysis
          // we don't have to cache the Message like in the cases
          // above, but we have to honor, that we have send a Message
          // and so can stop "producing" for analysis
          DEBUG_STREAM2("Disabling Analyse Production! WF: " << getWorkingFlag() << endl);
          setAnalyseProducingFlag(false); // stop producing
        }

      }
      else { // UPSTREAM
	if(curr_send_in_stream != 0) {
         if(curr_send_in_stream->sendMessage(out) == BLOCKED) {
	    ++upstr_blocked_queue_count;
	  }
	  // if more then one input (mux) then reset curr_send_in_stream to 0 after each send message
	  if(input_streams.size() > 1) {
	    curr_send_in_stream = 0;
	  }
	}
	else {
	  // send to all in jacks
	  InputStream_t *in_stream = 0;
	  for (InputStreamMap_t::iterator i = input_streams.begin(); i != input_streams.end(); i++) {
	    in_stream = i->second;
	    if(in_stream) {
	      if(in_stream->jack) {
		if(in_stream->needed) {
		  if(in_stream->jack->sendMessage(out->copy()) == BLOCKED) {
		    ++upstr_blocked_queue_count;
		  }
		  DEBUG_STREAM2("SENT EVENT, ref count " << out->getCount()
				<< " blocked " << upstr_blocked_queue_count << endl);
		}
		else {
		  DEBUG_STREAM2("not sending to unneded in_stream" << endl);
		}
	      }
	      else {
		ERROR_STREAM2("ARGH. in_stream->jack is 0 " << endl);
	      }
	    }
	    else {
	      ERROR_STREAM2("ARGH. in_stream is 0 " << endl);
	    }
	  }
	  // need to release since out->copy() was called for each jack
	  out->release();
	} // end (no curr_send_in_stream set)
      }

      // in all cases we can return SUCCESS, so return it here instead
      // of all cases explicit
      return SUCCESS;

    } // end if(out != 0}
    else {
      DEBUG_STREAM2( "tried to send to 0 message" << endl);
      // even if 0-message should have been sent : forget currend send-stream for demux/mux-nodes
      if(output_streams.size() > 1) {
	curr_send_out_stream = 0;
      }
      if(input_streams.size() > 1) {
	curr_send_in_stream = 0;
      }

      return FAILURE;
    }
  }

  Format* GenericNode::getInputStreamFormat(const string& jack_tag)
  {
    const Format* f = getInputFormat(jack_tag.c_str());
    if(f != 0) {
      return new Format(*f);
    }
    else {
      return 0;
    }
  }

  const Format* GenericNode::getInputFormat(const char * jack_tag) const
  {
    InputStreamMap_t::iterator i = input_streams.find(const_cast<char*>(jack_tag));
    if (i == input_streams.end()) return 0;
    if (i->second->needed == false) return 0;
    return i->second->format;
  }

  Format* GenericNode::getOutputStreamFormat(const string& jack_tag)
  {
    const Format* f = getOutputFormat(jack_tag.c_str());
    if(f != 0) {
      return new Format(*f);
    }
    else {
      return 0;
    }
  }
  const Format* GenericNode::getOutputFormat(const char * jack_tag) const
  {
    OutputStreamMap_t::const_iterator i = output_streams.find(jack_tag);
    if (i == output_streams.end()) return 0;  // not found!
    if (i->second->needed == false) return 0;
    return i->second->format;
  }

  list<string> GenericNode::getOutputStreamTags() const
  {
    list<string> tmp_list;
    for (OutputStreamMap_t::const_iterator i = output_streams.begin();
         i != output_streams.end();
         i++) {
      if (i->second->needed == true) {
	tmp_list.push_back(string(i->first));
      }
    }
    return tmp_list;
  }

  IOutputJack* GenericNode::cloneOutputJack(Format*& intersec_format, const string& jack_tag /* = "default" */ )
  {
    OutputStreamMap_t::const_iterator i = output_streams.find(jack_tag.c_str());
    if (i == output_streams.end()) return 0;  
    if (i->second->needed == false) return 0;

    const Format* f = getOutputFormat(jack_tag.c_str());
    if(f == 0) {
      ERROR_STREAM2("No output format for cloneOutputJack" << endl);
      return 0;
    }
    intersec_format = new Format(*f);

    string jack_name = jack_tag;
    jack_name = getName() + "> OutputJackClone> " + jack_name;

    QueuedOutputJack* clone = new QueuedOutputJack(this, jack_tag, jack_name, queue_mode, queue_size);

    i->second->jack_group->addJack(clone);

    return clone->getInterface<IOutputJack>();
  }

  InputJack* GenericNode::getOwnInputStream(const string& jack_tag /*= "default"*/ )
  {
    InputStreamMap_t::iterator i = input_streams.find(jack_tag.c_str());
    if (i == input_streams.end()) return 0;
    if (i->second->needed == false) return 0;
    return i->second->jack;
  }

  const InputJack* GenericNode::getOwnInputStream(const string& jack_tag /*= "default"*/ ) const
  {
    InputStreamMap_t::iterator i = input_streams.find(jack_tag.c_str());
    if (i == input_streams.end()) return 0;
    if (i->second->needed == false) return 0;
    return i->second->jack;
  }

  OutputJack* GenericNode::getOwnOutputStream(const string& jack_tag /*= "default"*/ )
  {
    OutputStreamMap_t::const_iterator i = output_streams.find(jack_tag.c_str());
    if (i == output_streams.end()) return 0;  // not found!
    if (i->second->needed == false) return 0;  
    return i->second->jack_group->getMaster();
  }

  const OutputJack* GenericNode::getOwnOutputStream(const string& jack_tag /*= "default"*/ ) const
  {
    OutputStreamMap_t::const_iterator i = output_streams.find(jack_tag.c_str());
    if (i == output_streams.end()) return 0;  // not found!
    if (i->second->needed == false) return 0;  
    return i->second->jack_group->getMaster();
  }

  // upstream
  Result GenericNode::sendEventUpstr(Event& event, const char * jack_tag)
  {
    CEvent* ce = new CEvent(event);
    return sendEventUpstr(ce, jack_tag);
  }

  Result GenericNode::sendEventUpstr(CEvent& cevent, const char * jack_tag)
  {
    CEvent* ce = new CEvent(cevent);
    return sendEventUpstr(ce, jack_tag);
  }

  Result GenericNode::sendEventUpstr(CEvent* cevent, const char * jack_tag)
  {
    OutputStreamMap_t::const_iterator i = output_streams.find(jack_tag);
    if (i == output_streams.end()) return FAILURE;
    if (i->second->needed == false) return FAILURE;
    cevent->setDirection(Message::UPSTREAM);
    return i->second->jack_group->getMaster()->receiveMessage(cevent, 0);
  }

  // downstream
  Result GenericNode::sendEventDownstr(CEvent& cevent, const char * jack_tag)
  {
    return sendEventDownstr(new CEvent(cevent), jack_tag);
  }

  Result GenericNode::sendEventDownstr(Event& event, const char * jack_tag)
  {
    return sendEventDownstr(new CEvent(event), jack_tag);
  }

  Result GenericNode::sendEventDownstr(CEvent* cevent, const char * jack_tag)
  {
    InputStreamMap_t::iterator i = input_streams.find(const_cast<char*>(jack_tag));
    if (i == input_streams.end()) return FAILURE;
    if (i->second->needed == false) return FAILURE;
    cevent->setDirection(Message::DOWNSTREAM);
    return i->second->jack->receiveMessage(cevent, 0);
  }

  const OutputJack* GenericNode::getOutputJack(const Format *format, const char *jack, bool output_initialized) const
  {
    OutputStreamMap_t::iterator i;

    if (jack == 0) {
      // Need a stream name
      DEBUG_STREAM2( "Output stream name null in getOutputJack, set to default" << endl);
      jack = "default";
    }

    if(format != 0) { // check if format is matching
      if(output_initialized) { // usual case: nod is in state OUTPUT_INITIALIZED
	if(!const_cast<GenericNode*>(this)->getOwnWorkingOutputProperty()->isFormatMatching(*format, jack)) {
	  ERROR_STREAM2("Given format does not match any in output property." << endl);
	  return 0;
	}
      }
      else { // not in state OUTPUT_INITIALIZED (but INITIALIZED) -> use 'static' output property
	if(!const_cast<GenericNode*>(this)->getOwnOutputProperty()->isFormatMatching(*format, jack)) {
	  ERROR_STREAM2("Given format does not match any in working output property." << endl);
	  return 0;
	}
      }
    }
    else { // no format given -> take default output format
      WARNING_STREAM2("No format given in getOutputJack, taking default output format" << endl);
      format = const_cast<GenericNode*>(this)->getOwnWorkingOutputProperty()->getDefaultFormat();
      if(format == 0) {
	ERROR_STREAM2("No default format in WorkingOutputProperty for getOutputJack" << endl);
	return 0;
      }
    }

    if(!output_initialized) { // when called from getUnitializedOutputJack (INITIALZIED) -> simply force adding of stream
      i = output_streams.find(jack);
      if (i == output_streams.end()) {
	const_cast<GenericNode*>(this)->addOutputStream(jack);
	WARNING_STREAM2("Forced addition of stream named " << jack << endl);
      }
    }

    // both cases, try to get stream (again)
    i = output_streams.find(jack);
    if (i == output_streams.end()) {
      // No such stream defined!
      ERROR_STREAM2("Stream named " << jack << " not defined" << endl);
      return 0;
    }
    
    if (i->second->needed == false) {
      // No such stream defined!
      ERROR_STREAM2("Stream named " << jack << " not defined" << endl);
      return 0;
    }

    /*Begin> delete the old format and create a copy of the new one (should not be NULL here) */
    delete i->second->format;
    i->second->format = new Format(*format);
    /*End< */

    return i->second->jack_group->getMaster();
  }

  unsigned int GenericNode::getQueueSize(const char* jack_tag /* = "default" */)
  {
    InputStreamMap_t::iterator i =
      input_streams.find(const_cast<char*>(jack_tag));
    if (i == input_streams.end()) return 0;
    if (i->second->needed == false) return 0;
    return i->second->jack->getQueueSize();
  }

  nmm_uint32 GenericNode::getDownstreamSize(const string& id /* = "default" */ ) const
  {
    InputStreamMap_t::iterator i = input_streams.find(id.c_str());
    if (i == input_streams.end()) return 0;
    if (i->second->needed == false) return 0;
    return i->second->jack->getSize();
  }

  nmm_uint32 GenericNode::getDownstreamMaxSize(const string& id /* = "default" */ ) const
  {
    InputStreamMap_t::iterator i = input_streams.find(id.c_str());
    if (i == input_streams.end()) return 0;
    if (i->second->needed == false) return 0;
    return i->second->jack->getMaxSize();
  }

  IQueued::Mode GenericNode::getDownstreamMode(const string& id /* = "default" */ ) const
  {
    InputStreamMap_t::iterator i = input_streams.find(id.c_str());
    if (i == input_streams.end()) return IQueued::DEFAULT_MODE;
    if (i->second->needed == false) return IQueued::DEFAULT_MODE;
    return (IQueued::Mode) (i->second->jack->getMode());
  }

  Result GenericNode::setDownstreamMaxSize(const nmm_uint32& size, const string& id /* = "default" */ ) /* limited(Node::CONSTRUCTED) */
  {
    InputStreamMap_t::iterator i = input_streams.find(id.c_str());
    if (i == input_streams.end()) return FAILURE;
    if (i->second->needed == false) return FAILURE;
    return i->second->jack->setMaxSize(size);
  }

  Result GenericNode::setDownstreamMaxDuration(const Interval& duration, const string& id /* = "default" */ ) /* limited(Node::CONSTRUCTED) */
  {
    InputStreamMap_t::iterator i = input_streams.find(id.c_str());
    if (i == input_streams.end()) return FAILURE;
    if (i->second->needed == false) return FAILURE;
    return i->second->jack->setMaxDuration(duration);
  }

  Result GenericNode::setDownstreamMode(const IQueued::Mode& mode, const string& id /* = "default" */ ) /* limited(Node::CONSTRUCTED) */
  {
    InputStreamMap_t::iterator i = input_streams.find(id.c_str());
    if (i == input_streams.end()) return FAILURE;
    if (i->second->needed == false) return FAILURE;
    return i->second->jack->setMode((StreamQueue::Mode) mode);
  }

  Result GenericNode::unlimitDownstreamSize() /* limited(INode::STARTED) */ /* in-stream */
  {
    for (InputStreamMap_t::iterator i = input_streams.begin(); i != input_streams.end(); i++) {
      i->second->jack->unlimitSize();
    }    
    return DELETE;
  }

  Result GenericNode::limitDownstreamSize() /* limited(INode::STARTED) */ /* in-stream */ 
  {
    for (InputStreamMap_t::iterator i = input_streams.begin(); i != input_streams.end(); i++) {
      i->second->jack->limitSize();
    }    
    return DELETE;
  }

  nmm_uint32 GenericNode::getUpstreamSize(const string& id /* = "default" */ ) const
  {
    OutputStreamMap_t::iterator i = output_streams.find(id.c_str());
    if (i == output_streams.end()) return 0;
    if (i->second->needed == false) return 0;
    return i->second->jack_group->getMaster()->getSize();
  }

  nmm_uint32 GenericNode::getUpstreamMaxSize(const string& id /* = "default" */ ) const
  {
    OutputStreamMap_t::iterator i = output_streams.find(id.c_str());
    if (i == output_streams.end()) return 0;
    if (i->second->needed == false) return 0;
    return i->second->jack_group->getMaster()->getMaxSize();
  }

  IQueued::Mode GenericNode::getUpstreamMode(const string& id /* = "default" */ ) const
  {
    OutputStreamMap_t::iterator i = output_streams.find(id.c_str());
    if (i == output_streams.end()) return IQueued::DEFAULT_MODE;
    if (i->second->needed == false) return IQueued::DEFAULT_MODE;
    return (IQueued::Mode) i->second->jack_group->getMaster()->getMode();
  }

  Result GenericNode::setUpstreamMaxSize(const nmm_uint32& size, const string& id /* = "default" */ ) /* limited(Node::CONSTRUCTED) */
  {
    OutputStreamMap_t::iterator i = output_streams.find(id.c_str());
    if (i == output_streams.end()) return FAILURE;
    if (i->second->needed == false) return FAILURE;
    return i->second->jack_group->getMaster()->setMaxSize(size);
  }

  Result GenericNode::setUpstreamMode(const IQueued::Mode& mode, const string& id /* = "default" */ ) /* limited(Node::CONSTRUCTED) */
  {
    OutputStreamMap_t::iterator i = output_streams.find(id.c_str());
    if (i == output_streams.end()) return FAILURE;
    if (i->second->needed == false) return FAILURE;
    return i->second->jack_group->getMaster()->setMode((StreamQueue::Mode) mode);
  }

  Result GenericNode::setMaxSize(const nmm_uint32& size, const string& id /* = "default" */ )
  {
    Result r = SUCCESS;
    if(setDownstreamMaxSize(size, id) != SUCCESS) {
      r = FAILURE;
    }
    if(setUpstreamMaxSize(size, id) != SUCCESS) {
      r = FAILURE;
    }
    return r;
  }

  Result GenericNode::setMode(const IQueued::Mode& mode, const string& id /* = "default" */ )
  {
    Result r = SUCCESS;
    if(setDownstreamMode(mode, id) != SUCCESS) {
      r = FAILURE;
    }
    if(setUpstreamMode(mode, id) != SUCCESS) {
      r = FAILURE;
    }
    return r;
  }

  /******************** analyseData & Co. ********************/
  void GenericNode::setAnalysingFlag(const bool w) 
  {
    cached_analysing_flag = w;
  }

  void GenericNode::updateAnalysingFlag()
  {
    // as the value of the analysing_flag must not be changed directly
    // if the concret Node set it to false (because it has finished
    // its analysis!), this method updates the value from the cached
    // one set in GenericNode::setAnalysingFlag(). This update is
    // done, every time, a iteration of the run()-method is finished
    // (in GenericNode::checkAnalysis(), see there).
    analysing_flag = cached_analysing_flag;
  }

  bool GenericNode::getAnalysingFlag() const
  {
    return analysing_flag;
  }

  void GenericNode::setAnalyseProducingFlag(const bool w) 
  {
    analyse_producing_flag = w;
  }

  bool GenericNode::getAnalyseProducingFlag() const
  {
    return analyse_producing_flag;
  }

  bool GenericNode::isAnyOutputStreamConnected(const string tag) const {
    OutputStreamMap_t::iterator i = output_streams.find(tag.c_str());
    if (i == output_streams.end()) {
      return false;
    }
    if (i->second->needed == false) {
      return false;  
    }
    if((i->second->jack_group) &&  (i->second->jack_group->isConnected())) {
      return true;
    }
    else {
      return false;
    }
  }

  void GenericNode::setMaxAnalysisIterations(const unsigned int it) 
  {
    max_analysis_iterations = it;
  }

  unsigned int GenericNode::getMaxAnalysisIterations() const
  {
    return max_analysis_iterations;
  }

  void GenericNode::setAnalyseProcessingFlag(const bool w) 
  {
    analyse_processing_flag = w;
  }

  bool GenericNode::getAnalyseProcessingFlag() const
  {
    return analyse_processing_flag;
  }

  bool GenericNode::analyseData(const string& in_jack_tag /*= "default"*/) {
    // the concret Node requests data for analysis from the input
    // stream in_jack_tag
    MESSAGE_STREAM2("Begin> analyseData" << endl);
    
    InputJack* in_jack = NULL;
    QueuedInputJack* queued_in_jack = NULL;
    if (getType() != INode::SOURCE) {
      /* get the input jack from which data should be requested */
      in_jack = getOwnInputStream(in_jack_tag);
      if(!in_jack) {
        ERROR_STREAM2("Selected InputJack " << in_jack_tag << " for analyseData is not available!" << endl);
        return false;
      }
      
      /* check if it is a queued input jack. otherwise, assume that no data is available */
      queued_in_jack = dynamic_cast<QueuedInputJack*>(in_jack);
      if(!in_jack) {
        DEBUG_STREAM2("Selected InputJack " << in_jack_tag << " for analyseData is not a QueuedJack!" << endl);
      }
    } else {
      /* a source node has no input jack */
      DEBUG_STREAM("analyseData(): SOURCE-Node, not requesting data from predecessor!" << endl);
    }
    
    for (unsigned int iteration = 0; iteration < getMaxAnalysisIterations(); iteration++) {
      // data analysis needs some iterations to come to a valid
      // result, so we iterate over the iterations. The upper limit is
      // given by the value of
      // GenericNode::getMaxAnalysisIterations(). A concret Node can
      // influent this value by using
      // GenericNode::setMaxAnalysisIterations() (e.g. the
      // MPEGDemuxNode increases the value to get its 100 demuxed
      // packets, which are taken as basis for its analysing result:
      // Which output streams does the incoming MPEG system stream
      // provide?)
      DEBUG_STREAM2("Data analysis: Iteration " << iteration << endl);
      
      /* we need to request data if we have an input jack and the input jack has no queue or the queue contains no data. */
      if(in_jack && (!queued_in_jack || queued_in_jack && queued_in_jack->getSize() == 0)) {
        if(!transmitSendData(*in_jack)) {
          /* failed to request data */
          ERROR_STREAM2("Could not send " << INode::sendData_event << " to " << in_jack_tag << ", maybe not connected?" << endl);
          ERROR_STREAM2("analyseData() failed!" << endl);
          return false;
        }
      }
      
      // we set the analysing_flag for the forthcoming analysis
      setAnalysingFlag(true);
      // don't forget to update the analying_flag, see
      // GenericNode::updateAnalysingFlag().
      updateAnalysingFlag();
      DEBUG_STREAM2("Begin> run" << endl);
      // let the Node run until GenericNode::checkAnalysis() breaks it
      // out of run()
      run();
      DEBUG_STREAM2("<End run" << endl);
      if (!getAnalysingFlag()) {
        // the Node has finished its analysis, so break out of this
        // for-iteration
        MESSAGE_STREAM2("Finished analysis!" << endl);
        break;
      }
    }
    
    // we store the result of the analysis for later use as we have to
    // disable the analysing flag now!
    bool analysis_failed(getAnalysingFlag());
    setAnalysingFlag(false);
    // don't forget to update the analying_flag, see
    // GenericNode::updateAnalysingFlag().
    updateAnalysingFlag();

    MESSAGE_STREAM2("Begin> finalizing analysing_cache" << endl);
    if (analysing_cache) {
      // now its time, to replace the empty lists of output streams in
      // the analysing_cache entries with lists with all streams. This
      // is neccessary as lists with all outputs cannot be generated
      // while analysing, see GenericNode::sendData() for more details.
      list<JackGroup*> all_output_jacks;
      for (OutputStreamMap_t::iterator stream = output_streams.begin();
           stream != output_streams.end();
           stream++) {
	if (stream->second->needed == true) {
	  all_output_jacks.push_back(stream->second->jack_group);
	}
      }

      // fit cached messages with appropriate output streams
      for (list<AnalysingCacheEntry>::iterator entry = analysing_cache->begin();
	   entry != analysing_cache->end();
	   entry++) {
	if (entry->jacks.empty()) {
          entry->jacks = all_output_jacks;
        }
      }
    } // end analysing_cache != NULL
    MESSAGE_STREAM2("End< finalizing analysing_cache" << endl);

    if (send_data_failed) {
      // we received a "INode::sendDataFailed" while analysing, so the
      // analysis was not successful terminated
      send_data_failed = false;
      ERROR_STREAM2("analysis failed because sendData failed!" << endl);
      return false;
    }

    if (analysis_failed) {
      // although we made all analysis iterations, we did not come to
      // a successful finished (analysing_flag was true after last
      // iteration, which means the Node still has not finished analysis)
      analysis_failed = false;
      ERROR_STREAM2("analysis failed because analysis failed!" << endl);
      return false;
    }

    MESSAGE_STREAM2("End< analyseData" << endl);
    // if we are here, the analysis of the input data was successful, so:
    return true;
  }

  Result GenericNode::sendData() {
    // we received a "INode::sendData" which instructs us to send
    // Messages for analysis to the Node, we received this Event
    // from.
    MESSAGE_STREAM2("Begin> sendData" << endl);

    // there are several (boolean) conditions, which decide, what's to
    // be done when receiving a request for data. We choose the right
    // case in a long if-then-else-if...
    if (Message::UPSTREAM != current_msg_direction) {
      ERROR_STREAM2("Got " << INode::sendData_event << " DOWNSTREAM!!!!" << endl);
      // not sending error msg upstream!
    }

    else if (!curr_recv_out_stream->second->enabled) {
      ERROR_STREAM2("curr_recv_out_stream is not enabled! How did we get this message?" << endl);
      // not sending error msg downstream!
    }

    else if (analysing_cache && sendAnalysingCacheEntryTo(*curr_recv_out_stream->second->jack_group)) {
      // we had a Message in the analysing_cache, which could be send
      // to the requesting Node
      DEBUG_STREAM("Successful send AnalysingCacheEntry to " << curr_recv_out_stream->first << endl);
    }
    
    else if (getType() == INode::SOURCE) {
      // this is a SourceNode. If it has to send a Message to its
      // predecessor, we simply have to enable the
      // analyse_producing_flag, then the Node's run()-method will do
      // the right stuff.
      MESSAGE_STREAM2("Enabling SourceNode's production!" << endl);
      setAnalyseProducingFlag(true);
    }

    else { // Begin> Non-SourceNode
      MESSAGE_STREAM2("Entering analysis-processing-state." << endl);
      // for Non-SourceNodes we enable the analyse_processing_flag ...
      setAnalyseProcessingFlag(true);
      
      // ... and request data round-robin from its input
      pair<string, InputJack*> next_analysis_input(getNextAnalysisInput());
      current_analysis_in_stream = next_analysis_input.second;
      current_analysis_out_stream = curr_recv_out_stream->second->jack_group;
      analysis_data_forwarded = false;

      if (!transmitSendData(*current_analysis_in_stream)) {
        ERROR_STREAM2("Could not send " << INode::sendData_event << " to input stream " << next_analysis_input.first << "." << endl);
        if (!transmitSendDataFailed(*current_analysis_out_stream)) {
          ERROR_STREAM2("Could not send " << INode::sendDataFailed_event << " to output stream although he requested data." << endl);
        }
        // sending data request failed, we disable analyse_processing again
        setAnalyseProcessingFlag(false);
      }
    } // End< Non-SourceNode

    MESSAGE_STREAM2("<End sendData" << endl);
    // in all cases, we have to delete this Event to avoid being send
    // further
    return DELETE;
  }

  Result GenericNode::sendDataFailed() {
    // we received a "INode::sendDataFailed" Event which signals, that
    // the Node, we requested data from, is not able to produce some
    // for us. So we stop analysing and set a flag for
    // GenericNode::checkAnalysis().
    ERROR_STREAM2("sendData failed!" << endl);
    setAnalysingFlag(false);
    send_data_failed = true;
    // in all cases, we have to delete this Event to avoid being send
    // further
    return DELETE;
  }

  bool GenericNode::transmitSendDataFailed(JackGroup& jack) {
    // this small helper is used to compact code, required at several
    // places (e.g. GenericNode::analyseData() or
    // GenericNode::checkAnalysis()).
    // it sends a "INode::sendDataFailed" Event through the given output.
    if (jack.isConnected()) {
      CEvent* cevent = new CEvent(INode::create_sendDataFailed());
      cevent->setDirection(Message::DOWNSTREAM);
      DEBUG_STREAM2("Begin> Send Error" << endl);
      if (BLOCKED == jack.sendMessage(cevent)) {
        WARNING_STREAM2("Jack returned blocked sending error!" << endl);
        ++downstr_blocked_queue_count;
      }
      DEBUG_STREAM2("End< Send Error" << endl);
      return true;
    } else {
      ERROR_STREAM2("Trying to send Event through not connected Jack!" << endl);
      return false;
    }
  }

  bool GenericNode::transmitSendData(InputJack& jack) {
    // this small helper is used to compact code, required at several
    // places (e.g. GenericNode::analyseData() or
    // GenericNode::checkAnalysis()).
    // it sends a "INode::sendData" Event through the given input.
    if (jack.isConnected()) {
      CEvent* cevent = new CEvent(INode::create_sendData());
      cevent->setDirection(Message::UPSTREAM);
      DEBUG_STREAM2("Begin> Send Message " << cevent << ">" << endl);
      if (BLOCKED == jack.sendMessage(cevent)) {
        WARNING_STREAM2("Jack returned blocked sending sendData!" << endl);
        ++upstr_blocked_queue_count;
      }
      DEBUG_STREAM2("End< Send Message" << endl);
      return true;
    } else {
      ERROR_STREAM2("Trying to send Event through not connected Jack!" << endl);
      return false;
    }
  }

  bool GenericNode::checkAnalysis() {
    // this method compacts code, required at two places in the
    // run()-method
    // it does all checks regarding data analysis, updates the flags
    // upon those checks and eventuallly returns true, signaling that
    // the run()-method should break out of its endless loop (this is
    // required for the concret analysis, where the run()-method is
    // trigged by GenericNode::analyseData() invoked from
    // doInitOutput() within the application thread)

    if (analyse_processing_flag) {
      DEBUG_STREAM2("in analysis procession!" << endl);
      if (!getWorkingFlag()) {
        if (analysis_data_forwarded) {
          // no more working flag and a Message was generated for the
          // requesting Node, so we disable analysis_processing.
          DEBUG_STREAM2("no more WorkingFlag! Stoping analysis production." << endl);
          analysis_data_forwarded = false;
          setAnalyseProcessingFlag(false);

          // now we can send the previously cached Message for the
          // requesting Node
          if (!sendAnalysingCacheEntryTo(*current_analysis_out_stream)) {
            cerr << "Could not send AnalysingCacheEntry to " << current_analysis_out_stream->getMaster()->getJackTag() << " although there should be at least one entry!" << endl;
          }
        }

        else if (send_data_failed) {
          // we received a "INode::sendDataFailed" Event as reply
          // for for our request for data (we are in
          // analyse_processing, which means, we don't have some
          // cached data from previous analysises and though have to
          // request some data from our predecessors, process it and
          // send it to the Node, requesting data from us).
          send_data_failed = false;
          // this cerr might be useful for debugging. If it does
          // apear, there might be a error in the graph communication
          cerr << "Processing Node " << getName() << " did not get data from its predecessor for its successor. What should we do? - sending failed!" << endl;
          // we inform our successor, that it won't get data from us ...
          if (!transmitSendDataFailed(*current_analysis_out_stream)) {
            ERROR_STREAM2("Could not send " << INode::sendDataFailed_event << " to output stream although he requested data." << endl);
          }
          // ... and stop analysis_processing
          setAnalyseProcessingFlag(false);
        }            

        else {
          // no more working_flag and no send_data_failed: 
          // -> we have to request more data
          // round-robin again like in GenericNode::sendData(), last
          // case
          pair<string, InputJack*> next_analysis_input(getNextAnalysisInput());
          current_analysis_in_stream = next_analysis_input.second;

          MESSAGE_STREAM2("Sending " << INode::sendData_event << " to input stream " << next_analysis_input.first << "." << endl);
          if (!transmitSendData(*current_analysis_in_stream)) {
            ERROR_STREAM2("Could not send " << INode::sendData_event << " to input stream " << next_analysis_input.first << "." << endl);
            if (!transmitSendDataFailed(*current_analysis_out_stream)) {
              ERROR_STREAM2("Could not send " << INode::sendDataFailed_event << " to output stream although he requested data." << endl);
            }
            // sending data request failed, we disable analyse_processing again
            setAnalyseProcessingFlag(false);
          }
        }
      } // end> !getWorkingFlag()

      else {
        DEBUG_STREAM("WorkingFlag in analysis-procssing. Let the Node finish its work!" << endl);
      }
    }

    else if (getAnalysingFlag()) {
      DEBUG_STREAM2("in analysis!" << endl);
      // if we are analysing, we update the analysing_flag here.
      updateAnalysingFlag();
      // after that, we can see, if the concret Node has finished its
      // analysis as after GenericNode::updateAnalysingFlag(), the
      // analysing_flag is false if the concret Node as set it so.
      if (!getAnalysingFlag()) {
        // we finished analysing
        DEBUG_STREAM2("Analysed Data!" << endl);
        return true;  // this will break out of run()!
        // this is ok, as run() was triggered by
        // GenericNode::analysData(), which is waiting for run()'s
        // termination
        // it will then recognize the finished analysis and break out
        // of the analysis iterations on its part.
      } else if (getWorkingFlag()) {
        // working flag and not yet analysed
        DEBUG_STREAM2("Continue Analysing!" << endl);
        // so we should continue analysing
      } else {
        // no working flag and not yet analysed, so we want MORE!
        DEBUG_STREAM2("Need more Data!" << endl);
        return true;  // this will break out of run()!
        // this is ok, as run() was triggered by
        // GenericNode::analysData(), which is waiting for run()'s
        // termination
        // it will then recognize the not finished analysis and continue
        // with the next analysis iterations (if there are some left!).
      }
    } else if (analyse_producing_flag) {
      // we have mad it complete through run() but this SourceNode has
      // not send a Message yet, so we retry it
      DEBUG_STREAM2("SourceNode needs more calls to produce data." << endl);
    }

    return false; // this will continue the run() loop!
  }

  bool GenericNode::sendAnalysingCacheEntryTo(JackGroup& out_jack) {
    // this method tries to send a Message, taken from the
    // analysing_cache, to the specified output
    DEBUG_STREAM2("Sending cached message to " << out_jack.getMaster()->getJackTag() << endl);
    // at first, we search the first analysing_cache entry, whose
    // Message should be send to the specified output (at least!).
    list<AnalysingCacheEntry>::iterator cache_entry;
    list<JackGroup*>::iterator cache_entry_jack;
    for (cache_entry = analysing_cache->begin();
         cache_entry != analysing_cache->end();
         cache_entry++) {
      for (cache_entry_jack = cache_entry->jacks.begin();
           cache_entry_jack != cache_entry->jacks.end();
           cache_entry_jack++) {
        if ((*cache_entry_jack) == &out_jack) {
          break;
        }
      }
      if (cache_entry_jack != cache_entry->jacks.end()) {
        break;
      }
    }

    if (cache_entry != analysing_cache->end()) {
      // we found a cached Message, so sending it
      Message* message(&(cache_entry->message));

      DEBUG_STREAM2("Erasing current send jack from list: " << cache_entry->jacks.size() << endl);
      cache_entry->jacks.erase(cache_entry_jack);

      if (cache_entry->jacks.size() > 0) {
        // the Message must be sent to multiple outputs, making a copy
        message = message->copy();
      } else {
        // no other jacks (any more), removing now unneeded cache entry
        analysing_cache->erase(cache_entry);
        if (analysing_cache->empty()) {
          // if this was the last analysing_cache entry at all, we
          // can destroy the analysing_cache object
          delete analysing_cache;
          analysing_cache = NULL;
        }
      }

      // now we can really send the found Message (through the
      // specified output)
      if (out_jack.sendMessage(message) == BLOCKED) {
        WARNING_STREAM2("Jack returned blocked!" << endl);
        ++downstr_blocked_queue_count;
      }
      DEBUG_STREAM2("Succeded!" << endl);
      return true;
    } else {
      WARNING_STREAM2("Failed to send cached message." << endl);
      // we did not find a analysing_cache entry with a Message to be
      // sent through the specified output.
      return false;
    }
  }

  pair<string, InputJack*> GenericNode::getNextAnalysisInput() {
    // we compute the next input for sending a "INode::sendData" Event
    // round-robin
    bool found(false);
    while (!found) {
      // searching for next input:
      if (next_analysis_in_stream == input_streams.end()) {
        // we are at the end of inputs, continue with beginning
        DEBUG_STREAM2("Continueing with first input stream!" << endl);
        next_analysis_in_stream = input_streams.begin();
        // this continue makes sure, the newly set
        // "nex_analysis_in_stream" fits all tests here!
        continue;
      }
      if (!next_analysis_in_stream->second->enabled) {
        // not enabled inputs are not suitable for requesting data
        // from, continueing with next one
        DEBUG_STREAM2("Stream \"" << next_analysis_in_stream->second->jack->getJackTag() << "\" is not enabled, skipping!" << endl);
        next_analysis_in_stream++;
        // this continue makes sure, the newly set
        // "nex_analysis_in_stream" fits all tests here!
        continue;
      }
      if (!next_analysis_in_stream->second->needed) {
        // not enabled inputs are not suitable for requesting data
        // from, continueing with next one
        DEBUG_STREAM2("Stream \"" << next_analysis_in_stream->second->jack->getJackTag() << "\" is not needed, skipping!" << endl);
        next_analysis_in_stream++;
        // this continue makes sure, the newly set
        // "nex_analysis_in_stream" fits all tests here!
        continue;
      }
      if (!next_analysis_in_stream->second->jack->isConnected()) {
        // not connected inputs are not suitable for requesting data
        // from, continueing with next one
        DEBUG_STREAM2("Stream \"" << next_analysis_in_stream->second->jack->getJackTag() << "\" is not connected, skipping!" << endl);
        next_analysis_in_stream++;
        // this continue makes sure, the newly set
        // "nex_analysis_in_stream" fits all tests here!
        continue;
      }
      // if we are here, next_analysis_in_stream fits all three above
      // conditions, so we can use it as next analysis input stream.
      found = true;
    }

    // compute the return value pair:
    // 1. the jack_tag of the next analysis input.
    // 2. the Jack object of the next analysis input.
    pair<string, InputJack*> retval(next_analysis_in_stream->first,
                                    next_analysis_in_stream->second->jack);

    // compute the next value for the next call of this method. If it
    // fits all three conditions (above), will be tested when it
    // should be used (maybe the concret Node removes or disables a
    // input meanwhile).
    if (++next_analysis_in_stream == input_streams.end()) {
      next_analysis_in_stream = input_streams.begin();
    }

    return retval;
  }
  /******************** analyseData & Co. ********************/

  Result GenericNode::sendMessageUpstream(const CEvent& cevent) {
    /* If no output_stream available no upstream message can be send */
    if (output_streams.empty()) {
      ERROR_STREAM("No OutputJack available to send instream message from application" << endl);
      return FAILURE;
    }

    //allocate a new CEvent to put in inside a queue
    CEvent* instream_event = new CEvent(cevent);
    instream_event->setDirection(Message::UPSTREAM);

    //Put the upstream message in the first output-stream.
    //Dont use the real sender because a Message is discarded in the Jacks if the sender is !=0 && != partner_jack
    output_streams.begin()->second->jack_group->getMaster()->receiveMessage(instream_event, 0);

    //Return always SUCCESS because the used queues have only one signalreceiver.
    //So an QUEUE_FREE_SLOT is never send to this sender.
    return SUCCESS;
  }

  Result GenericNode::sendMessageDownstream(const CEvent& cevent) {
    /* If no insput_stream available no downstream message can be send */
    if (input_streams.empty()) {
      ERROR_STREAM("No InputJack available to send instream message from application" << endl);
      return FAILURE;
    }

    //allocate a new CEvent to put in inside a queue
    CEvent* instream_event = new CEvent(cevent);
    instream_event->setDirection(Message::DOWNSTREAM);

    //Put the upstream message in the first output-stream.
    //Dont use the real sender because a Message is discarded in the Jacks if the sender is !=0 && != partner_jack
    input_streams.begin()->second->jack->receiveMessage(instream_event, 0);

    //Return always SUCCESS because the used queues have only one signalreceiver.
    //So an QUEUE_FREE_SLOT is never send to this sender.
    return SUCCESS;
  }

  NMMReference GenericNode::getBufferManager()
  {
    MESSAGE_STREAM("Giving out BufferManager!" << endl);
    return getReference();
  }

  Result GenericNode::useBufferManager(const IBufferManager& manager)
  {
    if(isLocalTo(manager)) {
      DEBUG_STREAM2("useBufferManager isLocalTo == true" << endl);
      // get NMMReference from interface
      NMMReference ref = (const_cast<IBufferManager&>(manager)).getBufferManager();
      // get instance
      IExternalBufferManager* bm = SReferenceManager::getInstance().getObject<IExternalBufferManager>(ref);
      if(bm) {
        // set buffer manager
        // (IExternalBufferManager::requestBufferManager will increase ref count of BufferManager-object,
        //  the ref-count will be decreased again in the next setBufferManager call or in the destructor
        //  of IExternalBufferManager, so now extra method for releasing is needed)
        setBufferManager(bm->requestBufferManager());
        DEBUG_STREAM2("useBufferManager SUCCESS " << bm << " " << bm->requestBufferManager() << endl);
        return SUCCESS;
      }
      else {
        WARNING_STREAM2("useBufferManager FAILURE could not get IExternalBufferManager" << bm);
        return FAILURE;
      }
    }
    else {
      DEBUG_STREAM2("useBufferManager FAILURE isLocalTo == false");
      return FAILURE;
    }
  }

  void GenericNode::releaseBufferManager()
  {
    // explicitly release current buffer manager (will decrease ref count of BufferManager-object)
    setBufferManager(0);
  }

  Buffer* GenericNode::getNewBuffer(const size_t size) {

    // get NMMReference for BufferManager
    NMMReference ref = getBufferManager();
    // get instance
    IExternalBufferManager* bm = SReferenceManager::getInstance().getObject<IExternalBufferManager>(ref);

    Buffer* out_buffer;

    if (bm) {
      out_buffer = bm->getNewBuffer(size);

      // Start delay measurement
      startDelayMeasurement(IDelayMeasurement::START_WITH_NEW_MESSAGE, out_buffer);
    } else {
      WARNING_STREAM2("useBufferManager FAILURE could not get IExternalBufferManager" << bm);
    }

    return out_buffer;
  }

  Buffer* GenericNode::getNewBuffer(const size_t size, const Buffer* buffer) {

    // get NMMReference for BufferManager
    NMMReference ref = getBufferManager();
    // get instance
    IExternalBufferManager* bm = SReferenceManager::getInstance().getObject<IExternalBufferManager>(ref);

    Buffer* out_buffer;

    if (bm) {
      out_buffer = bm->getNewBuffer(size, buffer);

      // Start delay measurement
      startDelayMeasurement(IDelayMeasurement::START_WITH_NEW_MESSAGE, out_buffer);
    } else {
      WARNING_STREAM2("useBufferManager FAILURE could not get IExternalBufferManager" << bm);
    }

    return out_buffer;
  }

  void GenericNode::setOutputFormat(const Format& format, const string& jack_tag)
    throw (const Exception&)
  {
    /* configure the new output format using the getOutputJack method */
    if ( getOutputJack(&format, jack_tag.c_str()) == 0 ) {
      ERROR_STREAM("No suitable output jack for given jack tag defined" << endl);
      throw Exception("No suitable output jack for given jack tag defined");
    }
  }

  void GenericNode::setDelayStartEnabled(const IDelayMeasurement::StartTimepoint& start_timepoint, const nmm_bool& enabled)
  {
    MutexGuard delay_guard(delay_mutex);
    unsigned int mask = (1 << start_timepoint);
    if(enabled)
    {
      delay_start_enabled = (delay_start_enabled & ~mask) | mask;
    }
    else
    {
      delay_start_enabled = (delay_start_enabled & ~mask);
    }
  }

  void GenericNode::setDelayStopEnabled(const IDelayMeasurement::StopTimepoint& stop_timepoint, const nmm_bool& enabled)
  {
    MutexGuard delay_guard(delay_mutex);
    unsigned int mask = (1 << stop_timepoint);
    if(enabled)
    {
      delay_stop_enabled = (delay_stop_enabled & ~mask) | mask;
    }
    else
    {
      delay_stop_enabled = (delay_stop_enabled & ~mask);
    }
  }

  void GenericNode::setDelayStartEnabled2(const string& start_timepoint2, const nmm_bool& enabled)
  {
    IDelayMeasurement::StartTimepoint start_timepoint;
    if(start_timepoint2 == "START_WITH_NEW_MESSAGE")
    {
      start_timepoint = IDelayMeasurement::START_WITH_NEW_MESSAGE;
    }
    else if(start_timepoint2 == "START_WITH_SEND_MESSAGE")
    {
      start_timepoint = IDelayMeasurement::START_WITH_SEND_MESSAGE;
    }
    else if(start_timepoint2 == "START_WITH_RECEIVE_MESSAGE")
    {
      start_timepoint = IDelayMeasurement::START_WITH_RECEIVE_MESSAGE;
    }
    else if(start_timepoint2 == "START_WITH_DEQUEUE_MESSAGE")
    {
      start_timepoint = IDelayMeasurement::START_WITH_DEQUEUE_MESSAGE;
    }
    else
    {
      ERROR_STREAM1("Invalid delay measurement start timepoint: " << start_timepoint2 << endl);
      return;
    }

    setDelayStartEnabled(start_timepoint, enabled);
  }

  void GenericNode::setDelayStopEnabled2(const string& stop_timepoint2, const nmm_bool& enabled)
  {
    IDelayMeasurement::StopTimepoint stop_timepoint;
    if(stop_timepoint2 == "STOP_WITH_SEND_MESSAGE")
    {
      stop_timepoint = IDelayMeasurement::STOP_WITH_SEND_MESSAGE;
    }
    else if(stop_timepoint2 == "STOP_WITH_RECEIVE_MESSAGE")
    {
      stop_timepoint = IDelayMeasurement::STOP_WITH_RECEIVE_MESSAGE;
    }
    else if(stop_timepoint2 == "STOP_WITH_DEQUEUE_MESSAGE")
    {
      stop_timepoint = IDelayMeasurement::STOP_WITH_DEQUEUE_MESSAGE;
    }
    else if(stop_timepoint2 == "STOP_WITH_PRESENT_MESSAGE")
    {
      stop_timepoint = IDelayMeasurement::STOP_WITH_PRESENT_MESSAGE;
    }
    else
    {
      ERROR_STREAM1("Invalid delay measurement stop timepoint: " << stop_timepoint2 << endl);
      return;
    }

    setDelayStopEnabled(stop_timepoint, enabled);
  }

  nmm_bool GenericNode::isDelayStartEnabled(const IDelayMeasurement::StartTimepoint& start_timepoint) const
  {
    MutexGuard delay_guard(delay_mutex);
    unsigned int mask = (1 << start_timepoint);
    return ((delay_start_enabled & mask) != 0);
  }

  nmm_bool GenericNode::isDelayStopEnabled(const IDelayMeasurement::StopTimepoint& stop_timepoint) const
  {
    MutexGuard delay_guard(delay_mutex);
    unsigned int mask = (1 << stop_timepoint);
    return ((delay_stop_enabled & mask) != 0);
  }

  Interval GenericNode::getAverageDelay(const IDelayMeasurement::StopTimepoint& stop_timepoint)
  {
    MutexGuard delay_guard(delay_mutex);
    return average_delay[stop_timepoint];
  }

  Interval GenericNode::getMinDelay(const IDelayMeasurement::StopTimepoint& stop_timepoint)
  {
    MutexGuard delay_guard(delay_mutex);
    return min_delay[stop_timepoint];
  }

  Interval GenericNode::getMaxDelay(const IDelayMeasurement::StopTimepoint& stop_timepoint)
  {
    MutexGuard delay_guard(delay_mutex);
    return max_delay[stop_timepoint];
  }

  void GenericNode::resetDelayMeasurement(const IDelayMeasurement::StopTimepoint& stop_timepoint)
  {
    MutexGuard delay_guard(delay_mutex);

    // Discard all samples
    delays[stop_timepoint].clear();

    // Reset minimum, maximum and average delays
    min_delay[stop_timepoint] = Interval();
    max_delay[stop_timepoint] = Interval();
    average_delay[stop_timepoint] = Interval();
  }

  void GenericNode::setNumberOfDelaySamples(const nmm_uint32& number_of_samples)
  {
    MutexGuard delay_guard(delay_mutex);
    max_delay_samples = number_of_samples;
  }

  void GenericNode::setDelayLoggingEnabled(const nmm_bool& enabled)
  {
    MutexGuard delay_guard(delay_mutex);
    delay_logging_enabled = enabled;
    delay_logging_counter = 0;
  }

  void GenericNode::setDelayLoggingInterval(const nmm_uint32& interval)
  {
    MutexGuard delay_guard(delay_mutex);
    delay_logging_interval = interval;
    delay_logging_counter = 0;
  }

  void GenericNode::startDelayMeasurement(IDelayMeasurement::StartTimepoint timepoint, Message* msg)
  {
    MutexGuard delay_guard(delay_mutex);

    // Check if delay measurement is enabled for this timepoint
    unsigned int mask = (1 << timepoint);
    if((delay_start_enabled & mask) == 0) return;

    // Measure current time
    Time delay_time = TimedElement::getTime();

    // Set delay timestamp
    msg->setDelayStartTime(delay_time);
  }

  void GenericNode::stopDelayMeasurement(IDelayMeasurement::StopTimepoint timepoint, Message* msg, Interval additional_delay)
  {
    // Only do the delay calculation if the delay timestamp is really set
    // (No need to lock the mutex yet)
    if(!msg->isDelayStartTimeSet()) return;

    MutexGuard delay_guard(delay_mutex);

    // Check if delay measurement is enabled for this timepoint
    unsigned int mask = (1 << timepoint);
    if((delay_stop_enabled & mask) == 0) return;

    // Get the delay start time of the message
    Time delay_time = msg->getDelayStartTime();

    // Measure current time
    Time now = TimedElement::getTime();

    // Compute delay and add additional delay
    Interval delay = (now - delay_time);
    delay += additional_delay;

    // Update minimum delay
    if (min_delay[timepoint] == Interval() || delay < min_delay[timepoint])
    {
      min_delay[timepoint] = delay;
    }

    // Update maximum delay
    if (delay > max_delay[timepoint])
    {
      max_delay[timepoint] = delay;
    }

    // Add delay to the delay array for average calculation
    delays[timepoint].push_front(delay);

    // Limit size of array
    while(delays[timepoint].size() > max_delay_samples)
    {
      delays[timepoint].pop_back();
    }

    // Recompute average delay
    size_t num_delays = 0;
    Interval total_delay;
    for(deque<Interval>::const_iterator i = delays[timepoint].begin(); i != delays[timepoint].end(); ++i)
    {
      ++num_delays;
      total_delay += *i;
    }

    // Average delay is only valid if the total number of samples is available
    // TODO: fw: Why?
    if(num_delays == 0)
    {
      average_delay[timepoint] = Interval();
    }
    else
    {
      average_delay[timepoint] = total_delay;
      average_delay[timepoint] *= (1.0 / double(num_delays));
    }

    // Reset the delay start time of the message
    msg->unsetDelayStartTime();

    // BEGIN> Delay logging
    if(delay_logging_enabled)
    {
      if(delay_logging_counter == delay_logging_interval)
      {
        // Increase counter
        delay_logging_counter = 0;

        // Describe start of measurement
        // DOESNTWORK: fw: We do not know when delay measurement of a message was started
        // DOESNTWORK: fw: (unless it was started at the same node)
        string from_string;
//        if(delay_start_timepoint == IDelayMeasurement::START_WITH_NEW_MESSAGE)
//        {
//          from_string = string("from new message");
//        }
//        else if(delay_start_timepoint == IDelayMeasurement::START_WITH_RECEIVE_MESSAGE)
//        {
//          from_string = string("from receive message");
//        }
//        else if(delay_start_timepoint == IDelayMeasurement::START_WITH_DEQUEUE_MESSAGE)
//        {
//          from_string = string("from dequeue message");
//        }
//        else if(delay_start_timepoint == IDelayMeasurement::START_WITH_SEND_MESSAGE)
//        {
//          from_string = string("from send message");
//        }

        // Describe end of measurement
        string to_string;
        if(timepoint == IDelayMeasurement::STOP_WITH_RECEIVE_MESSAGE)
        {
          from_string = string("to receive message");
        }
        else if(timepoint == IDelayMeasurement::STOP_WITH_DEQUEUE_MESSAGE)
        {
          from_string = string("to dequeue message");
        }
        else if(timepoint == IDelayMeasurement::STOP_WITH_SEND_MESSAGE)
        {
          from_string = string("to send message");
        }
        else if(timepoint == IDelayMeasurement::STOP_WITH_PRESENT_MESSAGE)
        {
          from_string = string("to present message");
        }

        // Display log message
        cerr << "LOGDELAY: " << from_string
             << " " << to_string
             << " delay: " << (double(delay) * 1000) << " ms"
             << " min: " << (double(min_delay[timepoint]) * 1000) << " ms"
             << " max: " << (double(max_delay[timepoint]) * 1000) << " ms"
             << " average: " << (double(average_delay[timepoint]) * 1000) << " ms"
             << endl;
      }
      else
      {
        // Increase delay logging counter
        ++delay_logging_counter;
      }
    }

    return;
  }

}


#include "nmm/nmm_suffix.hpp"
