/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "InputJack.hpp"
#include "Node.hpp"
#include "OutputJack.hpp"
#include "nmm/comm/InstreamChannel.hpp"
#include "INodeImpl.hpp"
#include "nmm/comm/IOutputJack.hpp"
#include "nmm/comm/TransportStrategy.hpp"
#include "nmm/comm/TSFactory.hpp"
#include "GenericNode.hpp"
#include <iostream>
#include "nmm/comm/ReferenceManager.hpp"
#include "nmm/comm/ConnectionManager.hpp"
#include "nmm/comm/FormatStrategyMap.hpp"
#include "nmm/comm/InstreamCompositeStrategy.hpp"
#include "nmm/comm/messaging/ITSOneWayBufferSender.hpp"
#include "nmm/comm/messaging/ITSOneWayBufferReceiver.hpp"
#include "nmm/comm/IObjectAcceptor.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  InputJack::InputJack(Node* parent, const string& jack_tag, const string& name)
    :  Jack(parent, jack_tag, name),
       __transport_strategy(0),
       __i_transport_strategy(0)
  {
    data_jack    = this;
  }

  InputJack::~InputJack() {
    if(__transport_strategy) {
      /* first delete transport strategy interface */
      if(__i_transport_strategy) {
        delete __i_transport_strategy;
        __i_transport_strategy = 0;
      }
      __transport_strategy->scheduleDestruction();
    }
  }

  nmm_bool InputJack::isConnected() const {
    return (getPartner() || data_partner.get()) &&
      (getMessageReceiver() || getOwnTransportStrategy());
  }
  
  void InputJack::connectTo(const OutputJack* out_jack) const
    throw (const BindingException&)
  {
    if(out_jack == 0) {
      throw BindingException(this, out_jack, "OutputJack is 0");
    }
    else {
      if (getPartner() == 0) {
        setPartner(const_cast<OutputJack*>(out_jack));
        const_cast<InputJack*>(this)->setMessageReceiver(const_cast<OutputJack*>(out_jack));
        NODE_DEBUG_STREAM2( (getParent()) , "Valid Partner connected" << endl);
      } else {
        NODE_WARNING_STREAM2( (getParent()), "Jack already connected" << endl);
        throw BindingException(this, out_jack, "InputJack already connected");
      }
    }
  }

  void InputJack::disconnect()
  {
    if(getParent()->getState() == INode::STARTED) {
    }
    setPartner(0);
    setMessageReceiver(0);
    if(__transport_strategy) {
      /* first delete transport strategy interface */
      if(__i_transport_strategy) {
        delete __i_transport_strategy;
        __i_transport_strategy = 0;
      }
      __transport_strategy->scheduleDestruction();
      
      __transport_strategy = 0;
    }
    data_partner.reset(0);

  }
  
  void InputJack::setTransportStrategy(InstreamCompositeStrategy* strategy)
    throw(const NMM::BindingException &)
  {
    /* Begin> If an old strategy is available, release it. */
    if(__transport_strategy) {
      /* first delete transport strategy interface */
      if(__i_transport_strategy) {
        delete __i_transport_strategy;
        __i_transport_strategy = 0;
      }
      __transport_strategy->scheduleDestruction();
    }
    /* End< */
    
    /* stores the new strategy */
    __transport_strategy = strategy;
    
    // XXX: fw: TransportStrategy is not a NamedObject
//     /* set name of strategy */
//     __transport_strategy->setName(getName() + "> " + strategy->getName());
    
    /* Begin> Handle downstream messages */
    /* If the strategy receives downstream-messages, the input_jack receives it*/
    strategy->setMessageReceiver(this);
    /* End< */

    /* Begin> Handle upstream messages. */
    /* If the the jack receives upstream-messages from its parent they are forwarded to the given strategy */
    setMessageReceiver(0);
    /* Signals from the stratgey, like FREE_QUEUE_SLOT are send to the jacks parent*/
    strategy->setSignalReceiver(this);
    /* End< */
    
    /* Get the buffer strategies to configure them */
    ITSOneWayBufferSender*   buffer_sender   = strategy->getOneWayBufferSender();
    ITSOneWayBufferReceiver* buffer_receiver = strategy->getOneWayBufferReceiver();
    if(!buffer_sender) {
      /* The TransportStrategy does not send buffers */
      NODE_ERROR_STREAM(getParent(), "TransportStrategy has no one-way buffer sender" << endl);        
      throw BindingException("No one-way buffer sender");
    }
    if(!buffer_receiver) {
      /* The TransportStrategy does not receive buffers */
      NODE_ERROR_STREAM(getParent(), "TransportStrategy has no one-way buffer receiver" << endl);        
      throw BindingException("No one-way buffer receiver");
    }
    
    /* Begin> Configure the TransportStrategie's Buffermanager */
    /* If a TransportStrategy receives a Buffer, it needs a BufferManager to request it. */
    if (getParent()) {
      /* Request the IExternalBuffermanager */
      IExternalBufferManager* buffer_manager = dynamic_cast<IExternalBufferManager*>(getParent());
      if(!buffer_manager) {
        /* The TransportStrategy can not receive messages */
        NODE_ERROR_STREAM(getParent(), "Parent does not implement IExternalManager" << endl);
        throw BindingException("No buffer manager");
      }
      
      /* and tell the strategy to use it */
      buffer_sender->setBufferManager(buffer_manager);
    }
    /* End< */
    
    /* Begin> Configure the TransportStrategy's Format */
    if (getParent()) {
      const Format* format = getParent()->getInputFormat(getJackTag().c_str());
      if(!format) {
        NODE_ERROR_STREAM(getParent(), "Jack has no format" << endl);
        throw BindingException("No format");
      }
      DEBUG_STREAM2("Configure InputJack with format " << format->getFormatType() << endl);
      buffer_sender->configure(*format);
      buffer_receiver->configure(*format);
    }
    /* End< */
  }
  
  void InputJack::setTransportStrategy(const TSAddress* address)
    throw(const NMM::BindingException &)
  {
    /* If the address is invalid throw an Exception */
    if (!address)
      throw BindingException("Invalid Address (0-ptr)");

    /* Begin> Handle local case!
     *        For performance reason, the jack communicates with its jack partner without a
     *        TransportStrategy in a local case.
     */
    /* If the address describes a local binding, its a NMMReference describes an object in this address space*/
    if (address->getBindingType() == TSAddress::LOCAL) {
      /* Request the NMMObject with NMMReference in the given address. The type of this Object must be
       * OutputJack
       */
      OutputJack* output_jack = SReferenceManager::getInstance().getObject<OutputJack>( address->getNMMReference() );

      /* If the address points not to an OutputJack, throw an Exception */
      if (!output_jack) {
        throw BindingException("No such OutputJack available");
      }
      
      /* Connect this jack with Call the connectTo method for this jack */
      connectTo(output_jack);
      /* Connect the output_jack with this jack */
      output_jack->connectTo(this);
      /* The output_jack is our partner and receives Messages from out parent */
      setMessageReceiver(output_jack);
      return;
    }
    /* End< */

    /* Begin> Handle distributed case */
    /* 1. Create a TransportStrategy which can connect to the given address */
    InstreamCompositeStrategy* strategy = new InstreamCompositeStrategy();
    
    /* If this jack can not create such a TransportStrategy throw a BindingException */
    if (!strategy)
      throw  BindingException("Can not create strategy with type "+ address->getStrategyType());

    /* 2. The strategy connects to the given address */
    strategy->setAddress(*address);

    /* 3. The connection between this object and the other object is used to send instream messagess.
     *    So configure the jack to the new strategy*/
    setTransportStrategy(strategy);

    /* 4. Tell the other jack to use it for instream message communication too.*/
    Event event(IObjectAcceptor::connectToJack_event, new TInValue<NMMReference>(address->getNMMReference()));
    if(strategy->sendReliableEvent(event) != SUCCESS) {
      /* Begin> Error handling */   
      if(event.getException()) {
        NODE_DEBUG_STREAM2( getParent() , *event.getException() << endl);
      } else {
        NODE_DEBUG_STREAM2( (getParent()) , "could not set TS in InputJack" << endl);
      }
      /* End< */
    }
  }
  
  InstreamChannel* InputJack::connectTo(const IOutputJack& partner)
      throw(const BindingException& )
  {
    // setPartner(0); // should not be needed
    data_partner.reset(dynamic_cast<IOutputJack*>(partner.copy()));
    IInputJack_var tmp_jack(getInterface<IInputJack>());
    return new InstreamChannel(*tmp_jack.get(), *data_partner.get());
  }

  IInstreamCompositeStrategy* InputJack::getTransportStrategy() const {
    /* if we don't already have an interface, request one */
    if(!__i_transport_strategy) {
      if(__transport_strategy) {
        __i_transport_strategy = __transport_strategy->getInterface<IInstreamCompositeStrategy>();
      }
    }
    
    /* return a copy of the interface */
    if(__i_transport_strategy) {
      return dynamic_cast<NMM::IInstreamCompositeStrategy*>(__i_transport_strategy->copy());
    } else {
      return 0;
    }
  }

  IOutputJack* InputJack::getDataPartner()
  {
    if (data_partner.get()) {
      return dynamic_cast<IOutputJack*>(data_partner.get()->copy()); 
    }
    else {
      return 0;
    }
  }
  
  TSAddress* InputJack::getDefaultAddress(const string& strategy_type) const {
    return NMMObject::getDefaultAddress(strategy_type);
  }
  
  Result InputJack::getSupportedStrategies(list<TSDescription>& strategy_list) const {
    const Format* format = getParent()->getInputFormat(getJackTag().c_str());
    
    strategy_list.clear();
    list<TSDescription> temp = STSFactory::getInstance().getSupportedStrategies();
    for(list<TSDescription>::const_iterator i=temp.begin(); i != temp.end(); ++i) {
      /* check if the format is supported. if the jack has no format,
         the format is not supported. (this case only happens during fast
         disconnect) */
      bool is_format_supported = false;
      if(format) {
        is_format_supported = SConnectionManager::getInstance().getFormatStrategyMap().isSupported(*format, *i);
      }
      
      /* modify the interfaces bitfield and conditionally exclude
         instream interfaces. */
      unsigned int interfaces = i->getInterfaces();
      if(!is_format_supported) {
        interfaces &= ~(TSDescription::ITS_ONE_WAY_BUFFER_SENDER | TSDescription::ITS_ONE_WAY_BUFFER_RECEIVER);
      }
      
      /* if any interfaces are supported, add the transport strategy */
      if(interfaces != 0) {
        TSDescription temp(*i);
        temp.setInterfaces(interfaces);
        strategy_list.push_back(temp);
      }
    }
    
    return SUCCESS;
  }
}

#include "nmm/nmm_suffix.hpp"
