/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "InputProducer_Linux.hpp"
#include "nmm/utils/NamedObject.hpp"

#include "nmm/Exception.hpp"

#include <sys/poll.h>

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  InputProducer::InputProducer()
    : fd(-1),
      timeout(-1)
  {
    pipefd[0] = -1;
    pipefd[1] = -1;
  }
  
  InputProducer::~InputProducer() {
    /* wake up poll and deactivate thread */
    InputProducer::stop();
  }
  
  void InputProducer::start() {
    GLOBAL_DEBUG_STREAM("InputProducer: start" << endl);
    
    /* create internal pipe for interrupting poll() */
    if ( pipe(&pipefd[0]) != 0 ) {
      GLOBAL_ERROR_STREAM("Could not create the internal pipe" << endl);
      throw Exception("Could not create the internal pipe");
    }
    
    /* start the thread */
    activateThread();
  }
  
  void InputProducer::stop() {
    GLOBAL_DEBUG_STREAM("InputProducer: stop" << endl);
    
    /* write to internal pipe to interrupt poll() */
    if (pipefd[1] >= 0) {
      write(pipefd[1], &pipefd[1], 1);
    }
    
    /* stop the thread */
    deactivateThread();
    
    /* close internal pipe */
    if (pipefd[1] >= 0) {
      close(pipefd[0]);
      pipefd[0] = -1;
      close(pipefd[1]);
      pipefd[1] = -1;
    }
  }
  
  void InputProducer::setTimeout(int timeout) {
    this->timeout = timeout;
  }
  
  void InputProducer::setFileDescriptor(int fd) {
    this->fd = fd;
  }
  
  void InputProducer::run() {
    
    /* setup array of poll FDs (fd and internal pipe fd) */
    pollfd pollfds[] = {
      { fd       , POLLIN, 0 },
      { pipefd[0], POLLIN, 0 }
    };
    
    int retval = 0;
    while(isRunning()) {
      
      /* call poll on fd */
      retval = poll(pollfds, 2, timeout);
      
      /* check for error */
      if (retval < 0) {
        GLOBAL_ERROR_STREAM("InputProducer: Something went wrong during the poll() call" << endl);
      }
      
      /* if change on a fd call handler */
      if (retval > 0 ) {
        if (pollfds[0].revents & POLLIN) {
          handleInput();
        }
        
        if (pollfds[1].revents & POLLIN) {
          char dummy;
          read(pipefd[0], &dummy, 1);
          break;
        }
      }
      
    }
  }
  
}

#include "nmm/nmm_suffix.hpp"
