/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "JackGroup.hpp"
#include "Node.hpp"
#include <iostream>
#include "nmm/comm/messaging/Message.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  JackGroup::JackGroup(Node* parent_node, QueuedOutputJack* jack) 
    : curr_dequeue_jack(0),
      parent(parent_node),
      jacks()
  {
    // FIXME: are added jacks always connected?
    InternalJack_t ijack(const_cast<QueuedOutputJack*>(jack), true, false);
    jacks.push_back(ijack);
  }

  JackGroup::~JackGroup() 
  {
    for(InternalJacks_t::iterator i = jacks.begin(); i != jacks.end(); ++i) {
      delete (*i).jack;
    }
  }

  int JackGroup::sendMessage(Message* message) const
  {
    int blocked_queue_count = 0;
    // save first (= master) jack of all jacks
    InternalJacks_t::const_iterator first = jacks.begin();
    // if more than one jack -> send to all other (=slave) jacks
    if(first != jacks.end()) {
      InternalJacks_t::const_iterator i = first;
      for(++i; i!= jacks.end(); ++i) {
	NODE_DEBUG_STREAM2(parent,  "JackGroup> Send Buffer to cloned jack "<< endl);
	// BLOCKED means increment blocked_queue_count;
	if((*i).jack->sendMessage(message->copy()) == BLOCKED) {
	  blocked_queue_count++;
	}
      }
    }
      
    // important: first need to send to slave list since this increases the ref count
    // (otherwise, master stream might decrease ref count to 0 *before* any slave gets the message!)
    // send to master
    if((*first).jack->sendMessage(message) == BLOCKED) { 
      blocked_queue_count++;
    }
    
    return blocked_queue_count;
  }

  bool JackGroup::checkNewConnections() 
  {
    bool new_connections = false;
    // if some output was *not* connected before, but is connected now -> syncReset
    for(InternalJacks_t::iterator i = jacks.begin(); i!= jacks.end(); ++i) {
      if((*i).jack->isConnected()) {
	if((*i).connected == false) {
	  new_connections = true;
	  (*i).connected = true; // set to 'connected' now
	  (*i).switch_to = true; // this jack will get the 'switch_to' message
	}
	else { // jack was already connected
	  // (*i).switch_to = false; // this jack will get the 'switch_from' message
	}
      }
      else {
	(*i).connected = false; // set to 'disconnected' now
      }
    }
    return new_connections;
  }


  Result JackGroup::addJack(const QueuedOutputJack *jack) const 
  {
    // FIXME: are added jacks always *NOT* connected?
    InternalJack_t ijack(const_cast<QueuedOutputJack*>(jack), false, false);
    jacks.push_back(ijack);
    return SUCCESS;
  }

  Result JackGroup::removeJack(const QueuedOutputJack* jack)
  {
    // try to find pointer in jacks
    InternalJacks_t::iterator i;
    for(i = jacks.begin(); i != jacks.end(); ++i) {
      if(jack == (*i).jack) {
	break;
      }
    }
    // if pointer was found -> remove jack
    if(i != jacks.end()) {
      jacks.erase(i);
      return SUCCESS;
    }
    else {
      return FAILURE;
    }
  }

  Message* JackGroup::dequeueMessage()
  {
    Message* msg = 0;
    // check master and all jacks
    for(size_t i = 0; i < jacks.size(); ++i) {
      msg = jacks[curr_dequeue_jack].jack->dequeueMessage();
      // round robin: move to next jack
      curr_dequeue_jack++;
      
      // wrap around if all jacks where checked
      if(curr_dequeue_jack == jacks.size()) {
	curr_dequeue_jack = 0;
      }

      // found message somewhere -> return this message
      if(msg) {
	return msg;
      }
    }
    return msg;
  }

  Result JackGroup::flush()
  {
    Result r = SUCCESS;
    for (InternalJacks_t::const_iterator i = jacks.begin(); i!= jacks.end(); i++) {
      Result tmp = (*i).jack->flush();
      if(tmp != SUCCESS) {
	r = FAILURE;
      }
    }
    return r;
  }

  bool JackGroup::isConnected() 
  { 
    for (InternalJacks_t::const_iterator i = jacks.begin(); i!= jacks.end(); i++) {
      if((*i).jack->isConnected()) {
	return true;
      }
    }
    return false;
  }

  int JackGroup::sendSwitchMessages(Message* switch_to_msg, Message* switch_from_msg)
  {
    int blocked_queue_count = 0;

    // send switch_to and switch_from messages and return blocked count
    for(InternalJacks_t::iterator i = jacks.begin(); i!= jacks.end(); ++i) {
      if((*i).switch_to == true) {
	if(switch_to_msg != 0) {
	  NODE_DEBUG_STREAM2(parent,  "JackGroup> HANDOVER: switchOutputs : switch_to" << endl);
	  if((*i).jack->sendMessage(switch_to_msg->copy()) == BLOCKED) {
	    blocked_queue_count++;
	  }
	}
      }
      else {
	if(switch_from_msg != 0) {
	  NODE_DEBUG_STREAM2(parent,  "JackGroup> HANDOVER: switchOutputs : switch_from" << endl);
	  if((*i).jack->sendMessage(switch_from_msg->copy()) == BLOCKED) {
	    blocked_queue_count++;
	  }
	}
      }
    }

    // release messages (they were copied before sending them)
    if(switch_to_msg) {
      switch_to_msg->release();
    }
    if(switch_from_msg) {
      switch_from_msg->release();
    }

    return blocked_queue_count;
  }

  void JackGroup::resetSwitchTo(bool status /* = false */ )
  {
    for(InternalJacks_t::iterator i = jacks.begin(); i!= jacks.end(); ++i) {
      (*i).switch_to = status;
    }
  }

}

#include "nmm/nmm_suffix.hpp"
