/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include "Node.hpp"
#include "InputJack.hpp"
#include "OutputJack.hpp"
#include "nmm/comm/format/Property.hpp"
#include "nmm/utils/thread/MutexGuard.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM 
{

  Node::Node(const char* name) : 
    StdNamedObject(name),
    state_mutex(ThreadMutex::RECURSIVE)
  {
    input_property  = 0; 
    output_property = 0; 
    working_input_property  = 0; 
    working_output_property = 0; 
    node_type = INode::UNKNOWN;
    getEventDispatcher().changeState(INode::CONSTRUCTED);

    computation_config = 0;
    working_computation_config = 0;
  }

  Node::~Node()  {
    delete input_property;
    delete output_property;
    delete working_input_property;
    delete working_output_property;

    if (computation_config != 0) {
      delete computation_config;
    }
    if (working_computation_config != 0) {
      delete working_computation_config;
    }
  }

  Property* Node::getInternalInputProperty()          const {
    if (!input_property) {
      internal_mutex.lock();
      if(!input_property) {
        input_property = new Property();
      }
      internal_mutex.unlock();
    }
    return input_property;
  }

  Property* Node::getInternalOutputProperty()         const {
    if (!output_property) {
      internal_mutex.lock();
      if(!output_property) {
        output_property = new Property();
      }
      internal_mutex.unlock();
    }
    return output_property;
  }

  Property* Node::getInternalWorkingInputProperty()   const {
    if (!working_input_property) {
      internal_mutex.lock();
      if(working_input_property == 0) {
        working_input_property = new Property();
      }
      internal_mutex.unlock();
    }
    return working_input_property;
  }

  Property* Node::getInternalWorkingOutputProperty()  const {
    if (!working_output_property) {
      internal_mutex.lock();
      if(working_output_property == 0) {
        working_output_property = new Property();
      }
      internal_mutex.unlock();
    }
    return working_output_property;
  }


  Property *Node::getOwnInputProperty()  {
    return getInternalInputProperty();
  }

  Property *Node::getOwnOutputProperty() {
    return getInternalOutputProperty();
  }
  
  const Property *Node::getInputProperty()  const  {
    return new Property(*getInternalInputProperty());
  }

  Property *Node::getInputProperty() {
    return new Property(*getInternalInputProperty());
  }

  const Property *Node::getOutputProperty() const  {
    return new Property(*getInternalOutputProperty());
  }

  Property *Node::getOutputProperty()   {
    return new Property(*getInternalOutputProperty());
  }

  // ** working **

  Property *Node::getOwnWorkingInputProperty()  {
    return getInternalWorkingInputProperty();
  }

  Property *Node::getOwnWorkingOutputProperty() {
    return getInternalWorkingOutputProperty();
  }
  
  const Property *Node::getWorkingInputProperty()  const  {
    return new Property(*getInternalWorkingInputProperty());
  }

  Property *Node::getWorkingInputProperty()    {
    return new Property(*getInternalWorkingInputProperty());
  }

  const Property *Node::getWorkingOutputProperty() const  {
    return new Property(*getInternalWorkingOutputProperty());
  }

  Property *Node::getWorkingOutputProperty()   {
    return new Property(*getInternalWorkingOutputProperty());
  }


  void Node::setType(INode::Type type) {
    node_type = type;
  }

  INode::Type Node::getType() const    {
    return node_type;
  }

  void Node::reachState(const INode::State& state) 
    throw(const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&) 
  {
    if(getState() == INode::CONSTRUCTED) {
      if(state == INode::CONSTRUCTED) {
	return;
      }
      if(state == INode::INITIALIZED) {
	return init();
      }
      if(state == INode::OUTPUT_INITIALIZED) {
	init();
	return initOutput();
      }
      if(state == INode::ACTIVATED) {
	init();
	initOutput();
	return activate();
      }
      if(state == INode::STARTED) {
	init();
	initOutput();
	activate();
	return start();
      }
    }
    if(getState() == INode::INITIALIZED) {
      if(state == INode::CONSTRUCTED) {
	return deinit();
      }
      if(state == INode::INITIALIZED) {
	return;
      }
      if(state == INode::OUTPUT_INITIALIZED) {
	return initOutput();
      }
      if(state == INode::ACTIVATED) {
	initOutput();
	return activate();
      }
      if(state == INode::STARTED) {
	initOutput();
	activate();
	return start();
      }
    }
    if(getState() == INode::OUTPUT_INITIALIZED) {
      if(state == INode::CONSTRUCTED) {
	deinitOutput();
	return deinit();
      }
      if(state == INode::INITIALIZED) {
	return deinitOutput();
      }
      if(state == INode::OUTPUT_INITIALIZED) {
	return;
      }
      if(state == INode::ACTIVATED) {
	return activate();
      }
      if(state == INode::STARTED) {
	activate();
	return start();
      }
    }
    if(getState() == INode::ACTIVATED) {
      if(state == INode::CONSTRUCTED) {
	deactivate();
	deinitOutput();
	return deinit();
      }
      if(state == INode::INITIALIZED) {
	deactivate();
	return deinitOutput();
      }
      if(state == INode::OUTPUT_INITIALIZED) {
	return deactivate();
      }
      if(state == INode::ACTIVATED) {
	return;
      }
      if(state == INode::STARTED) {
	return start();
      }
    }
    if(getState() == INode::STARTED) {
      if(state == INode::CONSTRUCTED) {
	stop();
	deactivate();
	deinitOutput();
	return deinit();
      }
      if(state == INode::INITIALIZED) {
	stop();
	deactivate();
	return deinitOutput();
      }
      if(state == INode::OUTPUT_INITIALIZED) {
	stop();
	return deactivate();
      }
      if(state == INode::ACTIVATED) {
	return stop();
      }
      if(state == INode::STARTED) {
	return;
      }
    }
    
    string s = string("") + state;
    throw StateTransitionFailedException(this, getName()+"> unknown state "+s);
  }

  void Node::reachConstructed() 
    throw(const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&)
  {
    return reachState(INode::CONSTRUCTED);
  }

  void Node::reachInitialized() 
    throw(const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&)
  {
    return reachState(INode::INITIALIZED);
  }

  void Node::reachOutputInitialized() 
    throw(const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&)
  {
    return reachState(INode::OUTPUT_INITIALIZED);
  }

  void Node::reachActivated() 
    throw(const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&)
  {
    return reachState(INode::ACTIVATED);
  }

  void Node::reachStarted() 
    throw(const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&)
  {
    return reachState(INode::STARTED);
  }

  void Node::init()
    throw (const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&)
  {
    MutexGuard mg(state_mutex);
    
    getEventDispatcher().checkMandatoryEvents(INode::INITIALIZED); //throws UnsuccessfulRequiredCallException
    if(doInit() == SUCCESS) {
      getEventDispatcher().changeState(INode::INITIALIZED); //throws UnsuccessfulRequiredCallException
    }
    else {
      ERROR_STREAM2("doInit failed" << endl);
      throw StateTransitionFailedException(this, getName()+"> doInit() failed");
    }
  }

  void Node::initOutput()
    throw (const StateTransitionFailedException&,  const UnsuccessfulRequiredCallException&)
  {
    MutexGuard mg(state_mutex);

    getEventDispatcher().checkMandatoryEvents(INode::OUTPUT_INITIALIZED);//throws UnsuccessfulRequiredCallException
    if(doInitOutput() == SUCCESS) {
      getEventDispatcher().changeState(INode::OUTPUT_INITIALIZED);//throws UnsuccessfulRequiredCallException
    }
    else {
      ERROR_STREAM2("doInitOutput failed" << endl);
      throw StateTransitionFailedException(this, getName()+"> doInitOutput() failed");
    }
  }

  void Node::activate()
    throw (const StateTransitionFailedException&,  const UnsuccessfulRequiredCallException&)
  {
    MutexGuard mg(state_mutex);

    getEventDispatcher().checkMandatoryEvents(INode::ACTIVATED);//throws UnsuccessfulRequiredCallException
    if(doActivate() == SUCCESS) {
      getEventDispatcher().changeState(INode::ACTIVATED);//throws UnsuccessfulRequiredCallException
    }
    else {
      ERROR_STREAM2("doActivate failed"  << endl);
      throw StateTransitionFailedException(this, getName()+"> doActivate() failed");
    }
  }

  void Node::deinit()
    throw (const StateTransitionFailedException&,  const UnsuccessfulRequiredCallException&)
  {
    MutexGuard mg(state_mutex);

    getEventDispatcher().checkMandatoryEvents(INode::CONSTRUCTED);//throws UnsuccessfulRequiredCallException

    if(doDeinit() == SUCCESS) {
      // if successful, erase formats in input and output-property
      getOwnInputProperty()->erase();
      getOwnOutputProperty()->erase();
      getEventDispatcher().changeState(INode::CONSTRUCTED);//throws UnsuccessfulRequiredCallException
    }
    else {
      ERROR_STREAM2("doDeinit failed" << endl);
      throw StateTransitionFailedException(this, getName()+"> doDeinit() failed");
    }
  }

  void Node::deinitOutput()
    throw (const StateTransitionFailedException&,  const UnsuccessfulRequiredCallException&)
  {
    MutexGuard mg(state_mutex);

    getEventDispatcher().checkMandatoryEvents(INode::INITIALIZED);//throws UnsuccessfulRequiredCallException

    if(doDeinitOutput() == SUCCESS) {
      // if successful, erase formats in input and output-property (working)
      getOwnWorkingInputProperty()->erase();
      getOwnWorkingOutputProperty()->erase();
      getEventDispatcher().changeState(INode::INITIALIZED);//throws UnsuccessfulRequiredCallException
    }
    else {
      ERROR_STREAM2("doDeinitOutput failed" << endl);
      throw StateTransitionFailedException(this, getName()+"> deinitOutput() failed");
    }
  }

  void Node::deactivate()
    throw (const StateTransitionFailedException&,  const UnsuccessfulRequiredCallException&)
  {
    MutexGuard mg(state_mutex);

    getEventDispatcher().checkMandatoryEvents(INode::OUTPUT_INITIALIZED);//throws UnsuccessfulRequiredCallException

    if(doDeactivate() == SUCCESS) {
      getEventDispatcher().changeState(INode::OUTPUT_INITIALIZED);//throws UnsuccessfulRequiredCallException
    }
    else {
      ERROR_STREAM2("doDeactivate failed" << endl);
      throw StateTransitionFailedException(this, getName()+"> doDeactivate() failed");
    }
  }

  void Node::start()
    throw (const StateTransitionFailedException&,  const UnsuccessfulRequiredCallException&)
  {
    MutexGuard mg(state_mutex);

    getEventDispatcher().checkMandatoryEvents(INode::STARTED);//throws UnsuccessfulRequiredCallException

    if(doStart() == SUCCESS) {
      getEventDispatcher().changeState(INode::STARTED);//throws UnsuccessfulRequiredCallException
    }
    else {
      ERROR_STREAM2("doStart failed" << endl);
      throw StateTransitionFailedException(this, "> doStart() failed"); 
    }
  }

  void Node::stop()
    throw (const StateTransitionFailedException&,  const UnsuccessfulRequiredCallException&)
  {
    MutexGuard mg(state_mutex);

    getEventDispatcher().checkMandatoryEvents(INode::ACTIVATED);//throws UnsuccessfulRequiredCallException

    if(doStop() == SUCCESS) {
      getEventDispatcher().changeState(INode::ACTIVATED);//throws UnsuccessfulRequiredCallException
    }
    else {
      ERROR_STREAM2("doStop failed" << endl);
      throw StateTransitionFailedException(this, getName()+"> doStop() failed");
    }
  }

  Result Node::flush() 
  {
    MutexGuard mg(state_mutex);
    return doFlush();
  }

  void printFormatList(const_format_list& l) 
  {
    for(const_format_list::const_iterator i = l.begin(); i != l.end(); ++i) {
      GLOBAL_DEBUG_STREAM1(*(*i) << endl);
    }
  }

  IInputJack* Node::getInputJack(const Format& format, const string& jack)  
  {
    const InputJack* t_jack = getInputJack(&format, jack.c_str());
    ERROR_STREAM(t_jack << endl);
    if (!t_jack) {
      ERROR_STREAM("InputJack is 0" << endl);
      return 0;
    }

    if (t_jack)
      return const_cast<IInputJack*>(t_jack->getInterface<IInputJack>());
    else {
      ERROR_STREAM("Jack is not an SimpleInputJack" << endl);
      return 0;  
    }

  }

  IOutputJack* Node::getOutputJack(const Format& format, const string& jack) 
  {
    const OutputJack* o_jack = getOutputJack(&format, jack.c_str());    

    if (o_jack)
      return const_cast<IOutputJack*>(o_jack->getInterface<IOutputJack>());
    else {
      ERROR_STREAM("Jack is not a SimpleOutputJack" << endl);
      return 0;
    }

  }

  nmm_bool Node::isOutputConnected(const string& jack_tag) const {
    const Jack* jack = getOwnOutputStream(jack_tag);
    if (!jack) {
      // no such jack, so not connected
      return false;
    }

    return jack->isConnected();
  }

  nmm_bool Node::isInputConnected(const string& jack_tag) const  {
    const Jack* jack = getOwnInputStream(jack_tag);
    if (!jack) {
      // no such jack, so not connected
      return false;
    }

    return jack->isConnected();
  }


  ComputationConfig* Node::getComputationConfig() const {
    return computation_config;
  }

  void Node::setComputationConfig(const ComputationConfig* comp_conf) {
    if (computation_config != 0) {
      delete computation_config;
    }
    computation_config = const_cast<ComputationConfig*>(comp_conf);
  }

  ComputationConfig* Node::getWorkingComputationConfig() const {
    return working_computation_config;
  }

  void Node::setWorkingComputationConfig(const ComputationConfig* comp_conf) {
    if (computation_config != 0) {
      delete working_computation_config;
    }
    working_computation_config = const_cast<ComputationConfig*>(comp_conf);
  }

#include "nmm/nmm_prefix.hpp"

}//end namespace NMM


#include "nmm/nmm_suffix.hpp"
