/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include "nmm/comm/messaging/Message.hpp"
#include "Node.hpp"
#include "QueuedInputJack.hpp"
#include "OutputJack.hpp"
#include "nmm/comm/InstreamCompositeStrategy.hpp"
#include "QueuedOutputJack.hpp"
#include "nmm/comm/CommunicationChannel.hpp"

#include "GenericNode.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  QueuedInputJack::QueuedInputJack(Node* parent, 
				   const string& jack_tag /* = "default */,
				   const string& name /* = "QueuedInputJack"*/,
				   const StreamQueue::Mode mode /* = StreamQueue::DEFAULT_MODE */,
				   const unsigned int size /* = StreamQueue::DEFAULT_SIZE */)
    : InputJack(parent, jack_tag, name)
  {
    stream_queue = 
      new StreamQueue( mode, size, name.c_str(), FREE_DOWNSTR_QUEUE_SLOT, INPUT_QUEUE_NOT_EMPTY);

    stream_queue->setParentSignalReceiver(dynamic_cast<ISignalReceiver *>(parent));
  }

  QueuedInputJack::~QueuedInputJack() 
  {
    stream_queue->setPartnerSignalReceiver(0);
    delete stream_queue;
  }

  void QueuedInputJack::connectTo(const OutputJack* out_jack) const
    throw (const BindingException&)
  {
    InputJack::connectTo(out_jack);
    stream_queue->setPartnerSignalReceiver(const_cast<Jack *>(getPartner()));
  }

  void QueuedInputJack::disconnect()
  {
    InputJack::disconnect();
    stream_queue->setPartnerSignalReceiver(0);
  }

  Result QueuedInputJack::receiveMessage(Message* message, const IMessageSender*  )
  {
    DEBUG_STREAM2("Forwarding buffer to parent" << endl);

    Node* parent = getParent();
    if(parent)
    {
      GenericNode* generic_parent = dynamic_cast<GenericNode*>(parent);
      if(generic_parent)
      {
        // Stop delay measurement
        generic_parent->stopDelayMeasurement(IDelayMeasurement::STOP_WITH_RECEIVE_MESSAGE, message);

        // Start delay measurement
        generic_parent->startDelayMeasurement(IDelayMeasurement::START_WITH_RECEIVE_MESSAGE, message);
      }
    }

    return stream_queue->receiveMessage(message);
  }

  Result QueuedInputJack::sendMessage(Message *message) {
    if(getMessageReceiver() != 0) {
      DEBUG_STREAM2("QueuedInputJack> Send Message UPSTREAM to partner jack" << endl);
      Result r = getMessageReceiver()->receiveMessage(message, this);
      if(r == BLOCKED) {
	DEBUG_STREAM2("QueuedInputJack> Partner jack returned BLOCKED which means its stream_queue is full now" << endl);
      }
      return r;
    } else if(getOwnTransportStrategy() != 0) {
      DEBUG_STREAM2("QueuedInputJack> Send Message UPSTREAM to transport strategy" << endl);
      Result r = getOwnTransportStrategy()->receiveCSMessage(message);
      if(r == BLOCKED) {
        DEBUG_STREAM2("QueuedInputJack> Transport strategy returned BLOCKED which means its stream_queue is full now" << endl);
      }
      return r;
    } else {
      DEBUG_STREAM2("QueuedInputJack> Jack not connected, discarding buffer" << endl);
      message->release();
      return FAILURE;
    }
  }

  void QueuedInputJack::setTransportStrategy(InstreamCompositeStrategy* strategy)
    throw(const NMM::BindingException &)
  {
    InputJack::setTransportStrategy(strategy);
    stream_queue->setPartnerSignalReceiver(strategy);
    strategy->setSignalReceiver(this);
  }
  
  Result QueuedInputJack::flush()
  { 
    return stream_queue->flush(); 
  }

  unsigned int QueuedInputJack::getSize() const 
  { 
    return stream_queue->getSize(); 
  }

  unsigned int QueuedInputJack::getMaxSize() const 
  { 
    return stream_queue->getMaxSize(); 
  }

  Result QueuedInputJack::setMaxSize(unsigned int size) 
  { 
    return stream_queue->setMaxSize(size); 
  }

  Result QueuedInputJack::setMaxDuration(const Interval& duration)
  {
    return stream_queue->setMaxDuration(duration);
  }

  StreamQueue::Mode QueuedInputJack::getMode() const 
  { 
    return stream_queue->getMode(); 
  }

  Result QueuedInputJack::setMode(StreamQueue::Mode mode) 
  { 
    return stream_queue->setMode(mode); 
  }

}

#include "nmm/nmm_suffix.hpp"
