/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include <iostream>
#include <string.h>
#include <sys/types.h>
#include <unistd.h>
#include "ThreadedNode.hpp"
#include "nmm/comm/messaging/Buffer.hpp"
#include "nmm/utils/time/Interval.hpp"
#include "nmm/utils/thread/MutexGuard.hpp"
#include "nmm/Exception.hpp"
#include "nmm/comm/serialize/SerializedException.hpp"

#include <sys/resource.h>
#include <sys/time.h>

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  ThreadedNode::ThreadedNode(const char* name) : Node(name)
  {
    thread_owner = true;
    activated_before = false;
    thread_pid = 0;
    thread_priority = 0;
  }

  ThreadedNode::~ThreadedNode()
  {
  }

  void ThreadedNode::initOutput()
    throw (const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&)
  {
    MutexGuard mg(state_mutex);

    Node::initOutput();

    if(thread_owner) {
      if(!activated_before) {
	activated_before = true;
	try {
	  thread.start(run_static, (void *)this); // create thread
	}
	catch(const std::runtime_error & e) {
	  ERROR_STREAM2("Starting thread failed : " << e.what() << endl);
	  throw StateTransitionFailedException("Starting thread failed");
	}
	DEBUG_STREAM2("Thread created with ID : " << thread.getID() << endl);
      }
      else {
	DEBUG_STREAM2("Already created thread with ID : " << thread.getID() << " reusing this one... " << endl);
      }
    }
    else {
      DEBUG_STREAM2("ThreadedNode does not own a thread and has to be scheduled manually." << endl);
    }
    if (thread_priority) {
#ifndef __CYGWIN__
      setpriority(PRIO_PROCESS,thread_pid,thread_priority);
#endif
    }
  }

  void ThreadedNode::deinitOutput()
    throw (const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&)
  {
    MutexGuard mg(state_mutex);

    Node::deinitOutput();

    DEBUG_STREAM2("Joining Thread in deinitOutput..." << endl);
    thread.join();
    activated_before = false;
    DEBUG_STREAM2("Joined!" << endl);

  }

  void* ThreadedNode::run_static(void *me) {
    try {
      static_cast<ThreadedNode*>(me)->setPID(getpid());


      NODE_DEBUG_STREAM2(static_cast<Node*>(me),
			"Thread PID : " << getpid() << endl);
      (static_cast<ThreadedNode*>(me))->run();

    } 
    catch (const SerializedException& e) {
      // All exceptions should be caught before this. See ThreadedNode.hpp
      cerr << NamedObject::color_array[NamedObject::RED_COLOR];
      cerr << "Node \"" << static_cast<ThreadedNode*>(me)->getName()
           << "\" (pid: " << getpid() 
           << ") caused a uncaught SerializedException in its internal thread:" << endl;
      cerr << e << endl;
      cerr << NamedObject::color_array[NamedObject::DEFAULT_COLOR];
    } 
    catch (const Exception& e) {
      // All exceptions should be caught before this. See ThreadedNode.hpp
      cerr << NamedObject::color_array[NamedObject::RED_COLOR];
      cerr << "Node \"" << static_cast<ThreadedNode*>(me)->getName()
           << "\" (pid: " << getpid() 
           << ") caused a uncaught Exception in its internal thread:" << endl;
      cerr << e << endl;
      cerr << NamedObject::color_array[NamedObject::DEFAULT_COLOR];
    } 
    catch (...) {
      cerr << NamedObject::color_array[NamedObject::RED_COLOR];
      cerr << "Node \"" << static_cast<ThreadedNode*>(me)->getName() 
           << "\" (pid: " << getpid() 
           << ") caused a uncaught Exception in its internal thread:" << endl;
      cerr << "Unknown Exception!" << endl;
      cerr << NamedObject::color_array[NamedObject::DEFAULT_COLOR];
    }
    return 0; // to satisfy void* requirements, only checked with pthread_join
  }

  pid_t ThreadedNode::getPID() {
    return thread_pid;
  }

  void ThreadedNode::setPID(pid_t new_pid) {
        thread_pid = new_pid;
  }

#ifndef __CYGWIN__
  void ThreadedNode::lowerPriority(const nmm_uint32& amount) {
    if (thread_pid == 0) {
        return;
    }
    int new_priority = getpriority(PRIO_PROCESS,thread_pid) + amount;
    if (new_priority > 20) {
        new_priority = 20;
    }
    if (setpriority(PRIO_PROCESS,thread_pid,new_priority) != 0) {
        cerr << "setting priority failed!";
    }
  }
#endif
#ifdef __CYGWIN__
  void ThreadedNode::lowerPriority(const unsigned int& ) {}
#endif

  Result ThreadedNode::setPriority(unsigned int level) {
    if (!isOutputInitialized()) {
      ERROR_STREAM2("Wrong state to set priority level! (state != INITIALIZED)" << endl);
      return FAILURE;
    }
    if ((level == 0) || (level > 20)) {
      ERROR_STREAM2("Priority level out of range! (1..20)" << endl);
      return FAILURE;
    }
    thread_priority = level;
    return SUCCESS;
  }


//------ pthread scheduling methods (only allowed for root)
  void ThreadedNode::enableCustomScheduling(Thread::Scheduling_Policy policy) {
    thread.enableCustomScheduling(policy);
  }
  void ThreadedNode::setMinPriority() {
    thread.setMinPriority();
  }
  void ThreadedNode::setMidPriority() {
    thread.setMidPriority();
  }
  void ThreadedNode::setMaxPriority() {
    thread.setMaxPriority();
  }

}

#include "nmm/nmm_suffix.hpp"
