/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "nmm/utils/thread/MutexGuard.hpp"
#include "Controller.hpp"
#include <iostream>

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  Controller::Controller(const IController::Mode mode)
    : TimedElement(), StdNamedObject("Controller"), invalidate_buffers(false), invalidate_buffers_changed(false), invalidate_buffers_ts()
  {
    setMode(mode);
    permitDiscardBuffers(true);
    DEBUG_STREAM2( "created." << endl );
  };

  Controller::Controller(const Node* parent, const IController::Mode mode) 
    : TimedElement(), StdNamedObject("Controller"), invalidate_buffers(false), invalidate_buffers_changed(false), invalidate_buffers_ts(){
    setMode(mode);
    permitDiscardBuffers(true);
    if(parent != 0) {
      setName(parent->getName() + " Controller");
    } 
    DEBUG_STREAM2( "created." << endl );
  };

  Controller::~Controller() {
    DEBUG_STREAM2( "destroyed." << endl );
  };
  
  Result Controller::setMode(const IController::Mode& mode) {
    this->mode = mode;
    if(mode == IController::ONLINE) {
      DEBUG_STREAM2("Set mode to ONLINE." << endl );
    } else if(mode == IController::OFFLINE) {
      DEBUG_STREAM2("Set mode to OFFLINE." << endl );
    }
    return SUCCESS;
  };
  
  bool Controller::isBufferValid(const Timestamp& timestamp) {
    
    if(mode == IController::OFFLINE) { 
      return true; 
    }
    else {
      if(timestamp.is_valid) {
	bool valid = checkBufferValid(timestamp); 

	{
	  MutexGuard mg(invalidate_mutex);
	  // status should be changed -> check timestamp -> change status
	  if(invalidate_buffers_changed) {
	    if(timestamp >= invalidate_buffers_ts) {
	      invalidate_buffers = !invalidate_buffers;
	      invalidate_buffers_changed = false;
	      if(invalidate_buffers) {
		cerr << "TIMEPOINT: setBuffersInvalidTimed -> mute     : " << getTime() << endl;
	      }
	      else {
		cerr << "TIMEPOINT: setBuffersInvalidTimed -> unmute   : " << getTime() << endl;
	      }
	    }
	  }
	}

	// if result of check should not used and timestamp for this behavior is reached -> buffer is invalid
	if(invalidate_buffers) {
	  return false;
	}
	else {  
	  return valid;
	}
      }
      else {
	return true; // invalid ts are always assumed to be true
      }
    }
  };
  
  Result Controller::permitDiscardBuffers(const nmm_bool &p) {
    discard_buffers = p;
    
    return(SUCCESS);
  };

  void Controller::setBuffersInvalid(const nmm_bool& status)
  {
    MutexGuard mg(invalidate_mutex);
    invalidate_buffers = status;
    invalidate_buffers_changed = false;
    if(invalidate_buffers) {
      cerr << "TIMEPOINT: setBuffersInvalid: mute            : " << getTime() << endl;
    }
    else {
      cerr << "TIMEPOINT: setBuffersInvalid: unmute          : " << getTime() << endl;
    }
  }

  void Controller::setBuffersInvalidTimed(const nmm_bool& status, const Timestamp& status_time)
  {
    MutexGuard mg(invalidate_mutex);
    // check if status really changed, otherwise only store timestamp
    if(status != invalidate_buffers) {
      invalidate_buffers_changed = true;
      invalidate_buffers_ts = status_time;
    }
    else {
      invalidate_buffers_ts = status_time;
    }
  }

  Time Controller::ping()
  {
    return getTime();
  }

}





#include "nmm/nmm_suffix.hpp"
