/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "GenericSyncSinkNode.hpp"
#include "SinkController.hpp"
#include "nmm/multimedia/interfaces/ITimeProgressListener.hpp"
#include "nmm/comm/messaging/CEvent.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  GenericSyncSinkNode::GenericSyncSinkNode(const char* name,
					   const StreamQueue::Mode queue_mode,
					   const int size)
    : GenericSinkNode(name, queue_mode, size)
  {
    setSynchronized(true);
    have_prepare_method = true;
    media_time = UserTime();
    controller = 0;
    
    total_buffers = 0;
    total_buffers_discarded = 0;
    total_wait_time = 0;
    start_time = Time();

    send_progress_update = false;
    progress_interval = Interval(1, 0);

    update_seek_time = false;
    seek_time = 0;
    track_duration = 0;
    first_time = Time();
    first_buffer = true;
    saved_buffer = 0;
  };

  GenericSyncSinkNode::~GenericSyncSinkNode() 
  {
    if (controller) {
      Interval total_time = controller->getTime() - start_time;
      float percent_discarded = total_buffers ? (float) total_buffers_discarded * 100 / total_buffers : 0;
      float percent_waited = ((float) total_time > 0) ? (float) total_wait_time * 100 / (float) total_time : 0;
      DEBUG_STREAM(": Total buffers          : " << total_buffers << endl);
      DEBUG_STREAM(": total_buffers_discarded: " << total_buffers_discarded <<
                   " (" << percent_discarded << "%)" << endl);
      DEBUG_STREAM(": total_time             : " << total_time << endl);
      DEBUG_STREAM(": total_wait_time        : " << total_wait_time <<
                   " (" << percent_waited << "%)" << endl);
    }
    
    if(controller != 0) {
      delete controller;
    }
  };

  Result GenericSyncSinkNode::setSynchronized(const nmm_bool& toggle) 
  {
    synchronized = toggle; 
    return SUCCESS;
  }

  nmm_bool GenericSyncSinkNode::isSynchronized()
  {
    return synchronized;
  }

  void GenericSyncSinkNode::prepareBuffer(Buffer*) {
    have_prepare_method = false;
  };

  Result GenericSyncSinkNode::syncReset() {
    update_seek_time = true;
    first_buffer = true;
    DEBUG_STREAM2("sync_reset begin." << endl);
    if (synchronized) {
      if(controller == 0) {
	DEBUG_STREAM2("eventSyncReset without having a controller." << endl);
      } else {
	controller->reset();
      }
    } else {
      DEBUG_STREAM2("syncronization disabled -> sync_reset irgnored." << endl);
    }
    DEBUG_STREAM2("sync_reset completed." << endl);
    
    if (controller) {
      total_buffers = 0;
      total_buffers_discarded = 0;
      total_wait_time = 0;
      start_time = controller->getTime();
    }
    
    return SUCCESS;
  };

  Result GenericSyncSinkNode::pause() {
    DEBUG_STREAM2(" pause begin." << endl);
    if(controller->pause() != SUCCESS) {
      ERROR_STREAM2("pause not possible..." << endl);
      return FAILURE;
    }
    
    //    DEBUG_STREAM2("Stopping the node..." << endl);
    //    stop();
    //    DEBUG_STREAM2("Node stopped, pause end." << endl);
    return SUCCESS;
  };

  Result GenericSyncSinkNode::wakeup() {
    DEBUG_STREAM2("wakeup begin." << endl);
    if(controller->wakeup() != SUCCESS) {
      ERROR_STREAM2("wakeup not possible..." << endl);
      return FAILURE;
    }

    //    DEBUG_STREAM2("Starting the node..." << endl);
    //    start();
    //    DEBUG_STREAM2("Node started, wakeup end." << endl);
    return SUCCESS;
  };

  Message* GenericSyncSinkNode::processBuffer(Buffer *in_buffer) {
    if (in_buffer != NULL) {
      ++total_buffers;
    }

    if(synchronized && !getWorkingFlag()) {
      // errror 20030512: Although we are synchronized, the working
      // flag might be set. In this case, in_buffer is a NULL-Ptr, so
      // we cannot access its methods. As the leaf Node has set the
      // working flag to be called again (e.g. the display to redraw
      // the last frame), we also should not discard this call.
      if(!controller->isBufferValid(in_buffer->getTimestamp())) {
	DEBUG_STREAM2("Release Buffer " 
		      << in_buffer->getTimestamp().stream_counter  
		      << endl);
        ++total_buffers_discarded;
	in_buffer->release();
	return 0;
      }
    }


    if (!getWorkingFlag()) {
      if(have_prepare_method) {  
	prepareBuffer(in_buffer);
      }
      // errror 20030512: Although we are synchronized, the working
      // flag might be set. In this case, in_buffer is a NULL-Ptr, so
      // we cannot access its methods. As the leaf Node has set the
      // working flag to be called again, this call is not really
      // synchronized as the is no buffer timestamp to synchronize on.
      if(synchronized) {
        Interval wait_time = controller->waitToPresent(in_buffer->getTimestamp());
        total_wait_time += wait_time;
        suspendThread(wait_time);
      }


      if (first_buffer) {
	first_buffer = false;
	if (!seek_time) {
	  first_time = in_buffer->getTimestamp().sync_time;
	}
	else {
	  if (in_buffer->getTimestamp().sync_time.sec) {
	    seek_time = 0;
	  }
	}
      }
      
      media_time = in_buffer->getTimestamp().sync_time;
      DEBUG_STREAM2("media time: " << media_time << endl);


      if (send_progress_update && track_duration) {
	Interval time = in_buffer->getTimestamp().sync_time - first_time + Interval(seek_time, 0);
	if (time - last_time >= progress_interval || last_time - time >= progress_interval) {
	  CEvent* cevent(new CEvent(ITimeProgressListener::create_setTimeProgress(Time(time.sec, time.nsec))));

	  last_time = time;
	  saved_buffer = in_buffer;
	  setWorkingFlag(true);

	  return cevent;
	}
      }
    }

    if (getWorkingFlag() && saved_buffer) {
      in_buffer = saved_buffer;
      saved_buffer = 0;
      setWorkingFlag(false);
    }
    
    return presentBuffer(in_buffer);
  };
  
  ISinkController* GenericSyncSinkNode::getController() {
    return controller->getCheckedInterface<ISinkController>();
  }

  void GenericSyncSinkNode::setProgressUpdateInterval(const Interval& update_interval) {
    progress_interval = update_interval;
  }

  void GenericSyncSinkNode::setProgressUpdateEnabled(const nmm_bool& enable_update) {
    send_progress_update = enable_update;
  }

  void GenericSyncSinkNode::setProgress(const nmm_uint64& numerator, const nmm_uint64& denominator) {
    if (update_seek_time) {
      seek_time = numerator * track_duration / denominator;
      update_seek_time = false;
    }
  }

  Result GenericSyncSinkNode::trackDuration(const Interval& i) {
    update_seek_time = false;
    first_buffer = true;
    saved_buffer = 0;
    track_duration = i.sec;
    seek_time = 0;
    last_time = Interval();
    return SUCCESS;
  }
}




#include "nmm/nmm_suffix.hpp"
