/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include <iostream>
#include "MasterSlaveSourceSynchronizer.hpp"
#include "nmm/comm/ProxyObject.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  MasterSlaveSourceSynchronizer::MasterSlaveSourceSynchronizer(const char* name)
    : SourceSynchronizer(name)
  {
    master_skew = 0;
    start_delay = 0;
    state = WORKING;
  }

  MasterSlaveSourceSynchronizer::~MasterSlaveSourceSynchronizer()
  {}

  Result MasterSlaveSourceSynchronizer::setStartDelay(const Interval& delay) { 
    start_delay = delay;
    DEBUG_STREAM2("start_delay set to " << start_delay << endl);

    // set start_time for controllers to 0 + delay:
    Time start_time = getTime() + delay;
    master_source_controller->setStartTime(start_time);
    list<ISourceController*>::iterator iter = slave_source_controllers.begin();
    while (iter != slave_source_controllers.end()) {
      (*iter)->setStartTime(start_time);
      iter++;
    }
    return SUCCESS;
  }

  Result MasterSlaveSourceSynchronizer::setMasterSkew(const Interval& time_skew) {
    master_skew = time_skew;
    DEBUG_STREAM2("Set master_skew to " << master_skew << endl);
    
    list<ISourceController*>::iterator iter = slave_source_controllers.begin();
    while (iter != slave_source_controllers.end()) {
      (*iter)->setMasterSkew(master_skew);
      iter++;
    }
    return SUCCESS;
  }

  Result MasterSlaveSourceSynchronizer::setMasterController(const ISourceController* master_source_controller)
  {
    if(master_source_controller == 0) {    // make sure that the pointer is not 0 
      DEBUG_STREAM2("setMasterController: Trying to set 0 as master source controller." << endl);
      return FAILURE;
    }

    // Set the master source controller.
    this->master_source_controller = dynamic_cast<ISourceController*>(master_source_controller->copy());
    INamed_var mscname(this->master_source_controller->getParentObject()->getCheckedInterface<INamed>());
    DEBUG_STREAM2("setMasterController to " << mscname->getName() << "." << endl);

    // set Synchronizer at master_source
    this->master_source_controller->setSourceSynchronizer(getCheckedInterface<ISourceSynchronizer>());

    // set syncMaster to true
    this->master_source_controller->setSyncMaster(true);

    return SUCCESS;
  };

  Result MasterSlaveSourceSynchronizer::addSlaveController (const ISourceController* slave_source_controller)
  {
    if(slave_source_controller == 0) {
      DEBUG_STREAM2("setSlaveController: Trying to set 0 as slave source controller." << endl);
      return FAILURE;
    }

    ISourceController* tmp_controller = dynamic_cast<ISourceController*>(slave_source_controller->copy());
    this->slave_source_controllers.push_back(tmp_controller); // Add the slave controller.
    INamed_var sscname(tmp_controller->getParentObject()->getCheckedInterface<INamed>());
    DEBUG_STREAM2("setMasterController to " << sscname->getName() << "." << endl);

    // set Synchronizer at slave_source
    tmp_controller->setSourceSynchronizer(getCheckedInterface<ISourceSynchronizer>());

    // set syncMaster to false
    tmp_controller->setSyncMaster(false);

    return SUCCESS;
  };

  Result MasterSlaveSourceSynchronizer::removeSlaveController (const ISourceController* slave_source_controller)
  {
    list<ISourceController*>::iterator iter = slave_source_controllers.begin();
    while (iter != slave_source_controllers.end()) {
      list<ISourceController*>::iterator next = iter;
      next++;

      // Compare parentObject as interfaces are newly generated each time and
      // will in general at different memory locations!
      if ((*iter)->getParentObject() == slave_source_controller->getParentObject()) {
	ISourceController* tmp_controller = *iter;
	slave_source_controllers.erase(iter);
	// We have to delete the interface, as we copied it in addSlaveController
	delete tmp_controller;
      }

      iter = next;
    }
    return SUCCESS;
  }

  Result MasterSlaveSourceSynchronizer::resetAll()
  {
    DEBUG_STREAM2("reset begin." << endl);

    state_mutex.lock();               // Lock the mutex for state changes.

    master_source_controller -> reset();    
    list<ISourceController*>::iterator iter = slave_source_controllers.begin();
    while (iter != slave_source_controllers.end()) {
      (*iter) -> reset();
      iter++;
    }

    state_mutex.unlock();       // State changes finished.

    DEBUG_STREAM2("reset end." << endl);
    return SUCCESS;
  };

  Result MasterSlaveSourceSynchronizer::pause()
  {
    state_mutex.lock();

    if(state == PAUSE) {
      DEBUG_STREAM2("Already in state PAUSE." << endl);
      state_mutex.unlock();
      return SUCCESS;
    }

    if(state != WORKING) {
      ERROR_STREAM2("Wrong state for pause!" << endl);
      state_mutex.unlock();
      return FAILURE;
    }

    // if we are here, the state is WORKING
    DEBUG_STREAM2("Switching to state PAUSE." << endl);
    state = PAUSE;
    // for debug output only
    pause_start = getTime();

    DEBUG_STREAM2("Pause the source controllers, too." << endl);
    master_source_controller -> pause();
    list<ISourceController*>::iterator iter = slave_source_controllers.begin();
    while (iter != slave_source_controllers.end()) {
      (*iter) -> pause();
      iter++;
    }

    state_mutex.unlock();

    return SUCCESS;
  };

  Result MasterSlaveSourceSynchronizer::wakeup()
  {
    state_mutex.lock();

    if(state != PAUSE) {
      ERROR_STREAM2("Wrong state for wakeup." << endl);
      state_mutex.unlock();
      return FAILURE;
    }

    // if we are here, the state is PAUSE

    DEBUG_STREAM2("Switching to state WORKING." << endl);
    state = WORKING;

    Time time_now = getTime();

    // for debug output only
    Interval pause_duration = time_now - pause_start;
    DEBUG_STREAM2("Pause duration was " << pause_duration
		  << ", setting new theo_latency." << endl);

    Time wakeuptime = time_now + start_delay;
    DEBUG_STREAM2("New start_time will be " << wakeuptime << endl);
    master_source_controller->setStartTime(wakeuptime);
    list<ISourceController*>::iterator iter = slave_source_controllers.begin();
    while (iter != slave_source_controllers.end()) {
      (*iter)->setStartTime(wakeuptime);
      iter++;
    }

    DEBUG_STREAM2("Wake up the sources, too." << endl);

    master_source_controller->wakeup();
    iter = slave_source_controllers.begin();
    while (iter != slave_source_controllers.end()) {
      (*iter)->wakeup();
      iter++;
    }

    state_mutex.unlock();

    return SUCCESS;
  };
  
  Result MasterSlaveSourceSynchronizer::reset(const nmm_int32& ) {
    ERROR_STREAM2("Reset not implemented" << std::endl);
    return FAILURE;
  }
}

#include "nmm/nmm_suffix.hpp"
