/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "MultiAudioVideoSynchronizer.hpp"
#include "nmm/utils/string_utils.hpp"
#include "nmm/utils/thread/MutexGuard.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  MultiAudioVideoSynchronizer::InternalSink::InternalSink()
    : controller(NULL),
      average_latency(0),
      skew(0),
      wait_for_reset(false)
  {
  }
  
  MultiAudioVideoSynchronizer::InternalSink::~InternalSink() {
    if (controller) {
      delete controller;
      controller = NULL;
    }
  }
  
  void MultiAudioVideoSynchronizer::InternalSink::setController(ISinkController* controller) {
    this->controller = controller;
  }
  
  ISinkController* MultiAudioVideoSynchronizer::InternalSink::getController() const {
    return controller;
  }
  
  Interval MultiAudioVideoSynchronizer::InternalSink::getAverageLatency() const {
    return average_latency;
  }
  
  void MultiAudioVideoSynchronizer::InternalSink::setAverageLatency(const Interval& average_latency) {
    this->average_latency = average_latency;
  }

  Interval MultiAudioVideoSynchronizer::InternalSink::getSkew() const {
    return skew;
  }
  
  void MultiAudioVideoSynchronizer::InternalSink::setSkew(const Interval& skew) {
    this->skew = skew;
  }
  
  bool MultiAudioVideoSynchronizer::InternalSink::getWaitForReset() const {
    return wait_for_reset;
  }
  
  void MultiAudioVideoSynchronizer::InternalSink::setWaitForReset(bool value) {
    wait_for_reset = value;
  }
  
  
  MultiAudioVideoSynchronizer::MultiAudioVideoSynchronizer(const char* name)
    : SinkSynchronizer(name),
      master(-1),
      master_average_latency_available(false),
      theo_latency(0),
      security_interval(0),
      pause_start(Time()),
      sink_count(0),
      reset_sink_count(0),
      state(WORKING),
      sinks_mutex(ThreadMutex::RECURSIVE),
      pause_cond(state_mutex)
  {
  }
  
  MultiAudioVideoSynchronizer::~MultiAudioVideoSynchronizer() {}
  
  Result MultiAudioVideoSynchronizer::setAudioSinkController(const int& id, const ISinkController* controller) {
    DEBUG_STREAM("setAudioSinkController " << id << ", master id " << master << endl);

    if (id < 0) {
      ERROR_STREAM("Invalid sink ID" << endl);
      return FAILURE;
    }
    
    //MutexGuard mg(sinks_mutex);
    
    /* check if sink already exists */
    sinks_t::iterator existing(audio_sinks.find(id));
    
    /* copy the interface */
    ISinkController* own_controller = dynamic_cast<ISinkController*>(controller->copy());
    
    /* set the sink */
    sinks_t::iterator it = audio_sinks.insert(pair<int, InternalSink>(id, InternalSink())).first;
    it->second.setController(own_controller);
    
    /* calculate internal id. */
    int internal_id = (id << 1) + 1;
    
    /* tell the audio we are the new Synchronizer. */
    {
      ISinkSynchronizer_var sync(getCheckedInterface<ISinkSynchronizer>());
      own_controller->setSinkSynchronizer(sync.get(), internal_id);
    }
    
    /* set this sink as new master if there is none */
    if (master < 0) {
      master = id;
    }
    
    if (id != master) { /* audio slave */
      /* set controller parameters */
      own_controller->setAverageCount(50);          // Set the internal parameters.
      own_controller->setMaxAverageDelay(2000*1000000); // 80 ms    // TEST: 2000 ms
      own_controller->setMaxDelay      (30*1000000); // 50 ms       // TEST: 30 ms
      own_controller->setMaxDiscardingCount(2000); // allow audio buffers to be thrown away (to keep up with master sink)
      own_controller->setMinWaitInterval(Interval(4*1000000)); // 4 ms 
      /* secondary audio may wait */
      own_controller->enableWaitToPresent(true);
      /* enable blocking mode (needed for seeking) */
      own_controller->enableBlocking(true);
    } else {
      /* this will set the parameters appropriate for the master audio sink */
      setMasterAudioSink(id);
    }

    /* tell the new controller the current theo_latency */
    it->second.getController()->setTheoLatency(theo_latency + it->second.getSkew());

    /* count the sink */
    if (existing == audio_sinks.end()) {
      ++sink_count;
    }
    
    return SUCCESS;
  }
  
  nmm_int32 MultiAudioVideoSynchronizer::addAudioSinkController(const ISinkController* controller)
  {
    int id = -1;

    /* Begin> Check if the controller was already registered */
    for(sinks_t::iterator it = audio_sinks.begin(); it != audio_sinks.end(); ++it) {
      if(it->second.getController()->getParentObject() == controller->getParentObject()) {
        WARNING_STREAM("AudioSink Controller already registered with id " << it->first << endl);
        /* Return the old ID in this case */
        id = it->first;
      }
    }
    /* End< */

    if (id == -1) {
      /* find first ID that can be (re)used */
      for(id = 0; ; ++id) {
        /* check if sink already exists */
        if(audio_sinks.find(id) == audio_sinks.end()) {
          break;
        }
      }
    }

    /* if successfully added, return ID */
    if(setAudioSinkController(id, controller) == SUCCESS) {
      return id;
    }
    else {
      return -1;
    }
  }

  Result MultiAudioVideoSynchronizer::setVideoSinkController(const int& id, const ISinkController* controller) {
    DEBUG_STREAM("setVideoSinkController " << id << ", master id " << master << endl);

    if (id < 0) {
      ERROR_STREAM("Invalid sink ID" << endl);
      return FAILURE;
    }
    
    //MutexGuard mg(sinks_mutex);
    
    /* check if sink already exists */
    sinks_t::iterator existing(video_sinks.find(id));
    
    /* copy the interface */
    ISinkController* own_controller = dynamic_cast<ISinkController*>(controller->copy());
    
    /* insert the bew sink */
    sinks_t::iterator it = video_sinks.insert( pair<int, InternalSink>(id, InternalSink()) ).first;
    it->second.setController(own_controller);
    
    /* calculate internal id. */
    int internal_id = (id << 1);
    
    /* tell the audio we are the new Synchronizer. */
    {
      ISinkSynchronizer_var sync(getCheckedInterface<ISinkSynchronizer>());
      own_controller->setSinkSynchronizer(sync.get(), internal_id);
    }
    
    /* set controller parameters : video sink */
    own_controller->setAverageCount(20); // 10 frames                 // TEST: 20 frames
    own_controller->setMaxAverageDelay(2000*1000000); // 40 ms        // TEST: 2000 ms
    own_controller->setMaxDiscardingCount(300); // 30 buffer at once  // TEST: 300
    own_controller->setMaxDelay(100*1000000); // 10 ms                // TEST: 100 ms
    own_controller->setMinWaitInterval(Interval(4*1000000)); // 4 ms  // TEST: new (FIXME: really good idea?)
    own_controller->enableWaitToPresent(true);

    /* enable blocking mode (needed for seeking) */
    own_controller->enableBlocking(true);

    /* tell the new controller the current theo_latency */
    it->second.getController()->setTheoLatency(theo_latency + it->second.getSkew());
    
    /* count the sink */
    if (existing == video_sinks.end()) {
      ++sink_count;
    }
    
    return SUCCESS;
  }
  
  nmm_int32 MultiAudioVideoSynchronizer::addVideoSinkController(const ISinkController* controller)
  {
    int id = -1; //set id to an invalid internal id value

    /* Begin> Check if the controller was already registered */
    for(sinks_t::iterator it = video_sinks.begin(); it != video_sinks.end(); ++it) {
      if(it->second.getController()->getParentObject() == controller->getParentObject()) {
        WARNING_STREAM("VideoSink Controller already registered with id " << it-> first << endl);
        /* reuse the old ID in this case */
        id = it->first;
      }
    }
    /* End< */

    /* Begin> If the controller is not registered, search an unused id */
    if (id == -1) {
      for(id = 0; ; ++id) {
        /* find first ID that can be (re)used */
        if(video_sinks.find(id) == video_sinks.end()) {
          break;
        }
      }
    }
    /* End< */

    /* if successfully added, return ID */
    if(setVideoSinkController(id, controller) == SUCCESS) {
      return id;
    }
    else {
      return -1;
    }
  }

  Result MultiAudioVideoSynchronizer::removeAudioSink(const int& id) {
    DEBUG_STREAM("removeAudioSink " << id << ", master id " << master << endl);
    /* check if sink exists */
    //MutexGuard mg(sinks_mutex);
    sinks_t::iterator existing(audio_sinks.find(id));
    if (existing != audio_sinks.end()) {
      audio_sinks.erase(id);
      --sink_count;

      /* if master was removed -> try to find new master */
      if(id == master) {
	/* must reset master to undefined (e.g. if master is the only audio controller and removed!) */
	master = -1;
	master_average_latency_available = false;

        for (sinks_t::iterator it = audio_sinks.begin(); it != audio_sinks.end(); ++it) {
          if(it->second.getController() != 0) {
	    /* will reset master id to new value */
            setMasterAudioSink(it->first);
            break;
          }
        }
      }

      return SUCCESS;
    } else {
      ERROR_STREAM("Invalid sink ID" << endl);
      return FAILURE;
    }
  }
  
  Result MultiAudioVideoSynchronizer::removeVideoSink(const int& id) {
    DEBUG_STREAM("removeVideoSink " << id << ", master id " << master << endl);
    /* check if sink exists */
    //MutexGuard mg(sinks_mutex);
    sinks_t::iterator existing(video_sinks.find(id));
    if (existing != video_sinks.end()) {
      video_sinks.erase(id);
      --sink_count;
      return SUCCESS;
    } else {
      ERROR_STREAM("Invalid sink ID" << endl);
      return FAILURE;
    }
  }
  
  Result MultiAudioVideoSynchronizer::removeSinkController(const ISinkController* controller)
  {
    for(sinks_t::iterator it = audio_sinks.begin(); it != audio_sinks.end(); ++it) {
      if(it->second.getController()->getParentObject() == controller->getParentObject()) {
        DEBUG_STREAM("removeSinkController: found audio sink " << it->first << ", master id " << master << endl);
        return removeAudioSink(it->first);
      }
    }
    for(sinks_t::iterator it = video_sinks.begin(); it != video_sinks.end(); ++it) {
      if(it->second.getController()->getParentObject() == controller->getParentObject()) {
        DEBUG_STREAM("removeSinkController: found video sink " << it->first << ", master id " << master << endl);
        return removeVideoSink(it->first);
      }
    }

    ERROR_STREAM("Could not find and remove controller" << endl);
    return FAILURE;
  }

  Result MultiAudioVideoSynchronizer::setAverageAudioLatency(const int& id, const Interval& average_latency) {
    /* find the sink */
    MutexGuard mg(sinks_mutex);
    sinks_t::iterator it(audio_sinks.find(id));
    if (it == audio_sinks.end()) {
      ERROR_STREAM("Invalid sink id: " << id << endl);
      return FAILURE;
    }
    
    DEBUG_STREAM2("setAverageAudioLatency: " << id << ": " << average_latency << endl);
    
    /* take the new average latency */
    it->second.setAverageLatency(average_latency);
    
    if(id == master) {
      master_average_latency_available = true;
    }

    /* and compute theoretical latency */
    if(computeTheoLatency()) {
    
      /* check if master (?) audio latency is smaller than theoretical latency */
      if(id == master && it->second.getAverageLatency() < theo_latency) {
	theo_latency = it->second.getAverageLatency();
	DEBUG_STREAM2("Master audio latency smaller than theo latency..." << endl);
      }
    
      /* tell all controllers the new theoretical latency */
      sendTheoLatency();
    }
    
    return SUCCESS;
  }
  
  Result MultiAudioVideoSynchronizer::setAverageVideoLatency(const int& id, const Interval& average_latency) {
    if(master < 0) { // no master set -> compute theo latency
      /* find the sink */
      MutexGuard mg(sinks_mutex);
      sinks_t::iterator it(video_sinks.find(id));
      if (it == video_sinks.end()) {
	ERROR_STREAM("Invalid sink id: " << id << endl);
	return FAILURE;
      }
    
      DEBUG_STREAM2("setAverageVideoLatency: " << id << ": " << average_latency << endl);
    
      /* take the new average latency */
      it->second.setAverageLatency(average_latency);
    
      /* and compute theoretical latency */
      if(computeTheoLatency()) {
	/* tell all controllers the new theoretical latency */
	sendTheoLatency();
      }
      return SUCCESS;
    }
    else { // synchronizer has master -> do not worry about video latency
      return SUCCESS;
    }
  }
  
  Result MultiAudioVideoSynchronizer::setMasterAudioSink(const nmm_int32& id) {
    DEBUG_STREAM2("setMasterAudioSink: " << id << endl);
    
    /* find the sink */
    //MutexGuard mg(sinks_mutex);
    sinks_t::iterator it(audio_sinks.find(id));
    if (it == audio_sinks.end()) {
      ERROR_STREAM("Invalid sink id: " << id << endl);
      return FAILURE;
    }
    
    /* set id of master */
    master = id;
    
    /* set default parameters for master audio sink */
    ISinkController* own_controller = it->second.getController();
    own_controller->setAverageCount(50);
    own_controller->setMaxAverageDelay(3*1000000); // 3 ms 
    own_controller->setMaxDiscardingCount(0);
    own_controller->setMaxDelay(150*1000000); // 150 ms, value does not matter so far...
    own_controller->setMinWaitInterval(Interval(4*1000000)); // 4 ms
    
    /* master audio should should also wait to account for increased latency of other sinks */
    own_controller->enableWaitToPresent(true);

    /* master audio sink controller should not discard buffers when waiting for a syncReset from another conroller */
    own_controller->permitDiscardBuffers(false);

    /* enable blocking mode (needed for seeking) */
    own_controller->enableBlocking(true);

    /* set skew to zero because it is always relative to master */
    it->second.setSkew(0);
    
    return SUCCESS;
  }
  
  Result MultiAudioVideoSynchronizer::setAudioSkew(const nmm_int32& id, const Interval& skew) {
    /* find the sink */
    //MutexGuard mg(sinks_mutex);
    sinks_t::iterator it(audio_sinks.find(id));
    if (it == audio_sinks.end()) {
      ERROR_STREAM("Invalid sink id: " << id << endl);
      return FAILURE;
    }
    
    /* can't set skew at master (always zero) */
    if (id == master) {
      ERROR_STREAM("Tried to set skew at master sink (which is always zero!)" << endl);
      return FAILURE;
    }
    
    /* set skew */
    it->second.setSkew(skew);
    
    /* send new theoretical latency with new skew values */
    sendTheoLatency();
    
    return SUCCESS;
  }
  
  Result MultiAudioVideoSynchronizer::setVideoSkew(const nmm_int32& id, const Interval& skew) {
    /* find the sink */
    //MutexGuard mg(sinks_mutex);
    sinks_t::iterator it(video_sinks.find(id));
    if (it == video_sinks.end()) {
      ERROR_STREAM("Invalid sink id: " << id << endl);
      return FAILURE;
    }
    
    /* set skew */
    it->second.setSkew(skew);
    
    /* send new theoretical latency with new skew values */
    sendTheoLatency();
    
    return SUCCESS;
  }
  
  void MultiAudioVideoSynchronizer::setAverageLatency(const Interval& value, const nmm_int32& sink) {
    DEBUG_STREAM2("setAverageLatency: " << sink << " = " << value << endl);
    
    /* resolve ID */
    int id;
    bool is_audio;
    resolveSinkID(sink, &id, &is_audio);
    
    /* call the right method */
    if (is_audio) {
      setAverageAudioLatency(id, value);
    } else {
      setAverageVideoLatency(id, value);
    }
  }
  
  Result MultiAudioVideoSynchronizer::setFirstBufferInTimeAverageLatency(const Interval& /* value */, 
      const nmm_int32& /* sink */,
      const Timestamp& /* first_ts */)
  {
    return SUCCESS;
  }

  Result MultiAudioVideoSynchronizer::pause() {
    MutexGuard mg(state_mutex);
    DEBUG_STREAM2("pause begin" << endl);
    
    if(state == PAUSE) {
      DEBUG_STREAM2("Already in state PAUSE." << endl);
      return SUCCESS;
    }
    
    if(state != WORKING) {
      DEBUG_STREAM2("Wrong state for pause. Waiting" << endl);
      pause_cond.wait();
    }
    
    DEBUG_STREAM2("Switching to state PAUSE." << endl);
    state = PAUSE;
    pause_start = getTime();
    
    /* pause the sinks */
    //MutexGuard mg_sinks(sinks_mutex);
    for (sinks_t::iterator it(audio_sinks.begin()); it != audio_sinks.end(); ++it) {
      DEBUG_STREAM2("pausing audio sink " << it->first << endl);
      it->second.getController()->pause();
    }
    for (sinks_t::iterator it(video_sinks.begin()); it != video_sinks.end(); ++it) {
      DEBUG_STREAM2("pausing video sink " << it->first << endl);
      it->second.getController()->pause();
    }
    
    DEBUG_STREAM2("pause end" << endl);
    return SUCCESS;
  }
  
  Result MultiAudioVideoSynchronizer::wakeup() {
    MutexGuard mg(state_mutex);
    DEBUG_STREAM2("wakeup" << endl);
    
    if(state != PAUSE) {
      ERROR_STREAM2("Wrong state for wakeup." << endl);
      return FAILURE;
    }
    
    /* add pause duration to theoretical latency */
    Interval pause_duration = getTime() - pause_start;
    theo_latency += pause_duration;
    
    DEBUG_STREAM2("Pause duration was " << pause_duration
		  << ", setting new theo_latency." << endl);
    
    /* send new theoretical latency */
    sendTheoLatency();    
    
    if (reset_sink_count == 0) {
      /* wake up the sinks */
      //MutexGuard mg_sinks(sinks_mutex);
      for (sinks_t::iterator it(audio_sinks.begin()); it != audio_sinks.end(); ++it) {
        it->second.getController()->wakeup();
      }
      for (sinks_t::iterator it(video_sinks.begin()); it != video_sinks.end(); ++it) {
        it->second.getController()->wakeup();
      }
      
      DEBUG_STREAM2("Switching to state WORKING." << endl);
      state = WORKING;
    } else {
      /* we have some resets to do */
      //MutexGuard mg_sinks(sinks_mutex);
      for (sinks_t::iterator it(audio_sinks.begin()); it != audio_sinks.end(); ++it) {
        if (it->second.getWaitForReset()) {
          it->second.getController()->waitForReset();
        }
      }
      for (sinks_t::iterator it(video_sinks.begin()); it != video_sinks.end(); ++it) {
        if (it->second.getWaitForReset()) {
          it->second.getController()->waitForReset();
        }
      }
      
      DEBUG_STREAM2("Switching to state RESET." << endl);
      state = RESET;
    }
    
    return SUCCESS;
  };
  
  Result MultiAudioVideoSynchronizer::reset(const nmm_int32& sink) {
    MutexGuard mg(state_mutex);
    DEBUG_STREAM2("reset begin" << endl);
    
    /* resolve ID */
    int id;
    bool is_audio;
    resolveSinkID(sink, &id, &is_audio);
    
    /* resolve sink */
    //MutexGuard mg_sinks(sinks_mutex);
    sinks_t::iterator it;
    if (is_audio) {
      it = audio_sinks.find(id);
      if (it == audio_sinks.end()) {
        ERROR_STREAM("Invalid sink ID" << endl);
        return FAILURE;
      }
    } else {
      it = video_sinks.find(id);
      if (it == video_sinks.end()) {
        ERROR_STREAM("Invalid sink ID" << endl);
        return FAILURE;
      }
    }
    
    string sink_id = string(is_audio ? "audio" : "video") + id;
    
    /* change the own state to RESET. */
    if (state == WORKING) {
      DEBUG_STREAM2("Switching from WORKING to RESET." << endl);
      state = RESET;
    }
    
    /* check state of the sink */
    if(!it->second.getWaitForReset()) {
      
      DEBUG_STREAM2("Initial reset on sink " << sink_id << endl);
      
      /* Tell all other sinks to wait for reset*/
      for (sinks_t::iterator i(audio_sinks.begin()); i != audio_sinks.end(); ++i) {
        string other_sink_id = string("audio") + i->first;
        if (is_audio && i == it) {
          continue;
        }
        if (!i->second.getWaitForReset()) {
          ++reset_sink_count;
        }
        
        if (state == RESET) {
          DEBUG_STREAM2("Tell sink to wait for reset " << other_sink_id << endl);
          i->second.getController()->waitForReset();
        }
        i->second.setWaitForReset(true);
        
      }
      for (sinks_t::iterator i(video_sinks.begin()); i != video_sinks.end(); ++i) {
        string other_sink_id = string("video") + i->first;
        if (!is_audio && i == it) {
          continue;
        }
        if (!i->second.getWaitForReset()) {
          ++reset_sink_count;
	}
        
        if (state == RESET) {
          DEBUG_STREAM2("Tell sink to wait for reset " << other_sink_id << endl);
          i->second.getController()->waitForReset();
        }
        i->second.setWaitForReset(true);
      }
      
    } else { // if(!it->second.getWaitForReset()) {
      
      /* This sink has now received its reset */
      DEBUG_STREAM2("Sink has received reset " << sink_id << endl);
      it->second.setWaitForReset(false);
      
      --reset_sink_count;
      
    }
    
    /* Check if all sinks have been reset */
    if (reset_sink_count == 0) {
      if (state == RESET) {
        DEBUG_STREAM2("All resets done -> restart all sinks" << endl);
	
	// HACK
	// cerr << "HACK: Received all syncs, waiting 5 second " << endl;
        // sleep(5);

        /* call restart at all sinks */
        for (sinks_t::iterator i(audio_sinks.begin()); i != audio_sinks.end(); ++i) {
          DEBUG_STREAM2("Restart on audio sink " << i->first << endl);
          i->second.getController()->restart();
        }
        for (sinks_t::iterator i(video_sinks.begin()); i != video_sinks.end(); ++i) {
          DEBUG_STREAM2("Restart on video sink " << i->first << endl);
          i->second.getController()->restart();
        }
        
        /* change the own state back to WORKING. */
        DEBUG_STREAM2("Switching from RESET to WORKING." << endl);
        state = WORKING;
        pause_cond.notify();
      }
    }
    
    /* reset all internal parameters */
    {
      theo_latency = 0;
      master_average_latency_available = false;
      for (sinks_t::iterator i(video_sinks.begin()); i != video_sinks.end(); ++i) {
        i->second.setAverageLatency(0);
      }
      for (sinks_t::iterator i(audio_sinks.begin()); i != audio_sinks.end(); ++i) {
        i->second.setAverageLatency(0);
      }
    }
    
    DEBUG_STREAM2("reset end" << endl);
    return SUCCESS;
  };
  
  void MultiAudioVideoSynchronizer::resolveSinkID(int internal_id, int* id, bool* is_audio) {
    /* check if it's an audio sink */
    if (internal_id & 1) {
      *is_audio = true;
    } else {
      *is_audio = false;
    }
    
    /* get our ID */
    *id = internal_id >> 1;
  }
  
  bool MultiAudioVideoSynchronizer::computeTheoLatency() {
    //MutexGuard mg(sinks_mutex);
    
    Interval new_theo_latency = 0;
    for (sinks_t::iterator i(audio_sinks.begin()); i != audio_sinks.end(); ++i) {
      // only if master average latency is available -> use this value as theo latency
      if(master_average_latency_available) {
	if(i->first == master) { // only consider latency of master audio sink
	  new_theo_latency = max(new_theo_latency, i->second.getAverageLatency());
	  break; 
	}
      }
      else { // otherwise, get largest value of all other audio sinks
	new_theo_latency = max(new_theo_latency, i->second.getAverageLatency());
      }
    }
    // if no audio sink at all -> only consider video sinks ( this time, consider all video sinks )
    if(audio_sinks.size() == 0) {
      for (sinks_t::iterator i(video_sinks.begin()); i != video_sinks.end(); ++i) {
	new_theo_latency = max(new_theo_latency, i->second.getAverageLatency());
      }
    }
        
    new_theo_latency += security_interval;
    if(theo_latency != new_theo_latency) {
      DEBUG_STREAM2("update theo_latency: old " << theo_latency << " != new " << new_theo_latency << endl);
      theo_latency = new_theo_latency;
      return true;
    }
    else {
      DEBUG_STREAM2("theo_latency: old " << theo_latency << " == new " << new_theo_latency << endl);
      return false;
    }
  }
  
  void MultiAudioVideoSynchronizer::sendTheoLatency() {
    //MutexGuard mg(sinks_mutex);
    
    for (sinks_t::iterator i(audio_sinks.begin()); i != audio_sinks.end(); ++i) {
      i->second.getController()->setTheoLatency(theo_latency + i->second.getSkew());
    }
    for (sinks_t::iterator i(video_sinks.begin()); i != video_sinks.end(); ++i) {
      i->second.getController()->setTheoLatency(theo_latency + i->second.getSkew());
    }
  }
  
}

#include "nmm/nmm_suffix.hpp"
