/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "nmm/comm/ProxyObject.hpp"
#include "nmm/utils/thread/MutexGuard.hpp"
#include "SinkController.hpp"
#include "GenericSyncSinkNode.hpp"
#include "nmm/multimedia/sync/interfaces/ISinkSynchronizer.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  void SinkController::reset_average() {
    n = 0;
    average_latency = 0;
    disc_count = 0;
  };

  void SinkController::init_variables() {
  
    max_average_delay = 0;
    max_delay = 0;
    default_theo_latency = 500000000L;  // Start with some default value.
    theo_latency = default_theo_latency;  // Start with default_theo_latency.
    average_count = 1;          // Talk to the synchronizer for each buffer.
    max_disc = 0;
    first_buffer = true;
    sink_id = -1;
    old_ts = Timestamp();

    reset_average();

    synchronizer = 0;

    state = WORKING;
    
    wait_to_present = true;

    min_wait_interv = 0L;

    notify_in_time = false;
    first_buffer_in_time = false;

    block_for_sync_reset = true;
  }

  SinkController::SinkController(const IController::Mode mode) : Controller(mode) { 
    parent_node = NULL;
    init_variables();
    DEBUG_STREAM2("created." << endl);
  };

  SinkController::SinkController(GenericSyncSinkNode* parent,
				 const IController::Mode mode) : Controller(parent, mode) {
    parent_node = parent;
    init_variables();
    DEBUG_STREAM2("created." << endl);
  };

  SinkController::~SinkController() 
  {};

  bool SinkController::checkBufferValid(const Timestamp ) {
    return true;
  };

  Interval SinkController::waitToPresent(const Timestamp ) {
    return Interval(0);
  };

  Result SinkController::setMaxDelay(const Interval& max_delay) {
    if(max_delay < Interval(0)) {
      DEBUG_STREAM2("setMaxDelay: Trying to set a negative delay." << endl);
      return FAILURE;
    }
    this->max_delay = max_delay;
    DEBUG_STREAM2("Setting max_delay to " << max_delay << "." << endl);
    return SUCCESS;
  };

  Result SinkController::setMaxAverageDelay(const Interval& max_average_delay) {
    if(max_average_delay < Interval(0)) {
      DEBUG_STREAM2("setMaxAverageDelay: Trying to set a negative delay." 
		    << endl);
      return FAILURE;
    }
    this->max_average_delay = max_average_delay;
    DEBUG_STREAM2("Setting max_average_delay to " << max_average_delay 
		  << "." << endl);
    return SUCCESS;
  };

  Result SinkController::setDefaultTheoLatency(const Interval& value)
  {
    default_theo_latency = value;
    theo_latency = default_theo_latency;
    return SUCCESS;
  }

  Result SinkController::setTheoLatency(const Interval& theo_latency) {
    //MutexGuard mg(state_mutex);
    int curr_state = state;
    if(curr_state == WORKING || curr_state == PAUSE || curr_state == WAIT_FOR_RESTART) {
      this->theo_latency = theo_latency;
      DEBUG_STREAM2("Setting theo_latency to " << theo_latency << endl);
    } else {
      DEBUG_STREAM2("setTheoLatency - call in wrong state." << endl);
    }
    return SUCCESS;
  };

  Result SinkController::setSinkSynchronizer(const ISinkSynchronizer* sync, const nmm_int32& sink_id)
  {
    if(sync == 0) {
      ERROR_STREAM("setSinkSynchronizer: Trying to set a null pointer..." << endl);
      synchronizer = 0;
      return FAILURE;
    }
    
    synchronizer = dynamic_cast<ISinkSynchronizer*>(sync->copy());
    synchronizer->getParentObject()->establishBinding();

    this->sink_id = sink_id;
    return SUCCESS;
  }

  Result SinkController::setAverageCount(const nmm_int32& average_count) {
    if(average_count < 0) {
      ERROR_STREAM("setAverageCount: Trying to set a negative average count." 
		   << endl);

      return FAILURE;
    }

    DEBUG_STREAM2("Setting average count to " << average_count << endl);
    this->average_count = average_count;
    return SUCCESS;
  };

  Result SinkController::setMaxDiscardingCount(const nmm_int32& max_disc) {
    if(max_disc < 0) { 
      DEBUG_STREAM2("setMaxDiscardingCount: Trying to set a negative "
		    "number." << endl);
      return FAILURE; 
    }

    DEBUG_STREAM2("Setting max_disc to " << max_disc << endl);
    this->max_disc = max_disc;
    return SUCCESS;
  };

  Result SinkController::reset() {
    DEBUG_STREAM2("reset begin." << endl);
    
    /* BEGIN> Lock the mutex for state changes. */
    {
      MutexGuard mg(state_mutex);
      switch (state) {
      case WORKING:
        {
          /** We have received an initial syncReset event.
              Now we must wait for the other sinks to receive a syncReset event, too */
          DEBUG_STREAM2("Switching from WORKING to WAIT_FOR_RESTART." << endl);
          state = WAIT_FOR_RESTART;
          
	  if(block_for_sync_reset) {
	    parent_node->block();
	  }
        }
        break;
      case WAIT_FOR_RESET:
        {
          /** We have received as many syncReset events as we have been waiting for.
              Now we must wait for the other sinks to receive all their syncReset events. */
          DEBUG_STREAM2("Switching from WAIT_FOR_RESET to WAIT_FOR_RESTART." << endl);
          state = WAIT_FOR_RESTART;
          
	  if(block_for_sync_reset) {
	    parent_node->block();
	  }
        }
        break;
      default:
        ERROR_STREAM("Wrong state for reset!!" << endl);
      }
    }
    /* END< Lock the mutex for state changes. */
    
    /* Reset internal parameters */
    theo_latency = default_theo_latency;
    reset_average();
    first_buffer = true;

    // syncReset -> first buffer that arrived in time has to be notified ?
    if(notify_in_time) {
      first_buffer_in_time = true; 
    }
    else {
      first_buffer_in_time = false; 
    }
    
    if(synchronizer) {
      /** Notify the synchronizer of the reset.
          The synchronizer must count the number of syncReset events, notify the other sinks
          and reset its own internal parameters */
      DEBUG_STREAM2("Reset the synchronizer." << endl);
      synchronizer->reset(sink_id);
    } else {
      /** If there is no synchronizer, there is no second stream to synchronize.
          So the reset is finished, and the state can be changed to working again. */
      DEBUG_STREAM2("There's no synchronizer to reset." << endl);
      if(state == WAIT_FOR_RESTART) {
        MutexGuard mg(state_mutex);
	state = WORKING;
	DEBUG_STREAM2("Switching to state WORKING again (nobody will wake us up.)" << endl);
        
	if(block_for_sync_reset) {
	  parent_node->unblock();
	}
      }
    }
    
    DEBUG_STREAM2("reset end." << endl);
    return SUCCESS;
  };

  Result SinkController::waitForReset() {
    DEBUG_STREAM2("waitForReset begin." << endl);
    
    /* BEGIN> Lock the mutex for state changes. */
    {
      MutexGuard mg(state_mutex);
      

      switch (state) {
      case WORKING:
        {
          /** Another sink has been reset and the synchronizer tells us to wait for a syncReset event.
              Therefore switch to WAIT_FOR_RESET and wait. */
          DEBUG_STREAM2("Switching from WORKING to WAIT_FOR_RESET." << endl);
          state = WAIT_FOR_RESET;
        }
        break;
      case WAIT_FOR_RESTART:
        {
          /** The synchronizer tells us to wait for a syncReset event we have alreadyy received.
              Do nothing. */
          DEBUG_STREAM2("Already received syncReset. Do nothing" << endl);
        }
        break;
      default:
        ERROR_STREAM("Wrong state for waitForReset!!" << endl);
      }
    }
    /* END< Lock the mutex for state changes. */
    
    DEBUG_STREAM2("waitForReset end." << endl);
    return SUCCESS;
  };

  Result SinkController::restart() {
    DEBUG_STREAM2("restart begin." << endl);
    
    /* BEGIN> Lock the mutex for state changes. */
    {
      MutexGuard mg(state_mutex);
      switch (state) {
      case WAIT_FOR_RESTART:
        {
          /** The synchronizer notifies us that all sinks have been reset and we can restart.
              Therefore switch back to WORKING. */
          DEBUG_STREAM2("Switching from state WAIT_FOR_RESTART to WORKING." << endl);
          state = WORKING;

	  if(block_for_sync_reset) {
	    parent_node->unblock();
	  }
        }
        break;
      case PAUSE:
        {
          DEBUG_STREAM2("Switching from state PAUSE to WORKING." << endl);
          state = WORKING;
          
          parent_node->start();
        }
      default:
        ERROR_STREAM("Wrong state for restart!!" << endl);
      }
    }
    /* END< Lock the mutex for state changes. */
    
    DEBUG_STREAM2("restart end." << endl);
    return SUCCESS;
  };

  Result SinkController::pause() {
    DEBUG_STREAM2("pause begin." << endl);
    
    /* BEGIN> Lock the mutex for state changes. */
    {
      MutexGuard mg(state_mutex);
      switch (state) {
      case PAUSE:
        {
          DEBUG_STREAM2("Already in state PAUSE, nothing to do." << endl);
        }
        break;
      case WORKING:
        {
          DEBUG_STREAM2("Switching from WORKING to PAUSE." << endl);
          state = PAUSE;
          parent_node->stop();
        }
        break;
      case WAIT_FOR_RESTART:
        {
          DEBUG_STREAM2("Switching from WAIT_FOR_RESTART to PAUSE." << endl);
          state = PAUSE;
          parent_node->stop();
        }
        break;
      case WAIT_FOR_RESET:
        {
          ERROR_STREAM2("BUG! The synchronizer should not call pause in this state!" << endl);
          return FAILURE;
        }
      }
    }
    /* END< Lock the mutex for state changes. */
    
    return SUCCESS;
  };

  Result SinkController::wakeup() {
    DEBUG_STREAM2("wakeup begin." << endl);
    MutexGuard mg(state_mutex);
    
    if(state != PAUSE) {
      ERROR_STREAM2("Wrong state for wakeup!" << endl);
      return FAILURE;
    }
    
    // if we are here, the state is PAUSE
    
    DEBUG_STREAM2("Switching from PAUSE to WORKING." << endl);
    state = WORKING;
    
    reset_average();
    
    parent_node->start();
    
    return SUCCESS;
  };
  
  Result SinkController::enableWaitToPresent(const nmm_bool& enable)
  {
    wait_to_present = enable;
    return SUCCESS;
  }

  Result SinkController::setMinWaitInterval(const Interval& min_wait)
  {
    min_wait_interv = min_wait;
    return SUCCESS;
  }

  void SinkController::enableFirstBufferInTime(const nmm_bool& enable)
  {
    notify_in_time = enable;
  }

  void SinkController::enableBlocking(const nmm_bool& enable)
  {
    MutexGuard mg(state_mutex);
    block_for_sync_reset = enable;
  }
}

#include "nmm/nmm_suffix.hpp"
