/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "nmm/comm/messaging/Buffer.hpp"
#include <iostream>
#include "SourceController.hpp"
#include "GenericSyncSourceNode.hpp"
#include "MasterSlaveSourceSynchronizer.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  SourceController::SourceController(GenericSyncSourceNode* parent,
				     const IController::Mode mode) 
    : Controller(parent, mode)
  {
    parent_skew = 0;
    master_skew = 0;
    start_time = Time(0,0);
    frame_time = 0;
    synchronizer = NULL;
    parent_node = parent;
    state = WORKING;
    DEBUG_STREAM2("created." << endl);
  }

  SourceController::~SourceController() 
  {}

  bool SourceController::checkBufferValid(const Timestamp timestamp){
    if (timestamp.sync_time < start_time) {
      DEBUG_STREAM("start time not reached! Buffer not valid." << endl);
      return false;
    }

    if (parent_node->isSyncMaster()) {
      // the policy at the moment is that the Master (sound card in all our cases)
      // does NOT discard any buffers (exept those before the start time)
      return true;
    }

    if (parent_skew - master_skew > frame_time) {
      // parent time is too much ahead of master time -> buffer is not valid
      return false;
    }
    return true;
  }

  bool SourceController::isBufferLate() {
    if (parent_node->isSyncMaster()) {
      // the policy at the moment is that the Master (sound card in all our cases)
      // is never too late
      return false;
    }
    
    if (master_skew - parent_skew > frame_time) {
      // parent time is too much behing master time -> buffer is too late
      DEBUG_STREAM(   "Buffer is too late: master_skew is" << master_skew << endl
		   << "                    parent_skew is" << parent_skew << endl
		   << "                    master_skew - parent_skew = " << master_skew - parent_skew << endl);
      return true;
    }
    
    return false;
  }

  void SourceController::setStartTime(const Time& time) {
    start_time = time;

    DEBUG_STREAM2("start_time set to " << start_time << endl);
  }

  Result SourceController::setSourceSynchronizer(const ISourceSynchronizer* synchronizer) {
    this->synchronizer = dynamic_cast<ISourceSynchronizer*>(synchronizer->copy());
    if(synchronizer) {
      DEBUG_STREAM2("Synchronizer set." << endl);
      return SUCCESS;
    }
    else {
      ERROR_STREAM2("No ISourceSynchronizer..." << endl);
      return FAILURE;
    }
  }

  void SourceController::setSyncMaster(const nmm_bool& master_) {
    parent_node->setSyncMaster(master_);
  }

  void SourceController::setFrameTime(Interval time_) {
    frame_time = time_;
    DEBUG_STREAM2("frame_time set to " << frame_time << endl);
  }

  Interval SourceController::getFrameTime() {
    return frame_time;
  }

  void SourceController::setMasterSkew(const Interval& skew) {
    master_skew = skew;
    DEBUG_STREAM("master_skew set to " << master_skew << endl);
  }

  void SourceController::setParentSkew(const Interval& skew) {
    /////////////////////////////////////////////////
    // THINK ABOUT THIS OUTCOMMENTED CODE... ////////
    /////////////////////////////////////////////////
//      Interval start_time_as_interval = start_time - Time(0,0);
//      parent_skew = skew - start_time_as_interval;
    parent_skew = skew;
    DEBUG_STREAM("parent_skew set to " << parent_skew << endl);
    if (parent_node->isSyncMaster()) {
      if (synchronizer) {
	synchronizer->setMasterSkew(parent_skew);
      }
      else {
	WARNING_STREAM("No synchronizer available, could not set MasterSkew!" << endl);
      }
    }
  }

  void SourceController::correctTimestamp(Buffer* buffer) {
    Timestamp t = buffer->getTimestamp();
    t.sync_time += master_skew;
    DEBUG_STREAM("corrTS: master_skew is " << master_skew << endl);
    buffer->setTimestamp(t);
  }

  Result SourceController::pause() {
    DEBUG_STREAM2("pause begin." << endl);

    state_mutex.lock();

    if(state == PAUSE) {
      DEBUG_STREAM2("Already in state PAUSE, nothing to do." << endl);
      state_mutex.unlock();
      return SUCCESS;

    } else if(state != WORKING) {
      ERROR_STREAM2("Wrong state for pause." << endl);
      state_mutex.unlock();
      return FAILURE;
    }

    // if we are here, the state is WORKING

    DEBUG_STREAM2("Switching from WORKING to PAUSE." << endl);
    state = PAUSE;
    parent_node->stop();

    state_mutex.unlock();

    return SUCCESS;
  }

  Result SourceController::wakeup() {
    DEBUG_STREAM2("wakeup begin." << endl);

    state_mutex.lock();

    if(state != PAUSE) {
      ERROR_STREAM2("Wrong state for wakeup!" << endl);
      state_mutex.unlock();
      return FAILURE;
    }

    // if we are here, the state is PAUSE
    DEBUG_STREAM2("Switching from PAUSE to WORKING." << endl);
    state = WORKING;
    state_mutex.unlock();
    
    parent_node->start();

    return SUCCESS;
  }

  Result SourceController::reset() {
    parent_node->resetTime();
    return SUCCESS;
  }

}

#include "nmm/nmm_suffix.hpp"
