/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include "StreamTimer.hpp"
#include "nmm/comm/messaging/Message.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  void StreamTimer::reset() {
    timestamp.is_valid                  = true;

    rate                                = 0;
    interval                            = 0;

    interval_set                        = false;
    delay_set                           = false;
    DEBUG_STREAM2 (" Reset. " << endl);
  };

  StreamTimer::StreamTimer(const StreamTimer::MODE mode)
    : TimedElement(), NamedObject("Timer") {
    this->reset();
    this->mode = mode;
  };

  StreamTimer::StreamTimer(const Node* parent, const StreamTimer::MODE mode) 
    : TimedElement(), NamedObject("Timer") {
    this->reset();
    this->mode = mode;                           
    if(parent != 0) {
      setName(parent->getName() + " Timer");
    }
  };

  StreamTimer::~StreamTimer() {};

  Result StreamTimer::setMode(const StreamTimer::MODE mode) {
    if(mode == REAL_TIME) { interval_set = true; }
    this->mode = mode;

    return SUCCESS;
  };

  Result StreamTimer::setRate(const float rate) {
    if(rate <= 0) return FAILURE;
    interval = Interval(static_cast<long int>(1000000000L / rate));
    DEBUG_STREAM2 ( "Interval is set to " << interval << endl );
    interval_set = true;
    return SUCCESS;
  };

  Result StreamTimer::setInterval(const Interval interval) {
    if(interval <= Interval(0)) return FAILURE; 
    this->interval = interval;
    DEBUG_STREAM2 ( "Interval is set to " << interval << endl );
    interval_set = true;
    return SUCCESS;
  };

  Result StreamTimer::setTimestamp(Message* message) {
    switch(mode) {
    case REAL_TIME:    // In real-time mode take the clock's time.
      timestamp.sync_time = this->getTime();
      break;

    case CONST_RATE:   // Compute the timestamp.
      if(!interval_set) {
	DEBUG_STREAM2 ( "Interval or rate has not been set!" << endl );
	return FAILURE;
      };
      if(timestamp.stream_counter == 0) break;
      
      // Begin with time 0 for message 0.
      // Only increase the timestamp for buffers.
      // Events simply get the timestamp of the last buffer.
      if (message->getType() == Message::BUFFER) {
        timestamp.sync_time += interval;
      }
      break;
      
    default:
      DEBUG_STREAM2 ( "Wrong mode for StreamTimer" << endl );
    };

    message->setTimestamp(timestamp);     // Set the buffer's timestamp.
    timestamp.stream_counter++;           // Count the buffer.

    return SUCCESS;
  };
}


#include "nmm/nmm_suffix.hpp"
