/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "USinkController.hpp"
#include "GenericUSyncSinkNode.hpp"
#include "Synchronizer.hpp"
#include <iostream>

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  USinkController::USinkController(const IController::Mode mode) 
    : SinkController(mode)
  {
    node = NULL;
  };

  USinkController::USinkController(GenericUSyncSinkNode* parent,
				   const IController::Mode mode) 
    : SinkController(parent, mode)
  {
    node = parent;
  };

  USinkController::~USinkController() 
  {};

  bool USinkController::checkBufferValid(const Timestamp timestamp) {
    DEBUG_STREAM2("Checking buffer " << timestamp << endl);

    //    state_mutex.lock();
    if(state == WAIT_FOR_RESET) {
      //state_mutex.unlock();
      
      if ( discard_buffers ) {
        DEBUG_STREAM2("wait_for_reset: Declare buffer " << timestamp.stream_counter
                      << " as invalid." << endl);
	cerr << "v" << flush;
        return false;
      } else {
        DEBUG_STREAM2("wait_for_reset: Declare buffer " << timestamp.stream_counter
                      << " as valid." << endl);
        return true;
      }
    }
    //state_mutex.unlock();

    DEBUG_STREAM2("Difference " 
		  << timestamp.sync_time - old_ts.sync_time << endl);
    old_ts = timestamp;

    Time dequeue_time = getTime(); // Get the time at the moment.
    Interval real_latency  = dequeue_time - timestamp.sync_time;
    // and compute the latency of this one buffer.
    DEBUG_STREAM2("real_latency " << real_latency << endl);

    average_latency += real_latency; // average latency computation.

    // Initialize the theo_latency.
    if(synchronizer) {
      if(first_buffer) { // If it is the first buffer, no average value
	// can be present yet. Take the value of this buffer
	// instead, it is the only information you have got.	
	DEBUG_STREAM2("FIRST BUFFER " << sink_id << " latency " << real_latency << " time " << getTime() << endl);
	synchronizer->setAverageLatency(real_latency, sink_id);
	
	first_buffer = false;
      }

      if(notify_in_time && first_buffer_in_time) { // if the first buffer in time should be notified (and has not been notified so far)
	DEBUG_STREAM2("FIRST BUFFER IN TIME sink " << sink_id << " latency " << real_latency << " time " << getTime() << endl);
	if(synchronizer->setFirstBufferInTimeAverageLatency(real_latency, sink_id, timestamp) == SUCCESS) {
	  first_buffer_in_time = false;
	}
      }
    }
    
    if(++n >= average_count) { // enough buffers to compute the average?

      average_latency /= average_count;  // Yes, then do it.
	
      DEBUG_STREAM2("average_latency " << average_latency << endl);
	
      // if average latency and theo latency differ a lot
      // (that means by more than max_average_delay)
      // tell it to the synchronizer
//       if(average_latency > theo_latency + max_average_delay ||
//          average_latency < theo_latency - max_average_delay) {
      if(average_latency > theo_latency + max_average_delay) {
	DEBUG_STREAM2("average_latency " << average_latency 
		      << " theo_latency " << theo_latency << endl);
	DEBUG_STREAM2("Average latency increasing..." << endl);
	if(synchronizer) {
	  // do not call setAverageLatency if the buffer will be invalidated
	  // (because the average_latency might be to small since the previous 
	  //  buffers have not been played out but discarded!)
	  if(!invalidate_buffers) {
	    synchronizer->setAverageLatency(average_latency, sink_id);
	  }
	} else {
	  theo_latency = average_latency;
	}
      }
      average_latency = 0;  // and start a new computation.
      n = 0;
    }

    // if this one buffer comes much to late 
    if(real_latency > theo_latency + max_delay)
      {
	DEBUG_STREAM2("Buffer " << timestamp.stream_counter << " too late, "
		      "declaring it as invalid: " << (real_latency - theo_latency - max_delay) << endl);
	if (++disc_count > max_disc) {  // if already max_disc buffer declared
	  disc_count = 0;               // as invalid: declare this as o.k.
	  return true;                  // that means it will be presented.
	} else {
	  cerr << "V" << flush;
	  return false;                 // else: declare it as invalid.
	}
      }
    disc_count = 0;  // this one buffer is o.k. and will be presented. 
    return true;
  };

  Interval USinkController::waitToPresent(const Timestamp timestamp) 
  {
    if(timestamp.is_valid) {
      if(wait_to_present) {  
	Time now = getTime();

	DEBUG_STREAM2("Wait to present call at " << now << " for buffer "
		      << timestamp << " theo_latency " << theo_latency << endl);
    
	if(now < timestamp.sync_time + theo_latency) {     // the buffer should be presented at
	  Interval wait_interv = timestamp.sync_time + theo_latency - now;
	  if(wait_interv < min_wait_interv) {
	    return Interval(0);
	  }
	  else {
	    DEBUG_STREAM2("Tell the node to wait " 
			  << wait_interv << "." << endl);  
	    return wait_interv; // so the node has to wait until then
	  }
	}
	else {
	  return Interval(0);
	}
      }
      else { // wait_to_present == false -> do not wait
	return Interval(0);
      }
    }
    else { // invalid timestamp -> do not wait
      return Interval(0);
    }
  };

}

#include "nmm/nmm_suffix.hpp"
