/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
 
/* 
 * Code is based on a52dec package that is
 * Copyrighted 2000-2001 by Michel Lespinasse <walken@zoy.org>
 * and copyrighted 1999-2000 by Aaron Holtzman <aholtzma@ess.engr.uvic.ca>
 */


#include "AC3DecodeNode.hpp"
#include "nmm/utils/time/Interval.hpp"

NMM_REGISTER_PLUGIN(AC3DecodeNode, NMM, "AC3DecodeNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  AC3DecodeNode::AC3DecodeNode( const char * name, StreamQueue::Mode qmode, int qsize )
    : GenericProcessorNode(name, qmode, qsize) {
  }
  
  AC3DecodeNode::~AC3DecodeNode() {
  }
  
  Result AC3DecodeNode::doInit() {
    // If another output format is added, think about the delay - value!
    // This has to be set to the duration of one output buffer to achieve
    // a correct synchronization.
    
    // the only input format
    Format* i_format;
    i_format = getOwnInputProperty()->addNewFormat("audio/ac3");
    i_format->addWildcard(Format::bitrate_param);
    
    i_format = getOwnInputProperty()->addNewFormat("audio/ac3");
    getOwnInputProperty()->setDefaultFormat(i_format);
    
    // the only output format
    Format* o_format = getOwnOutputProperty()->addNewFormat("audio/raw");
    o_format->addIntParamValue(Format::channels_param, 2);
    o_format->addIntParamValue(Format::sample_rate_param, 48000);
    o_format->addIntParamValue(Format::bitpersample_param, 16);
    o_format->addStringParamValue(Format::endian_param, "little");
    o_format->addStringParamValue(Format::encoding_param, "linear");
    o_format->addStringParamValue(Format::sign_param, "signed");
    getOwnOutputProperty()->setDefaultFormat(o_format);
    
    i_format->setIOPartner(o_format);
    o_format->setIOPartner(i_format);
    
    syncReset();
    
    return SUCCESS;
  }
  
  Result AC3DecodeNode::doDeinit() {
    getOwnOutputProperty()->erase();
    getOwnInputProperty()->erase();
    return SUCCESS;
  }
  
  Result AC3DecodeNode::doInitOutput() {    
    // the only input format
    Format* i_format = new Format( *getOwnInputProperty()->getDefaultFormat() );
    getOwnWorkingInputProperty()->addNewFormat(i_format);
    getOwnWorkingInputProperty()->setDefaultFormat(i_format);
    
    // the only output format
    Format* o_format = new Format( *getOwnOutputProperty()->getDefaultFormat() );
    getOwnWorkingOutputProperty()->addNewFormat(o_format);
    getOwnWorkingOutputProperty()->setDefaultFormat(o_format);
    
    // set I/O partners
    i_format->setIOPartner(o_format);
    o_format->setIOPartner(i_format);
    
    return SUCCESS;
  }
  
  Result AC3DecodeNode::doDeinitOutput() {
    getOwnWorkingOutputProperty()->erase();
    getOwnWorkingInputProperty()->erase();
    return SUCCESS;
  }
  
  Result AC3DecodeNode::doActivate() {
    /* Initialize buffer pointers for ac3 decode */
    bufptr = buf;
    bufpos = buf + 7;
    
    sample_rate = 0;
    flags       = 0;
    bias        = 384;
    channels    = 2;
    
    samples = 0;
    
    block = 0;
    
    a52_state = a52_init(0);
    if (NULL == a52_state) {
      return FAILURE;
    }
    
    old_in_buffer = NULL;
    
    return SUCCESS;
  }
  

  int16_t AC3DecodeNode::convert (int32_t i) {
    if (i > 0x43c07fff)
	return 32767;
    else if (i < 0x43bf8000)
	return -32768;
    else
	return i - 0x43c00000;
  }

  void AC3DecodeNode::float_to_int(float* _f, int16_t* s16, int flags) {
    int i;
    int32_t * f = (int32_t *) _f;

    switch (flags) {
    case A52_MONO:
	for (i = 0; i < 256; i++) {
	    s16[5*i] = s16[5*i+1] = s16[5*i+2] = s16[5*i+3] = 0;
	    s16[5*i+4] = convert (f[i]);
	}
	break;
    case A52_CHANNEL:
    case A52_STEREO:
    case A52_DOLBY:
	for (i = 0; i < 256; i++) {
	    s16[2*i] = convert (f[i]);
	    s16[2*i+1] = convert (f[i+256]);
	}
	break;
    case A52_3F:
	for (i = 0; i < 256; i++) {
	    s16[5*i] = convert (f[i]);
	    s16[5*i+1] = convert (f[i+512]);
	    s16[5*i+2] = s16[5*i+3] = 0;
	    s16[5*i+4] = convert (f[i+256]);
	}
	break;
    case A52_2F2R:
	for (i = 0; i < 256; i++) {
	    s16[4*i] = convert (f[i]);
	    s16[4*i+1] = convert (f[i+256]);
	    s16[4*i+2] = convert (f[i+512]);
	    s16[4*i+3] = convert (f[i+768]);
	}
	break;
    case A52_3F2R:
	for (i = 0; i < 256; i++) {
	    s16[5*i] = convert (f[i]);
	    s16[5*i+1] = convert (f[i+512]);
	    s16[5*i+2] = convert (f[i+768]);
	    s16[5*i+3] = convert (f[i+1024]);
	    s16[5*i+4] = convert (f[i+256]);
	}
	break;
    case A52_MONO | A52_LFE:
	for (i = 0; i < 256; i++) {
	    s16[6*i] = s16[6*i+1] = s16[6*i+2] = s16[6*i+3] = 0;
	    s16[6*i+4] = convert (f[i+256]);
	    s16[6*i+5] = convert (f[i]);
	}
	break;
    case A52_CHANNEL | A52_LFE:
    case A52_STEREO | A52_LFE:
    case A52_DOLBY | A52_LFE:
	for (i = 0; i < 256; i++) {
	    s16[6*i] = convert (f[i+256]);
	    s16[6*i+1] = convert (f[i+512]);
	    s16[6*i+2] = s16[6*i+3] = s16[6*i+4] = 0;
	    s16[6*i+5] = convert (f[i]);
	}
	break;
    case A52_3F | A52_LFE:
	for (i = 0; i < 256; i++) {
	    s16[6*i] = convert (f[i+256]);
	    s16[6*i+1] = convert (f[i+768]);
	    s16[6*i+2] = s16[6*i+3] = 0;
	    s16[6*i+4] = convert (f[i+512]);
	    s16[6*i+5] = convert (f[i]);
	}
	break;
    case A52_2F2R | A52_LFE:
	for (i = 0; i < 256; i++) {
	    s16[6*i] = convert (f[i+256]);
	    s16[6*i+1] = convert (f[i+512]);
	    s16[6*i+2] = convert (f[i+768]);
	    s16[6*i+3] = convert (f[i+1024]);
	    s16[6*i+4] = 0;
	    s16[6*i+5] = convert (f[i]);
	}
	break;
    case A52_3F2R | A52_LFE:
	for (i = 0; i < 256; i++) {
	    s16[6*i] = convert (f[i+256]);
	    s16[6*i+1] = convert (f[i+768]);
	    s16[6*i+2] = convert (f[i+1024]);
	    s16[6*i+3] = convert (f[i+1280]);
	    s16[6*i+4] = convert (f[i+512]);
	    s16[6*i+5] = convert (f[i]);
	}
	break;
    }
  }

  Message* AC3DecodeNode::decodeBlock() {
    /**
    * if the output device (soundcard driver) can handle special (dolby digital)
    * data, we must calculate the number of channels like this:
    */
    /*
    int aflags;
    aflags = flags & (A52_CHANNEL_MASK | A52_LFE);

    if (aflags & A52_LFE) channels = 6;
    else if (aflags & 1)
          channels = 5;
   	 else switch (aflags) {
    	   case A52_2F2R:
	     channels = 4;
	   break;
           case A52_CHANNEL:
           case A52_STEREO:
           case A52_DOLBY:
	     channels = 2;
	   break;
          }
   */
    block++;
    if ( block > 5 ) block = 0;

    if (a52_block(a52_state)) {
      return(0);
    }

    Buffer* out;
    out = getNewBuffer( 256 * sizeof (int16_t) * channels);
    float_to_int(a52_samples(a52_state), (int16_t*) out->getData() , flags);
    out->setUsed( 256 * sizeof (int16_t) * channels );
    
    samples += 256;
    
    // out->setMandatoryBit( true );
    
    return( out );

  }
  
  Message *AC3DecodeNode::processBuffer( Buffer *in_buffer ) {
    Message* out;
    bool in_buf_ts_taken = false;

    if( !getWorkingFlag() ) { // start processing new buffer
      buffer_start = (uint8_t*) in_buffer->getData();
      buffer_end   = buffer_start + in_buffer->getUsed();

      setWorkingFlag(true);
      old_in_buffer = in_buffer;
      
      Timestamp in_buf_stamp = in_buffer->getTimestamp();  // take the in_buffer's timestamp
      if(in_buf_stamp.is_valid) {                          // if it is valid
	timestamp.sync_time = in_buf_stamp.sync_time;
	in_buf_ts_taken = true;                            // set it for the next out_buffer
	DEBUG_STREAM("Take in_buffer timestamp: " << timestamp << endl);
        
        last_valid_ts = timestamp;
        samples = 0;
      }
    }

    if(!in_buf_ts_taken) {                  // if no timestamp has been taken from the in_buffer
      //      timestamp.sync_time += delay;         // simply add the duration of one buffer to the
                                            // last out_buffer's timestamp.
      
      double seconds = (double) samples / 48000.0;
      int sec = (int) seconds;
      int nsec = (int) ((seconds - (double) sec) * 1000000000.0);
      timestamp.sync_time = last_valid_ts.sync_time + Interval(sec, nsec);
      
      DEBUG_STREAM("Create own timestamp: " << timestamp << endl);
    }
    timestamp.is_valid = true;              // make timestamp valid
    timestamp.stream_counter++;             // don't forget to count the buffer.
    
    if ( (block > 0) && (block < 6) ) {
      out = decodeBlock();
      if ( out ) out->setTimestamp(timestamp); 
      return out;
      //return( decodeBlock() );
    } else {
      while (buffer_start < buffer_end) {
	*bufptr++ = *buffer_start++;
	if (bufptr == bufpos) {
	    if (bufpos == buf + 7) {
		int length;

                /**
                * Get Information about ac3 stream like samplerate, bitrate
                * and number of sound channels
                * Flags could contain this:
                *  A52_MONO      mono stream
                *  A52_STEREO    stereo stream
                *  A52_3F        3 front speakers
                *  A52_2F1R      2 front and 1 rear speaker
                *  A52_3F1R      3 front and 1 rear speaker
                *  A52_2F2R      2 front and 2 rear speaker
                *  A52_3F2R      3 front and 2 rear speaker
                *  A52_LFE       Low Frequency Effects (subwoofer)
                *
                *  A52_LFE with A52_3F2R means a dolby digital 5.1 stream
                */
		length = a52_syncinfo (buf, &flags, &sample_rate, &bit_rate);

		if (!length) {
		    for (bufptr = buf; bufptr < buf + 6; bufptr++)
			bufptr[0] = bufptr[1];
		} else bufpos = buf + length;
	    } else {
                bufptr = buf;
                bufpos = buf + 7;

                /* Here we must specify our output format and special adjustments with <flags> */
                /* for our purpose a stereo stream with volume level adjust is all right */
                level = 1;
		flags = A52_ADJUST_LEVEL | A52_STEREO;
		if ( !a52_frame (a52_state, buf, &flags, &level, bias))
		  {
		    out = decodeBlock();
		    if ( out ) out->setTimestamp(timestamp);
		    return out;
		  }
		  //return( decodeBlock() );
	    }
	}
     } //while

     setWorkingFlag(false);
     ///////////XXXXXXXXXXXXXXXXXXXXXX
     if (old_in_buffer)
       old_in_buffer->release();
     old_in_buffer = NULL;
     return(0);

    } //else
  }

  Result AC3DecodeNode::doFlush() {
    setWorkingFlag( false );
    if ( old_in_buffer ) {
      old_in_buffer->release();
      old_in_buffer = NULL;
    }

    block = 0;

    return( SUCCESS );

  }

  Result AC3DecodeNode::syncReset() {
    delay = 5333333;
    
    timestamp.sync_time = Time(0);
    timestamp.stream_counter = 0;
    
    DEBUG_STREAM("sync_reset completed." << endl);
    
    return SUCCESS;
  };
  
  Result AC3DecodeNode::startTrack(const string&) {
    DEBUG_STREAM("Message for me detected: startTrack" << endl);
    DEBUG_STREAM("Resetting AC3-Decoder-Lib" << endl);
    
    a52_state = a52_init(0);
    if (NULL == a52_state) {
      return FAILURE;
    }
    
    setWorkingFlag(false);
    return SUCCESS;
  }
  
  Result AC3DecodeNode::endTrack() {
    return SUCCESS;
  }
  
  Result AC3DecodeNode::started(const Interval&){
    return SUCCESS;
  }

  Result AC3DecodeNode::finished() {
    return SUCCESS;
  }
}




#include "nmm/nmm_suffix.hpp"
