/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

/*
 * Code is based on a52dec package that is
 * Copyrighted 2000-2001 by Michel Lespinasse <walken@zoy.org>
 * and copyrighted 1999-2000 by Aaron Holtzman <aholtzma@ess.engr.uvic.ca>
 */

#include "nmm/comm/format/Property.hpp"
#include "nmm/multimedia/Plugin.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"

#include "AC3ParseNode.hpp"

NMM_REGISTER_PLUGIN(NMM::AC3ParseNode, NMM, "AC3ParseNode")

namespace NMM {

  AC3ParseNode::AC3ParseNode( const char * name, StreamQueue::Mode qmode, int qsize )
    : GenericProcessorNode(name, qmode, qsize) {
  }
  
  AC3ParseNode::~AC3ParseNode() {
  }
  
  Result AC3ParseNode::doInit() {
    /* (in) Simple AC3 format */
    Format* i_format;
    i_format = getOwnInputProperty()->addNewFormat(Format::audio_ac3_type);
    getOwnInputProperty()->setDefaultFormat(i_format);
    
    /* (in) Simple AC3 format with bitrate known */
    i_format = getOwnInputProperty()->addNewFormat(Format::audio_ac3_type);
    i_format->addWildcard(Format::bitrate_param);
    
    /* (in) Fully specified AC3 format */
    i_format = getOwnInputProperty()->addNewFormat(Format::audio_ac3_type);
    i_format->addWildcard(Format::channels_param);
    i_format->addWildcard(Format::sample_rate_param);
    i_format->addWildcard(Format::bitrate_param);
    i_format->addWildcard(Format::speaker_setup_param);
    i_format->addIntParamValue(Format::lfe_param, 0);
    i_format->addIntParamValue(Format::lfe_param, 1);
    
    /* (out) Fully specified AC3 format */
    Format* o_format = getOwnOutputProperty()->addNewFormat(Format::audio_ac3_type);
    o_format->addWildcard(Format::channels_param);
    o_format->addWildcard(Format::sample_rate_param);
    o_format->addWildcard(Format::bitrate_param);
    o_format->addWildcard(Format::speaker_setup_param);
    o_format->addIntParamValue(Format::lfe_param, 0);
    o_format->addIntParamValue(Format::lfe_param, 1);
    getOwnOutputProperty()->setDefaultFormat(o_format);
    
    return SUCCESS;
  }
  
  Result AC3ParseNode::doInitOutput() {    
    const Format* in_format = getInputFormat();
    if (in_format == NULL) {
      ERROR_STREAM("No input format" << endl);
      return FAILURE;
    }
    
    /* initialize buffer pointers for ac3 decode */
    bufptr = buf;
    bufpos = buf + 7;
    
    sample_rate = 0;
    flags       = 0;
    channels    = 2;
    
    a52_state = a52_init(0);
    if (NULL == a52_state) {
      return FAILURE;
    }
    
    /* (in) Input connection format */
    Format* i_format = new Format(*in_format);
    getOwnWorkingInputProperty()->addNewFormat(i_format);
    getOwnWorkingInputProperty()->setDefaultFormat(i_format);
    
    /* analyse data to determine output format */
    if (!analyseData()) {
      ERROR_STREAM("Failed to analyse data" << endl);
      return FAILURE;
    }
    
    /* speaker configuration */
    string speaker_setup;
    bool has_lfe = flags & A52_LFE;
    int speaker_flags = has_lfe ? flags - A52_LFE : flags;
    switch(speaker_flags) {
    case A52_MONO:
      channels = 1;
      speaker_setup = "MONO";
      break;
    case A52_CHANNEL:
      speaker_setup = "CHANNEL";
      channels = 2;
      break;
    case A52_STEREO:
      speaker_setup = "STEREO";
      channels = 2;
      break;
    case A52_DOLBY:
      speaker_setup = "DOLBY";
      channels = 2;
      break;
    case A52_3F:
      channels = 3;
      speaker_setup = "3F";
      break;
    case A52_2F1R:
      channels = 3;
      speaker_setup = "2F1R";
      break;
    case A52_3F1R:
      channels = 4;
      speaker_setup = "3F1R";
      break;
    case A52_2F2R:
      channels = 4;
      speaker_setup = "2F2R";
      break;
    case A52_3F2R:
      channels = 5;
      speaker_setup = "3F2R";
      break;
    }
    
    /* (out) Fully specified AC3 format */
    Format* o_format = getOwnWorkingOutputProperty()->addNewFormat(Format::audio_ac3_type);
    o_format->addIntParamValue(Format::channels_param, channels);
    o_format->addIntParamValue(Format::sample_rate_param, sample_rate);
    o_format->addIntParamValue(Format::bitrate_param, bit_rate);
    o_format->addStringParamValue(Format::speaker_setup_param, speaker_setup);
    o_format->addIntParamValue(Format::lfe_param, has_lfe ? 1 : 0);
    getOwnWorkingOutputProperty()->setDefaultFormat(o_format);
    
    // set I/O partners
    i_format->setIOPartner(o_format);
    o_format->setIOPartner(i_format);
    
    return SUCCESS;
  }
  
  Message *AC3ParseNode::processBuffer( Buffer *in_buffer ) {
    /* just pass through buffers if we are not analysing them */
    if (!getAnalysingFlag()) {
      return in_buffer;
    }
    
    buffer_start = (uint8_t*) in_buffer->getData();
    buffer_end   = buffer_start + in_buffer->getUsed();
    
    while (buffer_start < buffer_end) {
      *bufptr++ = *buffer_start++;
      if (bufptr == bufpos) {
        if (bufpos == buf + 7) {
          /**
           * Get Information about ac3 stream like samplerate, bitrate
           * and number of sound channels
           * Flags could contain this:
           *  A52_MONO      mono stream
           *  A52_STEREO    stereo stream
           *  A52_3F        3 front speakers
           *  A52_2F1R      2 front and 1 rear speaker
           *  A52_3F1R      3 front and 1 rear speaker
           *  A52_2F2R      2 front and 2 rear speaker
           *  A52_3F2R      3 front and 2 rear speaker
           *  A52_LFE       Low Frequency Effects (subwoofer)
           *
           *  A52_LFE with A52_3F2R means a dolby digital 5.1 stream
           */
          int length = a52_syncinfo (buf, &flags, &sample_rate, &bit_rate);
          
          if (!length) {
            /* we need to resync. try again at next byte */
            for (bufptr = buf; bufptr < buf + 6; bufptr++)
              bufptr[0] = bufptr[1];
          } else {
            /* syncinfo read successfully, analysing finished */
            bufpos = buf + length;
            setAnalysingFlag(false);
            break;
          }
          
        } else {
          
          bufptr = buf;
          bufpos = buf + 7;
          
        }
      }
    } //while
    
    return in_buffer;
  }
  
  Result AC3ParseNode::doFlush() {
    return( SUCCESS );
  }
}



