/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "ALSAPlaybackNode.hpp"
#include "ALSAPlaybackNodePlugin.hpp"

#include "nmm/plugins/NMMRegisterPlugin.hpp"

#include "nmm/comm/messaging/TEDObject.hpp"

#include "nmm/utils/time/TimedElement.hpp"
#include "nmm/utils/time/Rational.hpp"
#include "nmm/utils/SecurityCheck.hpp"

#include <cmath>

// TODO: fw: ALSA doesn't use devices, so no need to use a device plugin.
// TODO: fw: however we DO need some sort of device management for ALSA!
// TODO: fw: use NMM_REGISTER_PLUGIN until ALSAPlaybackNodePlugin has been
// TODO: fw: correctly implemented
// static NMM::ALSAPlaybackNodePlugin<NMM::ALSAPlaybackNode> plugin("ALSAPlaybackNode",-1);
// NMM_REGISTER_SPECIAL_PLUGIN(ALSAPlaybackNode, NMM, plugin)
NMM_REGISTER_PLUGIN(ALSAPlaybackNode, NMM, "ALSAPlaybackNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  //! Formats to check in the doInit method.
  snd_pcm_format_t ALSAPlaybackNode::test_formats[] = {
    SND_PCM_FORMAT_MU_LAW,  //!< Logarithmic mu-law audio encoding
    SND_PCM_FORMAT_A_LAW,   //!< Logarithmic A-law audio encoding
    SND_PCM_FORMAT_U8,      //!< Linear encoding, unsigned 8-bit
    SND_PCM_FORMAT_S8,      //!< Linear encoding, signed 8-bit
    SND_PCM_FORMAT_U16_LE,  //!< Linear encoding, unsigned 16-bit, little endian
    SND_PCM_FORMAT_U16_BE,  //!< Linear encoding, unsigned 16-bit, big endian
    SND_PCM_FORMAT_S16_LE,  //!< Linear encoding, signed 16-bit, little endian
    SND_PCM_FORMAT_S16_BE,  //!< Linear encoding, signed 16-bit, big endian
    SND_PCM_FORMAT_U24_3LE,  //!< Linear encoding, unsigned 24-bit, little endian
    SND_PCM_FORMAT_U24_3BE,  //!< Linear encoding, unsigned 24-bit, big endian
    SND_PCM_FORMAT_S24_3LE,  //!< Linear encoding, signed 24-bit, little endian
    SND_PCM_FORMAT_S24_3BE,  //!< Linear encoding, signed 24-bit, big endian
    SND_PCM_FORMAT_UNKNOWN
    //!< End marker. If you add or remove some formats, make sure that the last one remains 0.
  };

  //! Channel numbers to check in the doInit method.
  const int ALSAPlaybackNode::test_channels[] = {
    1,      //!< Mono
    2,      //!< Stereo
    4,      //!< 4
    5,      //!< 5
    6,      //!< 5.1
    0
    //!< End marker. If you add or remove some values, make sure that the last one remains 0.
  };

  //! Sample rates to check in the doInit method.
  const int ALSAPlaybackNode::test_sample_rates[] = { 8000, 11025, 22050, 44100, 48000, 96000, 0 };

  Format& ALSAPlaybackNode::getNMMFormat(snd_pcm_format_t alsa_format, Format& nmm_format)
  {
    switch(alsa_format) {
    case SND_PCM_FORMAT_MU_LAW:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "mu_law");
      nmm_format.addIntParamValue(Format::bitpersample_param, 8);
      break;

    case SND_PCM_FORMAT_A_LAW:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "a_law");
      break;

    case SND_PCM_FORMAT_U8:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "linear");
      nmm_format.addStringParamValue(Format::sign_param, "unsigned");
      nmm_format.addIntParamValue(Format::bitpersample_param, 8);
      break;

    case SND_PCM_FORMAT_S8:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "linear");
      nmm_format.addStringParamValue(Format::sign_param, "signed");
      nmm_format.addIntParamValue(Format::bitpersample_param, 8);
      break;

    case SND_PCM_FORMAT_U16_LE:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "linear");
      nmm_format.addStringParamValue(Format::sign_param, "unsigned");
      nmm_format.addIntParamValue(Format::bitpersample_param, 16);
      nmm_format.addStringParamValue(Format::endian_param, "little");
      break;

    case SND_PCM_FORMAT_U16_BE:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "linear");
      nmm_format.addStringParamValue(Format::sign_param, "unsigned");
      nmm_format.addIntParamValue(Format::bitpersample_param, 16);
      nmm_format.addStringParamValue(Format::endian_param, "big");
      break;

    case SND_PCM_FORMAT_S16_LE:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "linear");
      nmm_format.addStringParamValue(Format::sign_param, "signed");
      nmm_format.addIntParamValue(Format::bitpersample_param, 16);
      nmm_format.addStringParamValue(Format::endian_param, "little");
      break;

    case SND_PCM_FORMAT_S16_BE:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "linear");
      nmm_format.addStringParamValue(Format::sign_param, "signed");
      nmm_format.addIntParamValue(Format::bitpersample_param, 16);
      nmm_format.addStringParamValue(Format::endian_param, "big");
      break;

    case SND_PCM_FORMAT_S24_3LE:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "linear");
      nmm_format.addStringParamValue(Format::sign_param, "signed");
      nmm_format.addIntParamValue(Format::bitpersample_param, 24);
      nmm_format.addStringParamValue(Format::endian_param, "little");
      break;
    case SND_PCM_FORMAT_S24_3BE:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "linear");
      nmm_format.addStringParamValue(Format::sign_param, "signed");
      nmm_format.addIntParamValue(Format::bitpersample_param, 24);
      nmm_format.addStringParamValue(Format::endian_param, "big");
      break;
    case SND_PCM_FORMAT_U24_3LE:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "linear");
      nmm_format.addStringParamValue(Format::sign_param, "unsigned");
      nmm_format.addIntParamValue(Format::bitpersample_param, 24);
      nmm_format.addStringParamValue(Format::endian_param, "little");
      break;
    case SND_PCM_FORMAT_U24_3BE:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "linear");
      nmm_format.addStringParamValue(Format::sign_param, "unsigned");
      nmm_format.addIntParamValue(Format::bitpersample_param, 24);
      nmm_format.addStringParamValue(Format::endian_param, "big");
      break;
    default:
      ERROR_STREAM("getNMMFormat: ALSA format " << alsa_format << " not supported." << endl);
    }

    return nmm_format;
  };

  snd_pcm_format_t ALSAPlaybackNode::getALSAFormat(const Format &nmm_format)
  {
    string encoding;
    try {
      encoding = nmm_format.getStringValue(Format::encoding_param);
    }
    catch (FormatException e) {
      ERROR_STREAM("getALSAFormat: Incomplete format (encoding not specified)." << endl);
      return SND_PCM_FORMAT_UNKNOWN;
    }

    if(encoding == "mu_law") return SND_PCM_FORMAT_MU_LAW;
    if(encoding == "a_law" ) return SND_PCM_FORMAT_A_LAW;
    if(encoding == "linear") {
      int bits;
      try {
        bits = nmm_format.getIntValue(Format::bitpersample_param);
      } catch (FormatException e) {
        ERROR_STREAM("getALSAFormat: Incomplete format (bitpersample not specified)." << endl);
        return SND_PCM_FORMAT_UNKNOWN;
      }
      string sign;
      try {
        sign = nmm_format.getStringValue(Format::sign_param);
      } catch (FormatException e) {
        ERROR_STREAM("getALSAFormat: Incomplete format (sign not specified)." << endl);
        return SND_PCM_FORMAT_UNKNOWN;
      }

      if(bits == 8) {
        if(sign == "unsigned") return SND_PCM_FORMAT_U8;
        if(sign == "signed"  ) return SND_PCM_FORMAT_S8;
        ERROR_STREAM("getALSAFormat: Incorrect format: sign is [" << sign << "]." << endl);
        ERROR_STREAM("getALSAFormat: Possible values for sign are: [unsigned|signed]." << endl);
      }

      if(bits == 16) {
        string endian;
        try {
          endian = nmm_format.getStringValue(Format::endian_param);
        } catch (FormatException e) {
          ERROR_STREAM("getALSAFormat: Incomplete format (endian not specified)." << endl);
          return SND_PCM_FORMAT_UNKNOWN;
        }

        if(sign == "unsigned") {
          if(endian == "little") return SND_PCM_FORMAT_U16_LE;
          if(endian == "big"   ) return SND_PCM_FORMAT_U16_BE;
          ERROR_STREAM("getALSAFormat: Incorrect format: endian is [" << endian << "]." << endl);
          ERROR_STREAM("getALSAFormat: Possible values for endian are [little|big]." << endl);
        }

        if(sign == "signed"  ) {
          if(endian == "little") return SND_PCM_FORMAT_S16_LE;
          if(endian == "big"   ) return SND_PCM_FORMAT_S16_BE;
          ERROR_STREAM("getALSAFormat: Incorrect format: endian is [" << endian << "]." << endl);
          ERROR_STREAM("getALSAFormat: Possible values for endian are [little|big]." << endl);
        }
        ERROR_STREAM("getALSAFormat: Incorrect format: sign is [" << sign << "]." << endl);
        ERROR_STREAM("getALSAFormat: Possible values for sign are [unsigend|signed]." << endl);
      }

      if(bits == 24) {
        string endian;
        try {
          endian = nmm_format.getStringValue(Format::endian_param);
        } catch (FormatException e) {
          ERROR_STREAM("getALSAFormat: Incomplete format (endian not specified)." << endl);
          return SND_PCM_FORMAT_UNKNOWN;
        }

        if(sign == "unsigned") {
          if(endian == "little") return SND_PCM_FORMAT_U24_3LE;
          if(endian == "big"   ) return SND_PCM_FORMAT_U24_3BE;
          ERROR_STREAM("getALSAFormat: Incorrect format: endian is [" << endian << "]." << endl);
          ERROR_STREAM("getALSAFormat: Possible values for endian are [little|big]." << endl);
        }
        if(sign == "signed"  ) {
          if(endian == "little") return SND_PCM_FORMAT_S24_3LE;
          if(endian == "big"   ) return SND_PCM_FORMAT_S24_3BE;
          ERROR_STREAM("getALSAFormat: Incorrect format: endian is [" << endian << "]." << endl);
          ERROR_STREAM("getALSAFormat: Possible values for endian are [little|big]." << endl);
        }
      }
      ERROR_STREAM("getALSAFormat: Only 8, 16 or 24 bitpersample are supported by ALSAPlaybackNode."
                  << endl);
    }
    return SND_PCM_FORMAT_UNKNOWN;
  };

  ALSAPlaybackNode::ALSAPlaybackNode(const char* name,
                                    const StreamQueue::Mode queue_mode,
                                    const int size)
    : GenericBSyncSinkNode(name, queue_mode, size),
      mixer_handle(0),
      mixer_pcm_element(0)
  {
    buffer_time = 0;
    buffer_frames = 0;
    period_time = 0;
    period_frames = 0;
    xfer_align = 0;

    pcm_access = SND_PCM_ACCESS_RW_INTERLEAVED;
    dev = "default";

    stream = SND_PCM_STREAM_PLAYBACK;

    byte_per_second = 0;
  }

  ALSAPlaybackNode::~ALSAPlaybackNode() {}

  Result ALSAPlaybackNode::initMixer() {
    /* reset mixer variables */
    mixer_handle = 0;
    mixer_pcm_element = 0;

    /* open mixer device */
    // HACK: fw: cardname is hardcoded (set to dev?)
    string cardname = "default";
    DEBUG_STREAM1("Opening mixer device for card '" << cardname << "'" << endl);
    if(snd_mixer_open(&mixer_handle, 0) < 0) {
      ERROR_STREAM("snd_mixer_open failed" << endl);
      return FAILURE;
    }
    if(snd_mixer_attach(mixer_handle, cardname.c_str()) < 0) {
      ERROR_STREAM("snd_mixer_attach failed for " << cardname << endl);
      deinitMixer();
      return FAILURE;
    }
    if(snd_mixer_selem_register(mixer_handle, NULL, NULL) < 0) {
      ERROR_STREAM("snd_mixer_selem_register failed" << endl);
      deinitMixer();
      return FAILURE;
    }
    if(snd_mixer_load(mixer_handle) < 0) {
      ERROR_STREAM("snd_mixer_load failed" << endl);
      deinitMixer();
      return FAILURE;
    }

    /* find mixer element for PCM output volume control */
    mixer_pcm_element = findMixerElementbyName(__mixer_element_name);
    if(!mixer_pcm_element) {
      ERROR_STREAM("No mixer control for output volume available" << endl);
      deinitMixer();
      return FAILURE;
    }

    /* set range of volume control */
    if(mixer_pcm_element) {
      // TODO: fw: Range for IAudioDevice interface
      snd_mixer_selem_set_playback_volume_range(mixer_pcm_element, 0, 255);
    }

    DEBUG_STREAM1("Mixer successfully initialized" << endl);
    return SUCCESS;
  }

  Result ALSAPlaybackNode::deinitMixer() {
    /* close mixer device */
    if(mixer_handle) {
      DEBUG_STREAM1("Closing mixer device" << endl);
      snd_mixer_close(mixer_handle);
    }

    /* reset variables */
    mixer_handle = 0;
    mixer_pcm_element = 0;

    return SUCCESS;
  }

  Result ALSAPlaybackNode::doInit() {
    DEBUG_STREAM("DoInit: testing devices"<<endl);

    // Default mixer device
    if(__mixer_device.empty()) {
      __mixer_device = "/dev/mixer";
    }

    // Default mixer element name
    if(__mixer_element_name.empty()) {
      __mixer_element_name = "PCM";
    }

    // Security check for mixer device
    if(!SSecurityCheck::getInstance().checkFilenameForWriting(__mixer_device)) {
      ERROR_STREAM("Security check not passed for file " << __mixer_device << ", check .nmmrc for options" << endl);
      return FAILURE;
    }

    int err;

    // Open PCM device
    if((err = snd_pcm_open(&pcm_handle, dev.c_str(), SND_PCM_STREAM_PLAYBACK, SND_PCM_ASYNC)) < 0) {
      ERROR_STREAM("Cannot open audio device " << dev <<" with error: "<< snd_strerror(err) << endl );
      return FAILURE;
    }
    DEBUG_STREAM("Audio device opened" << endl);

    // Allocate hardware parameters
    if((err = snd_pcm_hw_params_malloc(&hwparams)) < 0) {
      ERROR_STREAM("Cannot allocate hardware parameter structure. Error: " << snd_strerror(err) << endl);
      return FAILURE;
    }

    // Init hwparams with full configuration space
    if((err = snd_pcm_hw_params_any(pcm_handle, hwparams)) < 0) {
      ERROR_STREAM("Cannot initialize to configure this PCM device. Error: " << snd_strerror(err) << endl);
      return FAILURE;
    }

    // Set access hardware parameter
    if((err = snd_pcm_hw_params_set_access(pcm_handle, hwparams, pcm_access)) < 0) {
      ERROR_STREAM("Error setting access. Error: "<< snd_strerror(err) << endl);
      return FAILURE;
    }

    // Test all supported formats
    for(int i=0; test_formats[i] != SND_PCM_FORMAT_UNKNOWN; i++) {
      // HACK: marthi: Set of sample rates to skip in tests
      const size_t skip_sample_rates_size = 10;
      int skip_sample_rates[] = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

      // Reset to full configuration space
      snd_pcm_hw_params_any(pcm_handle, hwparams);
      snd_pcm_hw_params_set_access(pcm_handle, hwparams, pcm_access);

      // Test all the formats in test_formats which are supported by the hardware.
      if(snd_pcm_hw_params_set_format(pcm_handle, hwparams, test_formats[i]) < 0) {
        DEBUG_STREAM("audio device: - Format " << test_formats[i] << " not supported." << endl);
      }
      else {
        // This format is supported.
        DEBUG_STREAM("audio device: - Format " << test_formats[i] << endl);

        // Test the supported channel numbers for this format.
        // Note: If you need a channel number that is currently not supported by
        //       ALSAPlaybackNode, simply add it to the definition of the array
        //       test_channels in this file.
        for(int j=0; test_channels[j] != 0; j++) {
          int test_channel = test_channels[j];

          // Reset to full configuration space
          snd_pcm_hw_params_any(pcm_handle, hwparams);
          snd_pcm_hw_params_set_access(pcm_handle, hwparams, pcm_access);
          snd_pcm_hw_params_set_format(pcm_handle, hwparams, test_formats[i]);

          if(snd_pcm_hw_params_set_channels(pcm_handle,hwparams,test_channel) < 0) {
            DEBUG_STREAM("audio device: - Format " << test_formats[i] << " does not support " << test_channel << " channels."<<endl);
          }
          else {
            DEBUG_STREAM("audio device: - Format " << test_formats[i] << " supports " << test_channel << " channels."<<endl);

            // Create a new format.
            Format *input_format = new Format();

            // Configure the parameters of the format according to the actual
            // hardware configuration (since this configuration works.)
            getNMMFormat(test_formats[i], *input_format);

            // Add the format to the input property.
            getOwnInputProperty() -> addNewFormat(input_format);
            if (!getOwnInputProperty()->getDefaultFormat()) {
              // workaround for missing default format
              // setting first format found to default one
              getOwnInputProperty()->setDefaultFormat(input_format);
            }

            // Set channels and sample rate (not for ac3)
            input_format->addIntParamValue(Format::channels_param, test_channels[j]);

            // Last step: Test the supported sample rates for the current channel number
            //            and format.
            // Note: If you need a sample rate that is not tested here, simpy add it to
            //       the definition of the array test_sample_rates in this file.
            for(unsigned int k=0; test_sample_rates[k] != 0; k++) {
              if(k < skip_sample_rates_size) {
                // HACK: marthi: if once a sample_rate failed,
                // assume it will fail for other channels as well.
                if (skip_sample_rates[k] == test_sample_rates[k]) {
                  DEBUG_STREAM("Sample rate " << skip_sample_rates[k] << " skipped" << endl);
                  continue;
                }
              }

              int test_rate = test_sample_rates[k];

              // Reset to full configuration space
              snd_pcm_hw_params_any(pcm_handle, hwparams);
              snd_pcm_hw_params_set_access(pcm_handle, hwparams, pcm_access);
              snd_pcm_hw_params_set_format(pcm_handle, hwparams, test_formats[i]);
              snd_pcm_hw_params_set_channels(pcm_handle, hwparams, test_channel);

              if((err = snd_pcm_hw_params_test_rate(pcm_handle, hwparams, test_rate, 0)) < 0) {
                DEBUG_STREAM("audio device:     + Set sample rate to " << test_sample_rates[k]
                    << " failed. Error: "<< snd_strerror(err) << endl);
                if(k < skip_sample_rates_size) {
                  // HACK: marthi: if once a sample_rate failed,
                  // assume it will fail for other channels as well.
                  skip_sample_rates[k] = test_sample_rates[k];
                }
              }
              else {
                if(test_rate == test_sample_rates[k]) {
                  // Sample rate is supported with the actual configuration.
                  DEBUG_STREAM("audio device:     + Sample rate " << test_sample_rates[k] << endl);

                  // Add it to the format in the input property.
                  input_format->addIntParamValue(Format::sample_rate_param, test_sample_rates[k]);
                }
              }
            } // loop over the test_sample_rates
          }
        } //loop over test_channels
      }

    }

    initMixer();

    return SUCCESS;
  };

  Result ALSAPlaybackNode::doDeinit()
  {
    deinitMixer();

    snd_pcm_close(pcm_handle);
    return SUCCESS;
  };

  Result ALSAPlaybackNode::doInitOutput() {
    // Get the input format
    const Format* input_format = getInputFormat();
    if (input_format == NULL) {
      ERROR_STREAM("Input format not set."<<endl);
      return FAILURE;
    }

    // Set the working output property
    Format* format = new Format( *input_format );
    getOwnWorkingInputProperty()->addNewFormat(format);
    getOwnWorkingInputProperty()->setDefaultFormat(format);

    // Set the format at the device
    if(doSetFormat(*input_format) == FAILURE) {
      return FAILURE;
    }

    return SUCCESS;
  }

  Result ALSAPlaybackNode::doFlush()
  {
    // Drop all audio data.
    // This puts the device in unprepared state. The device needs to be prepared again before being started.
    // (We do not do this here, though, to guarantee that doFlush does not fail)
    snd_pcm_drop(pcm_handle);

    // Prepare the device again
    // (We do this here to guarantee that doStart fails if the device can not be prepared)
    snd_pcm_prepare(pcm_handle);
    flushed = false;

    // TEST: fw: Instead of snd_pcm_prepare, completely re-initialize the device
//    const Format* input_format = getInputFormat();
//    if(doSetFormat(*input_format) == FAILURE) {
//      return FAILURE;
//    }

    return SUCCESS;
  };

  Result ALSAPlaybackNode::doStart()
  {
    return SUCCESS;
  };

  Result ALSAPlaybackNode::doStop()
  {
    return SUCCESS;
  };

  Message* ALSAPlaybackNode::presentBuffer(Buffer *in_buffer) {
    unsigned int used = in_buffer->getUsed();
    unsigned int frames = used / (byte_per_sample*channels);
    unsigned int written = 0;
    ssize_t r;

    // HACK: fw: Without this call, no sound will be heard on Audigy 4 PRO
    // HACK: fw: This must be a bug in this node, but unfortunately I can not find it :(
    snd_pcm_start(pcm_handle);

    while(written < frames) {
      if (pcm_access == SND_PCM_ACCESS_RW_NONINTERLEAVED) {
        // NOT IMPLEMENTED
      }
      else {
        r = snd_pcm_writei(pcm_handle, in_buffer->getData() + (written*byte_per_sample*channels), frames - written);
        if (r == -EAGAIN || (r >= 0 && (size_t)r < frames)) {
          snd_pcm_wait(pcm_handle, 1000);
        }
        else if (r == -EPIPE) {
          xrun();
        }
        else if (r == -ESTRPIPE) {
          suspend();
        }
        else if (r < 0) {
          ERROR_STREAM("WRITE ERROR: " << snd_strerror(r) << endl);
          in_buffer->release();
          return 0;
        }

        if (r > 0) {
          DEBUG_STREAM("Wrote " << r <<" frames to buffer." << endl);
          written += r;
        }
      }
    }

    in_buffer->release();
    getOutputDelay();
    return 0;
  }

  Result ALSAPlaybackNode::xrun(void) {
    snd_pcm_status_t *status;
    int res;

    snd_pcm_status_alloca(&status);
    if ((res = snd_pcm_status(pcm_handle, status))<0) {
      ERROR_STREAM("xrun(): status error: "<< snd_strerror(res)<<endl);
      return FAILURE;
    }
    if (snd_pcm_status_get_state(status) == SND_PCM_STATE_XRUN) {
      struct timeval now, diff, tstamp;
      gettimeofday(&now, 0);
      snd_pcm_status_get_trigger_tstamp(status, &tstamp);
      timersub(&now, &tstamp, &diff);
      DEBUG_STREAM("xrun(): underrun!!! (at least " << diff.tv_sec * 1000 + diff.tv_usec / 1000.0 << " ms long."<<endl);

      if ((res = snd_pcm_prepare(pcm_handle))<0) {
        ERROR_STREAM("xrun(): prepare error: "<< snd_strerror(res)<<endl);
        return FAILURE;
      }

      // TODO: fw: Handle underrun

      return SUCCESS;         /* ok, data should be accepted again */
    }
    if (snd_pcm_status_get_state(status) == SND_PCM_STATE_DRAINING) {
      DEBUG_STREAM("xrun(): Status(DRAINING)"<<endl);
      // snd_pcm_status_dump(status, log);
    }
    ERROR_STREAM("xrun(): read/write error, state = "<< snd_pcm_state_name(snd_pcm_status_get_state(status)) <<endl );

    return FAILURE;
  }

  Result ALSAPlaybackNode::suspend(void)
  {
    int res;

    DEBUG_STREAM("Suspended. Trying resume. "<<endl);
    while ((res = snd_pcm_resume(pcm_handle)) == -EAGAIN) {
      sleep(1);       /* wait until suspend flag is released */
    }
    if (res < 0) {
      DEBUG_STREAM("Failed. Restarting stream. "<<endl);
      if ((res = snd_pcm_prepare(pcm_handle)) < 0) {
        ERROR_STREAM("suspend(): prepare error: "<< snd_strerror(res)<< endl);
        return FAILURE;
      }
    }
    DEBUG_STREAM("Suspend Done."<<endl);
    return SUCCESS;
  }

  void ALSAPlaybackNode::setPCMDevice(const string& device) {
    /* Begin> Check device */
    int err;
    if ((err = snd_pcm_open (&pcm_handle, device.c_str(), SND_PCM_STREAM_PLAYBACK, SND_PCM_ASYNC)) < 0) {
      ERROR_STREAM("Cannot open audio device " << device <<" with error: "<< snd_strerror(err) << endl );
      throw WrongDeviceException(string("Cannot open audio device ") + string( device ) + string ( " with error: " ) + string( snd_strerror(err) ));
    }
    snd_pcm_close (pcm_handle);
    /* End< */

    dev = device;

    DEBUG_STREAM("Set device to " << dev << endl);
  }

  string ALSAPlaybackNode::getPCMDevice() const {
    return dev;
  }

  void ALSAPlaybackNode::setMixerDevice(const string& mixer_device)
    throw (const DeviceNotFoundException&)
  {
    __mixer_device = mixer_device;
    return;
  }

  string ALSAPlaybackNode::getMixerDevice() const {
    return __mixer_device;
  }

  void ALSAPlaybackNode::setMixerVolumeElement(const string& value)
  {
    __mixer_element_name = value;
  }

  void ALSAPlaybackNode::setBufferTime(const nmm_uint32& buffertime) {
    buffer_time = buffertime;
  }

  void ALSAPlaybackNode::setPeriodTime(const nmm_uint32& periodtime) {
    period_time = periodtime;
  }

  void ALSAPlaybackNode::setXferAlign (const nmm_uint32& xferalign) {
    xfer_align = (size_t)xferalign;
  }

  Interval ALSAPlaybackNode::getOutputDelay()
  {
    snd_pcm_sframes_t delay;
    ssize_t bytes = 0;
    Interval delay_interval(0,0);
    Interval null_interval(0,0);

    int r = snd_pcm_delay(pcm_handle, &delay);
    if (r == 0) {
      if(delay > 0) {
      }
      else if(delay < 0) {
        return null_interval;
      }
      else { // delay == 0
        return null_interval;
      }

      bytes = snd_pcm_frames_to_bytes(pcm_handle, delay);
      delay_interval = Interval(Rational(bytes, byte_per_second));
      DEBUG_STREAM("ALSAPlaybackNode::getOutputDelay " << delay_interval << endl);

      if(delay_interval == null_interval) {
        return null_interval;
      }
      else {
        return delay_interval;
      }
    }
    else {
      return null_interval;
    }
  };

  Result ALSAPlaybackNode::setFormatInternal(Format& input_format) {
    doSetFormat(input_format);
    return SUCCESS;
  }

  Result ALSAPlaybackNode::doSetFormat(const Format& input_format) {
    // Reset the hwparams structure, so we have a full configuration space
    snd_pcm_hw_params_any(pcm_handle,hwparams);

    int err;
    buffer_time = period_time = buffer_frames = period_frames = 0;
    DEBUG_STREAM("Setting format at audio device." << endl);

    /* Get format parameters */
    snd_pcm_format_t format = getALSAFormat(input_format);
    if(format == SND_PCM_FORMAT_UNKNOWN) {
      return FAILURE;
    }
    channels    = input_format.getIntValue(Format::channels_param);
    sample_rate = input_format.getIntValue(Format::sample_rate_param);
    DEBUG_STREAM("Got Format: samplerate: " << sample_rate << ", channels: " << channels << endl);

    try {
      bit_per_sample = input_format.getIntValue(Format::bitpersample_param);
    }
    catch(...) {
      bit_per_sample = 16;
    }
    if (bit_per_sample == 0) bit_per_sample = 16;

    byte_per_sample = bit_per_sample / 8;
    byte_per_second = sample_rate * channels * byte_per_sample;

    // It is important always to keep the same order in setting the parameters:
    // At first the access, format, then the channel number and at last the sample rate.
    DEBUG_STREAM("audio device: Trying to set access. " << endl);
    if ( (err = snd_pcm_hw_params_set_access(pcm_handle, hwparams, pcm_access)) < 0) {
      ERROR_STREAM("Error setting access at init output. Error: "<< snd_strerror(err) << endl);
      return FAILURE;
    }

    // Set sample format
    DEBUG_STREAM("audio device: Trying to set format " << format << endl);
    if (snd_pcm_hw_params_set_format(pcm_handle, hwparams, format) < 0) {
      ERROR_STREAM("Error setting format " << format << "."<<endl);
      return FAILURE;
    }
    DEBUG_STREAM("audio device: Set format to " << format << endl);

    // Set number of channels
    DEBUG_STREAM("audio device: Trying to set channels to " << channels << endl);
    if (snd_pcm_hw_params_set_channels(pcm_handle, hwparams, channels) < 0) {
      ERROR_STREAM("Error setting channels."<<endl);
      return FAILURE;
    }
    DEBUG_STREAM("audio device: Set channels to " << channels << endl);

    // Set sample rate.
    // If the exact rate is not supported by the hardware, use nearest possible rate.
    DEBUG_STREAM("audio device: Trying to set sample_rate to " << sample_rate << "." << endl);
    unsigned int exact_rate = sample_rate;
    if(snd_pcm_hw_params_set_rate_near(pcm_handle, hwparams, &exact_rate, 0) < 0) {
      ERROR_STREAM("Error setting rate.");
      return FAILURE;
    }
    if(sample_rate != exact_rate) {
      DEBUG_STREAM("The rate " << sample_rate << " Hz is not supported by your hardware. " <<
                   "==> Using " << exact_rate << " Hz instead."<<endl);
      sample_rate = exact_rate;
    }
    DEBUG_STREAM("audio device: Set sample_rate to " << sample_rate << "." << endl);

    if (buffer_time == 0 && buffer_frames == 0) {
      if( snd_pcm_hw_params_get_buffer_time_max(hwparams, &buffer_time, 0) < 0) {
        ERROR_STREAM("audio device: I can't get  buffer time."<<endl);
        return FAILURE;
      }
      // if the hardware buffer is too big we set it to 500000 us
      if (buffer_time > 500000) {
        buffer_time = 500000;
      }
    }
    DEBUG_STREAM("audio device: buffer_time "<< buffer_time << " buffer_frames "<<buffer_frames<<endl);
    if (period_time == 0 && period_frames == 0) {
      if (buffer_time > 0) {
        period_time = buffer_time / 4;
      }
      else {
        period_frames = buffer_frames / 4;
      }
    }
    DEBUG_STREAM("audio device: period_time "<<period_time<<" period_frames "<<period_frames<<endl);
    if (period_time > 0) {
      if (snd_pcm_hw_params_set_period_time_near(pcm_handle, hwparams, &period_time, 0) < 0) {
        ERROR_STREAM("audio device: Error setting period time."<<endl);
        return FAILURE;
      }
      // after setting the period time we can get the corresponding period frames from the device
      if (snd_pcm_hw_params_get_period_size(hwparams, &period_frames, 0)) {
        ERROR_STREAM("audio device: Could not get period size from device." << endl);
      }
    }
    else {
      if ( snd_pcm_hw_params_set_period_size_near(pcm_handle, hwparams, &period_frames, 0) < 0) {
        ERROR_STREAM("audio device: Error setting period size."<<endl);
        return FAILURE;
      }
      // after setting the period size we can get the corresponding period time from the device
      if (snd_pcm_hw_params_get_period_time(hwparams, &period_time, 0)) {
        ERROR_STREAM("audio device: Could not get period time from device." << endl);
      }
    }
    DEBUG_STREAM("audio device: period buffer set"<<endl);
    if (buffer_time > 0) {
      if (snd_pcm_hw_params_set_buffer_time_near(pcm_handle, hwparams, &buffer_time, 0) < 0) {
        ERROR_STREAM("audio device: Error setting buffer time."<<endl);
        return FAILURE;
      }
      // after setting the buffer time we can get the corresponding buffer frames from the device
      if(snd_pcm_hw_params_get_buffer_size(hwparams, &buffer_frames)) {
        ERROR_STREAM("audio device: Could not get buffer size from device." << endl);
        return FAILURE;
      }
    }
    else {
      if(snd_pcm_hw_params_set_buffer_size_near(pcm_handle, hwparams, &buffer_frames) < 0) {
        ERROR_STREAM("audio device: Error setting buffer size."<<endl);
        return FAILURE;
      }
      // after setting the buffer size we can get the corresponding buffer time from the device
      if(snd_pcm_hw_params_get_buffer_time(hwparams, &buffer_time, 0)) {
        ERROR_STREAM("audio device: Could not get buffer time from device." << endl);
        return FAILURE;
      }
    }

    DEBUG_STREAM("audio device: Buffer set."<<endl);

    // Assign the hardware parameters to the device
    if(snd_pcm_hw_params(pcm_handle, hwparams) < 0) {
      ERROR_STREAM("audio device: Error setting HW params."<<endl);
      return FAILURE;
    }


    // Setting the software parameter on the device

    // Allocate software parameters
    if((err = snd_pcm_sw_params_malloc(&swparams)) < 0) {
      ERROR_STREAM("Cannot allocate software parameter structure with error " << snd_strerror(err) << endl);
      return FAILURE;
    }

    // Get current software parameters
    if((err = snd_pcm_sw_params_current(pcm_handle, swparams)) < 0) {
      ERROR_STREAM("Unable to determine current swparams for playback: " << snd_strerror(err) << endl);
      return FAILURE;
    }

    // Get the boundary from the software parameters
    snd_pcm_uframes_t boundary;
    if((err = snd_pcm_sw_params_get_boundary(swparams, &boundary)) < 0) {
      ERROR_STREAM("Couldn't get boundary from the device" << endl);
      boundary = 0;
    }

    if((err = snd_pcm_sw_params_set_start_threshold(pcm_handle, swparams, boundary)) < 0) {
      ERROR_STREAM("Unable to set start threshold mode for playback: " << snd_strerror(err)<< endl);
      return FAILURE;
    }

    // Allow the transfer when at least period_frames samples can be processed
    if(period_frames > 0) {
      // set to same value as period_frames queried from alsa (same approach is used in aplay)
      err = snd_pcm_sw_params_set_avail_min(pcm_handle, swparams, period_frames);
      if (err < 0) {
        ERROR_STREAM("Unable to set avail min for playback: " << snd_strerror(err) << endl);
        return FAILURE;
      }
    }

    // Set the transfer alignment
    if(xfer_align > 0) {
      err = snd_pcm_sw_params_set_xfer_align(pcm_handle, swparams, xfer_align);
      if (err < 0) {
        ERROR_STREAM("Unable to set transfer align for playback: " << snd_strerror(err) << endl);
        return FAILURE;
      }
    }

    // Set the software parameters
    err = snd_pcm_sw_params(pcm_handle, swparams);
    if (err < 0) {
      ERROR_STREAM("Unable to set sw params for playback: " << snd_strerror(err) << endl);
      return FAILURE;
    }

    return SUCCESS;
  };

  snd_mixer_elem_t* ALSAPlaybackNode::findMixerElementbyName(const string& name) {
    /* iterate over all mixer elements */
    for(snd_mixer_elem_t* mixer_elem = snd_mixer_first_elem(mixer_handle);
        mixer_elem != 0;
        mixer_elem = snd_mixer_elem_next(mixer_elem)) {
      if(0 == strcmp(snd_mixer_selem_get_name(mixer_elem), name.c_str() )) {
        DEBUG_STREAM1("Found mixer element '" << name << "'" << endl);
        return mixer_elem;
      }
    }

    /* no mixer element found */
    ERROR_STREAM1("Mixer element '" << name << "' not found" << endl);
    return 0;
  }

  Rational ALSAPlaybackNode::convertToLinearVolume(double logarithmic_volume) {
    return Rational(uint64_t(pow(10, logarithmic_volume/ 20) * 0x7FFFFFFFU ), uint64_t(0x7FFFFFFFU));
  }

  double ALSAPlaybackNode::convertToLogarithmicVolume(const Rational& linear_volume) {
    /* compute dezibel value as 20*lg(U2/U1)dB where U2/U1 is the volume in percent.
      note that "lg" stands for log_10 which in C/C++ is "log10".
      also note that log(U2)-log(U1) is numerically more stable than log(U2/U1).
      see http://de.wikipedia.org/wiki/Dezibel (in German) */
    return 20 * (log10(double(linear_volume.getNumerator())) - log10(double(linear_volume.getDenominator())));
  }

  void ALSAPlaybackNode::setLineInVolume( int /* vol */ ) {
    DEBUG_STREAM("setLineInVolume: mixer command isn't supported"<<endl);
  }

  int ALSAPlaybackNode::getLineInVolume() {
    DEBUG_STREAM("getLineInVolume: mixer command isn't supported"<<endl);
    return 0;
  }

  void ALSAPlaybackNode::setMicroInVolume(const nmm_int32&) {}
  nmm_int32 ALSAPlaybackNode::getMicroInVolume() { return 0;}

  void ALSAPlaybackNode::setVolume(const nmm_int32& volume) {
    /* set volume using some undocumented ALSA function */
    if(mixer_pcm_element) {
      DEBUG_STREAM1("Set volume to " << volume << std::endl);
      snd_mixer_selem_set_playback_volume_all(mixer_pcm_element, volume);
    } else {
      ERROR_STREAM1("Can not set volume (no mixer)" << std::endl);
    }
  }

  nmm_int32 ALSAPlaybackNode::getVolume() {
    /* get volume using some undocumented ALSA function */
    long volume;
    if(mixer_pcm_element) {
      snd_mixer_selem_get_playback_volume(mixer_pcm_element, SND_MIXER_SCHN_FRONT_LEFT, &volume);
      DEBUG_STREAM1("Get volume. Current volume is " << volume << std::endl);
      return static_cast<int>(volume);
    } else {
      ERROR_STREAM1("Can not get volume (no mixer)" << std::endl);
      return 0;
    }
  }

  void ALSAPlaybackNode::increaseVolume()
  {
    /* use setVolume/getVolume */
    setVolume(getVolume()+10);
  }

  void ALSAPlaybackNode::decreaseVolume()
  {
    /* use setVolume/getVolume */
    setVolume(getVolume()-10);
  }

  Rational ALSAPlaybackNode::setLinearVolume(const Rational& linear_volume) {
    /* set volume using some undocumented ALSA function */
    if(mixer_pcm_element) {
      DEBUG_STREAM1("Set volume to " << linear_volume << std::endl);
      snd_mixer_selem_set_playback_volume_all(mixer_pcm_element, linear_volume.getNumerator() * 255 / linear_volume.getDenominator());
    } else {
      ERROR_STREAM1("Can not set volume (no mixer)" << std::endl);
    }

    return linear_volume;
  }

  Rational ALSAPlaybackNode::setLinearVolumeRelative(const Rational& relative_linear_volume) {
    return setLinearVolume(getLinearVolume() + relative_linear_volume);
  }

  Rational ALSAPlaybackNode::getLinearVolume() const {
    /* get volume using some undocumented ALSA function */
    long volume;
    if(mixer_pcm_element) {
      snd_mixer_selem_get_playback_volume(mixer_pcm_element, SND_MIXER_SCHN_FRONT_LEFT, &volume);
      DEBUG_STREAM1("Get volume. Current volume is " << volume << std::endl);
      return Rational(volume, 255);
      return static_cast<int>(volume);
    } else {
      ERROR_STREAM1("Can not get volume (no mixer)" << std::endl);
      return Rational(0,1);
    }
  }

  nmm_double ALSAPlaybackNode::setLogarithmicVolume(const nmm_double& volume_in_db) {
    Rational linear_volume = convertToLinearVolume(volume_in_db);
    linear_volume = setLinearVolume(linear_volume);
    return convertToLogarithmicVolume(linear_volume);
  }

  nmm_double ALSAPlaybackNode::setLogarithmicVolumeRelative(const nmm_double& relative_volume_in_db) {
    return setLogarithmicVolume(getLogarithmicVolume() + relative_volume_in_db);
  }

  nmm_double ALSAPlaybackNode::getLogarithmicVolume() const {
    Rational linear_volume = getLinearVolume();
    return convertToLogarithmicVolume(linear_volume);
  }

  void ALSAPlaybackNode::getLinearVolumeRange(Rational& min_linear_volume, Rational& max_linear_volume) const {
    min_linear_volume = Rational(0,255);
    max_linear_volume = Rational(255,255);
  }

  void ALSAPlaybackNode::getLogarithmicVolumeRange(nmm_double& min_volume_in_db, nmm_double& max_volume_in_db) const {
    Rational min_linear_volume, max_linear_volume;
    getLinearVolumeRange(min_linear_volume, max_linear_volume);
    min_volume_in_db = convertToLogarithmicVolume(min_linear_volume);
    max_volume_in_db = convertToLogarithmicVolume(max_linear_volume);
  }

  void ALSAPlaybackNode::mute()
  {
    DEBUG_STREAM("MUTE" << endl);
    getController()->setBuffersInvalid(true);
  }

  void ALSAPlaybackNode::unmute()
  {
    DEBUG_STREAM("UN-MUTE" << endl);
    getController()->setBuffersInvalid(false);
  }

}

#include "nmm/nmm_suffix.hpp"
