/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#ifndef NMM_ALSAPLAYBACKNODE_HPP
#define NMM_ALSAPLAYBACKNODE_HPP NMM_ALSAPLAYBACKNODE_HPP

#include "nmm/nmm_prefix.hpp"

#include "nmm/multimedia/sync/GenericBSyncSinkNode.hpp"

#include "nmm/multimedia/DeviceNotFoundException.hpp"

#include "nmm/plugins/audio/interfaces/IAudioDeviceImpl.hpp"
#include "nmm/plugins/audio/interfaces/IVolumeControlImpl.hpp"
#include "nmm/multimedia/interfaces/IMuteImpl.hpp"

#include "nmm/plugins/audio/alsa/IALSADeviceImpl.hpp"

#include <alsa/asoundlib.h>

namespace NMM {

  /** This class plays an audio stream to a given output device.
   *
   * The ALSAPlaybackNode uses the ALSA audio driver.
   */
  class ALSAPlaybackNode :
    public GenericBSyncSinkNode,
    public IMuteImpl,
    public IALSADeviceImpl,
    public IAudioDeviceImpl,
    public IVolumeControlImpl
  {
  private:

    snd_pcm_t *pcm_handle; ///<  Handle for the PCM device 

    snd_pcm_stream_t stream; ///< Playback stream

    /* This structure contains information about    */
    /* the hardware and can be used to specify the  */
    /* configuration to be used for the PCM stream. */
    snd_pcm_hw_params_t *hwparams;

    snd_pcm_sw_params_t *swparams; ///< PCM software configuration container

    unsigned int buffer_time;  ///< length of a buffer used by alsa in us
    unsigned int period_time;  ///< length of a period used by alsa in us
    snd_pcm_uframes_t period_frames; ///< frames per period
    snd_pcm_uframes_t buffer_frames; ///< frames per buffer

    size_t xfer_align; ///< parameter of alsa
		
    snd_pcm_status_t *pcm_status; ///< status of device

    snd_pcm_access_t pcm_access; ///< Used to set the pcm access to INTERLEAVED

    static snd_pcm_format_t test_formats[]; ///< Formats to check in the doInit method.
		
    static const int test_channels[]; ///< Channels to check in the doInit method.
		
    static const int test_sample_rates[]; ///< Sample rates to check in the doInit method.
		
    /** Converts an ALSA_Format into an NMM::Format.
     *
     * @param alsa_format alsa format to convert
     * @param nmm_format Format variable to write converted alsa format to
     * @return reference to alsa format converted to nmm Format
     */
    Format& getNMMFormat(snd_pcm_format_t alsa_format, Format& nmm_format);
		
    /** Converts an NMM::Format into an ALSA_Format.
     *
     * @param nmm_format nmm Format to convert
     * @return nmm format converted to alsa format
     */
    snd_pcm_format_t getALSAFormat(const Format& nmm_format);

    /** Sets device string.
     *
     * @param value name of device to set
     */
    virtual void setDevice(const string& value) /* limited(INode::CONSTRUCTED) */
      throw(const DeviceNotFoundException&, const WrongDeviceException&);

    /** Returns device string.
     *
     * @return name of device used.
     */
    virtual string getDevice() const { return "default"; }

    /** @name IVolumeControl Interface
     */
    //@{
    /** Sets the linear volume of an audio output device.
     *
     * @param linear_volume volume absolute.
     * @return new volume
     */
    virtual Rational setLinearVolume(const Rational& linear_volume);

    /** Sets the linear volume of an audio output device relative to the current volume.
     *
     * @param relative_linear_volume volume relative to output device.
     * @return new volume
     */
    virtual Rational setLinearVolumeRelative(const Rational& relative_linear_volume);

    /** Returns the linear volume of an audio output device.
     *
     * @return linear volume
     */
    virtual Rational getLinearVolume() const;

    /** Sets the logarithmic volume of an audio output device.
     *
     * @param volume_in_db Volume logarithmic absolut in db. 
     * @return new logarithmic volume
     */
    virtual nmm_double setLogarithmicVolume(const nmm_double& volume_in_db);

    /** Sets the logarithmic volume of an audio output device relative to the current logarithmic volume.
     *
     * @param relative_volume_in_db Volume logarithmic in db relative to an audio output devie.
     * @return new logarithmic volume
     */
    virtual nmm_double setLogarithmicVolumeRelative(const nmm_double& relative_volume_in_db);

    /** Returns the logarithmic volume of an audio output device.
     *
     * @return Logartihmic volume of an audio output device.
     */
    virtual nmm_double getLogarithmicVolume() const;

    /** Returns the linear volume range of an audio output device.
     *
     * @param min_linear_volume variable to receive linear minimum volume
     * @param max_linear_volume variable to receive linear maximum volume
     */
    virtual void getLinearVolumeRange(Rational& min_linear_volume, Rational& max_linear_volume) const;

    /** Returns the logarithmic volume range of an audio output device.
     *
     * @param min_volume_in_db variable to receive minimum volume in db
     * @param max_volume_in_db variable to receive maximum volume in db
     */
    virtual void getLogarithmicVolumeRange(nmm_double& min_volume_in_db, nmm_double& max_volume_in_db) const;
    //@}

  private:
    unsigned int sample_rate;               ///< The sample_rate (depends on input_format)
    int channels;                           ///< The number of channels (depends on input_format)
    int bit_per_sample;                     ///< The bits per sample (depends on input format)
    int byte_per_sample;                    ///< The bytes per sample (depends on input format)
    int byte_per_second;                    ///< The bytes per second (depends on input format)

    snd_mixer_t *mixer_handle;              ///< Mixer handle
    snd_mixer_elem_t* mixer_pcm_element;    ///< handle for PCM mixer element

    bool flushed;                           //!< Device was flushed

    string dev;                             ///< The name of the sound device.
    string __mixer_device;                  ///< The name of the mixer device.
    string __mixer_element_name;            ///< The name of the mixer element.

    /** Handler method for xrun.
     *
     * @return SUCCESS if xrun could be handled successfully, false otherwise.
     */
    Result xrun(void);
		
    /** Handler method for suspend.
     *
     * @return SUCCESS if resume after suspend was successful, FAILURE otherwise.
     */
    Result suspend(void);

    /** @name Mixer Device Methods
     */
    //{
    /** Initializes the mixer device and searches PCM element.
     * 
     * This method opens the mixer device and finds an element for PCM output
     * volume control. If this succeeds, SUCCESS is returned, and mixer_pcm_element
     * is set. Otherwise, FAILURE is returned.
     * 
     * @return SUCCESS if successful, otherwise FAILURE
     */
    Result initMixer();

    /** Deinitialize the mixer device.
     * 
     * This method closes the mixer device if it has been opened and sets the
     * mixer_pcm_element to NULL.
     * 
     * @return SUCCESS
     */
    Result deinitMixer();

    /** Find a mixer element by name.
     * 
     * This method returns a mixer element with the given name, provided
     * that a mixer handle is available.
     * 
     * @param name Name of mixer element
     * @return Pointer to mixer element, if found, otherwise NULL
     */
    snd_mixer_elem_t* findMixerElementbyName(const string& name);
    //}
		
    /** The doFlush method from Node.
     *
     *  Flushes the audio device and checks the supported formats.
     *
     *  @return SUCCESS if the device could be flushed, FAILURE otherwise
     */
    virtual Result doFlush();

    /** The doInit method from Node.
     *
     * Initializes sound device.
     *
     * @return SUCCESS if sound device has been initialized successful, FAILURE otherwise.  
     */
    virtual Result doInit();

    /** The doStop method from Node.
     * 
     *  In this method the data in the internal buffer of the sound device will be played.
     *  The method returns when the internal buffer is empty. Depending on the fragment size,
     *  that means this method can take up to one second.
     *  Stops WavWriter and calls additionally GenericBSyncSinkNode::doStop().
     *
     * @return Result of GenericBSyncSinkNode::doStop()
     */
    virtual Result doStop();

    /* The doStart method from Node.
     *
     * Initializes and starts WavWriter if enabled and calls GenericBSyncSinkNode::doStart().
     *
     * @return Result of GenericBSyncSinkNode::doStart()
     */
    virtual Result doStart();

    /** The doInitOutput method from Node.
     *
     * In this method, the sound device is configured to the values determined in the formats.
     *
     * @return SUCCESS if the device has been configured correctly, FAILURE otherwise.
     */
    virtual Result doInitOutput();

        /** The doDeinit method from Node.
     *
     *  Closes sound device.
     *
     * @return SUCCESS
     */
    virtual Result doDeinit();

    //! The presentBuffer method from GenericSyncSinkNode.
    /*!
     *  This method writes the buffer to the device
     */
    virtual Message* presentBuffer(NMM::Buffer* buffer);

    //! The setFormatInternal method from GenericNode.
    /*!
     *  This method handles dynamic format changes
    */
    virtual Result setFormatInternal(Format& input_format);

    /** Returns output delay.
     *
     * The getOutputDelay - method from GenericSyncSinkNode.
     *
     * @return current output delay
     */
    virtual Interval getOutputDelay();

    /** @name IALSADevice interface
     */
    //@{
    /** Sets the length of the buffer used by ALSA in us.
     *
     * @param buffertime length of the buffer used by ALSA in us
     */
    virtual void setBufferTime(const nmm_uint32& buffertime);

     /** Sets the length of a period in us.
     *
     * @param periodtime length of a period in us
     */
    virtual void setPeriodTime(const nmm_uint32& periodtime);

    /** Sets xfer align.
     *
     * @param xferalign to set
     */	
    virtual void setXferAlign(const nmm_uint32& xferalign);

    /** Sets the output alsa device.
     *
     * @param device to set
     */
    virtual void setPCMDevice(const string& device);

    /** Returns the output alsa device.
     *
     * @return output alsa device.
     */
    virtual string getPCMDevice() const;
    //@}

    /** The old setVolume method.
     *
     * @param volume volume to set
     * @deprecated Please use the new IVolumeControl interface
     */
    void setVolume(const nmm_int32& volume);

    /** The old getVolume method.
     *
     * @return volume
     * @deprecated Please use the new IVolumeControl interface
     */ 
    nmm_int32 getVolume();

  public:
    /** The main constructor.
     *
     * @param name name of node
     * @param queue_mode operating mode of StreamQueue
     * @param size size of StreamQueue
     */
    ALSAPlaybackNode(const char* name = "ALSAPlaybackNode",
                     const StreamQueue::Mode queue_mode = StreamQueue::MODE_SUSPEND,
                     const int size = 24);

    /** The main destructor.
     */
    ~ALSAPlaybackNode();

    /** Sets line in volume.
     *
     * @param vol line in volume
     */
    virtual void setLineInVolume( int vol );

    /** Returns line in volume.
     *
     * @return line in volume
     */
    virtual int getLineInVolume();

    // methods from IALSADevice interface
    virtual void setMicroInVolume(const nmm_int32&);
    virtual nmm_int32 getMicroInVolume();

    /** Sets the mixer device.
     * 
     * @param mixer mixer device
     */
    virtual void setMixerDevice(const string& mixer) throw (const DeviceNotFoundException&);

    /** Sets the volume mixer element name.
     *
     * @param value mixer volume element
     */
    virtual void setMixerVolumeElement(const string& value);

    /** Returns the output device.
     * 
     * @return mixer device
     */
    virtual string getMixerDevice() const;

  protected:

    /**
     * Increases volume.
     */
    virtual void increaseVolume();

    /**
     * Decreases volume.
     */
    virtual void decreaseVolume();

    /**
     * Mutes device.
     */
    virtual void mute();

    /**
     * Un-mutes device.
     */
    virtual void unmute();

  private:
    /** Converts logarithmic to linear volume.
     *
     * @param logarithmic_volume logarithmic volume 
     * @return logarithmic volume converted to linear volume 
     */
    static Rational convertToLinearVolume(double logarithmic_volume);

    /** Converts linear to logarithmic volume.
     *
     * @param linear_volume linear volume 
     * @return linear volume converted to logarithmic volume 
     */
    static double convertToLogarithmicVolume(const Rational& linear_volume);

    /** Sets input format.
     *
     * Plugin-specific format change implementation.
     * This implementation sets the format on the ALSA device. It is also called by doInitOutput.
     */
    Result doSetFormat(const Format& input_format);
  };

  inline void ALSAPlaybackNode::setDevice(const string&) /* limited(INode::CONSTRUCTED) */
    throw(const DeviceNotFoundException&, const WrongDeviceException&) {}

}

#include "nmm/nmm_suffix.hpp"

#endif
