/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "ALSARecordNode.hpp"
#include "nmm/multimedia/sync/StreamTimer.hpp"
#include "nmm/multimedia/WrongDeviceException.hpp"
#include "nmm/utils/SecurityCheck.hpp"
#include "nmm/utils/time/Rational.hpp"
#include "nmm/comm/ProxyObject.hpp"
#include "nmm/multimedia/Plugin.hpp"

#include <sys/soundcard.h>
#include <alsa/asoundlib.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/stat.h>
#include <fcntl.h>

#include <iostream>
#include <iomanip>
#include <fstream>

NMM_REGISTER_PLUGIN(ALSARecordNode, NMM, "ALSARecordNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  //! Formats to check in the doInit method.
  snd_pcm_format_t ALSARecordNode::test_formats[] = {
    SND_PCM_FORMAT_MU_LAW,  //!< Logarithmic mu-law audio encoding
    SND_PCM_FORMAT_A_LAW,   //!< Logarithmic A-law audio encoding
    SND_PCM_FORMAT_U8,      //!< Linear encoding, unsigned 8-bit
    SND_PCM_FORMAT_S8,      //!< Linear encoding, signed 8-bit
    SND_PCM_FORMAT_U16_LE,  //!< Linear encoding, unsigned 16-bit, little endian
    SND_PCM_FORMAT_U16_BE,  //!< Linear encoding, unsigned 16-bit, big endian
    SND_PCM_FORMAT_S16_LE,  //!< Linear encoding, signed 16-bit, little endian
    SND_PCM_FORMAT_S16_BE,  //!< Linear encoding, signed 16-bit, big endian
    SND_PCM_FORMAT_U24_3LE,  //!< Linear encoding, unsigned 24-bit, little endian
    SND_PCM_FORMAT_U24_3BE,  //!< Linear encoding, unsigned 24-bit, big endian
    SND_PCM_FORMAT_S24_3LE,  //!< Linear encoding, signed 24-bit, little endian
    SND_PCM_FORMAT_S24_3BE,  //!< Linear encoding, signed 24-bit, big endian
    SND_PCM_FORMAT_UNKNOWN 
    //!< End marker. If you add or remove some formats, make sure that the last one remains 0.
  };

  //! Channel numbers to check in the doInit method.
  const int ALSARecordNode::test_channels[] = {
    1,      //!< Mono
    2,      //!< Stereo
    4,      //!< 4
    5,      //!< 5
    6,      //!< 5.1
    0
    //!< End marker. If you add or remove some values, make sure that the last one remains 0.
  };

  //! Sample rates to check in the doInit method.
  const int ALSARecordNode::test_sample_rates[] = { 8000, 11025, 22050, 44100, 48000, 96000, 0 };
  
  Format& ALSARecordNode::getNMMFormat(snd_pcm_format_t alsa_format, Format& nmm_format)
  {
    switch(alsa_format) {
    case SND_PCM_FORMAT_MU_LAW:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "mu_law");
      nmm_format.addIntParamValue(Format::bitpersample_param, 8);
      break;
      
    case SND_PCM_FORMAT_A_LAW:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "a_law");
      break;
      
    case SND_PCM_FORMAT_U8:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "linear");
      nmm_format.addStringParamValue(Format::sign_param, "unsigned");
      nmm_format.addIntParamValue(Format::bitpersample_param, 8);
      break;
      
    case SND_PCM_FORMAT_S8:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "linear");
      nmm_format.addStringParamValue(Format::sign_param, "signed");
      nmm_format.addIntParamValue(Format::bitpersample_param, 8);
      break;
      
    case SND_PCM_FORMAT_U16_LE:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "linear");
      nmm_format.addStringParamValue(Format::sign_param, "unsigned");
      nmm_format.addIntParamValue(Format::bitpersample_param, 16);
      nmm_format.addStringParamValue(Format::endian_param, "little");
      break;
      
    case SND_PCM_FORMAT_U16_BE:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "linear");
      nmm_format.addStringParamValue(Format::sign_param, "unsigned");
      nmm_format.addIntParamValue(Format::bitpersample_param, 16);
      nmm_format.addStringParamValue(Format::endian_param, "big");
      break;
      
    case SND_PCM_FORMAT_S16_LE:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "linear");
      nmm_format.addStringParamValue(Format::sign_param, "signed");
      nmm_format.addIntParamValue(Format::bitpersample_param, 16);
      nmm_format.addStringParamValue(Format::endian_param, "little");
      break;
      
    case SND_PCM_FORMAT_S16_BE:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "linear");
      nmm_format.addStringParamValue(Format::sign_param, "signed");
      nmm_format.addIntParamValue(Format::bitpersample_param, 16);
      nmm_format.addStringParamValue(Format::endian_param, "big");
      break;
    case SND_PCM_FORMAT_S24_3LE:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "linear");
      nmm_format.addStringParamValue(Format::sign_param, "signed");
      nmm_format.addIntParamValue(Format::bitpersample_param, 24);
      nmm_format.addStringParamValue(Format::endian_param, "little");
      break;
    case SND_PCM_FORMAT_S24_3BE:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "linear");
      nmm_format.addStringParamValue(Format::sign_param, "signed");
      nmm_format.addIntParamValue(Format::bitpersample_param, 24);
      nmm_format.addStringParamValue(Format::endian_param, "big");
      break;
    case SND_PCM_FORMAT_U24_3LE:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "linear");
      nmm_format.addStringParamValue(Format::sign_param, "unsigned");
      nmm_format.addIntParamValue(Format::bitpersample_param, 24);
      nmm_format.addStringParamValue(Format::endian_param, "little");
      break;
    case SND_PCM_FORMAT_U24_3BE:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "linear");
      nmm_format.addStringParamValue(Format::sign_param, "unsigned");
      nmm_format.addIntParamValue(Format::bitpersample_param, 24);
      nmm_format.addStringParamValue(Format::endian_param, "big");
      break;
    default:
      ERROR_STREAM("getNMMFormat: ALSA format " << alsa_format << " not supported." << endl);
    }
    
    return nmm_format;
  };
  
  snd_pcm_format_t ALSARecordNode::getALSAFormat(const Format &nmm_format)
  {
    string encoding;
    try {
      encoding = nmm_format.getStringValue(Format::encoding_param);
    } catch (FormatException e) {
      ERROR_STREAM("getALSAFormat: Incomplete format (encoding not specified)."
		   << endl);
      return SND_PCM_FORMAT_UNKNOWN;
    }
    if(encoding == "mu_law") return SND_PCM_FORMAT_MU_LAW;
    if(encoding == "a_law" ) return SND_PCM_FORMAT_A_LAW;
    if(encoding == "linear") {
      int bits;
      try {
	bits = nmm_format.getIntValue(Format::bitpersample_param);
      } catch (FormatException e) {
	ERROR_STREAM("getALSAFormat: Incomplete format (bitpersample not specified)." << endl);
	return SND_PCM_FORMAT_UNKNOWN;
      }
      string sign;
      try {
	sign = nmm_format.getStringValue(Format::sign_param);
      } catch (FormatException e) {
	ERROR_STREAM("getALSAFormat: Incomplete format (sign not specified)." << endl);
	return SND_PCM_FORMAT_UNKNOWN;
      }
      if(bits == 8) {
	if(sign == "unsigned") return SND_PCM_FORMAT_U8;
	if(sign == "signed"  ) return SND_PCM_FORMAT_S8;
	ERROR_STREAM("getALSAFormat: Incorrect format: sign is [" << sign << "]." << endl);
	ERROR_STREAM("getALSAFormat: Possible values for sign are: [unsigned|signed]." << endl);
      }
      if(bits == 16) {
	string endian;
	try {
	  endian = nmm_format.getStringValue(Format::endian_param);
	} catch (FormatException e) {
	  ERROR_STREAM("getALSAFormat: Incomplete format (endian not specified)." << endl);
	  return SND_PCM_FORMAT_UNKNOWN;
	}
	
	if(sign == "unsigned") {
	  if(endian == "little") return SND_PCM_FORMAT_U16_LE;
	  if(endian == "big"   ) return SND_PCM_FORMAT_U16_BE;
	  ERROR_STREAM("getALSAFormat: Incorrect format: endian is [" << endian << "]." << endl);
	  ERROR_STREAM("getALSAFormat: Possible values for endian are [little|big]." << endl);
	}
	
	if(sign == "signed"  ) {
	  if(endian == "little") return SND_PCM_FORMAT_S16_LE;
	  if(endian == "big"   ) return SND_PCM_FORMAT_S16_BE;
	  ERROR_STREAM("getALSAFormat: Incorrect format: endian is [" << endian << "]." << endl);
	  ERROR_STREAM("getALSAFormat: Possible values for endian are [little|big]." << endl);
	}
	ERROR_STREAM("getALSAFormat: Incorrect format: sign is [" << sign << "]." << endl);
	ERROR_STREAM("getALSAFormat: Possible values for sign are [unsigend|signed]." << endl);
      }
      if(bits == 24) {
	string endian;
	try {
	  endian = nmm_format.getStringValue(Format::endian_param);
	} catch (FormatException e) {
	  ERROR_STREAM("getALSAFormat: Incomplete format (endian not specified)." << endl);
	  return SND_PCM_FORMAT_UNKNOWN;
	}
	
	if(sign == "unsigned") {
	  if(endian == "little") return SND_PCM_FORMAT_U24_3LE;
	  if(endian == "big"   ) return SND_PCM_FORMAT_U24_3BE;
	  ERROR_STREAM("getALSAFormat: Incorrect format: endian is [" << endian << "]." << endl);
	  ERROR_STREAM("getALSAFormat: Possible values for endian are [little|big]." << endl);
	}
	if(sign == "signed"  ) {
	  if(endian == "little") return SND_PCM_FORMAT_S24_3LE;
	  if(endian == "big"   ) return SND_PCM_FORMAT_S24_3BE;
	  ERROR_STREAM("getALSAFormat: Incorrect format: endian is [" << endian << "]." << endl);
	  ERROR_STREAM("getALSAFormat: Possible values for endian are [little|big]." << endl);
	}
      }
      ERROR_STREAM("getALSAFormat: Only 8, 16 or 24 bitpersample are supported by ALSARecordNode."
		   << endl);
    }
    return SND_PCM_FORMAT_UNKNOWN;
  };
  
  ALSARecordNode::ALSARecordNode(const char* name,
				 const StreamQueue::Mode queue_mode,
				 const int size)
    : GenericSyncSourceNode(name, queue_mode, size),
      pcm_element(NULL)
  {
    buffer_time = 0;
    buffer_frames = 0;
    period_time = 0;
    period_frames = 0;
    
    buffer_size = 32*1024;

    timer = new StreamTimer(this, StreamTimer::REAL_TIME);

    pcm_access = SND_PCM_ACCESS_RW_INTERLEAVED;
    stream = SND_PCM_STREAM_CAPTURE;

    dev = "default";

    captured_first_buffer = false;
    captured_bytes = 0;
    system_start_time = Time();

    current_ts.sync_time = Time();
    current_ts.stream_counter = 0;
    total_read_bytes = 0;
  }
  
  ALSARecordNode::~ALSARecordNode() 
  {
    delete timer;
  }
  
  Result ALSARecordNode::doInit() {
    DEBUG_STREAM("DoInit: testing devices"<<endl);
    
    if (__mixer_device.empty())
      __mixer_device = "/dev/mixer";
    
    if(!SSecurityCheck::getInstance().checkFilenameForWriting(__mixer_device)) {
      ERROR_STREAM("Security check not passed for file " << __mixer_device << ", check .nmmrc for options" << endl);
      return FAILURE;
    }
    
    int err;
    //dev = "digital";
    // Open the device nonblocking, so the open call returns immediately.
    if ((err = snd_pcm_open (&pcm_handle, dev.c_str(), stream, SND_PCM_NONBLOCK)) < 0) {
      ERROR_STREAM("Cannot open audio device " << dev <<" with error: "<< snd_strerror(err) << endl );
      return FAILURE;
    }
    DEBUG_STREAM("Audio device open."<<endl); 
    if ((err = snd_pcm_hw_params_malloc (&hwparams)) < 0) {
      ERROR_STREAM("Cannot allocate hardware parameter structure with error " << snd_strerror (err) << endl);
      return FAILURE;
    }
    
    /*Init hwparams with full configuration space */
    if ((err = snd_pcm_hw_params_any(pcm_handle, hwparams)) < 0) {
      ERROR_STREAM("Cannot initialize to configure this PCM device. Error: " << snd_strerror(err) << endl);
      return FAILURE;
    }	
    
    if ( (err = snd_pcm_hw_params_set_access(pcm_handle, hwparams, pcm_access)) < 0) {
      ERROR_STREAM("Error setting access. Error: "<< snd_strerror(err) << endl);
      return FAILURE;
    }
    
    
    for(int i=0; test_formats[i] != SND_PCM_FORMAT_UNKNOWN; i++) {
      // First we need a full configuration space
      snd_pcm_hw_params_any(pcm_handle, hwparams);
      snd_pcm_hw_params_set_access(pcm_handle, hwparams, pcm_access);
      
      // Test all the formats in test_formats which are supported by the hardware.
      if(snd_pcm_hw_params_set_format(pcm_handle, hwparams, test_formats[i]) < 0) {
	DEBUG_STREAM("audio device: - Format " <<test_formats[i] << "not supported."<<endl);
      } else {
	// This format is supported.
	DEBUG_STREAM("audio device: - Format " << test_formats[i] << endl);
	
	// Test the supported channel numbers for this format.
	// Note: If you need a channel number that is currently not supported by
	//       ALSARecordNode, simply add it to the definition of the array
	//       test_channels in this file.
	for(int j=0; test_channels[j] != 0; j++) {
	  int test_channel = test_channels[j];
	  // First we need a full configuration space
	  snd_pcm_hw_params_any(pcm_handle, hwparams);
	  snd_pcm_hw_params_set_access(pcm_handle, hwparams, pcm_access);
	  snd_pcm_hw_params_set_format(pcm_handle, hwparams, test_formats[i]);
	  
	  if(snd_pcm_hw_params_set_channels(pcm_handle,hwparams,test_channel)<0) {
	    DEBUG_STREAM("audio device: - Format "<< test_formats[i]<< " not support "<< test_channel <<" channels."<<endl);
	  } else {
	    DEBUG_STREAM("audio device: - Format "<< test_formats[i]<< " support " << test_channel << " channels."<<endl);
	    // Create a new format.
	    Format *output_format = new Format();
	    
	    // Configure the parameters of the format according to the actual
	    // hardware configuration (since this configuration works.)
	    getNMMFormat(test_formats[i], *output_format);
	    
	    // Add the format to the output property.
	    getOwnOutputProperty() -> addNewFormat(output_format);
	    if (!getOwnOutputProperty()->getDefaultFormat()) {
	      // workaround for missing default format
	      // setting first format found to default one
	      getOwnOutputProperty()->setDefaultFormat(output_format);
	    }
	    
	    // Set channels and sample rate (not for ac3)
	    output_format->addIntParamValue(Format::channels_param, test_channels[j]);
	    
	    // Last step: Test the supported sample rates for the current channel number
	    //            and format.
	    // Note: If you need a sample rate that is not tested here, simpy add it to
	    //       the definition of the array test_sample_rates in this file.
	    for(int k=0; test_sample_rates[k] != 0; k++) {
	      int test_rate = test_sample_rates[k];
	      
	      // First we need a full configuration space
	      snd_pcm_hw_params_any(pcm_handle, hwparams);
	      snd_pcm_hw_params_set_access(pcm_handle, hwparams, pcm_access);
	      snd_pcm_hw_params_set_format(pcm_handle, hwparams, test_formats[i]);
	      snd_pcm_hw_params_set_channels(pcm_handle,hwparams,test_channel);

	      // Begin HACK for format handling
	      if( (err = snd_pcm_hw_params_test_rate(pcm_handle, hwparams, test_rate, 1)) < 0) {
		DEBUG_STREAM("audio device:     + Set sample rate to " << test_sample_rates[k]
			     << " failed. Error: "<< snd_strerror(err) << endl);
	      } else {
		if(test_rate == test_sample_rates[k]) {
		  // Sample rate is supported with the actual configuration.
		  DEBUG_STREAM("audio device:     + Sample rate " << test_sample_rates[k] << endl);
		  
		  // Add it to the format in the output property.
		  output_format->addIntParamValue(Format::sample_rate_param, test_sample_rates[k]);
		}
		}
	      // End HACK for format handling
	    } // loop over the test_sample_rates
	  }
	} //loop over test_channels
      }
      
    }
    
    /* open mixer device */
    string cardname = "default"; // TODO: cardname is hard coded
    if( snd_mixer_open (&mixer_handle, 0) < 0 ) {
      ERROR_STREAM("snd_mixer_open failed" << endl);
      return FAILURE;
    }
    
    if( snd_mixer_attach(mixer_handle, cardname.c_str()) < 0 ) {
      ERROR_STREAM("snd_mixer_attach failed for " << cardname << endl);
      return FAILURE;
    }
    
    if( snd_mixer_selem_register(mixer_handle, NULL, NULL) < 0 ) {
      ERROR_STREAM("snd_mixer_selem_register failed" << endl);
      return FAILURE;
    }
    
    if( snd_mixer_load(mixer_handle) < 0 ) {
      ERROR_STREAM("snd_mixer_load failed" << endl);
      return FAILURE;
    }
    
    /* find PCM element */
    pcm_element = findMixerElem("Mic");
    // set volume range to 0..100
    if(pcm_element) {
      snd_mixer_selem_set_playback_volume_range(pcm_element, 0, 100);
      
      long min, max;
      snd_mixer_selem_get_playback_volume_range(pcm_element, &min, &max);
      DEBUG_STREAM("volume range: " << "min " << min << " max " << max << endl);
    }


    // Begin HACK for format handling
    Format *format = getOwnOutputProperty()->addNewFormat("audio/raw");
    format->addIntParamValue(Format::sample_rate_param, 48000);
    format->addIntParamValue(Format::channels_param, 2);
    format->addIntParamValue(Format::bitpersample_param, 24);
    format->addStringParamValue(Format::endian_param, "little");
    format->addStringParamValue(Format::encoding_param, "linear");
    format->addStringParamValue(Format::sign_param, "signed");
    getOwnOutputProperty()->setDefaultFormat(format);
    // End HACK for format handling

    return SUCCESS;
  }
  
  snd_mixer_elem_t* ALSARecordNode::findMixerElem(string mixer_element)
  {
    snd_mixer_elem_t* element = NULL;
    
    snd_mixer_elem_t* mixer_elem = snd_mixer_first_elem(mixer_handle);
    while(mixer_elem) {
      if(!strcmp(snd_mixer_selem_get_name(mixer_elem), mixer_element.c_str() )) {
	DEBUG_STREAM("findMixerElem: found mixer elem for " << mixer_handle << endl);
	element = mixer_elem;
	break;
      }
      mixer_elem = snd_mixer_elem_next(mixer_elem);
    }
    
    return element;
  }
  
  Result ALSARecordNode::doDeinit()
  {
    if(mixer_handle)
      snd_mixer_close(mixer_handle);
    snd_pcm_close(pcm_handle);
    return SUCCESS;
  };
  
  Result ALSARecordNode::doInitOutput() {
    // copy formats from static property to working property
    list<Format*> formats = getOwnOutputProperty()->getFormatList();
    for (list<Format*>::iterator i(formats.begin()); i != formats.end(); ++i) {
      const Format* format = *i;
      Format* new_format = new Format( *format );
      getOwnWorkingOutputProperty()->addNewFormat(new_format);
      if (format == getOwnOutputProperty()->getDefaultFormat()) {
        getOwnWorkingOutputProperty()->setDefaultFormat(new_format);
      }
    }
    
    return SUCCESS;
  }

  Result ALSARecordNode::doActivate() 
  {
    
    DEBUG_STREAM("doInitActivate: Setting format device."<<endl);
    snd_pcm_hw_params_any(pcm_handle,hwparams);
    
    const Format* output_format = getOutputFormat();
    if (output_format == NULL) { 
      DEBUG_STREAM("Not setting output_format."<<endl);
      return FAILURE;
    }
    Format* format;
    
    format = new Format( *output_format );
    getOwnWorkingOutputProperty()->addNewFormat(format);
    getOwnWorkingOutputProperty()->setDefaultFormat(format);
    
    Format tmp_format = *output_format;
    
    if(setFormatInternal(tmp_format) == FAILURE) {
      ERROR_STREAM("setFormatInternal failed" << endl);
      return FAILURE;
    }
    
    if(snd_pcm_hw_params(pcm_handle, hwparams) < 0) {
      ERROR_STREAM("audio device: Error setting HW params."<<endl);
      return FAILURE;
    }

    timer->reset();
    return SUCCESS;
  }

  Message* ALSARecordNode::produceBuffer(Buffer* /*in_buffer*/)
  {
    
    Buffer *buffer = getNewBuffer(buffer_size);
    ssize_t r;

    //cerr << "Byte per sample: " << byte_per_sample << " Channels: " << channels << endl;
    //cerr << "Buffer size = " << buffer_size << endl;
    int frames = buffer_size/(byte_per_sample*channels);

    //cerr << "Frames to read: " << frames << endl;

    if (buffer != 0) {

      // BEGIN >
      if( !captured_first_buffer && (buffer->getTimestamp().stream_counter > 20) ) {
      // if( !captured_first_buffer && (buffer->getTimestamp().stream_counter > 200) ) {
	captured_first_buffer = true;
	captured_bytes = 0;
	system_start_time = TimedElement::getTime();
	// cerr << "System clock start time  : " << system_start_time << endl;
      }
      // < END

      r = pcm_read((u_char*)buffer->getData(), frames);
      if ( r > 0 ) {
	int used = r * byte_per_sample * channels;
	//cerr << used << " Bytes read" << endl;
	buffer->setUsed(used);
      } 
      if (buffer->getUsed() > 0) {

	current_ts.sync_time = getLocalTime(total_read_bytes);
	total_read_bytes += buffer->getUsed();
	current_ts.stream_counter++;
	current_ts.is_valid = true;
	buffer->setTimestamp(current_ts);

	// cerr << buffer->getTimestamp().stream_counter << " " << buffer->getTimestamp().sync_time  << endl;
	//cerr << "Returne buffer" << endl;
	return buffer;
      } else {
        // read error, release buffer
	DEBUG_STREAM("No data read from the device" << endl);
        buffer->release();
	return 0;
      }
    }
    return 0;

  }
  
  ssize_t ALSARecordNode::pcm_read(u_char *data, size_t rcount)
  {
    ssize_t r;
    size_t result = 0;
    size_t count = rcount;
    
    while (count > 0) {
      r = snd_pcm_readi(pcm_handle, data, count);
      if (r == -EAGAIN || (r >= 0 && (size_t)r < count)) {
	snd_pcm_wait(pcm_handle, 1000);
      } else if (r == -EPIPE) {
	xrun();
      } else if (r == -ESTRPIPE) {
	suspend();
      } else if (r < 0) {
	ERROR_STREAM("READ ERROR: "<< snd_strerror(r)<<endl);
	return 0;
      }
      if (r > 0) {
	result += r;
	count -= r;
	data += r * byte_per_sample * channels;
      }
    }
    return rcount;
  }

  /* I/O error handler */
  Result ALSARecordNode::xrun(void) {
    snd_pcm_status_t *status;
    int res;
    
    snd_pcm_status_alloca(&status);
    if ((res = snd_pcm_status(pcm_handle, status))<0) {
      ERROR_STREAM("xrun(): status error: "<< snd_strerror(res)<<endl);
      return FAILURE;
    }
    if (snd_pcm_status_get_state(status) == SND_PCM_STATE_XRUN) {
      struct timeval now, diff, tstamp;
      gettimeofday(&now, 0);
      snd_pcm_status_get_trigger_tstamp(status, &tstamp);
      timersub(&now, &tstamp, &diff);
      DEBUG_STREAM("xrun(): overrrun!!! (at least " << 
		   diff.tv_sec * 1000 + diff.tv_usec / 1000.0 << " ms long."<<endl);

      if ((res = snd_pcm_prepare(pcm_handle))<0) {
	ERROR_STREAM("xrun(): prepare error: "<< snd_strerror(res)<<endl);
	return FAILURE;
      }
      return SUCCESS;         /* ok, data should be accepted again */
    } if (snd_pcm_status_get_state(status) == SND_PCM_STATE_DRAINING) {
      DEBUG_STREAM("xrun(): Status(DRAINING)"<<endl);
      // snd_pcm_status_dump(status, log);
    }
    ERROR_STREAM("xrun(): read/write error, state = "<< snd_pcm_state_name(snd_pcm_status_get_state(status)) <<endl );
    return FAILURE;
  }
  
  /* I/O suspend handler */
  Result ALSARecordNode::suspend(void)
  {
    int res;
    
    DEBUG_STREAM("Suspended. Trying resume. "<<endl);
    while ((res = snd_pcm_resume(pcm_handle)) == -EAGAIN)
      sleep(1);       /* wait until suspend flag is released */
    if (res < 0) {
      DEBUG_STREAM("Failed. Restarting stream. "<<endl); 
      if ((res = snd_pcm_prepare(pcm_handle)) < 0) {
	ERROR_STREAM("suspend(): prepare error: "<< snd_strerror(res)<< endl);
	return FAILURE;
      }
    }
    DEBUG_STREAM("Suspend Done."<<endl);
    return SUCCESS;
  }

  Time ALSARecordNode::getLocalTime(u_int64_t valid_read_bytes) {
    return Time(Rational(valid_read_bytes, sample_rate * channels * bit_per_sample / 8));
  }

  void ALSARecordNode::setBufferSize(const nmm_int32 &size) {
    if (size > 0) {
      buffer_size = size;
      return;
    }
    ERROR_STREAM( "setValue: trying to set a negative buffer_size! " << endl );
  }
  
  nmm_int32 ALSARecordNode::getBufferSize() {
    return buffer_size;
  }

  void ALSARecordNode::setMicroInVolume(const nmm_int32& volume)
  {   /* set volume using some undocumented ALSA function */
    if(pcm_element) {
      DEBUG_STREAM1("Set volume to " << volume << std::endl);
      snd_mixer_selem_set_capture_volume_range(pcm_element, 0, 100);
      snd_mixer_selem_set_capture_volume_all(pcm_element, volume  );
    } else {
      ERROR_STREAM1("Can not set volume (no mixer)" << std::endl);
    }
  }

  nmm_int32 ALSARecordNode::getMicroInVolume()
  { /* get volume using some undocumented ALSA function */
    long volume = 0;
    if(pcm_element) {
      snd_mixer_selem_set_capture_volume_range(pcm_element, 0, 100);
      snd_mixer_selem_get_capture_volume(pcm_element, SND_MIXER_SCHN_FRONT_LEFT, &volume);
      DEBUG_STREAM1("Get volume. Current volume is " << volume << std::endl);
      return static_cast<int>(volume);
    } else {
      ERROR_STREAM1("Can not get volume (no mixer)" << std::endl);
      return 0;
    }
  }

  void ALSARecordNode::setPCMDevice(const string& device) {
    int err;
    dev = device;
    // Open the device nonblocking, so the open call returns immediately.
    if ((err = snd_pcm_open (&pcm_handle, dev.c_str(), stream, SND_PCM_NONBLOCK)) < 0) {
      ERROR_STREAM("Cannot open audio device " << dev <<" with error: "<< snd_strerror(err) << endl );
      throw WrongDeviceException(string("Cannot open audio device ") + string( dev ) + string(" with error: ") + string (snd_strerror(err)) );
    } else {
      DEBUG_STREAM("Audio device set."<<endl);
      snd_pcm_close(pcm_handle);
    }
    return;
  }

  string ALSARecordNode::getPCMDevice() const {
    return dev;
  }

  void ALSARecordNode::setBufferTime(const nmm_uint32& buffertime) {
    buffer_time = buffertime;
  }
  
  void ALSARecordNode::setPeriodTime(const nmm_uint32& periodtime) {
    period_time = periodtime;
  }

  void ALSARecordNode::setXferAlign(const nmm_uint32&) {
    // do nothing
  }

  Result ALSARecordNode::setFormatInternal(Format& input_format)
  {
    int err;
    snd_pcm_format_t format = SND_PCM_FORMAT_UNKNOWN;
    DEBUG_STREAM("Setting format to audio device."<<endl);
    
    // Get format parameters
    format  = getALSAFormat(input_format);
    if (format == SND_PCM_FORMAT_UNKNOWN) return FAILURE;
    channels    = input_format.getIntValue(Format::channels_param);
    sample_rate = input_format.getIntValue(Format::sample_rate_param);
    DEBUG_STREAM("Got Format: samplerate: " << sample_rate << ", channels: "
		 << channels << endl);
    
    try {
      bit_per_sample = input_format.getIntValue(Format::bitpersample_param);
    } catch(...) {
      bit_per_sample = 16;
    }
    if (bit_per_sample == 0) bit_per_sample = 16;
    
    byte_per_sample = bit_per_sample / 8;
    byte_per_second = sample_rate * channels * byte_per_sample;
    
    
    // It is important always to keep the same order in setting the parameters:
    // At first the access, format, then the channel number and at last the sample rate.
    DEBUG_STREAM("audio device: Trying to set access. " << endl);
    
    if ( (err = snd_pcm_hw_params_set_access(pcm_handle, hwparams, pcm_access)) < 0) {
      ERROR_STREAM("Error setting access at init output. Error: "<< snd_strerror(err) << endl);
      return FAILURE;
    }
    
    DEBUG_STREAM("audio device: Trying to set format " << format << endl);
    
    /* Set sample format */
    if (snd_pcm_hw_params_set_format(pcm_handle, hwparams, format) < 0) {
      ERROR_STREAM("Error setting format " << format << "."<<endl);
      return FAILURE;
    }
    
    DEBUG_STREAM("audio device: Set format to " << format << endl);
    
    /* Set number of channels */
    DEBUG_STREAM("audio device: Trying to set channels to " << channels << endl);
    if (snd_pcm_hw_params_set_channels(pcm_handle, hwparams, channels) < 0) {
      ERROR_STREAM("Error setting channels."<<endl);
      return FAILURE;
    }
    
    DEBUG_STREAM("audio device: Set channels to " << channels << endl);
    
    DEBUG_STREAM("audio device: Trying to set sample_rate to " << sample_rate << "." << endl);
    
    /* Set sample rate. If the exact rate is not supported */
    /* by the hardware, use nearest possible rate.         */
    unsigned int exact_rate = sample_rate;
    if (snd_pcm_hw_params_set_rate_near(pcm_handle, hwparams, &exact_rate, 0) < 0) {
      ERROR_STREAM("Error setting rate.");
      return FAILURE;
    }
    if (sample_rate != exact_rate) {
      DEBUG_STREAM("The rate " << sample_rate << " Hz is not supported by your hardware. " <<
		   "==> Using " << exact_rate << " Hz instead."<<endl);
      sample_rate = exact_rate;
    }
    
    DEBUG_STREAM("audio device: Set sample_rate to " << sample_rate << "." << endl);
    
    if (buffer_time == 0 && buffer_frames == 0) {
      if( snd_pcm_hw_params_get_buffer_time_max(hwparams, &buffer_time, 0) < 0) {
	ERROR_STREAM("audio device: I can't get  buffer time."<<endl);
	return FAILURE;
      }
      if (buffer_time > 500000)
	buffer_time = 500000;
    }
    DEBUG_STREAM("audio device: buffer_time "<< buffer_time << " buffer_frames "<<buffer_frames<<endl);
    if (period_time == 0 && period_frames == 0) {
      if (buffer_time > 0)
	period_time = buffer_time / 4;
      else
	period_frames = buffer_frames / 4;
    }
    DEBUG_STREAM("audio device: period_time "<<period_time<<" period_frames "<<period_frames<<endl);
    if (period_time > 0) {
      if (snd_pcm_hw_params_set_period_time_near(pcm_handle, hwparams, &period_time, 0) < 0) {
	ERROR_STREAM("audio device: Error setting period time."<<endl);
	return FAILURE;
      }
    } else {
      if ( snd_pcm_hw_params_set_period_size_near(pcm_handle, hwparams, &period_frames, 0) < 0) {
	ERROR_STREAM("audio device: Error setting period size."<<endl);
	return FAILURE;
      }
    }
    DEBUG_STREAM("audio device: period buffer set"<<endl);
    if (buffer_time > 0) {
      if (snd_pcm_hw_params_set_buffer_time_near(pcm_handle, hwparams, &buffer_time, 0) < 0) {
	ERROR_STREAM("audio device: Error setting buffer time."<<endl);
	return FAILURE;
      }
    } else {
      if ( snd_pcm_hw_params_set_buffer_size_near(pcm_handle, hwparams, &buffer_frames) < 0) {
	ERROR_STREAM("audio device: Error setting buffer size."<<endl);
	return FAILURE;
      }
    }
    
    DEBUG_STREAM("audio device: Buffer set."<<endl);
    
    buffer_frames = buffer_time * sample_rate;
    internal_buffer_size = buffer_frames * channels * byte_per_sample;
    internal_buffer_time = Time(static_cast<long int>(static_cast<double>(internal_buffer_size) / byte_per_second));
    DEBUG_STREAM("setFormatInternal: Internal buffer size: " << internal_buffer_size << " bytes." << endl);
    DEBUG_STREAM("setFormatInternal: Internal buffer time: " << internal_buffer_time << endl);
    
    return SUCCESS;
  };
  
  Result ALSARecordNode::doSetFormat(const Format& format) 
  {
    DEBUG_STREAM("ALSARecordNode::doSetFormat" << endl);
    try {
      DEBUG_STREAM("ALSARecordNode::doSetFormat, samplerate: " << format.getIntValue(Format::channels_param) << ", channels: " << format.getIntValue(Format::sample_rate_param) << endl);
    }
    catch(Exception e) {
      ERROR_STREAM("Exception in doSetFormat " << e << endl);
    }
    return SUCCESS;
  }

  void ALSARecordNode::setCLOEXEC(int fd){
    
    int oldflags = fcntl (fd, F_GETFD, 0);
    if (oldflags >= 0){
      oldflags |= FD_CLOEXEC;
      fcntl (fd, F_SETFD, oldflags);
      GLOBAL_DEBUG_STREAM("Setting CLOEXEC on audio_fd" << endl);
    }else{
      GLOBAL_ERROR_STREAM("Could not set CLOEXEC on audio_fd" << endl);
    }
  }
  
}

#include "nmm/nmm_suffix.hpp"
