/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#ifndef _NMM_ALSARECORDNODE_HPP_
#define _NMM_ALSARECORDNODE_HPP_

#include "nmm/nmm_prefix.hpp"

#include "nmm/plugins/NMMRegisterPlugin.hpp"

#include "nmm/multimedia/sync/GenericSyncSourceNode.hpp"
#include "nmm/utils/time/Clock.hpp"

#include "nmm/plugins/interfaces/IBufferSizeImpl.hpp"
#include "nmm/plugins/audio/alsa/IALSADeviceImpl.hpp"

#include <alsa/asoundlib.h>

namespace NMM {

  class StreamTimer;

  /** Node to record audio from an alsa device.
   */

  class ALSARecordNode :
    public GenericSyncSourceNode,
    public IBufferSizeImpl,
    public IALSADeviceImpl
  {
  private:
    StreamTimer* timer;

    snd_pcm_t *pcm_handle; ///< Handle for the PCM device
    
    snd_pcm_stream_t stream; ///< Playback stream

    /* This structure contains information about    */
    /* the hardware and can be used to specify the  */
    /* configuration to be used for the PCM stream. */
    snd_pcm_hw_params_t *hwparams;

    unsigned int buffer_time;  ///< length of buffer used by ALSA in us
    unsigned int period_time;  ///< length of period used by ALSA in us
    snd_pcm_uframes_t period_frames;  ///< frames per period
    snd_pcm_uframes_t buffer_frames;  ///< frames per buffer
		
    snd_pcm_status_t *pcm_status;  ///< status of audio device
    
    snd_pcm_access_t pcm_access;  ///< Used to set the pcm access to INTERLEAVED

    static snd_pcm_format_t test_formats[];  ///< Formats to check in the doInit method.
    
    static const int test_channels[];  ///< Channels to check in the doInit method. 
    
    static const int test_sample_rates[];  ///< Sample rates to check in the doInit method.
    
    /** Converts an ALSA_Format into an NMM::Format.
     * 
     * @param snd_pcm_format ALSA Format
     * @param nmm_format reference to nmm Format object to write to 
     * @return reference to converted format
     */
    Format& getNMMFormat(snd_pcm_format_t alsa_format, Format& nmm_format);
    
    /** Converts an NMM::Format into an ALSA_Format.
     * 
     * @param nmm_format NMM Format
     * @return snd_pcm_format ALSA Format
     */
    snd_pcm_format_t getALSAFormat(const Format& nmm_format);

    /** Reads audio data from alsa device.
     * 
     * @param data pointer to write buffer 
     * @param rcount data to read maximum
     * @return number of bytes read
     */
    ssize_t pcm_read(u_char *data, size_t rcount);

    
    /** I/O error handler method
     *
     * @return SUCCESS if error handling was successful, false otherwise
     */
    Result xrun(void);

    /** I/O suspend handler method
     *
     * @return SUCCESS if suspend handling was successful, false otherwise
     */
    Result suspend(void);
		
    unsigned int sample_rate;       ///< The sample_rate (depends on input_format)
    int channels;                   ///< The number of channels (depends on input_format)
    int bit_per_sample;             ///< The bits per sample (depends on input format)
    int byte_per_sample;            ///< The bytes per sample (depends on input format)
    int byte_per_second;            ///< The bytes per second (depends on input format)
    int buffer_size;                ///< Size of the buffers.
		
    int internal_buffer_size;       ///< size of internal buffer
    Time internal_buffer_time;      ///< duration of internal buffer
    
    snd_mixer_t *mixer_handle;      ///< Handle for mixer device

    snd_mixer_elem_t* pcm_element;  ///<  PCM element
    
    /** Finds mixer element for given string. 
     *
     * @return pointer to mixer element
     */
    snd_mixer_elem_t* findMixerElem(string);

    Clock clock;       ///< internal clock
    
    /**
     * Sets CLOEXEC. 
     *
     * @param fd Filedescriptor.
     */
    void setCLOEXEC(int fd);

    u_int64_t total_read_bytes;   ///< bytes read from device
    Timestamp current_ts;         ///< currently used timestamp
    bool captured_first_buffer;   ///< true if we created minimum one buffer
    u_int64_t captured_bytes;     ///< captured bytes since taken system start time ( needed for sync)
    Time system_start_time;       ///< system start time (needed for sync)

  protected:
    string dev;              ///< The name of the sound device.
    string __mixer_device;   ///< The name of the mixer device.
    
    
    /** Writes format to debug output.
     *
     * @param format Format to write to debug output
     * @return SUCCESS
     */
    Result doSetFormat(const Format& format);
    
    /** The doInit method from Node.
     *
     * Simply calls doInitInput.
     *
     * @return SUCCESS if method call was successful, FAILURE otherwise
     */
    virtual Result doInit();
    
    /** The doInitOutput method from Node.
    *
     * In this method, the sound device is configured to the values determined in the formats.
     *
     * @return SUCCESS if the device has been configured correctly, FAILURE otherwise.
     */
    virtual Result doInitOutput();
		
    /** The doDeinit method from Node.
     *
     *  Simply calls doDeinitInput.
     *
     * @return SUCCESS if method call was successful, FAILURE otherwise
     */
    virtual Result doDeinit();

    /** The doActivate method from Node.
     *
     * Sets working output property and sets format on audio device.
     *
     * @return SUCCESS if method call was successful, FAILURE otherwise
     */
    virtual Result doActivate();

    /** Sets the output device.
     *
     * @deprecated obsolete method
     * @param device device to set
     * @return SUCCESS if method call was successful, FAILURE otherwise
     */
    Result oldSetDevice(string& device);
    
    //! The presentBuffer method from GenericSyncSinkNode.
    virtual Message* produceBuffer(Buffer *in_buffer);

    /** Returns size of data read from audio device given in time.
     *
     * @param t bytes read from device
     * @return bytes read from device represented as playback time  
     */
    virtual Time getLocalTime(u_int64_t t);

    /** Sets internal audio format.
     *
     * This method configures the audio device to a given audio format.
     * It is called on activation and on format changes. It has been renamed to keep interface
     * and implementation seperate.
     *
     * @param input_format NMM audio format
     * @return SUCCESS if the device has been configured correctly, FAILURE otherwise.
     */
    Result setFormatInternal(Format& input_format);

  public:
    /** The main constructor.
     *
     * @param name name of node
     * @param queue_mode operating mode of StreamQueue
     * @param size size of StreamQueue
     */
    ALSARecordNode(const char* name = "ALSARecordNode",
		   const StreamQueue::Mode queue_mode = StreamQueue::MODE_SUSPEND,
		   const int size = StreamQueue::DEFAULT_SIZE);
    
    /** The main destructor.
     *
     */
    ~ALSARecordNode();
    
    /**
     * IBufferSize interface.
     */

    /** Sets buffer size to use.
     *
     *  Sets buffer size to be used by the node.
     *  
     *  @param size Buffer size to be used.
     */
    virtual void setBufferSize(const nmm_int32& size);

    /** Requests buffer size node is currently using.
     *  
     *  @return Buffer size used by node.
     */
    virtual nmm_int32 getBufferSize();

    /**
     * IALSADevice interface
     */

    /** Setting name of pcm device.
     *
     * @param device name of pcm device
     */
    virtual void setPCMDevice(const string& device);
    
    /** Returns name of pcm device.
     *
     * @return name of pcm device
     */
    virtual string getPCMDevice() const;

    /** Sets the length of the buffer used by ALSA in us
     *
     * @param buffertime Buffertime.
     */
    virtual void setBufferTime(const nmm_uint32& buffertime);

    /** Sets the length of a period in us.
     *
     * @param periodtime Periodtime.
     */
    virtual void setPeriodTime(const nmm_uint32& periodtime);

    /** Setting transfer alignment.
     *
     * @param xferalign transfer alignment
     */
    virtual void setXferAlign(const nmm_uint32& xferalign);

    virtual void setMicroInVolume(const nmm_int32&);
    virtual nmm_int32 getMicroInVolume();

  };
  
}

#include "nmm/nmm_suffix.hpp"

#endif // _NMM_ALSARECORDNODE_HPP_


