/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#ifndef always_inline
#define always_inline inline
#endif

#include <sys/time.h>
#include "FFMpegAudioEncodeNode.hpp"
#include "nmm/comm/format/Property.hpp"
#include "nmm/multimedia/Plugin.hpp"
#include "nmm/multimedia/TPlugin.hpp"
#include <math.h>
#include <iostream>
#include "nmm/comm/serialize/SerializeUtilsTime.hpp"
#include "nmm/comm/serialize/SerializeResult.hpp"
#include "nmm/utils/time/Rational.hpp"
#include "nmm/utils/time/Interval.hpp"

#include "nmm/utils/thread/MutexGuard.hpp"

NMM_REGISTER_PLUGIN(FFMpegAudioEncodeNode, NMM, "FFMpegAudioEncodeNode")

#define MIN(a,b) ((a<b)?a:b)

//used for translating amr_bitrate_modes (NB) to bitrates used in ffmpeg.
static const int amr_nb_bitrate_modes[8] = {4750,     // MR475 4.75 kbits
                                            5150,  // MR515 5.15 
                                            5900,  // MR59  5.90
                                            6700,  // MR67  6.70
                                            7400,  // MR74  7.40
                                            7950,  // MR795 7.95
                                            10200, // MR102 10.2
                                            12200  // MR122 12.2
                                           };

//used for translating amr_bitrate_modes (WB) to bitrates used in ffmpeg.
static const int amr_wb_bitrate_modes[9] = {6600,     //  6.60 kbits
                                            8850,  //  8.85
                                            12650, // 12.65
                                            14250, // 14.25
                                            15850, // 15.85
                                            18250, // 18.25
                                            19850, // 19.85
                                            23050, // 23.05
                                            23850, // 23.85
                                           };

#include "nmm/nmm_prefix.hpp"

namespace NMM {

static int bufcount= 0;
  FFMpegAudioEncodeNode::FFMpegHandle::FFMpegHandle() : ffAVCodec(NULL), ffavcc(NULL) {}

  FFMpegAudioEncodeNode::FFMpegAudioEncodeNode(const char *name, StreamQueue::Mode qmode, int qsize)
    : GenericProcessorNode(name, qmode, qsize),
      ffmpeg(new FFMpegHandle()),
      first_timestamp(true),
      timestamp(),
      in_buf_ts(),
/*      my_delay(0),
      my_delay_num(0),
      my_delay_den(0),
      num_samples(0), */
      out_buffer(0),
      work_buffer(0),
      work_buf(0),
      work_buf_ptr(0),
      inbuf(0),
      inbuf_ptr(0),
      sample_rate(0),
      channels(0),
      bitrate(0),
      size(0),
      byte_per_sample(0),
      frame_size(0),
      input_bytes_needed(0),
      enc_buf_size(0),
      compressor()
  {
  }
  
  FFMpegAudioEncodeNode::~FFMpegAudioEncodeNode() {
    delete ffmpeg;
  }

  Result FFMpegAudioEncodeNode::startTrack(const string&) {
    DEBUG_STREAM("Received Event: startTrack");
    setWorkingFlag(false);
    //cerr << "startTrack: set working flag (false)" << endl;
    return SUCCESS;
  }

  Result FFMpegAudioEncodeNode::endTrack() {
    DEBUG_STREAM("Received Event: endTrack");
    return SUCCESS;
  }

  Result FFMpegAudioEncodeNode::finished() {
    DEBUG_STREAM("Received Event: finished");
    return SUCCESS;
  }
  
  void FFMpegAudioEncodeNode::setQuality(const nmm_float&) {
    // quality not defined
  }

  void FFMpegAudioEncodeNode::setBitrate(const nmm_int32& bitrate) {
    {
      MutexGuard ffmpeg_mg(ffmpeg_mutex);
      // this is working only with AMR "on the fly"
      this->bitrate = bitrate;
      ffmpeg->ffavcc->bit_rate = bitrate;
    }
    DEBUG_STREAM("Received Event: setBitrate; bitrate changed to " << bitrate << endl);
  }

  nmm_float FFMpegAudioEncodeNode::getQuality() {
    // quality not defined
    return 1.0f;
  }

  nmm_int32 FFMpegAudioEncodeNode::getBitrate() {
    return bitrate;
  }

  Result FFMpegAudioEncodeNode::doInit() {
  
    {
      MutexGuard ffmpeg_guard(ffmpeg_mutex);
      DEBUG_STREAM("Got avcodec version: " << avcodec_version() << endl);
      DEBUG_STREAM("Got avcodec build: " << avcodec_build() << endl);
    
      enc_buf_size = 10000;    
      
      /* initialize library */
      avcodec_init();
    
      /* register all the codecs */
      avcodec_register_all();
    
      /* detect all codecs */
      codecs.clear();
      detectAllCodecs();
    
      /* check if any codec was found */
      if(codecs.empty()) {
        ERROR_STREAM("No codecs with known FOURCC code available" << std::endl);
        return FAILURE;
      }
    }
    
    // input format -> raw audio samples
    Format *in_format1 = getOwnInputProperty()->addNewFormat(Format::audio_raw_type);
    in_format1->addIntParamValue(Format::sample_rate_param, 44100);
    in_format1->addIntParamValue(Format::channels_param, 2);
    in_format1->addIntParamValue(Format::bitpersample_param, 16);
    in_format1->addStringParamValue(Format::endian_param, "little");
    in_format1->addStringParamValue(Format::encoding_param, "linear");
    in_format1->addStringParamValue(Format::sign_param, "signed");
  
    Format *in_format2 = getOwnInputProperty()->addNewFormat(Format::audio_raw_type);
    in_format2->addIntParamValue(Format::sample_rate_param, 48000);
    in_format2->addIntParamValue(Format::channels_param, 2);
    in_format2->addIntParamValue(Format::bitpersample_param, 16);
    in_format2->addStringParamValue(Format::endian_param, "little");
    in_format2->addStringParamValue(Format::encoding_param, "linear");
    in_format2->addStringParamValue(Format::sign_param, "signed");
    
    Format *in_format3 = getOwnInputProperty()->addNewFormat(Format::audio_raw_type);
    in_format3->addIntParamValue(Format::sample_rate_param, 8000);
    in_format3->addIntParamValue(Format::channels_param, 1);
    in_format3->addIntParamValue(Format::bitpersample_param,  16);
    in_format3->addStringParamValue(Format::endian_param, "little");
    in_format3->addStringParamValue(Format::encoding_param, "linear");
    in_format3->addStringParamValue(Format::sign_param, "signed");
    
    getOwnInputProperty()->setDefaultFormat(in_format3);
    
    Format *in_format4 = getOwnInputProperty()->addNewFormat(Format::audio_raw_type);
    in_format4->addIntParamValue(Format::sample_rate_param, 16000);
    in_format4->addIntParamValue(Format::channels_param, 1);
    in_format4->addIntParamValue(Format::bitpersample_param,  16);
    in_format4->addStringParamValue(Format::endian_param, "little");
    in_format4->addStringParamValue(Format::encoding_param, "linear");
    in_format4->addStringParamValue(Format::sign_param, "signed");
    
    Format *out_format1 = getOwnOutputProperty()->addNewFormat(Format::audio_mpeg_type);
    out_format1->addIntParamValue(Format::audio_codec_param, 0x55);
    out_format1->addIntParamValue(Format::layer_param, 3);
    out_format1->addWildcard(Format::bitrate_param);
    out_format1->addIntParamValue(Format::channels_param, 1);
    out_format1->addIntParamValue(Format::channels_param, 2);
    out_format1->addIntParamValue(Format::sample_rate_param, 8000);
    out_format1->addIntParamValue(Format::sample_rate_param, 11025);
    out_format1->addIntParamValue(Format::sample_rate_param, 12000);
    out_format1->addIntParamValue(Format::sample_rate_param, 16000);
    out_format1->addIntParamValue(Format::sample_rate_param, 22050);
    out_format1->addIntParamValue(Format::sample_rate_param, 44100);
    out_format1->addIntParamValue(Format::sample_rate_param, 48000);
//    out_format1->addIntParamValue(Format::bitpersample_param, 8);
    out_format1->addIntParamValue(Format::bitpersample_param, 16);
    
    Format *out_format2 = getOwnOutputProperty()->addNewFormat(Format::audio_amr_type);
    out_format2->addIntParamValue(Format::sample_rate_param, 8000);
    out_format2->addIntParamValue(Format::channels_param, 1);
    out_format2->addIntParamValue(Format::bitpersample_param, 16);
    out_format2->addIntParamValue(Format::bitrate_param, 4750);
    out_format2->addIntParamValue(Format::bitrate_param, 5150);
    out_format2->addIntParamValue(Format::bitrate_param, 5900);
    out_format2->addIntParamValue(Format::bitrate_param, 6700);
    out_format2->addIntParamValue(Format::bitrate_param, 7400);
    out_format2->addIntParamValue(Format::bitrate_param, 7950);
    out_format2->addIntParamValue(Format::bitrate_param, 10200);
    out_format2->addIntParamValue(Format::bitrate_param, 12200);
    out_format2->addStringParamValue(Format::encoding_param, "amr_nb");
        
    getOwnOutputProperty()->setDefaultFormat(out_format2);
    
    Format *out_format3 = getOwnOutputProperty()->addNewFormat(Format::audio_amr_type);
    out_format3->addIntParamValue(Format::sample_rate_param, 16000);
    out_format3->addIntParamValue(Format::channels_param, 1);
    out_format3->addIntParamValue(Format::bitpersample_param, 16);
    //out_format3->addWildcard(Format::bitrate_param);
    out_format3->addIntParamValue(Format::bitrate_param, 6600);
    out_format3->addIntParamValue(Format::bitrate_param, 8850);
    out_format3->addIntParamValue(Format::bitrate_param, 12650);
    out_format3->addIntParamValue(Format::bitrate_param, 14250);
    out_format3->addIntParamValue(Format::bitrate_param, 15850);
    out_format3->addIntParamValue(Format::bitrate_param, 18250);
    out_format3->addIntParamValue(Format::bitrate_param, 19850);
    out_format3->addIntParamValue(Format::bitrate_param, 23050);
    out_format3->addIntParamValue(Format::bitrate_param, 23850);
    out_format3->addStringParamValue(Format::encoding_param, "amr_wb");
    
    return SUCCESS;
  }

  Result FFMpegAudioEncodeNode::doInitOutput() {
    /* get input format */
    const Format* in_format = getInputFormat();
    if (in_format == NULL) {
      ERROR_STREAM("Missing input format" << std::endl);
      return FAILURE;
    }
    // set output format
    // put input format in working property
    Format* i_format = new Format(*in_format);
    getOwnWorkingInputProperty()->addNewFormat(i_format);
    getOwnWorkingInputProperty()->setDefaultFormat(i_format);
   
    // need channels and sample rate
    if (in_format->isParameterSpecified(Format::sample_rate_param) &&
      in_format->isParameterSpecified(Format::channels_param)){
      
      // build output format from input format.
      sample_rate = in_format->getIntValue(Format::sample_rate_param);
      channels = in_format->getIntValue(Format::channels_param);
      byte_per_sample = in_format->getIntValue(Format::bitpersample_param)/8;   
      
      DEBUG_STREAM("SAMPLE_RATE: " << sample_rate << endl);
      DEBUG_STREAM("channels: " << channels << endl);
      DEBUG_STREAM("byte_per_sample: " << byte_per_sample << endl);
      
      setWOPInternal();
     
     } else {
     
     	 ERROR_STREAM("Input Format not fully specified." << endl);
	 return FAILURE;
         /*
       sample_rate = 0;
       channels = 0;
       // build output format from data analysis
       DEBUG_STREAM("No fully specified input format, need to analyze data" << endl);
       setMaxAnalysisIterations(250);
       if (!analyseData()) {
	 ERROR_STREAM("analyseData failed in doInitOutput" << endl);
	 return FAILURE;
       }
       */
     }
     
    // set I/O partners
    const_cast<Format*>(getOwnWorkingInputProperty()->
			getDefaultFormat())->
      setIOPartner(const_cast<Format*>(getOwnWorkingOutputProperty()->
				       getDefaultFormat()));
    
    const_cast<Format*>(getOwnWorkingOutputProperty()->
			getDefaultFormat())->
      setIOPartner(const_cast<Format*>(getOwnWorkingInputProperty()->
				       getDefaultFormat()));

    return SUCCESS;
  }

  Result FFMpegAudioEncodeNode::doActivate() {
  
    /*eget input format */
    const Format* in_format = getInputFormat();
    if (in_format == NULL) {
      ERROR_STREAM("Missing input format" << std::endl);
      return FAILURE;
    }
  
    cerr << "FFMpegAudioEncodeNode: Encoder is: " << compressor << endl;

    // find the correct encoder
    CodecConstIterator found = codecs.find(compressor);
   
    if(found == codecs.end()) {
      /* should not happen */
      ERROR_STREAM("FATAL: No codec listed for FourCC code " << compressor << std::endl);
      return FAILURE;
    }
    
    // find a codec with this ID
    {
      MutexGuard ffmpeg_guard(ffmpeg_mutex);
      ffmpeg->ffAVCodec = avcodec_find_encoder(static_cast<CodecID>(found->second));
    }
    if (!ffmpeg->ffAVCodec) {
      /* should not happen */
      ERROR_STREAM("FATAL: Unsupported codec " << std::endl);
      return FAILURE;
    }
    
    {
      MutexGuard ffmpeg_guard(ffmpeg_mutex);
      /* allocate ffmpeg context */
      ffmpeg->ffavcc = avcodec_alloc_context();
    }

    ffmpeg->ffavcc->sample_rate = sample_rate; 
    ffmpeg->ffavcc->channels = channels; 
    
    if (compressor == "MP3") {
      if (in_format->isParameterSpecified(Format::bitrate_param))
        bitrate = in_format->getIntValue(Format::bitrate_param); 
      else
        bitrate = 128000;
              
    } else if (compressor == "AMR_NB") {
    
      // amr_nb bitrate mode between 0 and 7
      bitrate = amr_nb_bitrate_modes[7];
      
    } else if (compressor == "AMR_WB") {
      
      // amr_wb bitrate mode between 0 and 8
      bitrate = amr_wb_bitrate_modes[8];
      
    } else {
      ERROR_STREAM("Unknown Compressor." << endl);
      return FAILURE;
    }
    
    ffmpeg->ffavcc->bit_rate = bitrate;
      
    work_buffer = getNewBuffer(10000);  
    work_buffer->setUsed(0);
    work_buf = reinterpret_cast<UINT8*>(work_buffer->getData());
    work_buf_ptr = work_buf;
      
    /* open the codec */
    {
      MutexGuard ffmpeg_guard(ffmpeg_mutex);
      if(avcodec_open(ffmpeg->ffavcc, ffmpeg->ffAVCodec) < 0) {
        ERROR_STREAM("Could not open codec" << std::endl);
    
        if(ffmpeg->ffavcc) {
        /* release ffmpeg context */
          free(ffmpeg->ffavcc);
          ffmpeg->ffavcc = NULL;
        }
        
        DEBUG_STREAM("about to return FAILURE" << endl);
        return FAILURE;
      }
    }
    
    frame_size = ffmpeg->ffavcc->frame_size;
    DEBUG_STREAM("frame_size: " << frame_size << endl); 
    input_bytes_needed = frame_size*channels*byte_per_sample;
    
    return SUCCESS;
  }
  
  Result FFMpegAudioEncodeNode::doDeactivate() {
  
    free(ffmpeg->ffavcc->extradata);
	  
    /* flush ffmpeg buffers */
    if(ffmpeg->ffavcc) {
      MutexGuard ffmpeg_guard(ffmpeg_mutex);
      avcodec_flush_buffers(ffmpeg->ffavcc);
    }
    
    work_buffer->release();
    
    {
      MutexGuard ffmpeg_guard(ffmpeg_mutex);
      /* close codec */
      avcodec_close(ffmpeg->ffavcc);
    }
      
    /* release ffmpeg context */
    if(ffmpeg->ffavcc) {
      free(ffmpeg->ffavcc);
      ffmpeg->ffavcc = NULL;
    }
    
    return SUCCESS;
  }
  
  //for testing amr mode change
  //static int buffer_counter = 0;
     
  Message* FFMpegAudioEncodeNode::processBuffer(Buffer* in_buffer) {
    DEBUG_STREAM("Entered processBuffer!" << endl);
    
    // for testing amr mode change
    /* 
    buffer_counter = (buffer_counter + 1) % 1000;
    
    if (compressor == "AMR_NB") {
      
      if (buffer_counter == 0) {
        bitrate = amr_nb_bitrate_modes[0];
        cerr << "Set AMR Bitrate to mode: " << 0 << endl;
      }
      
      if (buffer_counter == 500) {
        bitrate = amr_nb_bitrate_modes[7];
        cerr << "Set AMR Bitrate to mode: " << 7 << endl;
      }
      ffmpeg->ffavcc->bit_rate = bitrate;
    }
    
    if (compressor == "AMR_WB") {
      if (buffer_counter == 0) {
        bitrate = amr_wb_bitrate_modes[0];
        cerr << "Set AMR Bitrate to mode: " << 0 << endl;
      }
      
      if (buffer_counter == 500) {
        bitrate = amr_wb_bitrate_modes[8];
        cerr << "Set AMR Bitrate to mode: " << 8 << endl;
      }
      ffmpeg->ffavcc->bit_rate = bitrate;
    }
    */
    
           
    if(!getWorkingFlag()) {    
      // new buffer received
      inbuf = reinterpret_cast<UINT8*>(in_buffer->getData());
      inbuf_ptr = inbuf;
      size = in_buffer->getUsed();
      
      //cerr << "FFMpegAudioEncodeNode: Size: " << size << endl;
      
      // append in_buffer to work_buffer 
      DEBUG_STREAM("about to memmove " << size << " bytes." <<  endl);
      memmove(work_buf+work_buffer->getUsed(), inbuf_ptr, size);     
    
      // update used
      work_buffer->setUsed(work_buffer->getUsed()+size);
      
      if (!in_buf_ts.is_valid) {
        // Get the timestamp of the incoming buffer
        // only if we don't have one yet
        in_buf_ts = in_buffer->getTimestamp();
      }
      in_buffer->release();
    }    
    
    if (work_buffer->getUsed() < input_bytes_needed) {
      // not enough data for encoding available -> get next in_buffer
      DEBUG_STREAM("Returning from processBuffer. Not enough data available (" << work_buffer->getUsed() << "). Wait for next Buffer." << endl);
      //cerr << "Returning from processBuffer. Not enough data available (" << work_buffer->getUsed() << "). Wait for next Buffer." << endl;
      setWorkingFlag(false);
      //cerr << "ProcessBuffer: set working flag (false)" << endl;
      return 0;
    }
    
    DEBUG_STREAM("Enough data available (" << work_buffer->getUsed() << "). Process Buffer." << endl);
    //cerr << "Enough data available (" << work_buffer->getUsed() << "). Process Buffer." << endl;
      
    // we have enough data for encoding -> encode as long as enough data is available
    setWorkingFlag(true);
    //cerr << "ProcessBuffer: set working flag (true)" << endl;
    out_buffer = getNewBuffer(enc_buf_size);
        
    // encode audio 
    long bytes_encoded;
    
    {
      MutexGuard ffmpeg_guard(ffmpeg_mutex);   
      bytes_encoded = avcodec_encode_audio(ffmpeg->ffavcc, reinterpret_cast<UINT8*>(out_buffer->getData()), enc_buf_size, reinterpret_cast<short *>(work_buf));
    }
    
    bool discard_frame = false;
                              
    if (bytes_encoded < 0) {
      ERROR_STREAM("Bytes encoded was negative!" << endl);
      discard_frame = true;
    }
    
    if (bytes_encoded == 0) {
      DEBUG_STREAM("Null bytes encoded!" << endl);
      discard_frame = true;
    }
            
    DEBUG_STREAM("FFMpegAudioEncodeNode: Bytes encoded: " << bytes_encoded  << endl);
    //cerr << "FFMpegAudioEncodeNode: Bytes encoded: " << bytes_encoded  << endl;
    DEBUG_STREAM("FFMpegAudioEncodeNode: Size: " << size  << endl);
    
    //cerr << "FFMpegAudioEncodeNode: Bytes encoded: " << bytes_encoded  << endl;
    
    //update buffer pointers
    DEBUG_STREAM("about to memmove " << size << " bytes." <<  endl);
    
    if (!discard_frame) {
      // shift remaining data in work_buffer to leftmost position 
      memmove(work_buf, work_buf+input_bytes_needed, work_buffer->getUsed() - input_bytes_needed);
      
      out_buffer->setUsed(bytes_encoded);
    }
    
    // update work_buffer used
    work_buffer->setUsed(work_buffer->getUsed() - input_bytes_needed);  
    
    //cerr << "FFMpegAudioEncodeNode: out_buffer->getUsed():" << out_buffer->getUsed() << endl;

/*        
    DEBUG_STREAM("about to read number of samples." <<  endl);
    //XXX: need to carry this value forward as a format parameter from the ASFReadNode, even though its always 1 (i think)
    //num_samples = ffmpeg->ffavcc->frame_size;
    num_samples = 1;
    if(num_samples < 0) {
      DEBUG_STREAM("Number of samples invalid. Can not compute timestamp." << endl);
    }
    
    else {
    
      my_delay_num = num_samples;
      DEBUG_STREAM("Number of samples: " << num_samples << endl);
      my_delay_den = ffmpeg->ffavcc->sample_rate;
      DEBUG_STREAM("Sample rate: " << my_delay_den << endl);
      my_delay = (my_delay_num * 1000000) / my_delay_den; 
      my_delay *= 1000;
      DEBUG_STREAM("Duration of this buffer is " << my_delay << " nanoseconds." << endl);
    }
    */

    // the creation of the timestamps is mostly taken from the MPEGAudioEncodeNode code
    if(in_buf_ts.is_set) { // if timestamp was set by predecessor (e.g. MPEGDemuxNode)
      if(in_buf_ts.is_valid) {                              // if the incoming timestamp is valid
        DEBUG_STREAM("in_buf_ts is VALID!!!" << endl);
        //cerr << "in_buf_ts is VALID!!!" << endl;
        timestamp.sync_time = in_buf_ts.sync_time;          // take it
        timestamp.is_valid  = true;                         // remember that it was valid
        in_buf_ts.is_valid = false;                         // and make sure that it is taken only once
        first_timestamp = false;                            // also, remember that we have a valid timestamp
      }
      else {
        // if the first incoming timestamp after syncReset is invalid
        if(first_timestamp) {
          DEBUG_STREAM("FIRST in_buf_ts is NOT VALID!!!" << endl);
          //cerr << "FIRST in_buf_ts is NOT VALID!!!" << endl;
          timestamp.sync_time = in_buf_ts.sync_time;       // take it anyway
          timestamp.is_valid  = false;                     // but remember that it was invalid
          first_timestamp = false;                         // also, remember that we have some first (invalid) timestamp
        }
        else { // not valid, but found a valid ts before -> new ts will be computed by taking valid ts
          DEBUG_STREAM("in_buf_ts is NOT VALID but we already had a valid" << endl);
          //cerr << "in_buf_ts is NOT VALID but we already had a valid" << endl;
        }
      }
    }
    else { // else: timestamp was *not* set by predecessor (e.g. GenericReadNode) :
      timestamp.is_valid = true; // will generate timestamps internally -> valid timestamps
      // do not need to do anything else : timestamp should be zero, and will be incremented for every outgoing buffer
    }
    out_buffer->setTimestamp(timestamp);           // Finally set the timestamp
    DEBUG_STREAM("Set timestamp to: " << timestamp << endl);
    timestamp.stream_counter++;             // and do already the computations for the
    
    double samples_per_sec = sample_rate * channels * byte_per_sample; //1561*2;
    
//     cerr << "Samples_per_sec: " << samples_per_sec << endl;
//     cerr << "Outbuffer_getUsed: " << out_buffer->getUsed() << endl;
//     cerr << "channels: " << channels << endl;
//     cerr << "byte_per_sample:" << byte_per_sample << endl;
//     cerr << "sample_rate : " << sample_rate << endl;
    
    if (compressor == "MP3") {
      timestamp.sync_time += static_cast<long int>(( static_cast<double>(out_buffer->getUsed())/samples_per_sec) * 1000000000);
      
      // bitrate / 8 = bytes_per_sec;
      // 1/bytes_per_sec = sec_per_bytes
      // x = out_buffer->getUsed() * sec_per_bytes * 1000000000 
      // => buffer holds x nanoseconds of sound
      timestamp.sync_time += static_cast<long int>(1000000000 * static_cast<double>(out_buffer->getUsed())/static_cast<double>(bitrate/8.0)); 
    }
    else if ((compressor == "AMR_NB") || (compressor == "AMR_WB")) {
      // AMR frames: 20 ms; we need nanosec.
      timestamp.sync_time += static_cast<long int>(20*1000000);
    }
    
    //cerr << "Sync time: " << timestamp.sync_time << endl;
    
    //cout << "Audio timestamp " << timestamp << endl;
    DEBUG_STREAM("Returning from processBuffer." << endl);
    
    if (discard_frame) {
      out_buffer->release();
      return 0;
    }
    
    bufcount++;
    
    DEBUG_STREAM("FFMpegAudioEncodeNode: OutBuffer.used: " << out_buffer->getUsed() << endl);
    //cerr << "FFMpegAudioEncodeNode: OutBuffer.used: " << out_buffer->getUsed() << endl;
    
    return out_buffer; 
  }

  
  Result FFMpegAudioEncodeNode::doFlush() {
    setWorkingFlag(false);
    //cerr << "doFlush: set working flag (false)" << endl;
    return SUCCESS;
  }


  void FFMpegAudioEncodeNode::setWOPInternal() {
  
    // amr_nb
    if ((sample_rate == 8000) && (channels == 1)) {
      Format* o_format = getOwnWorkingOutputProperty()->addNewFormat(Format::audio_amr_type);
      o_format->addIntParamValue(Format::channels_param, channels);
      o_format->addIntParamValue(Format::sample_rate_param, sample_rate);
      o_format->addIntParamValue(Format::bitpersample_param, 16); // marc always scales to 16
      o_format->addIntParamValue(Format::bitrate_param, 4750);
      o_format->addIntParamValue(Format::bitrate_param, 5150);
      o_format->addIntParamValue(Format::bitrate_param, 5900);
      o_format->addIntParamValue(Format::bitrate_param, 6700);
      o_format->addIntParamValue(Format::bitrate_param, 7400);
      o_format->addIntParamValue(Format::bitrate_param, 7950);
      o_format->addIntParamValue(Format::bitrate_param, 10200);
      o_format->addIntParamValue(Format::bitrate_param, 12200);
      o_format->addStringParamValue(Format::encoding_param, "amr_nb");
      getOwnWorkingOutputProperty()->setDefaultFormat(o_format);
      compressor = "AMR_NB";
      DEBUG_STREAM("Using amr_nb" << endl);
      return;
    }
    
    // amr_wb
    if ((sample_rate == 16000) && (channels == 1)) {
      Format* o_format = getOwnWorkingOutputProperty()->addNewFormat(Format::audio_amr_type);
      o_format->addIntParamValue(Format::channels_param, channels);
      o_format->addIntParamValue(Format::sample_rate_param, sample_rate);
      o_format->addIntParamValue(Format::bitpersample_param, 16); // marc always scales to 16
      //o_format->addIntParamValue(Format::bitrate_param, bitrate); 
      o_format->addIntParamValue(Format::bitrate_param, 6600);
      o_format->addIntParamValue(Format::bitrate_param, 8850);
      o_format->addIntParamValue(Format::bitrate_param, 12650);
      o_format->addIntParamValue(Format::bitrate_param, 14250);
      o_format->addIntParamValue(Format::bitrate_param, 15850);
      o_format->addIntParamValue(Format::bitrate_param, 18250);
      o_format->addIntParamValue(Format::bitrate_param, 19850);
      o_format->addIntParamValue(Format::bitrate_param, 23050);
      o_format->addIntParamValue(Format::bitrate_param, 23850);
      o_format->addStringParamValue(Format::encoding_param, "amr_wb");
      getOwnWorkingOutputProperty()->setDefaultFormat(o_format);
      compressor = "AMR_WB";
      DEBUG_STREAM("Using amr_wb" << endl);
      return;
    }
    
    // mp3
    Format* o_format = getOwnWorkingOutputProperty()->addNewFormat(Format::audio_mpeg_type);
    o_format->addIntParamValue(Format::audio_codec_param, 0x55);
    o_format->addIntParamValue(Format::layer_param, 3);
    o_format->addIntParamValue(Format::channels_param, channels);
    o_format->addIntParamValue(Format::sample_rate_param, sample_rate);
    o_format->addIntParamValue(Format::bitpersample_param, 16); // marc always scales to 16
    o_format->addIntParamValue(Format::bitrate_param, bitrate);
    getOwnWorkingOutputProperty()->setDefaultFormat(o_format);
    compressor = "MP3";
    DEBUG_STREAM("Using MP3" << endl);
  }
  
  
    void FFMpegAudioEncodeNode::detectAllCodecs() throw() {
    /* get all available codecs */
    AVCodec *p = first_avcodec;
    DEBUG_STREAM("Available encoders:" << endl);
    while (p) {
      if (p->encode) {
        DEBUG_STREAM("  " << p->name << endl);
        switch(p->id) {
        case CODEC_ID_AMR_NB:
          addCodec("AMR_NB", p->id);
          break;
	case CODEC_ID_AMR_WB:
	  addCodec("AMR_WB", p->id);
	  break;
        case CODEC_ID_MP3:
          addCodec("MP3", p->id);
          break;
        default:
          break;
        }
      }

      p = p->next;
    }
    DEBUG_STREAM("Reached end of detectAllCodecs." << endl);
  }

  void FFMpegAudioEncodeNode::addCodec(const std::string& fourcc, int id) {
    /* add codec to map */
    codecs.insert(std::pair<std::string, int>(fourcc, id));
    DEBUG_STREAM("    " << fourcc << std::endl);
  }
  
  
}


#include "nmm/nmm_suffix.hpp"
