/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "IcecastNode.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"

NMM_REGISTER_PLUGIN(IcecastNode, NMM, "IcecastNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  IcecastNode::IcecastNode(const char* name,
			   const StreamQueue::Mode queue_mode,
			   const int queue_size)
    : GenericSinkNode(name, queue_mode, queue_size )
  {
    // initialisation of members
    shout_init();
    shout_connection = shout_new();
    ignore_incoming = false;
    send_title = false;
    shout_metadata = 0;
  };

  IcecastNode::~IcecastNode() {
    if(0 != shout_metadata) {
      shout_metadata_free(shout_metadata);
      shout_metadata = 0;
    }
    shout_free(shout_connection);
  }
  
  Result IcecastNode::doInit() {    
    // XXX: fw: more formats ...
    
    /* input format */
    Format *in_format = getOwnInputProperty()->addNewFormat(Format::audio_mpeg_type);
    in_format->addIntParamValue(Format::layer_param, 3);
    in_format->addIntParamValue(Format::channels_param, 2);
    in_format->addIntParamValue(Format::audio_codec_param, 0x55);
    in_format->addIntParamValue(Format::sample_rate_param, 44100);
    in_format->addIntParamValue(Format::sample_rate_param, 48000);
    in_format->addIntParamValue(Format::bitpersample_param, 16);

    getOwnInputProperty()->setDefaultFormat(in_format);

    Format *in_format2 = getOwnInputProperty()->addNewFormat(Format::audio_mpeg_type);
    in_format2->addIntParamValue(Format::layer_param, 3);
    in_format2->addIntParamValue(Format::channels_param, 2);
    in_format2->addIntParamValue(Format::audio_codec_param, 0x55);
    in_format2->addIntParamValue(Format::sample_rate_param, 44100);
    in_format2->addIntParamValue(Format::sample_rate_param, 48000);
    in_format2->addIntParamValue(Format::bitpersample_param, 16);
    in_format2->addWildcard(Format::bitrate_param);

    return SUCCESS;
  }
  
  Result IcecastNode::doDeactivate() {
    /* disconnect from server */
    if(shout_close(shout_connection) != SHOUTERR_SUCCESS) {
      ERROR_STREAM("could not disconnect... maybe never connected?" << endl);
      return FAILURE;
    }
    MESSAGE_STREAM("Disconnected from server" << endl);
    
    return SUCCESS;
  }
  
  Result IcecastNode::doActivate() {
    /* set protocol (currently only http supported) */
    if(shout_set_protocol(shout_connection, SHOUT_PROTOCOL_HTTP) != SHOUTERR_SUCCESS) {
      ERROR_STREAM(shout_get_error(shout_connection) << endl);
      return FAILURE;
    }
    
    /* set format (currently only MP3 supported) */
    if(shout_set_format(shout_connection, SHOUT_FORMAT_MP3) != SHOUTERR_SUCCESS) {
      ERROR_STREAM(shout_get_error(shout_connection) << endl);
      return FAILURE;
    }
    
    if(shout_open(shout_connection) == SHOUTERR_SUCCESS) {
      MESSAGE_STREAM("Connected to server " << shout_get_host(shout_connection) << endl);
    } else {
      // error, connection failed
      ERROR_STREAM(shout_get_error(shout_connection) << endl);
      return FAILURE;
    }
    return SUCCESS;
  }
  
  // set server ip
  Result IcecastNode::setServerIP(const string &name) {
    if(shout_set_host(shout_connection, name.c_str()) == SHOUTERR_SUCCESS) {
      DEBUG_STREAM("server ip set to " << shout_get_host(shout_connection) << endl);
      return SUCCESS;
    } else {
      ERROR_STREAM(shout_get_error(shout_connection) << endl);
      return FAILURE;
    }
  }
  
  // set server port
  Result IcecastNode::setServerPort(const nmm_int32 &port) {
    if(shout_set_port(shout_connection, port) == SHOUTERR_SUCCESS) {
      DEBUG_STREAM("port set to " << shout_get_port(shout_connection) << endl);
      return SUCCESS;
    } else {
      ERROR_STREAM(shout_get_error(shout_connection) << endl);
      return FAILURE;
    }
  }
  
  // set server password
  Result IcecastNode::setServerPassword(const string &pass) {
    if(shout_set_password(shout_connection, pass.c_str()) == SHOUTERR_SUCCESS) {
      DEBUG_STREAM("password set to " << shout_get_password(shout_connection) << endl);
      return SUCCESS;
    } else {
      ERROR_STREAM(shout_get_error(shout_connection) << endl);
      return FAILURE;
    }
  }
  
  // set user
  Result IcecastNode::setUser(const string &user) {
    if(shout_set_user(shout_connection, user.c_str()) == SHOUTERR_SUCCESS) {
      DEBUG_STREAM("user set to " << shout_get_user(shout_connection) << endl);
      return SUCCESS;
    } else {
      ERROR_STREAM(shout_get_error(shout_connection) << endl);
      return FAILURE;
    }
  }
  
  // set mountpoint
  Result IcecastNode::setMountpoint(const string &mount) {
    if(shout_set_mount(shout_connection, mount.c_str()) == SHOUTERR_SUCCESS) {
      DEBUG_STREAM("mountpoint set to " << shout_get_mount(shout_connection) << endl);
      return SUCCESS;
    } else {
      ERROR_STREAM(shout_get_error(shout_connection) << endl);
      return FAILURE;
    }
  }
  
  // set title
  Result IcecastNode::setTitle(const string &title) {
    DEBUG_STREAM("got title " << title << endl);
    this->title = title;
    return SUCCESS;
  }
  
  // set name of stream
  Result IcecastNode::setStreamName(const string &name) {
    if(shout_set_name(shout_connection, name.c_str()) == SHOUTERR_SUCCESS) {
      DEBUG_STREAM("name set to " << shout_get_name(shout_connection) << endl);
      return SUCCESS;
    } else {
      ERROR_STREAM(shout_get_error(shout_connection) << endl);
      return FAILURE;
    }
  }

  // set genre of stream
  Result IcecastNode::setGenre(const string &genre) {
    if(shout_set_genre(shout_connection, genre.c_str()) == SHOUTERR_SUCCESS) {
      DEBUG_STREAM("genre set to " << shout_get_genre(shout_connection) << endl);
      return SUCCESS;
    } else {
      ERROR_STREAM(shout_get_error(shout_connection) << endl);
      return FAILURE;
    }
  }

  // set description of stream
  Result IcecastNode::setDescription(const string &descr) {
    if(shout_set_description(shout_connection, descr.c_str()) == SHOUTERR_SUCCESS) {
      DEBUG_STREAM("description set to " << shout_get_description(shout_connection) << endl);
      return SUCCESS;
    } else {
      ERROR_STREAM(shout_get_error(shout_connection) << endl);
      return FAILURE;
    }
  }

  // set url with additional info
  Result IcecastNode::setURL(const string &url) {
    if(shout_set_url(shout_connection, url.c_str()) == SHOUTERR_SUCCESS) {
      DEBUG_STREAM("url set to " << shout_get_url(shout_connection) << endl);
      return SUCCESS;
    } else {
      ERROR_STREAM(shout_get_error(shout_connection) << endl);
      return FAILURE;
    }
  }
  
  // publish stream in directory server (e.g. yp.soutcast.com)
  Result IcecastNode::setPublic(const nmm_int32 &public_flag) {
    if(shout_set_public(shout_connection, public_flag) == SHOUTERR_SUCCESS) {
      DEBUG_STREAM("public flag set to " << shout_get_public(shout_connection) << endl);
      return SUCCESS;
    } else {
      ERROR_STREAM(shout_get_error(shout_connection) << endl);
      return FAILURE;
    }
  }
  
  // send metadata information
  void IcecastNode::sendMetadata(const string &name, const string &value) {
    if(shout_metadata == 0) {
      shout_metadata = shout_metadata_new();
    }
    shout_metadata_add(shout_metadata, name.c_str(), value.c_str());
    DEBUG_STREAM("sending metadata " << shout_metadata << endl);
    if (!shout_set_metadata(shout_connection, shout_metadata)) {
      ERROR_STREAM(shout_get_error(shout_connection) << endl);
    }
  }
  
  // send title as metadata
  Result IcecastNode::setSendTitle(const nmm_bool &send) {
    if(send) {
      DEBUG_STREAM("sending title as metadata" << endl);
    } else {
      DEBUG_STREAM("not sending title as metadata" << endl);
    }
    send_title = send;
    return SUCCESS;
  }
  
  Result IcecastNode::startTrack(const string &track) {
    MESSAGE_STREAM("now playing " << track << endl);
    if(send_title) {
      sendMetadata("title", title);
    }
    return SUCCESS;
  }
    
  Result IcecastNode::endTrack() {
    DEBUG_STREAM("end of track" << endl);
    if(0 != shout_metadata) {
      shout_metadata_free(shout_metadata);
      shout_metadata = 0;
    }
    return SUCCESS;
  }
  
  Result IcecastNode::started(const Interval&) {
    return SUCCESS;
  }

  Result IcecastNode::finished() {
    DEBUG_STREAM("finished" << endl);
    return SUCCESS;
  }
  
  Buffer *IcecastNode::processBuffer(Buffer *buffer) {
    if (ignore_incoming) {
      // ignore data
      DEBUG_STREAM("ignoring " << buffer->getUsed() << " bytes" << endl);
      buffer->release();
      return 0;
    }
    // send data
    if (shout_send(shout_connection, (unsigned char*)(buffer->getData()), buffer->getUsed()) != SHOUTERR_SUCCESS) {
      ERROR_STREAM(shout_get_error(shout_connection) << endl);
    }

    // wait for a while
    shout_sync(shout_connection);
    DEBUG_STREAM("sent " << buffer->getUsed() << " bytes to server" << endl);

    // release buffer
    buffer->release();

    return 0;
  }
}

#include "nmm/nmm_suffix.hpp"
