/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "MacOSXPlaybackNode.hpp"
#include "nmm/utils/SecurityCheck.hpp"
#include "nmm/utils/time/Rational.hpp"

#include <iostream>

NMM_REGISTER_PLUGIN(MacOSXPlaybackNode, NMM, "MacOSXPlaybackNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {
	
 OSStatus MyAudioDeviceIOProc(
    AudioDeviceID inDevice, 
    const AudioTimeStamp*inNow, 
    const AudioBufferList*inInputData, 
    const AudioTimeStamp*inInputTime, 
    AudioBufferList*outOutputData, 
    const AudioTimeStamp*inOutputTime, 
    void*inClientData) {
       	outOutputData->mBuffers[0].mDataByteSize =
       	static_cast<MacOSXPlaybackNode*>(inClientData)->readBuffer((float *)outOutputData->mBuffers[0].mData, 1024);
  	  	return 0;  
    }  
	
	
MacOSXPlaybackNode::MacOSXPlaybackNode(const char* name,
                             const StreamQueue::Mode queue_mode,
                             const int _size)
    : GenericBSyncSinkNode(name, queue_mode, _size), __condition(__condition_mutex), __current_buffer(0), __read_pos(0), __is_running(false), __to_big_endian(false), format_list(0)
  {  	
     /* Begin> Get list of Audio Devices */
     UInt32 size    = 0;
     UInt32 status  = 0;
     format_size = 0;
     
     status = AudioHardwareGetPropertyInfo (
             				kAudioHardwarePropertyDevices,
     						&size, 
    						NULL );
     
     UInt32 number_of_devices = size / sizeof(AudioDeviceID);
     /* End */
     
     AudioDeviceID* device_list = new AudioDeviceID[ number_of_devices ];  	
  
     status = AudioHardwareGetProperty (
     		kAudioHardwarePropertyDevices,
     		&size, device_list );

     for (UInt32 i = 0; i < number_of_devices; ++i) {
        status = AudioDeviceGetPropertyInfo(
                  device_list[i], NULL, false,
                  kAudioDevicePropertyStreamFormats,
                  &size,
                  NULL );
 
         //get formats of the audio device
         format_size = size / sizeof(AudioStreamBasicDescription);
         if (format_size > 0) {
           __device = device_list[i];
           break;
         }
      }

     format_list = new AudioStreamBasicDescription[ format_size ];     
     status = AudioDeviceGetProperty(
      __device, NULL, false, 
      kAudioDevicePropertyStreamFormats,
      &size,
      format_list );
    
     UInt32 format_id = 0;
     for ( UInt32 i = 0; i < format_size; ++i ) {
       addFormat(format_list[i]);
     }
     
     format_size =  sizeof(AudioStreamBasicDescription);     
     int res = AudioDeviceSetProperty(
      __device, NULL, 0, false, 
      kAudioDevicePropertyStreamFormat,
      format_size,
      &format_list[format_id] );     

      delete [] device_list;
  }  
  
  MacOSXPlaybackNode::~MacOSXPlaybackNode() {
  	delete [] format_list;
  }
      
Result MacOSXPlaybackNode::doInit() {
    return SUCCESS;
}

Result MacOSXPlaybackNode::doInitOutput() { 
    const Format* input_format = getInputFormat();
    AudioStreamBasicDescription current_format;
    UInt32 param_size = sizeof(AudioStreamBasicDescription);
    
    __current_sample_rate    = input_format->getIntValue(Format::sample_rate_param);
    __current_channels       = input_format->getIntValue(Format::channels_param);
    
    
    UInt32 format_id = 0;
    for ( UInt32 i = 0; i < format_size; ++i ) {     
      if ( format_list[i].mSampleRate == __current_sample_rate && __current_channels == format_list[i].mChannelsPerFrame ){
         format_id = i;
         break;
      }
    }
         
    int res = AudioDeviceSetProperty(
      __device, NULL, 0, false, 
      kAudioDevicePropertyStreamFormat,
      format_size,
      &format_list[format_id] );     


  #ifdef BUILD_PPC
  __to_big_endian = true;
  #endif
  
  return SUCCESS;
}

Result MacOSXPlaybackNode::doActivate(){
	//Register audio handler 
   	OSStatus status = AudioDeviceAddIOProc(__device, MyAudioDeviceIOProc,this);
    	if (status) {
    		ERROR_STREAM("Could not register audio handler " << endl);	
		return FAILURE;
    	}
 
  	return SUCCESS;
}

Result MacOSXPlaybackNode::doStart() {
	__is_running = true;
	// Start the audio device   	
   	OSStatus status = AudioDeviceStart(__device, MyAudioDeviceIOProc);
   	if (status) {
   		ERROR_STREAM("Could not start audio device " << endl);
   		return FAILURE;
   	}
  	return SUCCESS;
}

Result MacOSXPlaybackNode::doStop() {
    __condition_mutex.lock();
    __is_running = false;
    __condition.notify();
    __condition_mutex.unlock();

	// Stop audio device 
 	OSStatus status = AudioDeviceStop(__device, MyAudioDeviceIOProc);
	
	if (status) {
		ERROR_STREAM("Could not stop audio device " << endl);
	  	return FAILURE;
	}
	 
	return SUCCESS;
}

Result MacOSXPlaybackNode::doDeactivate() {	
	/* Begin> Release already received buffer */
	if (__current_buffer) {
		__current_buffer->release();
		__read_pos = 0;
	}
	for (list<Buffer*>::iterator it = __buffer_list.begin(); it != __buffer_list.end(); ++it) {
		(*it)->release();
	}
	__buffer_list.clear();
	/* End< */
	
	// Remove audio handler
	OSStatus status = AudioDeviceRemoveIOProc(__device, MyAudioDeviceIOProc);
	
	if (status) {
		ERROR_STREAM("Could not remove audio handler " << endl);
	  	return FAILURE;
	}
	
	return SUCCESS;
}

void MacOSXPlaybackNode::mute() {
//  kAudioDevicePropertyMute = 'mute'
}
void MacOSXPlaybackNode::unmute() {
//  kAudioDevicePropertyMute = 'mute'
}

void MacOSXPlaybackNode::setVolume(const nmm_int32& volume) {

/*Boolean isWritable;
theStatus = AudioDeviceGetPropertyInfo (
              theDeviceID, 2, false,
              kAudioDevicePropertyVolumeScalar,
              NULL, &isWritable );

If ( (theStatus == kAudioHardwareNoError)
       && isWritable) {
       Float32 theValue;
       theStatus = AudioDeviceSetProperty (
       theDevice, NULL, 2, false,
       kAudioDevicePropertyVolumeScalar,
       sizeof(Float32), &theValue );
*/
}

Buffer* MacOSXPlaybackNode::presentBuffer(Buffer* buffer) {  
  __thread_mutex.lock();
  
  if (__to_big_endian)
  	toBigEndian(buffer);
 
  __buffer_list.push_back (buffer);

  if (__buffer_list.size() >= 20) { 
    __condition_mutex.lock();
    __thread_mutex.unlock();  
    if (__is_running) { 
      __condition.wait();
    }
    __condition_mutex.unlock();
  } else {
    __thread_mutex.unlock();	
  }
 
  return 0;	
}

void MacOSXPlaybackNode::toBigEndian(Buffer* buffer) {
  char* data = buffer->getData();
  char tmp;
  for (unsigned int i = 0; i < buffer->getUsed(); i+=2){
    tmp = data[i];
    data[i] = data[i+1];
    data[i+1]= tmp;  		
  }	
}

/*  void MacOSXPlaybackNode::mute()
  {
    DEBUG_STREAM("MUTE" << endl);
    getController()->setBuffersInvalid(true);
  }

  void MacOSXPlaybackNode::unmute()
  {
    DEBUG_STREAM("UN-MUTE" << endl);
    getController()->setBuffersInvalid(false);
  }*/
  void MacOSXPlaybackNode::addFormat(const AudioStreamBasicDescription& format)
  {
  	Format* nmm_format = new Format();
    switch(format.mFormatID) {
    case kAudioFormatAC3:
      nmm_format->setFormatType("audio/ac3");
      break;

    case kAudioFormat60958AC3:
      nmm_format->setFormatType("audio/raw");
      nmm_format->addStringParamValue(Format::encoding_param, "a_law");
      break;

    case kAudioFormatLinearPCM:
      nmm_format->setFormatType("audio/raw");
      nmm_format->addStringParamValue(Format::encoding_param, "linear");
      if ( format.mFormatFlags & kAudioFormatFlagIsSignedInteger )
        nmm_format->addStringParamValue(Format::sign_param, "unsigned");
      else 
        nmm_format->addStringParamValue(Format::sign_param, "signed");

      nmm_format->addIntParamValue(Format::bitpersample_param, format.mBitsPerChannel/format.mChannelsPerFrame);
      nmm_format->addIntParamValue(Format::sample_rate_param, static_cast<int32_t>(format.mSampleRate));
      nmm_format->addIntParamValue(Format::channels_param, format.mChannelsPerFrame);
      nmm_format->addStringParamValue(Format::endian_param, "little");
      break;
    }
   
 /*   if (format.mFormatFlags &     kAudioFormatFlagIsFloat) {
	  cerr << " FORMAT IS ao_data.format = AFMT_FLOAT; " << endl;
    } else {
      cerr << "Is not FLOAT !!!!!" << endl;
    }*/
    getOwnInputProperty()->addNewFormat(nmm_format);
    if (!getOwnInputProperty()->getDefaultFormat()) {
	  getOwnInputProperty()->setDefaultFormat(nmm_format);
    }
  }  
   

Result MacOSXPlaybackNode::getNextBuffer() {
  __thread_mutex.lock(); 
  if ( __current_buffer ) {
    __current_buffer->release();
    __current_buffer = 0;   
  }
  
  if (__buffer_list.empty()) {
    //WARNING_STREAM("Buffer underrun " << endl);
    __thread_mutex.unlock();
    return FAILURE; 
  }
  
  __current_buffer = *__buffer_list.begin();
  __buffer_list.erase(__buffer_list.begin());
  __thread_mutex.unlock();

  __condition_mutex.lock();
  __condition.notify();
  __condition_mutex.unlock();
  return SUCCESS;
}  

Interval MacOSXPlaybackNode::getOutputDelay() {

  __thread_mutex.lock();
  int byte_count = 0;
  for (list<Buffer*>::iterator it = __buffer_list.begin(); it != __buffer_list.end(); ++it ) {
    byte_count += (*it)->getUsed();
  }
  if (__current_buffer) {
    byte_count += __current_buffer->getUsed()- __read_pos;
//    byte_count += 4096; //this seems to be the internal size of the audio buffer 
  }
  __thread_mutex.unlock();
  return Interval(Rational(byte_count, __current_channels*__current_sample_rate*2));
}  

int MacOSXPlaybackNode::readBuffer(float* data,int len){
  if (!isStarted()) {
    return 0; 
  }

  if (!__current_buffer) {
    if (getNextBuffer() == FAILURE) {
      return 0;	
    }
  } 

  int j = 0;  
  int counter = 0;
  
  SInt16* int_array = (SInt16*) (__current_buffer->getData());
  int_array += __read_pos;
  while (1) {
    // if the audio buffer of the used device is full return the buffer size
    if ( counter >= len) {   
        return counter*4;
    }    
    
    // if we read all data from buffer, request a new one
    if( __read_pos >= __current_buffer->getUsed() / 2) {
      __read_pos = 0; 
      
      if (getNextBuffer() == FAILURE) {
        return counter*4;
      } 
   
      int_array = (SInt16*) (__current_buffer->getData());
      int_array += __read_pos;
    } 
       
    *data = *int_array;
    *data /= 32768.0;

    ++data;
    ++int_array;
    ++counter;    
    ++__read_pos;
  }
  return __buffer_size;
}
    
}
 



#include "nmm/nmm_suffix.hpp"
