/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


/*
 * scale and output functions are taken from minimad.c example of libmad package that is
 * copyrighted 2000-2001 by Robert Leslie
 */

#include "MPEGAudioDecodeNode.hpp"

// #include "nmm/base/format/Property.hpp"
// #include "nmm/base/registry/Plugin.hpp"

#include "nmm/plugins/NMMRegisterPlugin.hpp"
#include "nmm/comm/messaging/TEDObject.hpp"
#include "nmm/comm/serialize/SerializeResult.hpp"
#include "nmm/comm/serialize/SerializeUtilsTime.hpp"

#include "nmm/utils/time/Rational.hpp"
#include "nmm/utils/thread/MutexGuard.hpp"

NMM_REGISTER_PLUGIN(MPEGAudioDecodeNode, NMM, "MPEGAudioDecodeNode")

// adjust for windows gheyness
#undef DELETE

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  MPEGAudioDecodeNode::MPEGAudioDecodeNode( const char * name, StreamQueue::Mode qmode, int qsize)
    : GenericProcessorNode(name, qmode, qsize)
    , timestamp()
    , my_delay(0)  
    , in_buf_ts()
    , sample_rate(0)
    , channels(0)
    , bitrate(0)
    , seek_time()
    , mad_mutex()
    , first_timestamp(true)
    , track_size(0)
    , fragment_size(0)
    , rest_length(0)
    , cur_buffer_length(0)
    , cur_buffer_end(0)
    , tmpbuffer(0)
    , inbuffer_size(0)
    , track_duration(0)
    , send_track_duration(false)
  {    
    
    // !! flags that trigger something should be set in doActivate, not in the constructor !!
    tmpbuffer = NULL;
    fragment_size = 2048;
    inbuffer_size = fragment_size;
    track_duration = Interval(0,0);

    memset(&stream, 0, sizeof(mad_stream));
    memset(&frame, 0, sizeof(mad_frame));
    memset(&synth, 0, sizeof(mad_synth));

    syncReset();
    
    registerEvent("SEEK", new TEDObject3<MPEGAudioDecodeNode, Interval, Time, nmm_int32>(this, &MPEGAudioDecodeNode::eventSeek));
  };
  

  MPEGAudioDecodeNode::~MPEGAudioDecodeNode()
  {
    if (tmpbuffer) {
      delete[] tmpbuffer;
      tmpbuffer = NULL;
    }
  }
  
  Result MPEGAudioDecodeNode::doInit() {
    
    Format* in_format;
    
    // register all supported input formats
    // format without bitrate
    in_format = getOwnInputProperty()->addNewFormat("audio/mpeg");
    in_format->addIntParamValue(Format::audio_codec_param, 0x55);
    in_format->addWildcard(Format::layer_param);
    in_format->addIntParamValue(Format::channels_param, 1);
    in_format->addIntParamValue(Format::channels_param, 2);
    in_format->addIntParamValue(Format::sample_rate_param, 8000);
    in_format->addIntParamValue(Format::sample_rate_param, 11025);
    in_format->addIntParamValue(Format::sample_rate_param, 12000);
    in_format->addIntParamValue(Format::sample_rate_param, 16000);
    in_format->addIntParamValue(Format::sample_rate_param, 22050);
    in_format->addIntParamValue(Format::sample_rate_param, 44100);
    in_format->addIntParamValue(Format::sample_rate_param, 48000);
    in_format->addIntParamValue(Format::bitpersample_param, 8);
    in_format->addIntParamValue(Format::bitpersample_param, 16);
    getOwnInputProperty()->setDefaultFormat(in_format);
    
    // format with bitrate
    in_format = getOwnInputProperty()->addNewFormat("audio/mpeg");
    in_format->addIntParamValue(Format::audio_codec_param, 0x55);
    in_format->addWildcard(Format::layer_param);
    in_format->addWildcard(Format::bitrate_param);
    in_format->addIntParamValue(Format::channels_param, 1);
    in_format->addIntParamValue(Format::channels_param, 2);
    in_format->addIntParamValue(Format::sample_rate_param, 8000);
    in_format->addIntParamValue(Format::sample_rate_param, 11025);
    in_format->addIntParamValue(Format::sample_rate_param, 12000);
    in_format->addIntParamValue(Format::sample_rate_param, 16000);
    in_format->addIntParamValue(Format::sample_rate_param, 22050);
    in_format->addIntParamValue(Format::sample_rate_param, 44100);
    in_format->addIntParamValue(Format::sample_rate_param, 48000);
    in_format->addIntParamValue(Format::bitpersample_param, 8);
    in_format->addIntParamValue(Format::bitpersample_param, 16);
    
    in_format = getOwnInputProperty()->addNewFormat("audio/mpeg");
    in_format->addIntParamValue(Format::audio_codec_param, 0x55);
    in_format->addWildcard(Format::layer_param);
    in_format->addWildcard(Format::bitrate_param);

    in_format = getOwnInputProperty()->addNewFormat("audio/mpeg");
    in_format->addIntParamValue(Format::audio_codec_param, 0x55);
    in_format->addWildcard(Format::layer_param);
    
    // this format is needed so that a GenericReadNode with
    // output format = mime type of read file can be connected
    in_format = getOwnInputProperty()->addNewFormat("audio/mpeg");

    // register all supported output formats
    Format* out_format = getOwnOutputProperty()->addNewFormat("audio/raw");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::encoding_param, "linear");
    out_format->addStringParamValue(Format::sign_param, "signed");
    out_format->addIntParamValue(Format::sample_rate_param, 8000);
    out_format->addIntParamValue(Format::sample_rate_param, 11025);
    out_format->addIntParamValue(Format::sample_rate_param, 12000);
    out_format->addIntParamValue(Format::sample_rate_param, 16000);
    out_format->addIntParamValue(Format::sample_rate_param, 22050);
    out_format->addIntParamValue(Format::sample_rate_param, 44100);
    out_format->addIntParamValue(Format::sample_rate_param, 48000);
    out_format->addIntParamValue(Format::channels_param, 1);
    out_format->addIntParamValue(Format::channels_param, 2);
    out_format->addIntParamValue(Format::bitpersample_param, 16);
    getOwnOutputProperty()->setDefaultFormat(out_format);
    
    // set I/O partners
    in_format->setIOPartner(out_format);
    out_format->setIOPartner(in_format);
    
    // initialize MPEG Audio Decoder
    mad_stream_init(&stream);
    mad_frame_init(&frame);
    mad_synth_init(&synth);
    rest_length = 0;

    return SUCCESS;
  }
  
  Result MPEGAudioDecodeNode::doDeinit() {
    // deinitialize MPEG Audio Decoder
    mad_synth_finish(&synth);
    mad_frame_finish(&frame);
    mad_stream_finish(&stream);
    
    track_duration = Interval(0,0);
    
    return SUCCESS;
  }
  
  Result MPEGAudioDecodeNode::doInitOutput() {
    DEBUG_STREAM("MPEGAudioDecodeNode::doInitOutput" << endl);
    // get input format
    const Format* in_format = getInputFormat();

    if (in_format == NULL) {
      DEBUG_STREAM("in_format is NULL!" << endl);
      return FAILURE;
    }
    
    // set output format
    // put input format in working property
    Format* i_format = new Format(*in_format);
    getOwnWorkingInputProperty()->addNewFormat(i_format);
    getOwnWorkingInputProperty()->setDefaultFormat(i_format);
    
    if (in_format->isParameterSpecified(Format::sample_rate_param) && 
	in_format->isParameterSpecified(Format::channels_param)) {
      // build output format from input format.
      sample_rate = in_format->getIntValue(Format::sample_rate_param);
      channels = in_format->getIntValue(Format::channels_param);
      setWOPInternal();
    } else {
      sample_rate = 0;
      channels = 0;
      // build output format from data analysis
      if (!analyseData()) {
	ERROR_STREAM("analyseData failed in doInitOutput" << endl);
	return FAILURE;
      }
    }
    
    // set I/O partners
    const_cast<Format*>(getOwnWorkingInputProperty()->
			getDefaultFormat())->
      setIOPartner(const_cast<Format*>(getOwnWorkingOutputProperty()->
				       getDefaultFormat()));
    
    const_cast<Format*>(getOwnWorkingOutputProperty()->
			getDefaultFormat())->
      setIOPartner(const_cast<Format*>(getOwnWorkingInputProperty()->
				       getDefaultFormat()));

    send_track_duration = true;
    
    return SUCCESS;
  }
  
  Result MPEGAudioDecodeNode::doSetFormat(const Format& ) {
    DEBUG_STREAM("MPEGAudioDecodeNode::doSetFormat" << endl);
    return SUCCESS;
  }

  void MPEGAudioDecodeNode::setWOPInternal() {
    // clear output property
    // wrong: getOwnOutputProperty()->erase();
    // should be: getOwnWorkingOutputProperty()->erase();
    // and is also done in Generic-layer !!!
    
    Format* o_format = getOwnWorkingOutputProperty()->addNewFormat("audio/raw");
    o_format->addStringParamValue(Format::endian_param, "little");
    o_format->addStringParamValue(Format::encoding_param, "linear");
    o_format->addStringParamValue(Format::sign_param, "signed");
    o_format->addIntParamValue(Format::sample_rate_param, sample_rate);
    o_format->addIntParamValue(Format::channels_param, channels);
    o_format->addIntParamValue(Format::bitpersample_param, 16); // marc always scales to 16
    getOwnWorkingOutputProperty()->setDefaultFormat(o_format);
  }

  void MPEGAudioDecodeNode::resetMP3() {
    mad_synth_finish(&synth);
    mad_frame_finish(&frame);
    mad_stream_finish(&stream);

    mad_stream_init(&stream);
    mad_frame_init(&frame);
    mad_synth_init(&synth);
    rest_length = 0;
  }

//  Result MPEGAudioDecodeNode::seeked(const u_int64_t& filepos) {
//    seek_time = Time(Rational(filepos,bitrate/8));
//    return DELETE;
//  }

  Result MPEGAudioDecodeNode::startTrack(const string&) { // we don't need the filename here
    DEBUG_STREAM("Message for me detected: startTrack" << endl);
    DEBUG_STREAM("Resetting MP3-Decoder-Lib" << endl);
    resetMP3();
    setWorkingFlag(false);
    return SUCCESS;
  }
  
  Result MPEGAudioDecodeNode::endTrack() {
    return SUCCESS;
  }
   
  inline signed int scale(mad_fixed_t sample) {
    /* round */
    sample += (1L << (MAD_F_FRACBITS - 16));

    /* clip */
    if (sample >= MAD_F_ONE) sample = MAD_F_ONE - 1;
    else if (sample < -MAD_F_ONE) sample = -MAD_F_ONE;

    /* quantize */
    return sample >> (MAD_F_FRACBITS + 1 - 16);
  }

  Message *MPEGAudioDecodeNode::processBuffer( Buffer *in_buffer ) {
    
    /*
    if (in_buffer) {
      cout << "LOGGER processBuffer " << in_buffer->getTimestamp().sync_time 
	   << " is_valid " << in_buffer->getTimestamp().is_valid 
	   << " is_set " << in_buffer->getTimestamp().is_set << endl;
    }
    */

    MutexGuard mad_mg(mad_mutex);

    if(!getWorkingFlag()) { // start processing new buffer
      /* save incoming buffer size */
      cur_buffer_length = in_buffer->getUsed();

      /**
       * since we must copy some undecoded data from the last buffer
       * (because not enought data was available to decode the last frame)
       * we need some space before the incoming buffer to put this data in it
       * that's why we copy the incoming buffer in an extra buffer that is a bit larger
       * usually the incoming buffers has got the same size, so we adjust our extra
       * buffer only if the incoming buffer is larger, so we save some memory allocation
       */
      if ( cur_buffer_length > (inbuffer_size - fragment_size) ) {
        inbuffer_size = cur_buffer_length + fragment_size;

        if (tmpbuffer) {
          delete[] tmpbuffer;
        }
        /* our extra buffer with some extra space */
        tmpbuffer = new unsigned char[inbuffer_size];
        DEBUG_STREAM("New temp. buffer created with size:" << inbuffer_size << endl);
      }

      /* pointer to our end data */
      cur_buffer_end  = tmpbuffer + cur_buffer_length + rest_length;

      /**
      * copy incoming buffer into extra buffer and skip some bytes before
      * because here we find the undecoded data from the last buffer
      */
      memmove(tmpbuffer+rest_length, in_buffer->getData(), cur_buffer_length);

      if (!in_buf_ts.is_valid) {
        // Get the timestamp of the incoming buffer
        // only if we don't have one yet
        in_buf_ts = in_buffer->getTimestamp();
      }

      /* now we don't need the incoming buffer anymore */
      in_buffer->release();

      /* tell the mpeg audio lib where the audio stream can be found */
      mad_stream_buffer(&stream, tmpbuffer, cur_buffer_length+rest_length);

      setWorkingFlag( true );
    }

    /* try to decode audio frame */
    while (mad_frame_decode(&frame, &stream) == -1) {
	
      /* MAD_ERROR_BUFLEN occures, if we have not enought data to decode the frame */
      if (stream.error == MAD_ERROR_BUFLEN) {
	  
	/* calculate the size of the data that could not be decoded */
	rest_length = cur_buffer_end - stream.next_frame;
	  
	/**
	 * copy this piece of data to the begin of our extra buffer, so the next time
	 * enought data should be there to decode the frame
	 */
	memmove(tmpbuffer, stream.next_frame, rest_length);
	
	setWorkingFlag(false);
	
	return(0);
      }
      
      /* if we can not fix the broken stream data, hang up */
      if (!MAD_RECOVERABLE(stream.error)) {
	setWorkingFlag(false);
	return(0); 
      };
    } //while


    if (send_track_duration) {
      /* get VBR header ID, which is a 4-character string located 36 bytes after frame header */
      char s[] = {stream.this_frame[36], stream.this_frame[37], stream.this_frame[38], stream.this_frame[39], 0};
      string headerID(s);
      
      if (headerID == "Xing") {
	/* if there is a Xing header, get the number of frames to compute the track duration */
	u_int32_t samples_per_frame = 0;
    
	/* depending on the MPEG layer and type, the number of samples per frame varies. */
	switch (frame.header.layer) {
	case 1: samples_per_frame = 384; break;
	case 2: samples_per_frame = 1152; break;
	case 3:
	  if (frame.header.flags & MAD_FLAG_LSF_EXT || frame.header.flags & MAD_FLAG_MPEG_2_5_EXT) {
	    samples_per_frame = 576;
	  }
	  else {
	    samples_per_frame = 1152;
	  } break;
	default: break;
	}

	/* get the 32 bit frame size */
	u_int32_t frames =
	  stream.this_frame[47] |
	  (stream.this_frame[46] << 8) |
	  (stream.this_frame[45] << 16) |
	  (stream.this_frame[44] << 24);    

	/* compute the track duration */
	track_duration = frames * samples_per_frame / frame.header.samplerate;
      } else {
  	if ((bitrate != 0) && (track_duration == 0)) {
          track_duration = track_size / (bitrate/8);
	}
      }


      /* finally send the track duration */
      setWorkingFlag(true);
      send_track_duration = false;

      Event* e = ITrackDuration::create_trackDuration(Interval(track_duration,0));
      return new CEvent(e);      
    }

    // when we are here, a decodable frame is available
    mad_synth_frame(&synth, &frame);

    // read out the frame's bitrate
    bitrate = frame.header.bitrate;

    // mad stores channel infos as an enum, compute the number of output channels
    int header_channels(0);
    switch (frame.header.mode) {
    case MAD_MODE_SINGLE_CHANNEL:
      header_channels = 1;
      break;
    case MAD_MODE_DUAL_CHANNEL:
    case MAD_MODE_JOINT_STEREO:
    case MAD_MODE_STEREO:
      header_channels = 2;
      break;
    }

    if ((frame.header.samplerate != synth.pcm.samplerate) ||
        (header_channels != synth.pcm.channels)) {
      WARNING_STREAM("********************************************************************************" << endl);
      WARNING_STREAM("header says samplerate=" << frame.header.samplerate << ", channels=" << header_channels << endl);
      WARNING_STREAM("mad's pcm says samplerate=" << synth.pcm.samplerate << ", channels=" << synth.pcm.channels << endl);
      WARNING_STREAM("********************************************************************************" << endl);
      setWorkingFlag(false);
      resetMP3();
      return NULL;
    }

    // check for instream format changes and data analysis
    if ((frame.header.samplerate != sample_rate) || 
        (header_channels != channels)) {

      sample_rate = frame.header.samplerate;
      channels = header_channels;

      setWOPInternal();
      setWorkingFlag(true);

      if (getAnalysingFlag()) {
	DEBUG_STREAM("setAnalysingFlag(false)" << endl);
        setAnalysingFlag(false);
        return NULL;
      } 
    }

    // decode the data!
    /* pcm->samplerate contains the sampling frequency */
    unsigned int nchannels = synth.pcm.channels;
    unsigned int nsamples = synth.pcm.length;
    mad_fixed_t const* left_ch = synth.pcm.samples[0];
    mad_fixed_t const* right_ch  = synth.pcm.samples[1];

    // TODO: replace sample_rate with pcm.samplerate and remove sample_rate retrieving in doActivate
    my_delay = nsamples * 1000000 / sample_rate;     // Compute the duration of 
    my_delay *= 1000;                                // the content of this
                                                     // buffer in nanoseconds.
    DEBUG_STREAM("Duration of this buffer is " << my_delay << " nanoseconds." << endl);


    Buffer* out(getNewBuffer(nchannels * nsamples * 2));
    char* sbuffer(out->getData());
    
    // if we have two channels, one "stereo" sample (left&right) has 4
    // Bytes, as our mono samples are always scaled to 16 bit.
    unsigned int bytes_per_stereo_sample(2 * nchannels);
    // for all samples (left&right)
    for (unsigned int s = 0;
         s < (bytes_per_stereo_sample * nsamples);
         s += bytes_per_stereo_sample) {
      unsigned int channel_sample(s / bytes_per_stereo_sample);
      // scale ...
      // left or mono channel
      signed int sample(scale(left_ch[channel_sample]));
      sbuffer[s + 0] = (sample >> 0) & 0xff;
      sbuffer[s + 1] = (sample >> 8) & 0xff;
      if (nchannels == 2) {
        // ... and interleave them!
        // right channel
        signed int sample(scale(right_ch[channel_sample]));
        sbuffer[s + 2] = (sample >> 0) & 0xff;
        sbuffer[s + 3] = (sample >> 8) & 0xff;
      }
    }

    out->setUsed(bytes_per_stereo_sample * nsamples);
    
    if(in_buf_ts.is_set) { // if timestamp was set by predecessor (e.g. MPEGDemuxNode)
      if(first_timestamp && in_buf_ts.is_valid) {                              // if the incoming timestamp is valid
	DEBUG_STREAM("in_buf_ts is VALID!!!" << endl);
	// cerr << "in_buf_ts is VALID!!!" << endl;
	timestamp.sync_time = in_buf_ts.sync_time;          // take it 
	timestamp.is_valid  = true;                         // remember that it was valid
	in_buf_ts.is_valid = false;                         // and make sure that it is taken only once
	first_timestamp = false;                            // also, remember that we have a valid timestamp 
      }                                                    
      else {
	// if the first incoming timestamp after syncReset is invalid  
	if(first_timestamp) {
	  DEBUG_STREAM("FIRST in_buf_ts is NOT VALID!!!" << endl);
	  // cerr << "LOGGER FIRST in_buf_ts is NOT VALID!!!" << endl;
	  timestamp.sync_time = in_buf_ts.sync_time;       // take it anyway
	  timestamp.is_valid  = false;                     // but remember that it was invalid
	  first_timestamp = false;                         // also, remember that we have some first (invalid) timestamp
	}
	else { // not valid, but found a valid ts before -> new ts will be computed by taking valid ts
	  DEBUG_STREAM("in_buf_ts is NOT VALID but we already had a valid" << endl);
	  // cerr << "LOGGER in_buf_ts is NOT VALID but we already had a valid" << endl;
	}
      }
    }
    else { // else: timestamp was *not* set by predecessor (e.g. GenericReadNode) :
      timestamp.is_valid = true; // will generate timestamps internally -> valid timestamps
      // do not need to do anything else : timestamp should be zero, and will be incremented for every outgoing buffer
    }

    out->setTimestamp(timestamp);           // Finally set the timestamp
    DEBUG_STREAM("Set timestamp to: " << timestamp << endl);
    // cout << "LOGGER Set timestamp to: " << timestamp << " my_delay " << my_delay << " size " << out->getUsed() << endl;
    timestamp.stream_counter++;             // and do already the computations for the
    timestamp.sync_time += my_delay;        // next outgoing buffer's timestamp.

    return out;
  }

  Result MPEGAudioDecodeNode::doFlush() {
    setWorkingFlag( false );
    rest_length = 0;

    return( SUCCESS );
  }
  
  Result MPEGAudioDecodeNode::syncReset() {
    in_buf_ts.is_valid = false;    // make sure, we take the ts of the next incoming buffer
    in_buf_ts.is_set   = false;
    timestamp.stream_counter = 0;  // Reset the stream_counter
    timestamp.sync_time = seek_time;//Time(0);       // and the sync_time.
    seek_time = Time(0);
    timestamp.is_valid = true;     // MPEGAudioDecodeNode sets the timestamp for
                                   // each output buffer.
    first_timestamp = true;
    DEBUG_STREAM("sync_reset completed." << endl);

    return SUCCESS;
  };

  Result MPEGAudioDecodeNode::eventSeek(Interval& i, Time& t, nmm_int32& num_bytes) 
  {
    DEBUG_STREAM ("SEEK i:" << i << " t:" << t << " n:" << num_bytes << endl);
    return SUCCESS;
  }

  //from ITrackSizeImpl
  Result MPEGAudioDecodeNode::trackDuration(const Interval& i) {
    send_track_duration = true;
    track_duration = i.sec;
    return SUCCESS; //DELETE
  }
  
  Result MPEGAudioDecodeNode::trackSize(const nmm_uint64& bytes) {
    track_size = bytes;
    send_track_duration = true;
    return SUCCESS;    
  }

}


#include "nmm/nmm_suffix.hpp"
