/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "MPEGAudioEncodeNode.hpp"
#include "nmm/comm/format/Property.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"

#include "nmm/utils/thread/MutexGuard.hpp"

NMM_REGISTER_PLUGIN(NMM::MPEGAudioEncodeNode, NMM, "MPEGAudioEncodeNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  MPEGAudioEncodeNode::MPEGAudioEncodeNode(const char* name /*= "MPEGAudioEncodeNode"*/,
                                               const StreamQueue::Mode queue_mode /*= StreamQueue::MODE_SUSPEND*/,
                                               const int queue_size /*= StreamQueue::DEFAULT_SIZE*/)
    : GenericProcessorNode(name, queue_mode, queue_size), audio_codec(0x55)
  {
    filename = "";
    year = "unknown";
    genre = "unknown";
    artist = "unknown";
    title = "unknown";
    album = "unknown";
    encoding = false;
    event = DATA;
    lame_quality = 2; // default quality
  }

  MPEGAudioEncodeNode::~MPEGAudioEncodeNode() {}

  Result MPEGAudioEncodeNode::TagInfo(const string& _title, const string& _album, const string& _artist, const string& _year, const string& _genre) {
    title = _title;
    album = _album;
    year = _year;
    genre = _genre;
    artist = _artist;
    id3tag_init(gfp);
    id3tag_set_title(gfp, title.c_str());
    id3tag_set_artist(gfp, artist.c_str());
    id3tag_set_album(gfp, album.c_str());
    id3tag_set_year(gfp, year.c_str());
    int test = id3tag_set_genre(gfp,genre.c_str());
    if(test==-1)
      id3tag_set_genre(gfp,"Other");
    
    id3tag_set_comment(gfp, "encoded by MMBox");
    return SUCCESS;
  }

  Result MPEGAudioEncodeNode::startTrack(const string& name) {
    DEBUG_STREAM("MPEGAudioEncodeNode : Event startTrack " << name << endl);
    filename = name;
    event = DATA;
    return SUCCESS;
  }

  Result MPEGAudioEncodeNode::endTrack() {
    DEBUG_STREAM("MPEGAudioEncodeNode : Event end " << endl);
    setWorkingFlag(true);
    event = END;
    return DELETE;
  }

  Result MPEGAudioEncodeNode::started(const Interval&) {
    return SUCCESS;
  }
  
  Result MPEGAudioEncodeNode::finished() {
    return SUCCESS;
  }

  Result MPEGAudioEncodeNode::doInit() {
    // register all supported input formats
    Format* in_format = getOwnInputProperty()->addNewFormat("audio/raw");
    in_format->addIntParamValue(Format::sample_rate_param, 8000);
    in_format->addIntParamValue(Format::sample_rate_param, 11025);
    in_format->addIntParamValue(Format::sample_rate_param, 22050);
    in_format->addIntParamValue(Format::sample_rate_param, 44100);
    in_format->addIntParamValue(Format::sample_rate_param, 48000);
    in_format->addIntParamValue(Format::channels_param, 1);
    in_format->addIntParamValue(Format::channels_param, 2);
    in_format->addIntParamValue(Format::bitpersample_param, 16);
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::encoding_param, "linear");
    in_format->addStringParamValue(Format::sign_param, "signed");
    getOwnInputProperty()->setDefaultFormat(in_format);

    // register all supported output formats
    Format* out_format = getOwnOutputProperty()->addNewFormat("audio/mpeg");
    //    out_format->addWildcard(Format::sample_rate_param);
    out_format->addIntParamValue(Format::sample_rate_param, 8000);
    out_format->addIntParamValue(Format::sample_rate_param, 11025);
    out_format->addIntParamValue(Format::sample_rate_param, 22050);
    out_format->addIntParamValue(Format::sample_rate_param, 44100);
    out_format->addIntParamValue(Format::sample_rate_param, 48000);
    out_format->addIntParamValue(Format::channels_param, 1);
    out_format->addIntParamValue(Format::channels_param, 2);
    out_format->addIntParamValue(Format::bitpersample_param, 16);
    out_format->addIntParamValue(Format::audio_codec_param, audio_codec);
    out_format->addIntParamValue(Format::layer_param, 3);
    out_format->addWildcard(Format::bitrate_param);
    getOwnOutputProperty()->setDefaultFormat(out_format);

    //Outputformat without specified bitrate. A standard bitrate will be used
    out_format = getOwnOutputProperty()->addNewFormat("audio/mpeg");
    out_format->addIntParamValue(Format::sample_rate_param, 8000);
    out_format->addIntParamValue(Format::sample_rate_param, 11025);
    out_format->addIntParamValue(Format::sample_rate_param, 22050);
    out_format->addIntParamValue(Format::sample_rate_param, 44100);
    out_format->addIntParamValue(Format::sample_rate_param, 48000);
    out_format->addIntParamValue(Format::channels_param, 1);
    out_format->addIntParamValue(Format::channels_param, 2);
    out_format->addIntParamValue(Format::bitpersample_param, 16);
    out_format->addIntParamValue(Format::audio_codec_param, audio_codec);
    out_format->addIntParamValue(Format::layer_param, 3);

    ComputationConfig* comp_conf = new ComputationConfig;
    comp_conf->addRIntParamValue("quality",0,9);
    setComputationConfig(comp_conf);

    return SUCCESS;
  }

  Result MPEGAudioEncodeNode::doDeinit() {

    // unregister all input formats
    getOwnInputProperty()->erase();

    // unregister all output formats
    getOwnOutputProperty()->erase();

    return SUCCESS;
  }

  Result MPEGAudioEncodeNode::doInitOutput() {

    // get input format.
    // verify that it is not NULL. the node's input must be connected now, or it is a failure.
    const Format* in_format = getInputFormat();
    if (in_format == NULL) {
      ERROR_STREAM("No input format" << endl);
      return FAILURE;
    }

    // init working input property
    Format* i_format = new Format( *in_format );
    getOwnWorkingInputProperty()->addNewFormat(i_format);
    getOwnWorkingInputProperty()->setDefaultFormat(i_format);

    // get parameters.
    // make sure they exist. the user might specify a wrong format.
    if (!in_format->isParameterIncluded(Format::sample_rate_param)) { return FAILURE; }
    int sample_rate = in_format->getIntValue(Format::sample_rate_param);
    if (!in_format->isParameterIncluded(Format::channels_param)) { return FAILURE; }
    int channels = in_format->getIntValue(Format::channels_param);

    // build output format
    // the output format is defined by the input format
    // first, specify all possible samplerates
    Format* o_format = getOwnWorkingOutputProperty()->addNewFormat(Format::audio_mpeg_type);
    o_format->addIntParamValue(Format::layer_param, 3);
    //    o_format->addWildcard(Format::sample_rate_param);
    o_format->addIntParamValue(Format::sample_rate_param, 8000);
    o_format->addIntParamValue(Format::sample_rate_param, 11025);
    o_format->addIntParamValue(Format::sample_rate_param, 22050);
    o_format->addIntParamValue(Format::sample_rate_param, 44100);
    o_format->addIntParamValue(Format::sample_rate_param, 48000);
    o_format->addIntParamValue(Format::channels_param, channels);
    o_format->addIntParamValue(Format::bitpersample_param, 16);
    o_format->addIntParamValue(Format::audio_codec_param, audio_codec);
    o_format->addWildcard(Format::bitrate_param);

    // fully specified output format, also set as default format
    o_format = getOwnWorkingOutputProperty()->addNewFormat(Format::audio_mpeg_type);
    o_format->addIntParamValue(Format::layer_param, 3);
    o_format->addIntParamValue(Format::sample_rate_param, sample_rate);
    o_format->addIntParamValue(Format::channels_param, channels);
    o_format->addIntParamValue(Format::bitpersample_param, 16);
    o_format->addIntParamValue(Format::audio_codec_param, audio_codec);
    o_format->addWildcard(Format::bitrate_param);
    getOwnWorkingOutputProperty()->setDefaultFormat(o_format);
    
    //specified format without bitrate
    o_format = getOwnWorkingOutputProperty()->addNewFormat(Format::audio_mpeg_type);
    o_format->addIntParamValue(Format::layer_param, 3);
    o_format->addIntParamValue(Format::sample_rate_param, sample_rate);
    o_format->addIntParamValue(Format::channels_param, channels);
    o_format->addIntParamValue(Format::bitpersample_param, 16);
    o_format->addIntParamValue(Format::audio_codec_param, audio_codec);

    // set I/O partners
    i_format->setIOPartner(o_format);
    o_format->setIOPartner(i_format);

    // specify the computation configuration
    ComputationConfig* work_comp_conf = new ComputationConfig;
    work_comp_conf->addIntParamValue("quality",lame_quality); // the 'quality' has an heavy impact on the needed cpu-time
    setWorkingComputationConfig(work_comp_conf);


    return SUCCESS;
  }

  Result MPEGAudioEncodeNode::doDeinitOutput() {
    getOwnWorkingInputProperty()->erase();
    getOwnWorkingOutputProperty()->erase();
    return SUCCESS;
  }

  Result MPEGAudioEncodeNode::doActivate() {

    // init external encoder library.
    // we can do this here because it is not needed by the registry.
    // however, if we had to init the library to query for supported formats,
    // then we would have to do this in initInput.
    DEBUG_STREAM("Init lame" << endl);
    if ( NULL == (gfp=lame_init()) ) {
      ERROR_STREAM("Fatal error during initialization of lame" << endl);
      return FAILURE;
    }

    // get formats.
    // both formats should be known here, but we have to check it because the user might
    // call this before connecting the output.
    const Format *in_format = getInputFormat(); 
    const Format *out_format = getOutputFormat();
    if (in_format == NULL) {
      ERROR_STREAM("in_format is NULL!" << endl);
      return FAILURE;
    }
    if (out_format == NULL) {
      ERROR_STREAM("out_format is NULL!" << endl);
      return FAILURE;
    }
    
    
    /* --- setting Formats ------------------------------- */
    
    /* Input Format */
    DEBUG_STREAM("Set samplerate to " << in_format->getIntValue(Format::sample_rate_param) << endl);
    lame_set_in_samplerate(gfp, in_format->getIntValue(Format::sample_rate_param));
    
    channels = in_format->getIntValue(Format::channels_param);
    DEBUG_STREAM("Set channels number to " << channels << endl);    
    lame_set_num_channels(gfp, channels);  

    /* let lame do */
    lame_set_out_samplerate(gfp,0);

    /* Output Format */
    DEBUG_STREAM("Set OUTPUT samplerate to " << out_format->getIntValue(Format::sample_rate_param) << endl);
    lame_set_out_samplerate(gfp, out_format->getIntValue(Format::sample_rate_param));
    
    int bitrate;
    if (out_format->isParameterIncluded(Format::bitrate_param))
      bitrate = out_format->getIntValue(Format::bitrate_param);
    else
      bitrate = 128000;
    DEBUG_STREAM("Set bitrate to " << bitrate << " bit/s" << endl);
    lame_set_brate(gfp,bitrate / 1000);

    DEBUG_STREAM("Set quality to " << lame_quality << endl);
    lame_set_quality(gfp, lame_quality);


    /* VBR Header off */
    lame_set_bWriteVbrTag(gfp, 0);

    /* Init - Part 2 */
    DEBUG_STREAM("Init lame parameter" << endl);
    if (0 > lame_init_params(gfp) ) {
      ERROR_STREAM("Error in doActivate" << endl);
      return FAILURE;
    }

    return SUCCESS;
  }
  
  Result MPEGAudioEncodeNode::doDeactivate() {
    return SUCCESS;
  }
  
  
  Message* MPEGAudioEncodeNode::processBuffer(Buffer* in_buffer) {
    MutexGuard mae_mg(mae_mutex);
    int mp3_buffer_size = 0;
    int mp3_size = 0;
    Buffer *out_buffer;

    /* which event */
    switch(event) {
    case TEST:
      ERROR_STREAM("Ich darf nicht aufgerufen werden" << endl);
      return NULL;
      break;

    case START:
      ERROR_STREAM("Ich darf nicht aufgerufen werden" << endl);
      event = DATA;
      return NULL;
      break;
    
    case FINISHED:
      // now node is really finished and will return end track
      setWorkingFlag(false);
      
      return new CEvent(Event(ITrack::endTrack_event));
      break;

    case END:
      /* get the rest data from lame, put it in a new buffer, and send it  */
      mp3_buffer_size = 10000;
      out_buffer = getNewBuffer(mp3_buffer_size);

      if (out_buffer) {
        mp3_size = lame_encode_flush(gfp ,(unsigned char *)(out_buffer->getData()), mp3_buffer_size);
        
        if (0 > mp3_size) { 
          ERROR_STREAM ("Buffer too small, mp3 data will be corupt" << endl);          
        } else {
          out_buffer->setUsed(mp3_size);
        }

      } else {
        ERROR_STREAM("Can\'t get new buffer, mp3 data will be corupt" << endl); 
      }

      // set mode from END to FINISHED (working-flag is still true)
      event = FINISHED;

      DEBUG_STREAM("Flush " << out_buffer->getUsed() << endl);
      return out_buffer;
      break;

    case DATA:
      /* mp3_buffer_size = (int) ((in_buffer->getUsed() / 4) * 1.25) + 7200;  */
      mp3_buffer_size = in_buffer->getUsed();
      out_buffer = getNewBuffer(mp3_buffer_size); 
      
      /* fw: may be wrong due to latencies in the encoder but better than nothing */
      out_buffer->setTimestamp(in_buffer->getTimestamp());
      
      if (out_buffer) {
        if (2 == channels) {
          mp3_size = lame_encode_buffer_interleaved (gfp, 
                                                     (short int *)(in_buffer->getData()), 
                                                     in_buffer->getUsed() / 4, 
                                                     (unsigned char *)(out_buffer->getData()), 
                                                     mp3_buffer_size );
        } else {
          mp3_size = lame_encode_buffer(gfp, 
                                        (short int *)(in_buffer->getData()), 
                                        (short int *)(in_buffer->getData()), 
                                        in_buffer->getUsed() / 2, 
                                        (unsigned char *)(out_buffer->getData()), 
                                        mp3_buffer_size );     
        }
        if (-1 == mp3_size) { 
          ERROR_STREAM ("Buffer too small, mp3 data will be corupt " << endl);          
        } else {
          DEBUG_STREAM ("Encoded " << in_buffer->getUsed() << " bytes of audio to " << mp3_size << " bytes of mp3" << endl);
          out_buffer->setUsed(mp3_size);

        }
      } else {
        ERROR_STREAM ("Can\'t get new buffer!" << endl);          
        out_buffer = NULL;
      }
      //      cerr << "Encode " << in_buffer->getUsed() << "\t -> \t" << out_buffer->getUsed() << endl;

      in_buffer->release(); /* delete input data */
      
      // This has to be done because out_buffer->setUsed(0) does nothing, which is okay
      if (mp3_size > 0) {
	DEBUG_STREAM("Buffer requested and returned." << endl);
	return out_buffer;
      }
      else {
	out_buffer->release();
	DEBUG_STREAM("Buffer requested but not returned -> released." << endl);
	return NULL;
      }

      break;
      
    default:
      ERROR_STREAM ("No internal state found!" << endl);
      in_buffer->release();
      exit(1);
    }

    ERROR_STREAM ("Ich darf nicht erscheinen!!!" << endl);
    return NULL;
  } /* processBuffer */

  void MPEGAudioEncodeNode::setQuality(const nmm_float& q) {
    if ((q > 0) && (q < 9.1)) {
      lame_quality = (int) q;
    } else {
      ERROR_STREAM("Quality out of range (" << q << " not in [0...9]" << endl);
    }
  }

  nmm_float MPEGAudioEncodeNode::getQuality() {
    return ((float) lame_quality);
  }

  void MPEGAudioEncodeNode::setBitrate(const nmm_int32&) {
    DEBUG_STREAM("not implemented" << endl);
    return;
  }

  nmm_int32 MPEGAudioEncodeNode::getBitrate() {
    return (getOutputFormat()->getIntValue(Format::bitrate_param) / 1000);
  }


#include "nmm/nmm_prefix.hpp"

} /* namespace NMM */


#include "nmm/nmm_suffix.hpp"
