/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <pthread.h>
#include "AudioMuxNode.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"
#include "nmm/utils/thread/MutexGuard.hpp"

NMM_REGISTER_PLUGIN(AudioMuxNode, NMM, "AudioMuxNode")
    
#include "nmm/nmm_prefix.hpp"

namespace NMM {
	
  AudioMuxNode::AudioMuxNode(const char *name, StreamQueue::Mode qmode, int qsize) 
    : GenericMultiplexerNode(name, qmode, qsize)
    , in_buffer1(0)
    , in_buffer2(0)
    , volume1(80)
    , volume2(0)
    , received_first_buffer(false)
  {
    addInputStream("input1");
    addInputStream("input2");
  }
	
  AudioMuxNode::~AudioMuxNode() 
  {
  }

  Result AudioMuxNode::doInit()
  {
    // add new input format for input1 to property output format
    Format* in_format1 = getOwnInputProperty()->addNewFormat("audio/raw");
    in_format1->setJackTag("input1");

    in_format1->addWildcard(Format::sample_rate_param);
    in_format1->addIntParamValue(Format::channels_param, 2);
    in_format1->addIntParamValue(Format::bitpersample_param, 16);
    in_format1->addStringParamValue(Format::sign_param, "signed");
    in_format1->addStringParamValue(Format::endian_param, "little");
    in_format1->addStringParamValue(Format::encoding_param, "linear");
    getOwnInputProperty()->setDefaultFormat(in_format1, "input1");

    // add new input format for input1 to property output format
    Format* in_format2 = getOwnInputProperty()->addNewFormat("audio/raw");
    in_format2->setJackTag("input2");

    in_format2->addWildcard(Format::sample_rate_param);
    in_format2->addIntParamValue(Format::channels_param, 2);
    in_format2->addIntParamValue(Format::bitpersample_param, 16);
    in_format2->addStringParamValue(Format::sign_param, "signed");
    in_format2->addStringParamValue(Format::endian_param, "little");
    in_format2->addStringParamValue(Format::encoding_param, "linear");
    getOwnInputProperty()->setDefaultFormat(in_format2, "input2");

    // add new output format to property output format
    Format* out_format = getOwnOutputProperty()->addNewFormat("audio/raw");

    out_format->addWildcard(Format::sample_rate_param);
    out_format->addIntParamValue(Format::channels_param, 2);
    out_format->addIntParamValue(Format::bitpersample_param, 16);
    out_format->addStringParamValue(Format::sign_param, "signed");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::encoding_param, "linear");
    getOwnOutputProperty()->setDefaultFormat(out_format);
    
    // set I/O partners of formats
    in_format1->setIOPartner(out_format);
    out_format->setIOPartner(in_format1);

    in_format2->setIOPartner(out_format);
    out_format->setIOPartner(in_format2);
	    
    return SUCCESS;
  }
	
  Result AudioMuxNode::doInitOutput() 
  {
    // query the format currently set on input1 and add to working input property
    const Format* current_in_format1 = getInputFormat("input1");
    if (current_in_format1 == NULL) { 
      ERROR_STREAM("No input format on input1" << endl);
      return FAILURE; 
    }
	    
    Format* in_format1 = new Format(*current_in_format1);
    in_format1->setJackTag("input1");
    getOwnWorkingInputProperty()->addNewFormat(in_format1);
    getOwnWorkingInputProperty()->setDefaultFormat(in_format1, "input1");

    // query the format currently set on input2 and add to working input property
    const Format* current_in_format2 = getInputFormat("input2");
    if (current_in_format2 == NULL) { 
      ERROR_STREAM("No input format on input2" << endl);
      return FAILURE; 
    }
	    
    Format* in_format2 = new Format(*current_in_format2);
    in_format2->setJackTag("input2");
    getOwnWorkingInputProperty()->addNewFormat(in_format2);
    getOwnWorkingInputProperty()->setDefaultFormat(in_format2, "input2");

    // add the current format set on input1 to working output property
    Format* out_format = new Format(*current_in_format1);
    getOwnWorkingOutputProperty()->addNewFormat(out_format);
    getOwnWorkingOutputProperty()->setDefaultFormat(out_format);

    // set I/O partners of formats
    in_format1->setIOPartner(out_format);
    out_format->setIOPartner(in_format1);

    in_format2->setIOPartner(out_format);
    out_format->setIOPartner(in_format2);

    return SUCCESS;
  }

  Result AudioMuxNode::syncReset() 
  {
    // always delete sync reset from 2nd input
    if(!strcmp(getCurrentRecvInputStream(), "input2")) {
      return DELETE;
    }
    else {
      return SUCCESS;
    }
  }

  Time AudioMuxNode::getCurrentSyncTime()
  {
    MutexGuard mg(mutex);
    if(current_timestamp.is_set) {
      return current_timestamp.sync_time;
    }
    else {
      return Time();
    }
  }

  void AudioMuxNode::setVolumeControl(const string& input_jack_tag, const Time& start_time, const Time& end_time, const nmm_int32& start_value, const nmm_int32& end_value)
  {
    MutexGuard mg(mutex);
    if(input_jack_tag == "input1") {
      volume_interpolator1.setStartTime(start_time);
      volume_interpolator1.setEndTime(end_time); 
      volume_interpolator1.setStartValue(start_value);
      volume_interpolator1.setEndValue(end_value);
    }
    else if(input_jack_tag == "input2") {
      volume_interpolator2.setStartTime(start_time);
      volume_interpolator2.setEndTime(end_time); 
      volume_interpolator2.setStartValue(start_value);
      volume_interpolator2.setEndValue(end_value);
    }
    else {
      WARNING_STREAM("Unknown input_jack_tag " << input_jack_tag << " in method setVolumeInterpolation" << endl);
    }
  }

  nmm_int32 AudioMuxNode::setVolume(const string& input_jack_tag, const nmm_int32& volume)
  {
    MutexGuard mg(mutex);
    if(input_jack_tag == "input1") {
      return (volume1 = volume);
    }
    else if(input_jack_tag == "input2") {
      return (volume2 = volume);
    }
    else {
      WARNING_STREAM("Unknown input_jack_tag " << input_jack_tag << " in method getVolume" << endl);
      return 0;
    }
  }

  nmm_int32 AudioMuxNode::getVolume(const string& input_jack_tag) 
  {
    MutexGuard mg(mutex);
    if(input_jack_tag == "input1") {
      return volume1;
    }
    else if(input_jack_tag == "input2") {
      return volume2;
    }
    else {
      WARNING_STREAM("Unknown input_jack_tag " << input_jack_tag << " in method getVolume" << endl);
      return 0;
    }
  }

  Message* AudioMuxNode::processBuffer(Buffer* buffer) 
  {
    MutexGuard mg(mutex);
    Buffer* out_buffer = 0;
	  
    // collect buffer in a round-robin approach: input1, input2, input1, ..
    if(!strcmp(getCurrentRecvInputStream(), "input1")) {
      // disable input1, enable input2
      setRecvInputStreamEnabled("input1", false);
      setRecvInputStreamEnabled("input2", true);
	    
      in_buffer1 = buffer;

      // save latest valid and set timestamp of incoming stream on input 1 for method getCurrentTimestamp
      if(in_buffer1->getTimestamp().is_valid && in_buffer1->getTimestamp().is_set)
      {
	current_timestamp = in_buffer1->getTimestamp();
      }
    }
    else if(!strcmp(getCurrentRecvInputStream(), "input2")) {
      // disable input2, enable input1
      setRecvInputStreamEnabled("input2", false);
      setRecvInputStreamEnabled("input1", true);
	    
      in_buffer2 = buffer;
    }

    // if a frame from input1 and a frame from input2 is available
    // -> perform scaling/merging/muxing operation
    if(in_buffer1 && in_buffer2) {

      // set volume for input1 and input2 according to interpolator (both using timestamp of in_buffer1)
      if(volume_interpolator1.isConfigured()) {
	volume1 = static_cast<int32_t>( volume_interpolator1.getCurrentValue(in_buffer1->getTimestamp().sync_time) );
      }
      if(volume_interpolator2.isConfigured()) {
	volume2 = static_cast<int32_t>( volume_interpolator2.getCurrentValue(in_buffer1->getTimestamp().sync_time) );
      }

      // request new buffer for outgoing buffer
      // (always use size and timestamp of in_buffer1 received from first input)
      out_buffer = getNewBuffer(in_buffer1->getUsed(), in_buffer1);
      out_buffer->setUsed(in_buffer1->getUsed());

      // copy data from in_buffer1 to out_buffer
      int16_t* in_data_ptr1 = (int16_t*) (in_buffer1->getData());
      int16_t* in_data_ptr2 = (int16_t*) (in_buffer2->getData());
      int16_t* out_data_ptr = (int16_t*) (out_buffer->getData());

      int32_t tmp_data1;
      int32_t tmp_data2;

      for (size_t i = 0; i < in_buffer1->getUsed(); i+=2 ) {
	
	tmp_data1 = ( static_cast<int32_t>(*in_data_ptr1) * volume1 ) / 100;
	tmp_data2 = ( static_cast<int32_t>(*in_data_ptr2) * volume2 ) / 100;

	*out_data_ptr = (tmp_data1 + tmp_data2)/2;

	++out_data_ptr;
	++in_data_ptr1;
	++in_data_ptr2;
      }

      // release incoming buffers
      in_buffer1->release();
      in_buffer2->release();

      in_buffer1 = 0;
      in_buffer2 = 0;
    }

    // cerr << "out_buffer " << (out_buffer != 0) << endl;

    return out_buffer;
  }


};

#include "nmm/nmm_suffix.hpp"
