/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include "OggVorbisDecodeNode.hpp"
#include "nmm/plugins/interfaces/ITagInfo.hpp"
#include "nmm/multimedia/interfaces/ITrackDuration.hpp"

//#include "nmm/base/format/Property.hpp"
//#include "nmm/base/registry/Plugin.hpp"

#include "nmm/plugins/NMMRegisterPlugin.hpp"

NMM_REGISTER_PLUGIN(OggVorbisDecodeNode, NMM, "OggVorbisDecodeNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  OggVorbisDecodeNode::OggVorbisDecodeNode(const char* name /* = "OggVorbisDecodeNode" */, 
					   StreamQueue::Mode mode /* = StreamQueue::MODE_SUSPEND */, 
					   int queue_size /* = 20 */)
    : GenericProcessorNode(name, mode, queue_size),
      read_buffer(NULL),
      bytes_read(0),
      ov_initialized(false),
      merge_buffer(NULL),
      oggvorbis(),
      out_buffer_size(8192),
      info_user_comments(""),
      info_vendor(""),
      timestamp(), 
      in_timestamp(),
      old_pos(0),
      sample_rate(0),
      channels(0),
      send_track_duration(false)
{
    timestamp.stream_counter = 0;
    timestamp.sync_time = Time(0);
    timestamp.is_valid = true;

    in_timestamp.stream_counter = 0;
    in_timestamp.sync_time = Time(0);
    in_timestamp.is_valid = true;

    
  }

  OggVorbisDecodeNode::~OggVorbisDecodeNode() {

  }

  Result OggVorbisDecodeNode::doInit() {

    Format* in_format = getOwnInputProperty()->addNewFormat(Format::audio_x_ogg_type);
    in_format->setJackTag("default");
    getOwnInputProperty()->setDefaultFormat(in_format, "default");

    Format* out_format = getOwnOutputProperty()->addNewFormat("audio/raw");
    out_format->addIntParamValue(Format::sample_rate_param, 44100);
    out_format->addIntParamValue(Format::sample_rate_param, 48000);
    out_format->addIntParamValue(Format::channels_param, 1);
    out_format->addIntParamValue(Format::channels_param, 2);
    out_format->addIntParamValue(Format::bitpersample_param, 16);
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::encoding_param, "linear");
    out_format->addStringParamValue(Format::sign_param, "signed");
    getOwnOutputProperty()->setDefaultFormat(out_format);

    in_format->setIOPartner(out_format);
    out_format->setIOPartner(in_format);

    return SUCCESS;
  }

  Result OggVorbisDecodeNode::doInitOutput() {

    const Format* in_format = getInputFormat();

    if (!in_format) {
      GLOBAL_ERROR_STREAM("InputFormat is NULL!" << endl);
      return FAILURE;
    }

    if (!analyseData()) {
      ERROR_STREAM("Could not get data for analysis!" << endl);
      return FAILURE;
    }

    Format* myInputFormat = new Format(*in_format);
    getOwnWorkingInputProperty()->addNewFormat(myInputFormat);
    getOwnWorkingInputProperty()->setDefaultFormat(myInputFormat);

    Format* myout_format = getOwnWorkingOutputProperty()->addNewFormat("audio/raw");
    myout_format->addIntParamValue(Format::sample_rate_param, sample_rate);
    myout_format->addIntParamValue(Format::channels_param, channels);
    myout_format->addIntParamValue(Format::bitpersample_param, 16);  // fixed by the decoder!
    myout_format->addStringParamValue(Format::encoding_param, "linear");
    myout_format->addStringParamValue(Format::sign_param, "signed");
    myout_format->addStringParamValue(Format::endian_param, "little");
    getOwnWorkingOutputProperty()->setDefaultFormat(myout_format);

    myInputFormat->setIOPartner(myout_format);
    myout_format->setIOPartner(myInputFormat);

    return SUCCESS;
  }

  size_t OggVorbisDecodeNode::readFunc(void* ptr, size_t /*size*/, size_t size) {
    if (NULL == read_buffer) {
      return 0;
    }

    const size_t bytes_available = (read_buffer->getUsed() - bytes_read);
    if (size < bytes_available) {
      DEBUG_STREAM("Reading " << size << "Bytes!" << endl);
      memcpy(ptr, (read_buffer->getData() + bytes_read), size);
      bytes_read += size;
      DEBUG_STREAM(read_buffer->getUsed() - bytes_read << "Bytes left." << endl);
      return size;
    }

    DEBUG_STREAM("Reading " << bytes_available << "Bytes!" << endl);
    memcpy(ptr, (read_buffer->getData() + bytes_read), bytes_available);
    DEBUG_STREAM("Nothing left, releasing Buffer!" << endl);
    read_buffer->release();
    read_buffer = NULL;
    bytes_read += bytes_available;
    setWorkingFlag(false);
    return bytes_available;
  }

  int OggVorbisDecodeNode::seekFunc(ogg_int64_t, int) {
    return -1; // no seek implemented!
  }

  int OggVorbisDecodeNode::closeFunc() {
    return 0; // close should work
  }

  long int OggVorbisDecodeNode::tellFunc() {
    return -1; // error, but as no seek implemented, this should never happen!
  }

  bool OggVorbisDecodeNode::initOggVorbis() {
    ov_callbacks ov;
    ov.read_func  = &OggVorbisDecodeNode::readFuncOggStruct;
    ov.seek_func  = &OggVorbisDecodeNode::seekFuncOggStruct;
    ov.close_func = &OggVorbisDecodeNode::closeFuncOggStruct;
    ov.tell_func  = &OggVorbisDecodeNode::tellFuncOggStruct;

    if (NULL == merge_buffer) {
      merge_buffer = getNewBuffer(8192);
      if (NULL == merge_buffer) {
	ERROR_STREAM("Ey man, what's up. No Buffers abailable, when starting to initialize on incoming OggVorbis stream? Could not get merge_buffer!" << endl);
	return false;
      }
      merge_buffer->setUsed(0);
    }
    if ((merge_buffer->getUsed() + read_buffer->getUsed()) < 4096) {
      if ((merge_buffer->getSize() - merge_buffer->getUsed()) < read_buffer->getUsed()) {
	Buffer* tmp_buffer = getNewBuffer(merge_buffer->getUsed() + read_buffer->getUsed());
	if (NULL == tmp_buffer) {
	  ERROR_STREAM("Ey man, what's up. No Buffers abailable, when starting to initialize on incoming OggVorbis stream? Could not get tmp_buffer!" << endl);
	  return false;
	}
	memcpy(tmp_buffer->getData(), merge_buffer->getData(), merge_buffer->getUsed());
	tmp_buffer->setUsed(merge_buffer->getUsed());
	merge_buffer->release();
	merge_buffer = tmp_buffer;
      }
    
      memcpy(merge_buffer->getData()+merge_buffer->getUsed(), read_buffer->getData(), read_buffer->getUsed());
      merge_buffer->setUsed(merge_buffer->getUsed() + read_buffer->getUsed());
      read_buffer->release();
      read_buffer = NULL;
      setWorkingFlag(false);
      DEBUG_STREAM("Not enough Data to init OggStream, waiting for next buffer!" << endl);
      return false;
    }

    int ret = ov_open_callbacks((void*) this, &oggvorbis, merge_buffer->getData(), merge_buffer->getUsed(), ov);
    switch (ret) {
    case OV_EREAD:
      DEBUG_STREAM("Stream not jet opened: A read from media returned an error." << endl);
      return false;
      break;
    case OV_ENOTVORBIS: 
      DEBUG_STREAM("Stream not jet opened: Bitstream is not Vorbis data." << endl);
      return false;
      break;
    case OV_EVERSION: 
      DEBUG_STREAM("Stream not jet opened: Vorbis version mismatch." << endl);
      return false;
      break;
    case OV_EBADHEADER: 
      DEBUG_STREAM("Stream not jet opened: Invalid Vorbis bitstream header." << endl);
      return false;
      break;
    case OV_EFAULT: 
      DEBUG_STREAM("Stream not jet opened: Internal logic fault; indicates a bug or heap/stack corruption." << endl);
      return false;
      break;
    case 0:
      merge_buffer = NULL;
      ov_initialized = true;
      break;
    default:
      ERROR_STREAM("Unknown error on stream opening" << endl);
      return false;
      break;
    }

    return true;
  }


  Result OggVorbisDecodeNode::doDeactivate() {
    if (ov_initialized) {
      ov_clear(&oggvorbis);
    }
    ov_initialized = false;
    info_user_comments = string("");
    info_vendor = string("");

    return SUCCESS;
  }


  Message* OggVorbisDecodeNode::processBuffer(Buffer* in_buffer) {
    Buffer* out_buffer = NULL;
    int read_bytes = 0;

    if (! getWorkingFlag()) {
      setWorkingFlag(true);
      read_buffer = in_buffer;
      bytes_read = 0;
      Timestamp tmp_ts = in_buffer->getTimestamp();
      if (tmp_ts.is_valid) {
        in_timestamp.sync_time = tmp_ts.sync_time;
        in_timestamp.stream_counter = tmp_ts.stream_counter;
        in_timestamp.is_valid = true;
      }
      DEBUG_STREAM("Got new buffer: " << read_buffer->getUsed() << "Bytes." << endl);
    }
      
    if (!ov_initialized) {
      if (! initOggVorbis()) {
        return NULL;
      }
    }

    out_buffer = getNewBuffer(out_buffer_size);
    if (NULL == out_buffer) {
      ERROR_STREAM("Could not get a new buffer!" << endl);
      return NULL;
    }
      
    int bitstream = 0;
    read_bytes = ov_read(&oggvorbis, out_buffer->getData(), out_buffer->getSize(), 0, 2, 1, &bitstream);

    if (getAnalysingFlag()) {
      // get the vorbis_info struct with the format values
      vorbis_info* vorbisinfo = ov_info(&oggvorbis, bitstream);
      // read needed sample_rate and channels
      sample_rate = vorbisinfo->rate;
      channels = vorbisinfo->channels;

      // get vorbis_comment struct with meta informations
      vorbis_comment* vorbiscomment = ov_comment(&oggvorbis, bitstream);
      // read out the comments and vendor
      info_user_comments = string("");
      for (int i = 0; i < vorbiscomment->comments; i++) {
        if (0 != i) {
          info_user_comments += string("\n");
        }
        info_user_comments += string(vorbiscomment->user_comments[i]);
      }
      info_vendor = string(vorbiscomment->vendor);

      //parsing user_comments string to generate TrackInfo
      int equal,newline;
      if (info_user_comments.find("TITLE")!=string::npos) {
        equal = info_user_comments.find("=", info_user_comments.find("TITLE"));
        newline = info_user_comments.find("\n",equal);
        title = info_user_comments.substr(equal+1, newline-equal-1);
      } else title="unknown";
      if (info_user_comments.find("ALBUM")!=string::npos) {
        equal = info_user_comments.find("=", info_user_comments.find("ALBUM"));
        newline = info_user_comments.find("\n",equal);
        album = info_user_comments.substr(equal+1, newline-equal-1);
      } else album="unknown";
      if (info_user_comments.find("ARTIST")!=string::npos) {
        equal = info_user_comments.find("=", info_user_comments.find("ARTIST"));
        newline = info_user_comments.find("\n",equal);
        artist = info_user_comments.substr(equal+1, newline-equal-1);
      } else artist="unknown";
      if (info_user_comments.find("DATE")!=string::npos) {
        equal = info_user_comments.find("=", info_user_comments.find("DATE"));
        newline = info_user_comments.find("\n",equal);
        year = info_user_comments.substr(equal+1, newline-equal-1);
      } else year = "unknown";
      if (info_user_comments.find("GENRE")!=string::npos) {
        equal = info_user_comments.find("=", info_user_comments.find("GENRE"));
        newline = info_user_comments.find("\n",equal);
        genre = info_user_comments.substr(equal+1, newline-equal-1);
      } else genre="unknown";

      setAnalysingFlag(false);
      CEvent* cevent = new CEvent();

      cevent->insertEvent(ITagInfo::create_TagInfo(title,album,artist,year,genre));      

      if (send_track_duration) {
	send_track_duration = false;
	u_int64_t track_length = (track_size - read_bytes) * 8 * 1000 / ov_bitrate(&oggvorbis,-1);
	Interval track_duration = Interval(track_length / 1000, (track_length - track_length / 1000 * 1000) * 1000);
	cevent->insertEvent(ITrackDuration::create_trackDuration(track_duration));
	DEBUG_STREAM("sending trackDuration: " << track_duration << endl);
      }

      return cevent;
      //return NULL;
    }

    switch (read_bytes) {
    case OV_HOLE:
      DEBUG_STREAM("Indicates there was an interruption in the data. (one of: garbage between pages, loss of sync followed by recapture, or a corrupt page)" << endl);
      out_buffer->release();
      return NULL;
      break;
    case OV_EBADLINK:
      DEBUG_STREAM("Indicates that an invalid stream section was supplied to libvorbisfile, or the requested link is corrupt." << endl);
      out_buffer->release();
      return NULL;
      break;
    case 0:
      DEBUG_STREAM("Stream has reached EOF!" << endl);
      out_buffer->release();
      return NULL;
      break;
    default:
      break;
    }
    out_buffer->setUsed(read_bytes);

    long int audio_length = (long int) (((int64_t) (read_bytes * 8 /*bit per byte*/))
					* (int64_t) nano_per_sec
					/ (int64_t) (16 /*bits per sample*/ 
						     * 2 /*channels*/ 
						     * sample_rate));
    if (in_timestamp.is_valid) {
      timestamp.sync_time = in_timestamp.sync_time;
      timestamp.stream_counter = in_timestamp.stream_counter;
      in_timestamp.is_valid = false;
    } else {
      timestamp.sync_time += audio_length;
    }
    timestamp.is_valid = true;
    out_buffer->setTimestamp(timestamp);
    timestamp.stream_counter++;

    return out_buffer;
  }

  string OggVorbisDecodeNode::getUserComments() {
    return info_user_comments;
  }

  string OggVorbisDecodeNode::getVendor() {
    return info_vendor;
  }

  Result OggVorbisDecodeNode::syncReset() {
    timestamp.stream_counter = 0;
    timestamp.sync_time = Time(0);
    timestamp.is_valid = true;

    return SUCCESS;
  }

  Result OggVorbisDecodeNode::trackSize(const nmm_uint64& bytes) {
    track_size = bytes;
    send_track_duration = true;
    return DELETE;
  }
}

#include "nmm/nmm_suffix.hpp"
