/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "PlaybackNode.hpp"
#include "nmm/comm/ProxyObject.hpp"
#include "nmm/comm/messaging/TEDObject.hpp"
#include "nmm/comm/serialize/SerializeResult.hpp"
#include "nmm/comm/serialize/SerializeUtilsTime.hpp"
#include "nmm/utils/SecurityCheck.hpp"
#include "nmm/utils/time/Rational.hpp"


//#include "nmm/utils/utils.hpp"
//#include "nmm/utils/SecurityCheck.hpp"
//#include "nmm/base/ProxyObject.hpp"
//#include "nmm/base/registry/Plugin.hpp"

#include <sys/soundcard.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <iostream>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>

#ifndef __CYGWIN__
static NMM::PlaybackNodePlugin plugin("PlaybackNode",-1);
NMM_REGISTER_SPECIAL_PLUGIN(PlaybackNode, NMM, plugin)
#endif
#ifdef __CYGWIN__
NMM_REGISTER_PLUGIN(PlaybackNode, NMM, "PlaybackNode")
#endif

#if (__GNUC__ < 3)
/* normally the swab function is declared in <unistd.h>, but as the compiler can't find it,
I redeclared it here, the linker is satisfied with this */
extern "C" {
  void swab(const void *from, void *to, size_t n);
}
#endif


#include "nmm/nmm_prefix.hpp"

namespace NMM {
  //! Formats to check in the doInit method.
  const int PlaybackNode::test_formats[] = {
    AFMT_AC3,     //!< AC3 ecoded audio
    AFMT_MU_LAW,  //!< Logarithmic mu-law audio encoding
    AFMT_A_LAW,   //!< Logarithmic A-law audio encoding
    AFMT_U8,      //!< Linear encoding, unsigned 8-bit
    AFMT_S8,      //!< Linear encoding, signed 8-bit
    AFMT_U16_LE,  //!< Linear encoding, unsigned 16-bit, little endian
    AFMT_U16_BE,  //!< Linear encoding, unsigned 16-bit, big endian
    AFMT_S16_LE,  //!< Linear encoding, signed 16-bit, little endian
    AFMT_S16_BE,  //!< Linear encoding, signed 16-bit, big endian
    0
    //!< End marker. If you add or remove some formats, make sure that the last one remains 0.
  };

  //! Channel numbers to check in the doInit method.
  const int PlaybackNode::test_channels[] = {
    1,      //!< Mono
    2,      //!< Stereo
    4,      //!< 4
    5,      //!< 5
    6,      //!< 5.1
    0
    //!< End marker. If you add or remove some values, make sure that the last one remains 0.
  };

  //! Sample rates to check in the doInit method.
  const int PlaybackNode::test_sample_rates[] = { 8000, 11025, 16000, 22050, 32000, 44100, 48000, 96000, 0 };

  Format& PlaybackNode::getNMMFormat(const int oss_format, Format& nmm_format)
  {
    switch(oss_format) {
    case AFMT_MU_LAW:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "mu_law");
      nmm_format.addIntParamValue(Format::bitpersample_param, 8);
      break;

    case AFMT_A_LAW:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "a_law");
      break;

    case AFMT_U8:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "linear");
      nmm_format.addStringParamValue(Format::sign_param, "unsigned");
      nmm_format.addIntParamValue(Format::bitpersample_param, 8);
      break;

    case AFMT_S8:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "linear");
      nmm_format.addStringParamValue(Format::sign_param, "signed");
      nmm_format.addIntParamValue(Format::bitpersample_param, 8);
      break;

    case AFMT_U16_LE:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "linear");
      nmm_format.addStringParamValue(Format::sign_param, "unsigned");
      nmm_format.addIntParamValue(Format::bitpersample_param, 16);
      nmm_format.addStringParamValue(Format::endian_param, "little");
      break;

    case AFMT_U16_BE:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "linear");
      nmm_format.addStringParamValue(Format::sign_param, "unsigned");
      nmm_format.addIntParamValue(Format::bitpersample_param, 16);
      nmm_format.addStringParamValue(Format::endian_param, "big");
      break;

    case AFMT_S16_LE:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "linear");
      nmm_format.addStringParamValue(Format::sign_param, "signed");
      nmm_format.addIntParamValue(Format::bitpersample_param, 16);
      nmm_format.addStringParamValue(Format::endian_param, "little");
      break;

    case AFMT_S16_BE:
      nmm_format.setFormatType("audio/raw");
      nmm_format.addStringParamValue(Format::encoding_param, "linear");
      nmm_format.addStringParamValue(Format::sign_param, "signed");
      nmm_format.addIntParamValue(Format::bitpersample_param, 16);
      nmm_format.addStringParamValue(Format::endian_param, "big");
      break;
      
    case AFMT_AC3:
      nmm_format.setFormatType("audio/ac3");
      break;
      
    default:
      ERROR_STREAM("getNMMFormat: OSS format " << oss_format 
		   << " not supported." << endl);
    }

    return nmm_format;
  };

  int PlaybackNode::getOSSFormat(const Format &nmm_format)
  {
    string encoding;
    try {
      encoding = nmm_format.getStringValue(Format::encoding_param);
    } catch (FormatException e) {
      ERROR_STREAM("getOSSFormat: Incomplete format (encoding not specified)."
		   << endl);
      return 0;
    }

    if(encoding == "mu_law") return AFMT_MU_LAW;
    if(encoding == "a_law" ) return AFMT_A_LAW;
    if(encoding == "linear") {
      int bits;
      try {
	bits = nmm_format.getIntValue(Format::bitpersample_param);
      } catch (FormatException e) {
	ERROR_STREAM("getOSSFormat: Incomplete format (bitpersample not specified)." << endl);
	return 0;
      }
      string sign;
      try {
	sign = nmm_format.getStringValue(Format::sign_param);
      } catch (FormatException e) {
	ERROR_STREAM("getOSSFormat: Incomplete format (sign not specified)." << endl);
	return 0;
      }

      if(bits == 8) {
	if(sign == "unsigned") return AFMT_U8;
	if(sign == "signed"  ) return AFMT_S8;
	ERROR_STREAM("getOSSFormat: Incorrect format: sign is [" << sign << "]." << endl);
	ERROR_STREAM("getOSSFormat: Possible values for sign are: [unsigned|signed]." << endl);
      }

      if(bits == 16) {
	string endian;
	try {
	  endian = nmm_format.getStringValue(Format::endian_param);
	} catch (FormatException e) {
	  ERROR_STREAM("getOSSFormat: Incomplete format (endian not specified)." << endl);
	  return 0;
	}

	if(sign == "unsigned") {
	  if(endian == "little") return AFMT_U16_LE;
	  if(endian == "big"   ) return AFMT_U16_BE;
	  ERROR_STREAM("getOSSFormat: Incorrect format: endian is [" << endian << "]." << endl);
	  ERROR_STREAM("getOSSFormat: Possible values for endian are [little|big]." << endl);
	}

	if(sign == "signed"  ) {
	  if(endian == "little") return AFMT_S16_LE;
	  if(endian == "big"   ) return AFMT_S16_BE;
	  ERROR_STREAM("getOSSFormat: Incorrect format: endian is [" << endian << "]." << endl);
	  ERROR_STREAM("getOSSFormat: Possible values for endian are [little|big]." << endl);
	}
	ERROR_STREAM("getOSSFormat: Incorrect format: sign is [" << sign << "]." << endl);
	ERROR_STREAM("getOSSFormat: Possible values for sign are [unsigend|signed]." << endl);
      }
      ERROR_STREAM("getOSSFormat: Only 8 or 16 bitpersample are supported by PlaybackNode."
		   << endl);
    }
    return 0;
  };

  void PlaybackNode::InitFramesizeTable() {
    struct frmsize_s framesize_table[64] = {
        { 32  ,{64   ,69   ,96   } },
        { 32  ,{64   ,70   ,96   } },
        { 40  ,{80   ,87   ,120  } },
        { 40  ,{80   ,88   ,120  } },
        { 48  ,{96   ,104  ,144  } },
        { 48  ,{96   ,105  ,144  } },
        { 56  ,{112  ,121  ,168  } },
        { 56  ,{112  ,122  ,168  } },
        { 64  ,{128  ,139  ,192  } },
        { 64  ,{128  ,140  ,192  } },
        { 80  ,{160  ,174  ,240  } },
        { 80  ,{160  ,175  ,240  } },
        { 96  ,{192  ,208  ,288  } },
        { 96  ,{192  ,209  ,288  } },
        { 112 ,{224  ,243  ,336  } },
        { 112 ,{224  ,244  ,336  } },
        { 128 ,{256  ,278  ,384  } },
        { 128 ,{256  ,279  ,384  } },
        { 160 ,{320  ,348  ,480  } },
        { 160 ,{320  ,349  ,480  } },
        { 192 ,{384  ,417  ,576  } },
        { 192 ,{384  ,418  ,576  } },
        { 224 ,{448  ,487  ,672  } },
        { 224 ,{448  ,488  ,672  } },
        { 256 ,{512  ,557  ,768  } },
        { 256 ,{512  ,558  ,768  } },
        { 320 ,{640  ,696  ,960  } },
        { 320 ,{640  ,697  ,960  } },
        { 384 ,{768  ,835  ,1152 } },
        { 384 ,{768  ,836  ,1152 } },
        { 448 ,{896  ,975  ,1344 } },
        { 448 ,{896  ,976  ,1344 } },
        { 512 ,{1024 ,1114 ,1536 } },
        { 512 ,{1024 ,1115 ,1536 } },
        { 576 ,{1152 ,1253 ,1728 } },
        { 576 ,{1152 ,1254 ,1728 } },
        { 640 ,{1280 ,1393 ,1920 } },
        { 640 ,{1280 ,1394 ,1920 } }
    };
    memcpy( frmsizecod_tbl, framesize_table, sizeof( struct frmsize_s )*64 );

    unsigned char t_sync[4] = { 0x72, 0xF8, 0x1F, 0x4E };
    memcpy( sync, t_sync, sizeof( t_sync ) );
  }

  PlaybackNode::PlaybackNode(const char* name,
                             const StreamQueue::Mode queue_mode,
                             const int size)
    : GenericBSyncSinkNode(name, queue_mode, size)
      {
    InitFramesizeTable();
    decode_ac3 = false;

    audio_fd = -1;
    mixer_fd = -1;

    //dev = "/dev/dsp";
    
    // registerEvent("flush", new TEDObject0<PlaybackNode>(this, &PlaybackNode::doFlush));
    registerEvent("finished", new TEDObject1<PlaybackNode,nmm_int32>(this, &PlaybackNode::EventDebug));
    
    registerEvent("speed_up", new TEDObject1<PlaybackNode, nmm_int32>(this, &PlaybackNode::eventSpeedUp));

    registerEvent("SEEK", new TEDObject3<PlaybackNode, Interval, Time, nmm_int32>(this, &PlaybackNode::eventSeek));

    temp_ac3    = NULL;
    temp_size   = 0;

    //max 16 fragments @ 4096 bytes
    fragment_argument =  ( 0x000F << 16) + 0x000C;

  }

  PlaybackNode::~PlaybackNode() 
  {
    if (audio_fd != -1) {
      if (close(audio_fd) == -1) {
	ERROR_STREAM("destructor: Close audio device failed." << endl);
      } else {
	DEBUG_STREAM("destructor: Close audio device. Forget to call deinit?" << endl);
      }
    }

    if ( temp_ac3 ) delete[] temp_ac3;
  }
  
  Result PlaybackNode::doInit() {
    // Open the device nonblocking, so the open call returns immediately.
    if (dev.empty())
      dev = "/dev/dsp";
    
    if (__mixer_device.empty())
      __mixer_device = "/dev/mixer";
    
    if(!SSecurityCheck::getInstance().checkFilenameForWriting(dev)) {
      ERROR_STREAM("Security check not passed for file " << dev << ", check .nmmrc for options" << endl);
      return FAILURE;
    }
    if ((audio_fd = open(dev.c_str(), O_WRONLY | O_NONBLOCK)) == -1) {
      ERROR_STREAM("Open audio device " << dev << " failed." << endl);
      audio_fd = -1;
      return FAILURE;
    }

    if(!SSecurityCheck::getInstance().checkFilenameForWriting(__mixer_device)) {
      ERROR_STREAM("Security check not passed for file " << __mixer_device << ", check .nmmrc for options" << endl);
      return FAILURE;
    }
    mixer_fd = open( __mixer_device.c_str(), O_RDWR );

    int mask;
    // Querying hardware supported formats.
    if(ioctl(audio_fd, SNDCTL_DSP_GETFMTS, &mask) == -1) {
      DEBUG_STREAM("audio device: Getting supported formats failed." << endl);
    } else {
      DEBUG_STREAM("audio device: Getting supported formats..." << endl);

      for(int i=0; test_formats[i] != 0; i++) {
	// Test all the formats in test_formats which are supported by the hardware.

	if(mask & test_formats[i]) { 
	  // Test the format only if it is supported by the hardware.
	  // Note: We also could test formats that are not supported by the hardware.
	  //       In some cases there exist OSS software converter, so that some formats
	  //       work but are not reported by the above SNDCTL_DSP_GETFMTS.
	  
	  int fmt = test_formats[i];
	  // Try to set the format...
	  if(ioctl(audio_fd, SNDCTL_DSP_SETFMT, &fmt) == -1) {
	    DEBUG_STREAM("audio device: Set format " << test_formats[i] << " failed." << endl);
	  } else {
	    // and always check the variable after doing an ioctl!
	    if(fmt == test_formats[i]) {

	      // This format is supported.
	      DEBUG_STREAM("audio device: - Format " << test_formats[i] << endl);

	      // Test the supported channel numbers for this format.
	      // Note: If you need a channel number that is currently not supported by
	      //       PlaybackNode, simply add it to the definition of the array
	      //       test_channels in this file.
	      for(int j=0; test_channels[j] != 0; j++) {
		int test_channel = test_channels[j];

		// Try to set the channel number.
		if(ioctl(audio_fd, SNDCTL_DSP_CHANNELS, &test_channel) == -1) {
		  DEBUG_STREAM("audio device:   * Set channels to " << test_channels[j]
			       << " failed." << endl);
		} else {
		  if(test_channel == test_channels[j]) {
                    
		    // The channel number is supported together with the current format.
		    DEBUG_STREAM("audio device:   * Channels " << test_channels[j] << endl);
                    
		    // Create a new format.
		    Format *input_format = new Format();
                    
		    // Configure the parameters of the format according to the actual 
		    // hardware configuration (since this configuration works.)
		    getNMMFormat(test_formats[i], *input_format);
                    
                    // Add the format to the input property.
                    getOwnInputProperty() -> addNewFormat(input_format);
                    if (!getOwnInputProperty()->getDefaultFormat()) {
		      // workaround for missing default format
		      // setting first format found to default one
		      getOwnInputProperty()->setDefaultFormat(input_format);
		    }

                    // Set channels and sample rate (not for ac3)
                    if (input_format->getFormatType() != "audio/ac3") {
                      input_format->addIntParamValue(Format::channels_param, test_channels[j]);
                      
                      // Last step: Test the supported sample rates for the current channel number
                      //            and format.
                      // Note: If you need a sample rate that is not tested here, simpy add it to
                      //       the definition of the array test_sample_rates in this file.
                      for(int k=0; test_sample_rates[k] != 0; k++) {
                        int test_rate = test_sample_rates[k];
                        if(ioctl(audio_fd, SNDCTL_DSP_SPEED, &test_rate) == -1) {
                          DEBUG_STREAM("audio device:     + Set sample rate to " << test_sample_rates[k] 
                                       << " failed." << endl);
                        } else {
                          if(test_rate == test_sample_rates[k]) {
                            // Sample rate is supported with the actual configuration.
                            DEBUG_STREAM("audio device:     + Sample rate " << test_sample_rates[k] << endl);
                            
                            // Add it to the format in the input property.
                            input_format->addIntParamValue(Format::sample_rate_param, test_sample_rates[k]);
                          } 
                        }
                      } // loop over the test_sample_rates
                    } else {
                      
                      // if ac3 is supported, support both formats with and without bitrate
                      input_format = new Format( *input_format );
                      input_format->addWildcard(Format::bitrate_param);
                      getOwnInputProperty()->addNewFormat(input_format);
                      
                    } // format not ac3
		  } // channel set correctly ?
		} // ioctl for sample rate worked ?
	      } // loop over the test_channels
	    } // channel set correctly ?
 	  } // ioctl for channel worked ?
	  
	  // After testing all configurations for one format, the device is closed and reopened.
	  // This is necessary to configure it to another format in a secure way.
	  if(close(audio_fd) == -1) {
	    ERROR_STREAM("Close audio device " << dev << " failed." << endl);
	    return FAILURE;
	  }
          if(!SSecurityCheck::getInstance().checkFilenameForWriting(dev)) {
            ERROR_STREAM("Security check not passed for file " << dev << ", check .nmmrc for options" << endl);
            return FAILURE;
          }
	  if((audio_fd = open(dev.c_str(), O_WRONLY | O_NONBLOCK)) == -1) {
	    ERROR_STREAM("Open audio device " << dev << " failed." << endl);
	    audio_fd = -1;
	    return FAILURE;
	  }
	} // format set correctly ?
      } // loop over the test_formats

      // Close and reopen the device to make it forget the last format configurations.

      if(close(audio_fd) == -1) {
	ERROR_STREAM("Close audio device " << dev << " failed." << endl);
	return FAILURE;
      }

      // This is the last open in doInit, so it should be blocking.
      if(!SSecurityCheck::getInstance().checkFilenameForWriting(dev)) {
        ERROR_STREAM("Security check not passed for file " << dev << ", check .nmmrc for options" << endl);
        return FAILURE;
      }
      if((audio_fd = open(dev.c_str(), O_WRONLY)) == -1) {
	ERROR_STREAM("Open audio device " << dev << " failed." << endl);
	audio_fd = -1;
	return FAILURE;
      }

    }

    return SUCCESS;
  };

  Result PlaybackNode::doDeinit() {
    if (audio_fd != -1) { 
      if(close(audio_fd) == -1) {
	ERROR_STREAM("doDeinit: Close audio device failed." << endl);
      }
      DEBUG_STREAM("doDeinit: Close audio device." << endl);
    }
    audio_fd = -1;

    if ( mixer_fd != -1 ) close(mixer_fd);
    mixer_fd = -1;

    return SUCCESS;
  };
  
  Result PlaybackNode::doInitOutput() {
    const Format* input_format = getInputFormat();
    if (input_format == NULL) { return FAILURE; }
    Format* format;
    
    format = new Format( *input_format );
    getOwnWorkingInputProperty()->addNewFormat(format);
    getOwnWorkingInputProperty()->setDefaultFormat(format);
    
    // TODO: Try and change this if copy-constructors for format are cleanly implemented
    // if(setFormatInternal(*input_format) == FAILURE) {
    Format tmp_format = *input_format;


    if(setFormatInternal(tmp_format) == FAILURE) {
      return FAILURE;
    }

    int caps;
    DEBUG_STREAM("Trying to get device capabilities." << endl);
    if(ioctl(audio_fd, SNDCTL_DSP_GETCAPS, &caps) == -1) {
      DEBUG_STREAM("Getting device capabilities failed." << endl);
    } else {
      DEBUG_STREAM("Getting device capabilities:" << endl);

      if((caps & DSP_CAP_REALTIME) > 0) {
	DEBUG_STREAM("audio device: - Realtime" << endl);

	int buffered_bytes;
	if(ioctl(audio_fd, SNDCTL_DSP_GETODELAY, &buffered_bytes) == -1) {
	  DEBUG_STREAM("audio device:    * SNDCTL_DSP_GETODELAY failed." << endl);

	  audio_buf_info info;
	  if(ioctl(audio_fd, SNDCTL_DSP_GETOSPACE, &info) == -1) {
	    DEBUG_STREAM("audio device:    * SNDCTL_DSP_GETOSPACE failed." << endl);
	    sync_mode = SYSTEM_CLOCK;
	  } else {
	    DEBUG_STREAM("audio device:    * SNDCTL_DSP_GETOSPACE works." << endl);
	    sync_mode = GETOSPACE;
	  }
	} else {
	  DEBUG_STREAM("audio device:    * SNDCTL_DSP_GETODELAY works." << endl);
	  sync_mode = GETODELAY;
	}
      } else {
	DEBUG_STREAM("audio device: - No Realtime" << endl);
	sync_mode = SYSTEM_CLOCK;
      }

      if((caps & DSP_CAP_BATCH) > 0) {
	DEBUG_STREAM("audio device: - Batch" << endl);
      } else {
	DEBUG_STREAM("audio device: - No batch" << endl);
      }

      if((caps & DSP_CAP_TRIGGER) > 0) {
	DEBUG_STREAM("audio device: - Trigger" << endl);
      } else {
	DEBUG_STREAM("audio device: - No trigger" << endl);
      }

      if((caps & DSP_CAP_DUPLEX) > 0) {
	DEBUG_STREAM("audio device: - Full duplex" << endl);
      } else {
	DEBUG_STREAM("audio device: - No full duplex" << endl);
      }

      if((caps & DSP_CAP_MMAP) > 0) {
	DEBUG_STREAM("audio device: - Memory map" << endl);
      } else {
	DEBUG_STREAM("audio device: - No memory map" << endl);
      }
    }

    return SUCCESS;
  }

  Result PlaybackNode::doFlush() 
  {
    int sync(0);
    if(ioctl(audio_fd, SNDCTL_DSP_SYNC, &sync) == -1) {
      DEBUG_STREAM("SNDCTL_DSP_SYNC failed." << endl);
      return FAILURE;
    }
    DEBUG_STREAM("SNDCTL_DSP_SYNC" << endl);

    return SUCCESS;
  };

  Result PlaybackNode::doStart() 
  {
    // Initialize and reset the clock. 
    // (Perhaps create a clock with new and keep the pointer.)

    return SUCCESS;
  };

  Result PlaybackNode::doStop()
  {
//     int sync;
//     if(ioctl(audio_fd, SNDCTL_DSP_SYNC, &sync) == -1) {
//       DEBUG_STREAM("SNDCTL_DSP_SYNC failed." << endl);
//     } else {
//       DEBUG_STREAM("SNDCTL_DSP_SYNC" << endl);
//     }

    // Reset the clock. (Perhaps delete the clock.)
    return SUCCESS;
  };

  nmm_int32 PlaybackNode::getFragmentSize() {
    int frag_size = -1;
    if(ioctl(audio_fd, SNDCTL_DSP_GETBLKSIZE, &frag_size) == -1)
      perror("SNDCTL_DSP_GETBLKSIZE");
    return frag_size;
  }

  void PlaybackNode::setFragmentSize(const nmm_int32& value) {
    //maximum 16 fragments @ 2^value bytes each
    fragment_argument = ( 0x000F << 16) + value;
  }

  Message* PlaybackNode::presentBuffer(Buffer *in_buffer)
  {
    if ( audio_fd != -1 ) {
      unsigned int used = in_buffer->getUsed();
      unsigned int written = 0;

      if ( decode_ac3 ) {
        struct ac3info ai;
	int skipped;
	int i;
        unsigned char* ac3buf;

	/* create temporary buffer where ac3 data is placed
	 * add a little bit more space (6144) to hold broken frames
	 * create it only if its size was changed
	 */
	if ( temp_size < (used+6144) ) {
          if ( temp_ac3 ) delete[] temp_ac3;
	  temp_size   = used+6144;
	  temp_offset = 0;
	  temp_ac3 = new unsigned char[temp_size];
	  DEBUG_STREAM( "Created temp buffer with size: " << used+6144 << endl );
	}
	/* temp_offset is greater zero, if some frame data from the last
	 * buffer is stored in front of it
	 */
        memcpy( temp_ac3 + temp_offset, (unsigned char*) in_buffer->getData(), used );
	used += temp_offset;

	/* get ac3 data, especially the frame size and the bytes that are skipped to get the sync word */
	i = ac3_iec958_parse_syncinfo(temp_ac3, used, &ai, &skipped);
        if (i < 0) {
	  DEBUG_STREAM( "AC3 stream not valid." << endl );
	}

        //fprintf(stderr, "Samplerate: %i, Framesize: %i\n", ai.samplerate, ai.framesize);

	/* skip broken frame data */
	ac3buf = temp_ac3+skipped;
	used -= skipped;

	/* calculate number of frames */
	int frames  = used / ai.framesize;

	/* if the last frame is incomplete, save the number of bytes we already have */
	temp_offset = used % ai.framesize;

        //fprintf(stderr, "Frames: %i, Rest Bytes of last frame: %i, Sync Bytes skipped: %i\n", frames, temp_offset, skipped);

	/* repack all frames in well defined 6144 byte packets and send tehm to the sound device */
	for (i = 0; i < frames; i++) {
	  ac3_iec958_build_burst( ai.framesize, 1, 1, ac3buf + i*ai.framesize, iecbuf );
          write(audio_fd, iecbuf, 6144);
	}

	/* copy incomplete frame data to begin of the buffer */
	if ( temp_offset ) memcpy( temp_ac3, ac3buf+(frames*ai.framesize), temp_offset );

      }
      else {
        while(written < used) {
	  written += write(audio_fd, in_buffer->getData() + written, used - written);
        }
      }

      written_bytes += used;
      in_buffer->release();

      getOutputDelay();
    }
    return 0;
  }
  
  Result PlaybackNode::oldSetDevice(string &value)
  {
    dev = value;
    DEBUG_STREAM("Set device to " << dev << endl);

    // check dev here?
    return SUCCESS;
  }

  Result PlaybackNode::oldGetDevice(string &value)
  {
    value = string(dev);
    return SUCCESS;
  }


  void PlaybackNode::setDevice(const string& device)
    throw (const DeviceNotFoundException&)
  {
    try {
      /* to test if the device can be accessed, we try to open the device */
      if(!SSecurityCheck::getInstance().checkFilenameForWriting(device)) {
        DEBUG_STREAM("Security check not passed for file " << device << ", check .nmmrc for options" << endl);
        throw DeviceNotFoundException(this, "Security check failed for device", device);
      }
      if((audio_fd = ::open(device.c_str(), O_RDONLY | O_NONBLOCK)) < 0) {
        /* if the device can not be accessed, then we throw an exception */
        DEBUG_STREAM("Open audio device " << device << " failed" << endl);
        audio_fd = -1;
        throw DeviceNotFoundException(this, "Audio device not available or can not be accessed", device);
      } else {
        /* close the device again */
        ::close(audio_fd);
        audio_fd = -1;
      }
    } catch(...) {
      /* close the device again */
      if(audio_fd >= 0) {
        ::close(audio_fd);
        audio_fd = -1;
      }
      
      /* rethrow the exception */
      throw;
    }
    
    /* now we can set the device */
    this->dev = device;
  }

  void PlaybackNode::setMixerDevice(const string& mixer_device)
    throw (const DeviceNotFoundException&)
  {
    try {
      /* to test if the device can be accessed, we try to open the device */
      if(!SSecurityCheck::getInstance().checkFilenameForWriting(mixer_device)) {
        ERROR_STREAM("Security check not passed for file " << mixer_device << ", check .nmmrc for options" << endl);
        return;
      }
      if((mixer_fd = ::open(mixer_device.c_str(), O_RDWR)) == -1)  {
        /* if the device can not be accessed, then we throw an exception */
        ERROR_STREAM("Open mixer device " << mixer_device << " failed." << endl);
        throw DeviceNotFoundException(this, "Mixer device not available or can not be accessed", mixer_device);
      } else {
        /* close the device again */
        ::close(mixer_fd);
        mixer_fd = -1;
      }
    } catch(...) {
      /* close the device again */
      if(mixer_fd >= 0) {
        ::close(mixer_fd);
        mixer_fd = -1;
      }
      
      /* rethrow the exception */
      throw;
    }
    
    /* now we can set the device */
    __mixer_device = mixer_device;
  }

  void PlaybackNode::setMixerVolumeElement(const string&)
      {
        ERROR_STREAM("Can't set mixer volume element (not supported" << endl);
      }

  string PlaybackNode::getDevice() const {
    return dev;
  }

  string PlaybackNode::getMixerDevice() const {
    return __mixer_device;
  }


  Result PlaybackNode::eventSeek(Interval& /*i*/, Time& /*t*/, nmm_int32& /*num_bytes*/) {
    ERROR_STREAM2("Seeking not implemented" << std::endl);    
    return SUCCESS;
  }
  
  Result PlaybackNode::eventSpeedUp(int& speedup) 
  {
    DEBUG_STREAM ("SpeedUp " << speedup << endl);
    return SUCCESS;
  }

  Result PlaybackNode::EventDebug(int& what) 
  {
    DEBUG_STREAM ("************************************************" << what << endl);
    return SUCCESS;
  }

  Result PlaybackNode::startTrack(const string& what)
  {
    DEBUG_STREAM ("****************************************** StartTrack " << what << endl);
    return SUCCESS;
  }

  Result PlaybackNode::endTrack()
  {
    DEBUG_STREAM ("****************************************** EndTrack " << endl);
    return SUCCESS;
  }

  Result PlaybackNode::started(const Interval& what)
  {
    DEBUG_STREAM ("****************************************** Started " << what << endl);
    return SUCCESS;
  }
  
  Result PlaybackNode::finished()
  {
    DEBUG_STREAM ("****************************************** Finished " << endl);
    return SUCCESS;
  }

  Interval PlaybackNode::getOutputDelay()
  {
#ifdef __CYGWIN__
    return Interval();
#endif
#ifndef __CYGWIN__
    DEBUG_STREAM("PlaybackNode::getOutputDelay" << endl);

    int buffered_bytes;
    if(ioctl(audio_fd, SNDCTL_DSP_GETODELAY, &buffered_bytes) == -1) {
      DEBUG_STREAM("audio device: SNDCTL_DSP_GETODELAY failed." << endl);
      sync_mode = GETOSPACE;
    }
    DEBUG_STREAM("getOutputDelay: " << buffered_bytes 
		 << " in internal buffer (GETODELAY)." << endl);
    Interval odelay(Rational(buffered_bytes, byte_per_second));

    DEBUG_STREAM("getOutputDelay: " << odelay << " of delay (GETODELAY)." 
		 << endl);

    //        //    case GETOSPACE:
    //        audio_buf_info info;
    //        if(ioctl(audio_fd, SNDCTL_DSP_GETOSPACE, &info) == -1) {
    //  	DEBUG_STREAM("audio device: SNDCTL_DSP_GETOSPACE failed." << endl);
    //  	sync_mode = SYSTEM_CLOCK;
    //        }
    //        DEBUG_STREAM("getOutputDelay: " << internal_buffer_size - info.bytes 
    //  		   << " in internal buffer (GETOSPACE)." << endl);
    //        Time ospace(Rational(internal_buffer_size - info.bytes, byte_per_second));
    //        DEBUG_STREAM("getOutputDelay: " << ospace << " of delay (GETOSPACE)." << endl);

    //        //  case SYSTEM_CLOCK:
    //        {
    //  	//	Time playing(written_bytes / byte_per_second);
    //          Rational time(written_bytes, byte_per_second);

    //  	// long int sec = time.getNumerator() / time.getDenominator();
    //  	// double nanos = static_cast<double>(time.getNumerator()) / time.getDenominator();
    //  	// nanos -= sec;
    //  	// nanos *= nano_per_sec;

    //  	Time playing(time);

    //  	DEBUG_STREAM("getOutputDelay: Written bytes: " << written_bytes 
    //  		     << " theo_playing: " << playing << endl);

    //  	Time real = clock.getTime();

    //  	Interval diff = playing - real;
    //  	DEBUG_STREAM("getOutputDelay: Real time: " << real << " Diff: " << diff << endl);
    //        }
    //        //break;

    //        //default:
    //        //DEBUG_STREAM("No valid mode!" << endl);

    return odelay;
#endif
  };

  Result PlaybackNode::setFormatInternal(Format& input_format)
  {
    int format = 0;
    
    if(input_format.getFormatType() == "audio/ac3" ) {
      decode_ac3 = true;
      format = AFMT_AC3;  // that's the only thing, important

      // initialize with dummy values
      channels = 2;
      sample_rate = 48000;
      bit_per_sample = 16;
    } else {
      decode_ac3 = false;

      // Get format parameters
      format  = getOSSFormat(input_format);
      if (format == 0) return FAILURE;
      channels    = input_format.getIntValue(Format::channels_param);
      sample_rate = input_format.getIntValue(Format::sample_rate_param);
      DEBUG_STREAM("Got Format: samplerate: " << sample_rate << ", channels: "
		   << channels << endl);

      try {
	bit_per_sample = input_format.getIntValue(Format::bitpersample_param);
      } catch(...) {
	bit_per_sample = 16;
      }
      if (bit_per_sample == 0) bit_per_sample = 16;
    }

    byte_per_sample = bit_per_sample / 8;
    byte_per_second = sample_rate * channels * byte_per_sample;


    // The format should not be changed while the device is playing, so
    // simply close and reopen it. Then the internal buffers will be empty
    // when changing the format.
    //if(audio_fd == -1)
    if(close(audio_fd) == -1) {
      ERROR_STREAM("Close the audio device " << dev << " failed." << endl);
      return FAILURE;
    }

    if(!SSecurityCheck::getInstance().checkFilenameForWriting(dev)) {
      ERROR_STREAM("Security check not passed for file " << dev << ", check .nmmrc for options" << endl);
      return FAILURE;
    }

    //if(audio_fd ==-1)
    if((audio_fd = open(dev.c_str(), O_WRONLY)) == -1) {
      ERROR_STREAM("Open the audio device " << dev << " failed." << endl);
      audio_fd = -1;
      return FAILURE;
    }
    
    // It is important always to keep the same order in setting the parameters:
    // At first the format, then the channel number and at last the sample rate.
    DEBUG_STREAM("audio device: Trying to set format " << format << endl);

    int fmt = format;
    if (ioctl(audio_fd, SNDCTL_DSP_SETFMT, &fmt) == -1) {
      ERROR_STREAM("Set format " << format << " failed." << endl);
      if(close(audio_fd) == -1) {
	ERROR_STREAM("Close audio device failed." << endl);
      }
      audio_fd = -1;
      return FAILURE;
    }

    if(fmt != format) {
      DEBUG_STREAM("Format " << format << " not supported." << endl);
      DEBUG_STREAM("Nearest is " << fmt << "." << endl);
      if(close(audio_fd) == -1) {
	ERROR_STREAM("Close audio device failed." << endl);
      }
      audio_fd = -1;
      return FAILURE;
    }
    DEBUG_STREAM("audio device: Set format to " << format << endl);

    DEBUG_STREAM("audio device: Trying to set channels to " << channels << endl);
    int ch = channels;
    if (ioctl(audio_fd, SNDCTL_DSP_CHANNELS, &ch) == -1) {
      ERROR_STREAM("Set channel number " << channels << " failed." << endl);
      if(close(audio_fd) == -1) {
	ERROR_STREAM("Close audio device failed." << endl);
      }
      audio_fd = -1;
      return FAILURE;
    }
    if(ch != channels) {
      DEBUG_STREAM("Channels " << channels << " not supported with format "
		   << format << endl);
      if(close(audio_fd) == -1) {
	ERROR_STREAM("Close audio device failed." << endl);
      }
      audio_fd = -1;
      return FAILURE;
    }
    DEBUG_STREAM("audio device: Set channels to " << channels << endl);

    DEBUG_STREAM("audio device: Trying to set sample_rate to " << sample_rate << "." << endl);
    int rate = sample_rate;
    if (ioctl(audio_fd, SNDCTL_DSP_SPEED, &rate) == -1) {
      ERROR_STREAM("Set sample rate " << sample_rate << " failed." << endl);
      if(close(audio_fd) == -1) {
	ERROR_STREAM("Close audio device failed." << endl);
      }
      audio_fd = -1;
      return FAILURE;
    }
    if(rate != sample_rate) {
      DEBUG_STREAM("Sample rate " << sample_rate << " not supported with "
		   << format << " format and " << channels << " channels." << endl);
      if(close(audio_fd) == -1) {
	ERROR_STREAM("Close audio device failed." << endl);
      }
      audio_fd = -1;
      return FAILURE;
    }
    DEBUG_STREAM("audio device: Set sample_rate to " << sample_rate << "." << endl);

    // this means fragment size 2 ^ 0x000C (= 2^12 = 4096) bytes
    // and maximal 0x000F (= 16) fragments
    // That means 16*4096 / (48000*2*2) = 0.341 sec buffer at 48 kHz, stereo, 16 bit per sample
    // or         16*4096 / (44100*2*2) = 0.371 sec buffer at 44.1 kHz, stereo, 16 bit per sample

    if ( !decode_ac3 ) {
      if (ioctl(audio_fd, SNDCTL_DSP_SETFRAGMENT, &fragment_argument)) {
        // buffer setting failed
        perror(dev.c_str());
        close(audio_fd);
        audio_fd = -1;
        return FAILURE;
      }
    }

    // Get fragment size and number of fragments
    audio_buf_info ospace;

    if (ioctl(audio_fd, SNDCTL_DSP_GETOSPACE, &ospace) == -1) {
      DEBUG_STREAM("Error while taking the audio device's output space."
		   << endl);
    }

    DEBUG_STREAM("audio device: Totally allocated " << ospace.fragstotal
		 << " fragments with " << ospace.fragsize
		 << " bytes per fragment." << endl);
    DEBUG_STREAM("audio device: That means "
		 << (ospace.fragsize * ospace.fragstotal * 1000) / byte_per_second
		 << " milliseconds of audio can be in the internal buffer. "
		 << endl);

    internal_buffer_size = ospace.fragstotal * ospace.fragsize;
    internal_buffer_time = Time(static_cast<long int>(static_cast<double>(internal_buffer_size) / byte_per_second));

    DEBUG_STREAM("getOutputDelay: Internal buffer size: " << internal_buffer_size
		 << " bytes." << endl);
    DEBUG_STREAM("getOutputDelay: Internal buffer time: " << internal_buffer_time
		 << endl);

    return SUCCESS;
  };

  void PlaybackNode::ac3_iec958_build_burst( int length, int data_type, int big_endian, unsigned char* data, unsigned char* out ) {
    memcpy(out, sync, 4);
    if (length)
      out[4] = data_type; /* & 0x1F; */
    else
      out[4] = 0;

    out[5] = 0x00;
    out[6] = (length*8) & 0xFF;
    out[7] = ((length*8) >> 8) & 0xFF;
    if (big_endian)
      swab((void*) data, (void*) (out + 8), length);
    else
      memcpy(out + 8, data, length);

    memset(out + 8 + length, 0, 6144 - 8 - length);
  }

  int PlaybackNode::ac3_iec958_parse_syncinfo( unsigned char* buf, int size, struct ac3info* ai, int* skipped ) {
    int samplerates[4] = { 48000, 44100, 32000, -1 };
    unsigned short sync = 0;
    unsigned char *ptr = buf;
    int fscod, frmsizecod;
    struct syncframe *sf;

    sync = buf[0] << 8;
    sync |= buf[1];
    ptr = buf + 2;
    *skipped = 0;
    while (sync != 0xb77 && *skipped < size - 8) {
      sync <<= 8;
      sync |= *ptr;
      ptr++;
      *skipped += 1;
    }

    if (sync != 0xb77) return -1;
    ptr -= 2;
    sf = (struct syncframe *) ptr;
    fscod = (sf->syncinfo.code >> 6) & 0x03;
    ai->samplerate = samplerates[fscod];
    if (ai->samplerate == -1)
      return -1;
    frmsizecod = sf->syncinfo.code & 0x3f;
    ai->framesize = 2 * frmsizecod_tbl[frmsizecod].frm_size[fscod];
    ai->bitrate = frmsizecod_tbl[frmsizecod].bit_rate;
    if (((sf->bsi.bsidmod >> 3) & 0x1f) != 0x08) return -1;

    ai->bsmod = sf->bsi.bsidmod & 0x7;

    return 0;
  }

  void PlaybackNode::setLineInVolume( int vol ) {
    if ( mixer_fd != -1 ) {
      ioctl(mixer_fd, MIXER_WRITE(SOUND_MIXER_LINE), &vol);
    }
  }

  int PlaybackNode::getLineInVolume() {
    int vol = 0;
    if ( mixer_fd != -1 ) {
      ioctl(mixer_fd, MIXER_READ(SOUND_MIXER_LINE), &vol);
    }

    return( vol );
  }
  
  Result PlaybackNode::doSetFormat(const Format& format) 
  {
    DEBUG_STREAM("PlaybackNode::doSetFormat" << endl);
    try {
      DEBUG_STREAM("PlaybackNode::doSetFormat, samplerate: " << format.getIntValue(Format::channels_param) << ", channels: " << format.getIntValue(Format::sample_rate_param) << endl);
    }
    catch(Exception e) {
      ERROR_STREAM("Exception in doSetFormat " << e << endl);
    }
    return SUCCESS;
  }
  
  void PlaybackNode::increaseVolume()
  {
    setVolume(getVolume() + 10);
  }
  
  void PlaybackNode::decreaseVolume()
  {
    setVolume(getVolume() - 10);
  }
  
  void PlaybackNode::setVolume(const nmm_int32& volume)
  {
    if ( mixer_fd != -1 ) {
      int v = volume;
      if( v < 0)
        v = 0;
      v = (v << 8) | v;
      ioctl(mixer_fd, SOUND_MIXER_WRITE_PCM, &v);
    }
  }

  nmm_int32 PlaybackNode::getVolume()
  {
    int vol = 0;
    if ( mixer_fd != -1 ) {
      ioctl(mixer_fd, SOUND_MIXER_READ_PCM, &vol);
      vol = (((vol & 0xFF00) >> 8) + (vol & 0x00FF)) / 2;
    }
    return vol;
  }

  void PlaybackNode::mute()
  {
    DEBUG_STREAM("MUTE" << endl);
    getController()->setBuffersInvalid(true);
  }

  void PlaybackNode::unmute()
  {
    DEBUG_STREAM("UN-MUTE" << endl);
    getController()->setBuffersInvalid(false);
  }
  
  
  PlaybackNodePlugin::PlaybackNodePlugin (const string& node_name, int max_instance)
    : TDevicePlugin<PlaybackNode>("/dev/dsp", true, node_name, max_instance)
  {}
  
  void PlaybackNodePlugin::checkForStandardAuxDevice(IDevice& device) throw() {
    /* we get the IAudio device to specify the node capabilities on the system */
    IAudioDevice_var i_audio(device.getParentObject()->getCheckedInterface<IAudioDevice>());
    
    /* then we try to setup the mixer device. This device must not available for
       a soundcard, but if we wish to use an availble mixer we must setup the device. */
    try {
      i_audio->setMixerDevice("/dev/mixer");
    } catch (const DeviceNotFoundException& de ) {
      GLOBAL_WARNING_STREAM(de << endl);
    }
  }
  
  void PlaybackNodePlugin::checkForAuxDevice(IDevice& device, unsigned device_num) throw() {
    /* we get the IAudio device to specify the node capabilities on the system */
    IAudioDevice_var i_audio(device.getParentObject()->getCheckedInterface<IAudioDevice>());
    
    string next_mixer_device = string("/dev/mixer") + device_num;
    try {
      i_audio->setMixerDevice(next_mixer_device);
    } catch (const DeviceNotFoundException& de ) {
      GLOBAL_WARNING_STREAM(de << endl);
    }
  }

  void PlaybackNode::setCLOEXEC(int fd){

    int oldflags = fcntl (fd, F_GETFD, 0);
    if (oldflags >= 0){
      oldflags |= FD_CLOEXEC;
      fcntl (fd, F_SETFD, oldflags);
      GLOBAL_DEBUG_STREAM("Setting CLOEXEC on audio_fd" << endl);
    }else{
      GLOBAL_ERROR_STREAM("Could not set CLOEXEC on audio_fd" << endl);
    }
  }

}



#include "nmm/nmm_suffix.hpp"
