/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "RecordNode.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"
#include "nmm/multimedia/sync/StreamTimer.hpp"
#include "nmm/utils/SecurityCheck.hpp"
#include "nmm/utils/time/Rational.hpp"

#include <sys/fcntl.h>
#include <sys/soundcard.h>
#include <sys/ioctl.h>

#ifndef __CYGWIN__
static NMM::RecordNodePlugin plugin("RecordNode",-1);
NMM_REGISTER_SPECIAL_PLUGIN(RecordNode, NMM, plugin)
#endif
#ifdef __CYGWIN__
NMM_REGISTER_PLUGIN(RecordNode, NMM, "RecordNode")
#endif

//#define BUFFER_LENGTH 1024*1

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  RecordNode::RecordNode(const char* name /* = "RecordNode" */,
			 StreamQueue::Mode mode /* = StreamQueue::MODE_SUSPEND */,
			 int queue_size /* = StreamQueue::DEFAULT_SIZE */)
    : GenericSyncSourceNode(name, mode, queue_size) {

    // set the record node as master of synchronization
    setSyncMaster(true);

    // this node produces at a constant rate
    setConstantRateProducer(true);

    dev = "/dev/dsp";
    audio_fd = -1;
    buffer_size = 1024;
    
    timer = new StreamTimer(this, StreamTimer::REAL_TIME);
    sample_rate = 0;
    num_channels = 0;
    bits_per_sample = 0;

    // 4 buffers @ 2^10 bytes
    fragment_argument = 0x0008000A; 
  }
  
  RecordNode::~RecordNode() {
    if (audio_fd != -1) close(audio_fd);
    delete timer;
  }
  
  Result RecordNode::doInit() {
    if ((audio_fd = open(dev.c_str(), O_RDONLY, 0)) == -1) {
      // open failed
      ERROR_STREAM("Failed to open recording device " << dev << endl);
      audio_fd = -1;
      return FAILURE;
    }
    
    Format *format = getOwnOutputProperty()->addNewFormat("audio/raw");
    format->addIntParamValue(Format::sample_rate_param, 44100);
    format->addIntParamValue(Format::channels_param, 2);
    format->addIntParamValue(Format::bitpersample_param, 16);
    format->addStringParamValue(Format::endian_param, "little");
    format->addStringParamValue(Format::encoding_param, "linear");
    format->addStringParamValue(Format::sign_param, "signed");
    getOwnOutputProperty()->setDefaultFormat(format);
    
    format = getOwnOutputProperty()->addNewFormat("audio/raw");
    format->addIntParamValue(Format::sample_rate_param, 48000);
    format->addIntParamValue(Format::sample_rate_param, 44100);
    format->addIntParamValue(Format::sample_rate_param, 22050);
    format->addIntParamValue(Format::sample_rate_param, 11025);
    format->addIntParamValue(Format::sample_rate_param, 8000);
    format->addIntParamValue(Format::channels_param, 1);
    format->addIntParamValue(Format::channels_param, 2);
    format->addIntParamValue(Format::bitpersample_param, 16);
    format->addStringParamValue(Format::endian_param, "little");
    format->addStringParamValue(Format::encoding_param, "linear");
    format->addStringParamValue(Format::sign_param, "signed");
    
    return SUCCESS;
  }
  
  Result RecordNode::doDeinit() {
    if (audio_fd != -1) {
      close(audio_fd);
      audio_fd = -1;
    }
    return SUCCESS;
  }
  
  Result RecordNode::doInitOutput() {
    // copy formats from static property to working property
    list<Format*> formats = getOwnOutputProperty()->getFormatList();
    for (list<Format*>::iterator i(formats.begin()); i != formats.end(); ++i) {
      const Format* format = *i;
      Format* new_format = new Format( *format );
      getOwnWorkingOutputProperty()->addNewFormat(new_format);
      if (format == getOwnOutputProperty()->getDefaultFormat()) {
        getOwnWorkingOutputProperty()->setDefaultFormat(new_format);
      }
    }
    
    return SUCCESS;
  }
  
  void RecordNode::setDevice(const string& device)
      throw(const NMM::DeviceNotFoundException&)
  {   
    try {
      /* to test if the device can be accessed, we try to open the device */
      if(!SSecurityCheck::getInstance().checkFilenameForWriting(device)) {
        DEBUG_STREAM("Security check not passed for file " << device << ", check .nmmrc for options" << endl);
        throw DeviceNotFoundException(this, "Security check failed for device", device);
      }
      if((audio_fd = open(device.c_str(), O_RDONLY | O_NONBLOCK)) < 0) {
        /* if the device can not be accessed, then we throw an exception */
        DEBUG_STREAM("Open audio device " << device << " failed" << endl);
        audio_fd = -1;
        throw DeviceNotFoundException(this, "Audio device not available or can not be accessed", device);
      } else {
        /* close the device again */
        ::close(audio_fd);
        audio_fd = -1;
      }
    } catch(...) {
      /* close the device again */
      if(audio_fd >= 0) {
        ::close(audio_fd);
        audio_fd = -1;
      }
      
      /* rethrow the exception */
      throw;
    }
    
    /* now we can set the device */
    this->dev = device;
  }
  
  string RecordNode::getDevice() const {
    return dev;
  }
  
  void RecordNode::setBufferSize(const nmm_int32 &size) {
    if (size > 0) {
      buffer_size = size;
      return;
    }
    ERROR_STREAM( "setValue: trying to set a negative buffer_size! " << endl );
  }
  
  nmm_int32 RecordNode::getBufferSize() {
    return buffer_size;
  }
  
  nmm_int32 RecordNode::getFragmentSize() {
    int frag_size = -1;
    if(ioctl(audio_fd, SNDCTL_DSP_GETBLKSIZE, &frag_size) == -1)
      perror("SNDCTL_DSP_GETBLKSIZE");
    return frag_size;
  }

  void RecordNode::setFragmentSize(const nmm_int32& value) {
    // 4 fragments @ 2^value bytes
    fragment_argument = ( 0x0008 << 16) + value; 
  }

  Result RecordNode::doActivate() {
    /* We can only be sure of our formats here, not earlier.
     * So all ioctls will be called in this function.
     */
    const Format *out_f = getOutputFormat();    
    
    int format = AFMT_S16_LE;
    num_channels = out_f->getIntValue(Format::channels_param);
    sample_rate = out_f->getIntValue(Format::sample_rate_param);
    bits_per_sample = out_f->getIntValue(Format::bitpersample_param);
    
    if (ioctl(audio_fd, SNDCTL_DSP_SETFRAGMENT, &fragment_argument)) {
      // buffer setting failed
      perror(dev.c_str());
      close(audio_fd); audio_fd =-1;
      return FAILURE;
    }

    if (ioctl(audio_fd, SNDCTL_DSP_SETFMT, &format) == -1) {
     // set format failed
      perror("SNDCTL_DSP_SETFMT");
      close(audio_fd); audio_fd =-1;
      return FAILURE;
    }
    if (ioctl(audio_fd, SNDCTL_DSP_CHANNELS, &num_channels) == -1) {
      // set channels failed
      perror("SNDCTL_DSP_CHANNELS");
      close(audio_fd); audio_fd =-1;
      return FAILURE;
    }
    if (ioctl(audio_fd, SNDCTL_DSP_SPEED, &sample_rate) == -1) {
      // set speed failed
      perror("SNDCTL_DSP_SPEED");
      close(audio_fd); audio_fd =-1;
      return FAILURE;
    }

    timer->reset();
    return SUCCESS;
  }
  
  Time RecordNode::getLocalTime(u_int64_t valid_read_bytes) {
    return Time(Rational(valid_read_bytes, sample_rate * num_channels * bits_per_sample / 8));
  }
  
  Message* RecordNode::produceBuffer(Buffer *) {
    Buffer *buffer = getNewBuffer(buffer_size);
    if (buffer != 0) {
      buffer->setUsed(read(audio_fd, buffer->getData(), buffer_size));
      if (buffer->getUsed() > 0) {
	timer->setTimestamp(buffer);
	return buffer;
      } 
      else {
        // read error, release buffer
        buffer->release();
	return 0;
      }
    }
    return 0;
  }
  
  
  RecordNodePlugin::RecordNodePlugin (const string& node_name, int max_instance)
    : TDevicePlugin<RecordNode>("/dev/dsp", true, node_name, max_instance)
  {}
  
}

#include "nmm/nmm_suffix.hpp"
