/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "ScopeNode.hpp"
#include "nmm/comm/format/Property.hpp"
#include "nmm/multimedia/Plugin.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"
#include "nmm/utils/time/Rational.hpp"

NMM_REGISTER_PLUGIN(NMM::ScopeNode, NMM, "ScopeNode")

namespace NMM {

  ScopeNode::ScopeNode(const char *name,
		       const StreamQueue::Mode queue_mode,
		       const int size)
    : GenericProcessorNode(name, queue_mode, size) 
  {
    
    blur_enabled = false;
    blur_step    = 10;
  };
  
  ScopeNode::~ScopeNode() {}
  
  
  Result ScopeNode::doInit() {
    NMM::Format *input_format;
    input_format = getOwnInputProperty()->addNewFormat("audio/raw");
    input_format->addWildcard(Format::sample_rate_param);  
    input_format->addIntParamValue(Format::channels_param, 2);
    input_format->addIntParamValue(Format::bitpersample_param, 16);
    input_format->addStringParamValue(Format::encoding_param, "linear");
    input_format->addStringParamValue(Format::sign_param, "signed");
    input_format->addStringParamValue(Format::endian_param, "little");
    getOwnInputProperty()->setDefaultFormat(input_format);
    
    NMM::Format *output_format;
    output_format = getOwnOutputProperty()->addNewFormat("video/raw");
    output_format->addWildcard(Format::x_resolution_param);
    output_format->addWildcard(Format::y_resolution_param);
    output_format->addWildcard(Format::ratio_x_param);
    output_format->addWildcard(Format::ratio_y_param);
    output_format->addWildcard(Format::framerate_param);
    output_format->addStringParamValue(Format::colorspace_param, "rgb24");
    output_format->addStringParamValue(Format::channelorder_param, "rgb");
    output_format->addStringParamValue(Format::format_param, "packet");
    output_format->addStringParamValue(Format::endian_param, "little");
    output_format->addIntParamValue(Format::bitperpixel_param, 32);
    getOwnOutputProperty()->setDefaultFormat(output_format);
    
    input_format  -> setIOPartner(output_format);
    output_format -> setIOPartner(input_format);
    
    return SUCCESS;
  }
  
  Result ScopeNode::doDeinit() {
    getOwnOutputProperty()->erase();
    getOwnInputProperty()->erase();
    return SUCCESS;
  }
  
  Result ScopeNode::doInitOutput() {
    const Format* in_format = getInputFormat();
    if (in_format == NULL) { return FAILURE; }
    
    Format *input_format = new Format( *in_format );
    getOwnWorkingInputProperty()->addNewFormat( input_format );
    getOwnWorkingInputProperty()->setDefaultFormat(input_format);
    
    Format *output_format;
    output_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
    output_format->addIntParamValue(Format::x_resolution_param, 400);
    output_format->addIntParamValue(Format::y_resolution_param, 300);
    output_format->addIntParamValue(Format::ratio_x_param, 4);
    output_format->addIntParamValue(Format::ratio_y_param, 3);
    output_format->addFloatParamValue(Format::framerate_param, 25.0);
    output_format->addStringParamValue(Format::colorspace_param, "rgb24");
    output_format->addStringParamValue(Format::channelorder_param, "rgb");
    output_format->addStringParamValue(Format::format_param, "packet");
    output_format->addStringParamValue(Format::endian_param, "little");
    output_format->addIntParamValue(Format::bitperpixel_param, 32);
    getOwnWorkingOutputProperty()->setDefaultFormat(output_format);
    
    output_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
    output_format->addWildcard(Format::x_resolution_param);
    output_format->addWildcard(Format::y_resolution_param);
    output_format->addWildcard(Format::ratio_x_param);
    output_format->addWildcard(Format::ratio_y_param);
    output_format->addWildcard(Format::framerate_param);
    output_format->addStringParamValue(Format::colorspace_param, "rgb24");
    output_format->addStringParamValue(Format::channelorder_param, "rgb");
    output_format->addStringParamValue(Format::format_param, "packet");
    output_format->addStringParamValue(Format::endian_param, "little");
    output_format->addIntParamValue(Format::bitperpixel_param, 32);

    input_format  -> setIOPartner(output_format);
    output_format -> setIOPartner(input_format);
    
    return SUCCESS;
  }
  
  Result ScopeNode::doDeinitOutput() {
    getOwnWorkingOutputProperty()->erase();
    getOwnWorkingInputProperty()->erase();
    return SUCCESS;
  }
  
  Result ScopeNode::doActivate() {
    
    // ------------------------ Get the input parameters ------------------ 
    const Format* in_format = getInputFormat();
    if(in_format == 0) {
      ERROR_STREAM("doActivate failed: No valid input format." << endl);
      return FAILURE;
    }

    try {
      sample_rate = in_format->getIntValue(Format::sample_rate_param);
    } catch(FormatException e) {
      ERROR_STREAM("doActivate failed: Incomplete input format "
		   "- missing sample_rate." << endl);
      return FAILURE;
    }

    try {
      input_channels = in_format->getIntValue(Format::channels_param);
    } catch(FormatException e) {
      ERROR_STREAM("doActivate failed: Incomplete input format "
		   "- missing channel number." << endl);
      return FAILURE;
    }
    if(input_channels < 1) {
      ERROR_STREAM("doActivate failed: Input channel number less than one." 
		   << endl);
      return FAILURE;
    }

    int bit_per_sample = 0;
    try {
      bit_per_sample = in_format->getIntValue(Format::bitpersample_param);
    } catch(FormatException e) {
      ERROR_STREAM("doActivate failed: Incomplete input format "
		   "- bit per sample not specified." << endl);
      return FAILURE;
    }
    byte_per_sample = bit_per_sample / 8;


    // ------------------------ Get the output parameters ----------------- 
    const Format* out_format = getOutputFormat();
    if(out_format == 0) {
      ERROR_STREAM("doActivate failed: No valid output format." << endl);
      return FAILURE;
    }

    try {
      image_width = out_format->getIntValue(Format::x_resolution_param);
    } catch (FormatException e) {
      ERROR_STREAM("doActivate failed: Incomplete output format "
		   "- x_resolution not specified." << endl);
      return FAILURE;
    }
    DEBUG_STREAM("Output image width: " << image_width << endl);

    try {
      image_height = out_format->getIntValue(Format::y_resolution_param);
    } catch (FormatException e) {
      ERROR_STREAM("doActivate failed: Incomplete output format "
		   "- y_resolution not specified." << endl);
      return FAILURE;
    }
    DEBUG_STREAM("Output image height: " << image_height << endl);

    try {
      framerate = out_format->getFloatValue(Format::framerate_param);
    } catch (FormatException e) {
      ERROR_STREAM("doActivate failed: Incomplete output format "
		   "- framerate not specified." << endl);
      return FAILURE;
    }
    DEBUG_STREAM("Output framerate: " << framerate << endl);
    
    if ((image_width <= 0) || (image_height <= 0)) {
      DEBUG_STREAM("doActivate failed: Image width or height less than or "
		   "equal zero." << endl);
      return FAILURE;
    }

    int samples_per_image = static_cast<int>(sample_rate / framerate);
    gap = samples_per_image - image_width;
    gap *= input_channels * byte_per_sample;

    curr_audio = 0;
    curr_video = 0;

    range = image_width * image_height * 4;

    image = new unsigned char[image_height * image_width * 4];
    memset(image, 0, image_height * image_width * 4 * sizeof(unsigned char));

    // draw coordinate system
    int i = image_height/2; 
    for(int j = 0; j < image_width; j++) {
      setPoint(image, i, j);
    }
    
    i = image_width /2;
    for(int j = 0; j < image_height; j++) {
      setPoint(image, j, i);
    }

    return SUCCESS;
  };
  
  Result ScopeNode::doDeactivate() {
    delete[] image;

    range = 0;
    image_height = 0;
    image_width = 0;
    framerate = 0;
    input_channels = 0;
    sample_rate = 0;
    byte_per_sample = 0;

    return SUCCESS;
  };

  Result ScopeNode::setBlurEnabled(const nmm_bool& toggle) {
    blur_enabled = toggle;
    return SUCCESS;
  };

  Result ScopeNode::setBlurStep(const nmm_int32& step) {
    if(step > 0) {
      blur_step = step;
      DEBUG_STREAM("Set blur step to " << step << "." << endl);

      return SUCCESS;
    }
    ERROR_STREAM("Trying to set a blur step equal or less than zero." << endl);
    return FAILURE;
  };

  void ScopeNode::setPoint(unsigned char* image, int x, int y) {
    if(x > image_height) x = image_height - 1;
    if(x < 0) x = 0;
    if(y > image_width)  y = image_width  - 1;
    if(y < 0) y = 0;

    image[(x*image_width+y)*4    ] = static_cast<unsigned char>(128);
    image[(x*image_width+y)*4 + 1] = static_cast<unsigned char>(128);
    image[(x*image_width+y)*4 + 2] = static_cast<unsigned char>(255);
    image[(x*image_width+y)*4 + 3] = static_cast<unsigned char>(0);
  };

  Buffer* ScopeNode::processBuffer(Buffer* in_buffer) {
    if(!getWorkingFlag()) {
      /* get audio data */
      audio_end = in_buffer->getUsed();
      audio = in_buffer->getData();

      /* calculate a valid timestamp for current input buffer */
      in_ts = in_buffer->getTimestamp();
      if (in_ts.is_valid) {
        valid_ts = in_ts;
        valid_samples = 0;
      } else {
        in_ts.sync_time = valid_ts.sync_time + Rational(valid_samples, sample_rate);
      }
      
      /* count samples since last valid timestamp */
      valid_samples += audio_end / (input_channels * byte_per_sample);
      
      /* check if this buffer is too "old" */
      if(curr_audio >= audio_end) {
	curr_audio -= audio_end;
	in_buffer->release();
	return 0;
      }
      
      old_in_buffer = in_buffer;
      DEBUG_STREAM("proc: New audio buffer begins." << endl);
    }
    
    /* get starting position in this buffer in samples */
    in_samples = curr_audio / (input_channels * byte_per_sample);
    
    /* if we don't have a valid output timestamp, set it to the current input position */
    if (!out_ts.is_valid) {
      out_ts.is_valid = true;
      out_ts.sync_time = in_ts.sync_time + Rational(in_samples, sample_rate);
    }
    
    // How many samples of the audio buffer can we use?
    int audio_samples = (audio_end - curr_audio) / (input_channels * byte_per_sample);

    // How many samples do we need to finish the current out_buffer?
    int needed_samples = image_width - curr_video;
    
    // Enough samples to finish the current image?
    if(needed_samples <= audio_samples) {
      // Yes: Plot the rest of the current image.
      while(curr_video < image_width) {
	short int sample = audio[curr_audio];
	sample += audio[curr_audio + 1] * 256;
	setPoint(image, image_height * sample / 65536 + image_height/2, curr_video);
	curr_video++;
	curr_audio += input_channels * byte_per_sample;
      }

      // Compute the position in the audio stream to start
      // the next image.
      curr_audio += gap;
      if(curr_audio < audio_end) {
	// If this is inside the current audio buffer...
	// we don't need a new one.
	setWorkingFlag(true);
      } else {
	// The current audio buffer is not needed any more...
	curr_audio -= audio_end;

	old_in_buffer->release();
	setWorkingFlag(false);
      }

      curr_video = 0;

      Buffer* out_buffer = getNewBuffer(range);
      memcpy(out_buffer->getData(), image, range);
      
      /* add timestamp to output buffer */
      out_buffer->setTimestamp(out_ts);
      
      /* increment the stream counter and mark output timestamp as invalid (to calculate a new one next time) */
      out_ts.is_valid = false;
      out_ts.stream_counter++;
      
      if(blur_enabled) {
	// blur effect
	for(int i=0; i < range; i++) {
	  if (image[i] > blur_step) image[i] -= blur_step;
	  else image[i] = 0;
	}
      } else {
	memset(image, 0, range * sizeof(unsigned char));
      }

      // draw coordinate system
      int i = image_height/2; 
      for(int j = 0; j < image_width; j++) {
	setPoint(image, i, j);
      }

      i = image_width /2;
      for(int j = 0; j < image_height; j++) {
	setPoint(image, j, i);
      }

      return out_buffer;

    } else {
      // No: Plot till the end of the audio buffer ...
      while(curr_audio < audio_end) {
	short int sample = audio[curr_audio];
	sample += audio[curr_audio + 1] * 256;
	setPoint(image, image_height * sample / 65536 + image_height/2, curr_video);
	curr_video++;
	curr_audio += input_channels * byte_per_sample;
      }

      // The beginning of the next audio buffer is the next data
      // to plot.
      curr_audio = 0;

      // Release the old audio buffer.
      old_in_buffer->release();

      // More audio data is needed to finish the out_buffer.
      setWorkingFlag(false);
    }
    
    return 0;
  };

  Result ScopeNode::syncReset() {
    
    /* fw:
     * Timestamp generation is done in 2 steps.
     *
     * In the first step we simply calculate the starting time of the current input buffer.
     * This is simply the timestamp of the buffer, if it is valid.
     * Otherwise we use the last valid timestamp and add valid_samples / sample_rate to it.
     * valid_samples counts the number of samples since the last valid timestamp.
     * Whenever we reset the sync, we set the last valid timestamp to zero.
     *
     * In the second step we calculate the timestamp of the next output buffer by setting it
     * to the current input position whenever we start a new frame. The current input position
     * is the in buffer timestamp calculated in the first step + in_samples / sample_rate.
     * in_samples counts the number of samples since the beginning of the current input buffer.
     */
    
    if (getWorkingFlag()) {
      // if we are still working, release the old buffer to avoid memory leaks
      old_in_buffer->release();
    }
    
    // let's start fresh
    curr_audio = 0;
    curr_video = 0;

    valid_ts = Timestamp();
    valid_ts.sync_time = Time(0);
    valid_ts.is_valid = true;
    valid_samples = 0;
    
    in_ts = valid_ts;
    in_samples = 0;
    
    out_ts = Timestamp();
    out_ts.is_valid = false;
    
    return SUCCESS;
  }
}
