/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "AVMuxNode.hpp"
#include <assert.h>
#include "nmm/comm/format/Property.hpp"
#include "nmm/comm/messaging/Buffer.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"

NMM_REGISTER_PLUGIN(AVMuxNode, NMM, "AVMuxNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  AVMuxNode::AVMuxNode(const char *name, StreamQueue::Mode qmode, int qsize)
    : GenericMultiplexerNode(name, qmode, qsize),
      buf_count(0),
      track_start_count(0),
      send_event(true),
      last_buffer(NULL),
      audio_end_track(false),
      video_end_track(false),
      first_audio_buffer(false),
      first_video_buffer(false),
      audio_too_late(true),      
      set_audio_too_late(false),      
      discard_mode(true),      
      discard_mode_running(true),      
      sync_mode(true),
      sync_flush(false),
      last_audio(NULL),
      last_video(NULL),
      sync_end_track(false)
  {
  }
  
  AVMuxNode::~AVMuxNode() {
  }
  
  Result AVMuxNode::doInit() {
    /* Begin> Add input streams */
    if (FAILURE == addInputStream("audio")) {
      ERROR_STREAM("Failed to add audio input stream" << endl);
      return FAILURE;
    }
    if (FAILURE == addInputStream("video")) {
      ERROR_STREAM("Failed to add video input stream" << endl);
      return FAILURE;
    }
    /* End< Add input streams*/
    
    /* Begin> Video Subformats */
    /* DivX video with bitrate */
    Format v_format1(Format::video_mpeg4_type);
    v_format1.addWildcard(Format::x_resolution_param);
    v_format1.addWildcard(Format::y_resolution_param);
    v_format1.addWildcard(Format::ratio_x_param);
    v_format1.addWildcard(Format::ratio_y_param);
    v_format1.addWildcard(Format::framerate_param);
    v_format1.addWildcard(Format::compressor_param);
    v_format1.addWildcard(Format::bitrate_param);
    
    /* DivX video without bitrate */    
    Format v_format2(Format::video_mpeg4_type);
    v_format2.addWildcard(Format::x_resolution_param);
    v_format2.addWildcard(Format::y_resolution_param);
    v_format2.addWildcard(Format::ratio_x_param);
    v_format2.addWildcard(Format::ratio_y_param);
    v_format2.addWildcard(Format::framerate_param);
    v_format2.addWildcard(Format::compressor_param);
    /* End< Video Subformats */
    
    /* Begin> Audio Subformats */
    /* MPEG audio with bitrate and extra parameters */
    Format a_format1(Format::audio_mpeg_type);
    a_format1.addIntParamValue(Format::audio_codec_param, 0x55);
    a_format1.addWildcard(Format::layer_param);
    a_format1.addWildcard(Format::channels_param);
    a_format1.addWildcard(Format::sample_rate_param);
    a_format1.addWildcard(Format::bitpersample_param);
    a_format1.addWildcard(Format::bitrate_param);

    /* MPEG audio without bitrate and with extra parameters */
    Format a_format2(Format::audio_mpeg_type);
    a_format2.addWildcard(Format::layer_param);
    a_format2.addIntParamValue(Format::audio_codec_param, 0x55);
    a_format2.addWildcard(Format::sample_rate_param);
    a_format2.addWildcard(Format::channels_param);
    a_format2.addWildcard(Format::bitpersample_param);
    
    /* MPEG audio with sample rate only */
    Format a_format3(Format::audio_mpeg_type);
    a_format3.addIntParamValue(Format::audio_codec_param, 0x55);
    a_format3.addWildcard(Format::sample_rate_param);
    
    /* MPEG audio with sample rate and bitrate */
    Format a_format4(Format::audio_mpeg_type);
    a_format4.addIntParamValue(Format::audio_codec_param, 0x55);
    a_format4.addWildcard(Format::sample_rate_param);
    a_format4.addWildcard(Format::bitrate_param);

    /* AC3 audio without bitrate */    
    Format a_format5(Format::audio_ac3_type);

    /* AC3 audio with bitrate */    
    Format a_format6(Format::audio_ac3_type);
    a_format6.addWildcard(Format::bitrate_param);
    
    /* AC3 audio with bitrate and extra parameters */
    Format a_format7(Format::audio_ac3_type);
    a_format7.addWildcard(Format::channels_param);
    a_format7.addWildcard(Format::sample_rate_param);
    a_format7.addWildcard(Format::bitrate_param);
    a_format7.addWildcard(Format::speaker_setup_param);
    a_format7.addIntParamValue(Format::lfe_param, 0);
    a_format7.addIntParamValue(Format::lfe_param, 1);

    /* raw audio with endian */
    Format a_format8(Format::audio_raw_type);
    a_format8.addWildcard(Format::channels_param);
    a_format8.addWildcard(Format::sample_rate_param);
    a_format8.addWildcard(Format::bitpersample_param);
    a_format8.addStringParamValue(Format::endian_param, "little");
    a_format8.addStringParamValue(Format::encoding_param, "linear");
    a_format8.addStringParamValue(Format::sign_param, "unsigned");
    a_format8.addStringParamValue(Format::sign_param, "signed");

    /* raw audio without endian */
    Format a_format9(Format::audio_raw_type);
    a_format9.addWildcard(Format::channels_param);
    a_format9.addWildcard(Format::sample_rate_param);
    a_format9.addWildcard(Format::bitpersample_param);
    a_format9.addStringParamValue(Format::encoding_param, "linear");
    a_format9.addStringParamValue(Format::sign_param, "unsigned");
    a_format9.addStringParamValue(Format::sign_param, "signed");

    /* End< Audio Subformats */
    
    /* Begin> Video Input Formats */
    Format* in_v_format1 = new Format(v_format1);
    in_v_format1->setJackTag("video");
    getOwnInputProperty()->addNewFormat(in_v_format1);
    getOwnInputProperty()->setDefaultFormat(in_v_format1, "video");
    Format* in_v_format2 = new Format(v_format2);
    in_v_format2->setJackTag("video");
    getOwnInputProperty()->addNewFormat(in_v_format2);
    /* End< Video Input Formats */
    
    /* Begin> Audio Input Formats */
    Format* in_a_format1 = new Format(a_format1);
    in_a_format1->setJackTag("audio");
    getOwnInputProperty()->addNewFormat(in_a_format1);
    getOwnInputProperty()->setDefaultFormat(in_a_format1, "audio");
    Format* in_a_format2 = new Format(a_format2);
    in_a_format2->setJackTag("audio");
    getOwnInputProperty()->addNewFormat(in_a_format2);
    Format* in_a_format3 = new Format(a_format3);
    in_a_format3->setJackTag("audio");
    getOwnInputProperty()->addNewFormat(in_a_format3);
    Format* in_a_format4 = new Format(a_format4);
    in_a_format4->setJackTag("audio");
    getOwnInputProperty()->addNewFormat(in_a_format4);
    Format* in_a_format5 = new Format(a_format5);
    in_a_format5->setJackTag("audio");
    getOwnInputProperty()->addNewFormat(in_a_format5);
    Format* in_a_format6 = new Format(a_format6);
    in_a_format6->setJackTag("audio");
    getOwnInputProperty()->addNewFormat(in_a_format6);
    Format* in_a_format7 = new Format(a_format7);
    in_a_format7->setJackTag("audio");
    getOwnInputProperty()->addNewFormat(in_a_format7);
    Format* in_a_format8 = new Format(a_format8);
    in_a_format8->setJackTag("audio");
    getOwnInputProperty()->addNewFormat(in_a_format8);
    Format* in_a_format9 = new Format(a_format9);
    in_a_format9->setJackTag("audio");
    getOwnInputProperty()->addNewFormat(in_a_format9);
    /* End< Audio Input Formats */
    
    /* Begin> Composite Output Formats */
    /* Format with audio and video */
    Format* out_format1 = getOwnOutputProperty()->addNewFormat(Format::av_avi_type);
    out_format1->addFPointerParamValue(Format::video_format_param, new Format(v_format1));
    out_format1->addFPointerParamValue(Format::video_format_param, new Format(v_format2));
    out_format1->addFPointerParamValue(Format::audio_format_param, new Format(a_format1));
    out_format1->addFPointerParamValue(Format::audio_format_param, new Format(a_format2));
    out_format1->addFPointerParamValue(Format::audio_format_param, new Format(a_format3));
    out_format1->addFPointerParamValue(Format::audio_format_param, new Format(a_format4));
    out_format1->addFPointerParamValue(Format::audio_format_param, new Format(a_format5));
    out_format1->addFPointerParamValue(Format::audio_format_param, new Format(a_format6));
    out_format1->addFPointerParamValue(Format::audio_format_param, new Format(a_format7));
    out_format1->addFPointerParamValue(Format::audio_format_param, new Format(a_format8));
    out_format1->addFPointerParamValue(Format::audio_format_param, new Format(a_format9));
    getOwnInputProperty()->setDefaultFormat(out_format1);
    
    /* Format with video only */
    Format* out_format2 = getOwnOutputProperty()->addNewFormat(Format::av_avi_type);
    out_format2->addFPointerParamValue(Format::video_format_param, new Format(v_format1));
    out_format2->addFPointerParamValue(Format::video_format_param, new Format(v_format2));
    /* End< Composite Output Formats */
    
    return SUCCESS;
  }
  
  Result AVMuxNode::doInitOutput() {
    
    /* Get input formats.
     * If the audio jack is not connected, the output format is without audio.
     * Also disable synchronization if there is only one stream.
     * The video jack must be connected.
     */
    const Format* in_v_format = getInputFormat("video");
    const Format* in_a_format = getInputFormat("audio");
    if (in_a_format == NULL) {
      DEBUG_STREAM("No audio stream. Disabling synchronization" << endl);
      sync_mode = false;
    }
    if (in_v_format == NULL) {
      ERROR_STREAM("No video stream. Not supported" << endl);
      return FAILURE;
    }
    
    /* init working input property. */
    Format* i_v_format = NULL;
    Format* i_a_format = NULL;
    if (in_v_format) {
      i_v_format = new Format(*in_v_format);
      getOwnWorkingInputProperty()->addNewFormat(i_v_format);
      i_v_format->setJackTag("video");
      getOwnWorkingInputProperty()->setDefaultFormat(i_v_format, "video");
    }
    if (in_a_format) {
      i_a_format = new Format(*in_a_format);
      getOwnWorkingInputProperty()->addNewFormat(i_a_format);
      i_a_format->setJackTag("audio");
      getOwnWorkingInputProperty()->setDefaultFormat(i_a_format, "audio");
    }
    
    /* Set output format.
     * Simply set the format values here.
     * This is correct since we really just pass the buffers through unaltered.
     */
    Format* format = getOwnWorkingOutputProperty()->addNewFormat(Format::av_avi_type);
    if (i_v_format) {
      format->addFPointerParamValue(Format::video_format_param, new Format(*i_v_format));
    }
    if (i_a_format) {
      format->addFPointerParamValue(Format::audio_format_param, new Format(*i_a_format));
    }
    getOwnWorkingOutputProperty()->setDefaultFormat(format);
    
    /* re-enable discard_mode_running (only used if discard_mode is also set to true) 
     * and reset related status variables
     */
    discard_mode_running =  true;
    first_audio_buffer = false;
    first_video_buffer = false;
    audio_too_late = true;      
    set_audio_too_late = false;      

    return SUCCESS;
  }

  Message* AVMuxNode::processBuffer(Buffer* in_buffer) {
    MutexGuard mg_om(mutex);
    
    /* Check if we got a new buffer. */
    if (!getWorkingFlag()) {
      /* Set the buffer label. */
      if(strcmp(getCurrentRecvInputStream(), "audio") == 0) {
        in_buffer->setLabel(IAVMux::LABEL_AUDIO);
	
	/* store timestamp sync time of first audio buffer */
	if(discard_mode && !first_audio_buffer) {
	  first_audio_buffer = true;
	  first_audio_buffer_sync_time = in_buffer->getTimestamp().sync_time;
	}

      } else if(strcmp(getCurrentRecvInputStream(), "video") == 0) {
        in_buffer->setLabel(IAVMux::LABEL_VIDEO);

	/* store timestamp sync time of first video buffer */
	if(discard_mode && !first_video_buffer) {
	  first_video_buffer = true;
	  first_video_buffer_sync_time = in_buffer->getTimestamp().sync_time;
	}

      } else {
        /* This case is just for completeness and should never happen. */
        ERROR_STREAM2("THIS IS A BUG. THIS SHOULD NEVER HAVE HAPPENED" << endl);
        in_buffer->release();
        return NULL;
      }
    } else {
      /* If you did send an event last time, the incoming buffer
       * was remembered and has to be processed now.
       */
      setWorkingFlag(false);
      assert(in_buffer == 0);
      in_buffer = last_buffer;
      last_buffer = NULL;
    }
    
    /* Clear the event sending flag. */
    if (send_event) {
      send_event = false;
    }
    
    /* Send remembered endTrack_event after flushing sync. */
    if (sync_end_track && !sync_flush) {
      sync_end_track = false;
      last_buffer = in_buffer;
      setWorkingFlag(true);
      return new CEvent(Event(ITrack::endTrack_event));
    }
    
    // XXX: what to do with invalid timestamps ?
    
    /* Send buffers remembered for synchronization. */
    if (sync_flush) {
      /* Send remembered audio buffer. */
      if (last_audio) {
        last_buffer = in_buffer;
        setWorkingFlag(true);
        in_buffer = last_audio;
        last_audio = NULL;
        return in_buffer;
      }
      
      /* Send remembered video buffer. */
      if (last_video) {
        last_buffer = in_buffer;
        setWorkingFlag(true);
        in_buffer = last_video;
        last_video = NULL;
        return in_buffer;
      }
      
      /* Flushing done, all buffers sent. */
      sync_flush = false;
      
      /* Set working flag to send endTrack_event after this buffer. */
      if (sync_end_track) {
        last_buffer = NULL;
        setWorkingFlag(true);
      }
      
      /* Pass through in_buffer so that remembered buffers are not set again. */
      return in_buffer;
    }
    
    /* Synchronize audio and video. */
    if (sync_mode) {
      
      /* discard mode should also be enabled when sync mode is enabled */
      if(discard_mode && discard_mode_running) {

	/* still need to determine which stream is too late */
	if(!set_audio_too_late) { 
	  /* if first audio and video buffer was received -> check sync times */
	  if(first_audio_buffer && first_video_buffer) {
	    set_audio_too_late = true;
	    if(first_audio_buffer_sync_time < first_video_buffer_sync_time) {
	      audio_too_late = true;
	    }
	    else {
	      audio_too_late = false;
	    }
	  }
	  else { /* did not yet set audio_too_late -> discard in_buffer and switch inputs */
	    in_buffer->release();

	    if (in_buffer->getLabel() == IAVMux::LABEL_AUDIO) {
	      setRecvInputStreamEnabled("audio", false);
	      setRecvInputStreamEnabled("video", true);
	    }
	    else {
	      setRecvInputStreamEnabled("audio", true);
	      setRecvInputStreamEnabled("video", false);
	    }
	    return 0;
	  }
	}
	else { /* already determined which stream is too late -> either discard too late audio or video buffers */
	  if(audio_too_late) {
	    if (in_buffer->getLabel() == IAVMux::LABEL_AUDIO) {
	      if(in_buffer->getTimestamp().sync_time < first_video_buffer_sync_time) {
		in_buffer->release();
		return 0;
	      }
	      else { /* audio not too late anymore! */
		discard_mode_running = false;
	      }
	    }
	  }
	  else {
	    if (in_buffer->getLabel() == IAVMux::LABEL_VIDEO) {
	      if(in_buffer->getTimestamp().sync_time < first_audio_buffer_sync_time) {
		in_buffer->release();
		return 0;
	      }
	      else { /* video not too late anymore! */
		discard_mode_running = false;
	      }
	    }
	  }
	} 
      }

      if (in_buffer->getLabel() == IAVMux::LABEL_AUDIO) {
        
        if (last_audio) {
          /* Never happens, just for debugging. */
          ERROR_STREAM2("BUG: An audio frame was received but an audio frame is already waiting!");
          in_buffer->release();
          in_buffer = NULL;
          return NULL;
        }
        
        /* Check if audio frame is newer than the waiting video frame.
         * if no video frame is waiting, wait for a video frame and send no buffer.
         */
        if (!last_video || in_buffer->getTimestamp().sync_time >= last_video->getTimestamp().sync_time) {
          
          /* Remember the audio frame and block audio queue. */
          last_audio = in_buffer;
          setRecvInputStreamEnabled("audio", false);
          
          /* Send the video frame that is waiting and wait for a new one. */
          in_buffer = last_video;
          last_video = NULL;
          setRecvInputStreamEnabled("video", true);
          
        } else {
          
          /* Send the audio frame and wait for a new one. */
          
        }
        
      } else if (in_buffer->getLabel() == IAVMux::LABEL_VIDEO) {

        if (last_video) {
          /* Never happens, just for debugging. */
          ERROR_STREAM("BUG: A video frame was received but a video frame is already waiting!");
          in_buffer->release();
          in_buffer = NULL;
          return NULL;
        }
        
        /* Check if video frame is newer than the waiting audio frame.
         * If no audio frame is waiting, wait for an audio frame and send no buffer.
         */
        if (!last_audio || in_buffer->getTimestamp().sync_time >= last_audio->getTimestamp().sync_time) {
          
          /* Remember the video frame and block video queue. */
          last_video = in_buffer;
          setRecvInputStreamEnabled("video", false);
          
          /* Send the audio frame that is waiting and wait for a new one. */
          in_buffer = last_audio;
          last_audio = NULL;
          setRecvInputStreamEnabled("audio", true);
          
        } else {
          
          /* Send the audio frame and wait for a new one. */
          
        }
        
      } else {
        
        /* Should never happen, just for debugging. */
        ERROR_STREAM2("BUG! IN_BUFFER NOT LABELED" << endl);
        in_buffer->release();
        in_buffer = NULL;
        
      }
      
    }
    
    return in_buffer;
  }
  
  Result AVMuxNode::processCEvent(CEvent* event) {
    if ( !strcmp(getCurrentRecvInputStream(), "audio") ) {
      event->setLabel(IAVMux::LABEL_AUDIO);
    } else if ( !strcmp(getCurrentRecvInputStream(), "video") ) {
      event->setLabel(IAVMux::LABEL_VIDEO);
    } else {
      ERROR_STREAM("BUG! Current send input stream not set!" << endl);
    }
    
    return SUCCESS;
  }
  
  void AVMuxNode::setSyncMode(const nmm_bool& value) {
    bool old_sync_mode = sync_mode;
    sync_mode = value;
    discard_mode = value;
    
    /* make sure, remembered buffers are sent when we turn off sync mode while running */
    if (isStarted() && !sync_mode && old_sync_mode) {
      setRecvInputStreamEnabled("video", true);
      setRecvInputStreamEnabled("audio", true);
      sync_flush = true;
    }
  }
  
  Result AVMuxNode::endTrack() {
    MutexGuard mg_om(mutex);
    
    if (!strcmp(getCurrentRecvInputStream(), "audio")) {
      audio_end_track = true;
    } else if (!strcmp(getCurrentRecvInputStream(), "video")) {
      video_end_track = true;
    }
    
    /* check if we still have to wait for syncReset on other stream */
    if ((!audio_end_track && getInputFormat("audio") != NULL) || !video_end_track) {
      
      /* make sure both streams are enabled and disable sync_mode */
      setRecvInputStreamEnabled("video", true);
      setRecvInputStreamEnabled("audio", true);
      sync_flush = true;
      sync_mode = false;
      setWorkingFlag(true);
      
      return DELETE;
    }
    
    audio_end_track = false;
    video_end_track = false;
    
    if (sync_mode && (last_audio || last_video)) {
      sync_flush = true;
      sync_end_track = true;
      setWorkingFlag(true);
      return DELETE;
    }
    
    return SUCCESS;
  }
  
  Result AVMuxNode::started(const Interval&) {
    // XXX: this event makes no sense anyway at this point so delete it
    return DELETE;
  }
  
  Result AVMuxNode::finished() {
    // XXX: this event makes no sense anyway at this point so delete it
    return DELETE;
  }
  
  Result AVMuxNode::startTrack(const string&) {
    MutexGuard mg_om(mutex);
    // This is a special case for our Multiplexer:
    // we count how many events we receive, and delete all of them
    // but the first one
    // 2 should be replaced by number of InputJacks to be more generic
    track_start_count++;
    if (track_start_count > 1) {
      DEBUG_STREAM("Delete startTrack-event number " << track_start_count << endl); 
      return DELETE;
    }
    DEBUG_STREAM("Let startTrack-event number " << track_start_count << " pass." << endl); 
    return SUCCESS;
  }
  
}

#include "nmm/nmm_suffix.hpp"
