/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <vorbis/codec.h>

#include "nmm/comm/format/Property.hpp"
#include "nmm/comm/messaging/Buffer.hpp"
#include "nmm/multimedia/Plugin.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"
#include "nmm/plugins/interfaces/IKeyframe.hpp"

#include "OGMDemuxNode.hpp"

NMM_REGISTER_PLUGIN(NMM::OGMDemuxNode, NMM, "OGMDemuxNode")

namespace NMM {
  
  enum {
    ACVORBIS = 0xFFFF,
    ACPCM    = 0x0001,
    ACMP3    = 0x0055,
    ACAC3    = 0x2000
  };
  
  OGMDemuxNode::OGMDemuxNode(const char *name, StreamQueue::Mode qmode, int qsize) :
    GenericDemultiplexerNode(name, qmode, qsize),
    saved_buffer(NULL),
    saved_out_buffer(NULL),
    state(STATE_NEED_DATA),
    initial_packet(false),
    initial_page(false),
    analysing_done(false),
    current_stream(NULL),
    skip_until_syncpoint(false)
  {}
  
  OGMDemuxNode::~OGMDemuxNode() {}
  
  Result OGMDemuxNode::doInit() {
    Format*   format;
    
    // setup input property
    format = getOwnInputProperty()->addNewFormat("av/ogm");
    getOwnInputProperty()->setDefaultFormat(format, "default");
    
    syncReset();
    
    return SUCCESS;
  }
  
  Result OGMDemuxNode::doInitOutput() {
    ogg_sync_init(&sync);
    
    state = STATE_NEED_DATA;
    saved_out_buffer = NULL;
    
    analysing_done = false;
    if (!analyseData()) {
      ERROR_STREAM("analyseData() failed!" << endl);
      getWorkingInputProperty()->erase();
      return FAILURE;
    }
    
    return SUCCESS;
  }
  
  Result OGMDemuxNode::doDeinitOutput() {
    video_streams.clear();
    audio_streams.clear();
    
    return SUCCESS;
  }
  
  Message* OGMDemuxNode::processBuffer( Buffer* in_buffer ) {
    
    Message* out = NULL;
    
    /* Continue processing last buffer */
    if (getWorkingFlag()) {
      setWorkingFlag(false);
      in_buffer = saved_buffer;
      saved_buffer = NULL;
    }
    
    /* Send saved buffer */
    if (saved_out_buffer) {
      out = saved_out_buffer;
      saved_out_buffer = NULL;
      
      setWorkingFlag(true);
      saved_buffer = in_buffer;
      
      setCurrentSendOutputStream(current_stream->stream_name.c_str());
      return out;
    }
    
    switch(state) {
    case STATE_HAVE_PAGE: /* Read new packet and handle it */
      if (ogg_stream_packetout(&current_stream->instate, &pack) == 1) {
        if (initial_packet) {
          initial_packet = false;
          out = handleHeaderPacket();
        } else {
          out = handlePacket();
        }
      } else {
        state = STATE_NEED_DATA;
      }
      break;
      
    case STATE_NEED_DATA: /* Need more data to get a complete page */
      if (ogg_sync_pageseek(&sync, &page) <= 0) {
        if (in_buffer) {
          /* read data from in_buffer */
          char* buf = ogg_sync_buffer(&sync, in_buffer->getUsed()); // XXX: in_buffer may be too large !!
          int nread = in_buffer->getUsed();
          memcpy(buf, in_buffer->getData(), nread);
          in_buffer->release();
          in_buffer = NULL;
          ogg_sync_wrote(&sync, nread);
        } else {
          /* no more data available (working loop ends) */
          return NULL;
        }
      } else {
        /* check if initial page */
        initial_page = ogg_page_bos(&page);
        
        /* handle page */
        out = handlePage();
        
        /* check if analysing done */
        if (getAnalysingFlag() && !initial_page) {
          setAnalysingFlag(false);
        }
      }
      break;
      
    case STATE_WRITE_PAGES: /* Write out OGG pages to current stream */
      {
        ogg_page page;
        
        if (ogg_stream_pageout(&current_stream->outstate, &page)) {
          Buffer* buffer = getNewBuffer(page.header_len + page.body_len);
          buffer->setUsed(page.header_len + page.body_len);
          memcpy(buffer->getData(), page.header, page.header_len);
          memcpy(buffer->getData() + page.header_len, page.body, page.body_len);
          setCurrentSendOutputStream(current_stream->stream_name.c_str());
          out = buffer;
        } else {
          state = last_state;
        }
      }
      break;
      
    case STATE_FLUSH_PAGES: /* Write out OGG pages to current stream and flush */
      {
        ogg_page page;
        
        if (ogg_stream_pageout(&current_stream->outstate, &page)) {
          Buffer* buffer = getNewBuffer(page.header_len + page.body_len);
          buffer->setUsed(page.header_len + page.body_len);
          memcpy(buffer->getData(), page.header, page.header_len);
          memcpy(buffer->getData() + page.header_len, page.body, page.body_len);
          setCurrentSendOutputStream(current_stream->stream_name.c_str());
          out = buffer;
        } else {
          state = last_state;
        }
      }
      break;
      
    case STATE_END_TRACK:
      {
        /* Find next unfinished  stream */
        for (unsigned int i=0; i<audio_streams.size(); ++i) {
          current_stream = &audio_streams[i];
          
          if (!current_stream->end_track) {
            if (current_stream->is_vorbis && current_stream->eos) {
              pack.b_o_s = 0;
              pack.e_o_s = 1;
              pack.packet = NULL;
              pack.bytes = 0;
              pack.granulepos = current_stream->max_granulepos;
              pack.packetno = current_stream->packetno;
              ogg_stream_packetin(&current_stream->outstate, &pack);
              current_stream->eos = true;
            } else if (current_stream->is_vorbis && !current_stream->flushed) {
              last_state = state;
              state = STATE_FLUSH_PAGES;
              current_stream->flushed = true;
            } else {
              current_stream->end_track = true;
              setCurrentSendOutputStream(current_stream->stream_name.c_str());
              out = new CEvent( ITrack::create_endTrack() );
            }
            break;
          }
        }
        break;
      }
    }
    
    /* continue until buffer has been completely processed */
    setWorkingFlag(true);
    saved_buffer = in_buffer;
    return out;
  }
  
  Message* OGMDemuxNode::handlePage() {
    int sno = ogg_page_serialno(&page);
    
    Message* out = NULL;
    
    current_stream = NULL;
    initial_packet = false;
    
    if (!getAnalysingFlag() || !initial_page) {
      
      /* look up stream */
      Stream* stream = getStreamBySerial(sno);
      if (stream == NULL) {
        DEBUG_STREAM("Encountered page for an unknown serial: " << sno << endl);
        state = STATE_NEED_DATA;
        current_stream = NULL;
      } else {
        ogg_stream_pagein(&stream->instate, &page);
        current_stream = stream;
        state = STATE_HAVE_PAGE;
      }
      
    } else {
      
      /* set up analysing stream */
      analysing_stream.sno = sno;
      analysing_stream.stream_name = "analysing";
      analysing_stream.skip_until_valid_ts = false;
      analysing_stream.is_begin = true;

      current_stream = &analysing_stream;
      
      /* init stream */
      if (ogg_stream_init(&analysing_stream.instate, sno)) {
        ERROR_STREAM("ogg_stream_init failed" << endl);
        state = STATE_NEED_DATA;
      } else {
        ogg_stream_pagein(&analysing_stream.instate, &page);
        state = STATE_HAVE_PAGE;
        initial_packet = true;
      }
    }
    
    return out;
  }
  
  Buffer* OGMDemuxNode::createRawOutputBuffer(const char* data, ssize_t size) {
    Buffer* out = getNewBuffer(size);
    out->setUsed(size);
    memcpy(out->getData(), data, size);
    
    timestamp.stream_counter = current_stream->stream_counter++;
    out->setTimestamp(timestamp);
    return out;
  }
  
  Message* OGMDemuxNode::handlePacket() {
    if (pack.e_o_s) {
      current_stream->eos = true;
      pack.e_o_s = 1;
    }
    
    /* generate timestamp */
    Rational ts;
    timestamp.is_valid = false;
    if (pack.granulepos >= 0) {
      ts = Rational(pack.granulepos) / current_stream->sample_rate;
      if (ts > current_stream->highest_ts) { current_stream->highest_ts = ts; }
      timestamp.is_valid = true;
      DEBUG_STREAM("Stream: " << current_stream->stream_name << ": create timestamp: " << timestamp << endl);
    } else if (current_stream->is_begin) {
      current_stream->is_begin = false;
      ts = Rational(0);
      current_stream->highest_ts = ts;
      timestamp.is_valid = true;
      DEBUG_STREAM("Stream: " << current_stream->stream_name << ": create initial timestamp: " << timestamp << endl);
    } else {
      /* use last timestamp */
      ts = current_stream->highest_ts;
      timestamp.is_valid = false;
      DEBUG_STREAM("Stream: " << current_stream->stream_name << ": use last timestamp: " << timestamp << endl);
    }
    timestamp.sync_time = Time(ts);
    
    /* get length of packet header */
    int hdrlen = (*pack.packet & PACKET_LEN_BITS01) >> 6;
    hdrlen |= (*pack.packet & PACKET_LEN_BITS2) << 1;
    
    if ((*pack.packet & 3) == PACKET_TYPE_HEADER) {
      current_stream->is_begin = true;
    }
    
    /* handle packet, depending on type */
    switch(current_stream->stream_type) {
    case Stream::VIDEO:
      {
        if (((*pack.packet & 3) == PACKET_TYPE_HEADER) ||
            ((*pack.packet & 3) == PACKET_TYPE_COMMENT)) {
          DEBUG_STREAM("Stream: " << current_stream->stream_name << ": Ignoring header or comment packet" << endl);
          return NULL;
        }
        
        bool is_syncpoint = (0 != (pack.packet[0] & PACKET_IS_SYNCPOINT));
        
        if (skip_until_syncpoint) {
          if (!is_syncpoint) {
            DEBUG_STREAM("Skipping non-keyframe!" << endl);
            return NULL;
          } else {
            DEBUG_STREAM("Found KeyFrame, stop skipping!" << endl);
            if (!isOutputConnected(current_stream->stream_name)) {
              DEBUG_STREAM("Found key_frame, but according output is not connected! Continuing skip!" << endl);
              return NULL;
            }
            if (!timestamp.is_valid) {
              DEBUG_STREAM("Found key_frame, but its Timestamp is invalid! Continuing skip!" << endl);
              return NULL;
            }
            skip_until_syncpoint = false;
            DEBUG_STREAM("Taking KeyFrame #" << current_stream->stream_counter
                         << " - " << current_stream->stream_name << " (" << timestamp.sync_time << ") "
                         << pack.granulepos << " " << " " << is_syncpoint << endl);
          }
        }
        
        Buffer* out = createRawOutputBuffer((char*) (pack.packet + 1 + hdrlen), pack.bytes - 1 - hdrlen);
        
        if (is_syncpoint) {
          saved_out_buffer = out;
          setCurrentSendOutputStream(current_stream->stream_name.c_str());
          return new CEvent( IKeyframe::create_setKeyframe() );
        }
        
        setCurrentSendOutputStream(current_stream->stream_name.c_str());
        return out;
      }
      
    case Stream::AUDIO:
      {
        if (skip_until_syncpoint) {
          DEBUG_STREAM("Dropping Audio while searching KeyFrame!" << endl);
          return NULL;
        }
        
        if (current_stream->skip_until_valid_ts) {
          if (!timestamp.is_valid) {
            DEBUG_STREAM("Dropping Audio with invalid Timestamp after searching KeyFrame!" << endl);
            return NULL;
          }
          current_stream->skip_until_valid_ts = false;
        }
        
        if (current_stream->is_vorbis) {
          /* re-pack vorbis audio into ogg packets */
          current_stream->max_granulepos = (pack.granulepos > current_stream->max_granulepos ?
                                            pack.granulepos : current_stream->max_granulepos);
          if ((current_stream->packetno == 0) || (current_stream->packetno == 2)) {
            ogg_stream_packetin(&current_stream->outstate, &pack);
            last_state = state;
            state = STATE_FLUSH_PAGES;
          } else {
            ogg_stream_packetin(&current_stream->outstate, &pack);
            last_state = state;
            state = STATE_WRITE_PAGES;
          }
          current_stream->packetno++;
          
          return NULL;
        } else {
          if (((*pack.packet & 3) == PACKET_TYPE_HEADER) ||
              ((*pack.packet & 3) == PACKET_TYPE_COMMENT)) {
            DEBUG_STREAM("Stream: " << current_stream->stream_name << ": Ignoring header or comment packet" << endl);
            return NULL;
          }
          
          setCurrentSendOutputStream(current_stream->stream_name.c_str());
          return createRawOutputBuffer((char*) (pack.packet + 1 + hdrlen), pack.bytes - 1 - hdrlen);
        }
      }
      
    default:
      return NULL;
    } 
  }
  
  Message* OGMDemuxNode::handleHeaderPacket() {
    
    /* check stream type */
    if ((pack.bytes >= 7) && !strncmp((const char*) &pack.packet[1], "vorbis", 6)) {
      
      vorbis_info       vi;
      vorbis_comment    vc;
      vorbis_info_init(&vi);
      vorbis_comment_init(&vc);
      if (vorbis_synthesis_headerin(&vi, &vc, &pack) >= 0) {
        current_stream = addAudioStream(analysing_stream.sno, &analysing_stream.instate);
        current_stream->sample_rate = Rational(vi.rate);
        current_stream->is_vorbis = true;
        
        /* special ogg/vorbis stuff */
        current_stream->max_granulepos = 0;
        current_stream->packetno = 0;
        ogg_stream_init(&current_stream->outstate, 23);
        
        addOutputStream(current_stream->stream_name.c_str());
        
        Format* format = getOwnWorkingOutputProperty()->addNewFormat(Format::audio_x_ogg_type);
        format->setJackTag(current_stream->stream_name);
        
        getOwnWorkingOutputProperty()->setDefaultFormat(format, current_stream->stream_name);
      } else {
        DEBUG_STREAM("Can't get Vorbis stream information for stream with serial " << analysing_stream.sno << endl);
      }
      
    } else if ((pack.bytes >= 142) &&
               !strncmp((const char*) &pack.packet[1], "Direct Show Samples embedded in Ogg",
                        35) ) {
      
      if ((*(int32_t*) (&pack.packet[96]) == 0x05589f80) && (pack.bytes >= 184)) {
        DEBUG_STREAM("Found old video header. Not supported" << endl);
      } else if (*(int32_t*) (&pack.packet[96]) == 0x05589F81) {
        DEBUG_STREAM("Found old audio header. Not supported" << endl);
      } else {
        DEBUG_STREAM("OGG stream " << analysing_stream.sno << " has an old header with an unknown type." << endl);
      }
      
    } else if (((*pack.packet & PACKET_TYPE_BITS ) == PACKET_TYPE_HEADER) &&
               (pack.bytes >= (int)(sizeof(stream_header) + 1 - sizeof(int)))) {
      
      stream_header *sth = (stream_header *)(pack.packet + 1);
      if (!strncmp(sth->streamtype, "video", 5)) {
        unsigned long codec;
        char ccodec[5];
        strncpy(ccodec, sth->subtype, 4);
        ccodec[4] = 0;
        codec = (sth->subtype[0] << 24) + (sth->subtype[1] << 16) + (sth->subtype[2] << 8) + sth->subtype[3]; 
        DEBUG_STREAM("Found video bitstream" << endl);
        
        current_stream = addVideoStream(analysing_stream.sno, &analysing_stream.instate);
        current_stream->sample_rate = Rational(10000000, sth->time_unit);
        current_stream->sample_rate.reduce();
        
        addOutputStream(current_stream->stream_name.c_str());
        
        Format* format = getOwnWorkingOutputProperty()->addNewFormat("video/mpeg4");
        format->addIntParamValue(Format::x_resolution_param, sth->sh.video.width);
        format->addIntParamValue(Format::y_resolution_param, sth->sh.video.height);
        format->addIntParamValue(Format::ratio_x_param, sth->sh.video.width);
        format->addIntParamValue(Format::ratio_y_param, sth->sh.video.height);
        format->addFloatParamValue(Format::framerate_param, (float)10000000 / (float)sth->time_unit);
        format->addStringParamValue(Format::compressor_param, ccodec);
        
        format->setJackTag(current_stream->stream_name);
        getOwnWorkingOutputProperty()->setDefaultFormat(format, current_stream->stream_name);
        
      } else if (!strncmp(sth->streamtype, "audio", 5)) {
        
        int codec;
        char buf[5];
        memcpy(buf, sth->subtype, 4);
        buf[4] = 0;
        codec = strtoul(buf, NULL, 16);
        DEBUG_STREAM("Found audio bitstream" << endl);
        
        current_stream = addAudioStream(analysing_stream.sno, &analysing_stream.instate);
        current_stream->sample_rate = Rational(sth->samples_per_unit /** sth->sh.audio.channels*/);
        
        Format* format = NULL;
        if (codec == ACPCM) {
          format = getOwnWorkingOutputProperty()->addNewFormat("audio/raw");
          format->addIntParamValue(Format::sample_rate_param, sth->samples_per_unit);
          format->addIntParamValue(Format::channels_param, sth->sh.audio.channels);
          format->addIntParamValue(Format::bitpersample_param, sth->bits_per_sample);
          format->addStringParamValue(Format::endian_param, "little");
          format->addStringParamValue(Format::encoding_param, "linear");
          format->addStringParamValue(Format::sign_param, "signed");
        } else if (codec == ACMP3) {
          format = getOwnWorkingOutputProperty()->addNewFormat("audio/mpeg");
          format->addIntParamValue(Format::audio_codec_param, 0x55);
          format->addIntParamValue(Format::layer_param, 3);
          format->addIntParamValue(Format::sample_rate_param, sth->samples_per_unit);
          format->addIntParamValue(Format::channels_param, sth->sh.audio.channels);
          format->addIntParamValue(Format::bitpersample_param, sth->bits_per_sample);
        } else if (codec == ACAC3) {
          format = getOwnWorkingOutputProperty()->addNewFormat("audio/ac3");
        } else if (codec == ACVORBIS) {
          // XXX: fw: should not happen
        }
        
        addOutputStream(current_stream->stream_name.c_str());
        
        if (format) {
          format->setJackTag(current_stream->stream_name);
          getOwnWorkingOutputProperty()->setDefaultFormat(format, current_stream->stream_name);
        }
        
      } else {
        
        DEBUG_STREAM("found new header of unknown/unsupported type" << endl);
        
      }
      
    } else {
      
      DEBUG_STREAM("OGG stream " << analysing_stream.sno << " is of an unknown type (bad header?)" << endl);
      
    }
    
    return NULL;
  }
  
  OGMDemuxNode::Stream* OGMDemuxNode::getStreamBySerial(int sno) {
    for (unsigned int i=0; i<video_streams.size(); ++i) {
      if (video_streams[i].sno == sno) { return &video_streams[i]; }
    }
    for (unsigned int i=0; i<audio_streams.size(); ++i) {
      if (audio_streams[i].sno == sno) { return &audio_streams[i]; }
    }
    
    return NULL; // XXX: throw exception ?
  }
  
  OGMDemuxNode::Stream* OGMDemuxNode::addVideoStream(int sno, ogg_stream_state* sstate) {
    int stream_no =  video_streams.size();
    Stream stream;
    stream.sno = sno;
    stream.stream_counter = 0;
    stream.instate = *sstate;
    stream.highest_ts = 0;
    stream.eos = false;
    stream.flushed = false;
    stream.end_track = false;
    stream.is_vorbis = false;
    stream.stream_name = string("video") + stream_no;
    stream.stream_type = Stream::VIDEO;
    stream.skip_until_valid_ts = false;
    stream.is_begin = true;
    DEBUG_STREAM("Stream: " << stream.stream_name << ": Added" << endl);
    video_streams.push_back(stream);
    return &video_streams[stream_no];
  }
  
  OGMDemuxNode::Stream* OGMDemuxNode::addAudioStream(int sno, ogg_stream_state* sstate) {
    int stream_no =  audio_streams.size();
    Stream stream;
    stream.sno = sno;
    stream.stream_counter = 0;    
    stream.instate = *sstate;
    stream.highest_ts = 0;
    stream.eos = false;
    stream.flushed = false;
    stream.end_track = false;
    stream.is_vorbis = false;
    stream.stream_name = string("audio") + stream_no;
    stream.stream_type = Stream::AUDIO;
    stream.skip_until_valid_ts = false;
    stream.is_begin = true;
    DEBUG_STREAM("Stream: " << stream.stream_name << ": Added" << endl);
    audio_streams.push_back(stream);
    return &audio_streams[stream_no];
  }

  Result OGMDemuxNode::doFlush() {
    skip_until_syncpoint = true;
    for (unsigned int v = 0; v < video_streams.size(); v++) {
      video_streams[v].skip_until_valid_ts = true;
    }
    for (unsigned int a = 0; a < audio_streams.size(); a++) {
      audio_streams[a].skip_until_valid_ts = true;
    }
    if (getWorkingFlag()) {
      setWorkingFlag(false);
      if (saved_buffer) {
        saved_buffer->release();
        saved_buffer = NULL;
      }
    }
    if (saved_out_buffer) {
      saved_out_buffer->release();
      saved_out_buffer = NULL;
    }
    state = STATE_NEED_DATA;
    return SUCCESS;
  }
  
  /******************** ITrack interface ********************/
  
  Result OGMDemuxNode::startTrack(const string& /*filename*/) {
    return SUCCESS;
  }
  
  Result OGMDemuxNode::endTrack() {
    state = STATE_END_TRACK;
    saved_buffer = NULL;
    setWorkingFlag(true);
    
    return DELETE;
  }
  
  Result OGMDemuxNode::finished() {
    return SUCCESS;
  }
  
  /******************** ITrack interface ********************/
  
  /******************** ISyncReset interface ********************/
  
  Result OGMDemuxNode::syncReset() {
    ogg_sync_init(&sync);
    DEBUG_STREAM("sync_reset completed." << endl);
    
    return SUCCESS;
  };
  
  /******************** ISyncReset interface ********************/

}// namespace NMM
