/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 2 of the License, or 
 * (at your option) any later version. 
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#ifndef NMM_OGMDEMUXNODE_HPP
#define NMM_OGMDEMUXNODE_HPP NMM_OGMDEMUXNODE_HPP

#include <ogg/ogg.h>

#include "nmm/multimedia/GenericDemultiplexerNode.hpp"
#include "nmm/multimedia/ISyncResetImpl.hpp"
#include "nmm/multimedia/interfaces/ITrackImpl.hpp"
#include "nmm/utils/time/Rational.hpp"

#include "ogmstreams.h"

namespace NMM {
  
  /**
   * OGM Demux Node.
   * Demultiplexes an OGG stream.
   */
  class OGMDemuxNode
    : public GenericDemultiplexerNode
    , public ISyncResetImpl
    , public ITrackImpl
  {  
  private:
    /**
     * Logical OGG stream.
     * This structure represents a logical ogg stream.
     */
    struct Stream {
      /** Stream types */
      enum Type {
        AUDIO = 1, VIDEO = 2, OTHER = 4
      };
      
      int                 sno;                    /**< Stream serial number */
      Type                stream_type;            /**< Stream type */
      ogg_stream_state    instate;                /**< Stream state used for reading from the stream */
      string              stream_name;            /**< Jack tag this stream */
      bool                is_vorbis;              /**< Vorbis stream, write tot output as an ogg stream */
      
      Rational            sample_rate;            /**< Sample rate for calculating timestamps form granule_pos */
      Rational            highest_ts;             /**< Highest timestamp so far */
      int                 stream_counter;         /**< Current stream counter */
      
      bool                skip_until_valid_ts;    /**< Skip until valid timestamp (for seeking) */
      bool                is_begin;               /**< At beginning of stream */
      bool                eos;                    /**< At end of stream */
      bool                end_track;              /**< EndTrack has been sent on this stream */
      
      ogg_stream_state    outstate;               /**< Stream state used for writing to ogg packets (Vorbis) */
      int                 packetno;               /**< Output packet number */
      ogg_int64_t         max_granulepos;         /**< Maximum granule pos */
      bool                flushed;                /**< Output has been flushed */
    };
    
    /** Sub states of this node */
    enum State {
      STATE_NEED_DATA,                 /**< Need more data to read a complete OGG page */
      STATE_HAVE_PAGE,                 /**< Processing an OGG page */
      STATE_WRITE_PAGES,               /**< Write out pages to an OGG stream */
      STATE_FLUSH_PAGES,               /**< Flush pages to an OGG stream */
      STATE_END_TRACK                  /**< End of track, finish all streams */
    };
    
    Buffer* saved_buffer;              /**< Saved input buffer */
    Buffer* saved_out_buffer;          /**< Output buffer to send (after sending event) */
    
    ogg_sync_state    sync;            /**< OGG sync state */
    ogg_page          page;            /**< Current OGG page */
    ogg_packet        pack;            /**< Current OGG packet */
    
    int state;                         /**< Current substate of this node */
    int last_state;                    /**< Previous substate of this node */
    bool initial_packet;               /**< Initial packet of page */
    bool initial_page;                 /**< Current page is initial page of a stream */
    bool analysing_done;               /**< Internal flag which is set when analysing is done */
    bool is_video;                     /**< Current stream is a video stream */
    Stream* current_stream;            /**< stream the current page belongs to */
    
    Stream analysing_stream;           /**< Stream which is currently analysed */
    vector<Stream> audio_streams;      /**< Internal audio stream data */
    vector<Stream> video_streams;      /**< Internal video stream data */

    bool skip_until_syncpoint;         /**< All packets should be dropped until next key_frame */
    
    Timestamp timestamp;               /**< Current timestamp */
    
  private:
    /** Get the stream with the specified serial number */
    Stream* getStreamBySerial(int sno);
    
    /** Add an audio stream */
    Stream* addAudioStream(int sno, ogg_stream_state* sstate);
    
    /** Add a video stream */
    Stream* addVideoStream(int sno, ogg_stream_state* sstate);

    
    /**
     * Create raw output buffer.
     * Creates a new buffer containing raw output data. The current timestamp is attached
     * to the buffer and the current stream's counter is increased.
     */
    Buffer* createRawOutputBuffer(const char* data, ssize_t size);
    
    
    /** Handle packet */
    Message* handlePacket();
    
    /** Handle stream header packet */
    Message* handleHeaderPacket();
    
    /** Handle page */
    Message* handlePage();
    
  protected:
    virtual Result doInit();
    virtual Result doInitOutput();
    virtual Result doDeinitOutput();

    virtual Result doFlush();

    virtual Message *processBuffer (Buffer *);
    
    /** @name ISyncReset interface */
    //@{
    virtual Result syncReset();
    //@}
    
    /** @name ITrack interface */
    //@{
    virtual Result startTrack(const string& filename);
    virtual Result endTrack();
    virtual Result finished();
    //@}
    
  public:
    
    OGMDemuxNode(const char *name = "OGMDemuxNode", StreamQueue::Mode = StreamQueue::MODE_SUSPEND, int = 20);
    ~OGMDemuxNode();
  };
  
}

#endif //NMM_OGMDEMUXNODE_HPP



