/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
 
#include <math.h>
#include <fstream>

#include "AVIReadNode.hpp"
#include "nmm/comm/format/Property.hpp"
#include "nmm/multimedia/interfaces/ITrack.hpp"
#include "nmm/multimedia/FlushEvent.hpp"
#include "nmm/multimedia/interfaces/ITrackDuration.hpp"
#include "nmm/multimedia/interfaces/IProgressListener.hpp"
#include "nmm/plugins/interfaces/IAVMux.hpp"
#include "nmm/plugins/interfaces/IKeyframe.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"
#include "nmm/utils/thread/MutexGuard.hpp"
#include "nmm/utils/SecurityCheck.hpp"

extern "C" {
#include "avilib/avilib.h"
}

NMM_REGISTER_PLUGIN(AVIReadNode, NMM, "AVIReadNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  AVIReadNode::AVIReadNode(const char* name /* = "AVIReadNode" */,
			   StreamQueue::Mode mode /* = StreamQueue::MODE_SUSPEND */,
			   int queue_size /* = StreamQueue::DEFAULT_SIZE */)
    : GenericSourceNode(name, mode, queue_size), 
      send_format(false),
      seek_mutex(ThreadMutex::RECURSIVE),
      frame_rate_base(1000),
      vid_buf_bound(65536),
      __current_buffer(0),
      valid_audio(true)
  {
    av_offset = 0;
    av_pos = 0;
    read_size = 0;
    file_name  = "";
    seeked = false;
    end_of_file = true;
    begin_of_file = false;
    all_at_the_beginning = true;
    send_progress = false;
    avi_file = 0;
    width = 0;
    height = 0;
    num_frames = 0;
    frame_rate = 0.0;
    int_frame_rate  = 0;
    video_compressor = NULL;
    audio_channels = 0;
    audio_bits = 0;
    audio_format = 0;
    audio_rate = 0;
    audio_bytes = 0;
    audio_bitrate = 0;
    is_key_frame = 0;
    aud_buf_size = 0;
    buffers_before_progress = 5;
    buffers_progress        = 0;
    filesize=0;
    divisor = 2;
  };
  
  AVIReadNode::~AVIReadNode() {
    if (avi_file != 0) delete avi_file;
  }
  
  void AVIReadNode::setFilename(const string &name) {
    MutexGuard m(file_name_mutex);
    end_of_file = false;
    begin_of_file = true;
    valid_audio = true;
    file_name = name;
  }
  
  nmm_uint64 AVIReadNode::getFilesize() {
    return filesize;
  }
  
  Result AVIReadNode::syncReset() {
    MutexGuard smg(seek_mutex);
    av_offset = av_pos;
    return SUCCESS;
  }

  void AVIReadNode::sendProgressInformation(const nmm_bool& in) {
    send_progress = in;
  }

  void AVIReadNode::setProgressInterval(const nmm_int32& buffers) {
    buffers_before_progress = buffers;
  }

  Result AVIReadNode::doInit() {
    // register all supported output formats
    Format *out_format = getOwnOutputProperty()->addNewFormat(Format::av_avi_type);
    
    // video-part
    Format* out_v_format;
    out_v_format = new Format(Format::video_mpeg4_type);
    out_v_format->addWildcard(Format::x_resolution_param);
    out_v_format->addWildcard(Format::y_resolution_param);
    out_v_format->addWildcard(Format::ratio_x_param);
    out_v_format->addWildcard(Format::ratio_y_param);
    out_v_format->addWildcard(Format::framerate_param);
    out_v_format->addWildcard(Format::compressor_param);
    out_format->addFPointerParamValue(Format::video_format_param, out_v_format);

    out_v_format = new Format(Format::video_raw_type);
    out_v_format->addWildcard(Format::x_resolution_param);
    out_v_format->addWildcard(Format::y_resolution_param);
    out_v_format->addWildcard(Format::ratio_x_param);
    out_v_format->addWildcard(Format::ratio_y_param);
    out_v_format->addWildcard(Format::framerate_param);
    out_v_format->addStringParamValue(Format::channelorder_param,"bgr");
    out_v_format->addStringParamValue(Format::endian_param, "little");
    out_v_format->addStringParamValue(Format::colorspace_param,"rgb24");
    out_v_format->addIntParamValue(Format::bitperpixel_param,24);
    out_v_format->addStringParamValue(Format::format_param,"packet");
    out_format->addFPointerParamValue(Format::video_format_param, out_v_format);
 
    //audio-part
    Format* out_a_format;
    out_a_format = new Format(Format::audio_ac3_type);
    out_a_format->addWildcard(Format::bitrate_param);
    out_format->addFPointerParamValue(Format::audio_format_param, out_a_format);
    
    out_a_format = new Format(Format::audio_mpeg_type);
    out_a_format->addIntParamValue(Format::audio_codec_param, 0x55);
    out_a_format->addIntParamValue(Format::layer_param, 3); // XXX
    out_a_format->addWildcard(Format::sample_rate_param);
    out_a_format->addWildcard(Format::channels_param);
    out_a_format->addWildcard(Format::bitpersample_param);
    out_a_format->addWildcard(Format::bitrate_param);
    out_format->addFPointerParamValue(Format::audio_format_param, out_a_format);
    
    getOwnOutputProperty()->setDefaultFormat(out_format);
    
    return SUCCESS;
  }
  
  Result AVIReadNode::doInitOutput() {
    if(!SSecurityCheck::getInstance().checkFilenameForReading(file_name)) {
      ERROR_STREAM("Security check not passed for file " << file_name << ", check .nmmrc for options" << endl);
      return FAILURE;
    }
    
    ifstream in_fstream;
    in_fstream.clear();
    in_fstream.open(file_name.c_str(), ios::in);
    if(in_fstream.fail()) {
      ERROR_STREAM( "Error while opening file " << file_name << endl);
      return FAILURE;
    }
    
    // get filesize and seek to beginning again
    in_fstream.seekg(0, ios::end);
    filesize = in_fstream.tellg();
    in_fstream.close();
    
    
    if((avi_file = AVI_open_input_file(const_cast<char*>(file_name.c_str()), 1)) == 0) { // open file, 1 for flag "getIndex"
      DEBUG_STREAM( "Error Number " << errno << " while opening file " << file_name << endl);
      return FAILURE;
    }
    
    DEBUG_STREAM( "Opened file " << file_name << endl);
    // read all information from avi file
    width               = AVI_video_width(avi_file);
    height              = AVI_video_height(avi_file);
    num_frames          = AVI_video_frames(avi_file);
    frame_rate          = AVI_frame_rate(avi_file);
    int_frame_rate      = static_cast<int>(ceil(frame_rate * static_cast<double>(frame_rate_base)));
    video_compressor    = AVI_video_compressor(avi_file);
    audio_channels      = AVI_audio_channels(avi_file);
    audio_bits          = AVI_audio_bits(avi_file);
    audio_format        = AVI_audio_format(avi_file);
    audio_rate          = AVI_audio_rate(avi_file);
    
    if (audio_channels == 0) {
      divisor = 1;
    }
    else {
      divisor = 2;
    }

    //our DV camera produces mjpegs with an wrong samplerate of 11024, just round it to 11025
    if (audio_rate == 11024)
      audio_rate = 11025;

    audio_bytes         = AVI_audio_bytes(avi_file);
    audio_bitrate       = AVI_audio_mp3rate(avi_file);
    int num_achunks     = avi_file->audio_chunks;

        
    DEBUG_STREAM("Found width: " << width << endl);
    DEBUG_STREAM("Found height: " << height << endl);
    DEBUG_STREAM("Found frame_rate: " << frame_rate << endl);
    DEBUG_STREAM("Calculated int_frame_rate: " << int_frame_rate << endl);
    DEBUG_STREAM("Found num_frames: " << num_frames << endl);
    DEBUG_STREAM("Found video_compressor: " << video_compressor << endl);
    DEBUG_STREAM("Found audio_channels: " << audio_channels << endl);
    DEBUG_STREAM("Found audio_rate: " << audio_rate << endl);
    DEBUG_STREAM("Found audio_bits: " << audio_bits << endl);
    DEBUG_STREAM("Found audio_format: " << audio_format << endl);
    DEBUG_STREAM("Found audio_bytes: " << audio_bytes << endl);
    DEBUG_STREAM("Found audio chunks: " << num_achunks << endl);
    DEBUG_STREAM("Found audio_bitrate: " << audio_bitrate << " bit/s" << endl);

    // Set audio_bits to 16 if 0 is found
    if (audio_bits == 0) {
      DEBUG_STREAM("Found audio_bits == 0, setting to 16!" << endl);
      audio_bits = 16;
    }
    
    // Abort if no frames are found
    if (0 == num_frames) {
      DEBUG_STREAM("Found 0 as number of frames, will return FAILURE!" << endl);
      return FAILURE;
    }
    
    // Calculate how many audio bytes to be read per video frame
    aud_buf_size = audio_bytes / num_frames;
    DEBUG_STREAM("READING " << aud_buf_size << " BYTES OF AUDIO PER VIDEO FRAME." << endl);
    
    //we had another output-format before? If so -> send format_changed
    if(isOutputConnected("default")) {
      send_format = true;
    }
    
    // set the output format
    Format *out_format = getOwnWorkingOutputProperty()->addNewFormat("av/avi");
    
    // video-part
    // XXX: test if this is really video/mpeg
    Format* out_v_format = NULL;
    if (string(video_compressor) != string("")) {
      out_v_format = new Format("video/mpeg4");
      out_v_format->addIntParamValue(Format::x_resolution_param, width);
      out_v_format->addIntParamValue(Format::y_resolution_param, height);
      out_v_format->addIntParamValue(Format::ratio_x_param, width);
      out_v_format->addIntParamValue(Format::ratio_y_param, height);
      out_v_format->addFloatParamValue(Format::framerate_param, frame_rate);
      out_v_format->addStringParamValue(Format::compressor_param, video_compressor);
      out_format->addFPointerParamValue(Format::video_format_param, out_v_format);
    } else {
      out_v_format = new Format(Format::video_raw_type);
      out_v_format->addIntParamValue(Format::x_resolution_param, width);
      out_v_format->addIntParamValue(Format::y_resolution_param, height);
      out_v_format->addIntParamValue(Format::ratio_x_param, width);
      out_v_format->addIntParamValue(Format::ratio_y_param, height);
      out_v_format->addFloatParamValue(Format::framerate_param, frame_rate);
      out_v_format->addStringParamValue(Format::channelorder_param,"bgr");
      out_v_format->addStringParamValue(Format::endian_param, "little");
      out_v_format->addStringParamValue(Format::colorspace_param,"rgb24");
      out_v_format->addIntParamValue(Format::bitperpixel_param,24);
      out_v_format->addStringParamValue(Format::format_param,"packet");
      out_format->addFPointerParamValue(Format::video_format_param, out_v_format);
    }

    // audio-part
    // if audio_channels is zero, the AVI file definitely contains no audio
    // XXX: is there a better way to figure out whether an AVI file contains audio?
    Format* out_a_format = NULL;
    if (audio_channels != 0) {
      if (0x55 == audio_format) {
        out_a_format = new Format(Format::audio_mpeg_type);
        out_a_format->addIntParamValue(Format::audio_codec_param, audio_format);
        out_a_format->addIntParamValue(Format::layer_param, 3); // XXX
        out_a_format->addIntParamValue(Format::channels_param, audio_channels);
        out_a_format->addIntParamValue(Format::sample_rate_param, audio_rate);
        out_a_format->addIntParamValue(Format::bitpersample_param, audio_bits);
        out_a_format->addIntParamValue(Format::bitrate_param, audio_bitrate);
      } else if (0x2000 == audio_format) {
        out_a_format = new Format(Format::audio_ac3_type);
      } else if (0x1 == audio_format) {
        out_a_format = new Format(Format::audio_raw_type);
	out_a_format->addIntParamValue(Format::channels_param, audio_channels);
	out_a_format->addIntParamValue(Format::sample_rate_param, audio_rate);
	out_a_format->addIntParamValue(Format::bitpersample_param, audio_bits);
	if(audio_bits > 8) { // endian only exits if more than a single byte is used
	  out_a_format->addStringParamValue(Format::endian_param, "little");
          //12 or 16bit samples are signed
          out_a_format->addStringParamValue(Format::sign_param, "signed");
	}
        else {
          //8 bit is unsigned
          out_a_format->addStringParamValue(Format::sign_param, "unsigned");
        }
	out_a_format->addStringParamValue(Format::encoding_param, "linear");

      } else {
        ERROR_STREAM("Unknown audio codec!!!: " << audio_format << endl);
      }
      
      if (out_a_format != NULL) {
        out_format->addFPointerParamValue(Format::audio_format_param, out_a_format);
      } 
    }
    
    getOwnWorkingOutputProperty()->setDefaultFormat(out_format);
    setProducingFlag(true);
    return SUCCESS;
  }
  
  Message* AVIReadNode::processBuffer(Buffer *) {

    /* Begin> Handle a key frame stored in __current_buffer >*/
    if (__current_buffer) {
      setWorkingFlag(false);
      Buffer* tmp_buffer = __current_buffer;
      __current_buffer = 0;  //reset the buffer
      return tmp_buffer;
    }
    /* End< */


    // First thing to check is whether a seek was done - e.g. to
    // play the file not from the beginning but from a certain offset
    // This has to be done first, because e FlushEvent is returned,
    // which may otherwise delete the start_track or buffer_label event
    // in the predecessor nodes input queue.

    if ( send_progress ) {
      Event e;
      if ( buffers_progress > buffers_before_progress ) {
        buffers_progress = 0;
        CEvent* cevent(new CEvent(IProgressListener::create_setProgress(av_pos/divisor, num_frames)));
        cevent->setMandatoryBit(false);
        return cevent;
      }
      buffers_progress++;
    }

    if (seeked) {
      MutexGuard smg(seek_mutex);
      // search for next I-Frame
      is_key_frame = 0;
      while (is_key_frame == 0) {
	is_key_frame = AVI_is_keyframe(avi_file, av_pos/divisor);
	
	if (!is_key_frame) {
	  if (seek_backwards) {
	    av_pos -= 2;
	  }
	  else {
	    av_pos += 2;
	  }
	}
	if (av_pos/divisor > num_frames) {
	  DEBUG_STREAM("No keyframe found till EOF." << endl);
	  // set av_pos to last frame, this one is (further down)
	  // read and send, and EOF will be sent after that
	  av_pos = num_frames * 2;
	  break;
	}
      }

      float cursec = ((float)av_pos / 2) / frame_rate;
      int64_t abyte = (int64_t)(cursec * (float)audio_bitrate / 8);
      DEBUG_STREAM("Set audio position to: " << abyte << endl);
      AVI_set_audio_position(avi_file, abyte);
    
      // set position of video frame
      AVI_set_video_position(avi_file, av_pos/divisor);

      buffers_progress = buffers_before_progress +1;
      syncReset();

      seeked = false;
      DEBUG_STREAM("Seek complete - sending FlushEvent..." << endl);
      valid_audio = true;
      CEvent* cevent = new FlushEvent();
      cevent->insertEvent(cevent->size(), ISyncReset::create_syncReset());
      return cevent;
    }

    if (all_at_the_beginning) {
      all_at_the_beginning = false;
      
      // fw 20030710: no longer needed because labels are hardcoded
//       DEBUG_STREAM("Sending setBufferLabel events..." << endl);
//       Event video (ISetBufferLabel::setBufferLabel_event, new TInValue<string>(string("video")),
// 		   new TInValue<nmm_int32>(IAVMux::LABEL_VIDEO));
//       Event audio (ISetBufferLabel::setBufferLabel_event, new TInValue<string>(string("audio")),
// 		   new TInValue<nmm_int32>(IAVMux::LABEL_AUDIO));
//       return new CEvent(video, audio);
    }

    if (end_of_file) {
      setProducingFlag(false);
      
      DEBUG_STREAM("Sending endTrack..." << endl);
      Event e(ITrack::endTrack_event);
      return new CEvent(e);
    }

    if (begin_of_file) {
      begin_of_file = false;
      
      // XXX: (fw) no need to send this ?
//       Event event = Event( "aspect_ratio", 2 );
//       event.setValue( 0, new TValue<nmm_int32>( width ) );
//       event.setValue( 1, new TValue<nmm_int32>( height ) );
      
      // send start track event
      CEvent* ce = new CEvent();
      DEBUG_STREAM("Sending startTrack & syncReset"<<endl);
      ce->insertEvent(ITrack::create_startTrack(file_name)); 
      ce->insertEvent(ISyncReset::create_syncReset());

      // compute and send track duration
      double seconds = num_frames / frame_rate;
      Interval duration;
      duration.sec = static_cast<uint>(seconds);
      duration.nsec = static_cast<uint>((seconds - duration.sec) * 1000000);
      ce->insertEvent(ITrackDuration::create_trackDuration(duration));

      if(send_format) {
	send_format = false;

        //set new outputformat
        setOutputFormat(*getOwnWorkingOutputProperty()->getDefaultFormat());

	ERROR_STREAM("Sending setFormat" << endl);
	ce->insertEvent(IFormatChange::create_setFormat(*getOutputFormat()));
      }

      return ce;
    }

    Buffer* prod_buffer;

    MutexGuard smg(seek_mutex);
    if (av_pos % 2 != 1 || (audio_channels == 0)) {
      // read Video Frame
      // determine size of Frame
      read_size = AVI_frame_size(avi_file, av_pos/divisor);
      DEBUG_STREAM("Read frame with size: " << read_size << endl);
      // check if frame fits in our standard-video-buffers
      if (read_size <= vid_buf_bound) {
	  prod_buffer = getNewBuffer(vid_buf_bound);
      }
      else {
	DEBUG_STREAM("Video Buffer too large, getting extra big Buffer..." << endl);
	prod_buffer = getNewBuffer(read_size);
      }
      read_size = AVI_read_frame(avi_file, prod_buffer->getData(), &is_key_frame);
      DEBUG_STREAM("Read " << read_size << " bytes of Video (one frame)" <<
		   (is_key_frame ? " - KeyFrame" : "") << endl);
      if (read_size == -1) { //EOF
	// set end of file if we get the last video-frame
	end_of_file = true;
	prod_buffer->release();
	return 0;
      }
      prod_buffer->setUsed(read_size);
      prod_buffer->setLabel(IAVMux::LABEL_VIDEO);
    }
    else {
      if ( audio_channels != 0 ) {
      // read Audio data
      prod_buffer = getNewBuffer(aud_buf_size);
      
      read_size = AVI_read_audio(avi_file, prod_buffer->getData(), aud_buf_size);
      DEBUG_STREAM("Read " << read_size << " bytes of Audio" << endl);
      if (read_size == -1) { //EOF
	// set end of file if we get the last video-frame
	end_of_file = true;
	prod_buffer->release();
	return 0;
      }
      prod_buffer->setUsed(read_size);
      prod_buffer->setLabel(IAVMux::LABEL_AUDIO);
      }
    }
    
    // calculate time stamp: av_pos counts audio AND video samples, therefore /2
    // after seeking, timestamps begin with 0 again. This behaviour was copied from
    // MPEGAudioDecodeNode or MPEGVideoDecodeNode
    Timestamp t;
    int tmpint = av_pos - av_offset;
    if ( audio_channels != 0 ) {
      tmpint = tmpint/divisor;
    }
    Rational time(int(tmpint * frame_rate_base), int(int_frame_rate));

    t.sync_time = Time(time);
    t.stream_counter = tmpint;
    if (av_pos % 2 != 1 || (audio_channels == 0) ) {
      // read Video Frame
      t.is_valid = true;
    }
    else {
      // read Audio Frame
      //if (valid_audio) {
        t.is_valid = true;
        //valid_audio = false;
	//} else
        //t.is_valid = false;
    }

    DEBUG_STREAM("AVI: Set timestamp: " << t.sync_time << endl);
    prod_buffer->setTimestamp(t);

    av_pos++;

    /* Begin> If its a key frame, send the key_frame event first*/
    if (is_key_frame) {
      __current_buffer = prod_buffer; //store the key frame
      setWorkingFlag(true);
      is_key_frame = false;           //reset the is_key_frame flag
      CEvent* ce = new CEvent(Event(IKeyframe::setKeyframe_event));  // Create a keyframe event
      ce->setLabel( IAVMux::LABEL_VIDEO ); //set label to video to send it along the video branch
      return ce;
    }
    /* End< */

    return prod_buffer;
  }

  /******************** ISeekable interface ********************/
  nmm_bool AVIReadNode::hasPercentSeek() const {
    return true;
  }

  void AVIReadNode::seekPercent(const Rational& percent) {
    MutexGuard smg(seek_mutex);
    seekPercentTo(Rational(av_pos, num_frames*2) + percent);
  }
   
  void AVIReadNode::seekPercentTo(const Rational& percent) {
    MutexGuard smg(seek_mutex);
    if (seeked) {
      ERROR_STREAM("Still seeking" << endl);
      return;
    }

    Rational internal_percent(percent);
    if (internal_percent < Rational(0)) {
      ERROR_STREAM("Seeking to negativ not possible, defaulting to 0%!" << endl);
      internal_percent = Rational(0);
    }

    int old_av_pos = av_pos;
    Rational rat_av_pos = Rational(num_frames * 2) * internal_percent;
    av_pos = rat_av_pos.getNumerator() / rat_av_pos.getDenominator();
    if (av_pos % 2 == 1) {
      av_pos++;
    }

    if (av_pos/divisor > num_frames) {
      ERROR_STREAM("Tried to seek beyond EOF." << endl);
      av_pos = num_frames*2 - 2;
    }
    seeked = true;

    seek_backwards = av_pos < old_av_pos;
    DEBUG_STREAM(av_pos << " = " << internal_percent << " * " << num_frames << " * 2" << endl);
  }

  nmm_bool AVIReadNode::hasBytesSeek() const {
    return false;
  }

  void AVIReadNode::seekBytes(const nmm_int64&) {
  }

  void AVIReadNode::seekBytesTo(const nmm_uint64&) {
  }

  nmm_bool AVIReadNode::hasTimeSeek() const {
    return false;
  }

  void AVIReadNode::seekTime(const Interval& interval) {
    MutexGuard smg(seek_mutex);
    if (seeked) {
      ERROR_STREAM("Still seeking" << endl);
      return;
    }

    Interval internal_interval(interval);

    Rational rat_framerate(int_frame_rate, frame_rate_base);
    Rational rat_av_offset = (Rational(internal_interval.sec) + 
                              Rational(internal_interval.nsec, nano_per_sec)) * rat_framerate;
    rat_av_offset.reduce();
    int tmp_av_pos = av_pos + (rat_av_offset.getNumerator() / rat_av_offset.getDenominator());
    if ((tmp_av_pos % 2) == 1) {
      tmp_av_pos++;
    }
    if (tmp_av_pos/divisor > num_frames) {
      ERROR_STREAM("Tried to seek beyond EOF." << endl);
      tmp_av_pos = num_frames*2 - 2;
    } else if (tmp_av_pos < 0) {
      ERROR_STREAM("Seeking into negativ Time not possible, defaulting to Time 0!" << endl);
      tmp_av_pos = 0;
    }

    DEBUG_STREAM("Seeked: " << tmp_av_pos << " = " << av_pos << " + (" << rat_av_offset.getNumerator() << " / " << rat_av_offset.getDenominator() << ")" << endl);

    seek_backwards = tmp_av_pos < av_pos;
    av_pos = tmp_av_pos;
    seeked = true;
  }

  void AVIReadNode::seekTimeTo(const Time& time) {
    MutexGuard smg(seek_mutex);
    if (seeked) {
      ERROR_STREAM("Still seeking" << endl);
      return;
    }

    Time internal_time(time);
    if (internal_time < Time(0)) {
      ERROR_STREAM("Seeking to negativ Time not possible, defaulting to Time 0!" << endl);
    }

    Rational rat_framerate(int_frame_rate, frame_rate_base);
    Rational rat_av_pos = (Rational(internal_time.sec) + 
                           Rational(internal_time.nsec, nano_per_sec)) * rat_framerate;
    rat_av_pos.reduce();
    int tmp_av_pos = rat_av_pos.getNumerator() / rat_av_pos.getDenominator();
    if ((tmp_av_pos % 2) == 1) {
      tmp_av_pos++;
    }
    if (tmp_av_pos/divisor > num_frames) {
      ERROR_STREAM("Tried to seek beyond EOF." << endl);
      tmp_av_pos = num_frames*2 - 2;
    }

    DEBUG_STREAM("Seeked: " << tmp_av_pos << " = " << rat_av_pos.getNumerator() << " / " << rat_av_pos.getDenominator() << endl);
    av_pos = tmp_av_pos;
    seeked = true;
  }
  /******************** ISeekable interface ********************/
}

#include "nmm/nmm_suffix.hpp"
