/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include <iostream>
#include "AVIWriteNode.hpp"
#include "nmm/plugins/interfaces/IAVMux.hpp"
#include "nmm/comm/format/Property.hpp"
#include "nmm/comm/messaging/Timestamp.hpp"
#include "nmm/utils/SecurityCheck.hpp"
#include "nmm/utils/time/Time.hpp"
#include "nmm/utils/time/Interval.hpp"
#include "nmm/utils/thread/MutexGuard.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"
	

extern "C" {
#include "libmpdemux/muxer.h"
#include "libmpdemux/mmreg.h"
}

// windows is g.. ... doof :(
#undef DELETE

NMM_REGISTER_PLUGIN(AVIWriteNode, NMM, "AVIWriteNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  muxer_t* avimux;
  muxer_stream_t* mux_a;
  muxer_stream_t* mux_v;
  int out_file_format;
  FILE* avifile;
  long int bytes_written;
  long int frames_written;
  string filename;
  bool closed;
  bool is_key_frame;
  unsigned int labelnum_video;
  unsigned int labelnum_audio;
  ThreadMutex avw_mutex;
  
  bool first_audio;
  bool first_video;
  
  AVIWriteNode::AVIWriteNode(const char* name,
                             const StreamQueue::Mode queue_mode,
                             const int queue_size) : 
    GenericSinkNode(name, queue_mode, queue_size ),
    avimux(NULL),
    mux_a(NULL),
    mux_v(NULL),
    out_file_format(MUXER_TYPE_AVI),
    avifile(NULL),
    bytes_written(0),
    frames_written(0),
    __file_extension(""),
    __working_directory(""),
    closed(true),
    is_key_frame(false),
    labelnum_video(IAVMux::LABEL_VIDEO),
    labelnum_audio(IAVMux::LABEL_AUDIO),
    first_audio(false),
    first_video(false)
  {}
  
  AVIWriteNode::~AVIWriteNode() 
  {
    if(avifile!=NULL) {
      fclose(avifile);
      avifile=NULL;
    }
  }
  
  void AVIWriteNode::setWorkingDirectory(const string& dir) {
    MutexGuard avg_mg(avw_mutex);
    __working_directory = dir;
  }
  
  string AVIWriteNode::getWorkingDirectory() {
    return __working_directory;
  }
  
  void AVIWriteNode::setFileExtension(const string& ext) {
    MutexGuard avw_mg(avw_mutex);
    __file_extension = ext;
  }
  
  string AVIWriteNode::getFileExtension() {
    if (__file_extension.empty())
      return "avi";
    else return __file_extension;
  }
  
  void AVIWriteNode::setFilename(const string& filename) {
    MutexGuard avw_mg(avw_mutex);
    this->filename = filename;
  }
  
  Result AVIWriteNode::started(const Interval&) {
    return SUCCESS;
  }

  Result AVIWriteNode::finished() {
    return SUCCESS;
  }  
  
  Result AVIWriteNode::startTrack(const string&) {
    return SUCCESS;
  }
  
  Result AVIWriteNode::endTrack() {
    MutexGuard avw_mg(avw_mutex);
    
    // test if file is already closed (if we come from deinit and are already finished)
    if (avifile != NULL) {
      DEBUG_STREAM("Track ended, closing avi-file");

      if (mux_a) {  
        // fixup CBRmp3 audio header:
        mux_a->h.dwSampleSize=1;
        ((MPEGLAYER3WAVEFORMAT*)(mux_a->wf))->nBlockSize=
  	(mux_a->size+(mux_a->h.dwLength>>1))/mux_a->h.dwLength;
        mux_a->h.dwLength=mux_a->size;
        mux_a->h.dwRate=mux_a->wf->nAvgBytesPerSec;
        mux_a->h.dwScale=1;
        mux_a->wf->nBlockAlign=1;
      
//       float a_delay = (float) first_audio_ts.sync_time.sec + (float) first_audio_ts.sync_time.nsec / 1000000000.0f;
//       float a_srate = (float) mux_a->h.dwRate / (float) mux_a->h.dwScale;
//       mux_a->h.dwStart = 0 /*(int) (a_delay * a_srate) */;
//       MESSAGE_STREAM("audio delay=" << a_delay << " srate=" << a_srate << " start=" << a_delay * a_srate << endl);

//       float v_delay = (float) first_video_ts.sync_time.sec + (float) first_video_ts.sync_time.nsec / 1000000000.0f;
//       float v_srate = (float) mux_v->h.dwRate / (float) mux_v->h.dwScale;
//       mux_v->h.dwStart = INT_MAX/10-1; /*(int) (v_delay * v_srate); */
//       MESSAGE_STREAM("video delay=" << v_delay << " srate=" << v_srate << " start=" << v_delay * v_srate << endl);
      } else {
        DEBUG_STREAM("Write avi file without audio" << endl);
      }        
      muxer_write_index(avimux);
      fseek(avifile,0,SEEK_SET);
      muxer_write_header(avimux); // update header
      if(ferror(avifile) || fclose(avifile) != 0) {
	return FAILURE;
      }
      avifile=NULL;
      DEBUG_STREAM("Wrote " << bytes_written << " Bytes." << endl);
      closed = true;
    }
    return SUCCESS;
  }
  
  Result AVIWriteNode::setKeyframe() {
    MutexGuard avw_mg(avw_mutex);
    if (is_key_frame) {
      WARNING_STREAM("Keframe flage was not reset !!!!!!!!!!!!!!!" << endl);
    }
    is_key_frame = true;
    return DELETE;
  }
  
  Result AVIWriteNode::setBufferLabel(const string& label_name, const nmm_int32& label_number) {
    if ("video" == label_name) {
      labelnum_video = label_number;
    }
    else if ("audio" == label_name) {
      labelnum_audio = label_number;
    }
    else {
      ERROR_STREAM("setBufferLabel: FOUND UNKNOWN LABEL ON BUFFER!!!");
      return FAILURE;
    }
    return DELETE;
  }
  
  Result AVIWriteNode::doInit() {
    
    /* Begin> Video Subformats */
    /* DivX video with bitrate */
    Format in_v_format1(Format::video_mpeg4_type);
    in_v_format1.addWildcard(Format::x_resolution_param);
    in_v_format1.addWildcard(Format::y_resolution_param);
    in_v_format1.addWildcard(Format::ratio_x_param);
    in_v_format1.addWildcard(Format::ratio_y_param);
    in_v_format1.addWildcard(Format::framerate_param);
    in_v_format1.addWildcard(Format::compressor_param);
    in_v_format1.addWildcard(Format::bitrate_param);
    
    /* DivX video without bitrate */    
    Format in_v_format2(Format::video_mpeg4_type);
    in_v_format2.addWildcard(Format::x_resolution_param);
    in_v_format2.addWildcard(Format::y_resolution_param);
    in_v_format2.addWildcard(Format::ratio_x_param);
    in_v_format2.addWildcard(Format::ratio_y_param);
    in_v_format2.addWildcard(Format::framerate_param);
    in_v_format2.addWildcard(Format::compressor_param);
    /* End< Video Subformats */
    
    /* Begin> Audio Subformats */
    /* MPEG audio with bitrate and extra parameters */
    Format in_a_format1(Format::audio_mpeg_type);
    in_a_format1.addIntParamValue(Format::audio_codec_param, 0x55);
    in_a_format1.addWildcard(Format::layer_param);
    in_a_format1.addWildcard(Format::channels_param);
    in_a_format1.addWildcard(Format::sample_rate_param);
    in_a_format1.addWildcard(Format::bitpersample_param);
    in_a_format1.addWildcard(Format::bitrate_param);

    /* MPEG audio without bitrate and with extra parameters */
    Format in_a_format2(Format::audio_mpeg_type);
    in_a_format2.addWildcard(Format::layer_param);
    in_a_format2.addIntParamValue(Format::audio_codec_param, 0x55);
    in_a_format2.addWildcard(Format::sample_rate_param);
    in_a_format2.addWildcard(Format::channels_param);
    in_a_format2.addWildcard(Format::bitpersample_param);
    
    /* MPEG audio with sample rate only */
    Format in_a_format3(Format::audio_mpeg_type);
    in_a_format3.addIntParamValue(Format::audio_codec_param, 0x55);
    in_a_format3.addWildcard(Format::sample_rate_param);
    
    /* MPEG audio with sample rate and bitrate */
    Format in_a_format4(Format::audio_mpeg_type);
    in_a_format4.addIntParamValue(Format::audio_codec_param, 0x55);
    in_a_format4.addWildcard(Format::sample_rate_param);
    in_a_format4.addWildcard(Format::bitrate_param);

    /* AC3 audio without bitrate */    
    Format in_a_format5(Format::audio_ac3_type);

    /* AC3 audio with bitrate */    
    Format in_a_format6(Format::audio_ac3_type);
    in_a_format6.addWildcard(Format::bitrate_param);
    
    /* AC3 audio with bitrate and extra parameters */
    Format in_a_format7(Format::audio_ac3_type);
    in_a_format7.addWildcard(Format::channels_param);
    in_a_format7.addWildcard(Format::sample_rate_param);
    in_a_format7.addWildcard(Format::bitrate_param);
    in_a_format7.addWildcard(Format::speaker_setup_param);
    in_a_format7.addIntParamValue(Format::lfe_param, 0);
    in_a_format7.addIntParamValue(Format::lfe_param, 1);

    /* raw audio with endian */
    Format in_a_format8(Format::audio_raw_type);
    in_a_format8.addWildcard(Format::channels_param);
    in_a_format8.addWildcard(Format::sample_rate_param);
    in_a_format8.addWildcard(Format::bitpersample_param);
    in_a_format8.addStringParamValue(Format::endian_param, "little");
    in_a_format8.addStringParamValue(Format::encoding_param, "linear");
    in_a_format8.addStringParamValue(Format::sign_param, "unsigned");
    in_a_format8.addStringParamValue(Format::sign_param, "signed");

    /* raw audio without endian */
    Format in_a_format9(Format::audio_raw_type);
    in_a_format9.addWildcard(Format::channels_param);
    in_a_format9.addWildcard(Format::sample_rate_param);
    in_a_format9.addWildcard(Format::bitpersample_param);
    in_a_format9.addStringParamValue(Format::encoding_param, "linear");
    in_a_format9.addStringParamValue(Format::sign_param, "unsigned");
    in_a_format9.addStringParamValue(Format::sign_param, "signed");

    /* End< Audio Subformats */
    
    /* Begin> Composite Input Formats */
    /* Format with audio and video */
    Format* in_format1 = getOwnInputProperty()->addNewFormat(Format::av_avi_type);
    in_format1->addFPointerParamValue(Format::video_format_param, new Format(in_v_format1));
    in_format1->addFPointerParamValue(Format::video_format_param, new Format(in_v_format2));
    in_format1->addFPointerParamValue(Format::audio_format_param, new Format(in_a_format1));
    in_format1->addFPointerParamValue(Format::audio_format_param, new Format(in_a_format2));
    in_format1->addFPointerParamValue(Format::audio_format_param, new Format(in_a_format3));
    in_format1->addFPointerParamValue(Format::audio_format_param, new Format(in_a_format4));
    in_format1->addFPointerParamValue(Format::audio_format_param, new Format(in_a_format5));
    in_format1->addFPointerParamValue(Format::audio_format_param, new Format(in_a_format6));
    in_format1->addFPointerParamValue(Format::audio_format_param, new Format(in_a_format7));
    in_format1->addFPointerParamValue(Format::audio_format_param, new Format(in_a_format8));
    in_format1->addFPointerParamValue(Format::audio_format_param, new Format(in_a_format9));
    getOwnInputProperty()->setDefaultFormat(in_format1);
    
    /* Format with video only */
    Format* in_format2 = getOwnInputProperty()->addNewFormat(Format::av_avi_type);
    in_format2->addFPointerParamValue(Format::video_format_param, new Format(in_v_format1));
    in_format2->addFPointerParamValue(Format::video_format_param, new Format(in_v_format2));
    /* End< Composite Input Formats */
    
    return SUCCESS;
  }
  
  Result AVIWriteNode::doDeinit() {
    return SUCCESS;
  }
  
  Result AVIWriteNode::doActivate() {
    if (!__file_extension.empty())
      filename = filename+"."+__file_extension;

    if (!__working_directory.empty()) {
      filename = filename.substr(filename.find_last_of("/")+1,filename.size());
      filename = __working_directory + "/" + filename;
    }
    DEBUG_STREAM("Open File: " << filename << endl);

    first_audio = true;
    first_video = true;
    
    /**
     * open file: 
     * - binary mode
     * - delete old file with same name
     * - stream is for output
     */
    
    /* Security check */
    if (!SSecurityCheck::getInstance().checkFilenameForWriting(filename)) {
      ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return FAILURE;
    }
    
    if(avifile==NULL) {
      avifile=fopen(filename.c_str(),"wb");
      if(!avifile) {
 	ERROR_STREAM("avi open error" << endl);
 	return FAILURE;
      }
    } else {
      DEBUG_STREAM("Error: avifile not NULL, but should be NULL in doActivate()!");
      return FAILURE;
    }
    
    // set up muxer
    avimux=muxer_new_muxer(out_file_format, avifile);

    // get input format
    DEBUG_STREAM("Trying to get format infos...");
    const Format* in_format = getInputFormat();
    if (in_format == NULL) { return FAILURE; }
    
    // get video input format
    if (!in_format->isParameterIncluded(Format::video_format_param)) {
      ERROR_STREAM("No video format in input format!");
      return FAILURE;
    }
    const Format* in_v_format = in_format->getFPointerValue(Format::video_format_param);
    int v_width = in_v_format->getIntValue(Format::x_resolution_param);
    int v_height = in_v_format->getIntValue(Format::y_resolution_param);
    double v_fps = in_v_format->getFloatValue(Format::framerate_param);
    string compressor = in_v_format->getStringValue(Format::compressor_param);
    
    mux_v=muxer_new_stream(avimux,MUXER_TYPE_VIDEO);    

    mux_v->h.dwSampleSize=0; // VBR
    mux_v->h.dwScale=10000;
    mux_v->h.dwRate = (DWORD) (mux_v->h.dwScale * v_fps);

    mux_v->bih= new BITMAPINFOHEADER();// malloc(sizeof(BITMAPINFOHEADER));
    memset(mux_v->bih, 0, sizeof(BITMAPINFOHEADER));
    mux_v->bih->biSize=sizeof(BITMAPINFOHEADER);
    mux_v->bih->biWidth=v_width;
    mux_v->bih->biHeight=v_height;

    // fourcc code is expected in compressor string
    mux_v->bih->biCompression = mmioFOURCC(compressor[0], compressor[1],
					   compressor[2], compressor[3]);
    mux_v->bih->biPlanes = 1;
    mux_v->bih->biBitCount = 12; // FIXME!!!
    mux_v->bih->biSizeImage = mux_v->bih->biWidth*mux_v->bih->biHeight*(mux_v->bih->biBitCount/8);

    // get audio format
    // if audio_format is set to NULL, create an AVI file without audio
    if (in_format->isParameterIncluded(Format::audio_format_param)) {
      const Format* in_a_format = in_format->getFPointerValue(Format::audio_format_param);
      if (in_a_format != NULL) {
        // first set default values, based on format type
        // (for the case that the exact values are not known)
        int a_channels = 0;
        int a_srate = 0;
        int a_bitpersample = 0;
        int a_aud_codec = 0;
        if (in_a_format->getFormatType() == Format::audio_ac3_type) {
          a_aud_codec = 0x2000;
          a_channels = 5;//6;
          a_srate = 48000;
          a_bitpersample = 16;
        } else if (in_a_format->getFormatType() == Format::audio_mpeg_type) {
          a_aud_codec = 0x55;
        } else if (in_a_format->getFormatType() == Format::audio_raw_type) {
          a_aud_codec = 0x1;
          if (in_a_format->isParameterIncluded(Format::bitpersample_param)) {
            a_bitpersample = in_a_format->getIntValue(Format::bitpersample_param);
          }
        }
        
        // override values using input format
        if (in_a_format->isParameterIncluded(Format::audio_codec_param)) {
          a_aud_codec = in_a_format->getIntValue(Format::audio_codec_param);
        }
        if (in_a_format->isParameterIncluded(Format::channels_param)) {
          a_channels = in_a_format->getIntValue(Format::channels_param);
        }
        if (in_a_format->isParameterIncluded(Format::sample_rate_param)) {
          a_srate = in_a_format->getIntValue(Format::sample_rate_param);
        }
        if (in_a_format->isParameterIncluded(Format::bitpersample_param)) {
          a_bitpersample = in_a_format->getIntValue(Format::bitpersample_param);
        }
        
        // check if we have all value (we should, or it is a bug)
        if (!a_channels || !a_srate || !a_bitpersample || !a_aud_codec) {
          ERROR_STREAM("Can't determine audio format values (shouldn't happen!)" << endl);
        }
        // get the bitrate, if available
        int a_bitrate = 384000;
        if (in_a_format->isParameterIncluded(Format::bitrate_param)) {
          a_bitrate = in_a_format->getIntValue(Format::bitrate_param);
        }
        
        mux_a = new muxer_stream_t();
        memset(mux_a, 0, sizeof(muxer_stream_t));
        mux_a=muxer_new_stream(avimux,MUXER_TYPE_AUDIO);
        
        mux_a->h.dwSampleSize = 0; // VBR
        mux_a->h.dwRate = a_srate;
        mux_a->h.dwScale = (mux_a->h.dwRate<32000)?576:1152; // samples/frame
        mux_a->wf = (WAVEFORMATEX*) malloc(sizeof(MPEGLAYER3WAVEFORMAT)); // should be 30
        mux_a->wf->wFormatTag = a_aud_codec;
        mux_a->wf->nChannels = a_channels;
        mux_a->wf->nSamplesPerSec = mux_a->h.dwRate;
        mux_a->wf->nAvgBytesPerSec = a_bitrate / 8;
        mux_a->wf->nBlockAlign = (mux_a->h.dwRate<32000)?576:1152; // required for l3codeca.acm + WMP 6.4
        // obviously "0" helps some windows players to be able to play the file...
        // so just set it != 0 if we get audio_raw or audio_ac3
        mux_a->wf->wBitsPerSample = a_bitpersample;
        // from NaNdub:  (requires for l3codeca.acm)
        mux_a->wf->cbSize=12;
        ((MPEGLAYER3WAVEFORMAT*)(mux_a->wf))->wID=1;
        ((MPEGLAYER3WAVEFORMAT*)(mux_a->wf))->fdwFlags=2;
        ((MPEGLAYER3WAVEFORMAT*)(mux_a->wf))->nBlockSize=(mux_a->h.dwRate<32000)?576:1152; // ???
        ((MPEGLAYER3WAVEFORMAT*)(mux_a->wf))->nFramesPerBlock=1;
        ((MPEGLAYER3WAVEFORMAT*)(mux_a->wf))->nCodecDelay=1393;
      }
    }
    
    muxer_write_header(avimux);
 
    //do not force key frame at the very beginning of encoding, since
    //first frame will be a key fame anayway. Therefore key.quantizer
    //is well defined for any frame to follow
    //force_key_frame=(force_key_frame<0) ? 0:1;
    
    closed = false;
    return SUCCESS;
  }
  
  Result AVIWriteNode::doDeactivate() {
    endTrack();
    
    return SUCCESS;
  }

  nmm_int32 AVIWriteNode::getFrameCount() const {
    return frames_written;
  }
  
  Message* AVIWriteNode::processBuffer(Buffer *in_buffer) {
    MutexGuard avw_mg(avw_mutex);

    if (closed) {
      ERROR_STREAM("Error: Write to closed File" << endl);
      in_buffer->release();
      return NULL;
    }

    int used_bytes = in_buffer->getUsed();
    
    bool initial_frame = false;
    
    unsigned int label = in_buffer->getLabel();
    if (label == labelnum_video) {
      if (first_video) {
        initial_frame = true;
        first_video = false;
        first_video_ts = in_buffer->getTimestamp();
        MESSAGE_STREAM("First video TS:" << in_buffer->getTimestamp() << endl);
        if (!is_key_frame) {
          WARNING_STREAM("First video frame is not a key frame" << endl);
        }
      }
      DEBUG_STREAM("Writing Video Frame" << endl);
      // Write Frame, set key frame "flag" depending on is_key_frame
      mux_v->buffer = (unsigned char*) in_buffer->getData();
      muxer_write_chunk(mux_v,in_buffer->getUsed(),is_key_frame ? 0x10 : 0);
      if (is_key_frame) {
	DEBUG_STREAM("Wrote a key frame.");
      }
      // set is_key_frame to false always when we got a video frame.
      is_key_frame = false;
      frames_written++;
    }
    else if (label == labelnum_audio) {
      if (first_audio) {
        initial_frame = true;
        first_audio = false;
        first_audio_ts = in_buffer->getTimestamp();
        MESSAGE_STREAM("First audio TS:" << in_buffer->getTimestamp() << endl);
      }
      DEBUG_STREAM("Writing Audio Frame" << endl);
      mux_a->buffer = (unsigned char*) in_buffer->getData();
      muxer_write_chunk(mux_a, in_buffer->getUsed(), 0x10);
    }
    else {
      ERROR_STREAM("FOUND UNKNOWN LABEL!!! - Didn't do anything.");
    }
    
    /* Check if we have both initial frames */
    if (initial_frame && !first_audio && !first_video) {
      if (first_audio_ts.sync_time < first_video_ts.sync_time) {
        first_video_ts.sync_time = Time(0) + (first_video_ts.sync_time - first_audio_ts.sync_time);
        first_audio_ts.sync_time = Time(0);
        MESSAGE_STREAM("Video is delayed by " << first_video_ts.sync_time << endl);
      } else {
        first_audio_ts.sync_time = Time(0) + (first_audio_ts.sync_time - first_video_ts.sync_time);
        first_video_ts.sync_time = Time(0);
        MESSAGE_STREAM("Audio is delayed by " << first_audio_ts.sync_time << endl);
      }
    }
    
    bytes_written += used_bytes;
    DEBUG_STREAM("Written " << bytes_written << " Bytes" << endl);

    in_buffer->release();
    return NULL;
  }

#include "nmm/nmm_prefix.hpp"

} // namespace NMM

#include "nmm/nmm_suffix.hpp"
