/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <sstream>

#include "nmm/comm/format/Property.hpp"
#include "nmm/multimedia/FlushEvent.hpp"
#include "nmm/multimedia/ISyncReset.hpp"
#include "nmm/multimedia/Plugin.hpp"
#include "nmm/multimedia/interfaces/IProgressListener.hpp"
#include "nmm/multimedia/interfaces/ITrack.hpp"
#include "nmm/multimedia/interfaces/ITrackList.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"
#include "nmm/plugins/interfaces/ITagInfo.hpp"
#include "nmm/utils/thread/MutexGuard.hpp"

#include "CDDANode.hpp"

NMM_REGISTER_PLUGIN(NMM::CDDANode, NMM, "CDDANode")

namespace NMM {

  /********************************************************************************/
  /* Constructor */
  CDDANode::CDDANode(const char* name /* = "CDDANode" */,
		   StreamQueue::Mode mode /* = StreamQueue::MODE_SUSPEND */,
		   int queue_size /* = StreamQueue::DEFAULT_SIZE */)
    : GenericSourceNode(name, mode, queue_size),
      event(TOC),
      drive(NULL),
      paranoia(NULL),
      device(DEFAULT_CD_DEVICE),
      do_cddb_query(DEFAULT_DO_CDDB_QUERY),
      cddb_server_added(false),
      repeat(DEFAULT_CD_REPEAT),
      verbose(CDDA_MESSAGE_FORGETIT), // CDDA_MESSAGE_PRINTIT | CDDA_MESSAGE_LOGIT),
      paranoia_mode(PARANOIA_MODE_FULL^PARANOIA_MODE_NEVERSKIP),
      track_number(0),
      last_track_number(0),
      current_sector(0),
      stream_counter(0),
      paranoia_seek_mutex(),
      event_mutex(),
      toc(),
      send_progress(false),
      buffers_before_progress(10),
      buffers_progress(0),
      one_track_only(false),
      numberTrack(false),
      sendtaginfo(true) {

    if (do_cddb_query) {
      toc.addCDDBServer(DEFAULT_CDDB_SERVER, DEFAULT_CDDB_PORT);
      cddb_server_added = true;
    } else {
      cddb_server_added = false;
    }
  }
  /********************************************************************************/

  /********************************************************************************/
  /* Desctructor */
  CDDANode::~CDDANode() {
    /* this should have been done in doDeinit, but we make sure ... */
    if (NULL != paranoia) {
      paranoia_free(paranoia); //cause a segfault if paranoia is 0
      paranoia = NULL;
    }

    if (NULL != drive) {
      cdda_close(drive);
      drive = NULL;
    }
  }
  /********************************************************************************/


  /********************************************************************************/
  /* do* methods */
  Result CDDANode::doInit() {
    //old code from doInit()
    drive = cdda_identify(device.c_str(), verbose, NULL);
    if (NULL == drive) {                            
      drive = cdda_find_a_cdrom(verbose, NULL);
      if (NULL == drive) {                            
	ERROR_STREAM( "no CDROM device found" << endl);
	return FAILURE;
      }
    }

    
    Format* out_format = getOwnOutputProperty()->addNewFormat("audio/raw");
    out_format->addIntParamValue(Format::sample_rate_param, 44100);
    out_format->addIntParamValue(Format::channels_param, 2);
    out_format->addIntParamValue(Format::bitpersample_param, 16);
    out_format->addStringParamValue(Format::encoding_param, "linear");
    out_format->addStringParamValue(Format::sign_param, "signed");
    out_format->addStringParamValue(Format::endian_param, "little");
    getOwnOutputProperty()->setDefaultFormat(out_format);
    
    return SUCCESS;
  }

  Result CDDANode::doInitOutput() {
    
    if (0 != cdda_open(drive)) {                      
      ERROR_STREAM( "Can't open " << drive->cdda_device_name << " !" << endl);
      return FAILURE;
    }
    
    cdda_verbose_set(drive, verbose, verbose);

    paranoia = paranoia_init(drive);
    if (NULL == paranoia) {
      ERROR_STREAM( "Can't initialize cdparanoia library !" << endl);
      return FAILURE;
    }

    paranoia_modeset(paranoia, paranoia_mode);

    toc.setSize(drive->tracks);
    for(unsigned int i=0; i < toc.size(); i++) {
      if(cdda_track_audiop(drive, i)) {
        toc.setStartSector(i, cdda_track_firstsector(drive, i+1));
        toc.setEndSector(i, cdda_track_lastsector(drive, i+1));
      } else {
        toc.setAudio(i, false);
      }
    }
    current_sector = toc.getStartSector(0);
    
    if (do_cddb_query) {
      DEBUG_STREAM("Querying CDDB Server..." << endl);
      toc.queryCDDBServer();
    }

    paranoia_seek(paranoia, current_sector, SEEK_SET);

    Format* out_format = new Format( *getOwnOutputProperty()->getDefaultFormat() );
    getOwnWorkingOutputProperty()->addNewFormat(out_format);
    getOwnWorkingOutputProperty()->setDefaultFormat(out_format);

    return SUCCESS;
  }
  
  Result CDDANode::doDeinitOutput() {
    if (NULL != paranoia) {
      paranoia_free(paranoia); //cause a segfault if paranoia is 0
      paranoia = NULL;
    }
    
    if (NULL != drive) {
      cdda_close(drive);
      drive = NULL;
    }
    
    return SUCCESS;
  }

  /********************************************************************************/

  void CDDANode::sendProgressInformation(const nmm_bool& value) {
    send_progress = value;
  }

  void CDDANode::setProgressInterval(const nmm_int32& buffers) {
    buffers_before_progress = buffers;
  }
  
  /********************************************************************************/
  /* produceBuffer method */
  Message* CDDANode::processBuffer(Buffer*) {
    char* buf = NULL;
    string event_message;
    string name;
    string l1,l2;
    int i = 0;
    int test_size = 10;
    
    Event e;
    
    if ( send_progress ) {
      if ( buffers_progress > buffers_before_progress ) {
        buffers_progress = 0;
        long int current_samples = current_sector - toc.getStartSector(track_number);
        long int max_samples     = toc.getEndSector(track_number) - toc.getStartSector(track_number);
        CEvent* cevent(new CEvent(IProgressListener::create_setProgress(current_samples, max_samples)));
        cevent->setMandatoryBit(false);
        return cevent;
      }
      buffers_progress++;
    }
    
    Buffer* out = getNewBuffer(CD_FRAMESIZE_RAW * test_size);
    if (NULL == out) {
      ERROR_STREAM("Error creating Buffer" << endl);
      return out;
    }
    
    MutexGuard mg_em(event_mutex);
    switch (event) {
    case TOC: /* first we should introduce the playable tracks */
      e = Event("toc", new TInValue<string>(toc.toString()));
      event = START;
      return new CEvent(e);
      break;

    case START: /* Start of a new track */
      if (sendtaginfo) {
        Event* e = ITagInfo::create_TagInfo(toc.getName(track_number),toc.getTitle(), toc.getArtist(), toc.getYear(), toc.getGenre());
        sendtaginfo = false;
        return new CEvent(e);
      }
      if (numberTrack) {
        stringstream s;
        if ((track_number+1)<10)
          s << "0"<<track_number+1 <<" - " << toc.getName(track_number);
        else 
          s<< track_number+1 << " - " << toc.getName(track_number);
        e = Event(ITrack::startTrack_event, new TInValue<string>(s.str()));
      } else 
        e = Event(ITrack::startTrack_event, new TInValue<string>(toc.getName(track_number)));
      if (toc.getStartSector(track_number) < current_sector) {
        event = INCOMPLETE;
      } else {
        event = SYNC_RESET;
      }
      return new CEvent(e);
      break;

    case INCOMPLETE: /* current track was not started from the beginning */
      e = Event("truncated", new TInValue<string>(toc.getName(track_number)));
      event = SYNC_RESET;
      return new CEvent(e);
      break;

    case SYNC_RESET: /* timestamps are non-continues */
      e = Event(ISyncReset::syncReset_event);
      DEBUG_STREAM("Sending sync_reset" << endl);
      stream_counter = 0;
      event = DATA;
      return new CEvent(e);
      break;

    case DATA: /* ok, sending data message */
      {
        MutexGuard mg_psm(paranoia_seek_mutex);
 	long int current_samples =
	  (current_sector - toc.getStartSector(track_number))
	  * CD_FRAMESIZE_RAW 
	  * 8L /*BitperByte*/ 
	  / 2L /*Channels*/ 
	  / 16L /*Bitpersample*/;
	Timestamp t;
	t.sync_time = Time(Rational(current_samples, 44100L));
	t.stream_counter = stream_counter++;
	t.is_valid = true;
        while (i<test_size) {
          buf = reinterpret_cast<char *> (paranoia_read(paranoia, NULL));
          
          if (NULL == buf) {
            ERROR_STREAM( "Unrecoverable error in paranoia_read" << endl);
          } else {
            memcpy(out->getData() + CD_FRAMESIZE_RAW * i,buf,CD_FRAMESIZE_RAW);
            out->setUsed(CD_FRAMESIZE_RAW * (i+1)); 
          i++;
          }
          current_sector++;
          if (current_sector > toc.getEndSector(track_number)) {
            last_track_number = track_number;
            track_number++;
            event = END; /* end of current track */
            break; /* exit the while */
          }
        }
	out->setTimestamp(t);
      }
      return out;
      break;

    case SEEKED:
      e = Event("seeked", new TInValue<string>(toc.getName(track_number)));

      if (toc.getEndSector(track_number) < current_sector ||
          toc.getStartSector(track_number) > current_sector) {
        last_track_number = track_number;
        track_number = toc.getTrackBySector(current_sector);
        event = END;
      } else {
        event = FLUSH;
      }
      return new CEvent(e);
      break;

    case FLUSH:
      event = SYNC_RESET;
      return( new FlushEvent );
      break;

    case END: /* current Track doesn't continue */
      //      e = Event(ITrack::endTrack_event, new TInValue<string>(toc.getName(last_track_number)));
      e = Event(ITrack::endTrack_event);
      
      if (one_track_only) {   // one time only
	if (!repeat) {        // without repeat ?
	  event = FINISHED;   // easy: finish!
	} else {
	  MutexGuard mg_psm(paranoia_seek_mutex);
	  current_sector = toc.getStartSector(last_track_number); // play last track again 
	  paranoia_seek(paranoia, current_sector, SEEK_SET);
          sendtaginfo=true;
	  event = START;  // and start
	}
      } else {
	if (track_number >= toc.size()) {
	  MutexGuard mg_psm(paranoia_seek_mutex);
	  track_number = 0;
	  current_sector=toc.getStartSector(track_number);;
	  paranoia_seek(paranoia, current_sector, SEEK_SET);
	  if (!repeat) {
	    event = FINISHED;
	  } else { 
            sendtaginfo=true;
	    event = START;
	  }
	} else {
          sendtaginfo=true;
	  event = START;
	}
      }

      return new CEvent(e);
      break;
      
    case FINISHED: /* Complete CD is played */
      e = Event(ITrackList::endTrackList_event);
      event = FINISHED;
      setProducingFlag(false);
      return new CEvent(e);
      break;

    default:
      ERROR_STREAM( "Nonexisting internal EventState!" << endl);
      out->release();
      out = NULL;
      return out;
    }
  }
  /********************************************************************************/


  /********************************************************************************/
  /* Event Functions */
  void CDDANode::setDevice(const string& dev) throw(const DeviceNotFoundException&){
    device = dev;
  }
  
  string CDDANode::getDevice() const {
    return device;
  }
  
  Result CDDANode::setTrack(const int& no) {
    if ((1 > no) || ((int) toc.size() < no)) {
      ERROR_STREAM("FAILURE setting track " << no << endl);
      return FAILURE;
    }
    
    if (isStarted() && !getProducingFlag()) {
      setProducingFlag(true);
    }
    
    return jumpToSector(toc.getStartSector(no-1));
  }
  
  nmm_int32 CDDANode::getTrack() {
    return track_number + 1;
  }
  
  void CDDANode::setRepeat(const nmm_bool& r) {
    repeat = r;
  }
  
  nmm_bool CDDANode::getRepeat() {
    return repeat;
  }
  
  string CDDANode::getTrackName(const nmm_int32& no) {
    if ((1 > no) || ((int) toc.size() < no)) {
      return ""; // FAILURE
    }
    if(!numberTrack) return toc.getName(no-1);
    else {
      stringstream s;
      if ((no)<10)
        s << "0"<< no <<" - " << toc.getName(no-1);
      else 
        s<< no << " - " << toc.getName(no-1);
      return s.str();
    }
  }
  
  string CDDANode::getDiscTitle() {
    return toc.getTitle();
  }
  
  string CDDANode::getDiscArtist() {
    return toc.getArtist();
  }
  
  nmm_int32 CDDANode::getTotalTracks() {
    return toc.size();
  }
  
  Result CDDANode::setCDDBServer(const string& server, const nmm_int32& port) {
    toc.addCDDBServer(server, port);
    cddb_server_added = true;
    return SUCCESS;
  }
  
  void CDDANode::setCDDB(const nmm_bool& cddb) {
    if ( !cddb_server_added ) {
      ERROR_STREAM("You must add a cddb server first!" << endl);
      return; // FAILURE
    }
    
    do_cddb_query = cddb;
    
    if ((!isConstructed()) && do_cddb_query) {
      DEBUG_STREAM("Querying CDDB Server..." << endl);
      toc.queryCDDBServer();
    }
    return; // SUCCESS;
  }
  
  nmm_bool CDDANode::getCDDB(){
    return do_cddb_query;
  }
  
  Result CDDANode::jumpToSector(int sector) {
    if ( (sector < 0) || (sector > toc.getEndSector( toc.size()-1 )) ) {
      ERROR_STREAM("Out of seek range!" << endl);
      return FAILURE;
    }
    
    MutexGuard mg_em(event_mutex);
    
    {
      MutexGuard mg_psm(paranoia_seek_mutex);
      current_sector = sector;
      paranoia_seek(paranoia, current_sector, SEEK_SET);
    }
    
    if (isStarted()) {
      event = SEEKED;
    } else {
      sendtaginfo = true;
      event = START;
      track_number = toc.getTrackBySector(current_sector);
    }
    
    return SUCCESS;
  }
  
  void CDDANode::setOneTrackOnly(const nmm_bool& only) {
    one_track_only = only;
  }
  
  nmm_bool CDDANode::getOneTrackOnly() {
    return one_track_only;
  }
  
  nmm_bool CDDANode::isFinished() {
    return (event == FINISHED);
  }  

  void CDDANode::numberTracks(const nmm_bool& number) {
    numberTrack = number;
  }

  /******************** ISeekable interface ********************/
  nmm_bool CDDANode::hasPercentSeek() const {
    return true;
  }

  void CDDANode::seekPercent(const Rational& percent) {
    Rational temp;
    temp = Rational(current_sector)+(Rational(toc.getSectors(getTrack()))*percent);
    jumpToSector(temp.getNumerator() / temp.getDenominator());
  }
   
  void CDDANode::seekPercentTo(const Rational& percent) {
    Rational internal_percent(percent);
    if (internal_percent > Rational(1)) {
      ERROR_STREAM("Seeking to more than 100% not possible, defaulting to 100%!" << endl);
      internal_percent = Rational(1);
    }
    else if (internal_percent < Rational(0)) {
      ERROR_STREAM("Seeking to negativ not possible, defaulting to 0%!" << endl);
      internal_percent = Rational(0);
    }

    Rational rat_sector = Rational(toc.getSectors(getTrack())) * internal_percent;
    jumpToSector(rat_sector.getNumerator() / rat_sector.getDenominator());
  }

  nmm_bool CDDANode::hasBytesSeek() const {
    return true;
  }

  void CDDANode::seekBytes(const nmm_int64& offset) {
    if ((offset < 0) && (abs(offset) > static_cast<int64_t>(current_sector * CD_FRAMESIZE_RAW))) {
      ERROR_STREAM("Seeking to negativ byte-offset not possible, defaulting to beginning of track!" << endl);
      seekBytesTo(0);
    } else {
      seekBytesTo(current_sector * CD_FRAMESIZE_RAW + offset);
    }
  }

  void CDDANode::seekBytesTo(const nmm_uint64& bytes) {
    u_int64_t internal_bytes(bytes);

    if (internal_bytes > static_cast<u_int64_t>(toc.getSectors(getTrack()) * CD_FRAMESIZE_RAW)) {
      ERROR_STREAM("Seeking over EOT not possible, defaulting to EOT!" << endl);
      internal_bytes = toc.getSectors(getTrack()) * CD_FRAMESIZE_RAW;
    }

    jumpToSector(internal_bytes / CD_FRAMESIZE_RAW);
  }

  nmm_bool CDDANode::hasTimeSeek() const {
    return true;
  }

  void CDDANode::seekTime(const Interval& interval) {
    long new_sector(current_sector + (interval.sec*75 + interval.nsec*75/nano_per_sec));
    if (new_sector < 0) {
      ERROR_STREAM("Seeking into negativ not possible, defaulting to Beginning!" << endl);
      new_sector = 0;
    }
    //if we seeked into the next track start it from the beginning
    if (new_sector>toc.getEndSector(getTrack()-1)) 
      new_sector = toc.getEndSector(getTrack()-1);

    jumpToSector(new_sector);
  }

  void CDDANode::seekTimeTo(const Time& time) {
    jumpToSector(current_sector + (time.sec*75 + time.nsec*75/nano_per_sec));
  }
  /******************** ISeekable interface ********************/
}
