/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <arpa/inet.h>
#include <netdb.h>
#include <stdio.h>
#include <sys/socket.h>
#include <unistd.h>

#include "nmm/utils/NamedObject.hpp"

#include "CDDBServer.hpp"

namespace NMM {

  CDDBServer::CDDBServer(std::string s, int p) {
    server = s;
    port = p;
    toc = std::vector<CDDBTocEntry*>(0);
  } /* Constructor */

  CDDBServer::~CDDBServer() {
    for (unsigned int i=0; i<toc.size(); i++) {
      delete toc[i];
    }
  } /* Destructor */


  void CDDBServer::addTrack(long seconds, long frames) {
    CDDBTocEntry* entry = new CDDBTocEntry();
    entry->seconds = seconds;
    entry->frames = frames;
    if (toc.size() == 0) {
      entry->name = "Lead in";
    } else {
      entry->name = "unknown";
    }
    toc.push_back(entry);
  } /* addTrack */


  Result CDDBServer::doQuery() {
    /* 
     * 1) Local Query in directory ~/.cddb
     * 2) if Local Query fails then Remote Query
     *  
     * think: parameter todo always RemoteQuery
     */

    prepareQuery();

    try {
      if (remoteQuery() == FAILURE) {
        GLOBAL_ERROR_STREAM("Error in remote cddb query!" << std::endl);
        return FAILURE;
      }
      /* write the answer to local folder */
      /* WriteInfo(); */
    } catch (SocketError) {
      GLOBAL_ERROR_STREAM("A error occured while socket operations" << std::endl);
      return FAILURE;
    }

    if (parseAnswer() == FAILURE) {
      GLOBAL_ERROR_STREAM("Error while parsing cddb server answer!" << std::endl);
      return FAILURE;
    }

    return SUCCESS;
  } /* doQuery */


  string CDDBServer::getName(unsigned int tracknr) {
    if (toc.size() <= (tracknr+1)) {
      return "out of range track";
    }
    return toc[tracknr+1]->name;
  } /* getName */


  string CDDBServer::getTitle() {
    return title;
  } /* getTitle */


  string CDDBServer::getArtist() {
    return artist;
  } /* getArtist */

  string CDDBServer::getYear() {
    return year;
  }

  string CDDBServer::getGenre() {
    return genre;
  }


  unsigned int CDDBServer::cddbSum(int n) {
    unsigned int ret;
    
    ret = 0;
    while (n > 0) {
      ret += (n % 10);
      n /= 10;
    }
    return ret;
  } /* cddbSum */

  long CDDBServer::computeCDDBId() {
    unsigned int t = 0, n = 0;

    for (unsigned int i=0; i < toc.size()-1; i++) {
      n += cddbSum(toc[i]->seconds);
    }

    t = (toc[toc.size()-1]->seconds - toc[0]->seconds);

    return ((n % 0xff) << 24 | t << 8 | toc.size()-1);
  }

  void CDDBServer::prepareQuery() {
    char buffer[CDDB_BUFFER_SIZE];
    
    query_string="cddb query ";

    sprintf(buffer,"%08lx %d ", computeCDDBId(), toc.size()-1);

    query_string += buffer;

    for (unsigned int i = 0; i < toc.size()-1; i++) { 
      sprintf(buffer,"%ld ", toc[i]->frames);
      query_string += buffer; 
    }

    sprintf(buffer,"%ld", (toc[toc.size()-1]->frames)/75);
    query_string += buffer;
  } /* computeCDDBId */


  Result CDDBServer::parseAnswer() {
    /* analyse the answer an put it into the toc */
    if ("" == answer_string) {
      return FAILURE;
    }
    char buffer[CDDB_BUFFER_SIZE];
    string trackname;
    int equal,newline;

    equal = answer_string.find("=", answer_string.find("DTITLE"));
    newline = answer_string.find("\n",equal);
    string dtitle = answer_string.substr(equal+1, newline-equal-2);
    
    if (dtitle.find(" / ") == string::npos) {
      title = dtitle;
      artist = "variuos";
    } else {
      artist = dtitle.substr(0, dtitle.find(" / "));
      title = dtitle.substr(dtitle.find(" / ")+3);
    }
    //    GLOBAL_ERROR_STREAM("Found title [" << title << "] and artist [" << artist << "]." << std::endl);
    
    equal = answer_string.find("=", answer_string.find("DYEAR"));
    newline = answer_string.find("\n",equal);
    year = answer_string.substr(equal+1, newline-equal-2);
    
    equal = answer_string.find("=", answer_string.find("DGENRE"));
    newline = answer_string.find("\n",equal);
    genre = answer_string.substr(equal+1, newline-equal-2);

    for(unsigned int i=0; i < toc.size()-1; i++) {
      trackname = "";
      sprintf(buffer ,"TTITLE%d",i);      

      /* find pos "=" from "TTITLE*=" */
      equal = answer_string.find("=", answer_string.find(buffer)); 

      /* find newline */
      newline = answer_string.find("\n",equal); 

      trackname = answer_string.substr(equal+1, newline-equal-2);

      /* remove special chars */
      for(unsigned int j=0; j<trackname.length(); j++) {
        if (trackname[j] == '/') trackname[j] = '-';
      }

      /* there is a invisible \newline at the end of trackname!! */
      toc[i+1]->name = trackname;
    }

    return SUCCESS;
  } /* parseInfo */


  string CDDBServer::readSocket() {
    char buffer[CDDB_BUFFER_SIZE];
    int rc;

    /* read answer from server */
    rc = read(socketd, &buffer, CDDB_BUFFER_SIZE);
    if( 0 > rc ) {
      GLOBAL_ERROR_STREAM("CDDB-Client: Error during receive" << std::endl);
      throw SocketError();
    } 
    buffer[rc] = 0; /* NULL Terminate */
    
    return string(buffer);
  } /* readSocket */


  void CDDBServer::writeSocket(string question) {
    char buffer[CDDB_BUFFER_SIZE];
    unsigned int i;
    int rc;
    
    /* copy question to buffer */
    question += "\n";
    for(i=0; i < question.length(); i++) {
      buffer[i]=question[i]; 
    }
    
    /* Send buffer to server */
    rc = write(socketd, &buffer, question.length());
    if ( 0 > rc ) {
      GLOBAL_ERROR_STREAM("CDDB-Client: Error during send" << std::endl);
      throw SocketError();
    }
  } /* WriteSocket */


  Result CDDBServer::remoteQuery() {
    string question, answer;         
    struct sockaddr_in server_addr;  /* Server address descriptor    */
    struct hostent *host_ptr;        /* Server information structure */
    int rc;                          /* return code */
    int i;
 
    GLOBAL_DEBUG_STREAM("Starting cddb-query...." << std::endl); 

    /* Get the information for the target host */
    GLOBAL_DEBUG_STREAM("CDDB-Client: Getting addr for \"" << server << "\" " << std::endl);
    host_ptr = gethostbyname( server.c_str() );
    if ( NULL == host_ptr ) {
      GLOBAL_ERROR_STREAM("CDDB-Client: Error in gethostbyname" << std::endl);
      return FAILURE;
    }

    /* Fill in the address structure for the server */
    server_addr.sin_family = AF_INET;      /* Address family is Internet */
    server_addr.sin_port   = htons(port);  /* Port number */
    memcpy(&server_addr.sin_addr, host_ptr->h_addr, host_ptr->h_length ); /* IP address */
    for( i=0; i < 8; i++ ) {
      server_addr.sin_zero[i] = 0;
    }
    GLOBAL_DEBUG_STREAM("Addr for " << server << " is (" << inet_ntoa(server_addr.sin_addr) << ")" << std::endl);
    
    /* create the socket */
    GLOBAL_DEBUG_STREAM("CDDB-Client: Creating socket" << std::endl);
    socketd = socket( PF_INET, SOCK_STREAM, IPPROTO_TCP );
    if ( 0 > socketd ) {
      GLOBAL_ERROR_STREAM("CDDB-Client: Error creating socket" << std::endl);
      return FAILURE;
    }


    /* connect the server */
    GLOBAL_DEBUG_STREAM("CDDB-Client: Connecting to server" << std::endl);
    rc = connect( socketd, (struct sockaddr *)&server_addr, sizeof(server_addr) );
    if ( 0 > rc ) {
      GLOBAL_ERROR_STREAM("CDDB-Client: Error connecting server" << std::endl);
      close(socketd);
      return FAILURE;
    }

    /* server info/hello */
    try {
      answer = readSocket(); 
    } catch (...) {
      close(socketd);
      throw SocketError();
    }

    /* handshake */
    try {
      string user("");
      string host("");
      char* user_chrp = getenv("USER");
      if (NULL == user_chrp) {
	GLOBAL_ERROR_STREAM("Failure geting $USER from environment" << std::endl);
	user = string("dummyuser");
      } else {
	user = string(user_chrp);
      }
      char* host_chrp = getenv("HOST");
      if (NULL == host_chrp) {
	GLOBAL_ERROR_STREAM("Failure geting $HOST from environment" << std::endl);
	host = string("dummyhost");
      } else {
	host = string(host_chrp);
      }
      writeSocket("cddb hello " + user + " " + host + " CDDANode develop");
      answer = readSocket(); 
    } catch(...) { 
      close(socketd); 
      throw CDDBError(); 
    }


    /* set protocol level */
    try { 
      writeSocket("proto 5"); 
      answer = readSocket(); 
    } catch(...) { 
      close(socketd); 
      throw CDDBError(); 
    }

    /* the query */
    try { 
      writeSocket(query_string); 
      answer = readSocket(); 
    } catch(...) { 
      close(socketd); 
      throw CDDBError(); 
    }

    /* analyse the answer of the query */
    int ret_code = atoi((answer.substr(0,3)).c_str());
    switch (ret_code) {
    case 200: // Found exact match
      {
	int space1 = answer.find_first_of(' ');
	int space2 = answer.find(' ',space1+1);
	int space3 = answer.find(' ',space2+1);
	
	/* read cd info from server */
	question="cddb read " 
	  + answer.substr(space1+1,space2-space1-1) + " " /* categ  */
	  + answer.substr(space2+1,space3-space2-1);      /* discid */
	
	try {
	  writeSocket(question);
	  answer = readSocket(); 
	  answer += readSocket(); /* the "." need a second call */
	} catch(...) { 
	  close(socketd); 
	  throw CDDBError(); 
	}
      }
      break;
    case 210: // Found exact matches, list follows (until terminating marker)
      {
	/* 
	 * 210 Found exact matches, list follows (until terminating `.')
	 * misc d00fad10 Various Artists / The 80s - Vol. I
	 * rock d00fad10 Various Artists / The 80s (Media Markt) Vol. 1
	 * ...
	 */
	int cut1 = answer.find_first_of('\n')+1;
	int cut2 = answer.find('\n',cut1);
	
	answer = answer.substr(cut1,cut2-cut1-1);
	
	int space1 = answer.find_first_of(' ');
	int space2 = answer.find(' ',space1+1);
	
	question="cddb read " 
	  + answer.substr(0,space1) + " "            /* categ  */
	  + answer.substr(space1+1,space2-space1-1); /* discid */
	
	try {
	  writeSocket(question);
	  answer = readSocket(); 
	  answer += readSocket(); /* the "." need a second call */
	} catch(...) { 
	  close(socketd); 
	  throw CDDBError(); 
	}
      }
      break;
    default:
      GLOBAL_DEBUG_STREAM(REDOUT << "Nothing found!" << NORMALOUT << std::endl);
      //      GLOBAL_DEBUG_STREAM(REDOUT << answer << NORMALOUT << std::endl);
      answer = "";
      break;
    }
    answer_string = answer;

    /* say goodbye */
    try {
      writeSocket("quit");
      answer = readSocket(); 
    } catch(...) { 
      close(socketd); 
      return FAILURE; 
    }

    /* close socket */
    close(socketd);
  
    return SUCCESS;
  } /* Remote Query */

} /* namespace NMM */
