/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <stdio.h>

#include "GenericReadNode.hpp"
#include "nmm/comm/format/Property.hpp"
#include "nmm/comm/messaging/CEvent.hpp"
#include "nmm/plugins/interfaces/ITrackSize.hpp"
#include "nmm/plugins/interfaces/ISeekReceiver.hpp"
#include "nmm/multimedia/interfaces/ITrack.hpp"
#include "nmm/multimedia/interfaces/IProgressListener.hpp"
#include "nmm/multimedia/ISyncReset.hpp"
#include "nmm/multimedia/FlushEvent.hpp"
#include "nmm/plugins/file/generic/utils/FileUtils.hpp"
#include "nmm/utils/directory_utils.hpp"
#include "nmm/utils/SecurityCheck.hpp"

#include "nmm/plugins/NMMRegisterPlugin.hpp"

NMM_REGISTER_PLUGIN(GenericReadNode, NMM, "GenericReadNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  int64_t abs(int64_t x) { return (x >= 0) ? x : -x; }

  GenericReadNode::GenericReadNode(const char *name /* = "GenericReadNode" */,
                                   StreamQueue::Mode mode /* = StreamQueue::MODE_SUSPEND */,
                                   int queue_size /* = StreamQueue::DEFAULT_SIZE */)
    : GenericSourceNode(name, mode, queue_size)
  {
    sendEvent = false;
    send_format = false;
    end_track = false;
    
    filesize = 0;
    filepos = 0;
    seekto = 0;
    
    buffer_size = 10000;
    buffers_before_progress = 5;
    buffers_progress        = 0;
    send_progress = false;
  }
  
  void GenericReadNode::sendProgressInformation(const bool& in) {
    send_progress = in;
  }

  void GenericReadNode::setProgressInterval(const int& buffers) {
    buffers_before_progress = buffers;
  }
  
  nmm_uint64 GenericReadNode::getFilesize() {
    return filesize;
  }

  void GenericReadNode::setFilename(const string& name) {
    filename = name;
  }  
  
  void GenericReadNode::setBufferSize(const int& size) {
    buffer_size = size;
  }
  
  int GenericReadNode::getBufferSize() {
    return buffer_size;
  }
  
  Result GenericReadNode::doInit() {
    Format* out_format = getOwnOutputProperty()->addNewFormat("*/*");
    getOwnOutputProperty()->setDefaultFormat(out_format);
    
    return SUCCESS;
  }
  
  Result GenericReadNode::doDeinit() {
    sendEvent = false;
    send_format = false;
    end_track = false;
    
    filesize = 0;
    filepos = 0;
    seekto = 0;
    
    return SUCCESS;
  }
  
  Result GenericReadNode::doInitOutput() {
    /* Convert filename to an absolute path */
    filename = getAbsoluteNativePath(filename);

    if(!SSecurityCheck::getInstance().checkFilenameForReading(filename)) {
      ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return FAILURE;
    }
    
    in_fstream.clear();

    Format* out_format = FileUtils::getFormat(filename);
    DEBUG_STREAM("Open File: " << filename << endl);
    in_fstream.open(filename.c_str(), ios::in | ios::binary);
	filepos = 0; //reset file position to 0
    if(in_fstream.fail()) {
      ERROR_STREAM( "Error while opening file " << filename << endl);
      return FAILURE;
    }
    
    // get filesize and seek to beginning again
    in_fstream.seekg(0, ios::end);
    filesize = in_fstream.tellg();
    in_fstream.seekg(0, ios::beg);
	
    sendEvent = true;
    if(isOutputConnected("default")) {
      send_format = true;
    }
    
    getOwnWorkingOutputProperty()->addNewFormat(out_format);
    getOwnWorkingOutputProperty()->setDefaultFormat(out_format);

    setProducingFlag(true);
    
    return SUCCESS;
  }

  Result GenericReadNode::doDeinitOutput() {
    in_fstream.close();
    return SUCCESS;
  }
  
  Message *GenericReadNode::processBuffer(Buffer *) {
    if(end_track) {
      DEBUG_STREAM("Sending end_track..." << endl);
      setProducingFlag(false); // nothing left to do, so stop producing
      return new CEvent(Event(ITrack::endTrack_event));
    }
    if(send_progress) {
      Event e;
      if ( buffers_progress > buffers_before_progress ) {
        buffers_progress = 0;
        CEvent* cevent(new CEvent(IProgressListener::create_setProgress(filepos, filesize)));
        cevent->setMandatoryBit(false);
        return cevent;
      }
      buffers_progress++;
    }

    if(sendEvent) {
      CEvent* ce = new CEvent();

      sendEvent = false;
      DEBUG_STREAM("Sending startTrack..." << endl);
      ce->insertEvent(ITrack::create_startTrack(filename)); 
      DEBUG_STREAM("Sending syncReset" << endl);
      ce->insertEvent(ISyncReset::create_syncReset());
      DEBUG_STREAM("Sending setFileSize " << filesize << endl);
      ce->insertEvent(ITrackSize::create_trackSize(filesize));
      
      if(send_format) {
        send_format = false;

        //change outputformat
        setOutputFormat(*getOwnWorkingOutputProperty()->getDefaultFormat());

        DEBUG_STREAM("Sending setFormat" << endl);
        ce->insertEvent(IFormatChange::create_setFormat(*getOutputFormat()));
      }

      return ce;
    }
    else {
      if (seekto != 0) {
        // test if we would get negative
        if (seekto < 0 && abs(seekto) > static_cast<int64_t>(filepos)) {
          // if seek result would be negative go to beginning of file
          in_fstream.seekg(0, ios::beg);
          filepos = 0;
          DEBUG_STREAM("Seeked to beginning " << seekto << " from " << filepos << ", filesize " << filesize << endl);
        }
        else if (filepos + seekto > filesize) {
          // set back to short-before-file-end
          in_fstream.seekg(0, ios::end);
          filepos = filesize;
          DEBUG_STREAM("Seeked to end " << seekto << " from " << filepos << ", filesize " << filesize << endl);
        }
        else {
          // seek normally
          in_fstream.seekg(filepos + seekto);
          filepos = filepos+seekto;
          DEBUG_STREAM("Seeked normally " << seekto << " from " << filepos << ", filesize " << filesize << endl);
        }
        
        // create seek events
        CEvent* cevent = new FlushEvent();
        cevent->insertEvent(cevent->size(), ISyncReset::create_syncReset());
        cevent->insertEvent(ISeekReceiver::create_seeked(filepos));

        seekto = 0;
        return cevent;
      }
      
      // read data from file
      Buffer* out = getNewBuffer(buffer_size);
      if(out == 0) {
        ERROR_STREAM("Failed to allocate output buffer" << endl);
        return 0;
      }
      in_fstream.read(out->getData(), out->getSize());
      unsigned int bytes_read = in_fstream.gcount();
      out->setUsed(bytes_read);

      // check for end of file
      if (in_fstream.eof()) {
        end_track = true;
      }

      // update file position
      filepos+=bytes_read;
      return out;
    } // if(sendEvent) ... else
  }

  GenericReadNode::~GenericReadNode() {
    in_fstream.close();
  }

  /******************** ISeekable interface ********************/
  bool GenericReadNode::hasPercentSeek() const {
    return true;
  }

  void GenericReadNode::seekPercent(const Rational& percent) {
    Rational internal_percent(percent);
    if (internal_percent > Rational(1)) {
      ERROR_STREAM("Seeking more than 100% not possible, defaulting to end!" << endl);
      internal_percent = Rational(1);
    }
    else if (internal_percent < Rational(-1)) {
      ERROR_STREAM("Seeking less than -100% not possible, defaulting to beginning!" << endl);
      internal_percent = Rational(-1);
    }

    Rational rat_seekto = Rational(filesize) * internal_percent;
    seekto = rat_seekto.getNumerator() / rat_seekto.getDenominator();
    DEBUG_STREAM(seekto << " = " << internal_percent << " * " << filesize << endl);
  }
   
  void GenericReadNode::seekPercentTo(const Rational& percent) {
    seekPercent(percent - Rational(filepos, filesize));
  }

  bool GenericReadNode::hasBytesSeek() const {
    return true;
  }

  void GenericReadNode::seekBytes(const nmm_int64& offset) {
    if ((offset < 0) && (abs(offset) > static_cast<int64_t>(filepos))) {
      ERROR_STREAM("Seeking to negativ byte-offset not possible, defaulting to beginning of file!" << endl);
      seekBytesTo(0);
    } else {
      seekBytesTo(filepos + offset);
    }
  }

  void GenericReadNode::seekBytesTo(const nmm_uint64& bytes) {
    u_int64_t internal_bytes(bytes);

    if (internal_bytes > filesize) {
      ERROR_STREAM("Seeking over EOF not possible, defaulting to EOF!" << endl);
      internal_bytes = filesize;
    }

    seekto = internal_bytes - filepos;

    DEBUG_STREAM(seekto << " = " << internal_bytes << " - " << filepos << endl);
  }

  bool GenericReadNode::hasTimeSeek() const {
    return false;
  }

  void GenericReadNode::seekTime(const Interval&) {
  }

  void GenericReadNode::seekTimeTo(const Time&) {
  }
  /******************** ISeekable interface ********************/
}

#include "nmm/nmm_suffix.hpp"
