/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "GenericWriteNode.hpp"

#include "nmm/comm/format/Property.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"
#include "nmm/utils/SecurityCheck.hpp"
#include "nmm/utils/thread/MutexGuard.hpp"

NMM_REGISTER_PLUGIN(GenericWriteNode, NMM, "GenericWriteNode")

namespace NMM {

  GenericWriteNode::GenericWriteNode(const char* name,
                                     const StreamQueue::Mode queue_mode,
                                     const int queue_size)
    : GenericSinkNode(name, queue_mode, queue_size )
    , __file_extension()
    , __working_directory()
    , __override_default_filename(false)
    , __sub_state(CLOSED)
    , __filename()
    , __bytes_written(0)
    , __buffnum(0)
  {}

  GenericWriteNode::~GenericWriteNode() {
    __out_fstream.close();
  }

  Result GenericWriteNode::doInit() {
    Format* myInputFormat = getOwnInputProperty()->addNewFormat("*/*");
    getOwnInputProperty()->setDefaultFormat(myInputFormat);
    return SUCCESS;
  }

  Result GenericWriteNode::doInitOutput() {
    Format* myInputFormat = getOwnWorkingInputProperty()->addNewFormat("*/*");
    getOwnWorkingInputProperty()->setDefaultFormat(myInputFormat);
    return SUCCESS;
  }

  Result GenericWriteNode::doActivate() {
    return SUCCESS;
  }

  Result GenericWriteNode::doDeactivate() {
    MutexGuard mg(__substate_mutex);

    /* Close file if opened */
    if (__sub_state == OPENED) {
      closeFile();
      __sub_state = CLOSED;
    }

    /* Begin> Reset state of the node */
    __sub_state = CLOSED;
    __override_default_filename = false;
    __filename = "";
    __bytes_written = 0;
    __buffnum = 0;
    /* End< */

    return SUCCESS;
  }

  void GenericWriteNode::setFilename(const string& filename) {
    MutexGuard mg(__substate_mutex);

    switch(__sub_state) {
    case OPENED:
      WARNING_STREAM("Received setFilename event, but file already opened" << endl);
      return;

    case CLOSED:
    case STARTED:
      __filename = filename;
      __override_default_filename = true;
      return;
    }
  }

  void GenericWriteNode::setWorkingDirectory(const string& wd) {
    MutexGuard mg(__substate_mutex);
    __working_directory = wd;
  }

  string GenericWriteNode::getWorkingDirectory() {
    MutexGuard mg(__substate_mutex);
    return __working_directory;
  }

  void GenericWriteNode::setFileExtension(const string& ext) {
    MutexGuard mg(__substate_mutex);
    __file_extension = ext;
  }

  string GenericWriteNode::getFileExtension() {
    MutexGuard mg(__substate_mutex);
    return __file_extension;
  }

  Result GenericWriteNode::startTrack(const string& filename) {
    MutexGuard mg(__substate_mutex);

    switch (__sub_state) {
    case OPENED:
      ERROR_STREAM("Received startTrack event, but file already opened" << endl);
      return FAILURE;

    case STARTED:
      ERROR_STREAM("Received multiple startTrack events but no endTrack event" << endl);
      return FAILURE;

    case CLOSED:
      /* Make default filename if no other filename has been set */
      if (!__override_default_filename) {
        makeDefaultFilename(filename);
      }

      /* We are started now */
      __sub_state = STARTED;
    }

    return SUCCESS;
  }

  Result GenericWriteNode::endTrack() {
    MutexGuard mg(__substate_mutex);

    switch (__sub_state) {
    case CLOSED:
      ERROR_STREAM("Received endTrack event, but file not opened" << endl);
      return FAILURE;      

    case OPENED:
      closeFile();
      /* fall through */

    case STARTED:
      /* Reset the state of the node. Set the filename to empty to prevent previous file from */
      /* being overwritten */
      __sub_state = CLOSED;
      __filename = "";
      __override_default_filename = false;  
    }

    return SUCCESS;
  }

  Result GenericWriteNode::finished() {
    DEBUG_STREAM("Finished event." << endl);    
    return SUCCESS;
  }

  Result GenericWriteNode::openFile() {
    /* Reset state of stream */
    __out_fstream.clear();

    /* Check if a filename has been set */
    if (__filename.empty()) {
      ERROR_STREAM("No filename has been set!" << endl);
      return FAILURE;
    }

    /* Security check */
    if (!SSecurityCheck::getInstance().checkFilenameForWriting(__filename)) {
      ERROR_STREAM("Security check not passed for file " << __filename << ", check .nmmrc for options" << endl);
      return FAILURE;
    }

    /* Open stream */
    DEBUG_STREAM("Open File: " << __filename << endl);
    __out_fstream.open(__filename.c_str(), ios::binary | ios::trunc | ios::out);
    if (!__out_fstream) {      
      ERROR_STREAM("Failed to open file" << endl);
      return FAILURE;
    }

    return SUCCESS;
  }

  Result GenericWriteNode::closeFile() {
    /* Close stream */
    DEBUG_STREAM("Close File: " << __filename << endl);
    __out_fstream.close();

    return SUCCESS;
  }

  void GenericWriteNode::makeDefaultFilename(const string& filename) {
    __filename = filename;
    if (!__file_extension.empty()) {
      __filename += "." + __file_extension;
    }
    if (!__working_directory.empty()) {
      __filename = __working_directory + "/" + __filename;
    }
  }

  Message *GenericWriteNode::processBuffer(Buffer *in_buffer) {
    MutexGuard mg(__substate_mutex);

    switch (__sub_state) {
    case CLOSED:
    case STARTED:
      if (SUCCESS != openFile()) {
        __sub_state = CLOSED;
        break;
      }
      __sub_state = OPENED;
      /* fall through */

    case OPENED:
      {
        int use = in_buffer->getUsed();

        /* in_buffer is data */
        __out_fstream.write(in_buffer->getData(), in_buffer->getUsed());
        __bytes_written += use;
        DEBUG_STREAM("Written " << __bytes_written << " Bytes" << endl);

        break;
      }
    }

    in_buffer->release();
    return NULL;
  }

}
