/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include "FileUtils.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  Format* FileUtils::getFormat(const string& filename) {
    
    Format* out_format;

    // first try to determine a format out of the file extension
    string extension = string();
    string::size_type extension_pos = filename.rfind(".");
    if(extension_pos != string::npos && extension_pos != filename.length() - 1) {
      extension = filename.substr(extension_pos+1);
      for (unsigned int i = 0; i < extension.length(); i++) {
	extension[i] = tolower(extension[i]);
      }
    }
    
    out_format = getFormatFromExtension(extension);

    if (!out_format) out_format = getFormatFromContent(filename);
    if (!out_format) out_format = new Format(Format::wild_wild_type);

    return out_format;
  }

  Format* FileUtils::getFormatFromExtension(const string& extension)
  {
    string format_type = "";
   
    if(extension == "mp3") {
      format_type = Format::audio_mpeg_type;
    }
    else if((extension == "vob") || 
	    (extension == "mpg") ||
	    (extension == "mpeg") ||
	    (extension == "vdr")) {
      format_type = Format::av_mpegps_type;
    }
    else if(extension == "wav") {
      format_type = Format::audio_raw_type;
    }
    else if(extension == "ac3") {
      format_type = Format::audio_ac3_type;
    }
    else if(extension == "ogg") {
      format_type = Format::audio_x_ogg_type;
    }
    else if(extension == "ogm") {
      format_type = Format::av_ogm_type;
    }
    else if(extension == "wmv") {
      format_type = Format::av_asf_type;
    }
    else if(extension == "wma") {
      format_type = Format::audio_wma_type;
    }

    if(format_type.empty()) {
      return 0;
    }
    else {
      return new Format(format_type);
    }
  }

  Format* FileUtils::getFormatFromContent(const string& path)
  {
    string format_type = "";

    string mime_type = FileUtils::getMIMETypeFromFile(path);

    // 'file' will return audio/x-ogg for all ogg/ogm -> check again
    if(mime_type == Format::audio_x_ogg_type) {
      string mime_type2 = FileUtils::getMIMETypeFromFile(path, false);
      if(mime_type2.find("Vorbis audio") != string::npos) {
	format_type = Format::audio_x_ogg_type;
      }
      else {
	format_type = Format::av_ogm_type;
      }
    }

    // 'file' will return video/mpeg for all mpeg a/v files -> reset to av/mpegps
    else if(mime_type == Format::video_mpeg_type) {
      format_type = Format::av_mpegps_type;
    }

    // audio/x-wav -> audio/raw
    else if(mime_type.find(Format::audio_x_wav_type) != string::npos) {
      format_type = Format::audio_raw_type;
    }

    // avi files will output msvideo -> check again for AVI
    else if(mime_type.find("msvideo") != string::npos) {
      string mime_type2 = FileUtils::getMIMETypeFromFile(path, false);
      if(mime_type2.find("AVI") != string::npos) {
	format_type = Format::av_avi_type;
      }
    }

    if(format_type.empty()) {
      return 0;
    }
    else {
      return new Format(format_type);
    }
  }

  string FileUtils::getMIMETypeFromFile(const string& filename, bool mime_only /* = true */ )
  {
    string working_filename = urldecode(filename);

    std::FILE* read_fp;
    char buffer[256];
    int chars_read;
    string command;
    if(mime_only) {
      command = "/usr/bin/file --mime '" + working_filename+"'";
    }
    else {
      command = "/usr/bin/file '" + working_filename+"'";
    }
    string mime_type = "";

    memset(buffer, '\0', sizeof(buffer));

	read_fp = popen(command.c_str(), "r");

	if(read_fp != NULL) {
      chars_read = fread(buffer, sizeof(char), 255, read_fp);
      if(chars_read > 0) {
      
	string tmp(buffer);
	string::size_type pos = tmp.rfind(working_filename, 0);

	mime_type = tmp.substr(pos+working_filename.size()+2, tmp.size()-pos-working_filename.size()-3);
      }
	  pclose(read_fp);
    }
    
    return mime_type;
  }

}

#include "nmm/nmm_suffix.hpp"
