/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "MagickWriteNode.hpp"

#include "nmm/comm/format/Property.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"
#include "nmm/utils/SecurityCheck.hpp"

#include <fstream>
#include <string>

NMM_REGISTER_PLUGIN(MagickWriteNode, NMM, "MagickWriteNode")

namespace NMM{

  MagickWriteNode::MagickWriteNode(const char* name)
    : GenericSinkNode(name), image_resolution(0,0), bit_per_pixel(0)
  {
    
  }
  
  Result MagickWriteNode::doInit() {
    // Set the supported formats.
    // The formats rgb24 with 24 and 32 bits are supported.
    Format* rgb_format = getOwnInputProperty()->addNewFormat("video/raw");
    rgb_format->addWildcard(Format::x_resolution_param);
    rgb_format->addWildcard(Format::y_resolution_param);
    rgb_format->addWildcard(Format::ratio_x_param);
    rgb_format->addWildcard(Format::ratio_y_param);
    rgb_format->addStringParamValue(Format::channelorder_param,"rgb" );
    rgb_format->addStringParamValue(Format::endian_param, "little");
    rgb_format->addStringParamValue(Format::colorspace_param,"rgb24");
    rgb_format->addIntParamValue(Format::bitperpixel_param,32);
    rgb_format->addStringParamValue(Format::format_param, "packet");
    rgb_format->addWildcard(Format::framerate_param);

    // The default format has 32 bits.
    getOwnInputProperty()->setDefaultFormat(rgb_format);

    rgb_format = getOwnInputProperty()->addNewFormat("video/raw");
    rgb_format->addWildcard(Format::x_resolution_param);
    rgb_format->addWildcard(Format::y_resolution_param);
    rgb_format->addWildcard(Format::ratio_x_param);
    rgb_format->addWildcard(Format::ratio_y_param);
    rgb_format->addStringParamValue(Format::channelorder_param,"rgb");
    rgb_format->addStringParamValue(Format::endian_param, "little");
    rgb_format->addStringParamValue(Format::colorspace_param,"rgb24");
    rgb_format->addIntParamValue(Format::bitperpixel_param,24);
    rgb_format->addStringParamValue(Format::format_param, "packet");
    rgb_format->addWildcard(Format::framerate_param);

    return SUCCESS;
  } // doInit

  Result MagickWriteNode::doInitOutput() {
    // Initialize internal variables depending on the input format.

    const Format* in_format = getInputFormat();
    if (in_format == NULL)
      return FAILURE;

    try {
      image_resolution.first   = in_format->getIntValue(Format::x_resolution_param);
      image_resolution.second  = in_format->getIntValue(Format::y_resolution_param);
      bit_per_pixel            = in_format->getIntValue(Format::bitperpixel_param );
    } catch (...) {
      return FAILURE;
    }

    return SUCCESS;
  } // doInitOutput

  Message* MagickWriteNode::processBuffer(Buffer* buffer) {
    // Create a new Magick::Image object.
    Magick::Image image;

    // Read the pixel data from the buffer, depending on the number of bits per pixel.
    if (bit_per_pixel == 24) {
      image.read(image_resolution.first, image_resolution.second,
		 "BGR", Magick::CharPixel, buffer->getData());
    }
    else {
      image.read(image_resolution.first, image_resolution.second,
		 "BGRA", Magick::CharPixel, buffer->getData());
    }

    // Release the buffer.
    buffer->release();

    // Write the image to disk. The extension given in file_name determines the image format.
    try {
      /* Security check */
      if (!SSecurityCheck::getInstance().checkFilenameForWriting(file_name)) {
        ERROR_STREAM("Security check not passed for file " << file_name << ", check .nmmrc for options" << endl);
        return 0;
      }
      
      image.write(file_name);
    }
    catch (const Magick::Exception& e) {
      ERROR_STREAM(e.what() << endl);
    }
    catch (...) {
      ERROR_STREAM("Unknown exception." << endl);
    }


    return 0;
  } // processBuffer

  pair<int, int> MagickWriteNode::getImageResolution() {
    // Return the image resolution.
    return image_resolution;
  }

  void MagickWriteNode::setFilename(const string& file_name) {
    // Set the internal variable file_name.
    this->file_name = file_name;
  }

} //end namespace
