/* Network-Integrated Multimedia Middleware (NMM)
* http://www.motama.com/nmm.html
*  
* Copyright (C) 2006-2007
* Motama GmbH, Saarbruecken, Germany
* http://www.motama.com
*  
* See file LICENCE for terms and conditions of usage
*
*/

#include "MP3ReadNode.hpp"

#include <fcntl.h>

#include "nmm/comm/ProxyObject.hpp"
#include "nmm/multimedia/interfaces/ITrack.hpp"
#include "nmm/multimedia/interfaces/ITrackList.hpp"
#include "nmm/multimedia/FlushEvent.hpp"
#include "nmm/multimedia/IFormatChange.hpp"
#include "nmm/multimedia/interfaces/IProgressListener.hpp"
#include "nmm/multimedia/interfaces/ITrackDuration.hpp"
#include "nmm/multimedia/FileNotFoundException.hpp"
#include "nmm/plugins/interfaces/ITrackSize.hpp"
#include "nmm/multimedia/ISyncReset.hpp"
#include "nmm/plugins/interfaces/ITagInfo.hpp"
#include "nmm/utils/SecurityCheck.hpp"
#include "nmm/utils/time/Rational.hpp"
#include "nmm/utils/time/TimedElement.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"

#include <iostream>
#include <iomanip>

#include "genre.h"


#define ID3_HEADER_SIZE 10
#define ID3_TAG_SIZE 128

NMM_REGISTER_PLUGIN(MP3ReadNode, NMM, "MP3ReadNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  int64_t abs(int64_t x) { return (x >= 0) ? x : -x; }

  MP3ReadNode::MP3ReadNode(const char *name,
    const StreamQueue::Mode queue_mode,
    const int size)
    : GenericSourceNode(name, queue_mode, size),
    track_duration(0, 0)
  {
    filename      = "";
    buffer_size   = 1024 * 2;
      
    fmpeg       = 0;
    flayer      = 0;    
    fprotbit    = 0;
    fbitrate    = 0;
    fsample_rate = 0;
    fpadding    = 0;
    fnumchannels = 0;

    mp3_data_start = 0;
    mp3_data_end = 0;
    frames_pos = 0;
    frames_total = 0;
    have_mp3_frame_map = false;

    filepos          = 0;

    state = SET_TITLE;
    seekto = -1;
    seekto_frame = -1;
    send_progress = false;
    found_tag = false;
    send_format = false;

    buffers_before_progress = 5;
    buffers_progress        = 0;

    timestamp.sync_time = Time(0);
    timestamp.stream_counter = 0;
    timestamp.is_valid = false;
  }

  void MP3ReadNode::getMetaData(string& _title, string& _album, string& _artist,  string& _year,  string& _genre) {
    if (found_tag) {
      if (genre_count<genre || genre<0)
        _genre="Unknown";
      else
        _genre=genre_table[genre];
      _title = title;
      _album = album;
     _artist = interpret;
      _year = year;
   }
  }
  


  nmm_uint64 MP3ReadNode::getFilesize() {
    return filesize;
  }

  void MP3ReadNode::setFilename(const string& name) {
    filename = name;

    // if already playing one file, immediately open the new file.
    if (in.is_open()) {
      if (FAILURE == openFile()) {
        // XXX: in this case, setFilename can fail!
        setProducingFlag(false);
        return;
      } else {
        setProducingFlag(true);
      }

      // as an "end_track" causes the application to react (e.g. in setting the *next* filename)
      // we do *not* send an "end_track" after setting a filename
      //       if (state == END_TRACK) {
      state = START;
      //       } else {
      //         state = NEXT_TRACK;
      //       }
    }
  }

  void MP3ReadNode::setBufferSize(const nmm_int32& size) {
    buffer_size = size;
  }

  nmm_int32 MP3ReadNode::getBufferSize() {
    return buffer_size;
  }

  void MP3ReadNode::sendProgressInformation(const nmm_bool& f ) {
    send_progress = f;
  }

  void MP3ReadNode::setProgressInterval(const nmm_int32& buffers) {
    buffers_before_progress = buffers;
  }

  Result MP3ReadNode::doInit()
  {
    fmpeg       = 0;
    flayer      = 0;
    fprotbit    = 0;
    fbitrate    = 0;
    fsample_rate = 0;
    fpadding    = 0;
    fnumchannels = 0;
    filepos          = 0;

    timestamp.sync_time = Time(0);
    //timestamp.stream_counter = 0;
    timestamp.is_valid = false;

    // general output format
    Format* out_format = getOwnOutputProperty()->addNewFormat("audio/mpeg");
    out_format->addWildcard(Format::channels_param); //
    out_format->addWildcard(Format::layer_param);
    out_format->addIntParamValue(Format::audio_codec_param, 0x55);
    out_format->addIntParamValue(Format::sample_rate_param, 48000);
    out_format->addIntParamValue(Format::sample_rate_param, 44100);
    out_format->addIntParamValue(Format::sample_rate_param, 32000);
    out_format->addIntParamValue(Format::sample_rate_param, 24000);
    out_format->addIntParamValue(Format::sample_rate_param, 22050);
    out_format->addIntParamValue(Format::sample_rate_param, 16000);
    out_format->addIntParamValue(Format::sample_rate_param, 12000);
    out_format->addIntParamValue(Format::sample_rate_param, 11025);
    out_format->addIntParamValue(Format::sample_rate_param, 8000);
    out_format->addIntParamValue(Format::bitpersample_param, 16);

    out_format = getOwnOutputProperty()->addNewFormat("audio/mpeg");
    out_format->addWildcard(Format::layer_param);
    out_format->addIntParamValue(Format::audio_codec_param, 0x55);
    getOwnOutputProperty()->setDefaultFormat(out_format);

    return SUCCESS;
  }

  Result MP3ReadNode::openFile() 
  {
    int header_length = 0;
    filepos    = 0;
    frames_pos = 0;
    found_tag = false;

    if(!SSecurityCheck::getInstance().checkFilenameForReading(filename)) {
      ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      throw FileNotFoundException(this, "Security check not passed.", filename);
    }

    // close old file first
    if(in.is_open()){
      in.close();
    }
    in.clear();

    DEBUG_STREAM("Open File: " << filename << endl);
    in.open(filename.c_str(), fstream::in | ios::binary);

    if (!in.is_open()) {
      ERROR_STREAM("Could not open file!"<< endl);
      throw FileNotFoundException(filename);
    }

    // get filesize and seek to beginning
    in.seekg(0, ios::end);
    filesize = in.tellg();
    mp3_data_end = filesize;
    in.seekg(0, ios::beg);
    DEBUG_STREAM("File size is " << filesize << endl);
    header_length = isWav();
    mp3_data_start = header_length;
	in.seekg(0, ios::beg);
    
	// if not a wave file   
    if ( header_length == 0 )
    {
      // the ID3v2 header is at the begining of the file to avoid error
      // you must read just after the IDV3 Header if it exists
      header_length = isID3v2(); 
      mp3_data_start = header_length;
	  mp3_data_start = analyseFirstValidMP3Frame();
	}

    int id3v1_start = isID3v1(); 
    if(0 < id3v1_start){ 
      //ID3v1 header found
      mp3_data_end = id3v1_start;
    };

    // go to begin of mp3 data 
    in.seekg(mp3_data_start,ios::beg);

    // clear working output property
    getOwnWorkingOutputProperty()->erase();

    // initialize working output formats
    Format* out_format = getOwnWorkingOutputProperty()->addNewFormat("audio/mpeg");
    out_format->addIntParamValue(Format::layer_param, flayer);
    out_format->addIntParamValue(Format::audio_codec_param, 0x55);
    // getOwnWorkingOutputProperty()->setDefaultFormat(out_format);
    new_simple_format = *out_format;

    // complex format. uses values obtained in parseHeader
    out_format = getOwnWorkingOutputProperty()->addNewFormat("audio/mpeg");
    out_format->addIntParamValue(Format::channels_param, fnumchannels);
    out_format->addIntParamValue(Format::layer_param, flayer);
    out_format->addIntParamValue(Format::audio_codec_param, 0x55);
    out_format->addIntParamValue(Format::sample_rate_param, fsample_rate);
    out_format->addIntParamValue(Format::bitpersample_param, 16);
    getOwnWorkingOutputProperty()->setDefaultFormat(out_format);

    // store complex format as "old" format for testing when having to send format changes!
    old_format = out_format;

    state = START;

    return SUCCESS;
  }

  Result MP3ReadNode::doInitOutput() {
    // open first file.
    // we must have a file name before we can initialize the output
    if (FAILURE == openFile()/*filename == ""*/) {
      ERROR_STREAM("Need filename!");
      return FAILURE;
    }
	frames_pos = 0;
	
    // check if we are already connected
    if(isOutputConnected("default") && old_format != 0) {
      if(!((*getOutputFormat()).isFormatIdentical(old_format))){
        send_format = true;
      }
    }

  if(state == TRACK_ENDED || state == FINISHED) {
      DEBUG_STREAM1("WAVReadNode: START" << endl);
      state = START;
    } else {
      DEBUG_STREAM1("WAVReadNode: START_FLUSH" << endl);
      state = START_FLUSH;
    }

    //state = START_TRACK;
    //state = SET_TITLE;

    setProducingFlag(true);

    return SUCCESS;
  }

  Result MP3ReadNode::doDeinitOutput() {
    in.close();
    return SUCCESS;
  }

  Result MP3ReadNode::doActivate() {
    // get the connection format  
    const Format* out_format(getOutputFormat());

    if (!out_format) {
      ERROR_STREAM("Output format is NULL!" << endl);
      return FAILURE;
    }

    // store connection format as the "old" one
    old_connection_format = *out_format;

	//analyse opened mp3 file
    frames_total = initializeMP3FrameMap();
    track_duration= Rational(getSamplesPerFrame(fmpeg,flayer) * frames_total, fsample_rate);

    return SUCCESS;
  }
  Message *MP3ReadNode::processBuffer(Buffer *) {
    int bytes_read;

    /* check internal state */
    Message* out = NULL;
    switch(state) {
    case SET_TITLE:
      {
        if (found_tag) {
          state = TAGINFO;
        } else {
          state = START;
        }
      }
      break;
    case TAGINFO:
      {
        DEBUG_STREAM("Sending Taginfo..."<<endl);
        string _genre;
        if (genre_count<genre || genre<0)
          _genre="Unknown";
        else
          _genre=genre_table[genre];
        Event* e = ITagInfo::create_TagInfo(title,album,interpret,year,_genre);
        state = START;
        out = new CEvent(e);
      }
      break;
    case START_FLUSH:
    case START:
      {
        //estimate title from filename
        string::size_type pos1, pos2;
        pos1 = (int)filename.rfind("/");
        if (filename.npos == pos1) { 
          pos1 = (int)filename.rfind("\\");
          if(filename.npos == pos1){
            pos1=0; 
          }else{
            pos1++;
          }
        }
        pos2 = (int)filename.find(".mp3",pos1);

        CEvent* ce = 0;
        if(state == START) {
          ce = new CEvent();
        } else {
          ce = new FlushEvent();
        }

        ce->insertEvent(ITrack::create_startTrack(filename.substr(pos1,pos2-pos1)));
        DEBUG_STREAM("Sending startTrack..." << endl);
        ce->insertEvent(ITrackSize::create_trackSize(filesize));
        DEBUG_STREAM("Sending trackSize" <<endl);
        
        if(send_format) {
          send_format = false;
         
          //set new format
          setOutputFormat(*getOwnWorkingOutputProperty()->getDefaultFormat());

          DEBUG_STREAM("Sending setFormat" << endl);
          ce->insertEvent(IFormatChange::create_setFormat(*getOutputFormat()));  
        }

        ce->insertEvent(ISyncReset::create_syncReset());
        DEBUG_STREAM("Sending trackDuration" << endl);
        ce->insertEvent(ITrackDuration::create_trackDuration(track_duration));        

        
        
        state = DATA;
        out = ce;
      }
      break;
    case NEXT_TRACK:
      {
        DEBUG_STREAM("Sending endTrack..." << endl);
        Event e (ITrack::endTrack_event);
        state = START;
        out = new CEvent(e);
      }
      break;
    case DATA:
      {
        if (seekto >= 0) {
          // get current position in file
          filepos = in.tellg();

          // seek normally
          in.seekg(static_cast<streamoff>(seekto));
          DEBUG_STREAM("Seeked normally " << seekto << " from " << filepos << ", filesize " << filesize << endl);
          filepos = in.tellg();//lseek(fd, 0, SEEK_CUR);
          // make sure, progress is send next
          buffers_progress = buffers_before_progress +1;
          CEvent* cevent = new FlushEvent();
          cevent->insertEvent((unsigned int)cevent->size(), ISyncReset::create_syncReset());

          seekto=-1;
          return cevent;
        } else if(seekto_frame >=0){
          in.seekg(static_cast<streamoff>(mp3_header[seekto_frame]));
          frames_pos = seekto_frame;
          filepos = in.tellg();
          CEvent* cevent = new FlushEvent();
          cevent->insertEvent(cevent->size(), ISyncReset::create_syncReset());

          seekto_frame=-1;
          out = cevent;
        }else{
          Buffer* buffer = NULL;
          if ((buffer = getNewBuffer( buffer_size )) != 0) {
			      int frames_written = getNextFrames( buffer->getData(), buffer_size ); 
			      frames_pos += frames_written;
            bytes_read = in.gcount();
            filepos = in.tellg();
            if ( frames_written > 0 ) {
              buffer->setUsed( bytes_read );
              out = buffer;
              timestamp.sync_time = Time(Rational(frames_pos * getSamplesPerFrame(fmpeg,flayer), fsample_rate));
              timestamp.is_valid = true;
            } else {
              /* assume EOF */
              DEBUG_STREAM("No frames read. => assuming end of file... " << endl);
              setProducingFlag(false);
              state = TRACK_ENDED;
              Event e (ITrack::endTrack_event);
              buffer->release();
              out = new CEvent(e);
            }
          } else {
            out = NULL;
          }
        }
      }
      break;
    case TRACK_ENDED:
      if(true) { // TODO: if (file queue is empty)
        state = FINISHED;
        DEBUG_STREAM("Send finished event " << endl);
        setProducingFlag(false);
        out = NULL;
        return new CEvent(ITrackList::create_endTrackList());
      }
      break;
    default:
      ERROR_STREAM("Invalid internal state" << endl);
      state = DATA;
      out = NULL;
    }

    /* set correct timestamp */
    if(out) {
      out->setTimestamp(timestamp);
      ++timestamp.stream_counter;
    }

    return out;
  }

  MP3ReadNode::~MP3ReadNode() {
    in.close();
  }

  /**
  * Read the ID3v2 header at the begining of the file if it exists
  * it avoids an possible error, if you try to find the mp3 header 
  * on the IDV3v2 .  
  */
  int MP3ReadNode::isID3v2() {
    int   bytes_read;
    char  buffer[ID3_HEADER_SIZE];
    int   len;

    in.read( buffer, ID3_HEADER_SIZE );
    bytes_read = in.tellg();
    if ( bytes_read != ID3_HEADER_SIZE ) return(0);
    if ( strncmp( buffer, "ID3", 3 ) == 0 ) {
      if ( ( buffer[3] != (char)0xff ) && ( buffer[4]!=(char)0xff ) && 
        ((buffer[6] & 0x80 ) == 0 ) &&
        ((buffer[7] & 0x80 ) == 0 )&& 
        ((buffer[8] & 0x80 ) == 0 )&& 
        ((buffer[9] & 0x80 ) == 0 ) ) {

          len = ((buffer[6] & 0x7f ) << 21 ) |
            ((buffer[7] & 0x7f ) << 14 ) |
            ((buffer[8] & 0x7f ) << 7 ) |
            (buffer[9] & 0x7f );
          in.seekg(len, ios::cur); 

          MESSAGE_STREAM("ID3V2 Header found (size=" << ID3_HEADER_SIZE+len<< ")"  << endl );
          return(len+ID3_HEADER_SIZE);
      }
    }
    DEBUG_STREAM("ID3V2 Header not found" << endl );
    //with other case 
    //go back to the begining of the file
    in.seekg(0, ios::beg);
    return 0;
  }

  int MP3ReadNode::isID3v1(){
    char tag[4];
    int id3v1_pos = 0;

    in.seekg(-ID3_TAG_SIZE, ios::end);
    in.read(tag, 3);
    if (!strncmp(tag, "TAG", 3)) {
      // found id3 v1 tag
      MESSAGE_STREAM("Found ID3 V1 tag" << endl);
      int count;
      in.read(title, 30);
      count = in.gcount();
      title[count]='\0';
      DEBUG_STREAM("title: " << title << endl);

      in.read(interpret,30);
      count = in.gcount();
      interpret[count]='\0';
      DEBUG_STREAM("interpret: " << interpret << endl);

      in.read(album,30);
      count = in.gcount();
      album[count] = '\0';
      DEBUG_STREAM("album: " << album << endl);

      in.read(year,4);
      count = in.gcount();
      year[count] = '\0';
      DEBUG_STREAM("year: " << year << endl);

      in.read(comment,30);
      count = in.gcount();
      comment[count]='\0';
      DEBUG_STREAM("comment: " << comment << endl);

      char nr[2];
      in.read(nr,1);
      count = in.gcount(); 
      nr[count]='\0';
      genre=*nr;

      found_tag = true;
      mp3_data_end -= ID3_TAG_SIZE;
      id3v1_pos = mp3_data_end;
    }
    return mp3_data_end;
  }

  /**
  * Node: this function only skips the wav header if exists
  * it doesn't parse the header, for format chechking the parseHeader
  * function is used
  */
  int MP3ReadNode::isWav() {
    int   bytes_read;
    bool  data_found;
    char  buffer[4];
    int   mp3pos = 0;

    in.read(buffer, 4 );
    bytes_read = in.gcount();

    if ( bytes_read != 4 ) return(0);

    if ( strncmp( buffer, "RIFF", 4 ) == 0 ) {
      mp3pos = 4;

      MESSAGE_STREAM("RIFF Header found" << endl );

      data_found = false;

      /* look for data chunk and set mp3pos to end of data chunk */
      while( !data_found ) {
        in.seekg(mp3pos, ios::cur);
        filepos +=mp3pos;
        in.read(buffer, 4 );
        bytes_read = in.tellg();
        if ( bytes_read == 4 ) {
          mp3pos += 4;
          if ( strncmp( buffer, "data", 4 ) == 0 ) {

            data_found = true;
            //skip data chunk
            mp3pos += 4;

            DEBUG_STREAM("DATA chunk found, set fileposition to: " << mp3pos << endl);

          } else mp3pos -= 3;
        }
      }

    }
    else mp3pos = 0;

    return( mp3pos );

  }

  bool MP3ReadNode::parseHeader(char* _buffer) {
    /* MP3 Header 32 Bit
    example: 1111 1111  1111 1011  1011 0000  0000 0100
    |-b0----|  |-b1----|  |-b2----|  |-b3----|
    */

    char b  = 0;
    char b0 = *(_buffer + 0); // verbose(b0); cout << " ";
    char b1 = *(_buffer + 1); // verbose(b1); cout << " "; 
    char b2 = *(_buffer + 2); // verbose(b2); cout << " ";
    char b3 = *(_buffer + 3); // verbose(b3); cerr << endl;

    /* Bit 31-21 : Sync, should be 1111 1111 111 */
    if ( ((b0 & 0xff) != 0xff) || ((b1 & 0xe0) != 0xe0) ) {
      DEBUG_STREAM("No sync found" << endl);
      return false;
    } 

    /* Bit 20,19 : MPEG Audio version ID */
    switch ((b1 & 0x18) >> 3) {
    case 0 : fmpeg = 25; break; /* 00 = MPEG Version 2.5 */
    case 1 : fmpeg =  0; break; /* 01 = reserved */
    case 2 : fmpeg =  2; break; /* 10 = MPEG Version 2 (ISO/IEC 13818-3) */
    case 3 : fmpeg =  1; break; /* 11 = MPEG Version 1 (ISO/IEC 11172-3) */
    default: return false; //no correct header
    }
    //DEBUG_STREAM("MPEG Version : " << (int) fmpeg << endl);


    /* Bit 18,17 : Layer description */
    switch((b1 & 0x06) >> 1) {
    case 0 : flayer = 0; break; /* 00 = reserved */
    case 1 : flayer = 3; break; /* 01 = Layer III  */ 
    case 2 : flayer = 2; break; /* 10 = Layer II   */
    case 3 : flayer = 1; break; /* 11 = Layer I    */
    default: return false; //no correct header
    }
    //DEBUG_STREAM("Layer : " << (int) flayer << endl);

    /* Bit 16 : Protection bit */
    fprotbit = b1 & 0x01;
    //DEBUG_STREAM("Protection Bit : " << (int) fprotbit << endl);
    switch(fprotbit) {
    case 0: 
    case 1: 
      break;
    default: 
      return false; //no correct header
    }

    /* Bit 15,14,13,12 : Bitrate */
    static int bitrate_table[2][3][16] =  /* values in kBit */
    { 
      { 
        {0, 32, 64, 96, 128, 160, 192, 224, 256, 288, 320, 352, 384, 416, 448, -1}, /* MPEG 1 - Layer 1 */
        {0, 32, 48, 56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320, 384, -1}, /* MPEG 1 - Layer 2 */
        {0, 32, 40, 48,  56,  64,  80,  96, 112, 128, 160, 192, 224, 256, 320, -1}  /* MPEG 1 - Layer 3 */
      }, 
      { 
        {0, 32, 48, 56, 64, 80, 96, 112, 128, 144, 160, 176, 192, 224, 256, -1}, /* MPEG 2 - Layer 1 */ 
        {0,  8, 16, 24,  32,  40, 48, 56, 64,  80,  96, 112, 128, 144, 160, -1}, /* MPEG 2 - Layer 2 */ 
        {0,  8, 16, 24,  32,  40, 48, 56, 64,  80,  96, 112, 128, 144, 160, -1}  /* MPEG 2 - Layer 3 */
      }  
    };
    b = (b2 & 0xF0) >> 4;
    if ((b >= 0) && (b <= 15)) {
      fbitrate = bitrate_table[(int)(fmpeg-1)][(int)(flayer-1)][(int)(b)];
      if (fbitrate==0||fbitrate==-1) return false;
    } else {
      return false;
    }


    /* Bit 11,10: Frequenz */
    b = (b2 & 0x0c) >> 2;
    switch (fmpeg) {
    case 1: /* MPEG-1 */
      switch (b) {
    case 0: fsample_rate = 44100; break;
    case 1: fsample_rate = 48000; break;
    case 2: fsample_rate = 32000; break;
    case 3: fsample_rate =     0; break;
      }
      break; 
    case 2: /* MPEG-2 */
      switch (b) { 
    case 0: fsample_rate = 22050; break;
    case 1: fsample_rate = 24000; break;
    case 2: fsample_rate = 16000; break;
    case 3: fsample_rate =     0; break;
      }
      break;
    case 25: /* MPEG-2.5 */
      switch (b) { 
    case 0: fsample_rate = 11025; break;
    case 1: fsample_rate = 12000; break;
    case 2: fsample_rate =  8000; break;
    case 3: fsample_rate =     0; break;
      }
      break;
    default:
      return false;
    }
    if (0 >= fsample_rate) return false;

    //DEBUG_STREAM("fsample_rate : " << (int) fsample_rate << endl);

    /* Bit 9 : Padding Bit */
    fpadding = (b2 & 0x02) >> 1;

    //DEBUG_STREAM("fpadding : " << (int) fpadding << endl);

    switch (fpadding) {
    case 0:
    case 1:
      break;
    default:

      return false;
    }

    /* Bit 7,6 : Number of channels */
    switch((b3 & 0xc0) >> 6) {
    case 3:
      fnumchannels = 1;
      break;
    case 0:
    case 1:
    case 2:
      fnumchannels = 2;
      break;
    default:
      WARNING_STREAM("fnumchannel not found " << endl);
      return false;
      break;
    }
    //DEBUG_STREAM("fnumchannel : " << (int) fnumchannels << endl);


    /* Bit 3: Copyright
    Not used - but checked to verify complete header */
    int bitcopy=(b3&0x08) >> 3;
    switch (bitcopy) {
     case 0:
     case 1:
       break;
     default:
       WARNING_STREAM("bit Copyright not found : " << bitcopy << endl);
       return false;
    }
    //DEBUG_STREAM("bit Copyright : " << bitcopy << endl);

    /* Bit 2: Original media
    Not used - but checked to verify complete header */
    int original=(b3&0x04) >> 2;
    switch (original) {
     case 0:
     case 1:
       break;
     default:
       WARNING_STREAM("Original media bit not found : " << original << endl);
       return false;
    }
    //DEBUG_STREAM("Original media bit : " <<original << endl); 

    /* Bit 0/1: Emphasis
    Not used - but checked to verify complete header */
    int enphasys=b3&0x03;
    switch (enphasys) {
     case 0:
     case 1:
     case 2:
     case 3:
       break;
     default:
       WARNING_STREAM("bit enphasis not found" << endl);
       return false;
    }
    //DEBUG_STREAM("bit enphasis : "<< enphasys << endl);
    //DEBUG_STREAM("MP3 Header -> OK" << endl);

    return true;
  }
  
  int MP3ReadNode::getNextFrames(char* read_buffer,size_t max_bytes){
	DEBUG_STREAM("Trying to read " << max_bytes << endl);

    if((streampos)(mp3_data_end) == in.tellg()){ //end of mp3 data reached
      return 0;
    }

    int padding_size = 0;
    int first_valid_header = 0;  
    int valid_frames = 0;
    int valid_bytes = 0;
	int frame_size = 0;
    char test[4];

	in.seekg(mp3_header[frames_pos], ios::beg);
	//DEBUG_STREAM("Starting at frame " << frames_pos << " at byte position " << mp3_header[frames_pos] << endl);
	first_valid_header = in.tellg();
	
	while(static_cast<size_t>(valid_bytes) < max_bytes){

		in.read(test, 4 ); 
		if(parseHeader(test)){ 
			valid_frames++;

			if(flayer==1){
				padding_size = 4 * fpadding; 
			}else{
				padding_size = 1 * fpadding; 
			};

			frame_size = ( (getSamplesPerFrame(fmpeg,flayer) * fbitrate * 1000 / 8) / fsample_rate) + padding_size;
 
			if(frame_size>0){
				in.seekg((frame_size-4),ios::cur);
			}else{
				in.seekg(1,ios::cur); // if frame_size is invalid goto next byte
			}
			valid_bytes = in.tellg(); 
			valid_bytes -= first_valid_header;
			
		}else{
		  WARNING_STREAM("No frame header where should be one !!!! (at byte position: " << (-4 + in.tellg()) << ")"<< endl);
		  in.seekg(-3,ios::cur);
		};
          if(static_cast<size_t>(valid_bytes) > max_bytes || ((streampos)(mp3_data_end) <= in.tellg())){
			if(static_cast<size_t>(valid_bytes) > max_bytes){
        if(valid_frames==1 && (size_t)(valid_bytes)>max_bytes){
          ERROR_STREAM("Buffersize too small! Current frame is bigger than buffersize..." << endl);
        }
				valid_bytes -= frame_size; //last frame header + data
				valid_frames --;
				in.seekg(-frame_size,ios::cur);
			}
			break;
    }
	}
    in.seekg(-valid_bytes, ios::cur);
    in.read(read_buffer,valid_bytes);
    filepos = in.tellg();

    return valid_frames;
  }; 

  int MP3ReadNode::analyseFirstValidMP3Frame(){
	int bytes_read = 0;
	int first_valid_header = 0;
	char test[4];

	// Find first header an check for valid mp3 file
	in.seekg(mp3_data_start,ios::beg); 
	in.read(test, 4 );
    bytes_read = in.gcount(); 
    while ((bytes_read == 4) /* &&(!(parseHeader(test))*/) {
      if(parseHeader(test) && flayer == 3){
        filepos = in.tellg();
        break;
      }else{
        WARNING_STREAM("No frame header where expected!!!! (at byte position: " << (-4 + in.tellg()) << ")"<< endl);
      }
	  in.seekg(-3,ios::cur); //go 3 bytes back to search byte-wise
      in.read(test, 4 );
      bytes_read = in.gcount();
    }
	
	first_valid_header = in.tellg(); //position behind header
	first_valid_header -= 4; //position in front of header
	in.seekg(-4,ios::cur);
	
	return first_valid_header;
  }

  int MP3ReadNode::initializeMP3FrameMap(){
    have_mp3_frame_map = true;
    int frame_counter = 0;
    mp3_header.clear();
    in.seekg(static_cast<long>(mp3_data_start), ios::beg);
    filepos = mp3_data_start;
    
    int padding_size = 0;
    int first_valid_header = 0;    
    char test[4];

    // Find first header an check for valid mp3 file
    first_valid_header = analyseFirstValidMP3Frame();
	mp3_header[frame_counter] = first_valid_header;
   
	while(true){
		in.read(test, 4);
		int bytes_read = in.gcount();
		if(parseHeader(test)){
			frame_counter++;
			filepos = in.tellg();
			mp3_header[frame_counter] = filepos - 4;

			if(flayer==1){
				padding_size = 4 * fpadding; 
			}else{
				padding_size = 1 * fpadding; 
			};

			int frame_size = ( (getSamplesPerFrame(fmpeg,flayer) * fbitrate * 1000 / 8) / fsample_rate) + padding_size;
			DEBUG_STREAM("Calculated frame size = " << frame_size << "(bytes position:" << in.tellg() << ")" << endl);
			if(frame_size>0){
				in.seekg((frame_size-4),ios::cur); //goto expected start of next frame
			}else{
				in.seekg(1,ios::cur); //invalid frame_size 
			}


		}else{ 
			WARNING_STREAM("No frame header where one was expected!!!! (at byte position: " << (-4 + in.tellg()) << ")"<< endl); 
			in.seekg(-3,ios::cur);
		};
		filepos = in.tellg();
                if((streampos)(mp3_data_end) <= in.tellg() || bytes_read < 4){
			break;
		}
	}
    in.seekg(static_cast<long>(mp3_data_start), ios::beg);
    filepos = mp3_data_start;
    mp3_header[frame_counter+1] = mp3_data_end;

    return ++frame_counter;
  }

  int MP3ReadNode::getSamplesPerFrame(int mpeg, int layer){
    static int samples_per_frame[2][3] =  /* values in samples per frame */
    { 
      {384, 1152, 1152}, /* MPEG 1 - Layer 1 - 3 */
      {384, 1152, 576} /* MPEG 2 - Layer 1 -3 */   
    };

    return samples_per_frame[mpeg-1][layer-1];
  }

  /******************** ISeekable interface ********************/
  nmm_bool MP3ReadNode::hasPercentSeek() const {
    return true;
  }

  void MP3ReadNode::seekPercent(const Rational& percent) {
    Rational seek_percent(percent);
    Rational current_percent(filepos,filesize);
    Rational absolut_percent(current_percent + seek_percent);   
    if (absolut_percent > Rational(1)) {
      ERROR_STREAM("Seeking for " << (percent.getNumerator()*100/percent.getDenominator()) << " would seek over end of file" << endl);
      absolut_percent = Rational(1);
    }
    else if (absolut_percent < Rational(0)) {
      ERROR_STREAM("Seeking for " << (percent.getNumerator()*100/percent.getDenominator()) << " would seek over start of file" << endl);
      absolut_percent = Rational(0);
    }
    seekPercentTo(absolut_percent);  
  }

  void MP3ReadNode::seekPercentTo(const Rational& percent) {
    if (filename.substr(0,4) == string("/dev")) {
      // no seeking on devices!
      return;
    }

    Rational internal_percent(percent);
    if (internal_percent > Rational(1)) {
      ERROR_STREAM("Seeking more than 100% not possible, defaulting to end!" << endl);
      internal_percent = Rational(1);
    }
    else if (internal_percent < Rational(0)) {
      ERROR_STREAM("Seeking less than -100% not possible, defaulting to beginning!" << endl);
      internal_percent = Rational(0);
    }

    if(have_mp3_frame_map){
      Rational rat_seekto_frame = frames_total * internal_percent;
      seekto_frame = rat_seekto_frame.getNumerator() / rat_seekto_frame.getDenominator();
    }else{
      Rational rat_seekto = Rational(filesize) * internal_percent;
      seekto = rat_seekto.getNumerator() / rat_seekto.getDenominator();
    }
  }

  nmm_bool MP3ReadNode::hasBytesSeek() const {
    return true;
  }

  void MP3ReadNode::seekBytes(const nmm_int64& offset) {
    if ((offset < 0) && (abs(static_cast<long>(offset)) > static_cast<int64_t>(filepos))) {
      ERROR_STREAM("Seeking to negativ byte-offset not possible, defaulting to beginning of file!" << endl);
      seekBytesTo(0);
    } else {
      seekBytesTo(filepos + offset);
    }
  }

  void MP3ReadNode::seekBytesTo(const nmm_uint64& bytes) {
    u_int64_t internal_bytes(bytes);

    if (internal_bytes > filesize) {
      ERROR_STREAM("Seeking over EOF not possible, defaulting to EOF!" << endl);
      internal_bytes = filesize;
    }

    seekto = internal_bytes - filepos;

    DEBUG_STREAM(seekto << " = " << internal_bytes << " - " << filepos << endl);
  }

  nmm_bool MP3ReadNode::hasTimeSeek() const {
    return true;
  }

  void MP3ReadNode::seekTime(const Interval& seek_time) {
    Rational percent((int64_t)seek_time.sec * 1000000000 + seek_time.nsec, (int64_t)track_duration.sec * 1000000000 + track_duration.nsec);
    seekPercent(percent);
  }

  void MP3ReadNode::seekTimeTo(const Time& seek_time_to) {
    Rational percent_to((int64_t)seek_time_to.sec * 1000000000 + seek_time_to.nsec, (int64_t)track_duration.sec * 1000000000 + track_duration.nsec);
    seekPercentTo(percent_to);
  }

} /* namespace */

#include "nmm/nmm_suffix.hpp"
