/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "MPEGReadNode.hpp"

#include "nmm/plugins/NMMRegisterPlugin.hpp"
#include "nmm/utils/SecurityCheck.hpp"
#include "nmm/multimedia/interfaces/IProgressListener.hpp"
#include "nmm/multimedia/interfaces/ITrack.hpp"
#include "nmm/multimedia/ISyncReset.hpp"
#include "nmm/multimedia/FlushEvent.hpp"




NMM_REGISTER_PLUGIN(MPEGReadNode,NMM, "MPEGReadNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  MPEGReadNode::MPEGReadNode(const char * name /* = "MPEGReadNode" */,
                             StreamQueue::Mode mode /* = StreamQueue::MODE_SUSPEND */,
                             int queue_size /* = StreamQueue::DEFAULT_SIZE */)
    : GenericSourceNode(name, mode, queue_size),
      stream_timer() {
    buffer_size = 32*1024;
    filename = "";

    filesize = 0;
    filepos  = 0;
    seekto = 0;    

    send_start    = false;
    send_format   = false;
    send_progress = false;
    buffers_before_progress = 5;
    buffers_progress        = 0;
  }
  
  void MPEGReadNode::setFilename(const string& name) {
    filename = name;
  }
  
  void MPEGReadNode::sendProgressInformation(const nmm_bool& in) {
    send_progress = in;
  }

  void MPEGReadNode::setProgressInterval(const nmm_int32& buffers) {
    buffers_before_progress = buffers;
  }

  void MPEGReadNode::setBufferSize(const nmm_int32& size) {
    DEBUG_STREAM("Set buffer_size to " << size << endl);
    buffer_size = size;
  }
  
  nmm_int32 MPEGReadNode::getBufferSize() {
    return buffer_size;
  }
  
  Result MPEGReadNode::doInit() {
    Format* format;
    
    // mpeg program stream
    format = getOwnOutputProperty()->addNewFormat(Format::av_mpegps_type);
    format->addIntParamValue(Format::version_param, 1);
    format->addIntParamValue(Format::version_param, 2);
    getOwnOutputProperty()->setDefaultFormat(format);
    
    // mpeg elementary stream
    format = getOwnOutputProperty()->addNewFormat(Format::av_mpeg_type);
    format->addIntParamValue(Format::version_param, 1);
    format->addIntParamValue(Format::version_param, 2);
    
    // mpeg elementary stream with knowledge about format
    format = getOwnOutputProperty()->addNewFormat(Format::av_mpeg_type);
    format->addIntParamValue(Format::version_param, 1);
    format->addIntParamValue(Format::version_param, 2);
    format->addWildcard(Format::x_resolution_param);
    format->addWildcard(Format::y_resolution_param);
    format->addWildcard(Format::ratio_x_param);
    format->addWildcard(Format::ratio_y_param);
    format->addWildcard(Format::framerate_param);
    
    return SUCCESS;
  }
  
  Result MPEGReadNode::doInitOutput() {
    send_start = false;
    stream.clear();
    
    if(!SSecurityCheck::getInstance().checkFilenameForReading(filename)) {
      ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return FAILURE;
    }
    cerr << "her " << endl;
    stream.open(filename.c_str(), ios::binary);
    
    if (stream.fail()) {
      ERROR_STREAM("Could not open " << filename << endl);
      return FAILURE;
    }
    cerr << "her " << endl;

    if (filename.substr(0,4) == string("/dev")) {
      WARNING_STREAM("Seeking on devices not possible!" << endl);
      filesize = 0;
    } else {
      stream.seekg( 0, ios::end );
      filesize = stream.tellg();
      stream.seekg( 0, ios::beg );
    }
    cerr << "her " << endl;

    if (!CheckFormat(&file_format)) {
      ERROR_STREAM("Failure while checking format" << endl);
      return FAILURE;
    }
    cerr << "her " << endl;

    if (file_format.mpeg_audio) {
      DEBUG_STREAM("Found MPEG Audio Stream" << endl);
    }
    if (file_format.ac3_audio) {
      DEBUG_STREAM("Found AC3 Audio Stream" << endl);
    }
    if (file_format.mpeg2_video) {
      DEBUG_STREAM("Found MPEG2 Video Stream" << endl);
    }
    if (file_format.mpeg1_video) {
      DEBUG_STREAM("Found MPEG1 Video Stream" << endl);
    }
    cerr << "her 2" << endl;
    
    if (file_format.pes_stream) {
      // mpeg program stream
      DEBUG_STREAM("Found MPEG program stream" << endl);
      Format *format = getOwnWorkingOutputProperty()->addNewFormat("av/mpegps");
      format->addIntParamValue(Format::version_param, /*file_format.mpeg1_video ? 1 :*/ 2);
      getOwnWorkingOutputProperty()->setDefaultFormat(format);
cout << *format;
    } else {
      // mpeg elementary stream
      DEBUG_STREAM("Found MPEG elementary stream" << endl);
      Format* format = getOwnWorkingOutputProperty()->addNewFormat("video/mpeg");
      format->addIntParamValue(Format::version_param, file_format.mpeg1_video ? 1 : 2);
      getOwnWorkingOutputProperty()->setDefaultFormat(format);
cout << *format;
    }


    setProducingFlag(true);
    send_start = true;
    if(isOutputConnected("default")) {
      send_format = true;
    } 

    return SUCCESS;
  }
  
  
  int MPEGReadNode::MPEG_Read_Buffer( unsigned char *buf, int size ) {
    stream.read( reinterpret_cast<char*>(buf), size );

    return( stream.gcount() );
  }

  void MPEGReadNode::MPEG_Seek_Begin() {
    stream.close();

    stream.open( filename.c_str(), ios::binary );

    //stream.seekg( 0, ios::beg );
  }

  Message *MPEGReadNode::processBuffer(Buffer *) {
    Buffer *out;
    int bytes_read;

   if (send_progress && filesize) {
      Event e;
      if ( buffers_progress > buffers_before_progress ) {
        buffers_progress = 0;
        CEvent* cevent(new CEvent(IProgressListener::create_setProgress(filepos, filesize)));
        cevent->setMandatoryBit(false);
        return cevent;
      }
      buffers_progress++;
    }

    if (send_start) {
      DEBUG_STREAM("Sending start_track & sync_reset..." << endl);
      CEvent* ce = new CEvent();

      ce->insertEvent(Event(ITrack::startTrack_event, new TInValue<string>(filename)));
      ce->insertEvent(ISyncReset::create_syncReset());
      send_start = false;

      if(send_format) {
        send_format = false;

        //set new outputformat
        setOutputFormat(*getOwnWorkingOutputProperty()->getDefaultFormat());

        DEBUG_STREAM("Sending setFormat" << endl);
        ce->insertEvent(IFormatChange::create_setFormat(*getOutputFormat()));
      }

      return ce;
    }
    
    if ((out = getNewBuffer( buffer_size )) != 0) {
      if (seekto!=0) {
	if (seekto<0 && static_cast<unsigned int>(labs(seekto)) > filepos) {
	  stream.seekg( 0, ios::beg );
	  filepos = 0;
	  DEBUG_STREAM("Seeked to begining of file"<<endl);
	}
	else if ((filepos + seekto) >= filesize) {
	  DEBUG_STREAM("Stopping MPEGRead" << endl);
	  setProducingFlag(false); // nothing left to do, so stop producing
	  DEBUG_STREAM("Sending end_track..." << endl);
	  Event e (ITrack::endTrack_event);
	  return new CEvent(e);
	}
	else {
	  stream.seekg(filepos+seekto);
	  filepos = filepos+seekto;
	  DEBUG_STREAM("normal seek"<<endl);
	}
	seekto=0;
	buffers_progress = buffers_before_progress +1;
	CEvent* cevent = new FlushEvent();
        cevent->insertEvent(cevent->size(), ISyncReset::create_syncReset());
        // is this event really needed?
        // Isn't it possible to react on flush() instead of sending this special event?
        return cevent;
      }
      stream.read( out->getData(), buffer_size );
      bytes_read = stream.gcount();
      //cout << bytes_read << " Bytes read." << endl;
      if ( !stream.eof() ) {
        out->setUsed( bytes_read );
	filepos+=bytes_read;
        stream_timer.setTimestamp(out);
	return(out);

      } else { // assume EOF
        DEBUG_STREAM("Stopping MPEGRead" << endl);
	setProducingFlag(false); // nothing left to do, so stop producing
	DEBUG_STREAM("Sending end_track..." << endl);
	Event e (ITrack::endTrack_event);
	return new CEvent(e);
      }
    }
    return 0;
  }

  MPEGReadNode::~MPEGReadNode() {
    stream.close();
  }

  Result MPEGReadNode::doDeinitOutput() {
    stream.close();
    return SUCCESS;
  }

  /******************** ISeekable interface ********************/
  nmm_bool MPEGReadNode::hasPercentSeek() const {
    if (filename.substr(0,4) == string("/dev")) {
      // no seeking on devices!
      return false;
    }
    return true;
  }

  void MPEGReadNode::seekPercent(const Rational& percent) {
    if (filename.substr(0,4) == string("/dev")) {
      // no seeking on devices!
      return;
    }

    Rational internal_percent(percent);
    if (internal_percent > Rational(1)) {
      ERROR_STREAM("Seeking more than 100% not possible, defaulting to end!" << endl);
      internal_percent = Rational(1);
    }
    else if (internal_percent < Rational(-1)) {
      ERROR_STREAM("Seeking less than -100% not possible, defaulting to beginning!" << endl);
      internal_percent = Rational(-1);
    }

    Rational rat_seekto = Rational(filesize) * internal_percent;
    seekto = rat_seekto.getNumerator() / rat_seekto.getDenominator();
    DEBUG_STREAM(seekto << " = " << internal_percent << " * " << filesize << endl);
  }
   
  void MPEGReadNode::seekPercentTo(const Rational& percent) {
    seekPercent(percent - Rational(filepos, filesize));
  }

  nmm_bool MPEGReadNode::hasBytesSeek() const {
    if (filename.substr(0,4) == string("/dev")) {
      // no seeking on devices!
      return false;
    }
    return true;
  }

  void MPEGReadNode::seekBytes(const nmm_int64& offset) {
    if ((offset < 0) && (labs(offset) > static_cast<int64_t>(filepos))) {
      ERROR_STREAM("Seeking to negativ byte-offset not possible, defaulting to beginning of file!" << endl);
      seekBytesTo(0);
    }
    seekBytesTo(filepos + offset);
  }

  void MPEGReadNode::seekBytesTo(const nmm_uint64& bytes) {
    if (filename.substr(0,4) == string("/dev")) {
      // no seeking on devices!
      return;
    }

    u_int64_t internal_bytes(bytes);
    if (internal_bytes > filesize) {
      ERROR_STREAM("Seeking over EOF not possible, defaulting to EOF!" << endl);
      internal_bytes = filesize;
    }

    seekto = internal_bytes - filepos;

    DEBUG_STREAM(seekto << " = " << internal_bytes << " - " << filepos << endl);
  }

  nmm_bool MPEGReadNode::hasTimeSeek() const {
    return false;
  }

  void MPEGReadNode::seekTime(const Interval&) {
  }

  void MPEGReadNode::seekTimeTo(const Time&) {
  }
  /******************** ISeekable interface ********************/
}

#include "nmm/nmm_suffix.hpp"
