/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "PNGReadNode.hpp"

#include "nmm/utils/thread/MutexGuard.hpp"
//#include "nmm/interfaces/base/IFormatChange.hpp"
#include "nmm/multimedia/interfaces/ITrack.hpp"
//#include "nmm/base/format/Property.hpp"
#include "nmm/utils/SecurityCheck.hpp"
//#include "nmm/base/registry/Plugin.hpp"

#include <png.h>

#include "nmm/plugins/NMMRegisterPlugin.hpp"

#ifndef NULL
#define NULL 0
#endif

NMM_REGISTER_PLUGIN(PNGReadNode, NMM, "PNGReadNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  PNGReadNode::PNGReadNode(const char *name /* = "PNGReadNode" */,
			   StreamQueue::Mode mode /* = StreamQueue::MODE_SUSPEND */,
			   int queue_size /* = StreamQueue::DEFAULT_SIZE */)
    : GenericSourceNode(name, mode, queue_size),
      deffileext(".png"),
      format_changed(false) {
    
    end_track = false;
    image_width = 0;
    image_height = 0;
    frame_count  = 0;
    data         = NULL;
    buff_size = 0;
    fileext = deffileext;
    filename_mode = IFileModeHandler::FILENAME_ONLY;
    filenames.clear();
    cur_filename = string();
    last_image_done = true;
  }

  PNGReadNode::~PNGReadNode() {
    if ( data ) {
      delete[] data;
    }
  }

  void PNGReadNode::addFilename(const string& url)
  {
    MutexGuard filename_mg(filename_mutex);    
    filenames.push_back(url);
    DEBUG_STREAM("Added filename: " << url << endl);
    setProducingFlag(true);
  }

  void PNGReadNode::setFormat(const IFileModeHandler::FileFormat&)
  {
    ERROR_STREAM("setFormat not implemented here!");
  }
  
  IFileModeHandler::FileFormat PNGReadNode::getFormat() {
    ERROR_STREAM("getFormat not implemented here!");    
    ERROR_STREAM("Returning default value!");    
    return IFileModeHandler::DATA_ONLY;
  }

  void PNGReadNode::setMode(const IFileModeHandler::FileMode&)
  {
    ERROR_STREAM("setMode not implemented here!");
  }
  
  IFileModeHandler::FileMode PNGReadNode::getMode() {
    ERROR_STREAM("getMode not implemented here!");
    ERROR_STREAM("Returning default value!");    
    return IFileModeHandler::FILE_PER_FRAME;
  }

  void PNGReadNode::setFilename(const string& file) {
    addFilename(file);
  }

  void PNGReadNode::setFilenameMode(const IFileModeHandler::FilenameMode& mode)
  {
    filename_mode = mode;
  }

  IFileModeHandler::FilenameMode PNGReadNode::getFilenameMode() {
    return filename_mode;
  }

 void PNGReadNode::setLoop(const nmm_bool&) {
    ERROR_STREAM("setLoop not implemented here!");
  }

  Result PNGReadNode::createImage(const nmm_uint32& color, const nmm_uint32& w, const nmm_uint32& h) 
  {
    if ( data ) {
      delete[] data;
      data = NULL;
    }

    image_width  = w;
    image_height = h;

    /* load the new image */
    data = new unsigned char[image_width*image_height*4];
    if ( !data ) return(FAILURE);

    unsigned int* pdata = (unsigned int*) data;
    for( unsigned int i=0; i<(image_width*image_height); i++ ) {
      *pdata = color;
      pdata++;
    }
    
    setProducingFlag(true);

    return(SUCCESS);
  }

  unsigned char* PNGReadNode::loadBitmap( string fname, unsigned int* width, unsigned int* height ) {
    png_bytep*  png_rows;               // stores the PNG image
    png_uint_32 png_width, png_height;  // PNG image width and height

    png_structp png_ptr;
    png_infop info_ptr;
    int bit_depth, color_type, interlace_type;
    FILE *fp;

    png_rows = NULL;
    
    if(!SSecurityCheck::getInstance().checkFilenameForReading(fname)) {
      ERROR_STREAM("Security check not passed for file " << fname << ", check .nmmrc for options" << endl);
      return NULL;
    }
    
    if ((fp = fopen(fname.c_str(), "rb")) == NULL) {
      return NULL;
    }

    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);

    if (png_ptr == NULL) {
      fclose(fp);
      return NULL;
    }

    info_ptr = png_create_info_struct(png_ptr);
    if (info_ptr == NULL) {
      fclose(fp);
      png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
      return NULL;
    }

    if (setjmp(png_ptr->jmpbuf)) {
      png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
      fclose(fp);
      return NULL;
    }

    png_init_io(png_ptr, fp);

    png_read_info(png_ptr, info_ptr);

    png_get_IHDR(png_ptr, info_ptr, &png_width, &png_height, &bit_depth, &color_type, &interlace_type, NULL, NULL);


    if (bit_depth == 8 && (color_type == PNG_COLOR_TYPE_RGB || color_type == PNG_COLOR_TYPE_GRAY) ) {
      png_set_filler(png_ptr, 0xff, PNG_FILLER_AFTER);
      //png_set_bgr(png_ptr);
    }

    if ( color_type == PNG_COLOR_TYPE_GRAY ||
	 color_type == PNG_COLOR_TYPE_GRAY_ALPHA) {
      png_set_gray_to_rgb(png_ptr);
    };

    png_set_bgr(png_ptr);
    png_read_update_info(png_ptr, info_ptr);

    png_rows = new png_bytep[png_height];

    for (unsigned int row = 0; row < png_height; row++) {
      png_rows[row] = new png_byte[png_get_rowbytes(png_ptr, info_ptr)];
    }

    png_read_image(png_ptr, png_rows);

    png_read_end(png_ptr, info_ptr);

    png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);

    fclose(fp);

    *width  = (unsigned int) png_width;
    *height = (unsigned int) png_height;
    /* all bitamps we load are in rgba format, so mutliply width*height with 4 */
    unsigned char* retData = new unsigned char[png_width*png_height*4];
    unsigned char* temp = retData;

    for (unsigned int row = 0; row < png_height; row++) {
      if ( png_rows[row] ) {
	memcpy( temp, png_rows[row], png_width*4 );
	temp+=png_width*4;
	delete[] png_rows[row];
      }
    }

    return retData;
  }

  Result PNGReadNode::doInit() {
    DEBUG_STREAM("Creating output format, setting width and height: "
		 << image_width << ", " << image_height << endl);
    Format *out_format = getOwnOutputProperty()->addNewFormat("video/raw");
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addStringParamValue(Format::channelorder_param, "rgb");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::colorspace_param, "rgb24");
    out_format->addIntParamValue(Format::bitperpixel_param, 32);
    out_format->addStringParamValue(Format::format_param, "packet");
    
    getOwnOutputProperty()->setDefaultFormat(out_format);
    
    return SUCCESS;
  }

  Result PNGReadNode::doInitOutput() {
    // if createImage was *not* already called -> check for files
    if(data == 0) {
      // there must be at least one file added till INITIALIZED
      // (see IMultipleSourceFileHandler), so load the first file
      // if this is the first time doInitOutput is called (assumed that
      // it is called from application, then.
      while (!filenames.empty()) {
	if (openImage()) {
	  last_image_done = false;
	  break;
	}
      }
      if (last_image_done) {
	ERROR_STREAM("no file in the list could be opened!" << endl);
	return FAILURE;
      }
    }
    // openImage() sets format_changed to true as the resolution is
    // different to the old one (which was 0 after initialisation in
    // C'tor). 
    // If output is not connected, do not send setFormat
    if(!isOutputConnected("default")) {
      format_changed = false;
    }    
    setWOPInternal();

    return SUCCESS;
  }
  
  Result PNGReadNode::doActivate() {
    const Format* o_format = getOwnWorkingOutputProperty()->getDefaultFormat();
    if (o_format == 0) {
      ERROR_STREAM( "Invalid output format (NULL-ptr)" << endl);
      return FAILURE;
    }

    string colorspace = o_format->getStringValue(Format::colorspace_param);
    int bitperpixel = o_format->getIntValue(Format::bitperpixel_param);
    image_width  = o_format->getIntValue(Format::x_resolution_param);
    image_height = o_format->getIntValue(Format::y_resolution_param);

    if ((image_width <= 0) || (image_height <= 0)) {
      ERROR_STREAM( "Invalid Image-width or -height" << endl);
      return FAILURE;     
    }
    
    if(colorspace != "rgb24" || bitperpixel != 32) {
      ERROR_STREAM( "Invalid Colorspace!" << endl);
      return FAILURE;     
    }

    return SUCCESS;
  }

  void PNGReadNode::setWOPInternal() {
    // update working output property
    getOwnWorkingOutputProperty()->erase();

    DEBUG_STREAM("Creating working output format, setting width and height: "
		 << image_width << ", " << image_height << endl);
    Format *out_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
    out_format->addIntParamValue(Format::x_resolution_param, image_width);
    out_format->addIntParamValue(Format::y_resolution_param, image_height);
    out_format->addIntParamValue(Format::ratio_x_param, image_width);
    out_format->addIntParamValue(Format::ratio_y_param, image_height);
    out_format->addStringParamValue(Format::channelorder_param, "rgb");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::colorspace_param, "rgb24");
    out_format->addIntParamValue(Format::bitperpixel_param, 32);
    out_format->addStringParamValue(Format::format_param, "packet");
    out_format->addFloatParamValue(Format::framerate_param, 1.0);
    getOwnWorkingOutputProperty()->setDefaultFormat(out_format);
  }

  bool PNGReadNode::openImage() {
    // expects that filenames has at least 1 entry.
    if ( data ) {
      delete[] data;
      data = NULL;
    }
    
    /* load the new image */
    cur_filename = filenames.front();
    filenames.pop_front();
    unsigned int width(0);
    unsigned int height(0);
    data = loadBitmap( cur_filename, &width, &height );

    // Could get a failure here!
    if ( !data ) {
      ERROR_STREAM("Could not load image!!! filename = " << cur_filename << endl);
      return false;
    }

    if ((width != image_width) || (height != image_height)) {
      format_changed = true;
      image_width = width;
      image_height = height;
    }

    buff_size = image_width*image_height*4;
    return true;
  }
  
  Message *PNGReadNode::processBuffer(Buffer *) {

    if(data == 0) 
    {
      MutexGuard filename_mg(filename_mutex);    
      if (last_image_done && !filenames.empty()) {
        while (!filenames.empty()) {
          if (openImage()) {
            last_image_done = false;
            break;
          }
        }
      }

      if (last_image_done && filenames.empty()) {
        setProducingFlag(false);
        return NULL;
      }
    }

    if (format_changed) {
      format_changed = false;
      // send & set new format

      setOutputFormat(*getOwnWorkingOutputProperty()->getDefaultFormat());
      DEBUG_STREAM("Send setFormat " << endl << *getOwnWorkingOutputProperty()->getDefaultFormat() << endl);
      return new CEvent(IFormatChange::create_setFormat(*getOutputFormat()));
    }
      
    if (end_track) {
      end_track = false;
      last_image_done = true;
      
      DEBUG_STREAM("Sending endTrack..." << endl);
      Event e(ITrack::endTrack_event);
      return new CEvent(e);      
    } else {
      
      // send new format
      Buffer *out;
      
      if ((out = getNewBuffer( buff_size )) != 0) {
        out->setUsed( buff_size ); 
        
        if ( data ) {
	  memcpy( out->getData(), data, buff_size );
	  delete[] data;
	  data = 0;
	}
        DEBUG_STREAM("produceBuffer: IMAGE PAINTED" << endl);
	
	if (filename_mode == IFileModeHandler::TIMESTAMP_AS_NAME) {
	  string filename_only = cur_filename;
	  // cut away path to file
	  string::size_type lastbs = cur_filename.find_last_of('/');
	  if (lastbs != string::npos) {
	    filename_only = cur_filename.substr(lastbs + 1, cur_filename.length());
	  }
	  
	  // cut away file extension
	  string::size_type lpos = filename_only.rfind(fileext);
	  if (lpos != string::npos) {
	    filename_only = filename_only.substr(0, lpos);
	  }
	  
	  // cur_filename should now be in format xxx.xxx
	  // search for "."; digits before are seconds,
	  // digits after are interpreted as nanoseconds
	  // such that "2.5" means 2 seconds, 500 000 000 nanoseconds
	  string sec_str = "", nsec_str = "";
	  string::size_type ppos = filename_only.find_last_of('.');
	  if (ppos != string::npos) {
	    sec_str = filename_only.substr(0, ppos);
	    nsec_str = filename_only.substr(ppos + 1, filename_only.length());
	  }
	  
	  int secs = 0, nsecs = 0;
	  if (sec_str != "") secs = atoi(sec_str.c_str());
	  if (nsec_str != "") {
	    nsecs = atoi(nsec_str.c_str());
	    if (nsecs != 0) {
	      for (unsigned int i = 0; i < (9 - nsec_str.length()); i++) nsecs *= 10;
	    }
	  }
	  
	  DEBUG_STREAM("Extracted from filename " << filename_only << ": secs = "
		       << secs << ", nsecs = " << nsecs << endl);
	  Timestamp t;
	  t.sync_time = Time(secs, nsecs);
	  t.is_valid = true;
	  t.stream_counter = frame_count;
	  out->setTimestamp(t);
	  DEBUG_STREAM("Set buffer timestamp to: " << out->getTimestamp() << endl);
	}
      }
      else {
        ERROR_STREAM("Could get no Buffer!!!");
        return 0;
      }
      
      frame_count++;
      DEBUG_STREAM("produceBuffer: frame_count = " << frame_count << endl);
      
      // this node will create further frames only on demand
      end_track = true;
      
      return out;
    }
    return NULL; //should never be reached, as of if/elseif/else above
  }
  
}

#include "nmm/nmm_suffix.hpp"
