/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "PNGWriteNode.hpp"
#include "nmm/plugins/file/interfaces/ISendFilenameAndTimestamp.hpp"
#include "nmm/comm/format/Property.hpp"
#include "nmm/utils/SecurityCheck.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"

#ifndef NULL
#define NULL 0
#endif

NMM_REGISTER_PLUGIN(PNGWriteNode, NMM, "PNGWriteNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  PNGWriteNode::PNGWriteNode(const char *name /* = "PNGWriteNode" */,
			   StreamQueue::Mode mode /* = StreamQueue::MODE_SUSPEND */,
			   int queue_size /* = StreamQueue::DEFAULT_SIZE */)
    : GenericSinkNode(name, mode, queue_size), deffileext(".png") {

    image_width  = 0;
    image_height = 0;
    buf_count = 0;
    filename = string("");
    workdir = string("");
    fileext = deffileext;
    filename_mode = IFileModeHandler::FILENAME_ONLY;
    last_buf_timestamp = Timestamp();
    last_filename = string("");
    discardImages = false;
  }

  PNGWriteNode::~PNGWriteNode() {
  }

  void PNGWriteNode::setDiscardImages(const nmm_bool& disImg)
  {
    discardImages = disImg;
  }

  void PNGWriteNode::setFilename(const string& name)
  {
    filename = name;
  }

  void PNGWriteNode::setWorkingDirectory(const string& workdir_) {
    workdir = workdir_;
    // if no "/" at the end, add one.
    unsigned int lastPos = workdir.find_last_of('/');
    if (lastPos != workdir.length() - 1) {
      workdir += "/";
    }
  }
  
  string PNGWriteNode::getWorkingDirectory() {
    return workdir;
  }
  
  void PNGWriteNode::setFileExtension(const string& fileext_) {
    fileext = fileext_;
  }
  
  string PNGWriteNode::getFileExtension() {
    return fileext;
  }
  
  void PNGWriteNode::setFormat(const IFileModeHandler::FileFormat&)
  {
    ERROR_STREAM("setFormat not implemented here!");
  }
  
  IFileModeHandler::FileFormat PNGWriteNode::getFormat() {
    ERROR_STREAM("getFormat not implemented here!");    
    ERROR_STREAM("Returning default value!");    
    return IFileModeHandler::DATA_ONLY;
  }

  void PNGWriteNode::setMode(const IFileModeHandler::FileMode&)
  {
    ERROR_STREAM("setMode not implemented here!");
  }
  
  IFileModeHandler::FileMode PNGWriteNode::getMode() {
    ERROR_STREAM("getMode not implemented here!");
    ERROR_STREAM("Returning default value!");    
    return IFileModeHandler::FILE_PER_FRAME;
  }

  void PNGWriteNode::setFilenameMode(const IFileModeHandler::FilenameMode& mode)
  {
    filename_mode = mode;
  }

  IFileModeHandler::FilenameMode PNGWriteNode::getFilenameMode() {
    return filename_mode;
  }

 void PNGWriteNode::setLoop(const nmm_bool&) {
    ERROR_STREAM("setLoop not implemented here!");
  }

  Result PNGWriteNode::doInit()
  {
    Format *in_format = getOwnInputProperty()->addNewFormat("video/raw");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addStringParamValue(Format::channelorder_param,"rgb");
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::colorspace_param, "rgb24");
    in_format->addIntParamValue(Format::bitperpixel_param,32);
    in_format->addStringParamValue(Format::format_param, "packet");
    in_format->addWildcard(Format::framerate_param);

    getOwnInputProperty()->setDefaultFormat(in_format);

    in_format = getOwnInputProperty()->addNewFormat("video/raw");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addStringParamValue(Format::channelorder_param,"rgb");
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::colorspace_param, "rgb24");
    in_format->addIntParamValue(Format::bitperpixel_param,24);
    in_format->addStringParamValue(Format::format_param, "packet");
    in_format->addWildcard(Format::framerate_param);

    in_format = getOwnInputProperty()->addNewFormat("video/raw");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addStringParamValue(Format::channelorder_param,"yvu");
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::colorspace_param, "yv12");
    in_format->addIntParamValue(Format::bitperpixel_param, 12);
    in_format->addStringParamValue(Format::format_param, "planar");
    in_format->addWildcard(Format::framerate_param);

    return SUCCESS;
  }

  Result PNGWriteNode::doActivate() 
  {
    const Format* i_format = getInputFormat();
    if (i_format == 0) {
      ERROR_STREAM( "Invalid input format (NULL-ptr)" << endl);
      return FAILURE;
    }

    string colorspace = i_format->getStringValue(Format::colorspace_param);
    image_width  = i_format->getIntValue(Format::x_resolution_param);
    image_height = i_format->getIntValue(Format::y_resolution_param);
    int bpp = i_format->getIntValue(Format::bitperpixel_param);

    DEBUG_STREAM("Got width and height: (" << image_width << ", "
		 << image_height << ")" << endl);
    if ((image_width <= 0) || (image_height <= 0)) {
      ERROR_STREAM( "Invalid Image-width or -height" << endl);
      return FAILURE;     
    }
    png_utils.setDimension(image_width, image_height);

    if ("rgb24" == colorspace) {
      if (bpp == 24) {
	png_utils.setFormat(PNGUtils::RGB24);
      }
      else if (bpp ==32) {
	// our colorspace rgb24 with 32 bpp means that
	// only 24 are used, 8 bit are padding bits!!!
	png_utils.setFormat(PNGUtils::RGBX32);
      }
      else {
	ERROR_STREAM("Invalid bit_per_pixel parameter!!!" << endl);
	return FAILURE;
      }
    }
    else if ("yv12" == colorspace) {
      png_utils.setFormat(PNGUtils::YV12);
    }
    else {
      ERROR_STREAM("Invalid color format!!!" << endl);
      return FAILURE;
    }
    
    return SUCCESS;
  }
  
  Message* PNGWriteNode::processBuffer(Buffer *in_buffer) {
    
    string tmpfname = workdir + filename;
    Buffer* out_buffer;

    if (getWorkingFlag()) {
      setWorkingFlag(false);
      return new CEvent(ISendFilenameAndTimestamp::create_sendFilenameAndTimestamp(last_filename, last_buf_timestamp));
    }

    if(discardImages){
      in_buffer->release();
      return 0;
    }

    if (filename_mode == IFileModeHandler::FILENAME_BUFCOUNT) tmpfname = tmpfname + buf_count;
    else if (filename_mode == IFileModeHandler::TIMESTAMP_AS_NAME) {
      tmpfname = workdir;
      tmpfname += in_buffer->getTimestamp().sync_time.getPrettyString();
      tmpfname += fileext;
      setWorkingFlag(true);
      last_buf_timestamp = in_buffer->getTimestamp();
      last_filename = tmpfname;
    }

    DEBUG_STREAM ("Writing Image with name: " << tmpfname << endl);
    DEBUG_STREAM ("Writing Image with dimensions: " << image_width << ", " << image_height << endl);
    
    out_buffer = in_buffer->getWriteableInstance();
    
    /* Security check */
    if (!SSecurityCheck::getInstance().checkFilenameForWriting(tmpfname)) {
      ERROR_STREAM("Security check not passed for file " << tmpfname << ", check .nmmrc for options" << endl);
    } else {
      png_utils.writeImage(tmpfname, reinterpret_cast<unsigned char*>(out_buffer->getData()));
    }
    
    out_buffer->release();
    
    buf_count++;
    return 0;
  }

}

#include "nmm/nmm_suffix.hpp"
