/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "FileBuffer.hpp"

#include "nmm/utils/time/Interval.hpp"

#include <iostream>

using namespace std;

#include "nmm/nmm_prefix.hpp"

FileBuffer::FileBuffer(size_t size, unsigned int keep)
  : block_size(size),
    block_num(0),
    start_time(0),
    end_time(0),
    file_block_size(block_size + sizeof(size_t)),
    keep_min(keep),
    skiperror(false),
    read_pos(used_blocks.end()),
    keep_data(false),
    o_file(NULL),
    filename("tsbuffer"),
    ctemp(NULL)
{
  ctemp = new char[sizeof(size_t)];
  //debug = true;

  // open files for I/O
#if (__GNUC__ >= 3)
  o_file = new fstream(filename.c_str(),ios_base::binary|ios_base::in|ios_base::out|ios_base::trunc);//,0666);
#else
  o_file = new fstream(filename.c_str(),ios::binary|ios::in|ios::out|ios::trunc,0666);
#endif
  //d_file = new fstream("/video/tsdebug",ios::binary|ios::out|ios::trunc,0666);

  if (!o_file) {
    cerr << "unable to create file '" << filename << "'!" << endl;
    exit(-1);
  }

  // this allows reading of whitespace-characters using '>>'
  o_file->unsetf(ios::skipws);
}

FileBuffer::FileBuffer(size_t size, string name, unsigned int keep)
  : block_size(size),
    block_num(0),
    start_time(0),
    end_time(0),
    file_block_size(block_size + sizeof(size_t)),
    keep_min(keep),
    skiperror(false),
    read_pos(used_blocks.end()),
    keep_data(false),
    o_file(NULL),
    filename(name),
    ctemp(NULL)
{
  file_block_size = block_size + sizeof(size_t);
  ctemp = new char[sizeof(size_t)];
  //debug = true;
  
  // open file for I/O
#if (__GNUC__ >= 3)
  o_file = new fstream(filename.c_str(),ios_base::binary|ios_base::in|ios_base::out|ios_base::trunc);//,0666);
#else
  o_file = new fstream(filename.c_str(),ios::binary|ios::in|ios::out|ios::trunc,0666);
#endif
  //d_file = new fstream("/video/tsdebug",ios::binary|ios::out|ios::trunc,0666);
  
  if (!o_file) {
    cerr << "unable to create file '" << name << "'!" << endl;
    exit(-1);
  }

  // this allows reading of whitespace-characters using '>>'
  o_file->unsetf(ios::skipws);
}

FileBuffer::~FileBuffer() {
  // close file...
  o_file->close();
  //d_file->close();

  // ..delete it...
  if (! keep_data)
    unlink(filename.c_str());

  // ...and free used space
  delete o_file;
  //delete d_file;
  delete[] ctemp;
}


// DEBUG OUTPUT
void FileBuffer::printUsed() {
  //NMM::MutexGuard mg(buffer_mutex);
  
  cerr << "TSNode used list: (skiperror = " << skiperror << ")  keep_min =" << keep_min << endl;
  for (list<Block>::const_iterator i=used_blocks.begin(); i != used_blocks.end(); i++) {
    cerr << "TSNode id: " << i->id << " read: " << i->already_read << " size: " << i->size;
    if (i == read_pos)
      cerr << " (*)";
    cerr << endl;
  }
  //  cout << "TSNode free list: " << endl;
  //  for (list<size_t>::const_iterator i=free_blocks.begin(); i != free_blocks.end(); i++)
  //    cout << "TSNode id: " << *i << endl;
}


size_t FileBuffer::popData(char* data, bool& is_event, NMM::Time& timestamp,int skip_blocks) {
  //NMM::MutexGuard mg(buffer_mutex);
  
  size_t size;

  is_event = false;

  if ( (read_pos == used_blocks.end()) || skiperror || used_blocks.empty() ){ // EOF-check
    skiperror = false;
    return 0;
  }

  //printUsed();

  // move to correct position
  streamoff next_position = (read_pos->id * file_block_size);
  if (o_file->seekg(next_position,ios::beg)) {
  } else {
    cerr << "TSNode: NOT MOVED to "<< next_position << endl;
    return 0;
  }

  // get the data-size
  size = read_pos->size;
  o_file->read(ctemp,sizeof(size_t)); // read the size from the file
  // DEBUG CHECK
  char** pct = &ctemp;
  size_t sizeff = *(reinterpret_cast<size_t*>(*pct));
  if (sizeff != size) {
    cerr << "TSNode: size: " << read_pos->size << " != " << sizeff <<endl;
  }
  // END DEBUG

  // read the data
  o_file->read(data,size);

  read_pos->already_read = true;  // mark the block as beeing read
  is_event = read_pos->is_event;
  timestamp = read_pos->timestamp;
  read_pos++;                     // and setup read_pos

  // skip blocks
  if (skip_blocks)
    skipBlocks(skip_blocks);

  // DEBUG
  //      d_file->write(data,size);
  // END DEBUG

  return size;
}

bool FileBuffer::skipBlocks(int amount){
  if ((amount > 0) && (read_pos == used_blocks.end())) {
    return false; // cannot skip past EOF
  }

  bool reverse = false;
  if (amount < 0) {
    reverse = true;
    amount *= -1;
  }

  for (int i=0; i < amount; i++) {
    // setup read_pos
    if (reverse) {      // skip backward
      if (read_pos == used_blocks.begin()) { // BOF check
        read_pos->already_read = false;
        skiperror = true;
        return false;
      }
      read_pos--;
    } else {            // skip forward
      read_pos->already_read = true;       // mark the block as beeing read
      read_pos++;
      if  (read_pos == used_blocks.end()) { // EOF check
        skiperror = true;
        return false;
      }
    }
  }

  return true;
}

size_t FileBuffer::getNextBlockSize(bool& is_event) {
  size_t size;

  if ( (read_pos == used_blocks.end()) || skiperror){ // EOF-check
    return 0;
  }

  // get the data-size
  size = read_pos->size;

  is_event = read_pos->is_event;

  return size;
}

void FileBuffer::pushData(const char* data,size_t size,bool is_event, NMM::Time timestamp) {
  //NMM::MutexGuard mg(buffer_mutex);

  Block block_to_write;

  if (size > block_size) {
    cerr << "FILEBUFFER: unable to store " << size << " bytes in blocks of size " << block_size
         << " => setting size to zero!" << endl;
    size = 0;
  }

  // where should the next block be written?
  if ((keep_data == false) &&               // we dont keep all blocks..
      (used_blocks.size() > keep_min) &&    // there are enough blocks in the file..
      (read_pos != used_blocks.begin()) &&  // next popData() will not read the oldest block..
      (used_blocks.begin()->already_read = true)) {  // the oldest block has already been read..
    block_to_write = used_blocks.front();  // so we reuse the oldest block
    used_blocks.pop_front();
    if (used_blocks.front().timestamp != NMM::Time(0)) {
      start_time = used_blocks.front().timestamp; // and set the start_time of the buffer accordingly
    }
  } else {
    block_to_write.id = block_num;  // write block at EOF
    block_num++;                    // increase buffer_num
    if (timestamp != NMM::Time(0)) {
      end_time = timestamp; // and set the start_time of the buffer accordingly
      if (start_time == NMM::Time(0)) {
        start_time = timestamp;
      }
    }
  }
  block_to_write.already_read = false;
  block_to_write.is_event = is_event;
  block_to_write.timestamp = timestamp;
    
  // move to correct position
  streamoff write_position =  (block_to_write.id * file_block_size );
  if (!o_file->seekg( write_position,ios::beg)) {
    cerr << "Unable to expand file " << filename << " to size " << write_position <<
      ". Maybe you should free some space on your harddrive." << endl;
    //exit(-1);
    return;
  }

  // set data-size
  block_to_write.size = size;
  o_file->write(reinterpret_cast<char*>(&size),sizeof(size_t)); // and write it to disk

  // write the data
  o_file->write(data,size);

  used_blocks.push_back(block_to_write);

  // if the last valid block has already been read then
  // the block we wrote is the next block to be read
  if (read_pos == used_blocks.end()) {
    read_pos--; // set read_pos to one before end
  }

  // if there is just one valid block in the list this must be the read_position
  if (used_blocks.size() == 1) {
    read_pos = used_blocks.begin();
    if (read_pos->timestamp != NMM::Time(0)) {
      start_time = read_pos->timestamp;
      end_time = read_pos->timestamp;
    }
  }

  //    printUsed();
}

bool FileBuffer::setReadPosition(Position_t pos) {
  if (used_blocks.empty()) {   // file empty -> no read-position to set
    return false;
  }

  if (pos == BEGINNING)
    read_pos = used_blocks.begin();
  if (pos == END) {
    read_pos = used_blocks.end();
    read_pos--;
  }
  skiperror = false;

  return true;
}

void FileBuffer::setKeepData(bool keep) {
  keep_data = keep;
}

const bool FileBuffer::getKeepData() {
  return (keep_data);
}

bool FileBuffer::saveToFile(string filename) {
#if (__GNUC__ >= 3)
  fstream output(filename.c_str(),ios_base::binary|ios_base::out|ios_base::trunc);//,0666); // open output file
#else
  fstream output(filename.c_str(),ios::binary|ios::out|ios::trunc,0666); // open output file
#endif
  output.unsetf(ios::skipws);

  char* temp = new char[file_block_size];

  for (list<Block>::const_iterator i=used_blocks.begin(); i != used_blocks.end(); i++) { // write the blocks in correct order
    if (! i->is_event) {                                                                // don't write the events
      o_file->seekg( ((i->id * file_block_size) + sizeof(size_t)),ios::beg);          //move to correct position
      o_file->read(temp,i->size);
      output.write(temp,i->size);
    }
  }

  delete[] temp;
  return true;
}

NMM::Interval FileBuffer::getIntervalSize() const {
  NMM::Interval time_amount = end_time - start_time;

  return time_amount;
}

NMM::Time FileBuffer::getStartTime() const {
  //NMM::MutexGuard mg(buffer_mutex);
  
//   if (used_blocks.empty()) {
    return start_time;
//   } else {
//     for (list<Block>::const_iterator i(used_blocks.begin()); i != used_blocks.end(); ++i) {
//       if (i->timestamp.sec != 0) {
//         return i->timestamp;
//       }
//     }
    
//     return used_blocks.front().timestamp;
//   }
}

#include "nmm/nmm_suffix.hpp"
