/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "MPEGTimeShiftingNode.hpp"

#include "nmm/utils/SecurityCheck.hpp"
#include "nmm/utils/thread/MutexGuard.hpp"
#include "nmm/comm/messaging/CEvent.hpp"
#include "nmm/comm/serialize/XMLIValueStream.hpp"
#include "nmm/comm/serialize/XMLOValueStream.hpp"
#include "nmm/multimedia/FlushEvent.hpp"
#include "nmm/multimedia/ISyncReset.hpp"
#include "nmm/multimedia/sync/interfaces/ISynchronized.hpp"

#include "nmm/plugins/interfaces/IStreamJump.hpp"
#include "nmm/plugins/interfaces/IMPEGVideoDecoder.hpp"

#include "nmm/plugins/NMMRegisterPlugin.hpp"

NMM_REGISTER_PLUGIN(MPEGTimeShiftingNode, NMM, "MPEGTimeShiftingNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  MPEGTimeShiftingNode::MPEGTimeShiftingNode(const char * name, StreamQueue::Mode qmode, int qsize)
    : GenericFilterNode(name, qmode, qsize),
      mode(ITimeshifting::LIVE),
      playmode(ITimeshifting::PLAY),
      playback_speed(1),
      mpeg_buffer_size(512 * 1024), // enought space for a DVB-MPEG-GOP
      fbuf(NULL),
      filebuffer_threshold(0),
      use_membuf(false),
      max_membuf_size(0),
      membuf(),
      filename(),
      last_in_buffer(NULL),
      start_play(false),
      start_play2(false),
      start_seek(false),
      send_flush(false),
      send_play(false),
      send_progress(false),
      ready_out_buffers(),
      process_buffer_mutex(),
      keep_data(false),
      current_read_position(Time(0)),
      live_position(Time(0)),
      last_progress_time(Time(0))
  {}
  
  MPEGTimeShiftingNode::~MPEGTimeShiftingNode() {
    // free the memorybuffer
    IntData* temp;
    while (!membuf.empty()) {
      temp = &membuf.front();
      delete[] temp->data;
      membuf.pop();
    }
  }
  
  Result MPEGTimeShiftingNode::doFlush() {
    clearMemoryBuffer();
    return SUCCESS;
  }
  
  Result MPEGTimeShiftingNode::doInit() {
    /* Begin> MPEG audio subformat */
    Format mpeg_audio_format_1(Format::audio_mpeg_type);
    mpeg_audio_format_1.addIntParamValue(Format::audio_codec_param, 0x55);
    mpeg_audio_format_1.addIntParamValue(Format::layer_param, 2);
    mpeg_audio_format_1.addIntParamValue(Format::channels_param, 2);
    mpeg_audio_format_1.addIntParamValue(Format::sample_rate_param, 48000);
    mpeg_audio_format_1.addIntParamValue(Format::bitpersample_param, 16);
    /* End< MPEG audio subformat */
    
    /* Begin> Simple MPEG video subformat */
    Format mpeg_video_format_1(Format::video_mpeg_type);
    mpeg_video_format_1.addIntParamValue(Format::version_param, 2);
    /* End< Simple MPEG video subformat */
    
    /* Begin> Complex MPEG video subformat */    
    Format mpeg_video_format_2(Format::video_mpeg_type);
    mpeg_video_format_2.addIntParamValue(Format::version_param, 1);
    mpeg_video_format_2.addIntParamValue(Format::version_param, 2);
    mpeg_video_format_2.addWildcard(Format::x_resolution_param);
    mpeg_video_format_2.addWildcard(Format::y_resolution_param);
    mpeg_video_format_2.addWildcard(Format::ratio_x_param);
    mpeg_video_format_2.addWildcard(Format::ratio_y_param);
    mpeg_video_format_2.addWildcard(Format::framerate_param);
    /* End< Complex MPEG video subformat */    
    
    /* Begin> Complex MPEG format */
    Format mpeg_format_1(Format::av_mpegps_type);
    mpeg_format_1.addIntParamValue(Format::version_param, 2);
    mpeg_format_1.addIntParamValue(Format::map_mpeg_audio_streams_param, 1);
    mpeg_format_1.addIntParamValue(Format::map_mpeg_video_streams_param, 1);
    mpeg_format_1.addIntParamValue(Format::map_ac3_audio_streams_param, 0);
    mpeg_format_1.addIntParamValue(Format::map_dts_audio_streams_param, 0);
    mpeg_format_1.addIntParamValue(Format::map_lpcm_audio_streams_param, 0);
    mpeg_format_1.addIntParamValue(Format::map_spu_streams_param, 0);
    mpeg_format_1.addFPointerParamValue(Format::format_mpeg_audio_streams_param, new Format(mpeg_audio_format_1));
    mpeg_format_1.addFPointerParamValue(Format::format_mpeg_video_streams_param, new Format(mpeg_video_format_1));
    mpeg_format_1.addFPointerParamValue(Format::format_mpeg_video_streams_param, new Format(mpeg_video_format_2));
    mpeg_format_1.addFPointerParamValue(Format::format_ac3_audio_streams_param, new Format(Format::wild_wild_type));
    mpeg_format_1.addFPointerParamValue(Format::format_spu_streams_param, new Format(Format::wild_wild_type));
    mpeg_format_1.addFPointerParamValue(Format::format_lpcm_audio_streams_param, new Format(Format::wild_wild_type));
    mpeg_format_1.addFPointerParamValue(Format::format_dts_audio_streams_param, new Format(Format::wild_wild_type));
    /* End< Complex MPEG format */
    
    /* Begin> Simple MPEG format */
    Format mpeg_format_2(Format::av_mpegps_type);
    mpeg_format_2.addIntParamValue(Format::version_param, 2);
    /* End< Simple MPEG format */
    
    /* Begin> Init static properties */
    Format *in_f, *out_f;
    
    in_f = new Format(mpeg_format_1);
    out_f = new Format(mpeg_format_1);
    getOwnInputProperty()->addNewFormat(in_f);
    getOwnInputProperty()->setDefaultFormat(in_f);
    getOwnInputProperty()->addNewFormat(out_f);
    getOwnInputProperty()->setDefaultFormat(out_f);
    in_f->setIOPartner(out_f);
    out_f->setIOPartner(in_f);
    
    in_f = new Format(mpeg_format_2);
    out_f = new Format(mpeg_format_2);
    getOwnInputProperty()->addNewFormat(in_f);
    getOwnInputProperty()->addNewFormat(out_f);
    in_f->setIOPartner(out_f);
    out_f->setIOPartner(in_f);
    /* End< Init static properties */
    
    /* Security check for timeshifting buffer */
    if(!SSecurityCheck::getInstance().checkFilenameForReading(filename) ||
       !SSecurityCheck::getInstance().checkFilenameForWriting(filename)) {
      ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return FAILURE;
    }
    
    /* Create timeshifting buffer file */
    fbuf = new FileBuffer(mpeg_buffer_size, filename, filebuffer_threshold);
    
    return SUCCESS;
  }
  
  Result MPEGTimeShiftingNode::doDeinit() {
    if (fbuf) {
      delete fbuf; // release the file-buffer (the corresponding file is deleteted if we are not recording)
      fbuf = NULL;
    }
    
    return SUCCESS;
  }
  
  Result MPEGTimeShiftingNode::doInitOutput() {
    /* Get input format */
    const Format* in_format(getInputFormat());
    if (!in_format) {
      ERROR_STREAM("Input Format is NULL!" << endl);
      return FAILURE;
    }
    
    /* Begin> Init working properties */
    Format* in_f(new Format(*in_format));
    getOwnWorkingInputProperty()->addNewFormat(in_f);
    getOwnWorkingInputProperty()->setDefaultFormat(in_f);
    
    Format* out_f(new Format(*in_format));
    getOwnWorkingOutputProperty()->addNewFormat(out_f);
    getOwnWorkingOutputProperty()->setDefaultFormat(out_f);
    
    in_f->setIOPartner(out_f);
    out_f->setIOPartner(in_f);
    /* End< Init working properties */
    
    return SUCCESS;
  }
  
  Message* MPEGTimeShiftingNode::processBuffer(Buffer* in_buffer) {
    MutexGuard guard(process_buffer_mutex);
    
    /* Check if we need to send a progress event */
    if (getReadPosition().sec != last_progress_time.sec) {
      last_progress_time = getReadPosition();
      send_progress = true;
    }
    
    /* Check if we have a new incoming buffer */
    if (!getWorkingFlag()) {
      /* Remember the buffer for later use */
      last_in_buffer = in_buffer;
      DEBUG_STREAM("WorkingFlag: false" << endl);
      DEBUG_STREAM("incoming buffer ("<< in_buffer <<") has size: " << in_buffer->getUsed() << endl);
      
      /* Check if we are in file buffering mode */
      if (mode == ITimeshifting::FILEBUF) {
        /* put incoming data in the filebuffer */
        DEBUG_STREAM("Buffer-Mode: enqueuing " << last_in_buffer->getUsed() << " bytes" << endl);
        live_position = getTimestamp(in_buffer);
        fbuf->pushData(last_in_buffer->getData(), last_in_buffer->getUsed(), false, live_position);
        DEBUG_STREAM("Buffer-Mode: holding " << fbuf->getUsed() << " buffers" <<endl);
      }
    } else {
      DEBUG_STREAM("WorkingFlag: true" << endl);
    }
    
    /* Begin> Send special events and enqueued buffers */
    if (send_flush) {
      DEBUG_STREAM("Sending FlushEvent" << endl);
      send_flush = false;
      setWorkingFlag(true);
      return new FlushEvent;
    } else if (send_play) {
      DEBUG_STREAM("Sending send_play events" << endl);
      send_play = false;
      setWorkingFlag(true);
      return new CEvent(ITimeshiftingListener::create_timeshiftingBegin());
    } else if (start_play) {
      DEBUG_STREAM("Sending start_play events" << endl);
      start_play = false;
      CEvent* start_play_event = new CEvent();
      start_play_event->insertEvent(start_play_event->size(),
                                    IStreamJump::create_sendIFramesOnly(0));
      start_play_event->insertEvent(start_play_event->size(),
                                    IMPEGVideoDecoder::create_resyncMPEG());
      start_play_event->insertEvent(start_play_event->size(),
                                    ISynchronized::create_setSynchronized(true));
      start_play_event->insertEvent(start_play_event->size(),
                                    ISyncReset::create_syncReset());
      if (playmode == ITimeshifting::FB_LIVE) {
        start_play_event->insertEvent(start_play_event->size(),
                                      ITimeshiftingListener::create_live());
      }
      setWorkingFlag(true);
      return start_play_event;
    } else if (start_play2) {
      DEBUG_STREAM("Sending start_play2 events" << endl);
      start_play2 = false;
      CEvent* start_play_event = new CEvent();
      start_play_event->insertEvent(start_play_event->size(),
                                    IStreamJump::create_sendIFramesOnly(0));
      start_play_event->insertEvent(start_play_event->size(),
                                    IMPEGVideoDecoder::create_resyncMPEG());
      start_play_event->insertEvent(start_play_event->size(),
                                    ISynchronized::create_setSynchronized(true));
      start_play_event->insertEvent(start_play_event->size(),
                                    ISyncReset::create_syncReset());
      setWorkingFlag(true);
      return start_play_event;
    } else if (start_seek) {
      DEBUG_STREAM("Sending start_seek events" << endl);
      start_seek = false;
      CEvent* start_seek_event = new CEvent();
      start_seek_event->insertEvent(start_seek_event->size(),
                                    IStreamJump::create_sendIFramesOnly(1));
      start_seek_event->insertEvent(start_seek_event->size(),
                                    ISynchronized::create_setSynchronized(false));
      setWorkingFlag(true);
      return start_seek_event;
    } else if (send_progress) {
      DEBUG_STREAM("Sending progress event" << endl);
      send_progress = false;
      CEvent* progress_event = new CEvent();
      progress_event->insertEvent(progress_event->size(),
                                  ITimeshiftingListener::create_setProgress(last_progress_time));
      progress_event->setMandatoryBit(false);
      setWorkingFlag(true);
      return progress_event;
    } else if (!ready_out_buffers.empty()) {
      if (ready_out_buffers.size() == 1) {
        /* This is the last buffer. No more work to do */
        setWorkingFlag(false);
      } else {
        setWorkingFlag(true);
      }
      Message* out_message = ready_out_buffers.front();
      ready_out_buffers.pop_front();
      //DEBUG_STREAM("Sending prepared buffer (" << out_buffer << ") of size " << out_buffer->getUsed() << endl);
      return out_message;
    }
    /* End< Send special events and enqueued buffers*/
    
    /* Begin> Input/Output handling */
    switch(mode) {
      case ITimeshifting::LIVE:
        /* Begin> LIVE mode */
        {
          DEBUG_STREAM("LIVE-Mode" << endl);
          
          /* If memory buffering is enabled... */
          if (use_membuf) {
            /* Store incoming data in memory buffer */
            size_t buffer_size = last_in_buffer->getUsed();
            IntData temp;
            temp.size = buffer_size;
            temp.data = new char[buffer_size];
            temp.is_event = false;
            temp.timestamp = getTimestamp(last_in_buffer);
            memcpy(temp.data,last_in_buffer->getData(),buffer_size);
            membuf.push(temp);
            
            /* Pop data if queue is exhausted */
            while (membuf.size() > max_membuf_size) {
              delete[] membuf.front().data;
              membuf.pop();
            }
            
            DEBUG_STREAM("Size of memory-buffer: " << membuf.size() << " buffers" << endl);
          }
          
          /* If data should not be forwarded... */
          if (keep_data) {
            /* Release buffer and return */
            DEBUG_STREAM("keep_data -> releasing buffer (" << last_in_buffer << ") of size " << last_in_buffer->getUsed() << endl);
            last_in_buffer->release();
            setWorkingFlag(false);
            return NULL;
          }
          
          /* in LIVE-Mode simply return the incoming buffer */
          DEBUG_STREAM("Sending buffer(" << last_in_buffer << ") of size " << last_in_buffer->getUsed() << endl);
          setWorkingFlag(false);
          return last_in_buffer;
        }
        break;
        /* End< LIVE mode */
        
    case ITimeshifting::FILEBUF:
      /* Begin> FILEBUF mode */
      {
        switch (playmode) {
        case ITimeshifting::FB_LIVE:   // LIVE but buffered in file
          {
            DEBUG_STREAM("FB-LIVE-Mode" << endl);
            DEBUG_STREAM("Sending buffer(" << last_in_buffer << ") of size " << last_in_buffer->getUsed() << endl);
            setWorkingFlag(false);
            return last_in_buffer;
          }
          break;
          
        case ITimeshifting::PAUSE:     // if in PAUSE-Mode return nothing
          {
            DEBUG_STREAM("PAUSE-Mode: sending 0-buffer..." << endl);
            DEBUG_STREAM("Releasing buffer (" << last_in_buffer << ") of size " << last_in_buffer->getUsed() << endl);
            last_in_buffer->release();
            DEBUG_STREAM("sending 0" << endl);
            setWorkingFlag(false);
            return NULL;
          }
          break;
          
        case ITimeshifting::PLAY:      // if in PLAY or in FAST_FORWARD-Mode...
        case ITimeshifting::FAST_FORWARD:  // ...get the data out of the file-buffer
          {
            for (int i=0; i<playback_speed; i++) {
              /* Analyze next block */
              bool is_event;
              size_t next_buffer_size = fbuf->getNextBlockSize(is_event);
              if (is_event) {
                /* Read event and put it into the output queue */
                CEvent* buffered_event = popEvent(next_buffer_size, 0);
                if (buffered_event) {
                  ready_out_buffers.push_back(buffered_event);
                }
              } else {
                /* Check if we are at the end of the file buffer */
                if (next_buffer_size == 0) {
                  /* Switch back to FB_LIVE mode */
                  DEBUG_STREAM("End of filebuffer reached! ... switching back to FB_LIVE-Mode" << endl);
                  playmode = ITimeshifting::FB_LIVE;
                  
                  /* Send the last incoming buffer */
                  ready_out_buffers.push_back(last_in_buffer);
                  
                  /* Set event flags */
                  start_play = true;
                  send_flush = true;
                  
                  setWorkingFlag(true);
                  return NULL;
                }
                
                /* Read buffer and put it into the output queue */
                Buffer* out_buffer = popBuffer(next_buffer_size, 0);
                if (out_buffer) {
                  ready_out_buffers.push_back(out_buffer);
                }
              }
            }
            
            /* Get the next message from the output queue */
            Message* out_message = ready_out_buffers.front();
            ready_out_buffers.pop_front();
            
            /* Release incoming buffer */
            DEBUG_STREAM("Releasing buffer (" << last_in_buffer << ") of size " << last_in_buffer->getUsed() << endl);
            last_in_buffer->release();
            
            /* Return next output buffer */
            setWorkingFlag(false);
            return out_message;
          }
          break;
          
        case ITimeshifting::REWIND:
          {
            DEBUG_STREAM("REWIND-Mode (playback-speed = " << playback_speed << ")"<< endl);
            
            /* Skip events until we find a buffer */
            size_t next_buffer_size;
            bool is_event;
            do {
              next_buffer_size = fbuf->getNextBlockSize(is_event);
              if (is_event) fbuf->skipBlocks(-1);
            } while(is_event && (next_buffer_size > 0) );
            
            /* Check if we are at the beginning of the buffer */
            if (next_buffer_size == 0) {
              /* Switch to PLAY mode */
              DEBUG_STREAM("Beginning of filebuffer reached! ... switching back to PLAY-Mode" << endl);
              playmode = ITimeshifting::PLAY;
              
              /* Set playback speed */
              playback_speed = 1;
              
              /* Go to beginning of file buffer */
              fbuf->setReadPosition(FileBuffer::BEGINNING);
              next_buffer_size = fbuf->getNextBlockSize(is_event);
              
              /* Read next buffer and put it into the output queue */
              Buffer* out_buffer = popBuffer(next_buffer_size, playback_speed-1);
              if (out_buffer) {
                ready_out_buffers.push_back(out_buffer);
              }
              
              /* Set event flags */
              start_play2 = true;
              send_play = true;
              send_flush = true;
              
              setWorkingFlag(true);
              return NULL;
            }
            
            /* Read next buffer */
            Buffer* out_buffer = popBuffer(next_buffer_size, playback_speed-1);
            
            /* Release incoming buffer */
            DEBUG_STREAM("releasing buffer (" << last_in_buffer << ") of size " << last_in_buffer->getUsed() << endl);
            last_in_buffer->release();
            
            if (out_buffer) {
              DEBUG_STREAM("sending buffer(" << out_buffer << ") of size " << out_buffer->getUsed() << endl);
            }
            
            /* Return next output buffer */
            setWorkingFlag(false);
            return out_buffer;
          }
          break;
          
        }
        
      }
      break;
      /* End< FILEBUF mode */
      
    }
    /* End< Input/Output handling */
    
    ERROR_STREAM("process-buffer bottom: should never reach this point!" << endl);
    start_play = false;
    start_seek = false;
    setWorkingFlag(false);
    return NULL;
  }
  
  void MPEGTimeShiftingNode::setModeLive(const nmm_bool& enable_memory_buffer, const nmm_bool& send_sync_event) {
    DEBUG_STREAM("Switching to LIVE-Mode" << endl);
    MutexGuard process_guard(process_buffer_mutex);
    
    /* Purge the file buffer and create a new one */
    delete fbuf;
    fbuf = new FileBuffer(mpeg_buffer_size, filename, filebuffer_threshold);
    
    /* Set main mode to LIVE */
    mode = ITimeshifting::LIVE;
    
    /* Set play mode to PLAY */
    playmode = ITimeshifting::PLAY;
    
    /* Enable or disable memory buffering */
    use_membuf = enable_memory_buffer;
    
    /* Reset time variables */
    current_read_position = Time(0);
    live_position = Time(0);
    last_progress_time = Time(0);
    
    /* Set event flags */
    if (send_sync_event) {
      start_play = true;
      send_flush = true;
    }
  }
                         
  void MPEGTimeShiftingNode::setModeFBLive(const nmm_bool& send_sync_event) {
    DEBUG_STREAM("Switching to LIVE-Mode" << endl);
    MutexGuard process_guard(process_buffer_mutex);
    
    /* Set main mode to FILEBUF */
    mode = ITimeshifting::FILEBUF;
    
    /* Set play mode to FB_LIVE */
    playmode = ITimeshifting::FB_LIVE;
    
    /* Set event flags */
    if (send_sync_event) {
      start_play = true;
      send_flush = true;
    }
  }

  void MPEGTimeShiftingNode::setModePlay(const nmm_bool& send_sync_reset) {
    DEBUG_STREAM("Switching to PLAY-Mode (reset = " << send_sync_reset << ")" << endl);
    MutexGuard process_guard(process_buffer_mutex);
    
    /* Check if we are in the right state */
    if (mode != ITimeshifting::FILEBUF) {
      ERROR_STREAM("PLAY works only in PAUSE- or FAST_FORWARD-Mode!" << endl);
      return;
    }
    
    /* Set play mode to PLAY */
    playmode = ITimeshifting::PLAY;
    
    /* Set playback speed */
    playback_speed = 1;
    
    /* Set event flags */
    if (send_sync_reset) {
      start_play = true;
      send_flush = true;
    }
  }

  void MPEGTimeShiftingNode::setModePause() {
    DEBUG_STREAM("Switching to PAUSE-Mode" << endl);
    MutexGuard process_guard(process_buffer_mutex);
    
    /* If we are in buffered live mode... */
    if (playmode == ITimeshifting::FB_LIVE) {
      /* Seek to the end of the file buffer */
      DEBUG_STREAM("setting playback-position to end of filebuffer" << endl);
      fbuf->setReadPosition(FileBuffer::END);
    }
    
    /* Set play mode to PAUSE */
    playmode = ITimeshifting::PAUSE;
    
    /* Seek to end of file buffer */
    if (mode == ITimeshifting::LIVE) {
      mode = ITimeshifting::FILEBUF;
      
      /* Disable memory buffering */
      use_membuf = false;
      
      /* Flush memory buffer */
      int membuf_size = membuf.size();
      while (!membuf.empty()) {
        fbuf->pushData(membuf.front().data,membuf.front().size,membuf.front().is_event,membuf.front().timestamp);
        delete[] membuf.front().data;
        membuf.pop();
      }
      
      /* Adjust playback position */
      fbuf->skipBlocks(membuf_size-1);
    }
  }
  
  void MPEGTimeShiftingNode::setModeFastForward(const nmm_int32& speed) {
    DEBUG_STREAM("Switching to FAST-FORWARD-Mode" << endl);
    MutexGuard process_guard(process_buffer_mutex);
    
    /* Check if we are in the right state */
    if ((mode == ITimeshifting::LIVE) || (playmode == ITimeshifting::FB_LIVE)) {
      ERROR_STREAM("sorry, cannot fast-forward a live-picture!" << endl);
      return;
    }
    
    /* Check if playback speed is sane */
    if (speed < 2) {
      ERROR_STREAM("Fast_Forward: " << speed << " out of range!" << endl);
      return;
    }
    
    /* Set play mode to FAST_FORWARD */
    playmode = ITimeshifting::FAST_FORWARD;
    
    /* Set playback speed */
    DEBUG_STREAM("Setting playback-speed to " << speed << endl);
    playback_speed = speed;
    
    /* Set event flags */
    start_seek = true;
    send_flush = true;
  }

  void MPEGTimeShiftingNode::setModeRewind(const nmm_int32& speed) {
    DEBUG_STREAM("Switching to REWIND-Mode" << endl);
    MutexGuard process_guard(process_buffer_mutex);
    
    /* Check if we are in the right state */
    if (mode == ITimeshifting::LIVE) {
      ERROR_STREAM("sorry, cannot rewind a live-picture!" << endl);
      return;
    }
    
    /* Check if playback speed is sane */
    if (speed > 0) {
      ERROR_STREAM("Rewind: " << speed << " out of range!" << endl);
      return;
    }
    
    /* Seek to end of file buffer */
    if (playmode == ITimeshifting::FB_LIVE) {
      fbuf->setReadPosition(FileBuffer::END);
    }
    
    /* Set playback speed */
    DEBUG_STREAM("Setting playback-speed to "<< speed << endl);
    playback_speed = speed;
    
    /* Set play mode to REWIND */
    playmode = ITimeshifting::REWIND;
    
    /* Set event flags */
    start_seek = true;
    send_flush = true;
  }
  
  void MPEGTimeShiftingNode::enableMemBuf(const nmm_uint32& size) {
    MutexGuard process_guard(process_buffer_mutex);
    use_membuf = true;
    max_membuf_size = size;
  }
  
  void MPEGTimeShiftingNode::disableMemBuf(const nmm_bool& play) {
    MutexGuard process_guard(process_buffer_mutex);
    
    /* Flush memory buffering */
    while (!membuf.empty()) {
      if (play) {
        /* Store memory buffer contents in the file buffer */
        fbuf->pushData(membuf.front().data,membuf.front().size,membuf.front().is_event,membuf.front().timestamp);
      }
      delete[] membuf.front().data;
      membuf.pop();
    }
    
    use_membuf = false;
    
    if (play) {
      /* Start instant replay */
      mode = ITimeshifting::FILEBUF;
      playmode = ITimeshifting::PLAY;
      playback_speed = 1;
      start_play = true;
      send_flush = true;
    }
  }
  
  void MPEGTimeShiftingNode::setFilename(const string& s) {
    MutexGuard process_guard(process_buffer_mutex);
    
    if (isOutputInitialized()) {
      /* If file buffer is in use, we can't change the filename */
      if (fbuf->getUsed() == 0) {
        filename = s;
        
        /* Purge file buffer and create a new one */
        delete fbuf;
        fbuf = new FileBuffer(mpeg_buffer_size, filename, filebuffer_threshold);
      }
    } else {
      filename = s;
    }
  }
  
  void MPEGTimeShiftingNode::setFileBufferThreshhold(const nmm_uint32& s) {
    filebuffer_threshold = s;
  }

  void MPEGTimeShiftingNode::startRecord() {
    MutexGuard process_guard(process_buffer_mutex);
    fbuf->setKeepData(true);
  }

  void MPEGTimeShiftingNode::stopRecord() {
    MutexGuard process_guard(process_buffer_mutex);
    fbuf->setKeepData(false);
  }
  
  nmm_bool MPEGTimeShiftingNode::isRecording() {
    return fbuf->getKeepData();
  }
  
  void MPEGTimeShiftingNode::keepData(const nmm_bool& keepit) {
    MutexGuard process_guard(process_buffer_mutex);

    if (keepit) {
      DEBUG_STREAM("no longer forwarding any data" << endl );
    } else {
      DEBUG_STREAM("now sending data again" << endl);
    }
    keep_data = keepit;
  }
  
  void MPEGTimeShiftingNode::clearMemoryBuffer() {
    IntData* temp;
    MutexGuard process_guard(process_buffer_mutex);
    
    /* Purge memory buffer */
    while (!membuf.empty()) {
      temp = &membuf.front();
      delete[] temp->data;
      membuf.pop();
    }
  }
  
  Result MPEGTimeShiftingNode::processCEvent(CEvent* event) {
    DEBUG_STREAM ( "entering processCEvent" << endl);
    
    if (event->getDirection() == Message::UPSTREAM) {
      return SUCCESS;
    }
    
    Result r = GenericFilterNode::processCEvent(event);
    
    XMLOValueStream xov; //create a XML based OValueStream
    xov << *event;
    string s = xov.toString();
    
    if (mode == ITimeshifting::LIVE) { // MEMORYBUFFER
      if (use_membuf) { // if memorybuffering is activated store the event in the memorybuffer-queue
        IntData temp;
        temp.size = s.size()+1;
        temp.data = new char[temp.size];
        temp.is_event = true;
        memcpy(temp.data,s.c_str(),temp.size);
        membuf.push(temp);
        DEBUG_STREAM("enqueued event (memory-buffer)" << endl << s.c_str() << endl);
      }
    } else { // FILEBUFFER
      fbuf->pushData(s.c_str(),s.size()+1,true,Time(0));
      // delete events
      for (uint i = 0; i < event->size(); i++) {
        event->removeEvent(i);
      }
    }
    return r;
  }
  
  Result MPEGTimeShiftingNode::setBufferSize(const nmm_uint32& size) {
    mpeg_buffer_size = size;
    return SUCCESS;
  }

  Result MPEGTimeShiftingNode::saveToFile(const string& filename) {
    if (fbuf) {
      if (fbuf->saveToFile(filename))
        return SUCCESS;
      else
        return FAILURE;
    } else {
      return FAILURE;
    }
  }
  
  Time MPEGTimeShiftingNode::getTimestamp(Buffer* in_buffer) {
    Time timestamp = (in_buffer->getTimestamp()).sync_time;
    return timestamp;
  }
  
  Buffer* MPEGTimeShiftingNode::popBuffer(size_t next_buffer_size, int skip) {
    /* Allocate new buffer */
    Buffer* out_buffer = getNewBuffer(next_buffer_size);
    if(!out_buffer) {
      ERROR_STREAM("Unable to allocate a new buffer" << endl);
      return NULL;
    }    
    DEBUG_STREAM("Allocated buffer (" << out_buffer << ")" << endl);
    
    /* Read data form file buffer */
    bool is_event;
    size_t size = fbuf->popData(out_buffer->getData(), is_event, current_read_position, skip );
    if (size != next_buffer_size) {
      ERROR_STREAM("Filebuffer trouble: got " << size << " bytes instead of " << next_buffer_size << endl);
    }
    out_buffer->setUsed(size);
    
    return out_buffer;
  }
  
  CEvent* MPEGTimeShiftingNode::popEvent(size_t next_buffer_size, int skip) {
    /* Read data into buffer */
    bool is_event;
    vector<char> space(next_buffer_size);
    fbuf->popData(&space[0], is_event, current_read_position, skip);
    
    /* Convert buffer to string */
    string sp_str(&space[0]);
    
    /* Parse XML event */
    XMLIValueStream xiv(sp_str);
    CEvent* buffered_event = new CEvent;
    xiv >> *buffered_event;
    
    return buffered_event;
  }
  
  Time MPEGTimeShiftingNode::getReadPosition() const /* limited(INode::STARTED) */ {
    return current_read_position;
  }

  Interval MPEGTimeShiftingNode::getFileLength() const /* limited(INode::STARTED) */ {
    return fbuf->getIntervalSize();
  }

  Interval MPEGTimeShiftingNode::getLiveOffset() const /* limited(INode::STARTED) */ {
    return (live_position - current_read_position);
  }

  Time MPEGTimeShiftingNode::getStartTime() const /* limited(INode::STARTED) */ {
    if (mode == ITimeshifting::LIVE) { // MEMORYBUFFER
      return live_position;
    } else {
      return fbuf->getStartTime();
    }
  }
  
  Time MPEGTimeShiftingNode::getLivePosition() const /* limited(INode::STARTED) */ {
    return live_position;
  }

  bool MPEGTimeShiftingNode::isMemBuf() {
    return use_membuf;
  }

  ITimeshifting::PlayMode MPEGTimeShiftingNode::getPlayMode() const {
    return playmode;
  }

  ITimeshifting::Mode MPEGTimeShiftingNode::getMode() const {
    return mode;
  }
  
}

#include "nmm/nmm_suffix.hpp"
