/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "WavReadNode.hpp"

#include "nmm/plugins/NMMRegisterPlugin.hpp"
#include "nmm/comm/ProxyObject.hpp"
#include "nmm/multimedia/interfaces/ITrack.hpp"
#include "nmm/multimedia/interfaces/ITrackList.hpp"
#include "nmm/multimedia/ISyncReset.hpp"
#include "nmm/multimedia/interfaces/ITrackDuration.hpp"
#include "nmm/multimedia/FlushEvent.hpp"
#include "nmm/multimedia/FileNotFoundException.hpp"
#include "nmm/plugins/interfaces/ITrackSize.hpp"
#include "nmm/utils/SecurityCheck.hpp"
#include "nmm/utils/directory_utils.hpp"
#include "nmm/utils/time/Rational.hpp"
#include "nmm/utils/time/TimedElement.hpp"

#include <iomanip>

NMM_REGISTER_PLUGIN(WavReadNode, NMM, "WavReadNode");

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  int64_t abs(int64_t x) { return (x >= 0) ? x : -x; }

  WavReadNode::WavReadNode(const char *name /* = "WavReadNode" */,
                           StreamQueue::Mode mode /* = StreamQueue::MODE_SUSPEND */,
                           int queue_size /* = StreamQueue::DEFAULT_SIZE */)
    : GenericSourceNode(name, mode, queue_size),
      buffersize_changed(false),    
      buffer_size(8*1024),
      filename(),
      channels(0),
      bytes_per_sample(0),
      bytes_per_second(0),
      filesize(0),
      filepos(0),
      seekto(0),
      track_duration(0, 0),
      state(START_FLUSH),
      send_format(false),
      send_header(false)
  {}
  
  WavReadNode::~WavReadNode() {
    in.close();
  }
  
  void WavReadNode::setFilename(const string& name) {
    filename = name;
    filepos = 0;
    seekto = 0;
  }  
	
  nmm_uint64 WavReadNode::getFilesize() {
    return filesize;
  }
  
  void WavReadNode::setBufferSize(const nmm_int32& size) {
    buffer_size = size;
    buffersize_changed = true;
  }
  
  nmm_int32 WavReadNode::getBufferSize() {
    return buffer_size;
  }

  Result WavReadNode::doInitOutput() {
    
    unsigned char *b = new unsigned char[160];

    long int sample_rate;
    long int bits_per_sample;//, bitrate;
    long int pcm_length;
    
    /* --- open the file ----------------------------------- */
    in.clear(); 
    
    if ("" == filename) {
      ERROR_STREAM("Filename not set" << endl);
      throw FileNotFoundException(this, "Filename not specified.");
    }

    // Convert filename to an absolute path
    filename = getAbsoluteNativePath(filename);
    
    if(!SSecurityCheck::getInstance().checkFilenameForReading(filename)) {
      ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      throw FileNotFoundException(this, "Security check not passed.", filename);
    }
    
    in.open(filename.c_str(), fstream::in | ios::binary);
    
    if (!in.is_open()) {
      ERROR_STREAM("Can\'t open " << filename << endl);
      throw FileNotFoundException(filename);
    }
    
    
    // get filesize and seek to beginning again
    in.seekg(0, ios::end);
    filesize = in.tellg();
    in.seekg(0, ios::beg);

    /* --- analyse the header and set the out format ------- */

    /* read the first 160 Byte */
    in.read(reinterpret_cast<char*>(b), 160);

    /* 00 - 03 "RIFF" ASCII characters */
    if (!((b[0] == 'R') && (b[1] == 'I') && (b[2] = 'F') && (b[3] == 'F'))) {
      ERROR_STREAM("Error parsing wav header: can\'t find \'RIFF\' chunk" << endl);
      return FAILURE;
    }

    /* 12 - 15 "fmt_" ASCII characters */
    if (!((b[12] = 'f') && (b[13] = 'm') && (b[14] = 't') && (b[15] = '_'))) {
      ERROR_STREAM("Error parsing wav header: can\'t find \'FORMAT\' chunk" << endl);
      return FAILURE;
    }

    /* Byte 22-23 Channel numbers */
    switch (b[22] + (b[23] << 8)) {
    case 0x01: channels = 1; break; /* mono */
    case 0x02: channels = 2; break; /* stereo */
    default:
      ERROR_STREAM("Error can\'t determinate number of channels" << endl);
      return FAILURE;
    }

    /* 24 - 27 Sample rate (binary in Hz) */
    sample_rate = b[24] + (b[25] << 8) + (b[26] << 16) + (b[27] << 24);

    /* 28 - 31 Bytes per second */
    bytes_per_second = b[28] + (b[29] << 8) + (b[30] << 16) + (b[31] << 24);

    /* 32 - 33 Bytes per sample */
    bytes_per_sample = b[32] + (b[33] << 8);

    /* 34 - 35 Bits per sample */
    bits_per_sample = b[34] + (b[35] << 8);

    /* search 'DATA' Chunk */
    long int sdata = 36;
    while (!((b[sdata+0] == 'd') && (b[sdata+1] == 'a') && (b[sdata+2] == 't') && (b[sdata+3] == 'a'))) {
      sdata += 1;
      if (sdata > 155) {
        ERROR_STREAM("Error parsing wav header: can\'t find \'DATA\' chunk" << endl);
        return FAILURE;
      }
    }

    /* sdata+4 - sdata+7 Length of data to follow */
    pcm_length = b[sdata+4] + (b[sdata+5] << 8) + (b[sdata+6] << 16) + (b[sdata+7] << 24);

    // compute track duration
    track_duration = Interval(Rational(pcm_length, bytes_per_second));
    
    in.seekg(sdata+8, ios::beg);  /* skip the header */
    
    //the buffersize should be a multiple of the used bits per sample
    if(!buffersize_changed){
      buffer_size = 8192;
      while (buffer_size % (channels * bits_per_sample) != 0) {
        buffer_size += 2;
      }
      
      setBufferSize(buffer_size);

      buffersize_changed = false;
    }
    
    /* --- now define the formats -------------------------- */

    Format *out_format = getOwnWorkingOutputProperty()->addNewFormat("audio/raw");
    
    DEBUG_STREAM("Channels         : " << channels << endl);
    out_format ->addIntParamValue(Format::channels_param, channels);
    
    DEBUG_STREAM("Sample rate      : " << sample_rate << endl);
    out_format ->addIntParamValue(Format::sample_rate_param, sample_rate);
    
    DEBUG_STREAM("Bytes per second : " << bytes_per_second << endl);
    DEBUG_STREAM("Bytes per sample : " << bytes_per_sample << endl);
    DEBUG_STREAM("Bits per sample  : " << bits_per_sample << endl);
    out_format ->addIntParamValue(Format::bitpersample_param, bits_per_sample);
    
    DEBUG_STREAM("PCM length       : " << pcm_length << endl);
    
    int t = (int) pcm_length / bytes_per_second;
    DEBUG_STREAM("Playtime         : "
                 << setfill('0') << setw(2) << (int) (t / 3600) << ":"
                 << setfill('0') << setw(2) << (int) ((t % 3600) / 60) << ":"
                 << setfill('0') << setw(2) << (int) (t % 3600 % 60)
                 << endl);
    
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::encoding_param, "linear");
    out_format->addStringParamValue(Format::sign_param, bits_per_sample == 8 ? "unsigned" : "signed");
    
    getOwnWorkingOutputProperty()->setDefaultFormat(out_format);
    
    delete [] b;

    // check if we are already connected
    if(isOutputConnected("default") && !(*getOutputFormat() == *out_format)) {
      DEBUG_STREAM1("WAVReadode: send_format = true" << endl);
      send_format = true;
    } else {
      DEBUG_STREAM1("WAVReadode: send_format = false" << endl);
      send_format = false;
    }

    setProducingFlag(true);

    if(state == TRACK_ENDED || state == FINISHED) {
      DEBUG_STREAM1("WAVReadNode: START" << endl);
      state = START;
    }  
    else {
      DEBUG_STREAM1("WAVReadNode: START_FLUSH" << endl);
      state = START_FLUSH;
    }

    return SUCCESS;
  }

  Message *WavReadNode::processBuffer(Buffer *) {
    Event e;
    long int bytes_read;
    string event_str = "";
    string::size_type pos1;
    string::size_type pos2;

    switch (state) {
    case START_FLUSH:
    case START:
      {
        CEvent* ce = 0;
        if(state == START) {
          ce = new CEvent();
        } else {
          ce = new FlushEvent();
        }
        pos1 = filename.rfind("/");
        if (filename.npos == pos1) { 
          pos1=0; 
        } else {
          pos1++;
        }
        pos2 = filename.find(".wav",pos1);

        // Reset stream counter
        current_ts = Timestamp();

        // Add startTrack event
        DEBUG_STREAM("Send startTrack event" << endl);
        ce->insertEvent(ITrack::create_startTrack(filename.substr(pos1,pos2-pos1)));
        ce->insertEvent(ITrackSize::create_trackSize(filesize));

        // Add setFormat event if necessary
        if(send_format) {
          send_format = false;

          //set new format
          setOutputFormat(*getOwnWorkingOutputProperty()->getDefaultFormat());

          DEBUG_STREAM("Sending setFormat" << endl);
          ce->insertEvent(IFormatChange::create_setFormat(*getOutputFormat()));

        }

        // Add syncReset event
        DEBUG_STREAM("Sending syncReset" << endl);
        ce->insertEvent(ISyncReset::create_syncReset());

        // Add trackDuration event
        DEBUG_STREAM("Sending trackDuration" << endl);
        ce->insertEvent(ITrackDuration::create_trackDuration(track_duration));

        state = DATA;
        return ce;
      }

    case TRACK_ENDED:
      if(true) { // TODO: if (file queue is empty)
        state = FINISHED;
        DEBUG_STREAM("Send finished event " << endl);
        setProducingFlag(false);
        return new CEvent(ITrackList::create_endTrackList());
      }

    case DATA_STOPPED:
      {
        // Next state is DATA.
        state = DATA;
        return 0;
      }

    case DATA:
      {
        if (seekto != 0) {
          // test if we would get negative
          if (seekto < 44 && abs(seekto) > static_cast<int64_t>(filepos)) {
            // if seek result would be negative go to beginning of file
            in.seekg(44, ios::beg);
            filepos = 44;
            DEBUG_STREAM("Seeked to beginning " << seekto << " from " << filepos << ", filesize " << filesize << endl);
          }
          else if (filepos + seekto > filesize) {
            // set back to short-before-file-end
            in.seekg(0, ios::end);
            filepos = filesize;
            DEBUG_STREAM("Seeked to end " << seekto << " from " << filepos << ", filesize " << filesize << endl);
          }
          else {
            // seek normally
            in.seekg(filepos + seekto);
            filepos = filepos+seekto;
            DEBUG_STREAM("Seeked normally " << seekto << " from " << filepos << ", filesize " << filesize << endl);
          }
          seekto = 0;

          CEvent* cevent = new FlushEvent();
          cevent->insertEvent(cevent->size(), ISyncReset::create_syncReset());
          return cevent;
        }

        if (send_header) {
          in.seekg(0,ios::beg);
          char* tmp = new char[44];
          in.read(tmp,44);
          string retval;
          for (uint i=0;i<44;i++)
            retval+=tmp[i];
          CEvent* cevent = new CEvent();
          cevent->insertEvent(IHeader::create_FileHeader(retval));
          send_header = false;
          in.seekg(filepos);
          return cevent; 
        }

        Buffer* out_buffer = getNewBuffer(buffer_size);  
        if(!out_buffer) {
          ERROR_STREAM("Error allocating new buffer" << endl);
          return NULL;
        }

        in.read(out_buffer->getData(), buffer_size);
        bytes_read = in.gcount();
        current_ts.sync_time = Time(0) + Interval(Rational(filepos, bytes_per_second));
        current_ts.stream_counter++;
        current_ts.is_valid = true;

        out_buffer->setTimestamp(current_ts);

        if (0 < bytes_read) {
          out_buffer->setUsed(bytes_read); 
          filepos+=bytes_read;
        } 
        else if (0 == bytes_read) {
          out_buffer->release();

          DEBUG_STREAM("Send endTrack event " << endl);
          state = TRACK_ENDED;
          CEvent* cevent = new CEvent();
          cevent->insertEvent(ITrack::create_endTrack());
          return cevent;
        }
        else {
          ERROR_STREAM("Error reading file" << endl);
          out_buffer->setUsed(0); 
        }
        

        return out_buffer;
      }
    default:
      ERROR_STREAM("Internal state error" << endl);     
      exit(1);
    }
  } /* produceBuffer */
  
  Result WavReadNode::doInit() {
    /*Get the output property and erase the old formats*/
    Property* property = getOwnOutputProperty();
    
    /*The doInit method stores all formats in the input and output property which are
      supported by this node. These properties are used to register a Node by a registry.  */
    Format* out_format = property->addNewFormat("audio/raw");
    out_format->addWildcard(Format::channels_param);
    out_format->addWildcard(Format::sample_rate_param);
    out_format->addWildcard(Format::bitpersample_param);
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::encoding_param, "linear");
    out_format->addStringParamValue(Format::sign_param, "signed");
    
    property->setDefaultFormat(out_format);
    return SUCCESS;
  }
  
  Result WavReadNode::doDeinitOutput() {
    getOwnWorkingOutputProperty()->erase();
    in.close();
    return SUCCESS;
  }

  Result WavReadNode::doDeinit( ) {
    /*Erase the Formats added in doRegister*/
    getOwnOutputProperty()->erase();
    return SUCCESS;
  } 

  Result WavReadNode::doStop( ) {
    if(state == DATA) {
      state = DATA_STOPPED;
    }
    return SUCCESS;
  } 

  /******************** ISeekable interface ********************/
  nmm_bool WavReadNode::hasPercentSeek() const {
    return true;
  }

  void WavReadNode::seekPercent(const Rational& percent) {
    Rational internal_percent(percent);
    if (internal_percent > Rational(1)) {
      ERROR_STREAM("Seeking more than 100% not possible, defaulting to end!" << endl);
      internal_percent = Rational(1);
    }
    else if (internal_percent < Rational(-1)) {
      ERROR_STREAM("Seeking less than -100% not possible, defaulting to beginning!" << endl);
      internal_percent = Rational(-1);
    }

    Rational rat_seekto = Rational(filesize) * internal_percent;
    seekto = rat_seekto.getNumerator() / rat_seekto.getDenominator();

    /* FIXME !!!!!!!!*/
    int modulo(seekto % (channels * bytes_per_sample /2));
    /*    cerr << "channels " << channels << endl;
          cerr << "Bytes per sample " << bytes_per_sample << endl;
          cerr << "Modulo " << modulo << endl;
          cerr << "Seekto " << seekto << endl;*/
    if (modulo) {
      seekto += (/*channels **/ bytes_per_sample ) - modulo;
    }
    /*    cerr << "Seekto " << seekto << endl;

    cerr << "Setting seekto value to " << seekto << endl; 
    cerr << "=> " << ((filepos + seekto) - 44) % bytes_per_sample << endl; */
    if(((filepos + seekto) - 44) % bytes_per_sample) {
      seekto -= ((filepos + seekto) - 44) % bytes_per_sample;
      //       cerr << "=> " << ((filepos + seekto) - 44) % bytes_per_sample << endl; 
    }	
    DEBUG_STREAM(seekto << " = " << internal_percent << " * " << filesize << endl);
  }
   
  void WavReadNode::seekPercentTo(const Rational& percent) {
    seekPercent(percent - Rational(filepos, filesize));
  }

  nmm_bool WavReadNode::hasBytesSeek() const {
    return true;
  }

  void WavReadNode::seekBytes(const nmm_int64& offset) {
    if ((offset < 0) && (abs(offset) > static_cast<int64_t>(filepos))) {
      ERROR_STREAM("Seeking to negativ byte-offset not possible, defaulting to beginning of file!" << endl);
      seekBytesTo(0);
    } else {
      seekBytesTo(filepos + offset);
    }
  }

  void WavReadNode::seekBytesTo(const nmm_uint64& bytes) {
    u_int64_t internal_bytes(bytes);

    if (internal_bytes > filesize) {
      ERROR_STREAM("Seeking over EOF not possible, defaulting to EOF!" << endl);
      internal_bytes = filesize;
    }

    seekto = internal_bytes - filepos;

    int modulo(seekto % (channels * bytes_per_sample));
    if (modulo) {
      seekto += (channels * bytes_per_sample) - modulo;
    }

    DEBUG_STREAM(seekto << " = " << internal_bytes << " - " << filepos << endl);
  }

  nmm_bool WavReadNode::hasTimeSeek() const {
    return true;
  }

  void WavReadNode::seekTime(const Interval& interval) {
    Rational rat_interval(Rational(interval.sec) + Rational(interval.nsec, nano_per_sec));

    Rational rat_seekto(rat_interval * Rational(bytes_per_second));

    seekto = rat_seekto.getNumerator() / rat_seekto.getDenominator();

    int modulo(seekto % (channels * bytes_per_sample));
    if (modulo) {
      seekto += (channels * bytes_per_sample) - modulo;
    }

    DEBUG_STREAM(seekto << " = " << interval << " - " << filepos << endl);
  }

  void WavReadNode::seekTimeTo(const Time& time) {
    Rational rat_time(Rational(time.sec) + Rational(time.nsec, nano_per_sec));

    Rational rat_pos(rat_time * Rational(bytes_per_second));

    seekto = rat_pos.getNumerator() / rat_pos.getDenominator() - filepos;

    //int modulo(seekto % (channels * bytes_per_sample));
    int modulo(seekto % (channels * bytes_per_sample/2));
    if (modulo) {
      //seekto += (channels * bytes_per_sample) - modulo;
      seekto += bytes_per_sample - modulo;
    }

    if(((filepos + seekto) - 44) % bytes_per_sample) {
      seekto -= ((filepos + seekto) - 44) % bytes_per_sample;
      //       cerr << "=> " << ((filepos + seekto) - 44) % bytes_per_sample << endl; 
    }
    DEBUG_STREAM(seekto << " = " << time << " - " << filepos << endl);
  }
  /******************** ISeekable interface ********************/
}


#include "nmm/nmm_suffix.hpp"
