/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "WavWriteNode.hpp"

#include "nmm/comm/format/Property.hpp"
#include "nmm/utils/SecurityCheck.hpp"

#include "nmm/plugins/NMMRegisterPlugin.hpp"

#include <iostream>

NMM_REGISTER_PLUGIN(WavWriteNode, NMM, "WavWriteNode");

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  WavWriteNode::WavWriteNode(const char* name  /* = "WavWriteNode"*/,
                             const StreamQueue::Mode queue_mode /*= StreamQueue::MODE_SUSPEND*/,
                             const int size /*= StreamQueue::DEFAULT_SIZE*/)
  : GenericSinkNode(name, queue_mode, size) {

    bytes_receive = 0;
    filename = "";
    state = START;

    override_default_filename = false;
    file_extension = "wav";
  }

  WavWriteNode::~WavWriteNode() {}

  Result WavWriteNode::doInit() {
    Format* in_format = getOwnInputProperty()->addNewFormat("audio/raw");
    in_format->addWildcard(Format::channels_param);
    in_format->addWildcard(Format::sample_rate_param);
    in_format->addWildcard(Format::bitpersample_param);
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::encoding_param, "linear");
    in_format->addStringParamValue(Format::sign_param, "signed");

    return SUCCESS;
  }

  Result WavWriteNode::doActivate() {
    return SUCCESS;
  }

  Result WavWriteNode::doStop() {
    return SUCCESS;
  }

  Message* WavWriteNode::processBuffer(Buffer *in_buffer) {
    Event e;

    switch (state) {
      case START:
        break;
      case DATA:
        out_fstream.write(in_buffer->getData(), in_buffer->getUsed());
        bytes_receive += in_buffer->getUsed();
        in_buffer->release();  /* release buffer */
        break;

      default:
        ERROR_STREAM("internal state error" << endl);
    }

    return NULL;
  } /* processBuffer */


  bool WavWriteNode::isReady() {
    return ready;
  } /* isReady */

  void WavWriteNode::setFilename(const string& _filename) {
    // if this method is called, the user wants to set his own filename
    // and does not want to take the default (which would be the string
    // from the start_track event, see EventOpenFile)
    // So it must be called before the next start_track arrives
    filename = _filename;
    override_default_filename = true;
  }

  void WavWriteNode::setWorkingDirectory(const string& wd) {
    working_directory = wd;
  }

  string WavWriteNode::getWorkingDirectory() {
    return working_directory;
  }

  void WavWriteNode::setFileExtension(const string& ext) {
    file_extension = ext;
  }

  string WavWriteNode::getFileExtension() {
    return file_extension;
  }

  Result WavWriteNode::startTrack(const string& _filename) {

    out_fstream.clear();

    if (!override_default_filename) {
      filename = _filename;
      if ("" != file_extension) {
        filename += "." + file_extension;
      }
      if ("" != working_directory) {
        filename = working_directory + "/" + filename;
      }
      // set back override because next track should use trackname (provided
      // by start_track event) by default again
      override_default_filename = false;
    }

    DEBUG_STREAM("Open File: " << filename << " (" << _filename << ")" << endl);
    /**
     * open file:
     * - binary mode
     * - delete old file with same name
     * - stream is for output
     */
    if(!SSecurityCheck::getInstance().checkFilenameForWriting(filename)) {
      ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return FAILURE;
    }
    out_fstream.open(filename.c_str(), ios::binary | ios::trunc | ios::out);

    if (!out_fstream) {
      ERROR_STREAM2("File not opened" << std::endl);
    }

    const Format* in_format = getInputFormat();
    int sample_rate = in_format->getIntValue(Format::sample_rate_param);
    int num_channels = in_format->getIntValue(Format::channels_param);
    int bitpersample = in_format->getIntValue(Format::bitpersample_param);
    int blockalign = (bitpersample * num_channels) >> 3;
    int bps = blockalign * sample_rate;

    int sample_rate_low = sample_rate & 0xFF;
    int sample_rate_high = (sample_rate & 0xFF00) >> 8;

    int blockalign_low = blockalign & 0xFF;
    int blockalign_high = (blockalign & 0xFF00) >> 8;

    int bitpersample_low = bitpersample & 0xFF;
    int bitpersample_high = (bitpersample & 0xFF00) >> 8;

    int num_channels_low = num_channels & 0xFF;
    int num_channels_high = (num_channels & 0xFF00) >> 8;

    int bps_0 = bps & 0xFF;
    int bps_1 = (bps & 0xFF00) >> 8;
    int bps_2 = (bps & 0xFF0000) >> 16;
    int bps_3 = (bps & 0xFF000000) >> 24;

    unsigned char header[44] =
    { /* RIFF Chunk */
      0x52, 0x49, 0x46, 0x46, /* 00-03 : "RIFF" in ASCII */
      0x00, 0x00, 0x00, 0x00, /* 04-07 : Length of the Package to follow */
      0x57, 0x41, 0x56, 0x45, /* 08-11 : "WAVE" in ASCII */

      /* FORMAT Chunk */
      0x66, 0x6D, 0x74, 0x20, /* 12-15 : "fmt_" in ASCII */
      0x10, 0x00, 0x00, 0x00, /* 16-19 : Length of FORMAT Chunk = 0x10 */
      0x01, 0x00,             /* 20-21 : always 0x01 */
      0x00, 0x00,             /* 22-23 : channels 0x01=mono, 0x02=stereo */
      0x00, 0x00, 0x00, 0x00, /* 24-27 : sample rate in Hz */
      0x00, 0x00, 0x00, 0x00, /* 28-31 : Bytes per second */
      0x00, 0x00,             /* 32-33 : Bytes per sample */
      0x00, 0x00,             /* 34-35 : Bits per sample */

      /* Data Chunk */
      0x64, 0x61, 0x74, 0x61, /* 36-39 : "data" in ASCII */
      0x00, 0x00, 0x00, 0x00  /* 40-43 : Length of data to follow */
    };

    header[22] = num_channels_low; header[23] = num_channels_high;

    header[24] = sample_rate_low; header[25] = sample_rate_high;
    header[26] = 0x00; header[27] = 0x00;

    header[28] = bps_0; header[29] = bps_1; /* ->  176400 Bytes per second */
    header[30] = bps_2; header[31] = bps_3;

    header[32] = blockalign_low; header[33] = blockalign_high; /* -> 4 Bytes per sample -> 16 bit stereo */

    header[34] = bitpersample_low; header[35] = bitpersample_high; /* -> 16 Bites per sample */
    //cast to char*
    out_fstream.write(reinterpret_cast<char*>(header),44); /* write header to file */

    state = DATA;
    return SUCCESS;
  }

  Result WavWriteNode::endTrack() {
    long int l1 = bytes_receive + 44 - 8;
    long int l2 = bytes_receive;

    out_fstream.seekp(4, ios::beg); /* Start at bit 4 */
    out_fstream.put( (l1 & 0x000000FF)       );
    out_fstream.put( (l1 & 0x0000FF00) >>  8 );
    out_fstream.put( (l1 & 0x00FF0000) >> 16 );
    out_fstream.put( (l1 & 0xFF000000) >> 24 );

    out_fstream.seekp(40, ios::beg); /* Start at bit 40 */
    out_fstream.put( (l2 & 0x000000FF)       );
    out_fstream.put( (l2 & 0x0000FF00) >>  8 );
    out_fstream.put( (l2 & 0x00FF0000) >> 16 );
    out_fstream.put( (l2 & 0xFF000000) >> 24 );
    out_fstream.close();

    state = START;

    DEBUG_STREAM("Close File: " << filename << endl);
    out_fstream.close();

    return SUCCESS;
  }

  Result WavWriteNode::started(const Interval&){
    return SUCCESS;
  }


  Result WavWriteNode::finished() {
    DEBUG_STREAM("Finished event." << endl);
    return SUCCESS;
  }

}


#include "nmm/nmm_suffix.hpp"
