/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "BufferDropNode.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"
#include "nmm/comm/format/Property.hpp"
#include "nmm/comm/messaging/Buffer.hpp"
#include <stdio.h>

NMM_REGISTER_PLUGIN(BufferDropNode, NMM, "BufferDropNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  BufferDropNode::BufferDropNode(const char *name, StreamQueue::Mode qmode, int qsize) :
    GenericFilterNode(name, qmode, qsize),
    drop_num(0),
    drop_denom(0),
    drop_ratio(0),
    ratio_counter(0)
  {
    buffer_to_drop = 0;
    buffer_count = 0;
  }
  
  void BufferDropNode::setBufferToDrop(const nmm_uint32 &buffer_num)
  {
    buffer_to_drop = buffer_num;
  }
  
  Result BufferDropNode::doInit() 
  {
    Format* any_format_i = new Format("*/*");
    getOwnInputProperty()->addNewFormat(any_format_i);
    
    Format* any_format_o = new Format("*/*");
    getOwnOutputProperty()->addNewFormat(any_format_o);
    
    return SUCCESS;
  }
  
  Result BufferDropNode::doInitOutput() {
    /* get input format */
    if (!getInputFormat()) {
      GLOBAL_ERROR_STREAM("No Input Format! FAILURE" << endl);
      return FAILURE;
    }

    Format* i_format = new Format(*getInputFormat());
    getOwnWorkingInputProperty()->addNewFormat(i_format);
    getOwnWorkingInputProperty()->setDefaultFormat(i_format);
    
    Format* o_format = new Format(*i_format);
    getOwnWorkingOutputProperty()->addNewFormat(o_format);
    getOwnWorkingOutputProperty()->setDefaultFormat(o_format);
    
    return SUCCESS;
  }
  
  Message* BufferDropNode::processBuffer(Buffer *in) 
  {
    if(buffer_to_drop > 0) { 
      ++buffer_count;
      if(buffer_count == buffer_to_drop) {
	buffer_count = 0;
	in->release();
	return 0;
      }
    }

    else if (drop_num > 0) {
      ++buffer_count;
      
      int64_t old_integer_part = ratio_counter.getNumerator() / ratio_counter.getDenominator();
      ratio_counter += drop_ratio;
      int64_t new_integer_part = ratio_counter.getNumerator() / ratio_counter.getDenominator();
      if (old_integer_part < new_integer_part) {
        DEBUG_STREAM("DropDecision: " << buffer_count << "." << " Dropping!" << endl);
        in->release();
        in = NULL;
      } else {
        DEBUG_STREAM("DropDecision: " << buffer_count << "." << " Leaving!" << endl);
      }

      if (buffer_count == drop_denom) {
        buffer_count = 0;
      }
    }

    return in;
  }

  void BufferDropNode::setDropRatio(const nmm_uint32& numerator, const nmm_uint32& denominator) {
    //if ((denominator == 0) || (numerator == 0) || (numerator > denominator)) {
    if ((denominator == 0) || (numerator > denominator)) {
      ERROR_STREAM("setDropRatio() Invalid Args: " << numerator << " / " << denominator << endl);
      return;
    }

    if (buffer_to_drop > 0) {
      buffer_to_drop = 0;
    }

    drop_num = numerator;
    drop_denom = denominator;

    drop_ratio = Rational(drop_num, drop_denom);
    drop_ratio.reduce();

    buffer_count = 0;
  }
}

#include "nmm/nmm_suffix.hpp"
