/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "BufferShapingNode.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"
#include <unistd.h>
#include <iostream>

NMM_REGISTER_PLUGIN(BufferShapingNode, NMM, "BufferShapingNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  BufferShapingNode::BufferShapingNode(const char* name,
                 const StreamQueue::Mode queue_mode,
                 const int size)
    : GenericFilterNode(name, queue_mode, size),
      frame_size(1024),     // TODO: Fix problem: if(frame_size > in_buffer_>getUsed())
      incomplete_out_buffer(0)
  {
  }
  
  void BufferShapingNode::setBufferSize(const nmm_int32& value)
  {
    frame_size = value;
  }

  nmm_int32 BufferShapingNode::getBufferSize()
  {
    return frame_size;
  }

  Result BufferShapingNode::doInit() 
  {
    Format* input_format = new Format(Format::wild_wild_type);
    getOwnInputProperty()->addNewFormat(input_format);
    getOwnInputProperty()->setDefaultFormat(input_format);

    Format* output_format = new Format(Format::wild_wild_type);
    getOwnOutputProperty()->addNewFormat(output_format);
    getOwnOutputProperty()->setDefaultFormat(output_format);

    input_format->setIOPartner(output_format);
    output_format->setIOPartner(input_format);

    return SUCCESS;
  }
  
  Result BufferShapingNode::doInitOutput() 
  {
    const Format* in_format = getInputFormat();
    
    Format* input_format = new Format(*in_format);
    getOwnWorkingInputProperty()->addNewFormat(input_format);
    getOwnWorkingInputProperty()->setDefaultFormat(input_format);

    Format* output_format = new Format(*in_format);
    getOwnWorkingOutputProperty()->addNewFormat(output_format);
    getOwnWorkingOutputProperty()->setDefaultFormat(output_format);

    input_format->setIOPartner(output_format);
    output_format->setIOPartner(input_format);
    
    return SUCCESS;
  }
  
  Message* BufferShapingNode::processBuffer(Buffer *in_buffer) 
  {
    if(getWorkingFlag()) {
      DEBUG_STREAM("processBuffer(working-flag true) with " << complete_out_buffers.size() << " complete_out_buffers " << endl);
      if(complete_out_buffers.size()) {
	Buffer* out_buffer = complete_out_buffers.front();
	complete_out_buffers.pop();
	out_buffer->setTimestamp(current_ts);
	// cerr << "Shaping " << out_buffer->getUsed() << endl;
	return out_buffer;
      }
      else {
	setWorkingFlag(false);
	// cerr << "Shaping return 0; (1)" << endl;
	return 0;
      }
    }
    else {
      DEBUG_STREAM("processBuffer(working-flag false) with in_buffer used " << in_buffer->getUsed() << endl);

      current_ts = in_buffer->getTimestamp();

      // if incomplete out buffers exist -> fill rest of out buffer with in buffer
      if(incomplete_out_buffer) {
	// in buffer has too much data -> fill current incomplete_out_buffer and newly created (in)complete out buffer(s)
	if(in_buffer->getUsed() > (frame_size - incomplete_out_buffer->getUsed())) {
	  // compute size of data to be copied ( = offset of in buffer for newly created out buffers )
	  size_t offset_in_buffer = frame_size - incomplete_out_buffer->getUsed();
	  memcpy( incomplete_out_buffer->getData() + incomplete_out_buffer->getUsed(), in_buffer->getData(), offset_in_buffer );

	  // incomplete out buffer is now complete and can be returned
	  incomplete_out_buffer->setUsed( frame_size );
	  complete_out_buffers.push(incomplete_out_buffer);  // Buffer* complete_out_buffer = incomplete_out_buffer;
	  setWorkingFlag(true);
	  DEBUG_STREAM("INC YES : incomplete_buffer complete used " << incomplete_out_buffer->getUsed() << " with in_buffer used " << in_buffer->getUsed() << " - " << offset_in_buffer  << " more? " << (offset_in_buffer < in_buffer->getUsed()) << " offset in_buffer : " << offset_in_buffer << endl);
	  incomplete_out_buffer = 0;

	  // while still data in in buffer available -> create new out buffers
	  while(offset_in_buffer < in_buffer->getUsed()) {
	    size_t data_to_copy_from_in_buffer = 0;
	    if( (in_buffer->getUsed() - offset_in_buffer) > frame_size ) {
	      data_to_copy_from_in_buffer = frame_size;
	    }
	    else {
	      data_to_copy_from_in_buffer = in_buffer->getUsed() - offset_in_buffer;
	    }

	    Buffer* new_out_buffer = getNewBuffer(frame_size);
	    memcpy(new_out_buffer->getData(), in_buffer->getData() + offset_in_buffer, data_to_copy_from_in_buffer);
	    new_out_buffer->setUsed(data_to_copy_from_in_buffer);

	    if(data_to_copy_from_in_buffer == frame_size) { // filled new out buffer completely -> return in next iteration from complete out buffers
	      DEBUG_STREAM("INC YES : new complete_buffer" << endl);
	      complete_out_buffers.push(new_out_buffer);
	      offset_in_buffer += data_to_copy_from_in_buffer;
	    }
	    else { // not completed out buffer but in buffer is empty now -> need to get another in buffer 
	      incomplete_out_buffer = new_out_buffer;
	      offset_in_buffer += data_to_copy_from_in_buffer;
	      DEBUG_STREAM("INC YES : new incomplete_buffer used " << incomplete_out_buffer->getUsed() << " with in_buffer used " << in_buffer->getUsed() << " - " << data_to_copy_from_in_buffer << " more? " << (offset_in_buffer < in_buffer->getUsed()) << " offset in_buffer : " << offset_in_buffer << endl);
	      break;
	    }
	  }

	  // release completely emptied in buffer, then return old incomplete in buffer which is now really complete
	  in_buffer->release();
	  // cerr << "Shaping return 0; (2)" << endl;
	  return 0; //complete_out_buffer;
	} // end: in buffer has too much data
	else { // in buffer has just enough or not enough data -> fill incomplete out buffer with complete in buffer
	  memcpy( incomplete_out_buffer->getData() + incomplete_out_buffer->getUsed(), in_buffer->getData(), in_buffer->getUsed() );
	  incomplete_out_buffer->setUsed( incomplete_out_buffer->getUsed() +  in_buffer->getUsed() );
	  in_buffer->release();
	  if(incomplete_out_buffer->getUsed() == frame_size) { // we were lucky: incomplete out buffer is now filled -> return it
	    DEBUG_STREAM("INC YES : incomplete_buffer -> complete" << endl);
	    complete_out_buffers.push(incomplete_out_buffer);
	    incomplete_out_buffer = 0;
	    // cerr << "Shaping return 0; (3)" << endl;
	    return 0;
	  }
	  else { // not completed incomplete out buffer but in buffer is empty now -> need to get another in buffer 
	    DEBUG_STREAM("INC YES : incomplete_buffer -> incomplete" << endl);
	    // cerr << "Shaping return 0; (4)" << endl;
	    return 0;
	  }
	} // end: in buffer has just enough or not enough data
      } // end:  if(incomplete_out_buffer)
      else { // no incomplete out buffer -> create new (in)complete out buffer(s)

	if(in_buffer->getUsed() == frame_size) { // we are very lucky -> simply forward in buffer
	  DEBUG_STREAM("INC NO  : in buffer matches" << endl);
	  // cerr << "Shaping2 " << in_buffer->getUsed() << endl;
	  return in_buffer;
	}
	else { // manual handling of in buffer
	  size_t offset_in_buffer = 0;
	  // while still data in in buffer available -> create new out buffers
	  while(offset_in_buffer < in_buffer->getUsed()) {
	    size_t data_to_copy_from_in_buffer = 0;
	    if( (in_buffer->getUsed() - offset_in_buffer) > frame_size ) {
	      data_to_copy_from_in_buffer = frame_size;
	    }
	    else {
	      data_to_copy_from_in_buffer = in_buffer->getUsed() - offset_in_buffer;
	    }
	    
	    Buffer* new_out_buffer = getNewBuffer(frame_size);
	    memcpy(new_out_buffer->getData(), in_buffer->getData() + offset_in_buffer, data_to_copy_from_in_buffer);
	    new_out_buffer->setUsed(data_to_copy_from_in_buffer);
	    
	    if(data_to_copy_from_in_buffer == frame_size) { // filled new out buffer completely -> return in next iteration from complete out buffers
	      complete_out_buffers.push(new_out_buffer);
	      setWorkingFlag(true);
	      offset_in_buffer += data_to_copy_from_in_buffer;
	      DEBUG_STREAM("INC NO  : new complete_buffer" << endl);
	    }
	    else { // not completed out buffer but in buffer is empty now -> need to get another in buffer 
	      incomplete_out_buffer = new_out_buffer;
	      offset_in_buffer += data_to_copy_from_in_buffer;
	      DEBUG_STREAM("INC NO  : new incomplete_buffer used " << incomplete_out_buffer->getUsed() << " with in_buffer used " << in_buffer->getUsed() << " - " << data_to_copy_from_in_buffer << " more? " << (offset_in_buffer < in_buffer->getUsed()) << " offset in_buffer : " << offset_in_buffer << endl);
	      break;
	    }
	  }
	  
	  // release completely emptied in buffer, then return 0 (if working-flag was enabled, out buffers will be sent in next iteration)
	  in_buffer->release();
	  // cerr << "Shaping return 0; (5)" << endl;
	  return 0;
	} // end: manual handling of in buffer
      }      
    } // end: else (working-flag is false)
  }
  
}

#include "nmm/nmm_suffix.hpp"
