/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "GenericFireWireNode.hpp"
#include <iostream>
#include "nmm/comm/messaging/EventDispatcher.hpp"
#include "nmm/comm/format/Property.hpp"
#include "nmm/utils/thread/MutexGuard.hpp"
#include <l1394/l1394_session.h>


#include "nmm/nmm_prefix.hpp"

namespace NMM{


GenericFireWireNode::GenericFireWireNode(const char* name /* = "GenerocFireWireNode" */,
					 StreamQueue::Mode mode /* = StreamQueue::MODE_SUSPEND */,
					 int queue_size /* = StreamQueue::DEFAULT_SIZE */)
  : GenericSourceNode(name, mode, queue_size) {
  DEBUG_STREAM( "Create GenericFireWire object" << endl);

  test_count = 1;
  max_buffer = 20;
  number     = 0;

  setBufferManager(0);

  buffer_count = 0;

  timer = new StreamTimer(this, StreamTimer::REAL_TIME);
  DEBUG_STREAM( "Constructed" << endl);
}

GenericFireWireNode::~GenericFireWireNode() {
  DEBUG_STREAM("Destroy GenericFireWire object"<< endl);

  mutex.lock();
  for ( list<L1394::Frame*>::iterator it = frame_list.begin(); it!= frame_list.end(); it++)
    device->releaseFrame(*it);
  bufs.erase( bufs.begin(), bufs.end() );
  mutex.unlock();

  DEBUG_STREAM("< Destroy GenericFireWireNode object"<< endl);
}

void GenericFireWireNode::erase() const {

}

Result GenericFireWireNode::doInit() {
  if (device == 0) {
    ERROR_STREAM("No valid device set" << endl);
    return FAILURE;
  }
  initOutputFormat(*getOwnOutputProperty(), *device);
  return SUCCESS;
}

Result GenericFireWireNode::doInitOutput() {
  if (device == 0) {
    ERROR_STREAM("No valid device set" << endl);
    return FAILURE;
  }
  initOutputFormat(*getOwnWorkingOutputProperty(), *device);
  return SUCCESS;
}

Result GenericFireWireNode::doActivate() {
  //init the device with default values
  const Format* f = getOutputFormat();

  if (f == 0) {
    ERROR_STREAM( "Final Outputformat is NULL" << endl);
    return FAILURE;
  }

  float frame_rate = 0;
  int x_resolution = 0;
  int y_resolution = 0;
  L1394::Videomode  resolution       ;
  L1394::Framerate  l1394_frame_rate ;
  string colorspace;
/*Begin> Read the camera_parameter from final output-format*/
  frame_rate    = f->getFloatValue(Format::framerate_param);
  x_resolution  = f->getIntValue(Format::x_resolution_param);
  y_resolution  = f->getIntValue(Format::y_resolution_param);
  colorspace    = f->getStringValue(Format::colorspace_param);
/*End> Read the camera_parameter from final output-format*/

/*Begin> Configure the l1394 library */
  switch(x_resolution)
  {
    case 160:
      if (y_resolution != 120)
        return FAILURE;
      if (colorspace == "yuv444") {
        resolution = L1394::DEVICE_160x120_YUV_444;
        break;
      }
      ERROR_STREAM( "Unsupported colorspace" << colorspace << endl);
      return FAILURE;
      break;
    case 320:
      if (y_resolution != 240)
        return FAILURE;

      if (colorspace == "yuv422") {
        resolution = L1394::DEVICE_320x240_YUV_422;
        break;
      }
      ERROR_STREAM( "Unsupported colorspace " << colorspace << endl);
      return FAILURE;
      break;

    case 640:
      if (y_resolution != 480)
        return FAILURE;

      if (colorspace == "yuv422") {
        resolution = L1394::DEVICE_640x480_YUV_422;
        break;
      }

      if (colorspace == "yuv411") {
        resolution = L1394::DEVICE_640x480_YUV_411;
        break;
      }
      if (colorspace == "rgb24")  {
        resolution = L1394::DEVICE_640x480_RGB;
        break;
      }

      if (colorspace == "mono")   {
        resolution = L1394::DEVICE_640x480_MONO;
        break;
      }

      ERROR_STREAM( "Unsupported colorspace " << colorspace << endl);
      return FAILURE;
      break;

    case 800:
      if (y_resolution != 600)
        return FAILURE;
      if (colorspace == "yuv422")  {
        resolution = L1394::DEVICE_800x600_YUV_422;
        break;
      }
      if (colorspace == "rgb24")   {
        resolution = L1394::DEVICE_800x600_RGB;
        break;
      }

      if (colorspace == "mono")    {
        resolution = L1394::DEVICE_800x600_MONO;
        break;
      }

      ERROR_STREAM( "Unsupported colorspace " << colorspace << endl);
      return FAILURE;
      break;

    case 1024:
      if (y_resolution != 768)
        return FAILURE;
      if (colorspace == "yuv422")  {
        resolution = L1394::DEVICE_1024x768_YUV_422;
        break;
      }
      if (colorspace == "rgb24")   {
        resolution = L1394::DEVICE_1024x768_RGB;
        break;
      }

      if (colorspace == "mono")    {
        resolution = L1394::DEVICE_1024x768_MONO;
        break;
      }

      ERROR_STREAM( "Unsupported colorspace " << colorspace << endl);
      return FAILURE;
      break;
    case 1280:
      if (y_resolution != 960)
        return FAILURE;
      if (colorspace == "yuv422")  {
        resolution = L1394::DEVICE_1280x960_YUV_422;
        break;
      }
      if (colorspace == "rgb24")   {
        resolution = L1394::DEVICE_1280x960_RGB;
        break;
      }

      if (colorspace == "mono")    {
        resolution = L1394::DEVICE_1280x960_MONO;
        break;
      }

      ERROR_STREAM( "Unsupported colorspace " << colorspace << endl);
      return FAILURE;
      break;
    case 1600:
      if (y_resolution != 1200)
        return FAILURE;
      if (colorspace == "yuv422")  {
        resolution = L1394::DEVICE_1600x1200_YUV_422;
        break;
      }
      if (colorspace == "rgb24")   {
        resolution = L1394::DEVICE_1600x1200_RGB;
        break;
      }

      if (colorspace == "mono")    {
        resolution = L1394::DEVICE_1600x1200_MONO;
        break;
      }

      ERROR_STREAM( "Unsupported colorspace " << colorspace << endl);
      return FAILURE;
      break;

    default:
      ERROR_STREAM( "Unsupported x-resolution " <<  x_resolution << endl);
      return FAILURE;
      break;
  }

  if(frame_rate==3.75)
    l1394_frame_rate = L1394::DEVICE_FRAMES_3_75;
  if(frame_rate==7.5)
    l1394_frame_rate = L1394::DEVICE_FRAMES_7_5;
  if(frame_rate==15.0)
    l1394_frame_rate = L1394::DEVICE_FRAMES_15;
  if(frame_rate==30.0)
    l1394_frame_rate = L1394::DEVICE_FRAMES_30;
  if(frame_rate==60.0)
    l1394_frame_rate = L1394::DEVICE_FRAMES_60;
  if(frame_rate==0.0) {
    ERROR_STREAM( "Unsupported framerate " << frame_rate << endl);
    return FAILURE;
  }

  ++channel;

  device->setParameter(max_buffer, channel, L1394::DEVICE_ISO_RUN, resolution, l1394_frame_rate);

  device->init();
  device->setFrameMode(false);
/*End> Configure the l1394 library */

  timer->reset();

  return SUCCESS;
}

Message* GenericFireWireNode::processBuffer(Buffer*)
{
    DEBUG_STREAM( "produce buffer " << endl);
    test_count++;

    L1394::Frame* frame = device->getFrame();
    if (frame == 0) {
      ERROR_STREAM2("Received no frame from libl1394" << std::endl);
      return 0;
    }

    Buffer *buffer = requestBuffer(device->getFrameSize());
    if (buffer) {
      buffer->setData(device->getFrameSize(), frame->getBuffer());
      buffer->setUsed(device->getFrameSize());

      timer->setTimestamp(buffer);
      mutex.lock();
      frame_list.insert(frame_list.begin(), frame);
      mutex.unlock();
    } else {
     suspendThread(10);
    }
    return buffer;
}


Result GenericFireWireNode::doStart()
{
  if (device->start() == L1394_SUCCESS)
    return SUCCESS;

  return FAILURE;
}

Result GenericFireWireNode::doStop(){
  if(device->stop() == L1394_SUCCESS)
    return SUCCESS;

  return FAILURE;
}


void GenericFireWireNode::initOutputFormat(Property& p, L1394::Device& device)
{
  if (device.hasVideoFormat(L1394::DEVICE_DV) ) {
    GLOBAL_ERROR_STREAM2("Not implemented" << std::endl);
    return;
  }

  Format* o_format = 0;

  for (L1394::Videomode i = L1394::DEVICE_160x120_YUV_444; i < L1394::DEVICE_DV; ++(* reinterpret_cast<int*>(&i)))
  {
    if (device.hasVideoFormat(i))  {
      o_format = p.addNewFormat("video/raw");
      createOutputFormat(device, i, o_format);
    }
  }

  if (o_format)
    p.setDefaultFormat(o_format); //The last Format is the 'best' foramt
  else
    GLOBAL_ERROR_STREAM2("No default format available " << std::endl);
}

void GenericFireWireNode::createOutputFormat(L1394::Device& device, L1394::Videomode video_mode, Format* o_format)
{
  int x_resolution = 0;
  int y_resolution = 0;
  int bit_per_pixel = 16;
  string channelorder = "uyvy";
  string colorspace ;
  switch (video_mode) {
    case L1394::DEVICE_160x120_YUV_444 :
      x_resolution = 160;
      y_resolution = 120;
      colorspace = "yuv444";
      break;
    case L1394::DEVICE_320x240_YUV_422 :
      x_resolution = 320;
      y_resolution = 240;
      colorspace = "yuv422";
      break;
    case L1394::DEVICE_640x480_YUV_411 :
      x_resolution = 640;
      y_resolution = 480;
      colorspace = "yuv411";
      break;
    case L1394::DEVICE_640x480_YUV_422 :
      x_resolution = 640;
      y_resolution = 480;
      colorspace = "yuv422";
      break;
    case L1394::DEVICE_640x480_RGB :
      x_resolution = 640;
      y_resolution = 480;
      colorspace = "rgb24";
      bit_per_pixel = 24;
      channelorder = "rgb";
      break;
    case L1394::DEVICE_640x480_MONO :
      x_resolution = 640;
      y_resolution = 480;
      colorspace = "mono";
      bit_per_pixel = 8;
      //channelorder = "none";

    case L1394::DEVICE_800x600_YUV_422 :
      x_resolution = 800;
      y_resolution = 600;
      colorspace = "yuv422";
      bit_per_pixel = 16;
      break;

    case L1394::DEVICE_800x600_RGB :
      x_resolution = 800;
      y_resolution = 600;
      channelorder = "rgb";
      colorspace = "rgb24";
      bit_per_pixel = 24;
      break;

    case L1394::DEVICE_800x600_MONO :
      x_resolution = 800;
      y_resolution = 600;
      colorspace = "mono";
      bit_per_pixel = 8;
      break;

    case L1394::DEVICE_1024x768_YUV_422 :
      x_resolution = 1024;
      y_resolution = 768;
      colorspace = "yuv422";
      break;
    case L1394::DEVICE_1024x768_RGB :
      x_resolution = 1024;
      y_resolution = 768;
      colorspace = "rgb24";
      channelorder = "rgb";
      bit_per_pixel = 24;
      break;
    case L1394::DEVICE_1024x768_MONO :
      x_resolution = 1024;
      y_resolution = 786;
      colorspace = "mono";
      bit_per_pixel = 8;
      break;
    case L1394::DEVICE_1280x960_YUV_422 :
      x_resolution = 1280;
      y_resolution = 960;
      colorspace = "yuv422";
      break;
    case L1394::DEVICE_1280x960_RGB :
      x_resolution = 1280;
      y_resolution = 960;
      colorspace = "rgb24";
      channelorder = "rgb";
      bit_per_pixel = 24;
      break;
    case L1394::DEVICE_1280x960_MONO :
      x_resolution = 1280;
      y_resolution = 960;
      colorspace = "mono";
      bit_per_pixel = 8;
      break;

    case L1394::DEVICE_1600x1200_YUV_422 :
      x_resolution = 1600;
      y_resolution = 1200;
      colorspace = "yuv422";
      break;
    case L1394::DEVICE_1600x1200_RGB :
      x_resolution = 1600;
      y_resolution = 1200;
      bit_per_pixel = 24;
      colorspace = "rgb24";
      channelorder = "rgb";
      break;
    case L1394::DEVICE_1600x1200_MONO :
      x_resolution = 1600;
      y_resolution = 1200;
      colorspace = "mono";
      bit_per_pixel = 8;
      break;
    default:
      GLOBAL_WARNING_STREAM("Camera format not implemented in CamerNode : " << video_mode << endl);
      return;
      break;
  }

  o_format->addIntParamValue(Format::x_resolution_param,x_resolution);
  o_format->addIntParamValue(Format::y_resolution_param,y_resolution);
  o_format->addIntParamValue(Format::ratio_x_param, 4);
  o_format->addIntParamValue(Format::ratio_y_param, 3);
  o_format->addStringParamValue(Format::channelorder_param,channelorder);
  o_format->addStringParamValue(Format::endian_param, "little");
  o_format->addStringParamValue(Format::colorspace_param,colorspace);
  o_format->addIntParamValue(Format::bitperpixel_param,bit_per_pixel);
  o_format->addStringParamValue(Format::format_param, "packet");

  for (L1394::Framerate j = L1394::DEVICE_FRAMES_60; j>= L1394::DEVICE_FRAMES_3_75; --(*(reinterpret_cast<int*>(&j))))
  {
    if(device.hasFramerate(video_mode, j))
    {
      switch (j)
      {
        case L1394::DEVICE_FRAMES_3_75:
          o_format->addFloatParamValue(Format::framerate_param,3.75);
          break;

        case L1394::DEVICE_FRAMES_7_5:
          o_format->addFloatParamValue(Format::framerate_param,7.5);
          break;

        case L1394::DEVICE_FRAMES_15:
          o_format->addFloatParamValue(Format::framerate_param,15.0);
          break;
        case L1394::DEVICE_FRAMES_30:
          o_format->addFloatParamValue(Format::framerate_param,30.0);
          break;
        case L1394::DEVICE_FRAMES_60:
          o_format->addFloatParamValue(Format::framerate_param,60.0)          ;
          break;
        default:
          o_format->addFloatParamValue(Format::framerate_param,0.0);
          break;
      }
    }
  }
}

Buffer* GenericFireWireNode::requestBuffer (const size_t size) const
{
    bufset_t::iterator i;
    Buffer *b = 0;
    int new_size = size;

   MutexGuard m(mutex);
    if (size <= 0) new_size = 1;
    if ((i = bufs.upper_bound(new_size-1)) != bufs.end()) {
      b = (*i).second;
      bufs.erase(i);
      b->increaseCount(1);
    } else {
        if (!(number < max_buffer) )  {
          WARNING_STREAM( "Buffer-Allocator limit reached" << endl);
          return 0;
        }
      number++;
      b = new Buffer(size, 0 ,this); //create buffer with NULL-Ptr as data
      b->increaseCount(1);
    }

    return b;
}

void GenericFireWireNode::releaseBuffer (Buffer* buffer) const
{
  MutexGuard m(mutex);
  list<L1394::Frame*>::iterator tmp_iterator = frame_list.begin();
  for (tmp_iterator = frame_list.begin(); tmp_iterator != frame_list.end(); tmp_iterator++)
  {
    if( (*tmp_iterator)->getBuffer() == buffer->getData() )
    {
      bufs.insert(bufset_t::value_type(buffer->getSize(), buffer));
      device->releaseFrame(*tmp_iterator);
      frame_list.remove(*tmp_iterator);

      return;
    }
  }
  bufs.insert(bufset_t::value_type(buffer->getSize(), buffer));

  ERROR_STREAM("Invalid Buffer released" << endl);
}

void GenericFireWireNode::nodeDestroy  (const Node *)
{
  try{
    stop();
  } catch(const Exception& e) {
    ERROR_STREAM("Could not Stop node"<< endl);
    ERROR_STREAM(e << endl);
  } try {
    deactivate(); 
  } catch(const Exception& e) {
    ERROR_STREAM("Could not deactivate node"<< endl);
    ERROR_STREAM(e << endl);
  }
  try
  {
    deinit();
  } catch(const Exception& e) {
    ERROR_STREAM("Could not deinit node"<< endl);
    ERROR_STREAM(e << endl);
  }
}

Result GenericFireWireNode::getGuid(u_int64_t& guid) {
  if (device)
    guid = device->getParentNode()->getGuid();
  return SUCCESS;
}

Result GenericFireWireNode::getDevice(string& name) {
  if (device)
    name = device->getName();
  return SUCCESS;
}

void GenericFireWireNode::nodeDisabled (const Node *) {
  stop();
}

void GenericFireWireNode::nodeEnabled (const Node *) {
//  activated();
}

int GenericFireWireNode::channel = 0;

} //end namespace


#include "nmm/nmm_suffix.hpp"
