/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "NetSinkNode.hpp"

#include "nmm/comm/messaging/ITSOneWayBufferReceiver.hpp"
#include "nmm/comm/TransportStrategy.hpp"
#include "nmm/comm/TSFactory.hpp"
#include "nmm/comm/ConnectionManager.hpp"
#include "nmm/comm/FormatStrategyMap.hpp"

#include "nmm/utils/thread/MutexGuard.hpp"

// special includes for RTP
#include "nmm/comm/netstrategy/RTPAddress.hpp"
#include "nmm/comm/netstrategy/RTPSenderStrategy.hpp"
#include "nmm/comm/netstrategy/UnicastIPAddress.hpp"

#include "nmm/plugins/NMMRegisterPlugin.hpp"

#include "nmm/utils/time/Interval.hpp"

// #include "nmm/base/format/Property.hpp"
// #include "nmm/base/registry/Plugin.hpp"
// #include "nmm/base/ConnectionManager.hpp"
// #include "nmm/base/FormatStrategyMap.hpp"
// #include "nmm/base/TSFactory.hpp"

// #include "nmm/base/proxy/rtp/RTPSenderStrategy.hpp"

NMM_REGISTER_PLUGIN(NetSinkNode, NMM, "NetSinkNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  NetSinkNode::NetSinkNode(const char* name) : GenericSinkNode(name, StreamQueue::MODE_SUSPEND, 10), 
					       __transport_strategy(0),
                                               __buffer_receiver(0),
					       __internal_cond(__internal_mutex),
					       __payload_type(0),
					       __current_address(0),
                                               __first_buffer(false)
  {}
  
  NetSinkNode::~NetSinkNode() {
    if (__transport_strategy) {
      __transport_strategy->scheduleDestruction();
      __transport_strategy = 0;
    }
  }

  void NetSinkNode::setRTPStrategy(const string& _address, const nmm_int32& _rtp_port, const nmm_int32& _rtcp_port) {
    TSAddress* address = new RTPAddress(UnicastIPAddress(_address), _rtp_port, _rtcp_port);
    setTransportStrategy(address);    
  }

  void NetSinkNode::setTransportStrategy(TransportStrategy* transport_strategy) {
    /* get interfaces */
    ITSOneWayBufferReceiver* buffer_receiver = dynamic_cast<ITSOneWayBufferReceiver*>(transport_strategy);
    if(!buffer_receiver) {
      // XXX: fw: should throw an exception
      ERROR_STREAM("Transport strategy is not a one-way buffer receiver" << endl);
      return;
    }
    
    /* Begin> If an old strategy is available, release it. */
    if (__transport_strategy) {
      ITSOneWayBufferReceiver* receiver = dynamic_cast<ITSOneWayBufferReceiver*>( __transport_strategy );
      if ( receiver ) {
        receiver->flushITSReceiver();
      }
      __transport_strategy->scheduleDestruction();
    }
    /* End< */
    
    /* stores the new strategy */
    __transport_strategy = transport_strategy;
    __buffer_receiver = buffer_receiver;
    __buffer_receiver->setSignalReceiver(new NetSignalReceiver(this));
  }
  
  void NetSinkNode::setTransportStrategy(const TSAddress* address) throw(const BindingException&) {
    __current_address = address->copy();
    /* If the address is invalid throw an Exception */
    if (!address)
      throw BindingException("Invalid Address (0-ptr)");

    /* Create a TransportStrategy which can connect to the given address */
    TransportStrategy* transport_strategy = STSFactory::getInstance().createTransportStrategy(address->getStrategyType(),
                                                                                              TSDescription::ITS_ONE_WAY_BUFFER_RECEIVER);
    
    /* If it could not be created throw a BindingException */
    if (!transport_strategy)
      throw  BindingException("Can not create transport strategy with type "+ address->getStrategyType());
    
    /* The strategy connects to the given address */
    transport_strategy->setAddress(*address);
    
    setTransportStrategy(transport_strategy);
  }
  
  ITransportStrategy* NetSinkNode::getTransportStrategy() const {
    if (__transport_strategy)
      return __transport_strategy->getInterface<ITransportStrategy>();
    
    ERROR_STREAM("__transport_strategy is NULL." << endl);
    return NULL;
  }

  Result NetSinkNode::enableInputStream() {
    /* If a FREE_QUEUE_SLOT signal is received the input stream will be enabled */
    MutexGuard mg(__internal_mutex);
    __internal_cond.notify();

    return SUCCESS;
  }

  Message* NetSinkNode::processBuffer(Buffer* buffer) {
    if (buffer) {
      if (buffer->getUsed() == 0) {
        WARNING_STREAM("Received buffer without data! Releasing it." << endl);
        buffer->release();
        return NULL; 
      } 
      __internal_mutex.lock();
      
      /* for checking AMR MODE
          
      u_int8_t* data_ptr = reinterpret_cast<u_int8_t *>(buffer->getData());
      
      u_int8_t frameHeader = *data_ptr;
      
      // for testing of amr mode change
      
      int amr_mode = -1;
              
      switch (frameHeader) {
      case 0x04: { amr_mode = 0; break; }
      case 0x0C: { amr_mode = 1; break; }
      case 0x14: { amr_mode = 2; break; }
      case 0x1C: { amr_mode = 3; break; }
      case 0x24: { amr_mode = 4; break; }
      case 0x2C: { amr_mode = 5; break; }
      case 0x34: { amr_mode = 6; break; }
      case 0x3C: { amr_mode = 7; break; }
      case 0x44: { amr_mode = 8; break; }
      default:   { cerr << "Unrecognized Mode" << endl; amr_mode = -1; }
      }
      
      if (amr_mode != -1) {
      
        timeval tim;
        gettimeofday(&tim, NULL);
        double t1 = tim.tv_sec+(tim.tv_usec/1000000.0);
        printf("%.6lf stop\n", t1);
        cerr << "AMR-Mode: " << amr_mode << "; Header: " << static_cast<int>(frameHeader) << endl;
      }
      
      */
      
      
    ////////// HACK ///////////
    /// Send Reports more often

    RTPStrategy* rtp_strategy = dynamic_cast<RTPStrategy*>(__transport_strategy);
           
    if (__first_buffer) {
      __first_buffer = false;
      __last_report_time = __clock.getTime();
    }
    
    if ( rtp_strategy) {
      
      //if (__clock.getTime() > Time(__last_report + Interval(0, 500000000L) ) ) {
      if (__clock.getTime() > Time(__last_report_time + Interval(1,0) ) ) {
        __last_report_time = __clock.getTime();
        rtp_strategy->sendRTCPReport();
      }
      
    } else {
      cerr << "*** No rtp_strategy" << endl;
    }
    
    //////////  HACK ///////////
      
      
      Result status = __buffer_receiver->receiveOneWayBuffer(buffer);
      if (status == BLOCKED) __internal_cond.wait();
      __internal_mutex.unlock();
    }
    return NULL;
  }

  Result NetSinkNode::doInit()  {
    string strategy_type;
    if (__transport_strategy) {
      const TSDescription& tsd = __transport_strategy->getTSDescription();

      list<Format*> formats = SConnectionManager::getInstance().getFormatStrategyMap().getSupportedFormats(tsd);
      list<Format*>::iterator i;

      // set static in and out formats in own property
      for (i  = formats.begin();
	   i != formats.end();
	   ++i) {
	Format* inputFormat = (*i)->duplicateFormat();
        
	getOwnInputProperty()->addNewFormat(inputFormat);
      }
      
    } else {
      WARNING_STREAM("No transport strategy set. Using default format \"*/*\"." << endl);
      Format* inputFormat = getOwnInputProperty()->addNewFormat("*/*");
      getOwnInputProperty()->setDefaultFormat(inputFormat);
    }

    return SUCCESS;
  }
  
  Result NetSinkNode::doInitOutput() {
    
    const Format* in_format = getInputFormat();
    if (in_format == NULL) {
      ERROR_STREAM("Error in doInitOutput: no in_format specified" << endl);
      return FAILURE;
    }
     
    if (__current_address && !__transport_strategy) {
       setTransportStrategy(__current_address);
    }

    
    if (!__transport_strategy) {
      WARNING_STREAM("No transport strategy set. You should set a transport strategy before using this node" << endl);
      return SUCCESS;
    }
    
    __buffer_receiver->configure(*in_format);
    
    if (__transport_strategy->getStrategyType() == "RTPStrategy") {
      RTPSenderStrategy* rtp_strategy = dynamic_cast<RTPSenderStrategy*>(__transport_strategy);
      __payload_type = (int)rtp_strategy->getPayloadType();
    }
    return SUCCESS;
  }

  void NetSinkNode::setPayloadType(const nmm_int32& type) {
    __payload_type = type;
    DEBUG_STREAM("Now using NetSinkNode with payload type " << __payload_type << endl);
  }

  nmm_int32 NetSinkNode::getPayloadType() const {
    return __payload_type;
  }

  Result NetSinkNode::doDeinitOutput() {
    if (__transport_strategy) {
       __transport_strategy->scheduleDestruction();
       __transport_strategy = 0;
    }

    return SUCCESS;
  }
  
  /*****************************************/
  /* methods from IRTPParameters interface */
  /*****************************************/

  string NetSinkNode::getRTPSinkMediaType() const{
    if (__transport_strategy && __transport_strategy->getStrategyType()== string("RTPStrategy")) {
      RTPSenderStrategy* rtp_strategy = dynamic_cast<RTPSenderStrategy*>(__transport_strategy);
      if (rtp_strategy) {
	return (string(rtp_strategy->getMediaType()));
      }
    }
    return string("");
  }
  
  nmm_uchar NetSinkNode::getRTPSinkPayloadType() const {
    if (__transport_strategy && __transport_strategy->getStrategyType()== string("RTPStrategy")) {
      RTPSenderStrategy* rtp_strategy = dynamic_cast<RTPSenderStrategy*>(__transport_strategy);
      if (rtp_strategy) {
	return (rtp_strategy->getPayloadType());
      }
    }
    return 0;
  }
  
  string NetSinkNode::getRTPSinkMapLine() const{
    if (__transport_strategy && __transport_strategy->getStrategyType()== string("RTPStrategy")) {
      RTPSenderStrategy* rtp_strategy = dynamic_cast<RTPSenderStrategy*>(__transport_strategy);
      if (rtp_strategy) {
	return (string(rtp_strategy->getMapLine()));
      }
    }
    return string("");
  }
  
  string NetSinkNode::getRTPSinkAuxSDPLine() const{
    if (__transport_strategy && __transport_strategy->getStrategyType()== string("RTPStrategy")) {
      RTPSenderStrategy* rtp_strategy = dynamic_cast<RTPSenderStrategy*>(__transport_strategy);
      if (rtp_strategy) {
	return (string(rtp_strategy->getAuxSDPLine()));
      }
    }
    return string("");
  }

  nmm_uint16 NetSinkNode::getRTPSeqNum() const {
    if (__transport_strategy && __transport_strategy->getStrategyType()== string("RTPStrategy")) {
      RTPSenderStrategy* rtp_strategy = dynamic_cast<RTPSenderStrategy*>(__transport_strategy);
      if (rtp_strategy) {
	return (rtp_strategy->getRTPSeqNum());
      }
    }
    return 0;
  }
    
  nmm_uint32 NetSinkNode::getRTPTimestamp() const {
    if (__transport_strategy && __transport_strategy->getStrategyType()== string("RTPStrategy")) {
      RTPSenderStrategy* rtp_strategy = dynamic_cast<RTPSenderStrategy*>(__transport_strategy);
      if (rtp_strategy) {
	return (rtp_strategy->getRTPTimestamp());
      }
    }
    return 0;
  }

  nmm_uint32 NetSinkNode::getRTPPort() const {
    if (__transport_strategy && __transport_strategy->getStrategyType()== string("RTPStrategy")) {
      RTPSenderStrategy* rtp_strategy = dynamic_cast<RTPSenderStrategy*>(__transport_strategy);
      if (rtp_strategy) {
	return (rtp_strategy->getRTPPort());
      }
    }
    return 0;
  }

  nmm_uint32 NetSinkNode::getRTCPPort() const {
    if (__transport_strategy && __transport_strategy->getStrategyType()== string("RTPStrategy")) {
      RTPSenderStrategy* rtp_strategy = dynamic_cast<RTPSenderStrategy*>(__transport_strategy);
      if (rtp_strategy) {
	return (rtp_strategy->getRTCPPort());
      }
    }
    return 0;
  }

  /***************************/
  /* Begin NetSignalReceiver */
  /***************************/
  
  NetSignalReceiver::NetSignalReceiver(NetSinkNode* _netsink) {
    __netsink = _netsink;
  }

  NetSignalReceiver::~NetSignalReceiver() {
  }
    
  void NetSignalReceiver::receiveSignal(Signal signal){
    
    switch (signal) {
    case FREE_QUEUE_SLOT:
      GLOBAL_DEBUG_STREAM2("NetSignalReceiver: Received signal FREE_QUEUE_SLOT from send queue" << endl);
      __netsink->enableInputStream();
      break;
    default:
      break;
    } 
  }
  
  Result NetSinkNode::setFormat(const Format& format) {
    //return GenericSinkNode::setFormat(format);
    if ( !format.isParameterIncluded(Format::framerate_param) ){
      cerr << "Invoke standard setFormat" << endl;
      return GenericSinkNode::setFormat(format);
    }

    Format current_input_format = *getInputFormat();
    Format new_input_format     = format;
    
    if (current_input_format == new_input_format) {
      DEBUG_STREAM ("setFormat: Nothing todo" << endl);
      return DELETE;
    } 
    
    /* Begin> Check if only framerate || bitrate has been changed */
    current_input_format.deleteParameter(Format::framerate_param);
    new_input_format.deleteParameter(Format::framerate_param);
    
    current_input_format.deleteParameter(Format::bitrate_param);
    new_input_format.deleteParameter(Format::bitrate_param);
    
    if ( current_input_format == new_input_format) {
      cerr << "Only framerate has been changed. Don't restart network connection" << endl;
      const_cast<Format*>(getInputFormat())->deleteParameter(Format::framerate_param);
      const_cast<Format*>(getInputFormat())->addFloatParamValue(Format::framerate_param, current_input_format.getFloatValue(Format::framerate_param));
      const_cast<Format*>(getInputFormat())->deleteParameter(Format::bitrate_param);
      const_cast<Format*>(getInputFormat())->addIntParamValue(Format::bitrate_param, current_input_format.getIntValue(Format::bitrate_param));
      return DELETE;
    }
    
    cerr << "Invoke standard setFormat" << endl;
    return GenericSinkNode::setFormat(format);
    
  }
} //end namespace

#include "nmm/nmm_suffix.hpp"
