/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "NetSourceNode.hpp"

#include "nmm/comm/messaging/ITSOneWayBufferSender.hpp"
#include "nmm/comm/TransportStrategy.hpp"
#include "nmm/comm/TSFactory.hpp"
#include "nmm/comm/ProxyObject.hpp"

// special includes for RTP
#include "nmm/comm/netstrategy/RTPAddress.hpp"
#include "nmm/comm/netstrategy/UnicastIPAddress.hpp"

// HACK: fw: Required for sending reports more often
// HACK: fw: If this hack becomes a feature, we should use a IRTPControl here...
#include "nmm/comm/netstrategy/RTPStrategy.hpp"

#include <iostream>
#include <iomanip>
#include <fstream>
#include <cassert>

#include "nmm/plugins/NMMRegisterPlugin.hpp"

#include "nmm/utils/thread/MutexGuard.hpp"

NMM_REGISTER_PLUGIN(NMM::NetSourceNode, NMM, "NetSourceNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  NetSourceNode::NetSourceNode(const char* name,
                               IQueued::Mode queue_mode /*= IQueued::MODE_SUSPEND*/,
                               unsigned int queue_size /*= 100*/)
    : GenericSourceNode(name), 
      __transport_strategy(0),
      __buffer_sender(0),
      __current_message(0),
      __stream_queue((StreamQueue::Mode) queue_mode, queue_size, "BufferedSource"),
      __payload_type(0),
      __first_buffer(false)
  {
    first = true;
    
    /* this object reads from the queue, so it is the parent of the queue */
    __stream_queue.setParentSignalReceiver(this);
  }

  NetSourceNode::~NetSourceNode() {
    if (__transport_strategy) {
      ITSOneWayBufferReceiver* receiver = dynamic_cast<ITSOneWayBufferReceiver*>( __transport_strategy );
      if ( receiver ) {
        receiver->flushITSReceiver();
      }
      __transport_strategy->scheduleDestruction();
    }
  }

  void NetSourceNode::setRTPStrategy(const string& _address, const nmm_int32& _rtp_port, const nmm_int32& _rtcp_port) {
    TSAddress* address = new RTPAddress(UnicastIPAddress(_address), _rtp_port, _rtcp_port);
    setTransportStrategy(address);    
  }

  void NetSourceNode::setTransportStrategy(TransportStrategy* transport_strategy) {
    ITSOneWayBufferSender* buffer_sender = dynamic_cast<ITSOneWayBufferSender*>(transport_strategy);
    if(!buffer_sender) {
      // XXX: fw: should throw an exception
      ERROR_STREAM("Transport strategy is not a one-way buffer sender" << endl);
      return;
    }
    
    /* Begin> If an old strategy is available, release it. */
    if (__transport_strategy) {
      ITSOneWayBufferReceiver* receiver = dynamic_cast<ITSOneWayBufferReceiver*>( __transport_strategy );
      if ( receiver ) {
        receiver->flushITSReceiver();
      }
      __transport_strategy->scheduleDestruction();
    }
    /* End< */
    
    /* stores the new strategy */
    __transport_strategy = transport_strategy;
    __buffer_sender = buffer_sender;
    /* End< */
    
    /* the transport strategy writes to the queue, so it is the partner of the queue */
    __stream_queue.setPartnerSignalReceiver(__buffer_sender);
    
    buffer_sender->setOneWayBufferReceiver(this);
    buffer_sender->setBufferManager(this);
    if(getOwnWorkingOutputProperty()->getDefaultFormat() ) {
      buffer_sender->configure(*getOwnWorkingOutputProperty()->getDefaultFormat());
    }
  }

  ITransportStrategy* NetSourceNode::getTransportStrategy() const {
    if(__transport_strategy) {
      return __transport_strategy->getInterface<ITransportStrategy>();
    }
    else {
      ERROR_STREAM("No transport strategy!" << endl);
      return 0;
    }
  }

  void NetSourceNode::setTransportStrategy(const TSAddress* address) throw(const BindingException&) {
    /* If the address is invalid throw an Exception */
    if (!address)
      throw BindingException("Invalid Address (0-ptr)");

    /* Create a TransportStrategy which can connect to the given address */
    TransportStrategy* transport_strategy = STSFactory::getInstance().createTransportStrategy(address->getStrategyType(),
                                                                                              TSDescription::ITS_ONE_WAY_BUFFER_SENDER);
    
    /* If it could not be created throw a BindingException */
    if (!transport_strategy)
      throw  BindingException("Can not create strategy with type "+ address->getStrategyType());

    /* The strategy connects to the given address */
    
    transport_strategy->setAddress(*address);
    
    setTransportStrategy(transport_strategy);
  }

  void NetSourceNode::setPayloadType(const nmm_int32& type) {
    __payload_type = type;
  }

  nmm_int32 NetSourceNode::getPayloadType() const {
    return __payload_type;
  }

  // long FIRST_BUFFER = -1;

  Message* NetSourceNode::processBuffer(Buffer* /*buffer*/) {

    if (!__current_message) {
      setProducingFlag(false);
      return NULL;
    }

    MutexGuard current_message_guard(__current_message_mutex);

    // HACK: fw: Hack merged from trunk, see below
    // Set all timestamps to 'invalid' because RTP sends invalid timestamps first and
    // PlaybackNode stops playing after first valid frame is sent
//    if(static_cast<Buffer*>(__current_message)->getTimestamp().is_valid == true ) {
//      Timestamp ts = static_cast<Buffer*>(__current_message)->getTimestamp();
//      ts.is_valid = false;
//      static_cast<Buffer*>(__current_message)->setTimestamp(ts);
//    }

    // Message to return
    Message* out_message = 0;

    // HACK: fw: Discard all buffers with invalid timestamps
    // HACK: fw: This is for RTP. All buffers before the first RTCP sync are discarded
    // HACK: fw: This hack avoids a jump in the sync_time around the RTP sync
    if(static_cast<Buffer*>(__current_message)->getTimestamp().is_valid == false) {
      // Discard message
      __current_message->release();
    }
    else {
      // Return message
      out_message = __current_message;
    }

    // Start producing messages again
    // (wait for receiving more messages)
    __current_message = 0;
    setProducingFlag(false);
    assert(__buffer_sender);
    __buffer_sender->receiveSignal(FREE_QUEUE_SLOT);

    return out_message;
  }

  Result NetSourceNode::receiveOneWayBuffer(Buffer* buffer){
    if (!buffer) {
      return FAILURE;
    }

    if (__current_message) {
      ERROR_STREAM( "ERROR IN NetSourceNode __current_message != 0" << endl);
      return FAILURE;
    }

    // HACK: fw: Send RTCP reports more often
    RTPStrategy* rtp_strategy = dynamic_cast<RTPStrategy*>(__transport_strategy);
    if (__first_buffer) {
      __first_buffer = false;
      __last_report_time = __clock.getTime();
    }
    if ( rtp_strategy) {
      // TEST: fw: Was out-commented in trunk
      //if (__clock.getTime() > Time(__last_report + Interval(0, 500000000L) ) ) {
      if (__clock.getTime() > Time(__last_report_time + Interval(1,0) ) ) {
        __last_report_time = __clock.getTime();
        rtp_strategy->sendRTCPReport();
      }
    }

    /* Begin> Store the message pointer */
    __current_message_mutex.lock();
    __current_message = buffer;
    setProducingFlag(true);
    if (__current_message)
      __current_message_mutex.unlock();
    /* End< */

    /* Return blocked*/
    return BLOCKED;
  }

  Result NetSourceNode::doInit()  {
    
    return SUCCESS;
  }
  
  Result NetSourceNode::doInitOutput() {
    
    Format* strategy_format = 0;
    DEBUG_STREAM("__payload_type = " << __payload_type << endl);
    switch (__payload_type) {
    case 0:
      strategy_format = getOwnWorkingOutputProperty()->addNewFormat("audio/raw");
      strategy_format->addIntParamValue(Format::sample_rate_param,8000);
      strategy_format->addIntParamValue(Format::channels_param, 1);
      strategy_format->addIntParamValue(Format::bitpersample_param, 8);
      strategy_format->addStringParamValue(Format::endian_param, "little");
      strategy_format->addStringParamValue(Format::encoding_param, "mu_law");
      strategy_format->addStringParamValue(Format::sign_param, "signed");
      getOwnWorkingOutputProperty()->setDefaultFormat(strategy_format);
      break;
    case 8:
      strategy_format = getOwnWorkingOutputProperty()->addNewFormat("audio/raw");
      strategy_format->addIntParamValue(Format::sample_rate_param,8000);
      strategy_format->addIntParamValue(Format::channels_param, 1);
      strategy_format->addIntParamValue(Format::bitpersample_param, 8);
      strategy_format->addStringParamValue(Format::endian_param, "little");
      strategy_format->addStringParamValue(Format::encoding_param, "a_law");
      strategy_format->addStringParamValue(Format::sign_param, "signed");
      getOwnWorkingOutputProperty()->setDefaultFormat(strategy_format);
      break;
    case 10: 
      strategy_format = getOwnWorkingOutputProperty()->addNewFormat(Format::audio_raw_type);
      strategy_format->addIntParamValue(Format::sample_rate_param, 44100);
      strategy_format->addIntParamValue(Format::channels_param, 2);
      strategy_format->addIntParamValue(Format::bitpersample_param,  16);
      strategy_format->addStringParamValue(Format::endian_param, "little");
      strategy_format->addStringParamValue(Format::encoding_param, "linear");
      strategy_format->addStringParamValue(Format::sign_param, "signed");
      getOwnWorkingOutputProperty()->setDefaultFormat(strategy_format);
      break;
    case 11:
      strategy_format = getOwnWorkingOutputProperty()->addNewFormat(Format::audio_raw_type);
      strategy_format->addIntParamValue(Format::sample_rate_param, 44100);
      strategy_format->addIntParamValue(Format::channels_param, 1);
      strategy_format->addIntParamValue(Format::bitpersample_param,  16);
      strategy_format->addStringParamValue(Format::endian_param, "little");
      strategy_format->addStringParamValue(Format::encoding_param, "linear");
      strategy_format->addStringParamValue(Format::sign_param, "signed");
      getOwnWorkingOutputProperty()->setDefaultFormat(strategy_format);
      break;
    case 14:
      strategy_format = getOwnWorkingOutputProperty()->addNewFormat(Format::audio_mpeg_type);
      strategy_format->addIntParamValue(Format::audio_codec_param, 0x55);
      strategy_format->addWildcard(Format::layer_param);
      strategy_format->addWildcard(Format::channels_param);
      strategy_format->addWildcard(Format::sample_rate_param);
      strategy_format->addIntParamValue(Format::bitpersample_param, 8);
      strategy_format->addIntParamValue(Format::bitpersample_param, 16);
      getOwnWorkingOutputProperty()->setDefaultFormat(strategy_format);
      break;
    case 26:
      strategy_format = getOwnWorkingOutputProperty()->addNewFormat(Format::video_jpeg_type);
      strategy_format->addWildcard(Format::x_resolution_param);
      strategy_format->addWildcard(Format::y_resolution_param);
      strategy_format->addWildcard(Format::ratio_x_param);
      strategy_format->addWildcard(Format::ratio_y_param);
      strategy_format->addWildcard(Format::framerate_param);
      getOwnWorkingOutputProperty()->setDefaultFormat(strategy_format);
      break;  
    case 32:
      /* simple MPEG2 video format */
      strategy_format = getOwnWorkingOutputProperty()->addNewFormat(Format::video_mpeg_type);
      strategy_format->addIntParamValue(Format::version_param, 1);
      strategy_format->addIntParamValue(Format::version_param, 2);
      getOwnWorkingOutputProperty()->setDefaultFormat(strategy_format);

      /* detailed MPEG2 video format (needed for analyzeData) */
      strategy_format = getOwnWorkingOutputProperty()->addNewFormat(Format::video_mpeg_type);
      strategy_format->addIntParamValue(Format::version_param, 1);
      strategy_format->addIntParamValue(Format::version_param, 2);
      strategy_format->addWildcard(Format::x_resolution_param);
      strategy_format->addWildcard(Format::y_resolution_param);
      strategy_format->addWildcard(Format::ratio_x_param);
      strategy_format->addWildcard(Format::ratio_y_param);
      strategy_format->addWildcard(Format::framerate_param);
      break;
    case 34:
      strategy_format = getOwnWorkingOutputProperty()->addNewFormat(Format::video_h263plus_type);
      strategy_format->addWildcard(Format::x_resolution_param);
      strategy_format->addWildcard(Format::y_resolution_param);
      strategy_format->addWildcard(Format::ratio_x_param);
      strategy_format->addWildcard(Format::ratio_y_param);
      strategy_format->addWildcard(Format::framerate_param);
      strategy_format->addWildcard(Format::bitrate_param);
      getOwnWorkingOutputProperty()->setDefaultFormat(strategy_format);
      break;
    case 97:
      strategy_format = getOwnWorkingOutputProperty()->addNewFormat(Format::audio_ac3_type);
      // FIXME: is this needed: strategy_format->addWildcard(Format::bitrate_param);
      getOwnWorkingOutputProperty()->setDefaultFormat(strategy_format);
      break;
    case 99: 
      strategy_format = getOwnWorkingOutputProperty()->addNewFormat(Format::audio_raw_type);
      strategy_format->addIntParamValue(Format::sample_rate_param, 48000);
      strategy_format->addIntParamValue(Format::channels_param, 2);
      strategy_format->addIntParamValue(Format::bitpersample_param,  24);
      strategy_format->addStringParamValue(Format::endian_param, "little");
      strategy_format->addStringParamValue(Format::encoding_param, "linear");
      strategy_format->addStringParamValue(Format::sign_param, "signed");
      getOwnWorkingOutputProperty()->setDefaultFormat(strategy_format);
      break;
    case 100:
      strategy_format = getOwnWorkingOutputProperty()->addNewFormat(Format::audio_amr_type);
      strategy_format->addIntParamValue(Format::channels_param, 1);
      strategy_format->addIntParamValue(Format::sample_rate_param, 8000);
      strategy_format->addIntParamValue(Format::bitpersample_param, 16);
      strategy_format->addIntParamValue(Format::bitrate_param, 4750);
      strategy_format->addIntParamValue(Format::bitrate_param, 5150);
      strategy_format->addIntParamValue(Format::bitrate_param, 5900);
      strategy_format->addIntParamValue(Format::bitrate_param, 6700);
      strategy_format->addIntParamValue(Format::bitrate_param, 7400);
      strategy_format->addIntParamValue(Format::bitrate_param, 7950);
      strategy_format->addIntParamValue(Format::bitrate_param, 10200);
      strategy_format->addIntParamValue(Format::bitrate_param, 12200);
      strategy_format->addStringParamValue(Format::encoding_param, "amr_nb");
      getOwnWorkingOutputProperty()->setDefaultFormat(strategy_format);
      break;
    case 101:
      strategy_format = getOwnWorkingOutputProperty()->addNewFormat(Format::audio_amr_type);
      strategy_format->addIntParamValue(Format::channels_param, 1);
      strategy_format->addIntParamValue(Format::sample_rate_param, 16000);
      strategy_format->addIntParamValue(Format::bitpersample_param, 16);
      strategy_format->addIntParamValue(Format::bitrate_param, 6600);
      strategy_format->addIntParamValue(Format::bitrate_param, 8850);
      strategy_format->addIntParamValue(Format::bitrate_param, 12650);
      strategy_format->addIntParamValue(Format::bitrate_param, 14250);
      strategy_format->addIntParamValue(Format::bitrate_param, 15850);
      strategy_format->addIntParamValue(Format::bitrate_param, 18250);
      strategy_format->addIntParamValue(Format::bitrate_param, 19850);
      strategy_format->addIntParamValue(Format::bitrate_param, 23050);
      strategy_format->addIntParamValue(Format::bitrate_param, 23850);

      strategy_format->addStringParamValue(Format::encoding_param, "amr_wb");
      getOwnWorkingOutputProperty()->setDefaultFormat(strategy_format);
      break;
    default:
      Format* out_format = getOwnWorkingOutputProperty()->addNewFormat("*/*");
      getOwnWorkingOutputProperty()->setDefaultFormat(out_format);
      ERROR_STREAM("Unknown payload type specified. Only types audio/raw (10/11), "
                   << "pcmu (0), audio/mpeg (14), video/jpeg (26), video/mpeg (32)"
                   << "and video/h263 (34)  are supported!"  
                   << endl << "Taking */* as default format." << endl);
      
      return FAILURE;
    }
    
    if(__buffer_sender) {
      cerr << "TransportStrategy available" << endl;
      __buffer_sender->setOneWayBufferReceiver(this);
      __buffer_sender->setBufferManager(this);
      __buffer_sender->configure(*strategy_format);
    }
    
    if(__payload_type==0 || __payload_type==8 || __payload_type==10|| __payload_type==11 ){
      string type = strategy_format->getFormatType();
      int channels = strategy_format->getIntValue(Format::channels_param);
      int bitpersample = strategy_format->getIntValue(Format::bitpersample_param);
      int sample_rate = strategy_format->getIntValue(Format::sample_rate_param);
      string encoding = strategy_format->getStringValue(Format::encoding_param);
      
      DEBUG_STREAM(endl << endl << "NetSourceNode OwnWorkingOutputProperty" << endl 
                   << "type = " << type << endl
                   <<  "encoding = " << encoding << endl 
                   << "channels = " << channels<< endl
                   << "bitpersample = " << bitpersample << endl
                   << "sample_rate = " << sample_rate << endl << endl);
    }
    
    return SUCCESS;

    /*
    const Format* strategy_format = getOwnWorkingOutputProperty()->getDefaultFormat();

    if (strategy_format) {

      __transport_strategy->setMessageReceiver(this);
      __transport_strategy->setBufferManager(this);
      __transport_strategy->configureForInputJack(*strategy_format);
    
      return SUCCESS;
    }
    return FAILURE;*/
  }
  
  void NetSourceNode::flushITSReceiver() {
    if ( __transport_strategy ) {
      ITSOneWayBufferReceiver* receiver = dynamic_cast<ITSOneWayBufferReceiver*>( __transport_strategy );
      receiver->flushITSReceiver();
    }
  }

  
  Result NetSourceNode::doActivate() {
    cerr << "DoActivate" << endl;
    setProducingFlag(true);
    cerr << "DoActivate" << endl;
    return SUCCESS;
  }

}

#include "nmm/nmm_suffix.hpp"
