/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "HTTPProtocolReader.hpp"

#include "nmm/multimedia/Node.hpp"

#include <netdb.h>       /* for gethostbyname() */
#include <sys/types.h>   /* for socket(), connect()               */
#include <sys/socket.h>  /* for socket(), connect(), inet_ntoa()  */
#include <errno.h>       /* for error description */
#include <netinet/in.h>  /* for inet_ntoa()       */
#include <arpa/inet.h>   /* for inet_ntoa()       */
#include <unistd.h>

namespace NMM {

  HTTPProtocolReader::HTTPProtocolReader(Node* _parent, const string _name)
  : ProtocolReader(_parent, _name), 
    cache_buffer_size(0),
    host(""), 
    port(0), 
    uri(""), 
    title(""),
    content_type(""),
    content_length(0) {
    PROTO_DEBUG_STREAM("HTTPPR Constructor!" << endl);
  }; /* HTTPProtocolReader */

  HTTPProtocolReader::~HTTPProtocolReader() {
    close(socketd);
  }; /* ~HTTPProtocolReader */

  bool HTTPProtocolReader::parseURL(const string _url) {
    PROTO_DEBUG_STREAM("Test for HTTP URL!" << endl);
    if ("http" != _url.substr(0,4)) {
      PROTO_ERROR_STREAM("No 'http' head in [" << _url << "] but [" << _url.substr(0,3) << "]!" << endl);
      return false;
    }
    if ("://" != _url.substr(4,3)) {
      PROTO_ERROR_STREAM("No '://' after head!" << endl);
      return false;
    }
    PROTO_DEBUG_STREAM("This is a HTTP URL!" << endl);

    const unsigned int hostpos = 7; /* url must start with "http://" */
    const unsigned int colonpos = _url.find(":",hostpos); /* find the colon */
    const unsigned int uripos = _url.find("/",hostpos); /* find the first "/" after "http://" */
    if (_url.npos == colonpos) {
      /* no colon in url, so no port specified, take default */
      host = _url.substr(hostpos, uripos-hostpos);
      port = 80;
    } else {
      int portpos = colonpos + 1;
      host = _url.substr(hostpos, colonpos-hostpos);
      port = atoi( (_url.substr(portpos, uripos-portpos)).c_str() );
    }
    if (_url.npos == uripos) {
      /* no uri specified, take default */
      uri = "/";
    } else {
      uri  =  _url.substr(uripos);
    }
    
    if (("" == host) || (0 == port) || ("" == uri)) {
      /* parse problem */
      return false;
    }

    PROTO_DEBUG_STREAM("And it is valid: http://[" << host << "]:[" << port << "][" << uri << "]!" << endl);

    url = _url;
    return true;
  }; /* setURL */

  bool HTTPProtocolReader::prepareSocket() {
    /* get a host pointer */
    struct hostent* host_ptr;        /* host information structure */
    host_ptr = gethostbyname( host.c_str() );
    if (NULL == host_ptr) {
      PROTO_ERROR_STREAM("HTTPProtocolReader: Error looking up IP for host (" 
			  << h_errno << ")!" 
			  << endl); 
      return false;
    }
    PROTO_DEBUG_STREAM("Successful looked up hostname!" << endl);

    /* Fill in the address structure for the host */
    struct sockaddr_in host_addr;        /* host address descriptor    */
    host_addr.sin_family = AF_INET;      /* Address family is Internet */
    host_addr.sin_port   = htons(port);  /* Port number */
    memcpy(&host_addr.sin_addr, host_ptr->h_addr, host_ptr->h_length ); /* IP address */
    for(unsigned int i=0; i < 8; i++) {
      host_addr.sin_zero[i] = 0;
    }
    PROTO_DEBUG_STREAM("Successful constructed host socket struct with IP: [" <<  inet_ntoa(host_addr.sin_addr) << "]!" << endl);

    /* create the socket */
    socketd = socket( PF_INET, SOCK_STREAM, IPPROTO_TCP );
    if ( 0 > socketd ) {
      PROTO_ERROR_STREAM("HTTPProtocolReader: Error creating socket (" 
			  << errno 
			  << ": " << strerror(errno) << ")!"
			  << endl);
      return false;
    }
    PROTO_DEBUG_STREAM("Successful created socket!" << endl);

    /* connect to host */
    if ( 0 > connect( socketd, (struct sockaddr*) &host_addr, sizeof(host_addr) ) ) {
      PROTO_ERROR_STREAM("HTTPProtocolReader: Error connecting server ("
			  << errno 
			  << ": " << strerror(errno) << ")!"
			  << endl);
      close(socketd);
      return false;
    }
    PROTO_DEBUG_STREAM("Successful connected to host!" << endl);

    return true;
  }

  bool HTTPProtocolReader::writeSocket(string data) {
    char* buffer = new char[data.length()+1];

    /* copy data to buffer */
    for(unsigned int i=0; i < data.length(); i++) {
      buffer[i] = data[i]; 
    }
    buffer[data.length()] = '\n';

    /* Send buffer to data */
    if (0 > write(socketd, buffer, data.length())) {
      PROTO_ERROR_STREAM("HTTPProtocolReader: Error during send data." << endl);
      return false;
    }
    
    return true;
  } /* WriteSocket */

  string HTTPProtocolReader::computeRequest() {
    return "\
GET " + uri + " HTTP/1.1\n\
HOST: " + host + "\n\
\n\
";
  }

  bool HTTPProtocolReader::parseHeader(const string data, const string header, int& value) {
    string value_str = "";
    bool retval = parseHeader(data, header, value_str);
    value = atoi( value_str.c_str() );
    return retval;
  }

  bool HTTPProtocolReader::parseHeader(const string data, const string header, string& value) {
    unsigned int start = data.find(header);
    if (data.npos == start) {                         /* header field not found  */
      return false;
    }

    start += header.length();                       /* shift start after it    */
    int end = data.find("\n", start);
    if ('\r' == data[end-2]) { /* remove carriage return */
      value = data.substr(start, end - start - 2);
    } else {
      value = data.substr(start, end - start - 1);
    }

    return true;
  }

  bool HTTPProtocolReader::init() {
    PROTO_DEBUG_STREAM("Begin initialization!" << endl);
    if (! prepareSocket()) {
      return false;
    }

    PROTO_DEBUG_STREAM("Now starting communication!" << endl);
    if (! writeSocket(computeRequest())) {
      return false;
    }

    string answer = "";
    char* answer_ptr = NULL;
    int answer_size = 0;
    while (answer.npos == answer.find("\n\r\n")) { /* while no double-newline (== in header) */
      char* buf = new char[2048];
      int s = read(socketd, buf, 2048);
      if (0 > s) {
	PROTO_ERROR_STREAM("Read failed!" << endl);
	return false;
      }
      char* tmp = new char[s + answer_size];
      if (0 < answer_size) {
	memcpy(tmp, answer_ptr, answer_size);
	delete[] answer_ptr;
      }
      memcpy(tmp + answer_size, buf, s);
      delete[] buf;
      answer_ptr = tmp;
      answer_size += s;
      answer = string(answer_ptr, answer_size);
    }
    string header = answer.substr(0, answer.find("\n\r\n") + 3);
    cache_buffer = &answer_ptr[header.length()];
    cache_buffer_ptr = answer_ptr;
    cache_buffer_size = answer_size - header.length();
    string content = string(cache_buffer, cache_buffer_size);
    PROTO_DEBUG_STREAM("Read:" << endl <<
		       "========================================================" << endl << 
		       answer << endl <<
		       "========================================================" << endl << 
		       "========================================================" << endl << 
		       header << endl <<
		       "========================================================" << endl << 
		       content << endl <<
		       "========================================================" << endl);
    


    if ("ICY" == header.substr(0,3)) { /* this is an icecast server */
      content_type = Format::audio_mpeg_type;  /* icecast can only stream mp3! */
      content_length = 0;              /* streaming is endless! */
      if (! parseHeader(header, "icy-name:", title)) {
	title = uri;
      }
      PROTO_DEBUG_STREAM("Title of this SHOUTCast stream is [" << title << "]!" << endl);

    } else if ("HTTP" == header.substr(0,4)) { /* ok, a web server */
      if (string("404") == header.substr(9,3)) {
        PROTO_ERROR_STREAM("WebServer said: File not found!" << endl);
        return false;
      }
      title = uri;     /* no more meta data about the title of the data in the meta data */
      if (! parseHeader(header, "Content-Type: ", content_type)) {
	content_type = "";
      }
      else {
	if(content_type.find("ogg") != string::npos) {
	  content_type = Format::audio_x_ogg_type;
	}
      }
      PROTO_DEBUG_STREAM("Content-Type of this HTTP stream is [" << content_type << "]!" << endl);
      if (! parseHeader(header, "Content-Length: ", content_length)) {
	content_length = 0;
      }
      PROTO_DEBUG_STREAM("Content-Length of this HTTP stream is [" << content_length << "]!" << endl);

    }

    return true;
  }

  string HTTPProtocolReader::getMediaFormat() {
    return content_type;
  }

  int HTTPProtocolReader::readData(char* data, size_t amount) {
    if (eof()) {
      return 0;
    }

    char* read_point = data;
    unsigned int total_read = 0;
    if (0 < cache_buffer_size) {                        /* we still have some data from init    */
      PROTO_DEBUG_STREAM("There is still some content!" << endl);
      if (amount <= cache_buffer_size) {                /* more data left than to read          */
	PROTO_DEBUG_STREAM("And more than to read!" << endl);
	memcpy(read_point, cache_buffer, amount);            /* read part of the left data */
	cache_buffer = &cache_buffer[amount];          /* save the rest!                       */
	cache_buffer_size -= amount;
	return amount;
      } else {
	PROTO_DEBUG_STREAM("But not more than to read!" << endl);
	memcpy(read_point, cache_buffer, cache_buffer_size);   /* first get the left data              */
	amount -= cache_buffer_size;                      /* decrease the amount of data to read  */
	read_point = &read_point[cache_buffer_size];                  /* shift up the data pointer            */
	total_read = cache_buffer_size;
	cache_buffer_size = 0;
	if (cache_buffer_ptr) delete[] cache_buffer_ptr;
      }
    }

    int s = 0;
    while (2 * total_read < amount) {
      s  = read(socketd, read_point, amount);
      if (0 > s) {
	PROTO_ERROR_STREAM("Failed to read data!" << endl);
	return s;
      }
      if (0 == s) {
	PROTO_ERROR_STREAM("Failed to read on data!" << endl);
	is_eof = true;
	return total_read;
      }
      total_read += s;
      read_point = &read_point[s];
      amount -= s;
    }
    return total_read;
  }; /* readData */

}
