/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "URLNode.hpp"
#include "HTTPProtocolReader.hpp"

#include "nmm/comm/format/Property.hpp"
#include "nmm/multimedia/Plugin.hpp"
#include "nmm/multimedia/interfaces/ITrack.hpp"
#include "nmm/multimedia/interfaces/ITrackList.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"
#include "nmm/utils/thread/MutexGuard.hpp"

NMM_REGISTER_PLUGIN(NMM::URLNode, NMM, "URLNode")

namespace NMM {

  URLNode::URLNode(const char* name /* = "URLNode" */,
		   StreamQueue::Mode mode /* = StreamQueue::MODE_SUSPEND */,
		   int queue_size /* = StreamQueue::DEFAULT_SIZE */)
    : GenericSourceNode(name, mode, queue_size),
      event(START),
      event_mutex(),
      url(""),
      reader(NULL),
      buffer_size(32*1024) {
  }

  URLNode::~URLNode() {
    if (reader) {
      delete reader;
      reader = NULL;
    }
  }

  /* IURL interface */
  void URLNode::setURL(const string& _url) {
    url = _url;
  }

  string URLNode::getURL() const {
    return url;
  }


  void URLNode::setBufferSize(const nmm_int32& size) {
    buffer_size = size;
  }
  
  nmm_int32 URLNode::getBufferSize() {
    return buffer_size;
  }
  
  URLNode::URLProtocol URLNode::getProtocol(const string _url) {
    string proto = _url.substr(0,_url.find(":",0));
    DEBUG_STREAM("Found protocol string \"" << proto << "\"." << endl);
    if ("http" == proto) {
      return HTTP;
    } else if ("ftp" == proto) {
      return FTP;
    } else if ("smb" == proto) {
      return SMB;
    } else if ("fish" == proto) {
      return FISH;
    }

    return UNKNOWN;
  } /* getProtocol */


  Result URLNode::doInit() {
    Format* out_format = getOwnOutputProperty()->addNewFormat(Format::wild_wild_type);
    getOwnOutputProperty()->setDefaultFormat(out_format);

    return SUCCESS;
  }

  Result URLNode::doInitOutput() {
    switch (getProtocol(url)) {
    case HTTP:
      reader = new HTTPProtocolReader(this);
      break;
    case FTP:
      //      reader = new FTPProtocolReader();
      reader = NULL;
      ERROR_STREAM("FTP not yet implemented!" << endl);
      break;
    case SMB:
      //      reader = new SMBProtocolReader();
      reader = NULL;
      ERROR_STREAM("SMB not yet implemented!" << endl);
      break;
    case FISH:
      //      reader = new FISHProtocolReader();
      reader = NULL;
      ERROR_STREAM("FISH not yet implemented!" << endl);
      break;
    case UNKNOWN:
    default:
      reader = NULL;
      ERROR_STREAM("Protocol not recongnized/supported!" << endl);
      break;
    }

    if (!reader) {
      return FAILURE;
    }

    if (!reader->setURL(url)) {
      ERROR_STREAM("Unable to initialize reader with url " << url << endl);
      delete reader;
      reader = NULL;
      return FAILURE;
    }

    Format* out_format = getOwnWorkingOutputProperty()->addNewFormat(reader->getMediaFormat());
    getOwnWorkingOutputProperty()->setDefaultFormat(out_format);

    return SUCCESS;
  } /* doInitOutput */

  Result URLNode::doDeinitOutput() {
    if (reader) {
      delete reader;
      reader = NULL;
    }
    return SUCCESS;
  }

  Message* URLNode::processBuffer(Buffer*) {

    MutexGuard mg_em(event_mutex);
    switch (event) {
    case START: /* Start of a new Data */
      event = DATA;
      return new CEvent(ITrack::create_startTrack("url"));
      break;

    case DATA: /* ok, sending data message */
      {
	Buffer* out = getNewBuffer(buffer_size);
	int s = reader->readData(out->getData(), buffer_size);
	if (0 >= s) {
	  ERROR_STREAM("Could not read data!" << endl);
	  out->release();
	  out=NULL;
	} else {
	  out->setUsed(s);
	}
	if (reader->eof()) {
	  event = END;
	}
	return out;
      }
      break;

    case END: /* current Data doesn't continue */
      event = FINISHED;

      return new CEvent(ITrack::create_endTrack());
      break;

    case FINISHED: /* Complete Data is played */
      event = FINISHED;
      setProducingFlag(false);
      return new CEvent(ITrackList::create_endTrackList());
      break;
    }

    // as the above switch-case is complete, we should never get here
    // to avoid a compiler warning about "control reaches end of
    // non-void function", we
    return NULL;
  } /* produceBuffer */

}
