/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "DevNullNode.hpp"

#include "nmm/comm/format/Property.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"

#include "IBuffersPassedEventReceiver.hpp"

#include "nmm/utils/thread/Thread.hpp"

NMM_REGISTER_PLUGIN(DevNullNode, NMM, "DevNullNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  DevNullNode::DevNullNode(const char* name,
			   const StreamQueue::Mode queue_mode,
			   const int size)
    : GenericSinkNode(name, queue_mode, size)
  {
    buffers_passed = 0;
    sleep_time = 0;
    upstream_passed = 0;
    interval = 0;
  }
  
  void DevNullNode::setSleepTime(const nmm_int32& value) {
    sleep_time = value;
  }
  
  nmm_int32 DevNullNode::getSleepTime() const {
    return sleep_time;
  }
  
  nmm_int32 DevNullNode::getBuffersPassed() const {
    return buffers_passed;
  }
  
  void DevNullNode::resetBufferCounter() {
    buffers_passed = 0;
    upstream_passed = 0;
  }
  
  void DevNullNode::setBufferCountInterval(const nmm_int32& value) {
    if (interval == 0) {
      /*
       * If the interval was previously zero, we start counting buffers from now.
       * If we just set the interval to a new value, we don't reset the counter.
       */
      upstream_passed = 0;
    }
    
    interval = value;
  }
  
  nmm_int32 DevNullNode::getBufferCountInterval() const {
    return interval;
  }
  
  Result DevNullNode::doInit() {
    Format* format = getOwnInputProperty()->addNewFormat("*/*");
    getOwnInputProperty()->setDefaultFormat(format);
    
    return SUCCESS;
  }
  
  Result DevNullNode::doInitOutput() {
    const Format* in_format = getInputFormat();
    if (in_format == NULL) { return FAILURE; }
    Format* format;
    
    format = new Format( *in_format );
    getOwnWorkingInputProperty()->addNewFormat(format);
    getOwnWorkingInputProperty()->setDefaultFormat(format);
    
    return SUCCESS;
  }
  
  Message* DevNullNode::processBuffer(Buffer *in_buffer) {
    in_buffer->release();
    buffers_passed++;
    if (interval) { 
      upstream_passed++;
      if(upstream_passed >= interval) {
 	Event* e = new Event(IBuffersPassedEventReceiver::buffersPassed_event, new TInValue<nmm_int32>(upstream_passed));
 	CEvent* ce = new CEvent(e);
 	ce->setDirection(Message::UPSTREAM);
 	DEBUG_STREAM("Created UPSTREAM event " << IBuffersPassedEventReceiver::buffersPassed_event << " " << upstream_passed << endl);
	upstream_passed = 0;
 	return ce;
      }
    }

    if (sleep_time) {
      DEBUG_STREAM("Sleeping " << sleep_time << " usec" << endl);
	  Thread::suspendThread(sleep_time/1000);
    }
    return 0;
  }
}


#include "nmm/nmm_suffix.hpp"
