/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


/**
 * This NMM node provides the possibility to control a 
 * video camera over the VISCA-protocol from Sony.
 * Should work up to 7 cameras
 *
 * This NMM node was developed in the context of 
 * the multimedia course in WS2001/02 at 
 * University of Saarbruecken/Germany
 */

#ifndef _VISCACAMERAREADNODE_HPP_
#define _VISCACAMERAREADNODE_HPP_ _VISCACAMERAREADNODE_HPP_

#include <string>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <unistd.h>
#include <fstream>
#include <iomanip>
#include <termios.h>
#include <fcntl.h>


#include "nmm/plugins/video/tv/tvcard/TVCardReadNode2.hpp"
#include "nmm/plugins/video/camera/visca/IVISCACameraImpl.hpp"

#define MAX_VISCA_DEVICES         8 /* do NOT change this!! */
#define MAX_SLOTS_PER_CAMERA      5
#define MAX_VISCA_COMMAND_LENGTH 15

namespace NMM {

    class VISCACameraReadNode : 
	public TVCardReadNode2,
	public IVISCACameraImpl 
    {
    public:

	typedef struct struct_camerapos {
	    /* position in x direction */ 
	    signed short int x;       /* [-1440;1440] */
	    signed short int x_deg;   /* [-100�;100�] */
	    
	    /* position in y direction */ 
	    signed short int y;       /* [-360;360] */
	    signed short int y_deg;   /* [-25�;25�] */
	    
	    /* position of the zoom , 
	       optical zoom : [    0;16384]
	       digital zoom : [16384;28672] */
	    signed short int z;  
	};
	
	typedef struct struct_visca_command {
	    unsigned char command[MAX_VISCA_COMMAND_LENGTH];
	    unsigned char length;
	};
	
	struct_camerapos camerapos;  
	
	/**
	 * Constructor
	 * 
	 * input : string for the serial port device
	 *         eg. "/dev/ttyS1"
	 * ouput : none
	 */
	VISCACameraReadNode(const char *name = "VISCACameraReadNode",
			    StreamQueue::Mode = StreamQueue::MODE_SUSPEND,
			    int = StreamQueue::DEFAULT_SIZE);
	
	/**
	 * Destructor
	 */
	~VISCACameraReadNode();
	
	
	/**
	 * init function
	 *
	 * Call this function to initialize all variable and 
	 * constants needed for this library to do his job.
	 *
	 * input  : none
	 * output : SUCCESS when success, else FAILURE
	 */
	Result doInit();
	
	Result doDeinit();
	
	/**
	 * init_serialdevice function
	 * restore_serialdevice function
	 *
	 * This function sets the serial port paramater to 
	 * 8 data bits, 1 start and 1 stop bit, no parity and
	 * most significant byte first
	 *
	 * Please call this functions not directly, they are 
	 * called automaticly
	 *
	 * input  : none
	 * output : SUCCESS when success, else FAILURE
	 */
	Result init_serialdevice();
	Result restore_serialdevice();
	
	
	
	/**
	 * events functions 
	 */
	
	Result setDevice(string&, string&);
	Result getDevice(string&, string&);
	Result setChannel(nmm_int32&);
	
	/*******************************************************************
	 * EVI-D100P functions 
	 *******************************************************************/
	
	/**
	 * IF_Clear 
	 * 
	 * Clears the command buffers in the camera and 
	 * cancels the command currently beeing executed.
	 *
	 * this function will be called automaticly by the init
	 * function 
	 *
	 * input  : the number of the camera [1;7,8]
	 *          camera_id 8 means broadcast, all 
	 *          cameras will clear their command
	 *          buffers
	 * output : SUCCESS when success, else FAILURE
	 */
	Result IF_Clear( const nmm_uchar& );
	
	/**
	 * AddressSet
	 *
	 * Scan all cameras connected to the VISCA-bus 
	 * and dispatch one VISCA-address for each camera found.
	 *
	 * This function will be called automaticly by the 
	 * init function
	 *
	 * input  : none
	 * output : SUCCESS when success, else FAILURE
	 */
	Result AddressSet();
    
	/**
	 * getNumberOfCameras
	 *
	 * Returns the number of cameras connected to the VISCA-bus
	 *
	 * input  : none
	 * output : number of cameras found
	 */
        Result getNumberOfCameras(nmm_uchar&);
	
	
	/**
	 * CommandCancel
	 * 
	 * Cancel the command currently executed by the camera
	 * with the address <camera identificator> and the socket
	 * number <socket id>.
	 *
	 * input  : camera identificator and socket id
	 * output : SUCCESS, if the command is canceled
	 *          FAILURE, if not
	 */
        Result CommandCancel( const nmm_uchar&, const nmm_uchar& );
	
	
	/**
	 * Power functions
	 *
	 * The following functions controlls the powerstate 
	 * of the cameras. Each camera can be controlled seperatly,
	 * selectecd by the <camera id>. The AutoPowerOff-function sets 
	 * the timer after the camera goes in standby-mode, when no 
	 * VISCA-cammand arrived 
	 *
	 * input  : camera id
	 * output : SUCCESS if success, else FAILURE
	 */
        Result PowerOn( const nmm_uchar& );
        Result PowerOff( const nmm_uchar& );
        Result AutoPowerOff( const nmm_uchar&, const nmm_int32& );
	
	
	/**
	 * Zoom functions
	 *
	 * The following functions controls the zoom of the camera. 
	 * All that functions expect the <camera_id> as parameter.
	 * Some of them allow additional speed settings 
	 *
	 * - The Range for the Zoom:
	 *   optical zoom : [    0;16384]
	 *   digital zoom : [16384;28672]
	 * - ZoomDZoom(On|Off) turn the DigitalZoom On or Off.
	 */
        Result ZoomStop( const nmm_uchar& );
        Result ZoomTeleStd( const nmm_uchar& );
        Result ZoomWideStd( const nmm_uchar& );
        Result ZoomTeleVar( const nmm_uchar&, const nmm_uchar&);
        Result ZoomWideVar( const nmm_uchar&, const nmm_uchar&);
        Result ZoomDirect( const nmm_uchar&, const nmm_int32& );
        Result ZoomDigitalZoomOnOff(const nmm_uchar&, const nmm_bool&);
	
	
	/**
	 * Focus functions
	 * 
	 * the Focus* functions are controling the settings of te
	 * camera's focus. All functions expect the <camera_id> 
	 * as first parameter, some an additional parameter
	 */
        Result FocusStop(const nmm_uchar&);
        Result FocusFarStd(const nmm_uchar&);
        Result FocusNearStd(const nmm_uchar&);
        Result FocusFarVar(const nmm_uchar&,const nmm_uchar&);
        Result FocusNearVar(const nmm_uchar&,const nmm_uchar&);
        Result FocusDirect(const nmm_uchar&,const nmm_uint16&);
        Result FocusAutoFocus(const nmm_uchar&);
        Result FocusManualFocus(const nmm_uchar&);
        Result FocusAutoManual(const nmm_uchar&);
        Result FocusOnePushTrigger(const nmm_uchar&);
        Result FocusInfinity(const nmm_uchar&);
        Result FocusAFSensHigh(const nmm_uchar&);
        Result FocusAFSensLow(const nmm_uchar&);
        Result FocusNearLimit(const nmm_uchar&,const nmm_uint16&);
	
	
	/**
	 * WhiteBalance
	 */
        Result WhiteBalance(const nmm_uchar&,const nmm_uchar&,const nmm_uchar&);
        Result WhiteBalanceAuto(const nmm_uchar&);
        Result WhiteBalanceIndoor(const nmm_uchar&);
        Result WhiteBalanceOutdoor(const nmm_uchar&);
        Result WhiteBalanceOnePush(const nmm_uchar&);
        Result WhiteBalanceAT(const nmm_uchar&);
        Result WhiteBalanceManual(const nmm_uchar&);
        Result WhiteBalanceOnePushTrigger(const nmm_uchar&);
	
	
	/**
	 * RGain
	 */
        Result RGainReset(const nmm_uchar&);
        Result RGainUp(const nmm_uchar&);
        Result RGainDown(const nmm_uchar&);
        Result RGainDirect(const nmm_uchar&,const nmm_uint16&);
	
	
	
	/**
	 * BGain
	 */
        Result BGainReset(const nmm_uchar&);
        Result BGainUp(const nmm_uchar&);
        Result BGainDown(const nmm_uchar&);
        Result BGainDirect(const nmm_uchar&,const nmm_uint16&);
	
	
	/**
	 * Automatic Exposure 
	 */
        Result AutomaticExposureFull(const nmm_uchar&);
        Result AutomaticExposureManual(const nmm_uchar&);
        Result AutomaticExposureShutterPriority(const nmm_uchar&);
        Result AutomaticExposureIrisPriority(const nmm_uchar&);
        Result AutomaticExposureGainPriority(const nmm_uchar&);
        Result AutomaticExposureBright(const nmm_uchar&);
        Result AutomaticExposureShutterAuto(const nmm_uchar&);
        Result AutomaticExposureIrisAuto(const nmm_uchar&);
        Result AutomaticExposureGainAuto(const nmm_uchar&);
	
	/**
	 * Slow Shutter
	 */
        Result SlowShutter(const nmm_uchar&,const nmm_bool&);
	
	/**
	 * Shutter
	 */
        Result ShutterReset(const nmm_uchar&);
        Result ShutterUp(const nmm_uchar&);
        Result ShutterDown(const nmm_uchar&);
        Result ShutterDirect(const nmm_uchar&,const nmm_uint16&);
	
	
	/**
	 * Iris
	 */
        Result IrisReset(const nmm_uchar&);
        Result IrisUp(const nmm_uchar&);
        Result IrisDown(const nmm_uchar&);
        Result IrisDirect(const nmm_uchar&,const nmm_uint16&);
	
	
	/**
	 * Gain
	 */
        Result GainReset(const nmm_uchar&);
        Result GainUp(const nmm_uchar&);
        Result GainDown(const nmm_uchar&);
        Result GainDirect(const nmm_uchar&,const nmm_uint16&);
	
	/**
	 * Bright
	 */
        Result BrightReset(const nmm_uchar&);
        Result BrightUp(const nmm_uchar&);
        Result BrightDown(const nmm_uchar&);
        Result BrightDirect(const nmm_uchar&,const nmm_uint16&);
    
        /**
         * Exposure Compensation
         */
        Result ExposureCompensationOnOff(const nmm_uchar&,const nmm_bool&);
        Result ExposureCompensationReset(const nmm_uchar&);
        Result ExposureCompensationUp(const nmm_uchar&);
        Result ExposureCompensationDown(const nmm_uchar&);
        Result ExposureCompensationDirect(const nmm_uchar&,const nmm_uint16&);
    
        /** 
         * BackLight
         */
        Result BackLightOnOff(const nmm_uchar&,const nmm_bool&);
    
    
        /**
         * Aperture
         */
        Result ApertureReset(const nmm_uchar&);
        Result ApertureUp(const nmm_uchar&);
        Result ApertureDown(const nmm_uchar&);
        Result ApertureDirect(const nmm_uchar&,const nmm_uint16&);
    
    
        /**
         * Wide Mode
         */
        Result WideModeOff(const nmm_uchar&);
        Result WideModeCinema(const nmm_uchar&);
        Result WideMode16to9Full(const nmm_uchar&);
    
    
        /**
         *  Mirror image on/off
         */
        Result LR_Reverse(const nmm_uchar&, const nmm_bool&);
    
        /**
         *  Freeze image on/off
         */
        Result Freeze(const nmm_uchar&, const nmm_bool&);
    
    
    /**
     * Picture Effects
     *
     * These functions allows to turn internal effects on or off. 
     * They expect as parameter the <camera_id>.
     */
    Result PictureEffectOff( const nmm_uchar& );
    Result PictureEffectPastel( const nmm_uchar& );
    Result PictureEffectNegArt( const nmm_uchar& );
    Result PictureEffectSepia( const nmm_uchar& );
    Result PictureEffectBW( const nmm_uchar& );
    Result PictureEffectSolarize( const nmm_uchar& );
    Result PictureEffectMosaic( const nmm_uchar& );
    Result PictureEffectSlim( const nmm_uchar& );
    Result PictureEffectStretch( const nmm_uchar& );
    

    /**
     * Digital Effects
     *
     * These functions allows to turn internal digital effects 
     * on or off. They expect as parameter the <camera_id>.
     *
     * Please don't call DigitalEffect( const nmm_uchar&, const nmm_uchar& ), its a parent
     * functions for all the other effect functions  
     */
    Result DigitalEffectOff( const nmm_uchar& );
    Result DigitalEffectStill( const nmm_uchar& );
    Result DigitalEffectFlash( const nmm_uchar& );
    Result DigitalEffectLumi( const nmm_uchar& );
    Result DigitalEffectTrail( const nmm_uchar& );
    Result DigitalEffectLevel( const nmm_uchar& , const nmm_uchar&, const nmm_uchar& );
    

    /**
     * Memory (user defined presets)
     */
    Result Memory(const nmm_uchar&,const nmm_uchar&,const nmm_uchar&);
    Result MemoryReset( const nmm_uchar&, const nmm_uchar& );
    Result MemorySet( const nmm_uchar&, const nmm_uchar& );
    Result MemoryRecall( const nmm_uchar&, const nmm_uchar& );
    
    
    /**
     * DataScreen (OSD)
     */
    Result DataScreenOnOff(const nmm_uchar&,const nmm_bool&);
    

    /**
     * IR (remote controller)
     */
    Result IRReceiveOnOff(const nmm_uchar&,const nmm_bool&);
    Result IRReceiveReturn(const nmm_uchar&,const nmm_bool&);
    

    /** 
     * PanTiltDriveMotion
     *
     * input  : camera-identifier,
     *          motion x-direction, motion y-direction, 
     *          speed x-direction, speed y-direction
     * output : SUCCESS, if command send with success
     *          FAILURE, if an error occured
     * 
     * used for PanTiltDrive-motion functions
     */
    Result PanTiltDriveMotion(const nmm_uchar&, const nmm_uchar&, const nmm_uchar&, const nmm_uchar&, const nmm_uchar&);
    
    Result PanTiltDriveUp( const nmm_uchar&, const nmm_uchar&, const nmm_uchar&);
    Result PanTiltDriveDown( const nmm_uchar&, const nmm_uchar&, const nmm_uchar&);
    Result PanTiltDriveLeft( const nmm_uchar&, const nmm_uchar&, const nmm_uchar&);
    Result PanTiltDriveRight( const nmm_uchar&, const nmm_uchar&, const nmm_uchar&);
    Result PanTiltDriveUpLeft( const nmm_uchar&, const nmm_uchar&, const nmm_uchar&);
    Result PanTiltDriveUpRight( const nmm_uchar&, const nmm_uchar&, const nmm_uchar&);
    Result PanTiltDriveDownLeft( const nmm_uchar&, const nmm_uchar&, const nmm_uchar&);
    Result PanTiltDriveDownRight( const nmm_uchar&, const nmm_uchar&, const nmm_uchar&);
    Result PanTiltDriveStop( const nmm_uchar&, const nmm_uchar&, const nmm_uchar&);
    
    /** 
     * PanTiltDrivePosition
     *
     * input  : camera-identifier,
     *          position-mode (relative or absolute)
     *          motion x-direction in coordinates
     *          morion y-direction in coordinates
     *          speed x-direction, speed y-direction
     * output : SUCCESS, if command send with success
     *          FAILURE, if an error occured
     * 
     * used for PanTiltDrive-position functions
     */
    Result PanTiltDrivePosition( const nmm_uchar&, const nmm_uchar&, const nmm_int32&, const nmm_int32&, const nmm_uchar&, const nmm_uchar& );
    Result PanTiltDriveAbsolute( const nmm_uchar&, const nmm_int32&, const nmm_int32&, const nmm_uchar&, const nmm_uchar&);
    Result PanTiltDriveRelative( const nmm_uchar&, const nmm_int32&, const nmm_int32&, const nmm_uchar&, const nmm_uchar&);
    
    Result PanTiltDriveAbsoluteDeg( const nmm_uchar&, const nmm_int32&, const nmm_int32&, const nmm_uchar&, const nmm_uchar&);
    Result PanTiltDriveRelativeDeg( const nmm_uchar&, const nmm_int32&, const nmm_int32&, const nmm_uchar&, const nmm_uchar&);
    
    Result PanTiltDriveHome( const nmm_uchar& );
    Result PanTiltDriveReset( const nmm_uchar& );
    Result PanTiltLimitSet( const nmm_uchar&, const nmm_uchar&, const int&, const nmm_int32& );
    Result PanTiltLimitClear( const nmm_uchar&, const nmm_uchar& );
    
    
    /**
     * Inquiry Command List 
     */
    
    /**
     * PowerInq
     *
     * input : The camera ID
     * ouput : The status of the Camera
     *          true = On
     *          false =Off(Standby)
     */
    Result PowerInq( const nmm_uchar&,nmm_bool& );
    Result AutoPowerOffInq( const nmm_uchar&,nmm_uint16& );
    Result DZoomModeInq(const nmm_uchar&, nmm_bool&);
    Result ZoomPosInq(const nmm_uchar&, nmm_uint16&);
    Result FocusModeInq(const nmm_uchar&,nmm_bool&);
    Result FocusPosInq(const nmm_uchar&,nmm_uint16&);
    Result AFModeInq(const nmm_uchar&,nmm_bool&);
    Result FocusNearLimitInq(const nmm_uchar&,nmm_uint16&);
    Result WhiteBalanceModeInq(const nmm_uchar&,nmm_uchar&);
    //    Result RGainInq(const nmm_uchar&,nmm_uint16&);
    Result RGainInq(const nmm_uchar&,nmm_uchar&);
    Result BGainInq(const nmm_uchar&,nmm_uint16&);
    Result AutoExposureModeInq(const nmm_uchar&,nmm_uchar&);
    Result SlowShutterModeInq(const nmm_uchar&,nmm_bool&);
    Result ShutterPosInq(const nmm_uchar&,nmm_uint16&);
    Result IrisPosInq(const nmm_uchar&,nmm_uint16&);
    Result GainPosInq(const nmm_uchar&,nmm_uint16&);
    Result BrightPosInq(const nmm_uchar&,nmm_uint16&);
    Result ExposureCompensationModeInq(const nmm_uchar&,nmm_bool&);
    Result ExposureCompensationPosInq(const nmm_uchar&,nmm_uint16&);
    Result BackLightModeInq(const nmm_uchar&,nmm_bool&);
    Result ApertureInq(const nmm_uchar&,nmm_uint16&);
    Result WideModeInq(const nmm_uchar&,nmm_uchar&);
    Result LR_ReverseModeInq(const nmm_uchar&, nmm_bool&);
    Result FreezeModeInq(const nmm_uchar&, nmm_bool&);
    Result PictureEffectModeInq(const nmm_uchar&, nmm_uchar&);
    Result DigitalEffectModeInq(const nmm_uchar&, nmm_uchar&);
    Result DigitalEffectLevelInq(const nmm_uchar&, nmm_uchar&);
    Result MemoryInq(const nmm_uchar&,nmm_uchar&);
    Result DatascreenInq(const nmm_uchar&,nmm_bool&);
    Result PanTiltModeInq(const nmm_uchar&,nmm_uint16&);
    Result PanTiltMaxSpeedInq(const nmm_uchar&,nmm_uchar&,nmm_uchar&);
    Result PanTiltPosInq(const nmm_uchar&, nmm_int32&, nmm_int32&);
    //    Result PanTiltPosInq( const nmm_uchar& , struct_camerapos* );
    Result VideoSystemInq(const nmm_uchar&,nmm_uchar&);
    Result DeviceTypeInq(const nmm_uchar&, nmm_uint16&, nmm_uint16&, nmm_uint16&, nmm_uchar&);
    
    /**
     * misc stuff
     */
//     void verbose(const string&, struct_visca_command);
    void verbose(const string&, const string&);
    bool islocked(const nmm_uchar&);
    
  private:
    
    typedef struct struct_slots {
      bool active;
      unsigned char socket;
      struct_visca_command expected_answer;
      struct_visca_command comparison_mask;
      struct timeval expiredate;
    };
    
    typedef struct struct_camera_buffer {
      bool         locked;
      unsigned char         lockedslot;
      struct_slots slots[MAX_SLOTS_PER_CAMERA];
    };
    
    struct_visca_command visca_command;
    struct_visca_command visca_response1;
    struct_visca_command visca_response2;
    
    struct_camera_buffer slots[MAX_VISCA_DEVICES];
    
    
    /* stores port parameters */
    struct termios *old_termios, *new_termios;

    string serial_device;
    int spd; /* Serial Port Descriptor */
    
    int number_of_cameras;
    int number_of_tries;
    
    ThreadMutex mutex_command_send;

    /**
     * low level access functions to camera(s)
     */
    Result send_to_camera(struct_visca_command*);
    Result read_from_camera(struct_visca_command*);
    bool compare_commands( struct_visca_command*, struct_visca_command* );
    
    Result wait_for_ack(const nmm_uchar&,const nmm_uchar&);
    Result send_command_and_wait_for_ack(struct_visca_command *cmdsend, bool lockcam=false);
    Result send_inquiry_command(const nmm_uchar&, const nmm_uchar&, const nmm_uchar&, struct_visca_command*);
    
    
    /**
     * slot control
     */
    int check_slots_for_timeout(const nmm_uchar&);
    int find_free_slot(const nmm_uchar&);
    
    
    /**
     * misc stuff
     */
    void compose_command6(const nmm_uchar&,const nmm_uchar&,const nmm_uchar&,const nmm_uchar&,struct_visca_command*);
    int check_response(struct_visca_command *response, bool);
    Result process_response(struct_visca_command*);

  }; /* class */
  
}; /* namespace NMM */

#endif /* _CAMERAREADNODE_HPP_ */
