/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "FramerateConverterNode.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"
#include "nmm/comm/format/Property.hpp"
#include "nmm/comm/messaging/Buffer.hpp"
#include "nmm/utils/thread/MutexGuard.hpp"
#include <stdio.h>

NMM_REGISTER_PLUGIN(FramerateConverterNode, NMM, "FramerateConverterNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  FramerateConverterNode::FramerateConverterNode(const char *name, StreamQueue::Mode qmode, int qsize) :
    GenericProcessorNode(name, qmode, qsize),
    drop_num(0),
    drop_denom(0),
    drop_ratio(0),
    ratio_counter(0), 
    notify_framerate_changed(false),
    changed_framerate(false),
    first_timestamp(false),
    stored_buffer(0)
  {
    buffer_to_drop = 0;
    buffer_count = 0;
  }
  
  void FramerateConverterNode::setBufferToDrop(const nmm_uint32 &buffer_num)
  {
    buffer_to_drop = buffer_num;
  }
  
  Result FramerateConverterNode::doInit() 
  {
    Format* any_format_i = new Format(Format::video_raw_type);
    getOwnInputProperty()->addNewFormat(any_format_i);
    any_format_i->addWildcard(Format::x_resolution_param);
    any_format_i->addWildcard(Format::y_resolution_param);
    any_format_i->addWildcard(Format::ratio_x_param);
    any_format_i->addWildcard(Format::ratio_y_param);
    any_format_i->addWildcard(Format::framerate_param);
    any_format_i->addWildcard(Format::channelorder_param);
    any_format_i->addWildcard(Format::endian_param);
    any_format_i->addWildcard(Format::colorspace_param);
    any_format_i->addWildcard(Format::bitperpixel_param);
    any_format_i->addWildcard(Format::format_param);

    Format* any_format_o = new Format(Format::video_raw_type);
    getOwnOutputProperty()->addNewFormat(any_format_o);
    any_format_o->addWildcard(Format::x_resolution_param);
    any_format_o->addWildcard(Format::y_resolution_param);
    any_format_o->addWildcard(Format::ratio_x_param);
    any_format_o->addWildcard(Format::ratio_y_param);
    any_format_o->addWildcard(Format::framerate_param);
    any_format_o->addWildcard(Format::channelorder_param);
    any_format_o->addWildcard(Format::endian_param);
    any_format_o->addWildcard(Format::colorspace_param);
    any_format_o->addWildcard(Format::bitperpixel_param);
    any_format_o->addWildcard(Format::format_param);

    return SUCCESS;
  }
  
  Result FramerateConverterNode::doInitOutput() {
    /* get input format */
    if (!getInputFormat()) {
      GLOBAL_ERROR_STREAM("No Input Format! FAILURE" << endl);
      return FAILURE;
    }

    Format* i_format = new Format(*getInputFormat());
    getOwnWorkingInputProperty()->addNewFormat(i_format);
    getOwnWorkingInputProperty()->setDefaultFormat(i_format);
    
    Format* o_format = new Format(*i_format);
    getOwnWorkingOutputProperty()->addNewFormat(o_format);
    getOwnWorkingOutputProperty()->setDefaultFormat(o_format);

    /* Try to add framerate with wildcard. If available */
    if ( i_format->isParameterIncluded(Format::framerate_param) ) {
      o_format = new Format(*i_format);
      o_format->deleteParameter(Format::framerate_param);
      o_format->addWildcard(Format::framerate_param);
      getOwnWorkingOutputProperty()->addNewFormat(o_format);      
    }
        
    return SUCCESS;
  }
  
  Result FramerateConverterNode::doActivate() {
    notify_framerate_changed = true;
    changed_framerate = false;

    /* Check if input format and output format is valid */
    const Format* in_format  = getInputFormat();
    const Format* out_format = getOutputFormat();
    if (!in_format && !out_format) {
       ERROR_STREAM("No valid input or output format set"<< endl);
       return FAILURE;
    }
        
    /* Check if framrate is included */
    if ( in_format->isParameterSpecified(Format::framerate_param) && out_format->isParameterSpecified(Format::framerate_param)) {
      float in_framerate  = in_format->getFloatValue(Format::framerate_param);
      float out_framerate = out_format->getFloatValue(Format::framerate_param);

       if (out_framerate == in_framerate) {
        return SUCCESS; //nothing todo
      }
            
      if (out_framerate > in_framerate) {
        ERROR_STREAM("Output framerate is larger than input framerate !!"<< endl);
        return FAILURE;
      }
      int int_in_framerate = static_cast<int>(in_framerate * 1000.0);
      int int_out_framerate = static_cast<int>(out_framerate * 1000.0);
      setDropRatio(int_out_framerate, int_in_framerate);
    } else {
      ERROR_STREAM( "No input or  output formats available" << endl);
      return FAILURE;
    }
    
    return SUCCESS;
  }
  
  Message* FramerateConverterNode::processBuffer(Buffer *in) 
  {
    MutexGuard mg(internal_mutex);
    /* Begin> If there is a stored buffer, process it */
    if ( stored_buffer ) {
      if ( in ) {
        ERROR_STREAM( "ARGH!!! THIS MUST NOT HAPPEN" << endl);
      } else {
        in = stored_buffer;
	stored_buffer = 0;
	setWorkingFlag(false);
      }
    }
    /* End< */
    
    /* Begin> Store first initial timestamp */
    if ( in && first_timestamp) {
      if ( in->getTimestamp().is_valid )  {
        current_timestamp = in->getTimestamp();
      }
    } 
    /* End< */

    /* If the framerate has been changed, send it to all successor nodes */
    if ( notify_framerate_changed && changed_framerate ) {
      setWorkingFlag(true);
      stored_buffer = in;
      float new_framerate = getInputFormat()->getFloatValue(Format::framerate_param) - getInputFormat()->getFloatValue(Format::framerate_param) * drop_num / drop_denom;
      
      const_cast<Format*>(getOutputFormat())->deleteParameter(Format::framerate_param);
      const_cast<Format*>(getOutputFormat())->addFloatParamValue(Format::framerate_param, new_framerate);
      
      cerr << "Sending new format " << *getOutputFormat() << endl;
      changed_framerate = false;
      return new CEvent(IFormatChange::create_setFormat(*getOutputFormat()) );
    } 
      
    if(buffer_to_drop > 0) { 
      ++buffer_count;
      if(buffer_count == buffer_to_drop) {
	buffer_count = 0;
	in->release();
	return 0;
      }
    } else if (drop_num > 0) {
      ++buffer_count;
      int64_t old_integer_part = ratio_counter.getNumerator() / ratio_counter.getDenominator();
      ratio_counter += drop_ratio;
      int64_t new_integer_part = ratio_counter.getNumerator() / ratio_counter.getDenominator();
      if (old_integer_part < new_integer_part) {
        DEBUG_STREAM("DropDecision: " << buffer_count << "." << " Dropping!" << endl);
        in->release();
        return NULL;
      } else {
        DEBUG_STREAM("DropDecision: " << buffer_count << "." << " Leaving!" << endl);
      }

      if (buffer_count == drop_denom) {
        buffer_count = 0;
      }
    }
    
    /* Adjust timestamp */
    if ( adjust_timestamp_value != Interval() && !first_timestamp){
      in = in->getWriteableInstance(); //I think we need a writeable instance because we change the timestamp of the buffer which can affect other nodes that share this buffer. 
      current_timestamp.sync_time += adjust_timestamp_value;

      Timestamp new_timestamp  = in->getTimestamp();
      new_timestamp.sync_time += adjust_timestamp_value;
      
      in->setTimestamp( new_timestamp );
      //cerr << "New timestamp " << in->getTimestamp() << endl;
    }
    if (first_timestamp) {
      first_timestamp = false;
    }
    
    return in;
  }

  void FramerateConverterNode::setDropRatio(const nmm_uint32& numerator, const nmm_uint32& denominator) {
    MutexGuard mg(internal_mutex);
    if ((denominator == 0) || (numerator > denominator)) {
      ERROR_STREAM("setDropRatio() Invalid Args: " << numerator << " / " << denominator << endl);
      return;
    }

    if (buffer_to_drop > 0) {
      buffer_to_drop = 0;
    }

    /* Begin> Reset internal attributes */    
    buffer_count = 0;
    changed_framerate = true;
    first_timestamp = true;
    /* End< */
         
    drop_num   = numerator;
    drop_denom = denominator;
    drop_ratio = Rational(drop_num, drop_denom);
    drop_ratio.reduce();
    DEBUG_STREAM( "new dropratio " << drop_ratio << endl );
    float new_framerate =  getInputFormat()->getFloatValue(Format::framerate_param) - getInputFormat()->getFloatValue(Format::framerate_param) * numerator / denominator;
    DEBUG_STREAM( "New framerate " << new_framerate << endl);
    adjust_timestamp_value.nsec = static_cast<long long>(1000000000.0 / new_framerate);
  }
}

#include "nmm/nmm_suffix.hpp"
