/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "RGBtoRGBConverterNode.hpp"
#include "nmm/comm/format/Property.hpp"
#include "nmm/comm/messaging/Buffer.hpp"

#include "nmm/plugins/NMMRegisterPlugin.hpp"

extern "C" {
  void unpack422rgb32_mmx(char *src, int n, char *dst);
}

NMM_REGISTER_PLUGIN(NMM::RGBtoRGBConverterNode, NMM, "RGBtoRGBConverterNode")

namespace NMM {

  RGBtoRGBConverterNode::RGBtoRGBConverterNode( const char* name)
    : GenericConverterNode(name)
  {
    allocate();
  }

    /**
     * converts each pixel in cp_buffer (16bit) to a pixel in cp_target (16bit)
     * by switching red and blue channel
     */
  Buffer* RGBtoRGBConverterNode::BGR16_5_6_5_LIT_to_RGB16_5_6_5_LIT(Buffer* in_buffer, const GenericConverterNode* node) {
    Buffer* out_buffer = const_cast<GenericConverterNode*>(node)->getNewBuffer(in_buffer->getUsed(), in_buffer);
    if (out_buffer) {
      char* src=in_buffer->getData();
      char* dst=out_buffer->getData();
      unsigned int j=0;
	    
      int16_t s_pixel;
      for (unsigned int i=0; i<in_buffer->getUsed(); i+=2) {
	s_pixel = * ((int16_t *)(src+i));
		
	*((int16_t *)&dst[j]) = ((s_pixel & 0x001F) << 11)
	  | ((s_pixel & 0x07E0) )
	  | ((s_pixel & 0xF800) >> 11);
	j += 2;
      }
	    
    }
    else {
      NODE_ERROR_STREAM(node,  " Could not get out buffer." << endl);
    }
	
    in_buffer->release();
    return out_buffer;	
  }

  /**
   * converts each pixel in cp_buffer (16bit) to a pixel in cp_target (24bit)
   * by scaling rgb
   */
  Buffer* RGBtoRGBConverterNode::RGB16_5_6_5_LIT_to_RGB24_8_8_8_LIT(Buffer* in_buffer, const GenericConverterNode* node)
  {
    Buffer* out_buffer = const_cast<GenericConverterNode*>(node)->getNewBuffer(in_buffer->getUsed() + in_buffer->getUsed()/2 , in_buffer);
    if (out_buffer) {
      char* src = in_buffer->getData();
      char* dst = out_buffer->getData();
      out_buffer->setUsed(in_buffer->getUsed() + in_buffer->getUsed()/2 );
	    
      unsigned int j = 0;
      int16_t s_pixel;
      for (unsigned int i=0; i<in_buffer->getUsed(); i+=2) {
	s_pixel = * ((int16_t *)(src+i));
		
	dst[j++]  = (char)((s_pixel & 0x001F ) << 3);
	dst[j++]  = (char)((s_pixel & 0x07E0 ) >> 3);
	dst[j++]  = (char)((s_pixel & 0xF800 ) >> 8);
      }
    }
    else {
      NODE_ERROR_STREAM(node,  " Could not get out buffer." << endl);
    }
	
    in_buffer->release();
    return out_buffer;
  }

  /**
   * converts each pixel in in_buffer (16bit) to a pixel in out_buffer (32bit)
   * by switching red and blue channel, scaling rgb and adding 8 zero-bits to
   * alpha channel
   */

  Buffer* RGBtoRGBConverterNode::BGR16_5_6_5_LIT_to_RGB32_8_8_8_LIT(Buffer* in_buffer, const GenericConverterNode* node)
  {
    Buffer* out_buffer = const_cast<GenericConverterNode*>(node)->getNewBuffer(in_buffer->getUsed()*2, in_buffer);
    if (out_buffer) {
      char *data=in_buffer->getData();
      int cnt=0;
	    
      char *out_data=out_buffer->getData();
      int16_t s_pixel;
      for (unsigned int i=0; i<in_buffer->getUsed(); i+=2) {
	s_pixel = * ((int16_t *)(data+i));
		
	out_data[cnt++]  = (char)((s_pixel & 0xF800 ) >> 8);
	out_data[cnt++]  = (char)((s_pixel & 0x07E0 ) >> 3);
	out_data[cnt++]  = (char)((s_pixel & 0x001F ) << 3);
	out_data[cnt++]  = 0;
      }
    }
    else {
      NODE_ERROR_STREAM(node,  " Could not get out buffer." << endl);
    }
	
    in_buffer->release();
    return out_buffer;
  } 

  /**
   * converts each pixel in cp_buffer (16bit) to a pixel in cp_target (32bit)
   * by scaling rgb and adding 8 zero-bits in alpha channel
   */
  Buffer* RGBtoRGBConverterNode::RGB16_5_6_5_LIT_to_RGB32_8_8_8_LIT(Buffer* in_buffer, const GenericConverterNode* node)
  {
    Buffer* out_buffer = const_cast<GenericConverterNode*>(node)->getNewBuffer(in_buffer->getUsed()*2, in_buffer);
    if (out_buffer) {
      char *data=in_buffer->getData();
      int cnt=0;
	    
      char *out_data=out_buffer->getData();
      int16_t s_pixel;
      for (unsigned int i=0; i<in_buffer->getUsed(); i+=2) {
	s_pixel = * ((int16_t *)(data+i));
		
	out_data[cnt++]  = (char)((s_pixel & 0x001F ) << 3);
	out_data[cnt++]  = (char)((s_pixel & 0x07E0 ) >> 3);
	out_data[cnt++]  = (char)((s_pixel & 0xF800 ) >> 8);
	out_data[cnt++]  = 0;
      }
    }
    else {
      NODE_ERROR_STREAM(node,  " Could not get out buffer." << endl);
    }
	
    in_buffer->release();
    return out_buffer;
  }

  /**
   * converts each pixel in cp_buffer (24bit) to a pixel in cp_target (16bit)
   * by scaling rgb
   */
  Buffer* RGBtoRGBConverterNode::RGB24_8_8_8_LIT_to_RGB16_5_6_5_LIT(Buffer* in_buffer, const GenericConverterNode* node)
  {
    Buffer* out_buffer = const_cast<GenericConverterNode*>(node)->getNewBuffer(in_buffer->getUsed() - in_buffer->getUsed()/3 , in_buffer);
    if (out_buffer) {
      char* src = in_buffer->getData();
      char* dst = out_buffer->getData();
      out_buffer->setUsed(in_buffer->getUsed() - in_buffer->getUsed()/3 );
	    
      int32_t s_pixel;
      char* cur = src;
	    
      for (unsigned int j=0; j<in_buffer->getUsed()*2/3;j+=2) {
	s_pixel = 0;
	s_pixel |= (*(cur++) << 16)
	  | (*(cur++) << 8)
	  | (*(cur++));
		
	*((int16_t *)&dst[j]) = ((s_pixel & 0x000000f8) >> 3)
	  | ((s_pixel & 0x0000fc00) >> 5)
	  | ((s_pixel & 0x00f80000) >> 8);
      }
    }
    else {
      NODE_ERROR_STREAM(node,  " Could not get out buffer." << endl);
    }
	
    in_buffer->release();
    return out_buffer;
  }

  /**
   * converts each pixel in cp_buffer (24bit) to a pixel in cp_target (32bit)
   * by adding 8-zero-bits in alpha channel
   */
  Buffer* RGBtoRGBConverterNode::RGB24_8_8_8_LIT_to_RGBA32_8_8_8_8_LIT(Buffer* in_buffer, const GenericConverterNode* node)
  {
    Buffer* out_buffer = const_cast<GenericConverterNode*>(node)->getNewBuffer(in_buffer->getUsed() + in_buffer->getUsed()/3 , in_buffer);
    if (out_buffer) {
      char* src = in_buffer->getData();
      char* dst = out_buffer->getData();
      out_buffer->setUsed(in_buffer->getUsed() + in_buffer->getUsed()/3 );
	    
      int i = 0; int j = 0;
      int n = in_buffer->getUsed();
      while(i<n) {
	dst[j] = src[i];
	dst[j+1] = src[i+1];
	dst[j+2] = src[i+2];
	dst[j+3] = (char) 0x0;
	i+=3;
	j+=4;
      }
    }
    else {
      NODE_ERROR_STREAM(node,  " Could not get out buffer." << endl);
    }
	
    in_buffer->release();
    return out_buffer;
  }

  /**
   * converts each pixel in cp_buffer (24bit) to a pixel in cp_target (32bit)
   * by switching red and blue channel and adding 8-zero-bits in alpha channel
   */
  Buffer* RGBtoRGBConverterNode::RGB24_8_8_8_LIT_to_BGRA32_8_8_8_8_LIT(Buffer* in_buffer, const GenericConverterNode* node)
  {
    Buffer* out_buffer = const_cast<GenericConverterNode*>(node)->getNewBuffer(in_buffer->getUsed() + in_buffer->getUsed()/3 , in_buffer);

    if (out_buffer) {
      char* src = in_buffer->getData();
      char* dst = out_buffer->getData();
      out_buffer->setUsed(in_buffer->getUsed() + in_buffer->getUsed()/3 );
	    
      int i = 0; int j = 0;
      int n = in_buffer->getUsed();
      while(i<n) {
	dst[j] = src[i+2];
	dst[j+1] = src[i+1];
	dst[j+2] = src[i];
	dst[j+3] = (char)255;
	i+=3;
	j+=4;
      }
    }
    else {
      NODE_ERROR_STREAM(node,  " Could not get out buffer." << endl);
    }
	
    in_buffer->release();
    return out_buffer;
  }

  /**
   * converts each pixel in cp_buffer (32bit) to a pixel in cp_target (16bit)
   * by scaling rgb to 16 bit and omitting 8 bit in alpha channel
   */
  Buffer* RGBtoRGBConverterNode::RGB32_8_8_8_LIT_to_RGB16_5_6_5_LIT(Buffer* in_buffer, const GenericConverterNode* node)
  {
    Buffer* out_buffer = const_cast<GenericConverterNode*>(node)->getNewBuffer(in_buffer->getUsed()/2, in_buffer);
    if (out_buffer) {
      char *data=in_buffer->getData();
      int cnt=0;
	    
      char *out_data=out_buffer->getData();
      int32_t s_pixel;
      for (unsigned int i=0; i<in_buffer->getUsed(); i+=4) {
	s_pixel = * ((int32_t *)(data+i));
		
	*((int16_t *)&out_data[cnt]) = ((s_pixel & 0x000000f8) >> 3)
	  | ((s_pixel & 0x0000fc00) >> 5)
	  | ((s_pixel & 0x00f80000) >> 8);
	cnt += 2;
      }
    }
    else {
      NODE_ERROR_STREAM(node,  " Could not get out buffer." << endl);
    }

    in_buffer->release();
    return out_buffer;
  }

  /**
   * converts each pixel in cp_buffer (32bit) to a pixel in cp_target (16bit)
   * by scaling rgb to 16 bit, omitting 8 bit in alpha channel and switching
   * red and blue channel
   */
  Buffer* RGBtoRGBConverterNode::BGR32_8_8_8_LIT_to_RGB16_5_6_5_LIT(Buffer* in_buffer, const GenericConverterNode* node)
  {
    Buffer* out_buffer = const_cast<GenericConverterNode*>(node)->getNewBuffer(in_buffer->getUsed()/2, in_buffer);
    if (out_buffer) {
      char *data=in_buffer->getData();
      int cnt=0;
	    
      char *out_data=out_buffer->getData();
      int32_t s_pixel;
      for (unsigned int i=0; i<in_buffer->getUsed(); i+=4) {
	s_pixel = * ((int32_t *)(data+i));
		
	*((int16_t *)&out_data[cnt]) = ((s_pixel & 0x000000f8) << 8)
	  | ((s_pixel & 0x0000fc00) >> 5)
	  | ((s_pixel & 0x00f80000) >> 19);
	cnt += 2;
      }
    }
    else {
      NODE_ERROR_STREAM(node,  " Could not get out buffer." << endl);
    }

    in_buffer->release();
    return out_buffer;
  }

  /**
   * converts each pixel in in_buffer (32bit) to a pixel in out_buffer (32bit)
   * by switching red and blue channel and adding 8 zero-bits in alpha channel
   */
  Buffer* RGBtoRGBConverterNode::BGR32_8_8_8_LIT_to_RGB32_8_8_8_LIT(Buffer* in_buffer, const GenericConverterNode* node)
  {
    Buffer* out_buffer = const_cast<GenericConverterNode*>(node)->getNewBuffer(in_buffer->getUsed(), in_buffer);
    if (out_buffer) {
      char* src = in_buffer->getData();
      char* dst = out_buffer->getData();
      out_buffer->setUsed(in_buffer->getUsed());
	    
      int i = 0;
      int n = in_buffer->getUsed();
      while(i<n) {
	dst[i] = src[i+2];
	dst[i+1] = src[i+1];
	dst[i+2] = src[i];
	dst[i+3] = (char) 0x0;
	i+=4;
      }
    }
    else {
      NODE_ERROR_STREAM(node,  " Could not get out buffer." << endl);
    }
	
    in_buffer->release();
    return out_buffer;
  }

  /**
   * converts each pixel in cp_buffer (32bit) to a pixel in cp_target (16bit)
   * by scaling rgb to 16 bit
   */
  Buffer* RGBtoRGBConverterNode::RGB32_11_11_10_LIT_to_RGB16_5_6_5_LIT(Buffer* in_buffer, const GenericConverterNode* node)
  {
    Buffer* out_buffer = const_cast<GenericConverterNode*>(node)->getNewBuffer(in_buffer->getUsed()/2, in_buffer);
    if (out_buffer) {
      char *data=in_buffer->getData();
      int cnt=0;
	    
      char *out_data=out_buffer->getData();
      int32_t s_pixel;
      for (unsigned int i=0; i<in_buffer->getUsed(); i+=4) {
	s_pixel = * ((int32_t *)(data+i));
		
	*((int16_t *)&out_data[cnt]) = ((s_pixel & 0x000003E0) >> 5)
	  | ((s_pixel & 0x001F8000) >> 10)
	  | ((s_pixel & 0xF8000000) >> 16);
	cnt += 2;
      }
	    
    }
    else {
      NODE_ERROR_STREAM(node,  " Could not get out buffer." << endl);
    }
	
    in_buffer->release();
    return out_buffer;
  }

  /**
   * converts each pixel in cp_buffer (32bit) to a pixel in cp_target (16bit)
   * by scaling rgba to 16 bit, omitting 8 bit in alpha channel and switching
   * red and blue channel
   */
  Buffer* RGBtoRGBConverterNode::BGRA32_8_8_8_8_LIT_to_RGB16_5_6_5_LIT(Buffer* in_buffer, const GenericConverterNode* node)
  {
    Buffer* out_buffer = const_cast<GenericConverterNode*>(node)->getNewBuffer(in_buffer->getUsed()/2, in_buffer);
    if (out_buffer) {
      char *data=in_buffer->getData();
      int cnt=0;
	    
      char *out_data=out_buffer->getData();
      int32_t s_pixel;
      for (unsigned int i=0; i<in_buffer->getUsed(); i+=4) {
	s_pixel = * ((int32_t *)(data+i));
		
	*((int16_t *)&out_data[cnt]) = ((s_pixel & 0x000000f8) << 8)
	  | ((s_pixel & 0x0000fc00) >> 5)
	  | ((s_pixel & 0x00f80000) >> 19);
	cnt += 2;
      }
    }
    else {
      NODE_ERROR_STREAM(node,  " Could not get out buffer." << endl);
    }

    in_buffer->release();
    return out_buffer;
  }

  /**
   * converts each pixel in cp_buffer (32 bit) to a pixel in cp_target (24bit)
   * by omitting 8 bit in alpha channel
   */
  Buffer* RGBtoRGBConverterNode::RGBA32_8_8_8_8_LIT_to_RGB24_8_8_8_LIT(Buffer* in_buffer, const GenericConverterNode* node)
  {
    Buffer* out_buffer = const_cast<GenericConverterNode*>(node)->getNewBuffer(in_buffer->getUsed() - in_buffer->getUsed()/4 , in_buffer);
    if (out_buffer) {
      char* src = in_buffer->getData();
      char* dst = out_buffer->getData();
      out_buffer->setUsed(in_buffer->getUsed() - in_buffer->getUsed()/4 );
	    
      int i = 0; int j = 0;
      int n = in_buffer->getUsed();
      while(i<n) {
	dst[j++] = src[i++];
	dst[j++] = src[i++];
	dst[j++] = src[i++];
	++i; // discard alpha of in-buffer
      }
    }
    else {
      NODE_ERROR_STREAM(node,  " Could not get out buffer." << endl);
    }
	
    in_buffer->release();
    return out_buffer;
  }

  /**
   * converts each pixel in cp_buffer (32bit) to a pixel in cp_target (24bit)
   * by switching red and blue channel and omitting 8 bit in alpha channel
   */
  Buffer* RGBtoRGBConverterNode::RGBA32_8_8_8_8_LIT_to_BGR24_8_8_8_LIT(Buffer* in_buffer, const GenericConverterNode* node)
  {
    Buffer* out_buffer = const_cast<GenericConverterNode*>(node)->getNewBuffer(in_buffer->getUsed() - in_buffer->getUsed()/4 , in_buffer);
    if (out_buffer) {
      char* src = in_buffer->getData();
      char* dst = out_buffer->getData();
      out_buffer->setUsed(in_buffer->getUsed() - in_buffer->getUsed()/4 );
	    
      int i = 0; int j = 0;
      int n = in_buffer->getUsed();
      while(i<n) {
	dst[j+2] = src[i++];
	dst[j+1] = src[i++];
	dst[j] = src[i++];
	++i; // discard alpha of in-buffer
	j+=3; 
      }
    }
    else {
      NODE_ERROR_STREAM(node,  " Could not get out buffer." << endl);
    }
	
    in_buffer->release();
    return out_buffer;
  }

  /**
   * converts each pixel in cp_buffer (11bit R, 11bit G, 10bit B) to a pixel in cp_target (32bit)
   * by scaling rgb and adding 8 zero-bits 
   */
  Buffer* RGBtoRGBConverterNode::RGB32_11_11_10_LIT_to_RGB32_8_8_8_LIT(Buffer* in_buffer, const GenericConverterNode* node)
  {
    Buffer* out_buffer = const_cast<GenericConverterNode*>(node)->getNewBuffer(in_buffer->getUsed(), in_buffer);
    if (out_buffer) {
      char *data=in_buffer->getData();
      int cnt=0;
	    
      char *out_data=out_buffer->getData();
      int32_t s_pixel;
      for (unsigned int i=0; i<in_buffer->getUsed(); i+=4) {
	s_pixel = * ((int32_t *)(data+i));
		
	out_data[cnt++]  = (char)((s_pixel & 0x000003FC ) >> 2);
	out_data[cnt++]  = (char)((s_pixel & 0x001FE000 ) >> 13);
	out_data[cnt++]  = (char)((s_pixel & 0xFF000000 ) >> 24);
	out_data[cnt++]  = (char) 0x0;
      }
    }
    else {
      NODE_ERROR_STREAM(node,  " Could not get out buffer." << endl);
    }
	
    in_buffer->release();
    return out_buffer;
  } 

  /**
   * converts each pixel in in_buffer (10bit B, 11bit G, 11bit R) to a pixel in out_buffer (32bit)
   * by switching red and blue channel, scaling rgb and adding 8 zero-bits 
   */
  Buffer* RGBtoRGBConverterNode::BGR32_10_11_11_LIT_to_RGB32_8_8_8_LIT(Buffer* in_buffer, const GenericConverterNode* node)
  {
    Buffer* out_buffer = const_cast<GenericConverterNode*>(node)->getNewBuffer(in_buffer->getUsed(), in_buffer);
    if (out_buffer) {
      char *data=in_buffer->getData();
      int cnt=0;
	    
      char *out_data=out_buffer->getData();
      int32_t s_pixel;
      for (unsigned int i=0; i<in_buffer->getUsed(); i+=4) {
	s_pixel = * ((int32_t *)(data+i));
		
	out_data[cnt++]  = (char)((s_pixel & 0xFF000000 ) >> 24);
	out_data[cnt++]  = (char)((s_pixel & 0x003FC000 ) >> 14);
	out_data[cnt++]  = (char)((s_pixel & 0x000007F8 ) >> 3);
	out_data[cnt++]  = (char) 0x0;
      }
    }
    else {
      NODE_ERROR_STREAM(node,  " Could not get out buffer." << endl);
    }
	
    in_buffer->release();
    return out_buffer;
  }

  /**
   *  Converts BGR24 to RGB32
   */
  Buffer* RGBtoRGBConverterNode::BGR24_8_8_8_LIT_to_RGB32_8_8_8_8_LIT(Buffer* in_buffer, const GenericConverterNode* node)
  {
    Buffer* out_buffer = const_cast<GenericConverterNode*>(node)->getNewBuffer(in_buffer->getUsed() + in_buffer->getUsed()/3 , in_buffer);
    if (out_buffer) {
      char* src = in_buffer->getData();
      char* dst = out_buffer->getData();
      out_buffer->setUsed(in_buffer->getUsed() + in_buffer->getUsed()/3 );
	    
      int i = 0;int j=0;
      int n = in_buffer->getUsed();
      while(i<n) {
	dst[j] = src[i+2];
	dst[j+1] = src[i+1];
	dst[j+2] = src[i];
	dst[j+3] = (char) 0x0;
	i+=3;
        j+=4;
      }
    }
    else {
      NODE_ERROR_STREAM(node,  " Could not get out buffer." << endl);
    }
	
    in_buffer->release();
    return out_buffer;
  }


  /**
   * Dummy-converter-node wich just copies the buffer
   * Used to 'convert' RGBA32 to RGB32
   */
  Buffer* RGBtoRGBConverterNode::dummy(Buffer* in_buffer, const GenericConverterNode*)
  {
    return in_buffer;
  }    
  



  void RGBtoRGBConverterNode::allocate()
  {
    Format* in_format  = 0;
    Format* out_format = 0;

    /* Input format for function BGR16_5_6_5_LIT_to_RGB16_5_6_5_LIT */
    in_format = getOwnInputProperty()->addNewFormat("video/raw");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addWildcard(Format::framerate_param);
    in_format->addStringParamValue(Format::channelorder_param,"bgr");
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::colorspace_param,"rgb565");
    in_format->addIntParamValue(Format::bitperpixel_param,16);
    in_format->addStringParamValue(Format::format_param,"packet");
    getOwnInputProperty()->setDefaultFormat(in_format);
 
    /* Output format for function BGR16_5_6_5_LIT_to_RGB16_5_6_5_LIT */
    out_format = getOwnOutputProperty()->addNewFormat("video/raw");
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addStringParamValue(Format::channelorder_param,"rgb");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::colorspace_param,"rgb565");
    out_format->addIntParamValue(Format::bitperpixel_param,16);
    out_format->addStringParamValue(Format::format_param,"packet");
    getOwnOutputProperty()->setDefaultFormat(out_format);

    /* Add the converter*/
    addConverter(in_format, out_format, &BGR16_5_6_5_LIT_to_RGB16_5_6_5_LIT);


    /* Input format for function RGB16_5_6_5_LIT_to_RGB24_8_8_8_LIT */
    in_format = getOwnInputProperty()->addNewFormat("video/raw");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addWildcard(Format::framerate_param);
    in_format->addStringParamValue(Format::channelorder_param,"rgb");
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::colorspace_param,"rgb565");
    in_format->addIntParamValue(Format::bitperpixel_param,16);
    in_format->addStringParamValue(Format::format_param,"packet");
 
    /* Output format for function RGB16_5_6_5_LIT_to_RGB24_8_8_8_LIT */
    out_format = getOwnOutputProperty()->addNewFormat("video/raw");
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addStringParamValue(Format::channelorder_param,"rgb");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::colorspace_param,"rgb24");
    out_format->addIntParamValue(Format::bitperpixel_param,24);
    out_format->addStringParamValue(Format::format_param,"packet");

    /* Add the converter*/
    addConverter(in_format, out_format, &RGB16_5_6_5_LIT_to_RGB24_8_8_8_LIT);


    /* Input format for function BGR16_5_6_5_LIT_to_RGB32_8_8_8_LIT */
    in_format = getOwnInputProperty()->addNewFormat("video/raw");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addWildcard(Format::framerate_param);
    in_format->addStringParamValue(Format::channelorder_param,"bgr");
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::colorspace_param,"rgb565");
    in_format->addIntParamValue(Format::bitperpixel_param,16);
    in_format->addStringParamValue(Format::format_param,"packet");
 
    /* Output format for function BGR16_5_6_5_LIT_to_RGB32_8_8_8_LIT */
    out_format = getOwnOutputProperty()->addNewFormat("video/raw");
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addStringParamValue(Format::channelorder_param,"rgb");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::colorspace_param,"rgb24");
    out_format->addIntParamValue(Format::bitperpixel_param,32);
    out_format->addStringParamValue(Format::format_param,"packet");
    /* Add the converter*/
    addConverter(in_format, out_format, &BGR16_5_6_5_LIT_to_RGB32_8_8_8_LIT);


    /* Input format for function RGB16_5_6_5_LIT_to_RGB32_8_8_8_LIT */
    in_format = getOwnInputProperty()->addNewFormat("video/raw");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addWildcard(Format::framerate_param);
    in_format->addStringParamValue(Format::channelorder_param,"rgb");
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::colorspace_param,"rgb565");
    in_format->addIntParamValue(Format::bitperpixel_param,16);
    in_format->addStringParamValue(Format::format_param,"packet");
 
    /* Output format for function RGB16_5_6_5_LIT_to_RGB32_8_8_8_LIT */
    out_format = getOwnOutputProperty()->addNewFormat("video/raw");
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addStringParamValue(Format::channelorder_param,"rgb");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::colorspace_param,"rgb24");
    out_format->addIntParamValue(Format::bitperpixel_param,32);
    out_format->addStringParamValue(Format::format_param,"packet");
    /* Add the converter*/
    addConverter(in_format, out_format, &RGB16_5_6_5_LIT_to_RGB32_8_8_8_LIT);


    /* Input format for function RGB24_8_8_8_LIT_to_RGB16_5_6_5_LIT */
    in_format = getOwnInputProperty()->addNewFormat("video/raw");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addWildcard(Format::framerate_param);
    in_format->addStringParamValue(Format::channelorder_param,"rgb");
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::colorspace_param,"rgb24");
    in_format->addIntParamValue(Format::bitperpixel_param,24);
    in_format->addStringParamValue(Format::format_param,"packet");

    /* Output format for function RGB24_8_8_8_LIT_to_RGB16_5_6_5_LIT */
    out_format = getOwnOutputProperty()->addNewFormat("video/raw");
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addStringParamValue(Format::channelorder_param,"rgb");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::colorspace_param,"rgb565");
    out_format->addIntParamValue(Format::bitperpixel_param,16);
    out_format->addStringParamValue(Format::format_param,"packet");
    /* Add the converter*/
    addConverter(in_format, out_format, &RGB24_8_8_8_LIT_to_RGB16_5_6_5_LIT);


    /* Input format for function BGR32_8_8_8_LIT_to_RGB32_8_8_8_LIT */
    in_format = getOwnInputProperty()->addNewFormat("video/raw");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addWildcard(Format::framerate_param);
    in_format->addStringParamValue(Format::channelorder_param,"bgr");
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::colorspace_param,"rgb24");
    in_format->addIntParamValue(Format::bitperpixel_param,32);
    in_format->addStringParamValue(Format::format_param,"packet");

    /* Output format for function BGR32_8_8_8_LIT_to_RGB32_8_8_8_LIT */
    out_format = getOwnOutputProperty()->addNewFormat("video/raw");
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addStringParamValue(Format::channelorder_param,"rgb");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::colorspace_param,"rgb24");
    out_format->addIntParamValue(Format::bitperpixel_param,32);
    out_format->addStringParamValue(Format::format_param,"packet");
    /* Add the converter*/
    addConverter(in_format, out_format, &BGR32_8_8_8_LIT_to_RGB32_8_8_8_LIT);


    /* Input format for function RGB24_8_8_8_LIT_to_RGBA32_8_8_8_8_LIT */
    in_format = getOwnInputProperty()->addNewFormat("video/raw");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addWildcard(Format::framerate_param);
    in_format->addStringParamValue(Format::channelorder_param,"rgb");
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::colorspace_param,"rgb24");
    in_format->addIntParamValue(Format::bitperpixel_param,24);
    in_format->addStringParamValue(Format::format_param,"packet");

    /* Output format for function RGB24_8_8_8_LIT_to_RGBA32_8_8_8_8_LIT */
    out_format = getOwnOutputProperty()->addNewFormat("video/raw");
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addStringParamValue(Format::channelorder_param,"rgba");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::colorspace_param,"rgba32");
    out_format->addIntParamValue(Format::bitperpixel_param,32);
    out_format->addStringParamValue(Format::format_param,"packet");
    /* Add the converter*/
    addConverter(in_format, out_format, &RGB24_8_8_8_LIT_to_RGBA32_8_8_8_8_LIT);

    /* Input format for function RGB24_8_8_8_LIT_to_RGB32_8_8_8_8_LIT */
    in_format = getOwnInputProperty()->addNewFormat("video/raw");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addWildcard(Format::framerate_param);
    in_format->addStringParamValue(Format::channelorder_param,"rgb");
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::colorspace_param,"rgb24");
    in_format->addIntParamValue(Format::bitperpixel_param,24);
    in_format->addStringParamValue(Format::format_param,"packet");

    /* Output format for function RGB24_8_8_8_LIT_to_RGB32_8_8_8_8_LIT */
    out_format = getOwnOutputProperty()->addNewFormat("video/raw");
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addStringParamValue(Format::channelorder_param,"rgb");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::colorspace_param,"rgb24");
    out_format->addIntParamValue(Format::bitperpixel_param,32);
    out_format->addStringParamValue(Format::format_param,"packet");
    /* Add the converter*/
    addConverter(in_format, out_format, &RGB24_8_8_8_LIT_to_RGBA32_8_8_8_8_LIT);

    /* Input format for function BGR24_8_8_8_LIT_to_BGR32_8_8_8_8_LIT */
    in_format = getOwnInputProperty()->addNewFormat("video/raw");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addWildcard(Format::framerate_param);
    in_format->addStringParamValue(Format::channelorder_param,"bgr");
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::colorspace_param,"rgb24");
    in_format->addIntParamValue(Format::bitperpixel_param,24);
    in_format->addStringParamValue(Format::format_param,"packet");

    /* Output format for function BGR24_8_8_8_LIT_to_BGR32_8_8_8_8_LIT */
    out_format = getOwnOutputProperty()->addNewFormat("video/raw");
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addStringParamValue(Format::channelorder_param,"rgb");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::colorspace_param,"rgb24");
    out_format->addIntParamValue(Format::bitperpixel_param,32);
    out_format->addStringParamValue(Format::format_param,"packet");
    /* Add the converter*/
    addConverter(in_format, out_format, &RGB24_8_8_8_LIT_to_BGRA32_8_8_8_8_LIT);

    /* Input format for function BGR24_8_8_8_LIT_to_RGB32_8_8_8_8_LIT */
    in_format = getOwnInputProperty()->addNewFormat("video/raw");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addWildcard(Format::framerate_param);
    in_format->addStringParamValue(Format::channelorder_param,"bgr");
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::colorspace_param,"rgb24");
    in_format->addIntParamValue(Format::bitperpixel_param,24);
    in_format->addStringParamValue(Format::format_param,"packet");

    /* Output format for function BGR24_8_8_8_LIT_to_RGB32_8_8_8_8_LIT */
    out_format = getOwnOutputProperty()->addNewFormat("video/raw");
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addStringParamValue(Format::channelorder_param,"rgb");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::colorspace_param,"rgb24");
    out_format->addIntParamValue(Format::bitperpixel_param,32);
    out_format->addStringParamValue(Format::format_param,"packet");
    /* Add the converter*/
    addConverter(in_format, out_format, &BGR24_8_8_8_LIT_to_RGB32_8_8_8_8_LIT);

    /* Input format for function RGB24_8_8_8_LIT_to_BGRA32_8_8_8_8_LIT */
    in_format = getOwnInputProperty()->addNewFormat("video/raw");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addWildcard(Format::framerate_param);
    in_format->addStringParamValue(Format::channelorder_param,"rgb");
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::colorspace_param,"rgb24");
    in_format->addIntParamValue(Format::bitperpixel_param,24);
    in_format->addStringParamValue(Format::format_param,"packet");
 
    /* Output format for function RGB24_8_8_8_LIT_to_BGRA32_8_8_8_8_LIT */
    out_format = getOwnOutputProperty()->addNewFormat("video/raw");
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addStringParamValue(Format::channelorder_param,"bgra");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::colorspace_param,"rgba32");
    out_format->addIntParamValue(Format::bitperpixel_param,32);
    out_format->addStringParamValue(Format::format_param,"packet");
    /* Add the converter*/
    addConverter(in_format, out_format, &RGB24_8_8_8_LIT_to_BGRA32_8_8_8_8_LIT);


    /* Input format for function RGB32_8_8_8_LIT_to_RGB16_5_6_5_LIT */
    in_format = getOwnInputProperty()->addNewFormat("video/raw");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addWildcard(Format::framerate_param);
    in_format->addStringParamValue(Format::channelorder_param,"rgb");
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::colorspace_param,"rgb24");
    in_format->addIntParamValue(Format::bitperpixel_param,32);
    in_format->addStringParamValue(Format::format_param,"packet");
 
    /* Output format for function RGB32_8_8_8_LIT_to_RGB16_5_6_5_LIT */
    out_format = getOwnOutputProperty()->addNewFormat("video/raw");
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addStringParamValue(Format::channelorder_param,"rgb");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::colorspace_param,"rgb565");
    out_format->addIntParamValue(Format::bitperpixel_param,16);
    out_format->addStringParamValue(Format::format_param,"packet");
    /* Add the converter*/
    addConverter(in_format, out_format, &RGB32_8_8_8_LIT_to_RGB16_5_6_5_LIT);


    /* Input format for function RGB32_8_8_8_LIT_to_BGR24_8_8_8_LIT */
    in_format = getOwnInputProperty()->addNewFormat("video/raw");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addWildcard(Format::framerate_param);
    in_format->addStringParamValue(Format::channelorder_param,"rgb");
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::colorspace_param,"rgb24");
    in_format->addIntParamValue(Format::bitperpixel_param,32);
    in_format->addStringParamValue(Format::format_param,"packet");
 
    /* Output format for function RGB32_8_8_8_LIT_to_BGR24_8_8_8_LIT */
    out_format = getOwnOutputProperty()->addNewFormat("video/raw");
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addStringParamValue(Format::channelorder_param,"bgr");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::colorspace_param,"rgb24");
    out_format->addIntParamValue(Format::bitperpixel_param,24);
    out_format->addStringParamValue(Format::format_param,"packet");
    /* Add the converter*/
    /*Since RGBA32_8_8_8_8_LIT_to_BGR24_8_8_8_LIT already does the desired work, 
      this converter is used instead of implementing a new one */
    addConverter(in_format, out_format, &RGBA32_8_8_8_8_LIT_to_BGR24_8_8_8_LIT);

    /* Input format for function BGR32_8_8_8_LIT_to_RGB16_5_6_5_LIT */
    in_format = getOwnInputProperty()->addNewFormat("video/raw");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addWildcard(Format::framerate_param);
    in_format->addStringParamValue(Format::channelorder_param,"bgr");
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::colorspace_param,"rgb24");
    in_format->addIntParamValue(Format::bitperpixel_param,32);
    in_format->addStringParamValue(Format::format_param,"packet");
 
    /* Output format for function BGR32_8_8_8_LIT_to_RGB16_5_6_5_LIT */
    out_format = getOwnOutputProperty()->addNewFormat("video/raw");
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addStringParamValue(Format::channelorder_param,"rgb");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::colorspace_param,"rgb565");
    out_format->addIntParamValue(Format::bitperpixel_param,16);
    out_format->addStringParamValue(Format::format_param,"packet");
    /* Add the converter*/
    addConverter(in_format, out_format, &BGR32_8_8_8_LIT_to_RGB16_5_6_5_LIT);


    /* Input format for function RGB32_11_11_10_LIT_to_RGB16_5_6_5_LIT */
    in_format = getOwnInputProperty()->addNewFormat("video/raw");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addWildcard(Format::framerate_param);
    in_format->addStringParamValue(Format::channelorder_param,"rgb");
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::colorspace_param,"rgb32");
    in_format->addIntParamValue(Format::bitperpixel_param,32);
    in_format->addStringParamValue(Format::format_param,"packet");

    /* Output format for function RGB32_11_11_10_LIT_to_RGB16_5_6_5_LIT */
    out_format = getOwnOutputProperty()->addNewFormat("video/raw");
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addStringParamValue(Format::channelorder_param,"rgb");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::colorspace_param,"rgb565");
    out_format->addIntParamValue(Format::bitperpixel_param,16);
    out_format->addStringParamValue(Format::format_param,"packet");

    /* Add the converter*/
    addConverter(in_format, out_format, &RGB32_11_11_10_LIT_to_RGB16_5_6_5_LIT);


    /* Input format for function BGRA32_8_8_8_8_LIT_to_RGB16_5_6_5_LIT */
    in_format = getOwnInputProperty()->addNewFormat("video/raw");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addWildcard(Format::framerate_param);
    in_format->addStringParamValue(Format::channelorder_param,"bgra");
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::colorspace_param,"rgba32");
    in_format->addIntParamValue(Format::bitperpixel_param,32);
    in_format->addStringParamValue(Format::format_param,"packet");

    /* Output format for function BGRA32_8_8_8_8_LIT_to_RGB16_5_6_5_LIT */
    out_format = getOwnOutputProperty()->addNewFormat("video/raw");
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addStringParamValue(Format::channelorder_param,"rgb");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::colorspace_param,"rgb565");
    out_format->addIntParamValue(Format::bitperpixel_param,16);
    out_format->addStringParamValue(Format::format_param,"packet");
    /* Add the converter*/
    addConverter(in_format, out_format, &BGRA32_8_8_8_8_LIT_to_RGB16_5_6_5_LIT);


    /* Input format for function RGBA32_8_8_8_8_LIT_to_RGB24_8_8_8_LIT */
    in_format = getOwnInputProperty()->addNewFormat("video/raw");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addWildcard(Format::framerate_param);
    in_format->addStringParamValue(Format::channelorder_param,"rgba");
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::colorspace_param,"rgba32");
    in_format->addIntParamValue(Format::bitperpixel_param,32);
    in_format->addStringParamValue(Format::format_param,"packet");

    /* Output format for function RGBA32_8_8_8_8_LIT_to_RGB24_8_8_8_LIT */
    out_format = getOwnOutputProperty()->addNewFormat("video/raw");
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addStringParamValue(Format::channelorder_param,"rgb");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::colorspace_param,"rgb24");
    out_format->addIntParamValue(Format::bitperpixel_param,24);
    out_format->addStringParamValue(Format::format_param,"packet");
    /* Add the converter*/
    addConverter(in_format, out_format, &RGBA32_8_8_8_8_LIT_to_RGB24_8_8_8_LIT);

    /* Input format for function RGBA32_8_8_8_8_LIT_to_BGR24_8_8_8_LIT */
    in_format = getOwnInputProperty()->addNewFormat("video/raw");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addWildcard(Format::framerate_param);
    in_format->addStringParamValue(Format::channelorder_param,"rgba");
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::colorspace_param,"rgba32");
    in_format->addIntParamValue(Format::bitperpixel_param,32);
    in_format->addStringParamValue(Format::format_param,"packet");

    /* Output format for function RGBA32_8_8_8_8_LIT_to_BGR24_8_8_8_LIT */
    out_format = getOwnOutputProperty()->addNewFormat("video/raw");
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addStringParamValue(Format::channelorder_param,"bgr");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::colorspace_param,"rgb24");
    out_format->addIntParamValue(Format::bitperpixel_param,24);
    out_format->addStringParamValue(Format::format_param,"packet");
    /* Add the converter*/
    addConverter(in_format, out_format, &RGBA32_8_8_8_8_LIT_to_BGR24_8_8_8_LIT); 

    /* Input format for function RGBA32_8_8_8_8_LIT_to_BGR24_8_8_8_LIT */
    in_format = getOwnInputProperty()->addNewFormat("video/raw");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addWildcard(Format::framerate_param);
    in_format->addStringParamValue(Format::channelorder_param,"rgba");
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::colorspace_param,"rgba32");
    in_format->addIntParamValue(Format::bitperpixel_param,32);
    in_format->addStringParamValue(Format::format_param,"packet");
    /* Output format for function RGBA32_8_8_8_8_LIT_to_RGB32_8_8_8_LIT */
    out_format = getOwnOutputProperty()->addNewFormat("video/raw");
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addStringParamValue(Format::channelorder_param,"rgb");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::colorspace_param,"rgb24");
    out_format->addIntParamValue(Format::bitperpixel_param,32);
    out_format->addStringParamValue(Format::format_param,"packet");
    /* Add the converter*/
    addConverter(in_format, out_format, dummy);

    /* Input format for function RGB32_11_11_10_LIT_to_RGB32_8_8_8_LIT */
    in_format = getOwnInputProperty()->addNewFormat("video/raw");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addWildcard(Format::framerate_param);
    in_format->addStringParamValue(Format::channelorder_param,"rgb");
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::colorspace_param,"rgb32");
    in_format->addIntParamValue(Format::bitperpixel_param,32);
    in_format->addStringParamValue(Format::format_param,"packet");

    /* Output format for function RGB32_11_11_10_LIT_to_RGB32_8_8_8_LIT */
    out_format = getOwnOutputProperty()->addNewFormat("video/raw");
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addStringParamValue(Format::channelorder_param,"rgb");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::colorspace_param,"rgb24");
    out_format->addIntParamValue(Format::bitperpixel_param,32);
    out_format->addStringParamValue(Format::format_param,"packet");

    /* Add the converter*/
    addConverter(in_format, out_format, &RGB32_11_11_10_LIT_to_RGB32_8_8_8_LIT);


    /* Input format for function BGR32_10_11_11_LIT_to_RGB32_8_8_8_LIT */
    in_format = getOwnInputProperty()->addNewFormat("video/raw");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addWildcard(Format::framerate_param);
    in_format->addStringParamValue(Format::channelorder_param,"bgr");
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addStringParamValue(Format::colorspace_param,"rgb32");
    in_format->addIntParamValue(Format::bitperpixel_param,32);
    in_format->addStringParamValue(Format::format_param,"packet");

    /* Output format for function BGR32_10_11_11_LIT_to_RGB32_8_8_8_LIT */
    out_format = getOwnOutputProperty()->addNewFormat("video/raw");
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addStringParamValue(Format::channelorder_param,"rgb");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::colorspace_param,"rgb24");
    out_format->addIntParamValue(Format::bitperpixel_param,32);
    out_format->addStringParamValue(Format::format_param,"packet");

    /* Add the converter*/
    addConverter(in_format, out_format, &BGR32_10_11_11_LIT_to_RGB32_8_8_8_LIT);
  }

  Result RGBtoRGBConverterNode::doInitOutput() {
    // get input format
    const Format* input_format = getInputFormat();
    if (input_format == NULL) {
      ERROR_STREAM("input format is null"<< endl);
      return FAILURE; }

    // setup output format
    Format* in_format;
    Format* out_format;

    // get format parameters
    int x_resolution = input_format->getIntValue(Format::x_resolution_param);
    int y_resolution = input_format->getIntValue(Format::y_resolution_param);
    int ratio_x = input_format->getIntValue(Format::ratio_x_param);
    int ratio_y = input_format->getIntValue(Format::ratio_y_param);
    int bitperpixel = input_format->getIntValue(Format::bitperpixel_param);
    float framerate = input_format->getFloatValue(Format::framerate_param);
    string channelorder = input_format->getStringValue(Format::channelorder_param);
    string colorspace = input_format->getStringValue(Format::colorspace_param);
    string format = input_format->getStringValue(Format::format_param);
    string endian = input_format->getStringValue(Format::endian_param);


    if (colorspace == string("rgb24")) {
      if (bitperpixel == 24) {
	/* Output format for function RGB24_8_8_8_LIT_to_RGBA32_8_8_8_8_LIT */
	out_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
	out_format->addIntParamValue(Format::x_resolution_param,x_resolution);
	out_format->addIntParamValue(Format::y_resolution_param,y_resolution);
	out_format->addIntParamValue(Format::ratio_x_param,ratio_x);
	out_format->addIntParamValue(Format::ratio_y_param,ratio_y);
	out_format->addFloatParamValue(Format::framerate_param, framerate);
	out_format->addStringParamValue(Format::channelorder_param,"rgba");
	out_format->addStringParamValue(Format::endian_param, endian);
	out_format->addStringParamValue(Format::colorspace_param,"rgba32");
	out_format->addIntParamValue(Format::bitperpixel_param,32);
	out_format->addStringParamValue(Format::format_param,format);

	in_format = getInputFormat()->duplicateFormat();
	getOwnWorkingInputProperty()->addNewFormat(in_format);

	in_format->setIOPartner(out_format);
	out_format->setIOPartner(in_format);
	
	getOwnWorkingInputProperty()->setDefaultFormat(in_format);
	getOwnWorkingOutputProperty()->setDefaultFormat(out_format);

	/* Output format for function RGB24_8_8_8_LIT_to_BGRA32_8_8_8_8_LIT */
	out_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
	out_format->addIntParamValue(Format::x_resolution_param,x_resolution);
	out_format->addIntParamValue(Format::y_resolution_param,y_resolution);
	out_format->addIntParamValue(Format::ratio_x_param,ratio_x);
	out_format->addIntParamValue(Format::ratio_y_param,ratio_y);
	out_format->addFloatParamValue(Format::framerate_param, framerate);
	out_format->addStringParamValue(Format::channelorder_param,"bgra");
	out_format->addStringParamValue(Format::endian_param, endian);
	out_format->addStringParamValue(Format::colorspace_param,"rgba32");
	out_format->addIntParamValue(Format::bitperpixel_param,32);
	out_format->addStringParamValue(Format::format_param,format);

	in_format = getInputFormat()->duplicateFormat();
	getOwnWorkingInputProperty()->addNewFormat(in_format);
	
	in_format->setIOPartner(out_format);
	out_format->setIOPartner(in_format);
	
	/* Output format for function RGB24_8_8_8_LIT_to_RGB16_5_6_5_LIT */
	out_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
	out_format->addIntParamValue(Format::x_resolution_param,x_resolution);
	out_format->addIntParamValue(Format::y_resolution_param,y_resolution);
	out_format->addIntParamValue(Format::ratio_x_param,ratio_x);
	out_format->addIntParamValue(Format::ratio_y_param,ratio_y);
	out_format->addFloatParamValue(Format::framerate_param, framerate);
	out_format->addStringParamValue(Format::channelorder_param,"rgb");
	out_format->addStringParamValue(Format::endian_param, endian);
	out_format->addStringParamValue(Format::colorspace_param,"rgb565");
	out_format->addIntParamValue(Format::bitperpixel_param,16);
	out_format->addStringParamValue(Format::format_param,format);

	in_format = getInputFormat()->duplicateFormat();
	getOwnWorkingInputProperty()->addNewFormat(in_format);
	
	in_format->setIOPartner(out_format);
	out_format->setIOPartner(in_format);

        /* Output format for function RGB24_8_8_8_LIT_to_RGB32_8_8_8_8_LIT */
	out_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
	out_format->addIntParamValue(Format::x_resolution_param,x_resolution);
	out_format->addIntParamValue(Format::y_resolution_param,y_resolution);
	out_format->addIntParamValue(Format::ratio_x_param,ratio_x);
	out_format->addIntParamValue(Format::ratio_y_param,ratio_y);
	out_format->addFloatParamValue(Format::framerate_param, framerate);
	out_format->addStringParamValue(Format::channelorder_param,"rgb");
	out_format->addStringParamValue(Format::endian_param, "little");
	out_format->addStringParamValue(Format::colorspace_param,"rgb24");
	out_format->addIntParamValue(Format::bitperpixel_param,32);
	out_format->addStringParamValue(Format::format_param,"packet");

	in_format = getInputFormat()->duplicateFormat();
	getOwnWorkingInputProperty()->addNewFormat(in_format);

	in_format->setIOPartner(out_format);
	out_format->setIOPartner(in_format);
      }
      else if (bitperpixel == 32) {

	/* Output format for function RGB32_8_8_8_LIT_to_RGB16_5_6_5_LIT */
	  /* Output format for function BGR32_8_8_8_LIT_to_RGB16_5_6_5_LIT */
	out_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
	out_format->addIntParamValue(Format::x_resolution_param,x_resolution);
	out_format->addIntParamValue(Format::y_resolution_param,y_resolution);
	out_format->addIntParamValue(Format::ratio_x_param,ratio_x);
	out_format->addIntParamValue(Format::ratio_y_param,ratio_y);
	out_format->addFloatParamValue(Format::framerate_param, framerate);
	out_format->addStringParamValue(Format::channelorder_param,"rgb");
	out_format->addStringParamValue(Format::endian_param, endian);
	out_format->addStringParamValue(Format::colorspace_param,"rgb565");
	out_format->addIntParamValue(Format::bitperpixel_param,16);
	out_format->addStringParamValue(Format::format_param,format);


	in_format = getInputFormat()->duplicateFormat();
	getOwnWorkingInputProperty()->addNewFormat(in_format);
	
	in_format->setIOPartner(out_format);
	out_format->setIOPartner(in_format);

	getOwnWorkingInputProperty()->setDefaultFormat(in_format);
	getOwnWorkingOutputProperty()->setDefaultFormat(out_format);

	/* Output format for function BGR32_8_8_8_LIT_to_RGB32_8_8_8_LIT */
	out_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
	out_format->addIntParamValue(Format::x_resolution_param,x_resolution);
	out_format->addIntParamValue(Format::y_resolution_param,y_resolution);
	out_format->addIntParamValue(Format::ratio_x_param,ratio_x);
	out_format->addIntParamValue(Format::ratio_y_param,ratio_y);
	out_format->addFloatParamValue(Format::framerate_param, framerate);
	out_format->addStringParamValue(Format::channelorder_param,"rgb");
	out_format->addStringParamValue(Format::endian_param, endian);
	out_format->addStringParamValue(Format::colorspace_param,"rgb24");
	out_format->addIntParamValue(Format::bitperpixel_param,32);
	out_format->addStringParamValue(Format::format_param,format);

	in_format = getInputFormat()->duplicateFormat();
	getOwnWorkingInputProperty()->addNewFormat(in_format);
	
	in_format->setIOPartner(out_format);
	out_format->setIOPartner(in_format);

       
	/* Output format for function RGB32_8_8_8_LIT_to_BGR24_8_8_8_LIT */
	out_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
	out_format->addIntParamValue(Format::x_resolution_param,x_resolution);
	out_format->addIntParamValue(Format::y_resolution_param,y_resolution);
	out_format->addIntParamValue(Format::ratio_x_param,ratio_x);
	out_format->addIntParamValue(Format::ratio_y_param,ratio_y);
	out_format->addFloatParamValue(Format::framerate_param, framerate);
	out_format->addStringParamValue(Format::channelorder_param,"bgr");
	out_format->addStringParamValue(Format::endian_param, endian);
	out_format->addStringParamValue(Format::colorspace_param,"rgb24");
	out_format->addIntParamValue(Format::bitperpixel_param,24);
	out_format->addStringParamValue(Format::format_param,format);

	in_format = getInputFormat()->duplicateFormat();
	getOwnWorkingInputProperty()->addNewFormat(in_format);
	
	in_format->setIOPartner(out_format);
	out_format->setIOPartner(in_format);
      }
      
      return SUCCESS;
    }
    else if (colorspace == string("rgba32")) {
      /* Output format for function RGBA32_8_8_8_8_LIT_to_RGB24_8_8_8_LIT */
      out_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
      out_format->addIntParamValue(Format::x_resolution_param,x_resolution);
      out_format->addIntParamValue(Format::y_resolution_param,y_resolution);
      out_format->addIntParamValue(Format::ratio_x_param,ratio_x);
      out_format->addIntParamValue(Format::ratio_y_param,ratio_y);
      out_format->addFloatParamValue(Format::framerate_param, framerate);
      out_format->addStringParamValue(Format::channelorder_param,"rgb");
      out_format->addStringParamValue(Format::endian_param, endian);
      out_format->addStringParamValue(Format::colorspace_param,"rgb24");
      out_format->addIntParamValue(Format::bitperpixel_param,24);
      out_format->addStringParamValue(Format::format_param,format);

      in_format = getInputFormat()->duplicateFormat();
      getOwnWorkingInputProperty()->addNewFormat(in_format);
	
      in_format->setIOPartner(out_format);
      out_format->setIOPartner(in_format);

      getOwnWorkingInputProperty()->setDefaultFormat(in_format);
      getOwnWorkingOutputProperty()->setDefaultFormat(out_format);

      /* Output format for function BGRA32_8_8_8_8_LIT_to_RGB16_5_6_5_LIT */
      out_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
      out_format->addIntParamValue(Format::x_resolution_param,x_resolution);
      out_format->addIntParamValue(Format::y_resolution_param,y_resolution);
      out_format->addIntParamValue(Format::ratio_x_param,ratio_x);
      out_format->addIntParamValue(Format::ratio_y_param,ratio_y);
      out_format->addFloatParamValue(Format::framerate_param, framerate);
      out_format->addStringParamValue(Format::channelorder_param,"rgb");
      out_format->addStringParamValue(Format::endian_param, endian);
      out_format->addStringParamValue(Format::colorspace_param,"rgb565");
      out_format->addIntParamValue(Format::bitperpixel_param,16);
      out_format->addStringParamValue(Format::format_param,format);

      in_format = getInputFormat()->duplicateFormat();
      getOwnWorkingInputProperty()->addNewFormat(in_format);

      in_format->setIOPartner(out_format);
      out_format->setIOPartner(in_format);

      /* Output format for function RGBA32_8_8_8_8_LIT_to_BGR24_8_8_8_LIT */
      out_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
      out_format->addIntParamValue(Format::x_resolution_param,x_resolution);
      out_format->addIntParamValue(Format::y_resolution_param,y_resolution);
      out_format->addIntParamValue(Format::ratio_x_param,ratio_x);
      out_format->addIntParamValue(Format::ratio_y_param,ratio_y);
      out_format->addFloatParamValue(Format::framerate_param, framerate);
      out_format->addStringParamValue(Format::channelorder_param,"bgr");
      out_format->addStringParamValue(Format::endian_param, endian);
      out_format->addStringParamValue(Format::colorspace_param,"rgb24");
      out_format->addIntParamValue(Format::bitperpixel_param,24);
      out_format->addStringParamValue(Format::format_param,format);

      in_format = getInputFormat()->duplicateFormat();
      getOwnWorkingInputProperty()->addNewFormat(in_format);

      in_format->setIOPartner(out_format);
      out_format->setIOPartner(in_format);

      //chwellner
      /* Output format for function RGBA32_8_8_8_8_LIT_to_BGR24_8_8_8_LIT */
      out_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
      out_format->addIntParamValue(Format::x_resolution_param,x_resolution);
      out_format->addIntParamValue(Format::y_resolution_param,y_resolution);
      out_format->addIntParamValue(Format::ratio_x_param,ratio_x);
      out_format->addIntParamValue(Format::ratio_y_param,ratio_y);
      out_format->addFloatParamValue(Format::framerate_param, framerate);
      out_format->addStringParamValue(Format::channelorder_param,"rgb");
      out_format->addStringParamValue(Format::endian_param, endian);
      out_format->addStringParamValue(Format::colorspace_param,"rgb24");
      out_format->addIntParamValue(Format::bitperpixel_param,32);
      out_format->addStringParamValue(Format::format_param,format);

      in_format = getInputFormat()->duplicateFormat();
      getOwnWorkingInputProperty()->addNewFormat(in_format);

      in_format->setIOPartner(out_format);
      out_format->setIOPartner(in_format);

      return SUCCESS;
    }
    else if (colorspace == string("rgb565")) {
      /* Output format for function RGB16_5_6_5_LIT_to_RGB32_8_8_8_LIT */
      /* Output format for function BGR16_5_6_5_LIT_to_RGB32_8_8_8_LIT */
      out_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
      out_format->addIntParamValue(Format::x_resolution_param,x_resolution);
      out_format->addIntParamValue(Format::y_resolution_param,y_resolution);
      out_format->addIntParamValue(Format::ratio_x_param,ratio_x);
      out_format->addIntParamValue(Format::ratio_y_param,ratio_y);
      out_format->addFloatParamValue(Format::framerate_param, framerate);
      out_format->addStringParamValue(Format::channelorder_param,"rgb");
      out_format->addStringParamValue(Format::endian_param, endian);
      out_format->addStringParamValue(Format::colorspace_param,"rgb24");
      out_format->addIntParamValue(Format::bitperpixel_param,32);
      out_format->addStringParamValue(Format::format_param,format);

      in_format = getInputFormat()->duplicateFormat();
      getOwnWorkingInputProperty()->addNewFormat(in_format);

      in_format->setIOPartner(out_format);
      out_format->setIOPartner(in_format);

      getOwnWorkingInputProperty()->setDefaultFormat(in_format);
      getOwnWorkingOutputProperty()->setDefaultFormat(out_format);

      /* Output format for function RGB16_5_6_5_LIT_to_RGB24_8_8_8_LIT */
      out_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
      out_format->addIntParamValue(Format::x_resolution_param,x_resolution);
      out_format->addIntParamValue(Format::y_resolution_param,y_resolution);
      out_format->addIntParamValue(Format::ratio_x_param,ratio_x);
      out_format->addIntParamValue(Format::ratio_y_param,ratio_y);
      out_format->addFloatParamValue(Format::framerate_param, framerate);
      out_format->addStringParamValue(Format::channelorder_param,"rgb");
      out_format->addStringParamValue(Format::endian_param, endian);
      out_format->addStringParamValue(Format::colorspace_param,"rgb24");
      out_format->addIntParamValue(Format::bitperpixel_param,24);
      out_format->addStringParamValue(Format::format_param,format);

      in_format = getInputFormat()->duplicateFormat();
      getOwnWorkingInputProperty()->addNewFormat(in_format);

      in_format->setIOPartner(out_format);
      out_format->setIOPartner(in_format);

      /* Output format for function BGR16_5_6_5_LIT_to_RGB16_5_6_5_LIT */
      out_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
      out_format->addIntParamValue(Format::x_resolution_param,x_resolution);
      out_format->addIntParamValue(Format::y_resolution_param,y_resolution);
      out_format->addIntParamValue(Format::ratio_x_param,ratio_x);
      out_format->addIntParamValue(Format::ratio_y_param,ratio_y);
      out_format->addFloatParamValue(Format::framerate_param, framerate);
      out_format->addStringParamValue(Format::channelorder_param,"rgb");
      out_format->addStringParamValue(Format::endian_param, endian);
      out_format->addStringParamValue(Format::colorspace_param,"rgb565");
      out_format->addIntParamValue(Format::bitperpixel_param,16);
      out_format->addStringParamValue(Format::format_param,format);

      in_format = getInputFormat()->duplicateFormat();
      getOwnWorkingInputProperty()->addNewFormat(in_format);

      in_format->setIOPartner(out_format);
      out_format->setIOPartner(in_format);

      return SUCCESS;
    } else if (colorspace == string("rgb32")) {
      
      /* Output format for function RGB32_11_11_10_LIT_to_RGB32_8_8_8_LIT */
      /* Output format for function BGR32_10_11_11_LIT_to_RGB32_8_8_8_LIT */
      out_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
      out_format->addIntParamValue(Format::x_resolution_param,x_resolution);
      out_format->addIntParamValue(Format::y_resolution_param,y_resolution);
      out_format->addIntParamValue(Format::ratio_x_param,ratio_x);
      out_format->addIntParamValue(Format::ratio_y_param,ratio_y);
      out_format->addFloatParamValue(Format::framerate_param, framerate);
      out_format->addStringParamValue(Format::channelorder_param,"rgb");
      out_format->addStringParamValue(Format::endian_param, endian);
      out_format->addStringParamValue(Format::colorspace_param,"rgb24");
      out_format->addIntParamValue(Format::bitperpixel_param,32);
      out_format->addStringParamValue(Format::format_param,format);

      in_format = getInputFormat()->duplicateFormat();
      getOwnWorkingInputProperty()->addNewFormat(in_format);

      in_format->setIOPartner(out_format);
      out_format->setIOPartner(in_format);

      getOwnWorkingInputProperty()->setDefaultFormat(in_format);
      getOwnWorkingOutputProperty()->setDefaultFormat(out_format);

      /* Output format for function RGB32_11_11_10_LIT_to_RGB16_5_6_5_LIT */
      out_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
      out_format->addIntParamValue(Format::x_resolution_param,x_resolution);
      out_format->addIntParamValue(Format::y_resolution_param,y_resolution);
      out_format->addIntParamValue(Format::ratio_x_param,ratio_x);
      out_format->addIntParamValue(Format::ratio_y_param,ratio_y);
      out_format->addFloatParamValue(Format::framerate_param, framerate);
      out_format->addStringParamValue(Format::channelorder_param,"rgb");
      out_format->addStringParamValue(Format::endian_param, endian);
      out_format->addStringParamValue(Format::colorspace_param,"rgb565");
      out_format->addIntParamValue(Format::bitperpixel_param,16);
      out_format->addStringParamValue(Format::format_param,format);

      in_format = getInputFormat()->duplicateFormat();
      getOwnWorkingInputProperty()->addNewFormat(in_format);

      in_format->setIOPartner(out_format);
      out_format->setIOPartner(in_format);
      
      return SUCCESS;
    }

    return FAILURE;
  }

  Result RGBtoRGBConverterNode::doDeinitOutput() {
    // erase working properties
    getOwnWorkingInputProperty()->erase();
    getOwnWorkingOutputProperty()->erase();
    return SUCCESS;
  }

}
