/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
//#include <unistd.h>

#include "VideoCropNode.hpp"
#include "nmm/comm/format/Property.hpp"
#include "nmm/comm/messaging/Buffer.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"

NMM_REGISTER_PLUGIN(VideoCropNode, NMM, "VideoCropNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  VideoCropNode::VideoCropNode(const char *name,
				   const StreamQueue::Mode qmode,
				   const int qsize) :
    GenericProcessorNode(name, qmode, qsize)
  {
    send_format = false;
    saved_buffer = NULL;
    x_offset = 0;
    y_offset = 0;
    x_range = 0;
    y_range = 0;
  }
  
  VideoCropNode::~VideoCropNode() {}
  
  Result VideoCropNode::doInit() {
    // register all supported input formats
    Format* in_format_yv12 = getOwnInputProperty()->addNewFormat(Format::video_raw_type);
    in_format_yv12->addWildcard(Format::x_resolution_param);
    in_format_yv12->addWildcard(Format::y_resolution_param);
    in_format_yv12->addWildcard(Format::ratio_x_param);
    in_format_yv12->addWildcard(Format::ratio_y_param);
    in_format_yv12->addWildcard(Format::framerate_param);
    in_format_yv12->addStringParamValue(Format::endian_param,"little");
    in_format_yv12->addStringParamValue(Format::format_param,"planar");
    in_format_yv12->addStringParamValue(Format::channelorder_param,"yvu");
    in_format_yv12->addStringParamValue(Format::channelorder_param,"yuv");
    in_format_yv12->addStringParamValue(Format::colorspace_param,"yv12");
    in_format_yv12->addIntParamValue(Format::bitperpixel_param, 12);
    getOwnInputProperty()->setDefaultFormat(in_format_yv12);
    
    // register all supported output formats
    Format* out_format_yv12 = getOwnOutputProperty()->addNewFormat(Format::video_raw_type);
    out_format_yv12->addWildcard(Format::x_resolution_param);
    out_format_yv12->addWildcard(Format::y_resolution_param);
    out_format_yv12->addWildcard(Format::ratio_x_param);
    out_format_yv12->addWildcard(Format::ratio_y_param);
    out_format_yv12->addWildcard(Format::framerate_param);
    out_format_yv12->addStringParamValue(Format::endian_param,"little");
    out_format_yv12->addStringParamValue(Format::format_param,"planar");
    out_format_yv12->addStringParamValue(Format::channelorder_param,"yvu");
    out_format_yv12->addStringParamValue(Format::channelorder_param,"yuv");
    out_format_yv12->addStringParamValue(Format::colorspace_param,"yv12");
    out_format_yv12->addIntParamValue(Format::bitperpixel_param, 12);
    getOwnOutputProperty()->setDefaultFormat(out_format_yv12);
    
    // set I/O partners
    in_format_yv12->setIOPartner(out_format_yv12);
    out_format_yv12->setIOPartner(in_format_yv12);
    
    return SUCCESS;
  }
  
  Result VideoCropNode::doInitOutput() {
    // get input format
    const Format* in_format = getInputFormat();
    if (in_format == NULL) {
      ERROR_STREAM("No input format" << endl);
      return FAILURE;
    }

    // get format parameters
    x_resolution = in_format->getIntValue(Format::x_resolution_param);
    y_resolution = in_format->getIntValue(Format::y_resolution_param);
    ratio_x = in_format->getIntValue(Format::ratio_x_param);
    ratio_y = in_format->getIntValue(Format::ratio_y_param);
    framerate = in_format->getFloatValue(Format::framerate_param);
    channelorder = in_format->getStringValue(Format::channelorder_param);
    colorspace = in_format->getStringValue(Format::colorspace_param);
    format = in_format->getStringValue(Format::format_param);
    endian = in_format->getStringValue(Format::endian_param);
    bitperpixel = in_format->getIntValue(Format::bitperpixel_param);

    // clip crop window
    if (x_range <= 0) {
      x_range = x_resolution;
    }
    if (y_range <= 0) {
      y_range = y_resolution;
    }
    if (x_offset < 0) {
      x_offset = 0;
    }
    if (y_offset < 0) {
      y_offset = 0;
    }
    if (x_offset >= x_resolution) {
      x_offset = x_resolution - 1;
    }
    if (y_offset >= y_resolution) {
      y_offset = y_resolution - 1;
    }
    if (x_offset + x_range > x_resolution) {
      x_range = x_resolution - x_offset;
    }
    if (y_offset + y_range > y_resolution) {
      y_range = y_resolution - y_offset;
    }

    // erase working properties
    getOwnWorkingInputProperty()->erase();
    getOwnWorkingOutputProperty()->erase();
    
    // general working input format
    // (resolution and aspect ratio may change)
    Format* i_format_gen = getOwnWorkingInputProperty()->addNewFormat("video/raw");
    i_format_gen->addWildcard(Format::x_resolution_param);
    i_format_gen->addWildcard(Format::y_resolution_param);
    i_format_gen->addWildcard(Format::ratio_x_param);
    i_format_gen->addWildcard(Format::ratio_y_param);
    i_format_gen->addWildcard(Format::framerate_param);
    i_format_gen->addStringParamValue(Format::channelorder_param, channelorder);
    i_format_gen->addStringParamValue(Format::colorspace_param, colorspace);
    i_format_gen->addStringParamValue(Format::format_param, format);
    i_format_gen->addStringParamValue(Format::endian_param, endian);
    i_format_gen->addIntParamValue(Format::bitperpixel_param, bitperpixel);
    
    // current working input format
    // (the current connection format)
    Format* i_format = new Format(*in_format);
    getOwnWorkingInputProperty()->addNewFormat(i_format);
    getOwnWorkingInputProperty()->setDefaultFormat(i_format);
    
    // general working output format
    // (arbitrary resolution)
    Format* o_format_gen = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
    o_format_gen->addWildcard(Format::x_resolution_param);
    o_format_gen->addWildcard(Format::y_resolution_param);
    o_format_gen->addWildcard(Format::ratio_x_param);
    o_format_gen->addWildcard(Format::ratio_y_param);
    o_format_gen->addFloatParamValue(Format::framerate_param, framerate);
    o_format_gen->addStringParamValue(Format::channelorder_param, channelorder);
    o_format_gen->addStringParamValue(Format::colorspace_param, colorspace);
    o_format_gen->addStringParamValue(Format::format_param, format);
    o_format_gen->addStringParamValue(Format::endian_param, endian);
    o_format_gen->addIntParamValue(Format::bitperpixel_param, bitperpixel);

    
    // default output format
    // (cropped resolution)
    int cropped_ratio_x = (((x_range * ratio_x * 100) / x_resolution) / 2) * 2;
    int cropped_ratio_y = (((y_range * ratio_y * 100) / y_resolution) / 2) * 2;

    Format* o_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
    o_format->addIntParamValue(Format::x_resolution_param, x_range);
    o_format->addIntParamValue(Format::y_resolution_param, y_range);
    o_format->addIntParamValue(Format::ratio_x_param, cropped_ratio_x);
    o_format->addIntParamValue(Format::ratio_y_param, cropped_ratio_y);
    o_format->addFloatParamValue(Format::framerate_param, framerate);
    o_format->addStringParamValue(Format::channelorder_param, channelorder);
    o_format->addStringParamValue(Format::colorspace_param, colorspace);
    o_format->addStringParamValue(Format::format_param, format);
    o_format->addStringParamValue(Format::endian_param, endian);
    o_format->addIntParamValue(Format::bitperpixel_param, bitperpixel);
    getOwnWorkingOutputProperty()->setDefaultFormat(o_format);
    
    // set I/O partners
    i_format->setIOPartner(o_format);
    o_format->setIOPartner(i_format);
    i_format_gen->setIOPartner(o_format_gen);
    o_format_gen->setIOPartner(i_format_gen);
    
    return SUCCESS;
  }
  
  Message* VideoCropNode::processBuffer(Buffer *in_buffer) {
    if (getWorkingFlag()) {
      setWorkingFlag(false);
      in_buffer = saved_buffer;
    }
    
    if (send_format) {
      // clear flag
      send_format = false;
      
      // store buffer
      saved_buffer = in_buffer;
      setWorkingFlag(true);
      
      // send new output format
      Format new_format("video/raw");
      new_format.addIntParamValue(Format::x_resolution_param, x_range);
      new_format.addIntParamValue(Format::y_resolution_param, y_range);
      new_format.addIntParamValue(Format::ratio_x_param, ratio_x);
      new_format.addIntParamValue(Format::ratio_y_param, ratio_y);
      new_format.addFloatParamValue(Format::framerate_param, framerate);
      new_format.addStringParamValue(Format::channelorder_param, channelorder);
      new_format.addStringParamValue(Format::colorspace_param, colorspace);
      new_format.addStringParamValue(Format::format_param, format);
      new_format.addStringParamValue(Format::endian_param, endian);
      new_format.addIntParamValue(Format::bitperpixel_param, bitperpixel);
      Event event(IFormatChange::setFormat_event, new TInValue<Format>(new_format));
      return new CEvent(event);
    }
    
    // allocate buffer
    int out_buffer_size = x_range * y_range * bitperpixel / 8;
    Buffer* out_buffer = getNewBuffer(out_buffer_size, in_buffer);

    if (NULL == out_buffer) {
      DEBUG_STREAM("failed to allocate output buffer" << endl);
      in_buffer->release();
      return NULL;
    }

    out_buffer->setUsed(out_buffer_size);
    out_buffer->setTimestamp(in_buffer->getTimestamp());
    
    // process data
    char* in_data = in_buffer->getData();
    char* out_data = out_buffer->getData();

    // crop planes

    // first the Y plane
    for (int y = 0; y < y_range; y++) {
      memcpy(&out_data[y * x_range], &in_data[x_offset + (y + y_offset) * x_resolution], x_range);
    }

    // compute various offsets
    int crop_plane_offset1 = x_range * y_range;
    int plane_offset1 = x_resolution * y_resolution;
    int crop_plane_offset2 = crop_plane_offset1 + (crop_plane_offset1 >> 2);
    int plane_offset2 = plane_offset1 + (plane_offset1 >> 2);
    int uv_pitch = x_resolution >> 1;
    int crop_uv_pitch = x_range >> 1;
    int offset1 = (x_offset >> 1) + plane_offset1;
    int offset2 = (x_offset >> 1) + plane_offset2;

    // then the VU planes
    for (int y = 0; y < (y_range >> 1); y++) {
      memcpy(&out_data[y * crop_uv_pitch + crop_plane_offset1],
	     &in_data[offset1 + ((y + (y_offset >> 1)) * uv_pitch)], crop_uv_pitch);
      memcpy(&out_data[y * crop_uv_pitch + crop_plane_offset2],
	     &in_data[offset2 + ((y + (y_offset >> 1)) * uv_pitch)], crop_uv_pitch);
    }
    
    // done
    in_buffer->release();
    return out_buffer;
  }

  void VideoCropNode::setCrop(const nmm_int32& xOffset, const nmm_int32& yOffset, const nmm_int32& xRange, const nmm_int32& yRange) {
    // because of YV12 only multiples of 2 are allowed
    x_offset = xOffset / 2 * 2;
    y_offset = yOffset / 2 * 2;
    x_range = xRange / 2 * 2;
    y_range = yRange / 2 * 2;
  }
}

#include "nmm/nmm_suffix.hpp"
