/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#ifndef WIN32
#include <unistd.h>
#endif 
#include "VideoScalerNode.hpp"
#include "nmm/comm/format/Property.hpp"
#include "nmm/comm/messaging/Buffer.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"

NMM_REGISTER_PLUGIN(VideoScalerNode, NMM, "VideoScalerNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  VideoScalerNode::VideoScalerNode(const char *name,
				   const StreamQueue::Mode qmode,
				   const int qsize) :
    GenericProcessorNode(name, qmode, qsize)
  {
    keep_aspect_ratio = false;
    send_format = false;
    saved_buffer = NULL;
  }
  
  VideoScalerNode::~VideoScalerNode() {}
  
  Result VideoScalerNode::doInit() {
    // register all supported input formats
    Format* in_format_yv12 = getOwnInputProperty()->addNewFormat(Format::video_raw_type);
    in_format_yv12->addWildcard(Format::x_resolution_param);
    in_format_yv12->addWildcard(Format::y_resolution_param);
    in_format_yv12->addWildcard(Format::ratio_x_param);
    in_format_yv12->addWildcard(Format::ratio_y_param);
    in_format_yv12->addWildcard(Format::framerate_param);
    in_format_yv12->addStringParamValue(Format::endian_param,"little");
    in_format_yv12->addStringParamValue(Format::format_param,"planar");
    in_format_yv12->addStringParamValue(Format::channelorder_param,"yvu");
    in_format_yv12->addStringParamValue(Format::channelorder_param,"yuv");
    in_format_yv12->addStringParamValue(Format::colorspace_param,"yv12");
    in_format_yv12->addIntParamValue(Format::bitperpixel_param, 12);
    getOwnInputProperty()->setDefaultFormat(in_format_yv12);
    
    Format* in_format_yv12a = getOwnInputProperty()->addNewFormat(Format::video_raw_type);
    in_format_yv12a->addWildcard(Format::x_resolution_param);
    in_format_yv12a->addWildcard(Format::y_resolution_param);
    in_format_yv12a->addWildcard(Format::ratio_x_param);
    in_format_yv12a->addWildcard(Format::ratio_y_param);
    in_format_yv12a->addWildcard(Format::framerate_param);
    in_format_yv12a->addStringParamValue(Format::endian_param,"little");
    in_format_yv12a->addStringParamValue(Format::format_param,"planar");
    in_format_yv12a->addStringParamValue(Format::channelorder_param,"yvua");
    in_format_yv12a->addStringParamValue(Format::channelorder_param,"yuva");
    in_format_yv12a->addStringParamValue(Format::colorspace_param,"yv12a");
    in_format_yv12a->addIntParamValue(Format::bitperpixel_param, 20);
    
    Format* in_format_rgb = getOwnInputProperty()->addNewFormat(Format::video_raw_type);
    in_format_rgb->addWildcard(Format::x_resolution_param);
    in_format_rgb->addWildcard(Format::y_resolution_param);
    in_format_rgb->addWildcard(Format::ratio_x_param);
    in_format_rgb->addWildcard(Format::ratio_y_param);
    in_format_rgb->addWildcard(Format::framerate_param);
    in_format_rgb->addStringParamValue(Format::endian_param,"little");
    in_format_rgb->addStringParamValue(Format::format_param,"packet");
    in_format_rgb->addStringParamValue(Format::colorspace_param,"rgb");
    in_format_rgb->addStringParamValue(Format::channelorder_param,"rgb");
    in_format_rgb->addStringParamValue(Format::channelorder_param,"bgr");
    in_format_rgb->addIntParamValue(Format::bitperpixel_param, 24);
    
    Format* in_format_rgba = getOwnInputProperty()->addNewFormat(Format::video_raw_type);
    in_format_rgba->addWildcard(Format::x_resolution_param);
    in_format_rgba->addWildcard(Format::y_resolution_param);
    in_format_rgba->addWildcard(Format::ratio_x_param);
    in_format_rgba->addWildcard(Format::ratio_y_param);
    in_format_rgba->addWildcard(Format::framerate_param);
    in_format_rgba->addStringParamValue(Format::endian_param,"little");
    in_format_rgba->addStringParamValue(Format::format_param,"packet");
    in_format_rgba->addStringParamValue(Format::colorspace_param,"rgba");
    in_format_rgba->addStringParamValue(Format::channelorder_param,"rgba");
    in_format_rgba->addStringParamValue(Format::channelorder_param,"bgra");
    in_format_rgba->addStringParamValue(Format::channelorder_param,"argb");
    in_format_rgba->addStringParamValue(Format::channelorder_param,"abgr");
    in_format_rgba->addIntParamValue(Format::bitperpixel_param, 32);
    
    // register all supported output formats
    Format* out_format_yv12 = getOwnOutputProperty()->addNewFormat(Format::video_raw_type);
    out_format_yv12->addWildcard(Format::x_resolution_param);
    out_format_yv12->addWildcard(Format::y_resolution_param);
    out_format_yv12->addWildcard(Format::ratio_x_param);
    out_format_yv12->addWildcard(Format::ratio_y_param);
    out_format_yv12->addWildcard(Format::framerate_param);
    out_format_yv12->addStringParamValue(Format::endian_param,"little");
    out_format_yv12->addStringParamValue(Format::format_param,"planar");
    out_format_yv12->addStringParamValue(Format::channelorder_param,"yvu");
    out_format_yv12->addStringParamValue(Format::channelorder_param,"yuv");
    out_format_yv12->addStringParamValue(Format::colorspace_param,"yv12");
    out_format_yv12->addIntParamValue(Format::bitperpixel_param, 12);
    getOwnOutputProperty()->setDefaultFormat(out_format_yv12);
    
    Format* out_format_yv12a = getOwnOutputProperty()->addNewFormat(Format::video_raw_type);
    out_format_yv12a->addWildcard(Format::x_resolution_param);
    out_format_yv12a->addWildcard(Format::y_resolution_param);
    out_format_yv12a->addWildcard(Format::ratio_x_param);
    out_format_yv12a->addWildcard(Format::ratio_y_param);
    out_format_yv12a->addWildcard(Format::framerate_param);
    out_format_yv12a->addStringParamValue(Format::endian_param,"little");
    out_format_yv12a->addStringParamValue(Format::format_param,"planar");
    out_format_yv12a->addStringParamValue(Format::channelorder_param,"yvua");
    out_format_yv12a->addStringParamValue(Format::channelorder_param,"yuva");
    out_format_yv12a->addStringParamValue(Format::colorspace_param,"yv12a");
    out_format_yv12a->addIntParamValue(Format::bitperpixel_param, 20);
    
    Format* out_format_rgb = getOwnOutputProperty()->addNewFormat(Format::video_raw_type);
    out_format_rgb->addWildcard(Format::x_resolution_param);
    out_format_rgb->addWildcard(Format::y_resolution_param);
    out_format_rgb->addWildcard(Format::ratio_x_param);
    out_format_rgb->addWildcard(Format::ratio_y_param);
    out_format_rgb->addWildcard(Format::framerate_param);
    out_format_rgb->addStringParamValue(Format::endian_param,"little");
    out_format_rgb->addStringParamValue(Format::format_param,"packet");
    out_format_rgb->addStringParamValue(Format::colorspace_param,"rgb");
    out_format_rgb->addStringParamValue(Format::channelorder_param,"rgb");
    out_format_rgb->addStringParamValue(Format::channelorder_param,"bgr");
    out_format_rgb->addIntParamValue(Format::bitperpixel_param, 24);
    
    Format* out_format_rgba = getOwnOutputProperty()->addNewFormat(Format::video_raw_type);
    out_format_rgba->addWildcard(Format::x_resolution_param);
    out_format_rgba->addWildcard(Format::y_resolution_param);
    out_format_rgba->addWildcard(Format::ratio_x_param);
    out_format_rgba->addWildcard(Format::ratio_y_param);
    out_format_rgba->addWildcard(Format::framerate_param);
    out_format_rgba->addStringParamValue(Format::endian_param,"little");
    out_format_rgba->addStringParamValue(Format::format_param,"packet");
    out_format_rgba->addStringParamValue(Format::colorspace_param,"rgba");
    out_format_rgba->addStringParamValue(Format::channelorder_param,"rgba");
    out_format_rgba->addStringParamValue(Format::channelorder_param,"bgra");
    out_format_rgba->addStringParamValue(Format::channelorder_param,"argb");
    out_format_rgba->addStringParamValue(Format::channelorder_param,"abgr");
    out_format_rgba->addIntParamValue(Format::bitperpixel_param, 32);
    
    // set I/O partners
    in_format_yv12->setIOPartner(out_format_yv12);
    out_format_yv12->setIOPartner(in_format_yv12);
    in_format_yv12a->setIOPartner(out_format_yv12a);
    out_format_yv12a->setIOPartner(in_format_yv12a);
    in_format_rgb->setIOPartner(out_format_rgb);
    out_format_rgb->setIOPartner(in_format_rgb);
    in_format_rgba->setIOPartner(out_format_rgba);
    out_format_rgba->setIOPartner(in_format_rgba);
    
    return SUCCESS;
  }
  
  Result VideoScalerNode::doDeinit() {
    getOwnOutputProperty()->erase();
    getOwnInputProperty()->erase();
    return SUCCESS;
  }
  
  Result VideoScalerNode::doSetFormat(const Format& format) {
    // get relevant format parameters
    int x_res = format.getIntValue(Format::x_resolution_param);
    int y_res = format.getIntValue(Format::y_resolution_param);
    int ratio_x = format.getIntValue(Format::ratio_x_param);
    int ratio_y = format.getIntValue(Format::ratio_y_param);
    
    DEBUG_STREAM("SetFormat: " << x_res << " x " << y_res << ", ratio " << ratio_x << ":" << ratio_y << endl);
    
    return SUCCESS;
  }
  
  void VideoScalerNode::getAspectCorrection(int *w, int *h, int ratio_x, int ratio_y) {
    float desired_aspect = (float) ratio_x / (float) ratio_y;
    float current_aspect = (float) *w / (float) *h;
    
    // figure out whether to enlarge X or Y
    // example:
    // (1) input_size=640x480 aspect=16:9 ==> output_size=853x480
    // (2) input_size=400x200 aspect=16:9 ==> output_size=400x225
    // also make sure, X resolution is dividable by 4 and Y resolution is dividable by 2
    
    if (desired_aspect > current_aspect) { // need to enlarge X
      *w = (int) ((*w) * desired_aspect / current_aspect);
      *w = (*w + 3) & ~(int) 0x3;
    } else {
      *h = (int) ((*h) * current_aspect / desired_aspect);
      *h = (*h + 1) & ~(int) 0x1;
    }
  }
  
  void VideoScalerNode::reaspect() {
    // get input size
    int current_w = scaler.getInputWidth();
    int current_h = scaler.getInputHeight();
    
    int output_w = current_w;
    int output_h = current_h;
    if (keep_aspect_ratio) {
      // adjust resolution to aspect
      getAspectCorrection(&output_w, &output_h, ratio_x, ratio_y);
      DEBUG_STREAM("reaspect: adjust output size to input size " << current_w << "x" << current_h
		   << " with aspect ratio " << ratio_x << ":" << ratio_y
		   << " to " << output_w << "x" << output_h << endl);
    } else {
      // use previous output size of scaler
      output_w = scaler.getOutputWidth();
      output_h = scaler.getOutputHeight();
      
      DEBUG_STREAM("reaspect: keep fixed output resolution " << output_w << "x" << output_h << endl);
    }
    
    // set output size
    // only send the format if the output size has changed
    if (output_w != scaler.getOutputWidth() || output_h != scaler.getOutputHeight()) {
      send_format = true;
      scaler.setOutputSize(output_w, output_h);
    }
  }
  
  Result VideoScalerNode::setupOutputFormat(const Format* in_format) {
    // get format parameters
    int x_resolution;
    int y_resolution;
    x_resolution = in_format->getIntValue(Format::x_resolution_param);
    y_resolution = in_format->getIntValue(Format::y_resolution_param);
    ratio_x = in_format->getIntValue(Format::ratio_x_param);
    ratio_y = in_format->getIntValue(Format::ratio_y_param);
    framerate = in_format->getFloatValue(Format::framerate_param);
    channelorder = in_format->getStringValue(Format::channelorder_param);
    colorspace = in_format->getStringValue(Format::colorspace_param);
    format = in_format->getStringValue(Format::format_param);
    endian = in_format->getStringValue(Format::endian_param);
    bitperpixel = in_format->getIntValue(Format::bitperpixel_param);

    // erase working properties
    getOwnWorkingInputProperty()->erase();
    getOwnWorkingOutputProperty()->erase();
    
    // general working input format
    // (resolution and aspect ratio may change)
    Format* i_format_gen = getOwnWorkingInputProperty()->addNewFormat("video/raw");
    i_format_gen->addWildcard(Format::x_resolution_param);
    i_format_gen->addWildcard(Format::y_resolution_param);
    i_format_gen->addWildcard(Format::ratio_x_param);
    i_format_gen->addWildcard(Format::ratio_y_param);
    i_format_gen->addWildcard(Format::framerate_param);
    i_format_gen->addStringParamValue(Format::channelorder_param, channelorder);
    i_format_gen->addStringParamValue(Format::colorspace_param, colorspace);
    i_format_gen->addStringParamValue(Format::format_param, format);
    i_format_gen->addStringParamValue(Format::endian_param, endian);
    i_format_gen->addIntParamValue(Format::bitperpixel_param, bitperpixel);
    
    // current working input format
    // (the current connection format)
    Format* i_format = new Format(*in_format);
    getOwnWorkingInputProperty()->addNewFormat(i_format);
    getOwnWorkingInputProperty()->setDefaultFormat(i_format);
    
    // general working output format
    // (scaled to arbitrary resolution)
    Format* o_format_gen = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
    o_format_gen->addWildcard(Format::x_resolution_param);
    o_format_gen->addWildcard(Format::y_resolution_param);
    o_format_gen->addWildcard(Format::ratio_x_param);
    o_format_gen->addWildcard(Format::ratio_y_param);
    o_format_gen->addFloatParamValue(Format::framerate_param, framerate);
    o_format_gen->addStringParamValue(Format::channelorder_param, channelorder);
    o_format_gen->addStringParamValue(Format::colorspace_param, colorspace);
    o_format_gen->addStringParamValue(Format::format_param, format);
    o_format_gen->addStringParamValue(Format::endian_param, endian);
    o_format_gen->addIntParamValue(Format::bitperpixel_param, bitperpixel);
    
    // default output format
    // (recommended resolution with aspect correction)
    int scaled_x_resolution = x_resolution;
    int scaled_y_resolution = y_resolution;
    getAspectCorrection(&scaled_x_resolution, &scaled_y_resolution, ratio_x, ratio_y);
    Format* o_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
    o_format->addIntParamValue(Format::x_resolution_param, scaled_x_resolution);
    o_format->addIntParamValue(Format::y_resolution_param, scaled_y_resolution);
    o_format->addIntParamValue(Format::ratio_x_param, ratio_x);
    o_format->addIntParamValue(Format::ratio_y_param, ratio_y);
    o_format->addFloatParamValue(Format::framerate_param, framerate);
    o_format->addStringParamValue(Format::channelorder_param, channelorder);
    o_format->addStringParamValue(Format::colorspace_param, colorspace);
    o_format->addStringParamValue(Format::format_param, format);
    o_format->addStringParamValue(Format::endian_param, endian);
    o_format->addIntParamValue(Format::bitperpixel_param, bitperpixel);
    getOwnWorkingOutputProperty()->setDefaultFormat(o_format);
    
    // set I/O partners
    i_format->setIOPartner(o_format);
    o_format->setIOPartner(i_format);
    i_format_gen->setIOPartner(o_format_gen);
    o_format_gen->setIOPartner(i_format_gen);

    return SUCCESS;
  }
  
  Result VideoScalerNode::doInitOutput() {
    // get input format
    const Format* in_format = getInputFormat();
    if (in_format == NULL) {
      ERROR_STREAM("No input format" << endl);
      return FAILURE;
    }
    
    // setup output format
    return setupOutputFormat(in_format);
  }
  
  Result VideoScalerNode::doDeinitOutput() {
    // erase working properties
    getOwnWorkingInputProperty()->erase();
    getOwnWorkingOutputProperty()->erase();
    return SUCCESS;
  }
  
  Result VideoScalerNode::doActivate() {
    // get input and output formats
    const Format *in_f = getInputFormat();
    if (in_f == NULL) {
      ERROR_STREAM("Couldn't get input Format!!!" << endl);
      return FAILURE;
    }
    const Format *out_f = getOutputFormat();
    if (out_f == NULL) {
      ERROR_STREAM("Couldn't get output Format!!!" << endl);
      return FAILURE;
    }
    
    // set scaler pixel format
    // (this is fixed because it is set in doActivate only)
    string colorspace = in_f->getStringValue(Format::colorspace_param);
    if (colorspace == "yv12") {
      scaler.setFormat(VideoScaler::FORMAT_YV12);
    } else if (colorspace == "yv12a") {
      scaler.setFormat(VideoScaler::FORMAT_YV12A);
    } else if (colorspace == "rgb") {
      scaler.setFormat(VideoScaler::FORMAT_RGB);
    } else if (colorspace == "rgba") {
      scaler.setFormat(VideoScaler::FORMAT_RGBA);
    } else {
      ERROR_STREAM("Input format (colorspace) not supported" << endl);
      return FAILURE;
    }
    
    // set size
    // simply set the resolution specified on connection
    // auto aspect only affects instream format changes, and the initial default output format is
    // aspect corrected. if auto aspect mode is off, the initial output resolution is kept when the
    // input resolution changes.
    int w,h;
    w = in_f->getIntValue(Format::x_resolution_param);
    h = in_f->getIntValue(Format::y_resolution_param);
    scaler.setInputSize(w, h);
    w = out_f->getIntValue(Format::x_resolution_param);
    h = out_f->getIntValue(Format::y_resolution_param);
    scaler.setOutputSize(w, h);
    
    // clear send_format flag
    send_format = false;    
    
    return SUCCESS;
  }
  
  Message* VideoScalerNode::processBuffer(Buffer *in_buffer) {
    if ( getWorkingFlag() ) {
      setWorkingFlag(false);
      in_buffer = saved_buffer;
    }
    
    if (send_format) {
      // clear flag
      send_format = false;
      
      // store buffer
      saved_buffer = in_buffer;
      setWorkingFlag(true);
      
      // send new output format
      Format new_format("video/raw");
      new_format.addIntParamValue(Format::x_resolution_param, scaler.getOutputWidth());
      new_format.addIntParamValue(Format::y_resolution_param, scaler.getOutputHeight());
      new_format.addIntParamValue(Format::ratio_x_param, ratio_x);
      new_format.addIntParamValue(Format::ratio_y_param, ratio_y);
      new_format.addFloatParamValue(Format::framerate_param, framerate);
      new_format.addStringParamValue(Format::channelorder_param, channelorder);
      new_format.addStringParamValue(Format::colorspace_param, colorspace);
      new_format.addStringParamValue(Format::format_param, format);
      new_format.addStringParamValue(Format::endian_param, endian);
      new_format.addIntParamValue(Format::bitperpixel_param, bitperpixel);
      Event event(IFormatChange::setFormat_event, new TInValue<Format>(new_format));
      return new CEvent(event);
    }
    
    // allocate buffer
    // XXX: check if size calculation is correct
    int out_buffer_size = scaler.getOutputWidth() * scaler.getOutputHeight() * bitperpixel / 8;
    Buffer* out_buffer = getNewBuffer(out_buffer_size, in_buffer);
    if (NULL == out_buffer) {
      DEBUG_STREAM("failed to allocate output buffer" << endl);
      in_buffer->release();
      return NULL;
    }
    out_buffer->setUsed(out_buffer_size);
    out_buffer->setTimestamp(in_buffer->getTimestamp());
    
    // process data
    char* in_data = in_buffer->getData();
    char* out_data = out_buffer->getData();
    scaler.processData(out_data, in_data);
    
    // done
    in_buffer->release();
    return out_buffer;
  }
  
  void VideoScalerNode::setFiltering(const nmm_bool& filtering) {
    scaler.setFiltering(filtering);
  }
  
  void VideoScalerNode::setAutoAspect(const nmm_bool& auto_aspect) {
    DEBUG_STREAM("Received event: setAutoAspect: " << auto_aspect);
    
    if (keep_aspect_ratio != auto_aspect) {
      keep_aspect_ratio = auto_aspect;
      
      // calculate scaler output size
      reaspect();
    }
  }
  
  nmm_bool VideoScalerNode::getFiltering() const {
    return scaler.getFiltering();
  }
  
  nmm_bool VideoScalerNode::getAutoAspect() const {
    return keep_aspect_ratio;
  }
  
  void VideoScalerNode::setOutputSize(const nmm_int32& w, const nmm_int32& h) {
    if (w != scaler.getOutputWidth() || h != scaler.getOutputHeight()) {
      scaler.setOutputSize(w,h);
      send_format = true;
    }
  }
  
  void VideoScalerNode::getOutputSize(nmm_int32& w, nmm_int32& h) const {
    w = scaler.getOutputWidth();
    h = scaler.getOutputHeight();
  }
}

#include "nmm/nmm_suffix.hpp"
