/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "ShmBufferManager.hpp"

#include "nmm/comm/messaging/Buffer.hpp"
#include "nmm/comm/messaging/B_Allocator.hpp"

#include "nmm/utils/thread/MutexGuard.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  ShmBufferManager::ShmBufferManager(const string & name)
    : BufferManager(),
      NamedObject(name),
      buffers(),
      freebuffers(),
      num_of_buffers(0),
      erase_flag(false),
      shm_segment(NULL),
      shm_size(0),
      mutex(),
      buffer_size(0),
      fallback_bm(NULL) {
  }

  ShmBufferManager::~ShmBufferManager() {
    DEBUG_STREAM("Begin> Destructor" << endl);

    // as Destructor is protected, this object can only self-destruct,
    // so we don't need a mutex as the caller locked it already
    // see: releaseBuffer() and erase()

    if (fallback_bm) {
      fallback_bm->release();
      fallback_bm = NULL;
    }

    freebuffers.clear();
    
    /* delete all allocated buffers */
    for(bufferlist_t::iterator iter = buffers.begin();
        iter != buffers.end();
        iter++) {
      delete (*iter);
    }
    buffers.clear();

    DEBUG_STREAM("End< Destructor" << endl);
  }

  bool ShmBufferManager::setSharedMemory(char* memptr, size_t size) {
    DEBUG_STREAM("Begin> addSharedMemory()" << endl);

    MutexGuard mg(mutex);

    if (erase_flag) {
      // This BufferManager will be erased. Do not accept memory any more
      ERROR_STREAM("while erasing, no memory addition accepted any more!" << endl);
      return false;
    }

    if (shm_segment) {
      ERROR_STREAM("Shared Memory segment already set!" << endl);
      return false;
    }

    shm_segment = memptr;
    shm_size = size;

    DEBUG_STREAM("Added!" << endl);

    DEBUG_STREAM("End< addSharedMemory()" << endl);

    return true;
  }

  Buffer* ShmBufferManager::requestBuffer(const size_t size) const {
    DEBUG_STREAM("Begin> requestBuffer()" << endl);

    MutexGuard mg(mutex);

    if (erase_flag) {
      // This BufferManager will be erased. Deny all request
      ERROR_STREAM("while erasing, no Buffer requests accepted any more!" << endl);
      return NULL;
    }

    DEBUG_STREAM("Request buffer with size " << size << endl);

    if (buffers.empty() && shm_segment) {
      buffer_size = size;
      for (char* buf = shm_segment;
           buf + size < shm_segment + shm_size;
           buf += size) {
        DEBUG_STREAM("Adding new SHM Buffer with size " << buffer_size << endl);
        Buffer* buffer = new Buffer(buffer_size, buf, this);

        // add buffer to list
        buffers.push_front(buffer);

        // add buffer to freebuffers list
        freebuffers.push_front(buffer);
      }
    }

    if ((size > buffer_size)) {
      if (buffer_size > 0) {
        // we don't have such a large buffer. So we must create a
        // non-shared one!
        ERROR_STREAM("We don't have such a great shared memory buffer (" << size << "!" << endl);
        ERROR_STREAM("Largest Buffer size supported is " << buffer_size << "!" << endl);
        ERROR_STREAM("Giving out normal memory!" << endl);
        if (!fallback_bm) {
          fallback_bm = new B_Allocator(size, 4, 10, (getName() + string("-internal_B_Allocator")).c_str());
        }
        return fallback_bm->requestBuffer(size);
      } else {
        // we don't have ANY buffers yet. Now we have 2 possibilities:
        // 1. Wait until someone added buffers. This is what the old
        //    code did! But it forgot to notify() when adding memory!
        //    Anyway, this can cause a deadlock!
        // 2. Create a non-shared buffer! -> We decided to do this!
        ERROR_STREAM("We don't have any Buffers yet!" << endl);
        ERROR_STREAM("Giving out normal memory!" << endl);
        if (!fallback_bm) {
          fallback_bm = new B_Allocator(size, 4, 10, (getName() + string("-internal_B_Allocator")).c_str());
        }
        return fallback_bm->requestBuffer(size);
      }
    }

    // test, if buffer is available
    if (freebuffers.size() == 0) {
      ERROR_STREAM("No buffers available any more!" << endl);
      ERROR_STREAM("Giving out normal memory!" << endl);
      if (!fallback_bm) {
        fallback_bm = new B_Allocator(size, 4, 10, (getName() + string("-internal_B_Allocator")).c_str());
      }
      return fallback_bm->requestBuffer(size);
    }

    // take buffer out of the list
    Buffer* buf = freebuffers.front();

    // remove buffer from the list
    freebuffers.pop_front();

    // increase the RefCount as someone is using it now!
    buf->increaseCount(1);

    // increase the number of running buffers
    num_of_buffers++;

    DEBUG_STREAM("End< requestBuffer()" << endl);

    return buf;
  }

  void ShmBufferManager::releaseBuffer(Buffer* buf) const {
    DEBUG_STREAM("Begin> releaseBuffer()" << endl);

    MutexGuard mg(mutex);
    
    DEBUG_STREAM("Freeing buffer" << endl);
    // we got a Buffer back, decrease the count of running buffers
    num_of_buffers--;

    // add this buffer to the free ones!
    freebuffers.push_front(buf);

    DEBUG_STREAM("End< releaseBuffer()" << endl);

    // The object will delete itself only if all buffers given away have returned
    if(erase_flag && (num_of_buffers <= 0)) {
      DEBUG_STREAM("Deleting myself" << endl);
      delete this;
    }
  }

  void ShmBufferManager::erase() const {
    DEBUG_STREAM("Begin> erase()" << endl);

    MutexGuard mg(mutex);

    if (fallback_bm) {
      fallback_bm->release();
      fallback_bm = NULL;
    }

    // set flag, telling the object to try to delete itself
    erase_flag = true;

    DEBUG_STREAM("End< erase()" << endl);

    // The object will delete itself only if all buffers given away have returned
    if (num_of_buffers <= 0) {
      DEBUG_STREAM("deleting myself" << endl);
      delete this;
    }
  }


}

#include "nmm/nmm_suffix.hpp"
