/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "VideoGrabNode.hpp"

#include "nmm/comm/messaging/B_Allocator.hpp"
#include "nmm/comm/format/Property.hpp"
#include "nmm/multimedia/Plugin.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"

#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <iostream>

NMM_REGISTER_PLUGIN(NMM::VideoGrabNode, NMM, "VideoGrabNode")
 
namespace NMM {

  VideoGrabNode::VideoGrabNode(const char* name /* = "VideoGrabNode" */,
                                 StreamQueue::Mode mode /* = StreamQueue::MODE_SUSPEND */,
                                 int queue_size /* = StreamQueue::DEFAULT_SIZE */)
    : GenericSourceNode(name, mode, queue_size), 
      x_offset(0),
      y_offset(0),
      x_range(320),
      y_range(240),
      __last_timer(), 
      __condition(__cond_mutex)
  {
    stimer = new StreamTimer(this, StreamTimer::CONST_RATE);
  }
  
  VideoGrabNode::~VideoGrabNode() {
    delete stimer;
  }
  
  void VideoGrabNode::setCrop(const nmm_int32& xOffset, const nmm_int32& yOffset, const nmm_int32& xRange, const nmm_int32& yRange)
  {
    x_offset = xOffset;
    y_offset = yOffset;
    x_range  = xRange;
    y_range  = yRange;
  } 

  void VideoGrabNode::addSupportedFormats(Property* property) {
    NMM::Format *myOutputFormat;
    myOutputFormat=property->addNewFormat("video/raw");
    myOutputFormat->addIntParamValue(Format::x_resolution_param, x_range);
    myOutputFormat->addIntParamValue(Format::y_resolution_param, y_range);
    myOutputFormat->addIntParamValue(Format::ratio_x_param, 320);
    myOutputFormat->addIntParamValue(Format::ratio_y_param, 240);
    myOutputFormat->addStringParamValue(Format::colorspace_param,"rgb24");
    myOutputFormat->addStringParamValue(Format::channelorder_param,"rgb");
    myOutputFormat->addIntParamValue(Format::bitperpixel_param,32);
    myOutputFormat->addStringParamValue(Format::endian_param, "little");
    myOutputFormat->addStringParamValue(Format::format_param, "packet");
    myOutputFormat->addFloatParamValue(Format::framerate_param, 15.0);
    property->setDefaultFormat(myOutputFormat);
  }
  
  Result VideoGrabNode::doInit() {
    addSupportedFormats(getOwnOutputProperty());
    return SUCCESS;
  }
  
  Result VideoGrabNode::doDeinitOutput() {
    XShmDetach(_display, &_shminfo);
    shmdt(_shminfo.shmaddr);
    shmctl(_shminfo.shmid, IPC_RMID, 0);
    XDestroyImage(_ximage);

    return SUCCESS;
  }

  Result VideoGrabNode::doInitOutput() {
    addSupportedFormats(getOwnWorkingOutputProperty());
    return SUCCESS;
  }
 
  Result VideoGrabNode::doActivate() {
    if (getOutputFormat() == NULL) { return FAILURE; }
    image_width = getOutputFormat()->getIntValue(Format::x_resolution_param);
    image_height = getOutputFormat()->getIntValue(Format::y_resolution_param);
    image_bpp = getOutputFormat()->getIntValue(Format::bitperpixel_param);
    image_fps = getOutputFormat()->getFloatValue(Format::framerate_param);
    DEBUG_STREAM("fps is " << image_fps << endl);
    DEBUG_STREAM("bitsperpixel is " << image_bpp << endl);
    
    stimer->reset();
    stimer->setRate(image_fps);
    
    _display = XOpenDisplay(NULL);
    if(!_display) { 
      return FAILURE; 
    }

    if(!XShmQueryExtension(_display)) {
      return FAILURE;
    }

    _root = RootWindow(_display, 0);

    _ximage = XShmCreateImage(_display,
			      DefaultVisual(_display, DefaultScreen(_display)),
			      DefaultDepth(_display, DefaultScreen(_display)),
			      ZPixmap,
			      NULL,
			      &_shminfo,
			      image_width,
			      image_height);

    if(!_ximage) { return FAILURE; }

    _shminfo.shmid = shmget(IPC_PRIVATE, 
			    _ximage->bytes_per_line * _ximage->height,
			    IPC_CREAT|0777);

    if(_shminfo.shmid < 0) { return FAILURE; }

    _shminfo.shmaddr = _ximage->data = (char*)(void*)shmat(_shminfo.shmid, 0, 0);
    _shminfo.readOnly = False;

    if(!XShmAttach(_display, &_shminfo)) { return FAILURE; } 

    XSync(_display, False);

    shmctl(_shminfo.shmid, IPC_RMID, 0);

    return SUCCESS;
  }

  Result VideoGrabNode::doStop() {
    return SUCCESS;
  }
  
  Message *VideoGrabNode::processBuffer(Buffer*) {
    if (__last_timer == Time(0, 0))
    {
      /* initialize the timer with the current time */
      __last_timer = TimedElement::getTime();
    }

    Buffer* out_buffer = getNewBuffer((_ximage->bytes_per_line) * _ximage->height);
    
    XShmGetImage(_display, _root, _ximage, x_offset, y_offset, AllPlanes);

    memcpy(out_buffer->getData(), _ximage->data, out_buffer->getSize());

    /* Calculate the Timestamp of the next Buffer */
    __last_timer += Interval(0, static_cast<long>(1000000000L/image_fps));

    /* wait until next frame */
    __condition.timedWait(__last_timer.sec, __last_timer.nsec);

    /* set the new timestamp */
    stimer->setTimestamp(out_buffer);

    return(out_buffer);

  }

}

