/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "XDisplayNode.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"

#include "XImg.hpp"
#include "XShmImg.hpp"
#include "XvImg.hpp"
#include "XvShmImg.hpp"

#include "ShmBufferManager.hpp"

NMM_REGISTER_PLUGIN(XDisplayNode, NMM, "XDisplayNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  XDisplayNode::XDisplayNode(const char* name, StreamQueue::Mode queue_mode, int queue_size)
    : GenericUSyncSinkNode(name, queue_mode, queue_size),
      display_events_enabled(false),
      display_events_config(),
      align(CENTER),
      benchmarking_enabled(true)
  {
    /* This has to be called to make X-functions mutli-thread safe */
    XInitThreads();

    width = height = 0;
    depth = 0;
    last_in_buffer = 0;
    title_update_rate = 1;
    DEBUG_STREAM( "Begin > Create XDisplay Node" << endl);

    buffermanager            = NULL;
    image                    = 0;
    // todo: let SHM code determine the maximum number of shared memory pages
    shared_memory_pages      = 8;

    fullscreen = false;
    ratio_x = ratio_y = 0;

    frame_title = new char[100];
    
    benchmark.setNamedObject(this);
    
    xshm_supported = false;
    display_local = true;
    
    DEBUG_STREAM("End > Create XDisplayNode" << endl);
  }

  XDisplayNode::~XDisplayNode() {
    if (image) {
      delete image;
      image = NULL;
    }
    delete [] frame_title;
  }
  
  Result XDisplayNode::doStart() 
  {
    image->setTitle(getName());

    if(benchmarking_enabled) {
      benchmark.setEnabled(true);
    }
    return SUCCESS;
  }

  Result XDisplayNode::doStop() {
    if(benchmarking_enabled) {
      benchmark.setEnabled(false);
    }
    return SUCCESS;
  }

  Result XDisplayNode::doInit() {
    return allocate();
  }
  
  Result XDisplayNode::doDeinit() {
    getOwnInputProperty()->erase();
    return SUCCESS;
  }
  
  Result XDisplayNode::doDeactivate() {
    if (display_events_enabled) {
      // as image was created and initialized in doActivate,
      // deinitialize it here.
      disableDisplayEvents();
      // as the method sets the value to false we set it to true here
      // again to have it stored for the nex activation!
      display_events_enabled = true;
    }
    if (image) {
      image->deinit();
      delete image;
      image = NULL;
    }
    
    return SUCCESS;
  }

  Result XDisplayNode::getDepth(int& depth_) {
    depth_ = depth;
    return SUCCESS;
  }
  
  void XDisplayNode::useSharedMemoryManager( unsigned int shared_pages ) {
    shared_memory_pages = shared_pages;

    if (buffermanager) {
      setBufferManager(NULL);
    }
    if (shared_memory_pages > 1) {
      buffermanager = new ShmBufferManager();
    }
    else {
      buffermanager  = NULL;
    }
    setBufferManager(buffermanager);

    if (image) {
      image->updateMemoryManager(buffermanager, shared_memory_pages);
    }
  }
  
  Result XDisplayNode::allocate()
  {
    Display* display = XOpenDisplay(NULL);
    if (!display)
      {
        ERROR_STREAM("Cannot open display" << endl);
        return FAILURE;
      }
    
    depth = DefaultDepth(display, DefaultScreen(display));
    DEBUG_STREAM("Depth is " << depth << endl);

    bool lsbfirst = LSBFirst == ImageByteOrder(display);
    DEBUG_STREAM("Display is using " << (lsbfirst ? "LSBFirst" : "MSBFirst") << " -> " 
		 << (lsbfirst ? "little" : "big")<< " endian" << endl);

    Format* i_format;

    // find all supported pixmap formats
    int n;
    int bitperpixel;
    XPixmapFormatValues *pmf;
    pmf = XListPixmapFormats (display, &n);
    DEBUG_STREAM ("number of supported pixmap formats:  "  << n << endl);
    if (pmf) {
      DEBUG_STREAM ("supported pixmap formats:" << endl);
      int pixmap_depth;
      for (int i = 0; i < n; i++) {
	pixmap_depth = pmf[i].depth; 
	bitperpixel = pmf[i].bits_per_pixel;
       	if (pixmap_depth == depth) {  // only the format with the correct depth is added
	  switch (depth) {
	  case 16:
	    i_format = getOwnInputProperty()->addNewFormat("video/raw");
	    i_format->addWildcard(Format::x_resolution_param);
	    i_format->addWildcard(Format::y_resolution_param);
	    i_format->addWildcard(Format::ratio_x_param);
	    i_format->addWildcard(Format::ratio_y_param);
	    i_format->addWildcard(Format::framerate_param);
	    i_format->addStringParamValue(Format::channelorder_param, "rgb");
	    i_format->addStringParamValue(Format::endian_param, (lsbfirst ? "little" : "big"));
	    i_format->addStringParamValue(Format::colorspace_param,"rgb565");
	    i_format->addIntParamValue(Format::bitperpixel_param, bitperpixel);
	    i_format->addStringParamValue(Format::format_param, "packet");
	    i_format->setIOPartner(0);
	    break;
	  case 24:
	    i_format = getOwnInputProperty()->addNewFormat("video/raw");
	    i_format->addWildcard(Format::x_resolution_param);
	    i_format->addWildcard(Format::y_resolution_param);
	    i_format->addWildcard(Format::ratio_x_param);
	    i_format->addWildcard(Format::ratio_y_param);
	    i_format->addWildcard(Format::framerate_param);
	    i_format->addStringParamValue(Format::channelorder_param, "rgb");
	    i_format->addStringParamValue(Format::colorspace_param,"rgb24");
	    i_format->addIntParamValue(Format::bitperpixel_param, bitperpixel);
	    i_format->addStringParamValue(Format::format_param, "packet");
	    i_format->addStringParamValue(Format::endian_param, (lsbfirst ? "little" : "big"));
	    i_format->setIOPartner(0);
	    break;
	  case 32:
	    i_format = getOwnInputProperty()->addNewFormat("video/raw");
	    i_format->addWildcard(Format::x_resolution_param);
	    i_format->addWildcard(Format::y_resolution_param);
	    i_format->addWildcard(Format::ratio_x_param);
	    i_format->addWildcard(Format::ratio_y_param);
	    i_format->addWildcard(Format::framerate_param);
	    i_format->addStringParamValue(Format::channelorder_param, "rgba");
	    i_format->addStringParamValue(Format::endian_param, (lsbfirst ? "little" : "big"));
	    i_format->addStringParamValue(Format::colorspace_param,"rgba32");
	    i_format->addIntParamValue(Format::bitperpixel_param, bitperpixel);
	    i_format->addStringParamValue(Format::format_param, "packet");
	    i_format->setIOPartner(0);
	    break;  
	  default: break;
	  }
	  break;
	}
      }
      XFree ((char *) pmf);
    }

    DEBUG_STREAM("Used bitperpixel is " << bitperpixel << endl);

    // ask whether shared memory is supported
    xshm_supported = XShmQueryExtension(display);
    
    #ifdef BUILD_MACOSX 
    xshm_supported = false;
    #endif
    
    // test whether display is local
    // assumes that sshd is configured to set remote displays
    // with values >= 10 (ssh default)
    char* dispName = XDisplayName(NULL);
    DEBUG_STREAM("Got displayname: " << dispName << endl);
    if ( strncmp(dispName, "unix:", 5) == 0)
      dispName += 4;
    else if ( strncmp(dispName, "localhost:", 10) == 0)
      dispName += 9;
    display_local = (*dispName==':' && atoi(dispName+1)<10);
    DEBUG_STREAM("Deciding that we " << (display_local ? "" : "don't") << " have a local display." << endl);

    XCloseDisplay(display);
    
    i_format = getOwnInputProperty()->addNewFormat("video/raw");
    i_format->addWildcard(Format::x_resolution_param);
    i_format->addWildcard(Format::y_resolution_param);
    i_format->addWildcard(Format::ratio_x_param);
    i_format->addWildcard(Format::ratio_y_param);
    i_format->addWildcard(Format::framerate_param);
    i_format->addStringParamValue(Format::channelorder_param,"uyvy");
    i_format->addStringParamValue(Format::endian_param, (lsbfirst ? "little" : "big"));
    i_format->addStringParamValue(Format::colorspace_param,"yuv422");
    i_format->addIntParamValue(Format::bitperpixel_param,16);
    i_format->addStringParamValue(Format::format_param, "packet");
    i_format->setIOPartner(0);
    
    i_format = getOwnInputProperty()->addNewFormat("video/raw");
    i_format->addWildcard(Format::x_resolution_param);
    i_format->addWildcard(Format::y_resolution_param);
    i_format->addWildcard(Format::ratio_x_param);
    i_format->addWildcard(Format::ratio_y_param);
    i_format->addWildcard(Format::framerate_param);
    i_format->addStringParamValue(Format::channelorder_param, "yvu");
    i_format->addStringParamValue(Format::endian_param, (lsbfirst ? "little" : "big"));
    i_format->addStringParamValue(Format::colorspace_param,"yv12");
    i_format->addIntParamValue(Format::bitperpixel_param,12);
    i_format->addStringParamValue(Format::format_param, "planar");
    i_format->setIOPartner(0);
    getOwnInputProperty()->setDefaultFormat(i_format);
    return SUCCESS;
  }

  void XDisplayNode::setBenchmarkingEnabled(const nmm_bool& value)
  {
    benchmarking_enabled = value;
  }

  void XDisplayNode::setDisplayPosition(const nmm_int32& _xPos, const nmm_int32& _yPos) {
    align = CUSTOM;
    xPos = _xPos;
    yPos = _yPos;
  }

  void XDisplayNode::setFullscreen( const nmm_bool& f ) {
    if ( fullscreen != f ) {
      fullscreen = f;
      if ( image ) { 
        if (fullscreen) {
          sprintf(frame_title, "%s - %s", "Fullscreen", getName().c_str());
        } else {
          if (benchmark.getEnabled()) {
            sprintf(frame_title, "%s %.1f fps", getName().c_str(), benchmark.getFPS());
          } else {
            sprintf(frame_title, "%s", getName().c_str());
          }
        }
        image->setTitle(string(frame_title));
	image->setFullscreen( fullscreen );
      }
    }
  };
  
  nmm_bool XDisplayNode::getFullscreen() const {
    return fullscreen;
  }
  
  nmm_bool XDisplayNode::hasFullscreen() const {
    // XXX: test if fullscreen is REALLY supported
    return true;
  }
  
  nmm_bool XDisplayNode::isFullscreenOnly() const {
    return false;
  }
  
  void XDisplayNode::setRatio( int x, int y ) {
    DEBUG_STREAM("Set Ratio to x=" << x << ", y=" << y << endl);

    int a;
    int b;

    if ( ratio_y != 0 ) a = ratio_x*10000/ratio_y;
    else a = 0;
    b = x*10000/y;

    if ( a != b ) {
      if ( image ) image->setRatio( x, y );
    }

    ratio_x = x;
    ratio_y = y;
  };
  
  Result XDisplayNode::fastForward( const nmm_int32& speed ) {
    ERROR_STREAM("Can't get predeessor node!!! Please contact me!!! -fw" << endl);
    
    // XXX: fw: getPredecessorNode has been removed!!
//     /* get predecessor node */
//     INode_var predecessor(getPredecessorNode());
    
    /* stop itself */
    try{ stop(); }
    catch(...){ DEBUG_STREAM("Node already stopped" << endl); };
    
    // XXX: fw: getPredecessorNode has been removed!!
//     /* stop predecessor to avoid filling the queue */
//     if ( predecessor.get() ) {
//       try{ predecessor->stop(); }
//       catch(...){ DEBUG_STREAM("Predecessor node already stopped" << endl); };
//     }

    flush();

    if ( speed > 0 ) {
      DEBUG_STREAM("fast forward event: on" << endl);
      /* turn synchronisation off */
      setSynchronized( false );
    } else {
      DEBUG_STREAM("fast forward event: off" << endl);
      /* turn synchronisation on */
      setSynchronized( true );
    }
    
    /* send upstream event to predecessor */
    // removed sendSet|GetEvent (was implemented as return FAILURE; anyway!)
    // if ( predecessor ) predecessor->sendSetEvent(Event(IStreamJump::fastForward_event, new TInValue<nmm_int32>( speed )) );
    
    try{ start(); }
    catch(...){ DEBUG_STREAM("Node already started" << endl); };

    return( SUCCESS );
  }
  
  Result XDisplayNode::streamJump( const nmm_int32& speed ) {
    ERROR_STREAM("Can't get predeessor node!!! Please contact me!!! -fw" << endl);
    
    DEBUG_STREAM("stream jump event, speed = " << speed << endl);
    
    // XXX: fw: getPredecessorNode has been removed!!
//     /* get predecessor node */
//     INode_var predecessor(getPredecessorNode());

    /* stop itself */
    try{ stop(); }
    catch(...){ DEBUG_STREAM("Node already stopped" << endl); };

    // XXX: fw: getPredecessorNode has been removed!!
//     /* stop predecessor to avoid filling the queue */
//     if ( predecessor.get() ) {
//       try{ predecessor->stop(); }
//       catch(...){ DEBUG_STREAM("Predecessor node already stopped" << endl); };
//     }

    flush();
    
    /* send upstream event to predecessor */
    // removed sendSet|GetEvent (was implemented as return FAILURE; anyway!)
    // if ( predecessor ) predecessor->sendSetEvent(Event(IStreamJump::streamJump_event, new TInValue<nmm_int32>( speed )) );
    
    try{ start(); }
    catch(...){ DEBUG_STREAM("Node already started" << endl); };
    
    return( SUCCESS );
  }
  
  Result XDisplayNode::sendIFramesOnly( const nmm_int32& ) {
    return SUCCESS;
  }
  
  Result XDisplayNode::setResolution(int w, int h) {
    width = w;
    height = h;
    
    if ( image ) return( image->setSize( width, height ) );
    return( FAILURE );
  }
  
  Result XDisplayNode::doInitOutput() {
    const Format* i_format = getInputFormat();
    if (i_format == 0) {
      ERROR_STREAM( "Invalid input format (NULL-ptr)" << endl);
      return FAILURE;
    }
    Format* format;
    
    format = new Format( *i_format );
    getOwnWorkingInputProperty()->addNewFormat(format);
    getOwnWorkingInputProperty()->setDefaultFormat(format);
    
    try {
      title_update_rate = (int) (0.5 * i_format->getFloatValue(Format::framerate_param));
      if(title_update_rate == 0) {
	title_update_rate = 1;
      }
    }
    catch(ParameterException e) {
      title_update_rate = 10;
      WARNING_STREAM("No parameter framerate in input format" << endl);
    }

    // if new color space, save colorspace and delete image
    if(i_format->getStringValue(Format::colorspace_param) != colorspace) {
      colorspace = i_format->getStringValue(Format::colorspace_param);
      if(image) {
	delete image;
	image = NULL;
      }
    }

    if(image == NULL) {
      width  = i_format->getIntValue(Format::x_resolution_param);
      height = i_format->getIntValue(Format::y_resolution_param);


      if (i_format->isParameterIncluded("ratio_x") && i_format->isParameterIncluded("ratio_y")) {
	ratio_x = i_format->getIntValue("ratio_x");
	ratio_y = i_format->getIntValue("ratio_y");
      } else {
        ratio_x = width;
	ratio_y = height;
      }

      DEBUG_STREAM("colorspace in doActivate is: " << colorspace << endl);

      const char* title = getName().c_str();
      if (fullscreen) {
        title = (string("Fullscreen - ") + getName()).c_str();
      }

      if (xshm_supported && display_local) {
	DEBUG_STREAM("Shared X memory is used." << endl);
        useSharedMemoryManager(shared_memory_pages);
	if(colorspace == "yuv422") {
	  image = new XvShmImg(this, width, height, title, XvShmImg::XV_FORMAT_UYVY, fullscreen, ratio_x, ratio_y, buffermanager, shared_memory_pages);
	}
	else if(colorspace == "yv12") {
	  image = new XvShmImg(this, width, height, title, XvShmImg::XV_FORMAT_YV12, fullscreen, ratio_x, ratio_y, buffermanager, shared_memory_pages);
	}
	else if(colorspace == "rgba32") {
	  image = new XShmImg(this, width, height, title, buffermanager, shared_memory_pages);
	}
	else if(colorspace == "rgb32") {
	  image = new XShmImg(this, width, height, title, buffermanager, shared_memory_pages);
	}
	else if(colorspace == "rgb24") {
	  image = new XShmImg(this, width, height, title, buffermanager, shared_memory_pages);
	}
	else if(colorspace == "rgb565") {
	  image = new XShmImg(this, width, height, title, buffermanager, shared_memory_pages);
	}
	else {
	  ERROR_STREAM( "No matching colorspace found for " << colorspace << endl);
	  return FAILURE;
	}
      }

      else {
	DEBUG_STREAM("Shared X memory can not be used. Taking non-shared-memory XImg" << endl);
	if(colorspace == "yuv422") {
	  image = new XvImg(this, width, height, title, XvImg::XV_FORMAT_UYVY, fullscreen, ratio_x, ratio_y);
	}
	else if(colorspace == "yv12") {
	  image = new XvImg(this, width, height, title, XvImg::XV_FORMAT_YV12, fullscreen, ratio_x, ratio_y);
	}
	else if(colorspace == "rgba32") {
	  image = new XImg(this, width, height, title);
	}
	else if(colorspace == "rgb32") {
	  image = new XImg(this, width, height, title);
	}
	else if(colorspace == "rgb24") {
	  image = new XImg(this, width, height, title);
	}
	else if(colorspace == "rgb565") {
	  image = new XImg(this, width, height, title);
	}
	else {
	  return FAILURE;
	}
      }
     
      image->setAlignment(align,xPos,yPos);
      // we have some image now; initialize it.
      image->init();
    } // end if(image == NULL)
    else {
      // If new width or height, resize and store values
      if((width != i_format->getIntValue(Format::x_resolution_param)) ||
	 (height!= i_format->getIntValue(Format::y_resolution_param))) {
	width  = i_format->getIntValue(Format::x_resolution_param);
	height = i_format->getIntValue(Format::y_resolution_param);
	setResolution(width, height);
      }
    }

    if (display_events_enabled) {
      // if display events were enabled last time, we enable them
      // again this time
      enableDisplayEvents(display_events_config);
    }

    return SUCCESS;
  }
  
  void XDisplayNode::prepareBuffer(Buffer *in_buffer) {
    DEBUG_STREAM( " prepareBuffer, with used size " << in_buffer->getUsed() << endl );

    if( image != 0 ) {
      
      image->prepareOutput( in_buffer );
    }
  }

  Message* XDisplayNode::presentBuffer(Buffer *in_buffer) {
    
    DEBUG_STREAM( " presentBuffer " << endl );
    if( image != 0 ) {
      image->update( in_buffer );

      DEBUG_STREAM( " updated image " << endl );
      if(benchmark.getEnabled()) {
        benchmark.benchmark(in_buffer->getUsed());
        if(!fullscreen && (benchmark.getFrameCount() % title_update_rate) == 0) {
          sprintf(frame_title, "%s %.1f fps", getName().c_str(), benchmark.getFPS());
  	  image->setTitle(string(frame_title));
        }
      }
    }
    in_buffer->release();
    return 0;
  }

  void XDisplayNode::refresh() 
  {
  }
  
  Result XDisplayNode::doFlush() {
    return( SUCCESS );
  }
  
  Result XDisplayNode::setFormat(const Format& format) {
    DEBUG_STREAM("XDisplayNode::doSetFormat" << endl);

    int res_x = format.getIntValue(Format::x_resolution_param);
    int res_y = format.getIntValue(Format::y_resolution_param);
    setResolution(res_x, res_y);
    
    int ratio_x = format.getIntValue(Format::ratio_x_param);
    int ratio_y = format.getIntValue(Format::ratio_y_param);
    DEBUG_STREAM("XDisplayNode::doSetFormat : " << res_x << " x " << res_y << ", ratio " << ratio_x << ":" << ratio_y << endl);
    setRatio(ratio_x, ratio_y);

    return SUCCESS;
  }
  
  void XDisplayNode::enableDisplayEvents(const string& _config_file /* = "" */) {
    display_events_enabled = true;
    display_events_config = _config_file;
    image->enableDisplayEvents(_config_file);
  }
  
  void XDisplayNode::disableDisplayEvents() {
    display_events_enabled = false;
    image->disableDisplayEvents();
  }

  void XDisplayNode::setUsingKeySymKeyEvents(const nmm_bool& ukke) {
    image->setUsingKeySymKeyEvents(ukke);
  }

  nmm_bool XDisplayNode::isUsingKeySymKeyEvents() {
    return image->isUsingKeySymKeyEvents();
  }

  void XDisplayNode::setSendingMouseEventsDirect(const nmm_bool& smed) {
    image->setSendingMouseEventsDirect(smed);
  }

  nmm_bool XDisplayNode::isSendingMouseEventsDirect(){
    return image->isSendingMouseEventsDirect();
  }

  void XDisplayNode::setXAutoRepeat(const nmm_bool& on) {
    image->setXAutoRepeat(on);
  }

  nmm_bool XDisplayNode::isXAutoRepeat() {
    return image->isXAutoRepeat();
  }


  nmm_bool XDisplayNode::makeScreenshot(const string & /* filename */) const {
    if (!image) {
      ERROR_STREAM("No Image allocated yet!" << endl);
      return false;
    }

    char* data = image->getImage();
    if (!data) {
      ERROR_STREAM("Could not get image data!" << endl);
      return false;
    }

    MESSAGE_STREAM("Width: " << width << " Height: " << height << " Colorspace: " << colorspace << endl);
    
    ERROR_STREAM("OOPS!!! I forgot to port PNGUtils!!! Please contact me -fw" << endl);
    // XXX: fw: PNGUtils has not been ported yet
//     PNGUtils png;
//     if (colorspace == "yv12") {
//       png.setFormat(PNGUtils::YV12);
//     }
//     png.setDimension(width, height);
//     png.writeImage(filename, reinterpret_cast<unsigned char*>(data));

    return true;
  }

  void XDisplayNode::setRightWindowAlignment() {
    align = RIGHT;
  }

  void XDisplayNode::setLeftWindowAlignment() {
    align = LEFT;
  }

  void XDisplayNode::setCenterWindowAlignment() {
    align = CENTER;
  }
  
  void XDisplayNode::useDotCursor(const nmm_bool& value) {
    if (!image) {
      ERROR_STREAM("No Image allocated yet!" << endl);
    } else {
      image->useDotCursor(value);
    }
  }

  void XDisplayNode::useTriangleCursor(const nmm_bool& value) {
    if (!image) {
      ERROR_STREAM("No Image allocated yet!" << endl);
    } else {
      image->useTriangleCursor(value);
    }
  }

  void XDisplayNode::useStandardCursor() {
    if (!image) {
      ERROR_STREAM("No Image allocated yet!" << endl);
    } else {
      image->useStandardCursor();
    }
  }

  nmm_uint32 XDisplayNode::getWindowID() {
    return image->getWindow();
  }

}

#include "nmm/nmm_suffix.hpp"
