/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "XProducer.hpp"
#include "XVirtualImg.hpp"

#include "nmm/plugins/video/display/interfaces/IDisplayEvents.hpp"

#include "nmm/utils/xml/XMLTree.hpp"
#include "nmm/utils/xml/XMLNode.hpp"

#include "nmm/comm/messaging/Event.hpp"

#define XK_MISCELLANY
#include <X11/keysymdef.h>

#define BUF_LEN 80

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  XProducer::InternalReceiver::InternalReceiver()
    : dispatcher(NULL)
  {}
  
  void XProducer::InternalReceiver::connectTo(EventDispatcher* dispatcher) {
    this->dispatcher = dispatcher;
  }
  
  Result XProducer::InternalReceiver::receiveEvent(Event& event, const IEventSender* sender /* = 0 */) {
    if (!dispatcher) {
      return FAILURE;
    }
    
    return dispatcher->dispatch(event, sender);
  }
  
  XProducer::XProducer(XVirtualImg* xvi)
    : display_id(NULL),
      x_auto_repeat(true),
      SendingMouseEventsDirect(false),
      UsingKeySymKeyEvents(false),
      xvirtualimg(xvi)
  {
    /* Initialize Key-Mapping with NULL-Pointer */
    for( int x=0; x<KEY_NUM; x++ )
      keyevent[x] = NULL;
  }
  
  XProducer::~XProducer(){
    if (display_id != NULL) {
      XAutoRepeatOn(display_id);
    }
    
    delete_keyevents();
    delete_mousecoordinates();
  }
  
  bool XProducer::isXAutoRepeat() {
    return x_auto_repeat;
  }
  
  void XProducer::setXAutoRepeat(const bool& x_auto_repeat) {
    this->x_auto_repeat = x_auto_repeat;
  }
  
  void XProducer::setSendingMouseEventsDirect(bool samed){
    this->SendingMouseEventsDirect = samed;	
  }

  bool XProducer::isSendingMouseEventsDirect(){
    return this->SendingMouseEventsDirect;	
  }
  
  void XProducer::setUsingKeySymKeyEvents(bool ukke){
    this->UsingKeySymKeyEvents = ukke;	
  }
  
  bool XProducer::isUsingKeySymKeyEvents(){
    return this->UsingKeySymKeyEvents;	
  }
  
  void XProducer::delete_keyevents() {
    int x;

    for( x=0; x<KEY_NUM; x++ )
      if ( keyevent[x] ) { delete[] keyevent[x]; keyevent[x] = NULL; }
  }

  bool XProducer::add_key( int k, string event) {
    if ( (k < KEY_NUM) && (keyevent[k]==NULL) ) {
      keyevent[k] = new char[event.length() + 1];
      strcpy( keyevent[k], event.c_str() );
      return true;
    }

    return false;
  }

  void XProducer::delete_mousecoordinates() {
    vector<button_coordinate>::iterator mc = bc.begin();

    /* delete all allocated memory that was allocated for the event-string */
    while( mc != bc.end() ) {
      if ( mc->event ) delete[] mc->event;
      mc++;
    };

    /* delete vector */
    bc.erase( bc.begin(), bc.end() );

  }

  bool XProducer::add_mousecoordinates( int x1, int y1, int x2, int y2, string event) {
    button_coordinate b;

    b.x1 = x1;
    b.x2 = x2;
    b.y1 = y1;
    b.y2 = y2;
    b.event = new char[event.length() + 1];
    strcpy( b.event, event.c_str() );

    // Put entry into the list
    bc.push_back( b );
    return true;
  }

  char *XProducer::checkcoordinate( int x, int y ) {
    vector<button_coordinate>::iterator mc = bc.begin();

    /* Search until point is within a rectangle  */
    while( mc != bc.end() ) {
      if ( (mc->x1 < x) && (mc->x2 > x) )
	if ( (mc->y1 < y) && (mc->y2 > y) ) return( mc->event );
      mc++;
    };

    /* No rectangle found */
    return( NULL );
  }

  bool XProducer::readConfig( string fname ){
    /* open XML document */
    XMLTree* xml = new XMLTree( fname );

    /* get pointer to XML root */
    XMLNode* rootnode = xml->root();
    if ( rootnode == NULL ) return false;

    if ( rootnode->name() != "configuration" ) return false;

    XMLNode* node1;
  
    /**
     * First parse the keyboard configuration
     */
    /* Try to find configuration->keyboard entry */
    XMLNodeList nlist;
    XMLNodeIterator iter, stop;
    
    nlist = rootnode->children("keyboard");
    iter = nlist.begin();
    /* If not found return */
    if ( iter == nlist.end() ) return( true );

    /* now node1 points to the configuration->keyboard entry */
    node1 = *iter;

    /* Get a list with all <event> entries */
    nlist = node1->children( "event" );

    iter = nlist.begin();
    stop = nlist.end();

    /* Process each <event> entry */
    while ( iter != stop ) {
      XMLProperty* prop;
      XMLNode* node = *iter;
      iter++;

      /* Try to find <event> entry with "key" attribute */
      string key;
      prop = node->property( "key" );
      if ( prop ) key = prop->value();

      /* now the cdata-string of <event> tag containing the event-string */
      string event = node->getData();
      GLOBAL_DEBUG_STREAM("Found Key: " << key << " -> " << event << endl);

      int k;
      /* All "key" attributes with length=1 will correspond to the key itself */
      if ( key.length() == 1 ) {
	k = key[0];
	add_key( k, event);
      }
      else {  /* All other keys are keywords for the corresponding keys (like F1, SHIFT, ...) */
	if ( key == "SHIFT_L" )  add_key( XK_Shift_L , event );		
	if ( key == "SHIFT_R" )  add_key( XK_Shift_R , event );
	if ( key == "SHIFT_LOCK" )  add_key( XK_Shift_Lock , event );
	if ( key == "BACKSPACE" )  add_key( XK_BackSpace , event );
	if ( key == "CONTROL_L" )  add_key( XK_Control_L , event );
	if ( key == "CONTROL_R" )  add_key( XK_Control_R , event );
	if ( key == "ALT_L" )  add_key( XK_Alt_L , event );
	if ( key == "ALT_R" )  add_key( XK_Alt_R , event );
	if ( key == "TAB" )  add_key( XK_Tab , event );

	if ( key == "F1" )  add_key( XK_F1 , event );
	if ( key == "F2" )  add_key( XK_F2 , event );
	if ( key == "F3" )  add_key( XK_F3 , event );
	if ( key == "F4" )  add_key( XK_F4 , event );
	if ( key == "F5" )  add_key( XK_F5 , event );
	if ( key == "F6" )  add_key( XK_F6 , event );
	if ( key == "F7" )  add_key( XK_F7 , event );
	if ( key == "F8" )  add_key( XK_F8 , event );
	if ( key == "F9" )  add_key( XK_F9 , event );
	if ( key == "F10" ) add_key( XK_F10 , event );
	if ( key == "F11" ) add_key( XK_F11 , event );
	if ( key == "F12" ) add_key( XK_F12 , event );

	if ( key == "LEFT" )  add_key( XK_Left , event );
	if ( key == "RIGHT" ) add_key( XK_Right , event );
	if ( key == "UP" )    add_key( XK_Up , event );
	if ( key == "DOWN" )  add_key( XK_Down , event );

	if ( key == "ESC" )  add_key( XK_Escape , event );
	if ( key == "RETURN" )  add_key( XK_Return , event );

	if ( key == "PAGEUP"   ) { add_key ( XK_Prior , event ); }
	if ( key == "PAGEDOWN" ) { add_key ( XK_Next , event ); }
	if ( key == "HOME"     ) { add_key ( XK_Home , event ); }
	if ( key == "END"      ) { add_key ( XK_End , event ); }
	if ( key == "INSERT"   ) { add_key ( XK_Insert , event ); }
	if ( key == "DELETE"   ) { add_key ( 127 , event ); }

	/* KP- means KeyPad keys */
	if ( key == "KP-DIV"   ) { add_key ( XK_KP_Divide , event ); }
	if ( key == "KP-MUL"   ) { add_key ( XK_KP_Multiply , event ); }
	if ( key == "KP-SUB"   ) { add_key ( XK_KP_Subtract , event ); }
	if ( key == "KP-ADD"   ) { add_key ( XK_KP_Add , event ); }
	if ( key == "KP-ENTER" ) { add_key ( XK_KP_Enter , event ); }
      }

      //fprintf(stderr, "%d -> %s\n", k, keyevent[k]);
    }

    return(true);
  }

  //  void XProducer::registerXDisplayNode(XDisplayNode *d )
  //{
  //  display_node = d;
  //  registerDisplay( display_node->getID() );
  //}
  
  void XProducer::registerDisplay( Display *displayid ) {

    display_id = displayid;

    // Get filedescriptor for Xdisplay
    setFileDescriptor( ConnectionNumber(displayid) );
    
  }

  void XProducer::handleInput(){
    XEvent       event;
    XKeyEvent    *kevent;
    XButtonEvent *bevent;
    XMotionEvent *mevent;
    const string str;
    int          count;
    char         buffer[BUF_LEN];
    KeySym       keysym;
    int          ret;
    char         *cret;
    Event        ev("none");
    Event        ev1("none",1);
    Event        ev4("none",4); 
    Event        ev5("none",5);
    
    // Get event
    //XNextEvent(display_id, &event);
    while ( XCheckMaskEvent(display_id,
                            XVirtualImg::getEventMask(),
                            &event ) ) {
      
      // XSync(display_id, False);
      
      
      switch(event.type){
      case FocusIn:
        {
          if (!isXAutoRepeat()) {
            XAutoRepeatOff(display_id);
          }
          break;
        }
        
      case FocusOut:
        {
          if (!isXAutoRepeat()) {
            XAutoRepeatOn(display_id);
          }
          break;
        }
        
      case Expose:
        {
          auto_ptr<Event> event(IDisplayEvents::create_windowExposed());
          getEventReceiver()->receiveEvent(*event, this);
          
          //display_node->refresh();
          break;
        }
        
      case KeyRelease:
      case KeyPress:
        {
          if(!UsingKeySymKeyEvents) {
            
            kevent = (XKeyEvent *) &event;
            
            /* Map a key event to ASCII string and keysym */
            count = XLookupString( kevent, buffer, BUF_LEN, &keysym, NULL );
            
            /* use keysym to get code for function-keys */
            if ( buffer[0] < 48 ) ret = keysym;
            else ret = buffer[0];			  
            if ( ret > 0xffff ) ret = 0;
            
            if ( keyevent[ret] ) {
              // Push the event to the listener object
              if (event.type == KeyPress) {
                auto_ptr<Event> event(IDisplayEvents::create_keyPressed(string(keyevent[ret])));
                getEventReceiver()->receiveEvent(*event, this);
              } else { //KeyPressed
                auto_ptr<Event> event(IDisplayEvents::create_keyReleased(string(keyevent[ret])));
                getEventReceiver()->receiveEvent(*event, this);
              }
            }
          } else { //UsingKeySymKeyEvents
            
            kevent = (XKeyEvent *) &event;
            
            /* Map a key event to ASCII string and keysym */
            count = XLookupString( kevent, buffer, BUF_LEN, &keysym, NULL );
            
            /* use keysym to get code for function-keys */
            if ( buffer[0] < 48 ) ret = keysym;
            else ret = buffer[0];
            
            if ( ret > 0xffff ) ret = 0;
            
            if ( keyevent[ret] ) {
              // Push the event to the listener object
              if (event.type == KeyPress) {
                
                auto_ptr<Event> event(IDisplayEvents::create_keysym_keyPressed(keysym));
                getEventReceiver()->receiveEvent(*event, this);
                
              } else { //KeyRelease
                
                auto_ptr<Event> event(IDisplayEvents::create_keysym_keyReleased(keysym));
                getEventReceiver()->receiveEvent(*event, this);
                
              }							    		
            }
          }
          break;
        }
        
      default:
        
        if(!SendingMouseEventsDirect){
          
          switch(event.type){
	    
          case ButtonRelease:
            
            bevent = (XButtonEvent *) &event;
            cret = checkcoordinate( bevent->x, bevent->y );
            
            if ( cret ) {
              // Push the event to the listener object
              
              ev.setKey( cret );
              getEventReceiver()->receiveEvent(ev, this);
            }
            break;
            
          default:
            GLOBAL_ERROR_STREAM("Xproducer: got unknown event: " << event.type << endl);
            break;
            
            
          } //switch
          
	  
        } else { //SendingAllMouseEventsDirect
          
          switch(event.type){
            
          case MotionNotify:
            {
              mevent = (XMotionEvent *) &event;
              auto_ptr<Event> event(IDisplayEvents::create_mouseMoved
                                    (mevent->x, mevent->y, mevent->state, mevent->state & 1));
              
              getEventReceiver()->receiveEvent(*event, this);
              
              break;
            }
            
          case ButtonPress:
            {
              bevent = (XButtonEvent *) &event;
              auto_ptr<Event> event(IDisplayEvents::create_mouseButtonPressed
                                    (bevent->button, bevent->x, bevent->y, bevent->state, bevent->state & 1));
              
              getEventReceiver()->receiveEvent(*event, this);
              
              break;
            }
            
          case ButtonRelease:
            {
              bevent = (XButtonEvent *) &event;            
              auto_ptr<Event> event(IDisplayEvents::create_mouseButtonReleased
                                    (bevent->button, bevent->x, bevent->y, bevent->state, bevent->state & 1));
              
              getEventReceiver()->receiveEvent(*event, this);
              
              break;
            }
            
          default:
            GLOBAL_ERROR_STREAM("Xproducer: got unknown event: " << event.type << endl);
            break;
            
          }
        }
      }
    } 
  }
  
  bool XProducer::getKey(char& retu) {
    XEvent       event;
    XKeyEvent    *kevent;
    int          count;
    char         buffer[BUF_LEN];
    KeySym       keysym;
    int          ret=-1;
    
    if(XCheckMaskEvent(display_id, XVirtualImg::getEventMask(), &event)) {
      switch(event.type){
      case KeyPress:
	kevent = (XKeyEvent *) &event;
        
	/* Map a key event to ASCII string and keysym */
	count = XLookupString( kevent, buffer, BUF_LEN, &keysym, NULL );
        
	/* use keysym to get code for function-keys */
	if ( buffer[0] < 48 ) ret = keysym;
	else ret = buffer[0];
	
	if ( ret > 0xffff ) ret = 0;
	break;
      } // switch
      
      switch( ret ) {
      case XK_F1:
	ret = F1;
	break;
      case XK_F2:
	ret = F2;
	break;
      case XK_F3:
	ret = F3;
	break;
      case XK_F4:
	ret = F4;
	break;
      case XK_F5:
	ret = F5;
	break;
      case XK_F6:
	ret = F6;
	break;
      case XK_F7:
	ret = F7;
	break;
      case XK_F8:
	ret = F8;
	break;
      case XK_F9:
	ret = F9;
	break;
      case XK_F10:
	ret = F10;
	break;
      case XK_F11:
	ret = F11;
	break;
      case XK_F12:
	ret = F12;
	break;
      case XK_Left:
	ret = LEFT;
	break;
      case XK_Right:
	ret = RIGHT;
	break;
      case XK_Up:
	ret = UP;
	break;
      case XK_Down:
	ret = DOWN;
	break;
      case XK_Escape:
	ret = ESCAPE;
	break;
      case XK_Return:
	ret = RETURN;
	break;
      case XK_Prior:
	ret = PAGEUP;
	break;
      case XK_Next:
	ret = PAGEDOWN;
	break;
      case XK_Home:
	ret = HOME;
	break;
      case XK_End:
	ret = END;
	break;
      case XK_Insert:
	ret = INSERT;
	break;
      case XK_Delete:
	ret = DELETE;
	break;
      }
      retu = (char) ret;
      return true;
    }
    return false;
  }
  
  void XProducer::connectTo(EventDispatcher* dispatcher) {
    internal_receiver.connectTo(dispatcher);
    setEventReceiver(&internal_receiver);
  }
}

#include "nmm/nmm_suffix.hpp"
