/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "XShmImg.hpp"

#include "nmm/comm/messaging/Buffer.hpp"
#include "nmm/multimedia/Node.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  XShmImg::XShmImg(Node* parent,
                   int iwidth, int iheight,
		   string ititle,
                   ShmBufferManager* manager, unsigned int shared_pages)
    : XImg(parent, iwidth, iheight, ititle),
      shm(manager, shared_pages) {
  }

  XShmImg::~XShmImg() {
    // this is not very clean: as freeImage() is virtual and might be
    // overloaded in derived classes, we test here for a
    // NULL-Ptr and only call freeImage() if we still have a ximage
    if (ximage) {
      deinitImage();
    }
  }

  bool XShmImg::initImage() {
    // query XShmManager if XShm extension is available
    if(!shm.queryDisplay(getDisplay())) {
      NODE_ERROR_STREAM(getParent(), "XShmImg> no SHM extension available!" << endl);
      return false;
    }

    // allocate ximage using "Default*" macros for visual and depth
    // and NULL-Ptr for the image data (will be replaced on present
    // with the char-Ptr of a shared memory buffer). XShmManager
    // provides the needed XShmSegmentInfo struct-Ptr.Variables
    // starting with '_' are provided by XVirtualImg.
    ximage = XShmCreateImage(getDisplay(), //Display
                             DefaultVisual(getDisplay(), DefaultScreen(getDisplay())), //Visual
                             DefaultDepth(getDisplay(), DefaultScreen(getDisplay())), //depth
                             ZPixmap, //format
                             NULL, //data
                             shm.getInfo(), //XShmSegmentInfo
                             getWidth(), //width
                             getHeight()); //height

    if(!ximage) {
      NODE_ERROR_STREAM(getParent(), "XShmImg> cannot create XShmImage" << endl);
      return false;
    }

    // Let the XShmManager attach his shared memory segment to the
    // display
    // buffer size for one shared memory page is set to
    // (width * height * sizeof(RGBA))
    if (!shm.attachMemory(getDisplay(), getWidth()*getHeight()*4)) {
      NODE_ERROR_STREAM(getParent(), "XShmImg> cannot create SHM" << endl);
      XFree(ximage);
      return false;
    }

    // sync the display, to be sure image creation is forwarded to the
    // XServer, not discarding incoming events
    XSync(getDisplay(), False);

    return true;
  }

  void XShmImg::updateMemoryManager(ShmBufferManager* manager, unsigned int shared_pages) {
    // detach previous attached shared memory segment
    deinitImage();

    shm.updateMemoryManager(manager, shared_pages);

    // reattach new image & shared memory segment
    initImage();
  }

  void XShmImg::deinitImage() {
    if (ximage) {
      shm.detachMemory();
      XFree(ximage);
      ximage = NULL;
    }
  }

  void XShmImg::prepareOutput(Buffer* in_buffer) {
    // not simple forward pointer, but ask ShmManager to ensure
    // data-Ptr is from the shared memory segment.
    ximage->data = shm.getMem(in_buffer->getData());
  }

  void XShmImg::update(Buffer*) {
    NODE_DEBUG_STREAM(getParent(), "XShmImg> update begin" << endl);
    // put the image ximage to the Window getWindow() on Display getDisplay(),
    // using context getContext(), width getWidth() and height getHeight(). Variables
    // starting with '_' are provided by XVirtualImg. Presentation
    // will have zero offset in x and y direction (specified by src_x
    // and src_y parameters).
    XShmPutImage(getDisplay(), getWindow(), getContext(), ximage,
                 0, 0, 0, 0, // src_x, src_y, dest_x, dest_y
                 getWidth(), getHeight(),
                 False); // do not sent completion event, instead:
    // sync the display, to be sure image is forwarded to the
    // XServer, not discarding incoming events
    XSync(getDisplay(), False);
    // now, we can modify the data of the ximage again
    NODE_DEBUG_STREAM(getParent(), "XShmImg> update end" << endl);
  }

#include "nmm/nmm_prefix.hpp"

} // namespace NMM

#include "nmm/nmm_suffix.hpp"
