/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "XShmManager.hpp"

#include "nmm/plugins/video/display/X/ShmBufferManager.hpp"

#include <sys/ipc.h>
#include <sys/shm.h>
#include <errno.h>

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  ShmManager::ShmManager(ShmBufferManager* manager, unsigned int shared_pages)
    : buffermanager(manager),
      shmid(-1),
      shmaddr(0),
      buffer_size(0),
      shared_memory_pages(shared_pages) {
    updateMemoryManager(buffermanager, shared_memory_pages);
  }

  ShmManager::~ShmManager() {
    detachMemory();
  }

  bool ShmManager::attachMemory(const int size) {
    buffer_size = size;
    
    /* 1) get shared memory segment */
    int alloc_size = buffer_size;
    if (buffermanager) {
      alloc_size *= shared_memory_pages + 1; // reserve one page for memcpys in getMem()
      // allocated size must be lower than SHMMAX = 0x2000000
      // todo: determine maximum dynamically
      while (alloc_size > 0x2000000) {
        // allocate size to large, reduce it by one page
        alloc_size -= size;
        shared_memory_pages--;
      }
    }
    shmid = shmget(IPC_PRIVATE, alloc_size, IPC_CREAT|0777);
    if(shmid < 0) {
      GLOBAL_ERROR_STREAM("ShmManager> cannot get memory" << endl);
      switch (errno) {
      case EINVAL:
        GLOBAL_ERROR_STREAM("a new segment was to be created and size < SHMMIN or size > SHMMAX, or no new segment was  to  be  created,  a segment with given key existed, but size is greater than the size of that segment." << endl);
        break;

      case EEXIST:
        GLOBAL_ERROR_STREAM("IPC_CREAT | IPC_EXCL was specified and the segment exists." << endl);
        break;

      case EIDRM:
        GLOBAL_ERROR_STREAM("the segment is marked as destroyed, or was removed." << endl);
        break;

      case ENOSPC:
        GLOBAL_ERROR_STREAM("all  possible  shared memory id's have been taken (SHMMNI) or if allocating a segment of the requested size would cause the system to exceed the system-wide limit on shared memory (SHMALL)." << endl);
        break;

      case ENOENT:
        GLOBAL_ERROR_STREAM("no segment exists for the given key, and IPC_CREAT was not specified." << endl);
        break;

      case EACCES:
        GLOBAL_ERROR_STREAM("the user does not have permission to access the shared memory segment." << endl);
        break;

      case ENOMEM:
        GLOBAL_ERROR_STREAM("no memory could be allocated for segment overhead." << endl);
        break;

      default:
        GLOBAL_ERROR_STREAM("Unknown error!" << endl);
        break;
      }
      return false;
    }

    /* 2) attach memory segment to user data space and get pointer to it */
    shmaddr = reinterpret_cast<char*>(shmat(shmid, 0, 0));
    if (shmaddr < 0) {
      GLOBAL_ERROR_STREAM("ShmManager> cannot attach memory" << endl);
      switch (errno) {
      case EACCES:
        GLOBAL_ERROR_STREAM("The calling process has no access permissions for the requested attach type." << endl);
        break;

      case EINVAL:
        GLOBAL_ERROR_STREAM("Invalid shmid value, unaligned (i.e., not page-aligned and SHM_RND was not specified) or invalid shmaddr value, or failing attach at brk." << endl);
        break;

      case ENOMEM:
        GLOBAL_ERROR_STREAM("Could not allocate memory for the descriptor or for the page tables." << endl);
        break;

      default:
        GLOBAL_ERROR_STREAM("Unknown error!" << endl);
        break;
      }
      detachMemory();
      return false;
    }

    /* 3) add shared memory segment to buffermanager */
    if (buffermanager) {
      GLOBAL_DEBUG_STREAM("Adding " << shared_memory_pages << " pages SHM to Buffermanager" << endl);
      buffermanager->setSharedMemory(shmaddr + buffer_size, // reserve first page for memcpys in getMem()
                                     getShmEnd() - (shmaddr + buffer_size));
    }

    return true;
  }

  void ShmManager::detachMemory() {
    if (shmaddr > 0) {
      shmdt(shmaddr);
      shmaddr = 0;
    }
    if (shmid >= 0) {
      shmctl(shmid, IPC_RMID, 0);
      shmid = -1;
    }
  }

  void ShmManager::updateMemoryManager(ShmBufferManager* manager, unsigned int shared_pages) {
    buffermanager = manager;
    if (buffermanager) {
      shared_memory_pages = shared_pages;
    } else {
      shared_memory_pages = 1;
    }
  }

  bool ShmManager::isSharedMemory(const char* data) const {
    return ((data >= shmaddr) && (data < getShmEnd()));
  }

  char* ShmManager::getMem(char* data) const {
    char* retval = shmaddr;

    /* copy buffer ony if no shared memory buffer is used */
    if (!isSharedMemory(data)) {
      GLOBAL_DEBUG_STREAM("ShmManager::getMem()> Copy Buffer" << endl);
      /* copy in buffer data to display area */
      memcpy(retval, data, buffer_size);
    } else {
      GLOBAL_DEBUG_STREAM("ShmManager::getMem()> Using shm-Buffer" << endl);
      retval = data;
    }

    return retval;
  }

  const char* ShmManager::getShmEnd() const {
    if (buffermanager) {
      return shmaddr + (shared_memory_pages * buffer_size);
    } else {
      return shmaddr + buffer_size;
    }
  }

  int ShmManager::getID() {
    return shmid;
  }

  char* ShmManager::getShm() {
    return shmaddr;
  }


  XShmManager::XShmManager(ShmBufferManager* manager, unsigned int shared_pages)
    : ShmManager(manager, shared_pages),
      xshminfo(),
      attached_display(0) {
  }

  XShmManager::~XShmManager() {
    detachMemory();
  }

  bool XShmManager::queryDisplay(Display* display) {
    if(!XShmQueryExtension(display)) {
      GLOBAL_ERROR_STREAM("ShmManager> no SHM extension available!" << endl);
      return false;
    }
    return true;
  }

  bool XShmManager::attachMemory(Display* display, const int size) {
    if (!ShmManager::attachMemory(size)) {
      GLOBAL_ERROR_STREAM("Could not attach SHM in ShmManager!" << endl);
      return false;
    }

    xshminfo.shmid = getID();
    xshminfo.shmaddr = getShm();
    xshminfo.readOnly = false;

    /* 4) attach memory segment to X display */
    if(!XShmAttach(display, &xshminfo)) {
      GLOBAL_ERROR_STREAM("ShmManager> cannot attach SHM to X" << endl);
      detachMemory();
      return false;
    }

    attached_display = display;

    return true;
  }

  void XShmManager::detachMemory() {
    if (attached_display) {
      XShmDetach(attached_display, &xshminfo);
      attached_display = 0;
    }
    ShmManager::detachMemory();
  }

  XShmSegmentInfo* XShmManager::getInfo() {
    return &xshminfo;
  }

#include "nmm/nmm_prefix.hpp"

} // namespace NMM

#include "nmm/nmm_suffix.hpp"
