/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "XVirtualImg.hpp"
#include "XProducer.hpp"

#include "nmm/multimedia/Node.hpp"

#include "nmm/utils/time/Rational.hpp"

#include "nmm/utils/directory_utils.hpp"

#ifdef HAVE_MWMUTIL_H
#include <Xm/MwmUtil.h>
#endif

#include <cassert>

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  XVirtualImg::XVirtualImg(Node* parent, int width, int height, const std::string& title)
    : __parent(parent),
      __xproducer(0),
      __direct_mouse(false),
      __direct_keyboard(false),
      __xpos(0),
      __ypos(0),
      __width(width),
      __height(height),
      __ratio_x(width),
      __ratio_y(height),
      __scaled_width(0),
      __scaled_height(0),
      __full_screen_width(0),
      __full_screen_height(0),
      __screen_width(0),
      __screen_height(0),
      __x_blit_offset(0),
      __y_blit_offset(0),
      __title(title),
      __fullscreen(false),
      __screen(-1),
      __depth(0),
      __align(1),
      __have_ewmh(false),
      __display(0),
      __win(0),
      __gc(0),
      __visual_info(0)
  {
    initCommon();
  }
  
  
  XVirtualImg::XVirtualImg(Node* parent, int width, int height, int ratio_x, int ratio_y, bool fullscreen, const std::string& title)
    : __parent(parent),
      __xproducer(0),
      __direct_mouse(false),
      __direct_keyboard(false),
      __xpos(0),
      __ypos(0),
      __width(width),
      __height(height),
      __ratio_x(ratio_x),
      __ratio_y(ratio_y),
      __scaled_width(0),
      __scaled_height(0),
      __full_screen_width(0),
      __full_screen_height(0),
      __screen_width(0),
      __screen_height(0),
      __x_blit_offset(0),
      __y_blit_offset(0),
      __title(title),
      __fullscreen(fullscreen),
      __screen(-1),
      __depth(0),
      __have_ewmh(false),
      __display(0),
      __win(0),
      __gc(0),
      __visual_info(0)
  {
    initCommon();
  }
  
  void XVirtualImg::initCommon() {    
    memset(&__size_hints, 0, sizeof(__size_hints));
    memset(&__gc, 0, sizeof(__gc));
    
    /* open display */
    __display = XOpenDisplay(0);
    
    if(!__display) {
      NODE_ERROR_STREAM(getParent(), "Can't open display '" << XDisplayName(NULL) << "'" << endl);
    } else {
      /* get screen properties */
      __screen = DefaultScreen(__display);
      __full_screen_width  = DisplayWidth(__display, __screen);
      __full_screen_height = DisplayHeight(__display, __screen);
      
      /* check for EWMH support */
      Atom ewmh_support = XInternAtom(__display, "_NET_SUPPORTED", True);
      Atom ewmh_fullscreen = XInternAtom(__display, "_NET_WM_STATE_FULLSCREEN", True);
      if(ewmh_support && ewmh_fullscreen) {
        Atom* args;
        int format;
        unsigned long num_items;
        unsigned long bytes_after;
        
        /* check for supported states (currently only FULLLSCREEN is needed) */
        if(Success == XGetWindowProperty(__display, RootWindow(__display, __screen), ewmh_support, 0, 16384, False,
                                         AnyPropertyType, &ewmh_support, &format, &num_items,
                                         &bytes_after, reinterpret_cast<unsigned char**>(&args))) {
          for(unsigned long i=0; i < num_items; ++i) {
            if(args[i] == ewmh_fullscreen) {
              __have_ewmh = true;
              break;
            }
          }
          
          XFree(args);
        }
      }
      
      if(__have_ewmh) {
        NODE_DEBUG_STREAM(getParent(), "Detected support for EWMH fullscreen state" << std::endl);
      }
    }    
  }
  
  XVirtualImg::~XVirtualImg() {
    if(__display) {
      XCloseDisplay(__display);
    }
  }

  void XVirtualImg::setTitle(const string& title) {
    __title = title;
    if(__win) {
      XStoreName(__display, __win, __title.c_str());
      NODE_DEBUG_STREAM(getParent(), "XVirtualImg setTitle begin" << endl);
      XSync(__display, False);
      NODE_DEBUG_STREAM(getParent(), "XVirtualImg setTitle end" << endl);
    }
  }
  
  void XVirtualImg::waitForXConfigureNotifyEvent() {
    XEvent event;
    NODE_DEBUG_STREAM(getParent(), "Waiting for ConfigureNotify" << endl);
    for(unsigned int retry=0; retry < 10; ++retry) {
      XWindowEvent(__display, __win, StructureNotifyMask, &event);
      NODE_DEBUG_STREAM(getParent(), "Got Event: " << event.type << endl);
      if(event.type == ConfigureNotify) {
        XWindowAttributes xwa;
 	XGetWindowAttributes(__display, __win, &xwa);
 	NODE_DEBUG_STREAM(getParent(), "waitForXConfigure: got size " << xwa.width << "x"
                          << xwa.height << " should be " << __size_hints.width << "x" << __size_hints.height << endl);
 	if(__have_ewmh || xwa.width == __size_hints.width && xwa.height == __size_hints.height) {
	  break;
        } else {
 	  /* KDE Release 3.1.0 has some problems resizing from fullscreen back to normal window
             This forces KDE to resize and position the window correctely */
 	  XMoveResizeWindow(__display, __win, __size_hints.x, __size_hints.y, __size_hints.width, __size_hints.height);
        }
      }
    }
  }
  
  void XVirtualImg::waitForXUnmapEvent() {
    XEvent event;
    NODE_DEBUG_STREAM(getParent(), "Waiting for UnmapEvent" << endl);
    do {
      XWindowEvent(__display, __win, StructureNotifyMask, &event);
    } while(event.type != UnmapNotify);
    NODE_DEBUG_STREAM(getParent(), "done UnmapEvent" << endl);
  }

  void XVirtualImg::waitForXExposureEvent() {
    XEvent event;
    NODE_DEBUG_STREAM(getParent(), "waiting for Expose" << endl);
    do {
      XWindowEvent(__display, __win, ExposureMask, &event);
    } while(event.type != Expose);
    NODE_DEBUG_STREAM(getParent(), "done Expose" << endl);
  }
  
  void XVirtualImg::calcScreenWidthHeight() {
    /* compute non-fullscreen window size */
    __scaled_width = __width;
    __scaled_height = __height;
    if(__ratio_x && __ratio_y) {
      if(Rational(__scaled_width, __scaled_height) < Rational(__ratio_x, __ratio_y)) {
        __scaled_width = static_cast<int>((static_cast<double>(__ratio_x)/static_cast<double>(__ratio_y)) * __scaled_height);
      } else {
        __scaled_height = static_cast<int>((static_cast<double>(__ratio_y)/static_cast<double>(__ratio_x)) * __scaled_width);
      }
    }
    
    if(__fullscreen) {
      if(__ratio_x && __ratio_y) {
        /* compute screen aspect ratio */
        Rational screen_aspect(__full_screen_width, __full_screen_height);
        
        /* compute image aspect ratio */
        Rational image_aspect(__ratio_x, __ratio_y);
        
        /* check whether to fit to width or to height such that.
           - the image fits on the screen
           - the aspect ratio of the image is preserved
           - the displayed image is as large as possible (ideally it uses the full screen) 
        */
        if(image_aspect > screen_aspect) {
          /* fit to width (e.g. 16:9 video on 4:3 screen)
             screen_width = full_screen_width;
             screen_height = full_screen_width / image_aspect; 
          */
          __screen_width = __full_screen_width;
          __screen_height = static_cast<int>((static_cast<double>(__ratio_y)/static_cast<double>(__ratio_x)) * __screen_width);
        } else {
          /* fit to height (e.g. 4:3 video on 16:9 screen)
             screen_width = full_screen_height * image_aspect;
             screen_height = full_screen_height; 
          */
          __screen_height = __full_screen_height;
          __screen_width = static_cast<int>((static_cast<double>(__ratio_x)/static_cast<double>(__ratio_y)) * __screen_height);
        }
      } else {
        /* no aspect ratio given, scale to full screen */
        __screen_width = __full_screen_width;
        __screen_height = __full_screen_height;
      }
      
      switch(__align) {
      case 0:
        //left
        __x_blit_offset = 0;
        break;
      case 1:
        //center
        __x_blit_offset = (__full_screen_width - __screen_width)/2;
        break;
      case 2:
        //right
        __x_blit_offset = (__full_screen_width - __screen_width);      
        break;
      }
      __y_blit_offset = (__full_screen_height - __screen_height)/2;
      
      // XXX: fw: please leave here for further debugging
//       cerr << "FULLSCREEN SIZE:" << endl
//            << "w,h            :" << __width << ", " << __height << endl
//            << "x:y            :" << __ratio_x << ":" << __ratio_y << endl
//            << "xw,xh          :" << __scaled_width << ", " << __scaled_height << endl
//            << "fw,fh          :" << __full_screen_width << ", " << __full_screen_height << endl
//            << "sw,sh          :" << __screen_width << ", " << __screen_height << endl
//            << "xb,yb          :" << __x_blit_offset << ", " << __y_blit_offset << endl;
    } else {
      /* use the whole window */
      __screen_width  = __scaled_width;
      __screen_height = __scaled_height;
      __y_blit_offset = 0;
      __x_blit_offset = 0;
      
      // XXX: fw: please leave here for further debugging
//       cerr << "WINDOWED SIZE:" << endl
//            << "w,h            :" << __width << ", " << __height << endl
//            << "x:y            :" << __ratio_x << ":" << __ratio_y << endl
//            << "xw,xh          :" << __scaled_width << ", " << __scaled_height << endl
//            << "sw,sh          :" << __screen_width << ", " << __screen_height << endl
//            << "xb,yb          :" << __x_blit_offset << ", " << __y_blit_offset << endl;
    }    
  }
  
  void XVirtualImg::saveXYPosition() {
    //////////////////////////////////////
    // This method does not work yet!!! //
    //////////////////////////////////////

    //   // only save if going to fullscreen
    //   XSizeHints hints;
    //   long supplied = 0;
    //   long* suppliedp = &supplied;
    //   if (_fullscreen) {
    //     XGetWMNormalHints(__display, _win, &hints, suppliedp);
    //     _xpos = hints.x;
    //     _ypos = hints.y;
    //   }

    // Alternative: Doesn't work, too...
    //   XWindowAttributes xwa;
    //   if (_fullscreen) {
    //     XGetWindowAttributes(__display, _win, &xwa);
    //     _xpos = xwa.x;
    //     _ypos = xwa.y;
    //     cerr << "got position: " << xwa.x << ", " << xwa.y << endl;
    //   }
  }

  void XVirtualImg::setPosition() {
    /* assumes correct __real_screen_width and __real_screen_height */
    /* calculate correct __screen_width, __screen_height */
    calcScreenWidthHeight();
    if(__align != 3) {
    switch(__align) {
    case 0:
      //left
      __xpos = 0;
      break;
    case 1:
      //center
      __xpos = (__full_screen_width - __scaled_width) / 2;
    break;
    case 2:
      //right
      __xpos = (__full_screen_width - __scaled_width);
      cerr << __xpos << endl;
      break;
    }
    __ypos = (__full_screen_height - __scaled_height) / 2;
    }
  }

  void XVirtualImg::setSizeHints() {
    /* Expects that __screen_width, __screen_height, __scaled_width and __scaled_height are set correct */

    /* do not set PMaxSize and PMinSize, so window is resizable...
       ... but DO set size_hints.{min,max}_width and __height
       later, otherwise some window managers show a picture of
       much too small size (e.g. 83x20 pixels for a 640x480 image) */
    __size_hints.flags  = PSize | PPosition | PWinGravity;
    
    if(__fullscreen && !__have_ewmh) {
      __size_hints.x      = 0;
      __size_hints.y      = 0;
      __size_hints.width  = __size_hints.max_width  = __size_hints.min_width  = __full_screen_width;
      __size_hints.height = __size_hints.max_height = __size_hints.min_height = __full_screen_height;
    } else {
      __size_hints.x      = __xpos;
      __size_hints.y      = __ypos;
      __size_hints.width  = __size_hints.max_width  = __size_hints.min_width  = __scaled_width;
      __size_hints.height = __size_hints.max_height = __size_hints.min_height = __scaled_height;
    }
    __size_hints.win_gravity = StaticGravity;
  }

  void XVirtualImg::setCursorAndDecoration() {
    showCursor(!__fullscreen);
    
    /* if we have set the EWMH fullscreen state, the window decoration is already turned off. */
    if(!__have_ewmh) {
      showDecorations(!__fullscreen);
    }
  }
  
  void XVirtualImg::showCursor(bool show) {
    if(show) {
      XUndefineCursor(__display, __win);
    } else {
      Cursor no_ptr;
      Pixmap bm_no;
      XColor black,dummy;
      Colormap colormap;
      static char bm_no_data[] = { 0,0,0,0, 0,0,0,0  };
      
      colormap = DefaultColormap(__display, DefaultScreen(__display));
      XAllocNamedColor(__display, colormap, "black", &black, &dummy);
      bm_no  = XCreateBitmapFromData(__display, __win, bm_no_data, 8, 8);
      no_ptr = XCreatePixmapCursor(__display, bm_no, bm_no, &black, &black, 0, 0);
      XDefineCursor(__display, __win, no_ptr);
    }
  }
  
  void XVirtualImg::showDecorations(bool /*show*/) {
    Atom wm_hints = None;
    
    for(;;) {
      // XXX: fw: doesn't work with KDE and breaks WMAKER support ... however the motif WM hints work with KDE < 3.2
//       /* hide decoration for KDE */
//       wm_hints = XInternAtom(__display, "KWM_WIN_DECORATION", True);
//       if(wm_hints != None) {
//         long kwm_hints = show ? 1 : 0;
//         XChangeProperty(__display, __win,
//                         wm_hints, wm_hints, 32,
//                         PropModeReplace,
//                         reinterpret_cast<unsigned char*>(&kwm_hints),
//                         sizeof(kwm_hints)/sizeof(long));
//         break;
//       }
      
      /* hide decoration for MWM */
      wm_hints = XInternAtom( __display, "_MOTIF_WM_HINTS", True );
#ifdef HAVE_MWMUTIL_H
      if(wm_hints != None) {
        if(!show) {
          MotifWmHints motif_wm_hints;
          memset(&motif_wm_hints, 0, sizeof(motif_wm_hints));
          motif_wm_hints.flags = MWM_HINTS_DECORATIONS;
          if(show) {
            motif_wm_hints.decorations = MWM_DECOR_ALL;
          } else {
            motif_wm_hints.decorations = 0;
          }
          
          XChangeProperty(__display, __win, wm_hints, wm_hints, 32,
                          PropModeReplace, reinterpret_cast<unsigned char*>(&motif_wm_hints), sizeof(motif_wm_hints)/sizeof(long));
        } else {
          XDeleteProperty(__display, __win, wm_hints);
        }
        break;
      }
#endif
      
      NODE_ERROR_STREAM(getParent(), "Failed to turn off window decoration" << std::endl);
      break;
    }
  }

  void XVirtualImg::setRatio(int ratio_x, int ratio_y) {
    __ratio_x = ratio_x;
    __ratio_y = ratio_y;
    
    setPosition();
    changeWindow();
  }
  
  void XVirtualImg::setEWMHFullscreen(bool fullscreen) {
    NODE_DEBUG_STREAM(getParent(), "XVirtualImg> Set fullscreen for EWMH " << fullscreen << std::endl);
    
    /* EWMH state actions, see
       http://freedesktop.org/Standards/wm-spec/index.html#id2768769 */
    enum {
      _NET_WM_STATE_REMOVE        = 0,   /* remove/unset property */
      _NET_WM_STATE_ADD           = 1,   /* add/set property */
      _NET_WM_STATE_TOGGLE        = 2    /* toggle property  */
    };
    
    XEvent xev;
    
    /* init X event structure for _NET_WM_FULLSCREEN client msg */
    xev.xclient.type = ClientMessage;
    xev.xclient.serial = 0;
    xev.xclient.send_event = True;
    xev.xclient.message_type = XInternAtom(__display,
                                           "_NET_WM_STATE", True);
    assert(xev.xclient.message_type);
    xev.xclient.window = __win;
    xev.xclient.format = 32;
    xev.xclient.data.l[0] = fullscreen ? _NET_WM_STATE_ADD : _NET_WM_STATE_REMOVE;
    xev.xclient.data.l[1] = XInternAtom(__display,
                                        "_NET_WM_STATE_FULLSCREEN",
                                        True);
    assert(xev.xclient.data.l[1]);
    xev.xclient.data.l[2] = 0;
    xev.xclient.data.l[3] = 0;
    xev.xclient.data.l[4] = 0;
    
    /* finally send that damn thing */
    if(!XSendEvent(__display, DefaultRootWindow(__display), False,
                   SubstructureRedirectMask | SubstructureNotifyMask,
                   &xev))
      {
        /* SHOULD NOT HAPPEN (unless window manager is broken) */
        NODE_ERROR_STREAM(getParent(), "Failed to switch to fullscreen" << std::endl);
      }
  }
  
  void XVirtualImg::setFullscreen(bool fullscreen) {
    /* save old window position */
    if(!__fullscreen && fullscreen) {
      saveXYPosition();
    }
    
    __fullscreen = fullscreen;
    
    /* set EWMH fullscreen state if supported */
    if(__have_ewmh) {
      setEWMHFullscreen(__fullscreen);
    }
    
    /* unmap window */
    if(!__have_ewmh) {
      XSelectInput(__display, __win, StructureNotifyMask);
      XUnmapWindow(__display, __win);
      waitForXUnmapEvent();
    }
    
    /* calculate screen size */
    calcScreenWidthHeight();    

    /* set cursor and window decoration */
    setCursorAndDecoration();
    
    if(!__have_ewmh) {
      /* init size hints */
      setSizeHints();
      
      /* set window properties and move window */
      XSelectInput(__display, __win, StructureNotifyMask);
      XSetWMNormalHints(__display, __win, &__size_hints);
      XSetStandardProperties(__display, __win, __title.c_str(), NULL, None, NULL, 0, &__size_hints);
      XSetTransientForHint(__display, __win, None);
      XMoveResizeWindow(__display, __win, __size_hints.x, __size_hints.y, __size_hints.width, __size_hints.height);
      waitForXConfigureNotifyEvent();
      
      /* map window */
      XSelectInput(__display, __win, StructureNotifyMask | SubstructureNotifyMask | ExposureMask);
      XMapWindow(__display, __win);
      if(!fullscreen) {
        waitForXExposureEvent();
      }
    }
    
    XFlush(__display);
    XSync(__display, False);
    
    /* set standard event mask */
    XSelectInput(__display, __win, WINDOW_MASK);
    
    /* improve karma */
    // XXX: fw: this is a hack, btu it doesn't hurt, and it is necessary.
    // XXX: fw: don't remove the next line without COMPLETE testing
    // XXX: fw: (all window managers, fullscreen/window mode switching, resolution changes...)
    // XXX: fw: X11 is gay
    changeWindow();
  }
  
  void XVirtualImg::initWindow() {
    XWindowAttributes    attribs;
    XSetWindowAttributes xswa;
    unsigned long        xswamask;
    Colormap             color_map;
    
    NODE_DEBUG_STREAM(getParent(), "XVirtualImg> Init window" << std::endl);
    
    /* get root window attributes */
    XGetWindowAttributes(__display, DefaultRootWindow(__display), &attribs);
    __depth = attribs.depth;
    
    /* depth must be one of 15, 16, 24 and 32 */
    if(__depth != 15 && __depth != 16 && __depth != 24 && __depth != 32) {
      __depth = 24;
    }
    
    /* init visual */
    initVisual();
    XMatchVisualInfo(__display, __screen, __depth, TrueColor, __visual_info);
    
    /* create color map */
    color_map = XCreateColormap(__display, RootWindow(__display, __screen), __visual_info->visual, AllocNone);
    
    /* init window attributes */
    xswa.background_pixel  = 0;
    xswa.border_pixel      = 1;
    xswa.colormap          = color_map;
    xswamask = CWBackPixel | CWBorderPixel | CWColormap;

    /* set position for window in non-fullscreen state */
    setPosition();
    
    /* calculate screen size */
    calcScreenWidthHeight();
    
    /* init size hints */
    setSizeHints();
    
    /* create window */
    __win = XCreateWindow(__display, RootWindow(__display, __screen),
                          __size_hints.x, __size_hints.y,
                          __size_hints.width, __size_hints.height,
                          xswa.border_pixel, __visual_info->depth, CopyFromParent,
                          __visual_info->visual, xswamask, &xswa);
    
    /* set cursor and window decoration */
    setCursorAndDecoration();
    
    /* set window properties */
    XSetWMNormalHints(__display, __win, &__size_hints);
    XSetStandardProperties(__display, __win, __title.c_str(), NULL, None, NULL, 0, &__size_hints);
    XSetTransientForHint(__display, __win, None);
    
    /* raise window */
    XRaiseWindow(__display, __win);
    
    /* map window */
    XSelectInput(__display, __win, StructureNotifyMask | SubstructureNotifyMask | ExposureMask);
    XMapWindow(__display, __win);
    waitForXExposureEvent();
    
    /* set standard event mask */
    XSelectInput(__display, __win, WINDOW_MASK);
    
    /* init graphics context */
    initContext();
    
    /* set EWMH fullscreen state if supported */
    if(__have_ewmh) {
      setEWMHFullscreen(__fullscreen);
    }
    
    NODE_DEBUG_STREAM(getParent(), "XVirtualImg> Done init window" << std::endl);
  }
  
  void XVirtualImg::deinitWindow() {
    /* deinit graphics context */
    deinitContext();
    
    /* unmap and destroy window */
    XUnmapWindow(__display, __win);
    XDestroyWindow(__display, __win);
    
    /* deinit visual */
    deinitVisual();
    
    /* synchronize display */
    XSync(__display, True);
  }
  
  void XVirtualImg::initContext() {
    /* create a simple graphics context */
    __gc = XCreateGC(__display, __win, 0, NULL);
    XSetFunction(__display, __gc, GXcopy);
  }
  
  void XVirtualImg::deinitContext() {
    XFreeGC(__display, __gc);
  }
  
  void XVirtualImg::initVisual() {
    /* find a visual matching the given parameters */
    XMatchVisualInfo(__display, __screen, __depth, TrueColor, &__own_visual_info);
    setVisualInfo(&__own_visual_info);
  }
  
  void XVirtualImg::deinitVisual() {
    setVisualInfo(0);
  }
  
  void XVirtualImg::fillRectangle(int x, int y, int w, int h) {
    if(w <= 0 || h <= 0) {
      return;
    }
    
    XFillRectangle(__display, __win, __gc,
		   x,
		   y,
		   w,
		   h);
  }
  
  void XVirtualImg::changeWindow() {
    NODE_DEBUG_STREAM(getParent(), "Change window to: " << __width << ", " << __height <<
		      ", ratio: " << __ratio_x << "x" << __ratio_y <<
		      ", fullscreen: " << (__fullscreen ? "true" : "false") << endl);
    
    /* unmap window */
    if(!__have_ewmh) {
      XSelectInput(__display, __win, StructureNotifyMask);
      XUnmapWindow(__display, __win);
      waitForXUnmapEvent();
    }
    
    /* calculate screen size */
    calcScreenWidthHeight();
    
    /* init size hints */
    setSizeHints();
    
    if(!__have_ewmh) {
      /* set cursor and window decoration */
      setCursorAndDecoration();
    }
    
    /* set window properties and move window */
    XSelectInput(__display, __win, StructureNotifyMask);
    XSetWMNormalHints(__display, __win, &__size_hints);
    XSetStandardProperties(__display, __win, __title.c_str(), NULL, None, NULL, 0, &__size_hints);
    XSetTransientForHint(__display, __win, None);
    XMoveResizeWindow(__display, __win, __size_hints.x, __size_hints.y, __size_hints.width, __size_hints.height);
    waitForXConfigureNotifyEvent();
    
    /* for EWMH, set fullscreen mode again (seems to be needed by FVWM) */
    if(__have_ewmh) {
      setEWMHFullscreen(__fullscreen);
    }
    
    /* fill regions outside the video image */
    if(__fullscreen) {
      if(__have_ewmh) {
        /* top bar */
	fillRectangle(0,
		      0,
		      __full_screen_width,
		      __y_blit_offset);
        
        /* bottom bar */
	fillRectangle(0,
		      __y_blit_offset + __screen_height,
		      __full_screen_width,
		      __full_screen_height - __y_blit_offset - __screen_height);
        
        /* left bar */
	fillRectangle(0,
		      __y_blit_offset,
		      __x_blit_offset,
		      __screen_height);
        
        /* right bar */
	fillRectangle(__x_blit_offset + __screen_width,
		      __y_blit_offset,
		      __full_screen_width - __x_blit_offset - __screen_width,
		      __screen_height);
      } else {
	fillRectangle(0,
		      0,
		      __size_hints.width,
		      __size_hints.height);
      }
    }
    
    /* map window */
    if(!__have_ewmh) {
      XSelectInput(__display, __win, StructureNotifyMask | SubstructureNotifyMask | ExposureMask);
      XMapWindow(__display, __win);
      waitForXExposureEvent();
    }    
    
    XFlush(__display);
    XSync(__display, False);
    
    /* set standard event mask */
    XSelectInput(__display, __win, WINDOW_MASK);
  }
  
  void XVirtualImg::enableDisplayEvents(const string& config_file /*= ""*/) {
    string config = config_file;
    if(!__xproducer) {
      __xproducer = new XProducer(this);
    }
    __xproducer->registerDisplay(getDisplay());
    if(config.empty()) {
      config = getNMMResource("XProducer.xml");
    }
    if(!__xproducer->readConfig(config)) {
      /* if loading fails, try at least with default config */
      __xproducer->readConfig(getNMMResource("XProducer.xml"));
    }
    
    __xproducer->connectTo(&__parent->getEventDispatcher());
    __xproducer->start();
  }
  
  void XVirtualImg::disableDisplayEvents() {
    if(__xproducer) {
      __xproducer->stop();
      delete __xproducer;
      __xproducer = 0;
    }
  }
  
  void XVirtualImg::setUsingKeySymKeyEvents(const bool& ukke) {
    if(__xproducer != NULL) {
      __xproducer->setUsingKeySymKeyEvents(ukke);
      __direct_keyboard = ukke;
    } else {
      NODE_ERROR_STREAM(getParent(), "No X producer exists!");
    }
  }

  bool XVirtualImg::isUsingKeySymKeyEvents(){
      return __direct_keyboard;
  }

  void XVirtualImg::setSendingMouseEventsDirect(const bool& smed) {
    if(__xproducer != NULL) {
      __xproducer->setSendingMouseEventsDirect(smed);
      __direct_mouse = smed;
    } else {
      NODE_ERROR_STREAM(getParent(), "No X producer exists!");
    }
  }
  
  bool XVirtualImg::isSendingMouseEventsDirect(){
    return __direct_mouse;
  }
  
  void XVirtualImg::setXAutoRepeat(const bool& on) {
    if(__xproducer != NULL) {
      __xproducer->setXAutoRepeat(on);
    } else {
      NODE_ERROR_STREAM(getParent(), "No X producer exists!");
    }    
  }

  bool XVirtualImg::isXAutoRepeat() {
    if(__xproducer != NULL) {
      return __xproducer->isXAutoRepeat();
    } else {
      NODE_ERROR_STREAM(getParent(), "No X producer exists!");
    }
    return false;
  }
  
  Result XVirtualImg::setSize(int width, int height) {
    /* check if size is changed */
    if((width == __width) && (height == __height)) {
      return SUCCESS;
    }
    
    NODE_DEBUG_STREAM(getParent(), "XVirtualImg> resize to " << width << "x" << height << " begin" << endl);
    __width  = width;
    __height = height;
    
    /* deinit X image */
    deinitImage();
    
    /* change window attributes */
    changeWindow();
    
    /* init X image */
    if (!initImage()) {
      NODE_ERROR_STREAM(getParent(), "XVirtualImg> Could not allocate a Image while resizing!" << endl);
      return FAILURE;
    }
    
    NODE_DEBUG_STREAM(getParent(), "XVirtualImg> resize end" << endl);
    return SUCCESS;
  };

  void XVirtualImg::init() {
    initWindow();
    
    if(!initImage()) {
      NODE_ERROR_STREAM(getParent(), "XVirtualImg> Could not allocate a Image while initializing!" << endl);
      return;
    }
  }

  void XVirtualImg::deinit() {
    deinitImage();
    deinitWindow();
  }

  void XVirtualImg::useDotCursor(bool value) {
    if(value) {
      XSetWindowAttributes attr;
      static char srcBits[] = { 0, 14,14,14, 0 };
      static char mskBits[] = { 31,31,31,31,31 };
      attr.cursor = createCursor(srcBits, mskBits, 5, 5);
      XChangeWindowAttributes(__display, __win, CWCursor, &attr);
      
    } else {
      XUndefineCursor(__display, __win);
    }
  }

  void XVirtualImg::useTriangleCursor(bool value) {
    if (value) {
      XSetWindowAttributes attr;
      static char srcBits[] = {0,0, 6,0, 30,0, 124,0, 252,1, 248,7, 248,31, 240,127, 240,63, 224,31, 224,15, 192,7, 192,3, 128,1, 128,0, 0,0};
      static char mskBits[] = {7,0, 31,0, 127,0, 254,1, 254,7, 252,31, 252,127, 248,255, 248,127, 240,63, 240,31, 224,15, 224,7, 192,3, 192,1, 128,0};
      attr.cursor = createCursor(srcBits, mskBits, 16, 16);
      XChangeWindowAttributes(__display, __win, CWCursor, &attr);
    } else {
      XUndefineCursor(__display, __win);
    }
    
  }

  void XVirtualImg::useStandardCursor() {
    XUndefineCursor(__display, __win);
  }
  
  Cursor XVirtualImg::createCursor(char* src, char* msk, int width, int height) {
    Cursor cursor;
    Pixmap pixsrc, pixmsk;
    XColor fg, bg;
    
    pixsrc = XCreateBitmapFromData(__display, DefaultRootWindow(__display), src, width, height);
    pixmsk = XCreateBitmapFromData(__display, DefaultRootWindow(__display), msk, width, height);
    XAllocNamedColor(__display, DefaultColormap(__display,DefaultScreen(__display)), "black",
		     &fg, &fg);
    XAllocNamedColor(__display, DefaultColormap(__display,DefaultScreen(__display)), "white",
		     &bg, &bg);
    cursor = XCreatePixmapCursor(__display, pixsrc, pixmsk, &fg, &bg, 2, 2);
    XFreePixmap(__display, pixsrc);
    XFreePixmap(__display, pixmsk);
    return cursor;
  }
  
  void XVirtualImg::setVisualInfo(XVisualInfo* visual_info) throw() {
    __visual_info = visual_info;
  }

  void XVirtualImg::setContext(GC gc) throw() {
    __gc = gc;
  }
  
  void XVirtualImg::updateMemoryManager(ShmBufferManager*, unsigned int) {}

  void XVirtualImg::setAlignment(int i,int XPOS = 0, int YPOS = 0) {
    __align = i;
    if (__align == 3)
    {
	__xpos = XPOS;
	__ypos = YPOS;
    }
    setPosition();
  }
}

#include "nmm/nmm_suffix.hpp"
