/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "XvImg.hpp"

#include "nmm/comm/messaging/Buffer.hpp"
#include "nmm/multimedia/Node.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  int XvImg::XV_FORMAT_UYVY = 0x59565955;
  int XvImg::XV_FORMAT_YV12 = 0x32315659;
  int XvImg::XV_FORMAT_YUV2 = 0x32595559;
  int XvImg::XV_FORMAT_I420 = 0x30323449; 

  XvImg::XvImg(Node* parent,
               int iwidth, int iheight,
               string title,
               unsigned int format, bool fullscreen,
               int ratio_x, int ratio_y)
    : XVirtualImg(parent, iwidth, iheight, ratio_x, ratio_y, fullscreen, title),
      xvport(0),
      xvimage(NULL),
      xvformat(format)
  {}
  
  XvImg::~XvImg() {
    // this is not very clean: as freeImage() is virtual and might be
    // overloaded in derived classes (see XvShmImg), we test here for a
    // NULL-Ptr and only call freeImage() if we still have a xvimage
    if (xvimage) {
      deinitImage();
    }
  }

  bool XvImg::initImage() {
    // query if XVideo extension is available
    if (!queryAdaptors(getDisplay(), getWindow())) {
      NODE_ERROR_STREAM(getParent(), "XvImg> Could not find a Xv port!" << endl);
      return false;
    }
    
    // allocate xvimage using previously (in queryAdaptor()) computed
    // XVideo port, via C'tor specified XVideo format
    // and NULL-Ptr for the image data (will be replaced on present
    // with the char-Ptr of the incoming Buffer).
    xvimage = XvCreateImage(getDisplay(), //Display
                            xvport, //XvPortID
                            xvformat, //format
                            NULL, //data
                            getWidth(), //width
                            getHeight()); //height

    if (!xvimage) {
      NODE_ERROR_STREAM(getParent(), "XvImg> Could not allocate a XImage!" << endl);
      return false;
    }

    // sync the display, to be sure image creation is forwarded to the
    // XServer, not discarding incoming events
    XSync(getDisplay(), False);

    return true;
  }

  void XvImg::deinitImage() {
    if (xvport) {
      // todo: is ths call really needed? xvinfo says: "operations
      // supported: PutImage", nothing about "PutVideo", ...
      XvStopVideo(getDisplay(), xvport, getWindow());
      if (xvimage) {
        XFree(xvimage);
        xvimage = NULL;
      }
      // ungrab the previously grabed XVideo port
      XvUngrabPort(getDisplay(), xvport, CurrentTime);
      xvport = 0;
    }
  }

  void XvImg::prepareOutput(Buffer* in_buffer) {
    // simple forward pointer
    xvimage->data = in_buffer->getData();
  }

  void XvImg::update(Buffer*) {
    NODE_DEBUG_STREAM(getParent(), "XvImg> update begin" << endl);
    // put the image xvimage to the Window getWindow() on Display getDisplay(),
    // using XVideo port xvport, context _gc, width getWidth() and height
    // getHeight(). Variables starting with '_' are provided by
    // XVirtualImg. Presentation will have zero offset in x direction,
    // y_blit_offset (from XVirtualImg) in y direction to put 16/9
    // video frame in the middle of a 4/3 display.
    XvPutImage(getDisplay(), xvport, getWindow(), getContext(), xvimage,
               0, 0, getWidth(), getHeight(),
               getXBlitOffset(), getYBlitOffset(),
               getScreenWidth(), getScreenHeight());
    // sync the display, to be sure image is forwarded to the
    // XServer, not discarding incoming events
    XSync(getDisplay(), False);
    // now, we can modify the data of the ximage again
    NODE_DEBUG_STREAM(getParent(), "XvImg> update end" << endl);
  }

  char* XvImg::getImage() {
    if (!xvimage) {
      return NULL;
    }

    return xvimage->data;
  }


  bool XvImg::queryAdaptors(Display* display, const Window win) {
    unsigned int count;

    // query for XVideo adaptors
    XvAdaptorInfo* ai;
    XvQueryAdaptors(display, win, &count, &ai);

    NODE_MESSAGE_STREAM(getParent(), "XvImg> Adaptors: " << count << endl);
    // iterate over all found adaptors
    for(unsigned int i=0;
        i < count;
        i++) {
      NODE_MESSAGE_STREAM(getParent(), "XvImg>   --------------------------"      << endl);
      NODE_MESSAGE_STREAM(getParent(), "XvImg>   name: "     << ai[i].name        << endl);
      NODE_MESSAGE_STREAM(getParent(), "XvImg>   base: "     << ai[i].base_id     << endl);
      NODE_MESSAGE_STREAM(getParent(), "XvImg>   nports: "   << ai[i].num_ports   << endl);
      NODE_MESSAGE_STREAM(getParent(), "XvImg>   nformats: " << ai[i].num_formats << endl);

      // iterate over all formats of the current adaptor
      for(unsigned int j=0; j < ai[i].num_formats; j++) {
        NODE_MESSAGE_STREAM(getParent(), "XvImg>     format: " << j << 
                              ", depth: " << static_cast<short>(ai[i].formats[j].depth) << 
                              ", id: " << ai[i].formats[j].visual_id << endl);
      }

      // iterate over all available ports of the current adaptor
      for(XvPortID port = ai[i].base_id;
          port < ai[i].base_id + ai[i].num_ports;
          port++) {
        if(XvGrabPort(getDisplay(), port, CurrentTime) == 0) {
          // if a port could be grabed
          string name(ai[i].name);
          if (!xvport && name != "video4linux") {
            // use if as we don't havn't found one yet
            NODE_MESSAGE_STREAM(getParent(), "XvImg>   Port " << port << " is suitable, using it!" << endl);
            xvport = port;
          } else {
            // else ungrab it again
            NODE_MESSAGE_STREAM(getParent(), "XvImg>   Port " << port << " is suitable, but we have one already!" << endl);
            XvUngrabPort(getDisplay(), port, CurrentTime);
          }
        } else {
          NODE_MESSAGE_STREAM(getParent(), "XvImg>   Port " << port << " is NOT suitable!");
        }
      }
    }

    if (!xvport) {
      NODE_ERROR_STREAM(getParent(), "XvImg> Could not find a suitable Xv port!" << endl);
      return false;
    }

    // query for encodings of the found XVideo port
    XvEncodingInfo* ei;
    XvQueryEncodings(display, xvport, &count, &ei);

    NODE_MESSAGE_STREAM(getParent(), "XvImg> Encodings: " << count << endl);
    // iterate over all found encodings
    for(unsigned int i = 0;
        i < count;
        i++) {
      NODE_MESSAGE_STREAM(getParent(), "XvImg>   --------------------------"                                             << endl);
      NODE_MESSAGE_STREAM(getParent(), "XvImg>  id: "           << ei[i].encoding_id                                     << endl);
      NODE_MESSAGE_STREAM(getParent(), "XvImg>  name: "         << ei[i].name                                            << endl);
      NODE_MESSAGE_STREAM(getParent(), "XvImg>  resolution: "   << ei[i].width << "x" << ei[i].height                    << endl);
      NODE_MESSAGE_STREAM(getParent(), "XvImg>  rate: "         << ei[i].rate.numerator << "/" << ei[i].rate.denominator <<endl);
      NODE_MESSAGE_STREAM(getParent(), "XvImg>  num_encodings: "<< ei[i].num_encodings                                   << endl);
    }

    // query for formats of the found XVideo port
    XvImageFormatValues* fo;
    fo = XvListImageFormats(display, xvport, reinterpret_cast<int*>(&count));
    NODE_MESSAGE_STREAM(getParent(), "XvImg> ImageFormats: " << count << endl);
    // iterate over all found formats
    for(unsigned int i = 0;
        i < count;
        i++) {
	NODE_MESSAGE_STREAM(getParent(), "XvImg>   --------------------------"                        << endl);
	NODE_MESSAGE_STREAM(getParent(), "XvImg>   id: 0x"      << fo[i].id                           << endl);
	NODE_MESSAGE_STREAM(getParent(), "XvImg>   idname: "    << reinterpret_cast<char*>(&fo[i].id) << endl);
	NODE_MESSAGE_STREAM(getParent(), "XvImg>   byteorder: " << fo[i].byte_order                   << endl);
	NODE_MESSAGE_STREAM(getParent(), "XvImg>   guid: "      << fo[i].guid                         << endl);
	NODE_MESSAGE_STREAM(getParent(), "XvImg>   bpp: "       << fo[i].bits_per_pixel               << endl);
	NODE_MESSAGE_STREAM(getParent(), "XvImg>   pack: "      << fo[i].format                       << endl);
	NODE_MESSAGE_STREAM(getParent(), "XvImg>   nplanes: "   << fo[i].num_planes                   << endl);
	NODE_MESSAGE_STREAM(getParent(), "XvImg>   Y_bits: "    << fo[i].y_sample_bits                << endl);
	NODE_MESSAGE_STREAM(getParent(), "XvImg>   U_bits: "    << fo[i].u_sample_bits                << endl);
	NODE_MESSAGE_STREAM(getParent(), "XvImg>   V_bits: "    << fo[i].v_sample_bits                << endl);
	NODE_MESSAGE_STREAM(getParent(), "XvImg>   order: "     << fo[i].component_order              << endl);
      }

    // query for attributes of the found XVideo port
    XvAttribute* pa;
    pa = XvQueryPortAttributes(display, xvport, reinterpret_cast<int*>(&count));
    NODE_MESSAGE_STREAM(getParent(), "XvImg> Attributes: "<< count << endl);
    // iterate over all found attributes
    for(unsigned int i = 0;
        i < count;
        i++) {
      NODE_MESSAGE_STREAM(getParent(), "XvImg>   --------------------------" << endl);
      NODE_MESSAGE_STREAM(getParent(), "XvImg>   name: "  << pa[i].name      << endl);
      NODE_MESSAGE_STREAM(getParent(), "XvImg>   flags: " << pa[i].flags     << endl);
      NODE_MESSAGE_STREAM(getParent(), "XvImg>   min: "   << pa[i].min_value << endl);
      NODE_MESSAGE_STREAM(getParent(), "XvImg>   max: "   << pa[i].max_value << endl);

      if (pa[i].flags & XvGettable) {
        // query current attribute value only if the attribute is "getable"
        int current_value = 0;
        Atom atom = XInternAtom (display, pa[i].name, False);
        XvGetPortAttribute(display, xvport, atom, &current_value);
        NODE_MESSAGE_STREAM(getParent(), "XvImg>   current: " << current_value << endl);
      } else {
        NODE_MESSAGE_STREAM(getParent(), "XvImg>   UNGETABLE!!!" << endl);
      }
    }

    return true;
  }

}

#include "nmm/nmm_suffix.hpp"
