/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "XvShmImg.hpp"

#include "nmm/comm/messaging/Buffer.hpp"
#include "nmm/multimedia/Node.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  XvShmImg::XvShmImg(Node* parent,
                     int width, int height,
                     string title,
                     unsigned int format, bool fullscreen,
                     int ratio_x, int ratio_y,
                     ShmBufferManager* manager, unsigned int shared_pages)
    : XvImg(parent, width, height, title, format, fullscreen, ratio_x, ratio_y),
      shm(manager, shared_pages) {
  }

  XvShmImg::~XvShmImg() {
    // this is not very clean: as freeImage() is virtual and might be
    // overloaded in derived classes (see XvShmImg), we test here for a
    // NULL-Ptr and only call freeImage() if we still have a xvimage
    if (xvimage) {
      deinitImage();
    }
  }

  bool XvShmImg::initImage() {
    // query XShmManager if XShm extension is available
    if(!shm.queryDisplay(getDisplay())) {
      NODE_ERROR_STREAM(getParent(), "XvShmImg> no SHM extension available" << endl);
      return false;
    }

    // query if XVideo extension is available
    if (!queryAdaptors(getDisplay(), getWindow())) {
      NODE_ERROR_STREAM(getParent(), "XvImg> Could not find a Xv port!" << endl);
      return false;
    }
    
    // allocate xvimage using previously (in queryAdaptor()) computed
    // XVideo port, via C'tor specified XVideo format
    // and NULL-Ptr for the image data (will be replaced on present
    // with the char-Ptr of shared memory segment). XShmManager
    // provides the needed XShmSegmentInfo struct-Ptr.Variables
    // starting with '_' are provided by XVirtualImg.
    xvimage = XvShmCreateImage(getDisplay(), //Display
                               xvport, //XvPortID
                               xvformat, //format
                               NULL, //data
                               getWidth(), //width
                               getHeight(), //height
                               shm.getInfo()); //XShmSegmentInfo

    if(!xvimage) {
      NODE_ERROR_STREAM(getParent(), "XvShmImg> cannot create XvShmImage" << endl);
      return false;
    }

    // Let the XShmManager attach his shared memory segment to the
    // display buffer size for one shared memory page is set to
    // xvimage->data_size (specified by XVideo extension in
    // XvShmCreatImage above
    if (!shm.attachMemory(getDisplay(), xvimage->data_size)) {
      NODE_ERROR_STREAM(getParent(), "XvShmImg> cannot create SHM" << endl);
      XFree(xvimage);
      return false;
    }

    // sync the display, to be sure image creation is forwarded to the
    // XServer, not discarding incoming events
    XSync(getDisplay(), False);

    return true;
  }

  void XvShmImg::updateMemoryManager(ShmBufferManager* manager, unsigned int shared_pages) {
    // delete previous attached shared memory segment
    deinitImage();

    shm.updateMemoryManager(manager, shared_pages);

    // reattach new image & shared memory segment
    initImage();
  }

  void XvShmImg::deinitImage() {
    if (xvport) {
      // todo: is ths call really needed? xvinfo says: "operations
      // supported: PutImage", nothing about "PutVideo", ...
      XvStopVideo(getDisplay(), xvport, getWindow());
      if (xvimage) {
        shm.detachMemory();
        XFree(xvimage);
        xvimage = NULL;
      }
      // ungrab the previously grabed XVideo port
      XvUngrabPort(getDisplay(), xvport, CurrentTime);
      xvport = 0;
    }
  }

  void XvShmImg::prepareOutput(Buffer* in_buffer) {
    // not simple forward pointer, but ask ShmManager to ensure
    // data-Ptr is from the shared memory segment.
    xvimage->data = shm.getMem(in_buffer->getData());
  }

  void XvShmImg::update(Buffer*)
  {
    NODE_DEBUG_STREAM(getParent(), "XvShmImg> update begin" << endl);
    // put the image xvimage to the Window getWindow() on Display getDisplay(),
    // using XVideo port xvport, context getContext(), width getWidth() and height
    // getHeight(). Variables starting with '_' are provided by
    // XVirtualImg. Presentation will have zero offset in x direction,
    // getYBlitOffset() (from XVirtualImg) in y direction to put 16/9
    // video frame in the middle of a 4/3 display.
    XvShmPutImage(getDisplay(), xvport, getWindow(), getContext(), xvimage,
		  0, 0, getWidth(), getHeight(),
		  getXBlitOffset(), getYBlitOffset(),
                  getScreenWidth(), getScreenHeight(),
                  False); // do not sent completion event, instead:
    // sync the display, to be sure image is forwarded to the
    // XServer, not discarding incoming events
    XSync(getDisplay(), False);
    // now, we can modify the data of the ximage again
    NODE_DEBUG_STREAM(getParent(), "XvShmImg> update end" << endl);
  }

}

#include "nmm/nmm_suffix.hpp"
