/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "iPAQDisplayNode.hpp"
#include "nmm/comm/format/Property.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"
#include "nmm/utils/SecurityCheck.hpp"

#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/vt.h>
#include <linux/kd.h>
#include <termios.h>
#include <sys/mman.h>
#include <sys/time.h>

NMM_REGISTER_PLUGIN(iPAQDisplayNode, NMM, "iPAQDisplayNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  iPAQDisplayNode::iPAQDisplayNode(const char* name, StreamQueue::Mode queue_mode, int queue_size)
    : GenericUSyncSinkNode(name, queue_mode, queue_size) {
    DEBUG_STREAM( "Begin > Create iPAQDisplay Node" << endl);

    declareInputFormats();
    
    fb_name = "/dev/fb0";
    fb_handle = -2;
    
    fb_data   =  NULL; 
    pb_data   =  NULL;
    
    image_size   = 0;
    image_size_quarter = 0;
    image_width  = 0;
    image_height = 0;
    
    screen_width  = 0;
    screen_height = 0;
    
    offset = 0;
    pad    = 0;
    
    xpos=0;
    ypos=0;

    colorspace = "";
    channelorder = "";
    bitperpixel = 0;
    format = "";

    convmode = NONE;

    src_uint = NULL;
    pixel=0;
    src0_uchar = src1_uchar = NULL;
    src2_uchar = src3_uchar = NULL;
    y1 = y2 = u = v = c1 = c2 = c3 = c4 = r = g = b =0;

    dest1 = NULL;
    dest2 = NULL;
    asm_buffer = NULL;
    

    DEBUG_STREAM( "End > Create iPAQDisplay Node" << endl);
  } /* Constructor */





  iPAQDisplayNode::~iPAQDisplayNode() {

    /* disallocate mmemory */
    if (NULL != pb_data) {
      delete[] pb_data; 
    }
    if ((NULL != fb_data) && (0 != fb_screensize)) {
      munmap(fb_data, fb_screensize);
    }

  } /* Destructor */





  void iPAQDisplayNode::declareInputFormats() {
    Format* i_format;
    
    /* RGB32 - * 8R + 8G + 8B + 8fill bit e  */
    i_format = getOwnInputProperty()->addNewFormat(Format::video_raw_type);
    i_format->addWildcard(Format::x_resolution_param);
    i_format->addWildcard(Format::y_resolution_param);
    i_format->addWildcard(Format::ratio_x_param);
    i_format->addWildcard(Format::ratio_y_param);
    i_format->addWildcard(Format::framerate_param);
    i_format->addStringParamValue(Format::channelorder_param, "rgb");
    i_format->addStringParamValue(Format::endian_param, "little");
    i_format->addStringParamValue(Format::colorspace_param, "rgb24");
    i_format->addIntParamValue(Format::bitperpixel_param, 32);
    i_format->addStringParamValue(Format::format_param, "packet");
    i_format->setIOPartner(0);
    getOwnInputProperty()->setDefaultFormat(i_format);

    /* YV 12 */
    /* NxN Y plane followed by (N/4)x(N/4) U and V planes */
    i_format = getOwnInputProperty()->addNewFormat(Format::video_raw_type);
    i_format->addWildcard(Format::x_resolution_param);
    i_format->addWildcard(Format::y_resolution_param);
    i_format->addWildcard(Format::ratio_x_param);
    i_format->addWildcard(Format::ratio_y_param);
    i_format->addWildcard(Format::framerate_param);
    i_format->addStringParamValue(Format::channelorder_param, "yvu");
    i_format->addStringParamValue(Format::endian_param, "little");
    i_format->addStringParamValue(Format::colorspace_param, "yv12");
    i_format->addIntParamValue(Format::bitperpixel_param, 12);
    i_format->addStringParamValue(Format::format_param, "planar");
    i_format->setIOPartner(0);

    /* RGB565 - 5R + 6G + 5B */
    i_format = getOwnInputProperty()->addNewFormat(Format::video_raw_type);
    i_format->addWildcard(Format::x_resolution_param);
    i_format->addWildcard(Format::y_resolution_param);
    i_format->addWildcard(Format::ratio_x_param);
    i_format->addWildcard(Format::ratio_y_param);
    i_format->addWildcard(Format::framerate_param);
    i_format->addStringParamValue(Format::channelorder_param, "rgb");
    i_format->addStringParamValue(Format::endian_param, "little");
    i_format->addStringParamValue(Format::colorspace_param, "rgb565");
    i_format->addIntParamValue(Format::bitperpixel_param, 16);
    i_format->addStringParamValue(Format::format_param, "packet");
    i_format->setIOPartner(0);
  
  } /* declareInputFormats */





  Result iPAQDisplayNode::doInit() {
    /* open framebuffer */
    if(!SSecurityCheck::getInstance().checkFilenameForReading(fb_name)) {
      ERROR_STREAM("Security check not passed for file " << fb_name << ", check .nmmrc for options" << endl);
      return FAILURE;
    }
    fb_handle = open(fb_name.c_str(), O_RDWR);
    if (fb_handle < 0) {
      ERROR_STREAM("Cannot open framebuffer device " << fb_name.c_str() << "." << endl);
      return FAILURE;
    }
    if (ioctl(fb_handle, FBIOGET_FSCREENINFO, &fb_finfo)) {
      ERROR_STREAM("Cannot read framebuffer fixed information." << endl);
      return FAILURE;
    }
    if (ioctl(fb_handle, FBIOGET_VSCREENINFO, &fb_vinfo)) {
      ERROR_STREAM("Cannot read framebuffer variable information." << endl);
      return FAILURE;
    }
    if (strcmp(fb_finfo.id,"SA1100")) {
      ERROR_STREAM("Framebuffer id \'" << fb_finfo.id << "\' is not iPAQ framebuffer id." << endl);
      return FAILURE;
    }
    if ((fb_vinfo.xres != 320) || (fb_vinfo.yres != 240) ||
	(fb_vinfo.xres_virtual != 320) || (fb_vinfo.yres_virtual != 240)) {
      ERROR_STREAM("Framebuffer resolution is not 320x240." << endl);
      return FAILURE;
    } else {
      screen_width  = fb_vinfo.xres; /* 320 */
      screen_height = fb_vinfo.yres; /* 240 */
    }
    if (fb_vinfo.bits_per_pixel != 16) {
      ERROR_STREAM("Framebuffer depth is not 16 bits." << endl);
      return FAILURE;
    }

    if (ioctl(fb_handle,FBIOBLANK,VESA_NO_BLANKING)) {
      WARNING_STREAM("Turn LCD on failed." << endl);
    }

    if (!setBrightness(60)) {
      WARNING_STREAM("Could not set brightness" << endl);
    }

    /* close framebuffer*/
    close(fb_handle);
    fb_handle = -2;

    return SUCCESS;
  } /* doInit */





  Result iPAQDisplayNode::doDeinit() {
    if (fb_handle > 0) {
      close(fb_handle);
      fb_handle = -2;
    }
    return SUCCESS;
  } /* doDeinit */





  Result iPAQDisplayNode::doInitOutput() {
    const Format* in_format = getInputFormat();
    if (in_format == NULL) { return FAILURE; }
    Format* format;
    
    format = new Format( *in_format );
    getOwnWorkingInputProperty()->addNewFormat(format);
    getOwnWorkingInputProperty()->setDefaultFormat(format);
    
    return SUCCESS;
  } /* doInitOutput() */
  




  Result iPAQDisplayNode::doDeinitOutput() {
    getOwnWorkingInputProperty()->erase();
    getOwnWorkingOutputProperty()->erase();
    return SUCCESS;
  } /* doDeinitOutput() */
  




  Result iPAQDisplayNode::doActivate() {
    const Format* i_format = getInputFormat();

    /* analyse input format */
    if (i_format == 0) {
      ERROR_STREAM( "Invalid input format (NULL-ptr)" << endl);
      return FAILURE;
    }

    /* set all parameter depending on the input format */
    FormatChange();

    /* open framebuffer device */
    if (fb_handle < 0 ) {
      if(!SSecurityCheck::getInstance().checkFilenameForReading(fb_name)) {
        ERROR_STREAM("Security check not passed for file " << fb_name << ", check .nmmrc for options" << endl);
        return FAILURE;
      }
      fb_handle = open(fb_name.c_str(), O_RDWR);
    }
    if (fb_handle < 0) {
      ERROR_STREAM("Cannot open framebuffer device " << fb_name << ".");
      return FAILURE;
    }

    /* map framebuffer memory */
    fb_screensize = (fb_vinfo.xres * fb_vinfo.yres * fb_vinfo.bits_per_pixel) >> 3; /* divide 8 */
    fb_data = (UWORD *)mmap(0, fb_screensize, PROT_READ | PROT_WRITE, MAP_SHARED, fb_handle, 0);
    if (fb_data < 0) {
      ERROR_STREAM("Failed to map framebuffer device to memory.");
      fb_data = NULL;
      return FAILURE;
    }
    fb_data_offset = fb_data + offset;

    /* allocate memory for prepare buffer 320*240*2 Bytes */ 
    pb_data = new UWORD[76800]; /* UWORD is 2 Bytes long */
    if (NULL == pb_data) {
      ERROR_STREAM("Failed to allocate memory for prepare buffer.");
      return FAILURE;
    }

    if (fb_screensize > 0) {
      memset (fb_data, 0x00, fb_screensize);
      memset (pb_data, 0xFF, fb_screensize);
    }

    return SUCCESS;
  } /* doActivate() */





  Result iPAQDisplayNode::doDeactivate() {
    /* clear screen */
    if ((NULL != fb_data) && (fb_screensize >0)) {
      memset(fb_data, 0, fb_screensize);
    }
    if (pb_data != NULL) {
      delete[] pb_data; 
      pb_data = NULL;
    }
    if (fb_data != NULL) {
      munmap(fb_data, fb_screensize);
      fb_data = NULL;
    }

    return SUCCESS;
  }





  void iPAQDisplayNode::prepareBuffer(Buffer *in_buffer) {
    //    DEBUG_STREAM( " Enter prepareBuffer " << endl );
    if (in_buffer != NULL) {
      switch (convmode) {
      case NONE:
        /* nothing to do, data is already rgb565/rgb/16 */
        break;

      case YV12_to_RGB565c:
        /* read pointers */
        src0_uchar=reinterpret_cast<unsigned char*>(in_buffer->getData());
        src1_uchar=src0_uchar + image_width;
        src2_uchar=src0_uchar + image_size;
        src3_uchar=src2_uchar + image_size_quarter;

        /* write pointers */
        dest1=pb_data + offset;
        dest2=dest1 + screen_width;

        for (ypos=0; ypos<image_height; ypos += 2) {
          for (xpos=0; xpos<image_width; xpos += 2) {

            u = *src2_uchar++;
            v = *src3_uchar++;

            c1 = crv_tab[v];
            c2 = cgu_tab[u];
            c3 = cgv_tab[v];
            c4 = cbu_tab[u];
            
            /* up-left */
            y1 = tab_76309[*src0_uchar++];
            b = clpb[384+((y1 + c1)>>16)];
            g = clpg[384+((y1 - c2 - c3)>>16)];
            r = clpr[384+((y1 + c4)>>16)];
            *dest1++ = (UWORD) (r | g | b);

            /* down-left */
            y2 = tab_76309[*src1_uchar++];
            b = clpb[384+((y2 + c1)>>16)];
            g = clpg[384+((y2 - c2 - c3)>>16)];
            r = clpr[384+((y2 + c4)>>16)];
            *dest2++ = (UWORD) (r | g | b);

            /* up-right */
            y1 = tab_76309[*src0_uchar++];
            b = clpb[384+((y1 + c1)>>16)];
            g = clpg[384+((y1 - c2 - c3)>>16)];
            r = clpr[384+((y1 + c4)>>16)];
            *dest1++ = (UWORD) (r | g | b);

            /* down-right */
            y2 = tab_76309[*src1_uchar++];
            b = clpb[384+((y2 + c1)>>16)];
            g = clpg[384+((y2 - c2 - c3)>>16)];
            r = clpr[384+((y2 + c4)>>16)];
            *dest2++ = (UWORD) (r | g | b);
          }
         
          dest1 +=  screen_width + pad;
          dest2 +=  screen_width + pad;
          src0_uchar += image_width;
          src1_uchar += image_width;
        }
        break;
      
      case YV12_to_RGB565asm:
        pic.data[0] = reinterpret_cast<unsigned char*>(in_buffer->getData());
        pic.data[2] = pic.data[0] + image_size;
        pic.data[1] = pic.data[2] + image_size_quarter;
        /* !!! height must be a multiplier of 2 !!! */
        convert_yuv420_rgb565(&pic, reinterpret_cast<unsigned char*>(pb_data), image_width, image_height);
        break;

      case RGBX32_to_RGB565:
        dest1    = pb_data;
        src_uint = reinterpret_cast<unsigned int*>(in_buffer->getData());
        pixel=0;

        for(ypos=0; ypos<image_height; ypos++) {        
          for(xpos=0; xpos<image_width; xpos++) {
            
            /* Read PRGB32 */
            pixel = (*src_uint);
            
            /* Write RGB565 */
            *dest1 = (UWORD)
              ((pixel & 0x00F80000) >> 8) | /* Red */
              ((pixel & 0x0000FC00) >> 5) | /* Green */
              ((pixel & 0x000000F8) >> 3);  /* Blue */
            
            src_uint++;  /* + 32 Bit */
            dest1++;     /* + 16 Bit */                
          }
          dest1 += pad;
        }

        break;

      default:
        ERROR_STREAM("Only rgba32/rgb565/yv12");
        break;
      }
    }
    //    DEBUG_STREAM( " Leave prepareBuffer " << endl );
  } /* prepareBuffer */






  Message* iPAQDisplayNode::presentBuffer(Buffer *in_buffer) {
    //   DEBUG_STREAM( " Enter presentBuffer " << endl );
    if(NULL != in_buffer) {

      switch (convmode) {
      case NONE:
        /* take frame direct from the in_buffer */
        dest1 = reinterpret_cast<short unsigned int*>(in_buffer->getData());
        dest2 = fb_data_offset;
        break;

      case YV12_to_RGB565c:
      case YV12_to_RGB565asm:
      case RGBX32_to_RGB565:
        /* take frame from prepare buffer */
        dest1 = pb_data;
        dest2 = fb_data_offset;
        break;

      default:
        break;
      }

      /* copy picture to framebuffer */
      for(ypos=image_height; ypos!=0; ypos--) {
        memcpy(dest2, dest1, image_width<<1);
        dest1 += 320;
        dest2 += 320;
      }

      /* relase incoming buffer */
      in_buffer->release();
    }
    return 0;
    //    DEBUG_STREAM( " Leave presentBuffer " << endl );
  } /* presentBuffer */





  Result iPAQDisplayNode::FormatChange() {
    const Format* i_format = getInputFormat();

    /* analyse input format */
    if (0 == i_format) {
      ERROR_STREAM( "Invalid input format (NULL-ptr)" << endl);
      return FAILURE;
    }

    /* colorspacem, colororder and bitperpixel */
    colorspace   = i_format->getStringValue(Format::colorspace_param);
    channelorder = i_format->getStringValue(Format::channelorder_param);
    bitperpixel  = i_format->getIntValue   (Format::bitperpixel_param);
    format       = i_format->getStringValue(Format::format_param);

    /* pad  */
    image_width  = i_format->getIntValue(Format::x_resolution_param);
    image_height = i_format->getIntValue(Format::y_resolution_param);
    if (image_width  > 320) { image_width  = 320; }
    if (image_height > 240) { image_height = 240; }
    
    image_size = image_width * image_height;
    image_size_quarter = (image_size >> 2); /* shift 2 -> DIV 4 */

    offset = int( (((screen_height - image_height) >> 1) * screen_width) + 
                  ((screen_width - image_width) >> 1) ); /* shift 1 -> DIV 2 */
    pad = screen_width - image_width;

    DEBUG_STREAM("Offset is " << offset << "and pad is  " << pad << endl);

    pic.linesize[0] = image_width;
    pic.linesize[1] = pic.linesize[2] = (image_width >> 1);

    /* determinate converter mode*/
    if (colorspace == "rgb24") {
      if (channelorder == "rgb") {
        if (bitperpixel == 32) {
          if (format == "packet") {
            convmode = RGBX32_to_RGB565;
            DEBUG_STREAM("Converter is PRGB32 to RGB565" << endl);
          } 
        }
      }
    } 

    else if (colorspace == "yv12") {
      if (channelorder == "yvu") {
        if (bitperpixel == 12) {
          if (format == "planar") {
            convmode = YV12_to_RGB565asm;
            DEBUG_STREAM("Converter is YV12 to RGB565 asm-version" << endl);
          }
        }      
      }
    }

    else if (colorspace == "rgb565") {
      if (channelorder == "rgb") {
        if (bitperpixel == 16) {
          convmode = NONE; /* for rgb 565 nothing to do */
        }
      }
    }


    /* fill converter tables, if c version of YV12->RGR565 converter */
    if (convmode == YV12_to_RGB565c) {
      initTables();
    }

    return SUCCESS;
  } /* Format Change */





  Result iPAQDisplayNode::setBrightness(int level) {
    Result res=FAILURE;
    
    if (fb_handle > 0) {
      typedef struct {
        unsigned char mode;
        unsigned char pwr;
        unsigned char brightness;
      } brightness_struct;

      if (level > 255) { level = 255; }
      if (level <   0) { level =   0; }

      brightness_struct bl;
        
      bl. mode = 1;
      bl. pwr = 1;
      bl. brightness = (level *255) / 64;
      res = ((ioctl ( fb_handle, (( 1<<30 ) | ('f'<<8) | 7 | ( sizeof(brightness_struct) << 16 )), &bl ) == 0 ) ? SUCCESS : FAILURE);
    }

    return res;
  } /* setBrightness */





  void iPAQDisplayNode::initTables() {
    long int crv,cbu,cgu,cgv;
    int i,ind;

    crv = 104597; cbu = 132201;
    cgu = 25675;  cgv = 53279;

    for (i = 0; i < 256; i++) {
      crv_tab[i] = (i-128) * crv;
      cbu_tab[i] = (i-128) * cbu;
      cgu_tab[i] = (i-128) * cgu;
      cgv_tab[i] = (i-128) * cgv;
      tab_76309[i] = 76309*(i-16);
    }

    for (i=0;i<384;i++) { 
      clpr[i]=clpg[i]=clpb[i]=0;  
    }
    ind=384;
    for (i=0;i<256;i++) { 
      clpr[ind]=(i >> 3) << 11;  
      clpg[ind]=(i >> 2) <<  5;  
      clpb[ind++]=(i >> 3);  
    }
    ind=640;
    for (i=0;i<384;i++) { 
      clpr[ind]=0xF800;  
      clpg[ind]=0x07E0; 
      clpb[ind++]=0x001F;  
    }
  } /* initTables */ 

#include "nmm/nmm_prefix.hpp"

} /* namespace NMM */

#include "nmm/nmm_suffix.hpp"
