/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "FFMpegDecodeNode.hpp"


#include <sys/time.h>
#include <math.h>
#include <assert.h>
#include <ffmpeg/avcodec.h>

#include "nmm/comm/format/Property.hpp"
#include "nmm/multimedia/Plugin.hpp"
#include "nmm/multimedia/TPlugin.hpp"
#include "nmm/comm/messaging/CEvent.hpp"
#include "nmm/utils/time/Rational.hpp"
#include "nmm/utils/time/Interval.hpp"


#ifndef FRAME_RATE_BASE
#define FRAME_RATE_BASE 10000
#endif

#ifndef UINT8
#define UINT8 uint8_t
#endif

NMM_REGISTER_PLUGIN(FFMpegDecodeNode, NMM, "FFMpegDecodeNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  FFMpegDecodeNode::FFMpegHandle::FFMpegHandle() : ffAVCodec(NULL), ffavcc(NULL), avframe(NULL) {}

  FFMpegDecodeNode::FFMpegDecodeNode(const char *name, StreamQueue::Mode qmode, int qsize)
    : GenericProcessorNode(name, qmode, qsize),
      ffmpeg(new FFMpegHandle()),
      height(0),
      height_half(0),
      width(0),
      width_half(0),
      framerate(0),
      framerate_base(1000),
      int_framerate(0),
      bitrate(0),
      gop_size(0),
      avframe_size(0),
      out_size(0),
      frame_count(0),
      timestamp(),
      in_buf_ts(),
      startsec(0),
      neededsecs(0),
      encfps(0.0f),
      got_picture(0),
      saved_buffer(0),
      out_buffer_data1(0),
      out_buffer_data2(0),
      src1(0),
      src2(0),
      out_buffer_size(0),
      line(0),
      video_extradata_size(0),
      video_extradata(),
      vid_delay(0),
      skip_until_keyframe(false),
      codecs(),
      default_codec()
  {
    /* get and remember current time */
    gettimeofday(&tv, &tz);
    startsec = tv.tv_sec;
  }
  
  FFMpegDecodeNode::~FFMpegDecodeNode() {
    delete ffmpeg;
  }

  Result FFMpegDecodeNode::startTrack(const string&) {
    DEBUG_STREAM("Received Event: startTrack");
    return SUCCESS;
  }

  Result FFMpegDecodeNode::endTrack() {
    DEBUG_STREAM("Received Event: endTrack");
    return SUCCESS;
  }

  Result FFMpegDecodeNode::started(const Interval&) {
    return SUCCESS;
  }
  
  Result FFMpegDecodeNode::finished() {
    DEBUG_STREAM("Received Event: finished");
    return SUCCESS;
  }

  Result FFMpegDecodeNode::doInit() {
    DEBUG_STREAM("Got avcodec version: " << avcodec_version() << endl);
    DEBUG_STREAM("Got avcodec build: " << avcodec_build() << endl);

    /* initialize library */
    avcodec_init();

    /* register all the codecs */
    avcodec_register_all();

    /* detect all codecs */
    codecs.clear();

    default_codec = "";
    detectAllCodecs();

    /* check if any codec was found */
    if(default_codec.empty()) {
      ERROR_STREAM("No codecs with known FOURCC code available" << std::endl);
      return FAILURE;
    }

    /* general input format with bitrate known */
    Format* in_format = getOwnInputProperty()->addNewFormat(Format::video_mpeg4_type);
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addWildcard(Format::framerate_param);
    in_format->addWildcard(Format::bitrate_param);
    for(CodecConstIterator i = codecs.begin(); i != codecs.end(); ++i) {
      if(i->first != "WMV1" && i->first != "WMV2")
        in_format->addStringParamValue(Format::compressor_param, i->first);
    }

    in_format = getOwnInputProperty()->addNewFormat(Format::video_mpeg4_type);
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addWildcard(Format::framerate_param);
    in_format->addWildcard(Format::extradata_param);
    in_format->addWildcard(Format::extradata_size_param);
    in_format->addWildcard(Format::bitrate_param);
    for(CodecConstIterator i = codecs.begin(); i != codecs.end(); ++i) {
      if(i->first == "WMV1" || i->first == "WMV2")
	in_format->addStringParamValue(Format::compressor_param, i->first);
    }

    Format* in_format2 = getOwnInputProperty()->addNewFormat(Format::video_mpeg_type);
    in_format2->addIntParamValue(Format::version_param, 1);
    in_format2->addIntParamValue(Format::version_param, 2);
    in_format2->addWildcard(Format::x_resolution_param);
    in_format2->addWildcard(Format::y_resolution_param);
    in_format2->addWildcard(Format::ratio_x_param);
    in_format2->addWildcard(Format::ratio_y_param);
    in_format2->addWildcard(Format::bitrate_param);
    in_format2->addWildcard(Format::framerate_param);
    in_format2->addWildcard(Format::bitrate_param);
    
    Format* in_format3 = getOwnInputProperty()->addNewFormat(Format::video_jpeg_type);
    in_format3->addWildcard(Format::x_resolution_param);
    in_format3->addWildcard(Format::y_resolution_param);
    in_format3->addWildcard(Format::ratio_x_param);
    in_format3->addWildcard(Format::ratio_y_param);
    in_format3->addWildcard(Format::framerate_param);
    in_format3->addWildcard(Format::bitrate_param);

    Format* in_format4 = getOwnInputProperty()->addNewFormat(Format::video_h263_type);
    in_format4->addWildcard(Format::x_resolution_param);
    in_format4->addWildcard(Format::y_resolution_param);
    in_format4->addWildcard(Format::ratio_x_param);
    in_format4->addWildcard(Format::ratio_y_param);
    in_format4->addWildcard(Format::framerate_param);
    in_format4->addWildcard(Format::bitrate_param);

    Format* in_format44 = getOwnInputProperty()->addNewFormat(Format::video_h264_type);
    in_format44->addWildcard(Format::x_resolution_param);
    in_format44->addWildcard(Format::y_resolution_param);
    in_format44->addWildcard(Format::ratio_x_param);
    in_format44->addWildcard(Format::ratio_y_param);
    in_format44->addWildcard(Format::framerate_param);
    in_format44->addWildcard(Format::bitrate_param);

    Format* in_format5 = getOwnInputProperty()->addNewFormat(Format::video_h263plus_type);
    in_format5->addWildcard(Format::x_resolution_param);
    in_format5->addWildcard(Format::y_resolution_param);
    in_format5->addWildcard(Format::ratio_x_param);
    in_format5->addWildcard(Format::ratio_y_param);
    in_format5->addWildcard(Format::framerate_param);
    in_format5->addWildcard(Format::bitrate_param);

    /* general input format with bitrate unknown */
    in_format = getOwnInputProperty()->addNewFormat(Format::video_mpeg4_type);
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param);
    in_format->addWildcard(Format::ratio_y_param);
    in_format->addWildcard(Format::framerate_param);
    for(CodecConstIterator i = codecs.begin(); i != codecs.end(); ++i) {
      in_format->addStringParamValue(Format::compressor_param, i->first);
    }
    getOwnInputProperty()->setDefaultFormat(in_format);

    in_format2 = getOwnInputProperty()->addNewFormat(Format::video_mpeg_type);
    in_format2->addIntParamValue(Format::version_param, 1);
    in_format2->addIntParamValue(Format::version_param, 2);
    in_format2->addWildcard(Format::x_resolution_param);
    in_format2->addWildcard(Format::y_resolution_param);
    in_format2->addWildcard(Format::ratio_x_param);
    in_format2->addWildcard(Format::ratio_y_param);
    in_format2->addWildcard(Format::framerate_param);

    in_format3 = getOwnInputProperty()->addNewFormat(Format::video_jpeg_type);
    in_format3->addWildcard(Format::x_resolution_param);
    in_format3->addWildcard(Format::y_resolution_param);
    in_format3->addWildcard(Format::ratio_x_param);
    in_format3->addWildcard(Format::ratio_y_param);
    in_format3->addWildcard(Format::framerate_param);
    
    in_format4 = getOwnInputProperty()->addNewFormat(Format::video_h263_type);
    in_format4->addWildcard(Format::x_resolution_param);
    in_format4->addWildcard(Format::y_resolution_param);
    in_format4->addWildcard(Format::ratio_x_param);
    in_format4->addWildcard(Format::ratio_y_param);
    in_format4->addWildcard(Format::framerate_param);
    
    in_format44 = getOwnInputProperty()->addNewFormat(Format::video_h264_type);
    in_format44->addWildcard(Format::x_resolution_param);
    in_format44->addWildcard(Format::y_resolution_param);
    in_format44->addWildcard(Format::ratio_x_param);
    in_format44->addWildcard(Format::ratio_y_param);
    in_format44->addWildcard(Format::framerate_param);
    in_format44->addWildcard(Format::bitrate_param);

    in_format5 = getOwnInputProperty()->addNewFormat(Format::video_h263plus_type);
    in_format5->addWildcard(Format::x_resolution_param);
    in_format5->addWildcard(Format::y_resolution_param);
    in_format5->addWildcard(Format::ratio_x_param);
    in_format5->addWildcard(Format::ratio_y_param);
    in_format5->addWildcard(Format::framerate_param);
    
    /* general output format */
    Format* out_format = getOwnOutputProperty()->addNewFormat(Format::video_raw_type);
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addStringParamValue(Format::channelorder_param,"yvu");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::colorspace_param,"yv12");
    out_format->addIntParamValue(Format::bitperpixel_param,12);
    out_format->addStringParamValue(Format::format_param, "planar");
    getOwnOutputProperty()->setDefaultFormat(out_format);

    return SUCCESS;
  }

  Result FFMpegDecodeNode::doInitOutput() {
    /* get input format */
    const Format* in_format = getInputFormat();
    if (in_format == NULL) {
      ERROR_STREAM("Missing input format" << std::endl);
      return FAILURE;
    }

    /* get FourCC code */
    string compressor;
    if(in_format->getFormatType() == Format::video_mpeg_type) {
      unsigned int version = in_format->getIntValue(Format::version_param);
      if(version == 1) {
	compressor = "MPG1";
      } else if(version == 2) {
	compressor = "MPG2";
      } else {
	/* SHOULD NOT HAPPEN */
	ERROR_STREAM("Invalid MPEG video version" << endl);
	return FAILURE;
      }
    } else if(in_format->getFormatType() == Format::video_mpeg4_type) {
      compressor = in_format->getStringValue(Format::compressor_param);
    } else if(in_format->getFormatType() == Format::video_h263_type) {
      compressor = "H263";
    } else if(in_format->getFormatType() == Format::video_h264_type) {
      compressor = "H264";
    } else if(in_format->getFormatType() == Format::video_h263plus_type) {
      compressor = "H263";
    } else if(in_format->getFormatType() == Format::video_jpeg_type) {
      compressor = "MJPG";
    } else {
      /* SHOULD NOT HAPPEN */
      ERROR_STREAM("Invalid input format type" << endl);
      return FAILURE;
    }
    
    /* find the correct decoder */
    CodecConstIterator found = codecs.find(compressor);
    if(found == codecs.end()) {
      /* SHOULD NOT HAPPEN */
      ERROR_STREAM("FATAL: No codec listed for FourCC code " << compressor << std::endl);
      return FAILURE;
    }
    
    /* find a codec with this ID */
    ffmpeg->ffAVCodec = avcodec_find_decoder(static_cast<CodecID>(found->second));
    if (!ffmpeg->ffAVCodec) {
      /* SHOULD NOT HAPPEN */
      ERROR_STREAM("FATAL: Unsupported codec" << std::endl);
      return FAILURE;
    }

    /* setup working input property */
    Format* i_format = new Format( *in_format );
    getOwnWorkingInputProperty()->addNewFormat(i_format);
    getOwnWorkingInputProperty()->setDefaultFormat(i_format);
    
    /* determine output format */
    if(in_format->getIntValue(Format::x_resolution_param) &&
       in_format->getIntValue(Format::y_resolution_param) &&
       in_format->getIntValue(Format::ratio_x_param) &&
       in_format->getIntValue(Format::ratio_y_param) &&
       in_format->getFloatValue(Format::framerate_param)) {
      /* get parameters from input format */
      width = in_format->getIntValue(Format::x_resolution_param);
      height = in_format->getIntValue(Format::y_resolution_param);
      ratio_x = in_format->getIntValue(Format::ratio_x_param);
      ratio_y = in_format->getIntValue(Format::ratio_y_param);
      framerate = in_format->getFloatValue(Format::framerate_param);
      DEBUG_STREAM("Got fully specified input format" << endl);
      DEBUG_STREAM("width is " << width << endl);
      DEBUG_STREAM("height is " << height << endl);
      DEBUG_STREAM("framerate is " << framerate << endl);
      
      // wmv specific parameters
      if(in_format->isParameterSpecified(Format::extradata_size_param) &&
	 in_format->isParameterSpecified(Format::extradata_param)) {
	video_extradata_size = in_format->getIntValue(Format::extradata_size_param);
	video_extradata = in_format->getStringValue(Format::extradata_param);
      }
      
      /* init context */
      initContext();
      
      /* update working output property and internal variables */
      updateFormat();
    } else {
      /* set parameters to zero */
      width = 0;
      height = 0;
      ratio_x = 0;
      ratio_y = 0;
      framerate = 0;
      
      /* init context */
      initContext();
      
      /* analyze data */
      DEBUG_STREAM("No fully specified input format, need to analyze data" << endl);
      setMaxAnalysisIterations(250);
      if(!analyseData()) {
	ERROR_STREAM("Data analysis failed" << endl);
	return FAILURE;
      }
    }
    
    return SUCCESS;
  }

  void FFMpegDecodeNode::updateFormat() {
    /* set some variables */
    int_framerate = (int) ceil(framerate * (double) framerate_base);
    width_half = width >> 1;
    height_half = height >> 1;

    /* set some more variables */
    avframe_size = ffmpeg->ffavcc->width * ffmpeg->ffavcc->height;    
    out_buffer_size = (avframe_size * 3) >> 1;
    
    /* get general working input format */
    const Format* i_format = getOwnWorkingInputProperty()->getDefaultFormat();
    assert(i_format);
    
    /* set general working output format */
    Format* o_format = getOwnWorkingOutputProperty()->addNewFormat(Format::video_raw_type);
    o_format->addIntParamValue(Format::x_resolution_param, width);
    o_format->addIntParamValue(Format::y_resolution_param, height);
    o_format->addIntParamValue(Format::ratio_x_param, ratio_x);
    o_format->addIntParamValue(Format::ratio_y_param, ratio_y);
    o_format->addFloatParamValue(Format::framerate_param, framerate);
    o_format->addStringParamValue(Format::channelorder_param,"yvu");
    o_format->addStringParamValue(Format::endian_param, "little");
    o_format->addStringParamValue(Format::colorspace_param,"yv12");
    o_format->addIntParamValue(Format::bitperpixel_param,12);
    o_format->addStringParamValue(Format::format_param, "planar");
    getOwnWorkingOutputProperty()->setDefaultFormat(o_format);
    
    /* set I/O partners */
    // XXX: fw: const_cast ...
    const_cast<Format*>(i_format)->setIOPartner(o_format);
    o_format->setIOPartner(i_format);
  }

  Result FFMpegDecodeNode::initContext() {
    /* allocate ffmpeg context */
    ffmpeg->ffavcc = avcodec_alloc_context();
    
    /* allocate frame */
    ffmpeg->avframe = avcodec_alloc_frame();
    
    /* init context */
    ffmpeg->ffavcc->width = width;
    ffmpeg->ffavcc->height = height;
#if LIBAVCODEC_BUILD >= 4754
    ffmpeg->ffavcc->time_base.den = (int)(framerate * FRAME_RATE_BASE);
#else
    ffmpeg->ffavcc->frame_rate = (int)(framerate * FRAME_RATE_BASE);
#endif
    if(video_extradata_size !=0 && video_extradata != "") {
      ffmpeg->ffavcc->extradata_size = video_extradata_size;
      ffmpeg->ffavcc->extradata = static_cast<uint8_t*>(malloc(video_extradata_size));
      memmove(ffmpeg->ffavcc->extradata, const_cast<char *>(video_extradata.c_str()), video_extradata_size);
    }
    
    /* open the codec */
    if(avcodec_open(ffmpeg->ffavcc, ffmpeg->ffAVCodec) < 0) {
      ERROR_STREAM("Could not open codec" << std::endl);
      
      /* release ffmpeg frame */
      if(ffmpeg->avframe) {
	free(ffmpeg->avframe);
      }
      
      /* release ffmpeg context */
      if(ffmpeg->ffavcc) {
	free(ffmpeg->ffavcc);
	ffmpeg->ffavcc = 0;
      }
      
      return FAILURE;
    }
    return SUCCESS;
  }
  
  Result FFMpegDecodeNode::doDeinitOutput() { 
    /* close codec */
    avcodec_close(ffmpeg->ffavcc);
    
    /* release ffmpeg frame */
    if(ffmpeg->avframe != NULL) {
      free(ffmpeg->avframe);
    }
    
    /* release ffmpeg context */
    if(ffmpeg->ffavcc) {
      free(ffmpeg->ffavcc);
      ffmpeg->ffavcc = NULL;
    }
    return SUCCESS;
  }

  Result FFMpegDecodeNode::doDeactivate() {
    /* flush ffmpeg buffers */
    if(ffmpeg->ffavcc) {
      avcodec_flush_buffers(ffmpeg->ffavcc);
    }
    return SUCCESS;
  }
    
  Message* FFMpegDecodeNode::processBuffer(Buffer* in_buffer) {
    if(!getWorkingFlag()) {
      got_picture = 0;
      
      DEBUG_STREAM("in_buffer->getUsed() = " << in_buffer->getUsed() << endl);
      DEBUG_STREAM("ffmpeg->avframe->key_frame =" << ffmpeg->avframe->key_frame << endl);
      out_size = avcodec_decode_video(ffmpeg->ffavcc,
				      ffmpeg->avframe,
				      &got_picture,
				      reinterpret_cast<UINT8*>(in_buffer->getData()),
				      in_buffer->getUsed());
      
      /* check format */
#if LIBAVCODEC_BUILD >= 4754
      float new_framerate = float(ffmpeg->ffavcc->time_base.den) / float(ffmpeg->ffavcc->time_base.num);
#else
      float new_framerate = float(ffmpeg->ffavcc->frame_rate) / float(ffmpeg->ffavcc->frame_rate_base);
#endif
      if(ffmpeg->ffavcc->width != width ||
	 ffmpeg->ffavcc->height != height ||
	 new_framerate != framerate) {
	/* set new values */
	width = ffmpeg->ffavcc->width;
	height = ffmpeg->ffavcc->height;
	framerate = new_framerate;
	DEBUG_STREAM("New X resolution is " << width << endl);
	DEBUG_STREAM("New Y resolution is " << height << endl);
	DEBUG_STREAM("New framerate is " << framerate << endl);
      
	// XXX: fw: determine correct aspect ratio from ffmpeg !!!
	if(getOwnWorkingInputProperty()->getDefaultFormat()->getFormatType() == Format::video_mpeg_type) {
	  // XXX: fw: assume 16:9 for mpeg1/2
	  ratio_x = 16;
	  ratio_y = 9;
	} else {
	  // XXX: fw: assume width:height otherwise
	  ratio_x = width;
	  ratio_y = height;
	}
      
	/* update working output property and internal variables */
	updateFormat();
      
	if(getAnalysingFlag()) {
	  /* done with analysing */
	  DEBUG_STREAM("Finished analysing" << endl);
	  setAnalysingFlag(false);
	} else {
	  /* send format change event */
	  saved_buffer = in_buffer;
	  setWorkingFlag(true);
	  DEBUG_STREAM("SetFormat: " << *getOwnWorkingOutputProperty()->getDefaultFormat() << endl);
	  return new CEvent(IFormatChange::create_setFormat(*getOwnWorkingOutputProperty()->getDefaultFormat()));
	}
      }
    } else {
      in_buffer = saved_buffer;
    }
    setWorkingFlag(false);
    
    Buffer* out_buffer = getNewBuffer(out_buffer_size, in_buffer);

    DEBUG_STREAM("got_picture = " << got_picture << endl);
    if (!got_picture) {
      // no picture yet
      DEBUG_STREAM("got no picture!!!" << endl);
      out_buffer->release();
      return 0;
    }

    //Sometimes the lib procudes non YUV420p data (mjpeg) -> convert the data to YUV420p
    if (ffmpeg->ffavcc->pix_fmt != PIX_FMT_YUV420P)
      img_convert((AVPicture*)ffmpeg->avframe,
		  PIX_FMT_YUV420P,
		  (AVPicture*)ffmpeg->avframe,
		  ffmpeg->ffavcc->pix_fmt,
		  ffmpeg->ffavcc->width,
		  ffmpeg->ffavcc->height);
    
    if (skip_until_keyframe) {
      if (ffmpeg->avframe->key_frame) {
	DEBUG_STREAM("FFMpeg-KeyFrame: " << in_buffer->getTimestamp().stream_counter << endl);
	skip_until_keyframe = false;
      } else {
	DEBUG_STREAM("Dropping non-keyframe!" << endl);
	out_buffer->release();
	out_buffer = NULL;
	return NULL;
      }
    }

    // As the picture provided by the library has a linesize of e.g. 752 pixels for an
    // image that actually has 720 pixels width, we have to copy the image into our
    // buffer line by line, and repeat this for the Y-Plane as well as for the U- and V-Plane
    out_buffer_data1 = reinterpret_cast<UINT8*>(out_buffer->getData());
    src1 = ffmpeg->avframe->data[0];
    for (line = 0; line < height; line++) {
      memcpy(out_buffer_data1, src1, width);
      out_buffer_data1 += width;
      src1 += ffmpeg->avframe->linesize[0];
    }

    // We also have to swap U and V, so ffmpeg->avframe->data[2] comes first
    // out_buffer_data1 is at the right place
    out_buffer_data2 = out_buffer_data1 + (avframe_size >> 2);    
    src1 = ffmpeg->avframe->data[2];
    src2 = ffmpeg->avframe->data[1];

    for (line = 0; line < height_half; line++) {
      memcpy(out_buffer_data1, src1, width_half);
      memcpy(out_buffer_data2, src2, width_half);
      out_buffer_data1 += width_half;
      out_buffer_data2 += width_half;
      src1 += ffmpeg->avframe->linesize[2];
      src2 += ffmpeg->avframe->linesize[1];
    }
    
    out_buffer->setUsed(out_buffer_size);
    DEBUG_STREAM("out_buffer->setUsed(" << in_buffer->getUsed() << ");"<< endl);
    DEBUG_STREAM("out_size = " << out_size << endl);

    // Caculate time-stamp for current frame
    //    Rational time(frame_count * framerate_base, int_framerate);
    
    in_buf_ts = in_buffer->getTimestamp();
    if (!in_buf_ts.is_valid) {
      //      timestamp.sync_time = Time(time);
      //      timestamp.stream_counter = frame_count;
      //      timestamp.is_valid = true;
      timestamp.is_valid = true;
      
      // use last valid timestamp and increment it
      timestamp.stream_counter++;
      timestamp.sync_time += Interval(Rational(framerate_base, int_framerate));
      
      DEBUG_STREAM("Set Timestamp " << timestamp << " for frame no " << frame_count << endl);
      out_buffer->setTimestamp(timestamp);
    }
    else {
      DEBUG_STREAM("Found Timestamp " << in_buf_ts << " for frame no " << frame_count << endl);
      DEBUG_STREAM("Leaving it untouched." << endl);
      
      // store as last valid timestamp
      timestamp.sync_time = in_buf_ts.sync_time;
      timestamp.stream_counter = in_buf_ts.stream_counter;
    }
    
    frame_count++;
    
    // the compressed Frame is not needed any more
    in_buffer->release();
    return out_buffer;
  }
  
  Result FFMpegDecodeNode::setVideoDelay(const nmm_int32& delay) {
    DEBUG_STREAM("Received Event: video_delay. delay is " << delay << endl);
    vid_delay = delay;
    if (delay > 1000) {
      start_time.tv_sec = delay / 1000;
    }
    start_time.tv_usec = - (vid_delay % 1000) * 1000;
    DEBUG_STREAM("start_time set to " << start_time.tv_sec << ", " << start_time.tv_usec << endl);
    return SUCCESS;
  }

  Result FFMpegDecodeNode::doFlush() {
    skip_until_keyframe = true;
    return SUCCESS;
  }

  Result FFMpegDecodeNode::setKeyframe() {
    /* there are no keyframes in raw video streams, so delete event */
    return DELETE;
  }

  void FFMpegDecodeNode::detectAllCodecs() throw() {
    /* get all available codecs */
    AVCodec *p = first_avcodec;
    DEBUG_STREAM("Available Decoders:" << endl);
    while (p) {
      if (p->decode) {
	DEBUG_STREAM("  " << p->name << endl);
	switch(p->id) {
	case CODEC_ID_MSMPEG4V1:
	  addCodec("MP41", p->id);
	  break;
	case CODEC_ID_MSMPEG4V2:
	  addCodec("MP42", p->id);
	  break;
	case CODEC_ID_MSMPEG4V3:
	  addCodec("MP43", p->id);
	  addCodec("DIV3", p->id);
	  break;
	case CODEC_ID_MPEG1VIDEO:
	  addCodec("MPEG", p->id);
	  addCodec("MPG1", p->id);
	  break;
	case CODEC_ID_MPEG2VIDEO:
	  addCodec("MPEG", p->id);
	  addCodec("MP2V", p->id);
	  addCodec("MPG2", p->id);
	  break;
	case CODEC_ID_H263:
	  addCodec("H263", p->id);
	  break;
	case CODEC_ID_H264:
	  addCodec("H264", p->id);
   addCodec("h264", p->id);
	  break;
	case CODEC_ID_H263P:
	  addCodec("H263P", p->id);
	  break;
	case CODEC_ID_RV10:
	  addCodec("RV10", p->id);
	  break;
	case CODEC_ID_MJPEG:
	  addCodec("MJPG", p->id);
	  break;
	case CODEC_ID_WMV1:
	  addCodec("WMV1", p->id);
	  break;
	case CODEC_ID_WMV2:
	  addCodec("WMV2", p->id);
	  break;
	case CODEC_ID_HUFFYUV:
	  addCodec("HFYU", p->id);
	  break;
	case CODEC_ID_MPEG4:
	  addCodec("DIVX", p->id);
	  addCodec("DX50", p->id);
	  addCodec("XVID", p->id);
	  break;
	default:
	  break;
	}
      }

      p = p->next;
    }
  }

  void FFMpegDecodeNode::addCodec(const std::string& fourcc, int id) {
    if(default_codec.empty()) {
      /* if we have no default yet, use this one */
      default_codec = fourcc;
    } else if(fourcc == "DIVX") {
      /* if DIVX is available, override default */
      default_codec = fourcc;
    }

    /* add codec to map */
    codecs.insert(std::pair<std::string, int>(fourcc, id));
    DEBUG_STREAM("    " << fourcc << std::endl);
  }
}

#include "nmm/nmm_suffix.hpp"
