/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <iostream>
#include "BrightnessNode.hpp"
#include "nmm/comm/format/Property.hpp"
#include "nmm/multimedia/Plugin.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"

NMM_REGISTER_PLUGIN(NMM::BrightnessNode, NMM, "BrightnessNode")

namespace NMM {

  BrightnessNode::BrightnessNode(const char *name)
                    : GenericFilterNode(name) {
    width = 0;
    height = 0;
    brightness = 1.0;
  }
  
  BrightnessNode::~BrightnessNode() {}
  
  Result BrightnessNode::doInit() {
    // initialization code
    // check if all parameters have been set correctly
    if (brightness >= 0.0) return SUCCESS;
    else return FAILURE;
    
    // register all supported input formats
    Format *myInputFormat;
    myInputFormat = getOwnInputProperty()->addNewFormat("video/raw");
    myInputFormat->addWildcard(Format::x_resolution_param);
    myInputFormat->addWildcard(Format::y_resolution_param);
    myInputFormat->addStringParamValue(Format::colorspace_param, "rgba32");
    myInputFormat->addStringParamValue(Format::channelorder_param, "rgba");
    myInputFormat->addIntParamValue(Format::bitperpixel_param,32);
    myInputFormat->addStringParamValue(Format::endian_param, "little");
    myInputFormat->addWildcard(Format::framerate_param);
    getOwnInputProperty()->setDefaultFormat(myInputFormat);
    
    // register all supported output formats
    Format *myOutputFormat;
    myOutputFormat = getOwnOutputProperty()->addNewFormat("video/raw");
    myOutputFormat->addWildcard(Format::x_resolution_param);
    myOutputFormat->addWildcard(Format::y_resolution_param);
    myOutputFormat->addStringParamValue(Format::colorspace_param, "rgba32");
    myOutputFormat->addStringParamValue(Format::channelorder_param, "rgba");
    myOutputFormat->addIntParamValue(Format::bitperpixel_param,32);
    myOutputFormat->addStringParamValue(Format::endian_param, "little");
    myOutputFormat->addWildcard(Format::framerate_param);
    getOwnOutputProperty()->setDefaultFormat(myOutputFormat);
    
    // set I/O partners
    myInputFormat->setIOPartner(myOutputFormat);
    myOutputFormat->setIOPartner(myInputFormat);
    
    return SUCCESS;
  }
  
  Result BrightnessNode::doDeinit() {
    // unregister all formats
    getOwnInputProperty()->erase();
    getOwnOutputProperty()->erase();
    return SUCCESS;
  }
  
  Result BrightnessNode::doInitOutput() {
    // get input format
    // make sure we have one, i.e. an input jack must have been requested successfully
    const Format *in_f = getInputFormat();
    if (in_f == NULL) { return FAILURE; }
    
    // store values in internal variables.
    width = in_f->getIntValue(Format::x_resolution_param);
    height = in_f->getIntValue(Format::y_resolution_param);
    
    if ((width <= 0) || (height <= 0)) {
      return FAILURE;
    }
    
    // init working input property
    Format* i_format = new Format(*in_f);
    getOwnWorkingInputProperty()->addNewFormat(i_format);
    getOwnWorkingInputProperty()->setDefaultFormat(i_format);
    
    // set output format
    // this is equal to input format for all filter nodes
    Format* o_format = new Format(*i_format);
    getOwnWorkingOutputProperty()->addNewFormat(o_format);
    getOwnWorkingOutputProperty()->setDefaultFormat(o_format);
    
    // set I/O partners
    i_format->setIOPartner(o_format);
    o_format->setIOPartner(i_format);
    
    return SUCCESS;
  }
  
  Result BrightnessNode::doDeinitOutput() {
    getOwnWorkingInputProperty()->erase();
    getOwnWorkingOutputProperty()->erase();
    return SUCCESS;
  }
  
  Buffer* BrightnessNode::processBuffer(Buffer *in_buffer) {
    in_buffer = in_buffer->getWriteableInstance();
    unsigned char *data = (unsigned char*)in_buffer->getData();
    
    for (int y=0; y<height; y++) {
      for (int x=0; x<width; x++) {
        int r, g, b;
        r = (int)data[(x + width*y)*4 + 0];
        g = (int)data[(x + width*y)*4 + 1];
        b = (int)data[(x + width*y)*4 + 2];
        r = (int)(r * brightness);
        if (r > 255) r = 255; if (r < 0) r = 0;
        g = (int)(g * brightness);
        if (g > 255) g = 255; if (g < 0) g = 0;
        b = (int)(b * brightness);
        if (b > 255) b = 255; if (b < 0) b = 0;
        data[(x + width*y)*4 + 0] = (unsigned char)r;
        data[(x + width*y)*4 + 1] = (unsigned char)g;
        data[(x + width*y)*4 + 2] = (unsigned char)b;
        
      }
    }
    return in_buffer;
  }
  
  void BrightnessNode::setBrightness(const nmm_float& value) {
    brightness = value;
    cout << "*******Brightness = " << brightness << endl;
  }
}
