/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "LogoNode.hpp"
#include "nmm/comm/format/Property.hpp"
#include "nmm/multimedia/Plugin.hpp"
#include "nmm/utils/directory_utils.hpp"
#include "nmm/utils/SecurityCheck.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"

#include <stdio.h>
#include <iostream>
#include <assert.h>

#ifndef NULL
#define NULL 0
#endif

NMM_REGISTER_PLUGIN(NMM::LogoNode, NMM, "LogoNode")

namespace NMM {

  LogoNode::LogoNode(const char *name,
		     const StreamQueue::Mode queue_mode,
		     const int queue_size )
    : GenericFilterNode(name, queue_mode, queue_size) 
  {
    logo_filename = getNMMResource("logo.png");

    x_offset = 0;
    y_offset = 0;
    
    image_width = 0; 
    image_height = 0;

    png_rows = 0;
  }

  LogoNode::~LogoNode() {
    if (png_rows != 0) {
      for (unsigned int row = 0; row < png_height; row++) {
        delete [] png_rows[row];
      }
      delete [] png_rows;
    }
  }

  Result LogoNode::doInit() {
    // initialize formats
    NMM::Format *myDefaultInputFormat;
    myDefaultInputFormat=getOwnInputProperty()->addNewFormat("video/raw");
    myDefaultInputFormat->addIntParamValue(Format::x_resolution_param, 250);
    myDefaultInputFormat->addIntParamValue(Format::y_resolution_param, 150);
    myDefaultInputFormat->addIntParamValue(Format::ratio_x_param, 250);
    myDefaultInputFormat->addIntParamValue(Format::ratio_y_param, 150);
    myDefaultInputFormat->addStringParamValue(Format::colorspace_param,"rgb24");
    myDefaultInputFormat->addStringParamValue(Format::channelorder_param,"rgb");
    myDefaultInputFormat->addIntParamValue(Format::bitperpixel_param,32);
    myDefaultInputFormat->addStringParamValue(Format::endian_param, "little");
    myDefaultInputFormat->addStringParamValue(Format::format_param,"packet");
    myDefaultInputFormat->addFloatParamValue(Format::framerate_param, 15.0);
    getOwnInputProperty()->setDefaultFormat(myDefaultInputFormat);

    NMM::Format *myDefaultOutputFormat;
    myDefaultOutputFormat=getOwnInputProperty()->addNewFormat("video/raw");
    myDefaultOutputFormat->addIntParamValue(Format::x_resolution_param, 250);
    myDefaultOutputFormat->addIntParamValue(Format::y_resolution_param, 150);
    myDefaultOutputFormat->addIntParamValue(Format::ratio_x_param, 250);
    myDefaultOutputFormat->addIntParamValue(Format::ratio_y_param, 150);
    myDefaultOutputFormat->addStringParamValue(Format::colorspace_param,"rgb24");
    myDefaultOutputFormat->addStringParamValue(Format::channelorder_param,"rgb");
    myDefaultOutputFormat->addIntParamValue(Format::bitperpixel_param,32);
    myDefaultOutputFormat->addIntParamValue(Format::bitperpixel_param,24);
    myDefaultOutputFormat->addStringParamValue(Format::endian_param, "little");
    myDefaultOutputFormat->addStringParamValue(Format::format_param,"packet");
    myDefaultOutputFormat->addFloatParamValue(Format::framerate_param, 15.0);
    getOwnOutputProperty()->setDefaultFormat(myDefaultOutputFormat);

    myDefaultInputFormat->setIOPartner(myDefaultOutputFormat);
    myDefaultOutputFormat->setIOPartner(myDefaultInputFormat);

    NMM::Format *myInputFormat;
    myInputFormat=getOwnInputProperty()->addNewFormat("video/raw");
    myInputFormat->addWildcard(Format::x_resolution_param);
    myInputFormat->addWildcard(Format::y_resolution_param);
    myInputFormat->addWildcard(Format::ratio_x_param);
    myInputFormat->addWildcard(Format::ratio_y_param);
    myInputFormat->addStringParamValue(Format::colorspace_param,"rgb24");
    myInputFormat->addStringParamValue(Format::channelorder_param,"rgb");
    myInputFormat->addIntParamValue(Format::bitperpixel_param,32);
    myInputFormat->addStringParamValue(Format::endian_param, "little");
    myInputFormat->addStringParamValue(Format::format_param,"packet");
    myInputFormat->addWildcard(Format::framerate_param);

    NMM::Format *myOutputFormat;
    myOutputFormat=getOwnOutputProperty()->addNewFormat("video/raw");
    myOutputFormat->addWildcard(Format::x_resolution_param);
    myOutputFormat->addWildcard(Format::y_resolution_param);
    myOutputFormat->addWildcard(Format::ratio_x_param);
    myOutputFormat->addWildcard(Format::ratio_y_param);
    myOutputFormat->addStringParamValue(Format::colorspace_param,"rgb24");
    myOutputFormat->addStringParamValue(Format::channelorder_param,"rgb");
    myOutputFormat->addIntParamValue(Format::bitperpixel_param,32);
    myOutputFormat->addIntParamValue(Format::bitperpixel_param,24);
    myOutputFormat->addStringParamValue(Format::endian_param, "little");
    myOutputFormat->addWildcard(Format::framerate_param);
    myOutputFormat->addStringParamValue(Format::format_param,"packet");
    
    myInputFormat->setIOPartner(myOutputFormat);
    myOutputFormat->setIOPartner(myInputFormat);
    
    return SUCCESS;
  }
  
  Result LogoNode::doInitOutput() {
    const Format *format = getInputFormat();
    if (format == NULL) {
      return FAILURE; // error, not without a format!
    }
    
    if (FAILURE == preparePNGLogo(logo_filename)) {
      ERROR_STREAM("Failed to open logo file" << endl);
      return FAILURE;
    }
    
    image_width = format->getIntValue(Format::x_resolution_param);
    image_height = format->getIntValue(Format::y_resolution_param);
    
    if ((image_width <= 0) || (image_height <= 0)) {
       return FAILURE; // error!
    }
    
    // set working format
    // this is equal to input format for all filter nodes
    Format* in_f = new Format(*format);
    getOwnWorkingInputProperty()->addNewFormat(in_f);
    getOwnWorkingInputProperty()->setDefaultFormat(in_f);
    Format* out_f = new Format(*format);
    getOwnWorkingOutputProperty()->addNewFormat(out_f);
    getOwnWorkingOutputProperty()->setDefaultFormat(out_f);
    
    return SUCCESS;
  }

  Buffer* LogoNode::processBuffer(Buffer *in_buffer) {
    in_buffer = in_buffer->getWriteableInstance();
    paintPNGLogo(in_buffer->getData(), x_offset, y_offset);
    return in_buffer;
  }
  
  void LogoNode::setLogoImage(const string& path) {
    logo_filename = path;
  }
  
  void  LogoNode::setOffset(const nmm_int32& x, const nmm_int32& y) {
    x_offset = x;
    y_offset = y;
  }
  
  void  LogoNode::getOffset(nmm_int32& x, nmm_int32& y) const {
    x = x_offset;
    y = y_offset;
  }
  
  /* ==== Internal methods follow ==== */
  
  void LogoNode::setPixel(char *image, int x, int y,
			  unsigned char r, unsigned char g, unsigned char b) {
    unsigned char *ptr;
    ptr = (unsigned char *)image;
    ptr += 4 * (image_width * y + x);
    ptr[0] = b;
    ptr[1] = g;
    ptr[2] = r;
  }

  Result LogoNode::preparePNGLogo(string filename) {

    // SEE THE LIBPNG-EXAMPLE.c FOR DETAILS!

    png_structp png_ptr;
    png_infop info_ptr;
    //unsigned int sig_read = 0;
    //png_uint_32 width, height;
    int bit_depth, color_type, interlace_type;
    FILE *fp;
    
    png_rows = NULL;
    
    if(!SSecurityCheck::getInstance().checkFilenameForReading(filename)) {
      ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return FAILURE;
    }
    if ((fp = fopen(filename.c_str(), "rb")) == NULL) {
      ERROR_STREAM("Could not open file " << filename << endl);
      return FAILURE;
    }
    
    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING,
				     NULL, NULL, NULL);
    
    if (png_ptr == NULL) {
      fclose(fp);
      return FAILURE;
    }
    
    info_ptr = png_create_info_struct(png_ptr);
    if (info_ptr == NULL) {
      fclose(fp);
      png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
      return FAILURE;
    }

    if (setjmp(png_ptr->jmpbuf)) {
      png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
      fclose(fp);
      return FAILURE;
    }

    png_init_io(png_ptr, fp);

    png_read_info(png_ptr, info_ptr);
    
    png_get_IHDR(png_ptr, info_ptr, &png_width, &png_height,
		 &bit_depth, &color_type,
		 &interlace_type, NULL, NULL);
    
    png_rows = new png_bytep[png_height];

    for (unsigned int row = 0; row < png_height; row++) {
      png_rows[row] = new png_byte[png_get_rowbytes(png_ptr, info_ptr)];
    }
    
    png_read_image(png_ptr, png_rows);
    
    png_read_end(png_ptr, info_ptr);
    
    png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
    
    fclose(fp);
    
    return SUCCESS;
  }

  void LogoNode::paintPNGLogo(char *image, int xoff, int yoff){
    unsigned char *pngimg;
    unsigned char *img; 

    // neg. offsets mean offset from OTHER border

    if (xoff < 0) xoff = (image_width + xoff) - png_width;
    if (yoff < 0) yoff = (image_height + yoff) - png_height;

    // clip to visible area
    // following variables are in coords of the logo!

    int x_start, x_stop;
    int y_start, y_stop;
    
    x_start = 0; x_stop = png_width;
    if (xoff + x_start < 0) x_start = -xoff;    
    if (xoff + x_stop > image_width) x_stop = image_width - xoff;

    y_start = 0; y_stop = png_height;
    if (yoff + y_start < 0) y_start = -yoff;  
    if (yoff + y_stop > image_height) y_stop = image_height - yoff;

    if ((x_start >= x_stop) || (y_start >= y_stop)) return;
    // logo out of visible area

    assert((x_start >= 0) 
           && ("x_start negativ in LogoNode::paintPNGLogo!"));
    assert((x_stop <= (int)png_width)
           && ("x_stop Ueberlauf in LogoNode::paintPNGLogo!"));
    assert((y_start >= 0) 
           && ("y_start negativ in LogoNode::paintPNGLogo!"));
    assert((y_stop <= (int)png_height)
           && ("y_stop Ueberlauf in LogoNode::paintPNGLogo!"));
    assert((xoff + x_start >= 0)
           && ("Bild faengt zu weit links an in LogoNode::paintPNGLogo!"));
    assert((xoff + x_stop < image_width)
           && ("Bild hoert zu weit rechts auf in LogoNode::paintPNGLogo!"));
    assert((yoff + y_start >= 0)
           && ("Bild faengt zu weit oben an in LogoNode::paintPNGLogo!"));
    assert((yoff + y_stop < image_height)
           && ("Bild hoert zu weit unten auf in LogoNode::paintPNGLogo!"));

    for (int y = y_start; y < y_stop; y++) {
      assert ((y >= 0) && (y < (int)png_height)
              && ("y out of bounds in LogoNode::paintPNGLogo!"));
      pngimg = (unsigned char*) png_rows[y] + 4 * x_start;
      assert ((4 * (xoff + image_width * (yoff + y))
               <= image_width * image_height * 4)
              && ("image pointer out of bounds in LogoNode::paintPNGLogo!"));
      img = (unsigned char*)image
        + 4 * (xoff + x_start+ image_width * (yoff + y));
      for (int x = x_start; x < x_stop; x++) {
	unsigned char r, g, b, a;
	unsigned char r0, g0, b0;
        assert(((unsigned char*)pngimg - (unsigned char*)png_rows[y]
                < 4 * (int)png_width)
               && ("pngimg out of bounds in LogoNode::paintPNGLogo!"));
	r = pngimg[0];
	g = pngimg[1];
	b = pngimg[2];
	a = pngimg[3];
	
	if (a > 0) {
	  if (a == 255) {
	    setPixel(image, xoff+x, yoff+y, r, g, b);
	  } else {
	    r0 = img[2];
	    g0 = img[1];
	    b0 = img[0];
	    float t = a / 255.0;
	    r = (unsigned char) (r * t + r0 * (1-t));
	    g = (unsigned char) (g * t + g0 * (1-t));
	    b = (unsigned char) (b * t + b0 * (1-t));
	    setPixel(image, xoff+x, yoff+y, r, g, b);
	  }
	}
	pngimg += 4; img += 4;
      }
    }    
  }
}
