/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "nmm/plugins/video/filter/OverlayNode.hpp"
#include "nmm/utils/overlay/overlay.h"
#include "nmm/comm/format/Property.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"

#include "nmm/multimedia/IFormatChange.hpp"

NMM_REGISTER_PLUGIN(OverlayNode, NMM, "OverlayNode")

#ifdef DELETE
#undef DELETE
#endif
#define RGBA32 0
#define YV12 1

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  OverlayNode::OverlayNode(const char *name, const StreamQueue::Mode queue_mode, const int queue_size) :
    GenericMultiplexerNode(name, queue_mode, queue_size) {

    foreground_width = 0;
    foreground_height = 0;
    new_foreground_width = 0;
    new_foreground_height = 0;
    background_width = 0;
    background_height = 0;
    x_offset = 0;
    y_offset = 0;
    new_x_offset = 0;
    new_y_offset = 0;
    alpha = 256; // 1.0
    
    last_foreground = 0;
    last_background = 0;
    
    saved_background = NULL;
    
    send_format = false;
    saved_in_buffer = NULL;
  }

  OverlayNode::~OverlayNode() {
    if ( saved_background ) {
      saved_background->release();
      saved_background = NULL;
    }
  }
  
  Result OverlayNode::doInit() {
    
    /* FIXME: should be false */
    save_next_background = false;
    save_still_frame = false;
    send_format = false;
    
    Format *format;
    
    // input foreground stream
    addInputStream("foreground");
    
    // input foreground rgb input format
    format = getOwnInputProperty()->addNewFormat("video/raw");
    format->addWildcard(Format::x_resolution_param);
    format->addWildcard(Format::y_resolution_param);
    format->addWildcard(Format::ratio_x_param);
    format->addWildcard(Format::ratio_y_param);
    format->addWildcard(Format::framerate_param);
    format->addStringParamValue(Format::colorspace_param,"rgba");
    format->addStringParamValue(Format::channelorder_param,"bgra");
    format->addIntParamValue(Format::bitperpixel_param,32);
    format->addStringParamValue(Format::endian_param, "little");
    format->addWildcard(Format::framerate_param);
    format->addStringParamValue(Format::format_param,"packet");
    format->setJackTag("foreground");
    
    // input foreground yuv format (default)
    format = getOwnInputProperty()->addNewFormat("video/raw");
    format->addWildcard(Format::x_resolution_param);
    format->addWildcard(Format::y_resolution_param);
    format->addWildcard(Format::ratio_x_param);
    format->addWildcard(Format::ratio_y_param);
    format->addWildcard(Format::framerate_param);
    format->addStringParamValue(Format::colorspace_param,"yv12a");
    format->addStringParamValue(Format::channelorder_param,"yvua");
    format->addIntParamValue(Format::bitperpixel_param,20);
    format->addStringParamValue(Format::endian_param, "little");
    format->addWildcard(Format::framerate_param);
    format->addStringParamValue(Format::format_param,"planar");
    format->setJackTag("foreground");
    getOwnInputProperty()->setDefaultFormat(format, "foreground");
    
    // input background stream
    addInputStream("background");
    
    // input background rgb format
    in_b_rgba_format = getOwnInputProperty()->addNewFormat("video/raw");
    in_b_rgba_format->addWildcard(Format::x_resolution_param);
    in_b_rgba_format->addWildcard(Format::y_resolution_param);
    in_b_rgba_format->addWildcard(Format::ratio_x_param);
    in_b_rgba_format->addWildcard(Format::ratio_y_param);
    in_b_rgba_format->addWildcard(Format::framerate_param);
    in_b_rgba_format->addStringParamValue(Format::colorspace_param,"rgba32");
    in_b_rgba_format->addStringParamValue(Format::channelorder_param,"bgra");
    in_b_rgba_format->addIntParamValue(Format::bitperpixel_param,32);
    in_b_rgba_format->addStringParamValue(Format::endian_param, "little");
    in_b_rgba_format->addStringParamValue(Format::format_param,"packet");
    in_b_rgba_format->setJackTag("background");
    
    // input background yuv format (default)
    in_b_yv12_format=getOwnInputProperty()->addNewFormat("video/raw");
    in_b_yv12_format->addWildcard(Format::x_resolution_param);
    in_b_yv12_format->addWildcard(Format::y_resolution_param);
    in_b_yv12_format->addWildcard(Format::ratio_x_param);
    in_b_yv12_format->addWildcard(Format::ratio_y_param);
    in_b_yv12_format->addWildcard(Format::framerate_param);
    in_b_yv12_format->addStringParamValue(Format::colorspace_param,"yv12");
    in_b_yv12_format->addStringParamValue(Format::channelorder_param,"yvu");
    in_b_yv12_format->addIntParamValue(Format::bitperpixel_param,12);
    in_b_yv12_format->addStringParamValue(Format::endian_param, "little");
    in_b_yv12_format->addStringParamValue(Format::format_param,"planar");
    in_b_yv12_format->setJackTag("background");
    getOwnInputProperty()->setDefaultFormat(in_b_yv12_format, "background");
    
    // output rgb format
    out_rgba_format = getOwnOutputProperty()->addNewFormat("video/raw");
    out_rgba_format->addWildcard(Format::x_resolution_param);
    out_rgba_format->addWildcard(Format::y_resolution_param);
    out_rgba_format->addWildcard(Format::ratio_x_param);
    out_rgba_format->addWildcard(Format::ratio_y_param);
    out_rgba_format->addWildcard(Format::framerate_param);
    out_rgba_format->addStringParamValue(Format::colorspace_param,"rgba");
    out_rgba_format->addStringParamValue(Format::channelorder_param,"bgra");
    out_rgba_format->addIntParamValue(Format::bitperpixel_param,32);
    out_rgba_format->addStringParamValue(Format::endian_param,"little");
    out_rgba_format->addStringParamValue(Format::format_param,"packet");
    out_rgba_format->setIOPartner(in_b_rgba_format);
    in_b_rgba_format->setIOPartner(out_rgba_format);
    
    // output yuv format (default)
    out_yv12_format=getOwnOutputProperty()->addNewFormat("video/raw");
    out_yv12_format->addWildcard(Format::x_resolution_param);
    out_yv12_format->addWildcard(Format::y_resolution_param);
    out_yv12_format->addWildcard(Format::ratio_x_param);
    out_yv12_format->addWildcard(Format::ratio_y_param);
    out_yv12_format->addWildcard(Format::framerate_param);
    out_yv12_format->addStringParamValue(Format::colorspace_param,"yv12");
    out_yv12_format->addStringParamValue(Format::channelorder_param,"yvu");
    out_yv12_format->addIntParamValue(Format::bitperpixel_param,12);
    out_yv12_format->addStringParamValue(Format::endian_param, "little");
    out_yv12_format->addStringParamValue(Format::format_param,"planar");
    out_yv12_format->setIOPartner(in_b_yv12_format);
    in_b_yv12_format->setIOPartner(out_yv12_format);
    getOwnOutputProperty()->setDefaultFormat(out_yv12_format, "default");
    
    return SUCCESS;
  }
  
  Result OverlayNode::doDeinit() {
    getOwnInputProperty()->erase();
    getOwnOutputProperty()->erase();
    return SUCCESS;
  }
  
  Result OverlayNode::doInitOutput() {
    
    // get input formats and make sure they exist
    const Format *inf_f = getInputFormat("foreground");
    const Format *inb_f = getInputFormat("background");
    if (inf_f == NULL) {
      ERROR_STREAM("No foreground input format" << endl);
      return FAILURE;
    }
    if (inb_f == NULL) {
      ERROR_STREAM("No background input format" << endl);
      return FAILURE;
    }
    
    // set internal variables
    // check for required parameters
    foreground_width = inf_f->getIntValue(Format::x_resolution_param);
    foreground_height = inf_f->getIntValue(Format::y_resolution_param);
    background_width = inb_f->getIntValue(Format::x_resolution_param);
    background_height = inb_f->getIntValue(Format::y_resolution_param);
    
    // background specifies colorspace, ratio and framerate
    int ratio_x = inb_f->getIntValue(Format::ratio_x_param);
    int ratio_y = inb_f->getIntValue(Format::ratio_y_param);
    float framerate = inb_f->getFloatValue(Format::framerate_param);
    string colorspace = inb_f->getStringValue(Format::colorspace_param);
    
    // check if background and foreground format are compatible
    // broken. reason: foreground must be same as background PLUS alpha channel.
//     if (colorspace != inf_f->getStringValue(Format::colorspace_param)) {
//       ERROR_STREAM("Foreground and background stream have different colorspace" << endl);
//       return FAILURE;
//     }
    
    // determine colorspace
    if(colorspace == "rgba") {
      color = RGBA32;
    }
    if(colorspace == "yv12") {
      color = YV12;
    }
    
    // working output property
    Format* out_f = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
    out_f->addIntParamValue(Format::x_resolution_param, background_width);
    out_f->addIntParamValue(Format::y_resolution_param, background_height);
    out_f->addIntParamValue(Format::ratio_x_param, ratio_x);
    out_f->addIntParamValue(Format::ratio_y_param, ratio_y);
    out_f->addFloatParamValue(Format::framerate_param, framerate);
    if (color == YV12) {
      out_f->addStringParamValue(Format::colorspace_param,"yv12");
      out_f->addStringParamValue(Format::channelorder_param,"yvu");
      out_f->addIntParamValue(Format::bitperpixel_param,12);
      out_f->addStringParamValue(Format::format_param,"planar");
    } else {
      out_f->addStringParamValue(Format::colorspace_param,"rgba");
      out_f->addStringParamValue(Format::channelorder_param,"bgra");
      out_f->addIntParamValue(Format::bitperpixel_param,32);
      out_f->addStringParamValue(Format::format_param,"packet");
    }
    out_f->addStringParamValue(Format::endian_param, "little");
    out_f->setIOPartner(inb_f);
    getOwnWorkingOutputProperty()->setDefaultFormat(out_f);
    
    // working input property
    Format* ifb = new Format(*inf_f);
    Format* iff = new Format(*inb_f);
    getOwnWorkingInputProperty()->addNewFormat(ifb);
    getOwnWorkingInputProperty()->addNewFormat(iff);
    getOwnWorkingInputProperty()->setDefaultFormat(ifb);
    getOwnWorkingInputProperty()->setDefaultFormat(iff);
    return SUCCESS;
  }
  
  Result OverlayNode::doDeinitOutput() {
    getOwnWorkingOutputProperty()->erase();
    getOwnWorkingInputProperty()->erase();
    return SUCCESS;
  }
  
  Result OverlayNode::doActivate() {
    return SUCCESS;
  }
  
  Result OverlayNode::syncReset() {
    /* to avoid sending sync_reset twice, we delete it if we got it from foreground,
     * so the synchroniser is not confused
     */
    if ( strcmp( getCurrentRecvInputStream(), "foreground" ) == 0 ) return DELETE;

    return SUCCESS;
  }

  Result OverlayNode::setStillFrame(const nmm_bool& value) {
    // fw: still frame events apply to background only
    if ( strcmp( getCurrentRecvInputStream(), "foreground" ) == 0 ) return DELETE;
    
    save_next_background = value;
    return SUCCESS;
  }

  Result OverlayNode::disableOverlay() {
    if(last_foreground) last_foreground->release();
    last_foreground = NULL;
    
    return SUCCESS;
  }
  
  void OverlayNode::setOffset(const nmm_int32& x, const nmm_int32& y) {
    x_offset = x;
    y_offset = y;
  }
  
  void OverlayNode::getOffset(nmm_int32& x, nmm_int32& y) const {
    x = x_offset;
    y = y_offset;
  }
  
  Result OverlayNode::setFormat( const Format& format ) {
    
    if (!strcmp( getCurrentRecvInputStream(), "background" )) {
      
      // set new background size
      background_width  = format.getIntValue(Format::x_resolution_param);
      background_height = format.getIntValue(Format::y_resolution_param);
      
      // get other parameters
      int ratio_x = format.getIntValue(Format::ratio_x_param);
      int ratio_y = format.getIntValue(Format::ratio_y_param);
      float framerate = format.getFloatValue(Format::framerate_param);
      string colorspace = format.getStringValue(Format::colorspace_param);
      
      // rebuild working output property
      getOwnWorkingOutputProperty()->erase();
      Format* out_f = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
      out_f->addIntParamValue(Format::x_resolution_param, background_width);
      out_f->addIntParamValue(Format::y_resolution_param, background_height);
      out_f->addIntParamValue(Format::ratio_x_param, ratio_x);
      out_f->addIntParamValue(Format::ratio_y_param, ratio_y);
      out_f->addFloatParamValue(Format::framerate_param, framerate);
      if (color == YV12) {
        out_f->addStringParamValue(Format::colorspace_param,"yv12");
        out_f->addStringParamValue(Format::channelorder_param,"yvu");
        out_f->addIntParamValue(Format::bitperpixel_param,12);
        out_f->addStringParamValue(Format::format_param,"planar");
      } else {
        out_f->addStringParamValue(Format::colorspace_param,"rgba");
        out_f->addStringParamValue(Format::channelorder_param,"bgra");
        out_f->addIntParamValue(Format::bitperpixel_param,32);
        out_f->addStringParamValue(Format::format_param,"packet");
      }
      out_f->addStringParamValue(Format::endian_param, "little");
      getOwnWorkingOutputProperty()->setDefaultFormat(out_f);
      
      // send new output format
      send_format = true;
      
    } else if (!strcmp( getCurrentRecvInputStream(), "foreground" )) {
      
      // set new foreground size
      foreground_width  = format.getIntValue(Format::x_resolution_param);
      foreground_height = format.getIntValue(Format::y_resolution_param);
      
      // output format is not affected
      
    }
    
    // the event only applies to this node, so we must delete it!
    return DELETE;
  }
  
  void OverlayNode::setForeground( const nmm_int32& x, const nmm_int32& y, const nmm_int32& w, const nmm_int32& h ) {
    new_foreground_width  = new_segment_width = w;
    new_foreground_height = new_segment_height = h;
    
    new_x_offset = x;
    new_y_offset = y;
    
    new_segment_x_offset = 0;
    new_segment_y_offset = 0;
  }
  
  void OverlayNode::setForegroundSegment( const nmm_int32& _x, const nmm_int32& _y, const nmm_int32& _w, const nmm_int32& _h ) {
    int x = _x;
    int y = _y;
    int w = _w;
    int h = _h;
    
    if ( x%2 ) x++;
    if ( y%2 ) y++;
    if ( w%2 ) w--;
    if ( h%2 ) h--;
    
    new_segment_width  = w;
    new_segment_height = h;
    
    new_x_offset += x;
    new_y_offset += y;
    
    new_segment_x_offset = x;
    new_segment_y_offset = y;
  }
  
  void OverlayNode::setAlpha(const nmm_float& a) {
    alpha  = (int)(255.0 * a);
    if (alpha < 0) alpha = 0;
    if (alpha > 255) alpha = 255;
  }

  nmm_float OverlayNode::getAlpha() const {
    return alpha;
  }
  
  Message *OverlayNode::processBuffer(Buffer *in_buffer) {
    if (getWorkingFlag()) {
      setWorkingFlag(false);
      in_buffer = saved_in_buffer;
      saved_in_buffer = NULL;
    }
    
    // send format if necessary
    if (send_format) {
      send_format = false;
      
      setWorkingFlag(true);
      saved_in_buffer = in_buffer;
      
      const Format* out_f = getOwnWorkingOutputProperty()->getDefaultFormat();
      return ( new CEvent(Event(IFormatChange::setFormat_event, new TInValue<Format>(*out_f))) );
    }
    
    // get latest background
    // do not forget to request writeable instance
    // since background image will be used for overlay operation
    if(strcmp(getCurrentRecvInputStream(), "background") == 0) {
      last_background = in_buffer->getWriteableInstance();
      
      have_still_frame = false;
      
      if ( saved_background ) {
        saved_background->release();
        saved_background = NULL;
      }
      
      if ( save_next_background ) {
        DEBUG_STREAM( "Saving Background" << endl );
	//save_next_background = false;
        
	saved_background = last_background->copy();
	have_still_frame = true;
      }
    } else {
      // get latest foreground and return 0, since we
      // cannot do the overlay without a background
      if(strcmp(getCurrentRecvInputStream(), "foreground") == 0) {
	// do not forget to release last kept foreground
	if(last_foreground) {
          last_foreground->release();
        }
        
	last_foreground   = in_buffer;
	foreground_width  = new_foreground_width;
	foreground_height = new_foreground_height;
	x_offset          = new_x_offset;
	y_offset          = new_y_offset;
        
	segment_width     = new_segment_width;
        segment_height    = new_segment_height;
        segment_x_offset  = new_segment_x_offset;
        segment_y_offset  = new_segment_y_offset;
        
        // Save overlay as png
//         PNGUtils png;
//         png.setFormat( PNGUtils::YV12A );
//         png.setDimension( foreground_width, foreground_height );
//         static int count=0;
//         char temp[128];
//         sprintf( temp, "overlay%i.png", count++ ),
//           png.writeImage( temp, (unsigned char*) in_buffer->getData() );
        
	if ( !have_still_frame ) {
          return 0;
	} else {
	  DEBUG_STREAM( "Copying saved background" << endl );
          // get writeable instance of a copy
          last_background = saved_background->copy()->getWriteableInstance();
	}
      } else {
        ERROR_STREAM("BUG: wrong input sream (never happens!)" << endl);
	return 0;
      }
    }

    // if no foreground so far, return background (might be 0)
    if (!last_foreground) {
      return last_background;
    }

    // overlay

    unsigned char *foreground_data; // fixed pointer to data
    unsigned char *background_data; // fixed pointer to data
    unsigned char *foreground_img; // walking pointer to data
    unsigned char *background_img; // walking pointer to data

    // clipping box in foreground image coordinates
    int x_start, x_stop;
    int y_start, y_stop;

    x_start = 0; x_stop = foreground_width;
    if (x_offset + x_start < 0) x_start = -x_offset;
    if (x_offset + x_stop > background_width)
      x_stop = background_width - x_offset;

    y_start = 0; y_stop = foreground_height;
    if (y_offset + y_start < 0) y_start = -y_offset;
    if (y_offset + y_stop > background_height)
      y_stop = background_height - y_offset;

    if ((x_start >= x_stop) || (y_start >= y_stop)) {
      // foreground completely clipped out
      
      return last_background; 
    }

    foreground_data = (unsigned char*) last_foreground->getData();
    background_data = (unsigned char*) last_background->getData();    

    switch( color ) {
      case RGBA32:
        for (int y = y_start; y < y_stop; y++) {
          foreground_img = foreground_data + 4 * (x_start + foreground_width * y);
          background_img = background_data
            + 4 * (x_offset + x_start + background_width * (y_offset + y));
          for (int x = x_start; x < x_stop; x++) {
	    unsigned char r, g, b, a;
	    unsigned char r0, g0, b0;
	    r = foreground_img[0];
	    g = foreground_img[1];
	    b = foreground_img[2];
	    a = foreground_img[3] * alpha >> 8;

	    if (a > 0) {
	      if (a >= 255) {
	        setPixel(background_data, x_offset + x, y_offset + y, r, g, b);
	      } else {
	        r0 = background_img[2];
	        g0 = background_img[1];
	        b0 = background_img[0];
	        //float t = a / 255.0;
                int t = 256 - (int)a;
	        //r = (unsigned char) (r * t + r0 * (1-t));
	        //g = (unsigned char) (g * t + g0 * (1-t));
	        //b = (unsigned char) (b * t + b0 * (1-t));
                r = (unsigned char) ((r * a + r0 * t) >> 8);
                g = (unsigned char) ((g * a + g0 * t) >> 8);
                b = (unsigned char) ((b * a + b0 * t) >> 8);
	        setPixel(background_data, x_offset + x, y_offset + y, r, g, b);
	      }
	    }
	    foreground_img += 4; background_img += 4;
          }
        }
	break;

       case YV12:         
         if (segment_width > 0 && segment_height > 0 ) {
           YV12Aoverlay2( foreground_data, x_offset, y_offset,
                          segment_x_offset,  segment_y_offset,
                          segment_width, segment_height,
                          foreground_width, foreground_height,
                          background_data, background_width, background_height );
         }
	break;
    } //switch

    return last_background;
  }

  /* ==== Internal methods follow ==== */


  void OverlayNode::setPixel(unsigned char *image, int x, int y,
                       unsigned char r, unsigned char g, unsigned char b) {
    image += 4 * (background_width * y + x);
    image[0] = r;
    image[1] = g;
    image[2] = b;
  }

  void OverlayNode::setPixel_y(unsigned char *image, int x, int y,
                       unsigned char v) {
    image += (background_width * y + x);
    image[0] = v;
  }


}

#include "nmm/nmm_suffix.hpp"
