/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "YUVDeInterlaceNode.hpp"
#include "yv12_deinterlace.h"
#include "nmm/comm/format/Property.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"
#include "nmm/comm/messaging/TEDObject.hpp"
#include "nmm/comm/serialize/SerializeResult.hpp"

NMM_REGISTER_PLUGIN(YUVDeInterlaceNode, NMM, "YUVDeInterlaceNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  YUVDeInterlaceNode::YUVDeInterlaceNode(const char *name,
					 const StreamQueue::Mode qmode,
					 const int qsize) :
    GenericFilterNode(name, qmode, qsize)
  {
    width = 0;
    height = 0;
    
    registerEvent("resolution_changed", new TEDObject2<YUVDeInterlaceNode,nmm_int32,nmm_int32>(this, &YUVDeInterlaceNode::resize));
  }

  YUVDeInterlaceNode::~YUVDeInterlaceNode() {}

  Result YUVDeInterlaceNode::resize(nmm_int32& width_, nmm_int32& height_) {
    width = width_;
    height = height_;
    return SUCCESS;
  }
  
  
  Result YUVDeInterlaceNode::doInit() {
    // register general input format
    Format* in_format = getOwnInputProperty()->addNewFormat("video/raw");
    in_format->addWildcard(Format::x_resolution_param);
    in_format->addWildcard(Format::y_resolution_param);
    in_format->addWildcard(Format::ratio_x_param); //
    in_format->addWildcard(Format::ratio_y_param);//
    in_format->addWildcard(Format::framerate_param);//
    in_format->addStringParamValue(Format::channelorder_param,"yvu");
    in_format->addStringParamValue(Format::colorspace_param,"yv12");
    in_format->addStringParamValue(Format::endian_param, "little");
    in_format->addIntParamValue(Format::bitperpixel_param, 12);
    in_format->addStringParamValue(Format::format_param, "planar");
    
    // register general output format
    
    Format* o_format = getOwnOutputProperty()->addNewFormat("video/raw");
    o_format->addWildcard(Format::x_resolution_param);
    o_format->addWildcard(Format::y_resolution_param);
    o_format->addWildcard(Format::ratio_x_param);
    o_format->addWildcard(Format::ratio_y_param);
    o_format->addWildcard(Format::framerate_param);
    o_format->addStringParamValue(Format::channelorder_param,"yvu");
    o_format->addStringParamValue(Format::endian_param, "little");
    o_format->addStringParamValue(Format::colorspace_param,"yv12");
    o_format->addIntParamValue(Format::bitperpixel_param, 12);
    o_format->addStringParamValue(Format::format_param, "planar");

    // set each format as default format
    getOwnOutputProperty()->setDefaultFormat(o_format);
    getOwnInputProperty()->setDefaultFormat(in_format);
    
    return SUCCESS;
  }
  
  Result YUVDeInterlaceNode::doDeinit() {
    getOwnInputProperty()->erase();
    getOwnOutputProperty()->erase();
    return SUCCESS;
  }
  
  Result YUVDeInterlaceNode::doInitOutput() {
    // get input format and make sure we have one
    const Format *in_f = getInputFormat();
    if (in_f == NULL) {
      DEBUG_STREAM("Couldn't get input Format!!!" << endl);
      return FAILURE;
    }
    
    return setupOutputFormat(in_f);
  }
  
  Result YUVDeInterlaceNode::setupOutputFormat(const Format* in_f) {
    // check for required parameters
    if (!in_f->isParameterIncluded(Format::x_resolution_param)) { return FAILURE; }
    if (!in_f->isParameterIncluded(Format::y_resolution_param)) { return FAILURE; }
    width  = in_f->getIntValue(Format::x_resolution_param);
    height = in_f->getIntValue(Format::y_resolution_param);
    
    DEBUG_STREAM("width = " << width << endl);
    DEBUG_STREAM("height = " << height << endl);
    
    if ((width <= 0) || (height <= 0)) {
      DEBUG_STREAM("height or width < 0 !!!" << endl);
      return FAILURE;
    }
    
    // set output format
    // this is equal to input format for all filter nodes
    Format* out_f = new Format(*in_f);
    getOwnWorkingOutputProperty()->addNewFormat(out_f);
    getOwnWorkingOutputProperty()->setDefaultFormat(out_f);
    
    return SUCCESS;
  }
  
  Message* YUVDeInterlaceNode::processBuffer(Buffer *in_buffer) {
    in_buffer = in_buffer->getWriteableInstance();
    
    yv12_deinterlace(in_buffer->getData(), width, height);
    
    return in_buffer;
  }
  
  Result YUVDeInterlaceNode::setFormat(const Format& format) {
    getWorkingInputProperty()->erase();
    getWorkingOutputProperty()->erase();
    
    return setupOutputFormat(&format);
  }
}

#include "nmm/nmm_suffix.hpp"
