/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "JPEGDecodeNode.hpp"
#include "nmm/comm/format/Property.hpp"
#include "nmm/multimedia/Plugin.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

NMM_REGISTER_PLUGIN(NMM::JPEGDecodeNode, NMM, "JPEGDecodeNode")

namespace NMM {

  JPEGDecodeNode::JPEGDecodeNode(const char* name,
				 const StreamQueue::Mode queue_mode,
				 const int queue_size) : 
    GenericProcessorNode(name, queue_mode, queue_size ) 
  {
    myJpegSrcMgr.init_source = jpegdecInitSrc;
    myJpegSrcMgr.fill_input_buffer = jpegdecFillInputBuffer;
    myJpegSrcMgr.skip_input_data = jpegdecSkipInputData;
    myJpegSrcMgr.resync_to_restart = jpegdecResyncToRestart;
    myJpegSrcMgr.term_source = jpegdecTermSrc;

  };

  JPEGDecodeNode::~JPEGDecodeNode() {
  }
  
  Result JPEGDecodeNode::doInit() {
    Format *myInputFormat = getOwnInputProperty()->addNewFormat("video/jpeg");
    myInputFormat->addWildcard(Format::x_resolution_param);
    myInputFormat->addWildcard(Format::y_resolution_param);
    myInputFormat->addWildcard(Format::ratio_x_param);
    myInputFormat->addWildcard(Format::ratio_y_param);
    myInputFormat->addStringParamValue(Format::colorspace_param,"rgb24");
    myInputFormat->addStringParamValue(Format::channelorder_param,"bgr");
    myInputFormat->addIntParamValue(Format::bitperpixel_param,24);
    myInputFormat->addStringParamValue(Format::endian_param, "little");
    myInputFormat->addWildcard(Format::framerate_param);
    myInputFormat->addStringParamValue(Format::format_param, "packet");
    getOwnInputProperty()->setDefaultFormat(myInputFormat);
    
    NMM::Format *myOutputFormat = getOwnOutputProperty()->addNewFormat("video/raw");
    myOutputFormat->addWildcard(Format::x_resolution_param);
    myOutputFormat->addWildcard(Format::y_resolution_param);
    myOutputFormat->addWildcard(Format::ratio_x_param);
    myOutputFormat->addWildcard(Format::ratio_y_param);
    myOutputFormat->addStringParamValue(Format::colorspace_param,"rgb24");
    myOutputFormat->addStringParamValue(Format::channelorder_param,"bgr");
    myOutputFormat->addIntParamValue(Format::bitperpixel_param,24);
    myOutputFormat->addStringParamValue(Format::endian_param, "little");
    myOutputFormat->addWildcard(Format::framerate_param);
    myOutputFormat->addStringParamValue(Format::format_param, "packet");
    getOwnOutputProperty()->setDefaultFormat(myOutputFormat);
    
    myInputFormat->setIOPartner(myOutputFormat);
    myOutputFormat->setIOPartner(myInputFormat);
    
    return SUCCESS;
  }
  
  Result JPEGDecodeNode::doInitOutput() {
    const Format* in_format = getInputFormat();
    if (in_format == NULL) { return FAILURE; }
    
    int x_resolution = in_format->getIntValue(Format::x_resolution_param);
    int y_resolution = in_format->getIntValue(Format::y_resolution_param);
    float framerate = in_format->getFloatValue(Format::framerate_param);
    
    Format* myInputFormat = new Format( *in_format );
    getOwnWorkingInputProperty()->addNewFormat(myInputFormat);
    getOwnWorkingInputProperty()->setDefaultFormat(myInputFormat);
    
    Format *myOutputFormat = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
    myOutputFormat->addIntParamValue(Format::x_resolution_param, x_resolution);
    myOutputFormat->addIntParamValue(Format::y_resolution_param, y_resolution);
    myOutputFormat->addIntParamValue(Format::ratio_x_param, in_format->getIntValue(Format::ratio_x_param));
    myOutputFormat->addIntParamValue(Format::ratio_y_param, in_format->getIntValue(Format::ratio_y_param));
    myOutputFormat->addStringParamValue(Format::colorspace_param,"rgb24");
    myOutputFormat->addStringParamValue(Format::channelorder_param,"bgr");
    myOutputFormat->addIntParamValue(Format::bitperpixel_param,24);
    myOutputFormat->addStringParamValue(Format::endian_param, "little");
    myOutputFormat->addFloatParamValue(Format::framerate_param, framerate);
    myOutputFormat->addStringParamValue(Format::format_param, "packet");
    getOwnWorkingOutputProperty()->setDefaultFormat(myOutputFormat);
    
    myInputFormat->setIOPartner(myOutputFormat);
    myOutputFormat->setIOPartner(myInputFormat);
    
    return SUCCESS;
  }
  
  Result JPEGDecodeNode::doActivate() {
    const Format *in_f = getInputFormat();
    
    image_width = in_f->getIntValue(Format::x_resolution_param);
    image_height = in_f->getIntValue(Format::y_resolution_param);

    if ((image_width <= 0) || (image_height <= 0)) {
      ERROR_STREAM( "image size is wrong: " << endl);
      return FAILURE; 
    }

    return SUCCESS; 
  }
  
  Buffer* JPEGDecodeNode::processBuffer(Buffer *in) {
    if (isStarted()) {
	
      Buffer *out = getNewBuffer(image_width * image_height * 3, in);
	
      if (out) {
	out->setUsed(readFromBuffer((JSAMPLE *) in->getData(), in->getUsed(), (JSAMPLE *) out->getData()));
      }
      else {
	DEBUG_STREAM( "JPEGDecodeNode: Could not allocate out Buffer!" << endl);
      }
      in->release();
      return out;

    }
    return 0;
  }
  
  long JPEGDecodeNode::readFromBuffer( JSAMPLE * image_buffer, long image_size, JSAMPLE *dest_buffer) {
    struct jpeg_decompress_struct cinfo;
    struct jpeg_error_mgr jerr;// struct my_error_mgr jerr;
    /* More stuff */
    JSAMPARRAY buffer;		/* Output row buffer */
    int row_stride;		/* physical row width in output buffer */

    /* Step 1: allocate and initialize JPEG decompression object */
    cinfo.err = jpeg_std_error(&jerr);
 
    /* Now we can initialize the JPEG decompression object. */
    jpeg_create_decompress(&cinfo);

    myJpegSrcMgr.next_input_byte = image_buffer;
    myJpegSrcMgr.bytes_in_buffer = image_size;
    cinfo.src = &myJpegSrcMgr;

    /* Step 2: specify data source (eg, a file) */

    /* Step 3: read file parameters with jpeg_read_header() */

    (void) jpeg_read_header(&cinfo, TRUE);

    /* Step 4: set parameters for decompression */
    cinfo.out_color_space = JCS_RGB; // JCS_YCbCr;

    /* Step 5: Start decompressor */

    (void) jpeg_start_decompress(&cinfo);
    row_stride = cinfo.output_width * cinfo.output_components;
    /* Make a one-row-high sample array that will go away when done with image */
    buffer = (*cinfo.mem->alloc_sarray)
      ((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);

    /* Step 6: while (scan lines remain to be read) */
    /*           jpeg_read_scanlines(...); */

    while (cinfo.output_scanline < cinfo.output_height) {
      (void) jpeg_read_scanlines(&cinfo, buffer, 1);
      memcpy(dest_buffer, buffer[0], row_stride);
      dest_buffer+=row_stride;
    }

    /* Step 7: Finish decompression */
    (void) jpeg_finish_decompress(&cinfo);
    /* Step 8: Release JPEG decompression object */
    jpeg_destroy_decompress(&cinfo);
    return cinfo.output_height * cinfo.output_width * 3;
  }
}
