/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "JPEGEncodeNode.hpp"
#include "nmm/comm/format/Property.hpp"
#include "nmm/multimedia/Plugin.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

NMM_REGISTER_PLUGIN(NMM::JPEGEncodeNode, NMM, "JPEGEncodeNode")

namespace NMM {

  JPEGEncodeNode::JPEGEncodeNode(const char* name,
				 const StreamQueue::Mode queue_mode,
				 const int queue_size) : 
    GenericProcessorNode(name, queue_mode, queue_size )
  {
   
    benchmark.setEnabled(1);
  };

  Result JPEGEncodeNode::doInit() {
    Format *myOutputFormat =
      getOwnOutputProperty()->addNewFormat("video/jpeg");
    myOutputFormat->addWildcard(Format::x_resolution_param);
    myOutputFormat->addWildcard(Format::y_resolution_param);
    myOutputFormat->addWildcard(Format::ratio_x_param);
    myOutputFormat->addWildcard(Format::ratio_y_param);
    myOutputFormat->addStringParamValue(Format::colorspace_param,"rgb24");
    myOutputFormat->addStringParamValue(Format::channelorder_param,"bgr");
    myOutputFormat->addIntParamValue(Format::bitperpixel_param,24);
    myOutputFormat->addStringParamValue(Format::endian_param, "little");
    myOutputFormat->addStringParamValue(Format::format_param, "packet");
    myOutputFormat->addWildcard(Format::framerate_param);

    Format *myInputFormat =
      getOwnInputProperty()->addNewFormat("video/raw");
    myInputFormat->addWildcard(Format::x_resolution_param);
    myInputFormat->addWildcard(Format::y_resolution_param);
    myInputFormat->addWildcard(Format::ratio_x_param);
    myInputFormat->addWildcard(Format::ratio_y_param);
    myInputFormat->addStringParamValue(Format::colorspace_param,"rgb24");
    myInputFormat->addStringParamValue(Format::channelorder_param,"bgr");
    myInputFormat->addIntParamValue(Format::bitperpixel_param,24);
    myInputFormat->addStringParamValue(Format::endian_param, "little");
    myInputFormat->addWildcard(Format::framerate_param);
    myInputFormat->addStringParamValue(Format::format_param, "packet");
    
    myInputFormat->setIOPartner(myOutputFormat);
    myOutputFormat->setIOPartner(myInputFormat);

    getOwnOutputProperty()->setDefaultFormat(myOutputFormat);
    getOwnInputProperty()->setDefaultFormat(myInputFormat);

    quality = 40;

    myJpegDestMgr.init_destination = jpegencInitDest;
    myJpegDestMgr.empty_output_buffer = jpegencFlushDest;
    myJpegDestMgr.term_destination = jpegencTermDest;

    return SUCCESS;
  }

  Result JPEGEncodeNode::doInitOutput() {
    const Format *in_f = getInputFormat();

    image_width = in_f->getIntValue(Format::x_resolution_param);
    image_height = in_f->getIntValue(Format::y_resolution_param);
    
    float framerate = in_f->getFloatValue(Format::framerate_param);
    
    if ((image_width <= 0) || (image_height <= 0)) {
      ERROR_STREAM( "image size is wrong: " << endl);
      return FAILURE; 
    }

    Format *myOutputFormat =
      getOwnWorkingOutputProperty()->addNewFormat("video/jpeg");
    myOutputFormat->addIntParamValue(Format::x_resolution_param,image_width);
    myOutputFormat->addIntParamValue(Format::y_resolution_param,image_height);
    myOutputFormat->addIntParamValue(Format::ratio_x_param, in_f->getIntValue(Format::ratio_x_param));
    myOutputFormat->addIntParamValue(Format::ratio_y_param, in_f->getIntValue(Format::ratio_y_param));
    myOutputFormat->addStringParamValue(Format::colorspace_param,"rgb24");
    myOutputFormat->addStringParamValue(Format::channelorder_param,"bgr");
    myOutputFormat->addIntParamValue(Format::bitperpixel_param,24);
    myOutputFormat->addStringParamValue(Format::endian_param, "little");
    myOutputFormat->addFloatParamValue(Format::framerate_param, framerate);
    myOutputFormat->addStringParamValue(Format::format_param, "packet");

    getOwnWorkingOutputProperty()->setDefaultFormat(myOutputFormat);

    Format *myInputFormat =
      getOwnWorkingInputProperty()->addNewFormat("video/raw");
    myInputFormat->addIntParamValue(Format::x_resolution_param,image_width);
    myInputFormat->addIntParamValue(Format::y_resolution_param,image_height);
    myInputFormat->addIntParamValue(Format::ratio_x_param, in_f->getIntValue(Format::ratio_x_param));
    myInputFormat->addIntParamValue(Format::ratio_y_param, in_f->getIntValue(Format::ratio_y_param));
    myInputFormat->addStringParamValue(Format::colorspace_param,"rgb24");
    myInputFormat->addStringParamValue(Format::channelorder_param,"bgr");
    myInputFormat->addIntParamValue(Format::bitperpixel_param,24);
    myInputFormat->addStringParamValue(Format::endian_param, "little");
    myInputFormat->addFloatParamValue(Format::framerate_param, framerate);
    myInputFormat->addStringParamValue(Format::format_param, "packet");
    
    getOwnWorkingInputProperty()->setDefaultFormat(myInputFormat);
    myInputFormat->setIOPartner(myOutputFormat);
    myOutputFormat->setIOPartner(myInputFormat);
    return SUCCESS;
  }

  JPEGEncodeNode::~JPEGEncodeNode() 
  {
  }
 
  Buffer *JPEGEncodeNode::processBuffer(Buffer *in) 
  {
    if (isStarted()) {
	
      Buffer *out = getNewBuffer(image_width * image_height * 3, in);
	
      if (out) {
	out->setUsed(writeToBuffer(quality, (JSAMPLE *) in->getData(), image_width, image_height, (JSAMPLE *) out->getData()) );
	benchmark.benchmark(out->getUsed());
      }
      else {
	DEBUG_STREAM( "JPEGEncodeNode: Could not allocate out Buffer!" << endl);
      }
      in->release();
      return out;

    }
    return 0;
  } 

  void JPEGEncodeNode::setQuality(const nmm_float& q) {
    /// XXXXX Quality is set as an float - dunno if the JPEGLibrary likes the float XXXXX /////
    if (q > 100) quality = 100;
    else quality = q;
  }

  nmm_float JPEGEncodeNode::getQuality() {
    return quality;
  }

  void JPEGEncodeNode::setBitrate(const nmm_int32&) {}

  nmm_int32 JPEGEncodeNode::getBitrate() {
    return -1;
  }

  long JPEGEncodeNode::writeToBuffer(int quality, JSAMPLE * image_buffer, int image_width, int image_height, JSAMPLE *dest_buffer )
  {
    struct jpeg_compress_struct cinfo;
    struct jpeg_error_mgr jerr;
    /* More stuff */
    // FILE * outfile;		/* target file */
    JSAMPROW row_pointer[1];	/* pointer to JSAMPLE row[s] */
    int row_stride;		/* physical row width in image buffer */

    /* Step 1: allocate and initialize JPEG compression object */
    cinfo.err = jpeg_std_error(&jerr);
    /* Now we can initialize the JPEG compression object. */
    jpeg_create_compress(&cinfo);
  
    myJpegDestMgr.next_output_byte = dest_buffer;
    myJpegDestMgr.free_in_buffer   = image_width * image_height * 3;
    cinfo.dest = &myJpegDestMgr;

    /* Step 2: specify data destination (eg, a file) */
    /* Note: steps 2 and 3 can be done in either order. */

    /* Step 3: set parameters for compression */
    cinfo.image_width = image_width; 	/* image width and height, in pixels */
    cinfo.image_height = image_height;
    cinfo.input_components = 3; // 4; 		/* # of color components per pixel */
    cinfo.in_color_space =  JCS_RGB; //JCS_YCbCr; /* colorspace of input image */
    jpeg_set_defaults(&cinfo);
    jpeg_set_quality(&cinfo, quality, TRUE /* limit to baseline-JPEG values */);

    /* Step 4: Start compressor */
    jpeg_start_compress(&cinfo, TRUE);

    /* Step 5: while (scan lines remain to be written) */
    /*           jpeg_write_scanlines(...); */
    row_stride = image_width * 3;	/* JSAMPLEs per row in image_buffer */

    while (cinfo.next_scanline < cinfo.image_height) {
      /* jpeg_write_scanlines expects an array of pointers to scanlines.
       * Here the array is only one element long, but you could pass
       * more than one scanline at a time if that's more convenient.
       */
      row_pointer[0] = & image_buffer[cinfo.next_scanline * row_stride];
      (void) jpeg_write_scanlines(&cinfo, row_pointer, 1);
    }

    /* Step 6: Finish compression */

    jpeg_finish_compress(&cinfo);

    /* Step 7: release JPEG compression object */
    jpeg_destroy_compress(&cinfo);

    // return used bytes in dest_buffer, this value should be much less than image_width * image_height * 3 
    return ( (image_width * image_height * 3) - myJpegDestMgr.free_in_buffer);
  }


}
