/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
/*
 * Code is based on decode.c that is part of mpeg2dec, a free MPEG-2 video stream decoder.
 * Copyrighted 1999-2001 by Aaron Holtzman <aholtzma@ess.engr.uvic.ca>
 */


#include "MPEGVideoDecodeNode.hpp"

#include "nmm/plugins/NMMRegisterPlugin.hpp"
#include "nmm/utils/thread/MutexGuard.hpp"
#include "nmm/utils/time/Interval.hpp"
#include "nmm/multimedia/interfaces/ITrackDuration.hpp"

//#include "nmm/base/format/Property.hpp"
//#include "nmm/interfaces/base/IFormatChange.hpp"
//#include "nmm/base/memorymgr/Buffer.hpp"
//#include "nmm/base/registry/Plugin.hpp"
//#include "nmm/multimedia/interfaces/ITrackDuration.hpp"

//#ifdef HAVE_CONFIG_H
//#include "config.h"
//#endif

//#include <iostream>
#ifdef DELETE
#undef DELETE //DELETE is defined in some windows headers
#endif

NMM_REGISTER_PLUGIN(MPEGVideoDecodeNode, NMM, "MPEGVideoDecodeNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  MPEGVideoDecodeNode::MPEGVideoDecodeNode( const char * name, StreamQueue::Mode qmode, int qsize)
    : GenericProcessorNode(name, qmode, qsize),
      in_buffer_timestamp_taken(false),
      in_buffer_sync_time(0),
      frame_type(PIC_FLAG_CODING_TYPE_I),
      timestamp(),
      decoded_frames(0),
      mpeg2dec(NULL),
      out(NULL),
      buffer_size(0),
      old_in_buffer(NULL),
      image_height(0),
      image_width(0),
      aspect_ratio(0),
      framerate(0.0f),
      format_changed(false),
      i_frames_only(0),
      mpeg_resync(true), // make sure, playback starts without any broken frames
      drop_b_frames(true),
      frames_to_drop(0),
      iframes_to_drop(0),
      process_buffer_mutex(),
			first_iframe(0)
  {
  }


  MPEGVideoDecodeNode::~MPEGVideoDecodeNode()
  {
  }

  Result MPEGVideoDecodeNode::doInit() {
    DEBUG_STREAM("Initializing mpeg video decoder..." << endl);
    format_changed           = false;
    
    setWorkingFlag(false);
    
    in_buffer_timestamp_taken = false;    // No in_buffer timestamp taken yet.
    
    mpeg2dec = mpeg2_init();
    
    syncReset();
    
    // input formats
    Format* i_format;
    
    // specific input format
    i_format = getOwnInputProperty()->addNewFormat("video/mpeg");
    i_format->addIntParamValue(Format::version_param, 1);
    i_format->addIntParamValue(Format::version_param, 2);
    i_format->addWildcard(Format::x_resolution_param);
    i_format->addWildcard(Format::y_resolution_param);
    i_format->addWildcard(Format::ratio_x_param);
    i_format->addWildcard(Format::ratio_y_param);
    i_format->addWildcard(Format::framerate_param);
    
    // general input foramt
    i_format = getOwnInputProperty()->addNewFormat("video/mpeg");
    i_format->addIntParamValue(Format::version_param, 1);
    i_format->addIntParamValue(Format::version_param, 2);
    getOwnInputProperty()->setDefaultFormat(i_format);
    
    // general output format
    Format* o_format = getOwnOutputProperty()->addNewFormat("video/raw");
    o_format->addWildcard(Format::x_resolution_param);
    o_format->addWildcard(Format::y_resolution_param);
    o_format->addWildcard(Format::ratio_x_param);
    o_format->addWildcard(Format::ratio_y_param);
    o_format->addWildcard(Format::framerate_param);
    o_format->addStringParamValue(Format::channelorder_param,"yvu");
    o_format->addStringParamValue(Format::endian_param, "little");
    o_format->addStringParamValue(Format::colorspace_param,"yv12");
    o_format->addIntParamValue(Format::bitperpixel_param,12);
    o_format->addStringParamValue(Format::format_param, "planar");
    getOwnOutputProperty()->setDefaultFormat(o_format);
    
    // set I/O partners
    // XXX: 
    i_format->setIOPartner(o_format);
    o_format->setIOPartner(i_format);
    
    return SUCCESS;
  }
  
  Result MPEGVideoDecodeNode::doInitOutput() {
    // get input format
    // make sure input format exists
    const Format* in_format = getInputFormat();

    if (in_format == NULL) { 
      ERROR_STREAM("Input Format is NULL!" << endl);
      return FAILURE; 
    }
    
    // add input format to working input property
    // this is nothing else but the specified input format
    Format* i_format = new Format( *in_format );
    getOwnWorkingInputProperty()->addNewFormat(i_format);
    getOwnWorkingInputProperty()->setDefaultFormat(i_format);
    
    // get parameters
    if (in_format->isParameterIncluded(Format::x_resolution_param) &&
        in_format->isParameterIncluded(Format::y_resolution_param) &&
        in_format->isParameterIncluded(Format::ratio_x_param) &&
        in_format->isParameterIncluded(Format::ratio_y_param) &&
        in_format->isParameterIncluded(Format::framerate_param)) {
      // good: all needed params specified, no need to analyse data
      image_width = in_format->getIntValue(Format::x_resolution_param);
      image_height = in_format->getIntValue(Format::y_resolution_param);
      aspect_ratio = Rational(in_format->getIntValue(Format::ratio_x_param),
                              in_format->getIntValue(Format::ratio_y_param));
      aspect_ratio.reduce(); // aspect should always be stored reduced
      framerate = in_format->getFloatValue(Format::framerate_param);
      setWOPInternal();
    } else {
      // ok, we have to request and analyse some data
      setMaxAnalysisIterations(250);
      if (!analyseData()) {
	WARNING_STREAM("analyseData() failed! Setting 720X576 as resolution, 16:9 as aspect, 25.0 as framerate as workaround!" << endl);
        image_width = 720;
        image_height = 576;
        aspect_ratio = Rational(16, 9);
        aspect_ratio.reduce(); // aspect should always be stored reduced
        framerate = 25.0;
        setWOPInternal();
      }
    }
    
    // set I/O partners
    const_cast<Format*>(getOwnWorkingInputProperty()->
			getDefaultFormat())->
      setIOPartner(const_cast<Format*>(getOwnWorkingOutputProperty()->
				       getDefaultFormat()));
    
    const_cast<Format*>(getOwnWorkingOutputProperty()->
			getDefaultFormat())->
      setIOPartner(const_cast<Format*>(getOwnWorkingInputProperty()->
				       getDefaultFormat()));
    

    return SUCCESS;
  }

  Result MPEGVideoDecodeNode::doDeinit() {
    DEBUG_STREAM("exiting mpeg video decoder..." << endl);
    
    mpeg2_close(mpeg2dec);
    
    return SUCCESS;
  }
  
  void MPEGVideoDecodeNode::setWOPInternal() {
    // clear output property
    getOwnWorkingOutputProperty()->erase();
    
    // init working output property
    Format* o_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
    o_format->addIntParamValue(Format::x_resolution_param, image_width);
    o_format->addIntParamValue(Format::y_resolution_param, image_height);
    o_format->addIntParamValue(Format::ratio_x_param, aspect_ratio.getNumerator());
    o_format->addIntParamValue(Format::ratio_y_param, aspect_ratio.getDenominator());
    o_format->addFloatParamValue(Format::framerate_param, framerate);
    o_format->addStringParamValue(Format::channelorder_param,"yvu");
    o_format->addStringParamValue(Format::endian_param, "little");
    o_format->addStringParamValue(Format::colorspace_param,"yv12");
    o_format->addIntParamValue(Format::bitperpixel_param,12);
    o_format->addStringParamValue(Format::format_param, "planar");
    getOwnWorkingOutputProperty()->setDefaultFormat(o_format);
  }

  void MPEGVideoDecodeNode::copyBytes(char* dest, char* src, unsigned int nbytes) {
    unsigned int remainder = nbytes % 2048;
    if (nbytes != remainder) memcpy(dest, src, nbytes - remainder);
    if (remainder) memcpy(dest + nbytes - remainder, src + nbytes - remainder, remainder);
  }

  bool MPEGVideoDecodeNode::decodeMPEG2() {
    // resync
    bool is_frame_done = false;
    out = 0;
    
    const mpeg2_info_t* info = mpeg2_info(mpeg2dec);
    while (!is_frame_done) {
      int state = mpeg2_parse(mpeg2dec);
      switch (state) {
      case 0:
        // this case is for later versions of libmpeg2:
        // in the CVS version, mpeg2_parse returns no longer int but a
        // state-enum. The value for requesting more data is
        // "STATE_BUFFER", which is mapped to 0. As the enum nor
        // STATE_BUFFER is defined in earlier versions of mpeg2.h, we
        // use the mapped integer value here.

        // fallthrough
      case -1:
        return false;
        break;

      case STATE_SEQUENCE:
        {
          // calculate new aspect ratio
          Rational new_ratio(info->sequence->width * info->sequence->pixel_width, 
                             info->sequence->height * info->sequence->pixel_height);
          new_ratio.reduce(); // don't forget to reduce!
          
          if ((image_width  != info->sequence->width) ||
              (image_height != info->sequence->height) ||
              (aspect_ratio != new_ratio)) {
            
            image_width     = info->sequence->width;
            image_height    = info->sequence->height;

            aspect_ratio = new_ratio;

            // calculate the new frame rate and wait delay
            int frame_rate_code = info->sequence->frame_period;
            framerate = 1000000.0f / static_cast<float>(frame_rate_code/27);

            // set flag
            format_changed = true;

						buffer_size =     info->sequence->width        * info->sequence->height         + 
						              2 * info->sequence->chroma_width * info->sequence->chroma_height;
          }
        }
        break;
        
      case STATE_SEQUENCE_REPEATED:
        // nothing to do here!
        break;

      case STATE_GOP:
        // nothing to do here!
        break;

      case STATE_PICTURE:
        // this (mpeg2dec->decoder.coding_type) is the type of the
        // next frame to be read by the lib, not in all cases the next
        // frame to be decoded by the lib (do not forget the B-Frmaes
        // with temporal references in both directions!)
        if (i_frames_only) {
          if ((info->current_picture->flags & PIC_FLAG_CODING_TYPE_I) == 0) {
            mpeg2_skip(mpeg2dec,1);
	  } else {
            mpeg2_skip(mpeg2dec,0);
          }
        }
        if (mpeg_resync) {
          // if we have to resync, wait for the first incoming
          // I-Frame, as this is the first frame we can savely display
          // to be resynced!
          if (info->current_picture->flags & PIC_FLAG_CODING_TYPE_I) {
						// remember the picture pointer of this I-Frame
            first_iframe = info->current_picture;
          }
        }
        break;

      case STATE_SLICE_1ST:
        // nothing to do here!
        break;

      case STATE_PICTURE_2ND:
        /* should not do anything */
        break;

      case STATE_SLICE:
        // fallthrough
      case STATE_END:
        /* draw current picture */
        if (info->display_fbuf) {
	  frame_type = info->display_picture->flags & PIC_MASK_CODING_TYPE;        
          if (mpeg_resync) {
            // if we have to resync, we have to wait until the decoder
            // gives us the decoded first I-Frame that came in after
            // starting to resync (see case STATE_PICTURE!)
            if (info->display_picture == first_iframe) {
              // now we are synced!
              mpeg_resync = false;
              // forget the remembered I-Frame
              first_iframe = NULL;
            } else {
              // not resynced yet, not taking this decoded frame.
              break;
            }
          }
          
					int width  = info->sequence->width;
					int height = info->sequence->height;
					int chroma_width  = info->sequence->chroma_width;
					int chroma_height = info->sequence->chroma_height;
					
										
          if (!(i_frames_only && frame_type != PIC_FLAG_CODING_TYPE_I)) {
            out = getNewBuffer( buffer_size );
            out->setUsed( buffer_size );
						// U and V plane must be swapped and each plane *must* be copied seperately. They're not in a consecutive memory region!
            copyBytes(out->getData(), 															 								  reinterpret_cast<char*>(info->display_fbuf->buf[0]), width * height);
						copyBytes(out->getData() + width * height, 						 									  reinterpret_cast<char*>(info->display_fbuf->buf[2]), chroma_width * chroma_height);
						copyBytes(out->getData() + width * height + chroma_width * chroma_height, reinterpret_cast<char*>(info->display_fbuf->buf[1]), chroma_width * chroma_height);
          }
          
          /* increase for benchmarking */
          decoded_frames++;

          // the realy Frame-Type can only be looked up here. Storing
          // the Frame-Type of the incoming (encoded) frames is wrong
          // as they are not in the same temporal order as the decoded
          // ones:
          //
          // encoded stream:
          //   201534...
          //   IBBPBB...
          //
          // decoded stream:
          //   012345...
          //   BBIBBP...
          // 
          // (B-Frames have predictions in BOTH directions and so are
          // delivered AFTER the predicted frames!)
          //frame_type = info->display_picture->flags & PIC_MASK_CODING_TYPE;
          
          is_frame_done = true;
        }
        break;

      case STATE_INVALID:
        ERROR_STREAM("STATE_INVALID" << endl);
        // nothing to do here!
        break;
      } //end switch
    } // end while
    
    return true;
  }
  
  
  Message* MPEGVideoDecodeNode::processBuffer( Buffer *in_buffer ) {
    MutexGuard process_guard(process_buffer_mutex);

    if (!getWorkingFlag()) {
      uint8_t* current;
      uint8_t* end;

      current = reinterpret_cast<uint8_t*>(in_buffer->getData());
      end     = current + static_cast<int>(in_buffer->getUsed());

      mpeg2_buffer(mpeg2dec, current, end);

      setWorkingFlag( true );
      old_in_buffer = in_buffer;
      
      Timestamp in_buf_stamp = in_buffer->getTimestamp();      // If the incoming buffer's
      if(in_buf_stamp.is_valid) {                              // timstamp is valid
      	in_buffer_sync_time = in_buf_stamp.sync_time;          // take it
	in_buffer_timestamp_taken = true;                      // and remember that!
      }
    }

    if (format_changed) {      // if format (e.g. resolution) changed
      format_changed = false;  // send corresponding event

      // setup new output format
      setWOPInternal();
    
      // send the new format
      if (getAnalysingFlag()) {
        setAnalysingFlag(false);
	
				if (send_track_duration) {
		  		u_int64_t track_length = track_size / mpeg2_info(mpeg2dec)->sequence->byte_rate;
	  			Interval track_duration = Interval(track_length, 0);
	 		  	return new CEvent(ITrackDuration::create_trackDuration(track_duration));
				}
				return NULL;
				
      } else {
        DEBUG_STREAM("SetFormat: " << endl << *getOwnWorkingOutputProperty()->getDefaultFormat() << endl);
        return new CEvent(IFormatChange::create_setFormat(*getOwnWorkingOutputProperty()->getDefaultFormat()));
      }
    }

    /* try to decode a frame */
    if(decodeMPEG2()) {
      /* if we have a complete frame,
       *  put a timestamp to it and send it
       */
      
      DEBUG_STREAM("frametype: " << frame_type << endl);

      if(in_buffer_timestamp_taken && (frame_type == PIC_FLAG_CODING_TYPE_I)) {  // If there's an I-Frame and
	                                                         // a timestamp from an
	                                                         // incoming buffer
				timestamp.sync_time = in_buffer_sync_time;         // then: take it
				timestamp.is_valid = true;                             // and set to valid since only valid timestamps were considered
				in_buffer_timestamp_taken = false;                 //       and make sure that it is
	                                                         //       taken only once
				// cerr << "\nTAKE\n" << endl;
      } else {
	timestamp.sync_time += Interval(Rational(10000, static_cast<int>(framerate * 10000)));
	// cerr << "\nNOT TAKE taken " << in_buffer_timestamp_taken << " valid " << timestamp.is_valid << endl << endl;
        // otherwise add the duration of one frame to the last outgoing buffer's timestamp.
        // this is 1/framerate. To avoid rounding error's we use a Rational with
        // 10000 / (framerate*10000).
      }
      
			if (out) {
        out->setTimestamp(timestamp);
			}
      timestamp.stream_counter++;               // don't forget to count the buffer.
      
    } else {
      old_in_buffer->release();
      old_in_buffer = 0;
      setWorkingFlag( false );
      return( NULL );
    }

    /* drop frames */
    if ( frames_to_drop ) {
      frames_to_drop--;
      if (out) {
        out->release();
        out = 0;
      }
      return 0;
    }

    if (i_frames_only && (frame_type != PIC_FLAG_CODING_TYPE_I)) {
      if (out) {
        out->release();
        out = 0;
      }
      return 0;
    }
		
    Buffer* returned_buffer = out;
    out = 0;

    if(returned_buffer->getTimestamp().is_valid == false) {
      returned_buffer->release();
      return 0;
    }
    else {
      return( returned_buffer );
    }
  }

  Result MPEGVideoDecodeNode::doFlush() {
    setWorkingFlag(false);
    // we are flushed! Now let's resync on new stream again!
    resyncMPEG();

    return SUCCESS;
  }
  
  Result MPEGVideoDecodeNode::setStillFrame(const nmm_bool& ) {
    MutexGuard process_guard(process_buffer_mutex);
    
    // fw: if you uncomment this code, it won't work anymore because of changed setStillFrame semantics
    /*if ( send_current_frame_ok ) {
      saved_working_flag = getWorkingFlag();
      setWorkingFlag( true );
      send_current_frame_ok = false;
      send_current_frame    = true;
      }*/
    
    return SUCCESS;
  }

  Result MPEGVideoDecodeNode::syncReset() {
    MutexGuard process_guard(process_buffer_mutex);
    
    timestamp.stream_counter = 0;
    timestamp.sync_time = Time(0);
    timestamp.is_valid = false;

    in_buffer_timestamp_taken = false;

    DEBUG_STREAM("sync_reset completed." << endl);
    return SUCCESS;
  };

  Result MPEGVideoDecodeNode::sendIFramesOnly(const nmm_int32& trigger) {
    MutexGuard process_guard(process_buffer_mutex);

    DEBUG_STREAM("event i_frames_only: "<< trigger << endl);
    if (trigger == 0) {
      i_frames_only   = 0;
      iframes_to_drop = 0;
      mpeg2_skip(mpeg2dec,0);
      DEBUG_STREAM("normal processing mode" << endl);
      return SUCCESS;
    }

    i_frames_only   = trigger;
    iframes_to_drop = trigger;
    DEBUG_STREAM("sending I-frames only" << endl);

    return SUCCESS;
  };

  Result MPEGVideoDecodeNode::resyncMPEG() {
    MutexGuard process_guard(process_buffer_mutex);

    DEBUG_STREAM("event resync_mpeg_decoder" << endl);
    // enabling resyncing:
    // 1. wait for the first I-Frame coming in and remember, which
    //    picture it is stored as.
    // 2. Drop all decoded frames, whose picture-pointer is not the
    //    remembered I-Frame picture.
    // 3. Continue normaly with founf I-Frame.
    //
    // The libmpeg2 cannot be flushed for resyncing (at least I do not
    // have found a way to do so!). To avoid broken frames from the
    // internal caches of the lib, we wait for the first I-Frame
    // coming in as this is the first frame, that can be decoded
    // properly after resyncing.

    // When we found it, we have to remember and follow it through the
    // decoder as this might not be the first frame to be decoded
    // afterwards (remember: B-Frames after this I-Frame have
    // predictions to this I-Frame and the last non-B-Frame before
    // this I-Frame and consequently might be broken as we might not
    // have got this second Frame. As these B-Frames are temporal
    // ordered BEFORe the found I-Frame, the decoder will give us them
    // first, before the I-Frame.)

    // After finding the searched I-Frame in the decoded frames, we
    // can send it and all following frames and we are synced again!
    mpeg_resync = true;
    return SUCCESS;
  }

  Result MPEGVideoDecodeNode::fastForward( const nmm_int32& speed ) {
    /* stop itself */
    try{ stop(); }
    catch(...){ DEBUG_STREAM("Node already stopped" << endl); };

    flush();

    if ( speed > 0 ) {
      DEBUG_STREAM("fast forward event: on" << endl);
      
      sendIFramesOnly( speed );
      
    } else {
      DEBUG_STREAM("fast forward event: off" << endl);

      /* speed=0: decode all frames */
      sendIFramesOnly( speed );

      /* skip a few frames to avoid displaying crunched pictures */
      frames_to_drop = 6;
    }

    /* send upstream event to predecessor */
    // removed sendSet|GetEvent (was implemented as return FAILURE; anyway!)
    // if ( predecessor ) predecessor->sendSetEvent(Event("fast_forward", new TInValue<nmm_int32>( speed )) );

    try{ start(); }
    catch(...){ DEBUG_STREAM("Node already started" << endl); };

    return( SUCCESS );
  }

  Result MPEGVideoDecodeNode::streamJump( const nmm_int32& speed ) {
    DEBUG_STREAM("stream jump event, speed = " << speed << endl);

    /* stop itself */
    try{ stop(); }
    catch(...){ DEBUG_STREAM("Node already stopped" << endl); };

    flush();

    /* skip a few frames to avoid displaying crunched pictures */
    //frames_to_drop = 6;


    /* send upstream event to predecessor */
    // removed sendSet|GetEvent (was implemented as return FAILURE; anyway!)
    // if ( predecessor ) predecessor->sendSetEvent(Event("stream_jump", new TInValue<nmm_int32>( speed )) );

    try{ start(); }
    catch(...){ DEBUG_STREAM("Node already started" << endl); };

    return( SUCCESS );
  }

  Result MPEGVideoDecodeNode::trackSize(const nmm_uint64& bytes) {
    track_size = bytes;
    send_track_duration = true;
	return DELETE;
  }
	
}



#include "nmm/nmm_suffix.hpp"
