/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <pthread.h>
#include "VideoMuxNode.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"
#include "nmm/utils/thread/MutexGuard.hpp"
#include <math.h>

NMM_REGISTER_PLUGIN(VideoMuxNode, NMM, "VideoMuxNode")
    
#include "nmm/nmm_prefix.hpp"

namespace NMM {
	
  VideoMuxNode::VideoMuxNode(const char *name, StreamQueue::Mode qmode, int qsize) 
    : GenericMultiplexerNode(name, qmode, qsize)
    , in_buffer1(0)
    , in_buffer2(0)
    , counter(0)
    , input1_width(0)
    , input1_height(0)
    , input1_ratio_x(0)
    , input1_ratio_y(0)
    , input2_width(0)
    , input2_height(0)
    , input2_ratio_x(0)
    , input2_ratio_y(0)
    , width1(0)
    , height1(0)
    , width2(0)
    , height2(0)
    , x_pos1(0)
    , y_pos1(0)
    , x_pos2(0)
    , y_pos2(0)
    , alignment1(IVideoMux::NONE)
    , alignment2(IVideoMux::NONE)
    , x_alignment1(0)
    , y_alignment1(0)
    , x_alignment2(0)
    , y_alignment2(0)
    , output_width(0)
    , output_height(0)
    , width_height_coupling1(true)
    , width_height_coupling2(true)
  {
    addInputStream("input1");
    addInputStream("input2");
  }
	
  VideoMuxNode::~VideoMuxNode() 
  {
  }

  Result VideoMuxNode::doInit()
  {
    // add new input format for input1 to property output format
    Format* in_format1 = getOwnInputProperty()->addNewFormat("video/raw");
    in_format1->setJackTag("input1");

    in_format1->addWildcard(Format::x_resolution_param);
    in_format1->addWildcard(Format::y_resolution_param);
    in_format1->addWildcard(Format::ratio_x_param);
    in_format1->addWildcard(Format::ratio_y_param);
    in_format1->addWildcard(Format::framerate_param);
    in_format1->addStringParamValue(Format::channelorder_param,"yvu");
    in_format1->addStringParamValue(Format::endian_param, "little");
    in_format1->addStringParamValue(Format::colorspace_param,"yv12");
    in_format1->addIntParamValue(Format::bitperpixel_param,12);
    in_format1->addStringParamValue(Format::format_param, "planar");
    getOwnInputProperty()->setDefaultFormat(in_format1, "input1");

    // add new input format for input1 to property output format
    Format* in_format2 = getOwnInputProperty()->addNewFormat("video/raw");
    in_format2->setJackTag("input2");

    in_format2->addWildcard(Format::x_resolution_param);
    in_format2->addWildcard(Format::y_resolution_param);
    in_format2->addWildcard(Format::ratio_x_param);
    in_format2->addWildcard(Format::ratio_y_param);
    in_format2->addWildcard(Format::framerate_param);
    in_format2->addStringParamValue(Format::channelorder_param,"yvu");
    in_format2->addStringParamValue(Format::endian_param, "little");
    in_format2->addStringParamValue(Format::colorspace_param,"yv12");
    in_format2->addIntParamValue(Format::bitperpixel_param,12);
    in_format2->addStringParamValue(Format::format_param, "planar");
    getOwnInputProperty()->setDefaultFormat(in_format2, "input2");

    // add new output format to property output format
    Format* out_format = getOwnOutputProperty()->addNewFormat("video/raw");
    out_format->addWildcard(Format::x_resolution_param);
    out_format->addWildcard(Format::y_resolution_param);
    out_format->addWildcard(Format::ratio_x_param);
    out_format->addWildcard(Format::ratio_y_param);
    out_format->addWildcard(Format::framerate_param);
    out_format->addStringParamValue(Format::channelorder_param,"yvu");
    out_format->addStringParamValue(Format::endian_param, "little");
    out_format->addStringParamValue(Format::colorspace_param,"yv12");
    out_format->addIntParamValue(Format::bitperpixel_param,12);
    out_format->addStringParamValue(Format::format_param, "planar");
    getOwnOutputProperty()->setDefaultFormat(out_format);
    
    // set I/O partners of formats
    in_format1->setIOPartner(out_format);
    out_format->setIOPartner(in_format1);

    in_format2->setIOPartner(out_format);
    out_format->setIOPartner(in_format2);
	    
    return SUCCESS;
  }
	
  Result VideoMuxNode::doInitOutput() 
  {
    // query the format currently set on input1 and add to working input property
    const Format* current_in_format1 = getInputFormat("input1");
    if (current_in_format1 == NULL) { 
      ERROR_STREAM("No input format on input1" << endl);
      return FAILURE; 
    }
	    
    Format* in_format1 = new Format(*current_in_format1);
    in_format1->setJackTag("input1");
    getOwnWorkingInputProperty()->addNewFormat(in_format1);
    getOwnWorkingInputProperty()->setDefaultFormat(in_format1, "input1");

    // query the format currently set on input2 and add to working input property
    const Format* current_in_format2 = getInputFormat("input2");
    if (current_in_format2 == NULL) { 
      ERROR_STREAM("No input format on input2" << endl);
      return FAILURE; 
    }
	    
    Format* in_format2 = new Format(*current_in_format2);
    in_format2->setJackTag("input2");
    getOwnWorkingInputProperty()->addNewFormat(in_format2);
    getOwnWorkingInputProperty()->setDefaultFormat(in_format2, "input2");

    // add the current format set on input1 to working output property
    Format* out_format = new Format(*current_in_format1);
    getOwnWorkingOutputProperty()->addNewFormat(out_format);
    getOwnWorkingOutputProperty()->setDefaultFormat(out_format);

    // set I/O partners of formats
    in_format1->setIOPartner(out_format);
    out_format->setIOPartner(in_format1);

    in_format2->setIOPartner(out_format);
    out_format->setIOPartner(in_format2);

    // set scaler pixel format
    string colorspace1 = in_format1->getStringValue(Format::colorspace_param);
    string colorspace2 = in_format2->getStringValue(Format::colorspace_param);

    if(colorspace1 != colorspace2) {
      ERROR_STREAM("No matching colorspace for input1 and input2" << endl);
      return FAILURE; 
    }

    if (colorspace1 == "yv12") {
      scaler.setFormat(VideoScaler::FORMAT_YV12);
    } else {
      ERROR_STREAM("Input format (colorspace) not supported" << endl);
      return FAILURE;
    }
    
    // store input frame sizes for later scaling operations
    input1_width  = in_format1->getIntValue(Format::x_resolution_param);
    input1_height = in_format1->getIntValue(Format::y_resolution_param);

    input1_ratio_x = in_format1->getIntValue(Format::ratio_x_param);
    input1_ratio_y = in_format1->getIntValue(Format::ratio_y_param);

    input2_width  = in_format2->getIntValue(Format::x_resolution_param);
    input2_height = in_format2->getIntValue(Format::y_resolution_param);

    input2_ratio_x = in_format2->getIntValue(Format::ratio_x_param);
    input2_ratio_y = in_format2->getIntValue(Format::ratio_y_param);

    output_width  = out_format->getIntValue(Format::x_resolution_param);
    output_height = out_format->getIntValue(Format::y_resolution_param);

    return SUCCESS;
  }

  Result VideoMuxNode::syncReset() 
  {
    // always delete sync reset from 2nd input
    if(!strcmp(getCurrentRecvInputStream(), "input2")) {
      return DELETE;
    }
    else {
      return SUCCESS;
    }
  }

  void VideoMuxNode::copyVideoFrame(char*& out_data, size_t out_width, size_t out_height, size_t out_x, size_t out_y, 
				    char*& tmp_data, size_t tmp_width, size_t tmp_height) 
  {
    char* o;
    char* t;
	    
    // Y plane
    o = out_data + out_x + out_y*out_width;
    t = tmp_data;
	
    for(size_t i = 0; i < tmp_height; ++i) {
      memcpy(o, t, tmp_width);
      o += out_width;
      t += tmp_width;
    }
	
    // U plane
    o = out_data + out_width*out_height + out_x/2 + out_y*out_width/4;
    t = tmp_data + tmp_width*tmp_height;
	
    for(size_t i = 0; i < tmp_height/2; ++i) {
      memcpy(o, t, tmp_width/2);
      o += out_width/2;
      t += tmp_width/2;
    }
	
    // V plane
    o = out_data + out_width*out_height + out_width*out_height/4 + out_x/2 + out_y*out_width/4;
    t = tmp_data + tmp_width*tmp_height + tmp_width*tmp_height/4;
	
    for(size_t i = 0; i < tmp_height/2; ++i) {
      memcpy(o, t, tmp_width/2);
      o += out_width/2;
      t += tmp_width/2;
    }
  }

  Time VideoMuxNode::getCurrentSyncTime()
  {
    MutexGuard mg(mutex);
    if(current_timestamp.is_set) {
      return current_timestamp.sync_time;
    }
    else {
      return Time();
    }
  }

  void VideoMuxNode::setPosition(const string& input_jack_tag, const nmm_int32& x, const nmm_int32& y)
  {
    MutexGuard mg(mutex);
    if(input_jack_tag == "input1") {
      x_pos1 = x;
      y_pos1 = y;
    }
    else if(input_jack_tag == "input2") {
      x_pos2 = x;
      y_pos2 = y;
    }
    else {
      WARNING_STREAM("Unknown input_jack_tag " << input_jack_tag << " in method setPosition" << endl);
    }
  }

  void VideoMuxNode::setXPositionControl(const string& input_jack_tag, const Time& start_time, const Time& end_time, const nmm_int32& start_x, const nmm_int32& end_x)
  {
    MutexGuard mg(mutex);
    if(input_jack_tag == "input1") {
      x_pos_interpolator1.setStartTime(start_time);
      x_pos_interpolator1.setEndTime(end_time); 
      x_pos_interpolator1.setStartValue(start_x);
      x_pos_interpolator1.setEndValue(end_x);
    }
    else if(input_jack_tag == "input2") {
      x_pos_interpolator2.setStartTime(start_time);
      x_pos_interpolator2.setEndTime(end_time); 
      x_pos_interpolator2.setStartValue(start_x);
      x_pos_interpolator2.setEndValue(end_x);
    }
    else {
      WARNING_STREAM("Unknown input_jack_tag " << input_jack_tag << " in method setXPositionControl" << endl);
    }
  }

  void VideoMuxNode::setYPositionControl(const string& input_jack_tag, const Time& start_time, const Time& end_time, const nmm_int32& start_y, const nmm_int32& end_y)
  {
    MutexGuard mg(mutex);
    if(input_jack_tag == "input1") {
      y_pos_interpolator1.setStartTime(start_time);
      y_pos_interpolator1.setEndTime(end_time); 
      y_pos_interpolator1.setStartValue(start_y);
      y_pos_interpolator1.setEndValue(end_y);
    }
    else if(input_jack_tag == "input2") {
      y_pos_interpolator2.setStartTime(start_time);
      y_pos_interpolator2.setEndTime(end_time); 
      y_pos_interpolator2.setStartValue(start_y);
      y_pos_interpolator2.setEndValue(end_y);
    }
    else {
      WARNING_STREAM("Unknown input_jack_tag " << input_jack_tag << " in method setYPositionControl" << endl);
    }
  }

  void VideoMuxNode::setSize(const string& input_jack_tag, const nmm_int32& size_x, const nmm_int32& size_y)
  {
    MutexGuard mg(mutex);
    if(input_jack_tag == "input1") {
      width1  = size_x;
      height1 = size_y;
    }
    else if(input_jack_tag == "input2") {
      width2  = size_x;
      height2 = size_y;
    }
    else {
      WARNING_STREAM("Unknown input_jack_tag " << input_jack_tag << " in method setSize" << endl);
    }
  }

  void VideoMuxNode::setWidthControl(const string& input_jack_tag, const Time& start_time, const Time& end_time, const nmm_int32& start_width, const nmm_int32& end_width, const IVideoMux::Alignment& align)
  {
    MutexGuard mg(mutex);
    if(input_jack_tag == "input1") {
      width_interpolator1.setStartTime(start_time);
      width_interpolator1.setEndTime(end_time); 
      width_interpolator1.setStartValue(start_width);
      width_interpolator1.setEndValue(end_width);

      alignment1 = align;
      if(alignment1 == IVideoMux::TOP_LEFT) {
	x_alignment1 = x_pos1;
	y_alignment1 = y_pos1;
      }
      else if(alignment1 == IVideoMux::BOTTOM_RIGHT) {
	x_alignment1 = x_pos1 + width1;
	y_alignment1 = y_pos1 + height1;
      }

    }
    else if(input_jack_tag == "input2") {
      width_interpolator2.setStartTime(start_time);
      width_interpolator2.setEndTime(end_time); 
      width_interpolator2.setStartValue(start_width);
      width_interpolator2.setEndValue(end_width);

      alignment2 = align;
      if(alignment2 == IVideoMux::TOP_LEFT) {
	x_alignment2 = x_pos2;
	y_alignment2 = y_pos2;
      }
      else if(alignment2 == IVideoMux::BOTTOM_RIGHT) {
	x_alignment2 = x_pos2 + width2;
	y_alignment2 = y_pos2 + height2;
      }
    }
    else {
      WARNING_STREAM("Unknown input_jack_tag " << input_jack_tag << " in method setSizeControl" << endl);
    }
  }

  void VideoMuxNode::setHeightControl(const string& input_jack_tag, const Time& start_time, const Time& end_time, const nmm_int32& start_height, const nmm_int32& end_height, const IVideoMux::Alignment& align)
  {
    MutexGuard mg(mutex);
    if(input_jack_tag == "input1") {
      height_interpolator1.setStartTime(start_time);
      height_interpolator1.setEndTime(end_time); 
      height_interpolator1.setStartValue(start_height);
      height_interpolator1.setEndValue(end_height);

      alignment1 = align;
      if(alignment1 == IVideoMux::TOP_LEFT) {
	x_alignment1 = x_pos1;
	y_alignment1 = y_pos1;
      }
      else if(alignment1 == IVideoMux::BOTTOM_RIGHT) {
	x_alignment1 = x_pos1 + width1;
	y_alignment1 = y_pos1 + height1;
      }
    }
    else if(input_jack_tag == "input2") {
      height_interpolator2.setStartTime(start_time);
      height_interpolator2.setEndTime(end_time); 
      height_interpolator2.setStartValue(start_height);
      height_interpolator2.setEndValue(end_height);

      alignment2 = align;
      if(alignment2 == IVideoMux::TOP_LEFT) {
	x_alignment2 = x_pos2;
	y_alignment2 = y_pos2;
      }
      else if(alignment2 == IVideoMux::BOTTOM_RIGHT) {
	x_alignment2 = x_pos2 + width2;
	y_alignment2 = y_pos2 + height2;
      }
    }
    else {
      WARNING_STREAM("Unknown input_jack_tag " << input_jack_tag << " in method setSizeControl" << endl);
    }
  }

  void VideoMuxNode::setWidthHeightCouplingEnabled(const string& input_jack_tag, const nmm_bool& enable)
  {
    MutexGuard mg(mutex);
    if(input_jack_tag == "input1") {
      width_height_coupling1 = enable;
    }
    else if(input_jack_tag == "input2") {
      width_height_coupling2 = enable;
    }
    else {
      WARNING_STREAM("Unknown input_jack_tag " << input_jack_tag << " in method setWidthHeightCouplingEnabled" << endl);
    }
  }

  nmm_int32 VideoMuxNode::getWidth(const string& input_jack_tag)
  {
    MutexGuard mg(mutex);
    if(input_jack_tag == "input1") {
      return width1;
    }
    else if(input_jack_tag == "input2") {
      return width2;
    }
    else {
      WARNING_STREAM("Unknown input_jack_tag " << input_jack_tag << " in method setSizeControl" << endl);
      return 0;
    }
  }

  nmm_int32 VideoMuxNode::getHeight(const string& input_jack_tag) 
  {
    MutexGuard mg(mutex);
    if(input_jack_tag == "input1") {
      return height1;
    }
    else if(input_jack_tag == "input2") {
      return height2;
    }
    else {
      WARNING_STREAM("Unknown input_jack_tag " << input_jack_tag << " in method setSizeControl" << endl);
      return 0;
    }
  }

  nmm_int32 VideoMuxNode::getXPos(const string& input_jack_tag) 
  {
    MutexGuard mg(mutex);
    if(input_jack_tag == "input1") {
      return x_pos1;
    }
    else if(input_jack_tag == "input2") {
      return x_pos2;
    }
    else {
      WARNING_STREAM("Unknown input_jack_tag " << input_jack_tag << " in method setSizeControl" << endl);
      return 0;
    }
  }

  nmm_int32 VideoMuxNode::getYPos(const string& input_jack_tag) 
  {
    MutexGuard mg(mutex);
    if(input_jack_tag == "input1") {
      return y_pos1;
    }
    else if(input_jack_tag == "input2") {
      return y_pos2;
    }
    else {
      WARNING_STREAM("Unknown input_jack_tag " << input_jack_tag << " in method setSizeControl" << endl);
      return 0;
    }
  }

  Message* VideoMuxNode::processBuffer(Buffer* buffer) 
  {
    MutexGuard mg(mutex);
    Buffer* out_buffer = 0;
	  
    // collect video frames in a round-robin approach: input1, input2, input1, ..
    if(!strcmp(getCurrentRecvInputStream(), "input1")) {
      // disable input1, enable input2
      setRecvInputStreamEnabled("input1", false);
      setRecvInputStreamEnabled("input2", true);
	    
      in_buffer1 = buffer;

      // save latest valid and set timestamp of incoming stream on input 1 for method getCurrentTimestamp
      if(in_buffer1->getTimestamp().is_valid && in_buffer1->getTimestamp().is_set)
      {
	current_timestamp = in_buffer1->getTimestamp();
      }
    }
    else if(!strcmp(getCurrentRecvInputStream(), "input2")) {
      // disable input2, enable input1
      setRecvInputStreamEnabled("input2", false);
      setRecvInputStreamEnabled("input1", true);
	    
      in_buffer2 = buffer;
    }

    // if a frame from input1 and a frame from input2 is available
    // -> perform scaling/merging/muxing operation
    if(in_buffer1 && in_buffer2) {

      // request new buffer for outgoing video frame
      // (always use size and timestamp of in_buffer1 received from first input)
      out_buffer = getNewBuffer(in_buffer1->getUsed(), in_buffer1);
      out_buffer->setUsed(in_buffer1->getUsed());
      char* out_data  = out_buffer->getData();

      // Set Y, U, V planes of output frame to 'black'
      memset(out_data, 0,  output_width * output_height);
      memset(out_data + output_width * output_height, 127, output_width * output_height/4);
      memset(out_data + output_width * output_height + output_width * output_height/4, 127, output_width * output_height/4);

      // request temporary buffers required for scaling operation
      Buffer* tmp_buffer1 = getNewBuffer(in_buffer1->getUsed());
      Buffer* tmp_buffer2 = getNewBuffer(in_buffer2->getUsed());

      // get all data pointers
      char* in_data1 = in_buffer1->getData();
      char* in_data2 = in_buffer2->getData();

      char* tmp_data1 = tmp_buffer1->getData();
      char* tmp_data2 = tmp_buffer2->getData();

      if(x_pos_interpolator1.isConfigured()) {
	x_pos1 = static_cast<int32_t>( x_pos_interpolator1.getCurrentValue(in_buffer1->getTimestamp().sync_time) );
	x_pos1 = ((x_pos1)/2)*2;
      }
      if(y_pos_interpolator1.isConfigured()) {
	y_pos1 = static_cast<int32_t>( y_pos_interpolator1.getCurrentValue(in_buffer1->getTimestamp().sync_time) );
	y_pos1 = ((y_pos1)/2)*2;
      }

      // funny while-loop for finding width and height in coupling mode
      while(true) {
	// first, try to get width
	if(width_interpolator1.isConfigured()) {
	  width1 = static_cast<int32_t>( width_interpolator1.getCurrentValue(in_buffer1->getTimestamp().sync_time) );
	  width1  = ((width1/4)*4);
	  if(width_height_coupling1 == true) { // if coupling, compute height from aspect ratio and break;
	    height1 = ((width1*input1_ratio_y)/input1_ratio_x); 
	    height1 = ((height1/4)*4);
	    break;
	  }
	}
	// then, try to get height
	if(height_interpolator1.isConfigured()) {
	  height1 = static_cast<int32_t>( height_interpolator1.getCurrentValue(in_buffer1->getTimestamp().sync_time) );
	  height1 = ((height1/4)*4);
	  if(width_height_coupling1 == true) { // if coupling, compute width from aspect ratio and break;
	    width1 = ((height1*input1_ratio_x)/input1_ratio_y); 
	    width1  = ((width1/4)*4);
	    break;
	  }
	}
	break;
      }


      if(x_pos_interpolator2.isConfigured()) {
	x_pos2 = static_cast<int32_t>( x_pos_interpolator2.getCurrentValue(in_buffer1->getTimestamp().sync_time) );
	x_pos2 = ((x_pos2)/2)*2;
      }
      if(y_pos_interpolator2.isConfigured()) {
	y_pos2 = static_cast<int32_t>( y_pos_interpolator2.getCurrentValue(in_buffer1->getTimestamp().sync_time) );
	y_pos2 = ((y_pos2)/2)*2;
      }

      // funny while-loop for finding width and height in coupling mode
      while(true) {
	// first, try to get width
	if(width_interpolator2.isConfigured()) {
	  width2 = static_cast<int32_t>( width_interpolator2.getCurrentValue(in_buffer1->getTimestamp().sync_time) );
	  width2  = ((width2/4)*4);
	  if(width_height_coupling2 == true) { // if coupling, compute height from aspect ratio and break;
	    height2 = ((width2*input2_ratio_y)/input2_ratio_x); 
	    height2 = ((height2/4)*4);
	    break;
	  }
	}
	// then, try to get height
	if(height_interpolator2.isConfigured()) {
	  height2 = static_cast<int32_t>( height_interpolator2.getCurrentValue(in_buffer1->getTimestamp().sync_time) );
	  height2 = ((height2/4)*4);
	  if(width_height_coupling2 == true) { // if coupling, compute width from aspect ratio and break;
	    width2 = ((height2*input2_ratio_x)/input2_ratio_y); 
	    width2  = ((width2/4)*4);
	    break;
	  }
	}
	break;
      }

      /*
      cerr << width1 << " x " << height1 << " : " 
	   << (float) width1 / (float) height1 << ", " 
	   << width2 << " x " << height2 << " : " 
	   << (float) width2 / (float) height2 << endl;
      */

      // video frame needs to be visible, otherwise simply do not scale and copy
      if(width1 > 0 && height1 > 0) {
	// scale the incoming video frames 
	scaler.setInputSize(input1_width, input1_height);
	scaler.setOutputSize(width1, height1);
	scaler.processData(tmp_data1, in_data1);

	if(alignment1 == IVideoMux::BOTTOM_RIGHT) {
	  // align frame to bottom right corner
	  if(x_pos1 + width1 < x_alignment1) {
	    x_pos1 = x_alignment1 - width1;
	  }
	  if(y_pos1 + height1 < y_alignment1) {
	    y_pos1 = y_alignment1 - height1;
	  }
	}

	// and copy to correct position
	copyVideoFrame(out_data, output_width, output_height, x_pos1, y_pos1, tmp_data1, width1, height1);
      }

      // video frame needs to be visible, otherwise simply do not scale and copy
      if(width2 > 0 && height2 > 0) {
	// scale the incoming video frames 
	scaler.setInputSize(input2_width, input2_height);
	scaler.setOutputSize(width2, height2);
	scaler.processData(tmp_data2, in_data2);
	
	if(alignment2 == IVideoMux::BOTTOM_RIGHT) {
	  // align frame to bottom right corner
	  if(x_pos2 + width2 < x_alignment2) {
	    x_pos2 = x_alignment2 - width2;
	  }
	  if(y_pos2 + height2 < y_alignment2) {
	    y_pos2 = y_alignment2 - height2;
	  }
	  
	}

	// and copy to correct position
	copyVideoFrame(out_data, output_width, output_height, x_pos2, y_pos2, tmp_data2, width2, height2);
      }
	    
      // release buffers
      tmp_buffer1->release();
      tmp_buffer2->release();
      in_buffer1->release();
      in_buffer2->release();

      in_buffer1 = 0;
      in_buffer2 = 0;
    }


    return out_buffer;
  }


};

#include "nmm/nmm_suffix.hpp"
