/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include "BitmapFont.hpp"
#include "nmm/utils/xml/XMLTree.hpp"
#include "nmm/utils/xml/XMLNode.hpp"
#include <iostream>

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  BitmapFont::BitmapFont() {

    /* Initialize chars structure */
    for( int i=0; i<255; i++ )
      memset( &chars[i], 0, sizeof( char_t) );

    max_width  = 0;
    max_height = 0;

    setLineSpace( 2 );
  }

  BitmapFont::~BitmapFont() {

    /* deallocate memory of all characters */
    for( int i=0; i<255; i++ )
      freeChar( i );

  }

  void BitmapFont::setLineSpace( unsigned int ls ) {
    line_space = ls;
  }

  unsigned int BitmapFont::getLineSpace() {
    return( line_space );
  }

  unsigned int BitmapFont::getMaxWidth() {
    return( max_width );
  }

  unsigned int BitmapFont::getMaxHeight() {
    return( max_height );
  }

  Result BitmapFont::load( string xmlfilename, string fontname ) {
    /* open XML document */
    XMLTree* xml = new XMLTree( xmlfilename );

    /* get pointer to XML root */
    XMLNode* rootnode = xml->root();
    if ( rootnode == NULL ) return FAILURE;

    //if ( rootnode->name() != "FONTS" ) return FAILURE;
    /* get a list with all FONT nodes */
    XMLNodeList fontlist = rootnode->children("FONT");

    /* iterator to iterate FONT list */
    XMLNodeIterator iter, stop;
    iter = fontlist.begin();
    stop = fontlist.end();

    XMLNode* fontnode;

    /* indicates if any font or fontname was found */
    bool found = false;

    /* process font list */
    /* try to find FONTS->FONT entry with the attribute name=fontname */
    while ( (iter != stop) && !found ) {
      fontnode = *iter;
      iter++;

      /* if font name is not spezified, use the first that is found */
      if ( fontname == "" ) { found = true; break; }

      /* Try to find "name" attribute */
      XMLProperty* prop = fontnode->property("name");
      if ( prop != NULL ) {
	if ( prop->value() == fontname ) found = true;
      }

    } // while

    /* no font description was found, exit */
    if ( !found ) return FAILURE;

    /* now fontnode points to the font in the XML file we have selected */
    string bitmapfile = "";
    XMLProperty* prop = fontnode->property("bmapfile");
    if ( prop != NULL ) bitmapfile = prop->value();

    /* extend bitmap file name with full path */
    string bitmap_file_path = xmlfilename;
    string::size_type ls = bitmap_file_path.find_last_of("/");
    if (bitmap_file_path.npos == ls) {
      // No slash found => only filename
      // => replace from beginning to end!
      ls = 0;
    }
    bitmap_file_path.replace(ls+1 , bitmap_file_path.npos, bitmapfile );

    BitmapReader* bitmap = new BitmapReader();
    /* load the font bitmap, if no font bitmap was found, exit */
    if ( !bitmap->loadImage( "image", bitmap_file_path ) ) {
      cout << "Bitmap " << bitmapfile << "(" << bitmap_file_path << ") not found." << endl;
      return FAILURE;
    }

    /* width and height of the font bitmap */
    unsigned int   bitmapwidth, bitmapheight;

    /* get pixel data and dimension of the bitmap */
    unsigned char* bitmapdata = bitmap->getImage( "image", &bitmapwidth, &bitmapheight );

    /* get a list of all <CH> - tags */
    fontlist = fontnode->children( "CH" );

    iter = fontlist.begin();
    stop = fontlist.end();

    max_width = max_height = 0;

    XMLNode* charnode;
    /* process <ch> list */
    while( iter != stop ) {
      unsigned int width;
      unsigned int height;

      charnode = *iter;
      iter++;

      /* get all attributes of a <CH> tag */
      /* that means the description of a character */
      /* get character */
      string c;
      prop = charnode->property("char");
      if ( prop ) c = prop->value();

      /* get character ascii code (optional) */
      string asc;
      prop = charnode->property("ascii");
      if ( prop ) asc = prop->value();

      /* get x position (upper left corner) of this character in the bitmap */
      string x;
      prop = charnode->property("x");
      if ( prop ) x = prop->value();

      /* get y position (upper left corner) of this character in the bitmap */
      string y;
      prop = charnode->property("y");
      if ( prop ) y = prop->value();

      /* get width of this character in the bitmap */
      string w;
      prop = charnode->property("w");
      if ( prop ) w = prop->value();

      /* get height of this character in the bitmap */
      string h;
      prop = charnode->property("h");
      if ( prop ) h = prop->value();

      char ascii;
      if ( asc == "" ) ascii = *c.c_str();
      else ascii = atoi( asc.c_str() );

      width  = atoi( w.c_str() );
      height = atoi( h.c_str() );

      /* save the width or height if it is a new maximum */
      if ( width > max_width ) max_width = width;
      if ( height > max_height ) max_height = height;

      /* store Bitmapdata of the given character in the chars structure */
      captureChar( ascii, bitmapdata, bitmapwidth,
		   atoi( x.c_str() ), atoi( y.c_str() ),
		   width, height );
    }

    /* deallocate memory that was allocated when loading the Bitmap */
    //bitmap.freeImage( "image" );
    //bitmap.freeAll();

    return SUCCESS;
  }

  void BitmapFont::freeChar( unsigned char c ) {
    /* if pixeldata is present, delete it */
    if ( chars[c].pixeldata ) delete[] chars[c].pixeldata;

    /* Fill structure with zeros */
    memset( &chars[c], 0, sizeof( char_t) );
  }

  void BitmapFont::checkDimension( unsigned char c, unsigned int *x, unsigned int *y ) {
    *x = chars[c].width;
    *y = chars[c].height;
  }

  unsigned char* BitmapFont::getPixelData( unsigned char c ) {
    return( chars[c].pixeldata );
  }

  void BitmapFont::captureChar( unsigned char c, unsigned char* bitmapdata, unsigned int bitmapwidth, int xpos, int ypos, int width, int height ) {

    unsigned char  r, g, b, a;
    unsigned char* data;
    unsigned char* pixeldata;

    /* Free last character */
    freeChar( c );

    /* put width and height of the character in the struct */
    chars[c].width     = width;
    chars[c].height    = height;
    /* allocate memory for the rgba pixel data */
    chars[c].pixeldata = new unsigned char[width*height*4];  // Store character in RGBA format

    pixeldata = chars[c].pixeldata;
    /* process all lines of the character */
    for( int y=ypos; y < ypos+height; y++ ) {
      data = bitmapdata+(y*bitmapwidth*4);

      /* process all pixels per line (note: one pixel needs 4 bytes in rgba format) */
      for( int x=(xpos<<2); x < ( (xpos<<2) + (width<<2) ); x+=4 ) {

	r = data[x];
	g = data[x+1];
	b = data[x+2];
	a = data[x+3];

	/*if ( r || g || b ) cout << "*";
	  else cout << ".";*/

	/* save pixeldata in chars structure */
	pixeldata[0] = r;
	pixeldata[1] = g;
	pixeldata[2] = b;
	pixeldata[3] = a;
	pixeldata += 4;
      }

      //cout << endl;
    }

  }

}

#include "nmm/nmm_suffix.hpp"
