/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include "BitmapReader.hpp"
#include "nmm/utils/SecurityCheck.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  BitmapReader::BitmapReader() {
  }

  BitmapReader::~BitmapReader() {
    /* at the end, delete all images */
    freeAll();
  }

  unsigned char* BitmapReader::loadBitmap( string filename, unsigned int* width, unsigned int* height  ) {
    png_bytep*  png_rows;               // stores the PNG image
    png_uint_32 png_width, png_height;  // PNG image width and height

    png_structp png_ptr;
    png_infop info_ptr;
    int bit_depth, color_type, interlace_type;
    FILE *fp;

    png_rows = NULL;

    if(!SSecurityCheck::getInstance().checkFilenameForReading(filename)) {
      GLOBAL_ERROR_STREAM("Security check not passed for file " << filename << ", check .nmmrc for options" << endl);
      return NULL;
    }
    if ((fp = fopen(filename.c_str(), "rb")) == NULL) {
      return NULL;
    }

    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);

    if (png_ptr == NULL) {
      fclose(fp);
      return NULL;
    }

    info_ptr = png_create_info_struct(png_ptr);
    if (info_ptr == NULL) {
      fclose(fp);
      png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
      return NULL;
    }

    if (setjmp(png_ptr->jmpbuf)) {
      png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
      fclose(fp);
      return NULL;
    }

    png_init_io(png_ptr, fp);

    png_read_info(png_ptr, info_ptr);

    png_get_IHDR(png_ptr, info_ptr, &png_width, &png_height, &bit_depth, &color_type, &interlace_type, NULL, NULL);


    if (bit_depth == 8 && (color_type == PNG_COLOR_TYPE_RGB || color_type == PNG_COLOR_TYPE_GRAY) ) {
      //DEBUG_STREAM( "Adding alpha channel" <<  endl );
      png_set_filler(png_ptr, 0xFF, PNG_FILLER_AFTER);
      //png_set_bgr(png_ptr);
    }

    if ( color_type == PNG_COLOR_TYPE_GRAY ||
	 color_type == PNG_COLOR_TYPE_GRAY_ALPHA) {
      //DEBUG_STREAM( "Converting Grayscale to RGB" << endl );
      png_set_gray_to_rgb(png_ptr);
    };
    png_read_update_info(png_ptr, info_ptr);

    png_rows = new png_bytep[png_height];

    for (unsigned int row = 0; row < png_height; row++) {
      png_rows[row] = new png_byte[png_get_rowbytes(png_ptr, info_ptr)];
    }

    png_read_image(png_ptr, png_rows);

    png_read_end(png_ptr, info_ptr);

    png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);

    fclose(fp);

    *width  = (unsigned int) png_width;
    *height = (unsigned int) png_height;

    /* all bitamps we load are in rgba format, so mutliply width*height with 4 */
    unsigned char* data = new unsigned char[png_width*png_height*4];
    unsigned char* temp = data;

    for (unsigned int row = 0; row < png_height; row++) {
      if ( png_rows[row] ) {
	memcpy( temp, png_rows[row], png_width*4 );
	temp+=png_width*4;
	delete[] png_rows[row];
      }
    }


    return data;
  }

  void BitmapReader::freeBitmap( unsigned char* data ) {
    /* delete pixeldata of the image */
    if ( data ) delete[] data;
  }

  bool BitmapReader::loadImage( string id, string filename, format f ) {

    unsigned char* data;
    unsigned char* temp;
    unsigned int   width, height;

    /* first, try to delete a priviously opened image */
    freeImage( id );

    /* load the new image */
    temp = loadBitmap( filename, &width, &height );

    /* if there was an error, return */
    if ( !temp ) return false;

    if ( (f == YV12) || (f == YV12A) ) {
      /* convert RGBA data to YUV */
      data = convertToYUV( temp, width, height, f );

      /* delete RGBA data */
      freeBitmap( temp );
    } else {
      data = temp;
    }

    /* ok, bitmap could be read, create new bitmap structure */
    bitmap_t* bitmap = new bitmap_t;

    /* fill structure with image data */
    bitmap->width     = width;
    bitmap->height    = height;
    bitmap->cformat   = f;
    bitmap->pixeldata = data;

    /* create new map entry */
    bitmaplist[id]    = bitmap;

    return( true );
  }

  void BitmapReader::drawFilledBox( unsigned int* frame, unsigned int stride,
		         	    unsigned int xs, unsigned int ys,
				    unsigned int width, unsigned int height,
				    unsigned int color ) {

    for( unsigned int y=ys; y < (ys+height); y++ ) {
      for( unsigned int x=xs; x < (xs+width); x++ ) {
	frame[x+y*stride]  = color;
      }
    }

  }

  void BitmapReader::drawBox( unsigned int* frame, unsigned int stride,
		              unsigned int xs, unsigned int ys,
		              unsigned int width, unsigned int height,
		              unsigned int color1, unsigned int color2 ) {
    unsigned int  x, y;
    unsigned int  xd, yd;

    xd = xs+width-1;
    yd = ys+height-1;

    for(x=xs; x <= xd; x++) {
      frame[x+ys*stride]     = color1;
      frame[x+(ys+1)*stride] = color1;

      frame[x+(yd*stride)]     = color2;
      frame[x+((yd-1)*stride)] = color2;
    }

    for(y=ys+1; y <= yd; y++) {
      frame[xs+(y*stride)]   = color1;
      frame[xs+1+(y*stride)] = color1;

      frame[xd+(y*stride)]   = color2;
      frame[xd-1+(y*stride)] = color2;
    }

  }

  bool BitmapReader::createImage( string id, unsigned int width, unsigned int height,
                                  unsigned int color1, unsigned int color2, unsigned int color3, format f ) {

    unsigned char* data;
    unsigned char* temp;

    /* first, try to delete a priviously opened/generated image */
    freeImage( id );

    /* generate bitmap (RGBA format) */
    temp = new unsigned char[width*height*4];

    /* if there was an error, return */
    if ( !temp ) return false;

    drawFilledBox( (unsigned int*) temp, width, 0, 0,
		   width, height, color1 );

    drawBox( (unsigned int*) temp, width, 0, 0,
	     width, height, color2, color3 );

    if ( (f == YV12) || (f == YV12A) ) {
      /* convert RGBA data to YUV */
      data = convertToYUV( temp, width, height, f );

      /* delete RGBA data */
      freeBitmap( temp );
    } else {
      data = temp;
    }

    /* ok, create new bitmap structure */
    bitmap_t* bitmap = new bitmap_t;

    /* fill structure with image data */
    bitmap->width     = width;
    bitmap->height    = height;
    bitmap->cformat   = f;
    bitmap->pixeldata = data;

    /* create new map entry */
    bitmaplist[id]    = bitmap;

    return( true );
  }

  void BitmapReader::freeImage( string id ) {
    bitmap_t* bitmap;

    /* look for the given image */
    bitmap = bitmaplist[id];

    /* if the image was found, delete pixeldata and entry */
    if ( bitmap ) {
      //DEBUG_STREAM( "Bitmap: id found, deleting pixeldata" << endl );
      if ( bitmap->pixeldata ) freeBitmap( bitmap->pixeldata );

    }

    /* delete entry */
    bitmaplist.erase(id);
  }

  void BitmapReader::freeAll() {
    bitmaplist_t::iterator iter, stop;

    /* begin of the image list */
    iter = bitmaplist.begin();

    /* end of the image list */
    stop = bitmaplist.end();

    //DEBUG_STREAM( "Bitmap: deleting all images..." << endl );
    /* process the whole list and delete all items */
    while ( iter != stop ) {
      freeImage( iter->first );
      iter++;
    }

    bitmaplist.clear();

  }

  unsigned char* BitmapReader::getImage( string id, unsigned int* width, unsigned int* height ) {
    bitmap_t* bitmap;

    /* look for the given image */
    bitmap  = bitmaplist[id];

    /* if no image was found, return */
    if ( !bitmap ) return NULL;

    /* return the image data */
    *width  = bitmap->width;
    *height = bitmap->height;

    return( bitmap->pixeldata );
  }

  double BitmapReader::getYComponent( unsigned char R, unsigned char G, unsigned char B ) {
    return( R *  0.299 + G *  0.587 + B *  0.114 );
  }

  double BitmapReader::getUComponent( unsigned char R, unsigned char G, unsigned char B ) {
    return( R * -0.169 + G * -0.332 + B *  0.500 + 128 );
  }

  double BitmapReader::getVComponent( unsigned char R, unsigned char G, unsigned char B ) {
    return( R *  0.500 + G * -0.419 + B * -0.0813 + 128 );
  }

  unsigned char* BitmapReader::convertToYUV( unsigned char* rgbbuffer, int width, int height, format f ) {

    unsigned char* yuv_buffer;
    unsigned int framesize;

    if ( f == YV12A ) {
       /* 1 y-plane + 1/4 u-plane + 1/4 v-plane + 1 alpha plane */
       framesize = (unsigned int) ( width * height * 2.5 );
    } else
    if ( f == YV12 ) {
       /* 1 y-plane + 1/4 u-plane + 1/4 v-plane */
       framesize = (unsigned int) ( width * height * 1.5 );
    } else return 0;

    yuv_buffer = new unsigned char[framesize];

    int plane_size = width * height;

    unsigned char* y_plane = yuv_buffer;
    unsigned char* u_plane = yuv_buffer + plane_size;
    unsigned char* v_plane = u_plane + plane_size/4;

    unsigned char *rgba1, *yuv1;
    unsigned char *rgba2, *yuv2;
    unsigned char *rgba3, *yuv3;
    unsigned char *rgba4, *yuv4;
    unsigned char r1, r2, r3, r4, g1, g2, g3, g4, b1, b2, b3, b4;

    for (int row=0; row < height/2; row++) {
      rgba1 = rgbbuffer+(row*width << 3);             // rgbbuffer+(2*row*width*4);
      rgba2 = rgba1 + 4;
      rgba3 = rgbbuffer+((2*row+1)* (width<<2));      // rgbbuffer+((2*row+1)*width*4);
      rgba4 = rgba3 + 4;
      yuv1  = y_plane + 2*row * width;
      yuv2  = yuv1 + 1;
      yuv3  = y_plane + (2*row + 1) * width;
      yuv4  = yuv3 + 1;

      for (int col=0; col < width/2; col++) {
	b1 = rgba1[0];
	g1 = rgba1[1];
	r1 = rgba1[2];
	*yuv1 = (unsigned char) getYComponent( r1, g1, b1 );

	b2 = rgba2[0];
	g2 = rgba2[1];
	r2 = rgba2[2];
	*yuv2 = (unsigned char) getYComponent( r2, g2, b2 );

	b3 = rgba3[0];
	g3 = rgba3[1];
	r3 = rgba3[2];
	*yuv3 = (unsigned char) getYComponent( r3, g3, b3 );

	b4 = rgba4[0];
	g4 = rgba4[1];
	r4 = rgba4[2];
	*yuv4 = (unsigned char) getYComponent( r4, g4, b4 );

	r1 = (r1+r2+r3+r4)>>2;
	g1 = (g1+g2+g3+g4)>>2;
	b1 = (b1+b2+b3+b4)>>2;
	*u_plane = (unsigned char) getUComponent( r1, g1, b1 );
	*v_plane = (unsigned char) getVComponent( r1, g1, b1 );
	u_plane++; v_plane++;

	rgba1 += 8;
	rgba2 = rgba1 + 4;
	rgba3 += 8;
	rgba4 = rgba3 + 4;
	yuv1  += 2;
	yuv2  = yuv1 + 1;
	yuv3  += 2;
	yuv4  = yuv3 + 1;
      }
    }

    if ( f == YV12A ) {
      // copy alpha from rgb image to end of yuv_buffer
      unsigned char *yuv_alpha = yuv_buffer + plane_size + plane_size/2;
      unsigned char *alpha;
      for (int row=0; row < height; row++) {
        alpha = rgbbuffer+(row*width<<2);
        for (int col=0; col < width; col++) {
	  *yuv_alpha = alpha[3];
	  alpha += 4;
	  yuv_alpha++;
        }
      }
    } //if

    return( yuv_buffer );

  }
}

#include "nmm/nmm_suffix.hpp"
