/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include "DoubleListUnit.hpp"
#include "nmm/comm/ProxyObject.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  DoubleListUnit::DoubleListUnit( IOSDManager* manager, IWidgetFactory* factory, string _font ) : OSDElement(manager, factory),
                                                                                                  font(_font) {
    setColors( 0xff000000, 0xff100000, 0xff707070, 0xff303030, 0xff101010 );
    setFirstColumnWidth( 50 );
    
    composite.reset(factory->createWidgetComposite());
    
    plane.reset(factory->createPlaneDecorator( *composite ));
    plane->setPlaneColor( win_inactive_color );
    border.reset(factory->createBorderDecorator( *plane ));
    header.reset(factory->createHeaderDecorator( *border ));
    header->setFont(font);
    
    first.reset(factory->createTextView());
    first->setFont(font);
    first->setSelectionColor( sel_inactive_color );
    
    composite->add( *first );
    win->setContent( *header );
  }

  DoubleListUnit::~DoubleListUnit() {
    deleteAllLines();
    deleteWidgets();
  }

  unsigned int  DoubleListUnit::getNumLines() {
    return( first->getNumLines() );
  }
  
  ITextView* DoubleListUnit::getSubList( unsigned int line ) {
    IWidget_var second(first->getLineWidget( line ));
    if ( second.get() ) {
      return second->getParentObject()->getInterface<ITextView>();
    } else {
      return NULL;
    }
  }
  
  void DoubleListUnit::setFirstColumnWidth( unsigned int w ) {
    if ( w < 1 ) w = 1;
    if ( w > 100 ) w = 100;
    column_width = w;
  }

  void DoubleListUnit::deleteWidgets() {
    if ( composite.get() ) {
      manager->deleteWidget(*composite);
      composite.reset(NULL);
    }
    
    if ( first.get() ) {
      manager->deleteWidget(*first);
      first.reset(NULL);
    }
    
    if ( plane.get() ) {
      manager->deleteWidget(*plane);
      plane.reset(NULL);
    }
    
    if ( border.get() ) {
      manager->deleteWidget(*border);
      border.reset(NULL);
    }
    
    if ( header.get() ) {
      manager->deleteWidget(*header);
      header.reset(NULL);
    }
  }
  
  void DoubleListUnit::setColors( unsigned int _win_active_color, unsigned int _win_inactive_color,
                                  unsigned int _sel_active_color, unsigned int _sel_active_color2, unsigned int _sel_inactive_color ) {
    
    win_active_color   = _win_active_color;
    win_inactive_color = _win_inactive_color;
    sel_active_color   = _sel_active_color;
    sel_active_color2  = _sel_active_color2;
    sel_inactive_color = _sel_inactive_color;
  }

  void DoubleListUnit::selectWindow( unsigned int mode ) {
    if ( mode == 1 ) {
      first->setSelectionColor( sel_active_color2 );
      if (getNumLines()!=0) {
        for (uint i=1;i<=getNumLines();i++) {
          ITextView_var second(getSubList(i));
          if ( second.get() ) {
            if (first->getSelectedLine()==i) 
              second->setSelectionColor(sel_active_color2);
          }
        }
      }
    } else {
      first->setSelectionColor( sel_active_color );
      if (getNumLines()!=0) {
        for (uint i=1;i<=getNumLines();i++) {
          ITextView_var second(getSubList(i));
          if ( second.get() ) {
            if (first->getSelectedLine()==i) 
              second->setSelectionColor(sel_active_color);
            else second->setSelectionColor(win_active_color);
          }          
        }
      }
    }
    plane->setPlaneColor( win_active_color );
  }

  void DoubleListUnit::deSelectWindow() {
    first->setSelectionColor( sel_inactive_color );
    if (getNumLines()!=0) {
      for (uint i=1;i<=getNumLines();i++) {
        ITextView_var second(getSubList(i));
        if ( second.get() ) {
          if (first->getSelectedLine()==i) {
            second->setSelectionColor(sel_inactive_color);
          }
          else {
            second->setSelectionColor( win_inactive_color );
          }
        }
      }
    } 
    plane->setPlaneColor( win_inactive_color );
  }

  void DoubleListUnit::setHeader( string h ) {
    header->setText( h );
  }

  unsigned int DoubleListUnit::addLine( string s, line_t content, string url ) {
    ITextView_var second;
    second.reset(factory->createTextView());
    second->setFont(font);
    second->setSelectionColor( win_inactive_color );
    
    unsigned int size = content.size();
    for( unsigned i=0; i<size; i++ ) {
      second->addLine( content[i] );
    }
    
    unsigned int w1, w2,  h;
    win->getDimension( w1, h );
    w2 = (w1*column_width/100);
    w1 = w1-w2;
    
    first->setDimension( w2, h );
    unsigned int font_h = manager->getMaxHeight(font);
    second->setDimension( w1, font_h );
    second->selectLine( 1 );
    composite->add( *second );
    
    unsigned int line = first->addLine( s, url );
    first->addLineWidget( line, *second );
    
    if ( line ) second->setPosition( w2, (font_h+manager->getLineSpace(font))*(line-1) );
    
    return( line );
  }
  
  void DoubleListUnit::deleteLine( unsigned int pos ) {
    IWidget_var second(first->getLineWidget( pos ));
    if ( second.get() ) {
      composite->remove( *second );
      if ( second.get() ) { manager->deleteWidget( *second ); }
    }
    
    first->deleteLine( pos );
  }
  
  void DoubleListUnit::deleteAllLines() {
    unsigned int size = first->getNumLines();
    
    /* delete size times line 1 because if a line was deleted the following lines move up */
    for( unsigned int i=1; i<=size; i++ )
      deleteLine( 1 );
  }

  string DoubleListUnit::getLineUrl( unsigned int l ) {
    return( first->getLineURL( l ) );
  }

  string DoubleListUnit::getLineContent( unsigned int l ) {
    ITextView_var second(getSubList(l));
    if ( second.get() ) {
      return( second->getLineContent( second->getSelectedLine() ) );
    }
    
    return("");
  }

  void DoubleListUnit::changeSelectedLineContent( string b ) {
    unsigned int sel = first->getSelectedLine();
    changeLineContent( sel, b );
  }

  void DoubleListUnit::changeLineContent( unsigned int l, string b ) {
    ITextView_var second(getSubList(l));
    if ( second.get() ) {
      unsigned int pos = second->getSelectedLine();
      if ( pos > 0 ) {
        second->deleteLine( pos );
        second->insertLine( pos-1, b );
        second->selectLine( pos );
      }
    }
  }
  
  void DoubleListUnit::changeSelectedLineSubject(string b){
    unsigned int sel = first->getSelectedLine();
    changeLineSubject(sel, b);
  }

  void DoubleListUnit::changeLineSubject(unsigned int l, string b){
    first->changeLineContent(l,b);

  }

  string DoubleListUnit::getLineSubject(unsigned int l){
    return first->getLineContent(l);
  }

  void DoubleListUnit::chooseLineContent( unsigned int a, unsigned int b ) {
    ITextView_var second(getSubList(a));
    if ( second.get() ) {
      second->setSelectionColor( win_active_color );
      second->selectLine( b );
    }
  }
  
  void DoubleListUnit::chooseLineContent( unsigned int a, string b ) {
    ITextView_var second(getSubList(a));
    if ( second.get() ) {
      second->setSelectionColor( win_active_color );
      
      unsigned int l = second->getNumLines();
      unsigned int c = 1;
      string t;
      bool end;
      if ( l > 0 ) end = false;
      else end = true;
      
      while( !end ) {
        t = second->getLineContent( c );
        if ( t == b ) {
          second->selectLine( c );
          end = true;
        } else {
          c++;
        }
        
        if ( c>l ) end = true;
        
      }
    }
  }
  
  unsigned int DoubleListUnit::getSelectedLine() {
    return( first->getSelectedLine() );
  }
  
  void DoubleListUnit::reset() {
    unsigned int l = first->getNumLines();

    for( unsigned int i=1; i<=l; i++ ) {
      ITextView_var second(getSubList(i));
      if ( second.get() ) {
        second->setSelectionColor( win_active_color );
        second->selectLine( 1 );
      }
    }
  }
  
  Result DoubleListUnit::selectUp() {
    ITextView_var second;
    
    second.reset(getSubList( first->getSelectedLine() ));
    if ( second.get() ) { second->setSelectionColor( win_active_color ); }
    
    first->selectUp();
    
    second.reset(getSubList( first->getSelectedLine() ));
    if ( second.get() ) second->setSelectionColor( sel_active_color );
    return SUCCESS;
  }

  Result DoubleListUnit::selectDown() {
    ITextView_var second;
    
    second.reset(getSubList( first->getSelectedLine() ));
    if ( second.get() ) { second->setSelectionColor( win_active_color ); }
    
    first->selectDown();
    
    second.reset(getSubList( first->getSelectedLine() ));
    if ( second.get() ) second->setSelectionColor( sel_active_color );
    return SUCCESS;
  }
  
  Result DoubleListUnit::selectLeft() {
     ITextView_var second(getSubList( first->getSelectedLine() ));
     if ( second.get() ) {
       second->selectUp();
       return SUCCESS;
     }
     return FAILURE;
  }
  
  Result DoubleListUnit::selectRight() {
    ITextView_var second(getSubList( first->getSelectedLine() ));
    if ( second.get() ) {
      second->selectDown();
      return SUCCESS;
    }
    return FAILURE;
  }
  
  void DoubleListUnit::selectLine(const unsigned int& l){
    ITextView_var second;
    
    second.reset(getSubList( first->getSelectedLine() ));
    if ( second.get() ) { second->setSelectionColor( win_active_color ); }

    first->selectLine(l);
    first->setSelectionColor(sel_active_color);

    second.reset(getSubList( first->getSelectedLine() ));
    if ( second.get() ) second->setSelectionColor( sel_active_color );
  }  

  void DoubleListUnit::selectLine2(const unsigned int& l){
    ITextView_var second;
    
    second.reset(getSubList( first->getSelectedLine() ));
    if ( second.get() ) { second->setSelectionColor( win_active_color ); }

    first->selectLine(l);
    first->setSelectionColor(sel_active_color2);
    second.reset(getSubList( first->getSelectedLine() ));
    if ( second.get() ) second->setSelectionColor( sel_active_color2 );
  }  
}

#include "nmm/nmm_suffix.hpp"
