/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include "OSDManagerNode.hpp"

#include "nmm/utils/thread/MutexGuard.hpp"
#include "nmm/comm/messaging/CEvent.hpp"
#include "nmm/comm/format/Property.hpp"

#include "MessageUnit.hpp"
#include "TimeUnit.hpp"
#include "ProgressbarUnit.hpp"
#include "SymbolUnit.hpp"
#include "ColorBarUnit.hpp"

#include "Widget.hpp"
#include "TextView.hpp"
#include "ImageView.hpp"
#include "Buttons.hpp"
#include "RadioButtons.hpp"
#include "TimeView.hpp"
#include "ProgressBar.hpp"
#include "BitmapFont.hpp"
#include "overlay.h"

#include "Decorator.hpp"
#include "BorderDecorator.hpp"
#include "PlaneDecorator.hpp"
#include "ScrollDecorator.hpp"
#include "HeaderDecorator.hpp"

#include "interfaces/ISymbolChange.hpp"
#include "nmm/multimedia/IFormatChange.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"


NMM_REGISTER_PLUGIN(OSDManagerNode, NMM, "OSDManagerNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  OSDManagerNode::OSDManagerNode(const char *name, const StreamQueue::Mode queue_mode, const int queue_size) : 
    GenericMultiplexerNode(name, queue_mode, queue_size),
    lastBuffer(NULL),
    paused(false),
    max_window_id(0),
    max_widget_id(0),
    windows(),
    widgets(),
    fonts(),
    bmreader(new BitmapReader()),
    send_symbol_change(false),
    new_symbol(""),
    doupdate(false),
    message(NULL),
    time(NULL),
    bar(NULL),
    color_bar(NULL),
    symbol(NULL),
    create_time(false),
    osd_mutex(ThreadMutex::RECURSIVE),
    image_width(0),
    image_height(0),
    ratio_x(0),
    ratio_y(0),
    reference_width(720),
    reference_height(576),
    scale_factor_x(1),
    scale_factor_y(1),
    color(0),
    old_timestamp(),
    out_buffer(NULL),
    send_format_changed(false),
    format_changed_sent(false),
    save_working_flag(false),
    save_in_buffer(NULL)
  {
    addInputStream("stream1");
    addInputStream("stream2");
  }

  OSDManagerNode::~OSDManagerNode(){
    if ( message ) delete message;
    if ( time )    delete time;
    if ( bar )     delete bar;
    if ( symbol )  delete symbol;
    
    /*Begin> We own the fonts now, so we must delete them (fw) */
    for (fonts_t::iterator iter(fonts.begin()); iter != fonts.end(); ++iter) {
      delete iter->second;
    }
    /*End< */
    
    /*Begin> We own the windows now, so we must delete them (fw) */
    for (windows_t::iterator iter(windows.begin()); iter != windows.end(); ++iter) {
      delete iter->second;
    }
    /*End< */
    
    /*Begin> We own the widgets now, so we must delete them (fw) */
    for (widgets_t::iterator iter(widgets.begin()); iter != widgets.end(); ++iter) {
      delete iter->second;
    }
    /*End< */
    
    if (bmreader) {
      delete bmreader;
      bmreader = NULL;
    }
  }
  
  Result OSDManagerNode::doInit() {
    send_symbol_change = false;
    doupdate = false;
    
    
    // initialize formats
    Format* myInputFormat;
    myInputFormat=getOwnInputProperty()->addNewFormat("video/raw");
    myInputFormat->addWildcard(Format::x_resolution_param);
    myInputFormat->addWildcard(Format::y_resolution_param);
    myInputFormat->addWildcard(Format::ratio_x_param);
    myInputFormat->addWildcard(Format::ratio_y_param);
    myInputFormat->addStringParamValue(Format::colorspace_param,"yv12");
    myInputFormat->addStringParamValue(Format::channelorder_param,"yvu");
    myInputFormat->addIntParamValue(Format::bitperpixel_param,12);
    myInputFormat->addStringParamValue(Format::endian_param,"little");
    myInputFormat->addWildcard(Format::framerate_param);
    myInputFormat->addStringParamValue(Format::format_param,"planar");
    myInputFormat->setJackTag("stream1");
    getOwnInputProperty()->setDefaultFormat(myInputFormat, "stream1");
    
    myInputFormat=getOwnInputProperty()->addNewFormat("*/*");
    myInputFormat->setJackTag("stream2");
    getOwnInputProperty()->setDefaultFormat(myInputFormat, "stream2");
    
    Format* myOutputFormat;
    myOutputFormat=getOwnOutputProperty()->addNewFormat("video/raw");
    myOutputFormat->addWildcard(Format::x_resolution_param);
    myOutputFormat->addWildcard(Format::y_resolution_param);
    myOutputFormat->addWildcard(Format::ratio_x_param);
    myOutputFormat->addWildcard(Format::ratio_y_param);
    myOutputFormat->addStringParamValue(Format::colorspace_param,"yv12");
    myOutputFormat->addStringParamValue(Format::channelorder_param,"yvu");
    myOutputFormat->addIntParamValue(Format::bitperpixel_param,12);
    myOutputFormat->addStringParamValue(Format::endian_param,"little");
    myOutputFormat->addWildcard(Format::framerate_param);
    myOutputFormat->addStringParamValue(Format::format_param,"planar");
    
    myInputFormat->setIOPartner(myOutputFormat);
    myOutputFormat->setIOPartner(myInputFormat);
    
    return SUCCESS;
  }
  
  Result OSDManagerNode::doInitOutput() {
    const Format* in_format = getInputFormat("stream1");
    if (in_format == NULL) {
      ERROR_STREAM("doInitOutput: No input format");
      return FAILURE;
    }

    if ((ratio_x != 0) && (ratio_x != in_format->getIntValue(Format::ratio_x_param))) {
      send_format_changed = true;
    }
    ratio_x = in_format->getIntValue(Format::ratio_x_param);
    if ((ratio_y != 0) && (ratio_y != in_format->getIntValue(Format::ratio_y_param))) {
      send_format_changed = true;
    }
    ratio_y = in_format->getIntValue(Format::ratio_y_param);
    
    if ((image_width != 0) && (image_width != in_format->getIntValue(Format::x_resolution_param))) {
      send_format_changed = true;
    }
    image_width = in_format->getIntValue(Format::x_resolution_param);
    if ((image_height != 0) && (image_height != in_format->getIntValue(Format::y_resolution_param))) {
      send_format_changed = true;
    }
    image_height = in_format->getIntValue(Format::y_resolution_param);

    color = YVA12;

    Format* i_format = new Format( *in_format );
    getOwnWorkingInputProperty()->addNewFormat(i_format);
    
    Format* o_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
    o_format->addIntParamValue(Format::x_resolution_param, image_width);
    o_format->addIntParamValue(Format::y_resolution_param, image_height);
    o_format->addIntParamValue(Format::ratio_x_param, ratio_x);
    o_format->addIntParamValue(Format::ratio_y_param, ratio_y);
    o_format->addStringParamValue(Format::colorspace_param,"yv12");
    o_format->addStringParamValue(Format::channelorder_param,"yvu");
    o_format->addIntParamValue(Format::bitperpixel_param,12);
    o_format->addStringParamValue(Format::endian_param,"little");
    o_format->addFloatParamValue(Format::framerate_param, in_format->getFloatValue(Format::framerate_param));
    o_format->addStringParamValue(Format::format_param,"planar");
    
    if (!message) {
      string font("medium");
      if (getFont(font)) {
        IOSDManager_var manager(getCheckedInterface<IOSDManager>());
        IWidgetFactory_var factory(getCheckedInterface<IWidgetFactory>());
        message = new MessageUnit( manager.get(), factory.get(), font );
      }
    }

    return SUCCESS;
  }
  
  Result OSDManagerNode::doFlush() {
    MutexGuard mg_flush(osd_mutex);
    if (string(getCurrentRecvInputStream())!="stream2") {
      if (lastBuffer != 0) {
        lastBuffer->release();
        lastBuffer = NULL;
      }
    }
    return SUCCESS;
  }
  
  Result OSDManagerNode::doStart() {
    return SUCCESS;
  }
  
  Result OSDManagerNode::doStop() {
    return SUCCESS;
  }  
  
  void OSDManagerNode::addWindow( Window* win ) {
    osd_mutex.lock();
    /* add window to map */
    win->setID(max_window_id++);
    windows[win->getID()] = win;
    //win->setScaleFactor( 0.5, 0.5 );
    osd_mutex.unlock();
  }
  
  void OSDManagerNode::delWindow( Window* win ) {
    osd_mutex.lock();
    /* delete window from map */
    windows.erase(win->getID());
    delete win; // fw: we own the window, so we must delete it
    osd_mutex.unlock();
  }
  
  void OSDManagerNode::addWidget( Widget* win ) {
    win->setID(this, max_widget_id++);
    widgets[win->getID()] = win;
  }
  
  void OSDManagerNode::delWidget( Widget* win ) {
    widgets.erase(win->getID());
    delete win; // fw: we own the widget, so we must delete it
  }
  
  void OSDManagerNode::setFont( string id, BitmapFont* font  ) {
    /* (fw) delete old font if we replace one */
    fonts_t::iterator iter(fonts.find(id));
    if (iter != fonts.end()) {
      delete iter->second;
      iter->second = font;
    } else {
      /* add new font to map */
      fonts[id] = font;
    }
  }
  
  BitmapFont* OSDManagerNode::getFont( string id ) const {
    fonts_t::const_iterator iter(fonts.find(id));
    if (iter == fonts.end()) { return NULL; }
    return iter->second;
  }
  
  void OSDManagerNode::sendPauseEvent() {
    sendSymbolChange("PAUSE");
  }
  
  void OSDManagerNode::sendSymbolChange(const string& sym) {
    osd_mutex.lock();
    send_symbol_change = true;
    new_symbol = sym;
    if ( paused ) setWorkingFlag(true);
    osd_mutex.unlock();
  }
  
  Message* OSDManagerNode::processBuffer(Buffer* in_buffer){
    DEBUG_STREAM("BEGIN: processBuffer" << endl);
    if (send_format_changed) {
      send_format_changed = false;
      format_changed_sent = true;
      save_working_flag = getWorkingFlag();
      if (!save_working_flag) {
        save_in_buffer = in_buffer;
      }
      setWorkingFlag(true);
      setFormat(*getInputFormat("stream1"));
      return new CEvent(IFormatChange::create_setFormat(*getInputFormat("stream1")));
    }

    if (format_changed_sent) {
      format_changed_sent = false;
      if (!save_working_flag) {
        in_buffer = save_in_buffer;
        save_in_buffer = NULL;
      }
      setWorkingFlag(save_working_flag);
    }
    
    MutexGuard guard(osd_mutex);
    if ( send_symbol_change ) {
      if ( !doupdate ) {
        send_symbol_change = false;
        setWorkingFlag(true);
        return new CEvent( ISymbolChange::create_symbolChanged(new_symbol) );
      }
    }

    Timestamp tstamp;
    bool new_timestamp = false;
    if(string(getCurrentRecvInputStream()) == string("stream2")) { // buffer from the "timestamp" input
      if (in_buffer) {
        // get the new timestamp from the "timestamp" input
        tstamp = in_buffer->getTimestamp();
        
        // only update time if time display has changes
        if (old_timestamp.sync_time.sec == tstamp.sync_time.sec) {
          // time-display would not change, nothing to do any more
          in_buffer->release();
          setWorkingFlag(false);
          return NULL;
        } else {
          // save this timestamp as the old one
          old_timestamp = tstamp;
          // flag, that we have a new timestamp to be used
          new_timestamp = true;

          // time-display would have a new value so update it if available
          if (create_time && time) {
            time->setCurrentTime( tstamp.sync_time.sec );
            time->update();
          }
          // errror 20030701: Why setting WorkingFlag true here?
          //                  Answer: it is used internally as a flag,
          //                  that we got a Buffer from "stream2", see
          //                  below
          setWorkingFlag(true);
        }
      } // end if (in_buffer)

    } else if (string(getCurrentRecvInputStream()) == string("stream1")) {
      if (in_buffer) {
        // we got a new buffer, take its timestamp
        tstamp = in_buffer->getTimestamp();

        // only update time if time display has changes
        if (old_timestamp.sync_time.sec != tstamp.sync_time.sec) {
          // save this timestamp as the old one
          old_timestamp = tstamp;
          // we do not have to flag new_timestamp as the in_buffer's
          // timestamp is automatically copied to the outgoing one

          // time-display would have a new value so update it if available
          if (create_time && time) {
            time->setCurrentTime(tstamp.sync_time.sec);
            time->update();
          }
        }
      } // end if (in_buffer)
    }

    // Decide whether are triggered to foreground
    if (getWorkingFlag()) {
      if(in_buffer != 0) {
        in_buffer->release();
        DEBUG_STREAM("HAPPENED!"<<endl);
      }

      setWorkingFlag(false);
      if (lastBuffer != 0 ) {
         out_buffer  = getNewBuffer(lastBuffer->getSize());
         *out_buffer = *lastBuffer;
      }
      else {
        // this is needed if we got a static background from the beginning
        return 0;
      }
    }
    else // triggered to background and new background arrived
    {
      DEBUG_STREAM("getWorkingFlag is false" << endl);
      in_buffer = in_buffer->getWriteableInstance();

      if (lastBuffer != 0) {
        DEBUG_STREAM("REF COUNT before release ! " << lastBuffer->getCount() << endl);
        lastBuffer->release();
        lastBuffer = 0;
      }

      if (paused) {
        // create copy called lastBuffer from current buffer called in_buffer
        lastBuffer = getNewBuffer(in_buffer->getSize());
        *lastBuffer = *in_buffer;
      }

      // Send in_buffer as out_buffer
      out_buffer = in_buffer;
    }

    if ( paused ) {
      new_timestamp = true;
      tstamp = old_timestamp;

      if ( send_symbol_change ) {
        doupdate = false;
        setWorkingFlag(true);
      }
    }

    windows_t::reverse_iterator iter, stop;
    iter = windows.rbegin();
    stop = windows.rend();
    
    while ( iter != stop ) {
      Window* win = iter->second;
      iter++;

      // lock window update, so no window change can be done during overlay
      win->lock();
      int xp, yp, w, h;
      win->setScaleFactor( scale_factor_x, scale_factor_y );
      unsigned char* ws = win->get( &xp, &yp, &w, &h );

      if ( ws ) {
        YV12Aoverlay( ws, xp, yp, w, h, (unsigned char*) out_buffer->getData(), image_width, image_height );
      }
      win->unlock();
    }


    if ( new_timestamp ) out_buffer->setTimestamp( tstamp );

    doupdate = false;
    return out_buffer;
  }
  
  void OSDManagerNode::setPause(const nmm_bool& value) {
    DEBUG_STREAM("setPause " << value << endl );
    MutexGuard guard(osd_mutex);
    paused = value;
  }
  
  Result OSDManagerNode::setFormat(const Format& format) {
    image_width  = format.getIntValue(Format::x_resolution_param);
    image_height = format.getIntValue(Format::y_resolution_param);
    
    scale_factor_x = double(image_width)/reference_width;
    scale_factor_y = double(image_height)/reference_height;
    
    return SUCCESS;
  }
  
  Widget* OSDManagerNode::getOwnWidget(const IWidget& iwidget) {
    // XXX: if the widget or window was not created by this OSD manager, this will fail utterly
    // (generally we can only work with our own widgets and windows)
    return getOwnWidget(iwidget.getWidgetID());
  }
  
  Window* OSDManagerNode::getOwnWindow(const IWindow& iwindow) {
    // XXX: if the widget or window was not created by this OSD manager, this will fail utterly
    // (generally we can only work with our own widgets and windows)
    return getOwnWindow(iwindow.getWindowID());
  }  
  
  Widget* OSDManagerNode::getOwnWidget(int id) {
    widgets_t::iterator iter = widgets.find(id);
    if (iter == widgets.end()) {
      // invalid id
      return NULL;
    }
    return iter->second;
  }
  
  Window* OSDManagerNode::getOwnWindow(int id) {
    windows_t::iterator iter = windows.find(id);
    if (iter == windows.end()) {
      // invalid id
      return NULL;
    }
    return iter->second;
  }
  
  IWidgetComposite* OSDManagerNode::createWidgetComposite() {
    WidgetComposite* widget = new WidgetComposite();
    addWidget(widget);
    return widget->getCheckedInterface<IWidgetComposite>();
  }
  
  IProgressBar* OSDManagerNode::createProgressBar() {
    ProgressBar* widget = new ProgressBar();
    addWidget(widget);
    return widget->getCheckedInterface<IProgressBar>();
  }
  
  ITextView* OSDManagerNode::createTextView() {
    TextView* widget = new TextView();
    addWidget(widget);
    return widget->getCheckedInterface<ITextView>();
  }
  
  ITimeView* OSDManagerNode::createTimeView() {
    TimeView* widget = new TimeView();
    addWidget(widget);
    return widget->getCheckedInterface<ITimeView>();
  }
  
  IImageView* OSDManagerNode::createImageView() {
    ImageView* widget = new ImageView();
    widget->setBitmapReader(bmreader);
    addWidget(widget);
    return widget->getCheckedInterface<IImageView>();
  }
  
  IRadioButtons* OSDManagerNode::createRadioButtons() {
    RadioButtons* widget = new RadioButtons();
    widget->setBitmapReader(bmreader);
    addWidget(widget);
    return widget->getCheckedInterface<IRadioButtons>();
  }
  
  IButtons* OSDManagerNode::createButtons() {
    Buttons* widget = new Buttons();
    widget->setBitmapReader(bmreader);
    addWidget(widget);
    return widget->getCheckedInterface<IButtons>();
  }
  
  IBorderDecorator* OSDManagerNode::createBorderDecorator(const IWidget& internal) {
    BorderDecorator* widget = new BorderDecorator();
    widget->setOwnInternalWidget(getOwnWidget(internal));
    addWidget(widget);
    return widget->getCheckedInterface<IBorderDecorator>();
  }
  
  IHeaderDecorator* OSDManagerNode::createHeaderDecorator(const IWidget& internal) {
    HeaderDecorator* widget = new HeaderDecorator();
    widget->setOwnInternalWidget(getOwnWidget(internal));
    addWidget(widget);
    return widget->getCheckedInterface<IHeaderDecorator>();
  }
  
  IPlaneDecorator* OSDManagerNode::createPlaneDecorator(const IWidget& internal) {
    PlaneDecorator* widget = new PlaneDecorator();
    widget->setOwnInternalWidget(getOwnWidget(internal));
    addWidget(widget);
    return widget->getCheckedInterface<IPlaneDecorator>();
  }
  
  IScrollDecorator* OSDManagerNode::createScrollDecorator(const IWidget& internal) {
    ScrollDecorator* widget = new ScrollDecorator();
    widget->setOwnInternalWidget(getOwnWidget(internal));
    addWidget(widget);
    return widget->getCheckedInterface<IScrollDecorator>();
  }
  
  Result OSDManagerNode::loadFont(const string& id, const string& xmlfilename, const string& fontname) {
    cerr << "Load font " << endl;
    BitmapFont* font = new BitmapFont();
        cerr << "Load font 01" << endl;
    if (FAILURE == font->load(xmlfilename, fontname)) {
      delete font;
      return FAILURE;
    }
    cerr << "Load font2 " << endl;
    
    setFont(id, font);
    cerr << "Load font3 " << endl;

    return SUCCESS;
  }
  
  list<string> OSDManagerNode::getFontList() const {
    list<string> fontnames;
    for (fonts_t::const_iterator iter(fonts.begin()); iter != fonts.end(); ++iter) {
      fontnames.push_back(iter->first);
    }
    
    return fontnames;
  }
  
  Result OSDManagerNode::update(){
    osd_mutex.lock();
    
    doupdate = true;
    if (paused) {
      //if paused, set Flag to get into processBuffer again
      setCurrentSendInputStream( "stream1" );
      setWorkingFlag(true);
      DEBUG_STREAM("DRAW update, paused!" << endl);
    }
    else {
      DEBUG_STREAM("DRAW update, not paused!!!" << endl);
    }
    osd_mutex.unlock();
    
    return SUCCESS;
  }
  
  IWidget* OSDManagerNode::getWidget(const nmm_int32& id) {
    widgets_t::const_iterator iter = widgets.find(id);
    if (iter == widgets.end()) {
      // XXX: throw exception here
    }
    
    Widget* win = iter->second;
    return win->getCheckedInterface<IWidget>();
  }
  
  IWindow* OSDManagerNode::getWindow(const nmm_int32& id) {
    windows_t::const_iterator iter = windows.find(id);
    if (iter == windows.end()) {
      // XXX: throw exception here
    }
    
    Window* win = iter->second;
    return win->getCheckedInterface<IWindow>();
  }
  
  Result OSDManagerNode::displayMessage(const string& header, const string& text, const string& label1, const string& label2 /*= ""*/,
                                        const string& label3 /*= ""*/, const nmm_uint32& sel /*= 0*/) {
    MutexGuard mg(osd_mutex);
    if ( !message ) {
      string font("medium");
      if ( getFont(font) ) {
        IOSDManager_var manager(getCheckedInterface<IOSDManager>());
        IWidgetFactory_var factory(getCheckedInterface<IWidgetFactory>());
        message = new MessageUnit( manager.get(), factory.get(), font );
      } else {
        ERROR_STREAM("Could not instantiate a MessageUnit!" << endl);
        return FAILURE;
      }
    }
    
    message->print( header, text, label1, label2, label3, sel );
    
    return SUCCESS;
  }
  
  Result OSDManagerNode::deleteMessage() {
    if ( !message ) {
      ERROR_STREAM("No MessageUnit available" << endl);
      return FAILURE;
    }

    message->hide();  
    return SUCCESS;
  }
  
  Result OSDManagerNode::displayTime(const string& header /*= ""*/, const nmm_uint32& x /*= 380*/, const nmm_uint32& y /*= 406*/) {
    if ( !time ) {
      string font("medium");
      if (getFont(font)) {
        IOSDManager_var manager(getCheckedInterface<IOSDManager>());
        IWidgetFactory_var factory(getCheckedInterface<IWidgetFactory>());
        time = new TimeUnit( manager.get(), factory.get(), font );
      } else {
        ERROR_STREAM("No Font " << font << " available" << endl);
        create_time = false;
        return FAILURE;
      }
    }

    time->setHeader( header );
    time->setPosition( x, y );
    time->show();
    //    createTime( true );
    
    return SUCCESS;
  }
  
  Result OSDManagerNode::createTime(const nmm_bool& ct) {
    create_time = ct;
    return( SUCCESS );
  }
  
  Result OSDManagerNode::resetTime() {
    if (!time) {
      ERROR_STREAM("No TimeUnit available" << endl);
      return FAILURE;
    }

    time->setCurrentTime( 0 );
    time->update();
    
    return SUCCESS;
  }
  
  Result OSDManagerNode::hideTime() {
    if (!time) {
      ERROR_STREAM("No TimeUnit available" << endl);
      return FAILURE;
    }
    
    time->hide();
    return SUCCESS;
  }

  Result OSDManagerNode::setTime(const nmm_uint64& t) {
    time->setCurrentTime(t);
    time->update();
    return SUCCESS;
  }
  
  nmm_uint64 OSDManagerNode::getTime()
  {
    return time->getCurrentTime();
  }

  Result OSDManagerNode::displaySymbol(const string& symbolstr, const nmm_uint32& x /*= 600*/, const nmm_uint32& y /*= 50*/) {
    if ( !symbol ) {
      IOSDManager_var manager(getCheckedInterface<IOSDManager>());
      IWidgetFactory_var factory(getCheckedInterface<IWidgetFactory>());
      symbol = new SymbolUnit( manager.get(), factory.get(), 64, 64 );
    }

    if ( !symbolstr.empty() ) {
      symbol->setPosition( x, y );
      symbol->show();
      symbol->print( symbolstr );
      
      sendSymbolChange(symbolstr);
    } else {
      sendSymbolChange("");
    }
    
    return SUCCESS;
  }
  
  Result OSDManagerNode::hideSymbol() {
    if ( !symbol ) {
      return FAILURE;
    }
    
    symbol->hide();
    return SUCCESS;
  }
  
  Result OSDManagerNode::displayProgressBar(const string& header /*= ""*/,
                                            const nmm_uint32& x /*= 50*/,
					    const nmm_uint32& y /*= 470*/,
					    const nmm_uint32& width /*= 620*/) {

    string font("medium");

    if ( !bar ) {
      if (getFont(font)) {
        IOSDManager_var manager(getCheckedInterface<IOSDManager>());
        IWidgetFactory_var factory(getCheckedInterface<IWidgetFactory>());
        bar = new ProgressbarUnit( manager.get(), factory.get(), font );
      } else {
        ERROR_STREAM("No font \"" << font << "\" available. Could not create a ProgressBar!" << endl);
        return FAILURE;
      }
    }
      
    bar->setHeader( header );
    bar->setPosition( x, y );
    bar->setWidth(width);
    bar->show();
    return SUCCESS;
  }

  Result OSDManagerNode::hideProgressBar() {
    if ( !bar ) {
      ERROR_STREAM("No ProgressbarUnit available" << endl);
      return FAILURE;
    }

    bar->hide();
    return SUCCESS;
  }

  void OSDManagerNode::setMaxProgress(const nmm_uint64& max) {
    if ( !bar ) {
      ERROR_STREAM("No ProgressbarUnit available" << endl);
      return;
    }

    bar->setMaxProgress( max );
  }
  
  void OSDManagerNode::setProgress(const nmm_uint64& progress) {
    if ( !bar ) {
      ERROR_STREAM("No ProgressbarUnit available" << endl);
      return;
    }

    cerr << "setBarProgress" << progress << endl;
    bar->setProgress( progress );
  }
  
  void OSDManagerNode::updateProgress() {
    if ( !bar ) {
      ERROR_STREAM("No ProgressbarUnit available" << endl);
      return;
    }

    bar->update();
  }

  Result OSDManagerNode::displayColorBar(const nmm_uint32& x /*= 50*/,
					 const nmm_uint32& y /*= 530*/) {
    string font("medium");

    if ( !color_bar ) {
      if (getFont(font)) {
        IOSDManager_var manager(getCheckedInterface<IOSDManager>());
        IWidgetFactory_var factory(getCheckedInterface<IWidgetFactory>());
        color_bar = new ColorBarUnit( manager.get(), factory.get(), font );
      } else {
        ERROR_STREAM("No font \"" << font << "\" available. Could not create a ColorBar!" << endl);
        return FAILURE;
      }
    }
      
    color_bar->setPosition( x, y );
    color_bar->show();

    return SUCCESS;
  }
  
  Result OSDManagerNode::hideColorBar() {
    if ( !color_bar ) {
      ERROR_STREAM("No ColorBarUnit available" << endl);
      return FAILURE;
    }

    color_bar->hide();
    return SUCCESS;
  }

  void OSDManagerNode::setColorBarText(const string& red, const string& green, const string& yellow, const string& blue) {
    if (color_bar) {
      color_bar->setText(red, green, yellow, blue);
    }
  }

  void OSDManagerNode::setRedColorBarText(const string& text) {
    if (color_bar) {
      color_bar->setRedText(text);
    }
  }

  void OSDManagerNode::setGreenColorBarText(const string& text) {
    if (color_bar) {
      color_bar->setGreenText(text);
    }
  }

  void OSDManagerNode::setYellowColorBarText(const string& text) {
    if (color_bar) {
      color_bar->setYellowText(text);
    }
  }

  void OSDManagerNode::setBlueColorBarText(const string& text) {
    if (color_bar) {
      color_bar->setBlueText(text);
    }
  }

  string OSDManagerNode::getRedColorBarText() {
    if (color_bar) {
      return color_bar->getRedText();
    }
    return "";
  }

  string OSDManagerNode::getGreenColorBarText() {
    if (color_bar) {
      return color_bar->getGreenText();
    }
    return "";
  }

  string OSDManagerNode::getYellowColorBarText() {
    if (color_bar) {
      return color_bar->getYellowText();
    }
    return "";
  }

  string OSDManagerNode::getBlueColorBarText() {
    if (color_bar) {
      return color_bar->getBlueText();
    }
    return "";
  }

  Result OSDManagerNode::messageLeft() {
    if ( !message ) {
      ERROR_STREAM("No MessageUnit available" << endl);
      return FAILURE;
    }

    message->selectLeft();
    return( SUCCESS );
  }
  
  Result OSDManagerNode::messageRight() {
    if ( !message ) {
      ERROR_STREAM("No MessageUnit available" << endl);
      return FAILURE;
    }
     
    message->selectRight();
    return( SUCCESS );
  }
  
  string OSDManagerNode::messageID() {
    if ( message ) return message->messageID();
    return( string("") );
  }
  
  IWindow* OSDManagerNode::createWindow() {
    Window* window = new WindowYV12A(this);
    addWindow(window);
    return window->getCheckedInterface<IWindow>();
  }
  
  void OSDManagerNode::deleteWindow(const IWindow& window) {
    delWindow( getOwnWindow(window) );
  }
  
  void OSDManagerNode::deleteWidget(const IWidget& widget) {
    Widget* wg = getOwnWidget(widget);
    widgets.erase(wg->getID());
    delete wg;
  }
  
  IBitmapReader* OSDManagerNode::getBitmapReader() {
    // XXX: (fw) currently the OSDManagerNode implements IBitmapReader
    return getCheckedInterface<IBitmapReader>();
  }
  
  bool OSDManagerNode::loadImage(const string& id, const string& filename, const IBitmapReader::Format& f) {
    BitmapReader::format _f;
    switch (f) {
    case IBitmapReader::RGBA:
      _f = BitmapReader::RGBA;
      break;
    case IBitmapReader::YV12:
      _f = BitmapReader::YV12;
      break;
    case IBitmapReader::YV12A:
      _f = BitmapReader::YV12A;
      break;
    default:
      return false;
    }
    
    return bmreader->loadImage(id, filename, _f);
  }
  
  void OSDManagerNode::getImage(const string& id, nmm_uint32& w, nmm_uint32& h) {
    bmreader->getImage(id, &w, &h);
  }
  
  bool OSDManagerNode::createImage(const string& id, const nmm_uint32& width, const nmm_uint32& height,
                                   const nmm_uint32& color1, const nmm_uint32& color2, const nmm_uint32& color3,
                                   const IBitmapReader::Format& f) {
    BitmapReader::format _f;
    switch (f) {
    case IBitmapReader::RGBA:
      _f = BitmapReader::RGBA;
      break;
    case IBitmapReader::YV12:
      _f = BitmapReader::YV12;
      break;
    case IBitmapReader::YV12A:
      _f = BitmapReader::YV12A;
      break;
    default:
      return false;
    }
    
    return bmreader->createImage(id, width, height, color1, color2, color3, _f);
  }
  
  void OSDManagerNode::freeImage(const string& id) {
    bmreader->freeImage(id);
  }
  
  void OSDManagerNode::freeAll() {
    bmreader->freeAll();
  }
  
  void OSDManagerNode::checkCharDimension(const string& id, const nmm_uchar& c, nmm_uint32& x, nmm_uint32& y) const {
    BitmapFont* font = getFont(id);
    if (font == NULL) return; // XXX
    font->checkDimension(c, &x, &y);
  }
  
  nmm_uint32 OSDManagerNode::getMaxWidth(const string& id) const {
    BitmapFont* font = getFont(id);
    if (font == NULL) return 0; // XXX
    return font->getMaxWidth();
  }
  
  nmm_uint32 OSDManagerNode::getMaxHeight(const string& id) const {
    BitmapFont* font = getFont(id);
    if (font == NULL) return 0; // XXX
    return font->getMaxHeight();
  }
  
  void OSDManagerNode::setLineSpace(const string& id, const nmm_uint32& space) {
    BitmapFont* font = getFont(id);
    if (font == NULL) return; // XXX
    font->setLineSpace(space);
  }
  
  nmm_uint32 OSDManagerNode::getLineSpace(const string& id) const {
    BitmapFont* font = getFont(id);
    if (font == NULL) return 0; // XXX
    return font->getLineSpace();
  }
  
}

#include "nmm/nmm_suffix.hpp"
