/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include "TableUnit.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  TableUnit::TableUnit( IOSDManager* manager, IWidgetFactory* factory, string _font) 
    : OSDElement(manager, factory),
      font(_font) {

    scrollbarEnabled = true;
    createWindows();

  }
  
  TableUnit::~TableUnit() {
    deleteWidgets();
  }
  
  void TableUnit::deleteWidgets() {
    for(list<ITextView*>::iterator i = text.begin(); i != text.end(); i++){
      if ( (*i) ) {
        manager->deleteWidget(*(*i));
        (*i) = NULL;
      }
    }

    if ( composite.get() ) {
      manager->deleteWidget(*composite);
      composite.reset(NULL);
    }
  
    if ( plane.get() ) {
      manager->deleteWidget(*plane);
      plane.reset(NULL);
    }

    if ( border.get() ) {
      manager->deleteWidget(*border);
      border.reset(NULL);
    }

    if ( scroll.get() ) {
      manager->deleteWidget(*scroll);
      scroll.reset(NULL);
    }

    if ( header.get() ) {
      manager->deleteWidget(*header);
      header.reset(NULL);
    }
  }
  
  void TableUnit::setColors( unsigned int _win_active_color, unsigned int _win_inactive_color,
                            unsigned int _sel_active_color, unsigned int _sel_active_color2, unsigned int _sel_inactive_color ) {

    win_active_color   = _win_active_color;
    win_inactive_color = _win_inactive_color;
    sel_active_color   = _sel_active_color;
    sel_active_color2  = _sel_active_color2;
    sel_inactive_color = _sel_inactive_color;
  }
  
  void TableUnit::selectWindow( unsigned int mode ) {
    for(list<ITextView*>::iterator i = text.begin(); i != text.end(); i++){
      if ( mode == 1 ) {
        (*i)->setSelectionColor( sel_active_color2 );
      } else {
        (*i)->setSelectionColor( sel_active_color );
      }
    }
    plane->setPlaneColor( win_active_color );
  }

  void TableUnit::deSelectWindow() {
    for(list<ITextView*>::iterator i = text.begin(); i != text.end(); i++){
      (*i)->setSelectionColor( sel_inactive_color );
    }
    plane->setPlaneColor( win_inactive_color );
  }
  
  void TableUnit::setHeader( string h ) {
    header->setText( h );
  }
  
  void TableUnit::setColumnWidth( unsigned int c, unsigned int w ){
    ColumnList::iterator i = columns.begin();
    advance(i, (int)c); //cast to get rid of compiler-warning: "comparison of unsigned expression >= 0 is always true"
    columns.erase(i);
    columns.insert(i,w);
  }

  void TableUnit::addColumn(unsigned int w){
    ITextView* t = factory->createTextView();
    t->setFont(font);
    t->setSelectionColor( sel_inactive_color );
    text.push_back(t);
    composite->add(*t);

    if(scrollbarEnabled){
      if ( scroll.get() ) {
        composite->remove(*scroll);
        
        manager->deleteWidget(*scroll);
        scroll.reset(NULL);
      
        scroll.reset(factory->createScrollDecorator( *(text.back()) ));
        composite->add(*scroll);
      }
    }

    unsigned int col = columns.back();
    columns.pop_back();
    col -= w;
    columns.push_back(w);
    columns.push_back(col);
    //update the last column-width
  }

  string TableUnit::getLineURL(const unsigned int& i) {
    return text.front()->getLineURL(i);
  }

  void TableUnit::selectLine(const unsigned int& i){
    for(list<ITextView*>::iterator j = text.begin(); j != text.end(); j++){
      (*j)->selectLine(i);
    }
  }

  list<string> TableUnit::getLineContent(const unsigned int& i){
    list<string> ret;
    for(list<ITextView*>::iterator j = text.begin(); j != text.end(); j++){
      ret.push_back((*j)->getLineContent(i));
    }
    return ret;
  }

  string TableUnit::getLineContent(const unsigned int& c, const unsigned int& i){
    list<ITextView*>::iterator j = text.begin();
    advance(j, (int)c); //cast to get rid of compiler-warning: "comparison of unsigned expression >= 0 is always true"
    if(j != text.end()){
      return (*j)->getLineContent(i);
    }else{
      return string("");
    }    
  }

  void TableUnit::setLineContent(const unsigned int c, const unsigned int i, string s){
    list<ITextView*>::iterator j = text.begin();
    advance(j, (int)c); //cast to get rid of compiler-warning: "comparison of unsigned expression >= 0 is always true"
    if(j != text.end()){
      (*j)->deleteLine(i);
      (*j)->insertLine(i-1, s);
      (*j)->selectLine(getSelectedLine());
    }
  }

  void TableUnit::deleteAllLines(){
    //This is necessary to fix the following bug: 
    //If you have a widget containing x lines and selected line x, delete all lines and refill the widget with y lines
    //y < x, then the new lines will not be displayed
    selectLine(0);
    update();

    for(list<ITextView*>::iterator j = text.begin(); j != text.end(); j++){
      (*j)->deleteAllLines();
    }
  }
  unsigned int TableUnit::addLine(list<string> c, string url){
    unsigned int w, h, w1=0;
    unsigned int widgetOffset = 0;
    win->getDimension(w,h);
    int ret = 0;
    
    list<string>::iterator i = c.begin(); //list of text to add
    list<ITextView*>::iterator j = text.begin(); //list of text-widgets()
    for(list<unsigned int>::iterator k = columns.begin(); k != columns.end(); k++){
      w1 = (w * (*k) )/100;
      
      if(j != text.end() && i != c.end()){
        (*j)->setDimension(w1, h);
        (*j)->setPosition(widgetOffset, 0);
        ret = (*j)->addLine(*i, url);
        i++;j++;
        widgetOffset += w1;
      }
    }

    return ret;    
  }
  
  unsigned int TableUnit::addLine(list<string> c){
    return addLine(c, "");
  }

  unsigned int TableUnit::addLine(string c, string url){
    list<string> s;
    s.push_back(c);
    return addLine(s,url);
  }
  unsigned int TableUnit::addLine(string c){
    list<string> s;
    s.push_back(c);
    return addLine(s,"");
  }
  
  void TableUnit::deSelectLine() {
    for(list<ITextView*>::iterator j = text.begin(); j != text.end(); j++){
      (*j)->deSelectLine();
    }
  }

  unsigned int TableUnit::getSelectedLine(){
    return text.front()->getSelectedLine();
  }
  
  Result TableUnit::selectUp(){
    if(text.front()->getSelectedLine() > 0){
      for(list<ITextView*>::iterator j = text.begin(); j != text.end(); j++){
        (*j)->selectUp();
      }    
      return SUCCESS;
    }
    return FAILURE;
  }

  Result TableUnit::selectDown(){
    unsigned int l = text.front()->getNumLines();
    
    if(text.front()->getSelectedLine() < l){
      for(list<ITextView*>::iterator j = text.begin(); j != text.end(); j++){
        (*j)->selectDown();
      }
      return SUCCESS;
    }
    return FAILURE;
  }

  void TableUnit::createWindows(){
    
    setColors( 0xff000000, 0xff100000, 0xff707070, 0xff303030, 0xff101010 );
    
    composite.reset(factory->createWidgetComposite());
    ITextView* t = factory->createTextView();
    t->setFont(font);
    t->setSelectionColor( sel_inactive_color );
    composite->add(*t);
    text.push_back(t);
    columns.push_back(100);

    if(scrollbarEnabled){
      scroll.reset(factory->createScrollDecorator( *(text.back()) ));
      composite->add(*scroll);
    }

    plane.reset(factory->createPlaneDecorator( *composite ));
    plane->setPlaneColor( win_inactive_color );

    border.reset(factory->createBorderDecorator( *plane ));

    header.reset(factory->createHeaderDecorator( *border ));
    header->setFont(font);

    win->setContent( *header );
  }
  void TableUnit::setScrollbarEnabled(bool enabled){

    if(!enabled){
      if ( scroll.get() ) {
        composite->remove(*scroll);
        
        manager->deleteWidget(*scroll);
        scroll.reset(NULL);      
      }
    }

    scrollbarEnabled = enabled;
  }
  
  int TableUnit::getNumLines(){
    if(!text.empty()){
      return text.front()->getNumLines();
    }
    
    return 0;
  }
}

#include "nmm/nmm_suffix.hpp"
