/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include "TextView.hpp"
#include "BitmapFont.hpp"
#include "OSDManagerNode.hpp"

#include "overlay.h"

/*****************************************************************************************/
/* TextView									           */
/*****************************************************************************************/

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  TextView::TextView() {
    reset();
    setSelectionColor( 0xff808080 );
    setSelectionColor2( 0xff00ffff );
  }

  TextView::~TextView() {
    deleteAllLines();
  }
  

  int TextView::add( content_t con ) {
    lines.push_back( con );

    /* get number of lines */
    int s = lines.size();

    /* return index of added text */
    return(s);
  }

  void TextView::insert( int pos, content_t con ) {
    //line_t::iterator it = (line_t::iterator)&lines[pos];
    //lines.insert( it, con );
    line_t::iterator it = lines.begin();
    advance(it, pos);
    lines.insert(it, con);
  }
  
  TextView::content_t& TextView::getLine( unsigned int line ) {
    return( lines[line-1] );
  }
  
  void TextView::setLineSpace( unsigned int ls ) {
    line_space = ls;
  }
  
  unsigned int TextView::getLineSpace() {
    return( line_space );
  }
  
  void TextView::draw( workspace_t* ws ) {
    unsigned int ch;        // font height
    unsigned int ls;        // number of textlines
    unsigned int linespace; // space between 2 lines
    unsigned int v_width, v_height;

    v_width  = adjustDimension( xpos, width, ws->width );
    v_height = adjustDimension( ypos, height, ws->height );

    /* because the line selection routine needs the whole character */
    /* we need the maximum height of a character */
    ch = font->getMaxHeight();
    linespace = getLineSpace();

    setMaxLines( v_height/(ch+linespace) );
    if ( autofocus ) focusSelection();

    unsigned int end_line = getMaxLines() + getStartLine();

    /* get number of text lines */
    ls = lines.size();
    if ( ls > end_line ) ls = end_line;

    unsigned int  y = ypos;
    unsigned int* frame = (unsigned int*) ws->data;
    for( unsigned int l=getStartLine()-1; l<ls; l++ ) {

      /* BEGIN: LINE Selection */

      /* if selected_line is null, then no line will be marked */
      if ( (l+1)==selected_line ) {

	      unsigned int* dest = frame+y*ws->stride;
	      /* process the whole character height */
	      for( unsigned int g=0; g<ch; g++ ) {
	        if ( g+y >= (ypos+v_height) ) break;
	        /* fill the line with the selected color */
	        for( unsigned int k=xpos; k<(xpos+v_width); k++ ) {
	          dest[k+g*ws->stride] = selection_color;
	        }
	      }
      } else {
        /* if selected_line2 is null, then no (second) line will be highlighted */
        if ( (l+1)==selected_line2 ) {

	        unsigned int* dest = frame+y*ws->stride;
	        /* process the whole character height */
	        for( unsigned int g=0; g<ch; g++ ) {
	          if ( g+y >= (ypos+v_height) ) break;
	          /* fill the line with the selected color */
	          for( unsigned int k=xpos; k<(xpos+v_width); k++ ) {
	            dest[k+g*ws->stride] = selection_color2;
	          }
	        }
        }else{
          for(list<unsigned int>::iterator j = marked_lines.begin(); j != marked_lines.end(); j++){
            if( (l+1) == *j ){

	        unsigned int* dest = frame+y*ws->stride;
	        /* process the whole character height */
	        for( unsigned int g=0; g<ch; g++ ) {
	          if ( g+y >= (ypos+v_height) ) break;
	          /* fill the line with the selected color */
	          for( unsigned int k=xpos; k<(xpos+v_width); k++ ) {
	            dest[k+g*ws->stride] = selection_color2;
	          }
	        }
            } 
          }
        }
      }
      /* END: LINE Selection */

      if ( textdraw.draw( lines[l].text.c_str(), font, frame, ws->stride,
			  xpos, y,
			  v_width+xpos, v_height+ypos ) ) return;



      /* go to next line */
      y+=ch+linespace;
    }
  }
  
  
  nmm_int32 TextView::getScrollBarYPosition() const {
    unsigned int sl, nl;
    
    sl = getStartLine()-1; // get visible start line
    nl = getNumLines();    // get number of lines
    
    if ( (nl == 0) || (sl>nl) ) sl = 0;
    else sl = (sl*10000)/nl;
    return( sl );
  }
  
  int TextView::getScrollBarHeight() const {
    unsigned int nl, ml;
    
    ml = getMaxLines(); // get number of maximum visible lines
    nl = getNumLines(); // get number of lines
    
    if ( (nl == 0) || (ml>nl) ) ml = 10000;
    else ml = (ml*10000)/nl;
    
    return( ml );
  }
  
  void TextView::selectUp() {
    selectLine( selected_line-1 );
  }
  
  void TextView::selectDown() {
    selectLine( selected_line+1 );
  }
  
  void TextView::setFont(const string& value) {
    GLOBAL_DEBUG_STREAM("TextView::setFont()" << endl);
    font = manager->getFont(value);

    GLOBAL_DEBUG_STREAM("got font" << endl);

    /* set linespace to value taken from font */
    setLineSpace( font->getLineSpace() );
    GLOBAL_DEBUG_STREAM("TextView::setFont() finished" << endl);
  }
  
  nmm_uint32 TextView::selectLine(const nmm_uint32& _line) {
    unsigned int line = _line;
    
    unsigned int s = lines.size();
    
    if ( line > s ) line = s;
    if ( (line < 1) ) {
      if ( s > 0 ) line = 1;
      else line = 0;
    }
    
    selected_line = line;
    
    return( line );
  }
  
  nmm_uint32 TextView::selectLine2(const nmm_uint32& _line) {
    unsigned int line = _line;
    unsigned int s = lines.size();
    
    if ( line > s ) line = s;
    if ( (line < 1) ) {
      if ( s > 0 ) line = 1;
      else line = 0;
    }
    
    selected_line2 = line;
    
    return( line );
  }
  
  void TextView::deSelectLine() {
    selected_line = 0;
  }
  
  void TextView::deSelectLine2() {
    selected_line2 = 0;
  }
  
  nmm_uint32 TextView::getSelectedLine() const {
    return( selected_line );
  }
  
  nmm_uint32 TextView::getSelectedLine2() const {
    return( selected_line2 );
  }
  
  string TextView::getLineContent(const nmm_uint32& line) const {
    if ( line > 0 ) return( lines[line-1].text );
    else return( "" );
  }
  
  string TextView::getLineURL(const nmm_uint32& line) const {
    if ( line > 0 ) return( lines[line-1].url );
    else return( "" );
  }
  
  IWidget* TextView::getLineWidget(const nmm_uint32& line) const {
    if ( line > 0 ) {
      Widget* widget = (Widget*) lines[line-1].common;
      if (widget != NULL) {
        return( widget->getCheckedInterface<IWidget>() );
      } else {
        return( NULL );
      }
    } else {
      return NULL;
    }
  }
  
  void TextView::setAutoFocus(const nmm_bool& value) {
    autofocus = value;
  }
  
  void TextView::setSelectionColor(const nmm_uint32& color) {
    selection_color = color;
  }
  
  void TextView::setSelectionColor2(const nmm_uint32& color) {
    selection_color2 = color;
  }
  
  void TextView::reset() {
    deSelectLine();
    deSelectLine2();
    setAutoFocus( true );
    
    start_line = 1;
    setMaxLines( lines.size() );
  }
  
  void TextView::setMaxLines(const nmm_uint32& value) {
    unsigned int maxl = value;
    
    if ( maxl < 1 ) maxl = 1;
    max_lines = maxl-1;
  }
  
  nmm_uint32 TextView::getMaxLines() const {
    return( max_lines+1 );
  }
  
  nmm_uint32 TextView::setStartLine(const nmm_uint32& _start) {
    unsigned int start = _start;
    
    if ( (start+max_lines) > lines.size() ) start = lines.size()-max_lines;
    if ( start < 1 ) start = 1;
    
    start_line = start;
    return( start_line );
  }
  
  nmm_uint32 TextView::getStartLine() const {
    return( start_line );
  }
  
  nmm_uint32 TextView::getNumLines() const {
    return( lines.size() );
  }
  
  void TextView::focusSelection() {
    unsigned int end_line = max_lines + start_line;
    /* adjust the visible lines*/
    if ( selected_line ) {
      if ( selected_line < start_line ) {
	setStartLine( selected_line );
      }
      if ( selected_line >= end_line ) {
	setStartLine( selected_line - max_lines );
      }
    }
  }
  
  void TextView::insertLine(const nmm_int32& pos, const string& text, const string& url /*= ""*/) {
    content_t con;
    con.text   = text;
    con.url    = url;
    con.common = NULL;
    
    insert( pos, con );
  }
  
  nmm_int32 TextView::addLine(const string& text, const string& url /*= ""*/) {
    content_t con;
    con.text   = text;
    con.url    = url;
    con.common = NULL;
    
    return( add( con ) );
  }
  
  void TextView::addLineWidget(const nmm_uint32& line, const IWidget& widget) {
    if ( line > 0 ) {
      lines[line-1].common = manager->getOwnWidget( widget );
    }
  }
  
  void TextView::deleteLine(const nmm_uint32& line) {
    unsigned int pos = line;
    
    if ( pos > 0 ) {
      pos--;
      //line_t::iterator it = (line_t::iterator)&lines[pos];
      line_t::iterator it = lines.begin();
      advance(it, pos);
      lines.erase( it );
      
      if ( pos > 0 ) pos++;
      selectLine( pos );
    }
  }
  
  void TextView::deleteAllLines() {
    lines.clear();
    marked_lines.clear();
    selected_line  = 0;
    selected_line2 = 0;
  }

  void TextView::changeLineContent(const nmm_uint32& line, const string& text){
    if ( (line > 0) && (line <= lines.size()) )  lines[line-1].text = text;
  }

  void TextView::markLine(const nmm_uint32& line){
    marked_lines.push_back(line);
    marked_lines.unique();
  }
  
  void TextView::unmarkLine(const nmm_uint32& line){
    marked_lines.remove(line);
    marked_lines.unique();
  }
  
  list<nmm_uint32> TextView::getMarkedLines(){
    return marked_lines;
  }


}


#include "nmm/nmm_suffix.hpp"
