/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include "Widget.hpp"
#include "OSDManagerNode.hpp"

#include <limits.h>

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  Widget::Widget() {
    manager = NULL;
    setPosition( 0, 0 );
    setDimension( 0, 0 );
  }
  
  void Widget::connectTo( string id, class Callback* cb ) {
    /* add callback to the map */
    callbacks[id].push_back( cb );
  }
  
  void Widget::disconnect( string id ) {
    callbacklist_t*          cblist;

    /* try to get a list with callbacks */
    cblist = &callbacks[id];

    if ( cblist ) {
      /* remove all callbacks of the list */
      cblist->clear();

      /* remove list from the map */
      callbacks.erase(id);
    }


  }
  
  void Widget::startCallbacks( string id ) {
    callbacklist_t::iterator iter, stop;
    callbacklist_t*          cblist;
    Callback*		   cb;

    /* try to get a list with callbacks */
    cblist = &callbacks[id];

    if ( cblist ) {
      iter = cblist->begin();
      stop = cblist->end();

      /* start all callbacks of the list */
      while( iter != stop ) {
	cb = *iter;
	iter++;
	cb->run( (void*) this );

      }
    }
  }
  
  unsigned int Widget::adjustDimension( unsigned int xy, unsigned int wh, unsigned int wh_dest ) {
    /* if x/y coordinate gets out of width/height, return a width/height of null */
    if ( xy > wh_dest ) return( 0 );

    /* if no width/height is specified, return maximum width/height */
    if ( wh == 0 ) return( wh_dest-xy );

    if ( (xy + wh) < wh_dest ) return( wh );
    else return( wh_dest-xy );
  }
  
  nmm_int32 Widget::getScrollBarHeight() const {
    return 0;
  }
  
  nmm_int32 Widget::getScrollBarWidth() const {
    return 0;
  }
  
  nmm_int32 Widget::getScrollBarYPosition() const {
    return 0;
  }
  
  nmm_int32 Widget::getScrollBarXPosition() const {
    return 0;
  }
  
  nmm_int32 Widget::getWidgetID() const {
    return getID();
  }
  
  void Widget::selectLeft() {
    startCallbacks( "selectLeft" );
  }
  
  void Widget::selectRight() {
    startCallbacks( "selectRight" );
  }
  
  void Widget::selectUp() {
    startCallbacks( "selectUp" );
  }
  
  void Widget::selectDown() {
    startCallbacks( "selectDown" );
  }
  
  void Widget::selectPush() {
    startCallbacks( "selectPush" );
  }
  
  void Widget::setPosition( const nmm_uint32& x, const nmm_uint32& y ) {
    xpos = x;
    ypos = y;
  }
  
  void Widget::setDimension( const nmm_uint32& w, const nmm_uint32& h ) {
    width  = w;
    height = h;
  }
  
  void Widget::getPosition( nmm_uint32& x, nmm_uint32& y ) const {
    x = xpos;
    y = ypos;
  }
  
  void Widget::getDimension( nmm_uint32& w, nmm_uint32& h ) const {
    w = width;
    h = height;
  }
  
  
  WidgetComposite::WidgetComposite() {
    
  }
  
  WidgetComposite::~WidgetComposite() {
    removeAll();
  }
  
  void WidgetComposite::add( const IWidget& wg ) {
    widgetlist.push_back( manager->getOwnWidget(wg) );
  }

  void WidgetComposite::remove( const IWidget& wg ) {
    list<Widget*>::iterator i, end;
    
    Widget* widget = manager->getOwnWidget(wg);
    
    i   = widgetlist.begin();
    end = widgetlist.end();
    
    while( i != end ) {
      if ( (*i)== widget ) { widgetlist.erase( i ); break; }
      i++;
    }
  }
  
  void WidgetComposite::removeAll() {
    widgetlist.clear();
  }
  
  void WidgetComposite::draw( workspace_t* ws ) {
    list<Widget*>::iterator i, end;
    workspace_t workspace;

    i   = widgetlist.begin();
    end = widgetlist.end();

    while( i != end ) {
      workspace.data   = ws->data+(xpos<<2)+(ypos*ws->stride<<2);
      workspace.width  = ws->width - xpos;
      workspace.height = ws->height - ypos;
      workspace.stride = ws->stride;

      (*i)->draw( &workspace );
      i++;
    }
  }
}

#include "nmm/nmm_suffix.hpp"
