/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include "Window.hpp"
#include "OSDManagerNode.hpp"
#include "nmm/utils/time/Clock.hpp"
#include "nmm/utils/time/Time.hpp"
#include "converter.h"
#include "Widget.hpp"
#include "nmm/nmm_prefix.hpp"

namespace NMM {

  Window::Window(OSDManagerNode* _manager) {
    manager          = _manager;
    vc               = NULL;
    workspace.data   = NULL;
    workspace.width  = 0;
    workspace.height = 0;
    workspace.stride = 0;
    workspace.format = 0;
    xpos             = 0;
    ypos             = 0;
    id               = -1;
    setUpdateFlag( false );
    unHide();
    pthread_mutex_init(&mutex, 0);
  }

  Window::~Window() {
    remove();
    pthread_mutex_destroy(&mutex);
  }

  void Window::setContentInternal( Widget* mvc ) {
    vc = mvc;
  }

  bool Window::getUpdateFlag() {
    return( updateflag );
  }

  void Window::setUpdateFlag( bool b) {
    updateflag = b;
  }

  void Window::lock() {
    pthread_mutex_lock(&mutex);
  }

  void Window::unlock() {
    pthread_mutex_unlock(&mutex);
  }

  void Window::update() {
    pthread_mutex_lock(&mutex);
    
    setUpdateFlag( true );
    
    /* reset workspace content */
    memset( workspace.data, 0, workspace.width*workspace.height*4 );
    
    /* call draw function of the Widget */
    if ( vc ) vc->draw( &workspace );
    
    pthread_mutex_unlock(&mutex);
  }

  bool Window::isHidden() {
    return( hidden );
  }

  void Window::hide() {
    hidden = true;
  };

  void Window::unHide() {
    hidden = false;
  };

  bool Window::create( int w, int h ) {
    /* delete existing workspace */
    deleteWorkSpace();

    /* create new workspace */
    bool s = createWorkSpace( w, h );

    return( s );
  }

  void Window::remove() {
    deleteWorkSpace();
  }

  bool Window::createWorkSpace( int w, int h ) {
    int framesize = (int) (w*h*4);

    workspace.data = new unsigned char[framesize];

    if ( workspace.data ) {
      memset( workspace.data, 0, framesize );
      workspace.width  = w;
      workspace.height = h;
      workspace.stride = w;
      return true;
    }
    else return false;
  }

  void Window::deleteWorkSpace() {
    if ( workspace.data ) delete[] workspace.data;
    workspace.data   = NULL;
    workspace.width  = 0;
    workspace.height = 0;
    workspace.stride = 0;
    workspace.format = 0;
  }

  void Window::setPosition( const nmm_uint32& x, const nmm_uint32& y ) {
    xpos = x;
    ypos = y;
  }
  
  void Window::getPosition( nmm_uint32& x, nmm_uint32& y ) const {
    x = xpos;
    y = ypos;
  }
  
  void Window::setDimension(const nmm_uint32& w, const nmm_uint32& h) {
    create(w,h);
  }
  
  void Window::getDimension( nmm_uint32& w, nmm_uint32& h ) const {
    w = workspace.width;
    h = workspace.height;
  }
  
  nmm_bool Window::getVisible() const {
    return !hidden;
  }
  
  void Window::setVisible(const nmm_bool& value) {
    if (value != !hidden) {
      if ( value ) {
        unHide();
      } else {
        hide();
      }
    }
  }
  
  unsigned char* Window::get( int* x, int* y, int* w, int *h ) {
    *w = workspace.width;
    *h = workspace.height;
    *x = xpos;
    *y = ypos;

    if ( isHidden() ) return NULL;
    else return( workspace.data );
  }

  WindowYV12A::WindowYV12A(OSDManagerNode* manager) : Window(manager) {
    yv12a_workspace.data   = NULL;
    yv12a_workspace.width  = 0;
    yv12a_workspace.height = 0;
    yv12a_workspace.stride = 0;
    yv12a_workspace.format = 0;

    scaled_data          = NULL;
    scale_factor_x       = 1;
    scale_factor_y       = 1;
    scale_factor_changed = false;
    scaler.setFormat( VideoScaler::FORMAT_YV12A );
    //scaler.setFiltering( true );
  }

  WindowYV12A::~WindowYV12A() {
    if ( scaled_data ) delete[] scaled_data;
  }

  void WindowYV12A::setScaleFactor( const nmm_double& scale_x, const nmm_double& scale_y ) {
    if ( (scale_factor_x != scale_x) || (scale_factor_y != scale_y) ) scale_factor_changed = true;

    scale_factor_x = scale_x;
    scale_factor_y = scale_y;

    if ( (scale_factor_x == 1) && (scale_factor_y == 1) ) {
      if ( scaled_data ) {
        delete[] scaled_data;
        scaled_data = NULL;
      }
    }
  }

  unsigned char* WindowYV12A::get( int* x, int* y, int* w, int *h ) {
    if ( isHidden() ) return NULL;
    unsigned char* data;

    if ( (scale_factor_x != 1) || (scale_factor_y != 1) ) {
      *x = (int) (xpos*scale_factor_x);
      *y = (int) (ypos*scale_factor_y);

      *w = (unsigned int) (yv12a_workspace.width*scale_factor_x);
      *h = (unsigned int) (yv12a_workspace.height*scale_factor_y);

      if ( (*w)%2 != 0 ) (*w)++;
      if ( (*h)%2 != 0 ) (*h)++;
      data = scaled_data;
    } else {
      *x = xpos;
      *y = ypos;

      *w = yv12a_workspace.width;
      *h = yv12a_workspace.height;

      data = yv12a_workspace.data;
    }

    if ( yv12a_workspace.data && ((getUpdateFlag() == true) || scale_factor_changed) ) {
      scale_factor_changed = false;
      //cout << "WINDOW: Converting Window to YV12A..." << endl;
      convertRGBAtoYV12A( yv12a_workspace.data, workspace.data, workspace.width, workspace.height );
      setUpdateFlag( false );

      if ( (scale_factor_x != 1) || (scale_factor_y != 1) ) {
        if ( scaled_data ) delete[] scaled_data;
        int framesize = (int) ( (*w) * (*h) * 2.5 );
        scaled_data = new unsigned char[framesize];
        data = scaled_data;

        scaler.setInputSize( yv12a_workspace.width, yv12a_workspace.height );
        scaler.setOutputSize( *w, *h );
        scaler.processData( (char*) scaled_data, (char*) yv12a_workspace.data );
      }
    }

    if ( (*x)%2 != 0 ) (*x)++;
    if ( (*y)%2 != 0 ) (*y)++;

    return( data );

  }


  bool WindowYV12A::createWorkSpace( int width, int height ) {
    /* create "normal" rgba workspace */
    if ( Window::createWorkSpace( width, height ) == false ) return false;

    /* 1 y-plane + 1/4 u-plane + 1/4 v-plane + 1 alpha plane */
    int framesize = (int) ( width * height * 2.5 );

    /* get memory for yv12a workspace */
    yv12a_workspace.data = new unsigned char[framesize];

    if ( yv12a_workspace.data ) {
      yv12a_workspace.width  = width;
      yv12a_workspace.height = height;
      yv12a_workspace.stride = width;
      yv12a_workspace.format = 0;
      return true;
    }
    else return false;

  }

  void WindowYV12A::deleteWorkSpace() {
    /* delete rgba workspace */
    Window::deleteWorkSpace();

    /* delete yuv12a workspace */
    if ( yv12a_workspace.data ) delete[] yv12a_workspace.data;

    yv12a_workspace.data   = NULL;
    yv12a_workspace.width  = 0;
    yv12a_workspace.height = 0;
    yv12a_workspace.stride = 0;
    yv12a_workspace.format = 0;
  }
  
  void WindowYV12A::convertRGBAtoYV12A( unsigned char* yuv_buffer, unsigned char* rgbbuffer, int width, int height ) {
    
    NMM::Time t1, t2; 
    Clock clock;
    t1 = clock.getTime();
    GLOBAL_DEBUG_STREAM("WindowYV12A::convertRGBAtoYV12A: " << t1 <<endl);
    
    RGBAtoYV12A(yuv_buffer, rgbbuffer, width, height);
   
    t2 = clock.getTime();
    GLOBAL_DEBUG_STREAM("WindowYV12A::convertRGBAtoYV12A: " << t2 << endl);
//    GLOBAL_DEBUG_STREAM("WindowYV12A::convertRGBAtoYV12A: timeelapsed = " << t2-t1 << endl); 
    
  }
  
  int Window::getWindowID() const {
    return getID();
  }
  
  void Window::setContent(const IWidget& value) {
    // convert to local widget pointer and set content internally
    setContentInternal(manager->getOwnWidget(value));
  }
  
  IWidget* Window::getContent() const {
    // return an interface to the widget
    if (vc == NULL) {
      return NULL;
    } else {
      return vc->getInterface<IWidget>();
    }
  }
}

#include "nmm/nmm_suffix.hpp"
