/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "IBitmapReader.hpp"
#include "nmm/comm/serialize/SerializeResult.hpp"
#include "nmm/comm/serialize/InvalidTypeException.hpp"
#include "nmm/comm/serialize/TConstructor.hpp"

#include "nmm/nmm_prefix.hpp"


namespace NMM
{
  void IBitmapReader::used() {} 
  
  
  nmm_bool IBitmapReader::loadImage(const string& id, const string& filename, const IBitmapReader::Format& f /* = IBitmapReader::RGBA */)
      throw(const NMM::SerializedException&)
  {
    NMM::TOutValue<nmm_bool > return_value;
    NMM::TInValue<string > id_value(id, false);
    NMM::TInValue<string > filename_value(filename, false);
    NMM::TInValue<IBitmapReader::Format > f_value(f, false);
    NMM::Event event("IBitmapReader::loadImage", &id_value, &filename_value, &f_value);
    event.setReturnValue(&return_value);
    if(const_cast<IBitmapReader*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
    return return_value.getRValue();
  }
  
  void IBitmapReader::getImage(const string& id, nmm_uint32& w, nmm_uint32& h)
      throw(const NMM::SerializedException&)
  {
    NMM::TInValue<string > id_value(id, false);
    NMM::TOutValue<nmm_uint32 > w_value(w);
    NMM::TOutValue<nmm_uint32 > h_value(h);
    NMM::Event event("IBitmapReader::getImage", &id_value, &w_value, &h_value);
    if(const_cast<IBitmapReader*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
    w = w_value.getRValue();
    h = h_value.getRValue();
  }
  
  nmm_bool IBitmapReader::createImage(const string& id, const nmm_uint32& width, const nmm_uint32& height, const nmm_uint32& color1, const nmm_uint32& color2, const nmm_uint32& color3, const IBitmapReader::Format& f /* = IBitmapReader::RGBA */)
      throw(const NMM::SerializedException&)
  {
    NMM::TOutValue<nmm_bool > return_value;
    NMM::TInValue<string > id_value(id, false);
    NMM::TInValue<nmm_uint32 > width_value(width, false);
    NMM::TInValue<nmm_uint32 > height_value(height, false);
    NMM::TInValue<nmm_uint32 > color1_value(color1, false);
    NMM::TInValue<nmm_uint32 > color2_value(color2, false);
    NMM::TInValue<nmm_uint32 > color3_value(color3, false);
    NMM::TInValue<IBitmapReader::Format > f_value(f, false);
    NMM::Event event("IBitmapReader::createImage", &id_value, &width_value, &height_value, &color1_value, &color2_value, &color3_value, &f_value);
    event.setReturnValue(&return_value);
    if(const_cast<IBitmapReader*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
    return return_value.getRValue();
  }
  
  void IBitmapReader::freeImage(const string& id)
      throw(const NMM::SerializedException&)
  {
    NMM::TInValue<string > id_value(id, false);
    NMM::Event event("IBitmapReader::freeImage", &id_value);
    if(const_cast<IBitmapReader*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
  }
  
  void IBitmapReader::freeAll()
      throw(const NMM::SerializedException&)
  {
    NMM::Event event("IBitmapReader::freeAll");
    if(const_cast<IBitmapReader*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
  }

  const char* IBitmapReader::IBitmapReader_type = "IBitmapReader";
  NMM::TConstructor<IBitmapReader > __IBitmapReader_prototype("IBitmapReader","NMM");
  TValueConstructor< IBitmapReader::Format > __IBitmapReader_Format_prototype(TypeInfo("IBitmapReader::Format", "NMM"));

  NMM::OValueStream& operator<<(NMM::OValueStream& ov, const IBitmapReader::Format& t)
  {
    ov.beginType(TypeInfo("IBitmapReader::Format", "NMM"));
    nmm_int32 temp = static_cast<nmm_int32>(t);
    ov << temp;
    ov.endType();
    return ov;
  }

  NMM::IValueStream& operator>>(NMM::IValueStream& iv, IBitmapReader::Format& t)
  {
    if(iv.getCurrentTypeInfo() != TypeInfo("IBitmapReader::Format", "NMM")) {
      throw NMM::InvalidTypeException(TypeInfo("IBitmapReader::Format", "NMM"), iv.getCurrentTypeInfo());
    }
    iv.goDown();
    nmm_int32 temp;
    iv >> temp;
    t = static_cast<IBitmapReader::Format>(temp);
    iv.goUp();
    return iv;
  }



  const char* IBitmapReader::loadImage_event = "IBitmapReader::loadImage";
  const char* IBitmapReader::getImage_event = "IBitmapReader::getImage";
  const char* IBitmapReader::createImage_event = "IBitmapReader::createImage";
  const char* IBitmapReader::freeImage_event = "IBitmapReader::freeImage";
  const char* IBitmapReader::freeAll_event = "IBitmapReader::freeAll";


}


#include "nmm/nmm_suffix.hpp"

