/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "nmm/multimedia/interfaces/ITrack.hpp"
#include "nmm/multimedia/sync/ISynchronized.hpp"
#include "nmm/multimedia/ISyncReset.hpp"
#include "nmm/utils/SecurityCheck.hpp"
#include "nmm/multimedia/IDevice.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"
#include "nmm/utils/directory_utils.hpp"
#include "nmm/multimedia/FlushEvent.hpp"
#include "nmm/multimedia/TDevicePlugin.hpp"

// #include "nmm/misc/dvb/channel.hpp"

#include "DVBReadNode.hpp"
#include "vdr/DVBReadNodeTVGuideManager.hpp"
#include <string>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <fstream>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include <sys/mman.h>

#include <sys/ioctl.h>

#include <linux/videodev.h>

// Primary DVB Control-Funktions
#include "vdr/dvbapi.h"
// Needed to setup LNB-Frequency, DiSeq and the channels
#include "vdr/vdrconfig.h"
#include "vdr/recording.h"
#include "vdr/tools.h"
#include "vdr/videodir.h"
//Eit-Facilities
#include "vdr/eit.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#define ATTMP(a) DEBUG_STREAM( "Attempting " << a << endl);

static NMM::TDevicePlugin< NMM::DVBReadNode > plugin("/dev/video", false, "DVBReadNode", 1);
NMM_REGISTER_SPECIAL_PLUGIN(DVBReadNode, NMM, plugin)

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  DVBReadNode::DVBReadNode(const char* name /* = "DVBReadNode" */,
			   StreamQueue::Mode mode /* = StreamQueue::MODE_SUSPEND */,
			   int queue_size /* = StreamQueue::DEFAULT_SIZE */)
    : GenericSourceNode(name, mode, queue_size),
      dvb_is_initialized(false),
      node_count(0),
      sync_event_sent(false),
      frame(0),
      v_fd(-1),
      device("/dev/video0"),
      stream_timer()
  {
    setBufferManager(0);

    // member initialization
    format_mpeg = false;
    format_raw  = false;
    is_producing_mpeg_buffers = switched_channel = dvb_raw_is_initialized = false;
    buffers_sent = 0;

    send_channel_name_yuv = false;

    // count node-instances
    init_mutex.lock();
    node_count++;
    init_mutex.unlock();
    running = false;

    eitScanner = new VDR::cEITScanner();

    channelConfigFile = getNMMResourceDir() + "/dvb/channels.conf";
  }

  DVBReadNode::~DVBReadNode() {

    // mute audio
    VDR::cDvbApi::muteAudio(true);

    init_mutex.lock();

    // release video-4-linux device
    if (dvb_raw_is_initialized) {
      DEBUG_STREAM("closing video4linux device" << endl);
      if (-1 != v_fd) {
        close(v_fd);
        v_fd = -1;
      }
      dvb_raw_is_initialized = false;
    }

    // stop dvbapi
    if (is_producing_mpeg_buffers) {
      produceMPEG();
    }
    // release buffers that have not been sent
   clearQueue();

    if (node_count == 1) {  // only the last node-instance should do cleanup
      // Cleanup
      VDR::cDvbApi::Cleanup();
      dvb_is_initialized = false;
    }
    node_count--;
    init_mutex.unlock();
  }

  Result DVBReadNode::doDeactivate() {
    DEBUG_STREAM("doDeactivate" << endl);
    if (is_producing_mpeg_buffers) {  // stop dvbapi
      produceMPEG();
    }
    running = false;


    return SUCCESS;
  }

  void DVBReadNode::save() {
    VDR::Setup.CurrentChannel = VDR::cDvbApi::CurrentChannel();
    VDR::Setup.Save();
  }

  Result DVBReadNode::doFlush() {
    clearQueue();
    return SUCCESS;
  }

  Result DVBReadNode::doActivate() {

    MutexGuard init_guard(init_mutex);

    int one=1;
    int i;
    int frame_size;

    DEBUG_STREAM("doActivate" << endl);

    const Format* out_format = getOutputFormat();
    if (!out_format) {
      return FAILURE;
    }
    if (out_format->getFormatType() == string("video/raw")) {
      width = getOutputFormat()->getIntValue(Format::x_resolution_param);
      height = getOutputFormat()->getIntValue(Format::y_resolution_param);
    } else {
      width = 720;
      height = 576;
    }

    vidwin.width = width;
    vidwin.height = height;

    DEBUG_STREAM("picture-size = " << width << " x " << height << endl);



    DEBUG_STREAM("Format set to \"" << getOutputFormat()->getFormatType() << "\"" << endl);

    if (getOutputFormat()->getFormatType() != "av/mpegps")  {
      // YUV
      if (!dvb_raw_is_initialized) {
        DEBUG_STREAM( "using RAW YUV output" << endl);

        format_raw = true;

        // no need to collect MPEG-Buffers if this is the only DVBReadNode
        if (node_count == 1) {
          VDR::cDvbApi::PrimaryDvbApi->disableBufferCollecting();
        }

        // open and setup video-4-linux device
        ATTMP("open");

        vidwin.x=0;
        vidwin.y=0;
        vidwin.chromakey=0;
        vidwin.flags=0;
        vidwin.clipcount=0;
        ATTMP("set win");
        if (ioctl(v_fd, VIDIOCSWIN, &vidwin) < 0) {
          return FAILURE;
        }
        ATTMP("get pict");
        if ( ioctl(v_fd, VIDIOCGPICT, &vidpict) < 0) {
          return FAILURE;
        }
        vidpict.palette=VIDEO_PALETTE_YUV422;
        vidpict.depth=16;
        ATTMP("set pict");
        if ( ioctl(v_fd, VIDIOCSPICT, &vidpict) < 0) {
          return FAILURE;
        }
        ATTMP("get mbuf");
        if ( ioctl(v_fd, VIDIOCGMBUF, &vidmbuf) < 0) {
          return FAILURE;
        }

        maxframes=2;
        cp_buffer = (unsigned char *)mmap(0, vidmbuf.size, PROT_READ | PROT_WRITE, MAP_SHARED, v_fd, 0);
        ATTMP("memory map");
        if (cp_buffer < 0) {
          return FAILURE;
        }

        ATTMP("set capture");
        if ( ioctl(v_fd, VIDIOCCAPTURE, &one) < 0) {
          return FAILURE;
        }

        vidmmap.frame=0;
        vidmmap.width=vidwin.width;
        vidmmap.height=vidwin.height;
        vidmmap.format=vidpict.palette;
        ATTMP("pre-capture");
        if ( ioctl(v_fd, VIDIOCMCAPTURE, &vidmmap) < 0) {
          return FAILURE;
        }

        frame_size = vidwin.width * vidwin.height;
        size = frame_size + (frame_size/4)*2; // 4Y + 1U + 1V
        for (i=0; i<maxframes; i++) {
          vidmmap.frame=i;

          ATTMP("capture");
          if ( ioctl(v_fd, VIDIOCMCAPTURE, &vidmmap) < 0) {
            return FAILURE;
          }
        }

        if (!dvb_is_initialized) {
          if(VDR::Channels.SwitchTo(VDR::Setup.CurrentChannel)){
            dvb_is_initialized = true;
          }else{
            return FAILURE;
          }
        }

        send_format = true;
        dvb_raw_is_initialized = true;

        // unmute audio
        VDR::cDvbApi::muteAudio(false);
      }
    } else {
      DEBUG_STREAM( "using mpeg output" << endl);

      // setup a pointer to this node (needed to transfer buffers from dvbapi)
      VDR::cDvbApi::PrimaryDvbApi->setParentNode(this);

      format_mpeg = true;

      // this tells the VDR's Recordbuffer to collect MPEG-data in NMM-buffers
      VDR::cDvbApi::PrimaryDvbApi->enableBufferCollecting();

      //  switch to last active channel (saved in 'setup.conf' on destruction by a calling Setup.Save()
      if (!dvb_is_initialized) {
        if(VDR::Channels.SwitchTo(VDR::cDvbApi::CurrentChannel())){
          dvb_is_initialized = true;
        }else{
          return FAILURE;
        }

      }

      if (!is_producing_mpeg_buffers) {
        produceMPEG(); // this calls VDR's StartRecord without writing a file to disk
      }

      // mute audio
      VDR::cDvbApi::muteAudio(true);
    }

    return SUCCESS;
  }

  Message* DVBReadNode::processBuffer(Buffer *) {


    if (format_raw)
      return produceYUVBuffer();
    else
      return produceMPEGBuffer();
  }


  Message* DVBReadNode::produceMPEGBuffer() {
    //static bool sync_event_sent = false;
    MutexGuard mg(tuner_mutex);
    

    if ( !running ) {
      running = true;
      sync_event_sent = false;
      switched_channel = true;
      return new FlushEvent;
    }

    if (!sync_event_sent) {
	sync_event_sent = true;
	Event channel_sync_event(ISynchronized::setSynchronized_event, new TInValue<nmm_bool>(true));
	CEvent* channel_event=new CEvent(channel_sync_event);
        return channel_event;
    }

    if (buffer_queue.size() > 1)
      DEBUG_STREAM("Size of internal queue: " << buffer_queue.size() << endl);
    
    if (switched_channel && !buffer_queue.empty()) {
      // channel switched
      buffers_sent = 0;
      string channel_name = string(VDR::Channels.GetChannelNameByNumber(VDR::cDvbApi::PrimaryDvbApi->CurrentChannel()));
      DEBUG_STREAM("switched channel -> sending event start_track (" << channel_name << ")" << endl);
      DEBUG_STREAM("switched channel -> sending event sync_reset... " << endl);
      switched_channel = false;
      CEvent* channel_event = new FlushEvent();
      channel_event->insertEvent(channel_event->size(), ISyncReset::create_syncReset());
      channel_event->insertEvent(channel_event->size(), ITrack::create_startTrack(channel_name));
      return channel_event;
    }

    // discard first buffer after channel_switch occured
    if (buffers_sent == 0 && !buffer_queue.empty()) {
      buffer_queue.front()->release();
      buffer_queue.pop();
      buffers_sent++;
    }
    
    if (buffer_queue.empty()) {
      setProducingFlag(false);
      return 0;
    }

    Buffer* out_buffer = buffer_queue.front();
    buffer_queue.pop();
    buffers_sent++;

    DEBUG_STREAM("Sending mpeg buffer...("<< out_buffer->getUsed() <<")" << endl);

    //usleep(1);

    stream_timer.setTimestamp(out_buffer);

    return out_buffer;
  }

  Message* DVBReadNode::produceYUVBuffer() {
    Buffer *out;

    if (send_channel_name_yuv) {
      send_channel_name_yuv = false;
      string channel_name = string(VDR::Channels.GetChannelNameByNumber(VDR::cDvbApi::PrimaryDvbApi->CurrentChannel()));
      DEBUG_STREAM("send channel name -> sending event start_track (" << channel_name << ")" << endl);
      return new CEvent(ITrack::create_startTrack(channel_name));
    }

    MutexGuard mg(tuner_mutex);
    if ((out = getNewBuffer(size)) != 0) {
      // capture and convert frame
      ioctl(v_fd, VIDIOCSYNC, &frame);
      packed422_to_planar411( (char *) out->getData(), (char *) cp_buffer + vidmbuf.offsets[frame], width, height) ;
      vidmmap.frame=frame;
      ioctl(v_fd, VIDIOCMCAPTURE, &vidmmap);
      out->setUsed(size);
      frame++;  frame=frame%maxframes;
      return out;
    }
    frame++;  frame=frame%maxframes;
    ERROR_STREAM("unable to allocate buffer" << endl);
    return 0;
  }

  void DVBReadNode::produceMPEG() {
    if (is_producing_mpeg_buffers) {
      VDR::cDvbApi::PrimaryDvbApi->StopRecord();
      is_producing_mpeg_buffers = false;
      DEBUG_STREAM("Buffer production stopped!" << endl);
    } else {
      // is the current channel encrypted?
      //int channel_ca = VDR::Channels.GetByNumber(VDR::cDvbApi::PrimaryDvbApi->CurrentChannel())->ca;
      VDR::cDvbApi::PrimaryDvbApi->StartRecord(0,0);
      is_producing_mpeg_buffers = true;
      DEBUG_STREAM("Buffer production started!" << endl);
    }
  }

  void DVBReadNode::clearQueue() {
    buffer_queue_mutex.lock();
    while (!buffer_queue.empty()) {
      (buffer_queue.front())->release();
      buffer_queue.pop();
    }
    buffer_queue_mutex.unlock();
  }

  void DVBReadNode::sendFormat() {
    send_format = true;
  }

  void DVBReadNode::packed422_to_planar411(char *dest, char *src, int width_, int height_)
  {
    int  a1,b;
    char *s, *y,*u,*v;
    unsigned int a, *srca;

    srca = (unsigned int *)src;

    s = src;
    y = dest;
    u = y + width_ * height_;
    v = u + width_ * height_ / 4;

    for (a1 = height_; a1 > 0; a1 -= 2) {
      for (b = width_; b > 0; b -= 2) {
        a = *(srca++);
        *(v++) = a & 0xff; a >>= 8;
        *(y++) = a & 0xff; a >>= 8;
        *(u++) = a & 0xff; a >>= 8;
        *(y++) = a & 0xff;
      }
      for (b = width_; b > 0; b -= 2) {
        a = *(srca++); a >>= 8;
        *(y++) = a & 0xff; a >>= 16;
        *(y++) = a & 0xff;
      }
    }
  }

  Result DVBReadNode::doInit() {
    if (!dvb_is_initialized) { // setup LNB and load channels if not already done
      ATTMP("loading channels and keymap");

      // setup primary parameters (DiSeq,LNB,...)
      VDR::Setup.Load(VDR::AddDirectory(getNMMResource("dvb").c_str(), "setup.conf"));
      // load the channel configuration
      VDR::Channels.Load(channelConfigFile.c_str());

      ATTMP("init DVBAPI");
      if (!(VDR::cDvbApi::Init())) {
        ERROR_STREAM("Unable to initialize DVBAPI !" << endl);
        return FAILURE;
      } else {
      }
    } else {
    }

    DEBUG_STREAM("NUMBER of DVB-Devices: " << VDR::cDvbApi::NumDvbApis << endl);
    if(VDR::cDvbApi::NumDvbApis == 0) return FAILURE;

    // setup a pointer to this node (needed to transfer buffers from dvbapi)
    DEBUG_STREAM("Setting up a pointer to this node" << endl);
    VDR::cDvbApi::PrimaryDvbApi->setParentNode(this);
    
    if(!SSecurityCheck::getInstance().checkFilenameForReading(device)) {
      ERROR_STREAM("Security check not passed for file " << device << ", check .nmmrc for options" << endl);
      return FAILURE;
    }

    if ((v_fd = open(device.c_str(), O_RDWR)) < 0) {// /dev/video0 might be right
      WARNING_STREAM("unable to open video4linux device" << endl);
    } else {

      //query capabilities to test, if we can provide raw-video
      int index;
      if (-1 != ioctl (v_fd, VIDEO_GET_CAPABILITIES, &index)) {
      
        // format definition
        NMM::Format* myOutputFormat=getOwnOutputProperty()->addNewFormat(Format::video_raw_type);
        myOutputFormat->addWildcard(Format::x_resolution_param);
        myOutputFormat->addWildcard(Format::y_resolution_param);
        myOutputFormat->addWildcard(Format::ratio_x_param);
        myOutputFormat->addWildcard(Format::ratio_y_param);
        myOutputFormat->addFloatParamValue(Format::framerate_param, 25);
        myOutputFormat->addStringParamValue(Format::channelorder_param,"yvu");
        myOutputFormat->addStringParamValue(Format::endian_param, "little");
        myOutputFormat->addStringParamValue(Format::colorspace_param,"yv12");
        myOutputFormat->addIntParamValue(Format::bitperpixel_param,12);
        myOutputFormat->addStringParamValue(Format::format_param, "planar");
        myOutputFormat->setIOPartner(0);
      }
    }

    Format* o_format = getOwnOutputProperty()->addNewFormat("av/mpegps");
    o_format->addIntParamValue(Format::map_mpeg_audio_streams_param, 1); // only first mpeg audio stream
    o_format->addIntParamValue(Format::map_mpeg_video_streams_param, 1); // only first mpeg audio stream
    o_format->addIntParamValue(Format::map_ac3_audio_streams_param, 0); // no ac3
    o_format->addIntParamValue(Format::map_dts_audio_streams_param, 0); // no dts
    o_format->addIntParamValue(Format::map_lpcm_audio_streams_param, 0); // no lpcm
    o_format->addIntParamValue(Format::map_spu_streams_param, 0); // no spu

    Format* sub_format = new Format(Format::audio_mpeg_type);
    sub_format->addIntParamValue(Format::audio_codec_param, 0x55);
    sub_format->addIntParamValue(Format::layer_param, 2);
    sub_format->addIntParamValue(Format::channels_param, 2);
    sub_format->addIntParamValue(Format::sample_rate_param, 48000);
    sub_format->addIntParamValue(Format::bitpersample_param, 16);
    o_format->addFPointerParamValue(Format::format_mpeg_audio_streams_param, sub_format);

    sub_format = new Format(Format::video_mpeg_type);
    sub_format->addIntParamValue(Format::version_param, 2);
    o_format->addFPointerParamValue(Format::format_mpeg_video_streams_param, sub_format);

    // add dummy formats!
    o_format->addFPointerParamValue(Format::format_mpeg_video_streams_param, new Format(Format::wild_wild_type));
    o_format->addFPointerParamValue(Format::format_mpeg_video_streams_param, new Format(Format::wild_wild_type));
    o_format->addFPointerParamValue(Format::format_mpeg_video_streams_param, new Format(Format::wild_wild_type));
    o_format->addFPointerParamValue(Format::format_mpeg_video_streams_param, new Format(Format::wild_wild_type));

    getOwnOutputProperty()->setDefaultFormat(o_format);

    return SUCCESS;
  }

  Result DVBReadNode::doDeinit() {
    if (-1 != v_fd) {
      DEBUG_STREAM("Closing raw video device" << endl);
      close(v_fd);
      v_fd = -1;
    }
    return SUCCESS;
  }

  Result DVBReadNode::doInitOutput() {
    if (-1 != v_fd) {
      int index;
      if (-1 != ioctl (v_fd, VIDEO_GET_CAPABILITIES, &index)) {

        // format definition
        Format* myOutputFormat = getOwnWorkingOutputProperty()->addNewFormat(Format::video_raw_type);
        myOutputFormat->addIntParamValue(Format::x_resolution_param, 768);
        myOutputFormat->addIntParamValue(Format::y_resolution_param, 576);
        myOutputFormat->addIntParamValue(Format::ratio_x_param, 4);
        myOutputFormat->addIntParamValue(Format::ratio_y_param, 3);
        myOutputFormat->addFloatParamValue(Format::framerate_param, 25);
        myOutputFormat->addStringParamValue(Format::channelorder_param,"yvu");
        myOutputFormat->addStringParamValue(Format::endian_param, "little");
        myOutputFormat->addStringParamValue(Format::colorspace_param,"yv12");
        myOutputFormat->addIntParamValue(Format::bitperpixel_param,12);
        myOutputFormat->addStringParamValue(Format::format_param, "planar");
        myOutputFormat->setIOPartner(0);
        
        myOutputFormat = getOwnWorkingOutputProperty()->addNewFormat(Format::video_raw_type);
        myOutputFormat->addWildcard(Format::x_resolution_param);
        myOutputFormat->addWildcard(Format::y_resolution_param);
        myOutputFormat->addWildcard(Format::ratio_x_param);
        myOutputFormat->addWildcard(Format::ratio_y_param);
        myOutputFormat->addFloatParamValue(Format::framerate_param, 25);
        myOutputFormat->addStringParamValue(Format::channelorder_param,"yvu");
        myOutputFormat->addStringParamValue(Format::endian_param, "little");
        myOutputFormat->addStringParamValue(Format::colorspace_param,"yv12");
        myOutputFormat->addIntParamValue(Format::bitperpixel_param,12);
        myOutputFormat->addStringParamValue(Format::format_param, "planar");
        myOutputFormat->setIOPartner(0);
      }
    }

    Format* o_format = getOwnWorkingOutputProperty()->addNewFormat("av/mpegps");
    o_format->addIntParamValue(Format::version_param, 2);
    o_format->addIntParamValue(Format::map_mpeg_audio_streams_param, 1); // only first mpeg audio stream
    o_format->addIntParamValue(Format::map_mpeg_video_streams_param, 1); // only first mpeg audio stream
    o_format->addIntParamValue(Format::map_ac3_audio_streams_param, 0); // no ac3
    o_format->addIntParamValue(Format::map_dts_audio_streams_param, 0); // no dts
    o_format->addIntParamValue(Format::map_lpcm_audio_streams_param, 0); // no lpcm
    o_format->addIntParamValue(Format::map_spu_streams_param, 0); // no spu

    Format* sub_format = new Format(Format::audio_mpeg_type);
    sub_format->addIntParamValue(Format::audio_codec_param, 0x55);
    sub_format->addIntParamValue(Format::layer_param, 2);
    sub_format->addIntParamValue(Format::channels_param, 2);
    sub_format->addIntParamValue(Format::sample_rate_param, 48000);
    sub_format->addIntParamValue(Format::bitpersample_param, 16);
    o_format->addFPointerParamValue(Format::format_mpeg_audio_streams_param, sub_format);

    sub_format = new Format(Format::video_mpeg_type);
    sub_format->addIntParamValue(Format::version_param, 2);
    o_format->addFPointerParamValue(Format::format_mpeg_video_streams_param, sub_format);

    // add dummy formats!
    o_format->addFPointerParamValue(Format::format_ac3_audio_streams_param, new Format(Format::wild_wild_type));
    o_format->addFPointerParamValue(Format::format_spu_streams_param, new Format(Format::wild_wild_type));
    o_format->addFPointerParamValue(Format::format_lpcm_audio_streams_param, new Format(Format::wild_wild_type));
    o_format->addFPointerParamValue(Format::format_dts_audio_streams_param, new Format(Format::wild_wild_type));

    getOwnWorkingOutputProperty()->setDefaultFormat(o_format);

    return SUCCESS;
  }

  Result DVBReadNode::setChannel(const nmm_int32& ch_nr) {
    MutexGuard mg(tuner_mutex);
    MutexGuard mg1(eit_mutex);

    VDR::cChannel *channel = VDR::Channels.GetByNumber(ch_nr);

    if (channel) {
      if(channel->Switch()){
        setProducingFlag(true);
      }else{
        setProducingFlag(false);
        switched_channel = true;
        send_channel_name_yuv = true;
        return FAILURE;
      }
    } else {
      GLOBAL_ERROR_STREAM("Channel out of range : " << ch_nr << endl);
      return FAILURE;
    }
    switched_channel = true;
    send_channel_name_yuv = true;
    return SUCCESS;
  }

  Result DVBReadNode::getChannel(nmm_int32& channel, string& channelname) {
    channel = VDR::cDvbApi::CurrentChannel();
    if (!channel) {
      return FAILURE;
    }
    channelname = string(VDR::Channels.GetChannelNameByNumber(VDR::cDvbApi::PrimaryDvbApi->CurrentChannel()));
    
    return SUCCESS;
  }

  void DVBReadNode::setDevice(const string& device) throw(const DeviceNotFoundException&) {
    /* try to get the number of the device 
       regular device is /dev/video0, but DVB need /dev/dvb/adapterX */
    string devNumber = device.substr(10);
    int number = atoi(devNumber.c_str());
    if(!(VDR::cDvbApi::Probe(string("/dev/dvb/adapter"+itoa(number)+"/dvr"+itoa(number)).c_str()))){
      throw DeviceNotFoundException("/dev/dvb/adapter"+itoa(number)+"/dvr" + itoa(number));
    }

    /* now we can set the device */
    this->device = device;
  }
  
  string DVBReadNode::getDevice() const {
    return device;
  }
  
  // --------------- Inmplementation IEPG Interface ------------------------------
  
  string DVBReadNode::getCurrentShow()
  {
    
    MutexGuard mg(eit_mutex);
    string show("No Information Available");
    GLOBAL_DEBUG_STREAM(" start reading currentshow" << endl);
    int j =0;

    VDR::cThreadLock *lock;
    const VDR::cSchedules *schedule = VDR::cDvbApi::PrimaryDvbApi->Schedules(lock);
    if(schedule){
      const VDR::cSchedule *s  = schedule->GetSchedule(); 
      if(s)
        {
          const VDR::cEventInfo *i = NULL;
          while((i == NULL) && (j <= 100))
            {
              i = s->GetPresentEvent();
              if(i)
                {
                  show = string(i->GetTimeString())
                    + " - " + string(i->GetEndTimeString())
                    + " : " + string(i->GetTitle());
             
                }
              j++;
            }
        }
    }
    return show;
    
  }

  string DVBReadNode::getFollowingShow()
  {
    
    MutexGuard mg(tuner_mutex);
    string show("No Information Available");
    GLOBAL_DEBUG_STREAM(" start reading followingshow" << endl);
    int j = 0;
    
    VDR::cThreadLock *lock;
    const VDR::cSchedules *schedule = VDR::cDvbApi::PrimaryDvbApi->Schedules(lock);
    if(schedule){
      const VDR::cSchedule *s  = schedule->GetSchedule(); 
      if(s)
        {
          const VDR::cEventInfo *i = NULL;
          while((i == NULL) && (j <= 100))
            {
              i = s->GetFollowingEvent();
              if(i){
              show = string(i->GetTimeString())
                + " - " + string(i->GetEndTimeString())
                + " : " + string(i->GetTitle());
              }

              j++;
            }
        }
    }
    return show;
    
  }

  string DVBReadNode::getDetailedDescription()
  {
    
    MutexGuard mg(tuner_mutex);
    string show("No Information Available");
      
    VDR::cThreadLock *lock;
    const VDR::cSchedules *schedule = VDR::cDvbApi::PrimaryDvbApi->Schedules(lock);
    if(schedule){
      const VDR::cSchedule *s  = schedule->GetSchedule(); 
      if(s)
        {
          const VDR::cEventInfo *i  = s->GetPresentEvent();
          if(i){
            const char* sh = i->GetExtendedDescription();
            if(sh)
              show = string(sh);
          }
        }
    }
    return show; 
    
  } 

  void DVBReadNode::setEPGDirectory(const string& directory){
    DEBUG_STREAM2("setEPGDirectory: " << directory << std::endl);
    VDR::VideoDirectory = NULL;
    __directory=directory;
    VDR::VideoDirectory = __directory.c_str();
  }
  
  void DVBReadNode::getActivity(){
    eitScanner->Activity();

  }
  void DVBReadNode::scan(){
    eitScanner->Process();
  }
  
  void DVBReadNode::registerTVGuide(const ITVGuide* itvgs){
    SDVBReadNodeTVGuideManager::getInstance().addTVGuide(itvgs);
  }

  void DVBReadNode::unregisterTVGuide(const ITVGuide* itvgs){
    SDVBReadNodeTVGuideManager::getInstance().removeTVGuide(itvgs);
  }


  TVChannelList DVBReadNode::getChannelList(){
    channels.clear();
    loadChannelconfig(); 
    return channels;
  }

  void DVBReadNode::setChannelConfigFile(const string &filename){
    channelConfigFile = filename;
    channels.clear();
    loadChannelconfig();
  }

  void DVBReadNode::loadChannelconfig(){

    ifstream configfile(channelConfigFile.c_str());
    char line[101];

    if (!configfile) {
        GLOBAL_ERROR_STREAM("Channel file " << channelConfigFile << " not found" << endl);
        return;
    }
    string dummy;
    string freq;
    int channelCount = 1;

    while (!configfile.eof()) {
      configfile.getline(line, 100);
      if(line){
        string tmp(line);
        size_t pos = tmp.find_first_of(':');

        if(pos != string::npos){
          TVChannel channel;
          string channelName = tmp.substr(0,pos);
          channel.setChannelName(channelName);
          channel.setChannelNumber(channelCount);
          channelCount++;
          channel.readAliases();
          channels.addChannel(channel);
        }
      }
    }
  }
}

#include "nmm/nmm_suffix.hpp"
