/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "DVBSimulatorNode.hpp"

#include "nmm/multimedia/IDevice.hpp"
#include "nmm/multimedia/ISyncReset.hpp"
#include "nmm/multimedia/interfaces/ITrack.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"
#include "nmm/utils/thread/MutexGuard.hpp"
#include "nmm/utils/thread/Thread.hpp"
#include "nmm/utils/time/Interval.hpp"
#include "nmm/utils/directory_utils.hpp"
#include "nmm/multimedia/FlushEvent.hpp"



#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sstream>
#include <regex.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


NMM_REGISTER_PLUGIN(DVBSimulatorNode, NMM, "DVBSimulatorNode")

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  DVBSimulatorNode::DVBSimulatorNode(const char* name /* = "DVBSimulatorNode" */,
			   StreamQueue::Mode mode /* = StreamQueue::MODE_SUSPEND */,
			   int queue_size /* = StreamQueue::DEFAULT_SIZE */)
    : GenericSourceNode(name, mode, queue_size),
      old_ts(0L),
      old_pt(0L),
      have_to_sleep(false),
      te() {
    file_name   = "";
    send_event  = false;
    channelConfigFile = getNMMResourceDir() + "/dvb/channels.conf";
    channel = 0;
  };

  DVBSimulatorNode::~DVBSimulatorNode() {
  }
  

  /********************************************************************************/
  /** do methods **/
  Result DVBSimulatorNode::doInit() {
    Format* o_format = getOwnOutputProperty()->addNewFormat("av/mpegps");
    o_format->addIntParamValue(Format::map_mpeg_audio_streams_param, 1); // only first mpeg audio stream
    o_format->addIntParamValue(Format::map_mpeg_video_streams_param, 1); // only first mpeg audio stream
    o_format->addIntParamValue(Format::map_ac3_audio_streams_param, 0); // no ac3
    o_format->addIntParamValue(Format::map_dts_audio_streams_param, 0); // no dts
    o_format->addIntParamValue(Format::map_lpcm_audio_streams_param, 0); // no lpcm
    o_format->addIntParamValue(Format::map_spu_streams_param, 0); // no spu

    Format* sub_format = new Format(Format::audio_mpeg_type);
    sub_format->addIntParamValue(Format::audio_codec_param, 0x55);
    sub_format->addIntParamValue(Format::layer_param, 2);
    sub_format->addIntParamValue(Format::channels_param, 2);
    sub_format->addIntParamValue(Format::sample_rate_param, 48000);
    sub_format->addIntParamValue(Format::bitpersample_param, 16);
    o_format->addFPointerParamValue(Format::format_mpeg_audio_streams_param, sub_format);

    sub_format = new Format(Format::video_mpeg_type);
    sub_format->addIntParamValue(Format::version_param, 2);
    o_format->addFPointerParamValue(Format::format_mpeg_video_streams_param, sub_format);

    // add dummy formats!
    o_format->addFPointerParamValue(Format::format_mpeg_video_streams_param, new Format(Format::wild_wild_type));
    o_format->addFPointerParamValue(Format::format_mpeg_video_streams_param, new Format(Format::wild_wild_type));
    o_format->addFPointerParamValue(Format::format_mpeg_video_streams_param, new Format(Format::wild_wild_type));
    o_format->addFPointerParamValue(Format::format_mpeg_video_streams_param, new Format(Format::wild_wild_type));

    getOwnOutputProperty()->setDefaultFormat(o_format);
    
    return SUCCESS;
  }
  
  Result DVBSimulatorNode::doActivate(){
    send_event=true; 
    return SUCCESS;
  }

  Result DVBSimulatorNode::doInitOutput() {
    file_stream.open(file_name.c_str(), ios::in);
    if(file_stream.fail()) {
      DEBUG_STREAM( "Error while opening file " << file_name << endl);
      return FAILURE;
    }

    DEBUG_STREAM( "Opened file " << file_name << endl);
    old_ts = Time(0L);
    old_pt = Time(0L);
    have_to_sleep = false;
    send_event = true;

    Format* o_format = getOwnWorkingOutputProperty()->addNewFormat("av/mpegps");
    o_format->addIntParamValue(Format::version_param, 2);
    o_format->addIntParamValue(Format::map_mpeg_audio_streams_param, 1); // only first mpeg audio stream
    o_format->addIntParamValue(Format::map_mpeg_video_streams_param, 1); // only first mpeg audio stream
    o_format->addIntParamValue(Format::map_ac3_audio_streams_param, 0); // no ac3
    o_format->addIntParamValue(Format::map_dts_audio_streams_param, 0); // no dts
    o_format->addIntParamValue(Format::map_lpcm_audio_streams_param, 0); // no lpcm
    o_format->addIntParamValue(Format::map_spu_streams_param, 0); // no spu

    Format* sub_format = new Format(Format::audio_mpeg_type);
    sub_format->addIntParamValue(Format::audio_codec_param, 0x55);
    sub_format->addIntParamValue(Format::layer_param, 2);
    sub_format->addIntParamValue(Format::channels_param, 2);
    sub_format->addIntParamValue(Format::sample_rate_param, 48000);
    sub_format->addIntParamValue(Format::bitpersample_param, 16);
    o_format->addFPointerParamValue(Format::format_mpeg_audio_streams_param, sub_format);

    sub_format = new Format(Format::video_mpeg_type);
    sub_format->addIntParamValue(Format::version_param, 2);
    o_format->addFPointerParamValue(Format::format_mpeg_video_streams_param, sub_format);

    // add dummy formats!
    o_format->addFPointerParamValue(Format::format_ac3_audio_streams_param, new Format(Format::wild_wild_type));
    o_format->addFPointerParamValue(Format::format_spu_streams_param, new Format(Format::wild_wild_type));
    o_format->addFPointerParamValue(Format::format_lpcm_audio_streams_param, new Format(Format::wild_wild_type));
    o_format->addFPointerParamValue(Format::format_dts_audio_streams_param, new Format(Format::wild_wild_type));

    getOwnWorkingOutputProperty()->setDefaultFormat(o_format);

    return SUCCESS;
  }

  Result DVBSimulatorNode::doDeinitOutput() 
  {
    file_stream.close();
    return SUCCESS;
  }
  /********************************************************************************/
  

  /********************************************************************************/
  /** processBuffer **/
  Message* DVBSimulatorNode::processBuffer(Buffer *) 
  {
    MutexGuard m(file_name_mutex);
    Buffer *buffer;
    int read_status;
    Timestamp ts;
    
    // send start_track event
    if (send_event) {
      send_event = false;
      CEvent* channel_event = new FlushEvent();
      channel_event->insertEvent(channel_event->size(), ISyncReset::create_syncReset());
      channel_event->insertEvent(channel_event->size(), ITrack::create_startTrack(file_name));
      return channel_event;
    }
    
    // IFileModeHandler::SIZE_DATA or IFileModeHandler::SIZE_TIMESTAMP_DATA, first read size
    file_stream.read(reinterpret_cast<char*>(&frame_size), sizeof(size_t));
    read_status = sizeof(size_t);
    if(file_stream.fail()) {
      file_stream.seekg(0, ios::beg);
      DEBUG_STREAM( "Done with file but looping ..." << endl);
      old_ts = Time(0L);
      old_pt = Time(0L);
      have_to_sleep = false;
      return 0;
    }

    // IFileModeHandler::SIZE_TIMESTAMP_DATA, read timestamp
    file_stream.read(reinterpret_cast<char*>(&ts), sizeof(Timestamp));
    read_status = sizeof(Timestamp);
    if(file_stream.fail()) {
      file_stream.seekg(0, ios::beg);
      DEBUG_STREAM( "Done with file but looping ..." << endl);
      old_ts = Time(0L);
      old_pt = Time(0L);
      have_to_sleep = false;
      return 0;
    }

    DEBUG_STREAM("Timestamp " << ts << endl);
    
    // Get new Buffer, and read data
    if ((buffer = getNewBuffer(frame_size)) != 0) {
      DEBUG_STREAM("SIZE_TIMESTAMP_DATA used=" << frame_size << " Timestamp=" << ts << " Data" << endl);
      file_stream.read(buffer->getData(), frame_size);
      read_status = frame_size;
      buffer->setTimestamp(ts);
      buffer->setUsed(read_status);
      if (have_to_sleep) {
        // we only have to sleep if have_to_sleep
        Thread::suspendThread((ts.sync_time - old_ts) - (te.getTime() - old_pt));
      } else {
        have_to_sleep = true;
      }
      old_ts = ts.sync_time;
      old_pt = te.getTime();
    } else {
      ERROR_STREAM("Could not get new buffer" << endl);
      return 0;
    }

    // IFileModeHandler::FILE_PER_STREAM
    if(file_stream.fail() || (static_cast<unsigned int>(read_status) != frame_size)) {
      file_stream.seekg(0, ios::beg);
      buffer->release();
      DEBUG_STREAM( "Done with file but looping ..." << endl);
      old_ts = Time(0L);
      old_pt = Time(0L);
      have_to_sleep = false;
      return 0;
    }

    DEBUG_STREAM( "Ok. Read " << buffer->getUsed() << " bytes. " << endl);
    return buffer;
  }
  /********************************************************************************/


  /********************************************************************************/
  /** ISourceFileHandler Interface **/
  void DVBSimulatorNode::setFilename(const string& name) {
    if(isConstructed()) {
      file_name = name;
      TVChannel cd;
      cd.setChannelName(name);
      cd.setChannelNumber(0);
      list<string> a; a.push_back("DVBSim"); cd.setChannelAliases(a);
      channels.addChannel(cd);
    }      
    else {
      ERROR_STREAM("Node not in CONSTRUCTED state, setFileName illegal in current state and with FILE_PER_STREAM mode.");
    }
  }
  /********************************************************************************/
  
  /********************************************************************************/
  /** ITVCard Interface **/
  void DVBSimulatorNode::setDevice(const string& ) throw(const DeviceNotFoundException&) {
  } // ignored for simulation
  
  string DVBSimulatorNode::getDevice() const {
    return string("/dev/simulator"); 
  }
  
  void DVBSimulatorNode::setResolution(const nmm_int32&, const nmm_int32&) {} // ignored for simulation

  void DVBSimulatorNode::getResolution(nmm_int32& x, nmm_int32& y) {
    x = 768;
    y = 576;
  }

  Result DVBSimulatorNode::setChannel(const nmm_int32& c) {
    channel = c;
    return SUCCESS;
  }

  Result DVBSimulatorNode::getChannel(nmm_int32& c1, string& /*c2*/) {
    c1 = channel;
    return SUCCESS;
  }

  /********************************************************************************/


  /********************************************************************************/
  /** IDVBSimulator Interface **/
   void DVBSimulatorNode::setWorkingDirectory(const string&) {
     /// XXX not implemented
     /// shall be used in future for simulated channelswitching
   }
  /********************************************************************************/

  TVChannelList DVBSimulatorNode::getChannelList(){
    TVChannelList cl;
    TVChannel c;
    string channelName = file_name.substr(file_name.find_last_of('/')+1);
    c.setChannelName(channelName);
    cl.addChannel(c);
    return cl;
  }
  void DVBSimulatorNode::setChannelConfigFile(const string &){}


}

#include "nmm/nmm_suffix.hpp"
