/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

/*
 * $Id: dvbapi.cpp,v 1.23.8.2 2005/12/22 01:40:40 fw Exp $
 */

/*
 * dvbapi.c: Interface to the DVB driver
 *
 * See the main source file 'dvbapi.h' for original copyright information and
 * how to reach the original author.
 *
 * $Id: dvbapi.cpp,v 1.23.8.2 2005/12/22 01:40:40 fw Exp $
 */

#include "dvbapi.h"
#include <dirent.h>
#include <errno.h>
#include <fcntl.h>
#include <iostream>
/*
extern "C" {
#define HAVE_BOOLEAN
#include <jpeglib.h>
}
*/
#include <stdlib.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include "vdrconfig.h"
#include "recording.h"
#include "remux.h"
#include "ringbuffer.h"
#include "tools.h"
#include "videodir.h"

namespace VDR {
// NMM-Buffers of this size will be allocated
// 512k seems to be enought for a complete GOP of a DVB-Stream
#define MAXBUFFERSIZE (512 * 1024)

#define DEV_VIDEO      "/dev/video"
#define DEV_OST_OSD    "/dev/dvb/adapter0/osd"
  //#define DEV_OST_QAMFE  "/dev/ost/qamfe"
  //#define DEV_OST_QPSKFE "/dev/ost/qpskfe"
  //#define DEV_OST_SEC    "/dev/ost/sec"
#define DEV_OST_DVR    "/dev/dvb/adapter0/dvr"
#define DEV_OST_DEMUX  "/dev/dvb/adapter0/demux"
#define DEV_OST_VIDEO  "/dev/dvb/adapter0/video"
#define DEV_OST_AUDIO  "/dev/dvb/adapter0/audio"
#define DEV_OST_FRONT  "/dev/dvb/adapter0/frontend"

// The size of the array used to buffer video data:
// (must be larger than MINVIDEODATA - see remux.h)
#define VIDEOBUFSIZE (1024*1024)

// The maximum size of a single frame:
#define MAXFRAMESIZE (128*1024)

#define FRAMESPERSEC 25

// The maximum file size is limited by the range that can be covered
// with 'int'. 4GB might be possible (if the range is considered
// 'unsigned'), 2GB should be possible (even if the range is considered
// 'signed'), so let's use 1GB for absolute safety (the actual file size
// may be slightly higher because we stop recording only before the next
// 'I' frame, to have a complete Group Of Pictures):
#define MAXVIDEOFILESIZE (1024*1024*1024) // Byte
#define MAXFILESPERRECORDING 255

#define MINFREEDISKSPACE    (512) // MB
#define DISKCHECKINTERVAL   100 // seconds

#define INDEXFILESUFFIX     "/index.vdr"
#define RECORDFILESUFFIX    "/%03d.vdr"
#define RECORDFILESUFFIXLEN 20 // some additional bytes for safety...

// The number of frames to back up when resuming an interrupted replay session:
#define RESUMEBACKUP (10 * FRAMESPERSEC)

#define CHECK(s) { if ((s) < 0) LOG_ERROR; } // used for 'ioctl()' calls

typedef unsigned char uchar;

const char *IndexToHMSF(int Index, bool WithFrame)
{
  static char buffer[16];
  int f = (Index % FRAMESPERSEC) + 1;
  int s = (Index / FRAMESPERSEC);
  int m = s / 60 % 60;
  int h = s / 3600;
  s %= 60;
  snprintf(buffer, sizeof(buffer), WithFrame ? "%d:%02d:%02d.%02d" : "%d:%02d:%02d", h, m, s, f);
  return buffer;
}

int HMSFToIndex(const char *HMSF)
{
  int h, m, s, f = 0;
  if (3 <= sscanf(HMSF, "%d:%d:%d.%d", &h, &m, &s, &f))
     return (h * 3600 + m * 60 + s) * FRAMESPERSEC + f - 1;
  return 0;
}

// --- cIndexFile ------------------------------------------------------------

class cIndexFile {
private:
  struct tIndex { int offset; uchar type; uchar number; short reserved; };
  int f;
  char *fileName, *pFileExt;
  int size, last;
  tIndex *index;
  cResumeFile resumeFile;
  bool CatchUp(void);
public:
  cIndexFile(const char *FileName, bool Record);
  ~cIndexFile();
  bool Ok(void) { return index != NULL; }
  void Write(uchar PictureType, uchar FileNumber, int FileOffset);
  bool Get(int Index, uchar *FileNumber, int *FileOffset, uchar *PictureType = NULL, int *Length = NULL);
  int GetNextIFrame(int Index, bool Forward, uchar *FileNumber = NULL, int *FileOffset = NULL, int *Length = NULL);
  int Get(uchar FileNumber, int FileOffset);
  int Last(void) { CatchUp(); return last; }
  int GetResume(void) { return resumeFile.Read(); }
  bool StoreResume(int Index) { return resumeFile.Save(Index); }
  };

cIndexFile::cIndexFile(const char *FileName, bool Record)
:resumeFile(FileName)
{
  f = -1;
  fileName = pFileExt = NULL;
  size = 0;
  last = -1;
  index = NULL;
  if (FileName) {
     fileName = new char[strlen(FileName) + strlen(INDEXFILESUFFIX) + 1];
     if (fileName) {
        strcpy(fileName, FileName);
        pFileExt = fileName + strlen(fileName);
        strcpy(pFileExt, INDEXFILESUFFIX);
        int delta = 0;
        if (access(fileName, R_OK) == 0) {
           struct stat buf;
           if (stat(fileName, &buf) == 0) {
              delta = buf.st_size % sizeof(tIndex);
              if (delta) {
                 delta = sizeof(tIndex) - delta;
                 GLOBAL_ERROR_STREAM("ERROR: invalid file size " << buf.st_size <<" in '" << fileName <<"'"<<endl);
                 }
              last = (buf.st_size + delta) / sizeof(tIndex) - 1;
              if (!Record && last >= 0) {
                 size = last + 1;
                 index = new tIndex[size];
                 if (index) {
                    f = open(fileName, O_RDONLY);
                    if (f >= 0) {
                       if ((int)read(f, index, buf.st_size) != buf.st_size) {
                          GLOBAL_ERROR_STREAM("ERROR: can't read from file '" << fileName << "'" << endl);
                          delete index;
                          index = NULL;
                          close(f);
                          f = -1;
                          }
                       // we don't close f here, see CatchUp()!
                       }
                    else
                       GLOBAL_ERROR_STREAM(fileName << endl);
                    }
                 else
                    GLOBAL_ERROR_STREAM("ERROR: can't allocate "<< size * sizeof(tIndex) <<" bytes for index '"<< fileName << "'" << endl);
                 }
              }
           else
              LOG_ERROR;
           }
        else if (!Record)
           GLOBAL_DEBUG_STREAM("missing index file " << fileName << endl);
        if (Record) {
           if ((f = open(fileName, O_WRONLY | O_CREAT | O_APPEND, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) >= 0) {
              if (delta) {
                 GLOBAL_ERROR_STREAM("ERROR: padding index file with "<< delta << " '0' bytes" << endl);
                 while (delta--)
                       writechar(f, 0);
                 }
              }
           else
              GLOBAL_ERROR_STREAM(fileName << endl);
           delete fileName;
           fileName = pFileExt = NULL;
           }
        }
     else
        GLOBAL_ERROR_STREAM("ERROR: can't copy file name " << FileName << endl);
     }
}

cIndexFile::~cIndexFile()
{
  if (f >= 0)
     close(f);
  delete fileName;
}

bool cIndexFile::CatchUp(void)
{
  if (index && f >= 0) {
     struct stat buf;
     if (fstat(f, &buf) == 0) {
        int newLast = buf.st_size / sizeof(tIndex) - 1;
        if (newLast > last) {
           if (size <= newLast) {
              size *= 2;
              if (size <= newLast)
                 size = newLast + 1;
              }
           index = (tIndex *)realloc(index, size * sizeof(tIndex));
           if (index) {
              int offset = (last + 1) * sizeof(tIndex);
              int delta = (newLast - last) * sizeof(tIndex);
              if (lseek(f, offset, SEEK_SET) == offset) {
                 if (read(f, &index[last + 1], delta) != delta) {
                    esyslog(LOG_ERR, "ERROR: can't read from index");
                    delete index;
                    index = NULL;
                    close(f);
                    f = -1;
                    }
                 last = newLast;
                 return true;
                 }
              else
                 LOG_ERROR;
              }
           else
              GLOBAL_ERROR_STREAM("ERROR: can't realloc() index");
           }
        }
     else
        LOG_ERROR;
     }
  return false;
}

void cIndexFile::Write(uchar PictureType, uchar FileNumber, int FileOffset)
{
  if (f >= 0) {
     tIndex i = { FileOffset, PictureType, FileNumber, 0 };
     if (write(f, &i, sizeof(i)) != sizeof(i)) {
        GLOBAL_ERROR_STREAM("ERROR: can't write to index file");
        close(f);
        f = -1;
        return;
        }
     last++;
     }
}

bool cIndexFile::Get(int Index, uchar *FileNumber, int *FileOffset, uchar *PictureType, int *Length)
{
  if (index) {
     CatchUp();
     if (Index >= 0 && Index <= last) {
        *FileNumber = index[Index].number;
        *FileOffset = index[Index].offset;
        if (PictureType)
           *PictureType = index[Index].type;
        if (Length) {
           int fn = index[Index + 1].number;
           int fo = index[Index + 1].offset;
           if (fn == *FileNumber)
              *Length = fo - *FileOffset;
           else
              *Length = -1; // this means "everything up to EOF" (the buffer's Read function will act accordingly)
           }
        return true;
        }
     }
  return false;
}

int cIndexFile::GetNextIFrame(int Index, bool Forward, uchar *FileNumber, int *FileOffset, int *Length)
{
  if (index) {
     if (Forward)
        CatchUp();
     int d = Forward ? 1 : -1;
     for (;;) {
         Index += d;
         if (Index >= 0 && Index <= last - 100) { // '- 100': need to stay off the end!
            if (index[Index].type == I_FRAME) {
               if (FileNumber)
                  *FileNumber = index[Index].number;
               else
                  FileNumber = &index[Index].number;
               if (FileOffset)
                  *FileOffset = index[Index].offset;
               else
                  FileOffset = &index[Index].offset;
               if (Length) {
                  // all recordings end with a non-I_FRAME, so the following should be safe:
                  int fn = index[Index + 1].number;
                  int fo = index[Index + 1].offset;
                  if (fn == *FileNumber)
                     *Length = fo - *FileOffset;
                  else {
                     GLOBAL_ERROR_STREAM("ERROR: 'I' frame at end of file #" << *FileNumber << endl);
                     *Length = -1;
                     }
                  }
               return Index;
               }
            }
         else
            break;
         }
     }
  return -1;
}

int cIndexFile::Get(uchar FileNumber, int FileOffset)
{
  if (index) {
     CatchUp();
     //TODO implement binary search!
     int i;
     for (i = 0; i < last; i++) {
         if (index[i].number > FileNumber || (index[i].number == FileNumber) && index[i].offset >= FileOffset)
            break;
         }
     return i;
     }
  return -1;
}

// --- cFileName -------------------------------------------------------------

class cFileName {
private:
  int file;
  int fileNumber;
  char *fileName, *pFileNumber;
  bool record;
  bool blocking;
public:
  cFileName(const char *FileName, bool Record, bool Blocking = false);
  ~cFileName();
  const char *Name(void) { return fileName; }
  int Number(void) { return fileNumber; }
  int Open(void);
  void Close(void);
  int SetOffset(int Number, int Offset = 0);
  int NextFile(void);
  };

cFileName::cFileName(const char *FileName, bool Record, bool Blocking)
{
  file = -1;
  fileNumber = 0;
  record = Record;
  blocking = Blocking;
  // Prepare the file name:
  if (*FileName != '\0') {
    fileName = new char[strlen(FileName) + RECORDFILESUFFIXLEN];
    if (!fileName) {
      GLOBAL_ERROR_STREAM("ERROR: can't copy file name '" << fileName<< "'" << endl);
      return;
    }
    strcpy(fileName, FileName);
    pFileNumber = fileName + strlen(fileName);
    SetOffset(1);
  } else {
    fileName = 0;
  }
}

cFileName::~cFileName()
{
  Close();
  if (fileName) delete[] fileName;
}

int cFileName::Open(void)
{
  if (file < 0) {
     int BlockingFlag = blocking ? 0 : O_NONBLOCK;
     if (record) {
        GLOBAL_DEBUG_STREAM("DVBAPI(filename): recording to '"<< fileName <<"'" << endl );
        file = OpenVideoFile(fileName, O_RDWR | O_CREAT | BlockingFlag);
        if (file < 0)
           GLOBAL_ERROR_STREAM(fileName << endl);
        }
     else {
        if (access(fileName, R_OK) == 0) {
           GLOBAL_DEBUG_STREAM("playing '"<< fileName<<"'"<< endl);
           file = open(fileName, O_RDONLY | BlockingFlag);
           if (file < 0)
              GLOBAL_ERROR_STREAM(fileName << endl);
           }
        else if (errno != ENOENT)
           GLOBAL_ERROR_STREAM(fileName << endl);
        }
     }
  return file;
}

void cFileName::Close(void)
{
  if (file >= 0) {
     if ((record && CloseVideoFile(file) < 0) || (!record && close(file) < 0))
        GLOBAL_ERROR_STREAM(fileName << endl);
     file = -1;
     }
}

int cFileName::SetOffset(int Number, int Offset)
{
  if (fileNumber != Number)
     Close();
  if (0 < Number && Number <= MAXFILESPERRECORDING) {
     fileNumber = Number;
     sprintf(pFileNumber, RECORDFILESUFFIX, fileNumber);
     if (record) {
        if (access(fileName, F_OK) == 0) // file exists, let's try next suffix
           return SetOffset(Number + 1);
        else if (errno != ENOENT) { // something serious has happened
           GLOBAL_ERROR_STREAM(fileName << endl);
           return -1;
           }
        // found a non existing file suffix
        }
     if (Open() >= 0) {
        if (!record && Offset >= 0 && lseek(file, Offset, SEEK_SET) != Offset) {
           GLOBAL_ERROR_STREAM(fileName << endl);
           return -1;
           }
        }
     return file;
     }
  GLOBAL_ERROR_STREAM("ERROR: max number of files "<< MAXFILESPERRECORDING<<" exceeded"<< endl );
  return -1;
}

int cFileName::NextFile(void)
{
  return SetOffset(fileNumber + 1);
}

// --- cRecordBuffer ---------------------------------------------------------

class cRecordBuffer : public cRingBuffer {
private:
  cDvbApi *dvbApi;
  cFileName fileName;
  cIndexFile *index;
  cRemux remux;
  uchar pictureType;
  int fileSize;
  bool todisk;          // true if saving to disk
  int nmm_buffer_size;  // size of generated NMM-Buffer
  NMM::Buffer* nmm_buffer;
  int full_size;
  bool collecting_buffers;
  int videoDev;
  int recordFile;
  bool recording;
  bool leave;
  time_t lastDiskSpaceCheck;
  bool RunningLowOnDiskSpace(void);
  bool NextFile(void);
protected:
  virtual void Input(void);
  virtual void Output(void);
public:
  cRecordBuffer(cDvbApi *DvbApi, const char *FileName, int VPid, int APid);
  virtual ~cRecordBuffer();
  };

cRecordBuffer::cRecordBuffer(cDvbApi* DvbApi, const char *FileName, int VPid, int APid)
:cRingBuffer(VIDEOBUFSIZE, true)
,fileName(FileName, true)
,remux(VPid, APid, true)
{
  dvbApi = DvbApi;
  index = NULL;
  pictureType = NO_PICTURE;
  fileSize = 0;
  nmm_buffer_size = 0;
  full_size = 0;

  todisk = true;
  if (*FileName == '\0') {
    todisk = false;
  }
  if (todisk) {
    GLOBAL_DEBUG_STREAM("DVBAPI(recordbuffer): writing to file " << FileName << endl);
    recordFile = fileName.Open();
  } else {
    recordFile = 1;
    nmm_buffer = DvbApi->parent_node->getNewBuffer(MAXBUFFERSIZE); // allocate a first buffer
  }
  recording = false;
  leave = false;
  collecting_buffers = false;
  lastDiskSpaceCheck = time(NULL);
  if (todisk) {
    if (!fileName.Name())
       return;
  // Create the index file:
    index = new cIndexFile(FileName, true);
    if (!index)
       GLOBAL_ERROR_STREAM("ERROR: can't allocate index"<<endl);
     // let's continue without index, so we'll at least have the recording
  }
  videoDev = dvbApi->SetModeRecord();
  Start();
}

cRecordBuffer::~cRecordBuffer()
{
  Stop();
  dvbApi->SetModeNormal(true);
  if (index)
    delete index;
  nmm_buffer->release();
}

bool cRecordBuffer::RunningLowOnDiskSpace(void)
{
  if (time(NULL) > lastDiskSpaceCheck + DISKCHECKINTERVAL) {
     uint Free = FreeDiskSpaceMB(fileName.Name());
     lastDiskSpaceCheck = time(NULL);
     if (Free < MINFREEDISKSPACE) {
        GLOBAL_DEBUG_STREAM("low disk space ("<< Free <<" MB, limit is "<< MINFREEDISKSPACE <<" MB)"<<endl);
        return true;
        }
     }
  return false;
}

bool cRecordBuffer::NextFile(void) {
  if (recordFile >= 0 && pictureType == I_FRAME) { // every file shall start with an I_FRAME
    if (todisk && (fileSize > MAXVIDEOFILESIZE || RunningLowOnDiskSpace())) {
      recordFile = fileName.NextFile();
      fileSize = 0;
    }

    if (collecting_buffers && (nmm_buffer_size != 0)) {
      if (nmm_buffer_size >= MAXBUFFERSIZE) {
        GLOBAL_ERROR_STREAM("!!!DVBAPI: WARNING!! BUFFERSIZE ("<< MAXBUFFERSIZE <<") TO SMALL!!!" << endl);
      }
      nmm_buffer->setUsed(nmm_buffer_size); // set the size of the buffer
      dvbApi->parent_node->enqueue(nmm_buffer);    // and add the buffer to the queue
      GLOBAL_DEBUG_STREAM(" DVBAPI: enqueueing buffer(" << nmm_buffer_size << ")" << endl);
      nmm_buffer_size = 0;
      nmm_buffer = dvbApi->parent_node->getNewBuffer(MAXBUFFERSIZE);
    }
  }
  return recordFile >= 0;
}

void cRecordBuffer::Input(void)
{
  GLOBAL_DEBUG_STREAM("DVBAPI(RecordBuffer): input thread started (pid="<< getpid() << ")"<< endl);

  uchar b[MINVIDEODATA];
  time_t t = time(NULL);
  recording = true;
  if (dvbApi->collecting_buffers == true)
    collecting_buffers = true;

  for (;;) {
    int r = read(videoDev, b, sizeof(b));
    if (r > 0) {
      uchar *p = b;
      while (r > 0) {
        int w = Put(p, r);
        p += w;
        r -= w;
      }
      t = time(NULL);
    }
    else if (r < 0) {
      if (errno != EAGAIN) {
        GLOBAL_ERROR_STREAM("DVBAPI: ERROR reading dvr-device");
        if (errno != EBUFFEROVERFLOW)
          break;
      }
    }
    else if (r == 0) {
      GLOBAL_ERROR_STREAM("DVBAPI: reading no data from dvr-device");
      
    }
    else if (time(NULL) - t > 5) {
      GLOBAL_ERROR_STREAM("DVBAPI(RecordBuffer) ERROR: video data stream broken" << endl);
      cThread::EmergencyExit(true);
    }
    
    cFile::FileReady(videoDev, 100);
    if (!recording){
      break;
    }
  }
  

  GLOBAL_DEBUG_STREAM("DVBAPI(RecordBuffer): input thread ended (pid="<< getpid()<<")" <<endl );
}

void cRecordBuffer::Output(void)
{
  GLOBAL_DEBUG_STREAM("DVBAPI(RecordBuffer): output thread started (pid="<< getpid() <<")"<<endl);

  uchar b[MINVIDEODATA];
  int r = 0;
  for (;;) {
      int g = Get(b + r, sizeof(b) - r);
      if (g > 0) {
         r += g;
         int Count = r, Result;
         const uchar *p = remux.Process(b, Count, Result, &pictureType);
         //GLOBAL_DEBUG_STREAM("DVBAPI: remuxed " << Result << " bytes" << endl);
         if (p) {
            if (!Busy() && pictureType == I_FRAME){ // finish the recording before the next 'I' frame
               break;
            }
            if (NextFile()) {
               if (index && pictureType != NO_PICTURE && todisk)
                  index->Write(pictureType, fileName.Number(), fileSize);
               while (Result > 0) {
                     int w;
                     if (todisk) {
                       w = write(recordFile, p, Result);
                       fileSize += w;
                    //   GLOBAL_ERROR_STREAM("DVBAPI: wrote " << w << " Bytes to file of size "<< fileSize << endl);
                       if (w < 0) {
                         GLOBAL_ERROR_STREAM("DVBAPI: ERROR writing to file: "<< fileName.Name() << endl);
                         recording = false;
                         return;
                       }
                     } else {
                       w = Result;
                     }
                     if (collecting_buffers) {
                       // Check if there is enough space
                       if ((nmm_buffer_size +w) > MAXBUFFERSIZE) {
                         nmm_buffer->setUsed(nmm_buffer_size);     // if not
                         dvbApi->parent_node->enqueue(nmm_buffer); // send the buffer
                         nmm_buffer_size = 0;                      // and get a new one
                         nmm_buffer = dvbApi->parent_node->getNewBuffer(MAXBUFFERSIZE);
                       }
                       memcpy((nmm_buffer->getData())+nmm_buffer_size,p,w);
                       //GLOBAL_DEBUG_STREAM("DVBAPI: wrote " << w << " Bytes to buffer of size "<< nmm_buffer_size << endl);
                       nmm_buffer_size += w;
                     }
                     p += w;
                     Result -= w;
                     full_size += w;
               }
            } else{
               break;
            }
         }
         if (Count > 0) {
            r -= Count;
            memmove(b, b + Count, r);
         }
         if (!recording){
            break;
         }
      } else {
        if(!Busy()){ break;} //If the device can't be read and we are not busy, break. Otherwise it loops forever
        usleep(1); // this keeps the CPU load low
      }
  }
  recording = false;

  GLOBAL_DEBUG_STREAM("DVBAPI(RecordBuffer): output thread ended (pid=" << getpid() <<")" << endl);
}

// --- cReplayBuffer ---------------------------------------------------------

class cReplayBuffer : public cRingBuffer {
private:
  cDvbApi *dvbApi;
  cIndexFile *index;
  cFileName fileName;
  int fileOffset;
  int videoDev, audioDev;
  int replayFile;
  bool eof;
  int blockInput, blockOutput;
  bool paused, fastForward, fastRewind;
  int lastIndex, stillIndex;
  bool NextFile(uchar FileNumber = 0, int FileOffset = -1);
  void Clear(bool Block = false);
  void Close(void);
  int ReadFrame(uchar *b, int Length, int Max);
  void StripAudioPackets(uchar *b, int Length);
  void DisplayFrame(uchar *b, int Length);
  int Resume(void);
  bool Save(void);
protected:
  virtual void Input(void);
  virtual void Output(void);
public:
  cReplayBuffer(cDvbApi *DvbApi, int VideoDev, int AudioDev, const char *FileName);
  virtual ~cReplayBuffer();
  void Pause(void);
  void Play(void);
  void Forward(void);
  void Backward(void);
  int SkipFrames(int Frames);
  void SkipSeconds(int Seconds);
  void Goto(int Position, bool Still = false);
  void GetIndex(int &Current, int &Total, bool SnapToIFrame = false);
  };

cReplayBuffer::cReplayBuffer(cDvbApi *DvbApi, int VideoDev, int AudioDev, const char *FileName)
:cRingBuffer(VIDEOBUFSIZE)
,fileName(FileName, false)
{
  dvbApi = DvbApi;
  index = NULL;
  fileOffset = 0;
  videoDev = VideoDev;
  audioDev = AudioDev;
  replayFile = fileName.Open();
  eof = false;
  blockInput = blockOutput = false;
  paused = fastForward = fastRewind = false;
  lastIndex = stillIndex = -1;
  if (!fileName.Name())
     return;
  // Create the index file:
  index = new cIndexFile(FileName, false);
  if (!index) {
     GLOBAL_ERROR_STREAM("ERROR: can't allocate index"<< endl);
     }
  else if (!index->Ok()) {
     delete index;
     index = NULL;
     }
  dvbApi->SetModeReplay();
  Start();
}

cReplayBuffer::~cReplayBuffer()
{
  Stop();
  Save();
  Close();
  dvbApi->SetModeNormal(false);
  delete index;
}

void cReplayBuffer::Input(void)
{
  GLOBAL_DEBUG_STREAM("DVBAPI(ReplayBuffer): input thread started (pid="<< getpid() <<")" << endl);

  int ResumeIndex = Resume();
  if (ResumeIndex >= 0)
     GLOBAL_DEBUG_STREAM("DVBAPI(ReplayBuffer): resuming replay at index "<< ResumeIndex<< " ("<< IndexToHMSF(ResumeIndex, true)<<")" << endl );

  int lastIndex = -1;
  int brakeCounter = 0;
  uchar b[MAXFRAMESIZE];
  while (Busy() && (blockInput || NextFile())) {
        if (!blockInput && stillIndex < 0) {
           int r = 0;
           if (fastForward && !paused || fastRewind) {
              int Index = (lastIndex >= 0) ? lastIndex : index->Get(fileName.Number(), fileOffset);
              uchar FileNumber;
              int FileOffset, Length;
              if (!paused || (brakeCounter++ % 24) == 0) // show every I_FRAME 24 times in rmSlowRewind mode to achieve roughly the same speed as in slow forward mode
                 Index = index->GetNextIFrame(Index, fastForward, &FileNumber, &FileOffset, &Length);
              if (Index >= 0) {
                 if (!NextFile(FileNumber, FileOffset))
                    break;
                 }
              else {
                 paused = fastForward = fastRewind = false;
                 Play();
                 continue;
                 }
              lastIndex = Index;
              r = ReadFrame(b, Length, sizeof(b));
              StripAudioPackets(b, Length);
              }
           else {
              lastIndex = -1;
              r = read(replayFile, b, sizeof(b));
              }
           if (r > 0) {
              uchar *p = b;
              while (r > 0 && Busy() && !blockInput) {
                    int w = Put(p, r);
                    p += w;
                    r -= w;
                    usleep(1); // this keeps the CPU load low
                    }
              }
           else if (r ==0)
              eof = true;
           else if (r < 0 && errno != EAGAIN) {
              LOG_ERROR;
              break;
              }
           }
        if (blockInput > 1)
           blockInput = 1;
        }

  GLOBAL_DEBUG_STREAM("DVBAPI(ReplayBuffer): input thread ended (pid="<< getpid() <<")" << endl);
}

void cReplayBuffer::Output(void)
{
  GLOBAL_DEBUG_STREAM("DVBAPI(ReplayBuffer): output thread started (pid="<< getpid()<<")" << endl);

  uchar b[MINVIDEODATA];
  while (Busy()) {
        int r = blockOutput ? 0 : Get(b, sizeof(b));
        if (r > 0) {
           uchar *p = b;
           while (r > 0 && Busy() && !blockOutput) {
                 cFile::FileReadyForWriting(videoDev, 100);
                 int w = write(videoDev, p, r);
                 if (w < 0) {
                    LOG_ERROR;
                    Stop();
                    return;
                    }
                 p += w;
                 r -= w;
                 fileOffset += w;
                 }
           }
        if (blockOutput > 1)
           blockOutput = 1;
        }

  GLOBAL_DEBUG_STREAM("DVBAPI(ReplayBuffer): output thread ended (pid="<<getpid()<<")"<< endl );
}

int cReplayBuffer::ReadFrame(uchar *b, int Length, int Max)
{
  if (Length > Max) {
     GLOBAL_ERROR_STREAM("DVBAPI(ReplayBuffer): ERROR: frame larger than buffer ("<< Length <<" > " << Max << ")" <<endl);
     Length = Max;
     }
  int r = read(replayFile, b, Length);
  if (r >= 0) {
     if (r != Length)
        GLOBAL_ERROR_STREAM("DVBAPI(ReplayBuffer): ERROR: got "<< r<<" byte while reading " << Length << endl);
     return r;
     }
  LOG_ERROR;
  return -1;
}

void cReplayBuffer::StripAudioPackets(uchar *b, int Length)
{
  for (int i = 0; i < Length - 6; i++) {
      if (b[i] == 0x00 && b[i + 1] == 0x00 && b[i + 2] == 0x01) {
         if ((b[i + 3] >= 0xC0) && (b[i + 3] <= 0xDF)) {
            // audio
            int n = b[i + 4] * 256 + b[i + 5];
            for (int j = i; j < Length && n--; j++)
              b[j] = 0x00;
         } else {
           if ((b[i + 3] >= 0xE0) && (b[i + 3] <= 0xEF)) // video
             i += b[i + 4] * 256 + b[i + 5];
         }
      }
  }
}

void cReplayBuffer::DisplayFrame(uchar *b, int Length)
{
  StripAudioPackets(b, Length);
  video_still_picture sp = { (char *)b, Length };
  CHECK(ioctl(audioDev, AUDIO_SET_AV_SYNC, false));
  CHECK(ioctl(audioDev, AUDIO_SET_MUTE, true));
  CHECK(ioctl(videoDev, VIDEO_STILLPICTURE, &sp));
}

void cReplayBuffer::Clear(bool Block)
{
  if (!(blockInput || blockOutput)) {
     blockInput = blockOutput = 2;
     time_t t0 = time(NULL);
     while ((blockInput > 1 || blockOutput > 1) && time(NULL) - t0 < 2)
           usleep(1);
     Lock();
     cRingBuffer::Clear();
     CHECK(ioctl(videoDev, VIDEO_FREEZE));
     CHECK(ioctl(videoDev, VIDEO_CLEAR_BUFFER));
     CHECK(ioctl(audioDev, AUDIO_CLEAR_BUFFER));
     }
  if (!Block) {
     blockInput = blockOutput = 0;
     Unlock();
     }
}

void cReplayBuffer::Pause(void)
{
  paused = !paused;
  CHECK(ioctl(videoDev, paused ? VIDEO_FREEZE : VIDEO_CONTINUE));
  if (fastForward || fastRewind) {
     if (paused)
        Clear();
     fastForward = fastRewind = false;
     }
  CHECK(ioctl(audioDev, AUDIO_SET_MUTE, paused));
  stillIndex = -1;
}

void cReplayBuffer::Play(void)
{
  if (fastForward || fastRewind || paused) {
     if (!paused)
        Clear();
     stillIndex = -1;
     CHECK(ioctl(videoDev, paused ? VIDEO_CONTINUE : VIDEO_PLAY));
     CHECK(ioctl(audioDev, AUDIO_SET_AV_SYNC, true));
     CHECK(ioctl(audioDev, AUDIO_SET_MUTE, false));
     fastForward = fastRewind = paused = false;
     }
}

void cReplayBuffer::Forward(void)
{
  if (index || paused) {
     if (!paused)
        Clear(true);
     stillIndex = -1;
     fastForward = !fastForward;
     fastRewind = false;
     if (paused)
        CHECK(ioctl(videoDev, fastForward ? VIDEO_SLOWMOTION : VIDEO_FREEZE, 2));
     CHECK(ioctl(audioDev, AUDIO_SET_AV_SYNC, !fastForward));
     CHECK(ioctl(audioDev, AUDIO_SET_MUTE, fastForward || paused));
     if (!paused)
        Clear(false);
     }
}

void cReplayBuffer::Backward(void)
{
  if (index) {
     Clear(true);
     stillIndex = -1;
     fastRewind = !fastRewind;
     fastForward = false;
     CHECK(ioctl(audioDev, AUDIO_SET_AV_SYNC, !fastRewind));
     CHECK(ioctl(audioDev, AUDIO_SET_MUTE, fastRewind || paused));
     Clear(false);
     }
}

void cReplayBuffer::Close(void)
{
  if (replayFile >= 0) {
     fileName.Close();
     replayFile = -1;
     fileOffset = 0;
     }
}

int cReplayBuffer::Resume(void)
{
  if (index) {
     int Index = index->GetResume();
     if (Index >= 0) {
        uchar FileNumber;
        int FileOffset;
        if (index->Get(Index, &FileNumber, &FileOffset) && NextFile(FileNumber, FileOffset))
           return Index;
        }
     }
  return -1;
}

bool cReplayBuffer::Save(void)
{
  if (index) {
     int Index = index->Get(fileName.Number(), fileOffset);
     if (Index >= 0) {
        Index -= RESUMEBACKUP;
        if (Index > 0)
           Index = index->GetNextIFrame(Index, false);
        else
           Index = 0;
        if (Index >= 0)
           return index->StoreResume(Index);
        }
     }
  return false;
}

int cReplayBuffer::SkipFrames(int Frames)
{
  if (index && Frames) {
     int Current, Total;
     GetIndex(Current, Total, true);
     int OldCurrent = Current;
     Current = index->GetNextIFrame(Current + Frames, Frames > 0);
     return Current >= 0 ? Current : OldCurrent;
     }
  return -1;
}

void cReplayBuffer::SkipSeconds(int Seconds)
{
  if (index && Seconds) {
     Clear(true);
     int Index = index->Get(fileName.Number(), fileOffset);
     if (Index >= 0) {
        if (Seconds < 0) {
           int sec = index->Last() / FRAMESPERSEC;
           if (Seconds < -sec)
              Seconds = -sec;
           }
        Index += Seconds * FRAMESPERSEC;
        if (Index < 0)
           Index = 1; // not '0', to allow GetNextIFrame() below to work!
        uchar FileNumber;
        int FileOffset;
        if (index->GetNextIFrame(Index, false, &FileNumber, &FileOffset) >= 0)
           NextFile(FileNumber, FileOffset);
        }
     Clear(false);
     Play();
     }
}

void cReplayBuffer::Goto(int Index, bool Still)
{
  if (index) {
     Clear(true);
     if (++Index <= 0)
        Index = 1; // not '0', to allow GetNextIFrame() below to work!
     uchar FileNumber;
     int FileOffset, Length;
     Index = index->GetNextIFrame(Index, false, &FileNumber, &FileOffset, &Length);
     if (Index >= 0 && NextFile(FileNumber, FileOffset) && Still) {
        stillIndex = Index;
        uchar b[MAXFRAMESIZE];
        int r = ReadFrame(b, Length, sizeof(b));
        if (r > 0)
           DisplayFrame(b, r);
        fileOffset += Length;
        paused = true;
        }
     else
        stillIndex = -1;
     Clear(false);
     }
}

void cReplayBuffer::GetIndex(int &Current, int &Total, bool SnapToIFrame)
{
  if (index) {
     if (stillIndex >= 0)
        Current = stillIndex;
     else {
        Current = index->Get(fileName.Number(), fileOffset);
        if (SnapToIFrame) {
           int i1 = index->GetNextIFrame(Current + 1, false);
           int i2 = index->GetNextIFrame(Current, true);
           Current = (abs(Current - i1) <= abs(Current - i2)) ? i1 : i2;
           }
        }
     Total = index->Last();
     }
  else
     Current = Total = -1;
}

bool cReplayBuffer::NextFile(uchar FileNumber, int FileOffset)
{
  if (FileNumber > 0) {
     fileOffset = FileOffset;
     replayFile = fileName.SetOffset(FileNumber, FileOffset);
     }
  else if (replayFile >= 0 && eof) {
     Close();
     replayFile = fileName.NextFile();
     }
  eof = false;
  return replayFile >= 0;
}

// --- cTransferBuffer -------------------------------------------------------

class cTransferBuffer : public cRingBuffer {
private:
  cDvbApi *dvbApi;
  int fromDevice, toDevice;
  cRemux remux;
protected:
  virtual void Input(void);
  virtual void Output(void);
public:
  cTransferBuffer(cDvbApi *DvbApi, int ToDevice, int VPid, int APid);
  virtual ~cTransferBuffer();
  };

cTransferBuffer::cTransferBuffer(cDvbApi *DvbApi, int ToDevice, int VPid, int APid)
:cRingBuffer(VIDEOBUFSIZE, true)
,remux(VPid, APid)
{
  dvbApi = DvbApi;
  fromDevice = dvbApi->SetModeRecord();
  toDevice = ToDevice;
  Start();
}

cTransferBuffer::~cTransferBuffer()
{
  Stop();
  dvbApi->SetModeNormal(true);
}

void cTransferBuffer::Input(void)
{
  GLOBAL_DEBUG_STREAM("DVBAPI(TransferBuffer): input thread started (pid="<< getpid()<< ")" << endl);

  uchar b[MINVIDEODATA];
  int n = 0;
  while (Busy()) {
        cFile::FileReady(fromDevice, 100);
        int r = read(fromDevice, b + n, sizeof(b) - n);
        if (r > 0) {
           n += r;
           int Count = n, Result;
           const uchar *p = remux.Process(b, Count, Result);
           if (p) {
              while (Result > 0 && Busy()) {
                    int w = Put(p, Result);
                    p += w;
                    Result -= w;
                    }
              }
           if (Count > 0) {
              n -= Count;
              memmove(b, b + Count, n);
              }
           }
        else if (r < 0) {
           if (errno != EAGAIN) {
              LOG_ERROR;
              if (errno != EBUFFEROVERFLOW)
                 break;
              }
           }
        }

  GLOBAL_DEBUG_STREAM("DVBAPI(TransferBuffer): input thread ended (pid="<< getpid()<< ")" << endl);
}

void cTransferBuffer::Output(void)
{
  GLOBAL_DEBUG_STREAM("DVBAPI(TransferBuffer): output thread started (pid="<< getpid()<< ")" << endl);

  bool GotBufferReserve = false;
  uchar b[MINVIDEODATA];
  while (Busy()) {
        if (!GotBufferReserve) {
           if (Available() < MAXFRAMESIZE)
              usleep(100000); // allow the buffer to collect some reserve
           else
              GotBufferReserve = true;
           }
        int r = Get(b, sizeof(b));
        if (r > 0) {
           uchar *p = b;
           while (r > 0 && Busy()) {
                 int w = write(toDevice, p, r);
                 if (w < 0) {
                    LOG_ERROR;
                    Stop();
                    return;
                    }
                 p += w;
                 r -= w;
                 }
           }
        else
           usleep(1); // this keeps the CPU load low
        }

  GLOBAL_DEBUG_STREAM("DVBAPI(TransferBuffer): output thread ended (pid="<< getpid()<< ")" << endl);
}

/*
// --- cCuttingBuffer --------------------------------------------------------

class cCuttingBuffer : public cThread {
private:
  bool active;
  int fromFile, toFile;
  cFileName *fromFileName, *toFileName;
  cIndexFile *fromIndex, *toIndex;
  cMarks fromMarks, toMarks;
protected:
  virtual void Action(void);
public:
  cCuttingBuffer(const char *FromFileName, const char *ToFileName);
  virtual ~cCuttingBuffer();
  };

cCuttingBuffer::cCuttingBuffer(const char *FromFileName, const char *ToFileName)
{
  active = false;
  fromFile = toFile = -1;
  fromFileName = toFileName = NULL;
  fromIndex = toIndex = NULL;
  if (fromMarks.Load(FromFileName) && fromMarks.Count()) {
     fromFileName = new cFileName(FromFileName, false, true);
     toFileName = new cFileName(ToFileName, true, true);
     fromIndex = new cIndexFile(FromFileName, false);
     toIndex = new cIndexFile(ToFileName, true);
     toMarks.Load(ToFileName); // doesn't actually load marks, just sets the file name
     Start();
     }
  else
     esyslog(LOG_ERR, "no editing marks found for %s", FromFileName);
}

cCuttingBuffer::~cCuttingBuffer()
{
  active = false;
  Cancel(3);
  delete fromFileName;
  delete toFileName;
  delete fromIndex;
  delete toIndex;
}

void cCuttingBuffer::Action(void)
{
  dsyslog(LOG_INFO, "video cutting thread started (pid=%d)", getpid());

  cMark *Mark = fromMarks.First();
  if (Mark) {
     fromFile = fromFileName->Open();
     toFile = toFileName->Open();
     active = fromFile >= 0 && toFile >= 0;
     int Index = Mark->position;
     Mark = fromMarks.Next(Mark);
     int FileSize = 0;
     int CurrentFileNumber = 0;
     int LastIFrame = 0;
     toMarks.Add(0);
     toMarks.Save();
     uchar buffer[MAXFRAMESIZE];
     while (active) {
           uchar FileNumber;
           int FileOffset, Length;
           uchar PictureType;

           // Read one frame:

           if (fromIndex->Get(Index++, &FileNumber, &FileOffset, &PictureType, &Length)) {
              if (FileNumber != CurrentFileNumber) {
                 fromFile = fromFileName->SetOffset(FileNumber, FileOffset);
                 CurrentFileNumber = FileNumber;
                 }
              if (fromFile >= 0) {
                 if (Length <= (int)sizeof(buffer)) {
                    if (read(fromFile, buffer, Length) < 0) {
                       LOG_ERROR;
                       break;
                       }
                    }
                 else {
                    esyslog(LOG_ERR, "ERROR: frame larger than buffer (%d > %d)", Length, sizeof(buffer));
                    break;
                    }
                 }
              else
                 break;
              }
           else
              break;

           // Write one frame:

           if (PictureType == I_FRAME) { // every file shall start with an I_FRAME
              if (FileSize > MAXVIDEOFILESIZE) {
                 toFile = toFileName->NextFile();
                 if (toFile < 0)
                    break;
                 FileSize = 0;
                 }
              LastIFrame = 0;
              }
           write(toFile, buffer, Length);
           toIndex->Write(PictureType, toFileName->Number(), FileSize);
           FileSize += Length;
           if (!LastIFrame)
              LastIFrame = toIndex->Last();

           // Check editing marks:

           if (Mark && Index >= Mark->position) {
              Mark = fromMarks.Next(Mark);
              if (Mark) {
                 Index = Mark->position;
                 Mark = fromMarks.Next(Mark);
                 CurrentFileNumber = 0; // triggers SetOffset before reading next frame
                 toMarks.Add(LastIFrame);
                 toMarks.Add(toIndex->Last() + 1);
                 toMarks.Save();
                 }
              else
                 break; // final end mark reached
              }
           }
     }
  else
     esyslog(LOG_ERR, "no editing marks found!");
  dsyslog(LOG_INFO, "end video cutting thread");
}

// --- cVideoCutter ----------------------------------------------------------

cCuttingBuffer *cVideoCutter::cuttingBuffer = NULL;

bool cVideoCutter::Start(const char *FileName)
{
  if (!cuttingBuffer) {
     const char *EditedVersionName = PrefixVideoFileName(FileName, '%');
     if (EditedVersionName && RemoveVideoFile(EditedVersionName) && MakeDirs(EditedVersionName, true)) {
        cuttingBuffer = new cCuttingBuffer(FileName, EditedVersionName);
        return true;
        }
     }
  return false;
}

void cVideoCutter::Stop(void)
{
  delete cuttingBuffer;
  cuttingBuffer = NULL;
}

bool cVideoCutter::Active(void)
{
  if (cuttingBuffer) {
     if (cuttingBuffer->Active())
        return true;
     Stop();
     }
  return false;
}
*/
// --- cDvbApi ---------------------------------------------------------------


static const char *OstName(const char *Name, int n)
{
  static char buffer[_POSIX_PATH_MAX];
  snprintf(buffer, sizeof(buffer), "%s%d", Name, n);
  return buffer;
}

static int OstOpen
(const char *Name, int n, int Mode, bool ReportError = false)
{
  const char *FileName = OstName(Name, n);
  int fd = open(FileName, Mode);
  if (fd < 0 && ReportError)
     LOG_ERROR_STR(FileName);
  return fd;
}

int cDvbApi::NumDvbApis = 0;
int cDvbApi::useDvbApi = 0;
cDvbApi *cDvbApi::dvbApi[MAXDVBAPI] = { NULL };
cDvbApi *cDvbApi::PrimaryDvbApi = NULL;

cDvbApi::cDvbApi(int n)
{
  fd_osd = fd_qpskfe = fd_qamfe = fd_sec = fd_dvr = fd_audio = fd_video = fd_demuxa = fd_demuxv = fd_demuxt = fd_frontend = -1;

  vPid = aPid = 0;
  siProcessor = NULL;
  recordBuffer = NULL;
  replayBuffer = NULL;
  transferBuffer = NULL;
  transferringFromDvbApi = NULL;
  ca = 0;
  priority = -1;

  // Devices that are only present on DVB-C or DVB-S cards:

  //fd_qamfe  = OstOpen(DEV_OST_QAMFE,  n, O_RDWR);
  //fd_qpskfe = OstOpen(DEV_OST_QPSKFE, n, O_RDWR);
  //fd_sec    = OstOpen(DEV_OST_SEC,    n, O_RDWR);

  fd_frontend = OstOpen(DEV_OST_FRONT, n, O_RDWR | O_NONBLOCK);

  // Devices that all DVB cards must have:

  fd_demuxv = OstOpen(DEV_OST_DEMUX,  n, O_RDWR | O_NONBLOCK, true);
  fd_demuxa = OstOpen(DEV_OST_DEMUX,  n, O_RDWR | O_NONBLOCK, true);
  fd_demuxt = OstOpen(DEV_OST_DEMUX,  n, O_RDWR | O_NONBLOCK, true);

  // Devices not present on "budget" cards:

  fd_osd    = OstOpen(DEV_OST_OSD,    n, O_RDWR);
  fd_video  = OstOpen(DEV_OST_VIDEO,  n, O_RDWR | O_NONBLOCK);
  fd_audio  = OstOpen(DEV_OST_AUDIO,  n, O_RDWR | O_NONBLOCK);

  // Devices that may not be available, and are not necessary for normal operation:

  //videoDev  = OstOpen(DEV_VIDEO,      n, O_RDWR);

  // Devices that will be dynamically opened and closed when necessary:

  fd_dvr    = -1;

  // We only check the devices that must be present - the others will be checked before accessing them:

  //if (! ((fd_qpskfe >= 0 && fd_sec >= 0) || fd_qamfe >= 0) && fd_demuxv >= 0 && fd_demuxa >= 0 && fd_demuxt >= 0) {
     siProcessor = new cSIProcessor(OstName(DEV_OST_DEMUX, n));
     if (!dvbApi[0]) // only the first one shall set the system time
        siProcessor->SetUseTSTime(Setup.SetSystemTime);
     //}else{ 
     //  GLOBAL_ERROR_STREAM("ERROR: can't open video device "<< n <<endl);
     //}
  currentChannel = 1;
}

cDvbApi::~cDvbApi()
{
  delete siProcessor;
//  Close();
  StopReplay();
  StopRecord();
  StopTransfer();

  // We're not explicitly closing any device files here, since this sometimes
  // caused segfaults. Besides, the program is about to terminate anyway...

  GLOBAL_DEBUG_STREAM("Closing Devices" << endl);
  close(fd_qamfe);  
  close(fd_qpskfe); 
  close(fd_sec);    
  close(fd_demuxv);
  close(fd_demuxa);
  close(fd_demuxt);
  close(fd_osd);    
  close(fd_video);
  close(fd_audio);  
  close(fd_frontend);
}

void cDvbApi::SetUseDvbApi(int n)
{
  if (n < MAXDVBAPI)
     useDvbApi |= (1 << n);
}

bool cDvbApi::SetPrimaryDvbApi(int n)
{
  n--;
  if (0 <= n && n < NumDvbApis && dvbApi[n]) {
     GLOBAL_DEBUG_STREAM("setting primary DVB to "<< (n + 1) <<endl);
     PrimaryDvbApi = dvbApi[n];
     return true;
     }
  GLOBAL_ERROR_STREAM("invalid DVB interface: " << (n + 1) << endl);
  return false;
}

cDvbApi *cDvbApi::GetDvbApi(int Ca, int Priority)
{
  cDvbApi *d = NULL, *dMinPriority = NULL;
  int index = Ca - 1;
  for (int i = 0; i < MAXDVBAPI; i++) {
      if (dvbApi[i]) {
         if (i == index) { // means we need exactly _this_ device
            d = dvbApi[i];
            break;
            }
         else if (Ca == 0) { // means any device would be acceptable
            if (!d || !dvbApi[i]->Recording() || (d->Recording() && d->Priority() > dvbApi[i]->Priority()))
               d = dvbApi[i]; // this is one that is either not currently recording or has the lowest priority
            if (d && d != PrimaryDvbApi && !d->Recording()) // avoids the PrimaryDvbApi if possible
               break;
            if (d && d->Recording() && d->Priority() < Setup.PrimaryLimit && (!dMinPriority || d->Priority() < dMinPriority->Priority()))
               dMinPriority = d; // this is the one with the lowest priority below Setup.PrimaryLimit
            }
         }
      }
  if (d == PrimaryDvbApi) { // the PrimaryDvbApi was the only one that was free
     if (Priority < Setup.PrimaryLimit)
        return NULL;        // not enough priority to use the PrimaryDvbApi
     if (dMinPriority)      // there's one that must not use the PrimaryDvbApi...
        d = dMinPriority;   // ...so let's kick out that one
     }
  return (d                           // we found one...
      && (!d->Recording()             // ...that's either not currently recording...
          || d->Priority() < Priority // ...or has a lower priority...
          || (!d->Ca() && Ca)))       // ...or doesn't need this card
          ? d : NULL;
}

int cDvbApi::Index(void)
{
  for (int i = 0; i < MAXDVBAPI; i++) {
      if (dvbApi[i] == this)
         return i;
      }
  return -1;
}

bool cDvbApi::Probe(const char *FileName)
{
  if (access(FileName, F_OK) == 0) {
     GLOBAL_DEBUG_STREAM("DVBAPI: probing "<< FileName<< endl);
     int f = open(FileName, O_RDONLY);
     if (f >= 0) {
        close(f);
        return true;
        }
     else if (errno != ENODEV && errno != EINVAL)
        GLOBAL_ERROR_STREAM(FileName<<endl);
     }
  else if (errno != ENOENT)
     GLOBAL_ERROR_STREAM(FileName<<endl);
  return false;
}

static bool is_initialized = false;
bool cDvbApi::Init(void)
{
  if (!is_initialized) {
    NumDvbApis = 0;
    for (int i = 0; i < MAXDVBAPI; i++) {
      if (useDvbApi == 0 || (useDvbApi & (1 << i)) != 0) {

        if (Probe(OstName(DEV_OST_DVR, i)))
          dvbApi[NumDvbApis++] = new cDvbApi(i);
        else
          break; 
      }
    }
    PrimaryDvbApi = dvbApi[0];
    if (NumDvbApis > 0) {
      GLOBAL_DEBUG_STREAM("DVBAPI: found "<< NumDvbApis <<" video devices" << endl);
    } else {
      GLOBAL_ERROR_STREAM("DVBAPI: ERROR: no video device found, giving up!" << endl);
    }
    is_initialized = true;
  }
  return NumDvbApis > 0;
}

void cDvbApi::Cleanup(void)
{

  for (int i = 0; i < MAXDVBAPI; i++) {
      delete dvbApi[i];
      dvbApi[i] = NULL;
      }
  PrimaryDvbApi = NULL;
  is_initialized = false;
}
int cDvbApi::muteAudio(bool mode) {
    if (mode) {
      GLOBAL_DEBUG_STREAM("muting audio..." << endl);
    } else {
      GLOBAL_DEBUG_STREAM("unmuting audio..." << endl);
    }
    if (PrimaryDvbApi) {
      if (ioctl(PrimaryDvbApi->fd_audio,AUDIO_SET_MUTE,mode) < 0) {
        return 0;
      }
    }
    return 1;
}

const cSchedules *cDvbApi::Schedules(cThreadLock* ) const
{
  if (siProcessor)// && ThreadLock->Lock(siProcessor))
     return siProcessor->Schedules();
  return NULL;
}
/*
bool cDvbApi::GrabImage(const char *FileName, bool Jpeg, int Quality, int SizeX, int SizeY)
{
  if (videoDev < 0)
     return false;
  int result = 0;
  // just do this once?
  struct video_mbuf mbuf;
  result |= ioctl(videoDev, VIDIOCGMBUF, &mbuf);
  int msize = mbuf.size;
  // gf: this needs to be a protected member of cDvbApi! //XXX kls: WHY???
  unsigned char *mem = (unsigned char *)mmap(0, msize, PROT_READ | PROT_WRITE, MAP_SHARED, videoDev, 0);
  if (!mem || mem == (unsigned char *)-1)
     return false;
  // set up the size and RGB
  struct video_capability vc;
  result |= ioctl(videoDev, VIDIOCGCAP, &vc);
  struct video_mmap vm;
  vm.frame = 0;
  if ((SizeX > 0) && (SizeX <= vc.maxwidth) &&
      (SizeY > 0) && (SizeY <= vc.maxheight)) {
     vm.width = SizeX;
     vm.height = SizeY;
     }
  else {
     vm.width = vc.maxwidth;
     vm.height = vc.maxheight;
     }
  vm.format = VIDEO_PALETTE_RGB24;
  // this needs to be done every time:
  result |= ioctl(videoDev, VIDIOCMCAPTURE, &vm);
  result |= ioctl(videoDev, VIDIOCSYNC, &vm.frame);
  // make RGB out of BGR:
  int memsize = vm.width * vm.height;
  unsigned char *mem1 = mem;
  for (int i = 0; i < memsize; i++) {
      unsigned char tmp = mem1[2];
      mem1[2] = mem1[0];
      mem1[0] = tmp;
      mem1 += 3;
      }

  if (Quality < 0)
     Quality = 255; //XXX is this 'best'???

  isyslog(LOG_INFO, "grabbing to %s (%s %d %d %d)", FileName, Jpeg ? "JPEG" : "PNM", Quality, vm.width, vm.height);
  FILE *f = fopen(FileName, "wb");
  if (f) {
     if (Jpeg) {
        // write JPEG file:
        struct jpeg_compress_struct cinfo;
        struct jpeg_error_mgr jerr;
        cinfo.err = jpeg_std_error(&jerr);
        jpeg_create_compress(&cinfo);
        jpeg_stdio_dest(&cinfo, f);
        cinfo.image_width = vm.width;
        cinfo.image_height = vm.height;
        cinfo.input_components = 3;
        cinfo.in_color_space = JCS_RGB;

        jpeg_set_defaults(&cinfo);
        jpeg_set_quality(&cinfo, Quality, true);
        jpeg_start_compress(&cinfo, true);

        int rs = vm.width * 3;
        JSAMPROW rp[vm.height];
        for (int k = 0; k < vm.height; k++)
            rp[k] = &mem[rs * k];
        jpeg_write_scanlines(&cinfo, rp, vm.height);
        jpeg_finish_compress(&cinfo);
        jpeg_destroy_compress(&cinfo);
        }
     else {
        // write PNM file:
        if (fprintf(f, "P6\n%d\n%d\n255\n", vm.width, vm.height) < 0 ||
            fwrite(mem, vm.width * vm.height * 3, 1, f) < 0) {
           LOG_ERROR_STR(FileName);
           result |= 1;
           }
        }
     fclose(f);
     }
  else {
     LOG_ERROR_STR(FileName);
     result |= 1;
     }

  if (ovlStat && ovlGeoSet) {
     // switch the Overlay on again (gf: why have i to do anything again?)
     OvlG(ovlSizeX, ovlSizeY, ovlPosX, ovlPosY);
     }
  if (ovlFbSet)
     OvlP(ovlBrightness, ovlColour, ovlHue, ovlContrast);

  munmap(mem, msize);
  return result == 0;
}
*/

int cDvbApi::SetModeRecord(void)
{
  // Sets up the DVB device for recording

  SetPids(true);
  if (fd_dvr >= 0){
     close(fd_dvr);
  }
  fd_dvr = OstOpen(DEV_OST_DVR, Index(), O_RDONLY | O_NONBLOCK);
  //fd_dvr = OstOpen(DEV_OST_DVR, Index(), O_RDONLY);
  if (fd_dvr < 0) {
     GLOBAL_ERROR_STREAM("DVBAPI: unable to open dvr device " << endl);
  }
  return fd_dvr;
}

void cDvbApi::SetModeReplay(void)
{
  // Sets up the DVB device for replay

  if (fd_video >= 0 && fd_audio >= 0) {
    if (siProcessor)
      siProcessor->SetStatus(false);
     CHECK(ioctl(fd_video, VIDEO_SET_BLANK, true));
     CHECK(ioctl(fd_audio, AUDIO_SELECT_SOURCE, AUDIO_SOURCE_MEMORY));
     CHECK(ioctl(fd_audio, AUDIO_SET_AV_SYNC, true));
     CHECK(ioctl(fd_audio, AUDIO_PLAY));
     CHECK(ioctl(fd_video, VIDEO_SELECT_SOURCE, VIDEO_SOURCE_MEMORY));
     CHECK(ioctl(fd_video, VIDEO_PLAY));
     }
}

void cDvbApi::SetModeNormal(bool FromRecording)
{
  // Puts the DVB device back into "normal" viewing mode (after replay or recording)

  if (FromRecording) {
     close(fd_dvr);
     fd_dvr = -1;
     SetPids(false);
     }
  else {
     if (fd_video >= 0 && fd_audio >= 0) {
        CHECK(ioctl(fd_video, VIDEO_STOP, true));
        CHECK(ioctl(fd_audio, AUDIO_STOP, true));
        CHECK(ioctl(fd_video, VIDEO_CLEAR_BUFFER));
        CHECK(ioctl(fd_audio, AUDIO_CLEAR_BUFFER));
        CHECK(ioctl(fd_video, VIDEO_SELECT_SOURCE, VIDEO_SOURCE_DEMUX));
        CHECK(ioctl(fd_audio, AUDIO_SELECT_SOURCE, AUDIO_SOURCE_DEMUX));
        CHECK(ioctl(fd_audio, AUDIO_SET_AV_SYNC, true));
//        CHECK(ioctl(fd_audio, AUDIO_SET_MUTE, false));
        if (siProcessor)
          siProcessor->SetStatus(true);
        }
     }
}

bool cDvbApi::SetPid(int fd, dmx_pes_type_t PesType, int Pid, dmx_output_t Output)
{

  if(Output == DMX_OUT_TS_TAP)
    {
      int buffersize = 64 * 1024;
      if(ioctl(fd, DMX_SET_BUFFER_SIZE, buffersize) == -1)
        GLOBAL_ERROR_STREAM("DMX_SET_BUFFER_SIZE failed" << endl);
    }

  dmx_pes_filter_params pesFilterParams;
  if (Pid <= 0 || Pid >= 0xFFFF) {
    //CHECK(ioctl(fd, DMX_STOP, Pid));
     return true;
     }
  pesFilterParams.pid     = Pid;
  pesFilterParams.input   = DMX_IN_FRONTEND;
  pesFilterParams.output  = Output;
  pesFilterParams.pes_type = PesType;
  pesFilterParams.flags   = DMX_IMMEDIATE_START;
  if (ioctl(fd, DMX_SET_PES_FILTER, &pesFilterParams) < 0) {
    GLOBAL_ERROR_STREAM("failed to set pid: " << PesType << endl);
     return false;
     }
  return true;
}

bool cDvbApi::SetPids(bool ForRecording)
{
  return SetVpid(vPid, ForRecording ? DMX_OUT_TS_TAP : DMX_OUT_DECODER) &&
         SetApid(aPid, ForRecording ? DMX_OUT_TS_TAP : DMX_OUT_DECODER);
}



bool cDvbApi::SetChannel(int ChannelNumber, int FrequencyMHz, char Polarization, int Diseqc, int Srate, int Vpid, int Apid, int , int Ca, int Pnr)
{
  // Make sure the siProcessor won't access the device while switching
  cThreadLock ThreadLock(siProcessor);

  bool recordingflag = Recording();
  if (recordingflag) {
    StopRecord();
  }

  StopTransfer();
  StopReplay();

  // Avoid noise while switching:

  if (fd_video >= 0 && fd_audio >= 0) {
     //CHECK(ioctl(fd_audio, AUDIO_SET_MUTE, true));
     CHECK(ioctl(fd_video, VIDEO_SET_BLANK, true));
     CHECK(ioctl(fd_video, VIDEO_CLEAR_BUFFER));
     CHECK(ioctl(fd_audio, AUDIO_CLEAR_BUFFER));
     }

  //bool ChannelSynced = false;
  unsigned int freq = FrequencyMHz;
  int tone = SEC_TONE_OFF;
  int pol;

  if((Polarization == 'v') || (Polarization == 'V'))
    {
      pol = SEC_VOLTAGE_13;
    }else{
      pol = SEC_VOLTAGE_18;
    }

  if (freq < (unsigned int)Setup.LnbSLOF) {
    freq -= Setup.LnbFrequLo;
    tone = SEC_TONE_OFF;
  }
  else {
    freq -= Setup.LnbFrequHi;
    tone = SEC_TONE_ON;
  }
  //!!!!!!!!!! NEW CODE !!!!!!!!!!!!!!!!!!! start
  
  dvb_frontend_parameters frontParam;
  dvb_frontend_info info;
  //dvb_frontend_event ev;

  if (ioctl(fd_frontend, FE_GET_INFO, &info) < 0)
    GLOBAL_ERROR_STREAM(" error getting fd-info" << endl);

  frontParam.frequency = freq * 1000UL;
  frontParam.inversion = INVERSION_OFF;

  if(info.type == FE_QPSK)
    {
      //DiSeqC
      struct dvb_diseqc_master_cmd cmd = {{0xe0, 0x10, 0x38, 0xf0, 0x00, 0x00}, 4};
      cmd.msg[3] = 0xf0 | (((Diseqc * 4) & 0x0f) | (tone == SEC_TONE_ON ? 1 : 0) | (pol == SEC_VOLTAGE_13 ? 0 : 2));

      if(ioctl(fd_frontend, FE_SET_TONE, SEC_TONE_OFF) < 0)
        return false;

      if(ioctl(fd_frontend, FE_SET_VOLTAGE, pol) < 0)
        return false;
      usleep(15 * 1000);

      if(ioctl(fd_frontend, FE_DISEQC_SEND_MASTER_CMD, &cmd) < 0)
        return false;
      usleep(15 * 1000);

      if(ioctl(fd_frontend, FE_DISEQC_SEND_BURST, (Diseqc / 4) % 2 ? SEC_MINI_B : SEC_MINI_A) < 0)
        return false;
      usleep(15 * 1000);

      if(ioctl(fd_frontend, FE_SET_TONE, tone) < 0)
        return false;

      //while(false){
      //if(ioctl(fd_frontend, FE_GET_EVENT, &ev) == -1)
      //  cerr << "event received" << endl;
      //    break;
      //}

      frontParam.u.qpsk.symbol_rate = Srate * 1000UL;
      frontParam.u.qpsk.fec_inner = FEC_AUTO;


    }
  else if(info.type == FE_QAM)
    {
      frontParam.u.qam.symbol_rate = Srate * 1000UL;
      frontParam.u.qam.fec_inner = FEC_AUTO;
      frontParam.u.qam.modulation = QAM_64;
    }

  if(ioctl(fd_frontend, FE_SET_FRONTEND, &frontParam) < 0)
    GLOBAL_ERROR_STREAM("Error setting DVB-Frontend" << endl);


  
  vPid = Vpid;
  aPid = Apid;
  if (!SetPids(true)) {  //org:if (!SetPids(false)) {
    GLOBAL_ERROR_STREAM("DVBAPI: ERROR: failed to set PIDs for channel "<< ChannelNumber << endl);
    return false;
  }

  usleep(1000000);


  fe_status_t status;
  if(ioctl(fd_frontend, FE_READ_STATUS, &status) < 0)
    {
      GLOBAL_ERROR_STREAM("Error getting Frontend-staus" << endl);
      return false;
    }
  currentChannel = ChannelNumber;
  GLOBAL_DEBUG_STREAM( "setchannel done" << endl);

  if (siProcessor)
    siProcessor->SetCurrentServiceID(Pnr);
  

  uint16_t snr, signal;
  uint32_t ber, uncorrected_blocks;
    
  ioctl(fd_frontend, FE_READ_STATUS, &status);
  ioctl(fd_frontend, FE_READ_SIGNAL_STRENGTH, &signal);
  ioctl(fd_frontend, FE_READ_SNR, &snr);
  ioctl(fd_frontend, FE_READ_BER, &ber);
  ioctl(fd_frontend, FE_READ_UNCORRECTED_BLOCKS, &uncorrected_blocks);
  
  printf ("status %02x | signal %04x | snr %04x | ber %08x | unc %08x | ",
          status, signal, snr, ber, uncorrected_blocks);
  
  if (status && FE_HAS_LOCK){
    printf("FE_HAS_LOCK");
  }else{
    return false;
  }
  
  printf("\n");

  StartRecord(Ca, 0);

  return true;
  //!!!!!!!!! NEW CODE !!!!!!!!!!!!!!!!!!!! end

}

bool cDvbApi::Transferring(void)
{
  return transferBuffer;
}

cDvbApi *cDvbApi::StartTransfer(int TransferToVideoDev)
{
  StopTransfer();
  transferBuffer = new cTransferBuffer(this, TransferToVideoDev, vPid, aPid);
  return this;
}

void cDvbApi::StopTransfer(void)
{
  if (transferBuffer) {
     delete transferBuffer;
     transferBuffer = NULL;
     }
  if (transferringFromDvbApi) {
     transferringFromDvbApi->StopTransfer();
     transferringFromDvbApi = NULL;
     }
}

int cDvbApi::SecondsToFrames(int Seconds)
{
  return Seconds * FRAMESPERSEC;
}

bool cDvbApi::Recording(void)
{
  if (recordBuffer && !recordBuffer->Active())
     StopRecord();
  return recordBuffer != NULL;
}

bool cDvbApi::collectingBuffers(void) {
	return collecting_buffers;
}

bool cDvbApi::Replaying(void)
{
  if (replayBuffer && !replayBuffer->Active())
     StopReplay();
  return replayBuffer != NULL;
}

bool cDvbApi::StartRecord(const char *FileName, int Ca, int Priority)
{
  if (Recording()) {
    GLOBAL_ERROR_STREAM("DVBAPI ERROR: StartRecord() called while recording - ignored!" << endl);
    return false;
  }
  GLOBAL_DEBUG_STREAM("DVBAPI: recording to disk" << endl);
  StopTransfer();

  StopReplay(); // TODO: remove this if the driver is able to do record and replay at the same time

  // Check FileName:

  if (!FileName) {
     GLOBAL_ERROR_STREAM( "DVBAPI ERROR: StartRecord: file name is (null)" << endl);
     return false;
     }
  GLOBAL_DEBUG_STREAM("DVBAPI: record "<< FileName<< endl);

  // Create directories if necessary:

  if (!MakeDirs(FileName, true))
     return false;

  // Create recording buffer:

  recordBuffer = new cRecordBuffer(this, FileName, vPid, aPid);

  if (recordBuffer) {
     ca = Ca;
     priority = Priority;
     return true;
     }
  else {
     GLOBAL_ERROR_STREAM("DVBAPI: ERROR: can't allocate recording buffer"<< endl);
  }
  return false;
}

bool cDvbApi::StartRecord(int Ca, int Priority)
{
  if (Recording()) {
    GLOBAL_ERROR_STREAM("DVBAPI: ERROR: StartRecord() called while recording - ignored!" << endl);
    return false;
  }
  GLOBAL_DEBUG_STREAM("DVBAPI: recording to buffer" << endl);

  StopTransfer();

  StopReplay(); // TODO: remove this if the driver is able to do record and replay at the same time

  const char* FileName = "";

  // Create recording buffer:
  recordBuffer = new cRecordBuffer(this, FileName, vPid, aPid);

  if (recordBuffer) {
     ca = Ca;
     priority = Priority;
     return true;
     }
  else {
     GLOBAL_ERROR_STREAM("DVBAPI: ERROR: can't allocate recording buffer"<< endl);
  }
  return false;
}

void cDvbApi::StopRecord(void)
{
  if (recordBuffer) {
     delete recordBuffer;
     recordBuffer = NULL;
     ca = 0;
     priority = -1;
  }
}

bool cDvbApi::StartReplay(const char *FileName)
{
  if (Recording()) {
     GLOBAL_ERROR_STREAM("DVBAPI: ERROR: StartReplay() called while recording - ignored!"<< endl);
     return false;
     }
  StopTransfer();
  StopReplay();
  if (fd_video >= 0 && fd_audio >= 0) {

     // Check FileName:

     if (!FileName) {
        GLOBAL_ERROR_STREAM("DVBAPI: ERROR: StartReplay: file name is (null)"<< endl);
        return false;
        }
     GLOBAL_DEBUG_STREAM("DVBAPI: replay "<< FileName<< endl);

     // Create replay buffer:

     replayBuffer = new cReplayBuffer(this, fd_video, fd_audio, FileName);
     if (replayBuffer)
        return true;
     else
        GLOBAL_ERROR_STREAM("DVBAPI: ERROR: can't allocate replaying buffer" << endl);
     }
  return false;
}

void cDvbApi::StopReplay(void)
{
  if (replayBuffer) {
     delete replayBuffer;
     replayBuffer = NULL;
     }
}

void cDvbApi::Pause(void)
{
  if (replayBuffer)
     replayBuffer->Pause();
}

void cDvbApi::Play(void)
{
  if (replayBuffer)
     replayBuffer->Play();
}

void cDvbApi::Forward(void)
{
  if (replayBuffer)
     replayBuffer->Forward();
}

void cDvbApi::Backward(void)
{
  if (replayBuffer)
     replayBuffer->Backward();
}

void cDvbApi::SkipSeconds(int Seconds)
{
  if (replayBuffer)
     replayBuffer->SkipSeconds(Seconds);
}

int cDvbApi::SkipFrames(int Frames)
{
  if (replayBuffer)
     return replayBuffer->SkipFrames(Frames);
  return -1;
}

bool cDvbApi::GetIndex(int &Current, int &Total, bool SnapToIFrame)
{
  if (replayBuffer) {
     replayBuffer->GetIndex(Current, Total, SnapToIFrame);
     return true;
     }
  return false;
}

void cDvbApi::Goto(int Position, bool Still)
{
  if (replayBuffer)
     replayBuffer->Goto(Position, Still);
}

// --- cEITScanner -----------------------------------------------------------

cEITScanner::cEITScanner(void)
{
  lastScan = lastActivity = time(NULL);
  currentChannel = 0;
  lastChannel = 0;
  numTransponders = 0;
  transponders = NULL;
}

cEITScanner::~cEITScanner()
{
  delete transponders;
}

bool cEITScanner::TransponderScanned(cChannel *Channel)
{
  for (int i = 0; i < numTransponders; i++) {
      if (transponders[i] == Channel->frequency)
         return true;
      }
  transponders = (int *)realloc(transponders, ++numTransponders * sizeof(int));
  transponders[numTransponders - 1] = Channel->frequency;
  return false;
}

void cEITScanner::Activity(void)
{
  if (currentChannel) {
     Channels.SwitchTo(currentChannel);
     currentChannel = 0;
     }
  lastActivity = time(NULL);
}

void cEITScanner::Process(void)
{
  
  if (Setup.EPGScanTimeout && Channels.MaxNumber() > 1) {
     time_t now = time(NULL);

     if (now - lastScan > ScanTimeout && now - lastActivity > ActivityTimeout) {
        for (int i = 0; i < cDvbApi::NumDvbApis; i++) {
            cDvbApi *DvbApi = cDvbApi::GetDvbApi(i + 1, 0);
            if (DvbApi) {
               if (DvbApi != cDvbApi::PrimaryDvbApi || (cDvbApi::NumDvbApis == 1 && Setup.EPGScanTimeout && now - lastActivity > Setup.EPGScanTimeout * 3600)) {
                  if (!(DvbApi->Recording() || DvbApi->Replaying() || DvbApi->Transferring())) {
                     int oldCh = lastChannel;
                     int ch = oldCh + 1;

                     while (ch != oldCh) {
                           if (ch > Channels.MaxNumber()) {
                              ch = 1;
                              numTransponders = 0;
                              }
                           cChannel *Channel = Channels.GetByNumber(ch);
                           if (Channel && Channel->pnr && !TransponderScanned(Channel)) {
                              if (DvbApi == cDvbApi::PrimaryDvbApi && !currentChannel)
                                 currentChannel = DvbApi->Channel();
                              Channel->Switch(DvbApi, false);
                              lastChannel = ch;
                              break;
                              }
                           ch++;
                           }
                     }
                  }
               }
            }
        lastScan = time(NULL);
        }
     }
}

}

#include "nmm/nmm_suffix.hpp"
