/*
 * ringbuffer.cpp: A threaded ring buffer
 *
 * Copyright (C) 2000 Klaus Schmidinger
 *
 * Modified by Motama GmbH,
 * Copyright of modifications (C) 2006-2007 Motama GmbH
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 *
 * The author can be reached at kls@cadsoft.de
 *
 * The project's page is at http://www.cadsoft.de/people/kls/vdr
 *
 */


/*
 * $Id: ringbuffer.cpp,v 1.3.8.2 2005/12/22 01:40:50 fw Exp $
 */

/*
 * ringbuffer.c: A threaded ring buffer
 *
 * See the main source file 'ringbuffer.h' for original copyright information and
 * how to reach the original author.
 *
 * Parts of this file were inspired by the 'ringbuffy.c' from the
 * LinuxDVB driver (see linuxtv.org).
 *
 * $Id: ringbuffer.cpp,v 1.3.8.2 2005/12/22 01:40:50 fw Exp $
 */

#include "ringbuffer.h"
#include "tools.h"
#include <iostream>
namespace VDR {

// --- cRingBufferInputThread -------------------------------------------------

class cRingBufferInputThread : public cThread {
private:
  cRingBuffer *ringBuffer;
protected:
  virtual void Action(void) { ringBuffer->Input(); }
public:
  cRingBufferInputThread(cRingBuffer *RingBuffer) { ringBuffer = RingBuffer; }
 
 ~cRingBufferInputThread() { pthread_join( thread , NULL); }
}; 
// --- cRingBufferOutputThread ------------------------------------------------

class cRingBufferOutputThread : public cThread {
private:
  cRingBuffer *ringBuffer;
protected:
  virtual void Action(void) { ringBuffer->Output(); }
public:
  cRingBufferOutputThread(cRingBuffer *RingBuffer) { ringBuffer = RingBuffer; }
 
  ~cRingBufferOutputThread() { pthread_join( thread, NULL); }
};
// --- cRingBuffer ------------------------------------------------------------

cRingBuffer::cRingBuffer(int Size, bool Statistics)
{
  size = Size;
  statistics = Statistics;
  buffer = NULL;
  inputThread = NULL;
  outputThread = NULL;
  maxFill = 0;
  busy = false;
  if (size > 1) { // 'size - 1' must not be 0!
     buffer = new uchar[size];
     if (!buffer)
        esyslog(LOG_ERR, "ERROR: can't allocate ring buffer (size=%d)", size);
     Clear();
     }
  else
     esyslog(LOG_ERR, "ERROR: illegal size for ring buffer (%d)", size);
}

cRingBuffer::~cRingBuffer()
{
  delete inputThread;
  delete outputThread;
  delete buffer;
  if (statistics)
     dsyslog(LOG_INFO, "buffer stats: %d (%d%%) used", maxFill, maxFill * 100 / (size - 1));
}

int cRingBuffer::Available(void)
{
  mutex.Lock();
  int diff = head - tail;
  int cont = (diff >= 0) ? diff : size + diff;
  mutex.Unlock();
  return cont;
}

void cRingBuffer::Clear(void)
{
  mutex.Lock();
  head = tail = 0;
  mutex.Unlock();
}

int cRingBuffer::Put(const uchar *Data, int Count)
{
  if (Count > 0) {
     mutex.Lock();
     int rest = size - head;
     int diff = tail - head;
     mutex.Unlock();
     int free = (diff > 0) ? diff - 1 : size + diff - 1;
     if (statistics) {
        int fill = size - free - 1 + Count;
        if (fill >= size)
           fill = size - 1;
        if (fill > maxFill) {
           maxFill = fill;
           int percent = maxFill * 100 / (size - 1);
           if (percent > 75)
              dsyslog(LOG_INFO, "buffer usage: %d%%", percent);
           }
        }
     if (free <= 0)
        return 0;
     if (free < Count)
        Count = free;
     if (Count > maxFill)
        maxFill = Count;
     if (Count >= rest) {
        memcpy(buffer + head, Data, rest);
        if (Count - rest)
           memcpy(buffer, Data + rest, Count - rest);
        head = Count - rest;
        }
     else {
        memcpy(buffer + head, Data, Count);
        head += Count;
        }
     }
  return Count;
}

int cRingBuffer::Get(uchar *Data, int Count)
{
  if (Count > 0) {
     mutex.Lock();
     int rest = size - tail;
     int diff = head - tail;
     mutex.Unlock();
     int cont = (diff >= 0) ? diff : size + diff;
     if (rest <= 0)
        return 0;
     if (cont < Count)
        Count = cont;
     if (Count >= rest) {
        memcpy(Data, buffer + tail, rest);
        if (Count - rest)
           memcpy(Data + rest, buffer, Count - rest);
        tail = Count - rest;
        }
     else {
        memcpy(Data, buffer + tail, Count);
        tail += Count;
        }
     }
  return Count;
}

bool cRingBuffer::Start(void)
{
  if (!busy) {
     busy = true;
     outputThread = new cRingBufferOutputThread(this);
     if (!outputThread->Start())
        DELETENULL(outputThread);
     inputThread = new cRingBufferInputThread(this);
     if (!inputThread->Start()) {
        DELETENULL(inputThread);
        DELETENULL(outputThread);
        }
     busy = outputThread && inputThread;
     }
  return busy;
}

bool cRingBuffer::Active(void)
{
  return outputThread && outputThread->Active() && inputThread && inputThread->Active();
}

void cRingBuffer::Stop(void)
{
  busy = false;
  /* repplix: As far as i understand this code, this is an evil hack to wait until a thread terminates!!!!!*/
 /* for (time_t t0 = time(NULL) + 3; time(NULL) < t0; ) {
      if (!((outputThread && outputThread->Active()) || (inputThread && inputThread->Active())))
         break;
      }*/
  /*repplix: If a thread is deleted, it destroyes its internal thread by invoking pthread_join*/
  DELETENULL(inputThread);
  DELETENULL(outputThread);
}

}

#include "nmm/nmm_suffix.hpp"
