/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include "TVChannelList.hpp"

#include "nmm/Exception.hpp"

#include "nmm/utils/xml/XMLTree.hpp"
#include "nmm/utils/xml/XMLNode.hpp"

#include "nmm/comm/serialize/StdValue.hpp"
#include "nmm/comm/serialize/TConstructor.hpp"
#include "nmm/comm/serialize/TValueConstructor.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM{

  /* ****************************************
   *           TVChannelList                *
   **************************************** */

  TVChannelList::TVChannelList()
    :SerializedValue("TVChannelList", "NMM"),
     channelNumber(0)
  {
    registerAttributes();
  }

  TVChannelList::TVChannelList(string filename)
    :SerializedValue("TVChannelList", "NMM"),
     channelNumber(0)
  {
    load(filename);
    registerAttributes();
  }

  TVChannelList::TVChannelList(const TVChannelList& t)
    :SerializedValue("TVChannelList", "NMM"),
     channelNumber(0)
  {
    registerAttributes();
    *this=t;
  }

  TVChannelList::~TVChannelList(){}

  void TVChannelList::registerAttributes(){
    registerAttribute(&channels);
    registerAttribute(&channelNumber);
  }

  void TVChannelList::addChannel(TVChannel cd){
    channels.getRValue().push_back(cd);
  }
  
  TVChannel& TVChannelList::getChannelByNumber(unsigned int number){

    if(number >= channels.getRValue().size())     throw ChannelNotFoundException(number);

    list<TVChannel>::iterator i = channels.getRValue().begin();
    advance(i,number);
    return *i;
    
  }

  TVChannel& TVChannelList::getChannelByName(string name){


    for(list<TVChannel>::iterator i = channels.getRValue().begin(); i != channels.getRValue().end(); i++){
      string s1(i->getChannelName()); transform(s1);
      string s2(name); transform(s2);
      if(s1==s2) return *i;
      list<string> aliases = i->getChannelAliases();
      for(list<string>::iterator j = aliases.begin(); j != aliases.end(); j++){
        string s1(*j); transform(s1);
        string s2(name); transform(s2);
        if(s1==s2) return *i;
      }

      /*
        
      Commented out because to slow in TVGuide-State

      //If everything failes, try to find the name in Channel-Alias-file      
      string aliasFilename = getNMMResourceDir() + "/tvguide/channelAliases.xml";
      //      XMLTree *tree = new XMLTree(aliasFilename);
      XMLTree tree(aliasFilename);
      //BEGIN> read Aliases
      //if(tree){
      //XMLNode *root = tree->root();
        XMLNode *root = tree.root();
        if(root){
          XMLNodeList nodeList = root->children("channel");
          for(XMLNodeIterator k = nodeList.begin(); k != nodeList.end(); k++){
            list<string> aliases;
            XMLNodeList childList = (*k)->children("alias");
            for(XMLNodeIterator j = childList.begin(); j != childList.end(); j++){
              string tmpName((*j)->getData()); transform(tmpName);
              cerr << "TVChannellist checking: " << tmpName << " " << name << endl;
              if(tmpName == name){
                //cerr << "TVChannellist matchFound" << endl;
                i->addAlias((*j)->getData());
                //delete tree; tree = NULL;
                return *i;
              }
            }// end for(childList)
            aliases.clear();
          }//end for(nodeList)
        }//end if(root)
        //delete tree;
        //tree = NULL;
        // }//end if(tree)          
        */
      
    }
    throw ChannelNotFoundException(name);
  }

  list<TVChannel> TVChannelList::getChannels(){
    return channels.getRValue();
  }


  list<string> TVChannelList::getTVCardURLs(){
    list<string> returnlist;
    for(list<TVChannel>::iterator i = channels.getRValue().begin();
        i != channels.getRValue().end(); i++){
      TVChannel::cardList_t cards = (i->getCards());
      for(TVChannel::cardList_t::iterator j = cards.begin(); j != cards.end(); j++){
        returnlist.push_back(j->first);
      }
    }
    returnlist.sort();
    returnlist.unique();
    return returnlist;
  }

  Result TVChannelList::load(string filename){
    /* open XML document */
    XMLTree* xml = new XMLTree(filename);
    
    /* get root node */
    XMLNode* root = xml->root();
    if ( root == NULL ) {
      GLOBAL_WARNING_STREAM("Could not parse " << filename << endl);
      delete xml;
      return FAILURE;
    }


    XMLNodeList channels = root->children("tvchannel");

    for(XMLNodeIterator i = channels.begin(); i != channels.end(); i++){
    
      XMLProperty* prop = (*i)->property("name");
      string channelName;
      if ( prop != NULL )  channelName = prop->value();
      if(channelName.empty()) continue;
      TVChannel channel(channelName);
      
      XMLNodeList cards = (*i)->children("card");
      for(XMLNodeIterator j = cards.begin(); j != cards.end(); j++){
        
        unsigned int cnumber=0;
        prop = (*j)->property("number");
        if(prop) cnumber=atoi((prop->value()).c_str());

        string location = (*j)->getData();

        channel.addCard(location, cnumber);
      }

      XMLNodeList aliases = (*i)->children("alias");
      for(XMLNodeIterator j = aliases.begin(); j != aliases.end(); j++){
        channel.addAlias((*j)->getData());
      }
     
      channel.setChannelNumber(channelNumber.getRValue());
      channelNumber.getRValue()++;

      addChannel(channel);
    } 
    return SUCCESS;

  }

  Result TVChannelList::write(string filename){
    try{
      
      XMLTree *tree = new XMLTree();
      XMLNode *root = new XMLNode("tvchannellist");
      tree->set_root(root);

      for(list<TVChannel>::iterator i = channels.getRValue().begin(); i != channels.getRValue().end(); i++){
        XMLNode *n = new XMLNode(i->toXMLNode());
        root->add_child(n);
      }

      cerr << "written" << endl;
      tree->write(filename);
      delete tree; tree = NULL;
      
    }catch(const Exception &e){
      GLOBAL_ERROR_STREAM("TVChannelList: Error writing channelList: e" << endl);
      return FAILURE;
    }

    return SUCCESS;

  }
  OValueStream& operator<<(OValueStream& os, const list<TVChannelList>& nmm_tvchannellist){
    return writeList<TVChannelList>(os, nmm_tvchannellist, "TVChannelList");
  }
  IValueStream& operator>>(IValueStream& is, list<TVChannelList>& nmm_tvchannellist){
    return readList<TVChannelList>(is, nmm_tvchannellist, "TVChannelList");
  }

  TConstructor<TVChannelList> nmm_tvchannellist("TVChannelList", "NMM");
  TValueConstructor<list<TVChannelList> > nmm_tvchannellist_list("list_TVChannelList","NMM");
}

#include "nmm/nmm_suffix.hpp"
