/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "ProgramDescription.hpp"
#include "nmm/utils/time/timeutils.hpp"
#include "nmm/comm/serialize/StdValue.hpp"
#include "nmm/comm/serialize/SerializeLocalTime.hpp"
#include "nmm/comm/serialize/TConstructor.hpp"
#include "nmm/comm/serialize/TValueConstructor.hpp"
#include <stdio.h>

#include "nmm/nmm_prefix.hpp"

namespace NMM{

  ProgramDescription::ProgramDescription()
    : NMM::SerializedValue("ProgramDescription","NMM"),
           sourcePriority(0),
           startTimeEmpty(true),
           endTimeEmpty(true)
  {
    registerValues();
  }

  ProgramDescription::ProgramDescription(unsigned int sourcePrio)
    : NMM::SerializedValue("ProgramDescription","NMM"),
           sourcePriority(sourcePrio),
           startTimeEmpty(true),
           endTimeEmpty(true)
  {
    registerValues();
  }

  ProgramDescription::ProgramDescription(const ProgramDescription& pd)
    : NMM::SerializedValue("ProgramDescription","NMM"),
      sourcePriority(0)

  {
    registerValues();
    *this=pd;
  }

  ProgramDescription::~ProgramDescription(){}
  
  bool ProgramDescription::empty(){
    if(startTimeEmpty.getRValue() && endTimeEmpty.getRValue() &&
       channel.getRValue().empty() && title.getRValue().empty()){
      return true;
    }else{
      return false;
    }
  }

  void ProgramDescription::setStartTime(const LocalTime &st){
    startTimeEmpty.setValue(false);
    startTime.setValue(st);
    startTime.getRValue().toGMT();
  }
  void ProgramDescription::setStartTime(string st){
    startTimeEmpty.setValue(false);
    startTime.setValue(stringToLocalTime(st));
  }
  LocalTime ProgramDescription::getStartTime(timeLocale_t l) const {
    if(l == GMT){
      return startTime.getRValue();
    }

    LocalTime ret(startTime.getRValue());
    ret.toLocalTimeZone();
    return ret;
  }
  
  void ProgramDescription::setEndTime(const LocalTime &et) {
    endTimeEmpty.setValue(false);
    endTime.setValue(et);
    endTime.getRValue().toGMT();
  }

  void ProgramDescription::setEndTime(string et) {
    endTimeEmpty.setValue(false);
    endTime.setValue(stringToLocalTime(et));
  }
  LocalTime ProgramDescription::getEndTime(timeLocale_t l) const {
    if(l == GMT){
      return endTime.getRValue();
    }
    LocalTime ret(endTime.getRValue());
    ret.toLocalTimeZone();
    return ret;
  }

  void ProgramDescription::setChannelID(string id){
    channel.getRValue().setID(id);
  }
  string ProgramDescription::getChannelID() const{
    return channel.getRValue().getID();
  }
  
  void ProgramDescription::setChannelDescription(ChannelDescription cd){
    channel = cd;
  }
  ChannelDescription ProgramDescription::getChannelDescription() const{
    return channel.getRValue();
  }
  
  void ProgramDescription::setDuration(unsigned int d){
    duration = d;
  }
  unsigned int ProgramDescription::getDuration() const {
    return duration.getRValue();
  }
  
  void ProgramDescription::setTitles(list<string> titles){
    title = titles;
  }
  void ProgramDescription::addTitle(string t){
    (title.getRValue()).push_back(t);
  }
  list<string> ProgramDescription::getTitles() const {
    return title.getRValue();
  }

  void ProgramDescription::setDescription(string d){
    desc = d;
  }
  string ProgramDescription::getDescription() const {
    return desc.getRValue();
  }
  
  void ProgramDescription::addCategory(string c){
    category.getRValue().push_back(c);
  }
  list<string> ProgramDescription::getCategory() const {
    return category.getRValue();
  }

  ProgramDescription* ProgramDescription::copy(){
    return new ProgramDescription(*this);
  }

  void ProgramDescription::registerValues(){
    insertArgument(&startTime);
    insertArgument(&endTime);
    insertArgument(&channel);
    insertArgument(&duration);
    insertArgument(&title);
    insertArgument(&desc);
    insertArgument(&sourcePriority);
    insertArgument(&category);
    insertArgument(&startTimeEmpty);
    insertArgument(&endTimeEmpty);
  }

  bool ProgramDescription::operator<(const ProgramDescription &pd) const{

    if(startTime.getRValue() < pd.getStartTime()){
      return true;
    }else{
      if(startTime.getRValue() == pd.getStartTime() && channel.getRValue().getID() < pd.getChannelID()){
        return true;
      }
    }
        
    return false;
  }

  bool ProgramDescription::operator==(const ProgramDescription &pd) const{
    if(startTime.getRValue() == pd.getStartTime() && channel.getRValue().getID() == pd.getChannelID()){
      return true;
    }else{
      return false;

    }
    return false;
  }
 

  LocalTime ProgramDescription::stringToLocalTime(string s){
    
    if(s.empty()) LocalTime t;

    unsigned int pos = s.find_first_of(' ');
    int xmlDifference = 0;
    if(pos != string::npos){
      //if no timezone-info is available, assume UTC
      xmlDifference = atoi(s.substr(pos+1,5).c_str())/100;
    }
    
    time_t xmlTime = stringToTime(s);
    struct tm *t = new tm();
    localtime_r(&xmlTime, t);
    //Compute the difference between the xml-Timezone and our timezone and update the outputtime accordingly

    xmlTime +=   (-(xmlDifference*3600)) + (t->tm_gmtoff);
    
    gmtime_r(&xmlTime, t);
    LocalTime lt(*t);
    delete t; t = NULL;
    return lt; 
  }

  OValueStream& operator<<(OValueStream& os, const list<ProgramDescription>& pd){
    return writeList<ProgramDescription>(os, pd, "ProgramDescription");
  }
  IValueStream& operator>>(IValueStream& is, list<ProgramDescription>& pd){
    return readList<ProgramDescription>(is, pd, "ProgramDescription");
  }
  
  TConstructor<ProgramDescription> nmm_pd("ProgramDescription", "NMM");
  TValueConstructor<list<ProgramDescription> > nmm_pd_list("list_ProgramDescription","NMM");


  OValueStream& operator<<(OValueStream& os, const list<ProgramDescriptionList>& pl){
    return writeList<ProgramDescriptionList>(os, pl, "ProgramDescriptionList");
  }
  IValueStream& operator>>(IValueStream& is, list<ProgramDescriptionList>& pl){
    return readList<ProgramDescriptionList>(is, pl, "ProgramDescriptionList");
  }
  
  TConstructor<ProgramDescriptionList> nmm_pl("ProgramDescriptionList", "NMM");
  TValueConstructor<list<ProgramDescriptionList> > nmm_pl_list("list_ProgramDescriptionList","NMM");
}

#include "nmm/nmm_suffix.hpp"
