/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "ProgramFilter.hpp"
#include "nmm/comm/serialize/StdValue.hpp"
#include "nmm/comm/serialize/TValueConstructor.hpp"
#include "nmm/comm/serialize/SerializeLocalTime.hpp"
#include "nmm/comm/serialize/TConstructor.hpp"
#include "nmm/utils/time/timeutils.hpp"
#include <stdio.h>

#include "nmm/nmm_prefix.hpp"

namespace NMM{

  ProgramFilter::ProgramFilter()
    : NMM::SerializedValue("ProgramFilter","NMM"),
           filtername(""),
           phrase(""),
           starttimeEdited(false),
           getCurrentEvents(false),
           getFollowingEvents(false),
           startTimeEmpty(true),
           endTimeEmpty(true),
           id(0),
           hops(0)
  {
    registerValues();
  }

  ProgramFilter::ProgramFilter(const ProgramFilter& pd)
    : NMM::SerializedValue("ProgramFilter","NMM")
  {
    registerValues();
    *this=pd;
  }

  ProgramFilter::~ProgramFilter(){}
  
  void ProgramFilter::setStartTime(const LocalTime &t){
    startTimeEmpty.setValue(false);
    starttime.setValue(t);
    starttime.getRValue().toGMT();
  }
  void ProgramFilter::setStartTime(string st){
    startTimeEmpty.setValue(false);
    starttime.setValue(stringToLocalTime(st));
  }

  LocalTime ProgramFilter::getStartTime() const {
    return starttime.getRValue();
  }
  
  
  void ProgramFilter::setEndTime(const LocalTime &t) {
    endTimeEmpty.setValue(false);
    endtime.setValue(t);
    endtime.getRValue().toGMT();
  }
  void ProgramFilter::setEndTime(string et) {
    endTimeEmpty.setValue(false);
    endtime.setValue(stringToLocalTime(et));
  }
  LocalTime ProgramFilter::getEndTime() const {
    return endtime.getRValue();
  }

  void ProgramFilter::setSearchPhrase(string p){
    phrase = p;
  }
  string ProgramFilter::getSearchPhrase() const{
    return phrase.getRValue();
  }
    

  void ProgramFilter::addCategory(string c){
    (categories.getRValue()).push_back(c);
  }
  list<string> ProgramFilter::getCategories() const {
    return categories.getRValue();
  }
    
  void ProgramFilter::addChannel(ChannelDescription c){
    (channels.getRValue()).push_back(c);
    (channels.getRValue()).sort();
    (channels.getRValue()).unique();    
  }
  list<ChannelDescription> ProgramFilter::getChannels() const {
    return channels.getRValue();
  }

  list<ChannelDescription> ProgramFilter::getChannels() {
    return channels.getRValue();
  }

  ProgramFilter* ProgramFilter::copy(){
    return new ProgramFilter(*this);
  }

  void ProgramFilter::registerValues(){
    insertArgument(&filtername);
    insertArgument(&starttime);
    insertArgument(&endtime);
    insertArgument(&phrase);
    insertArgument(&categories);
    insertArgument(&channels);
    insertArgument(&starttimeEdited);
    insertArgument(&visitedPeerHosts);
    insertArgument(&id);
    insertArgument(&hops);
    insertArgument(&getCurrentEvents);
    insertArgument(&getFollowingEvents);
    insertArgument(&startTimeEmpty);
    insertArgument(&endTimeEmpty);
  }

  void ProgramFilter::addPeerHost(string name){
    visitedPeerHosts.getRValue().push_back(name);
    visitedPeerHosts.getRValue().sort();
    visitedPeerHosts.getRValue().unique();
  }
  
  bool ProgramFilter::hasPeerHost(string name){
    for(list<string>::iterator i = visitedPeerHosts.getRValue().begin(); i != visitedPeerHosts.getRValue().end(); i++){
      if (*i == name) return true;
    }
    return false;
  }
 
  void ProgramFilter::clearPeerHosts(){
    visitedPeerHosts.getRValue().clear();
  }

  bool ProgramFilter::operator==(const ProgramFilter &pf) const{
    if(starttime.getRValue() == pf.getStartTime()
       && endtime.getRValue() == pf.getEndTime() 
       && phrase.getRValue() == pf.getSearchPhrase()
       && categories.getRValue() == pf.getCategories()
       && channels.getRValue() == pf.getChannels()){
      return true;
    }else{
      return false;
    }
    
  }

  LocalTime ProgramFilter::stringToLocalTime(string s){
    
    if(s.empty()) LocalTime t;

    unsigned int pos = s.find_first_of(' ');
    int xmlDifference = 0;
    if(pos != string::npos){
      //if no timezone-info is available, assume UTC
      xmlDifference = atoi(s.substr(pos+1,5).c_str())/100;
    }
    
    time_t xmlTime = stringToTime(s);
    struct tm *t = new tm();
    localtime_r(&xmlTime, t);
    //Compute the difference between the xml-Timezone and our timezone and update the outputtime accordingly

    xmlTime +=   (-(xmlDifference*3600)) + (t->tm_gmtoff);
    
    gmtime_r(&xmlTime, t);
    LocalTime lt(*t);
    delete t; t = NULL;
    return lt; 
  }


  OValueStream& operator<<(OValueStream& os, const list<ProgramFilter>& pf){
    return writeList<ProgramFilter>(os, pf, "ProgramFilter");
  }
  IValueStream& operator>>(IValueStream& is, list<ProgramFilter>& pf){
    return readList<ProgramFilter>(is, pf, "ProgramFilter");
  }
  
  TConstructor<ProgramFilter> nmm_pf("ProgramFilter", "NMM");
  TValueConstructor<list<ProgramFilter> > nmm_pf_list("list_ProgramFilter","NMM");

}

#include "nmm/nmm_suffix.hpp"
