/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#ifndef __NMM_IVTV_STRUCTS_HPP
#define __NMM_IVTV_STRUCTS_HPP

#include "nmm/nmm_prefix.hpp"

#include "nmm/comm/serialize/TValue.hpp"

namespace NMM {  
  
  /** IVTV Codec properties.
   * This structure contains all the codec parameters for an IVTV based TV card.
   */
  class IVTVCodec
    : public SerializedValue
  {
  public:
    /** @name Constructor & Destructor
     */
    //@{
    /** Default constructor.
     */
    IVTVCodec();
    
    /** Constructor.
     * @param audio Audio bitmask
     * @param bframes Number of B-Frames
     * @param bitrate_mode Bitrate mode
     * @param bitrate Bitrate
     * @param bitrate_peak Bitrate peak
     * @param dnr_mode
     * @param dnr_spatial
     * @param dnr_temporal
     * @param dnr_type
     * @param framerate Framerate type
     * @param frames_per_gop Number of frames per gop
     * @param gop_closure GOP closure mode
     * @param pulldown
     * @param stream_type Stream type
     */
    IVTVCodec(unsigned int audio, unsigned int bframes,
              unsigned int bitrate_mode, unsigned int bitrate, unsigned int bitrate_peak,
              unsigned int dnr_mode, unsigned int dnr_spatial, unsigned int dnr_temporal, unsigned int dnr_type,
              unsigned int framerate, unsigned int frames_per_gop, unsigned int gop_closure,
              unsigned int pulldown, unsigned int stream_type);
    
    /** virtual copy constructor
     */
    virtual IVTVCodec* copy()                     const { return new IVTVCodec(*this); }
    //@}
    
    /** @name Access Methods
     */
    //@{
    /** This method returns the audio bitmask of the codec.
     */
    inline unsigned int  getAudioBitmask()        const { return __audio.getRValue(); }
    
    /** This method returns the number of B-frames of the codec.
     */
    inline unsigned int  getBFrames()             const { return __bframes.getRValue(); }
    
    /** This method returns the bitrate mode of the codec.
     */
    inline unsigned int  getBitrateMode()         const { return __bitrate_mode.getRValue(); }
    
    /** This method returns the bitrate of the codec.
     */
    inline unsigned int  getBitrate()             const { return __bitrate.getRValue(); }
    
    /** This method returns the peak bitrate of the codec.
     */
    inline unsigned int  getBitratePeak()         const { return __bitrate_peak.getRValue(); }
    
    /** This method returns the DNR mode of the codec.
     */
    inline unsigned int  getDNRMode()             const { return __dnr_mode.getRValue(); }
    
    /**
     */
    inline unsigned int  getDNRSpatial()          const { return __dnr_spatial.getRValue(); }
    
    /**
     */
    inline unsigned int  getDNRTemporal()         const { return __dnr_temporal.getRValue(); }
    
    /**
     */
    inline unsigned int  getDNRType()             const { return __dnr_type.getRValue(); }
    
    /** This method returns the framerate ID of the codec
     */
    inline unsigned int  getFramerate()           const { return __framerate.getRValue(); }
    
    /** This method returns number of frames per GOP of the codec
     */
    inline unsigned int  getFramesPerGOP()        const { return __frames_per_gop.getRValue(); }
    
    /** This method returns the GOP closure mode of the codec
     */
    inline unsigned int  getGOPClosure()          const { return __gop_closure.getRValue(); }
    
    /**
     */
    inline unsigned int  getPulldown()            const { return __pulldown.getRValue(); }
    
    /** This method returns the stream type of the codec
     */
    inline unsigned int  getStreamType()          const { return __stream_type.getRValue(); }

    //////////////////////

    /** This method sets the audio bitmask of the codec.
     */
    inline void  setAudioBitmask(unsigned int value)        {__audio.setValue(value); }
    
    /** This method sets the number of B-frames of the codec.
     */
    inline void  setBFrames(unsigned int value)             {__bframes.setValue(value); }
    
    /** This method sets the bitrate mode of the codec.
     */
    inline void  setBitrateMode(unsigned int value)         {__bitrate_mode.setValue(value); }
    
    /** This method sets the bitrate of the codec.
     */
    inline void  setBitrate(unsigned int value)             {__bitrate.setValue(value); }
    
    /** This method sets the peak bitrate of the codec.
     */
    inline void  setBitratePeak(unsigned int value)         {__bitrate_peak.setValue(value); }
    
    /** This method sets the DNR mode of the codec.
     */
    inline void  setDNRMode(unsigned int value)             {__dnr_mode.setValue(value); }
    
    /**
     */
    inline void  setDNRSpatial(unsigned int value)          {__dnr_spatial.setValue(value); }
    
    /**
     */
    inline void  setDNRTemporal(unsigned int value)         {__dnr_temporal.setValue(value); }
    
    /**
     */
    inline void  setDNRType(unsigned int value)             {__dnr_type.setValue(value); }
    
    /** This method sets the framerate ID of the codec
     */
    inline void  setFramerate(unsigned int value)           {__framerate.setValue(value); }
    
    /** This method sets number of frames per GOP of the codec
     */
    inline void  setFramesPerGOP(unsigned int value)        {__frames_per_gop.setValue(value); }
    
    /** This method sets the GOP closure mode of the codec
     */
    inline void  setGOPClosure(unsigned int value)          {__gop_closure.setValue(value); }
    
    /**
     */
    inline void  setPulldown(unsigned int value)            {__pulldown.setValue(value); }
    
    /** This method sets the stream type of the codec
     */
    inline void  setStreamType(unsigned int value)          {__stream_type.setValue(value); }
    //@}
    
  private:
    /** This method adds all attributes by the base class.
     */
    void registerAttributes();
    
  private:
    TValue<nmm_uint32> __audio;                 ///< Audio bitmask
    TValue<nmm_uint32> __bframes;               ///< B-Frames
    TValue<nmm_uint32> __bitrate_mode;          ///< Bitrate mode
    TValue<nmm_uint32> __bitrate;               ///< Bitrate
    TValue<nmm_uint32> __bitrate_peak;          ///< Bitrate peak
    TValue<nmm_uint32> __dnr_mode;              ///< ???
    TValue<nmm_uint32> __dnr_spatial;           ///< ???
    TValue<nmm_uint32> __dnr_temporal;          ///< ???
    TValue<nmm_uint32> __dnr_type;              ///< ???
    TValue<nmm_uint32> __framerate;             ///< Framerate id
    TValue<nmm_uint32> __frames_per_gop;        ///< Number of frames per GOP
    TValue<nmm_uint32> __gop_closure;           ///< Open/Closed GOP
    TValue<nmm_uint32> __pulldown;
    TValue<nmm_uint32> __stream_type;
  };
  
} // namespace NMM

#include "nmm/nmm_suffix.hpp"

#endif // __NMM_IVTV_STRUCTS_HPP
