/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#ifndef _NMM_FREQUENCY_TABLE_HPP_
#define _NMM_FREQUENCY_TABLE_HPP_

#include "nmm/nmm_prefix.hpp"

#include <vector>
#include <map>

#include "nmm/Exception.hpp"
#include "nmm/utils/string_utils.hpp"

using namespace std;

namespace NMM {
  
  /** Frequency table.
   * A frequency table maps channels to frequencies. Channels are uniquely identified by their index
   * in the frequency table (a zero based integer), the channel name (a string) or the frequency
   * (an integer which is the frequency in units of 62.5 kHz).
   */
  class FrequencyTable {
  public:
    
    /** Channel entry.
     * An entry in the frequency table, contains the channel name and frequency
     */
    struct Channel {
      Channel()
        : channel_frequency(0)
      {}
      
      Channel(const string& name, unsigned int frequency)
        : channel_name(name),
          channel_frequency(frequency)
      {}
      
      string       channel_name;        ///< Channel name
      unsigned int channel_frequency;   ///< Channel frequency
    };
    
  public:
    /** @name Constructors
     */
    //@{
    /** Default constructor.
     */
    FrequencyTable();
    
    /** Constructor.
     * @param name Name of the frequency table
     */
    FrequencyTable(const string& name);
    //@}
    
    /** @name Access Methods
     */
    //@{
    /** This method returns the name of the frequency table.
     */
    inline const string& getName        () const { return __name; }
    
    /** This method returns the number of channels in the frequency table.
     */
    inline unsigned int  getChannelCount() const { return __channels.size(); }
    
    /** This method sets the name of the frequency table.
     */
    inline void setName(const string& name) {
      __name = name;
    }
    
    /** This method returns the frequency of the given channel.
     * @param channel Channel number, ranges from 0 to getChannelCount()-1
     * @exception Exception, if the channel number was invalid
     */
    unsigned int getChannelFrequency(unsigned int channel) const
      throw (const Exception&);
    
    /** This method returns the name of the given channel.
     * @param channel Channel number, ranges from 0 to getChannelCount()-1
     * @exception Exception, if the channel number was invalid
     */
    const string& getChannelName(unsigned int channel) const
      throw (const Exception&);
    
    /** This method returns the channel number of the channel given by its name.
     * @param The channel name
     * @return The channel number, if the channel exists, otherwise -1
     */
    int findChannel(const string& name) const;
    
    /** This method adds a new channel to the frequency table.
     * @param name Name of the channel, must be unique
     * @param frequency Frequency of the channel
     * @exception Exception, if the channel name was not unique
     */
    void addChannel(const string& name, unsigned int frequency)
      throw (const Exception&);
    //@}
    
  private:
    string          __name;             ///< Name of the frequency table
    vector<Channel> __channels;         ///< Channel list
    
    map<string, unsigned int> __map;    ///< Channel name map
  };
  
} // namespace NMM

#include "nmm/nmm_suffix.hpp"

#endif // _NMM_FREQUENCY_TABLE_HPP_
