/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include "IVTVReadNode.hpp"

#include "nmm/utils/directory_utils.hpp"
#include "nmm/utils/SecurityCheck.hpp"
#include "nmm/multimedia/TDevicePlugin.hpp"

#include "nmm/multimedia/ISyncReset.hpp"
#include "nmm/multimedia/FlushEvent.hpp"

#include "nmm/utils/xml/XMLTree.hpp"
#include "nmm/utils/xml/XMLNode.hpp"

#include "nmm/utils/time/TimedElement.hpp"
#include "nmm/utils/time/Rational.hpp"

#include <fstream>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#include <iomanip>
#include <regex.h>

#include "videodev2.h"
#include "ivtv-ext-api.h"

#define IVTV_IOC_FRAMESYNC      0xFFEE7783
#define IVTV_IOC_GET_TIMING     0xFFEE7784
#define MPEG_CLOCK_FREQ 90000 

static NMM::TDevicePlugin<NMM::IVTVReadNode> plugin("/dev/video", false, "IVTVReadNode", 1);
NMM_REGISTER_SPECIAL_PLUGIN(IVTVReadNode, NMM,plugin)

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  /*============================== Constructor & Destructor ==============================*/
  IVTVReadNode::IVTVReadNode(const char *name /*= "IVTVReadNode"*/,
                             StreamQueue::Mode q_mode /*= StreamQueue::MODE_SUSPEND*/,
                             int q_size /*= StreamQueue::DEFAULT_SIZE*/)
    : GenericSourceNode(name, q_mode, q_size),
      __device("/dev/video0"),
      __fd(-1),
      __buffer_size(512*1024),
      __do_flush(false),
      __stream_timer(this),
      __cur_channel(0),
      __system_start_time(),
      __scr_start(0),
      __buffer_counter(0),
      __received_first_buffer(false)
  {
    channelConfigFile = getNMMResourceDir() + "/ivtv/ivtv.conf";
  }
  
  IVTVReadNode::~IVTVReadNode() {
    /* close device */
    if(__fd >= 0) {
      DEBUG_STREAM("Closing ivtv device" << std::endl);
      ::close(__fd);
      __fd = -1;
    }    
  }
  /*============================== Constructor & Destructor ==============================*/
  
  
  /*============================== IBufferSize Interface ==============================*/
  void IVTVReadNode::setBufferSize(const nmm_int32& value) {
    __buffer_size = value;
  }
  
  nmm_int32 IVTVReadNode::getBufferSize() {
    return __buffer_size;
  }
  /*============================== IBufferSize Interface ==============================*/
  
  
  /*============================== ITVCard Interface ==============================*/
  void IVTVReadNode::setDevice(const string& device) throw(const DeviceNotFoundException&, const WrongDeviceException&) {
    /* open device */
    try {
      if(!SSecurityCheck::getInstance().checkFilenameForReading(device)) {
        ERROR_STREAM("Security check not passed for file " << device << ", check .nmmrc for options" << std::endl);
        throw DeviceNotFoundException(this, "Security check failed for device", device);
      }
      DEBUG_STREAM("Probing ivtv device " << device << std::endl);
      __fd = ::open(device.c_str(), 0);
      if(__fd < 0) {
        string error(strerror(errno));
        ERROR_STREAM("Could not open device " << device << " (" << error << ")" << std::endl);
        throw DeviceNotFoundException(this, "Failed to open device", device);
      }
      
      /* query capabilities and check if this is really an ivtv device */
      v4l2_capability vcap;
      if(ioctl(__fd, VIDIOC_QUERYCAP, &vcap) < 0) {
        ERROR_STREAM("Failed to query device capabilities (ioctl: VIDIOC_QUERYCAP failed)" << std::endl);
        ERROR_STREAM("Maybe the device is not a video4linux2 TV card" << std::endl);
        throw DeviceNotFoundException(this, "Failed to query device capabilities", device);
      } else {
        /* print out information */
        DEBUG_STREAM("Driver name   : " << vcap.driver << std::endl);
        DEBUG_STREAM("Card type     : " << vcap.card << std::endl);
        DEBUG_STREAM("Bus info      : " << vcap.bus_info << std::endl);
        DEBUG_STREAM("Driver version: " << vcap.version << std::endl);
        DEBUG_STREAM("Capabilities  : 0x" << std::setfill('0') << std::setw(8) << std::hex <<
                     vcap.capabilities << std::dec << std::endl);
        
        /* check for ivtv driver */
        if(std::string(reinterpret_cast<const char*>(vcap.driver)) != "ivtv") {
          ERROR_STREAM("Device is not an ivtv device" << std::endl);
          throw WrongDeviceException(this, "Device is not an ivtv device", device);
        }
      }
      
      /* close device */
      DEBUG_STREAM("Closing ivtv device" << std::endl);
      if(__fd >= 0) {
        ::close(__fd);
        __fd = -1;
      }
    } catch(...) {
      /* close device */
      DEBUG_STREAM("Closing ivtv device" << std::endl);
      if(__fd >= 0) {
        ::close(__fd);
        __fd = -1;
      }
      
      /* rethrow exception */
      throw;
    }
    
    /* now set the device name */
    __device = device;
  }
  
  string IVTVReadNode::getDevice() const {
    return __device;
  }
  
  void IVTVReadNode::setResolution(const nmm_int32& x, const nmm_int32& y) throw(const Exception&) {
    struct v4l2_format vfmt;
    vfmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if (ioctl(__fd, VIDIOC_G_FMT, &vfmt) < 0) {
      ERROR_STREAM("Couldn't query old video format"<<endl);
      throw Exception("Couldn't query old video format (VIDIOC_G_FMT failed)");
    }

    vfmt.fmt.pix.width  = x;
    vfmt.fmt.pix.height = y;

    if (ioctl(__fd, VIDIOC_S_FMT, &vfmt) < 0) {
      ERROR_STREAM("Couldn't set Size to "<<x<<"-"<<y<<endl);
      throw Exception("Couldn't set video format (VIDIOC_S_FMT failed)");
    }
  }
  
  void IVTVReadNode::getResolution(nmm_int32& x, nmm_int32& y) throw(const Exception&) {
    struct v4l2_format vfmt;
    vfmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    if(ioctl(__fd, VIDIOC_G_FMT, &vfmt) < 0) {
      ERROR_STREAM("ioctl: VIDIOC_G_FMT failed" << endl);
      throw Exception("Couldn't get video format (VIDIOC_G_FMT failed)");
    }
    x = vfmt.fmt.pix.width;
    y = vfmt.fmt.pix.height;
  }
  
  Result IVTVReadNode::setChannel(const nmm_int32& channel) {
    try {
      /* Begin> Check if current input is a tuner */
      unsigned int cur_input = getCurrentInput();
      if (cur_input >= __inputs.size()) {
        ERROR_STREAM("Can't set channel (no information about current input)" << endl);
        return FAILURE;
      }
      if (__inputs[cur_input].getType() != TVInput::TYPE_TUNER) {
        ERROR_STREAM("Can't set channel (current input is not a tuner)" << endl);
      }
      unsigned int cur_tuner = __inputs[cur_input].getTuner();
      /* End< */
      
      /* Begin> Get current frequency table */
      if (__cur_freqtable.empty()) {
        ERROR_STREAM("No frequency table set");
        return FAILURE;
      }
      const FrequencyTable& table = __frequencies[__cur_freqtable];
      /* End< */
      
      /* Begin> Determine frequency */
      Channel c  = ivtv.getChannel(channel);
      int cnr = table.findChannel(c.frequency);

      if (cnr < 0 || cnr >= static_cast<int>(table.getChannelCount())) {
        ERROR_STREAM("Channel not found: " << c.name << endl);
        return FAILURE;
      }

      unsigned int frequency = table.getChannelFrequency(cnr);
      /* End< */
      
      /* Set frequency at current tuner */
      setFrequency(cur_tuner, static_cast<float>(frequency) / 1000);
      __cur_channel = channel;

      __do_flush = true;

      return SUCCESS;
    } catch( const Exception& e) {
      ERROR_STREAM("SetChannel failed: " << e.getComment() << endl);
      return FAILURE;
    }
  }
  
  Result IVTVReadNode::getChannel(nmm_int32& channel, string& name) {
    /* Begin> Get current frequency table */
    if (__cur_freqtable.empty()) {
      ERROR_STREAM("No frequency table set");
      return FAILURE;
    }
    
    channel = __cur_channel;
    name = ivtv.getChannel(channel).name;
    
    return SUCCESS;
  }


  /*============================== ITVCard Interface ==============================*/
  
  
  /*============================== ITV Interface ==============================*/
  unsigned int IVTVReadNode::getAudioCount() const {
    return __audio_inputs.size();
  }
  
  TVAudio IVTVReadNode::getAudio(const nmm_uint32& num) const {
    if (num >= __audio_inputs.size()) {
      throw Exception(string("Invalid audio input: ") + num);
    }
    
    return __audio_inputs[num];
  }
  
  list<TVAudio> IVTVReadNode::getAudioList() const {
    return list<TVAudio>(__audio_inputs.begin(), __audio_inputs.end());
  }
  
  nmm_uint32 IVTVReadNode::getCurrentAudio() const {
    struct v4l2_audio audio;
    if (ioctl(__fd, VIDIOC_G_AUDIO, &audio) < 0) {
      throw Exception("Failed to get current audio input (ioctl: VIDIOC_G_AUDIO failed)");
    } else {
      return audio.index;
    }
  }
  
  void IVTVReadNode::setCurrentAudio(const nmm_uint32& num) {
    struct v4l2_audio audio;
    memset (&audio, 0, sizeof (audio));
    audio.index = num;
    if (ioctl(__fd, VIDIOC_S_AUDIO, &audio) < 0) {
      throw Exception("Failed to set current audio input (ioctl: VIDIOC_S_AUDIO failed)");
    } else {
      DEBUG_STREAM("Set current audio to " << num << endl);
      __do_flush = true;
    }
  }
  
  nmm_uint32 IVTVReadNode::getInputCount() const {
    return __inputs.size();
  }
  
  TVInput IVTVReadNode::getInput(const nmm_uint32& num) const {
    if (num >= __inputs.size()) {
      throw Exception(string("Invalid input: ") + num);
    }
    
    return __inputs[num];
  }
  
  list<TVInput> IVTVReadNode::getInputList() const {
    return list<TVInput>(__inputs.begin(), __inputs.end());
  }
  
  nmm_uint32 IVTVReadNode::getCurrentInput() const {
    int input;
    if (ioctl(__fd, VIDIOC_G_INPUT, &input) < 0) {
      throw Exception("Failed to get current input (ioctl: VIDIOC_G_INPUT failed)");
    } else {
      return input;
    }
  }
  
  void IVTVReadNode::setCurrentInput(const nmm_uint32& num) {
    int input = num;
    if (ioctl(__fd, VIDIOC_S_INPUT, &input) < 0) {
      throw Exception("Failed to set current input (ioctl: VIDIOC_S_INPUT failed)");
    } else {
      DEBUG_STREAM("Set current input to " << num << endl);
      __do_flush = true;
    }
  }
  
  nmm_uint32 IVTVReadNode::getStandardByName(const string& name) const {
    for (StandardConstIterator i(__standards.begin());
         i != __standards.end();
         ++i) {
      if (i->second.getName() == name) {
        return i->first;
      }
    }
    
    return 0;
  }
  
  TVStandard IVTVReadNode::getStandard(const nmm_uint32& num) const {
    StandardConstIterator found(__standards.find(num));
    if (found == __standards.end()) {
      throw Exception(string("Standard ") + num + " not supported");
    }
    return found->second;
  }
  
  list<TVStandard> IVTVReadNode::getStandardList() const {
    list<TVStandard> std_list;
    for (StandardConstIterator i(__standards.begin());
         i != __standards.end();
         ++i) {
      std_list.push_back(i->second);
    }
    return std_list;
  }
  
  nmm_uint32 IVTVReadNode::getCurrentStandard() const {
    v4l2_std_id std;
    if (ioctl(__fd, VIDIOC_G_STD, &std) < 0) {
      ERROR_STREAM("Failed to get current standard (ioctl: VIDIOC_G_STD failed)" << std::endl);
      throw Exception("Failed to get current standard (ioctl: VIDIOC_G_STD failed)");
    }
    return std;
  }
  
  void IVTVReadNode::setCurrentStandard(const nmm_uint32& standard) {
    v4l2_std_id std = standard;
    if (ioctl(__fd, VIDIOC_S_STD, &std) < 0) {
      throw Exception("Failed to set current standard (ioctl: VIDIOC_S_STD failed)");
    } else {
      DEBUG_STREAM("Set standard to " << standard << " for current input" << endl);
      __do_flush = true;
    }
  }
  
  void IVTVReadNode::setStandardByName(const string& name) {
    unsigned int standard = getStandardByName(name);
    if (standard == 0) {
      throw Exception("Standard " + name + " does not exist");
    }
    setCurrentStandard(standard);
  }
  
  nmm_uint32 IVTVReadNode::getTunerCount() const {
    return __tuners.size();
  }
  
  TVTuner IVTVReadNode::getTuner(const nmm_uint32& index) const {
    if (index >= __tuners.size()) {
      throw Exception(string("Invalid tuner index ") + index);
    }
    return __tuners[index];
  }
  
  list<TVTuner> IVTVReadNode::getTunerList() const {
    return list<TVTuner>(__tuners.begin(), __tuners.end());
  }
  
  nmm_float IVTVReadNode::getFrequency(const nmm_uint32& tuner) const {
    struct v4l2_frequency vf;
    vf.tuner = tuner;
    if (ioctl(__fd, VIDIOC_G_FREQUENCY, &vf) < 0) {
      throw Exception("Failed to get frequency (ioctl: VIDIOC_G_FREQUENCY failed)");
    } else {
      float granularity = 0.0625;
      if (tuner < __tuners.size()) {
        granularity = __tuners[tuner].getGranularity();
      }
      return static_cast<float>(vf.frequency) * granularity;
    }
  }
  
  void IVTVReadNode::setFrequency(const nmm_uint32& tuner, const nmm_float& frequency) {
    struct v4l2_frequency vf;
    vf.tuner     = tuner;
    
    float granularity = 0.0625;
    if (tuner < __tuners.size()) {
      granularity = __tuners[tuner].getGranularity();
    }
    vf.frequency = static_cast<unsigned int>(frequency / granularity);
    
    if (ioctl(__fd, VIDIOC_S_FREQUENCY, &vf) < 0) {
      throw Exception("Failed to set frequency (ioctl: VIDIOC_S_FREQUENCY failed)");
    } else {
      DEBUG_STREAM("Set frequency to " << frequency << " on tuner " << tuner << endl);
    }
  }
  
  nmm_uint32 IVTVReadNode::getBrightness() const {
    return getControl(V4L2_CID_BRIGHTNESS, "V4L2_CID_BRIGHTNESS");
  }
  
  nmm_uint32 IVTVReadNode::getContrast() const {
    return getControl(V4L2_CID_CONTRAST, "V4L2_CID_CONTRAST");
  }
  
  nmm_uint32 IVTVReadNode::getSaturation() const {
    return getControl(V4L2_CID_SATURATION, "V4L2_CID_SATURATION");
  }
  
  nmm_int32 IVTVReadNode::getHue() const {
    return getControl(V4L2_CID_HUE, "V4L2_CID_HUE");
  }
  
  nmm_uint32 IVTVReadNode::getVolume() const {
    return getControl(V4L2_CID_AUDIO_VOLUME, "V4L2_CID_AUDIO_VOLUME");
  }
  
  nmm_bool IVTVReadNode::getMute() const {
    return (getControl(V4L2_CID_AUDIO_MUTE, "V4L2_CID_AUDIO_MUTE") != 0);
  }
  
  void IVTVReadNode::setBrightness(const nmm_uint32& brightness) {
    setControl(V4L2_CID_BRIGHTNESS, brightness, "V4L2_CID_BRIGHTNESS");
  }
  
  void IVTVReadNode::setContrast(const nmm_uint32& contrast) {
    setControl(V4L2_CID_CONTRAST, contrast, "V4L2_CID_CONTRAST");
  }
  
  void IVTVReadNode::setSaturation(const nmm_uint32& saturation) {
    setControl(V4L2_CID_SATURATION, saturation, "V4L2_CID_SATURATION");
  }
  
  void IVTVReadNode::setHue(const nmm_int32& hue) {
    setControl(V4L2_CID_HUE, hue, "V4L2_CID_HUE");
  }
  
  void IVTVReadNode::setVolume(const nmm_uint32& volume) {
    setControl(V4L2_CID_AUDIO_VOLUME, volume, "V4L2_CID_AUDIO_VOLUME");
  }
  
  void IVTVReadNode::setMute(const nmm_bool& mute) {
    setControl(V4L2_CID_AUDIO_MUTE, mute ? 1 : 0, "V4L2_CID_AUDIO_MUTE");
  }
  
  list<string> IVTVReadNode::getFrequencyTableList() const {
    list<string> outlist;
    for (map<string, FrequencyTable>::const_iterator i(__frequencies.begin());
         i != __frequencies.end();
         ++i) {
      outlist.push_back(i->first);
    }
    return outlist;
  }
  
  string IVTVReadNode::getFrequencyTable() const {
    return __cur_freqtable;
  }
  
  void IVTVReadNode::setFrequencyTable(const string& table) {
    /* Check if this frequency table exists */
    map<string, FrequencyTable>::const_iterator found(__frequencies.find(table));
    if (found == __frequencies.end()) {
      throw Exception("Frequency table " + table + " does not exist");
    }
    
    /* Set channel */
    __cur_freqtable = table;
    setChannel(0);
  }

  void IVTVReadNode::setInputOutput(const nmm_int32& input, const nmm_int32& output){
    struct msp_matrix mm;
    mm.input = input;
    mm.output = output;
    
    if(ioctl(__fd, MSP_SET_MATRIX, &mm)){
      DEBUG_STREAM("Setting MSP-Matrix: input=" << input << ",output=" << output << endl);
    }else{
      throw Exception("Values in MSP-Matrix not set!");
    }
    
    
  }

  /*============================== IIVTV Interface ==============================*/
  IVTVCodec IVTVReadNode::getCodecParams() const {
    struct ivtv_ioctl_codec in_codec;
    
    if(ioctl(__fd, IVTV_IOC_G_CODEC, &in_codec) < 0) {
      throw Exception("Failed to get Codecparametners (IVTV_IOC_G_CODEC failed)");
    } else {
      IVTVCodec current_codec;
      current_codec.setAudioBitmask(in_codec.audio_bitmask);
      current_codec.setBFrames(in_codec.bframes);
      current_codec.setBitrateMode(in_codec.bitrate_mode);
      current_codec.setBitrate(in_codec.bitrate);
      current_codec.setBitratePeak(in_codec.bitrate_peak);
      current_codec.setDNRMode(in_codec.dnr_mode);
      current_codec.setDNRSpatial(in_codec.dnr_spatial);
      current_codec.setDNRTemporal(in_codec.dnr_temporal);
      current_codec.setDNRType(in_codec.dnr_type);
      current_codec.setFramerate(in_codec.framerate);
      current_codec.setFramesPerGOP(in_codec.framespergop);
      current_codec.setGOPClosure(in_codec.gop_closure);
      current_codec.setPulldown(in_codec.pulldown);
      current_codec.setStreamType(in_codec.stream_type);
      
      return current_codec;
    }
  }
  
  void IVTVReadNode::setCodecParams(const IVTVCodec& codec) {
    struct ivtv_ioctl_codec out_codec;
    
    out_codec.audio_bitmask = codec.getAudioBitmask();
    out_codec.bframes = codec.getBFrames();
    out_codec.bitrate_mode = codec.getBitrateMode();
    out_codec.bitrate = codec.getBitrate();
    out_codec.bitrate_peak = codec.getBitratePeak();
    out_codec.dnr_mode = codec.getDNRMode();
    out_codec.dnr_spatial = codec.getDNRSpatial();
    out_codec.dnr_temporal = codec.getDNRTemporal();
    out_codec.dnr_type = codec.getDNRType();
    out_codec.framerate = codec.getFramerate();
    out_codec.framespergop = codec.getFramesPerGOP();
    out_codec.gop_closure = codec.getGOPClosure();
    out_codec.pulldown = codec.getPulldown();
    out_codec.stream_type = codec.getStreamType();
    
    if(ioctl(__fd, IVTV_IOC_S_CODEC, &out_codec) < 0) {
      throw Exception("Failed to set Codecparametners (IVTV_IOC_S_CODEC failed)");
    }
  }
  
  void IVTVReadNode::setGOPSize(const nmm_uint32& value) {
    struct ivtv_ioctl_codec in_codec;
    if (ioctl(__fd, IVTV_IOC_G_CODEC, &in_codec) < 0)
      throw Exception("Error getting Codec Params");
    else {
      in_codec.framespergop = value;
      if (ioctl(__fd, IVTV_IOC_S_CODEC, &in_codec) < 0)
        throw Exception("Error setting Codec Param GOPSize");
    }
  }
  
  void IVTVReadNode::setBFrames(const nmm_uint32& value) {
    struct ivtv_ioctl_codec in_codec;
    if (ioctl(__fd, IVTV_IOC_G_CODEC, &in_codec) < 0)
      throw Exception("Error getting Codec Params");
    else {
      in_codec.bframes = value;
      if (ioctl(__fd, IVTV_IOC_S_CODEC, &in_codec) < 0)
        throw Exception("Error setting Codec Param bframes");
    }
  }
  
  void IVTVReadNode::setGOPClosure(const nmm_uint32& value) {
    struct ivtv_ioctl_codec in_codec;
    if (ioctl(__fd, IVTV_IOC_G_CODEC, &in_codec) < 0)
      throw Exception("Error getting Codec Params");
    else {
      in_codec.gop_closure = value;
      if (ioctl(__fd, IVTV_IOC_S_CODEC, &in_codec) < 0)
        throw Exception("Error setting Codec Param GOPClosure");
    }
  }

  void IVTVReadNode::setPulldown(const nmm_uint32& value) {
    struct ivtv_ioctl_codec in_codec;
    if (ioctl(__fd, IVTV_IOC_G_CODEC, &in_codec) < 0)
      throw Exception("Error getting Codec Params");
    else {
      in_codec.pulldown = value;
      if (ioctl(__fd, IVTV_IOC_S_CODEC, &in_codec) < 0)
        throw Exception("Error setting Codec Param pulldown");
    }
  }

  void IVTVReadNode::setBitrateMode(const nmm_uint32& value) {
    struct ivtv_ioctl_codec in_codec;
    if (ioctl(__fd, IVTV_IOC_G_CODEC, &in_codec) < 0)
      throw Exception("Error getting Codec Params");
    else {
      in_codec.bitrate_mode = value;
      if (ioctl(__fd, IVTV_IOC_S_CODEC, &in_codec) < 0)
        throw Exception("Error setting Codec Param Bitrate Mode");
    }
  }

  void IVTVReadNode::setBitrate(const nmm_uint32& value) {
    struct ivtv_ioctl_codec in_codec;
    if (ioctl(__fd, IVTV_IOC_G_CODEC, &in_codec) < 0)
      throw Exception("Error getting Codec Params");
    else {
      //first: look at PeakBitrate! PeakBitrate _must_ be >= Bitrate
      //if peakbitrate < new bitrate --> raise PeakBitrate to new Bitrate
      if (in_codec.bitrate_peak < value)
        in_codec.bitrate_peak = value;
      in_codec.bitrate = value;
      if (ioctl(__fd, IVTV_IOC_S_CODEC, &in_codec) < 0)
        throw Exception("Error setting Codec Param bitrate");
    }
  }

  void IVTVReadNode::setPeakBitrate(const nmm_uint32& value) {
    struct ivtv_ioctl_codec in_codec;
    if (ioctl(__fd, IVTV_IOC_G_CODEC, &in_codec) < 0)
      throw Exception("Error getting Codec Params");
    else {
      //first: look at Bitrate! Bitrate _must_ be <= peak_Bitrate
      //if bitrate > peak_bitrate --> lower bitrate to peak_bitrate
      if (in_codec.bitrate > value)
        in_codec.bitrate = value;
      in_codec.bitrate_peak = value;
      if (ioctl(__fd, IVTV_IOC_S_CODEC, &in_codec) < 0)
        throw Exception("Error setting Codec Param peak_bitrate");
    }
  }

  void IVTVReadNode::setStreamType(const nmm_uint32& value) {
    struct ivtv_ioctl_codec in_codec;
    if (ioctl(__fd, IVTV_IOC_G_CODEC, &in_codec) < 0)
      throw Exception("Error getting Codec Params");
    else {
      in_codec.stream_type = value;
      if (ioctl(__fd, IVTV_IOC_S_CODEC, &in_codec) < 0)
        throw Exception("Error setting Codec Param stream_type");
    }
  }

  void IVTVReadNode::setDNRMode(const nmm_uint32& value) {
    struct ivtv_ioctl_codec in_codec;
    if (ioctl(__fd, IVTV_IOC_G_CODEC, &in_codec) < 0)
      throw Exception("Error getting Codec Params");
    else {
      in_codec.dnr_mode = value;
      if (ioctl(__fd, IVTV_IOC_S_CODEC, &in_codec) < 0)
        throw Exception("Error setting Codec Param DNR Mode");
    }
  }

  void IVTVReadNode::setDNRType(const nmm_uint32& value) {
    struct ivtv_ioctl_codec in_codec;
    if (ioctl(__fd, IVTV_IOC_G_CODEC, &in_codec) < 0)
      throw Exception("Error getting Codec Params");
    else {
      in_codec.dnr_type = value;
      if (ioctl(__fd, IVTV_IOC_S_CODEC, &in_codec) < 0)
        throw Exception("Error setting Codec Param DNR Type");
    }
  }

  void IVTVReadNode::setDNRSpatial(const nmm_uint32& value) {
    struct ivtv_ioctl_codec in_codec;
    if (ioctl(__fd, IVTV_IOC_G_CODEC, &in_codec) < 0)
      throw Exception("Error getting Codec Params");
    else {
      in_codec.dnr_spatial = value;
      if (ioctl(__fd, IVTV_IOC_S_CODEC, &in_codec) < 0)
        throw Exception("Error setting Codec Param DNR Spatial");
    }
  }

  void IVTVReadNode::setDNRTemporal(const nmm_uint32& value) {
    struct ivtv_ioctl_codec in_codec;
    if (ioctl(__fd, IVTV_IOC_G_CODEC, &in_codec) < 0)
      throw Exception("Error getting Codec Params");
    else {
      in_codec.dnr_temporal = value;
      if (ioctl(__fd, IVTV_IOC_S_CODEC, &in_codec) < 0)
        throw Exception("Error setting Codec Param DNR Temporal");
    }
  }
  

  
  /*============================== GenericNode Methods ==============================*/
  Result IVTVReadNode::doInit() {
    /* Begin> Open device */
    if(!SSecurityCheck::getInstance().checkFilenameForReading(__device)) {
      ERROR_STREAM("Security check not passed for file " << __device << ", check .nmmrc for options" << endl);
      return FAILURE;
    }
    DEBUG_STREAM("Opening ivtv device " << __device << endl);
    __fd = ::open(__device.c_str(), O_RDWR/*O_RDONLY*/);
    if (__fd < 0) {
      string error(strerror(errno));
      ERROR_STREAM("Could not open device " << __device << " (" << error << ")" << endl);
      return FAILURE;
    }
    /* End< */

    /* Begin> Enumerate audio inputs */
    struct v4l2_audio vaudio;
    memset (&vaudio, 0, sizeof (vaudio));
    vaudio.index = 0;
    __audio_inputs.clear();
    DEBUG_STREAM("Available audio inputs:" << endl);
    while (ioctl(__fd, VIDIOC_ENUMAUDIO, &vaudio) >= 0) {
      TVAudio audio(vaudio.index, string(reinterpret_cast<const char*>(vaudio.name)),
                    vaudio.capability, vaudio.mode);
      memset (&vaudio, 0, sizeof (vaudio));
      vaudio.index = audio.getIndex()+1;
      __audio_inputs.push_back(audio);
      DEBUG_STREAM(" Index     : " << audio.getIndex() << endl <<
                   "  Name     : " << audio.getName() << endl <<
                   "  Caps     : " << audio.getCapabilities() << endl <<
                   "  Mode     : " << audio.getMode() << endl);
    }
    /* End< */
    
    /* Begin> Enumerate inputs */
    struct v4l2_input vin;
    vin.index = 0;
    __inputs.clear();
    DEBUG_STREAM("Available inputs:" << endl);
    while (ioctl(__fd, VIDIOC_ENUMINPUT, &vin) >= 0) {
      TVInput input(vin.index, string(reinterpret_cast<const char*>(vin.name)),
                    vin.type, vin.audioset, vin.tuner, vin.std, vin.status);
      __inputs.push_back(input);
      DEBUG_STREAM(" Index     : " << input.getNum() << endl <<
                   "  Name     : " << input.getName() << endl <<
                   "  Type     : " << input.getType() << endl <<
                   "  AudioSet : " << input.getAudioSet() << endl <<
                   "  Tuner    : " << input.getTuner() << endl <<
                   "  Standard : " << input.getStandard() << endl <<
                   "  Status   : " << input.getStatus() << endl);
      vin.index++;
    }
    /* End< */
    
    /* Begin> Enumerate standards */
    struct v4l2_standard vs;
    vs.index = 0;
    __standards.clear();
    DEBUG_STREAM("Supported standards:" << endl);
    while(ioctl(__fd, VIDIOC_ENUMSTD, &vs) >= 0) {
      TVStandard standard(vs.id, string(reinterpret_cast<const char*>(vs.name)),
                          vs.frameperiod.numerator, vs.frameperiod.denominator,
                          vs.framelines);
      __standards.insert(pair<unsigned int, TVStandard>(standard.getID(), standard));
      DEBUG_STREAM(" ID            : " << standard.getID() << endl <<
                   "  Name         : " << standard.getName() << endl <<
                   "  Frame period : " << standard.getFramePeriodNumerator() << "/" <<
                   standard.getFramePeriodDenominator() << endl <<
                   "  Frame lines  : " << standard.getFrameLines() << endl);
      vs.index++;
    }
    /* End< */
    
    /* Begin> Enumerate tuners */
    struct v4l2_tuner vt;
    memset (&vt, 0, sizeof (vt));
    vt.index = 0;
    __tuners.clear();
    DEBUG_STREAM("Supported tuners:" << endl);
    while(ioctl(__fd, VIDIOC_G_TUNER, &vt) >= 0) {
      /* Begin> Determine tuner type */
      unsigned int type;
      switch(vt.type) {
      case V4L2_TUNER_RADIO:
        type = TVTuner::TYPE_RADIO;
        break;
      case V4L2_TUNER_ANALOG_TV:
        type = TVTuner::TYPE_ANALOG_TV;
        break;
      default:
        type = TVTuner::TYPE_UNKNOWN;
      }
      /* End< Determine tuner type */
      
      /* Begin> Determine frequency scale */
      float granularity;
      if (vt.capability & V4L2_TUNER_CAP_LOW) {
        granularity = 0.0000625; /* 62.5 Hz */
      } else {
        granularity = 0.0625;    /* 62.5 kHz */
      }
      float range_min = static_cast<float>(vt.rangelow) * granularity;
      float range_max = static_cast<float>(vt.rangehigh) * granularity;
      /* End< */
      
      TVTuner tuner(vt.index, string(reinterpret_cast<const char*>(vt.name)), type,
                    range_min, range_max, granularity);
      __tuners.push_back(tuner);
      
      DEBUG_STREAM(" Index        : " << vt.index << endl <<
                   "  Name        : " << vt.name << endl <<
                   "  Type        : " << ((type == TVTuner::TYPE_RADIO) ? "RADIO" :
                                          (type == TVTuner::TYPE_ANALOG_TV) ? "ANALOG TV" :
                                          "UNKNOWN") << endl <<
                   "  Range       : " << range_min << "~" << range_max << " MHz" << endl <<
                   "  Granularity : " << granularity << " MHz" << endl);
      memset (&vt, 0, sizeof (vt));
      vt.index = __tuners.size();
    }
    /* End< */
    
    /* Begin> Initialize frequency tables */
    try {
      initFrequencyTables();
      ivtv.load(ChannelConfig::IVTV, channelConfigFile);
    } catch(const Exception& e) {
      ERROR_STREAM("Failed to initialize frequency tables" << endl);
      ERROR_STREAM(e << endl);
      return FAILURE;
    }
    /* End< */
    
    /* BEGIN> detailed mpeg program stream */
    Format* o_format = getOwnOutputProperty()->addNewFormat(Format::av_mpegps_type);
    o_format->addIntParamValue(Format::version_param, 2);
    o_format->addIntParamValue(Format::map_mpeg_audio_streams_param, 1); // only first mpeg audio stream
    o_format->addIntParamValue(Format::map_mpeg_video_streams_param, 1); // only first mpeg audio stream
    o_format->addIntParamValue(Format::map_ac3_audio_streams_param, 0); // no ac3
    o_format->addIntParamValue(Format::map_dts_audio_streams_param, 0); // no dts
    o_format->addIntParamValue(Format::map_lpcm_audio_streams_param, 0); // no lpcm
    o_format->addIntParamValue(Format::map_spu_streams_param, 0); // no spu
    
    /* Begin> Audio sub format */
    {
      Format* sub_format = new Format(Format::audio_mpeg_type);
      sub_format->addIntParamValue(Format::audio_codec_param, 0x55);
      sub_format->addIntParamValue(Format::layer_param, 2);
      sub_format->addIntParamValue(Format::channels_param, 2);
      sub_format->addIntParamValue(Format::sample_rate_param, 48000);
      sub_format->addIntParamValue(Format::bitpersample_param, 16);
      o_format->addFPointerParamValue(Format::format_mpeg_audio_streams_param, sub_format);
    }
    /* End< Audio sub format */
    
    /* Begin> Video sub format */
    {
      list<TVStandard> standards = getStandardList();
      
      Format* sub_format = new Format(Format::video_mpeg_type);
      sub_format->addIntParamValue(Format::version_param, 2);
      sub_format->addWildcard(Format::x_resolution_param);
      sub_format->addWildcard(Format::y_resolution_param);
      sub_format->addWildcard(Format::ratio_x_param);
      sub_format->addWildcard(Format::ratio_y_param);
      for(list<TVStandard>::const_iterator i(standards.begin());
          i != standards.end();
          ++i) {
        float framerate = float(i->getFramePeriodDenominator()) / float(i->getFramePeriodNumerator());
        DEBUG_STREAM("Found framerate " << framerate << endl);
        sub_format->addFloatParamValue(Format::framerate_param, framerate);
      }
      o_format->addFPointerParamValue(Format::format_mpeg_video_streams_param, sub_format);
    }
    /* End< Video sub format */
    
    /* add dummy formats */
    o_format->addFPointerParamValue(Format::format_ac3_audio_streams_param, new Format(Format::wild_wild_type));
    o_format->addFPointerParamValue(Format::format_spu_streams_param, new Format(Format::wild_wild_type));
    o_format->addFPointerParamValue(Format::format_lpcm_audio_streams_param, new Format(Format::wild_wild_type));
    o_format->addFPointerParamValue(Format::format_dts_audio_streams_param, new Format(Format::wild_wild_type));
    /* END< */
    
    getOwnOutputProperty()->setDefaultFormat(o_format);
    
    return SUCCESS;
  }
  
  Result IVTVReadNode::doDeinit() {
    /* Erase property */
    getOwnOutputProperty()->erase();
    
    /* Close device */
    if (__fd >= 0) {
      DEBUG_STREAM("Closing ivtv device" << endl);
      ::close(__fd);
      __fd = -1;
    }
    
    /* Clear lists */
    __audio_inputs.clear();
    __inputs.clear();
    __standards.clear();
    __tuners.clear();
    
    /* Clear frequency tables */
    __cur_channel = 0;
    __cur_freqtable = "";
    __frequencies.clear();
    
    return SUCCESS;
  }
  
  Result IVTVReadNode::doInitOutput() {
    /* determine resolution */
    int x_resolution;
    int y_resolution;
    try {
      getResolution(x_resolution, y_resolution);
      DEBUG_STREAM("Current resolution is " << x_resolution << "x" << y_resolution << endl);
    } catch(const Exception& e) {
      ERROR_STREAM("Failed to determine resolution" << endl);
      ERROR_STREAM(e << endl);
      return FAILURE;
    }
    
    /* determine framerate */
    float framerate;
    try {
      int standard = getCurrentStandard();
      TVStandard standard_info = getStandard(standard);
      framerate = float(standard_info.getFramePeriodDenominator()) / float(standard_info.getFramePeriodNumerator());
      DEBUG_STREAM("Current framerate is " << framerate << endl);
    } catch(const Exception& e) {
      ERROR_STREAM("Failed to determine framerate" << endl);
      ERROR_STREAM(e << endl);
      return FAILURE;
    }
    
    /* BEGIN> detailed mpeg program stream */
    Format* o_format = getOwnWorkingOutputProperty()->addNewFormat(Format::av_mpegps_type);
    o_format->addIntParamValue(Format::version_param, 2);
    o_format->addIntParamValue(Format::map_mpeg_audio_streams_param, 1); // only first mpeg audio stream
    o_format->addIntParamValue(Format::map_mpeg_video_streams_param, 1); // only first mpeg audio stream
    o_format->addIntParamValue(Format::map_ac3_audio_streams_param, 0); // no ac3
    o_format->addIntParamValue(Format::map_dts_audio_streams_param, 0); // no dts
    o_format->addIntParamValue(Format::map_lpcm_audio_streams_param, 0); // no lpcm
    o_format->addIntParamValue(Format::map_spu_streams_param, 0); // no spu
    
    /* Begin> Audio sub format */
    {
      Format* sub_format = new Format(Format::audio_mpeg_type);
      sub_format->addIntParamValue(Format::audio_codec_param, 0x55);
      sub_format->addIntParamValue(Format::layer_param, 2);
      sub_format->addIntParamValue(Format::channels_param, 2);
      sub_format->addIntParamValue(Format::sample_rate_param, 48000);
      sub_format->addIntParamValue(Format::bitpersample_param, 16);
      o_format->addFPointerParamValue(Format::format_mpeg_audio_streams_param, sub_format);
    }
    /* End< Audio sub format */
    
    /* Begin> Video sub format */
    {
      Format* sub_format = new Format(Format::video_mpeg_type);
      sub_format->addIntParamValue(Format::version_param, 2);
      sub_format->addIntParamValue(Format::x_resolution_param, x_resolution);
      sub_format->addIntParamValue(Format::y_resolution_param, y_resolution);
      sub_format->addIntParamValue(Format::ratio_x_param, x_resolution);
      sub_format->addIntParamValue(Format::ratio_y_param, y_resolution);
      sub_format->addFloatParamValue(Format::framerate_param, framerate);
      o_format->addFPointerParamValue(Format::format_mpeg_video_streams_param, sub_format);
    }
    /* End< Video sub format */
    
    /* add dummy formats */
    o_format->addFPointerParamValue(Format::format_ac3_audio_streams_param, new Format(Format::wild_wild_type));
    o_format->addFPointerParamValue(Format::format_spu_streams_param, new Format(Format::wild_wild_type));
    o_format->addFPointerParamValue(Format::format_lpcm_audio_streams_param, new Format(Format::wild_wild_type));
    o_format->addFPointerParamValue(Format::format_dts_audio_streams_param, new Format(Format::wild_wild_type));
    /* END< */

    getOwnWorkingOutputProperty()->setDefaultFormat(o_format);
    
    return SUCCESS;
  }
  
  Result IVTVReadNode::doActivate() {
        
    __do_flush = false;
    return SUCCESS;
  }

  Message* IVTVReadNode::processBuffer(Buffer *) {
    /* Check if we need to flush */
    if (__do_flush) {
      __do_flush = false;
      CEvent* cevent = new FlushEvent();
      cevent->insertEvent(cevent->size(), ISyncReset::create_syncReset());
      __stream_timer.setTimestamp(cevent);
      return cevent;
    }
    
    /* Get output buffer */
    Buffer* out = getNewBuffer(__buffer_size);
    if(!out) {
      ERROR_STREAM("Failed to allocate buffer");
      setProducingFlag(false);
      return 0;
    }
    
    struct ivtv_ioctl_framesync frame_info;
    int ioctl_result    = 0;

    Time ioctl_time = TimedElement::getTime();

    // TODO: which one is better? framesync or timing?
    ioctl_result    = ioctl(__fd, IVTV_IOC_GET_TIMING, &frame_info);
    // ioctl_result = ioctl(__fd, IVTV_IOC_FRAMESYNC , &framesync_frame_info);

    /* Read from device */
    ssize_t bytes_read = ::read(__fd, out->getData(), __buffer_size);

    // if( !__received_first_buffer && ( __buffer_counter > 200)) {
    if( !__received_first_buffer) {
      __received_first_buffer = true;
      
      // System start time
      __system_start_time = ioctl_time;
      cout << "System clock start time  : " << __system_start_time << endl;

      if ( ioctl_result < 0 ) {
        cerr << "ioctl: IVTV_IOC_FRAMESYNC or IVTV_IOC_GET_TIMING failed" << endl;
      } else {
	__scr_start = frame_info.scr;
	cout << "SCR start time    : " << Time(Rational(__scr_start, MPEG_CLOCK_FREQ)) << endl;
      }
    }   
    
    if (bytes_read > 0) {

      uint64_t scr_now    = frame_info.scr;
      Rational scr_time(scr_now - __scr_start, MPEG_CLOCK_FREQ);

      out->setUsed(bytes_read);
      Timestamp timestamp;
      timestamp.sync_time = Time(scr_time);
      timestamp.stream_counter = __buffer_counter;
      timestamp.is_valid = true;
      timestamp.is_set = true;
      out->setTimestamp(timestamp);

      // __stream_timer.setTimestamp(out);

      ++__buffer_counter;
      return out;
    } else {
      ERROR_STREAM("Read 0 bytes from device" << endl);
      out->release();
      return 0;
    }
  }
  /*============================== GenericNode Methods ==============================*/


  /*============================== Helper Methods ==============================*/
  void  IVTVReadNode::setControl(unsigned int id, unsigned int value, const string& name) {
    struct v4l2_control ctrl;
    ctrl.id = id;
    ctrl.value = value;
    if (ioctl(__fd, VIDIOC_S_CTRL, &ctrl) == 0) {
      DEBUG_STREAM("Set control " << name << " to " << value << endl);
    } else {
      throw Exception("Error setting control value (VIDIOC_S_CTRL with ctrl.id=" + name + " failed)");
    }
  }

  unsigned int IVTVReadNode::getControl(unsigned int id, const string& name) const {
    struct v4l2_control ctrl;
    ctrl.id = id;
    if (ioctl(__fd, VIDIOC_G_CTRL, &ctrl) == 0) {
      return id;
    } else {
      throw Exception("Error getting control value (VIDIOC_G_CTRL with ctrl.id=" + name + " failed)");
    }
  }

  void IVTVReadNode::setFramerate(const unsigned int& value) {
    struct ivtv_ioctl_codec in_codec;
    if (ioctl(__fd, IVTV_IOC_G_CODEC, &in_codec) < 0)
      throw Exception("Error getting Codec Params");
    else {
      in_codec.framerate = value;
      if (ioctl(__fd, IVTV_IOC_S_CODEC, &in_codec) < 0)
        throw Exception("Error setting Codec Param framerate");
    }
  }

  void IVTVReadNode::setAspect(const unsigned int& value) {
    struct ivtv_ioctl_codec in_codec;
    if (ioctl(__fd, IVTV_IOC_G_CODEC, &in_codec) < 0)
      throw Exception("Error getting Codec Params");
    else {
      in_codec.aspect = value;
      if (ioctl(__fd, IVTV_IOC_S_CODEC, &in_codec) < 0)
        throw Exception("Error setting Codec Param aspect");
    }
  }
  
  void IVTVReadNode::initFrequencyTables()
    throw (const Exception&)
  {
    string filename;
    
    /* Default filename */
    filename = getNMMResource("ivtv/tuner.xml");
    MESSAGE_STREAM("Parsing frequency file " << filename << endl);
    
    /* Begin> Parse XML file */
    XMLTree* xml = new XMLTree( filename );
    XMLNode* root = xml->root();
    if (!root) {
      throw Exception("Failed to parse XML file " + filename);
    }
    if (root->name() != "frequencies") {
      throw Exception("XMl file " + filename + " is not a frequency table file");
    }
    /* End< */
    
    /* Begin> Parse channel lists */
    XMLNodeList nlist_chanlist(root->children("chanlist"));
    for (XMLNodeIterator i(nlist_chanlist.begin()); i != nlist_chanlist.end(); ++i) {
      
      /* Get name of channel list */
      XMLProperty* prop_chanlist_name((*i)->property("name"));
      if (!prop_chanlist_name) {
        throw Exception("Error in frequency table: Channel list has no name");
      }
      
      /* Create new frequency table */
      FrequencyTable table(prop_chanlist_name->value());
      
      /* Begin> Parse sub nodes of channel list */
      XMLNodeList nlist((*i)->children());
      for (XMLNodeIterator j(nlist.begin()); j != nlist.end(); ++j) {
        
        /* Begin> Parse channel entry */
        if ((*j)->name() == "channel") {
          
          /* Get channel name */
          XMLProperty* prop_name((*j)->property("name"));
          if (!prop_name) {
            throw Exception("Error in frequency table: Channel has no name");
          }
          
          /* Get channel frequency */
          XMLProperty* prop_freq((*j)->property("frequency"));
          if (!prop_freq) {
            throw Exception("Error in frequency table: Channel has no frequency");
          }
          
          /* Add channel entry */
          //          for(int i = 0; i < ivtv.size(); i++){
          //  if(prop_name->value() == string(ivtv.getChannel(i).frequency)){
          //    table.addChannel(string(ivtv.getChannel(i).name), atoi(prop_freq->value().c_str()));
          table.addChannel(prop_name->value(), atoi(prop_freq->value().c_str()));
          //     break;
          //}
          //}
        }
        /* End< */
        
        /* Begin> Parse include entry */
        else if ((*j)->name() == "include") {
          
          /* Get include  name */
          XMLProperty* prop_name((*j)->property("name"));
          if (!prop_name) {
            throw Exception("Error in frequency table: Include has no name");
          }
          
          /* Check if that channel list exists */
          map<string, FrequencyTable>::const_iterator found(__frequencies.find(prop_name->value()));
          if (found == __frequencies.end()) {
            throw Exception("Error in frequency table: Include unknown frequency table " + prop_name->value());
          }
          
          /* Add channels from that frequency table to current */
          for (unsigned int i=0; i<found->second.getChannelCount(); ++i) {
            table.addChannel(found->second.getChannelName(i), found->second.getChannelFrequency(i));
          }
          
        }
        /* End< */
        
      }
      /* End< Parse sub nodes of channel list */
      
      /* Check if table is non-empty */
      /*
      if (table.getChannelCount() == 0) {
        throw Exception("Channel list " + prop_chanlist_name->value() + " is empty");
      }
      */
      /* Add table to map */
      DEBUG_STREAM("Adding channel list " << prop_chanlist_name->value() << endl);
      __frequencies[prop_chanlist_name->value()] = table;
    }
    if(__frequencies.empty())
      throw Exception("no Frequency-Tables available");
  }
  /*============================== Helper Methods ==============================*/
  
  /* *** IChannelConfig-Interface *** */

  TVChannelList IVTVReadNode::getChannelList(){
    channels.clear();
    loadChannelconfig();
    return channels;
  }

  void IVTVReadNode::setChannelConfigFile(const string &filename){
    channelConfigFile = filename;
    channels.clear();
    loadChannelconfig();
  }
  void IVTVReadNode::loadChannelconfig(){
    ifstream configfile(channelConfigFile.c_str());
    char line[41];

    if (!configfile) {
      GLOBAL_ERROR_STREAM("Channel file " << channelConfigFile << " not found" << endl);
      return;
    }

    int channelCounter = 0;

    while (!configfile.eof()) {
      configfile.getline(line, 40);
      if(line){
        regex_t exp;
        regmatch_t pm;

        if (!regcomp(&exp, "\\[.*?\\]", REG_EXTENDED)){
          if(regexec(&exp, line , 1 , &pm, 0) != REG_NOMATCH){
            //wich tag do we have?
            string res(line);

            if(res.substr(pm.rm_so+1, pm.rm_eo - pm.rm_so -2) == "defaults") {
              //if we are in defaults-section
              continue;
            }else if( res.substr(pm.rm_so+1, pm.rm_eo - pm.rm_so-2) == "global"){
              //if we are in global-section
              continue;
            }else{
              //ok, we have found a channel
              char name[20];
              name[0]='\0';
              strcat(name, res.substr(pm.rm_so+1, pm.rm_eo - pm.rm_so-2).c_str()) ;
              name[pm.rm_eo - pm.rm_so-2] = '\0';
              TVChannel channel;
              string channelName(name);
              channel.setChannelName(channelName);
              channel.setChannelNumber(channelCounter);
              channelCounter++;
              channel.readAliases();
              channels.addChannel(channel);
            }
          }//end if(regexec)
        }//end if(regcomp)
      }//end if(line)
    }//end while(!eof)
  }
}

#include "nmm/nmm_suffix.hpp"
