/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#ifndef _NMM_IVTVREADNODE_HPP_
#define _NMM_IVTVREADNODE_HPP_ _NMM_IVTVREADNODE_HPP_

#include "nmm/nmm_prefix.hpp"

#include "nmm/multimedia/GenericSourceNode.hpp"

#include "nmm/multimedia/IDeviceImpl.hpp"
#include "nmm/plugins/interfaces/IBufferSizeImpl.hpp"
#include "nmm/plugins/video/tv/ivtv/IIVTVImpl.hpp"
#include "nmm/plugins/video/tv/interfaces/ITVCardImpl.hpp"
#include "nmm/plugins/video/tv/interfaces/IChannelInfoImpl.hpp"
#include "nmm/plugins/video/tv/utils/ChannelConfig.hpp"

#include "nmm/multimedia/sync/StreamTimer.hpp"

#include "nmm/plugins/video/tv/ivtv/FrequencyTable.hpp"


namespace NMM {
  
  /** IVTV Read Node.
   */
  class IVTVReadNode :
    public GenericSourceNode,
    public IBufferSizeImpl,
    public ITVCardImpl,
    public IIVTVImpl,
    public IChannelInfoImpl
  {
  public:
    /** @name Constructor & Destductor
     */
    //@{
    /** Constructor.
     */
    IVTVReadNode(const char *name = "IVTVReadNode",
                 StreamQueue::Mode = StreamQueue::MODE_SUSPEND,
                 int = StreamQueue::DEFAULT_SIZE);
    
    /** Virtual destructor.
     */
    virtual ~IVTVReadNode();
    //@}
    
  protected:
    /** @name GenericNode Methods
     */
    //@{
    virtual Result doInit();
    virtual Result doDeinit();
    virtual Result doInitOutput();
    virtual Result doActivate();
    
    virtual Message* processBuffer(Buffer*);
    //@}
    
    /** @name IBufferSize interface
     */
    //@{
    /** Set buffer size.
     */
    virtual void setBufferSize(const nmm_int32& value);
    
    /** Get buffer size.
     */
    virtual nmm_int32 getBufferSize();
    //@}
    
    /** @name ITVCard Interface
     */
    //@{
    /** Set device.
     */
    virtual void setDevice(const string& value) throw(const DeviceNotFoundException&, const WrongDeviceException&);
    
    /** Get device.
     */
    virtual string getDevice() const;
    
    /** Set resolution.
     */
    virtual void setResolution(const nmm_int32& x, const nmm_int32& y) throw(const Exception&);

    /** Get resolution.
     */
    virtual void getResolution(nmm_int32& x, nmm_int32& y) throw(const Exception&);
    
    /** Set channel.
     */
    virtual Result setChannel(const nmm_int32& channel);
    
    /** Get channel.
     */
    virtual Result getChannel(nmm_int32& channel, string& name);
    //@}
    
    
    /** @name Audio Inputs
     */
    //@{
    /** This method returns the number of available inputs.
     */
    virtual nmm_uint32 getAudioCount() const;

    /** This method returns the properties of the specified input.
     * @param index Audio index, ranges from 0 to getAudioCount()-1
     */
    virtual TVAudio getAudio(const nmm_uint32& index) const;
    
    /** This method returns the list of available inputs.
     */
    virtual list<TVAudio> getAudioList() const;
    
    /** This method returns the current input.
     */
    virtual nmm_uint32 getCurrentAudio() const;
    
    /** This method sets the current input.
     * @param index Audio index, ranges from 0 to getAudioCount()-1
     */
    virtual void setCurrentAudio(const nmm_uint32& index);
    //@}
    
    
    /** @name Inputs
     */
    //@{
    /** This method returns the number of available inputs.
     */
    virtual nmm_uint32 getInputCount() const;

    /** This method returns the properties of the specified input.
     * @param num Input number, ranges from 0 to getInputCount()-1
     */
    virtual TVInput getInput(const nmm_uint32& num) const;
    
    /** This method returns the list of available inputs.
     */
    virtual list<TVInput> getInputList() const;
    
    /** This method returns the current input.
     */
    virtual nmm_uint32 getCurrentInput() const;

    /** This method sets the current input.
     * @param num Input number, ranges from 0 to getInputCount()-1
     */
    virtual void setCurrentInput(const nmm_uint32& num);
    //@}
    
    
    /** @name Standards
     */
    //@{
    /** This method returns the identifier of the standard which matches the given name.
     * If no such standard is found, zero is returned.
     */
    virtual nmm_uint32 getStandardByName(const string& name) const;
    
    /** This method returns the properties of the specified standard.
     * @param num Standard number
     */
    virtual TVStandard getStandard(const nmm_uint32& num) const;
    
    /** This method returns the list of supported standards.
     */
    virtual list<TVStandard> getStandardList() const;
    
    /** This method returns the standard set for the current input.
     */
    virtual nmm_uint32 getCurrentStandard() const;
    
    /** This method sets the standard for the current input
     * @param num Standard number, ranges from 0 to getStandardCount()-1
     */
    virtual void setCurrentStandard(const nmm_uint32& standard);
    
    /** This method sets the standard for the current input by name.
     */
    virtual void setStandardByName(const string& name);
    //@}
    
    /** @name Tuners
     */
    //@{
    /** This method returns the number of tuners.
     */
    virtual nmm_uint32 getTunerCount() const;
    
    /** This method returns information about a tuner.
     * @param index Index of tuner, ranges from 0 to getTunerCount()-1
     */
    virtual TVTuner getTuner(const nmm_uint32& index) const;
    
    /** This method returns the list of tuners.
     */
    virtual list<TVTuner> getTunerList() const;
    
    /** This method returns the current frequency of a tuner.
     * @param tuner Number of the tuner
     * @return Current frequency of the tuner in MHz
     */
    virtual nmm_float getFrequency(const nmm_uint32& tuner) const;
    
    /** This method sets the frequency of a tuner.
     * @param tuner Number of the tuner
     * @param frequency The frequency to set in MHz
     */
    virtual void setFrequency(const nmm_uint32& tuner, const nmm_float& frequency);
    //@}
    
    /** @name Codecs
     */
    //@{
    /** This method returns the codec parameters.
     */
    virtual IVTVCodec getCodecParams() const;
    
    /** This method sets the codec parameters.
     */
    virtual void setCodecParams(const IVTVCodec& codec);

    /** This method sets the GOP size
     */
    virtual void setGOPSize(const nmm_uint32& value);

    /** This method sets the number of B-Frames
     */
    virtual void setBFrames(const nmm_uint32& value);

    /** This method sets the GOP closure mode of the codec (1=close GOP, 0=open GOP)
     */
    virtual void setGOPClosure(const nmm_uint32& value);

    /** This method sets inverse telecine (1=ON, 0=OFF)
     */
    virtual void setPulldown(const nmm_uint32& value);

    /** This method sets bitrate
     */
    virtual void setBitrate(const nmm_uint32& value);

    /** This method sets peak bitrate (must be greater than bitrate)
     */
    virtual void setPeakBitrate(const nmm_uint32& value);

    /** This method sets the Streamtype
     *   IVTV_STREAM_PS     = 0
     *   IVTV_STREAM_TS     = 1
     *   IVTV_STREAM_MPEG1  = 2
     *   IVTV_STREAM_PES_AV = 3
     *   IVTV_STREAM_PES_V  = 5
     *   IVTV_STREAM_PES_A  = 7
     *   IVTV_STREAM_DVD    = 10
    */
    virtual void setStreamType(const nmm_uint32& value);

    virtual void setBitrateMode(const nmm_uint32& value);
    virtual void setDNRMode(const nmm_uint32& value);
    virtual void setDNRType(const nmm_uint32& value);
    virtual void setDNRSpatial(const nmm_uint32& value);
    virtual void setDNRTemporal(const nmm_uint32& value);
    virtual void setAspect(const nmm_uint32& value);
    //@}
    
    /** @name Controls.
     */
    //@{
    /** This method gets the current brightness.
     */
    virtual nmm_uint32 getBrightness() const;
    
    /** This method gets the current contrast.
     */
    virtual nmm_uint32 getContrast() const;
    
    /** This method gets the current saturation.
     */
    virtual nmm_uint32 getSaturation() const;
    
    /** This method gets the current hue.
     */
    virtual nmm_int32 getHue() const;
    
    /** This method gets the current volume.
     */
    virtual nmm_uint32 getVolume() const;
    
    /** This method gets the mute status.
     */
    virtual nmm_bool getMute() const;
    
    /** This method sets the brightness.
     * @param brightness Brightness (0 to 255)
     */
    virtual void setBrightness(const nmm_uint32& brightness);
    
    /** This method sets the contrast.
     * @param contrast Contrast (0 to 127)
     */
    virtual void setContrast(const nmm_uint32& contrast);
    
    /** This method sets the saturation.
     * @param saturation Saturation (0 to 127)
     */
    virtual void setSaturation(const nmm_uint32& saturation);
    
    /** This method sets the hue.
     * @param hue Hue offset (-128 to 127)
     */
    virtual void setHue(const nmm_int32& hue);
    
    /** This method sets the volume.
     * @param volume Volume (0 to 65535)
     */
    virtual void setVolume(const nmm_uint32& volume);
    
    /** This method sets the mute state.
     */
    virtual void setMute(const nmm_bool& mute);
    //@}
    
    /** @name Frequency tables
     */
    //@{
    /** This method returns the available frequency tables.
     */
    virtual list<string> getFrequencyTableList() const;
    
    /** This method returns the current frequency table.
     */
    virtual string getFrequencyTable() const;
    
    /** This method sets the current frequency table.
     * @param name Name of the frequency table
     */
    virtual void setFrequencyTable(const string& table);

    /**This method sets the input/output-value in MSP-matrix
     * @param input input-value
     * @param output output-value
     */
    virtual void setInputOutput(const nmm_int32& input, const nmm_int32& output);

  private:
    /** This method sets the value of a control.
     */
    void setControl(unsigned int id, unsigned int value, const string& name);
    
    /** This method returns the value of a control.
     */
    unsigned int getControl(unsigned int id, const string& name) const;

    /** This method sets the Framerate
     */
    void setFramerate(const unsigned int& value);

    /** This method initializes the frequency tables.
     */
    void initFrequencyTables()
      throw (const Exception&);

    //IChannelInfo interface
    virtual Result hasChannel(nmm_int32& , const string& ){return FAILURE;}
    virtual TVChannelList getChannelList();
    virtual void setChannelConfigFile(const string &filename);

  private:
    TVChannelList channels;
    string channelConfigFile;
    void loadChannelconfig();

  private:
    typedef vector<TVInput>                 Inputs;
    typedef Inputs::iterator                InputIterator;
    typedef Inputs::const_iterator          InputConstIterator;
    
    typedef vector<TVAudio>                 AudioInputs;
    typedef AudioInputs::iterator           AudioInputIterator;
    typedef AudioInputs::const_iterator     AudioInputConstIterator;
    
    typedef vector<TVTuner>                 Tuners;
    typedef Tuners::iterator                TunerIterator;
    typedef Tuners::const_iterator          TunerConstIterator;
    
    typedef map<unsigned int, TVStandard>   Standards;
    typedef Standards::iterator             StandardIterator;
    typedef Standards::const_iterator       StandardConstIterator;
    
    /* Begin> Device */
    string                      __device;            ///< Device filename
    int                         __fd;                ///< File descriptor to read from device
    /* End< */
    
    /* Begin> Information */
    AudioInputs                 __audio_inputs;      ///< Available audio inputs
    Inputs                      __inputs;            ///< Available inputs
    Tuners                      __tuners;            ///< Available tuners
    Standards                   __standards;         ///< Supported standards
    /* End< */
    
    /* Begin> Output */
    ssize_t                     __buffer_size;       ///< Read buffer size
    bool                        __do_flush;          ///< Need to flush the stream
    StreamTimer                 __stream_timer;      ///< Stream timer
    /* End< */
    
    /* Begin> Channels */
    unsigned int                __cur_channel;       ///< Current channel
    string                      __cur_freqtable;     ///< Current frequency table
    map<string, FrequencyTable> __frequencies;       ///< Frequency tables
    /* End< */

    ChannelConfig ivtv;

    struct ivtv_ioctl_framesync {
      uint32_t frame;
      uint64_t pts;
      uint64_t scr;
    };
    
    Time __system_start_time;
    uint64_t __scr_start;
    
    uint64_t __buffer_counter;
    bool __received_first_buffer;
 
  };
  
}

#include "nmm/nmm_suffix.hpp"

#endif
