/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

 
#include <errno.h>

#include "TVCardReadNode.hpp"
#include "nmm/utils/SecurityCheck.hpp"
#include "nmm/utils/NMMConfig.hpp"
#include "nmm/utils/directory_utils.hpp"
#include "nmm/utils/time/Rational.hpp"
#include "nmm/utils/xml/XMLTree.hpp"
#include "nmm/utils/xml/XMLNode.hpp"
#include "nmm/comm/messaging/CEvent.hpp"
#include "nmm/multimedia/FlushEvent.hpp"
#include "nmm/multimedia/TDevicePlugin.hpp"
#include "nmm/multimedia/ISyncReset.hpp"

#include "nmm/utils/thread/MutexGuard.hpp"


#include <fcntl.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <linux/types.h>



static NMM::TDevicePlugin<NMM::TVCardReadNode> plugin("/dev/video", false, "TVCardReadNode");
NMM_REGISTER_SPECIAL_PLUGIN(TVCardReadNode, NMM, plugin)

namespace NMM {

  TVCardReadNode::TVCardReadNode(const char *name /* = "TVCardReadNode" */,
				 StreamQueue::Mode q_mode /* = StreamQueue::MODE_SUSPEND */,
				 int q_size /* = StreamQueue::DEFAULT_SIZE */)
	:GenericSyncSourceNode(name, q_mode, q_size),
	_framesize(0),
	_frame_rate_base(1000),
	__device("/dev/video0"),
	__fd(-1),
	__frame_rate(25),
	__buffer_size(720*576),
	__do_flush(false),
	__stream_timer(this, StreamTimer::REAL_TIME),
	__cur_channel(0),
	buffers(NULL)
  {

    /* set default values */
    setSyncMaster(false);

    // this node produces at a constant rate
    setConstantRateProducer(true);

    
    //next two have to die - they set the image size but NOT the canvas size. 
    width = 320;
    height = 240;

    int_frame_rate = static_cast<int>(__frame_rate * _frame_rate_base);

    NMMConfig device;
    
    const char* home(getenv("HOME"));
    string nmmrcfile(string(home) + string("/.nmmrc"));
    device.parseFile(nmmrcfile);
     
    if(!device.getValue("tv_device", __device))
      __device = "/dev/video0";
     
    int input=0;
    if(!device.getValue("tv_input_channel",input))
      __cur_channel = 0;
    else
      __cur_channel = input;
  
    //next lines have to die
    memset(&_capability,  0, sizeof(video_capability));
    memset(&_vid_buffer,  0, sizeof(video_buffer));
    memset(&_vid_mbuf,    0, sizeof(video_mbuf));
    memset(&_vid_mmap,    0, sizeof(video_mmap));
    memset(&_vid_picture, 0, sizeof(video_picture));
    memset(&_vid_win,     0, sizeof(video_window));
    memset(&_vid_channel, 0, sizeof(video_channel));
    memset(&_vid_tuner,   0, sizeof(video_tuner));

    _frame = 0;
    grab_format = VIDEO_PALETTE_YUV420P;

    _vid_mmap = video_mmap();
    _vid_mmap.frame = _frame;
    _vid_mmap.width = width;
    _vid_mmap.height = height; 
  }

  TVCardReadNode::~TVCardReadNode() 
  {   
    if(__fd >= 0) 
    {
      DEBUG_STREAM("Closing tv device" << std::endl);
      ::close(__fd);
      __fd = -1;
    }
  }

  /*============================== IBufferSize Interface ==============================*/
  void TVCardReadNode::setBufferSize(const int& value) {
    __buffer_size = value;
  }

  int TVCardReadNode::getBufferSize() {
    return __buffer_size;
  }
  /*============================== IBufferSize Interface ==============================*/


  /* Event functions */
  void TVCardReadNode::setDevice(const string& device) throw(const DeviceNotFoundException&) 
  {
    try 
    {
      /* to test if the __device can be accessed, we try to open the __device */
      if(!SSecurityCheck::getInstance().checkFilenameForWriting(device)) {
        DEBUG_STREAM("Security check not passed for file " << device << ", check .nmmrc for options" << endl);
        throw DeviceNotFoundException(this, "Security check failed for device", device);
      }
     
      DEBUG_STREAM("Probing tv device " << device << std::endl);
      
      __fd = ::open(device.c_str(), 0);
      
      if(__fd < 0) {
        string error(strerror(errno));
        ERROR_STREAM("Could not open device " << device << " (" << error << ")" << std::endl);
        throw DeviceNotFoundException(this, "Failed to open device", device);
      }
      /* query capabilities and check if this is really an ivtv device */
      video_capability vcap;
      if(-1 == ioctl(__fd, VIDIOCGCAP, &vcap)) 
      {
        ERROR_STREAM("Failed to query device capabilities (ioctl: VIDIOC_QUERYCAP failed)" << std::endl);
        ERROR_STREAM("Maybe the device is not a video4linux1 TV card" << std::endl);
        throw DeviceNotFoundException(this, "Failed to query device capabilities", device);
      } 
      else 
      {
        /* print out information */
        DEBUG_STREAM("Driver name         : " << vcap.name << std::endl);
	DEBUG_STREAM("Capability.width    : " << vcap.minwidth  <<
		                  " <= x <= " << vcap.maxwidth  << std::endl);
    	DEBUG_STREAM("Capability.height   : " << vcap.minheight <<
				  " <= y <= " << vcap.maxheight << std::endl);
	DEBUG_STREAM("Capability.channels : " << vcap.channels << std::endl);
        DEBUG_STREAM("Driver type         : " << vcap.type << std::endl);
        
      }

      /* close device */
      DEBUG_STREAM("Closing tv device" << std::endl);
      if(__fd >= 0) {
        ::close(__fd);
        __fd = -1;
      }
   } catch(...) {
      /* close device */
      DEBUG_STREAM("Closing tv device" << std::endl);
      if(__fd >= 0) {
        ::close(__fd);
        __fd = -1;
      }
      
      /* rethrow exception */
      throw;
    }
    
    /* now set the device name */
    __device = device;
  } /* setDevice */

  string TVCardReadNode::getDevice() const {
    return __device;
  } /* getDevice */

  void TVCardReadNode::setSize(const int& /*x*/, const int& /*y*/) {
    // XXX: not implemented 
  }
  
  void TVCardReadNode::getSize(int& /*x*/, int& /*y*/) {
    // XXX: not implemented
  }


  
  Result TVCardReadNode::setChannel(const int &channel) {
    MutexGuard mg_v4l(mutex_v4l);

    DEBUG_STREAM("Setting video channel to " << channel << endl);

    _vid_channel.channel = channel;
    if(-1 == ioctl(__fd, VIDIOCSCHAN, &_vid_channel)) {
      ERROR_STREAM("ERROR : Can\'t ioctl set video channel to " << _vid_channel.channel << endl);
      return FAILURE;
    }
    return SUCCESS;
  } /* TVCardReadNode::setChannel */


    Result TVCardReadNode::getChannel(int &channel, string &name) {
  //to be checked!!!
    if(-1 == ioctl(__fd, VIDIOCGCHAN, &_vid_channel)) {
      ERROR_STREAM("Can\'t get video channel");
      return FAILURE;
    }
    channel = _vid_channel.channel;
    name   = string(_vid_channel.name);

    return SUCCESS;
  }
  
  void TVCardReadNode::setResolution(const int& x, const int& y) {
    width = x;
    height = y;
  } /* getSize */

  void TVCardReadNode::getResolution(int& x, int& y) {
    x =  _vid_win.width;
    y =  _vid_win.height;
  } /*  */

  
  Result TVCardReadNode::setOutputFormat(int &format) {
    switch (format) 
    {
      case 0: grab_format = VIDEO_PALETTE_YUV420P; break;
      case 1: grab_format = VIDEO_PALETTE_RGB32; break;
      default : grab_format = VIDEO_PALETTE_YUV420P;
    }
    

    return SUCCESS;
  } /* setOutputFormat */
  


  /**
   * NMM  functions 
   */
  Result TVCardReadNode::doInit() {
    // set the FrameTime at the Controller
    
    //controller->setFrameTime(Interval(Rational(1, __frame_rate)));

  //  o_format->addFloatParamValue(Format::framerate_param,__frame_rate);

    Format* o_format;



    /* setting up supported formats */
    o_format = getOwnOutputProperty()->addNewFormat("video/raw");
    o_format->addWildcard(Format::x_resolution_param);
    o_format->addWildcard(Format::y_resolution_param);	
    o_format->addWildcard(Format::ratio_x_param);
    o_format->addWildcard(Format::ratio_y_param);
    o_format->addWildcard(Format::framerate_param);
    o_format->addStringParamValue(Format::channelorder_param,"yvu");
    o_format->addStringParamValue(Format::endian_param, "little");
    o_format->addStringParamValue(Format::colorspace_param,"yv12");
    o_format->addIntParamValue(Format::bitperpixel_param, 12);
    o_format->addStringParamValue(Format::format_param, "planar");
  
    getOwnOutputProperty()->setDefaultFormat(o_format);
    
    o_format = getOwnOutputProperty()->addNewFormat("video/raw");
    o_format->addWildcard(Format::x_resolution_param);
    o_format->addWildcard(Format::y_resolution_param);
    o_format->addWildcard(Format::ratio_x_param);
    o_format->addWildcard(Format::ratio_y_param);
    o_format->addFloatParamValue(Format::framerate_param, __frame_rate);
    o_format->addStringParamValue(Format::channelorder_param,"rgba");
    o_format->addStringParamValue(Format::endian_param, "little");
    o_format->addStringParamValue(Format::colorspace_param,"rgba32");
    o_format->addIntParamValue(Format::bitperpixel_param, 32);
    o_format->addStringParamValue(Format::format_param, "planar");
  

    string nmm_v4l_conf = "v4l-conf > /dev/null 2>&1";
#ifdef __linux__
    switch (system(nmm_v4l_conf.c_str())) {
    case -1: /* can't run */
      ERROR_STREAM("could'nt start v4l-conf - make sure you have v4l-conf installed!" << endl);
      break;
    case 0: /* ok */
      DEBUG_STREAM("v4l-conf ok." << endl);
      break;
    default: /* non-zero return */
      ERROR_STREAM("v4l-conf had some trouble, trying to continue anyway" << endl);
      break;
    }
#endif

    /* try to open the device */
    if (-1 != __fd) {
      DEBUG_STREAM("closing device_fd before opening it..." << endl);
      close(__fd);
    }
    if(!SSecurityCheck::getInstance().checkFilenameForWriting(__device)) {
      DEBUG_STREAM("Security check not passed for file " << __device << ", check .nmmrc for options" << endl);
      return FAILURE;
    }
    __fd = open(__device.c_str(), O_RDWR);
    if(__fd < 0) {
      ERROR_STREAM("Error: Can\'t open " << __device << " for reading." << endl);
      return FAILURE;
    }
    DEBUG_STREAM("Open device " << __device << endl);


    /**
     * get caps 
     */
    if(-1 == ioctl(__fd, VIDIOCGCAP, &_capability)) {
      ERROR_STREAM("Error: Can\'t get video capability");
      close (__fd);
      return FAILURE;
    }
    DEBUG_STREAM("  Capability.name     : " << _capability.name << endl);
    DEBUG_STREAM("  Capability.width    : " << _capability.minwidth  << " <= x <= " << _capability.maxwidth  << endl);
    DEBUG_STREAM("  Capability.height   : " << _capability.minheight << " <= y <= " << _capability.maxheight << endl);
    DEBUG_STREAM("  Capability.channels : " << _capability.channels << endl);


    /** 
     * set channel aka input
     */
    // to set a channel, you first have to do a get-channel ioctl:
    // have to make sure that _vid_channel is lower than _capability.channels here
    // (strange, but see bttv-driver.c in linux-2.4.19)
    if(-1 == ioctl(__fd, VIDIOCGCHAN, &_vid_channel)) {
      ERROR_STREAM("Can\'t get video channel");
      // you could not set a channel, so return FAILURE
      return FAILURE;
    }
    setChannel(0); 

    if (_capability.type & VID_TYPE_TUNER) {
      if(-1 == ioctl(__fd, VIDIOCGTUNER, &_vid_tuner)) {
	ERROR_STREAM("Error: Can\'t get video capability");
	close (__fd);
	return FAILURE;
      }
    }

    return SUCCESS;
  }

  Result TVCardReadNode::doDeinit() {
    DEBUG_STREAM("doDeinit..." << endl);
    if (__fd != -1) {
      DEBUG_STREAM("Closing device_fd" << endl);
      close (__fd);
      __fd = -1;
    }
    return SUCCESS;
  }

  
  
  
  Result TVCardReadNode::doInitOutput() 
  {
    Format* o_format;
       
    if (grab_format == VIDEO_PALETTE_YUV420P) {
      o_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
      o_format->addWildcard(Format::x_resolution_param);
      o_format->addWildcard(Format::y_resolution_param);
      o_format->addWildcard(Format::ratio_x_param);
      o_format->addWildcard(Format::ratio_y_param);
      o_format->addWildcard(Format::framerate_param);
      o_format->addStringParamValue(Format::channelorder_param,"yvu");
      o_format->addStringParamValue(Format::endian_param, "little");
      o_format->addStringParamValue(Format::colorspace_param,"yv12");
      o_format->addIntParamValue(Format::bitperpixel_param, 12);
      o_format->addStringParamValue(Format::format_param, "planar");
      
      getOwnWorkingOutputProperty()->setDefaultFormat(o_format);
    } else {
      o_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
      o_format->addIntParamValue(Format::x_resolution_param,width);
      o_format->addIntParamValue(Format::y_resolution_param,height);
      o_format->addFloatParamValue(Format::framerate_param, __frame_rate);
      o_format->addStringParamValue(Format::channelorder_param,"rgba");
      o_format->addStringParamValue(Format::endian_param, "little");
      o_format->addStringParamValue(Format::colorspace_param,"rgba32");
      o_format->addIntParamValue(Format::bitperpixel_param, 32);
      o_format->addStringParamValue(Format::format_param, "planar");
      
      getOwnWorkingOutputProperty()->setDefaultFormat(o_format);
    }
    return SUCCESS;
  }
      


  Result TVCardReadNode::doActivate() {
    int    zero   = 0;
    string colorspace = "";

    const Format* o_format = getOutputFormat();
    if (NULL == o_format) {
      ERROR_STREAM("out_format is NULL!!" << endl);
      return FAILURE;
    }

    try
    {//if(fmt->getIntValue(Format::bitperpixel_param)==16)
    width = o_format->getIntValue(Format::x_resolution_param);
    height = o_format->getIntValue(Format::y_resolution_param);
    __frame_rate = o_format->getFloatValue(Format::framerate_param);
    }
    catch(...)
    {
    cout << "got shit 'mon" << endl;
    }
    /**
     * clamp width and height 
     */
    width = (width > _capability.maxwidth ? _capability.maxwidth :
             (width < _capability.minwidth ? _capability.minwidth : width));
    height = (height > _capability.maxheight ? _capability.maxheight :
              (height < _capability.minheight ? _capability.minheight : height));
    DEBUG_STREAM("Try to grab with resolution " << width << "x" << height << endl);


    /**
     * get/set picture properties 
     */
    if(-1 == ioctl(__fd, VIDIOCGPICT, &_vid_picture)) {
      ERROR_STREAM("Error: Can\'t get picture properties" << endl);
      close (__fd);
      return FAILURE;
    }
  
    /**
     * choose colorspace for grabbing
     */
    colorspace = o_format->getStringValue(Format::colorspace_param);
    if ("yv12" == colorspace) {
      grab_format = VIDEO_PALETTE_YUV420P;
    } else if ("rgba32" == colorspace) {
      grab_format = VIDEO_PALETTE_RGB32;
    }
    _vid_picture.palette = grab_format;
    if(-1 == ioctl(__fd, VIDIOCSPICT, &_vid_picture)) { 
      ERROR_STREAM("ERROR: Can'\t set picture properties!" << endl);
      close (__fd);
      return FAILURE;
    }


    /*
     * get video_mbuf for mmapping
     */
    if(-1 == ioctl(__fd, VIDIOCGMBUF, &_vid_mbuf)) {
      ERROR_STREAM("ERROR: Can'\t get mbuf!" << endl);
      close (__fd);
      return FAILURE;
    }
    _mmap_buffer = (unsigned char*) mmap(0, _vid_mbuf.size,
                                         PROT_READ | PROT_WRITE,
                                         MAP_SHARED, __fd, 0);
    if(_mmap_buffer < 0) {
      ERROR_STREAM("ERROR: Can\'t get mmap!" << endl);
      close (__fd);
      return FAILURE;
    }
      
    if (-1 == ioctl(__fd, VIDIOCGWIN, &_vid_win)) {
      ERROR_STREAM("ERROR: Can\'t get overlay window!" << endl);
      close (__fd);
      return FAILURE;
    }

    _vid_win.width = width;
    _vid_win.height = height;

    if(-1 == ioctl(__fd, VIDIOCSWIN, &_vid_win)) {
      ERROR_STREAM("ERROR: VIDIOCSWIN failed!" << endl);
      //        close (__fd);
      //        return FAILURE;
    }

    DEBUG_STREAM("_vid.win : " 
                 << _vid_win.x << " " << _vid_win.y << " " 
                 << _vid_win.width << " " << _vid_win.height << endl);

    if (-1 == ioctl(__fd, VIDIOCGWIN, &_vid_win)) {
      ERROR_STREAM("ERROR: VIDIOCGWIN (2) failed!" << endl);
      close (__fd);
      return FAILURE;
    } else {
      DEBUG_STREAM("_vid.win : " 
                   << _vid_win.x << " " << _vid_win.y << " " 
                   << _vid_win.width << " " << _vid_win.height << endl);
      DEBUG_STREAM("Got resolution " << _vid_win.width << "x" << _vid_win.height << endl);
    }
 

    /**
     * pre-capture
     */
    if(-1 == ioctl(__fd, VIDIOCCAPTURE, &zero)) {
      ERROR_STREAM("ERROR: Can't set capturing!" << endl);
      close (__fd);
      return FAILURE;
    }
    
    /**
     * init mmapped frames
     */
     
     DEBUG_STREAM("width and height have been set as  " << width << " by " << height << endl);
     
    _vid_mmap.width  = width;
    _vid_mmap.height = height;
    _vid_mmap.format = _vid_picture.palette;
    
    _vid_win.width = width;
    _vid_win.height = height;

    picture_size = _vid_win.width * _vid_win.height; 
    DEBUG_STREAM("picture size to a whopping " << picture_size << endl);
    offset1   = picture_size;
    offset2   = (offset1 * 10) / 8; //was 8 statt 1        
    uv_size   = offset1 / 4;

    switch (grab_format) {
    case VIDEO_PALETTE_YUV420P:
//       DEBUG_STREAM("offset1: " << offset1 << " and as a newly converted " << (offset1 * 15) / 10 << endl);
      _framesize = (offset1 * 15) / 10; //hier war vorher 10 statt 1
      DEBUG_STREAM("-----------------set framesize to " << _framesize << endl);
      break;
    case VIDEO_PALETTE_RGB32: 
      _framesize = offset1 * 4;
      break;
    }
    
    // reset the stream __stream_timer
    __stream_timer.reset();

    return SUCCESS;
  }

  Time TVCardReadNode::getLocalTime(u_int64_t valid_read_bytes) {
    // calcluate Rational exactely as
    // valid_read_bytes * _frame_rate_base / int_frame_rate; _framesize
    // to avoid overflow (_framesize * int_frame_rate would create one)
    return Time(Rational(valid_read_bytes * _frame_rate_base / int_frame_rate, _framesize));
  }  
  
  Result TVCardReadNode::doDeinitOutput()
  {

    return SUCCESS;

  } 
  
  
  void TVCardReadNode::setInputOutput(const int& /*input*/, const int& /*output*/){
    //Not implemented

  }
  
    
  Result TVCardReadNode::doDeactivate()
  {
    return SUCCESS;
  }
  
  Message* TVCardReadNode::produceBuffer(Buffer *) {
    Buffer *out;
    char* target;
    char* source;

    DEBUG_STREAM("Grab with framesize " << _framesize << endl);

    out = getNewBuffer(_framesize);

    if ( out ) {
      MutexGuard mg_v4l(mutex_v4l);

      DEBUG_STREAM("Got buffer, try to grab " << _framesize << endl);

      /* Capture */
      DEBUG_STREAM("Capture frame " << _vid_mmap.frame << " (" << width << "x" << height << ")" << endl);
      if (0 > ioctl(__fd, VIDIOCMCAPTURE, &_vid_mmap)) {
        ERROR_STREAM("ERROR: VIDIOCMCAPTURE frame " << _vid_mmap.frame << endl);
        close (__fd);
        out->release();
        return NULL;
      }
      // switch capture buffer
      _vid_mmap.frame = 1-_frame;
      _frame = 1-_frame;

      /* Sync */
      DEBUG_STREAM("Sync frame" << _frame << endl);
      ioctl(__fd, VIDIOCSYNC, &_frame);

      /* get new frame */
      source = (char*) _mmap_buffer + _vid_mbuf.offsets[_frame];
      target = out->getData();

      switch (grab_format) {
      case VIDEO_PALETTE_YUV420P:  
        /* in order : copy Y part then V and U values */
        memcpy(target, source, offset1);
        memcpy(target + offset2, source + offset1, uv_size );
        memcpy(target + offset1, source + offset2, uv_size );
        out->setUsed(_framesize);        
        break;
      case VIDEO_PALETTE_RGB32: 
        DEBUG_STREAM("Copy RGB" << endl);    
        /* all in right order, only one copy step needed */
        memcpy(out->getData(), source, _framesize);
        out->setUsed(_framesize);
        break;
      }

      // Set the correct time stamp for the buffer
      __stream_timer.setRate(__frame_rate);
      __stream_timer.setTimestamp(out);

      return out;
    } 

    return NULL;
  } /* produceBuffer */

 
  
} /* namespace */
