/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

 /* This source code was developed by Andrés Otón Urbano in 2004 to 2005.
  * Copyright and Rights of Use were assigned to Motama GmbH by Andrés in
  * 2007.
  */

#include <errno.h>

#include "TVCardReadNode2.hpp"

#include "nmm/utils/SecurityCheck.hpp"
#include "nmm/utils/NMMConfig.hpp"
#include "nmm/utils/directory_utils.hpp"
#include "nmm/utils/time/Rational.hpp"
#include "nmm/utils/xml/XMLTree.hpp"
#include "nmm/utils/xml/XMLNode.hpp"
#include "nmm/comm/messaging/CEvent.hpp"
#include "nmm/multimedia/FlushEvent.hpp"
#include "nmm/multimedia/TDevicePlugin.hpp"
#include "nmm/multimedia/ISyncReset.hpp"

#include <errno.h>
#include <iomanip>

#define CLEAR(x) memset (&(x), 0, sizeof (x))

#ifndef __CYGWIN__
static NMM::TDevicePlugin<NMM::TVCardReadNode2> plugin("/dev/video", false, "TVCardReadNode2");
NMM_REGISTER_SPECIAL_PLUGIN(TVCardReadNode2, NMM, plugin)
#endif
#ifdef __CYGWIN__
NMM_REGISTER_PLUGIN(TVCardReadNode2, NMM, "TVCardReadNode2")
#endif


#include "nmm/nmm_prefix.hpp"

namespace NMM {

  /*============================== Constructor & Destructor ==============================*/
  TVCardReadNode2::TVCardReadNode2(const char *name /*= "TVCardReadNode2"*/,
                                   StreamQueue::Mode q_mode /*= StreamQueue::MODE_SUSPEND*/,
                                   int q_size /*= StreamQueue::DEFAULT_SIZE*/)
    : GenericSyncSourceNode(name, q_mode, q_size), 
      _framesize(0),
      _frame_rate_base(1000), 
      __device("/dev/video0"),
      __fd(-1),
      __frame_rate(25),
      __buffer_size(720*576),
      __do_flush(false),
      __stream_timer(this, StreamTimer::REAL_TIME),
      __cur_channel(0),
      buffers(NULL)
  {
    /* set default values */
    setSyncMaster(false);
    
    // this node produces at a constant rate
    setConstantRateProducer(true);

    _int_frame_rate = static_cast<int>(__frame_rate * _frame_rate_base);
    
    NMMConfig device;
    
     const char* home(getenv("HOME"));
     string nmmrcfile(string(home) + string("/.nmmrc"));
     device.parseFile(nmmrcfile);
     
     if(!device.getValue("tv2_device", __device))
       __device = "/dev/video0";
     
     int input=0;
     if(!device.getValue("tv2_input_channel",input))
       __cur_channel = 0;
    else
      __cur_channel = input;
  }
  
  TVCardReadNode2::~TVCardReadNode2() {
    /* close device */
    if(__fd >= 0) {
      DEBUG_STREAM("Closing tv device" << std::endl);
      ::close(__fd);
      __fd = -1;
    }
  }
  /*============================== Constructor & Destructor ==============================*/

  /*============================== IBufferSize Interface ==============================*/
  void TVCardReadNode2::setBufferSize(const nmm_int32& value) {
    __buffer_size = value;
  }

  int TVCardReadNode2::getBufferSize() {
    return __buffer_size;
  }
  /*============================== IBufferSize Interface ==============================*/

  
  /*============================== ITVCard Interface ==============================*/
  void TVCardReadNode2::setDevice(const string& device) throw(const DeviceNotFoundException&) {
    /* open device */
    try {
      if(!SSecurityCheck::getInstance().checkFilenameForReading(device)) {
        ERROR_STREAM("Security check not passed for file " << device << ", check .nmmrc for options" << std::endl);
        throw DeviceNotFoundException(this, "Security check failed for device", device);
      }
      DEBUG_STREAM("Probing tv device " << device << std::endl);
      __fd = ::open(device.c_str(), 0);
      if(__fd < 0) {
        string error(strerror(errno));
        ERROR_STREAM("Could not open device " << device << " (" << error << ")" << std::endl);
        throw DeviceNotFoundException(this, "Failed to open device", device);
      }
      
      /* query capabilities and check if this is really an ivtv device */
      v4l2_capability vcap;
      if(ioctl(__fd, VIDIOC_QUERYCAP, &vcap) < 0) {
        ERROR_STREAM("Failed to query device capabilities (ioctl: VIDIOC_QUERYCAP failed)" << std::endl);
        ERROR_STREAM("Maybe the device is not a video4linux2 TV card" << std::endl);
        throw DeviceNotFoundException(this, "Failed to query device capabilities", device);
      } else {
        /* print out information */
        DEBUG_STREAM("Driver name   : " << vcap.driver << std::endl);
        DEBUG_STREAM("Card type     : " << vcap.card << std::endl);
        DEBUG_STREAM("Bus info      : " << vcap.bus_info << std::endl);
        DEBUG_STREAM("Driver version: " << vcap.version << std::endl);
        DEBUG_STREAM("Capabilities  : 0x" << std::setfill('0') << std::setw(8) << std::hex <<
                     vcap.capabilities << std::dec << std::endl);
        
        /* check for STREAMING capability */
	if(!(vcap.capabilities & V4L2_CAP_STREAMING)) {
          ERROR_STREAM("Device is not an tv device with STREAMING support" << std::endl);
          throw DeviceNotFoundException(this, "Device is not an tv device with STREAMING support", device);
        }
      }
      
      /* close device */
      DEBUG_STREAM("Closing tv device" << std::endl);
      if(__fd >= 0) {
        ::close(__fd);
        __fd = -1;
      }
    } catch(...) {
      /* close device */
      DEBUG_STREAM("Closing tv device" << std::endl);
      if(__fd >= 0) {
        ::close(__fd);
        __fd = -1;
      }
      
      /* rethrow exception */
      throw;
    }
    
    /* now set the device name */
    __device = device;
  }
  
  string TVCardReadNode2::getDevice() const {
    return __device;
  }
  
  void TVCardReadNode2::setSize(const nmm_int32& /*x*/, const nmm_int32& /*y*/) {
    // XXX: not implemented 
  }
  
  void TVCardReadNode2::getSize(nmm_int32& /*x*/, nmm_int32& /*y*/) {
    // XXX: not implemented
  }
  
  Result TVCardReadNode2::setChannel(const nmm_int32& channel) {
    try {
      /* Begin> Check if current input is a tuner */
      unsigned int cur_input = getCurrentInput();
      if (cur_input >= __inputs.size()) {
        ERROR_STREAM("Can't set channel (no information about current input)" << endl);
        return FAILURE;
      }
      if (__inputs[cur_input].getType() != TVInput::TYPE_TUNER) {
        ERROR_STREAM("Can't set channel (current input is not a tuner)" << endl);
      }
      unsigned int cur_tuner = __inputs[cur_input].getTuner();
      /* End< */
      /* Begin> Get current frequency table */
      if (__cur_freqtable.empty()) {
        ERROR_STREAM("No frequency table set");
        return FAILURE;
      }
      const FrequencyTable& table = __frequencies[__cur_freqtable];
      /* End< */
      
      /* Begin> Determine frequency */
      if (channel < 0 || channel >= static_cast<int>(table.getChannelCount())) {
        ERROR_STREAM("Channel out of range: " << channel << endl);
        return FAILURE;
      }
      unsigned int frequency = table.getChannelFrequency(channel);
      /* End< */
      
      /* Set frequency at current tuner */
      setFrequency(cur_tuner, static_cast<float>(frequency) / 1000);
      __cur_channel = channel;
      
      return SUCCESS;
    } catch( const Exception& e) {
      ERROR_STREAM("SetChannel failed: " << e.getComment() << endl);
      return FAILURE;
    }
  }
  
  Result TVCardReadNode2::getChannel(nmm_int32& channel, string& name) {
    /* Begin> Get current frequency table */
    if (__cur_freqtable.empty()) {
      ERROR_STREAM("No frequency table set");
      return FAILURE;
    }
    const FrequencyTable& table = __frequencies[__cur_freqtable];
    /* End< */
    
    channel = __cur_channel;
    name = table.getChannelName(channel);
    
    return SUCCESS;
  }

  void TVCardReadNode2::setInputOutput(const nmm_int32& /*input*/, const nmm_int32& /*output*/){
    //Not implemented

  }

  void TVCardReadNode2::getResolution(nmm_int32& x, nmm_int32& y){
    struct v4l2_format format;
    if(-1 == ioctl (__fd, VIDIOC_G_FMT, &format))
      throw Exception(string("I cann't get video format at getResolution"));
    x = format.fmt.pix.width;
    y = format.fmt.pix.height;
  }

  void TVCardReadNode2::setResolution(const nmm_int32& /*x*/, const nmm_int32& /*y*/){
    //Not implemented
  }

  /*============================== ITVCard Interface ==============================*/

  
  /*============================== ITV Interface ==============================*/
  nmm_uint32 TVCardReadNode2::getAudioCount() const {
    return __audio_inputs.size();
  }
  
  TVAudio TVCardReadNode2::getAudio(const nmm_uint32& num) const {
    if (num >= __audio_inputs.size()) {
      throw Exception(string("Invalid audio input: ") + num);
    }
    
    return __audio_inputs[num];
  }
  
  list<TVAudio> TVCardReadNode2::getAudioList() const {
    return list<TVAudio>(__audio_inputs.begin(), __audio_inputs.end());
  }
  
  nmm_uint32 TVCardReadNode2::getCurrentAudio() const {
    struct v4l2_audio audio;
    if (ioctl(__fd, VIDIOC_G_AUDIO, &audio) < 0) {
      throw Exception("Failed to get current audio input (ioctl: VIDIOC_G_AUDIO failed)");
    } else {
      return audio.index;
    }
  }
  
  void TVCardReadNode2::setCurrentAudio(const nmm_uint32& num) {
    struct v4l2_audio audio;
    memset (&audio, 0, sizeof (audio));
    audio.index = num;
    if (ioctl(__fd, VIDIOC_S_AUDIO, &audio) < 0) {
      throw Exception("Failed to set current audio input (ioctl: VIDIOC_S_AUDIO failed)");
    } else {
      DEBUG_STREAM("Set current audio to " << num << endl);
      __do_flush = true;
    }
  }
  
  nmm_uint32 TVCardReadNode2::getInputCount() const {
    return __inputs.size();
  }
  
  TVInput TVCardReadNode2::getInput(const nmm_uint32& num) const {
    if (num >= __inputs.size()) {
      throw Exception(string("Invalid input: ") + num);
    }
    
    return __inputs[num];
  }
  
  list<TVInput> TVCardReadNode2::getInputList() const {
    return list<TVInput>(__inputs.begin(), __inputs.end());
  }
  
  nmm_uint32 TVCardReadNode2::getCurrentInput() const {
    int input;
    if (ioctl(__fd, VIDIOC_G_INPUT, &input) < 0) {
      throw Exception("Failed to get current input (ioctl: VIDIOC_G_INPUT failed)");
    } else {
      return input;
    }
  }
  
  void TVCardReadNode2::setCurrentInput(const nmm_uint32& num) {
    int input = num;
    if (ioctl(__fd, VIDIOC_S_INPUT, &input) < 0) {
      throw Exception("Failed to set current input (ioctl: VIDIOC_S_INPUT failed)");
    } else {
      DEBUG_STREAM("Set current input to " << num << endl);
      __do_flush = true;
    }
  }
  
  nmm_uint32 TVCardReadNode2::getStandardByName(const string& name) const {
    for (StandardConstIterator i(__standards.begin());
         i != __standards.end();
         ++i) {
      if (i->second.getName() == name) {
        return i->first;
      }
    }
    
    return 0;
  }
  
  TVStandard TVCardReadNode2::getStandard(const nmm_uint32& num) const {
    StandardConstIterator found(__standards.find(num));
    if (found == __standards.end()) {
      throw Exception(string("Standard ") + num + " not supported");
    }
    return found->second;
  }
  
  list<TVStandard> TVCardReadNode2::getStandardList() const {
    list<TVStandard> std_list;
    for (StandardConstIterator i(__standards.begin());
         i != __standards.end();
         ++i) {
      std_list.push_back(i->second);
    }
    return std_list;
  }
  
  nmm_uint32 TVCardReadNode2::getCurrentStandard() const {
    int std;
    if (ioctl(__fd, VIDIOC_G_STD, &std) < 0) {
      throw Exception("Failed to get current standard (ioctl: VIDIOC_G_STD failed)");
    }
    return std;
  }
  
  void TVCardReadNode2::setCurrentStandard(const nmm_uint32& standard) {
    int std = standard;
    if (ioctl(__fd, VIDIOC_S_STD, &std) < 0) {
      throw Exception("Failed to set current standard (ioctl: VIDIOC_S_STD failed)");
    } else {
      DEBUG_STREAM("Set standard to " << standard << " for current input" << endl);
      __do_flush = true;
      TVStandard tvstandard = getStandard(std);
      //__frame_rate = tvstandard.getFramePeriodNumerator() / tvstandard.getFramePeriodDenominator();
      if(tvstandard.getFramePeriodNumerator()==1001)
        __frame_rate = 30;
      else
        __frame_rate = 25;

    }
  }
  
  void TVCardReadNode2::setStandardByName(const string& name) {
    unsigned int standard = getStandardByName(name);
    if (standard == 0) {
      throw Exception("Standard " + name + " does not exist");
    }
    setCurrentStandard(standard);
  }
  
  nmm_uint32 TVCardReadNode2::getTunerCount() const {
    return __tuners.size();
  }
  
  TVTuner TVCardReadNode2::getTuner(const nmm_uint32& index) const {
    if (index >= __tuners.size()) {
      throw Exception(string("Invalid tuner index ") + index);
    }
    return __tuners[index];
  }
  
  list<TVTuner> TVCardReadNode2::getTunerList() const {
    return list<TVTuner>(__tuners.begin(), __tuners.end());
  }
  
  float TVCardReadNode2::getFrequency(const nmm_uint32& tuner) const {
    struct v4l2_frequency vf;
    vf.tuner = tuner;
    if (ioctl(__fd, VIDIOC_G_FREQUENCY, &vf) < 0) {
      throw Exception("Failed to get frequency (ioctl: VIDIOC_G_FREQUENCY failed)");
    } else {
      float granularity = 0.0625;
      if (tuner < __tuners.size()) {
        granularity = __tuners[tuner].getGranularity();
      }
      return static_cast<float>(vf.frequency) * granularity;
    }
  }
  
  void TVCardReadNode2::setFrequency(const nmm_uint32& tuner, const nmm_float& frequency) {
    struct v4l2_frequency vf;
    vf.tuner     = tuner;
    
    float granularity = 0.0625;
    if (tuner < __tuners.size()) {
      granularity = __tuners[tuner].getGranularity();
    }
    //vf.frequency = static_cast<unsigned int>(frequency / granularity);
    vf.frequency = static_cast<unsigned int>(frequency * 16 );
    vf.type = V4L2_TUNER_ANALOG_TV;
    if (ioctl(__fd, VIDIOC_S_FREQUENCY, &vf) < 0) {
      throw Exception("Failed to set frequency (ioctl: VIDIOC_S_FREQUENCY failed)");
    } else {
      DEBUG_STREAM("Set frequency to " << frequency << " on tuner " << tuner << endl);
    }
  }
  
  nmm_uint32 TVCardReadNode2::getBrightness() const {
    return getControl(V4L2_CID_BRIGHTNESS, "V4L2_CID_BRIGHTNESS");
  }
  
  nmm_uint32 TVCardReadNode2::getContrast() const {
    return getControl(V4L2_CID_CONTRAST, "V4L2_CID_CONTRAST");
  }
  
  nmm_uint32 TVCardReadNode2::getSaturation() const {
    return getControl(V4L2_CID_SATURATION, "V4L2_CID_SATURATION");
  }
  
  nmm_int32 TVCardReadNode2::getHue() const {
    return getControl(V4L2_CID_HUE, "V4L2_CID_HUE");
  }
  
  nmm_uint32 TVCardReadNode2::getVolume() const {
    return getControl(V4L2_CID_AUDIO_VOLUME, "V4L2_CID_AUDIO_VOLUME");
  }
  
  nmm_bool TVCardReadNode2::getMute() const {
    return (getControl(V4L2_CID_AUDIO_MUTE, "V4L2_CID_AUDIO_MUTE") != 0);
  }
  
  void TVCardReadNode2::setBrightness(const nmm_uint32& brightness) {
    setControl(V4L2_CID_BRIGHTNESS, brightness, "V4L2_CID_BRIGHTNESS");
  }
  
  void TVCardReadNode2::setContrast(const nmm_uint32& contrast) {
    setControl(V4L2_CID_CONTRAST, contrast, "V4L2_CID_CONTRAST");
  }
  
  void TVCardReadNode2::setSaturation(const nmm_uint32& saturation) {
    setControl(V4L2_CID_SATURATION, saturation, "V4L2_CID_SATURATION");
  }
  
  void TVCardReadNode2::setHue(const nmm_int32& hue) {
    setControl(V4L2_CID_HUE, hue, "V4L2_CID_HUE");
  }
  
  void TVCardReadNode2::setVolume(const nmm_uint32& volume) {
    setControl(V4L2_CID_AUDIO_VOLUME, volume, "V4L2_CID_AUDIO_VOLUME");
  }
  
  void TVCardReadNode2::setMute(const nmm_bool& mute) {
    setControl(V4L2_CID_AUDIO_MUTE, mute ? 1 : 0, "V4L2_CID_AUDIO_MUTE");
  }
  
  list<string> TVCardReadNode2::getFrequencyTableList() const {
    list<string> outlist;
    for (map<string, FrequencyTable>::const_iterator i(__frequencies.begin());
         i != __frequencies.end();
         ++i) {
      outlist.push_back(i->first);
    }
    return outlist;
  }
  
  string TVCardReadNode2::getFrequencyTable() const {
    return __cur_freqtable;
  }
  
  void TVCardReadNode2::setFrequencyTable(const string& table) {
    /* Check if this frequency table exists */
    map<string, FrequencyTable>::const_iterator found(__frequencies.find(table));
    if (found == __frequencies.end()) {
      throw Exception("Frequency table " + table + " does not exist");
    }
    
    /* Set channel */
    __cur_freqtable = table;
    setChannel(0);
  }

  /*============================== ITV Interface ==============================*/
  
  
  /*============================== GenericSyncNode Methods ==============================*/
  Time TVCardReadNode2::getLocalTime(u_int64_t valid_read_bytes) {
    // calcluate Rational exactely as
    // valid_read_bytes * frame_rate_base / int_frame_rate; framesize
    // to avoid overflow (framesize * int_frame_rate would create one)
    return Time(Rational(valid_read_bytes * _frame_rate_base / _int_frame_rate, _framesize));
  } 

  Result TVCardReadNode2::doInit() {
    /* Begin> Open device */
    if(!SSecurityCheck::getInstance().checkFilenameForReading(__device)) {
      ERROR_STREAM("Security check not passed for file " << __device << ", check .nmmrc for options" << endl);
      return FAILURE;
    }
    DEBUG_STREAM("Opening tv device " << __device << endl);
    //    __fd = ::open(__device.c_str(), 0/*O_RDONLY*/);
    __fd = ::open(__device.c_str(), O_RDWR/*O_RDONLY*/);
    if (__fd < 0) {
      string error(strerror(errno));
      ERROR_STREAM("Could not open device " << __device << " (" << error << ")" << endl);
      return FAILURE;
    }
    /* End< */

    /* Begin> Enumerate audio inputs */
    struct v4l2_audio vaudio;
    memset (&vaudio, 0, sizeof (vaudio));
    vaudio.index = 0;
    __audio_inputs.clear();
    DEBUG_STREAM("Available audio inputs:" << endl);
    while (ioctl(__fd, VIDIOC_ENUMAUDIO, &vaudio) >= 0) {
      TVAudio audio(vaudio.index, string(reinterpret_cast<const char*>(vaudio.name)),
                    vaudio.capability, vaudio.mode);
      memset (&vaudio, 0, sizeof (vaudio));
      vaudio.index = audio.getIndex()+1;
      __audio_inputs.push_back(audio);
      DEBUG_STREAM(" Index     : " << audio.getIndex() << endl <<
                   "  Name     : " << audio.getName() << endl <<
                   "  Caps     : " << audio.getCapabilities() << endl <<
                   "  Mode     : " << audio.getMode() << endl);
    }
    /* End< */

    /* Begin> Enumerate audio outputs */
    int index=0;
    struct v4l2_audioout vaudioout;
    memset (&vaudioout, 0, sizeof (vaudioout));
    vaudioout.index = 0;
    DEBUG_STREAM("Available audio outputs:" << endl);
    while (ioctl(__fd, VIDIOC_ENUMAUDOUT, &vaudioout) >= 0) {
      TVAudio audio(vaudioout.index, string(reinterpret_cast<const char*>(vaudioout.name)),
                    vaudioout.capability, vaudioout.mode);
      memset (&vaudioout, 0, sizeof (vaudioout));
      index++;
      vaudioout.index = index;;
      __audio_inputs.push_back(audio);
      DEBUG_STREAM(" Index     : " << audio.getIndex() << endl <<
                   "  Name     : " << audio.getName() << endl <<
                   "  Caps     : " << audio.getCapabilities() << endl <<
                   "  Mode     : " << audio.getMode() << endl);
    }
    /* End< */
    
    /* Begin> Enumerate inputs */
    struct v4l2_input vin;
    vin.index = 0;
    __inputs.clear();
    DEBUG_STREAM("Available inputs:" << endl);
    while (ioctl(__fd, VIDIOC_ENUMINPUT, &vin) >= 0) {
      TVInput input(vin.index, string(reinterpret_cast<const char*>(vin.name)),
                    vin.type, vin.audioset, vin.tuner, vin.std, vin.status);
      __inputs.push_back(input);
      DEBUG_STREAM(" Index     : " << input.getNum() << endl <<
                   "  Name     : " << input.getName() << endl <<
                   "  Type     : " << input.getType() << endl <<
                   "  AudioSet : " << input.getAudioSet() << endl <<
                   "  Tuner    : " << input.getTuner() << endl <<
                   "  Standard : " << input.getStandard() << endl <<
                   "  Status   : " << input.getStatus() << endl);
      vin.index++;
    }
    /* End< */
    
    /* Begin> Enumerate standards */
    struct v4l2_standard vs;
    vs.index = 0;
    __standards.clear();
    DEBUG_STREAM("Supported standards:" << endl);
    while(ioctl(__fd, VIDIOC_ENUMSTD, &vs) >= 0) {
      TVStandard standard(vs.id, string(reinterpret_cast<const char*>(vs.name)),
                          vs.frameperiod.numerator, vs.frameperiod.denominator,
                          vs.framelines);
      __standards.insert(pair<unsigned int, TVStandard>(standard.getID(), standard));
      DEBUG_STREAM(" ID            : " << standard.getID() << endl <<
                   "  Name         : " << standard.getName() << endl <<
                   "  Frame period : " << standard.getFramePeriodNumerator() << "/" <<
                   standard.getFramePeriodDenominator() << endl <<
                   "  Frame lines  : " << standard.getFrameLines() << endl);
      vs.index++;
    }
    /* End< */
    
    /* Begin> Enumerate tuners */
    struct v4l2_tuner vt;
    memset (&vt, 0, sizeof (vt));
    vt.index = 0;
    __tuners.clear();
    DEBUG_STREAM("Supported tuners:" << endl);
    while(ioctl(__fd, VIDIOC_G_TUNER, &vt) >= 0) {
      
      /* Begin> Determine tuner type */
      unsigned int type;
      switch(vt.type) {
      case V4L2_TUNER_RADIO:
        type = TVTuner::TYPE_RADIO;
        break;
      case V4L2_TUNER_ANALOG_TV:
        type = TVTuner::TYPE_ANALOG_TV;
        break;
      default:
        type = TVTuner::TYPE_UNKNOWN;
      }
      /* End< Determine tuner type */
      
      /* Begin> Determine frequency scale */
      float granularity;
      if (vt.capability & V4L2_TUNER_CAP_LOW) {
        granularity = 0.0000625; /* 62.5 Hz */
      } else {
        granularity = 0.0625;    /* 62.5 kHz */
      }
      float range_min = static_cast<float>(vt.rangelow) * granularity;
      float range_max = static_cast<float>(vt.rangehigh) * granularity;
      /* End< */
      
      TVTuner tuner(vt.index, string(reinterpret_cast<const char*>(vt.name)), type,
                    range_min, range_max, granularity);
      __tuners.push_back(tuner);
      
      DEBUG_STREAM(" Index        : " << vt.index << endl <<
                   "  Name        : " << vt.name << endl <<
                   "  Type        : " << ((type == TVTuner::TYPE_RADIO) ? "RADIO" :
                                          (type == TVTuner::TYPE_ANALOG_TV) ? "ANALOG TV" :
                                          "UNKNOWN") << endl <<
                   "  Capability  : " << vt.capability << endl <<
                   "  Range       : " << range_min << "~" << range_max << " MHz" << endl <<
                   "  Granularity : " << granularity << " MHz" << endl);
      memset (&vt, 0, sizeof (vt));
      vt.index = __tuners.size();
    }
    /* End< */
   
    /* Begin> Initialize frequency tables */
    try {
      initFrequencyTables();
    } catch(const Exception& e) {
      ERROR_STREAM("Failed to initialize frequency tables" << endl);
      ERROR_STREAM(e << endl);
      return FAILURE;
    }
    /* End< */
    
    DEBUG_STREAM("Init output formats"<<endl);
    int nfmts=0;
    Format* o_format;
    struct v4l2_fmtdesc fmt;
    fmt.index = nfmts;
    fmt.type  = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    while ( ioctl(__fd, VIDIOC_ENUM_FMT, &fmt) >= 0 )
      {
        nfmts++;
        fmt.index = nfmts;
        fmt.type  = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        DEBUG_STREAM("Format description " << fmt.description << endl);;
        o_format = pixelformatToFormat(fmt.pixelformat);
        if (o_format != NULL)
          if(fmt.pixelformat == V4L2_PIX_FMT_UYVY)
            {
              getOwnOutputProperty()->addNewFormat(o_format);
              getOwnOutputProperty()->setDefaultFormat(o_format);
            }
          else
            getOwnOutputProperty()->addNewFormat(o_format);
      }
    DEBUG_STREAM("TVCardReadNode2 doInit success" << endl);
    return SUCCESS;
  }
  
  Result TVCardReadNode2::doDeinit() {
    /* Erasae property */
    getOwnOutputProperty()->erase();
    
    /* Close device */
    if (__fd >= 0) {
      DEBUG_STREAM("Closing tv device" << endl);
      ::close(__fd);
      __fd = -1;
    }
    
    /* Clear lists */
    __audio_inputs.clear();
    __inputs.clear();
    __standards.clear();
    __tuners.clear();
    
    /* Clear frequency tables */
    __cur_channel = 0;
    __cur_freqtable = "";
    __frequencies.clear();
    
    return SUCCESS;
  }
  
  Result TVCardReadNode2::doInitOutput() {
    
    Format* orig_format;
    Format* o_format;
    Format* o2_format;
            
    struct v4l2_format fmt;
    format_list list = getOwnOutputProperty()->getFormatList();
    format_iterator it;
    it = list.begin();
    for(unsigned int i=0;i<list.size();i++)
      {
        orig_format = (*it);
        fmt.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        fmt.fmt.pix.width       = 720;
        fmt.fmt.pix.height      = 576;
    
        fmt.fmt.pix.pixelformat = FormatTopixelformat(orig_format);
        //      fmt.fmt.pix.colorspace  = V4L2_COLORSPACE_BT878;
        fmt.fmt.pix.field       = V4L2_FIELD_ANY;
        if (ioctl(__fd, VIDIOC_S_FMT, &fmt) >=0)
          {
            o_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
            o_format->addIntParamValue(Format::x_resolution_param,fmt.fmt.pix.width);
            o_format->addIntParamValue(Format::y_resolution_param,fmt.fmt.pix.height);
            o_format->addIntParamValue(Format::ratio_x_param, 4);
            o_format->addIntParamValue(Format::ratio_y_param, 3);
            o_format->addFloatParamValue(Format::framerate_param,__frame_rate);
            o_format->addStringParamValue(Format::channelorder_param, orig_format->getStringValue(Format::channelorder_param));
            o_format->addStringParamValue(Format::endian_param, orig_format->getStringValue(Format::endian_param));
            o_format->addStringParamValue(Format::colorspace_param,orig_format->getStringValue(Format::colorspace_param));
            o_format->addIntParamValue(Format::bitperpixel_param, orig_format->getIntValue(Format::bitperpixel_param));
            o_format->addStringParamValue(Format::format_param, orig_format->getStringValue(Format::format_param));
        

            o2_format = getOwnWorkingOutputProperty()->addNewFormat("video/raw");
            o2_format->addWildcard(Format::x_resolution_param);
            o2_format->addWildcard(Format::y_resolution_param);
            o2_format->addWildcard(Format::ratio_x_param);
            o2_format->addWildcard(Format::ratio_y_param);
            if(__frame_rate == 30){
              o2_format->addFloatParamValue(Format::framerate_param,__frame_rate);
            }else{
              o2_format->addWildcard(Format::framerate_param);
            }
            o2_format->addStringParamValue(Format::channelorder_param, orig_format->getStringValue(Format::channelorder_param));
            o2_format->addStringParamValue(Format::endian_param, orig_format->getStringValue(Format::endian_param));
            o2_format->addStringParamValue(Format::colorspace_param,orig_format->getStringValue(Format::colorspace_param));
            o2_format->addIntParamValue(Format::bitperpixel_param, orig_format->getIntValue(Format::bitperpixel_param));
            o2_format->addStringParamValue(Format::format_param, orig_format->getStringValue(Format::format_param));
        
            if(fmt.fmt.pix.pixelformat == V4L2_PIX_FMT_YVU420/*V4L2_PIX_FMT_UYVY*/)
              {
                getOwnWorkingOutputProperty()->setDefaultFormat(o_format);
              }
          }
        it++;
      }        
    /* END< */
    return SUCCESS;        
  }

  Result TVCardReadNode2::doDeinitOutput()
  {
    for (unsigned int i = 0; i < n_buffers; ++i)
      if (-1 == munmap (buffers[i].start, buffers[i].length))
        {
          ERROR_STREAM("Error while freeing buffers"<<endl);
          return FAILURE;
        }

    return SUCCESS;

  } 

  Result TVCardReadNode2::doActivate() {
    struct v4l2_format fmt;
    int pixelformat;
    const Format* orig_format = getOutputFormat();
    orig_format->showParameter();
    if (orig_format == NULL) {
      ERROR_STREAM("Can't set output format at doInitOutput" << endl);
      return FAILURE;
    }

    fmt.type                = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    fmt.fmt.pix.width       = orig_format->getIntValue(Format::x_resolution_param);
    fmt.fmt.pix.height      = orig_format->getIntValue(Format::y_resolution_param);
    if((pixelformat=FormatTopixelformat(orig_format))==-1)
      {
        ERROR_STREAM("Can't set output format at doInitOutput from getOutputFormat" << endl);
        return FAILURE;
      }
    
    fmt.fmt.pix.pixelformat = pixelformat;
    fmt.fmt.pix.field       = V4L2_FIELD_ANY;

    if (ioctl(__fd, VIDIOC_S_FMT, &fmt) >=0)
      {

        /* Buggy driver paranoia. */
        unsigned int min = fmt.fmt.pix.width * 2;
        if (fmt.fmt.pix.bytesperline < min)
          fmt.fmt.pix.bytesperline = min;
        min = fmt.fmt.pix.bytesperline * fmt.fmt.pix.height;
        if (fmt.fmt.pix.sizeimage < min)
          fmt.fmt.pix.sizeimage = min;

        setBufferSize(fmt.fmt.pix.sizeimage);
        
        //We use MMAP method

        struct v4l2_requestbuffers reqbuf;
        unsigned int i;

        memset (&reqbuf, 0, sizeof (reqbuf));
        reqbuf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        reqbuf.memory = V4L2_MEMORY_MMAP;
        reqbuf.count = 10;
      
        if ( ioctl(__fd, VIDIOC_REQBUFS, &reqbuf) >=0 ) {
          if (reqbuf.count < 4) {
            ERROR_STREAM ("Insufficient buffer memory on device"<<endl);
            return FAILURE;
          }

          buffers =(buffer*) calloc (reqbuf.count, sizeof (*buffers));

          if (!buffers) {
            ERROR_STREAM("Out of memory"<<endl);
            return FAILURE;
          }
          DEBUG_STREAM("Allocated "<<reqbuf.count<<" buffers"<<endl);

          for (n_buffers = 0; n_buffers < reqbuf.count; ++n_buffers) {

            struct v4l2_buffer buf;

            CLEAR (buf);

            buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            buf.memory      = V4L2_MEMORY_MMAP;
            buf.index       = n_buffers;
            if ((ioctl (__fd, VIDIOC_QUERYBUF, &buf,0))>=0){

              buffers[n_buffers].length = buf.length;
              buffers[n_buffers].start =
                mmap (NULL /* start anywhere */,
                      buf.length,
                      PROT_READ | PROT_WRITE /* required */,
                      MAP_SHARED /* recommended */,
                      __fd, buf.m.offset);

              if (MAP_FAILED == buffers[n_buffers].start){
                ERROR_STREAM("Error to query mmap (QUERYBUF)"<<endl);
                return FAILURE;
              }
            } else {
              ERROR_STREAM("Error to query buffer (QUERYBUF)"<<endl);
              return FAILURE;
            }
          }

          for (i = 0; i < n_buffers; ++i) {
            struct v4l2_buffer buf;
                                                                                                                                                             
            CLEAR (buf);
                                                                                                                                                            
            buf.type        = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            buf.memory      = V4L2_MEMORY_MMAP;
            buf.index       = i;
                                                                                                                                                             
            if (-1 == ioctl (__fd, VIDIOC_QBUF, &buf)){
              ERROR_STREAM("Error to query buffer (QBUF)"<<endl);
              return FAILURE;
            }
          }
                                                                                                                                                             
          int type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
                                                                                                                                                             
          if (-1 == ioctl (__fd, VIDIOC_STREAMON, &type)){
            ERROR_STREAM("Error to activate STREAMON"<<endl);
            return FAILURE;
          }
                                                                                                                                                             
          __do_flush = false;
          return SUCCESS;
        
        }
        ERROR_STREAM("Error at REQBUFFER"<<endl);    
        return FAILURE;
      }
    ERROR_STREAM("Error at setformat"<<endl);
    return FAILURE;
  }
  
  Result TVCardReadNode2::doDeactivate()
  {
    enum v4l2_buf_type type;

    type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
                                                                                                                                                             
    if (-1 == ioctl (__fd, VIDIOC_STREAMOFF, &type)){
      ERROR_STREAM("Error doing STREAMOFF"<<endl);
      return FAILURE;
    }
    return SUCCESS;

  }

  Message* TVCardReadNode2::produceBuffer(Buffer *) {
    Buffer* out;
    /* Check if we need to flush */
    if (__do_flush) {
      __do_flush = false;
      CEvent* cevent = new FlushEvent();
      cevent->insertEvent(cevent->size(), ISyncReset::create_syncReset());
      __stream_timer.setTimestamp(cevent);
      return cevent;
    }
    struct v4l2_buffer buf;
    //    Buffer* out = getNewBuffer(buffers[buf.index].length);
    CLEAR (buf);
                                                                                                                                                             
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.memory = V4L2_MEMORY_MMAP;
    if (ioctl (__fd, VIDIOC_DQBUF, &buf)>=0)
      {
        if(buf.index >= n_buffers)
          {
            ERROR_STREAM("Failed to manipulate buffer"<<endl);
            setProducingFlag(false);
            return NULL;
          }
        out = getNewBuffer(buffers[buf.index].length);
        /* Get output buffer */
        if (!out) 
          {
            ERROR_STREAM("Failed to allocate buffer"<<endl);
            setProducingFlag(false);
            return NULL;
          }
        //void *data = malloc(buffers[buf.index].length);
        //memcpy(data,buffers[buf.index].start,buffers[buf.index].length);
        //out->setData(buffers[buf.index].length, (char*)data);
        out->setData(buffers[buf.index].length, (char*)buffers[buf.index].start);
        _framesize = buffers[buf.index].length;
      }
    
    if (-1 == ioctl (__fd, VIDIOC_QBUF, &buf))
      {
        ERROR_STREAM("Error while enqueuing buffer"<<endl);
        setProducingFlag(false);
        return NULL;
      }

    __stream_timer.setRate(__frame_rate);
    __stream_timer.setTimestamp(out);
    return out;
    
  }
  /*============================== GenericNode Methods ==============================*/

  
  /*============================== Helper Methods ==============================*/
  void  TVCardReadNode2::setControl(unsigned int id, unsigned int value, const string& name) {
    struct v4l2_control ctrl;
    ctrl.id = id;
    ctrl.value = value;
    if (ioctl(__fd, VIDIOC_S_CTRL, &ctrl) == 0) {
      DEBUG_STREAM("Set control " << name << " to " << value << endl);
    } else {
      throw Exception("Error setting control value (VIDIOC_S_CTRL with ctrl.id=" + name + " failed)");
    }
  }
  
  unsigned int TVCardReadNode2::getControl(unsigned int id, const string& name) const {
    struct v4l2_control ctrl;
    ctrl.id = id;
    if (ioctl(__fd, VIDIOC_G_CTRL, &ctrl) == 0) {
      return id;
    } else {
      throw Exception("Error getting control value (VIDIOC_G_CTRL with ctrl.id=" + name + " failed)");
    }
  }
  
  void TVCardReadNode2::initFrequencyTables()
    throw (const Exception&)
  {
    string filename;
    
    /* Default filename */
    filename = getNMMResource("ivtv/tuner.xml");
    MESSAGE_STREAM("Parsing frequency file " << filename << endl);
    
    /* Begin> Parse XML file */
    XMLTree* xml = new XMLTree( filename );
    XMLNode* root = xml->root();
    if (!root) {
      throw Exception("Failed to parse XML file " + filename);
    }
    if (root->name() != "frequencies") {
      throw Exception("XMl file " + filename + " is not a frequency table file");
    }
    /* End< */
    
    /* Begin> Parse channel lists */
    XMLNodeList nlist_chanlist(root->children("chanlist"));
    for (XMLNodeIterator i(nlist_chanlist.begin()); i != nlist_chanlist.end(); ++i) {
      
      /* Get name of channel list */
      XMLProperty* prop_chanlist_name((*i)->property("name"));
      if (!prop_chanlist_name) {
        throw Exception("Error in frequency table: Channel list has no name");
      }
      
      /* Create new frequency table */
      FrequencyTable table(prop_chanlist_name->value());
      
      /* Begin> Parse sub nodes of channel list */
      XMLNodeList nlist((*i)->children());
      for (XMLNodeIterator j(nlist.begin()); j != nlist.end(); ++j) {
        
        /* Begin> Parse channel entry */
        if ((*j)->name() == "channel") {
          
          /* Get channel name */
          XMLProperty* prop_name((*j)->property("name"));
          if (!prop_name) {
            throw Exception("Error in frequency table: Channel has no name");
          }
          
          /* Get channel frequency */
          XMLProperty* prop_freq((*j)->property("frequency"));
          if (!prop_freq) {
            throw Exception("Error in frequency table: Channel has no frequency");
          }
          
          /* Add channel entry */
          table.addChannel(prop_name->value(), atoi(prop_freq->value().c_str()));
        } 
        /* End< */
        
        /* Begin> Parse include entry */
        else if ((*j)->name() == "include") {
          
          /* Get include  name */
          XMLProperty* prop_name((*j)->property("name"));
          if (!prop_name) {
            throw Exception("Error in frequency table: Include has no name");
          }
          
          /* Check if that channel list exists */
          map<string, FrequencyTable>::const_iterator found(__frequencies.find(prop_name->value()));
          if (found == __frequencies.end()) {
            throw Exception("Error in frequency table: Include unknown frequency table " + prop_name->value());
          }
          
          /* Add channels from that frequency table to current */
          for (unsigned int i=0; i<found->second.getChannelCount(); ++i) {
            table.addChannel(found->second.getChannelName(i), found->second.getChannelFrequency(i));
          }
          
        }
        /* End< */
        
      }
      /* End< Parse sub nodes of channel list */
      
      /* Check if table is non-empty */
      /*
        if (table.getChannelCount() == 0) {
        throw Exception("Channel list " + prop_chanlist_name->value() + " is empty");
        }
      */
      /* Add table to map */
      DEBUG_STREAM("Adding channel list " << prop_chanlist_name->value() << endl);
      __frequencies[prop_chanlist_name->value()] = table;
    }
    if(__frequencies.empty())
      throw Exception("no Frequency-Tables available");
  }
  
  Format *TVCardReadNode2::pixelformatToFormat(long pixelfmt)
  {
    Format *o_format = new Format();
    switch(pixelfmt)
      {
      case V4L2_PIX_FMT_YUYV:
        o_format->setFormatType("video/raw");
        o_format->addWildcard(Format::x_resolution_param);
        o_format->addWildcard(Format::y_resolution_param);
        o_format->addWildcard(Format::ratio_x_param);
        o_format->addWildcard(Format::ratio_y_param);
        o_format->addFloatParamValue(Format::framerate_param, __frame_rate);
        o_format->addStringParamValue(Format::channelorder_param,"yuyv");
        o_format->addStringParamValue(Format::endian_param, "little");
        o_format->addStringParamValue(Format::colorspace_param,"yuv422");
        o_format->addIntParamValue(Format::bitperpixel_param, 16);
        o_format->addStringParamValue(Format::format_param, "packet");
        break;
        
      case V4L2_PIX_FMT_UYVY:
        o_format->setFormatType("video/raw");
        o_format->addWildcard(Format::x_resolution_param);
        o_format->addWildcard(Format::y_resolution_param);
        o_format->addWildcard(Format::ratio_x_param);
        o_format->addWildcard(Format::ratio_y_param);
        o_format->addFloatParamValue(Format::framerate_param, __frame_rate);
        o_format->addStringParamValue(Format::channelorder_param,"uyvy");
        o_format->addStringParamValue(Format::endian_param, "little");
        o_format->addStringParamValue(Format::colorspace_param,"yuv422");
        o_format->addIntParamValue(Format::bitperpixel_param, 16);
        o_format->addStringParamValue(Format::format_param, "packet");
        break;
          
      case V4L2_PIX_FMT_YUV422P:
        o_format->setFormatType("video/raw");
        o_format->addWildcard(Format::x_resolution_param);
        o_format->addWildcard(Format::y_resolution_param);
        o_format->addWildcard(Format::ratio_x_param);
        o_format->addWildcard(Format::ratio_y_param);
        o_format->addFloatParamValue(Format::framerate_param, __frame_rate);
        o_format->addStringParamValue(Format::channelorder_param,"yuyv");
        o_format->addStringParamValue(Format::endian_param, "little");
        o_format->addStringParamValue(Format::colorspace_param,"yuv422");
        o_format->addIntParamValue(Format::bitperpixel_param, 16);
        o_format->addStringParamValue(Format::format_param, "planar");
        break;
          
      case V4L2_PIX_FMT_YUV411P:
        o_format->setFormatType("video/raw");
        o_format->addWildcard(Format::x_resolution_param);
        o_format->addWildcard(Format::y_resolution_param);
        o_format->addWildcard(Format::ratio_x_param);
        o_format->addWildcard(Format::ratio_y_param);
        o_format->addFloatParamValue(Format::framerate_param, __frame_rate);
        o_format->addStringParamValue(Format::channelorder_param,"yuv");
        o_format->addStringParamValue(Format::endian_param, "little");
        o_format->addStringParamValue(Format::colorspace_param,"yuv411");
        o_format->addIntParamValue(Format::bitperpixel_param, 12);
        o_format->addStringParamValue(Format::format_param, "planar");
        break;
          
      case V4L2_PIX_FMT_Y41P:
        o_format->setFormatType("video/raw");
        o_format->addWildcard(Format::x_resolution_param);
        o_format->addWildcard(Format::y_resolution_param);
        o_format->addWildcard(Format::ratio_x_param);
        o_format->addWildcard(Format::ratio_y_param);
        o_format->addFloatParamValue(Format::framerate_param, __frame_rate);
        o_format->addStringParamValue(Format::channelorder_param,"yuv");
        o_format->addStringParamValue(Format::endian_param, "little");
        o_format->addStringParamValue(Format::colorspace_param,"yuv411");
        o_format->addIntParamValue(Format::bitperpixel_param, 12);
        o_format->addStringParamValue(Format::format_param, "packet");
        break;
          
      case V4L2_PIX_FMT_YUV420:
        o_format->setFormatType("video/raw");
        o_format->addWildcard(Format::x_resolution_param);
        o_format->addWildcard(Format::y_resolution_param);
        o_format->addWildcard(Format::ratio_x_param);
        o_format->addWildcard(Format::ratio_y_param);
        o_format->addFloatParamValue(Format::framerate_param, __frame_rate);
        o_format->addStringParamValue(Format::channelorder_param,"yuv");
        o_format->addStringParamValue(Format::endian_param, "little");
        o_format->addStringParamValue(Format::colorspace_param,"yv12");
        o_format->addIntParamValue(Format::bitperpixel_param, 12);
        o_format->addStringParamValue(Format::format_param, "planar");
        break;
          
      case V4L2_PIX_FMT_YVU420:
        o_format->setFormatType("video/raw");
        o_format->addWildcard(Format::x_resolution_param);
        o_format->addWildcard(Format::y_resolution_param);
        o_format->addWildcard(Format::ratio_x_param);
        o_format->addWildcard(Format::ratio_y_param);
        o_format->addFloatParamValue(Format::framerate_param, __frame_rate);
        o_format->addStringParamValue(Format::channelorder_param,"yvu");
        o_format->addStringParamValue(Format::endian_param, "little");
        o_format->addStringParamValue(Format::colorspace_param,"yv12");
        o_format->addIntParamValue(Format::bitperpixel_param, 12);
        o_format->addStringParamValue(Format::format_param, "planar");
        break;
          
      case V4L2_PIX_FMT_RGB565:
        o_format->setFormatType("video/raw");
        o_format->addWildcard(Format::x_resolution_param);
        o_format->addWildcard(Format::y_resolution_param);
        o_format->addWildcard(Format::ratio_x_param);
        o_format->addWildcard(Format::ratio_y_param);
        o_format->addFloatParamValue(Format::framerate_param, __frame_rate);
        o_format->addStringParamValue(Format::channelorder_param,"rgb");
        o_format->addStringParamValue(Format::endian_param, "little");
        o_format->addStringParamValue(Format::colorspace_param,"rgb565");
        o_format->addIntParamValue(Format::bitperpixel_param, 16);
        o_format->addStringParamValue(Format::format_param, "packet");
        break;
          
      case V4L2_PIX_FMT_RGB24:
        o_format->setFormatType("video/raw");
        o_format->addWildcard(Format::x_resolution_param);
        o_format->addWildcard(Format::y_resolution_param);
        o_format->addWildcard(Format::ratio_x_param);
        o_format->addWildcard(Format::ratio_y_param);
        o_format->addFloatParamValue(Format::framerate_param, __frame_rate);
        o_format->addStringParamValue(Format::channelorder_param,"rgb");
        o_format->addStringParamValue(Format::endian_param, "little");
        o_format->addStringParamValue(Format::colorspace_param,"rgb24");
        o_format->addIntParamValue(Format::bitperpixel_param, 24);
        o_format->addStringParamValue(Format::format_param, "packet");
        break;
          
      case V4L2_PIX_FMT_BGR24:
        o_format->setFormatType("video/raw");
        o_format->addWildcard(Format::x_resolution_param);
        o_format->addWildcard(Format::y_resolution_param);
        o_format->addWildcard(Format::ratio_x_param);
        o_format->addWildcard(Format::ratio_y_param);
        o_format->addFloatParamValue(Format::framerate_param, __frame_rate);
        o_format->addStringParamValue(Format::channelorder_param,"bgr");
        o_format->addStringParamValue(Format::endian_param, "little");
        o_format->addStringParamValue(Format::colorspace_param,"rgb24");
        o_format->addIntParamValue(Format::bitperpixel_param, 24);
        o_format->addStringParamValue(Format::format_param, "packet");
        break;
      case V4L2_PIX_FMT_RGB32:
        o_format->setFormatType("video/raw");
        o_format->addWildcard(Format::x_resolution_param);
        o_format->addWildcard(Format::y_resolution_param);
        o_format->addWildcard(Format::ratio_x_param);
        o_format->addWildcard(Format::ratio_y_param);
        o_format->addFloatParamValue(Format::framerate_param, __frame_rate);
        o_format->addStringParamValue(Format::channelorder_param,"rgba");
        o_format->addStringParamValue(Format::endian_param, "little");
        o_format->addStringParamValue(Format::colorspace_param,"rgba32");
        o_format->addIntParamValue(Format::bitperpixel_param, 32);
        o_format->addStringParamValue(Format::format_param, "packet");
        break;
      case V4L2_PIX_FMT_BGR32:
        o_format->setFormatType("video/raw");
        o_format->addWildcard(Format::x_resolution_param);
        o_format->addWildcard(Format::y_resolution_param);
        o_format->addWildcard(Format::ratio_x_param);
        o_format->addWildcard(Format::ratio_y_param);
        o_format->addFloatParamValue(Format::framerate_param, __frame_rate);
        o_format->addStringParamValue(Format::channelorder_param,"bgra");
        o_format->addStringParamValue(Format::endian_param, "little");
        o_format->addStringParamValue(Format::colorspace_param,"rgba32");
        o_format->addIntParamValue(Format::bitperpixel_param, 32);
        o_format->addStringParamValue(Format::format_param, "packet");
        break;
      default:
        o_format=NULL;
        break;
      }
    return o_format;
  }
  
  int TVCardReadNode2::FormatTopixelformat(const Format *fmt)
  {
    
    string order = fmt->getStringValue(Format::channelorder_param);
    
    if(order=="yuyv")
      {
        if(fmt->getStringValue(Format::format_param) == "packet")
          return V4L2_PIX_FMT_YUYV;
        else
          return V4L2_PIX_FMT_YUV422P;
      }

    if(order=="uyvy")
      return V4L2_PIX_FMT_UYVY;

    if(order=="yuv")
      {
        if(fmt->getStringValue(Format::format_param) == "packet")
          return V4L2_PIX_FMT_Y41P;
        else {
          if(fmt->getStringValue(Format::colorspace_param)=="yuv411")
            return V4L2_PIX_FMT_YUV411P;
          else
            return V4L2_PIX_FMT_YUV420;
        }
      }

    if(order=="yvu")
      return V4L2_PIX_FMT_YVU420;

    if(order=="rgb")
      {
        if(fmt->getIntValue(Format::bitperpixel_param)==16)
          return V4L2_PIX_FMT_RGB565;
        else
          return V4L2_PIX_FMT_RGB24;
      }

    if(order=="bgr")
      return V4L2_PIX_FMT_BGR24;

    if(order=="rgba")
      return V4L2_PIX_FMT_RGB32;

    if(order=="bgra")
      return V4L2_PIX_FMT_BGR32;

    return -1;

  }
  /*============================== Helper Methods ==============================*/
  
}


#include "nmm/nmm_suffix.hpp"
