/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

 /* This source code was developed by Andrés Otón Urbano in 2004 to 2005.
  * Copyright and Rights of Use were assigned to Motama GmbH by Andrés in
  * 2007.
  */

#ifndef _NMM_TVCARDREADNODE2_HPP_
#define _NMM_TVCARDREADNODE2_HPP_ _NMM_TVCARDREADNODE2_HPP_

#include "nmm/nmm_prefix.hpp"

#include "nmm/multimedia/sync/GenericSyncSourceNode.hpp"
#include "nmm/multimedia/sync/StreamTimer.hpp"
#include "nmm/plugins/interfaces/IBufferSizeImpl.hpp"
#include "nmm/plugins/video/tv/interfaces/ITVCardImpl.hpp"
#include "nmm/plugins/video/tv/interfaces/ITVImpl.hpp"

#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>

extern "C"
{
#include <linux/types.h>
}
#include "videodev2.h"

#include "FrequencyTable.hpp"

namespace NMM {

 /** TVCardReadNode2 Read Node.
  */
 class TVCardReadNode2 :
   public GenericSyncSourceNode,
   public IBufferSizeImpl,
   public ITVCardImpl,
   public ITVImpl
 {
 public:
   /** @name Constructor & Destductor
    */
   //@{
   /** Constructor.
    */
   TVCardReadNode2(const char *name = "TVCardReadNode2",
                StreamQueue::Mode = StreamQueue::MODE_SUSPEND,
                int = StreamQueue::DEFAULT_SIZE);

   /** Virtual destructor.
    */
   virtual ~TVCardReadNode2();
   //@}

 protected:
   /** @name GenericNode Methods
    */
   //@{
   virtual Result doInit();
   virtual Result doDeinit();
   virtual Result doInitOutput();
   virtual Result doDeinitOutput();
   virtual Result doActivate();
   virtual Result doDeactivate();

   virtual Message* produceBuffer(Buffer*);
   //@}

   /** @name IBufferSize interface
    */
   //@{
   /** Set buffer size.
    */
   virtual void setBufferSize(const nmm_int32& value);

   /** Get buffer size.
    */
   virtual nmm_int32 getBufferSize();
   //@}

   /** @name ITVCard Interface
    */
   //@{
   /** Set device.
    */
   virtual void setDevice(const string& value) throw(const DeviceNotFoundException&);

   /** Get device.
    */
   virtual string getDevice() const;

   /** Set size.
    */
   virtual void setSize(const nmm_int32& x, const nmm_int32& y);

   /** Get size.
    */
   virtual void getSize(nmm_int32& x, nmm_int32& y);

   /** Get resolution
    */
   virtual void getResolution(nmm_int32& x, nmm_int32& y);

   /** Set resolution
    */
   virtual void setResolution(const nmm_int32& x, const nmm_int32& y);

   /** Set channel.
    */
   virtual Result setChannel(const nmm_int32& channel);

   /** Get channel.
    */
   virtual Result getChannel(nmm_int32& channel, string& name);
   //@}

   virtual Result hasChannel(nmm_int32&, const string&){return FAILURE;}

   /** @name Audio Inputs
    */
   //@{
   /** This method returns the number of available inputs.
    */
   virtual nmm_uint32 getAudioCount() const;

   /** This method returns the properties of the specified input.
    * @param index Audio index, ranges from 0 to getAudioCount()-1
    */
   virtual TVAudio getAudio(const nmm_uint32& index) const;

   /** This method returns the list of available inputs.
    */
   virtual list<TVAudio> getAudioList() const;

   /** This method returns the current input.
    */
   virtual nmm_uint32 getCurrentAudio() const;

   /** This method sets the current input.
    * @param index Audio index, ranges from 0 to getAudioCount()-1
    */
   virtual void setCurrentAudio(const nmm_uint32& index);
   //@}

   /** @name Inputs
    */
   //@{
   /** This method returns the number of available inputs.
    */
   virtual nmm_uint32 getInputCount() const;

   /** This method returns the properties of the specified input.
    * @param num Input number, ranges from 0 to getInputCount()-1
    */
   virtual TVInput getInput(const nmm_uint32& num) const;

   /** This method returns the list of available inputs.
    */
   virtual list<TVInput> getInputList() const;

   /** This method returns the current input.
    */
   virtual nmm_uint32 getCurrentInput() const;

   /** This method sets the current input.
    * @param num Input number, ranges from 0 to getInputCount()-1
    */
   virtual void setCurrentInput(const nmm_uint32& num);
   //@}

   /** @name Standards
    */
   //@{
   /** This method returns the identifier of the standard which matches the given name.
    * If no such standard is found, zero is returned.
    */
   virtual nmm_uint32 getStandardByName(const string& name) const;

   /** This method returns the properties of the specified standard.
    * @param num Standard number
    */
   virtual TVStandard getStandard(const nmm_uint32& num) const;

   /** This method returns the list of supported standards.
    */
   virtual list<TVStandard> getStandardList() const;

   /** This method returns the standard set for the current input.
    */
   virtual nmm_uint32 getCurrentStandard() const;

   /** This method sets the standard for the current input
    * @param num Standard number, ranges from 0 to getStandardCount()-1
    */
   virtual void setCurrentStandard(const nmm_uint32& standard);

   /** This method sets the standard for the current input by name.
    */
   virtual void setStandardByName(const string& name);
   //@}

   /** @name Tuners
    */
   //@{
   /** This method returns the number of tuners.
    */
   virtual nmm_uint32 getTunerCount() const;

   /** This method returns information about a tuner.
    * @param index Index of tuner, ranges from 0 to getTunerCount()-1
    */
   virtual TVTuner getTuner(const nmm_uint32& index) const;

   /** This method returns the list of tuners.
    */
   virtual list<TVTuner> getTunerList() const;

   /** This method returns the current frequency of a tuner.
    * @param tuner Number of the tuner
    * @return Current frequency of the tuner in MHz
    */
   virtual nmm_float getFrequency(const nmm_uint32& tuner) const;

   /** This method sets the frequency of a tuner.
    * @param tuner Number of the tuner
    * @param frequency The frequency to set in MHz
    */
   virtual void setFrequency(const nmm_uint32& tuner, const nmm_float& frequency);
   //@}

   /** @name Controls.
    */
   //@{
   /** This method gets the current brightness.
    */
   virtual nmm_uint32 getBrightness() const;

   /** This method gets the current contrast.
    */
   virtual nmm_uint32 getContrast() const;

   /** This method gets the current saturation.
    */
   virtual nmm_uint32 getSaturation() const;

   /** This method gets the current hue.
    */
   virtual nmm_int32 getHue() const;

   /** This method gets the current volume.
    */
   virtual nmm_uint32 getVolume() const;

   /** This method gets the mute status.
    */
   virtual nmm_bool getMute() const;

   /** This method sets the brightness.
    * @param brightness Brightness (0 to 255)
    */
   virtual void setBrightness(const nmm_uint32& brightness);

   /** This method sets the contrast.
    * @param contrast Contrast (0 to 127)
    */
   virtual void setContrast(const nmm_uint32& contrast);

   /** This method sets the saturation.
    * @param saturation Saturation (0 to 127)
    */
   virtual void setSaturation(const nmm_uint32& saturation);

   /** This method sets the hue.
    * @param hue Hue offset (-128 to 127)
    */
   virtual void setHue(const nmm_int32& hue);

   /** This method sets the volume.
    * @param volume Volume (0 to 65535)
    */
   virtual void setVolume(const nmm_uint32& volume);

   /** This method sets the mute state.
    */
   virtual void setMute(const nmm_bool& mute);
   //@}

   /** @name Frequency tables
    */
   //@{
   /** This method returns the available frequency tables.
    */
   virtual list<string> getFrequencyTableList() const;

   /** This method returns the current frequency table.
    */
   virtual string getFrequencyTable() const;

   /** This method sets the current frequency table.
    * @param name Name of the frequency table
    */
   virtual void setFrequencyTable(const string& table);

   /** virtual function inherited from GenericSyncSourceNode
    */
   virtual Time getLocalTime(u_int64_t);

 private:
   /** This method sets the value of a control.
    */
   void setControl(unsigned int id, unsigned int value, const string& name);

   /** This method returns the value of a control.
    */
   unsigned int getControl(unsigned int id, const string& name) const;

   /** This method convert pixelformat from V4L2 to Format of nmm.
    */
   Format *pixelformatToFormat(long pixelfmt);

   /** This method convert a Format of nmm to pixelformat.
    */
   int FormatTopixelformat(const Format *fmt);

   /** This method initializes the frequency tables.
    */
   void initFrequencyTables()
     throw (const Exception&);

   virtual void setInputOutput(const int& input, const int& output);

 protected:
   int _framesize;
   int _int_frame_rate;
   const int _frame_rate_base;
   
 private:
   typedef vector<TVInput>                 Inputs;
   typedef Inputs::iterator                InputIterator;
   typedef Inputs::const_iterator          InputConstIterator;

   typedef vector<TVAudio>                 AudioInputs;
   typedef AudioInputs::iterator           AudioInputIterator;
   typedef AudioInputs::const_iterator     AudioInputConstIterator;

   typedef vector<TVTuner>                 Tuners;
   typedef Tuners::iterator                TunerIterator;
   typedef Tuners::const_iterator          TunerConstIterator;

   typedef map<unsigned int, TVStandard>   Standards;
   typedef Standards::iterator             StandardIterator;
   typedef Standards::const_iterator       StandardConstIterator;

   /* Begin> Device */
   string                      __device;            ///< Device filename
   int                         __fd;                ///< File descriptor to read from device
   /* End< */

   /* Begin> Information */
   AudioInputs                 __audio_inputs;      ///< Available audio inputs
   Inputs                      __inputs;            ///< Available inputs
   Tuners                      __tuners;            ///< Available tuners
   Standards                   __standards;         ///< Supported standards
   float                       __frame_rate;        ///< Frame rate of current standard
   /* End< */

   /* Begin> Output */
   ssize_t                     __buffer_size;       ///< Read buffer size
   bool                        __do_flush;          ///< Need to flush the stream
   StreamTimer                 __stream_timer;      ///< Stream timer
   /* End< */

   /* Begin> Channels */
   unsigned int                __cur_channel;       ///< Current channel
   string                      __cur_freqtable;     ///< Current frequency table
   map<string, FrequencyTable> __frequencies;       ///< Frequency tables
   /* End< */

  /* Begin> Buffers of device */
  struct buffer {
    void *start;
    size_t length;
  };
  struct buffer * buffers;
  unsigned int n_buffers;
 };

}

#include "nmm/nmm_suffix.hpp"

#endif
