/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <fstream>
#include <regex.h>

#include "ChannelConfig.hpp"

#include "nmm/utils/directory_utils.hpp"
#include "nmm/utils/NamedObject.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  ChannelConfig::ChannelConfig() {
    config.resize(0);
  }

  ChannelConfig& ChannelConfig::load(mode_e m, string file) {

    config.clear();
    switch(m){
    case DVB:
      {
        string filename;
        if(file.empty()){
          filename = getNMMResource("dvb/channels.conf");
        }else{
          filename = file;
        }
        readDVBChannels(filename);
      }
      break;
    case IVTV:
      {
        string filename;
        if(file.empty()){
          filename = getNMMResource("ivtv/ivtv.conf");
        }else{
          filename = file;
        }
        readIVTVChannels(filename);
      }
      break;
    }
    return *this;
  }
  
  unsigned int ChannelConfig::getChannelNumber(string ch){
       

    for(unsigned int i=0; i < config.size(); i++){
      string tmp (config[i].name);
      if(compare(ch, tmp)) return i+1;
    } 
    return 0;
  }

  bool ChannelConfig::compare(string s1, string s2){

    transform(s1); transform(s2);
    if(s1 == s2) return true;

    string bufferLine;
    string c1, c2;

    ifstream channelAlias(aliasFilename.c_str());
    
    if(!channelAlias) {
      GLOBAL_ERROR_STREAM("channel.alias not found" << std::endl);
      return false;
    }
    
    while(!channelAlias.eof()){
      getline(channelAlias, bufferLine);
      size_t pos = bufferLine.find_first_of(':');
      if(pos != string::npos){
        c1 = bufferLine.substr(0, pos);
        c2 = bufferLine.substr(pos+1);
        transform(c1); transform(c2);
        
        if(((s1 == c1) || (s1 == c2)) && ((s2 == c1) || (s2 == c2))) return true;
      }
    }
    
    return false;

  }

  void ChannelConfig::readDVBChannels(string file){
    Channel chan;
    ifstream configfile(file.c_str());
    char line[101];

    if (!configfile) {
    	GLOBAL_ERROR_STREAM("Channel file " << file << " not found" << endl);
        return;
    }
    string dummy;
    string freq;
    while (!configfile.eof()) {
      configfile.getline(line, 100);
      if(line){
        regex_t exp;
        regmatch_t pm;
        if(!regcomp(&exp, ":",REG_EXTENDED)){
          if(regexec(&exp, line, 1, &pm, 0) == REG_NOMATCH) continue;
          string res(line);

          chan.name[0] = '\0';
          strcat(chan.name, res.substr(0, pm.rm_so).c_str());
          int freq_start = pm.rm_eo;
          *line += pm.rm_eo;

          if(regexec(&exp, line, 1, &pm, 0) == REG_NOMATCH) continue;

          chan.frequency[0] = '\0';
          strcat(chan.frequency, res.substr(freq_start,5).c_str());

          config.push_back(chan);
        }
      }

    }

    int i = file.find_last_of("/");
    
    aliasFilename = file.substr(0, i+1) + "channel.alias";

  }

  void ChannelConfig::readIVTVChannels(string file){
    Channel chan;
    ifstream configfile(file.c_str());
    char line[41];

    if (!configfile) {
      GLOBAL_ERROR_STREAM("Channel file " << file << " not found" << endl);
      return;
    }

    while (!configfile.eof()) {
      configfile.getline(line, 40);
      if(line){
        regex_t exp;
        regmatch_t pm;

        if (!regcomp(&exp, "\\[.*?\\]", REG_EXTENDED)){
          if(regexec(&exp, line , 1 , &pm, 0) != REG_NOMATCH){
            //wich tag do we have?
            string res(line);
              
            if(res.substr(pm.rm_so+1, pm.rm_eo - pm.rm_so -2) == "defaults") {
              //if we are in defaults-section
              continue;
            }else if( res.substr(pm.rm_so+1, pm.rm_eo - pm.rm_so-2) == "global"){
              //if we are in global-section
              continue;
            }else{
              //ok, we have found a channel
              chan.name[0]='\0';
              strcat(chan.name, res.substr(pm.rm_so+1, pm.rm_eo - pm.rm_so-2).c_str()) ;
              chan.name[pm.rm_eo - pm.rm_so-2] = '\0';
              char freq[41];
              while(!configfile.eof()){
                configfile.getline(freq,40);
                regex_t freg;
                if(!regcomp(&freg, "channel = ", REG_EXTENDED)){
                  regmatch_t fpm;
                  if(regexec(&freg, freq, 1, &fpm, 0) != REG_NOMATCH){
                    //now we have found the frequency-part
                    res = string(freq);
                    chan.frequency[0] = '\0';
                    strcat(chan.frequency, res.substr(fpm.rm_eo).c_str());

                    config.push_back(chan);      
                    break;
                  }
                }   
              }
            }
          }
        }
      }  

    }  
    
    int i = file.find_last_of("/");
    
    aliasFilename = file.substr(0, i+1) + "channel.alias";
    //cerr << aliasFilename << endl;

      
  }
}

#include "nmm/nmm_suffix.hpp"
