/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "WhiteNoiseNode.hpp"

#include "nmm/comm/messaging/B_Allocator.hpp"
#include "nmm/comm/format/Property.hpp"
#include "nmm/plugins/NMMRegisterPlugin.hpp"

#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <iostream>

#define WAIT_SECS 0
#define WAIT_USECS 200000
#define WAIT_MILLIS 100

NMM_REGISTER_PLUGIN(WhiteNoiseNode, NMM, "WhiteNoiseNode")

namespace NMM {

  WhiteNoiseNode::WhiteNoiseNode(const char* name /* = "WhiteNoiseNode" */,
                                 StreamQueue::Mode mode /* = StreamQueue::MODE_SUSPEND */,
                                 int queue_size /* = StreamQueue::DEFAULT_SIZE */)
    : GenericSourceNode(name, mode, queue_size), __last_timer(), __condition(__cond_mutex)
  {
    stimer = new StreamTimer(this, StreamTimer::CONST_RATE);
  }
  
  WhiteNoiseNode::~WhiteNoiseNode() {
    delete stimer;
  }
  
  void WhiteNoiseNode::addSupportedFormats(Property* property) {
    NMM::Format *myOutputFormat;
    myOutputFormat=property->addNewFormat("video/raw");
    myOutputFormat->addIntParamValue(Format::x_resolution_param, 320);
    myOutputFormat->addIntParamValue(Format::y_resolution_param, 240);
    myOutputFormat->addIntParamValue(Format::ratio_x_param, 320);
    myOutputFormat->addIntParamValue(Format::ratio_y_param, 240);
    myOutputFormat->addStringParamValue(Format::colorspace_param,"rgb24");
    myOutputFormat->addStringParamValue(Format::channelorder_param,"rgb");
    myOutputFormat->addIntParamValue(Format::bitperpixel_param,32);
    myOutputFormat->addStringParamValue(Format::endian_param, "little");
    myOutputFormat->addStringParamValue(Format::format_param, "packet");
    myOutputFormat->addFloatParamValue(Format::framerate_param, 15.0);
    property->setDefaultFormat(myOutputFormat);

    myOutputFormat=property->addNewFormat("video/raw");
    myOutputFormat->addWildcard(Format::x_resolution_param);
    myOutputFormat->addWildcard(Format::y_resolution_param);
    myOutputFormat->addWildcard(Format::ratio_x_param);
    myOutputFormat->addWildcard(Format::ratio_y_param);
    myOutputFormat->addStringParamValue(Format::colorspace_param,"rgb24");
    myOutputFormat->addStringParamValue(Format::channelorder_param,"rgb");
    myOutputFormat->addIntParamValue(Format::bitperpixel_param,32);
    myOutputFormat->addIntParamValue(Format::bitperpixel_param,24);
    myOutputFormat->addStringParamValue(Format::endian_param, "little");
    myOutputFormat->addStringParamValue(Format::format_param, "packet");
    myOutputFormat->addRFloatParamValue(Format::framerate_param, 0.0, 100.0);
    
    myOutputFormat=property->addNewFormat("video/raw");
    myOutputFormat->addWildcard(Format::x_resolution_param);
    myOutputFormat->addWildcard(Format::y_resolution_param);
    myOutputFormat->addWildcard(Format::ratio_x_param);
    myOutputFormat->addWildcard(Format::ratio_y_param);
    myOutputFormat->addStringParamValue(Format::colorspace_param,"yuv422");
    myOutputFormat->addStringParamValue(Format::channelorder_param,"uyvy");
    myOutputFormat->addIntParamValue(Format::bitperpixel_param,16);
    myOutputFormat->addStringParamValue(Format::endian_param, "little");
    myOutputFormat->addStringParamValue(Format::format_param, "packet");
    myOutputFormat->addRFloatParamValue(Format::framerate_param, 0.0, 100.0);
    
    myOutputFormat=property->addNewFormat("video/raw");
    myOutputFormat->addWildcard(Format::x_resolution_param);
    myOutputFormat->addWildcard(Format::y_resolution_param);
    myOutputFormat->addWildcard(Format::ratio_x_param);
    myOutputFormat->addWildcard(Format::ratio_y_param);
    myOutputFormat->addStringParamValue(Format::colorspace_param,"yuv411");
    myOutputFormat->addStringParamValue(Format::channelorder_param,"yvu");
    myOutputFormat->addIntParamValue(Format::bitperpixel_param,16);
    myOutputFormat->addStringParamValue(Format::endian_param, "little");
    myOutputFormat->addStringParamValue(Format::format_param, "planar");
    myOutputFormat->addRFloatParamValue(Format::framerate_param, 0.0, 100.0);
  }
  
  Result WhiteNoiseNode::doInit() {
    addSupportedFormats(getOwnOutputProperty());
    return SUCCESS;
  }
  
  Result WhiteNoiseNode::doInitOutput() {
    addSupportedFormats(getOwnWorkingOutputProperty());
    return SUCCESS;
  }
  
  Result WhiteNoiseNode::doActivate() {
    if (getOutputFormat() == NULL) { return FAILURE; }
    image_width = getOutputFormat()->getIntValue(Format::x_resolution_param);
    image_height = getOutputFormat()->getIntValue(Format::y_resolution_param);
    image_bpp = getOutputFormat()->getIntValue(Format::bitperpixel_param);
    image_fps = getOutputFormat()->getFloatValue(Format::framerate_param);
    DEBUG_STREAM("fps is " << image_fps << endl);
    DEBUG_STREAM("bitsperpixel is " << image_bpp << endl);
    
    stimer->reset();
    stimer->setRate(image_fps);
    
    return SUCCESS;
  }

  Result WhiteNoiseNode::doStop() {
    return SUCCESS;
  }
  
  Message *WhiteNoiseNode::processBuffer(Buffer*) {
    if(__last_timer == Time()) {
      /* reset the timer */
      __last_timer = getTime();
    }

    if(getOutputFormat()->getStringValue(Format::colorspace_param) == "yuv422") {
      return produceYUVBuffer();
    } else {
      if(getOutputFormat()->getStringValue(Format::colorspace_param) == "yuv411" ||
	 getOutputFormat()->getStringValue(Format::colorspace_param) == "yv12") {
	return produceYV12Buffer();
      } else {
	return produceRGBBuffer();
      }
    }
  }

  void WhiteNoiseNode::waitToSendBuffer() {
    /* Begin> Calculate the Timestamp of the next Buffer */
    Interval delta;
    delta.nsec = static_cast<nmm_int64>(1000000000L/image_fps);
    /* End< */

    /* wait until next frame */
    __condition.timedWait(__last_timer.sec, __last_timer.nsec);
  }

  Buffer *WhiteNoiseNode::produceRGBBuffer() {
    Buffer *out = NULL;
    long seed = 119;
    int bpp = (image_bpp = 32)?4:3;


    if ((out = getNewBuffer(image_width*image_height*bpp))!=0) {
      char *out_buffer = out->getData();
      for (int i=0; i<image_width*image_height; i++) {
	//char value = (char)(rand() & 0xff);
	char value = (char) seed;
        seed = ((seed * seed) >> 8) + i + out_buffer - (char*)(0);
	out_buffer[0] = value;
	out_buffer[1] = value;
	out_buffer[2] = value;
	if (bpp == 4) ((unsigned char*)out_buffer)[3] = 255;
	out_buffer += bpp;
      }
      out->setUsed(image_width*image_height*bpp);
    }
    else {
      ERROR_STREAM( " Could not get out buffer." << endl);
    }

    /* wait until next frame must be send */
    waitToSendBuffer();

    /* set the new timestamp */
    stimer->setTimestamp(out);
   
    return out;
  }

  Buffer * WhiteNoiseNode::produceYUVBuffer() 
  {
    Buffer *out = NULL;
    int seed = 119;

    if ((out = getNewBuffer(image_width*image_height*2)) != 0) {
      char *out_buffer = out->getData();
      for (int i=0; i<image_width*image_height/2; i++) {
	//char value = (char)(rand() % 255);
	char value = (char) seed;
        seed = ((seed * seed) >> 8) + i + out_buffer - (char*)(0);
	out_buffer[0] = 127;  // U/V
	out_buffer[1] = value;  // Y0
	out_buffer[2] = 127;  // V/U
	out_buffer[3] = value;  // Y1
	out_buffer += 4; // filled TWO pixels!
      }
      out->setUsed(image_width*image_height*2);
    }
    else {
      ERROR_STREAM( " Could not get out buffer." << endl);
    }

    /* wait until next frame must be send */
    waitToSendBuffer();

    /* set the new timestamp */
    stimer->setTimestamp(out);

    return out;
  }

  Buffer * WhiteNoiseNode::produceYV12Buffer() 
  {
    Buffer *out = NULL;
    int seed = 119;

    if ((out = getNewBuffer(image_width*image_height + (image_width*image_height)/2)) != 0) {
      char *out_buffer = out->getData();
      // fill u and v plane with 127, use 0 for green white-noise
      memset(out_buffer + (image_width*image_height), 127, (image_width*image_height)/2);
      for (int i=0; i<image_width*image_height; i++) {
	//char value = (char)(rand() % 255);
	char value = (char) seed;
        seed = ((seed * seed) >> 8) + i + out_buffer - (char*)(0);
	out_buffer[0] = value;  // Y0
	out_buffer++;
      }
      out->setUsed(image_width*image_height*2);
    }
    else {
      ERROR_STREAM( " Could not get out buffer." << endl);
    }

    /* wait until next frame must be send */
    waitToSendBuffer();

    /* set the new timestamp */
    stimer->setTimestamp(out);

    return out;
  }
}
