/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include "CompositeNode.hpp"

#include "nmm/utils/thread/MutexGuard.hpp"
#include "nmm/comm/ProxyObject.hpp"
#include "nmm/comm/serialize/ValueFactory.hpp"
#include "nmm/comm/OutOfBandCompositeStrategy.hpp"
#include "nmm/multimedia/NodeDescription.hpp"
#include "nmm/services/registry/GraphDescription.hpp"
#include "nmm/services/registry/ClientRegistry.hpp"

// XXX: fw: needed for native windows
#undef interface

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  /****************************** local helpers ******************************/
  ostream& operator<<(ostream& out, const CompositeNode::TransitionTargetState & state) {
    switch (state) {
    case CompositeNode::INITIALIZED:
      return out << string("INITIALIZED");
      break;
    case CompositeNode::OUTPUT_INITIALIZED:
      return out << string("OUTPUT_INITIALIZED");
      break;
    case CompositeNode::ACTIVATED:
      return out << string("ACTIVATED");
      break;
    case CompositeNode::STARTED:
      return out << string("STARTED");
      break;
    case CompositeNode::STOPPED:
      return out << string("STOPPED");
      break;
    case CompositeNode::DEACTIVATED:
      return out << string("DEACTIVATED");
      break;
    case CompositeNode::OUTPUT_DEINITIALIZED:
      return out << string("OUTPUT_DEINITIALIZED");
      break;
    case CompositeNode::DEINITIALIZED:
      return out << string("DEINITIALIZED");
      break;
    }

    return out;
  }

  ostream& operator<<(ostream& out, const CompositeNode::JackType & dir) {
    switch (dir) {
    case CompositeNode::STATIC_INPUT:
      return out << string("STATIC_INPUT");
      break;
    case CompositeNode::STATIC_OUTPUT:
      return out << string("STATIC_OUTPUT");
      break;
    case CompositeNode::WORKING_INPUT:
      return out << string("WORKING_INPUT");
      break;
    case CompositeNode::WORKING_OUTPUT:
      return out << string("WORKING_OUTPUT");
      break;
    }

    return out;
  }

  CompositeNode::JackType operator++(CompositeNode::JackType& type, int) {
    CompositeNode::JackType retval = type;

    switch (type) {
    case CompositeNode::STATIC_INPUT:
      type = CompositeNode::WORKING_INPUT;
      break;
    case CompositeNode::STATIC_OUTPUT:
      type = CompositeNode::WORKING_OUTPUT;
      break;
    case CompositeNode::WORKING_INPUT:
      type = CompositeNode::STATIC_OUTPUT;
      break;
    case CompositeNode::WORKING_OUTPUT:
      type = CompositeNode::STATIC_OUTPUT;
      break;
    }

    return retval;
  }

  list<string> get_event_list(const list<string>& event_list, const list<Interface*> iface_list) {
    list<string> retval;
    for (list<Interface*>::const_iterator iface = iface_list.begin();
         iface != iface_list.end();
         iface++) {
      string iface_type((*iface)->getInterfaceType());
      for (list<string>::const_iterator event = event_list.begin();
           event != event_list.end();
           event++) {
        if (event->substr(0, iface_type.size()) == iface_type) {
          retval.push_back(*event);
        }
      }
    }
    return retval;
  }
  /****************************** local helpers ******************************/

  /****************************** C'tor/D'tor ******************************/
  CompositeNode::CompositeNode(const char* name /*= "CompositeNode"*/, ClientRegistry* reg /*= NULL*/)
    : Node(name),
      ICompositeNodeImpl(),
      registry(reg),
      nodes(),
      input_jacks(),
      output_jacks(),
      register_receiver_mutex(),
      forwarding_events(),
      reach_state(false),
      __graph_description(NULL) {

    setType(INode::COMPOSITE);
  }

  CompositeNode::~CompositeNode() {
    for (map<unsigned int, CNNode>::iterator i = nodes.begin();
         i != nodes.end();
         i++) {
      removeNode(i->first, true);
    }
    nodes.clear();

    delete __graph_description;
  }
  /****************************** C'tor/D'tor ******************************/


  /****************************** ICompositeNode Interface ******************************/
  nmm_uint32 CompositeNode::addNode(const INode& node) {
    return addNodeWithInterfaces(node, node.getParentObject()->getInterfaceList());
  }

  bool CompositeNode::addNodeWithID(const INode& node, const unsigned int& node_id) {
    return addNodeWithIDWithInterfaces(node, node_id, node.getParentObject()->getInterfaceList());
  }

  nmm_uint32 CompositeNode::addNodeWithInterfaces(const INode & node, const list<Interface*> & iface_list) {
    static unsigned int current_new_node_id(0);

    current_new_node_id++;

    if (!addNodeWithIDWithInterfaces(node, current_new_node_id, iface_list)) {
      ERROR_STREAM("Adding new Node with generated ID " << current_new_node_id << " failed!" << endl);
      current_new_node_id--;
      return 0;
    }

    return current_new_node_id;
  }

  bool CompositeNode::addNodeWithIDWithInterfaces(const INode& node, const unsigned int& id, const list<Interface*> & iface_list) {
    if (nodes.find(id) != nodes.end()) {
      ERROR_STREAM("Node with ID " << id << " already exists in this CN!" << endl);
      return false;
    }

    /* Begin> If node was already added return false */
    for (map<unsigned int, CNNode>::iterator it = nodes.begin(); it != nodes.end(); ++it) {
      if ( it->second.getNode().getParentObject() == node.getParentObject() ) {
        WARNING_STREAM("Node already inserted into CompositeNode and is not added again!" << endl);
        return false;
      }
    }
    /* End< */
    
    // make a copy of the INode. This is important as the given INode&
    // is only valid within this method. As we have to save it in our
    // internal nodes map (as addNode may suggest :-) ), we have to
    // copy it.
    INode& real_node(*(dynamic_cast<INode*>(node.copy())));
    list<Interface*> node_iface_list(iface_list);
    list<string> node_event_list(get_event_list(real_node.getParentObject()->getEventList(), iface_list));
    unsigned int node_id(id);

    CNNode cnnode(real_node, node_iface_list, node_event_list);

    pair<map<unsigned int, CNNode>::iterator, bool> res =
      nodes.insert(pair<unsigned int, CNNode>(node_id, cnnode));

    if (!res.second) {
      ERROR_STREAM("Could not add Node!" << endl);
      return false;
    }

    for (list<Interface*>::const_iterator iface = node_iface_list.begin();
         iface != node_iface_list.end();
         iface++) {
      addCNInterface((*iface)->getTypeInfo());
    }

    for (list<string>::const_iterator event = node_event_list.begin();
         event != node_event_list.end();
         event++) {
      addCNEvent(*event);
    }

    updateRemoteInterfaceLists();

    return true;
  }

  Result CompositeNode::renameInputJackTag(const string & old_cn_jack_tag, const string & new_cn_jack_tag) {
    return renameJackTag(INPUT, old_cn_jack_tag, new_cn_jack_tag);    
  }

  Result CompositeNode::renameOutputJackTag(const string & old_cn_jack_tag, const string & new_cn_jack_tag) {
    return renameJackTag(OUTPUT, old_cn_jack_tag, new_cn_jack_tag);
  }


  void CompositeNode::registerEventListener(IEventReceiver* receiver) {
    DEBUG_STREAM("CompositeNode::registerEventListener(IEventReceiver*)" << endl);
    MutexGuard mg(register_receiver_mutex);
    Node::registerEventListener(receiver);

    IEventReceiver* real_receiver(receiver);
    if (receiver) {
      multimap<IEventReceiver*, ProxyEventReceiver>::iterator insert = 
        proxy_receivers.insert(pair<IEventReceiver*, ProxyEventReceiver>(receiver, 
                                                                         ProxyEventReceiver(*receiver)));
      real_receiver = &insert->second;
    }

    set<INode*> registered_nodes;
    for (CNJackMap_t::iterator i = input_jacks.begin();
         i != input_jacks.end();
         i++) {
      if (registered_nodes.find(i->second.node) == registered_nodes.end()) {
        registered_nodes.insert(i->second.node);
        i->second.node->getParentObject()->registerEventListener(real_receiver);
      } else {
        WARNING_STREAM("Already registered " << i->first << endl);
      }
    }
    for (CNJackMap_t::iterator i = output_jacks.begin();
         i != output_jacks.end();
         i++) {
      if (registered_nodes.find(i->second.node) == registered_nodes.end()) {
        registered_nodes.insert(i->second.node);
        i->second.node->getParentObject()->registerEventListener(real_receiver);
      } else {
        WARNING_STREAM("Already registered " << i->first << endl);
      }
      break; //BUG:break after first iteration ?!?!
    }
  }

  void CompositeNode::registerEventListener(const string& event_key, IEventReceiver* receiver) {
    DEBUG_STREAM("CompositeNode::registerEventListener(const string& " << event_key << ", IEventReceiver* receiver);" << endl);
    
    Node::registerEventListener(event_key, receiver);

    /* Forward event listener to all nodes */
    for( map<unsigned int, CNNode>::iterator it = nodes.begin(); it != nodes.end(); ++it) {
      it->second.getNode().getParentObject()->registerEventListener(event_key, receiver);
    }
  }

  void CompositeNode::registerEventListener(const string& event_key, EDObject* edo) {
    DEBUG_STREAM("CompositeNode::registerEventListener(const string& " << event_key << ", EDObject* edo);" << endl);
    Node::registerEventListener(event_key, edo);

    /* Forward event listener to all nodes */
    for( map<unsigned int, CNNode>::iterator it = nodes.begin(); it != nodes.end(); ++it) {
      it->second.getNode().getParentObject()->registerEventListener(event_key, edo);
    }
  }

  void CompositeNode::removeEventListener(IEventReceiver* receiver) {
    MutexGuard mg(register_receiver_mutex);
    Node::removeEventListener(receiver);

    IEventReceiver* real_receiver(receiver);
    multimap<IEventReceiver*, ProxyEventReceiver>::iterator search = proxy_receivers.find(receiver);
    if (proxy_receivers.end() != search) {
      real_receiver = &search->second;
    } else {
      ERROR_STREAM("Trying to remove a non-registered receiver! Ignoring!" << endl);
      return;
    }

    DEBUG_STREAM("CompositeNode::removeEventListener(IEventReceiver*)" << endl);
    for (CNJackMap_t::iterator i = input_jacks.begin();
         i != input_jacks.end();
         i++) {
      i->second.node->getParentObject()->removeEventListener(real_receiver);
    }
    for (CNJackMap_t::iterator i = output_jacks.begin();
         i!= output_jacks.end();
         i++) {
      i->second.node->getParentObject()->removeEventListener(real_receiver);
    }

    proxy_receivers.erase(search);
  }

  void CompositeNode::removeEventListener(const string& event_key, IEventReceiver* receiver) {
    DEBUG_STREAM("CompositeNode::removeEventListener(const string& event_key, IEventReceiver* receiver);" << endl);
    Node::removeEventListener(event_key, receiver);

    /* remove event listener from all nodes */
    for( map<unsigned int, CNNode>::iterator it = nodes.begin(); it != nodes.end(); ++it) {
      it->second.getNode().getParentObject()->removeEventListener(event_key, receiver);
    }
  }

  void CompositeNode::removeEventListener(EDObject* edo) {
    DEBUG_STREAM("CompositeNode::removeEventListener(EDObject* edo);" << endl);
    Node::removeEventListener(edo);

    /* remove event listener from all nodes */
    for( map<unsigned int, CNNode>::iterator it = nodes.begin(); it != nodes.end(); ++it) {
      it->second.getNode().getParentObject()->removeEventListener(edo);
    }
  }

  Result CompositeNode::receiveEvent(Event& event, const IEventSender* sender) {
    return receiveEventInternal(event, sender);
  }
  
  Event* CompositeNode::getEventPrototype(const string& event_key) const {
    if (forwarding_events.find(event_key) == forwarding_events.end()) {
      return Node::getEventPrototype(event_key);
    } else {
      DEBUG_STREAM("CompositeNode::getEventPrototype(" << event_key << ")" << endl);

      for (map<unsigned int, CNNode>::const_iterator n = nodes.begin();
           n != nodes.end();
           n++) {
        ProxyObject& nodepo = *n->second.getNode().getParentObject();
        list<string> node_event_list(n->second.getEvents());
        list<string>::const_iterator search;
        for (search = node_event_list.begin(); search != node_event_list.end(); search++) {
          if (*search == event_key) {
            break;
          }
        }
        if (search != node_event_list.end()) {
          return nodepo.getEventPrototype(event_key);
        }
      }

      ERROR_STREAM("CompositeNode::getEventPrototype(" << event_key << "): failed" << endl);

      return NULL;
    }
  }

  list<string> CompositeNode::getEventList() const {
    const list<string> internal_events(Node::getEventList());
    list<string> retval(internal_events);
    for (set<string>::const_iterator event = forwarding_events.begin();
         event != forwarding_events.end();
         event++) {
      retval.push_back(*event);
    }
    return retval;
  }
  /****************************** ICompositeNode Interface ******************************/



  /****************************** private methods ******************************/
  Result CompositeNode::doStateTransitionTo(TransitionTargetState target_state)
    throw (const StateTransitionFailedException&,
	   const IllegalStateException&,
           const UnsuccessfulRequiredCallException&) {
    if (!reach_state) {
      // determine the state, the nodes must have after the
      // transition is complete
      INode::State node_state;
    
      switch (target_state) {
      case INITIALIZED:
        node_state = INode::INITIALIZED;
        break;
      case OUTPUT_INITIALIZED:
        node_state = INode::OUTPUT_INITIALIZED;
        break;
      case ACTIVATED:
        node_state = INode::ACTIVATED;
        break;
      case STARTED:
        node_state = INode::STARTED;
        break;
      case STOPPED:
        node_state = INode::ACTIVATED;
        break;
      case DEACTIVATED:
        node_state = INode::OUTPUT_INITIALIZED;
        break;
      case OUTPUT_DEINITIALIZED:
        node_state = INode::INITIALIZED;
        break;
      case DEINITIALIZED:
        node_state = INode::CONSTRUCTED;
        break;
      }
    
      for (map<unsigned int, CNNode>::iterator i = nodes.begin(); i!= nodes.end(); i++) {
        INode& current_node = i->second.getNode();  // for convinience

        if (current_node.getState() == node_state) { // ok, node is already ready
          continue;
        }

        try {
          // determine the state transition method to call from the node
          switch (target_state) {
          case INITIALIZED:
            if (current_node.isConstructed()) {
              current_node.init();
            }
            break;
          case OUTPUT_INITIALIZED:
            if (current_node.isInitialized()) {
              current_node.initOutput();
            }
            break;
          case ACTIVATED:
            if (current_node.isOutputInitialized()) {
              current_node.activate();
            }
            break;
          case STARTED:
            current_node.start();
            break;
          case STOPPED:
            current_node.stop();
            break;
          case DEACTIVATED:
            current_node.deactivate();
            break;
          case OUTPUT_DEINITIALIZED:
            current_node.deinitOutput();
            break;
          case DEINITIALIZED:
            current_node.deinit();
            break;
          }
        } catch (const StateTransitionFailedException & e) {
          ERROR_STREAM("in doStateTransitionTo(): " << e << endl);
          // One of the nodes can't reach state node_state, so the CN
          // can't also hence do<StateTransition> must return FAILURE.
          return FAILURE;
        } catch (const IllegalStateException & e) {
          ERROR_STREAM("in doStateTransitionTo(): " << e << endl);
          // oops, one of the nodes is cracy:
          // perhaps someone has externly changed its state?
          ERROR_STREAM("Node " << 
                       i->first << 
                       " is not in its expected state: " << 
                       current_node.getState() << endl);
          throw IllegalStateException(this, 
                                      getName() + 
                                      string(">Wrong state for this transition in subnode ") + 
                                      i->first);
        } catch (...) {
          // oops, what's going up now, unexpected exception?
          ERROR_STREAM("Unexpected exception!" << endl);
          throw StateTransitionFailedException("CompositeNode: Unexpected exception in method CompositeNode::doStateTransitionTo() ");
        }
      }
    } // end> (!reach_state)

    return SUCCESS;
  }


  void CompositeNode::removeNode(const unsigned int & node_id, bool do_release /* = false */) {
    map<unsigned int, CNNode>::iterator search = nodes.find(node_id);
    
    INode& node = search->second.getNode();
    // just to make sure, delete will work
    node.reachConstructed();

    if (do_release) { // this node is managed by the registry, so release it!
      if (registry) {
        DEBUG_STREAM("Releasing " << node_id << endl);
        registry->releaseNode(node);
        DEBUG_STREAM("Released " << node_id << endl);
      }
    }

    DEBUG_STREAM("Deleting " << node_id <<"'s INode" << endl);
    delete &node;
  }


  void CompositeNode::addAllFormats(const JackType dir, const unsigned int & node_id) {
    DEBUG_STREAM("Adding formats in for " << dir << " of this " << node_id << endl);

    map<unsigned int, CNNode>::iterator search = nodes.find(node_id);
    if (search == nodes.end()) {
      ERROR_STREAM("Node " << node_id << " is not a part of this CN!" << endl);
      return;
    }
    INode& node = search->second.getNode();

    Property_var node_prop;
    list<string> node_jack_tags;
    Property* own_prop(NULL);
    CNJackMap_t* jack_map(NULL);

    switch (dir) {
    case STATIC_INPUT:
      node_prop.reset(node.getInputProperty());
      node_jack_tags = node.getInputStreamTags();
      own_prop = getOwnInputProperty();
      jack_map = &input_jacks;
      break;
    case STATIC_OUTPUT:
      node_prop.reset(node.getOutputProperty());
      node_jack_tags = node.getOutputStreamTags();
      own_prop = getOwnOutputProperty();
      //jack_map = &output_jacks;
      // do not add output streams when not using working objects
      break;
    case WORKING_INPUT:
      node_prop.reset(node.getWorkingInputProperty());
      node_jack_tags = node.getInputStreamTags();
      own_prop = getOwnWorkingInputProperty();
      //jack_map = &input_jacks;
      // do not add input streams when using working objects
      break;
    case WORKING_OUTPUT:
      node_prop.reset(node.getWorkingOutputProperty());
      node_jack_tags = node.getOutputStreamTags();
      own_prop = getOwnWorkingOutputProperty();
      jack_map = &output_jacks;
      break;
    }

    if (NULL == node_prop.get()) {
      DEBUG_STREAM("Node has no Property of Type " << dir << ". Skipping !" << endl);
      return;
    }

    const format_list fmts = node_prop->getFormatList();
    DEBUG_STREAM("Node has " << fmts.size() << " Formats!" << endl);

    if (node_jack_tags.empty() && (fmts.size() != 0)) {
      WARNING_STREAM("Node " << node_id << " has formats but no jack_tags for Type " << dir << "!" << endl);
      return;
    }

    for (format_list::const_iterator f = fmts.begin(); f != fmts.end(); f++) {
      Format* fmt = *f; // for convenience
      string jack_tag("");
      if (fmt->hasJackTag()) {
	jack_tag = fmt->getJackTag();
      } else {
	jack_tag = string("");
      }
      string cn_jack_tag = string() + node_id + string(".") + jack_tag;
      DEBUG_STREAM("Trying to add format with CNJackTag " << cn_jack_tag << endl);

      {
        bool jack_tag_found = false;
        for (list<string>::const_iterator i = node_jack_tags.begin();
             i != node_jack_tags.end();
             i++) {
          if (*i == jack_tag) {
            jack_tag_found = true;
            break;
          }
        }
        if (!jack_tag_found) {
          ERROR_STREAM("Jack " << jack_tag << " from " << node_id << " specified in Format, but not in stream tags of Type " << dir << ". Not using in CN" << endl);
          continue;
        }

        switch (dir) {
        case STATIC_INPUT:
        case WORKING_INPUT:
          if (node.isInputConnected(jack_tag)) {
            ERROR_STREAM("Jack " << jack_tag << " in direction " << dir << " from " << node_id << " already connected. Not exporting in CN!" << endl);
            continue;
          }
          break;
        case STATIC_OUTPUT:
        case WORKING_OUTPUT:
          if (node.isOutputConnected(jack_tag)) {
            ERROR_STREAM("Jack " << jack_tag << " in direction " << dir << " from " << node_id << " already connected. Not exporting in CN!" << endl);
            continue;
          }
          break;
        }
      }

      Format* added_fmt = new Format(*fmt);
      added_fmt->setJackTag(cn_jack_tag);
      own_prop->addNewFormat(added_fmt);
      const Format* def_fmt = node_prop->getDefaultFormat(jack_tag.c_str());
      if (NULL == def_fmt) {
	ERROR_STREAM("********************************************************************************" << endl);
	ERROR_STREAM("No default format for jack tag " << jack_tag << " of node " << node_id << " found! This will lastly fail!" << endl);
	ERROR_STREAM("********************************************************************************" << endl);
	cerr << "No default format for jack tag " << jack_tag << " of node " << node_id << " found! This will lastly fail!" << endl;
      } else {
	DEBUG_STREAM("Default format for jack tag " << jack_tag << ":" << endl << *def_fmt << endl);
	DEBUG_STREAM("Actual Format: " << endl << *fmt << endl);
	if (def_fmt == fmt) {
	  own_prop->setDefaultFormat(added_fmt, cn_jack_tag);
	  DEBUG_STREAM("DefaultFormat!" << endl);
        }
      }
      DEBUG_STREAM("Format added. Now adding jack to CN's jack list!" << endl);

      // insert the cn_jack in the map and test, if successful
      if (jack_map) {
        if (jack_map->end() == jack_map->find(cn_jack_tag)) {
          pair<CNJackMap_t::iterator, bool> res = 
            jack_map->insert(make_pair(cn_jack_tag, CNJack_t(jack_tag, &node)));
          if (!res.second) {
            ERROR_STREAM("Problems inserting Jack with tag " << cn_jack_tag << "!" << endl);
            continue;
          }
        }
      }
    } // for (format_list::const_iterator f = fmts.begin(); f != fmts.end(); f++)
  }


  void CompositeNode::removeAllFormats(const JackType dir, const unsigned int & node_id) {
    DEBUG_STREAM("Removing formats for Type " << dir << " of this Node " << node_id << endl);

    map<unsigned int, CNNode>::iterator search = nodes.find(node_id);
    if (search == nodes.end()) {
      ERROR_STREAM("Node " << node_id << " is not a part of this CN!" << endl);
      return;
    }
    INode& node = search->second.getNode();

    Property_var node_prop;
    Property* own_prop(NULL);
    CNJackMap_t* jack_map(NULL);
    switch (dir) {
    case STATIC_INPUT:
      node_prop.reset(node.getInputProperty());
      own_prop = getOwnInputProperty();
      jack_map = &input_jacks;
      break;
    case STATIC_OUTPUT:
      node_prop.reset(node.getOutputProperty());
      own_prop = getOwnOutputProperty();
      jack_map = &output_jacks;
      break;
    case WORKING_INPUT:
      node_prop.reset(node.getWorkingInputProperty());
      own_prop = getOwnWorkingInputProperty();
      jack_map = &input_jacks;
      break;
    case WORKING_OUTPUT:
      node_prop.reset(node.getWorkingOutputProperty());
      own_prop = getOwnWorkingOutputProperty();
      jack_map = &output_jacks;
      break;
    }

    if (NULL == node_prop.get()) {
      DEBUG_STREAM("Node has no Property of Type " << dir << ". Skipping !" << endl);
      return;
    }

    for (CNJackMap_t::iterator i = jack_map->begin(); i != jack_map->end(); i++) {
      const string cn_jack_tag = string(i->first); // for convenience
      CNJack_t cn_jack = i->second; // for convenience

      if (cn_jack.node != &node) {
	continue;
      }

      const format_list fmts = own_prop->getFormatList();
      for (format_list::const_iterator f = fmts.begin(); f != fmts.end(); f++) {
	Format* fmt = *f; // for convenience
	if (fmt->getJackTag() == cn_jack_tag) {
	  own_prop->deleteFormat(fmt);
	}
      }

      DEBUG_STREAM("Erasing Map-Entry..." << endl);
      jack_map->erase(i);
    }
  }


  const CompositeNode::CNJack_t CompositeNode::getCNJackFromJackTag(const JackType dir, const string & jack_tag) const {
    CNJackMap_t* jack_map(NULL);
    switch (dir) {
    case STATIC_INPUT:
    case WORKING_INPUT:
      jack_map = &input_jacks;
      break;
    case STATIC_OUTPUT:
    case WORKING_OUTPUT:
      jack_map = &output_jacks;
      break;
      break;
    }

    CNJackMap_t::const_iterator i = jack_map->find(jack_tag);
    if (jack_map->end() == i) {
      // ERRRORTODO: error handling
      return CNJack_t();
    }

    return i->second;
  }


  Result CompositeNode::renameJackTag(JackType jack_type, const string & old_cn_jack_tag, const string & new_cn_jack_tag) {
    // determine, in which map the jack should be added
    CNJackMap_t* jack_map;
    switch (jack_type) {
    case STATIC_INPUT:
    case WORKING_INPUT:
      jack_map = &input_jacks;
      break;
    case STATIC_OUTPUT:
    case WORKING_OUTPUT:
      jack_map = &output_jacks;
      break;
    }

    CNJackMap_t::iterator i = jack_map->find(old_cn_jack_tag);
    if (jack_map->end() == i) {
      ERROR_STREAM(old_cn_jack_tag << " not found!" << endl);
      return FAILURE;
    }

    // insert the cn_jack in the map and test, if successful
    DEBUG_STREAM("Adding new jack tag " << new_cn_jack_tag << " in direction " << jack_type << endl);
    pair<CNJackMap_t::iterator, bool> res = 
      jack_map->insert(make_pair(new_cn_jack_tag, i->second));
    if (!res.second) {
      ERROR_STREAM(new_cn_jack_tag << " not inserted!" << endl);
      return FAILURE;
    }
    DEBUG_STREAM("Removing old jack tag " << old_cn_jack_tag << endl);
    jack_map->erase(i);

    // iterate over STATIC_{IN,OUT}PUT and WORKING_{IN,OUT}PUT
    JackType dir = jack_type;
    for (unsigned int i = 0; i < 2; i++) {
      DEBUG_STREAM("Working on Property in direction " << dir << endl);
      const format_list* format_lst;
      Property* own_prop;
      switch (dir++) {
      case STATIC_INPUT:
        format_lst = &(getOwnInputProperty()->getFormatList());
        own_prop = getOwnInputProperty();
        break;
      case WORKING_INPUT:
        format_lst = &(getOwnWorkingInputProperty()->getFormatList());
        own_prop = getOwnWorkingInputProperty();
        break;
      case STATIC_OUTPUT:
        format_lst = &(getOwnOutputProperty()->getFormatList());
        own_prop = getOwnOutputProperty();
        break;
      case WORKING_OUTPUT:
        format_lst = &(getOwnWorkingOutputProperty()->getFormatList());
        own_prop = getOwnWorkingOutputProperty();
        break;
      }

      for (format_list::const_iterator fmt = format_lst->begin();
           fmt != format_lst->end();
           fmt++) {
        if ((*fmt)->getJackTag() == old_cn_jack_tag) {
          DEBUG_STREAM("Setting jack tag of Format " << (*fmt)->getFormatType() << " from " << old_cn_jack_tag << " to " << new_cn_jack_tag << endl);
          (*fmt)->setJackTag(new_cn_jack_tag);
        }
      }
      // don't forget the default formats!!!
      const Format* old_def_fmt = own_prop->getDefaultFormat(old_cn_jack_tag);
      if (old_def_fmt) {
        DEBUG_STREAM("Replacing Default Format!" << endl);
        Format* ren_fmt = new Format(*old_def_fmt);
        own_prop->deleteFormat(const_cast<Format*>(old_def_fmt));
        own_prop->addNewFormat(ren_fmt);
        own_prop->setDefaultFormat(ren_fmt, string(new_cn_jack_tag));
      } else {
        WARNING_STREAM("No default Format for " << old_cn_jack_tag << " in " << dir << endl);
      }
    }

    return SUCCESS;
  }

  void CompositeNode::addCNInterface(const TypeInfo & iface_type) {
    bool found_iface(false);

    const list<Interface*> own_interfaces(getInterfaceList());
    for (list<Interface*>::const_iterator i = own_interfaces.begin();
         i != own_interfaces.end();
         i++) {
      if ((*i)->getTypeInfo() == iface_type) {
        found_iface = true;
        break;
      }
    }

    if (!found_iface) {
      Interface* interface = SValueFactory::getInstance().createType<Interface>(iface_type);
      if (interface) {
        DEBUG_STREAM("Adding Interface " << iface_type << endl);
        interface->setEventReceiver(*this);
        addInterface(interface);
      }
      else {
        ERROR_STREAM("Failed to create Interface with type " << iface_type << endl);
      }
    }
  }

  void CompositeNode::addCNEvent(const string & event_key) {
    bool found_event(false);
    const list<string> own_events(Node::getEventList());
    for (list<string>::const_iterator i = own_events.begin();
         i != own_events.end();
         i++) {
      if ((*i) == event_key) {
        found_event = true;
        break;
      }
    }

    if (!found_event && (forwarding_events.find(event_key) == forwarding_events.end())) {
      forwarding_events.insert(event_key);
    }
  }

  Result CompositeNode::receiveEventInternal(Event & event, const IEventSender* sender) {
    if (forwarding_events.find(event.getKey()) == forwarding_events.end()) {
      return Node::receiveEvent(event, sender);
    } else {
      DEBUG_STREAM("Forwarding " << event.getKey() << endl);

      Result retval(SUCCESS);
      for (map<unsigned int, CNNode>::const_iterator n = nodes.begin();
           n != nodes.end();
           n++) {
        ProxyObject& nodepo = *n->second.getNode().getParentObject();
        list<string> node_event_list(n->second.getEvents());
        list<string>::const_iterator search;
        for (search = node_event_list.begin(); search != node_event_list.end(); search++) {
          if (*search == event.getKey()) {
            break;
          }
        }
        if (search != node_event_list.end()) {
          Result tmpresult = nodepo.receiveEvent(event, sender);
          if (tmpresult != SUCCESS) {
            ERROR_STREAM("Forwarding " << event.getKey() << " to " << n->first << " leaded to " << tmpresult << endl);
            if (retval == SUCCESS) {
              retval = tmpresult;
            }
          }
        }
      }

      return retval;
    }
  }

  void CompositeNode::updateRemoteInterfaceLists() const {
    lockStrategyList();

    try {
      const list<OutOfBandCompositeStrategy*>& strategy_list(getStrategyList());
      if (strategy_list.size() == 0) {
        cerr << "No ProxyObject's to update!" << endl;
      }
      for (list<OutOfBandCompositeStrategy*>::const_iterator channel = strategy_list.begin();
           channel != strategy_list.end();
           channel++) {
        if (!*channel) {
          WARNING_STREAM("Got a NULL-Ptr OutOfBandChannel in NMMObject::getStrategyList()!" << endl);
        } else {
          const OutOfBandCompositeStrategy* ts = *channel;
          if (!ts) {
            WARNING_STREAM("TransportStrategy of OutOfBandChannel is NULL, skipping!" << endl);
          } else {
            DEBUG_STREAM("Updating ProxyObject's InterfaceList" << endl);
            Event event("ProxyObject::updateInterfaces");
            const_cast<OutOfBandCompositeStrategy*>(ts)->receiveEvent(event);
            DEBUG_STREAM("Updated ProxyObject's InterfaceList" << endl);
          }
        }
      }
    } catch (const Exception& e) {
      cerr << "Something when wrong, when updating Interface lists at connected ProxyObjects:" << endl
           << e << endl
           << "Ignoring!" << endl;
    } catch (...) {
      cerr << "Something STRANGE when wrong, when updating Interface lists at connected ProxyObjects!" << endl
           << "Ignoring!" << endl;
    }

    unlockStrategyList();
  }
  /****************************** private methods ******************************/


  /************************* INode Interface ******************/
  /*INode* CompositeNode::getPredecessorNode(const string & in_jack) const {
    const CNJack_t cn_jack = getCNJackFromJackTag(INPUT, in_jack);
    return cn_jack.node->getPredecessorNode(cn_jack.jack_tag);
    }*/

  /*INode* CompositeNode::getSuccessorNode(const string & out_jack) const {
    const CNJack_t cn_jack = getCNJackFromJackTag(OUTPUT, out_jack);
    return cn_jack.node->getSuccessorNode(cn_jack.jack_tag);
    }*/

  IInputJack* CompositeNode::getInputJack(const Format & format, const string & tag) { 
    CNJack_t cn_jack = getCNJackFromJackTag(INPUT, tag); 
    return cn_jack.node->getInputJack(format, cn_jack.jack_tag); 
  }

  IOutputJack* CompositeNode::getOutputJack(const Format & format, const string & tag) { 
    CNJack_t cn_jack = getCNJackFromJackTag(OUTPUT, tag); 
    return cn_jack.node->getOutputJack(format, cn_jack.jack_tag);
  }

  void CompositeNode::block()
  {
    // FIXME
    ERROR_STREAM2("Not implemented! ..." << endl);
  }
  
  void CompositeNode::unblock()
  {
    // FIXME
    ERROR_STREAM2("Not implemented! ..." << endl);
  }

  list<string> CompositeNode::getInputStreamTags() const {
    list<string> retval;
    for (CNJackMap_t::iterator i = input_jacks.begin();
         i != input_jacks.end();
         i++) {
      retval.push_back(i->first);
    }
    return retval;
  }

  list<string> CompositeNode::getOutputStreamTags() const {
    list<string> retval;
    for (CNJackMap_t::iterator i = output_jacks.begin();
         i != output_jacks.end();
         i++) {
      retval.push_back(i->first);
    }
    return retval;
  }

  IOutputJack* CompositeNode::cloneOutputJack(Format*& intersec_format, const string& tag)
  {
    const CNJack_t cn_jack = getCNJackFromJackTag(INPUT, tag);
    return cn_jack.node->cloneOutputJack(intersec_format, cn_jack.jack_tag); 
  }

  IInputJack* CompositeNode::getInputStream(const string & tag) {
    const CNJack_t cn_jack = getCNJackFromJackTag(INPUT, tag); 
    return cn_jack.node->getInputStream(cn_jack.jack_tag); 
  }

  IOutputJack* CompositeNode::getOutputStream(const string & tag) {
    const CNJack_t cn_jack = getCNJackFromJackTag(OUTPUT, tag); 
    return cn_jack.node->getOutputStream(cn_jack.jack_tag); 
  }

  void CompositeNode::reachState(const INode::State& state) 
    throw(const StateTransitionFailedException&, const UnsuccessfulRequiredCallException&)
  {
    DEBUG_STREAM("Begin> reachState(" << state << ")" << endl);
    reach_state = true;
    if(state != INode::STARTED) {
      for (map<unsigned int, CNNode>::iterator node = nodes.begin(); node != nodes.end(); node++) {
	node->second.getNode().reachState(state);
      }
    }
    else { // handle prefetching for started state
      list<INode*> prefetching_nodes;

      // first, start all nodes that have no prefetching enabled 
      for (map<unsigned int, CNNode>::iterator node = nodes.begin(); node != nodes.end(); node++) {
	INode* n = &(node->second.getNode());
	if(n->getPrefetchQueueSize() == 0) {
	  n->reachState(state);
	}
	else {
	  prefetching_nodes.push_back(n);
	}
      }

      // then, wait for prefetching nodes to finish prefetching
      for(list<INode*>::iterator j = prefetching_nodes.begin(); j != prefetching_nodes.end(); ++j) {
	(*j)->waitPrefetchFinished();
      }

      // finally, if all these nodes have finished prefetching, start them 
      for(list<INode*>::iterator j = prefetching_nodes.begin(); j != prefetching_nodes.end(); ++j) {
	(*j)->reachState(state);
      }
    }
    Node::reachState(state);
    reach_state = false;
    DEBUG_STREAM("End< reachState(" << state << ")" << endl);
  }
  /************************* INode Interface ******************/


  /*************************   Node interface; do-methods   ******************/
  Result CompositeNode::doInit()
    throw (const StateTransitionFailedException&, 
	   const IllegalStateException&, const UnsuccessfulRequiredCallException&) { 
    Result retval = doStateTransitionTo(INITIALIZED);

    for (map<unsigned int, CNNode>::const_iterator n = nodes.begin(); n != nodes.end(); n++) {
      addAllFormats(STATIC_INPUT,  n->first);
      addAllFormats(STATIC_OUTPUT, n->first);
    }

    return retval;
  }


  Result CompositeNode::doInitOutput()
    throw (const StateTransitionFailedException&, 
	   const IllegalStateException&,
           const UnsuccessfulRequiredCallException&) { 
    Result retval = doStateTransitionTo(OUTPUT_INITIALIZED);

    for (map<unsigned int, CNNode>::const_iterator n = nodes.begin(); n != nodes.end(); n++) {
      addAllFormats(WORKING_INPUT, n->first);
      addAllFormats(WORKING_OUTPUT, n->first);
    }

    return retval;
  }


  Result CompositeNode::doActivate()
    throw (const StateTransitionFailedException&, 
           const IllegalStateException&,
           const UnsuccessfulRequiredCallException&) { 
    return doStateTransitionTo(ACTIVATED); 
  }

  Result CompositeNode::doStart()
    throw (const StateTransitionFailedException&, 
           const IllegalStateException&,
           const UnsuccessfulRequiredCallException&) { 
    return doStateTransitionTo(STARTED); 
  }

  Result CompositeNode::doStop()
    throw (const StateTransitionFailedException&, 
           const IllegalStateException&,
           const UnsuccessfulRequiredCallException&) { 
    return doStateTransitionTo(STOPPED); 
  }

  Result CompositeNode::doDeactivate()
    throw (const StateTransitionFailedException&, 
           const IllegalStateException&,
           const UnsuccessfulRequiredCallException&) { 
    return doStateTransitionTo(DEACTIVATED); 
  }


  Result CompositeNode::doDeinitOutput()
    throw (const StateTransitionFailedException&, 
	   const IllegalStateException&,
           const UnsuccessfulRequiredCallException&) { 
    Result retval = doStateTransitionTo(OUTPUT_DEINITIALIZED); 

    for (map<unsigned int, CNNode>::const_iterator n = nodes.begin(); n != nodes.end(); n++) {
      removeAllFormats(WORKING_INPUT,  n->first);
      removeAllFormats(WORKING_OUTPUT, n->first);
    }

    DEBUG_STREAM("StateTrasition! Available Jacks are: " << endl);
    for (CNJackMap_t::const_iterator i = input_jacks.begin(); i != input_jacks.end(); i++) {
      DEBUG_STREAM("Input:" << i->first << endl);
    }
    for (CNJackMap_t::const_iterator i = output_jacks.begin(); i != output_jacks.end(); i++) {
      DEBUG_STREAM("Output:" << i->first << endl);
    }
    DEBUG_STREAM("Available Formats:" << endl);
    DEBUG_STREAM("Input: " << getOwnInputProperty()->getFormatList().size() << endl);
//     for (format_list::const_iterator i = getOwnInputProperty()->getFormatList().begin(); i != getOwnInputProperty()->getFormatList().end(); i++) {
//       DEBUG_STREAM((*(*i)) << endl);
//     }
    DEBUG_STREAM("Output: " << getOwnInputProperty()->getFormatList().size() << endl);
    for (format_list::const_iterator i = getOwnOutputProperty()->getFormatList().begin(); i != getOwnOutputProperty()->getFormatList().end(); i++) {
      DEBUG_STREAM((*(*i)) << endl);
    }

    return retval;
  }


  Result CompositeNode::doDeinit()
    throw (const StateTransitionFailedException&, 
	   const IllegalStateException&, const UnsuccessfulRequiredCallException&) { 
    Result retval = doStateTransitionTo(DEINITIALIZED); 

    for (map<unsigned int, CNNode>::const_iterator n = nodes.begin(); n != nodes.end(); n++) {
      removeAllFormats(STATIC_INPUT, n->first);
      removeAllFormats(STATIC_OUTPUT, n->first);
    }

    DEBUG_STREAM("StateTrasition! Available Jacks are: " << endl);
    for (CNJackMap_t::const_iterator i = input_jacks.begin(); i != input_jacks.end(); i++) {
      DEBUG_STREAM("Input:" << i->first << endl);
    }
    for (CNJackMap_t::const_iterator i = output_jacks.begin(); i != output_jacks.end(); i++) {
      DEBUG_STREAM("Output:" << i->first << endl);
    }
    DEBUG_STREAM("Available Formats:" << endl);
    DEBUG_STREAM("Input: " << getOwnInputProperty()->getFormatList().size() << endl);
//     for (format_list::const_iterator i = getOwnInputProperty()->getFormatList().begin(); i != getOwnInputProperty()->getFormatList().end(); i++) {
//       DEBUG_STREAM((*(*i)) << endl);
//     }
    DEBUG_STREAM("Output: " << getOwnInputProperty()->getFormatList().size() << endl);
    for (format_list::const_iterator i = getOwnOutputProperty()->getFormatList().begin(); i != getOwnOutputProperty()->getFormatList().end(); i++) {
      DEBUG_STREAM((*(*i)) << endl);
    }

    return retval;
  }


  Result CompositeNode::doFlush() 
    throw (const IllegalStateException&) {
    Result retval = SUCCESS;

    for (map<unsigned int, CNNode>::iterator i = nodes.begin(); i!= nodes.end(); i++) {
      INode& current_node = i->second.getNode();  // for convinience

      try {
	Result tmpresult(current_node.flush());
        if ((SUCCESS == retval) && (SUCCESS != tmpresult)) {
          retval = tmpresult;
        }
      } catch (const IllegalStateException & e) {
	// oops, one of the nodes is cracy
	// perhaps someone has externly changed its state?
	ERROR_STREAM("Node " << 
		     i->first << 
		     " is not in its expected state: " << 
		     current_node.getState() << endl);
	throw IllegalStateException(this, 
				    getName() + 
				    string(">Wrong state for flush() in subnode ") + 
				    i->first);
      } catch (...) {
	// oops, what's going up now, unexpected exception?
	ERROR_STREAM("Unexpected exception from subnode " << 
		     i->first << "!" << endl);
	throw IllegalStateException(string("CompositeNode: Unexpceted exception in method CompositeNode::doFlush(), caused by node ") + i->first);
      }
    }

    return retval;
    
  }
  /*************************   Node interface; do-methods   ******************/


  /************************* the internal ProxyEventSender ******************/
  CompositeNode::ProxyEventReceiver::ProxyEventReceiver(IEventReceiver& recv)
    : IEventReceiver(),
      forwarding_receiver(recv) {
  }

  CompositeNode::ProxyEventReceiver::~ProxyEventReceiver() {
  }
  
  Result CompositeNode::ProxyEventReceiver::receiveEventInternal(Event& event, const IEventSender* sender) {
    if (!isForwardingAllowed(event)) {
      // this behaviour is wanted, so no error
      return SUCCESS;
    }

    return forwarding_receiver.receiveEvent(event, sender);
  }
  /************************* the internal ProxyEventSender ******************/



  void CompositeNode::setGraphDescription(const GraphDescription& gd) {
    __graph_description = gd.copy();
  }


  void CompositeNode::realizeGraph() {
  }

  const Format* CompositeNode::getInputFormat(const char * ) const {
    // FIXME
    ERROR_STREAM2("getInputFormat Not implemented! ..." << endl);
    return 0;
  }

  const Format* CompositeNode::getOutputFormat(const char * ) const{
    // FIXME
    ERROR_STREAM2("getOutputFormat Not implemented! ..." << endl);
    return 0;
  }
 
  void CompositeNode::setPrefetchQueueSize(const nmm_uint32 &){
    ERROR_STREAM("setPrefetchQueueSize not implemented" << endl);
  }

  void CompositeNode::setPrefetchQueueDuration(const Interval &){
    ERROR_STREAM("setPrefetchQueueDuration not implemented" << endl);
  }
  
  nmm_uint32 CompositeNode::getPrefetchQueueSize() {
    ERROR_STREAM("setPrefetchQueueSize not implemented" << endl);
    return 0;
  }
  
  void CompositeNode::waitPrefetchFinished() {
    ERROR_STREAM("waitPrefetchFinished() not implemented " << endl);
  }


 }



#include "nmm/nmm_suffix.hpp"
