/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include "GraphBuilder.hpp"

#include "nmm/comm/ProxyObject.hpp"
#include "nmm/multimedia/connect.hpp"
#include "nmm/multimedia/IBiDirQueuedNode.hpp"
#include "nmm/multimedia/sync/interfaces/ISynchronizedSink.hpp"
#include "nmm/services/registry/RegistryLock.hpp"

#include "nmm/services/DefaultSinks.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  GraphBuilder::GraphBuilder()
    : __use_vis(false),
      __url(),
      __app(0),
      __audio_sink(0),
      __audio_sink_nd(),
      __audio_sink_format(0),
      __audio_sink_property(0),
      __audio_demux_out_jack_tag("audio"),
      __audio_sink_in_jack_tag("default"),
      __audio_sink_prefetch_queue_size(0),
      __is_audio_sink_set_from_external(false),
      __video_sink(0),
      __video_sink_nd(),
      __video_sink_format(0),
      __video_sink_property(0),
      __video_demux_out_jack_tag("video"),
      __video_sink_in_jack_tag("default"),
      __video_sink_prefetch_queue_size(0),
      __is_video_sink_set_from_external(false),
      __sync(0),
      __sharing_type(NodeDescription::UNDEFINED)
  {
    /* default audio sink */
    NodeDescription default_audio_sink_nd(DEFAULT_AUDIO_SINK);
    setAudioSink(default_audio_sink_nd);
    
    /* default video sink */
    NodeDescription default_video_sink_nd(DEFAULT_VIDEO_SINK);
    setVideoSink(default_video_sink_nd);
  }

  GraphBuilder::~GraphBuilder() 
  {
    if( __audio_sink_format ) {
      delete __audio_sink_format;
    }
    if( __video_sink_format ) {
      delete __video_sink_format;
    }
    if( __audio_sink_property ) {
      delete __audio_sink_property;
    }
    if( __video_sink_property ) {
      delete __video_sink_property;
    }
  }

  bool GraphBuilder::setURL(const string& url) 
  {
    __url = GraphURL(url);
    return __url.isValid();
  }

  void GraphBuilder::setURL(const GraphURL& url) 
  {
    __url = url;
  }

  void GraphBuilder::setVisualization(const bool vis) 
  {
    __use_vis = vis;
  }

  void GraphBuilder::setAudioSink(INode* sink, const string& input_jack) 
  {
    __audio_sink = sink;
    __audio_sink_in_jack_tag = input_jack;
    __is_audio_sink_set_from_external = true;
  }

  void GraphBuilder::setAudioSink(const NodeDescription& sink, const string& input_jack) 
  {
    __audio_sink_nd = sink;
    __audio_sink_in_jack_tag = input_jack;
  }

  const NodeDescription& GraphBuilder::getAudioSinkNodeDescription() {
    return __graph_nodes[__audio_sink];
  }

  void GraphBuilder::setAudioSinkFormat(const Format& format)
  {
    __audio_sink_format = new Format(format);
  }

  void GraphBuilder::setDemuxAudioJackTag(const string& tag) 
  {
    __audio_demux_out_jack_tag = tag;
  }

  void GraphBuilder::setAudioPrefetchQueueSize(const unsigned int& prefetch_queue_size)
  {
    __audio_sink_prefetch_queue_size = prefetch_queue_size;
  }

  void GraphBuilder::setVideoSink(INode* sink, const string& input_jack) 
  {
    __video_sink = sink;
    __video_sink_in_jack_tag = input_jack;
    __is_video_sink_set_from_external = true;
  }

  void GraphBuilder::setVideoSink(const NodeDescription& sink, const string& input_jack) 
  {
    __video_sink_nd = sink;
    __video_sink_in_jack_tag = input_jack;
  }

  const NodeDescription& GraphBuilder::getVideoSinkNodeDescription() {
    return __graph_nodes[__video_sink];
  }

  void GraphBuilder::setVideoSinkFormat(const Format& format)
  {
    __video_sink_format = new Format(format);
  }

  void GraphBuilder::setDemuxVideoJackTag(const string& tag) 
  {
    __video_demux_out_jack_tag = tag;
  }

  void GraphBuilder::setVideoPrefetchQueueSize(const unsigned int& prefetch_queue_size)
  {
    __video_sink_prefetch_queue_size = prefetch_queue_size;
  }

  void GraphBuilder::setAudioVideoPrefetchingEnabled(bool enable)
  {
    if(enable) {
      if(__audio_sink_prefetch_queue_size == 0) {
	setAudioPrefetchQueueSize(4);
      }
      if(__video_sink_prefetch_queue_size == 0) {
	setVideoPrefetchQueueSize(4);
      }
    }
    else {
      setAudioPrefetchQueueSize(0);
      setVideoPrefetchQueueSize(0);
    }
  }

  void GraphBuilder::setMultiAudioVideoSynchronizer(IMultiAudioVideoSynchronizer* sync) 
  {
    __sync = sync;
  }

  void GraphBuilder::setSharingType(NodeDescription::SharingType type)
  {
    __sharing_type = type;
  }
  
  const GraphDescription& GraphBuilder::getGraphDescription()
  {
    return __graph_descr;
  }

  list<Response> GraphBuilder::initRequest(ClientRegistry& registry, NodeDescription& nd) 
  {
    if(__sharing_type != NodeDescription::UNDEFINED) {
      nd.setSharingType(__sharing_type);
      GLOBAL_DEBUG_STREAM2("GB: SharingType: " << __sharing_type << endl);
    }
    return registry.initRequest(nd);
  }

  INode* GraphBuilder::requestNode(ClientRegistry& registry, const Response& response)
  {
    NodeDescription nd;
    INode* n = registry.requestNode(response, &nd);
    if(n) {
      __graph_nodes[n] = nd;
    }
    return n;
  }

  INode* GraphBuilder::requestNode(ClientRegistry& registry, const list<Response>& response_list)
  {
    NodeDescription nd;
    INode* n = registry.requestNode(response_list, &nd);
    if(n) {
      __graph_nodes[n] = nd;
    }
    return n;
  }

  INode* GraphBuilder::requestNode(ClientRegistry& registry, const RegistryID& reg_id, const ObjectID& node_id)
  {
    NodeDescription nd;
    INode* n = registry.requestNode(reg_id, node_id, &nd);
    if(n) {
      __graph_nodes[n] = nd;
    }
    return n;
  }

  INode* GraphBuilder::requestSinkNode(const NodeDescription& sink_nd, NMMApplication* app) 
  {
    INode* sink = 0;
    ClientRegistry& registry = app->getRegistry();

    {
      RegistryLock lock(registry);
      // sink nodes are not requested by calling member initRequest since they should not be shared
      list<Response> node_response = registry.initRequest(sink_nd);

      if (node_response.empty()) {
        GLOBAL_ERROR_STREAM1("GB: The specified sink node " << sink_nd.getNodeName() << " is not available!" << endl);
	      throw Exception("The specified sink node is not available");
      }
      
      sink = requestNode(registry, node_response.front());
      GLOBAL_DEBUG_STREAM2("GB: Successfully requested node " << sink_nd.getNodeName() << " @ " << sink_nd.getLocation() << endl);

    }

    return sink;
  }

  void GraphBuilder::connectNodes(INode* n1, const string& out_jack, INode* n2, const string& in_jack, const Format* format)
  {
    Format connection_format;
    connect(n1, out_jack, n2, in_jack, format, &connection_format);

    if ( ((n2 == __audio_sink && !__is_audio_sink_set_from_external) || n2 != __audio_sink) ||
	 ((n2 == __video_sink && !__is_video_sink_set_from_external) || n2 != __video_sink) ) {
      __graph_descr.addEdge(__graph_nodes[n1], out_jack, __graph_nodes[n2], in_jack, &connection_format);
      __graph_descr.setINode(__graph_nodes[n1], *n1);
      __graph_descr.setINode(__graph_nodes[n2], *n2);
      cerr << "connect(" << __graph_nodes[n1].getNodeName() << ", " << __graph_nodes[n2].getNodeName() << ")" << endl;
    } else {
      if (!__graph_descr.getINode(__graph_nodes[n1])) {
	__graph_descr.addNode(__graph_nodes[n1]);
	__graph_descr.setINode(__graph_nodes[n1], *n1);
	cerr << "added node " << __graph_nodes[n1].getNodeName() << endl;
      }
    }
  }
  
  CompositeNode* GraphBuilder::createGraph(NMMApplication& app)
  {
    __app = &app;
    list<INode*> nodes_to_release;
    CompositeNode* cnode = new CompositeNode("Player", &__app->getRegistry());
    ICompositeNode_var composite(cnode->getCheckedInterface<ICompositeNode>());
    INode_var composite_node(cnode->getCheckedInterface<INode>());
    try {
      // if no INode*s were set, try request to request from given node descriptions
      if( !__audio_sink ) {
	__audio_sink = requestSinkNode(__audio_sink_nd, __app);

	if ( __audio_sink_nd.getSharingType() != NodeDescription::SHARED ) {
	  if (__audio_sink_nd.getNodeEvent(INode::CONSTRUCTED)) {
	    __audio_sink->reachConstructed();
	    CEvent& ce = *__audio_sink_nd.getNodeEvent(INode::CONSTRUCTED);
	    for(unsigned int i=0; i<ce.size(); ++i) {
	      __audio_sink->sendEvent(ce[i]);
	    }
	  }
	}
	
	nodes_to_release.push_front(__audio_sink);
      }
      if( !__video_sink ) {
	__video_sink = requestSinkNode(__video_sink_nd, __app);

	if ( __video_sink_nd.getSharingType() != NodeDescription::SHARED ) {
	  if (__video_sink_nd.getNodeEvent(INode::CONSTRUCTED)) {
	    __video_sink->reachConstructed();
	    CEvent& ce = *__video_sink_nd.getNodeEvent(INode::CONSTRUCTED);
	    for(unsigned int i=0; i<ce.size(); ++i) {
	      __video_sink->sendEvent(ce[i]);
	    }
	  }
	}

	nodes_to_release.push_front(__video_sink);
      }
      if(__audio_sink_prefetch_queue_size > 0) {
	__audio_sink->setPrefetchQueueSize(__audio_sink_prefetch_queue_size);
      }
      if(__video_sink_prefetch_queue_size > 0) {
	__video_sink->setPrefetchQueueSize(__video_sink_prefetch_queue_size);
      }

      __audio_sink->reachInitialized();
      __video_sink->reachInitialized();

      if ( __audio_sink_nd.getSharingType() != NodeDescription::SHARED ) {
	if (__audio_sink_nd.getNodeEvent(INode::INITIALIZED)) {
	  CEvent& ce = *__audio_sink_nd.getNodeEvent(INode::INITIALIZED);
	  for(unsigned int i=0; i<ce.size(); ++i) {
	    __audio_sink->sendEvent(ce[i]);
	  }
	}
      }
      
      if ( __video_sink_nd.getSharingType() != NodeDescription::SHARED ) {
	if (__video_sink_nd.getNodeEvent(INode::INITIALIZED)) {
	  CEvent& ce = *__video_sink_nd.getNodeEvent(INode::INITIALIZED);
	  for(unsigned int i=0; i<ce.size(); ++i) {
	    __video_sink->sendEvent(ce[i]);
	  }
	}
      }
      
      // only need input property of sinks if no input format was given
      if( !__audio_sink_format) {
	__audio_sink_property = __audio_sink->getInputProperty();
      }
      if( !__video_sink_format) {
	__video_sink_property = __video_sink->getInputProperty();
      }

      SinkNodes_t sink_nodes;
      sink_nodes.push_front( SinkNode_t(__audio_sink, __audio_sink_in_jack_tag, __audio_sink_format, __audio_sink_property) );
      sink_nodes.push_front( SinkNode_t(__video_sink, __video_sink_in_jack_tag, __video_sink_format, __video_sink_property) );

      // request source node
      NodeDescription source_nd = __url.getNodeDescription(); 
      GLOBAL_DEBUG_STREAM2("GB: source node: name: " << source_nd.getNodeName() << endl);
      GLOBAL_DEBUG_STREAM2("GB: source node: number of interfaces: " << source_nd.getInterfaceList().size() << endl);
      GLOBAL_DEBUG_STREAM2("GB: source node: format: " << *(source_nd.getOutputProperty().getFormatList().front()) << endl);
      // check if meaningful node description was generated
      if(source_nd.getNodeName().empty()) {
	if(source_nd.getOutputProperty().getFormat(Format::wild_wild_type).size() == 
	   source_nd.getOutputProperty().getFormatList().size()) {
	  throw Exception("No matching source node available for URL");
	}
      }
      Format* source_format = source_nd.getOutputProperty().getFormatList().front();

      INode* source = 0;
      ClientRegistry& registry = __app->getRegistry();
      
      {
	RegistryLock lock(registry);

	list<Response> response = initRequest(registry, source_nd);

	if (response.empty()) {
	  throw Exception("No matching source node found");
	}
	Response best_response;
	for(list<Response>::const_iterator response_iter = response.begin(); 
	    response_iter != response.end(); 
	    ++response_iter) 
	{
	  for(list<ObjectID>::const_iterator objid_iter = response_iter->second.begin(); 
	      objid_iter != response_iter->second.end(); 
	      ++objid_iter) 
	  {
	    NodeDescription nd(registry.getNodeDescription(response_iter->first, *objid_iter));
	    bool includes_wild_format  = !(nd.getOutputProperty().getFormat(Format::wild_wild_type).empty());
	    bool includes_exact_format = !(nd.getOutputProperty().getFormat(source_format).empty());
	    bool includes_format = false; 
	    for (list<Format*>::const_iterator it = nd.getOutputProperty().getFormatList().begin(); 
		 it != nd.getOutputProperty().getFormatList().end(); 
		 ++it) 
	    {
	      if ((*it)->isFormatIncluded(source_format)) {
		includes_format = true;
		break;
	      }
	    }
	    GLOBAL_DEBUG_STREAM2("GB: NodeName: " << nd.getNodeName() /* << endl */ 
				 << "  includes wildcard format? " << includes_wild_format /* << endl */ 
				 << "  includes exact format? " << includes_exact_format /* << endl */ 
				 << "  includes format? " << includes_format << endl);
	    if(!includes_wild_format) {
	      if(includes_exact_format) { // really best response
		list<ObjectID> list;
		list.push_back(*objid_iter);
		best_response = make_pair(response_iter->first,  list);
		break;
	      }
	      else if(includes_format) { // last response: at least includes format
		list<ObjectID> list;
		list.push_back(*objid_iter);
		best_response = make_pair(response_iter->first,  list);
	      }
	    }
	    else if(best_response.second.empty()) { // response with wildcard if no response without wildcard found so far
	      list<ObjectID> list;
	      list.push_back(*objid_iter);
	      best_response = make_pair(response_iter->first,  list);
	    }
	  }
	  if(!best_response.second.empty()) { // already found some best response?
	    break;
	  }
	}
	source = requestNode(registry, best_response);
	
	nodes_to_release.push_front(source);
      }
      INamed_var named_source(source->getParentObject()->getCheckedInterface<INamed>());
      GLOBAL_DEBUG_STREAM2("GB: Successfully requested source node: " << named_source->getName() 
			   << " @ " << source_nd.getLocation() << endl);

      // setup source node
      //      __url.configureConstructed(source);
      if (source_nd.getNodeEvent(INode::CONSTRUCTED)) {
	CEvent& ce = *source_nd.getNodeEvent(INode::CONSTRUCTED);
	for(unsigned int i=0; i<ce.size(); ++i) {
	  source->sendEvent(ce[i]);
	}
      }

      source->init();
      //      __url.configureInitialized(source);
      if (source_nd.getNodeEvent(INode::INITIALIZED)) {
	CEvent& ce = *source_nd.getNodeEvent(INode::INITIALIZED);
	for(unsigned int i=0; i<ce.size(); ++i) {
	  source->sendEvent(ce[i]);
	}
      }

      source->initOutput();
      //__url.configureOutputInitialized(source);
      if (source_nd.getNodeEvent(INode::OUTPUT_INITIALIZED)) {
	CEvent& ce = *source_nd.getNodeEvent(INode::OUTPUT_INITIALIZED);
	for(unsigned int i=0; i<ce.size(); ++i) {
	  source->sendEvent(ce[i]);
	}
      }

      // setup list of nodes that need to be connected
      list< pair<INode*, string> > left_nodes;
      left_nodes.push_front( make_pair(source, source->getOutputStreamTags().front()) );

      INode* right_node = 0;
      INode* final_right_node = 0;
      INode* best_right_node = 0;
      INode* some_right_node = 0;
      int num_connected_demux_outputs = 0;
      int num_demux_outputs = 0;
      int depth = 0;

      list< pair<INode*, string> >::iterator left_nodes_iter = left_nodes.begin(); 
      while(left_nodes_iter != left_nodes.end()) {

	INode* left_node = (*left_nodes_iter).first;
	string left_node_jack_tag = (*left_nodes_iter).second;
      
	// test if left node is demultiplexer -> then all downstream queues of left_nodes need to be increased
	bool is_left_node_demux = false;
	if(left_node->getType() == INode::DEMULTIPLEXER) {
	  is_left_node_demux = true;
	  num_connected_demux_outputs = 0;
	}

	size_t num_connected_sink_nodes = 0;
	SinkNodes_t::iterator sink_nodes_iter = sink_nodes.begin();
        { 
	  for(sink_nodes_iter = sink_nodes.begin(); sink_nodes_iter != sink_nodes.end(); ++sink_nodes_iter) { 
	    try {
	      INode* sink_node = (*sink_nodes_iter).node;
	      if( !((*sink_nodes_iter).connected) ) { // if current sink is not connected yet -> try to connect with current left_node

		left_node->reachOutputInitialized();

		// connect left_node and some sink node
		if(left_node == source) { // only connect source to sink using default output format of sink
		  Property_var wop(source->getWorkingOutputProperty());
		  const Format* default_wop_format = wop->getDefaultFormat();
		  if(default_wop_format) {
		    if((*sink_nodes_iter).format != 0) {
		      Format_var intersection_format(default_wop_format->createIntersectionFormat((*sink_nodes_iter).format));
		      if(intersection_format.get()) {
			connectNodes(left_node, left_node_jack_tag, sink_node, (*sink_nodes_iter).jack_tag, intersection_format.get());
			(*sink_nodes_iter).connected = true;
		      }
		      else {
			if(default_wop_format->isFormatIncluded((*sink_nodes_iter).format)) {
			  connectNodes(left_node, left_node_jack_tag, sink_node, (*sink_nodes_iter).jack_tag, (*sink_nodes_iter).format);
			  (*sink_nodes_iter).connected = true;
			}
			else {
			  throw Exception("Default output format of source and specified input format of sink cannot be intersected");
			}
		      }
		    }
		    else {
		      connectNodes(left_node, left_node_jack_tag, sink_node, (*sink_nodes_iter).jack_tag,  default_wop_format);
		      (*sink_nodes_iter).connected = true;
		    }
		  }
		  else {
		    GLOBAL_DEBUG_STREAM("GB: Chosen source node does not provide a default format" << endl);
		    connectNodes(left_node, left_node_jack_tag, sink_node,(*sink_nodes_iter).jack_tag, (*sink_nodes_iter).format);
		    (*sink_nodes_iter).connected = true;
		  }
		} // end: if(left_node == source)
		else {
		  connectNodes(left_node, left_node_jack_tag, sink_node,(*sink_nodes_iter).jack_tag, (*sink_nodes_iter).format);
		  (*sink_nodes_iter).connected = true;
		}

		// add to synchronizer
		if(__sync) {
		  ISynchronizedSink_var sync_sink(sink_node->getParentObject()->getInterface<ISynchronizedSink>());
		  if(sync_sink.get()) {
		    sync_sink->setSynchronized(true);
		    ISinkController_var controller(sync_sink->getController());
		    if(sync_sink->isBuffered()) {
		      __sync->addAudioSinkController(controller.get());
		    }
		    else {
		      __sync->addVideoSinkController(controller.get());
		    }      
		  }
		}
	      
		// do not activate demux node here, all outputs need to be connected first
		if(!is_left_node_demux) {
		  left_node->activate();
		  sink_node->initOutput();
		  sink_node->activate();
		}
	    
		// if sink node is in list of nodes to release -> allocate node in graph builder -> add to composite node
		for(list<INode*>::iterator find = nodes_to_release.begin(); find != nodes_to_release.end(); ++find) {
		  if(*find == sink_node) {
		    composite->addNode(*sink_node);
		    break;
		  }
		}
		composite->addNode(*left_node);
	    
		left_nodes.erase(left_nodes_iter);
		GLOBAL_DEBUG_STREAM2("GB: Finished some part of the graph :)" << endl);
	    
		// if all other sink nodes are connected and activated -> finished graph building
		num_connected_sink_nodes = 0;
		for(SinkNodes_t::const_iterator j = sink_nodes.begin(); j != sink_nodes.end(); ++j) {
		  if( (*j).node->isActivated() ) {
		    ++num_connected_sink_nodes;
		  }
		}
		if(num_connected_sink_nodes == sink_nodes.size()) {
		  break;
		}

		// nothing to do anymore 
		if(left_nodes.empty()) { 
		  break;
		}
	    
		// if we came that far in try-block, some sink node was connected -> break for-loop (and continue while-loop!)
		left_nodes_iter = left_nodes.begin();
		break;
	      } // end: if(sink_node->isInitialized()
	    }
	    catch(const SerializedException& se) {
	      GLOBAL_DEBUG_STREAM2("GB: Could not finish this part of graph yet..." << endl);
	    }
	    catch(const Exception& e) {
	      GLOBAL_DEBUG_STREAM2("GB: Could not finish this part of graph yet..." << endl);
	    }
	  } // end: for all sink_nodes

	  // if some condition for final termination is given -> break while-loop here
	  if( (num_connected_sink_nodes == sink_nodes.size()) || (left_nodes.empty()) ) {
	    break;
	  }
	  // if not all sink_nodes were tested, some connection setup was successful -> continue while-loop!
	  if( sink_nodes_iter != sink_nodes.end() ) {
	    continue;
	  }
	} // end: if(!is_left_node_demux)

	// get default output format of left node
	Property_var wop(left_node->getWorkingOutputProperty());
	Format* default_wop_format = const_cast<Format*>(wop->getDefaultFormat( left_node_jack_tag ));
	// need to set jack tag to default since all converter and demultiplexer have an input named default
	default_wop_format->setJackTag("default");
	// cerr << *default_wop_format << endl;

	for(int node_type = 0; node_type < 2; ++node_type) { // try CONVERTER and DEMULTIPLEXER
	  NodeDescription right_nd;
	  right_nd.addInputFormat(*default_wop_format);
	  if(node_type == 0) {
	    GLOBAL_DEBUG_STREAM2("GB: try CONVERTER" << endl);
	    right_nd.setNodeType(INode::CONVERTER);
	    string format_type = default_wop_format->getFormatType();
	    
	    if(format_type.find("audio") != string::npos) {
	      const RegistryName& reg_name = registry.determineRegistryName(*__audio_sink);
	      right_nd.setLocation(reg_name.first);
	      right_nd.setPort(reg_name.second);
	    }
	    else if(format_type.find("video") != string::npos) {
	      const RegistryName& reg_name = registry.determineRegistryName(*__video_sink);
	      right_nd.setLocation(reg_name.first);
	      right_nd.setPort(reg_name.second);
	    }
	  }
	  else {
	    GLOBAL_DEBUG_STREAM2("GB: try DEMULTIPLEXER" << endl);
	    right_nd.setNodeType(INode::DEMULTIPLEXER);
	    // Locate demux on same host as source node
	    right_nd.setLocation(__url.getHost());
	    right_nd.setPort(__url.getPort());
	  }

	  // Response best_response;            // holds the 'best' response of all possible right_nodes
	  float best_response_weight = -1.0; // start with quality that can only get better
	  {
	    RegistryLock lock(registry);
	    list<Response> response = initRequest(registry, right_nd);
	
	    list<Response>::const_iterator response_iter = response.begin();
	    
	    // only one possiblity found
	    if((response.size() == 1) && (response_iter->second.size() == 1)) {
	      GLOBAL_DEBUG_STREAM2("GB: Only one possibility found" << endl);
	      Response best_response = *(response_iter); 
	      right_node = requestNode(registry, best_response);
	    }
	    else { // more than a single response
	      
	      list<Response> response = initRequest(registry, right_nd);

	      // FIRST STEP **************************************************

	      list< pair< RegistryID, ObjectID> > very_good_responses;
	      list< pair< RegistryID, ObjectID> > good_responses;

	      for(list<Response>::const_iterator response_iter = response.begin(); 
		  response_iter != response.end(); 
		  ++response_iter) 
	      {
		for(list<ObjectID>::const_iterator objid_iter = response_iter->second.begin(); 
		    objid_iter != response_iter->second.end(); 
		    ++objid_iter) 
		{
		  NodeDescription nd(registry.getNodeDescription(response_iter->first, *objid_iter));
		  bool includes_wild_format  = !(nd.getInputProperty().getFormat(Format::wild_wild_type).empty());
		  GLOBAL_DEBUG_STREAM2("GB: Node: " << nd.getNodeName() << 
				       ": includes wild format? "<< includes_wild_format << endl);

		  if(!includes_wild_format) {   // no wildcard
		    bool includes_exact_format = !(nd.getInputProperty().getFormat(default_wop_format).empty());
		    GLOBAL_DEBUG_STREAM2("GB: Node: " << nd.getNodeName() << 
					 ": includes exact format? " << includes_exact_format << endl);

		    if(includes_exact_format) { // exact format -> very good
		      very_good_responses.push_front( make_pair( response_iter->first, *objid_iter ) );
		    }
		    else { // else: not exact
		      bool includes_format = false; 
		      for (list<Format*>::const_iterator it = nd.getInputProperty().getFormatList().begin(); 
			   it != nd.getInputProperty().getFormatList().end(); 
			   ++it) 
		      {
			if ((*it)->isFormatIncluded(default_wop_format)) {
			  includes_format = true;
			  break;
			}
		      }
		      GLOBAL_DEBUG_STREAM2("GB: Node: " << nd.getNodeName() << 
					   ": includes format? " << includes_format << endl);
		      if(includes_format) { // at least includes format -> good
			good_responses.push_front( make_pair( response_iter->first, *objid_iter ) );
		      }
		    } // end: else: not exact
		  } // end: no wildcard 
		} // end: for all ObjectID of current Response
	      } // end: for all Response objects

	      if(very_good_responses.size() == 0) {
		very_good_responses = good_responses;
	      }
	      if(very_good_responses.size() == 1) {
		GLOBAL_DEBUG_STREAM2("GB: Only single good or very good possibility found" << endl);
		list< pair< RegistryID, ObjectID> >::iterator r = very_good_responses.begin();
		right_node = requestNode(registry, (*r).first, (*r).second );
	      }
	      else {
		// SECOND STEP **************************************************

		GLOBAL_DEBUG_STREAM2("GB: SECOND STEP: " << very_good_responses.size() << endl);
		list< pair<RegistryID, ObjectID> > great_responses;

		for(list< pair<RegistryID, ObjectID> >::const_iterator iter = very_good_responses.begin(); 
		    iter != very_good_responses.end(); 
		    ++iter) 
		{
		  NodeDescription nd( registry.getNodeDescription((*iter).first, (*iter).second) );
		
		  GLOBAL_DEBUG_STREAM2("GB: very good: " << nd.getNodeName() << endl);

		  if(node_type == 0) { // CONVERTER
		    list<INode*> possible_right_nodes;
		    for(SinkNodes_t::const_iterator s = sink_nodes.begin(); s != sink_nodes.end(); ++s) { // for all sinks
		      const_format_list l; // FIXME: delete Format*s in list
		      if( (*s).format ) {
			if( nd.getOutputProperty().isFormatIncluded( *((*s).format), "default") ) {
			  l.push_back( new Format( *((*s).format) ) );
			}
		      }
		      else {
			l =  nd.getOutputProperty().createIntersectionFormats("default", (*s).property, (*s).jack_tag, true);
		      }
		      GLOBAL_DEBUG_STREAM2("GB: very good: " << nd.getNodeName() << ": SOP->sink: " << l.size() << endl);

		      if(l.size()) { // direct connection could be possible (if some format in wop can be intersected!)
			cleanupFormatList(l);
			some_right_node = requestNode(registry, (*iter).first, (*iter).second );
			possible_right_nodes.push_back(some_right_node);
			some_right_node->init();
			IInputJack_var jack(some_right_node->getInputJack(*default_wop_format, default_wop_format->getJackTag()));

			// try to get intersection with wop and input format of current sink
			const_format_list l2; 
			try {
			  some_right_node->initOutput();
			  Property_var right_wop_property(some_right_node->getWorkingOutputProperty());

			  if( (*s).format ) {
			    if( right_wop_property->isFormatIncluded( *((*s).format), "default") ) {
			      l2.push_back( new Format( *((*s).format) ) );
			    }
			  }
			  else {
			    l2 =  right_wop_property->createIntersectionFormats("default", (*s).property, (*s).jack_tag, true);
			  }
			  GLOBAL_DEBUG_STREAM2("GB: very good: " << nd.getNodeName() << ": WOP->sink: " << l2.size() << endl);

			  if(l2.size()) { // can be connected to current sink 
			    cleanupFormatList(l2);
			    if(!final_right_node) { // and no such node was found yet -> final_right_node
			      final_right_node = some_right_node;
			      GLOBAL_DEBUG_STREAM2("GB: final_right_node: " << nd.getNodeName() << endl);
			    }
			  }
			}
			
			catch(const SerializedException& se) { // doInitOutputFailed failed
			  GLOBAL_DEBUG_STREAM2("GB: doInitOutput failed: unable to get wop for unconnected node: using node anyway if no better solution is found" << endl);
			}
			catch(const Exception& e) { // something else failed
			  GLOBAL_ERROR_STREAM("GB: " << e << endl);
			  throw e;
			}

			// find best quality of current format in input property of possible right_node
			format_list nl = nd.getInputProperty().getFormat(default_wop_format);
			float best_weight = 0.0f;
			for(format_list::const_iterator f = nl.begin(); f != nl.end(); ++f) {
			  float f_weight = (*f)->getStandardizedWeight();
			  if( f_weight > best_weight ) {
			    best_weight = f_weight;
			    GLOBAL_DEBUG_STREAM2("GB: very good: " << nd.getNodeName() << ": best_weight: " << best_weight << endl);
			  }
			}		

			// if this quality is better than the best quality found so far -> update best_right_node
			// if this quality is equally good than best quality so far and a final_right_node -> update best_right_node
			if((  best_weight > best_response_weight) || 
			   ( (best_weight == best_response_weight) && (final_right_node == some_right_node) )) {
			  best_response_weight = best_weight;
			  best_right_node = some_right_node;
			  GLOBAL_DEBUG_STREAM2("GB: UPDATE: best_response, best_response_weight " << best_response_weight << endl);
			}
			// already found best possible solution?
			if((best_right_node == final_right_node) && (best_response_weight == 1.0f)) {
			  GLOBAL_DEBUG_STREAM2("GB: break1: optimal direct connection found" << endl);
			  break; // break for all sink nodes since optimal direct connection to current sink was found
			}
		      } // end: l.size() 
		    } // end: for all sink nodes

		    // determine right_node from final_right_node and best_right_node
		    if(final_right_node) {
		      right_node = final_right_node;
		    }
		    else if(best_right_node) {
		      right_node = best_right_node;
		    }
		    // release all other nodes
		    for(list<INode*>::iterator k = possible_right_nodes.begin(); k != possible_right_nodes.end(); ++k) {
		      if((*k) == right_node) {
			// check is_left_node_demux -> need to increase downstream queue size later-on -> reach CONSTRUCTED
			if(is_left_node_demux) {
			  right_node->reachConstructed();
			}
		      }
		      else {
			(*k)->reachConstructed();
			registry.releaseNode(**k);
		      }
		    }
		  } // end: CONVERTER 
		  else { // DEMULTIPLEXER (in this case, the first node is taken)
		    list<ObjectID> list;
		    list.push_back( (*iter).second );
		    Response best_response = make_pair( (*iter).first,  list);
		    right_node = requestNode(registry, best_response);
		    break; // break for-loop: for all very_best_responses
		  }

		  // already found best possible solution?
		  if((best_right_node == final_right_node) && (best_response_weight == 1.0f)) {
		    GLOBAL_DEBUG_STREAM2("GB: break2: optimal direct connection found" << endl);
		    break; // break for all sink nodes since optimal direct connection to current sink was found
		  }

		} // end: for all very_good_responses
	      } // end: else more than a single good or very good response
	    } // end else:  more than a single response

	    if(right_node) {
	      nodes_to_release.push_front(right_node);
	      INamed_var named_right(right_node->getParentObject()->getCheckedInterface<INamed>());
	      GLOBAL_DEBUG_STREAM2("GB: Successfully requested right node: " << named_right->getName() 
				   << " @ " << right_nd.getLocation() << endl);
	      if(depth > 10) {
		throw Exception("searched too deep");
	      }
	      else {
		++depth;
	      }
	      break; // break: while-loop for CONVERTER and DEMULTIPLEXER
	    }
	    else if(node_type > 0) {
	      throw Exception("no matching converter or demultiplexer found");
	    }

	  } // end: scope of registry lock
	} // end: for-loop: try CONVERTER and DEMULTIPLEXER

	// check is_left_node_demux -> need to increase downstream queue size of right node
	if(is_left_node_demux) {
	  IBiDirQueuedNode_var queued(right_node->getParentObject()->getInterface<IBiDirQueuedNode>());
	  if(queued.get()) {
	    queued->setDownstreamMaxSize(100);
	  }
	}

	// setup right_node, connect, add to composite node, ...
	right_node->reachInitialized();
	left_node->reachOutputInitialized();
	connectNodes(left_node, left_node_jack_tag, right_node, "default", default_wop_format);

	if(is_left_node_demux) { // only activate demux if all outputs are connected (otherwise syncReset will get lost!)
	  ++num_connected_demux_outputs;
	  if(num_connected_demux_outputs == num_demux_outputs) { // if all outputs of demux are connected -> activate
	    left_node->activate();
	  }
	}
	else {
	  left_node->activate();
	  right_node->initOutput();
	}

	composite->addNode(*left_node);

	left_nodes.erase(left_nodes_iter);

	// test, which output jacks of right_node should be handled 
	list<string> right_output_jack_tags = right_node->getOutputStreamTags();
	if(right_output_jack_tags.size() == 1) {
	  left_nodes.push_back( make_pair(right_node, right_output_jack_tags.front()) );
	}
	else {
	  bool added_audio_jack_tag = false;
	  bool added_video_jack_tag = false;
	  num_demux_outputs = 0;
	  for(list<string>::const_iterator tag = right_output_jack_tags.begin(); tag != right_output_jack_tags.end(); ++tag) {
	    if( ((*tag).find(__audio_demux_out_jack_tag) != string::npos) && (!added_audio_jack_tag) ) {
	      left_nodes.push_back( make_pair(right_node, *tag) );
	      added_audio_jack_tag = true;
	      ++num_demux_outputs;
	    }
	    else if( ((*tag).find(__video_demux_out_jack_tag) != string::npos) && (!added_video_jack_tag)) {
	      left_nodes.push_back( make_pair(right_node, *tag) );
	      added_video_jack_tag = true;
	      ++num_demux_outputs;
	    }
	  }
	}

	// set iterator to begin of all nodes again
	left_nodes_iter = left_nodes.begin();

	right_node = best_right_node = final_right_node = 0;
      }
    }
    catch (const SerializedException& se) {
      GLOBAL_ERROR_STREAM2("GB: " << se << endl);
      delete cnode;
      cnode = 0;
    }
    catch (const Exception& e) {
      GLOBAL_ERROR_STREAM2("GB: " << e << endl);
      delete cnode;
      cnode = 0;
    }
    catch(...) {
      GLOBAL_ERROR_STREAM2("GB: Something went wrong..." << endl);
      delete cnode;
      cnode = 0;
    }
    if( !cnode ) { // something went wrong -> clean-up
      ClientRegistry& registry = __app->getRegistry();
      RegistryLock lock(registry);
      for(list<INode*>::iterator i = nodes_to_release.begin(); i != nodes_to_release.end(); ++i) {
	if(*i) {
	  registry.releaseNode(**i);
	}
      }
      throw Exception("Could not create graph");
    }

    return cnode;
  }

  void GraphBuilder::cleanupFormatList(const const_format_list& l)
  {
    for(const_format_list::const_iterator i = l.begin(); i != l.end(); ++i) {
      delete (*i);
    }
  }
}

#include "nmm/nmm_suffix.hpp"
