/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include <iostream>
#include "GraphDescription.hpp"
#include "ISession.hpp"

#include "nmm/multimedia/connect.hpp"
#include "nmm/comm/InstreamChannel.hpp"
#include "nmm/comm/OutOfBandChannel.hpp"
#include "nmm/comm/ProxyObject.hpp"
#include "nmm/comm/serialize/TConstructor.hpp"
#include "nmm/comm/serialize/InvalidTypeException.hpp"
#include "nmm/comm/serialize/StdValue.hpp"
#include "nmm/multimedia/sync/interfaces/IMultiAudioVideoSynchronizer.hpp"
#include "nmm/multimedia/sync/interfaces/ISynchronizedSink.hpp"
#include "nmm/multimedia/IBiDirQueuedNode.hpp"
#include "nmm/comm/IInstreamCompositeStrategy.hpp"

#include <cassert>

#include "nmm/nmm_prefix.hpp"

namespace NMM {

GraphDescription::GraphDescription()
  : SerializedValue(GraphDescription_type_name, GraphDescription_type_ns), __internal_counter(0), __i_session(0), __sink_synch_descr()
{
  __interface_list.useTValue(false);
  __node_list.useTValue(false);
  registerAttributes();
}


GraphDescription::~GraphDescription()
{
  delete __i_session.getValue();

  for(SequenceValue< NodeDescription >::iterator i = __node_list.begin(); i != __node_list.end(); ++i) {
    delete (*i);
  }
  for(SequenceValue< INode >::iterator j = __interface_list.begin(); j != __interface_list.end(); ++j) {
    delete (*j);
  }
}

GraphDescription GraphDescription::getSkeleton() const {
  
  GraphDescription gd;
  
  /* delete all internal attributes */
  gd.clear();

  /*Begin> assign the needed values */
  gd.__edges             = __edges;
  gd.__out_edges         = __out_edges;
  gd.__in_edges          = __in_edges;

  gd.__node_list         = __node_list;
  gd.__request_flags     = __request_flags;
  gd.__release_flags     = __release_flags;

  gd.__internal_counter  = __internal_counter;
  /* End< */

  return gd;
}

bool GraphDescription::isIdenticalOverlap(const GraphDescription& gd) const
{
  if(getNumEdges() == gd.getNumEdges()) {
    unsigned int num_matching_edges = 0;
    for(unsigned int i = 0; i < getNumEdges(); ++i) {
      const EdgeDescription& edge = getEdge(i);
      for(unsigned int j = 0; j < gd.getNumEdges(); ++j) {
	const EdgeDescription& other = gd.getEdge(j);
	if(edge.isIdenticalOverlap(other)) {
	  ++num_matching_edges;
	  break;
	}
      }
    }
    if(num_matching_edges == getNumEdges()) {
      return true;
    }
    else {
      return false;
    }
  }
  else {
    return false;
  }
}

void GraphDescription::clear()
{
  /* Begin> Delete and reset the session */
  delete __i_session.getValue();
  __i_session.setValue(0);
  /* End< */

  /* Clear the object description of the sink synchronizer */
  __sink_synch_descr = ObjectDescription();

  /* Begin> Delete the NodeDescriptions and clear the vector */
  for(SequenceValue< NodeDescription >::iterator i = __node_list.begin(); i != __node_list.end(); ++i) {
    delete (*i);
  }
  __node_list.clear();
  /*End< */

  /* Begin> Delete the INodes and clear the vector */
  for(SequenceValue< INode >::iterator j = __interface_list.begin(); j != __interface_list.end(); ++j) {
    delete (*j);
  }
  __interface_list.clear();
  /* End< */
  
  /* Begin> Clear the remaing container */  
  __empty_list.clear();
  __edges.getRValue().clear();
  __out_edges.getRValue().clear();
  __in_edges.getRValue().clear();
  __request_flags.getRValue().clear();
  __release_flags.getRValue().clear();
  __internal_counter = 0;
  /* End< */
}

GraphDescription::GraphDescription(const GraphDescription& nd)
  :  SerializedValue(nd)
{
  __interface_list.useTValue(false);
  __node_list.useTValue(false);
  registerAttributes();
  *this = nd;
}


void GraphDescription::registerAttributes()
{
  insertArgument(&__edges);
  insertArgument(&__out_edges);
  insertArgument(&__in_edges);
  insertArgument(&__node_list);
  insertArgument(&__interface_list);
  insertArgument(&__request_flags);
  insertArgument(&__release_flags);
  insertArgument(&__i_session);
  insertArgument(&__sink_synch_descr);
}

void GraphDescription::setINode(const NodeDescription& nd, const INode& inode) 
{
  if (nd.getInternalID() < 0) {
    GLOBAL_ERROR_STREAM2("gd: "<< "Illegal NodeDescription" << endl);
    return;
  }

  int current_size = __interface_list.size();

  if (static_cast<unsigned int>(nd.getInternalID()) >= __interface_list.size()) {
    __interface_list.resize(nd.getInternalID()+1);

    /* Initialize the new fields */
    for (unsigned int i = current_size; i < __interface_list.size(); ++i) {
      __interface_list[i] = 0;
    }
  }

  if (__interface_list[nd.getInternalID()]) return;
  __interface_list[nd.getInternalID()] = dynamic_cast<INode*>(inode.copy());
}

const INode* GraphDescription::getINode(const NodeDescription& nd) const
{
  if (nd.getInternalID() < 0 )
    return 0;

  if (static_cast<unsigned int>(nd.getInternalID()) < __interface_list.size())
    return __interface_list[nd.getInternalID()];

  return 0;
}

INode* GraphDescription::getINode(const NodeDescription& nd) 
{
  if (nd.getInternalID() < 0 ) {
    return 0;
  }

  if (static_cast<unsigned int>(nd.getInternalID()) < __interface_list.size()) {
    return __interface_list[nd.getInternalID()];
  }

  return 0;
}

INode* GraphDescription::getINode(const NodeDescription* nd)
{
  return getINode(*nd);
}
 
const INode* GraphDescription::getINode(const NodeDescription* nd) const
{
  return getINode(*nd);
}

const list<int>& GraphDescription::getOutEdgeList(int i) const 
{ 
  CInternalIt iter = __out_edges.getRValue().find(i);
  if(iter == __out_edges.getRValue().end()) {
    return __empty_list;
  }
  else {
    return iter->second;  
  }
}

const list<int>& GraphDescription::getInEdgeList(int i) const 
{ 
  CInternalIt iter = __in_edges.getRValue().find(i);
  if(iter == __in_edges.getRValue().end()) {
    return __empty_list;
  }
  else {
    return iter->second;  
  }
}

const NodeDescription& GraphDescription::getFirstNodeDescription() const 
{ 
  if( (__node_list.size() > 0) && __node_list[0]) {
    return *(__node_list[0]); 
  }
  else {
    GLOBAL_ERROR_STREAM2("gd: "<< "GraphDescription::getFirstNodeDescription() called with size " << __node_list.size() << endl);
    throw Exception(" GraphDescription::getFirstNodeDescription() > Graph description is empty or 1. element is NULL");
  }
}

const NodeDescription& GraphDescription::getNodeDescription(unsigned int i) const 
{ 
  if(__node_list.size() > i) {
    return *(__node_list[i]); 
  }
  else {
    GLOBAL_ERROR_STREAM2("gd: "<< "GraphDescription::getNodeDescription( " << i << " ) const called with size " << __node_list.size() << endl);
    exit(-1);
  }
}

NodeDescription& GraphDescription::getNodeDescription(unsigned int i)
{
  if(__node_list.size() > i) {
    return *(__node_list[i]); 
  }
  else {
    GLOBAL_ERROR_STREAM2("gd: "<< "GraphDescription::getNodeDescription( " << i << " ) called with size " << __node_list.size() << endl);
    exit(-1);
  }
}

void GraphDescription::addNode(NodeDescription& nd) {
 // node descr has no id yet -> add to end 
  if(nd.getInternalID() < 0) {
    GLOBAL_DEBUG_STREAM2("gd: "<< "Adding new NodeDescription " << nd.getNodeName() << " with new ID " << __internal_counter << endl);

    nd.setInternalID(__internal_counter);
    NodeDescription* new_nd = new NodeDescription(nd);

    /* resize the node list */
    __node_list.resize(nd.getInternalID() + 1);

    /* resize the request flag */
    __request_flags.getRValue().resize(nd.getInternalID() + 1);
    /* resize the release flag */
    __release_flags.getRValue().resize(nd.getInternalID() + 1);

    __node_list[nd.getInternalID()] = new_nd;
    __request_flags.getRValue()[nd.getInternalID()] = true;
    __release_flags.getRValue()[nd.getInternalID()] = true;

    ++__internal_counter;
  }
  // node descr already has id >= size -> increase size and add node descr
  else if(static_cast<unsigned int>(nd.getInternalID()) >= __node_list.size()) { 
    unsigned int current_size = __node_list.size();

    /* resize the node list */
    __node_list.resize(nd.getInternalID() + 1);

    /* resize the request flag */
    __request_flags.getRValue().resize(nd.getInternalID() + 1);
    __release_flags.getRValue().resize(nd.getInternalID() + 1);

    /* Initialize the node list and request flags */
    for(unsigned int i = current_size; i < __node_list.size(); ++i) {
      __node_list[i] = 0;
      __request_flags.getRValue()[i] = true;
      __release_flags.getRValue()[i] = true;
    }

    NodeDescription* new_nd = new NodeDescription(nd);
    __node_list[nd.getInternalID()] = new_nd;
    __internal_counter = __node_list.size();
    GLOBAL_DEBUG_STREAM2("gd: "<< "Adding new NodeDescription " << nd.getNodeName() << " with preset ID " << nd.getInternalID() << " >= size." << endl);
  }
  // node descr already has id < size -> add node desrc at position of id if no other node descr exists
  else { 
    if(__node_list[nd.getInternalID()] == 0) {
      NodeDescription* new_nd = new NodeDescription(nd);
      __node_list[nd.getInternalID()] = new_nd;
      GLOBAL_DEBUG_STREAM2("gd: "<< "Adding new NodeDescription " << nd.getNodeName() << " with preset ID " << nd.getInternalID() << " < size." << endl);
    }
    else {
      GLOBAL_DEBUG_STREAM2("A NodeDescription " << __node_list[nd.getInternalID()]->getNodeName() << " with ID " << nd.getInternalID() << " already exists for " << nd.getNodeName() << endl);
    }
  }
}

unsigned int GraphDescription::addEdge(const NodeDescription& nd_from, 
			       const NodeDescription& nd_to  , const string& in_tag,
			       const Format* format)
{
  return addEdge(nd_from, "default", nd_to, in_tag, format);
}

unsigned int GraphDescription::addEdge(const NodeDescription& nd_from, const string& out_tag, 
			       const NodeDescription& nd_to  ,  
			       const Format* format)
{
  return addEdge(nd_from, out_tag, nd_to, "default", format);
}

unsigned int GraphDescription::addEdge(const NodeDescription& nd_from,  
			       const NodeDescription& nd_to  ,  
			       const Format* format)
{
  return addEdge(nd_from, "default", nd_to, "default", format);
}

unsigned int GraphDescription::addEdge(const NodeDescription& nd_from, const string& out_tag, 
			       const NodeDescription& nd_to  , const string& in_tag,
			       const Format* format)
{
  addNode(const_cast<NodeDescription& >(nd_from));
  addNode(const_cast<NodeDescription& >(nd_to));
  
  int current_id =  __edges.getRValue().size();
  __edges.getRValue().push_back(EdgeDescription(nd_from.getInternalID(), out_tag, nd_to.getInternalID(), in_tag, format, current_id));
  __out_edges.getRValue()[nd_from.getInternalID()].push_back(__edges.getRValue().size() - 1);
  __in_edges.getRValue() [nd_to.getInternalID()]  .push_back(__edges.getRValue().size() - 1);
  
  GLOBAL_DEBUG_STREAM2("gd: "<< "addEdge " << __edges.getRValue().size() - 1 << " " << nd_from.getInternalID() << "::" << out_tag << " -> " << nd_to.getInternalID() << "::" << in_tag << endl);
  
  return ( __edges.getRValue().size() - 1 );
}

unsigned int GraphDescription::addEdgeFromIDs(int nd_from, const string& out_tag, 
					int nd_to  , const string& in_tag,
					const Format* format)
{
  int current_id =  __edges.getRValue().size();
  __edges.getRValue().push_back(EdgeDescription(nd_from, out_tag, nd_to, in_tag, format, current_id));
  __out_edges.getRValue()[nd_from].push_back(__edges.getRValue().size() - 1);
  __in_edges.getRValue() [nd_to]  .push_back(__edges.getRValue().size() - 1);
  
  GLOBAL_DEBUG_STREAM2("gd: "<< "addEdge " << __edges.getRValue().size() - 1 << " " << nd_from << "::" << out_tag << " -> " << nd_to << "::" << in_tag<< ":: " << endl);
  
  return ( __edges.getRValue().size() - 1 );
}

unsigned int GraphDescription::addInEdge(const NodeDescription& nd_to, const string& in_tag, 
				 const Format* format)
{
  addNode(const_cast<NodeDescription& >(nd_to));

  int current_id =  __edges.getRValue().size();
  __edges.getRValue().push_back(EdgeDescription(EdgeDescription::UNCONNECTED_IN_ID, "default", nd_to.getInternalID(), in_tag, format, current_id));
  __in_edges.getRValue() [nd_to.getInternalID()].push_back(__edges.getRValue().size() - 1);
		    
  GLOBAL_DEBUG_STREAM2("gd: "<< "addInEdge " << __edges.getRValue().size() - 1 << " UNCONNECTED_IN_ID::default -> " << nd_to.getInternalID() << "::" << in_tag<< endl);

  return ( __edges.getRValue().size() - 1 );
}

unsigned int GraphDescription::addInEdge(const NodeDescription& nd_to,  
				 const Format* format)
{
  return addInEdge(nd_to, "default", format);
}

unsigned int GraphDescription::addOutEdge(const NodeDescription& nd_from, const string& out_tag, 
				  const Format* format)
{
  addNode(const_cast<NodeDescription& >(nd_from));

  int current_id =  __edges.getRValue().size();
  __edges.getRValue().push_back(EdgeDescription(nd_from.getInternalID(), out_tag, EdgeDescription::UNCONNECTED_OUT_ID, "default", format, current_id));
  __out_edges.getRValue()[nd_from.getInternalID()].push_back(__edges.getRValue().size() - 1);

  GLOBAL_DEBUG_STREAM2("gd: "<< "addEdge " << __edges.getRValue().size() - 1 << " " << nd_from.getInternalID() << "::" << out_tag << " -> UNCONNECTED_OUT_ID::default" << endl);

  return ( __edges.getRValue().size() - 1 );
}

unsigned int GraphDescription::addOutEdge(const NodeDescription& nd_from,  
				  const Format* format)
{
  return addOutEdge(nd_from, "default", format);
}

unsigned int GraphDescription::addOutHelpEdge(const NodeDescription& nd_from, const string& out_tag, 
					const Format* format)
{
  addNode(const_cast<NodeDescription& >(nd_from));

  int current_id =  __edges.getRValue().size();
  __edges.getRValue().push_back(EdgeDescription(nd_from.getInternalID(), out_tag, EdgeDescription::HELP_EDGE_ID, "default", format, current_id));
  __out_edges.getRValue()[nd_from.getInternalID()].push_back(__edges.getRValue().size() - 1);

  GLOBAL_DEBUG_STREAM2("gd: "<< "addEdge " << __edges.getRValue().size() - 1 << " " << nd_from.getInternalID() << "::" << out_tag << " -> UNCONNECTED_OUT_ID::default" << endl);

  return ( __edges.getRValue().size() - 1 );
}

int GraphDescription::getEdgeID(const NodeDescription& nd_from, const string& out_tag,
                                    const NodeDescription& nd_to  , const string&  in_tag) const
{
  for(unsigned int i = 0; i < __edges.getRValue().size(); ++i) {
    const EdgeDescription& e = __edges.getRValue()[i];
    if( (e.getOutputNodeID()  == nd_from.getInternalID()) && 
	(e.getInputNodeID() == nd_to.getInternalID()) && 
	(e.getOutputJackTag() == out_tag) && 
	(e.getInputJackTag() == in_tag) ) 
    {
      return i;
    }
  }
  
  return -1;
}

int GraphDescription::getEdgeID(const NodeDescription& nd_from, 
                                    const NodeDescription& nd_to  , const string&  in_tag) const
{
  return getEdgeID(nd_from, "default", nd_to, in_tag);
}

int GraphDescription::getEdgeID(const NodeDescription& nd_from, const string& out_tag,
                                    const NodeDescription& nd_to) const
{
  return getEdgeID(nd_from, out_tag, nd_to, "default");
}

int GraphDescription::getEdgeID(const NodeDescription& nd_from, 
                                    const NodeDescription& nd_to) const
{
  return getEdgeID(nd_from, "default", nd_to, "default");
}

InstreamChannel* GraphDescription::getInstreamChannel(unsigned int edge_id) const {
  assert(edge_id < __edges.getRValue().size());
  return __edges.getRValue()[edge_id].getInstreamChannel();
}

void GraphDescription::showAll()
{
  showEdges();
  
  for(unsigned int j = 0; j < __internal_counter; ++j) {
    GLOBAL_DEBUG_STREAM2("gd: "<< "node " << j << endl);
    GLOBAL_DEBUG_STREAM2("gd: "<< "  out" << endl);
    const list<int>& out_l = __out_edges.getRValue()[j];
    for(list<int>::const_iterator iter = out_l.begin(); iter != out_l.end(); ++iter) {
      GLOBAL_DEBUG_STREAM2("gd: "<< "    " << *iter << endl);
    }
    GLOBAL_DEBUG_STREAM2("gd: "<< "  in" << endl);
    const list<int>& in_l = __in_edges.getRValue()[j];
    for(list<int>::const_iterator iter = in_l.begin(); iter != in_l.end(); ++iter) {
      GLOBAL_DEBUG_STREAM2("gd: "<< "    " << *iter << endl);
    }
  }
}

void GraphDescription::showEdges()
{
  for(unsigned int i = 0; i < __edges.getRValue().size(); ++i) {
    const EdgeDescription& e = __edges.getRValue()[i];
    GLOBAL_DEBUG_STREAM2("gd: "<< "edge " << i << " " << e.getOutputNodeID() << "::" << e.getOutputJackTag() << " -> " << e.getInputNodeID() << "::" << e.getInputJackTag() << endl);
  }

}


void GraphDescription::removeEdge(const int edge_num)
{
  // get edge to remove
  const EdgeDescription& e = __edges.getRValue()[edge_num];
  //GLOBAL_DEBUG_STREAM2("gd: "<< "removeEdge " << edge_num << " " << e.getOutputNodeID() << "::" << e.getOutputJackTag() << " -> " << e.getInputNodeID() << "::" << e.getInputJackTag() << endl);

  // get id of node description for outgoing edge
  int out_num = e.getOutputNodeID();
  // check if out_num represents a connected id
  if(out_num != EdgeDescription::UNCONNECTED_OUT_ID) {
    // try to get list of outgoing edges for node descr with id out_num
    InternalIt out_iter = __out_edges.getRValue().find(out_num);
    // delete edge
    if(out_iter != __out_edges.getRValue().end()) {
      (*out_iter).second.remove(edge_num);
    }
    else {
      GLOBAL_ERROR_STREAM2("gd: "<< "ERROR out_iter == __out_edges.getRValue().end()" << endl);
    }
  }
  
  // get id of node description for ingoing edge
  int in_num  = e.getInputNodeID();
  // check if in_num represents a connected id
  if(in_num != EdgeDescription::UNCONNECTED_IN_ID) {
    // try to get list of ingoing edges for node descr with id in_num
    InternalIt in_iter = __in_edges.getRValue().find(in_num);
    // delete edge
    if(in_iter != __in_edges.getRValue().end()) {
      (*in_iter).second.remove(edge_num);
    }
    else {
      GLOBAL_ERROR_STREAM2("gd: "<< "ERROR in_iter == __in_edges.getRValue().end()" << endl);
    }
  }

  // erase edge (do not, otherwise indices also have to be updated!)
  // __edges.getRValue().erase( ((__edges.getRValue()).begin() + edge_num) );
}

void GraphDescription::removeSubgraph(const int edge_num)
{
  // get edge to remove
  const EdgeDescription& e = __edges.getRValue()[edge_num];

  // get id of node description for ingoing edge
  int in_num  = e.getInputNodeID();
  // check if in_num represents a connected id
  if(in_num != EdgeDescription::UNCONNECTED_IN_ID) {
    // try to get list of ingoing edges for node descr with id in_num
    InternalIt in_iter = __in_edges.getRValue().find(in_num);
    // delete all incoming edges 
    if(in_iter != __in_edges.getRValue().end()) {
      list<int> l = (*in_iter).second;
      for(list<int>::const_iterator i = l.begin(); i != l.end(); ++i) {
	removeEdge(*i);
	GLOBAL_DEBUG_STREAM2("gd: "<< "edge_num " << edge_num << " removeEdge " << *i << endl);
      }
    }
    else {
      GLOBAL_ERROR_STREAM2("gd: "<< "ERROR in_iter == __in_edges.getRValue().end()" << endl);
    }

    // call removeSubgraph for all outgoing edges
    // try to get list of outgoing edges for node descr with id in_num
    InternalIt out_iter = __out_edges.getRValue().find(in_num);
    if(out_iter != __out_edges.getRValue().end()) {
      list<int> l = (*out_iter).second;
      for(list<int>::const_iterator i = l.begin(); i != l.end(); ++i) {
        removeSubgraph(*i);
      }
    }
    else {
      // GLOBAL_ERROR_STREAM2("gd: "<< "No more outgoing edges" << endl);
    }
  }
}

void GraphDescription::addEdges(const NodeDescription* n1, const NodeDescription* n2, const NodeDescription* n3, 
				const NodeDescription* n4, const NodeDescription* n5, const NodeDescription* n6,
				const NodeDescription* n7, const NodeDescription* n8, const NodeDescription* n9,
				const NodeDescription* n10,const NodeDescription* n11,const NodeDescription* n12)
{
  if(n1 && n2) addEdge(*n1, *n2);
  if(n2 && n3) addEdge(*n2, *n3);
  if(n3 && n4) addEdge(*n3, *n4);
  if(n4 && n5) addEdge(*n4, *n5);
  if(n5 && n6) addEdge(*n5, *n6);
  if(n6 && n7) addEdge(*n6, *n7);
  if(n7 && n8) addEdge(*n7, *n8);
  if(n8 && n9) addEdge(*n8, *n9);
  if(n9 && n10) addEdge(*n9, *n10);
  if(n10 && n11) addEdge(*n10, *n11);
  if(n11 && n12) addEdge(*n11, *n12);
}

EdgeDescription::EdgeDescription()
  : SerializedValue(__EdgeDescription_type_name, __EdgeDescription_type_ns)
{
  registerAttributes();
  __out_id.setValue(UNCONNECTED_OUT_ID);
  __out_tag.setValue("");
  __in_id.setValue(UNCONNECTED_IN_ID);
  __in_tag.setValue("");
  __format.setValue(0);
  __overlap_type.setValue(NONE);
  __q_out_id.setValue(UNCONNECTED_OUT_ID);
  __q_in_id.setValue(UNCONNECTED_IN_ID);
  __id.setValue(UNDEF_EDGE_ID);
  __q_id.setValue(UNDEF_EDGE_ID);
  __r_id.setValue(UNDEF_EDGE_ID);
  __instream_channel.setValue(0);
}

EdgeDescription::EdgeDescription(int out_id, const string& out_tag, int in_id, const string& in_tag, const Format* format, int id)
  : SerializedValue(__EdgeDescription_type_name, __EdgeDescription_type_ns)
{
  registerAttributes();
  __out_id.setValue(out_id);
  __out_tag.setValue(out_tag);
  __in_id.setValue(in_id);
  __in_tag.setValue(in_tag);
  if (format) {
    __format.setValue(new TValue<Format>(*format));
  } else {
    __format.setValue(0);
  }
  __overlap_type.setValue(NONE);
  __q_out_id.setValue(UNCONNECTED_OUT_ID);
  __q_in_id.setValue(UNCONNECTED_IN_ID);
  __id.setValue(id);
  __q_id.setValue(UNDEF_EDGE_ID);
  __r_id.setValue(UNDEF_EDGE_ID);
  
  __instream_channel.setValue(0);
}

EdgeDescription::EdgeDescription(const EdgeDescription& edge)
  : SerializedValue(edge), __instream_channel(0)
{
  registerAttributes();
  *this = edge;
}

EdgeDescription::~EdgeDescription() {
  delete __format.getValue();
  delete __instream_channel.getValue();
}

void EdgeDescription::setFormat(const Format* fom) {
  delete __format.getValue();
  
  if(fom) {
    __format.setValue(new TValue<Format>(*fom));
  } else {
    __format.setValue(0);
  }
}

EdgeDescription& EdgeDescription::operator=(const EdgeDescription& edge) {
  if(this != &edge) {
    __out_id    = edge.__out_id;
    __out_tag   = edge.__out_tag;
    __in_id     = edge.__in_id;
    __in_tag    = edge.__in_tag;
    
    /* delete the old format */
    delete __format.getValue ();
    
    if (edge.__format.getValue() != 0) {
      __format.setValue(new TValue<Format>(*edge.__format.getValue()));
    }
    else {
      __format.setValue(0);
    }
    
    __overlap_type = edge.__overlap_type;
    __q_out_id     = edge.__q_out_id;
    __q_in_id      = edge.__q_in_id;
    
    __id = edge.__id;
    
    __q_id = edge.__q_id;
    __r_id = edge.__r_id;
    
    __binding_description = edge.__binding_description;
    
    if(edge.__instream_channel.getValue())
      __instream_channel.setValue(edge.__instream_channel.getValue()->copy());
    else
      __instream_channel.setValue(0);
  }
  
  return *this;
}

bool EdgeDescription::isIdenticalOverlap(const EdgeDescription& edge) const {
  if(getLeftOutputNodeID() != edge.getLeftOutputNodeID()) {
    return false;
  }
  if(getLeftInputNodeID() != edge.getLeftInputNodeID()) {
    return false;
  }
  if(getLeftEdgeID() != edge.getLeftEdgeID()) {
    return false;
  }
  if(getRightEdgeID() != edge.getRightEdgeID()) {
    return false;
  }
  return true;
}

void EdgeDescription::registerAttributes() {
  insertArgument(&__out_id);
  insertArgument(&__out_tag);
  insertArgument(&__in_id);
  insertArgument(&__in_tag);
  insertArgument(&__format);
  insertArgument(&__overlap_type);
  insertArgument(&__q_out_id);
  insertArgument(&__q_in_id);
  insertArgument(&__id);
  insertArgument(&__q_id);
  insertArgument(&__r_id);
  insertArgument(&__binding_description);
  insertArgument(&__instream_channel);
}

void GraphDescription::setISession(const ISession& session) {
  delete __i_session.getValue();
  __i_session.setValue(dynamic_cast<ISession*>(session.copy()));
}

void GraphDescription::setSinkSynchronizerDescription(const ObjectDescription& od)
{
  __sink_synch_descr.setValue(od);
}

void GraphDescription::realizeCommunicationChannel(InstreamChannel* c_channel, EdgeDescription& edge)
{
  if(c_channel == 0) {
    GLOBAL_ERROR_STREAM2("gd: CommunicationChannel is 0" << endl);
  }
  
  /* add all transport strategies */
  for(unsigned int i=0; i<edge.getBindingDescription().getNumberOfTransportStrategies(); ++i) {
    const InstreamStrategyDescription& ts_desc = edge.getBindingDescription().getTransportStrategy(i);
    
    if(ts_desc.getRoles() == 0) {
      GLOBAL_WARNING_STREAM("GraphDescription: Edge " << edge.getEdgeID() <<
                            " has strategy " << ts_desc.getStrategyType() <<
                            " without any communication roles" << endl);
      continue;
    }
    
    if(ts_desc.isExclusive()) {
      c_channel->addExplicitTransportStrategy(ts_desc.getStrategyType(), ts_desc.getRoles());
    } else {
      c_channel->addTransportStrategy(ts_desc.getStrategyType(), ts_desc.getRoles());
    }
  }  
  
  /* establish the binding, so we can configure the strategies */
  c_channel->establishBinding();
  
  /* BEGIN> configure the top-level strategy */
  /* upstream */
  const CEvent& upstream_top_level_config = edge.getBindingDescription().getUpstreamTopLevelConfig();
  if(upstream_top_level_config.size() != 0) {
    ITransportStrategy_var upstream_top_level_ts(c_channel->getUpstreamInterface<ITransportStrategy>());
    for(unsigned j=0; j<upstream_top_level_config.size(); ++j) {
      Event event = upstream_top_level_config[j];
      upstream_top_level_ts->getParentObject()->receiveEvent(event);
    }
  }
  
  /* downstream */
  const CEvent& downstream_top_level_config = edge.getBindingDescription().getDownstreamTopLevelConfig();
  if(downstream_top_level_config.size() != 0) {
    ITransportStrategy_var downstream_top_level_ts(c_channel->getDownstreamInterface<ITransportStrategy>());
    for(unsigned j=0; j<downstream_top_level_config.size(); ++j) {
      Event event = downstream_top_level_config[j];
      downstream_top_level_ts->getParentObject()->receiveEvent(event);
    }
  }
  /* END< configure the top-level strategy */
  
  /* BEGIN> configure all transport strategies */
  for(unsigned int i=0; i<edge.getBindingDescription().getNumberOfTransportStrategies(); ++i) {
    const InstreamStrategyDescription& ts_desc = edge.getBindingDescription().getTransportStrategy(i);
    
    if(ts_desc.getRoles() == 0) {
      GLOBAL_WARNING_STREAM("GraphDescription: Edge " << edge.getEdgeID() <<
                            " has strategy " << ts_desc.getStrategyType() <<
                            " without any communication roles" << endl);
      continue;
    }    
    
    /* get upstream and downstream strategy config */
    const CEvent& upstream_config = ts_desc.getUpstreamConfig();
    const CEvent& downstream_config = ts_desc.getDownstreamConfig();
    
    /* get one role so we can request an interface */
    unsigned int roles = ts_desc.getRoles();
    InstreamChannel::CommunicationRole role;
    if(roles & InstreamChannel::DOWNSTREAM_BUFFER) {
      role = InstreamChannel::DOWNSTREAM_BUFFER;
    } else if(roles & InstreamChannel::DOWNSTREAM_CEVENT) {
      role = InstreamChannel::DOWNSTREAM_CEVENT;
    } else if(roles & InstreamChannel::UPSTREAM_BUFFER) {
      role = InstreamChannel::UPSTREAM_BUFFER;
    } else if(roles & InstreamChannel::UPSTREAM_CEVENT) {
      role = InstreamChannel::UPSTREAM_CEVENT;
    } else {
      // NEVER HAPPENS
      assert(false);
    }
    
    /* upstream */
    if(upstream_config.size() != 0) {
      ITransportStrategy_var ts(c_channel->getUpstreamInterface<ITransportStrategy>(role));
      for(unsigned j=0; j<upstream_config.size(); ++j) {
        Event event = upstream_config[j];
        ts->getParentObject()->receiveEvent(event);
      }
    }
    
    /* downstream */
    if(downstream_config.size() != 0) {
      ITransportStrategy_var ts(c_channel->getDownstreamInterface<ITransportStrategy>(role));
      for(unsigned j=0; j<downstream_config.size(); ++j) {
        Event event = downstream_config[j];
        ts->getParentObject()->receiveEvent(event);
      }
    }
    
    /* set value streams */
    if(!ts_desc.getIValueStream().empty() || !ts_desc.getOValueStream().empty()) {
      /* set value stream by type */
      c_channel->setValueStream(roles, ts_desc.getIValueStream(), ts_desc.getOValueStream());
    } else if(ts_desc.getIValueStreamStack() || ts_desc.getOValueStreamStack()) {
      /* set value stream by ValueStream object */
      c_channel->setValueStreamStack(roles, ts_desc.getIValueStreamStack(), ts_desc.getOValueStreamStack());
    }
  }
  /* END< configure all transport strategies */
}

void GraphDescription::realizeGraph() throw(const BindingException&, const SerializedException&)
{
  /* Begin> Send all events for state CONSTRUCTED to the nodes */
  for (unsigned int i = 0; i < __node_list.size(); ++i) {
    if ( __node_list[i]->getSharingType() != NodeDescription::SHARED ) {
      if (__node_list[i]->getNodeEvent(INode::CONSTRUCTED)) {
        __interface_list[i]->reachConstructed();
        sendNodeEvent(__interface_list[i], *__node_list[i]->getNodeEvent(INode::CONSTRUCTED));
      }
    }
  }
  /* End< */

  /* Begin> Send all events for state INITIALIZED to the nodes */
  for (unsigned int i = 0; i < __node_list.size(); ++i) {
    if ( __node_list[i]->getSharingType() != NodeDescription::SHARED ) {
      if (__node_list[i]->getNodeEvent(INode::INITIALIZED)) {
        __interface_list[i]->reachInitialized();
        sendNodeEvent(__interface_list[i], *__node_list[i]->getNodeEvent(INode::INITIALIZED));
      }
    }
  }
  /* End< */

  GraphDescription gd(*this);

  list<int> graph_sources;

  // get all sources of query graph
  const SequenceValue< NodeDescription >& nds = gd.getNodeDescriptionList();
  for(unsigned int n = 0; n < nds.size(); ++n) {
    // node with no incoming edge and some outgoing edges -> source that needs to be connected
    if(gd.getInEdgeList(n).empty() && !gd.getOutEdgeList(n).empty()) {
      graph_sources.push_back(n);
    }
    // node with only unconnected incoming edges and some outgoing edges -> node that needs to be connected
    if(!gd.getOutEdgeList(n).empty()) {
      list<int>::const_iterator in_edge_iter;
      for(in_edge_iter = gd.getInEdgeList(n).begin(); in_edge_iter != gd.getInEdgeList(n).end(); ++in_edge_iter) {
	EdgeDescription& in_edge = gd.getEdge(*in_edge_iter);
	if(in_edge.isOutConnected()) { // -> outgoind part is connected -> connected edge -> no 'source' node
	  break;
	}
      }
      if(in_edge_iter == gd.getInEdgeList(n).end()) {
	graph_sources.push_back(n);
      }
    }
  }

  // for all sources in graph
  for(list<int>::iterator i = graph_sources.begin(); i != graph_sources.end(); ++i) {
    INode* left_node = gd.getINode((nds[*i]));
    if (!left_node) {
      throw BindingException("Node is not available in the GraphDescription");
    }

    // used to store all unconnected out jacks and add them to this graph descrition in the end
    // this is needed to connect to unconnected out jacks in shared graph descriptions
    list<string> left_node_unconnected_out_jacks;

    // stores edges that have been connected and that can be removed from copy of this graph description (namely gd)
    list<int> edges_to_remove;

    // try to connect all outgoing edges
    const list<int>& out_list = gd.getOutEdgeList(*i);
    for(list<int>::const_iterator iter = out_list.begin(); iter != out_list.end(); ++iter) {
      EdgeDescription& edge = (gd.getEdge(*iter));

      // check if edge is connected at all
      if(edge.isInConnected() && edge.isOutConnected()) {

	if(edge.getOverlapType() == EdgeDescription::COMPLETE) { // nothing to do, simply remove edge
	  edges_to_remove.push_back(*iter);
	}
	else if(edge.getOverlapType() == EdgeDescription::PARTIAL_COPY || 
		edge.getOverlapType() == EdgeDescription::PARTIAL_OUTPUT) { // add new connection

	  INode* right_node = gd.getINode(nds[edge.getInputNodeID()]);
	  if (!right_node) {
	    throw BindingException("Node is not available in the GraphDescription");
	  }
	  right_node->reachInitialized();
	  left_node->reachOutputInitialized();
	  /* Begin> Check if events in state OUTPUT_INITIALIZED must be send to the left_node */
          if ( nds[edge.getInputNodeID()]->getSharingType() != NodeDescription::SHARED ) {
	    if (nds[edge.getInputNodeID()]->getNodeEvent(INode::OUTPUT_INITIALIZED)) {
              sendNodeEvent(left_node, *nds[edge.getInputNodeID()]->getNodeEvent(INode::OUTPUT_INITIALIZED));
	    }
          }
	  /* End< */
	  // first time: get all unconnected out jacks of left node
	  if(left_node_unconnected_out_jacks.empty()) {
	    left_node_unconnected_out_jacks = left_node->getOutputStreamTags();
	  }
          
          /* if no format is set, create one */
          if(getEdge(edge.getEdgeID()).getFormat() == 0) {
            Format dummy_format;
            getEdge(edge.getEdgeID()).setFormat(&dummy_format);
          }
          
	  // create shared (copied) connection or connection to unconnect output
	  if(edge.getOverlapType() == EdgeDescription::PARTIAL_COPY) {
            if(edge.getBindingDescription().isImplicit()) {
	      /* If no specific transport strategy should be used, use the default s_connect method */
              InstreamChannel* channel = dynamic_cast<InstreamChannel*>(c_s_connect(left_node, edge.getOutputJackTag(), right_node, edge.getInputJackTag(), getEdge( edge.getEdgeID() ).getFormat()));
              /* Store the returned instream channel */
              getEdge(*iter).setInstreamChannel( channel );
	      
              cout << "s_connect( " << (*i) << ", " << edge.getInputNodeID() << " )" << endl;
            } else {
	      /* Otherwise use the c_s_connect method */
 	      InstreamChannel_var c_channel(c_s_connect(left_node, edge.getOutputJackTag(), right_node, edge.getInputJackTag(), getEdge( edge.getEdgeID() ).getFormat()));
 	      realizeCommunicationChannel(c_channel.get(), edge);
 	      cout << "c_s_connect( " << (*i) << ", " << edge.getInputNodeID() << " ) using explicit binding" << endl;
              
              /* Store the returned instream channel */
              getEdge(*iter).setInstreamChannel(c_channel.get());
 	    }
            realizeCommunicationChannel(getEdge(*iter).getInstreamChannel(), edge);
	  }
	  else if(edge.getOverlapType() == EdgeDescription::PARTIAL_OUTPUT) {
	    Format* f = edge.getFormat();
            
            /* If no connection format is set for this edge, we create it */
            if(!getEdge(edge.getEdgeID()).getFormat()) {
              Format dummy_format;
              getEdge(edge.getEdgeID()).setFormat(&dummy_format);
            }
            
            if(edge.getBindingDescription().isImplicit()) {
	      InstreamChannel* channel = dynamic_cast<InstreamChannel*>(c_connect(left_node, edge.getOutputJackTag(), right_node, edge.getInputJackTag(), f, getEdge( edge.getEdgeID() ).getFormat()));
              /* Store the returned instream channel */
              getEdge(*iter).setInstreamChannel( channel );
              
	      cout << "connect( " << (*i) << ", " << edge.getInputNodeID() << " ) (other output)" << endl;
 	    } else {
 	      InstreamChannel_var c_channel(c_connect(left_node, edge.getOutputJackTag(), right_node, edge.getInputJackTag(), f, getEdge( edge.getEdgeID() ).getFormat()));
 	      realizeCommunicationChannel(c_channel.get(), edge);
 	      cout << "connect( " << (*i) << ", " << edge.getInputNodeID() << " ) (other output) using explicit binding" << endl;
              
              /* Store the returned instream channel */
              getEdge(*iter).setInstreamChannel( dynamic_cast<InstreamChannel*>(c_channel->copy() ));
 	    }
	    realizeCommunicationChannel(getEdge(*iter).getInstreamChannel(), edge);
	  }
	  edges_to_remove.push_back(*iter);
	  left_node_unconnected_out_jacks.remove(edge.getOutputJackTag());
	} // end: if(PARTIAL_COPY or PARTIAL_OUTPUT)
	else if(edge.getOverlapType() == EdgeDescription::NONE) { // no overlap -> normal connection
	  INode* right_node = gd.getINode(nds[edge.getInputNodeID()]);
	  if (!right_node) {
	    throw BindingException("Node is not available in the GraphDescription");
	  }
	  right_node->reachInitialized();

	  left_node->reachOutputInitialized();
	  /* Begin> Check if events in state OUTPUT_INITIALIZED must be send to the node */
          if (nds[*i]->getSharingType() != NodeDescription::SHARED) {
	    if (nds[*i]->getNodeEvent(INode::OUTPUT_INITIALIZED)) {
              sendNodeEvent(left_node, *nds[*i]->getNodeEvent(INode::OUTPUT_INITIALIZED));
            }
          }
	  /* End< */
          
	  // first time: get all unconnected out jacks of left node
	  if(left_node_unconnected_out_jacks.empty()) {
	    left_node_unconnected_out_jacks = left_node->getOutputStreamTags();
	  }
          
	  Format* f = edge.getFormat();
          
          /* If no connection format is set for this edge, we create it */
          if(!getEdge(edge.getEdgeID()).getFormat()) {
            Format dummy_format;
            getEdge(edge.getEdgeID()).setFormat(&dummy_format);
          }
          
          if(edge.getBindingDescription().isImplicit()) {
	    try {
              InstreamChannel* channel = dynamic_cast<InstreamChannel*>(c_connect(left_node, edge.getOutputJackTag(), right_node, edge.getInputJackTag(), f, getEdge( edge.getEdgeID() ).getFormat()));
              /* Store the returned instream channel */
              getEdge(*iter).setInstreamChannel( channel );
              
              realizeCommunicationChannel(getEdge(*iter).getInstreamChannel(), edge);
              cout << "connect( " << (*i) << ", " << edge.getInputNodeID() << " )" << endl;
	    }
	    catch(const BindingException& e) { // jack already connected? => try to connect within jackgroup
	      InstreamChannel* channel = dynamic_cast<InstreamChannel*>(c_s_connect(left_node, edge.getOutputJackTag(), right_node, edge.getInputJackTag(), getEdge( edge.getEdgeID() ).getFormat()));
              /* Store the returned instream channel */
              getEdge(*iter).setInstreamChannel( channel );

              realizeCommunicationChannel(getEdge(*iter).getInstreamChannel(), edge);
	      cout << "could not connect( " << (*i) << ", " << edge.getInputNodeID() << " ), "
		   << "used s_connect( " << (*i) << ", " << edge.getInputNodeID() << " ) instead." << endl;
	    }
 	  } else {
 	    try {
 	      InstreamChannel_var c_channel(c_connect(left_node, edge.getOutputJackTag(), right_node, edge.getInputJackTag(), f, getEdge( edge.getEdgeID() ).getFormat()));
 	      realizeCommunicationChannel(c_channel.get(), edge);
              getEdge(*iter).setInstreamChannel( dynamic_cast<InstreamChannel*>(c_channel->copy()) );
              cout << "c_connect( " << (*i) << ", " << edge.getInputNodeID() << " ) using using explicit binding " << endl;
 	    }
 	    catch(const BindingException& e) { // jack already connected? => try to connect within jackgroup
 	      InstreamChannel_var c_channel(c_s_connect(left_node, edge.getOutputJackTag(), right_node, edge.getInputJackTag(), getEdge( edge.getEdgeID() ).getFormat()));
 	      realizeCommunicationChannel(c_channel.get(), edge);
              getEdge(*iter).setInstreamChannel( dynamic_cast<InstreamChannel*>(c_channel->copy()) );
 	      cout << "could not c_connect( " << (*i) << ", " << edge.getInputNodeID() << " ), "
 		   << "used c_s_connect( " << (*i) << ", " << edge.getInputNodeID() << " ) instead "
 		   << "using explicit binding" << endl;
 	    }
 	  }
          
	  edges_to_remove.push_back(*iter);
	  left_node_unconnected_out_jacks.remove(edge.getOutputJackTag());
	} // end: if(NONE): no overlap -> normal connection
	else {
	  GLOBAL_ERROR_STREAM2("gd: "<< "Unhandled case in realize graph!" << endl);
	  throw BindingException("Unhandled case in realize graph!");
	}
      } // end: check if edge is connected at all
    } // end: try to connect all outgoing edges

    // since all outputs are now connected -> activate , if out_list is not empty() (not empty() means no sink node?)
    if (!out_list.empty()) {
      // need to activate node since its successor(s) needs data
      left_node->reachActivated(); 
    }

    // remove edges from list of edge indices for nodes
    for(list<int>::iterator remover = edges_to_remove.begin(); remover != edges_to_remove.end(); ++remover) {
      // check, if node with in_id becomes new source node
      int n = gd.getEdge(*remover).getInputNodeID();
      // node with no incoming edge (after remove) and some outgoing edges -> source (or unconnected node) that needs to be connected
      if(gd.getInEdgeList(n).size() == 1 && !gd.getOutEdgeList(n).empty()) {
	graph_sources.push_back(n);
      }

      gd.removeEdge(*remover);
    }

    // add all unconnceted out edges to this graph description
    SequenceValue< NodeDescription >& my_nds = this->getNodeDescriptionList();
    for(list<string>::iterator adder = left_node_unconnected_out_jacks.begin();
	adder != left_node_unconnected_out_jacks.end();
	++adder) {
      this->addOutEdge(*(my_nds[*i]), *adder);
      GLOBAL_DEBUG_STREAM2("gd: " << "Adding unconnected out edge " << *adder << " to node " << my_nds[*i]->getNodeName() << endl);
    }
  } // end: for all sources in graph

  /* Finally update the graphdescription stored in the session */
  if (__i_session.getValue()) {
    __i_session.getValue()->setGraphDescription(*this);
  }
}
// END realizeGraph *******************************************************************

void GraphDescription::realizeSharedGraph() throw(const BindingException&,const SerializedException&)
{
  /* Begin> Send all events for state CONSTRUCTED to the nodes */
  for (unsigned int i = 0; i < __node_list.size(); ++i) {
    if ( __node_list[i]->getSharingType() != NodeDescription::SHARED ) {
      if (__node_list[i]->getNodeEvent(INode::CONSTRUCTED)) {
        __interface_list[i]->reachConstructed();
        sendNodeEvent(__interface_list[i], *__node_list[i]->getNodeEvent(INode::CONSTRUCTED));
      }
    }
  }
  /* End< */

  /* Begin> Send all events for state INITIALIZED to the nodes */
  for (unsigned int i = 0; i < __node_list.size(); ++i) {
    if ( __node_list[i]->getSharingType() != NodeDescription::SHARED ) {
      if (__node_list[i]->getNodeEvent(INode::INITIALIZED)) {
        __interface_list[i]->reachInitialized();
        sendNodeEvent(__interface_list[i], *__node_list[i]->getNodeEvent(INode::INITIALIZED));
      }
    }
  }
  /* End< */

  GraphDescription gd(*this);

  list<int> graph_sources;

  // get all sources of query graph
  const SequenceValue< NodeDescription >& nds = gd.getNodeDescriptionList();
  for(unsigned int n = 0; n < nds.size(); ++n) {
    // node with no incoming edge and some outgoing edges -> source that needs to be connected
    if(gd.getInEdgeList(n).empty() && !gd.getOutEdgeList(n).empty()) {
      graph_sources.push_back(n);
    }
    // node with only unconnected incoming edges and some outgoing edges -> node that needs to be connected
    if(!gd.getOutEdgeList(n).empty()) {
      list<int>::const_iterator in_edge_iter;
      for(in_edge_iter = gd.getInEdgeList(n).begin(); in_edge_iter != gd.getInEdgeList(n).end(); ++in_edge_iter) {
	EdgeDescription& in_edge = gd.getEdge(*in_edge_iter);
	if(in_edge.isOutConnected()) { // -> outgoind part is connected -> connected edge -> no 'source' node
	  break;
	}
      }
      if(in_edge_iter == gd.getInEdgeList(n).end()) {
	graph_sources.push_back(n);
      }
    }
  }

  // for all sources in graph
  for(list<int>::iterator i = graph_sources.begin(); i != graph_sources.end(); ++i) {
    INode* left_node = gd.getINode((nds[*i]));
    if (!left_node) {
      throw BindingException("Node is not available in the GraphDescription");
    }

    // used to store all unconnected out jacks and add them to this graph descrition in the end
    // this is needed to connect to unconnected out jacks in shared graph descriptions
    list<string> left_node_unconnected_out_jacks;

    // stores edges that have been connected and that can be removed from copy of this graph description (namely gd)
    list<int> edges_to_remove;

    // try to connect all outgoing edges
    const list<int>& out_list = gd.getOutEdgeList(*i);
    for(list<int>::const_iterator iter = out_list.begin(); iter != out_list.end(); ++iter) {
      EdgeDescription& edge = (gd.getEdge(*iter));

      // check if edge is connected at all
      if(edge.isInConnected() && edge.isOutConnected()) {

	if(edge.getOverlapType() == EdgeDescription::COMPLETE) { // nothing to do, simply remove edge
	  edges_to_remove.push_back(*iter);
	}
	else if(edge.getOverlapType() == EdgeDescription::PARTIAL_COPY || 
		edge.getOverlapType() == EdgeDescription::PARTIAL_OUTPUT) { // add new connection

	  INode* right_node = gd.getINode(nds[edge.getInputNodeID()]);
	  if (!right_node) {
	    throw BindingException("Node is not available in the GraphDescription");
	  }
	  // unlimit the queue size for right nodes of all partial shared edges
	  // -> input jacks of these nodes will buffer all downstream messages during realizing of graph
	  // -> no blocking, master graph can continue to work
	  right_node->getParentObject()->getInterface<IBiDirQueuedNode>()->unlimitDownstreamSize();

	  right_node->reachInitialized();

	  left_node->reachOutputInitialized();
	  /* Begin> Check if events in state OUTPUT_INITIALIZED must be send to the left_node */
	  if (nds[edge.getInputNodeID()]->getSharingType() != NodeDescription::SHARED ) {
	    if (nds[edge.getInputNodeID()]->getNodeEvent(INode::OUTPUT_INITIALIZED)) {
              sendNodeEvent(left_node, *nds[edge.getInputNodeID()]->getNodeEvent(INode::OUTPUT_INITIALIZED));
	    }
          }
	  /* End< */
	  // first time: get all unconnected out jacks of left node
	  if(left_node_unconnected_out_jacks.empty()) {
	    left_node_unconnected_out_jacks = left_node->getOutputStreamTags();
	  }

          /* If no connection format is set for this edge, we create it */
          if(!getEdge(edge.getEdgeID()).getFormat()) {
            Format dummy_format;
            getEdge(edge.getEdgeID()).setFormat(&dummy_format);
          }
          
	  // create shared (copied) connection or connection to unconnect output
	  if(edge.getOverlapType() == EdgeDescription::PARTIAL_COPY) {
            if(edge.getBindingDescription().isImplicit()) {
	      /* If no specific transport strategy should be used, use the default s_connect method */
	      InstreamChannel* channel = dynamic_cast<InstreamChannel*>(c_s_connect(left_node, edge.getOutputJackTag(), right_node, edge.getInputJackTag(), getEdge( edge.getEdgeID() ).getFormat() ));
              /* Store the returned instream channel */
              getEdge(*iter).setInstreamChannel( channel );

              realizeCommunicationChannel(getEdge(*iter).getInstreamChannel(), edge);
              cout << "s_connect( " << (*i) << ", " << edge.getInputNodeID() << " )" << endl;
	    } else {
	      /* Otherwise use the c_s_connect method */
	      InstreamChannel_var c_channel(c_s_connect(left_node, edge.getOutputJackTag(), right_node, edge.getInputJackTag(), getEdge( edge.getEdgeID() ).getFormat() ));
	      realizeCommunicationChannel(c_channel.get(), edge);
              getEdge(*iter).setInstreamChannel( dynamic_cast<InstreamChannel*>(c_channel->copy()));
              cout << "c_s_connect( " << (*i) << ", " << edge.getInputNodeID() << " ) using explicit binding" << endl;
	    }
	  }
	  else if(edge.getOverlapType() == EdgeDescription::PARTIAL_OUTPUT) {
	    Format* f = edge.getFormat();
            
            /* If no connection format is set for this edge, we create it */
            if(!getEdge(edge.getEdgeID()).getFormat()) {
              Format dummy_format;
              getEdge(edge.getEdgeID()).setFormat(&dummy_format);
            }
            
            if(edge.getBindingDescription().isImplicit()) {
	      InstreamChannel* channel = dynamic_cast<InstreamChannel*>(c_connect(left_node, edge.getOutputJackTag(), right_node, edge.getInputJackTag(), f, getEdge( edge.getEdgeID() ).getFormat()));
              /* Store the returned instream channel */
              getEdge(*iter).setInstreamChannel( channel );

	      cout << "connect( " << (*i) << ", " << edge.getInputNodeID() << " ) (other output)" << endl;
 	    } else {
 	      InstreamChannel_var c_channel(c_connect(left_node, edge.getOutputJackTag(), right_node, edge.getInputJackTag(), f, getEdge( edge.getEdgeID() ).getFormat()));
 	      realizeCommunicationChannel(c_channel.get(), edge);
              getEdge(*iter).setInstreamChannel( dynamic_cast<InstreamChannel*>(c_channel->copy()) );
              cout << "connect( " << (*i) << ", " << edge.getInputNodeID() << " ) (other output) using explicit binding" << endl;
 	    }
	    realizeCommunicationChannel(getEdge(*iter).getInstreamChannel(), edge);
	  }
	  edges_to_remove.push_back(*iter);
	  left_node_unconnected_out_jacks.remove(edge.getOutputJackTag());
	} // end: if(PARTIAL_COPY or PARTIAL_OUTPUT)
	else if(edge.getOverlapType() == EdgeDescription::NONE) { // no overlap -> normal connection
	  INode* right_node = gd.getINode(nds[edge.getInputNodeID()]);
	  if (!right_node) {
	    throw BindingException("Node is not available in the GraphDescription");
	  }
	  right_node->reachInitialized();

	  left_node->reachOutputInitialized();
	  /* Begin> Check if events in state OUTPUT_INITIALIZED must be send to the node */
          if ( nds[*i]->getSharingType() != NodeDescription::SHARED ) {
	    if (nds[*i]->getNodeEvent(INode::OUTPUT_INITIALIZED)) {
              sendNodeEvent(left_node, *nds[*i]->getNodeEvent(INode::OUTPUT_INITIALIZED));
            }
          }
	  /* End< */

	  // first time: get all unconnected out jacks of left node
	  if(left_node_unconnected_out_jacks.empty()) {
	    left_node_unconnected_out_jacks = left_node->getOutputStreamTags();
	  }
          
	  Format* f = edge.getFormat();
          
          /* If no connection format is set for this edge, we create it */
          if(!getEdge(edge.getEdgeID()).getFormat()) {
            Format dummy_format;
            getEdge(edge.getEdgeID()).setFormat(&dummy_format);
          }
          
          if(edge.getBindingDescription().isImplicit()) {
            try {
	      InstreamChannel* channel = dynamic_cast<InstreamChannel*>(c_connect(left_node, edge.getOutputJackTag(), right_node, edge.getInputJackTag(), f, getEdge( edge.getEdgeID() ).getFormat()));
              /* Store the returned instream channel */
              getEdge(*iter).setInstreamChannel( channel );

	      realizeCommunicationChannel(getEdge(*iter).getInstreamChannel(), edge);
	      cout << "connect( " << (*i) << ", " << edge.getInputNodeID() << " )" << endl;
	    }
	    catch(const BindingException& e) { // jack already connected? => try to connect within jackgroup
	      InstreamChannel* channel = dynamic_cast<InstreamChannel*>(c_s_connect(left_node, edge.getOutputJackTag(), right_node, edge.getInputJackTag(), getEdge( edge.getEdgeID() ).getFormat()));
              /* Store the returned instream channel */
              getEdge(*iter).setInstreamChannel( channel );
	      realizeCommunicationChannel(getEdge(*iter).getInstreamChannel(), edge);
	      cout << "could not connect( " << (*i) << ", " << edge.getInputNodeID() << " ), "
		   << "used s_connect( " << (*i) << ", " << edge.getInputNodeID() << " ) instead." << endl;
	    }
	  } else {
	    try {
	      InstreamChannel_var c_channel(c_connect(left_node, edge.getOutputJackTag(), right_node, edge.getInputJackTag(), f,getEdge( edge.getEdgeID() ).getFormat()));
	      realizeCommunicationChannel(c_channel.get(), edge);
              getEdge(*iter).setInstreamChannel( dynamic_cast<InstreamChannel*>(c_channel->copy()) );
	      cout << "c_connect( " << (*i) << ", " << edge.getInputNodeID() << " ) using explicit binding" << endl;
	    }
	    catch(const BindingException& e) { // jack already connected? => try to connect within jackgroup
	      InstreamChannel_var c_channel(c_s_connect(left_node, edge.getOutputJackTag(), right_node, edge.getInputJackTag(), getEdge( edge.getEdgeID() ).getFormat()));
	      realizeCommunicationChannel(c_channel.get(), edge);
              getEdge(*iter).setInstreamChannel( dynamic_cast<InstreamChannel*>(c_channel->copy()) );
              cout << "could not c_connect( " << (*i) << ", " << edge.getInputNodeID() << " ), "
		   << "used c_s_connect( " << (*i) << ", " << edge.getInputNodeID() << " ) instead "
		   << "using explicit binding" << endl;
	    }
	  }
          
	  edges_to_remove.push_back(*iter);
	  left_node_unconnected_out_jacks.remove(edge.getOutputJackTag());
	} // end: if(NONE): no overlap -> normal connection
	else {
	  GLOBAL_ERROR_STREAM2("gd: "<< "Unhandled case in realize graph!" << endl);
	  throw BindingException("Unhandled case in realize graph!");
	}
      } // end: check if edge is connected at all
    } // end: try to connect all outgoing edges

    // stores if left node can already be started after this iteration
    INode* left_node_to_start = 0;

    // remove edges from list of edge indices for nodes
    for(list<int>::iterator remover = edges_to_remove.begin(); remover != edges_to_remove.end(); ++remover) {
      // check, if node with out_id ( = left_node ) has no more unconnected outputs
      int left_node_id = gd.getEdge(*remover).getOutputNodeID();
      list<int> left_out_edges = gd.getOutEdgeList(left_node_id);
      // node with no outgoing edge (after remove) -> reach ACTIVATED
      if(left_out_edges.size() == 1) {
	left_node_to_start = gd.getINodeList()[left_node_id];
	left_node_to_start->reachActivated();
      }
      else if(edges_to_remove.size() == 1) { // node with only unconnected outgoing edges (after remove!) -> reach ACTIVATED
	// tricky: count connected edges -> must be equal to size of edges_to_remove list
	unsigned int connected_out_edges = 0;
	for(list<int>::iterator iter = left_out_edges.begin(); iter != left_out_edges.end(); ++iter) {
	  if(gd.getEdge(*iter).isOutConnected()) {
	    ++connected_out_edges;
	  }
	}
	if(connected_out_edges == edges_to_remove.size()) {
	  left_node_to_start = gd.getINodeList()[left_node_id];
	  left_node_to_start->reachActivated();
	}
      }

      // check, if node with in_id ( = right node ) becomes new source node
      int right_node_id = gd.getEdge(*remover).getInputNodeID();
      // node with no incoming edge (after remove) ... 
      if(gd.getInEdgeList(right_node_id).size() == 1) {
	// ... and some outgoing edges -> source (or unconnected node) that needs to be connected
	if(!gd.getOutEdgeList(left_node_id).empty()) {
	  graph_sources.push_back(right_node_id);
	}
	
	// in any case, node has no more incoming edges that need to be connected -> reach OUTPUT_INITIALIZED
	INode* right_node = gd.getINodeList()[right_node_id];
	right_node->reachOutputInitialized();
      }

      // finally, really remove the edge
      gd.removeEdge(*remover);
    }

    if(left_node_to_start) {
      left_node_to_start->reachStarted();
      cerr << "NOWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWWW " << endl << endl;
      string s;
      cin >> s;
    }

    // add all unconnceted out edges to this graph description
    SequenceValue< NodeDescription >& my_nds = this->getNodeDescriptionList();
    for(list<string>::iterator adder = left_node_unconnected_out_jacks.begin();
	adder != left_node_unconnected_out_jacks.end();
	++adder) {
      this->addOutEdge(*(my_nds[*i]), *adder);
      GLOBAL_DEBUG_STREAM2("gd: " << "Adding unconnected out edge " << *adder << " to node " << my_nds[*i]->getNodeName() << endl);
    }
  } // end: for all sources in graph


  /* Finally update the graphdescription stored in the session */
  if (__i_session.getValue()) {
    __i_session.getValue()->setGraphDescription(*this);
  }
}
// END realizeSharedGraph *******************************************************************

void GraphDescription::realizeHandoverGraph(const GraphDescription& master_gd) throw(const BindingException&,const SerializedException&)
{
  /* Begin> Send all events for state CONSTRUCTED to the nodes */
  for (unsigned int i = 0; i < __node_list.size(); ++i) {
    if (__node_list[i]->getNodeEvent(INode::CONSTRUCTED)) {
      __interface_list[i]->reachConstructed();
      sendNodeEvent(__interface_list[i], *__node_list[i]->getNodeEvent(INode::CONSTRUCTED));
    }
  }
  /* End< */

  /* Begin> Send all events for state INITIALIZED to the nodes */
  for (unsigned int i = 0; i < __node_list.size(); ++i)
    if (__node_list[i]->getNodeEvent(INode::INITIALIZED)) {
      __interface_list[i]->reachInitialized();
      sendNodeEvent(__interface_list[i], *__node_list[i]->getNodeEvent(INode::INITIALIZED));
    }
  /* End< */

  GraphDescription gd(*this);

  // request all connection formats from master and set them in graph description gd
  const SequenceValue<INode >& master_inodes = master_gd.getINodeList();
  for(unsigned int x = 0; x < master_gd.getNumEdges(); ++x) {
    EdgeDescription& gd_edge = gd.getEdge(x);
    if(gd_edge.getFormat() == 0) { // no format set in gd -> take from master
      const EdgeDescription& master_edge = master_gd.getEdge(x);
      if(master_edge.isOutConnected()) { // and edge really connected -> get output format of 'left' node
	INode* left  = master_inodes[master_edge.getOutputNodeID()];
	Format_var c_format(left->getOutputStreamFormat(master_edge.getOutputJackTag()));
	gd_edge.setFormat(c_format.get());
	GLOBAL_DEBUG_STREAM2("gd: " << "Update format x " << *(c_format.get()) << endl);
      }
    }
  }

  list<int> graph_sources;

  // get all sources of query graph
  const SequenceValue< NodeDescription >& nds = gd.getNodeDescriptionList();
  for(unsigned int n = 0; n < nds.size(); ++n) {
    // node with no incoming edge and some outgoing edges -> source that needs to be connected
    if(gd.getInEdgeList(n).empty() && !gd.getOutEdgeList(n).empty()) {
      graph_sources.push_back(n);
    }
    // node with only unconnected incoming edges and some outgoing edges -> node that needs to be connected
    if(!gd.getOutEdgeList(n).empty()) {
      list<int>::const_iterator in_edge_iter;
      for(in_edge_iter = gd.getInEdgeList(n).begin(); in_edge_iter != gd.getInEdgeList(n).end(); ++in_edge_iter) {
	EdgeDescription& in_edge = gd.getEdge(*in_edge_iter);
	if(in_edge.isOutConnected()) { // -> outgoind part is connected -> connected edge -> no 'source' node
	  break;
	}
      }
      if(in_edge_iter == gd.getInEdgeList(n).end()) {
	graph_sources.push_back(n);
      }
    }
  }

  // all shared edges that will first be connected in the very end
  list<int> edges_to_connect_later;

  // for all sources in graph
  for(list<int>::iterator i = graph_sources.begin(); i != graph_sources.end(); ++i) {
    INode* left_node = gd.getINode((nds[*i]));
    if (!left_node) {
      throw BindingException("Node is not available in the GraphDescription");
    }

    // used to store all unconnected out jacks and add them to this graph descrition in the end
    // this is needed to connect to unconnected out jacks in shared graph descriptions
    list<string> left_node_unconnected_out_jacks;

    // stores edges that have been connected and that can be removed from copy of this graph description (namely gd)
    list<int> edges_to_remove;

    // try to connect all outgoing edges
    const list<int>& out_list = gd.getOutEdgeList(*i);
    for(list<int>::const_iterator iter = out_list.begin(); iter != out_list.end(); ++iter) {
      EdgeDescription& edge = (gd.getEdge(*iter));

      // check if edge is connected at all
      if(edge.isInConnected() && edge.isOutConnected()) {

	if(edge.getOverlapType() == EdgeDescription::COMPLETE) { // nothing to do, simply remove edge
	  edges_to_remove.push_back(*iter);
	}
	else if(edge.getOverlapType() == EdgeDescription::PARTIAL_COPY) {
	  // only remember edge for later connection setup
	  edges_to_connect_later.push_back(*iter);
	  edges_to_remove.push_back(*iter);
	} // end: if(PARTIAL_COPY)
	else if(edge.getOverlapType() == EdgeDescription::NONE) { // no overlap -> unchecked_connect
	  INode* right_node = gd.getINode(nds[edge.getInputNodeID()]);
	  if (!right_node) {
	    throw BindingException("Node is not available in the GraphDescription");
	  }
	  left_node->reachInitialized(); 
	  right_node->reachInitialized();
          
	  Format* f = edge.getFormat();
          
          /* If no connection format is set for this edge, we create it */
          if(!getEdge(edge.getEdgeID()).getFormat()) {
            Format dummy_format;
            getEdge(edge.getEdgeID()).setFormat(&dummy_format);
          }
          
          if(edge.getBindingDescription().isImplicit()) {
            try {
	      InstreamChannel* channel = dynamic_cast<InstreamChannel*>(unchecked_c_connect(left_node, edge.getOutputJackTag(), right_node, edge.getInputJackTag(), f));
              getEdge( edge.getEdgeID() ).setFormat(f);
              /* Store the returned instream channel */
              getEdge(*iter).setInstreamChannel( channel );
              realizeCommunicationChannel(channel, getEdge(*iter));

	      cout << "unchecked_connect( " << (*i) << ", " << edge.getInputNodeID() << " )" << endl;
	    }
	    catch(const BindingException& e) { // jack already connected? => try to connect within jackgroup
	      InstreamChannel* channel = dynamic_cast<InstreamChannel*>(c_s_connect(left_node, edge.getOutputJackTag(), right_node, edge.getInputJackTag(), getEdge( edge.getEdgeID() ).getFormat()));
              /* Store the returned instream channel */
              getEdge(*iter).setInstreamChannel( channel );
              realizeCommunicationChannel(channel, getEdge(*iter));

	      cout << "could not unchecked_connect( " << (*i) << ", " << edge.getInputNodeID() << " ), "
		   << "used s_connect( " << (*i) << ", " << edge.getInputNodeID() << " ) instead." << endl;
	    }
	  } else {
	    try {
	      InstreamChannel_var c_channel(unchecked_c_connect(left_node, edge.getOutputJackTag(), 
                                                                right_node, edge.getInputJackTag(), f));
              getEdge( edge.getEdgeID() ).setFormat(f);
              /* Store the returned instream channel */
              getEdge(*iter).setInstreamChannel( dynamic_cast<InstreamChannel*>(c_channel.get()->copy() ));
	      realizeCommunicationChannel(c_channel.get(), edge);
	      cout << "unchecked_c_connect( " << (*i) << ", " << edge.getInputNodeID() << " ) using explicit binding" << endl;
	    }
	    catch(const BindingException& e) { // jack already connected? => try to connect within jackgroup
	      InstreamChannel_var c_channel(c_s_connect(left_node, edge.getOutputJackTag(), right_node, edge.getInputJackTag(), getEdge( edge.getEdgeID() ).getFormat()));
              /* Store the returned instream channel */
              getEdge(*iter).setInstreamChannel( dynamic_cast<InstreamChannel*>(c_channel.get()->copy() ));
	      realizeCommunicationChannel(c_channel.get(), edge);
	      cout << "could not c_connect( " << (*i) << ", " << edge.getInputNodeID() << " ), "
		   << "used c_s_connect( " << (*i) << ", " << edge.getInputNodeID() << " ) instead "
		   << "using explicit binding" << endl;
	    }
	  }
          
	  edges_to_remove.push_back(*iter);
	  left_node_unconnected_out_jacks.remove(edge.getOutputJackTag());
	} // end: if(NONE): no overlap -> normal connection
	else if(edge.getOverlapType() == EdgeDescription::PARTIAL_OUTPUT) {
	  GLOBAL_ERROR_STREAM2("gd: "<< "PARTIAL_OUTPUT overlap not possible for realizeHandoverGraph()" << endl);
	  throw BindingException("PARTIAL_OUTPUT overlap not possible for realizeHandoverGraph()");
	} // end: if(PARTIAL_OUTPUT)
	else {
	  GLOBAL_ERROR_STREAM2("gd: "<< "Unhandled case in realize graph!" << endl);
	  throw BindingException("Unhandled case in realize graph!");
	}
      } // end: check if edge is connected at all
    } // end: try to connect all outgoing edges

    // remove edges from list of edge indices for nodes
    for(list<int>::iterator remover = edges_to_remove.begin(); remover != edges_to_remove.end(); ++remover) {
      // check, if node with in_id becomes new source node
      int n = gd.getEdge(*remover).getInputNodeID();
      // node with no incoming edge (after remove) and some outgoing edges -> source (or unconnected node) that needs to be connected
      if(gd.getInEdgeList(n).size() == 1 && !gd.getOutEdgeList(n).empty()) {
	graph_sources.push_back(n);
      }

      gd.removeEdge(*remover);
    }

    // add all unconnceted out edges to this graph description
    SequenceValue< NodeDescription >& my_nds = this->getNodeDescriptionList();
    for(list<string>::iterator adder = left_node_unconnected_out_jacks.begin();
	adder != left_node_unconnected_out_jacks.end();
	++adder) {
      this->addOutEdge(*(my_nds[*i]), *adder);
      GLOBAL_DEBUG_STREAM2("gd: " << "Adding unconnected out edge " << *adder << " to node " << my_nds[*i]->getNodeName() << endl);
    }
  } // end: for all sources in graph

  // connect all shared edges in the end
  for(list<int>::iterator connector = edges_to_connect_later.begin(); connector != edges_to_connect_later.end(); ++connector) {
    EdgeDescription& edge = (gd.getEdge(*connector));
    // check if edge is connected at all
    if(edge.isInConnected() && edge.isOutConnected()) {
      INode* left_node = gd.getINode((nds[edge.getOutputNodeID()]));
      INode* right_node = gd.getINode(nds[edge.getInputNodeID()]);
      if (!right_node) {
	throw BindingException("Node is not available in the GraphDescription");
      }
      right_node->reachInitialized();
      left_node->reachActivated(); // only reach ACTIVATED

      /* If no connection format is set for this edge, we create it */
      if(!getEdge(edge.getEdgeID()).getFormat()) {
        Format dummy_format;
        getEdge(edge.getEdgeID()).setFormat(&dummy_format);
      }
      
      if(edge.getBindingDescription().isImplicit()) {
	/* If no specific transport strategy should be used, use the default s_connect method */
	InstreamChannel* channel = dynamic_cast<InstreamChannel*>(c_s_connect(left_node, edge.getOutputJackTag(), right_node, edge.getInputJackTag(), getEdge( edge.getEdgeID() ).getFormat()));
        /* Store the returned instream channel */
        getEdge(edge.getEdgeID()).setInstreamChannel( channel );
        realizeCommunicationChannel( channel, getEdge(edge.getEdgeID()) );
	cout << "s_connect( " << edge.getOutputNodeID() << ", " << edge.getInputNodeID() << " )" << endl;
      } else {
        /* Otherwise use the c_s_connect method */
 	InstreamChannel_var c_channel(c_s_connect(left_node, edge.getOutputJackTag(), right_node, edge.getInputJackTag(), getEdge( edge.getEdgeID() ).getFormat()));
 	realizeCommunicationChannel(c_channel.get(), edge);
 	cout << "c_s_connect( " << edge.getOutputNodeID() << ", " << edge.getInputNodeID() << " ) using explicit binding" << endl;
        /* Store the returned instream channel */
        getEdge(edge.getEdgeID()).setInstreamChannel( dynamic_cast<InstreamChannel*>(c_channel.get()->copy() ) );
      }
      gd.removeEdge(*connector);
    }
  }
  
  /* Finally update the graphdescription stored in the session */
  if (__i_session.getValue()) {
    __i_session.getValue()->setGraphDescription(*this);
  }
}
// END realizeHandoverGraph ******************************************************

void GraphDescription::activateGraph() {
 for (unsigned int i = 0; i < __interface_list.size(); ++i)
   if (__interface_list[i]) {
     __interface_list[i]->reachActivated();
     /* Begin> Check if events in state ACTIVATED must be send to the node */
     if (__node_list[i]->getNodeEvent(INode::ACTIVATED)) {
       sendNodeEvent(__interface_list[i], *__node_list[i]->getNodeEvent(INode::ACTIVATED));
     }
     /* End< */
   }
}

void GraphDescription::startGraph()
{
  list< pair<INode*, unsigned int> > prefetching_nodes;

  const SequenceValue< NodeDescription >& nds = getNodeDescriptionList();

  // first, start all nodes that have no prefetching enabled 
  for(unsigned int i = 0; i < nds.size(); ++i) {
    INode* n = getINode(nds[i]);
    if(n) {
      if(n->getPrefetchQueueSize() == 0) { // start nodes that do not prefetch 
	n->reachStarted();
        /* Begin> Check if events in state STARTED must be send to the node */
        if (__node_list[i]->getNodeEvent(INode::STARTED)) {
          sendNodeEvent(__interface_list[i], *__node_list[i]->getNodeEvent(INode::STARTED));
        }
        /* End< */
      }
      else { // add to list of nodes that prefetch 
	prefetching_nodes.push_back(make_pair<INode*, unsigned int>(n,i));
      }
    }
  }
  
  // then, wait for prefetching nodes to finish prefetching
//  for(list< pair<INode*, unsigned int> >::iterator j = prefetching_nodes.begin(); j != prefetching_nodes.end(); ++j) {
//    (*j).first->waitPrefetchFinished();
//  }
  
  // finally, if all these nodes have finished prefetching, start them 
  for(list< pair<INode*, unsigned int> >::iterator k = prefetching_nodes.begin(); k != prefetching_nodes.end(); ++k) {
    INode* n = (*k).first;
    unsigned int i = (*k).second;
    if(n) {
      n->reachStarted();
      // Begin> Check if events in state STARTED must be send to the node 
      if (__node_list[i]->getNodeEvent(INode::STARTED)) {
        sendNodeEvent(__interface_list[i], *__node_list[i]->getNodeEvent(INode::STARTED));
      }
      // End< 
    }
  }
  
}

void GraphDescription::stopGraph() {
  const SequenceValue< NodeDescription >& nds = getNodeDescriptionList();
  for(unsigned int i = 0; i < nds.size(); ++i) {
    if (getINode(nds[i]) ) {
      getINode(nds[i])->reachActivated();
    }
  }
}

void GraphDescription::unrealizeGraph() {
  const SequenceValue< NodeDescription >& nds = getNodeDescriptionList();
  for(unsigned int i = 0; i < nds.size(); ++i) {
    if (getINode(nds[i]) ) 
      getINode(nds[i])->reachConstructed();
  }
}

void GraphDescription::sendNodeEvent(INode* node, CEvent& cevent) {
  for(unsigned int j=0; j<cevent.size(); ++j) {
    node->sendEvent(cevent[j]);
  }
}

bool GraphDescription::isSharingType(NodeDescription::SharingType s) {
  /* Begin> Check if all NodeDescriptions have sharing type s*/
  for (unsigned int i = 0; i < __node_list.size(); ++i) {
    if ( __node_list[i]->getSharingType() != s ) {
      /* If not return false */
      return false;
    }
  }
  /* End< */

  /* Otherwise return true */
  return true;
}

bool GraphDescription::getRequestFlag(const NodeDescription& nd) const {
  return __request_flags.getRValue()[nd.getInternalID()];
}

bool GraphDescription::getReleaseFlag(const NodeDescription& nd) const {
  return __release_flags.getRValue()[nd.getInternalID()];
}

void GraphDescription::setRequestFlag(const NodeDescription& nd, bool flag) {
  __request_flags.getRValue()[nd.getInternalID()] = flag;
}

void GraphDescription::setReleaseFlag(const NodeDescription& nd, bool flag) {
  __release_flags.getRValue()[nd.getInternalID()] = flag;
}

void GraphDescription::configureOutOfBandBindings() {
  /* Begin> Configure the TransportStrategies between ProxyObjects and NMMObjects */
  for (unsigned int i = 0; i < __node_list.size(); ++i) {
    if ( !__node_list[i]->getTransportStrategy().empty() ) {
      INode* inode = getINode( *__node_list[i] );
      if (! inode ) {
        GLOBAL_ERROR_STREAM2("gd: No INode interface for node " << __node_list[i]->getNodeName() << " in GraphDescription available " << endl ) ;
        continue;
      }
      
      inode->getParentObject()->getOutOfBandChannel().addTransportStrategy(__node_list[i]->getTransportStrategy(),
                                                                           OutOfBandChannel::REQUEST | OutOfBandChannel::REPLY);
      inode->getParentObject()->getOutOfBandChannel().establishBinding();
      
      GLOBAL_DEBUG_STREAM2("gd: " << "use TransportStragey " <<        __node_list[i]->getTransportStrategy() << " for node " << __node_list[i]->getNodeName() << endl);
    }
  }
  /* End< */
}

void GraphDescription::connectSinkSynchronizer() {
  if ( __i_session.getValue() ) {
    ISinkSynchronizer_var tmp_sync(__i_session->getSinkSynchronizer());
    IMultiAudioSinkSynchronizer_var ia_sync(tmp_sync->getParentObject()->getInterface<IMultiAudioSinkSynchronizer>());
    IMultiVideoSinkSynchronizer_var iv_sync(tmp_sync->getParentObject()->getInterface<IMultiVideoSinkSynchronizer>());

    /* Begin> Connect all sinks with the synchronizer */
    for (unsigned int i = 0; i< getINodeList().size(); ++i) {
      string node_name = getNodeDescriptionList()[i]->getNodeName();

      /* If the out edge list is empty we assume a sink node. IF NODE_DESCRIPTIONS ARE UPDATED USE THIS INFORMATION TO
       * DECIDE WHETHER A NODE IS A SINK OR NOT.
       */
      if (! getOutEdgeList( getNodeDescriptionList()[i]->getInternalID() ).empty() ){
        continue;
      } 

      /* Begin> Add the controller to the synchronizer */
      ISynchronizedSink_var av(getINodeList()[i]->getParentObject()->getInterface<ISynchronizedSink>());
      if (av.get()) {
        ISinkController_var controller (av->getController());
        av->setSynchronized(true);
        if(av->isBuffered() && ia_sync.get()) { // add audio controller ...
          ia_sync->addAudioSinkController(controller.get());
        }
        else if(iv_sync.get()) { // ... or video (unbuffered)
          iv_sync->addVideoSinkController(controller.get());
        }
      }
      /* End< */
    }
  } else {
    throw Exception("No session available for requesting synchronizer");
  }
}

void GraphDescription::setSinkSynchronizationEnabled(bool enable) {
  /* Begin> Enable or disable synchronization for all sinks */
  for (unsigned int i = 0; i< getINodeList().size(); ++i) {
    string node_name = getNodeDescriptionList()[i]->getNodeName();
    
    /* If the out edge list is empty we assume a sink node. IF NODE_DESCRIPTIONS ARE UPDATED USE THIS INFORMATION TO
     * DECIDE WHETHER A NODE IS A SINK OR NOT.
     */
    if (! getOutEdgeList( getNodeDescriptionList()[i]->getInternalID() ).empty() ){
      continue;
    } 
    
    /* Begin> Enable or disable synchronization */
    ISynchronizedSink_var av(getINodeList()[i]->getParentObject()->getInterface<ISynchronizedSink>());
    if (av.get()) {
      ISinkController_var controller (av->getController());
      av->setSynchronized(enable);
    }
    /* End< */
  }
}

void GraphDescription::setPrefetchingEnabled(bool enable) 
{
  /* Begin> enable prefetching for all sinks */
  for (unsigned int i = 0; i< getINodeList().size(); ++i) {
    string node_name = getNodeDescriptionList()[i]->getNodeName();

    /* If the out edge list is empty we assume a sink node. IF NODE_DESCRIPTIONS ARE UPDATED USE THIS INFORMATION TO
     * DECIDE WHETHER A NODE IS A SINK OR NOT.
     */
    if (! getOutEdgeList( getNodeDescriptionList()[i]->getInternalID() ).empty() ){
      continue;
    } 

    if(enable) {
      if(getINodeList()[i]->getPrefetchQueueSize() == 0) {
	getINodeList()[i]->setPrefetchQueueSize(4);
      }
    }
    else {
      getINodeList()[i]->setPrefetchQueueSize(0);
    }
  }
}

void GraphDescription::disconnectSinkSynchronizer() {
  /* Begin> remove the synchronizer */
  for (unsigned int i = 0; i< getINodeList().size(); ++i) {
    /* If the out edge list is empty we assume a sink node. IF NODE_DESCRIPTIONS ARE UPDATED USE THIS INFORMATION TO
     * DECIDE WHETHER A NODE IS A SINK OR NOT.
     */
    if (! getOutEdgeList( getNodeDescriptionList()[i]->getInternalID() ).empty() ){
      continue;
    }

    if ( __i_session.getValue() ) {
      ISinkSynchronizer_var tmp_sync(__i_session->getSinkSynchronizer());
      ISinkSynchronizer_var isync(tmp_sync->getParentObject()->getCheckedInterface<ISinkSynchronizer>());

      if (isync.get()) {
        ISynchronizedSink_var av(getINodeList()[i]->getParentObject()->getInterface<ISynchronizedSink>());
        if (av.get()) {
          ISinkController_var controller (av->getController());
          isync->removeSinkController(controller.get());
        }
      }
    }
  }
  /*End< */
}

const list<INode*> GraphDescription::getPartialOverlappedNodes() 
{
  list<INode*> l;
  for(unsigned int j = 0; j < __internal_counter; ++j) {
    const list<int>& out_l = __out_edges.getRValue()[j];
    for(list<int>::const_iterator iter = out_l.begin(); iter != out_l.end(); ++iter) {
      EdgeDescription& e = __edges.getRValue()[*iter];
      // found some node with parial overlapped outgoing edge? -> add to list and go to next node until __internal_counter
      if( (e.getOverlapType() == EdgeDescription::PARTIAL_COPY) || (e.getOverlapType() == EdgeDescription::PARTIAL_OUTPUT) ) {
	INode* n = __interface_list[e.getOutputNodeID()];
	l.push_back(n);
	break;
      }
    }
  }
  return l;
}

const GraphDescription& GraphDescription::operator=(const GraphDescription& gd) {
  /* delete our internal attributes */
  clear();

  /* call operator of base class */
  SerializedValue::operator=(gd);

  /*Begin> assign all values */
  __edges             = gd.__edges;
  __out_edges         = gd.__out_edges;
  __in_edges          = gd.__in_edges;

  __node_list         = gd.__node_list;
  __interface_list    = gd.__interface_list;
  __request_flags     = gd.__request_flags;
  __release_flags     = gd.__release_flags;

  __internal_counter  = gd.__internal_counter;

  __sink_synch_descr  = gd.__sink_synch_descr;

  if (gd.getISession())
    setISession(*gd.getISession());

  /* End< */
  return *this;
}

const char* GraphDescription::GraphDescription_type_name = "GraphDescription";
const char* GraphDescription::GraphDescription_type_ns   = "NMM";

const char* EdgeDescription::__EdgeDescription_type_name = "EdgeDescription";
const char* EdgeDescription::__EdgeDescription_type_ns   = "NMM";

TConstructor<GraphDescription>    __nmm__GraphDescription_type(GraphDescription::GraphDescription_type_name, GraphDescription::GraphDescription_type_ns );


OValueStream& operator<<(OValueStream& ov, const EdgeDescription::OverlapStatus& eo)
{
  ov.beginType(TypeInfo("EdgeDescription::OverlapStatus", "NMM"));
  nmm_int32 temp = static_cast<nmm_int32>(eo);
  ov << temp;
  ov.endType();
  return ov;
}

IValueStream& operator>>(IValueStream& iv, EdgeDescription::OverlapStatus& eo)
{
  if(iv.getCurrentTypeInfo() != TypeInfo("EdgeDescription::OverlapStatus", "NMM")) {
    throw InvalidTypeException(TypeInfo("EdgeDescription::OverlapStatus", "NMM"), iv.getCurrentTypeInfo());
  }
  iv.goDown();
  nmm_int32 temp;
  iv >> temp;
  eo = static_cast<EdgeDescription::OverlapStatus>(temp);
  iv.goUp();
  return iv;
}



NMM::OValueStream& operator<<(NMM::OValueStream& ov, const vector<NMM::EdgeDescription> & string_map) 
{
  return writeVector<NMM::EdgeDescription>(ov, string_map, "EdgeDescription");
}

NMM::IValueStream& operator>>(NMM::IValueStream& ov, vector<NMM::EdgeDescription> & string_map) 
{
  return readVector<NMM::EdgeDescription>(ov, string_map, "EdgeDescription");
}



} //end namespace

#include "nmm/nmm_suffix.hpp"
