/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "IPluginManager.hpp"
#include "nmm/comm/serialize/SerializeResult.hpp"
#include "nmm/comm/serialize/InvalidTypeException.hpp"
#include "nmm/comm/serialize/TConstructor.hpp"

#include "nmm/nmm_prefix.hpp"


namespace NMM
{
  void IPluginManager::used() {} 
  
  
  void IPluginManager::addPluginPath(const string& path)
      throw(const NMM::SerializedException&)
  {
    NMM::TInValue<string > path_value(path, false);
    NMM::Event event("IPluginManager::addPluginPath", &path_value);
    if(const_cast<IPluginManager*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
  }
  
  void IPluginManager::loadLibrary(const string& library)
      throw(const NMM::SerializedException&)
  {
    NMM::TInValue<string > library_value(library, false);
    NMM::Event event("IPluginManager::loadLibrary", &library_value);
    if(const_cast<IPluginManager*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
  }
  
  void IPluginManager::loadPlugin(const string& plugin_name) 
      throw(const NMM::SerializedException&, const SerializedException&)
  {
    NMM::TInValue<string > plugin_name_value(plugin_name, false);
    NMM::Event event("IPluginManager::loadPlugin", &plugin_name_value);
    if(const_cast<IPluginManager*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        if(SerializedException* ex = dynamic_cast<SerializedException*>(event.getException())) {
          throw SerializedException(*ex);
        }
        throw NMM::SerializedException(*event.getException());
      }
    }
  }
  
  void IPluginManager::loadExternalPlugins(const string& directory) 
      throw(const NMM::SerializedException&, const SerializedException&)
  {
    NMM::TInValue<string > directory_value(directory, false);
    NMM::Event event("IPluginManager::loadExternalPlugins", &directory_value);
    if(const_cast<IPluginManager*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        if(SerializedException* ex = dynamic_cast<SerializedException*>(event.getException())) {
          throw SerializedException(*ex);
        }
        throw NMM::SerializedException(*event.getException());
      }
    }
  }
  
  void IPluginManager::removeExternalPlugins(const string& directory) 
      throw(const NMM::SerializedException&, const SerializedException&)
  {
    NMM::TInValue<string > directory_value(directory, false);
    NMM::Event event("IPluginManager::removeExternalPlugins", &directory_value);
    if(const_cast<IPluginManager*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        if(SerializedException* ex = dynamic_cast<SerializedException*>(event.getException())) {
          throw SerializedException(*ex);
        }
        throw NMM::SerializedException(*event.getException());
      }
    }
  }
  
  void IPluginManager::activateRegistryPart(const string& library) 
      throw(const NMM::SerializedException&, const SerializedException&)
  {
    NMM::TInValue<string > library_value(library, false);
    NMM::Event event("IPluginManager::activateRegistryPart", &library_value);
    if(const_cast<IPluginManager*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        if(SerializedException* ex = dynamic_cast<SerializedException*>(event.getException())) {
          throw SerializedException(*ex);
        }
        throw NMM::SerializedException(*event.getException());
      }
    }
  }
  
  void IPluginManager::activateFactoryPart(const string& library) 
      throw(const NMM::SerializedException&, const SerializedException&)
  {
    NMM::TInValue<string > library_value(library, false);
    NMM::Event event("IPluginManager::activateFactoryPart", &library_value);
    if(const_cast<IPluginManager*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        if(SerializedException* ex = dynamic_cast<SerializedException*>(event.getException())) {
          throw SerializedException(*ex);
        }
        throw NMM::SerializedException(*event.getException());
      }
    }
  }
  
  void IPluginManager::deactivateRegistryPart(const string& library) 
      throw(const NMM::SerializedException&, const SerializedException&)
  {
    NMM::TInValue<string > library_value(library, false);
    NMM::Event event("IPluginManager::deactivateRegistryPart", &library_value);
    if(const_cast<IPluginManager*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        if(SerializedException* ex = dynamic_cast<SerializedException*>(event.getException())) {
          throw SerializedException(*ex);
        }
        throw NMM::SerializedException(*event.getException());
      }
    }
  }
  
  void IPluginManager::deactivateFactoryPart(const string& library) 
      throw(const NMM::SerializedException&, const SerializedException&)
  {
    NMM::TInValue<string > library_value(library, false);
    NMM::Event event("IPluginManager::deactivateFactoryPart", &library_value);
    if(const_cast<IPluginManager*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        if(SerializedException* ex = dynamic_cast<SerializedException*>(event.getException())) {
          throw SerializedException(*ex);
        }
        throw NMM::SerializedException(*event.getException());
      }
    }
  }
  
  void IPluginManager::removeLibrary(const string& library) 
      throw(const NMM::SerializedException&, const SerializedException&)
  {
    NMM::TInValue<string > library_value(library, false);
    NMM::Event event("IPluginManager::removeLibrary", &library_value);
    if(const_cast<IPluginManager*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        if(SerializedException* ex = dynamic_cast<SerializedException*>(event.getException())) {
          throw SerializedException(*ex);
        }
        throw NMM::SerializedException(*event.getException());
      }
    }
  }
  
  list<ObjectID >  IPluginManager::getPluginID(const string& library, const nmm_uint32& index)
      throw(const NMM::SerializedException&)
  {
    NMM::TOutValue<list<ObjectID >  > return_value;
    NMM::TInValue<string > library_value(library, false);
    NMM::TInValue<nmm_uint32 > index_value(index, false);
    NMM::Event event("IPluginManager::getPluginID", &library_value, &index_value);
    event.setReturnValue(&return_value);
    if(const_cast<IPluginManager*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
    return return_value.getRValue();
  }
  
  list<string >  IPluginManager::getPlugins() const
      throw(const NMM::SerializedException&)
  {
    NMM::TOutValue<list<string >  > return_value;
    NMM::Event event("IPluginManager::getPlugins");
    event.setReturnValue(&return_value);
    if(const_cast<IPluginManager*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
    return return_value.getRValue();
  }
  
  list<string >  IPluginManager::getUsedPlugins() const
      throw(const NMM::SerializedException&)
  {
    NMM::TOutValue<list<string >  > return_value;
    NMM::Event event("IPluginManager::getUsedPlugins");
    event.setReturnValue(&return_value);
    if(const_cast<IPluginManager*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
    return return_value.getRValue();
  }
  
  list<string >  IPluginManager::getPluginPath() const
      throw(const NMM::SerializedException&)
  {
    NMM::TOutValue<list<string >  > return_value;
    NMM::Event event("IPluginManager::getPluginPath");
    event.setReturnValue(&return_value);
    if(const_cast<IPluginManager*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
    return return_value.getRValue();
  }

  const char* IPluginManager::IPluginManager_type = "IPluginManager";
  NMM::TConstructor<IPluginManager > __IPluginManager_prototype("IPluginManager","NMM");


  const char* IPluginManager::addPluginPath_event = "IPluginManager::addPluginPath";
  const char* IPluginManager::loadLibrary_event = "IPluginManager::loadLibrary";
  const char* IPluginManager::loadPlugin_event = "IPluginManager::loadPlugin";
  const char* IPluginManager::loadExternalPlugins_event = "IPluginManager::loadExternalPlugins";
  const char* IPluginManager::removeExternalPlugins_event = "IPluginManager::removeExternalPlugins";
  const char* IPluginManager::activateRegistryPart_event = "IPluginManager::activateRegistryPart";
  const char* IPluginManager::activateFactoryPart_event = "IPluginManager::activateFactoryPart";
  const char* IPluginManager::deactivateRegistryPart_event = "IPluginManager::deactivateRegistryPart";
  const char* IPluginManager::deactivateFactoryPart_event = "IPluginManager::deactivateFactoryPart";
  const char* IPluginManager::removeLibrary_event = "IPluginManager::removeLibrary";
  const char* IPluginManager::getPluginID_event = "IPluginManager::getPluginID";
  const char* IPluginManager::getPlugins_event = "IPluginManager::getPlugins";
  const char* IPluginManager::getUsedPlugins_event = "IPluginManager::getUsedPlugins";
  const char* IPluginManager::getPluginPath_event = "IPluginManager::getPluginPath";


}


#include "nmm/nmm_suffix.hpp"

