/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "IRegistry.hpp"
#include "nmm/comm/serialize/SerializeResult.hpp"
#include "nmm/comm/serialize/InvalidTypeException.hpp"
#include "nmm/comm/serialize/TConstructor.hpp"

#include "nmm/nmm_prefix.hpp"


namespace NMM
{
  void IRegistry::used() {} 
  
  
  INode* IRegistry::requestNode(const RegistryID& reg, const ObjectID& node_id, const UserInfo& info, NodeDescription& nd) 
      throw(const NMM::SerializedException&, const RegistryException&)
  {
    NMM::TOutValue<INode* > return_value;
    NMM::TInValue<RegistryID > reg_value(reg, false);
    NMM::TInValue<ObjectID > node_id_value(node_id, false);
    NMM::TInValue<UserInfo > info_value(info, false);
    NMM::TInOutValue<NodeDescription > nd_value;
    nd_value.setValue(nd);
    NMM::Event event("IRegistry::requestNode", &reg_value, &node_id_value, &info_value, &nd_value);
    event.setReturnValue(&return_value);
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        if(RegistryException* ex = dynamic_cast<RegistryException*>(event.getException())) {
          throw RegistryException(*ex);
        }
        throw NMM::SerializedException(*event.getException());
      }
    }
    nd = nd_value.getRValue();
    return return_value.getRValue();
  }
  
  Interface* IRegistry::requestObject(const RegistryID& reg, const ObjectID& node_id, const UserInfo& info) 
      throw(const NMM::SerializedException&, const RegistryException&)
  {
    NMM::TOutValue<Interface* > return_value;
    NMM::TInValue<RegistryID > reg_value(reg, false);
    NMM::TInValue<ObjectID > node_id_value(node_id, false);
    NMM::TInValue<UserInfo > info_value(info, false);
    NMM::Event event("IRegistry::requestObject", &reg_value, &node_id_value, &info_value);
    event.setReturnValue(&return_value);
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        if(RegistryException* ex = dynamic_cast<RegistryException*>(event.getException())) {
          throw RegistryException(*ex);
        }
        throw NMM::SerializedException(*event.getException());
      }
    }
    return return_value.getRValue();
  }
  
  void IRegistry::releaseNode(const RegistryID& reg_id, const NMMReference& ref, const ObjectID& node_id, const UserInfo& user_info)
      throw(const NMM::SerializedException&)
  {
    NMM::TInValue<RegistryID > reg_id_value(reg_id, false);
    NMM::TInValue<NMMReference > ref_value(ref, false);
    NMM::TInValue<ObjectID > node_id_value(node_id, false);
    NMM::TInValue<UserInfo > user_info_value(user_info, false);
    NMM::Event event("IRegistry::releaseNode", &reg_id_value, &ref_value, &node_id_value, &user_info_value);
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
  }
  
  void IRegistry::releaseObject(const RegistryID& reg_id, const NMMReference& ref, const ObjectID& node_id, const UserInfo& user_info)
      throw(const NMM::SerializedException&)
  {
    NMM::TInValue<RegistryID > reg_id_value(reg_id, false);
    NMM::TInValue<NMMReference > ref_value(ref, false);
    NMM::TInValue<ObjectID > node_id_value(node_id, false);
    NMM::TInValue<UserInfo > user_info_value(user_info, false);
    NMM::Event event("IRegistry::releaseObject", &reg_id_value, &ref_value, &node_id_value, &user_info_value);
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
  }
  
  list<Response >  IRegistry::initRequest(const NodeDescription& node_desc) const
      throw(const NMM::SerializedException&)
  {
    NMM::TOutValue<list<Response >  > return_value;
    NMM::TInValue<NodeDescription > node_desc_value(node_desc, false);
    NMM::Event event("IRegistry::initRequest", &node_desc_value);
    event.setReturnValue(&return_value);
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
    return return_value.getRValue();
  }
  
  list<Response >  IRegistry::initObjectRequest(const ObjectDescription& node_desc) const
      throw(const NMM::SerializedException&)
  {
    NMM::TOutValue<list<Response >  > return_value;
    NMM::TInValue<ObjectDescription > node_desc_value(node_desc, false);
    NMM::Event event("IRegistry::initObjectRequest", &node_desc_value);
    event.setReturnValue(&return_value);
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
    return return_value.getRValue();
  }
  
  list<ObjectID >  IRegistry::searchNode(const RegistryID& registry_id, const NodeDescription& node_desc) const
      throw(const NMM::SerializedException&)
  {
    NMM::TOutValue<list<ObjectID >  > return_value;
    NMM::TInValue<RegistryID > registry_id_value(registry_id, false);
    NMM::TInValue<NodeDescription > node_desc_value(node_desc, false);
    NMM::Event event("IRegistry::searchNode", &registry_id_value, &node_desc_value);
    event.setReturnValue(&return_value);
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
    return return_value.getRValue();
  }
  
  FactoryTicket IRegistry::reserveNode(const RegistryID& reg_id, const ObjectID& node_id, const UserInfo& info) 
      throw(const NMM::SerializedException&, const RegistryException&)
  {
    NMM::TOutValue<FactoryTicket > return_value;
    NMM::TInValue<RegistryID > reg_id_value(reg_id, false);
    NMM::TInValue<ObjectID > node_id_value(node_id, false);
    NMM::TInValue<UserInfo > info_value(info, false);
    NMM::Event event("IRegistry::reserveNode", &reg_id_value, &node_id_value, &info_value);
    event.setReturnValue(&return_value);
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        if(RegistryException* ex = dynamic_cast<RegistryException*>(event.getException())) {
          throw RegistryException(*ex);
        }
        throw NMM::SerializedException(*event.getException());
      }
    }
    return return_value.getRValue();
  }
  
  INode* IRegistry::instantiateNode(const RegistryID& reg_id, const FactoryTicket& ticket, const UserInfo& info) 
      throw(const NMM::SerializedException&, const RegistryException&)
  {
    NMM::TOutValue<INode* > return_value;
    NMM::TInValue<RegistryID > reg_id_value(reg_id, false);
    NMM::TInValue<FactoryTicket > ticket_value(ticket, false);
    NMM::TInValue<UserInfo > info_value(info, false);
    NMM::Event event("IRegistry::instantiateNode", &reg_id_value, &ticket_value, &info_value);
    event.setReturnValue(&return_value);
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        if(RegistryException* ex = dynamic_cast<RegistryException*>(event.getException())) {
          throw RegistryException(*ex);
        }
        throw NMM::SerializedException(*event.getException());
      }
    }
    return return_value.getRValue();
  }
  
  void IRegistry::releaseReservation(const RegistryID& reg_id, const ObjectID& node_id, const UserInfo& user_info)
      throw(const NMM::SerializedException&)
  {
    NMM::TInValue<RegistryID > reg_id_value(reg_id, false);
    NMM::TInValue<ObjectID > node_id_value(node_id, false);
    NMM::TInValue<UserInfo > user_info_value(user_info, false);
    NMM::Event event("IRegistry::releaseReservation", &reg_id_value, &node_id_value, &user_info_value);
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
  }
  
  void IRegistry::destroyNode(const RegistryID& reg_id, const NMMReference& reference, const UserInfo& info)
      throw(const NMM::SerializedException&)
  {
    NMM::TInValue<RegistryID > reg_id_value(reg_id, false);
    NMM::TInValue<NMMReference > reference_value(reference, false);
    NMM::TInValue<UserInfo > info_value(info, false);
    NMM::Event event("IRegistry::destroyNode", &reg_id_value, &reference_value, &info_value);
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
  }
  
  nmm_int32 IRegistry::getMaxInstanceCount(const RegistryID& reg_id, const ObjectID& node_id) const
      throw(const NMM::SerializedException&)
  {
    NMM::TOutValue<nmm_int32 > return_value;
    NMM::TInValue<RegistryID > reg_id_value(reg_id, false);
    NMM::TInValue<ObjectID > node_id_value(node_id, false);
    NMM::Event event("IRegistry::getMaxInstanceCount", &reg_id_value, &node_id_value);
    event.setReturnValue(&return_value);
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
    return return_value.getRValue();
  }
  
  nmm_int32 IRegistry::getInstanceCount(const RegistryID& reg_id, const ObjectID& node_id) const
      throw(const NMM::SerializedException&)
  {
    NMM::TOutValue<nmm_int32 > return_value;
    NMM::TInValue<RegistryID > reg_id_value(reg_id, false);
    NMM::TInValue<ObjectID > node_id_value(node_id, false);
    NMM::Event event("IRegistry::getInstanceCount", &reg_id_value, &node_id_value);
    event.setReturnValue(&return_value);
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
    return return_value.getRValue();
  }
  
  string IRegistry::getPluginInfo(const RegistryID& reg_id, const ObjectID& node_id) const
      throw(const NMM::SerializedException&)
  {
    NMM::TOutValue<string > return_value;
    NMM::TInValue<RegistryID > reg_id_value(reg_id, false);
    NMM::TInValue<ObjectID > node_id_value(node_id, false);
    NMM::Event event("IRegistry::getPluginInfo", &reg_id_value, &node_id_value);
    event.setReturnValue(&return_value);
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
    return return_value.getRValue();
  }
  
  list<ObjectID >  IRegistry::getNodeList(const RegistryID& reg_id) const
      throw(const NMM::SerializedException&)
  {
    NMM::TOutValue<list<ObjectID >  > return_value;
    NMM::TInValue<RegistryID > reg_id_value(reg_id, false);
    NMM::Event event("IRegistry::getNodeList", &reg_id_value);
    event.setReturnValue(&return_value);
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
    return return_value.getRValue();
  }
  
  list<ObjectID >  IRegistry::getReservedNodeList(const RegistryID& reg_id) const
      throw(const NMM::SerializedException&)
  {
    NMM::TOutValue<list<ObjectID >  > return_value;
    NMM::TInValue<RegistryID > reg_id_value(reg_id, false);
    NMM::Event event("IRegistry::getReservedNodeList", &reg_id_value);
    event.setReturnValue(&return_value);
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
    return return_value.getRValue();
  }
  
  list<NodeInfo >  IRegistry::getNodeInfoList(const RegistryID& reg_id) const
      throw(const NMM::SerializedException&)
  {
    NMM::TOutValue<list<NodeInfo >  > return_value;
    NMM::TInValue<RegistryID > reg_id_value(reg_id, false);
    NMM::Event event("IRegistry::getNodeInfoList", &reg_id_value);
    event.setReturnValue(&return_value);
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
    return return_value.getRValue();
  }
  
  NodeDescription IRegistry::getNodeDescription(const RegistryID& reg_id, const ObjectID& node_id) const
      throw(const NMM::SerializedException&)
  {
    NMM::TOutValue<NodeDescription > return_value;
    NMM::TInValue<RegistryID > reg_id_value(reg_id, false);
    NMM::TInValue<ObjectID > node_id_value(node_id, false);
    NMM::Event event("IRegistry::getNodeDescription", &reg_id_value, &node_id_value);
    event.setReturnValue(&return_value);
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
    return return_value.getRValue();
  }
  
  list<NodeDescription >  IRegistry::getNodeDescriptionList(const RegistryID& reg_id)
      throw(const NMM::SerializedException&)
  {
    NMM::TOutValue<list<NodeDescription >  > return_value;
    NMM::TInValue<RegistryID > reg_id_value(reg_id, false);
    NMM::Event event("IRegistry::getNodeDescriptionList", &reg_id_value);
    event.setReturnValue(&return_value);
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
    return return_value.getRValue();
  }
  
  NodeFactoryID IRegistry::getNodeFactoryID(const RegistryID& reg_id, const ObjectID& node_id) const
      throw(const NMM::SerializedException&)
  {
    NMM::TOutValue<NodeFactoryID > return_value;
    NMM::TInValue<RegistryID > reg_id_value(reg_id, false);
    NMM::TInValue<ObjectID > node_id_value(node_id, false);
    NMM::Event event("IRegistry::getNodeFactoryID", &reg_id_value, &node_id_value);
    event.setReturnValue(&return_value);
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
    return return_value.getRValue();
  }
  
  void IRegistry::resetInternalReservation()
      throw(const NMM::SerializedException&)
  {
    NMM::Event event("IRegistry::resetInternalReservation");
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
  }
  
  void IRegistry::resetSpecificInternalReservation(const UserInfo& user_info)
      throw(const NMM::SerializedException&)
  {
    NMM::TInValue<UserInfo > user_info_value(user_info, false);
    NMM::Event event("IRegistry::resetSpecificInternalReservation", &user_info_value);
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
  }
  
  void IRegistry::initTransaction(const UserInfo& info) 
      throw(const NMM::SerializedException&, const RegistryException&)
  {
    NMM::TInValue<UserInfo > info_value(info, false);
    NMM::Event event("IRegistry::initTransaction", &info_value);
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        if(RegistryException* ex = dynamic_cast<RegistryException*>(event.getException())) {
          throw RegistryException(*ex);
        }
        throw NMM::SerializedException(*event.getException());
      }
    }
  }
  
  void IRegistry::closeTransaction(const UserInfo& info)
      throw(const NMM::SerializedException&)
  {
    NMM::TInValue<UserInfo > info_value(info, false);
    NMM::Event event("IRegistry::closeTransaction", &info_value);
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
  }
  
  RegistryID IRegistry::getRegistryID() const
      throw(const NMM::SerializedException&)
  {
    NMM::TOutValue<RegistryID > return_value;
    NMM::Event event("IRegistry::getRegistryID");
    event.setReturnValue(&return_value);
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
    return return_value.getRValue();
  }
  
  list<RegistryID >  IRegistry::getChildList(const RegistryID& registry_id) const
      throw(const NMM::SerializedException&)
  {
    NMM::TOutValue<list<RegistryID >  > return_value;
    NMM::TInValue<RegistryID > registry_id_value(registry_id, false);
    NMM::Event event("IRegistry::getChildList", &registry_id_value);
    event.setReturnValue(&return_value);
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
    return return_value.getRValue();
  }
  
  nmm_bool IRegistry::isSubRegistry(const RegistryID& reg_id) const
      throw(const NMM::SerializedException&)
  {
    NMM::TOutValue<nmm_bool > return_value;
    NMM::TInValue<RegistryID > reg_id_value(reg_id, false);
    NMM::Event event("IRegistry::isSubRegistry", &reg_id_value);
    event.setReturnValue(&return_value);
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
    return return_value.getRValue();
  }
  
  RegistryMemento IRegistry::getMemento()
      throw(const NMM::SerializedException&)
  {
    NMM::TOutValue<RegistryMemento > return_value;
    NMM::Event event("IRegistry::getMemento");
    event.setReturnValue(&return_value);
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
    return return_value.getRValue();
  }
  
  void IRegistry::setMemento(const RegistryMemento& rm)
      throw(const NMM::SerializedException&)
  {
    NMM::TInValue<RegistryMemento > rm_value(rm, false);
    NMM::Event event("IRegistry::setMemento", &rm_value);
    if(const_cast<IRegistry*>(this)->sendEvent(event) == NMM::FAILURE) {
      if (event.getException()) {
        throw NMM::SerializedException(*event.getException());
      }
    }
  }

  const char* IRegistry::IRegistry_type = "IRegistry";
  NMM::TConstructor<IRegistry > __IRegistry_prototype("IRegistry","NMM");


  const char* IRegistry::requestNode_event = "IRegistry::requestNode";
  const char* IRegistry::requestObject_event = "IRegistry::requestObject";
  const char* IRegistry::releaseNode_event = "IRegistry::releaseNode";
  const char* IRegistry::releaseObject_event = "IRegistry::releaseObject";
  const char* IRegistry::initRequest_event = "IRegistry::initRequest";
  const char* IRegistry::initObjectRequest_event = "IRegistry::initObjectRequest";
  const char* IRegistry::searchNode_event = "IRegistry::searchNode";
  const char* IRegistry::reserveNode_event = "IRegistry::reserveNode";
  const char* IRegistry::instantiateNode_event = "IRegistry::instantiateNode";
  const char* IRegistry::releaseReservation_event = "IRegistry::releaseReservation";
  const char* IRegistry::destroyNode_event = "IRegistry::destroyNode";
  const char* IRegistry::getMaxInstanceCount_event = "IRegistry::getMaxInstanceCount";
  const char* IRegistry::getInstanceCount_event = "IRegistry::getInstanceCount";
  const char* IRegistry::getPluginInfo_event = "IRegistry::getPluginInfo";
  const char* IRegistry::getNodeList_event = "IRegistry::getNodeList";
  const char* IRegistry::getReservedNodeList_event = "IRegistry::getReservedNodeList";
  const char* IRegistry::getNodeInfoList_event = "IRegistry::getNodeInfoList";
  const char* IRegistry::getNodeDescription_event = "IRegistry::getNodeDescription";
  const char* IRegistry::getNodeDescriptionList_event = "IRegistry::getNodeDescriptionList";
  const char* IRegistry::getNodeFactoryID_event = "IRegistry::getNodeFactoryID";
  const char* IRegistry::resetInternalReservation_event = "IRegistry::resetInternalReservation";
  const char* IRegistry::resetSpecificInternalReservation_event = "IRegistry::resetSpecificInternalReservation";
  const char* IRegistry::initTransaction_event = "IRegistry::initTransaction";
  const char* IRegistry::closeTransaction_event = "IRegistry::closeTransaction";
  const char* IRegistry::getRegistryID_event = "IRegistry::getRegistryID";
  const char* IRegistry::getChildList_event = "IRegistry::getChildList";
  const char* IRegistry::isSubRegistry_event = "IRegistry::isSubRegistry";
  const char* IRegistry::getMemento_event = "IRegistry::getMemento";
  const char* IRegistry::setMemento_event = "IRegistry::setMemento";


}


#include "nmm/nmm_suffix.hpp"

