/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "LocalNodeFactory.hpp"
#include "nmm/multimedia/Node.hpp"
#include "nmm/multimedia/Plugin.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {


LocalNodeFactory::LocalNodeFactory()
  : NodeFactory("LocalNodeFactory",0)
{ }

LocalNodeFactory::~LocalNodeFactory() {
  /* destroy the FacotorObjects */
  for (map<ObjectID, FactoryObject*>::iterator it = __factory_map.begin(); it != __factory_map.end(); ++it)
    delete it->second;
}

void LocalNodeFactory::registerNode(const ObjectID& node_id, const FactoryObject& fo) {
  /* Begin> Check if the ObjectID is already registered */
  map<ObjectID, FactoryObject*>::iterator it = __factory_map.find(node_id);
  if (it != __factory_map.end()) {
    WARNING_STREAM("A FactoryObject for plugin " << fo.getPluginName() << " and Node " << fo.getNodeName() << " is already registered" << endl);
    return;
  }
  /* End< */

  /* insert a copy of the FactoryObject */
  __factory_map.insert(pair<ObjectID, FactoryObject*>( node_id, fo.copy()));
}

void LocalNodeFactory::unregisterNode(const ObjectID& node_id) {
  /* Search the iterator for the given ObjectID */
  map<ObjectID, FactoryObject*>::iterator it = __factory_map.find(node_id);

  /* If no such ObjectID was registered return */
  if (it == __factory_map.end()) {
    WARNING_STREAM("No FactoryObject for plugin "<< node_id <<  endl);
    return;
  }

  /* Delete the FactoryObject */
  delete it->second;

  /* remove the entry from the map */
  __factory_map.erase(it);
}


Node& LocalNodeFactory::createNode (const FactoryTicket& fp)
    throw (const FactoryException& )
{
  /* search the FactoryObject for the Node */
  map<ObjectID, FactoryObject*>::iterator it = __factory_map.find(fp.getObjectID());

  /* If no such FactoryObject is registered, throw an exception */
  if (it == __factory_map.end())
    throw FactoryException("Node FactoryObject available of this Node");

  /* The FactoryTicket must be valid, otherwise an exception is thrown */
  if (!fp.isValid())
    throw FactoryException("FactoryTicket is invalid");

  /* create the node */
  Node* tmp_node = it->second->createNode();

  /* Check if the node is successfully created */
  if (!tmp_node)
    throw FactoryException("Could not create an instance of the Node");

  //Send the configuration event to the node. This event restores the configuration
  //which is used for the node registration.
  {
    auto_ptr<CEvent> cevent(fp.getConfigEvent().copy());
    for(unsigned int i=0; i<cevent->size(); ++i) {
      tmp_node->receiveEvent((*cevent)[i]);
    }
  }
  
  /*Begin> Create an entry in node_list for the new Node*/
  __node_list.insert(pair<Node*, FactoryObject*>(tmp_node, it->second));
  /*End< */

  /* The FactoryTicket is invalid now */
  setTicketInvalid(fp);

  /* return the node */
  return *tmp_node;
}

void LocalNodeFactory::releaseNode (const Node& node)
    throw (const FactoryException& )
{
  /* seach the FactoryObject for the given node */
  map<Node*, FactoryObject*>::iterator it = __node_list.find(const_cast<Node*>(&node));

  /* if no such FactoryObject available throw an exception */
  if (it == __node_list.end()) {
    throw FactoryException("No such node created");
  }

  /* release the node */
  it->second->releaseNode(const_cast<Node*>(&node));

  /* remove the entry from map */
  __node_list.erase(it);
}


void LocalNodeFactory::registerPlugin(Plugin* plugin) {
  if (plugin->getSize() < 1) {
      WARNING_STREAM2("No ObjectID available for node " + plugin->getNodeName() << endl);
    return;
  }

  for (unsigned int i = 0; i < plugin->getSize(); ++i) {
    /* If a specific NodeFactory is set use this one */
    FactoryObject* fo = plugin->getFactoryObject();
    if (fo) {
      registerNode(plugin->getObjectID(i), *fo);
      delete fo;    
    } else {
      delete fo;
      throw Exception("No FactoryObject available for node " + plugin->getNodeName());
    }
  }
}

void LocalNodeFactory::unregisterPlugin(const Plugin* plugin) {
  for (unsigned int i = 0; i< plugin->getSize(); ++i) {
    unregisterNode(plugin->getObjectID(i));
  }
}

}//end namespace

DEFINE_SINGLETON(NMM::LocalNodeFactory);

#include "nmm/nmm_suffix.hpp"
