/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include "LocalRegistry.hpp"
#include "LocalNodeFactory.hpp"
#include "PluginManager.hpp"
#include "nmm/comm/ReferenceManager.hpp" 
#include "nmm/multimedia/Node.hpp" 
#include "nmm/multimedia/Plugin.hpp"

#include <iostream>

#include "nmm/nmm_prefix.hpp"

namespace NMM {

LocalRegistry::LocalRegistry()
  : Registry("LocalRegistry", 0), __node_id_counter(0), __use_memento(false)
{
  __internal_plugin_manager.reset( SPluginManager::getInstance().getInterface<IPluginManager>() );
  propagateRegistry();
  propagateRegistry();
}

LocalRegistry::~LocalRegistry() {
  DEBUG_STREAM2(">Destructor" << endl);

  /* Begin> Unregister all registered nodes */
  try {
    while(!__registered_nodes.empty()) {
      unregisterNode(__registered_nodes.begin()->first);
    }
  } catch (const RegistryException& re) {
    ERROR_STREAM( re << endl);
  }
  /* End< */

  DEBUG_STREAM2("<Destructor" << endl);
}


ObjectID LocalRegistry::registerNode(const NodeDescription& np, int max_instance, unsigned int index)
{
  /* create an InternalStruct that stores the information */
  InternalStruct *tmp = new InternalStruct();

  /* set the NodeDescription */
  tmp->property = np;

  /* set the max_instance count */
  tmp->max_instance_count = max_instance;

  /* store the index */
  tmp->plugin_index = index;

  /* create the ObjectID */
  ObjectID node_id = __node_id_counter;

  /* increment the node-id counter */
  ++__node_id_counter; 

  /* register the nodedescription */
  __registered_nodes.insert(pair<ObjectID, InternalStruct*>(node_id, tmp));

  /* return the ObjectID */
  return node_id;
}


void LocalRegistry::unregisterNode(const ObjectID& node_id)
  throw (const RegistryException& )
{
  /* Begin > Check if Node is currently in use*/
  InternalIt it = __registered_nodes.find(node_id);
  if ( it != __registered_nodes.end()) {
    if ( it->second->instance_count > 0) {
      throw RegistryException("Node " + it->second->property.getNodeName() + " is currently in use and can be unregistered");
    }
  /* End < */
    DEBUG_STREAM("Unregister node " << it->second->property.getNodeName() << endl);

    /* Delete the InternalStruct */
    delete it->second;
    /* remove the entry from the map*/
    __registered_nodes.erase(node_id);
  } else {
    WARNING_STREAM2("No such node registered" << endl);
  }
}


FactoryTicket LocalRegistry::reserveNode(const RegistryID& reg_id, const ObjectID& node_id, const UserInfo& info)
  throw(const RegistryException&)
{
  /* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->reserveNode(reg_id, node_id, info);
    else
      throw RegistryException("Unknown Registr");
  /* End< */

  InternalIt it = __registered_nodes.find(node_id);
  if ( it != __registered_nodes.end() ) {
    /* Begin > Handle nodes availability and create a FactoryProperty*/
    if ( (it->second->max_instance_count == -1) || (it->second->max_instance_count - it->second->instance_count > 0)) {
      /*Begin> Create a ticket to create the node*/
      FactoryTicket fp;
      /* Set the NodeFactoryID */
      fp.setNodeFactoryID(NodeFactoryID("LocalNodeFactory"));
      /* Set the ObjectID */
      fp.setObjectID(node_id);
      /* Set the config event which is stored in the NodeDescription */
      fp.setConfigEvent(it->second->property.getConfigEvent());
      /* The Registry base class sets the ticket as valid */
      setTicketValid(fp);
      /*End<*/

      /* increment the instance counter */
      ++it->second->instance_count;
      /* Mark it as reserved from a specific user */
      __reserved_nodes.insert(pair<ObjectID, UserInfo>(node_id, info));
      /* return the FactoryTicket */
      return fp;
    }
    /* Begin> No more instances of this node available, so throw an exception */
    else {
      throw RegistryException("No more instances available");
    }
    /* End < */
  }
  /* If the ObjectID is unknown throw an exception */
  throw RegistryException("Invalid ObjectID. ObjectID of given Node is not stored in this registry");
}


void LocalRegistry::releaseReservation(const RegistryID& reg_id, const ObjectID& node_id, const UserInfo& user_info)
{
  /* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->releaseReservation(reg_id, node_id, user_info);
    else
      /* The Registry is unkown and we throw an Exception */
      throw Exception("Unknwon RegistryID");
  /* End< */

  multimap<ObjectID, UserInfo>::iterator it = __reserved_nodes.find(node_id);
  if (it == __reserved_nodes.end()) {
    ERROR_STREAM("No such node available to release" << endl);
    return;
  }

  /* Begin> Check if the caller of this method reserved the node. In this case the is_valid flag
   *        to true, otherwise to false
   */
  bool is_valid = false;
  while (it->first == node_id && it != __reserved_nodes.end()) {
    if (it->second == user_info) {
      is_valid = true;
      break;
    }
    ++it;
  }
  /* End< */

  /* Begin> If the caller has not reserved the node return */
  if (!is_valid) {
    ERROR_STREAM("Node is not reserved by user" << endl);
    return;
  }
  /* End< */

  /* remove the iterator from the __reserved_nodes */
  __reserved_nodes.erase(it);

  /* Begin> Decrement the instance-counter */
  InternalIt internal_it = __registered_nodes.find(node_id);
  if (internal_it != __registered_nodes.end()) {
    --internal_it->second->instance_count;
  } else {
    ERROR_STREAM("No such node available to release" << endl);
    return;
  }
  /* End< */

}


list<ObjectID >  LocalRegistry::getNodeList(const RegistryID& reg_id) const {
  /* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->getNodeList(reg_id);
    else
      /* The Registry is unkown and we throw an Exception */
      throw Exception("Unknwon RegistryID");
  /* End< */

  /* create the return value */
  list<ObjectID> node_list;

  /* Begin> Store all ObjectID's that can be instantiated. A node can be instantiated if the
   *        max_instance_count = -1 or (max_instance_count - instance_count > 0)
   */
  for (CInternalIt it = __registered_nodes.begin(); it != __registered_nodes.end(); ++it)
  {
    if ( (it->second->max_instance_count == -1) || (it->second->max_instance_count - it->second->instance_count))
      node_list.push_back(it->first);
  }
  /* End< */

  /* return the node_list */
  return node_list;
}

list<ObjectID >  LocalRegistry::getReservedNodeList(const RegistryID& reg_id) const {
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->getReservedNodeList(reg_id);
    else
      /* The Registry is unkown and we throw an Exception */
      throw Exception("Unknwon RegistryID");
  /* End< */

  /* create the return value */
  list<ObjectID> node_list;

  /* Begin> Store all ObjectID's that can not be instantiated because all instances are reserved. A
   *        node can not be instantiated if the max_instance_count != -1 or
   *        (max_instance_count - instance_count == 0)
   */
  for (CInternalIt it = __registered_nodes.begin(); it != __registered_nodes.end(); ++it)
  {
    if ( (it->second->max_instance_count != -1) && !(it->second->max_instance_count - it->second->instance_count))
      node_list.push_back(it->first);
  }

  /* return the node_list */
  return node_list;
}

NodeDescription LocalRegistry::getNodeDescription(const RegistryID& reg_id, const ObjectID& node_id) const {
  /* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->getNodeDescription(reg_id, node_id);
    else
      /* The Registry is unkown and we throw an Exception */
      throw Exception("Unknwon RegistryID");
  /* End< */

  /* find the iterator in the current list */
  CInternalIt it = __registered_nodes.find(node_id);
  /* If the ObjectID is unknown throw an exception */
  if (it == __registered_nodes.end())
    throw RegistryException("Unknown ObjectID");

  /* return the NodeDescription */
  return it->second->property;
}

list<ObjectID >  LocalRegistry::searchNode(const RegistryID& reg_id, const NodeDescription& node_desc) const {
  list<ObjectID> node_list;
  /* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->searchNode(reg_id, node_desc);
    else
      /* The Registry is unkown and we throw an Exception */
      throw Exception("Unknwon RegistryID");
  /* End< */

  for (CInternalIt it = __registered_nodes.begin(); it  != __registered_nodes.end(); ++it)
  {
    if ( (it->second->max_instance_count == -1) || (it->second->max_instance_count - it->second->instance_count > 0))
      if (it->second->property.isSubset(node_desc))
        node_list.push_back(it->first);
  }
  return node_list;
}

nmm_int32 LocalRegistry::getMaxInstanceCount(const RegistryID& reg_id, const ObjectID& node_id) const {
  /* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID()) 
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->getMaxInstanceCount(reg_id, node_id);
    else
      /* The Registry is unkown and we throw an Exception */
      throw Exception("Unknwon RegistryID");
  /* End< */

  CInternalIt it = __registered_nodes.find(node_id);
  if ( it == __registered_nodes.end())
    return 0;

  return it->second->max_instance_count;
}

nmm_int32 LocalRegistry::getInstanceCount(const RegistryID& reg_id, const ObjectID& node_id) const {
  /* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->getInstanceCount(reg_id, node_id);
    else
      /* The Registry is unkown and we throw an Exception */
      throw Exception("Unknwon RegistryID");
  /* End< */

  CInternalIt it = __registered_nodes.find(node_id);
  if ( it == __registered_nodes.end())
    return FAILURE;

  return it->second->instance_count;
}


string LocalRegistry::getPluginInfo(const RegistryID& reg_id, const ObjectID& node_id) const {
  /* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->getPluginInfo(reg_id, node_id);
    else
      /* The Registry is unkown and we throw an Exception */
      throw Exception("Unknwon RegistryID");
  /* End< */

  CInternalIt it = __registered_nodes.find(node_id);
  if (it == __registered_nodes.end()) {
    cerr << "Error in plugin info" << endl;
    return string();
  }

  return it->second->property.getLibraryName();
}

list<NodeDescription> LocalRegistry::getNodeDescriptionList(const RegistryID& reg_id)  {
/* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->getNodeDescriptionList(reg_id);
    else
      /* The Registry is unkown and we throw an Exception */
      throw Exception("Unknwon RegistryID");
  /* End< */

  list<NodeDescription>  result;
  for ( CInternalIt it = __registered_nodes.begin(); it != __registered_nodes.end(); ++it)
    result.push_back(it->second->property);
  
  return result;
}


NodeFactoryID LocalRegistry::getNodeFactoryID(const RegistryID& reg_id, const ObjectID& node_id) const {
  /* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->getNodeFactoryID(reg_id, node_id);
    else
      throw RegistryException("Unknown NodeFactory.");
  /* End< */

  /* return the FactoryID of the LocalNodeFactory */
  return SLocalNodeFactory::getInstance().getFactoryID();
}


INode* LocalRegistry::instantiateNode(const RegistryID& reg_id, const FactoryTicket& nf_id, const UserInfo& user_info)
  throw (const RegistryException&)
{
  /* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->instantiateNode(reg_id, nf_id, user_info);
    else
      /* The Registry is unkown and we throw an Exception */
      throw Exception("Unknwon RegistryID");
  /* End< */

  try {
     if (__use_memento)  {
       InternalIt it = __registered_nodes.find(nf_id.getObjectID());
       if ( it != __registered_nodes.end()
         && !SPluginManager::getInstance().isPluginUsed(it->second->property.getLibraryName()))
       {
         string plugin = it->second->property.getLibraryName();
         /* 1. load the library in our address space*/
         __internal_plugin_manager->loadPlugin(plugin);


         /* 3. If the plugin is not registered by a NodeFactory, register it */
         if (!SPluginManager::getInstance().isFactoryPartActive(plugin))
           __internal_plugin_manager->activateFactoryPart(plugin);
       } 
     }

     Node& node = SLocalNodeFactory::getInstance().createNode(nf_id);

     __allocated_nodes.insert(pair<NMMReference, UserInfo>(node.getReference(), user_info));

     return node.getInterface<INode>();
   } catch (const Exception& e) {
     ERROR_STREAM( e << endl) ;
     throw RegistryException(e.getComment());
   }
}

void LocalRegistry::destroyNode(const RegistryID& reg_id, const NMMReference& reference, const UserInfo& user_info) {
  /* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->destroyNode(reg_id, reference, user_info);
    else
      /* The Registry is unkown and we throw an Exception */
      throw Exception("Unknwon RegistryID");
  /* End< */


  Node* node = SReferenceManager::getInstance().getObject<Node>(reference);
  if (!node)
    return;

  map<NMMReference, UserInfo>::iterator it = __allocated_nodes.find(reference);

  /* Begin> Error handling */
  if (it == __allocated_nodes.end()) {
    ERROR_STREAM("Registry didn't create this node");
    return;
  }

  if (user_info != it->second) {
    ERROR_STREAM("User don't allocate Node");
    return;
  }
  /* End< */

  __allocated_nodes.erase(it);

  /* Release the node*/
  SLocalNodeFactory::getInstance().releaseNode(*node);
}

void LocalRegistry::resetInternalReservation() {
  /* Begin> first release the allocated nodes */
  /* Create a temporary copy to iterate through. So changes on the original internal structs can be performed */
  map<NMMReference, UserInfo> t_allocated_nodes = __allocated_nodes;
  for (map<NMMReference, UserInfo>::iterator it = t_allocated_nodes.begin(); it != t_allocated_nodes.end(); ++it)
  {
    /* Release the node, wich calls releaseReservation and destroyNode */
    destroyNode(getRegistryID(), it->first, it->second);
  }
  /* End< */

  /* Begin> release the nodes which were only reserved */
  /* Create a temporary copy to iterate through. So changes on the original internal structs can be performed */
  multimap<ObjectID, UserInfo> t_reserved_nodes = __reserved_nodes;
  for (multimap<ObjectID, UserInfo>::iterator it = t_reserved_nodes.begin(); it != t_reserved_nodes.end(); ++it)
  {
    /* The Registry is only responsible for the reservation not for destruction. So call
     * only releaseReservation
     */
    releaseReservation(getRegistryID(), it->first, it->second);
  }
  /* End */

  /* Forward the method call to all children */
  for (list<IRegistry*>::iterator it = getIChildList().begin(); it != getIChildList().end(); ++it)
    (*it)->resetInternalReservation();
}

void LocalRegistry::resetSpecificInternalReservation(const UserInfo& user_info) {
  /* Begin> first release the allocated nodes */
  /* Create a temporary copy to iterate through. So changes on the original internal structs can be performed */
  map<NMMReference, UserInfo> t_allocated_nodes = __allocated_nodes;
  for (map<NMMReference, UserInfo>::iterator it = t_allocated_nodes.begin(); it != t_allocated_nodes.end(); ++it)
  {
    /* If the UserInfo's match */
    if (it->second.getUsername() == user_info.getUsername())
      /* Release the node, wich calls releaseReservation and destroyNode */
      destroyNode(getRegistryID(), it->first, it->second);
  }
  /* End< */

  /* Begin> release the nodes which were only reserved */
  /* Create a temporary copy to iterate through. So changes on the original internal structs can be performed */
  multimap<ObjectID, UserInfo> t_reserved_nodes = __reserved_nodes;
  for (multimap<ObjectID, UserInfo>::iterator it = t_reserved_nodes.begin(); it != t_reserved_nodes.end(); ++it)
  {
    /* If the UserInfo's match */
    if (it->second.getUsername() == user_info.getUsername())
      /* The Registry is only responsible for the reservation not for destruction. So call
       * only releaseReservation
       */
      releaseReservation(getRegistryID(), it->first, it->second);
  }
  /* End */

  /* Forward the method call to all children */
  for (list<IRegistry*>::iterator it = getIChildList().begin(); it != getIChildList().end(); ++it)
    (*it)->resetSpecificInternalReservation(user_info);
}

list<NodeInfo> LocalRegistry::getNodeInfoList(const RegistryID& reg_id) const {
  /* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->getNodeInfoList(reg_id);
    else
      /* The Registry is unkown and we throw an Exception */
      throw Exception("Unknwon RegistryID");
  /* End< */

  /* create the return value */
  list<NodeInfo> node_info_list;

  /* Begin> create a NodeInfo for all registered nodes from the stored nodedescription */
  for (CInternalIt it = __registered_nodes.begin(); it != __registered_nodes.end();  ++it) {
    node_info_list.push_back( NodeInfo(it->first, it->second->property.getLibraryName(),
        it->second->property.getNodeName(), it->second->instance_count, it->second->max_instance_count,
        it->second->property.getNodeType())
    );
  }
  /* End< */

  /* return the list */
  return node_info_list;
}

RegistryMemento LocalRegistry::getMemento() {
  RegistryMemento memento;
  memento.setRegistryID( getRegistryID() );

  for (InternalIt it = __registered_nodes.begin(); it != __registered_nodes.end(); ++it)
  {
    MementoStruct ms;
    ms.setNodeDescription(it->second->property);
    ms.setObjectID( it->first );
    ms.setMaxInstanceCount( it->second->max_instance_count );
    ms.setPluginIndex( it->second->plugin_index );
    memento.addMementoStruct( ms );
  }
  return memento;
}

void LocalRegistry::setMemento( const RegistryMemento& rm ) {
  list<string> plugin_list;
  for (list<MementoStruct>::const_iterator it = rm.getMementoList().begin(); it != rm.getMementoList().end(); ++it)
  {
    InternalStruct* is      = new InternalStruct();
    is->property            = it->getNodeDescription();
    is->max_instance_count  = it->getMaxInstanceCount();
    is->plugin_index        = it->getPluginIndex();
    __registered_nodes.insert(pair<ObjectID, InternalStruct*>(it->getObjectID(), is));
    if ( __node_id_counter < it->getObjectID())
      __node_id_counter = it->getObjectID();
    plugin_list.push_back( it->getNodeDescription().getLibraryName() );
  }

  plugin_list.unique();

  for ( list<string>::iterator it = plugin_list.begin(); it != plugin_list.end(); ++it ) {
    SPluginManager::getInstance().defineObjectIDs( *it, getObjectIDList(*it) );
  }
  
  RegistryID reg_id (rm.getRegistryID().getRegistryName(), rm.getRegistryID().getReference());
  setRegistryID( reg_id );
  __use_memento = true;   
}

vector<list<ObjectID> > LocalRegistry::getObjectIDList(const string& plugin) {
  vector< list<ObjectID> > result; 
  string last_node;
  
  for (InternalIt it = __registered_nodes.begin(); it != __registered_nodes.end(); ++it)
    if (it->second->property.getLibraryName() == plugin) {
      if ( last_node == it->second->property.getNodeName() ) {
        if ( it->second->plugin_index + 1 > result.size() )
          result.resize(it->second->plugin_index + 1);
          
        result[it->second->plugin_index].push_back( it->first );        
      } else {
        if ( it->second->plugin_index + 1 > result.size() )
          result.resize(it->second->plugin_index + 1);

        result[it->second->plugin_index].push_back( it->first );
        last_node = it->second->property.getNodeName();
      }
    }
      
  for (vector<list<ObjectID> >::iterator it = result.begin();
       it != result.end(); ++it)
  {
    it->sort();
  }
  
  return result;
}

void LocalRegistry::registerPlugin(Plugin* plugin) {
  const list<pair<NodeDescription, int> >& nd_list = plugin->getNodeDescriptionList();
  for (list<pair<NodeDescription, int> >::const_iterator it  = nd_list.begin(); it != nd_list.end(); ++it) {
    plugin->addObjectID(registerNode(it->first, it->second, plugin->getPluginIndex()));
  }
}

void LocalRegistry::unregisterPlugin(const Plugin* plugin){
  for ( unsigned int i = 0; i< plugin->getSize(); ++i ) {
    unregisterNode(plugin->getObjectID(i));
  }
}

} //end namespace

DEFINE_SINGLETON(NMM::LocalRegistry);

#include "nmm/nmm_suffix.hpp"
