/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include "NMMApplication.hpp"

#include "ServerRegistry.hpp"
#include "PluginManager.hpp"

#include "nmm/comm/LocalAddress.hpp"
#include "nmm/comm/ReferenceManager.hpp"
#include "nmm/comm/ProcessUserInfo.hpp"

#include "nmm/comm/netstrategy/TCPAcceptor.hpp"

#include "nmm/comm/serialize/StdIStream.hpp"
#include "nmm/comm/serialize/StdOStream.hpp"
#include "nmm/comm/serialize/MagicNumberIValueStream.hpp"
#include "nmm/comm/serialize/MagicNumberOValueStream.hpp"

#include "nmm/utils/directory_utils.hpp"

/* guarantee linking against transport strategies when building static
   NMM libraries. */
#include "nmm/comm/netstrategy/StaticTransportStrategyList.hpp"
#include "nmm/comm/StaticLocalStrategy.hpp"

#include <fstream>

// XXX: fw: platform specific longjmp
#include <setjmp.h>

// XXX: fw: platform specific directory handling
#include <sys/types.h>
#include <sys/stat.h>

// XXX: fw: platform specific signal handling
#include <signal.h>
#include <cerrno>
#include <errno.h>

#ifndef VERSION
#define VERSION "1.0.0"
#endif

#include "nmm/nmm_prefix.hpp"

static const std::string NMM_VERSION_NUMBER = VERSION;

#ifdef BUILD_LINUX
static const std::string license_info = string("Network-Integrated Multimedia Middleware (NMM) Version ") + NMM_VERSION_NUMBER + std::string("\n\nCopyright (C) 2006-2007\nMotama GmbH, Saarbruecken, Germany\nhttp://www.motama.com \n\nNMM is free software; you can redistribute it and/or modify it\nunder the terms of the GNU General Public License as published by\nthe Free Software Foundation; only version 2 of the License.\n\nNMM is distributed in the hope that it will be useful, but WITHOUT\nANY WARRANTY; without even the implied warranty of MERCHANTABILITY\nor FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public \nLicense for more details.\n\nYou should have received a copy of the GNU General Public License\nalong with NMM; if not, write to the Free Software Foundation,\nInc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.");
#endif
#ifdef BUILD_WINXP
static const std::string license_info = std::string("Network-Integrated Multimedia Middleware (NMM) Version ") + NMM_VERSION_NUMBER + string("\n\nCopyright (C) 2006-2007\nMotama GmbH, Saarbruecken, Germany\nhttp://www.motama.com \n\nNMM is released under the terms and conditions of the \nNMM Non-Commercial License (NMM-NCL). You should have received \na copy of the NMM Non-Commercial License along with this program.\nA copy of the NMM Non-Commercial License can also be found at \nhttp://www.motama.com/NMM_Non_Commercial_License.txt ");
#endif

namespace NMM {
  
  static jmp_buf      plugin_loop_position;        ///<stack position to the loop which loads all plugins
  static string       current_plugin;              ///<current plugin
  static bool         seg_fault_occured = false;   ///<flag which is set to true if a segmentaion fault occurs during plugin loading
  
  
  /** This method catch segmentation faults during plugin loading */
  void segfaultHandler( int ) {
    seg_fault_occured = true;
    GLOBAL_ERROR_STREAM(" ********************************************** " << endl);
    GLOBAL_ERROR_STREAM(" Plugin " << current_plugin << " causes a segfault " << endl);
    GLOBAL_ERROR_STREAM(" ********************************************** " << endl << endl << endl);

    longjmp(plugin_loop_position, 0);
  }

  NMMApplication::NMMApplication(int argc_, char** argv_, bool is_local, int port, bool enable_port_listening) {
    argc = argc_;
    argv = argv_;
    if (is_local) {
      __client_registry.reset(new ClientRegistry());
      __server_registry.reset(new ServerRegistry());
      __plugin_manager.reset(SPluginManager::getInstance().getInterface<IPluginManager>());
      LocalAddress la(__server_registry->getReference());
      la.setPort(port);
      __client_registry->addRegistry(la, true);
      try {
        readRegistryConfig();
      } catch (const SerializedException& e) {
        GLOBAL_WARNING_STREAM( e << endl);
      } catch (const Exception& e) {
        GLOBAL_WARNING_STREAM( e << endl);
      }
    }
    if (enable_port_listening)
      __default_acceptor.reset(new TCPAcceptor(5000, 8000));
  }

  NMMApplication::~NMMApplication() {
  IPluginManager_var pm(SPluginManager::getInstance().getInterface<IPluginManager>());
  list<string> tmp_list = pm->getUsedPlugins();
  for (list<string>::iterator it = tmp_list.begin(); it != tmp_list.end(); ++it) {
    try {
      pm->removeLibrary(*it);
    } catch (const Exception& e) {
      GLOBAL_ERROR_STREAM("Could not remove the plugin " << *it << endl);
    }
  }
  }

  ClientRegistry& NMMApplication::getRegistry() {
    return *__client_registry.get();
  }
  const ClientRegistry& NMMApplication::getRegistry() const {
    return *__client_registry.get();
  }

  IPluginManager& NMMApplication::getPluginManager(){
    return *__plugin_manager.get();
  }
  const IPluginManager& NMMApplication::getPluginManager() const{
    return *__plugin_manager.get();
  }

  IServerRegistryImpl* NMMApplication::getServerRegistry() {
    return __server_registry.get();
  }

  const IServerRegistryImpl* NMMApplication::getServerRegistry() const {
    return __server_registry.get();
  }

  void NMMApplication::enableServerRegistry(int port) {
    if ( getServerRegistry() ) {
      /* The TCPAcceptor accpets incoming tcp connections on port 28001
       * This port is used by clients to connect to objcects running in this address space */
      __registry_acceptor.reset( new TCPAcceptor(port) );

      SReferenceManager::getInstance().registerService("Registry", getServerRegistry());
      SReferenceManager::getInstance().registerService("PluginManager", &SPluginManager::getInstance());
      /* End< */
    } else {
      GLOBAL_WARNING_STREAM ("No local ServerRegistry available that can be enabled incoming requests. Maybe a global ServerRegistry is running!" << endl);
    }
  }

  void NMMApplication::setServerRegistry(IServerRegistryImpl* registry) {
    __server_registry.reset(registry);
  }

  void NMMApplication::readRegistryConfig(const string& file) {
    string plugin_file;
    if (file.empty()) {
      string path = getNMMDir();
      //path = path.substr(1, path.size()-2);
	  while ( path[path.size()-1] == '\\' || path[path.size()-1] == '/') {
         path = path.substr(0,path.size() - 1); 
      }

	  for (unsigned int i = 0; i < path.size(); ++i) {
        if (path[i] == '/' || path[i] == '\\' || path[i] == ':') {
	      path[i] = '_';
		 }
	   }

      plugin_file =
        SProcessUserInfo::getInstance().getHomeDirectory() +
        string("/.nmm/plugins.") +
        string(VERSION) +
        string(".") +
        SProcessUserInfo::getInstance().getHostname() +
        string(".") +
        path;
    } else {
      plugin_file = SProcessUserInfo::getInstance().getHomeDirectory() + string("/.nmm/") + file;
    }
      
    ifstream* stream = new ifstream();
    stream->open(plugin_file.c_str());
    /* Begin> If no plugin information available, create it*/
    if (!(*stream)) {
      cerr << endl << "No plugin information available! If you start NMM for the first time this information is created automatically." << endl;
      cerr << "Note: If you ever change your hardware configuration or update the NMM version you must delete the file " << plugin_file << endl << endl;

      writeRegistryConfig(plugin_file);
      delete stream;
      stream = new ifstream();
      stream->open(plugin_file.c_str());
      if (!(*stream))
        throw Exception(string("Could not load plugin information from file ")+ plugin_file);
    }
    /* End< */
    
    /* Begin> Write the information to the stream */
    StdIStream s_stream(stream);
    MagicNumberIValueStream m_stream;
    m_stream.setIPartnerStream(&s_stream);
    m_stream.readData();
    ServerRegistry* reg = dynamic_cast<ServerRegistry*>(getServerRegistry());
    if (reg) {
      reg->readFromStream(m_stream);
    } else {
      GLOBAL_WARNING_STREAM("NMMApplication does not use a ServerRegistry" << file << endl);
    }
    /* End< */
    delete stream;
  }

  void NMMApplication::writeRegistryConfig(const string& file) {
    /* Begin> Check if directoy exist ~/.nmm exist, if not create it */
    string directory = SProcessUserInfo::getInstance().getHomeDirectory() + string("/.nmm") ;

	//create directory
	createDirectory(directory);

    string filename;
    if (file.empty()) {
      filename =
        directory +
        string("/plugins.") +
        string(VERSION); /*+
        string(".") +
        SProcessUserInfo::getInstance().getHostname();*/
    } else {
      filename = file;
    }

    ofstream stream(filename.c_str());
    if (!stream) {
      throw Exception(string( "Could not open file " ) + filename );
      return;
    }
    cerr << "Create config file with plugin information ..." << endl;
    /* Load Plugins */
    loadPlugins();

    /* Create the stream */
    StdOStream s_stream(&stream);
    MagicNumberOValueStream m_stream;
    m_stream.setOPartnerStream(&s_stream);

    /* write the plugin configuration */
    dynamic_cast<ServerRegistry*>(getServerRegistry())->writeToStream(m_stream);
    m_stream.writeData();
    stream.close();
    cerr << endl << "Config file successfully written. " << endl;
  }

  void NMMApplication::loadPlugins () {
    /*Begin> Load all plugins available*/
    IPluginManager_var pm(SPluginManager::getInstance().getInterface<IPluginManager>());
    if (!pm.get()) {
      GLOBAL_ERROR_STREAM("Could not get PluginManager interface" << endl);
      return;
    }

    typedef void (*sigmethod_t)(int);
    sigmethod_t old_sig_handler = ::signal(SIGSEGV, NMM::segfaultHandler); //register segfault handler

    list<string> plugin_list = pm->getPlugins();
    if (plugin_list.empty()) {
      throw Exception(string("No NMM plugins found. Please set the NMM_DEV_DIR environment variable to your NMM installation"));
    }
  
    cerr << "Loading plugins... " << endl;
    for (list<string>::const_iterator it = plugin_list.begin(); it != plugin_list.end(); ++it)
      {
		if (*it == "." || *it == "..") 
          continue;

        setjmp(plugin_loop_position); //stores the current position

        if (seg_fault_occured) {
          seg_fault_occured = false;
          continue;
        }
        int max_size = 48;
        current_plugin = *it;
        string plugin_name;
        if ( it->find_last_of('_') < it->size() ) {
          // For windows .dll names
          plugin_name = it->substr(it->find_last_of('_')+1, it->size());
          plugin_name = plugin_name.substr(0, plugin_name.find_last_of('.'));
        }
        else {
          // For Linux .so files
          plugin_name =  it->substr(6,it->size()-9);
        }
        int plugin_size = max_size - plugin_name.size();

        cerr << plugin_name;
        while (plugin_size > 0 ) {
          cerr << "\t";
          plugin_size -= 8;
        }

        // Plugin exclusion
        if(SPluginManager::getInstance().isPluginExcluded(plugin_name)) {
          cerr << " not available (excluded) " << endl;
          continue;
        }

        try {
          pm->loadLibrary(*it);
          if ( SPluginManager::getInstance().isFactoryPartActive(*it) ) {
            cerr << " available " << endl;
          } else {
            cerr << " not available " << endl;
          }
        } catch (const SerializedException& e) {
          cerr << " not available " << endl;
          GLOBAL_ERROR_STREAM(e <<  endl);
          GLOBAL_ERROR_STREAM("Error loading library " << *it << endl << endl << endl);
        }
      }
    cerr << "Finished loading plugins ..." << endl;
    /* If the old signal handler was vaild, use it. Otherwise set the SIG_DFL handler*/
    if (old_sig_handler != SIG_ERR)
      ::signal(SIGSEGV, old_sig_handler);
    else
      ::signal(SIGSEGV, SIG_DFL);
  }

  void NMMApplication::setUDPBroadcaster(UDPBroadcaster* broadcaster, int port) {
    ServerRegistry* reg = dynamic_cast<ServerRegistry*>(getServerRegistry());
    if ( reg ) {
      reg->setUDPBroadcaster( broadcaster, port );  
    } else {
      cerr << "No serverRegistry " << endl;
    }
  }

  void NMMApplication::printLicense(const string & application) {
    cerr << endl << application << " and " << license_info << endl;
  }

  int            NMMApplication::argc = 0;
  char**         NMMApplication::argv = 0;
  
}


#include "nmm/nmm_suffix.hpp"
