/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include "PluginManager.hpp"
#include "StaticPluginList.hpp"
#include "LocalNodeFactory.hpp"
#include "LocalRegistry.hpp"

#include "nmm/multimedia/Plugin.hpp"

#include "nmm/comm/serialize/SerializedException.hpp"

#include "nmm/utils/directory_utils.hpp"

#include <iostream>
#include <algorithm>

#ifndef ENABLE_STATIC_PLUGINS 
#include "DynamicLibrary.hpp"
#endif

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "nmm/nmm_prefix.hpp"

namespace NMM {

PluginManager::PluginManager()
  : NamedObject("PluginManager")
{
  addPluginPath(getNMMPluginDir());
}


PluginManager::~PluginManager() {
  DEBUG_STREAM(">Destroy"<< endl);
  /*Begin> Try to remove all used plugins */
  list<string> tmp_list = getUsedPlugins();
  for (list<string>::iterator it = tmp_list.begin(); it != tmp_list.end(); ++it) {
    try {
      removeLibrary(*it);
    } catch (const Exception& e) {
      ERROR_STREAM("Could not remove the plugin " << *it << endl);
    }
  }
  /*End< */
  DEBUG_STREAM("<Destroy"<< endl);
}

void PluginManager::removeRegistry(const Registry* registry) {
  for (map<string, PluginStruct>::iterator it = __plugin_map.begin(); it != __plugin_map.end(); ++it) {
    if (it->second.reg_active == registry){
      it->second.reg_active = 0;  
    }
  }
}

void PluginManager::removeNodeFactory(const NodeFactory* factory){
  for (map<string, PluginStruct>::iterator it = __plugin_map.begin(); it != __plugin_map.end(); ++it) {
    if (it->second.fac_active == factory){
      it->second.fac_active = 0;  
    }
  }
}


void PluginManager::addPluginPath(const string& path) {
  /*Begin> insert the path if it is not stored already in __plugin_path*/
  if( find(__plugin_path.begin(), __plugin_path.end(), path) == __plugin_path.end() ) {
    __plugin_path.push_back(path);
  } else {
    MESSAGE_STREAM("Plugin path " << path << "already set"<<endl);
  }
  /*End< */
}

int PluginManager::getPluginSize(const string& plugin)            const {
  Plugin* const plugin_array = requestPlugin( plugin );

  if (! plugin_array )
    return -1;

  return 1;
}

void PluginManager::loadLibrary(const string& library)
  throw (const SerializedException&)
{
  loadPlugin(library);
  activateRegistryPart(library);
  activateFactoryPart(library);
}

void PluginManager::unregisterPlugin(const std::string& plugin) {
  Plugin* p_array = requestPlugin(plugin);
  if ( p_array )  {
    map<string, PluginStruct>::iterator it = __plugin_map.find(plugin);

    /*Begin> Check if the plugin is loaded */
    if (it == __plugin_map.end()) {
      WARNING_STREAM("A plugin with name " << plugin << " was not loaded" << endl);
      return;
    }
    /*End< */
    it->second.plugin=0;
  }
}


void PluginManager::loadExternalPlugins(const string& path)
    throw (const SerializedException& )
{
  #ifndef ENABLE_STATIC_PLUGINS 
  list<string> file_list = readFolder(path);
  if ( file_list.empty() ) {
    WARNING_STREAM("Directory " << path << " provides no plugins. "<< endl);  
    return;
  }  
  
  /*Begin> Try to load the libraries */
  for (list<string>::iterator it = file_list.begin(); it != file_list.end(); ++it )
  {
    if ( __plugin_map.find( *it ) == __plugin_map.end() ) {
      try {
        PluginStruct ps;
        /* allocate the DynamicLibrary object */
        ps.dll_object= new DynamicLibrary();
        /* open the library */
        ps.dll_object->open(path+"/"+*it);
        /* Store the plugin object */ 
        __plugin_map.insert(pair<string, PluginStruct>(*it, ps));
      } catch (const Exception& e) {
        WARNING_STREAM ("Could not load external plugin " << *it << endl);
	WARNING_STREAM ("Error message: " << e << endl);        
      }
    } else {
      WARNING_STREAM ("External plugin " << *it << " already loaded " << endl);    
    }
  }
  #else
  WARNING_STREAM("Can not load/remove external plugin " << path << " with a static version of NMM");
  #endif
}

void PluginManager::removeExternalPlugins(const string& path)
    throw (const SerializedException& )
{
  #ifndef ENABLE_STATIC_PLUGINS 
  list<string> file_list = readFolder(path);
  if ( file_list.empty() ) {
    WARNING_STREAM("Directory " << path << " provides no plugins to remove. "<< endl);  
    return;
  }  
  
  for ( list<string>::iterator it = file_list.begin(); it != file_list.end(); ++it) {
    map<string, PluginStruct>::iterator p_it = __plugin_map.find(*it);

    /*Begin> Check if the plugin was loaded */
    if (p_it == __plugin_map.end()) {
      WARNING_STREAM("The external plugin with name " << *it << " was not loaded" << endl);
      continue;
    }
    /*End< */  
    
    if (p_it->second.dll_object) {
      try {
        p_it->second.dll_object->close();
      } catch (const Exception& e) {
        WARNING_STREAM("Could not close external library " << p_it->first << endl);  
      }
      delete p_it->second.dll_object;
      __plugin_map.erase(*it);
    }    
  }
  #else
  WARNING_STREAM("Can not load/remove external plugin " << path << " with a static version of NMM");  
  #endif 
}


void PluginManager::loadPlugin(const string& plugin) 
    throw (const SerializedException& )
{
  #ifdef ENABLE_STATIC_PLUGINS 
  loadStaticPlugin(plugin);
  #else
  loadDllPlugin(plugin);
  #endif
}

#ifdef ENABLE_STATIC_PLUGINS 

void PluginManager::loadStaticPlugin(const string& plugin) 
    throw (const SerializedException& )
{
  if (isPluginUsed(plugin)) {
    MESSAGE_STREAM("Plugin " << plugin << " is already loaded" << endl);
    return;
  }

  Plugin** tmp_plugins = nmm_static_plugins;
  Plugin*  tmp_plugin = 0;
  while(*tmp_plugins) {
    if ( (*tmp_plugins)->getLibraryName() == plugin ) {
      tmp_plugin = *tmp_plugins;
      break;
    }
    ++tmp_plugins;
  }
  
  if ( tmp_plugin ) {
    PluginStruct ps;
    ps.plugin = tmp_plugin;
    ps.plugin->setPluginManager(this);
    try {
      ps.plugin->nmmInit();
      __plugin_map.insert(pair<string, PluginStruct>(plugin, ps));
      applyDefinedObjectIDs( plugin );
    } catch (const Exception& e) {
      WARNING_STREAM( e << endl);
    }
  }
}
#else 

void PluginManager::loadDllPlugin(const string& plugin)
    throw (const SerializedException&)
{
  if (isPluginUsed(plugin)) {
    MESSAGE_STREAM("Plugin " << plugin << " is already loaded" << endl);
    return;
  }

  list<string> file_list;
  string folder;
  /*Begin> Find the folder for the given library*/
  for (list<string>::iterator it = __plugin_path.begin(); it != __plugin_path.end(); ++it)
  {
    try {
      file_list = readFolder(*it);
      if (find(file_list.begin(), file_list.end(), plugin) != file_list.end()) {
        folder = *it;
        break; //leave the loop
      }
    } catch (const Exception& e) {
      ERROR_STREAM(e << endl);
    }
  }
  /*End< */

  if (folder.empty()) {
    ERROR_STREAM (" Could not find the library for plugin " << plugin << endl);
    return;
  }

  /*Begin> Try to load the library */
  try {
    PluginStruct ps;
    /* allocate the DynamicLibrary object */
    ps.dll_object= new DynamicLibrary();
    /* open the library */
    ps.dll_object->open(folder+"/"+plugin);

    /* Begin> Search for the plugin array with name nmm_plugins and initialize the plugins */
    Plugin** tmp = static_cast<Plugin**>(ps.dll_object->resolve(__plugin_symbol));
    if (!tmp) {
      throw Exception("Could not resolve symbol " + __plugin_symbol);
    }
    ps.plugin = tmp[0];

    /* Begin> Check if the plugin was already loaded. */
    for ( map<string, PluginStruct>::iterator it = __plugin_map.begin(); it != __plugin_map.end(); ++it) {
      if ( (it->second.plugin == ps.plugin) && (ps.plugin != 0) )  {
        /* store the name as alias */
        it->second.alias.push_back(plugin);
        /* close the library again */
        ps.dll_object->close();
        /* delete the dll_object */
        delete ps.dll_object;
        return;
      }
    }
    /* End<*/
    
    if (ps.plugin) {
        ps.plugin->setPluginManager(this);
        ps.plugin->setLibraryName(plugin);
        ps.plugin->nmmInit();
      } else {
      WARNING_STREAM("Library " << plugin << " has no NMM plugins" << endl);
    }
    /*End< */
    __plugin_map.insert(pair<string, PluginStruct>(plugin, ps));
    applyDefinedObjectIDs( plugin );
  }
  /*Begin> Error handling: If an excetpion is thrown, remove the plugin */
  catch (const Exception& e) {
    removeLibrary(plugin);
    throw SerializedException(e.getComment());
  } catch (const SerializedException& e) {
    removeLibrary(plugin);
    throw SerializedException(e.getComment());
  }
  /*End< */
  DEBUG_STREAM("Plugin " << plugin << " successfully loaded" << endl);
}
#endif

void PluginManager::removeLibrary(const string& name)
  throw (const SerializedException&)
{
  map<string, PluginStruct>::iterator it = __plugin_map.find(name);

  /*Begin> Check if the plugin is loaded */
  if (it == __plugin_map.end()) {
    DEBUG_STREAM("A plugin with name " << name << " was not loaded, or was already removed" << endl);
    return;
  }
  /*End< */

  /*Begin> Call the unregister method for the Registry */
  deactivateRegistryPart(name);
  /*End< */

  /*Begin> Call the unregister method for the NodeFactory */
  deactivateFactoryPart(name);
  /*End< */

  /*Begin> Call the generic unregister method for nmm plugins */
  if (it->second.plugin) {
    it->second.plugin->nmmFini();
  } else {
    WARNING_STREAM("Library " << name << " has no NMM plugins" << endl);
  }
  /*End< */


  #ifndef ENABLE_STATIC_PLUGINS 
  /*Begin> Close the Dll Object and remove it from the map */
  try {
    if (it->second.dll_object) {
      it->second.dll_object->close(); //throws Exception
    }
    else
      GLOBAL_WARNING_STREAM("No DLL-Object available for library " << it->first << ". Maybe it causes a segfault" << endl);
  } catch (const Exception& e) {
    ERROR_STREAM(e << endl);
  }
  delete it->second.dll_object;   //delete the dll_object
  #endif

  DEBUG_STREAM("Plugin " << it->first << " successfully removed" << endl);
  __plugin_map.erase(it);
  /*End< */
}


void PluginManager::deactivateRegistryPart(const string& plugin)
    throw (const SerializedException&)
{
  map<string, PluginStruct>::iterator it = __plugin_map.find(plugin);

  /*Begin> Check if the plugin is loaded */
  if (it == __plugin_map.end()) {
    WARNING_STREAM("A plugin with name " << plugin << " was not loaded" << endl);
    return;
  }
  /*End< */

  /*Begin> Call the unregister method for the Registry */
  if (it->second.reg_active && it->second.plugin && it->second.plugin) {
    it->second.plugin->nmmRegistryFini();
    if (it->second.reg_active) {
      it->second.reg_active->unregisterPlugin(it->second.plugin);
    } else {
      WARNING_STREAM("Library " << plugin << " is not managed by a Registry!" << endl);  
    }
  } else {
    WARNING_STREAM("Library " << plugin << " has no NMM plugins" << endl);
  }
  it->second.reg_active = 0; //reset the reg_active to ensure it is called only once.
  /*End< */
}

void PluginManager::deactivateFactoryPart(const string& plugin)
    throw (const SerializedException&)
{
  map<string, PluginStruct>::iterator it = __plugin_map.find(plugin);

  /*Begin> Check if the plugin is loaded */
  if (it == __plugin_map.end()) {
    WARNING_STREAM("A plugin with name " << plugin << " was not loaded" << endl);
    return;
  }
  /*End< */

  /*Begin> Call the unregister method for the NodeFactory */
  if (it->second.fac_active && it->second.plugin && it->second.plugin) {
    it->second.plugin->nmmFactoryFini();
    if ( it->second.fac_active ) {
      it->second.fac_active->unregisterPlugin(it->second.plugin);
    }
  } else {
    WARNING_STREAM("Library " << plugin << " has no NMM plugins" << endl);
  }
  it->second.fac_active = 0; //reset the fac_active to ensure it is called once only.
  /*End< */
}

bool PluginManager::isPluginUsed(const string& plugin) const {
  /* If the plugin is not in the internal map, it is not used */
  if (__plugin_map.find(plugin) == __plugin_map.end())
    return false;
  return true;
}

bool PluginManager::isFactoryPartActive(const string& plugin)  const {
  map<string, PluginStruct>::const_iterator it = __plugin_map.find(plugin);


  if (it == __plugin_map.end()) { //plugin not loaded
    return false;
}

  if (it->second.fac_active)
    return true;

  return false;
}

bool PluginManager::isRegistryPartActive(const string& plugin)  const {
  map<string, PluginStruct>::const_iterator it = __plugin_map.find(plugin);
  if (it == __plugin_map.end()) //if plugin not load return false
    return false;

  if (it->second.reg_active)
    return true;

  return false;
}

void PluginManager::activateRegistryPart(const string& plugin, Registry& registry)
  throw (const SerializedException& )
{
  /*Begin> Check if plugin is load and not active */
  map<string, PluginStruct>::iterator it = __plugin_map.find(plugin);
  if (it == __plugin_map.end()) { //plugin not loaded
    DEBUG_STREAM("Plugin " << plugin << " not loaded" << endl);
    return;
  }

  if (it->second.reg_active) {
    DEBUG_STREAM("Plugin " << plugin << " is already activated by a Registry" << endl);
    return;
  }

  if (__defined_id_map.find(plugin) != __defined_id_map.end() ) {
    WARNING_STREAM(" plugin " << plugin << " has predefined IDs -> don't call activate Registry part" << endl);
    return;
  }
  /*End< */

  /* Store the used Registry */
  // TODO: fw: This should be cleaned up if an error occurs in nmmRegistryInit ?
  it->second.reg_active = &registry;

  if (it->second.plugin) {
    try {
      /* register the plugins with the library */
      it->second.plugin->nmmRegistryInit();
      registry.registerPlugin(it->second.plugin);
    }
    catch (const SerializedException& e) {
      ERROR_STREAM("Could not activate registry part of plugin " << plugin << endl);
      ERROR_STREAM("Reason: " << e << endl);
      throw;
    }
    catch (const Exception& e) {
      ERROR_STREAM("Could not activate registry part of plugin " << plugin << endl);
      ERROR_STREAM("Reason: " << e << endl);
      throw RegistryException(e.getComment());
    }
  }
  else {
    WARNING_STREAM("Library " << plugin << " has no NMM plugins" << endl);
  }
  DEBUG_STREAM("Registry part of plugin " << plugin << " successfully activated" << endl);
  /*End< */
}

void PluginManager::activateFactoryPart(const string& plugin, NodeFactory& factory)
  throw (const SerializedException&)
{
  /*Begin> Check if plugin is load and not activate */
  map<string, PluginStruct>::iterator it = __plugin_map.find(plugin);
  if (it == __plugin_map.end()) { //plugin not loaded
    DEBUG_STREAM("Plugin " << plugin << " not loaded" << endl);
    return;
  }

  if (it->second.fac_active) {
    DEBUG_STREAM("Plugin " << plugin << " is already activated by a NodeFactory" << endl);
    return;
  }
  /*End< */

  /* Store the used NodeFactory */
  it->second.fac_active = &factory;
  if(it->second.plugin) {
    try {
      /* register the plugins with a node factory */
      it->second.plugin->nmmFactoryInit();
      factory.registerPlugin(it->second.plugin);
    } catch (const Exception& e) {
      it->second.fac_active = 0;
      ERROR_STREAM("Could not activate factory part of plugin " << plugin << endl);
      ERROR_STREAM(e << endl);
    } catch (const SerializedException& e) {
      it->second.fac_active = 0;
      ERROR_STREAM("Could not activate factory part of plugin " << plugin << endl);
      ERROR_STREAM(e << endl);
    }
  } else {
    WARNING_STREAM("Library " << plugin << " has no NMM plugins" << endl);
  }
  /*End< */
}

list<string> PluginManager::getUsedPlugins() const {
  list<string> tmp_list;
  for (map<string, PluginStruct>::const_iterator it = __plugin_map.begin(); it != __plugin_map.end(); ++it)
  { 
    tmp_list.push_back(it->first);
  }

  return tmp_list;
}

list<string> PluginManager::getPlugins() const {
  #ifdef ENABLE_STATIC_PLUGINS
  return getStaticPlugins();
  #else
  return getDllPlugins();
  #endif
}

#ifndef ENABLE_STATIC_PLUGINS
list<string> PluginManager::getDllPlugins() const { 
  list<string> tmp_list;
  list<string> plugin_list;
  /*Begin> Find the folder for the given plugin*/
  for (list<string>::const_iterator it = __plugin_path.begin(); it != __plugin_path.end(); ++it)
  {
    try {
      tmp_list = readFolder(*it);
	  tmp_list.remove(".");
	  tmp_list.remove("..");
      for (list<string>::const_iterator it2 = tmp_list.begin(); it2 != tmp_list.end(); ++it2) {
		if ( (it2->find(".so") < it2->size()) || ( it2->find(".dll") < it2->size()) ) {
	      plugin_list.push_back(*it2);
		}
	  }
    } catch (const Exception& e) {
      ERROR_STREAM(e << endl);
    }
  }
  /*End> */
  return plugin_list;
}
#endif

#ifdef ENABLE_STATIC_PLUGINS
list<string> PluginManager::getStaticPlugins() const { 
  list<string> tmp_list;
  /*Begin> Find the folder for the given plugin*/
  Plugin** tmp_plugins = nmm_static_plugins;
  while(*tmp_plugins) {
    tmp_list.push_back((*tmp_plugins)->getLibraryName());
    ++tmp_plugins;
  }
  return tmp_list;
}
#endif

list<string> PluginManager::readFolder(const string&  folder) const
  throw (const SerializedException& )
{
  try {
    return NMM::getAllFilesInDirectory(folder);
  } catch(const Exception& e) {
    throw SerializedException(e.getComment());
  }
}


Plugin* PluginManager::requestPlugin(const string& plugin) {
  /* search the library */
  InternalIt it = __plugin_map.find(plugin);
  if (it == __plugin_map.end()) {
    for ( map<string, PluginStruct>::iterator it = __plugin_map.begin(); it != __plugin_map.end(); ++it) {
      if ( find(it->second.alias.begin(), it->second.alias.end(), plugin ) != it->second.alias.end() )  {
        return it->second.plugin;
      }
    }
    return 0;
  }

  return  it->second.plugin;
}

Plugin* const  PluginManager::requestPlugin(const string& plugin) const {
  /* search the library */
  CInternalIt it = __plugin_map.find(plugin);
  if (it == __plugin_map.end()) {
    for ( map<string, PluginStruct>::const_iterator it = __plugin_map.begin(); it != __plugin_map.end(); ++it) {
	  if ( std::find(it->second.alias.begin(), it->second.alias.end(), plugin ) != it->second.alias.end() )  {
        return it->second.plugin;
      }
    }
    return 0;
  }

  return  it->second.plugin;
}

list<ObjectID> PluginManager::getPluginID(const string& plugin, const nmm_uint32& index ){
  /* Begin> If predefined IDs are available, use them */
  map<string, vector<list<ObjectID> > >::iterator i_it = __defined_id_map.find( plugin );
  if ( i_it != __defined_id_map.end() ) {
    if ( index < i_it->second.size() ) {
      return i_it->second[index];
    }
  }
  
  /* allocate the return value */
  list<ObjectID> node_id;

  /* Begin> If the Plugin is not load, return an empty list */
  InternalIt it = __plugin_map.find(plugin);
  if (it == __plugin_map.end()) {        
    return node_id;
  }
  /* End< */

  /* Begin> If no Plugin-Array is available, return an empty list */
  if(!it->second.plugin)
    return node_id;
  /* End< */

  /* insert the ObjectID's of a */
  for (unsigned int i = 0; i< it->second.plugin->getSize(); ++i)
    node_id.push_back(it->second.plugin->getObjectID(i));

  return node_id;      
}


list<string> PluginManager::getPluginPath()      const {
  return __plugin_path;
}

void PluginManager::activateRegistryPart(const string& plugin)
  throw (const SerializedException&)
{
  activateRegistryPart(plugin, SLocalRegistry::getInstance());
}

void PluginManager::activateFactoryPart(const string& plugin)
  throw (const SerializedException&)
{
  activateFactoryPart(plugin, SLocalNodeFactory::getInstance());
}

void PluginManager::defineObjectIDs(const string& plugin, const vector<list<ObjectID> >& id_list) {
  __defined_id_map.insert(pair<string, vector<list<ObjectID> > >(plugin, id_list));
  applyDefinedObjectIDs(plugin);
}

void PluginManager::applyDefinedObjectIDs(const string& plugin) {
  InternalIt p_it = __plugin_map.find(plugin);
  if ( p_it == __plugin_map.end() ) {
    return;
  }

  if ( !p_it->second.plugin ) {
    return;
  }
 
  /* Check if predefined IDs are available and set them */
  map<string, vector<list<ObjectID> > >::iterator it = __defined_id_map.find(plugin);

  if (it != __defined_id_map.end()) {
    for ( unsigned int i = 0; i< it->second.size(); ++i) {
      if (p_it->second.plugin == 0) {
        break;
      }
      for (list<ObjectID>::iterator it2 = it->second[i].begin(); it2 != it->second[i].end(); ++it2) {
        p_it->second.plugin->addObjectID(*it2);
      }
    }
  }
}

const char* PluginManager::static_library_name = "static_library";
const string PluginManager::__plugin_symbol = "nmm_plugins";

}//end namespace

DEFINE_SINGLETON(NMM::PluginManager);

#include "nmm/nmm_suffix.hpp"
