/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include "Registry.hpp"
#include "nmm/comm/ReferenceManager.hpp"
#include "PluginManager.hpp"

#include <iostream>
#include <algorithm>

#include "nmm/nmm_prefix.hpp"

namespace NMM {

Registry::Registry(const char* name, Registry* parent)
  : NamedObject(name), __parent(parent), __registry_id(name, getReference()), __transaction_sender(0),
    __transaction_condition(__transaction_mutex)
{
  if (parent)
    /* Add this registry as child */
    parent->addRegistry(*this);
}


Registry::~Registry() {
  if (getParent())
    /* Remove the child from parent */
    getParent()->removeRegistry(*this);

  /* Begin> Delete the interfaces to children */
  for (list<IRegistry*>::iterator it = __i_child_list.begin(); it != __i_child_list.end(); ++it) {
    delete *it;
  }
  __i_child_list.clear();
  /* End< */

  /*Begin> Reset all registered children to 0*/
  list<Registry*> child_list = getChildList();

  /*Begin> set the parent of all children to NULL */
  for (list<Registry*>::iterator it = child_list.begin(); it != child_list.end(); ++it)
    (*it)->setParent(0);
  /*End< */

  SPluginManager::getInstance().removeRegistry(this);
}

void Registry::removeRegistry(Registry& reg) {
  DEBUG_STREAM("Remove Registry" << reg.getName() << endl);
  __child_list.remove(&reg);
}

void Registry::setParent(Registry* parent) {
  /* If an old parent exist, tell him that this Registry is no longer its child */
  if (getParent())
    getParent()->removeRegistry(*this);

  /* set the new parent */
  __parent = parent;
}

IRegistry* Registry::findRegistry(const RegistryID& registry_id) {
  /*Begin> If this is the searched registry, return it*/
  if (__registry_id == registry_id)
    return getOwnInterface<IRegistry>();
  /*End< */

  /*Begin> Search in the child list for the given registry_id */
  IRegistry* tmp_registry = 0;
  for (list<Registry*>::iterator it = __child_list.begin(); it != __child_list.end(); ++it)
  {
    tmp_registry = (*it)->findRegistry(registry_id);
    if (tmp_registry)
      return tmp_registry;
  }
  /*End< */
  return 0;
}


const IRegistry* Registry::findRegistry(const RegistryID& registry_id) const {
  /*Begin> If this is the searched registry, return it*/
  if (__registry_id == registry_id)
    return getOwnInterface<IRegistry>();
  /*End< */

  /*Begin> Search in the child list for the given registry_id */
  const IRegistry* tmp_registry = 0;
  for (list<Registry*>::const_iterator it = __child_list.begin(); it != __child_list.end(); ++it)
  {
    tmp_registry = (*it)->findRegistry(registry_id);
    if (tmp_registry)
      return tmp_registry;
  }
  /*End< */
  return 0;
}

void Registry::addRegistry(Registry& reg) {
  if (this == &reg)
    return;
  if (std::find(__child_list.begin(), __child_list.end(), &reg) == __child_list.end()) {
    /* Store a Pointer to the reference */
    __child_list.push_back(&reg);
    /* Store the Interface to the Registry */
    __i_child_list.push_back(reg.getInterface<IRegistry>());
  } else {
    WARNING_STREAM("Registry " << reg.getName() << " is already child of this registry" << endl);
  }
}

Result Registry::receiveEvent(Event& e, const IEventSender* sender) {
  /* Wait until access control is disabled */
  waitForTransaction(sender);
  /* Set the current sender */
  setCurrentSender(const_cast<IEventSender*>(sender));

  /* Dispatch the event */
  return getEventDispatcher().dispatch(e, sender);
}

void Registry::propagateRegistry() {
  /* Begin> Inform all Registries about the new Registry */
  Registry* master = this;
  RegistryID reg = getRegistryID();
  announceRegistry( reg );
  while (master->getParent() != 0 ) {
    master = master->getParent();
    master->announceRegistry( reg );
  }
  /* End< */
}



/********Event methods*********************/
void Registry::initTransaction(const UserInfo& u_info)
  throw(const RegistryException&)
{
  if (!getCurrentSender()) 
    throw RegistryException("Could not init transaction. No Sender available ");

  /* Begin> set the TransactionSender */
  setTransactionSender(getCurrentSender());

  __locked_user = u_info;

  DEBUG_STREAM(">Begin Start a Transaction: " << endl);
  DEBUG_STREAM("User : " << u_info.getUsername() << endl);
  DEBUG_STREAM("Host : " << u_info.getHostname() << endl);
  DEBUG_STREAM("Pid  : " << u_info.getPid() << endl);
}


void Registry::closeTransaction(const UserInfo& u_info) {
  if (__locked_user != u_info) {
    throw RegistryException("Could not close transaction. Invalid User");
  }

  setTransactionSender(0);
  /* Begin> Notify waiting threads */
  __transaction_mutex.lock();
  __transaction_condition.notify();
  __transaction_mutex.unlock();
  /* End<*/

  DEBUG_STREAM(">Close a Transaction: " << endl);
  DEBUG_STREAM("User : " << u_info.getUsername() << endl);
  DEBUG_STREAM("Host : " << u_info.getHostname() << endl);
  DEBUG_STREAM("Pid  : " << u_info.getPid() << endl);
}


list<RegistryID> Registry::getChildList(const RegistryID& reg_id) const {
  list<RegistryID> reg_list;

  /* Begin> Froward the method call if this RegistryID is != given reg_id*/
  if (getOwnRegistryID() != reg_id) {
    const IRegistry* registry = findRegistry(reg_id);
    if (registry)
      return registry->getChildList(reg_id);
    else
      throw RegistryException("getChildList: Unknown Registry" );
  }
  /* End< */

  /* Store the RegistryID's in the reg_list */
  for (list<Registry*>::const_iterator it = getChildList().begin(); it != getChildList().end(); ++it)
    reg_list.push_back((*it)->getRegistryID());

  /* Return the list with all RegistryID's */
  return reg_list;
}



void Registry::releaseNode(const RegistryID& reg_id, const NMMReference& ref, const ObjectID& node_id, const UserInfo& user_info) {
  /* Begin> Froward the method call if this RegistryID is != given reg_id*/
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->releaseNode(reg_id, ref, node_id, user_info);
    else
      return;
  /* End< */

  NMMObject* object = SReferenceManager::getInstance().getNMMObject(ref);
  if (!object) {
    ERROR_STREAM("No object with reference " << ref << " available to release " << endl);
    return;
  } 

  /* 1. Destroy the node */
  destroyNode(reg_id, ref, user_info);
  /* 2. Release the reservation */
  releaseReservation(reg_id, node_id, user_info);
}

void Registry::releaseObject(const RegistryID& , const NMMReference& , const ObjectID& , const UserInfo& )
{
  cerr << "NOT IMPLEMENTED YET" << endl;
}

nmm_bool Registry::isSubRegistry(const RegistryID& reg_id) const {
  /* If it is the RegistryID of this Registry return true */
  if (getOwnRegistryID() == reg_id)
    return true;

  /* If it is the RegistryID of a child Registry return true */
  for (list<IRegistry*>::const_iterator it = __i_child_list.begin(); it != __i_child_list.end(); ++it)
    if ((*it)->isSubRegistry(reg_id))
      return true;

  /* Otherwise return false */
  return false;
}

INode* Registry::requestNode(const RegistryID& reg_id, const ObjectID& node_id, const UserInfo& info, NodeDescription& nd)
    throw(const RegistryException&)
{
  /* Begin> Froward the method call if this RegistryID is != given reg_id*/
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->requestNode(reg_id, node_id, info, nd);
    else {
      throw RegistryException("requestNode: Unknown Registry");
    }
  /* End< */
  
  /* reserve the Node and use the returned FactoryTicket to instantiate the node. */
  INode* inode = instantiateNode(reg_id, reserveNode(reg_id, node_id, info), info);
  updateNodeDescription(nd, getNodeDescription(reg_id, node_id) );

   return inode;
}

Interface* Registry::requestObject(const RegistryID& , const ObjectID& , const UserInfo& ) 
  throw(const RegistryException&) 
{
  cerr << "NOT IMPLEMENTED YET" << endl;
  return 0;
}

list<Response >  Registry::initRequest(const NodeDescription& node_desc) const {
  list<Response> resp_list;

  /* Search this Registry */
  list<ObjectID> node_list = searchNode(__registry_id, node_desc);
  if (!node_list.empty())
    /* store the result in the resp_list */
    resp_list.push_back(Response(__registry_id, node_list));

  /* Begin> Forward the request to all children */
  for (list<IRegistry*>::const_iterator it = __i_child_list.begin(); it != __i_child_list.end(); ++it)
  {
    list<Response> tmp_list = (*it)->initRequest(node_desc);
    if (!tmp_list.empty()) {
      resp_list.insert(resp_list.end(), tmp_list.begin(), tmp_list.end());
    }
  }
  /* End< */

  return resp_list;
}

list<Response > Registry::initObjectRequest(const ObjectDescription& ) const
{ 
  cerr << "NOT IMPLEMENTED YET" << endl;
  list<Response> resp_list;
  return resp_list;
}

void Registry::waitForTransaction(const IEventSender* sender) const {
  int wait_counter = 0;

  /* Begin> If the __transaction_sender is != 0 and another sender enables transaction_lock
   * we must wait.
   */
  while (wait_counter < 10 && (__transaction_sender && __transaction_sender != sender)) {
    DEBUG_STREAM2("Registry is locked" << endl);
    /* wait for one second or the notify() call*/
    __transaction_mutex.lock();
    __transaction_condition.relativeTimedWait(1, 0);
    __transaction_mutex.unlock();
    /* End< */
    ++wait_counter;
  }

  /* The Transaction already exist after 10 retries and we throw an Excetption */
  if (__transaction_sender && __transaction_sender != sender) {
    throw RegistryException("A maximum timout occurs in the Registry. The access control is enabled by another process");
  }
}

void Registry::updateNodeDescription(NodeDescription& nd_app, const NodeDescription& nd_reg) {
  /* copy the values from the ObjectDescritption */
  nd_app.setObjectName(nd_reg.getObjectName());
  nd_app.setLibraryName( nd_reg.getLibraryName () );
  nd_app.setInterfaceList( nd_reg.getInterfaceList() );
  nd_app.setServiceDescription( nd_reg.getServiceDescription() );
  nd_app.setLocation( nd_reg.getLocation() );

  /* copy the values from the NodeDescription */
  nd_app.setNodeType(nd_reg.getNodeType());
  nd_app.setInputProperty(nd_reg.getInputProperty());
  nd_app.setOutputProperty(nd_reg.getOutputProperty());
}

}//end namespace




#include "nmm/nmm_suffix.hpp"
