/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "Registry1394.hpp"
#include "nmm/plugins/ieee1394/GenericFireWireNode.hpp"
#include "nmm/plugins/ieee1394/Camera1394.hpp"
#include "nmm/services/registry/LocalRegistry.hpp"
#include "nmm/services/registry/LocalNodeFactory.hpp"
#include "nmm/multimedia/NodeDescription.hpp"
#include "nmm/comm/ReferenceManager.hpp"
#include "nmm/utils/thread/MutexGuard.hpp"
#include <l1394/l1394_session.h>
#include <iostream>
#include <stdlib.h>


#include "nmm/nmm_prefix.hpp"

namespace NMM{

Registry1394::Registry1394( Registry* parent)
  : Registry("Registry1394", parent), __internal_mutex(ThreadMutex::RECURSIVE), session(0)
{
  session = L1394::SSession::getInstance();
  if (session) {
    scanNodes();
  }
  session->addEventHandle(this);
  propagateRegistry();
}

Registry1394::~Registry1394()
{
  session = L1394::SSession::getInstance();
  if (session)
    session->removeEventHandle(this);
}


void Registry1394::erase() {

}

FactoryTicket Registry1394::reserveNode(const RegistryID& reg_id, const ObjectID& node_id, const UserInfo& info)
   throw(const RegistryException&)
{
  /* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->reserveNode(reg_id, node_id, info);
    else
      throw RegistryException("Unknown Registry");
  /* End< */

  MutexGuard m(__internal_mutex);
  /*Begin> Check if node is currently in use */
  if (__used_map.find(node_id) != __used_map.end())
    throw RegistryException("Node is not available");
  /*End< */

  /*Begin> Check if the node is available*/
  map<ObjectID, InternalDeviceStruct*>::iterator it = __device_map.find(node_id);
  if (it == __device_map.end()) {
    throw RegistryException("Unknown Node");
  }
  /*End< */

  /*Begin> reserve the node*/
  InternalDeviceStruct* tmp_struct = it->second;
  tmp_struct->user_info = info;
  __used_map.insert(pair<ObjectID, InternalDeviceStruct*>(it->first, it->second));
  __device_map.erase(it);

  /*Begin> Create a ticket to create the node*/
  FactoryTicket fp;
  fp.setNodeFactoryID(NodeFactoryID("NodeFactory1394"));
  fp.setObjectID(node_id);
  setTicketValid(fp);
  /*End<*/

  return fp;
  /*End <*/
}


void Registry1394::releaseReservation(const RegistryID& reg_id, const ObjectID& node_id, const UserInfo& user_info)
{
  /* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->releaseReservation(reg_id, node_id, user_info);
    else
      throw RegistryException("Unknown Registry");
  /* End< */


  /*Begin> Check if the node was reserved */
  map<ObjectID, InternalDeviceStruct*>::iterator it = __used_map.find(node_id);
  if (__used_map.find(node_id) == __used_map.end()) {
    ERROR_STREAM("The node with " << node_id << " was not reserved" << endl);
    return;
  }
  /*End< */

  /*Begin> Remove the node from __used_map and store it in __device_map*/
  __device_map.insert(pair<ObjectID, InternalDeviceStruct*>(it->first, it->second));
  __used_map.erase(it);
  /*End< */
}


void Registry1394::scanNodes() {
  /* Begin> Cleanup the __device_map */
  for (InternalIt it = __device_map.begin(); it != __device_map.end(); ++it)
    delete it->second;

  __device_map.clear();
  /* End< */


  typedef list<L1394::Device*>::iterator DeviceIt;

  list<L1394::Device*> device_list(session->getAllDevices());

  InternalDeviceStruct *tmp_struct = 0;
  /* Begin> Create an entry for each device */
  for (DeviceIt it = device_list.begin(); it != device_list.end(); ++it) {
    DEBUG_STREAM("Scan Node " << hex << (*it)->getParentNode()->getGuid() << dec << endl);
    /* Create the struct */
    tmp_struct = new InternalDeviceStruct();

    /* Use the 1394-GUID as ObjectID */
    ObjectID tmp_id;
    tmp_id = (*it)->getParentNode()->getGuid();

    __device_map.insert(pair<ObjectID, InternalDeviceStruct*>(tmp_id, tmp_struct));
    map<ObjectID, InternalDeviceStruct*>::iterator tmp_it = __device_map.find(tmp_id);
    if (tmp_it == __device_map.end() ) {
      ERROR_STREAM("Internal Error in scanNode" << endl);
      continue;
    }
    
    createNodeProperty(tmp_it->second->node_property, **it);
  }
  /* End< */
}

void Registry1394::createNodeProperty(NodeDescription& np, L1394::Device& device) {
    /* set LibraryName */
    np.setLibraryName("libnmmieee1394.so");
    /* set the base class as name (it is overwritten for supported devices, like cameras, in the
     * switch statement).
     */
    np.setNodeName("GenericFireWireNode");

    GenericFireWireNode::initOutputFormat(np.getOutputProperty(), device);
    switch(device.getDeviceType()) {
      case L1394::_Camera: np.addInterface("ICamera"); np.setNodeName("Camera1394"); np.setNodeType(INode::SOURCE); break;
      case L1394::_Vcr   : break;
      case L1394::_VideoMonitor: break;
      case L1394::_Disc  : break;
      case L1394::_Tuner : break;
    }
}

list<ObjectID >  Registry1394::getNodeList(const RegistryID& reg_id) const
{
  /* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->getNodeList(reg_id);
    else
      throw RegistryException("Unknown Registry");
  /* End< */

  list<ObjectID> node_list;
  for (CInternalIt it = __device_map.begin(); it != __device_map.end(); ++it)
    node_list.push_back(it->first);

  return node_list;
}

list<ObjectID >  Registry1394::getReservedNodeList(const RegistryID& reg_id) const {
  /* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID()) {
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->getReservedNodeList(reg_id);
    else {
      throw RegistryException("Unknown Registry");
    }
  }
  /* End< */

  list<ObjectID> node_list;
  for (CInternalIt it = __used_map.begin(); it != __used_map.end(); ++it)
    node_list.push_back(it->first);

  return node_list;
}

NodeDescription Registry1394::getNodeDescription(const RegistryID& reg_id, const ObjectID& node_id) const {
  /* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->getNodeDescription(reg_id, node_id);
    else
      throw RegistryException("Unknown Registry");
  /* End< */

  CInternalIt it = __device_map.find(node_id);
  if (it != __device_map.end()) {
    return it->second->node_property;
  }

  it = __used_map.find(node_id);

  if (it != __used_map.end()) {
    return it->second->node_property;
  }

  throw RegistryException("Unknown Node");
}

nmm_int32 Registry1394::getMaxInstanceCount(const RegistryID& reg_id, const ObjectID& node_id) const{
  /* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->getMaxInstanceCount(reg_id, node_id);
    else
      throw RegistryException("Unknown Registry");

  return 1;
}

nmm_int32 Registry1394::getInstanceCount(const RegistryID& reg_id, const ObjectID& node_id) const {
  /* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->getInstanceCount(reg_id, node_id);
    else
      return 0;
  /* End< */

  /* Search the NodeDescription for given ObjectID */
  CInternalIt it = __device_map.find(node_id);
  if (it != __device_map.end()) {
   return 0;
  }

  it = __used_map.find(node_id);
  if (it != __device_map.end()) {
    return 1;
  }
  return 0;
}

list<ObjectID > Registry1394::searchNode(const RegistryID& reg_id, const NodeDescription& node_desc) const
{
  /* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->searchNode(reg_id, node_desc);
    else
      throw RegistryException("Unknown Registry");
  /* End< */

  list<ObjectID> node_list;

  for (CInternalIt it = __device_map.begin(); it != __device_map.end(); ++it) {
    if (it->second->node_property.isSubset(node_desc))
      node_list.push_back(it->first);
  }
  return node_list;
}

string Registry1394::getPluginInfo(const RegistryID& reg_id, const ObjectID& node_id) const{
  /* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->getPluginInfo(reg_id, node_id);
    else
      throw RegistryException("Unknown Registry");
  /* End< */

  CInternalIt it = __device_map.find(node_id);
  if (it == __device_map.end()) {
    it = __used_map.find(node_id);
    if (it == __used_map.end())
      return string();
  }

  return it->second->node_property.getLibraryName();
}


INode* Registry1394::instantiateNode(const RegistryID& reg_id, const FactoryTicket& nf_id, const UserInfo& user_info)
    throw (const RegistryException&)
{
  /* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->instantiateNode(reg_id, nf_id, user_info);
    else
      throw RegistryException("Unknown Registry");
  /* End< */

  try {
    /* Get the NodeFactory */
    NodeFactory* factory = SLocalNodeFactory::getInstance().findNodeFactory(nf_id.getNodeFactoryID());
    if (!factory)
      throw RegistryException("Unknown Registry");

     /* Create the node */
     Node& node = factory->createNode(nf_id);

     /* Store its reference in the allocated map */
     __allocated_nodes.insert(pair<NMMReference, UserInfo>(node.getReference(), user_info));

     /* return its interface */
     return node.getInterface<INode>();
   } catch (const Exception& e) {
     ERROR_STREAM( e << endl) ;
     throw RegistryException(e.getComment());
   }
}

void Registry1394::destroyNode(const RegistryID& reg_id, const NMMReference& ref, const UserInfo& user_info)
{
  /* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->destroyNode(reg_id, ref, user_info);
    else
      throw RegistryException("Unknown Registry");
  /* End< */

  /* Get the Node pointer */
  Node* node = SReferenceManager::getInstance().getObject<Node>(ref);

  if (!node)
    return;

  /* Get the NodeFactory */
  NodeFactory* factory = SLocalNodeFactory::getInstance().findNodeFactory(NodeFactoryID("NodeFactory1394"));
  if (!factory)
    return;

  /* Begin> Check if UserInfo and NMMReference are valid */
  map<NMMReference, UserInfo>::iterator it = __allocated_nodes.find(ref);
  if (it == __allocated_nodes.end()) {
    ERROR_STREAM("Node is not created by this registry" << endl);
    return;
  }

  if (it->second != user_info) {
    ERROR_STREAM("Node is not reserved by the user " << endl);
    return;
  }
  /* End< */

  /* remove from the allocated map */
  __allocated_nodes.erase(it);

  /* Call the Factory to release the node */
  factory->releaseNode(*node);
}

NodeFactoryID Registry1394::getNodeFactoryID(const RegistryID& reg_id, const ObjectID& node_id) const
{
  /* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->getNodeFactoryID(reg_id, node_id);
    else
      throw RegistryException("Unknown Registry");
  /* End< */

  NodeFactory* factory = SLocalNodeFactory::getInstance().findNodeFactory(NodeFactoryID("NodeFactory1394"));
  if (!factory)
    return NodeFactoryID();

  return factory->getFactoryID();
}

void Registry1394::resetInternalReservation() {
  /* Begin> first release the allocated nodes */
  /* Create a temporary copy to iterate through. So changes on the original internal structs can be performed */
  map<NMMReference, UserInfo> t_allocated_nodes = __allocated_nodes;
  for (map<NMMReference, UserInfo>::iterator it = t_allocated_nodes.begin(); it != t_allocated_nodes.end(); ++it)
  {
    /* Release the node, wich calls releaseReservation and destroyNode */
    destroyNode(getRegistryID(), it->first, it->second);
  }
  /* End< */

  /* Begin> release the nodes which were only reserved */
  /* Create a temporary copy to iterate through. So changes on the original internal structs can be performed */
  map<ObjectID, InternalDeviceStruct*> t_reserved_nodes = __used_map;
  for (map<ObjectID, InternalDeviceStruct*>::iterator it = t_reserved_nodes.begin(); it != t_reserved_nodes.end(); ++it)
  {
    /* The Registry is only responsible for the reservation not for destruction. So call
     * only releaseReservation
     */
    releaseReservation(getRegistryID(), it->first, it->second->user_info);
  }
  /* End */

  /* Forward the method call to all children */
  for (list<IRegistry*>::iterator it = getIChildList().begin(); it != getIChildList().end(); ++it)
    (*it)->resetInternalReservation();
}

void Registry1394::resetSpecificInternalReservation(const UserInfo& user_info) {
  /* Begin> first release the allocated nodes */
  /* Create a temporary copy to iterate through. So changes on the original internal structs can be performed */
  map<NMMReference, UserInfo> t_allocated_nodes = __allocated_nodes;
  for (map<NMMReference, UserInfo>::iterator it = t_allocated_nodes.begin(); it != t_allocated_nodes.end(); ++it)
  {
    /* If the UserInfo's name match */
    if (it->second.getUsername() == user_info.getUsername())
      /* Release the node, wich calls releaseReservation and destroyNode */
      destroyNode(getRegistryID(), it->first, it->second);
  }
  /* End< */

  /* Begin> release the nodes which were only reserved */
  /* Create a temporary copy to iterate through. So changes on the original internal structs can be performed */
  map<ObjectID, InternalDeviceStruct*> t_reserved_nodes = __used_map;
  for (map<ObjectID, InternalDeviceStruct*>::iterator it = t_reserved_nodes.begin(); it != t_reserved_nodes.end(); ++it)
  {
    /* If the UserInfo's match */
    if (it->second->user_info.getUsername() == user_info.getUsername())
      /* The Registry is only responsible for the reservation not for destruction. So call
       * only releaseReservation
       */
      releaseReservation(getRegistryID(), it->first, it->second->user_info);
  }
  /* End */

  /* Forward the method call to all children */
  for (list<IRegistry*>::iterator it = getIChildList().begin(); it != getIChildList().end(); ++it)
    (*it)->resetSpecificInternalReservation(user_info);
}

list<NodeInfo> Registry1394::getNodeInfoList(const RegistryID& reg_id) const {
  /* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->getNodeInfoList(reg_id);
    else
      throw Exception("Unknown RegistryID");
  /* End< */

  list<NodeInfo> node_info_list;

  /* Begin> Get the NodeInfo of the free devices */
  for (CInternalIt it = __device_map.begin(); it != __device_map.end();  ++it) {
    node_info_list.push_back( NodeInfo(it->first, it->second->node_property.getLibraryName(),
        it->second->node_property.getNodeName(), 0, 1,
        it->second->node_property.getNodeType())
    );
  }
  /* End< */

  /* Begin> Get the NodeInfo of the used devices */
  for (CInternalIt it = __used_map.begin(); it != __used_map.end();  ++it) {
    node_info_list.push_back( NodeInfo(it->first, it->second->node_property.getNodeName(),
        it->second->node_property.getLibraryName(),1, 1,
        it->second->node_property.getNodeType())
    );
  }
  /* End< */
  return node_info_list;
}

void Registry1394::busreset(const L1394::Card*){
  /* save access */
  MutexGuard mg(__internal_mutex);

  /* Scan the new nodes */
  scanNodes();

  /* Begin> remove the used nodes from the new __device_map */
  for (map<ObjectID, InternalDeviceStruct*>::iterator it = __used_map.begin(); it != __used_map.end(); ++it)
  {
    map<ObjectID, InternalDeviceStruct*>::iterator tmp_it = __device_map.find(it->first);
    if (tmp_it != __device_map.end()) {
      /* delete the internal struct */
      delete tmp_it->second;
      /* remove it from the map */
      __device_map.erase(tmp_it);
    } else {
      ERROR_STREAM ("A used device was removed, this currently not supported" << endl);
    }
  }
  /* End< */
}

list<NodeDescription> Registry1394::getNodeDescriptionList(const RegistryID& reg_id)  {
/* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->getNodeDescriptionList(reg_id);
    else
      /* The Registry is unkown and we throw an Exception */
      throw Exception("Unknwon RegistryID");
  /* End< */

  list<NodeDescription>  result;
  for (CInternalIt it = __device_map.begin(); it != __device_map.end();  ++it) 
    result.push_back(it->second->node_property);

  return result;
}

RegistryMemento Registry1394::getMemento() {
  RegistryMemento rm;
  rm.setRegistryID(getRegistryID());
  rm.setPluginName(__plugin_name);

  /* Begin> add all unused devices to the memento object */
  for ( map<ObjectID, InternalDeviceStruct*>::iterator it = __device_map.begin(); it != __device_map.end(); ++it)
  {
    MementoStruct ms;
    ms.setNodeDescription(it->second->node_property);
    ms.setObjectID( it->first );
    ms.setMaxInstanceCount( 1 );
    rm.addMementoStruct( ms );
  }
  /* End< */

  /* Begin> add all used devices to the memento object */
  for ( map<ObjectID, InternalDeviceStruct*>::iterator it = __used_map.begin(); it != __used_map.end(); ++it)
  {
    MementoStruct ms;
    ms.setNodeDescription(it->second->node_property);
    ms.setObjectID( it->first );
    ms.setMaxInstanceCount( 1 );
    rm.addMementoStruct( ms );
  }
  /* End< */
  return rm;
}

void Registry1394::setMemento(const RegistryMemento& rm){
    DEBUG_STREAM("setMemento called with " << rm.getRegistryID().getRegistryName() << ", " << rm.getRegistryID().getReference() << std::endl)
	RegistryID reg_id (rm.getRegistryID().getRegistryName(), rm.getRegistryID().getReference());
    setRegistryID( reg_id );
}

const char* Registry1394::__plugin_name = "libnmmieee1394.so";
} //end namespace


#include "nmm/nmm_suffix.hpp"
