/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#ifndef NMM_REGISTRY1394_HPP
#define NMM_REGISTRY1394_HPP

#include "nmm/nmm_prefix.hpp"


#include "nmm/services/registry/Registry.hpp"
#include "nmm/utils/thread/Thread.hpp"
#include "nmm/utils/thread/ThreadMutex.hpp"
#include "nmm/services/registry/NodeFactory.hpp"
#include "nmm/utils/NamedObject.hpp"
#include "nmm/utils/Singleton.hpp"
#include <l1394/l1394_eventhandle.h>

namespace L1394 {
  class Session;
  class Device;
}

namespace NMM{
class GenericFireWireNode;
class Camera1394;
class RegistryException;
class Property;
class NMMApplication;

class Registry1394 : public Registry, public L1394::EventHandle {
public:

/** Constructor
  */
  Registry1394(Registry* parent);

/** Destructor
  */
  virtual ~Registry1394();

protected:
/** @name Node Reservation
  * These methods can be used to search and reserve Nodes.
  */
//@{
/** This method searches for a specific node that matches the given NodeDescription. The NodeDescription
  * is created by an application and stores the information, like formats and interfaces, that must be
  * supported by the node.
  * \param registry_id The RegistryID of the Registry that should be searched.
  * \param node_desc The NodeDescription that must be supported by the nodes
  * \return A list of ObjectID's from nodes that support the requirements in the NodeDescription
  */
  virtual list<ObjectID >  searchNode(const RegistryID& registry_id, const NodeDescription& node_desc) const;

/** This method reserves a node which is managed by a Registry. If the reservation is succesfull,
  * the Registry returns a FactoryTickt that allows the caller to instantiate a Node by a NodeFactory.
  * \param reg_id The RegistryID of the Registry that should be searched.
  * \param node_id The ObjectID of the Node which is used to identify the Node inside the Registry
  * \param user_info The UserInfo identifies the user that requests the Node. If the node is released
  *        this parameter must be specified again, whereby no other user can release the reserved node.
  */
  virtual FactoryTicket reserveNode(const RegistryID& reg_id, const ObjectID& node_id, const UserInfo& user_info)
    throw(const RegistryException&);

/** This method creates an instance of a reserved node. If the node can be created its INode interface is
  * returned. Otherwise an exception is thrown.
  * \param reg_id The RegistryID of the Registry that should be searched.
  * \param ticket The FactoryTicket must be send to allow the Registry to instantiate the node
  * \param user_info The UserInfo identifies the user that requests the Node. If the node is released
  *        this parameter must be specified again, whereby no other user can release the reserved node.
  * \sa destroyNode
  */
  virtual INode* instantiateNode(const RegistryID& reg_id, const FactoryTicket& ticket, const UserInfo&)
    throw (const RegistryException&);

/** This method releases a reserved node. The node is not destroyed, only th reservation is released.
  * \param reg_id The RegistryID of the Registry that should be searched.
  * \param node_id The ObjectID of the reserved node
  * \param user_info The UserInfo identifies the user that requests the Node. If the node is released
  *        this parameter must be specified again, whereby no other user can release the reserved node.
  * \sa reserveNode
  */
  virtual void releaseReservation(const RegistryID& reg_id, const ObjectID& node_id, const UserInfo& user_info);

/** This method destroys a created node
  * \param reg_id The RegistryID of the Registry that should be searched.
  * \param reference The NMMReference of the instantiated node.
  * \param user_info The UserInfo identifies the user that requests the Node. If the node is released
  *        this parameter must be specified again, whereby no other user can release the reserved node.
  * \sa instantiateNode
  */
  virtual void destroyNode(const RegistryID& reg_id, const NMMReference& reference, const UserInfo&);
//@}


/** @name Node information
  * These methods can be used to access additional node information
  */
//@{
  RegistryMemento getMemento();
  void setMemento(const RegistryMemento& rm);
/** This method returns the maximum instance count for a specific node.
  * \param reg_id The RegistryID of the Registry that should be searched.
  * \param node_id The ObjectID of the node.
  */
  virtual nmm_int32 getMaxInstanceCount(const RegistryID& reg_id, const ObjectID& node_id) const;

/** This method returns the current instance count for a specific node.
  * \param reg_id The RegistryID of the Registry that should be searched.
  * \param node_id The ObjectID of the node.
  */
  virtual nmm_int32 getInstanceCount(const RegistryID& reg_id, const ObjectID& node_id) const;

/** This method returns the library that must be load for a specific node.
  * \param reg_id The RegistryID of the Registry that should be searched.
  * \param node_id The ObjectID of the node.
  */
  virtual string getPluginInfo(const RegistryID& reg_id, const ObjectID& node_id) const;

/** This method returns all ObjectIDs of nodes that are available.
  * \param reg_id The RegistryID of the Registry that should be searched.
  * \param node_id The ObjectID of the node.
  */
  virtual list<ObjectID >  getNodeList(const RegistryID& reg_id) const;

/** This method returns all ObjectIDs of nodes that are currently reserved and not available.
  * \param reg_id The RegistryID of the Registry that should be searched.
  */
  virtual list<ObjectID >  getReservedNodeList(const RegistryID& reg_id) const;

  virtual void registerPlugin(NMM::Plugin*) { } // Nothing to do

  virtual void unregisterPlugin(const NMM::Plugin*) { } // Nothing to do

/** This method returns a short NodeInfo for all nodes that are managed in a specific registry.
  * \param reg_id The RegistryID of the Registry that should be searched.
  */
  virtual list<NodeInfo> getNodeInfoList(const RegistryID& reg_id) const;

  virtual list<NodeDescription> getNodeDescriptionList(const RegistryID& reg_id); 
  
/** This method returns the complete NodeDescription of aspecifc node.
  * \param reg_id The RegistryID of the Registry that should be searched.
  * \param node_id The ObjectID of the node.
  */
  virtual NodeDescription getNodeDescription(const RegistryID& reg_id, const ObjectID& node_id) const;

/** This method returns the NodeFactoryID of the NodeFactory that can instantiate a specific node.
  * \param reg_id The RegistryID of the Registry that should be searched.
  * \param node_id The ObjectID of the node.
  */
  virtual NodeFactoryID getNodeFactoryID(const RegistryID& reg_id, const ObjectID& node_id) const;
//@}

/** @name Reset Registry
  * These methods can be used to reset a registry
  */
//@{
/** This method releases ALL Nodes.
 */
  virtual void resetInternalReservation();

/** This method releases ALL Nodes reserved by specifc user.
 */
  virtual void resetSpecificInternalReservation(const UserInfo& user_info);
//@}
protected:
/** @name L1394 Eventhandling
  */
//@{
/** This method is called if the configration of the 1394 bus changes.
  */
  virtual void busreset(const L1394::Card*);
//@}
private:

  void erase();
  void scanNodes();
  void createNodeProperty(NodeDescription& np, L1394::Device& device);
  Camera1394* createCamera(const string& );
private:
  struct InternalDeviceStruct {
    InternalDeviceStruct() : device(0) {}
    NodeDescription       node_property;    ///< stores the NodeDescription
    L1394::Device*        device;           ///< stores a pointer to the l1394 device
    UserInfo              user_info;        ///< stores the application that requests the node
  };

  typedef map<ObjectID, InternalDeviceStruct*>::iterator InternalIt;
  typedef map<ObjectID, InternalDeviceStruct*>::const_iterator CInternalIt;
  map<ObjectID, InternalDeviceStruct*> __device_map;            ///< stores all devices
  map<ObjectID, InternalDeviceStruct*> __used_map;              ///< stores the used devices
  map<NMMReference, UserInfo>        __allocated_nodes;       ///< stores the created nodes

  ThreadMutex                        __internal_mutex;        ///< mutex for safe search and busreset.

  L1394::Session* session;                                    ///< stores a pointer to the L1394 Session
  static const char* __plugin_name;
};


}

#include "nmm/nmm_suffix.hpp"

#endif

