/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include <iostream>
#include "ServerRegistry.hpp"
#include "LocalRegistry.hpp"
#include "nmm/comm/ReferenceManager.hpp"
#include "PluginManager.hpp"
#include "ObjectFactory.hpp"
#include "nmm/multimedia/NodeDescription.hpp"
#include "nmm/comm/ProxyObject.hpp"
#include "nmm/comm/InstreamChannel.hpp"
#include "nmm/utils/SecurityCheck.hpp"
#include "nmm/utils/directory_utils.hpp"
#include "nmm/utils/thread/MutexGuard.hpp"
#include "nmm/utils/time/Clock.hpp"
#include "nmm/utils/time/TimedElement.hpp"
#include "nmm/multimedia/Node.hpp"
#include "nmm/comm/ProcessUserInfo.hpp"
#include "nmm/comm/netstrategy/UDPBroadcaster.hpp"
#include "nmm/services/registry/ClientRegistry.hpp"
#include "DefaultSynchronizer.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {

ServerRegistry::ServerRegistry(bool security_check /* = true */ )
  : Registry("ServerRegistry", 0),  __is_running(true), __default_reservation_time(20), __default_transaction_time(10),
    __time_interval(5), __member_mutex(ThreadMutex::RECURSIVE), __thread_condition(__thread_mutex),
    __security_enabled(security_check), __broadcaster(0), __port(ClientRegistry::default_port),
    __synchronizer_name(default_synchronizer_name)
{
  __local_registry.reset(SLocalRegistry::getInstance().getInterface<IRegistry>());
  SLocalRegistry::getInstance().setParent(this);
  addRegistry(SLocalRegistry::getInstance());
//  activateThread();

  // Get synchronizer name from environment variable
  string synchronizer_name_from_env = getNMMEnvVar("NMM_SINK_SYNCHRONIZER");
  if(!synchronizer_name_from_env.empty()) __synchronizer_name = synchronizer_name_from_env;
}

ServerRegistry::~ServerRegistry()
{
  /* Begin> Deactivate the running thread */
  __thread_mutex.lock();
  __is_running = false;
  __thread_condition.notify();
  __thread_mutex.unlock();

  deactivateThread();
  /* End< */
}

void ServerRegistry::run() {
  while(__is_running) {
    __thread_mutex.lock();
    __thread_condition.relativeTimedWait(__time_interval, 0);

    /* Check the Timeout for access control */
    checkTransactionLock();

    /* Check the Timeout for reserved nodes */
    checkNodeReservation();

    __thread_mutex.unlock();
  }
}

list<ObjectID >  ServerRegistry::searchNode(const RegistryID& registry_id, const NodeDescription& node_desc) const
{
  /* Begin> Forward the method call */
  const IRegistry* tmp_registry = findRegistry(registry_id);
  if (tmp_registry)
    return tmp_registry->searchNode(registry_id, node_desc);

  throw RegistryException("Unknown Registry");
}

list<Response >  ServerRegistry::initRequest(const NodeDescription& node_desc) const {
  list<Response> resp = __local_registry->initRequest(node_desc);
  list<ObjectID> id_list;
     
  for (list<RegistryMemento>::const_iterator it = __memento_list.begin(); it != __memento_list.end(); ++it) {
    for (list<MementoStruct>::const_iterator it2 = it->getMementoList().begin(); it2 != it->getMementoList().end(); ++it2) {
      if (it2->getNodeDescription().isSubset(node_desc) ) {
        id_list.push_back( it2->getObjectID() );
      }
    }
    if ( !id_list.empty() ) {
      resp.push_back(Response( it->getRegistryID(), id_list));
      id_list.clear();
    }
  }
  return resp;
}

list<NodeInfo> ServerRegistry::getNodeInfoList(const RegistryID& reg_id) const {
  if (reg_id == getOwnRegistryID())
    return list<NodeInfo>();

  /* Forward the method call */
  return __local_registry->getNodeInfoList(reg_id);
}

bool ServerRegistry::securityChecks(const UserInfo& info, bool throw_exception /* = true */ )
  throw(const RegistryException&)
{
  if(__security_enabled && !SSecurityCheck::getInstance().checkPasswd(info.getPasswd())) {
    ERROR_STREAM("Security check not passed for password " << endl << info << endl << "check .nmmrc for options" << endl);
    if(throw_exception) {
      throw RegistryException("Security check not passed for password " + info.getPasswd());
    }
    else {
      return false;
    }
  }
  else {
    return true;
  }
}

void ServerRegistry::initTransaction(const UserInfo& info)
      throw(const RegistryException&)
{
  securityChecks(info);

  try {
    Registry::initTransaction(info);
    __transaction_lock.user_info        = info;
    __transaction_lock.reservation_time = __default_transaction_time;
    __transaction_lock.is_locked        = true;
  } catch ( const RegistryException& e) {
    throw e;
  }
}

void ServerRegistry::closeTransaction(const UserInfo& info) {
  if(!securityChecks(info, false)) return;

  try {
    Registry::closeTransaction(info);
    __transaction_lock.user_info        = info;
    __transaction_lock.reservation_time = __default_transaction_time;
    __transaction_lock.is_locked        = false;
  } catch ( const RegistryException& e) {
    throw e;
  }
}

nmm_int32 ServerRegistry::getMaxInstanceCount(const RegistryID& reg_id, const ObjectID& node_id) const {
  /* Begin> Forward the method call */
  const IRegistry* tmp_registry = findRegistry(reg_id);
  if (tmp_registry)
    return tmp_registry->getMaxInstanceCount(reg_id, node_id);

  throw RegistryException("Unknown Registry");
}

nmm_int32 ServerRegistry::getInstanceCount(const RegistryID& reg_id, const ObjectID& node_id) const {
  /* Begin> Forward the method call */
  const IRegistry* tmp_registry = findRegistry(reg_id);
  if (tmp_registry)
    return tmp_registry->getInstanceCount(reg_id, node_id);

  throw RegistryException("Unknown Registry");
}

string ServerRegistry::getPluginInfo(const RegistryID& reg_id, const ObjectID& node_id) const {
  /* Begin> Forward the method call */
  const IRegistry* tmp_registry = findRegistry(reg_id);
  if (tmp_registry)
    return tmp_registry->getPluginInfo(reg_id, node_id);

  throw RegistryException("Unknown Registry");
  /* End< */
}

INode* ServerRegistry::requestNode(const RegistryID& reg, const ObjectID& node_id, const UserInfo& info, NodeDescription& nd)
      throw(const RegistryException&)
{
  securityChecks(info);

  /* Lock the __reservation_list */
  MutexGuard guard(__member_mutex);

  /* Search the corresponding Registry */
  IRegistry* tmp_registry = findRegistry(reg);

  /* Begin> If the registry is not available, search the memento list */
  if (!tmp_registry) {
    for ( list<RegistryMemento>::iterator it = __memento_list.begin(); it != __memento_list.end(); ++it)
    {
      if ( it->getRegistryID() == reg ) {
        if (it->isPluginRequired()) {
          try {
            IPluginManager_var i_plugin_mgr ( SPluginManager::getInstance().getCheckedInterface<IPluginManager>());
            i_plugin_mgr->loadLibrary(it->getPluginName());
          } catch (const Exception& e) {
            throw RegistryException(e.getComment());
          }
        } else {
          ERROR_STREAM("No plugin specified!!!" << endl);
        }
        /* Search the corresponding Registry */
        tmp_registry = findRegistry(reg);
        break;
      }
    }
  }


  if (tmp_registry) {
    /* request the node */
    INode* tmp_node = tmp_registry->requestNode(reg, node_id, info, nd);
    /* Add the reservation entry */
    __reservation_list.push_back(Reservation(reg, node_id, info, __default_reservation_time, tmp_node->getParentReference()));

    /* Add the registry interface to the node */
    tmp_node->getParentObject()->setRegistry(*getOwnInterface<IServerRegistry>());

    return tmp_node;
  }

  throw RegistryException("Unknown Registry");
}

void ServerRegistry::releaseNode(const RegistryID& reg_id, const NMMReference& ref, const ObjectID& node_id, const UserInfo& user_info) {
  if(!securityChecks(user_info, false)) return;

  /* Lock the __reservation_list */
  MutexGuard guard(__member_mutex);

  /* Search the corresponing Registry */
  IRegistry* tmp_registry = findRegistry(reg_id);
  if (tmp_registry) {
    /* release the node */
    tmp_registry->releaseNode(reg_id, ref, node_id, user_info);
    /* Begin> remove the reservation entry */
    for (list<Reservation>::iterator it = __reservation_list.begin(); it != __reservation_list.end(); ++it)
      if (it->reference == ref) {
        __reservation_list.erase(it);
        break;
      }
    /* End< */
    return;
  } else
    throw RegistryException("Unknown Registry");
}

void ServerRegistry::releaseSharedNode(const INode& i_node) {
  list<Reservation>::iterator pos = __reservation_list.end();
  for (list<Reservation>::iterator it = __reservation_list.begin(); it != __reservation_list.end(); ++it) {
    if ( (it->reference == i_node.getParentReference()) ) {
      pos = it;
      break;
    }
  }

  if ( pos != __reservation_list.end() ) {
    releaseNode(pos->reg_id, pos->reference, pos->node_id, pos->user_info);
  } else {
    ERROR_STREAM("Node is not reserved by this ServerRegistry" << endl);
  }
}

INode* ServerRegistry::instantiateNode(const RegistryID& reg_id, const FactoryTicket& ticket, const UserInfo& info)
      throw(const RegistryException&)
{
  securityChecks(info);

  /* Search the corresponing Registry */
  IRegistry* tmp_registry = findRegistry(reg_id);
  if (tmp_registry)
    /* create the node */
    return tmp_registry->instantiateNode(reg_id, ticket, info);

  throw RegistryException("Unknown Registry");
}

FactoryTicket ServerRegistry::reserveNode(const RegistryID& reg_id, const ObjectID& node_id, const UserInfo& info)
      throw(const RegistryException&)
{
  securityChecks(info);

  /* Lock the __reservation_list */
  MutexGuard guard(__member_mutex);

  /* Search the corresponing Registry */
  IRegistry* tmp_registry = findRegistry(reg_id);
  if (tmp_registry) {
    /* Reserve the node */
    FactoryTicket tmp_ticket = __local_registry->reserveNode(reg_id, node_id, info);
    /* Add the reservation entry */
    __reservation_list.push_back(Reservation(reg_id, node_id, info, __default_reservation_time));
    return tmp_ticket;
  }

  throw RegistryException("Unknown Registry");
}

void ServerRegistry::releaseReservation(const RegistryID& reg_id, const ObjectID& node_id, const UserInfo& info)
{
  if(!securityChecks(info, false)) return;

  /* Lock the __reservation_list */
  MutexGuard guard(__member_mutex);

  /* Search the corresponing Registry */
  IRegistry* tmp_registry = findRegistry(reg_id);
  if (tmp_registry) {
    /* release the reservation in the Registry */
    tmp_registry->releaseReservation(reg_id, node_id, info);
    /* Begin> Remove the reservation entry from the internal list */
    for (list<Reservation>::iterator it = __reservation_list.begin(); it != __reservation_list.end(); ++it)
      if ((it->node_id == node_id) && (it->reg_id == reg_id)) { // The ServerRegistry dont need to check against the user_info because it is the job of the other Registry
        __reservation_list.erase(it);
        break;
      }
    /* End< */
  } else
    throw RegistryException("Unknown Registry");
}

void ServerRegistry::destroyNode(const RegistryID& reg_id, const NMMReference& reference, const UserInfo& info)
{
  if(!securityChecks(info, false)) return;

  /* Forward the method call to LocalRegistry */
  return __local_registry->destroyNode(reg_id, reference, info);
}

list<ObjectID >  ServerRegistry::getNodeList(const RegistryID& reg_id) const
{
  /* Forward the method call to LocalRegistry */
  return __local_registry->getNodeList(reg_id);
}

list<ObjectID >  ServerRegistry::getReservedNodeList(const RegistryID& reg_id) const
{
  /* Forward the method call to LocalRegistry */
  return __local_registry->getReservedNodeList(reg_id);
}

NodeDescription ServerRegistry::getNodeDescription(const RegistryID& reg_id, const ObjectID& node_id) const{
  /* Forward the method call to LocalRegistry */
  return __local_registry->getNodeDescription(reg_id, node_id);
}

NodeFactoryID ServerRegistry::getNodeFactoryID(const RegistryID& reg_id, const ObjectID& node_id) const {
  /* Forward the method call to LocalRegistry */
  return __local_registry->getNodeFactoryID(reg_id, node_id);
}


void ServerRegistry::resetInternalReservation() {
  /* Forward the method call to LocalRegistry */
  __local_registry->resetInternalReservation();
}

void ServerRegistry::resetSpecificInternalReservation(const UserInfo& user_info){
  if(!securityChecks(user_info, false)) return;

  /* Forward the method call to LocalRegistry */
  __local_registry->resetSpecificInternalReservation(user_info);
}
/* End<            IRegistry interface              */


Time ServerRegistry::getSystemTime() {
  return TimedElement::getTime();
}

/* Begin>        IServerRegistry interface  */
nmm_uint32 ServerRegistry::getTransactionTimeout() {
  return __default_transaction_time;
}

void ServerRegistry::setTransactionTimeout(const  nmm_uint32& time_out) {
  __default_transaction_time = time_out;
}

 nmm_uint32 ServerRegistry::getNodeLeasingTime() {
  return __default_reservation_time;
}

void ServerRegistry::setNodeLeasingTime(const  nmm_uint32& time_out) {
  __default_reservation_time = time_out;
}

 nmm_uint32 ServerRegistry::refreshNodeLeasing(const RegistryID& reg_id, const ObjectID& node_id, const UserInfo& user_info)
{
  if(!securityChecks(user_info, false)) return 0;

  __member_mutex.lock();
  /* Search the reservation entry and reset the reservation time again */
  for (list<Reservation>::iterator it = __reservation_list.begin(); it != __reservation_list.end(); ++it)
  {
    if (it->node_id == node_id)
      if (it->reg_id == reg_id)
        if (it->user_info == user_info)
          it->reservation_time = __default_reservation_time;
  }
  /* End< */
  __member_mutex.unlock();
  return __default_reservation_time;
}

  nmm_uint32 ServerRegistry::refreshTransaction(const UserInfo& user_info)
{
  if(!securityChecks(user_info, false)) return 0;

  if (__transaction_lock.user_info == user_info) {
    /* refresh the timeout for the access control */
    __transaction_lock.reservation_time = __default_reservation_time;
    return __default_reservation_time;
  }
  return 0;
}

void ServerRegistry::checkTransactionLock() {
  __member_mutex.lock();
  if (__transaction_lock.is_locked) {
    /* decrement the transaction about the time wait in the thread */
    __transaction_lock.reservation_time -= __time_interval;
    /* Begin> If the reserved time is < 0, disable access control */
    if (__transaction_lock.reservation_time < 0) {
      WARNING_STREAM("Transaction Timeout! Close Transaction "<< endl);
      try {
        closeTransaction(__transaction_lock.user_info);
      } catch (const Exception& e) {
        ERROR_STREAM( e << endl);
      }
    }
    /* End< */
  }
  __member_mutex.unlock();
}

void ServerRegistry::checkNodeReservation() {
  __member_mutex.lock();
  list< list<Reservation>::iterator > it_list;
  /* Begin> Decrement the reservation time and prepare nodes for destruction if the timeout is < 0*/
  for (list<Reservation>::iterator it = __reservation_list.begin(); it != __reservation_list.end(); ++it)
  {
    /* decrement the reserved time with the time wait in the run-method */
    it->reservation_time -= __time_interval;

    /* Begin> If the reservation time is < 0, release the node */
    if (it->reservation_time < 0) {
      WARNING_STREAM("Timeout for reserved node " << endl);
      if (it->reference != 0) {
        /* change the node's state to CONSTRUCTED, so it can be destroyed */
        prepareNodeForDestruction(it->reference);
      } else {
      }
      it_list.push_back(it);
    }
    /* End< */
  }
  /* End< */

  /* Begin> Release the node by the corresponding registry */
  for (list< list<Reservation>::iterator >::iterator it = it_list.begin(); it != it_list.end(); ++it) {
    list<Reservation>::iterator t_it = *it;
    /* If its reference != 0 the node is allocated in this address space and the releaseNode method must be called to destroy it*/
    if (t_it->reference != 0)
      __local_registry->releaseNode(t_it->reg_id, t_it->reference, t_it->node_id, t_it->user_info);

    /* If its reference == 0 the node is not allocated in this address space and it must be released with the method releaseReservation */
    else
      __local_registry->releaseReservation(t_it->reg_id, t_it->node_id, t_it->user_info);

    /* remove the reservation entry */
    __reservation_list.erase(t_it);
  }
  /* End< */
  __member_mutex.unlock();
}

void ServerRegistry::prepareNodeForDestruction(const NMMReference& ref) {
  Node* tmp_node = SReferenceManager::getInstance().getObject<Node>(ref);
  if (!tmp_node) {
    ERROR_STREAM( "Invalid address " << endl);
    return;
  }

  WARNING_STREAM("Prepare node for destruction " << endl);

  try {
    INode_var i_node(tmp_node->getCheckedInterface<INode>());
    switch( i_node->getState() ) {
      case INode::STARTED :
        i_node->stop(); 
        i_node->flush();            //don't call break. So the other state changes are called automatically.

      case INode::ACTIVATED :
        i_node->deactivate();       //don't call break. So the other state changes are called automatically.

      case INode::OUTPUT_INITIALIZED :
        i_node->deinitOutput();     //don't call break. So the other state changes are called automatically.

      case INode::INITIALIZED :
        i_node->deinit();           //don't call break. So the other state changes are called automatically.

      case INode::CONSTRUCTED :     //To avoid warnings
        break;
    }
  } catch ( const Exception& e ) {
    ERROR_STREAM("Could not finalize the node " << endl);
  }
}

ISession* ServerRegistry::initSession(GraphDescription& gd, const UserInfo& u_info, const ISession* given_session) {
  if(!securityChecks(u_info, false)) return 0;

  ERROR_STREAM("sshar: "<< " Init a new Session" << endl);
  /* allocate the session object */
  Session* session = new Session(gd);

  /* Get the session interface */
  ISession* i_session = session->getCheckedInterface<ISession>();

  /* Begin> Init a new session for the shared graph */
  ISinkSynchronizer_var i_sync_sink;
  ISourceSynchronizer_var i_sync_source;

  if (given_session) {
    /* Store the synchronizer of given session in the new one to enable synchronized playpback */
    i_sync_sink.reset(given_session->getSinkSynchronizer());
    i_sync_source.reset(given_session->getSourceSynchronizer());

    /* Begin> store the GraphDescription of the given session as master graph of the new session */
    GraphDescription_var master_graph(given_session->getGraphDescription());
    if (master_graph.get()) {
      i_session->setMasterGraph(*master_graph.get());
    }
    /* End< */
  } else {
    /* allocate a synchronizer if required */
    const ObjectDescription& od = gd.getSinkSynchronizerDescription();
    Interface_var sink_interface;
    if(od.getObjectName().empty()) {
      sink_interface.reset(requestNMMObject(ObjectDescription(__synchronizer_name)));
    }
    else {
      sink_interface.reset(requestNMMObject(od));
    }
    i_sync_sink.reset(sink_interface->getParentObject()->getCheckedInterface<ISinkSynchronizer>());

    Interface_var source_interface(requestNMMObject(ObjectDescription("MasterSlaveSourceSynchronizer")));
    i_sync_source.reset(source_interface->getParentObject()->getCheckedInterface<ISourceSynchronizer>());
  }

  /* store the objects inside the session */
  i_session->setSinkSynchronizer(*i_sync_sink);
  i_session->setSourceSynchronizer(*i_sync_source);

  SessionIt_t it = __session_list.insert(pair<ProxyObject*, SessionStruct>(i_session->getParentObject(), SessionStruct())).first;

  /* Store the session specific attributes */
  it->second.session = session;
  it->second.i_session = dynamic_cast<ISession*>(i_session->copy());
  it->second.user_info = u_info;

  gd.setISession(*i_session);

  return i_session;
}

Interface* ServerRegistry::requestNMMObject(const ObjectDescription& desc)
    throw(const Exception&)
{
  DEBUG_STREAM( "Allocate object " << desc.getObjectName() << " in local address space" << endl);

  // Let the factory create the object
  NMMObject* object = SObjectFactory::getInstance().createObject(desc.getObjectName());
  if(!object) {
    throw Exception("Unknown NMMObject " + desc.getObjectName());
  }

  Interface* _interface = object->getInterface<Interface>();

  __object_map.insert(pair<const ProxyObject*, NMMObject*>(_interface->getParentObject(), object));

  return _interface;
}

void ServerRegistry::releaseNMMObject(const Interface& i_face)
    throw(const Exception&)
{
  ObjectIt_t it = __object_map.find( i_face.getParentObject() );
  if ( it != __object_map.end() ) {
    delete it->second;
    __object_map.erase(it);
  } else
    ERROR_STREAM( "Illegal NMMObject" << endl);
}
/* End<        IServerRegistry interface  */

/*****************************************************/

EdgeDescription::OverlapStatus ServerRegistry::overlapComplete(const NodeDescription& nr_out, const EdgeDescription& nr_edge, const NodeDescription& nr_in,
                                         const NodeDescription& nq_out, const EdgeDescription& nq_edge, const NodeDescription& nq_in )
{
  // *********************** COMPLETE ************************
  // test 1: all allow sharing
  if(nr_out.allowsSharing() && nr_in.allowsSharing() && nq_out.allowsSharing() && nq_in.allowsSharing()) {
    // test 2: node descrs are subsets
    if(nr_out.isSubset(nq_out) && nr_in.isSubset(nq_in)) {
      // test 3: same out tag and in tag
      if(nr_edge.getOutputJackTag() == nq_edge.getOutputJackTag() && nr_edge.getInputJackTag() == nq_edge.getInputJackTag()) {
        // test 4: same connection format
        if(nr_edge.getFormat() != 0) {
          if(nq_edge.getFormat() != 0) {
            const Format* fr = nr_edge.getFormat();
            const Format* fq = nq_edge.getFormat();
            if(fr->isFormatIncluded(fq)) { 
              return EdgeDescription::COMPLETE;
            }
          }
          else { // no format for query -> format is identical
            return EdgeDescription::COMPLETE;
          }
        }
        else { // no format for running edge
          if(nq_edge.getFormat() == 0) { // no format for query edge -> formats are identical (FIXME: work-around)
            return EdgeDescription::COMPLETE;
          }
        }
      }
    }
  }
  return EdgeDescription::NONE;
}

EdgeDescription::OverlapStatus ServerRegistry::overlapPartialCopy(const NodeDescription& nr_out, const EdgeDescription& nr_edge,
                                            const NodeDescription& nq_out, const EdgeDescription& nq_edge)
{
  // *********************** PARTIAL_COPY ************************
  // test 1: all allow sharing (only out)
  if(nr_out.allowsSharing() && nq_out.allowsSharing()) {
    // test 2: node descrs are subsets (only out)
    if(nr_out.isSubset(nq_out)) {
      // test 3: same out tag only
      if(nr_edge.getOutputJackTag() == nq_edge.getOutputJackTag()) {
        // test 4: same connection format
        if(nr_edge.getFormat() != 0) {
          if(nq_edge.getFormat() != 0) {
            const Format* fr = nr_edge.getFormat();
            const Format* fq = nq_edge.getFormat();
            if(fr->isFormatIncluded(fq)) { 
              return EdgeDescription::PARTIAL_COPY;
            }
          }
          else { // no format for query -> format is identical
            return EdgeDescription::PARTIAL_COPY;
          }
        }
        else { // no format for running edge
          if(nq_edge.getFormat() == 0) { // no format for query edge -> formats are identical (FIXME: work-around)
            return EdgeDescription::PARTIAL_COPY;
          }
        }
      }
    }
  }
  return EdgeDescription::NONE;
}

EdgeDescription::OverlapStatus ServerRegistry::overlapPartialOutput(const NodeDescription& nr_out, const EdgeDescription& nr_edge,
                                              const NodeDescription& nq_out, const EdgeDescription& nq_edge)
{
  // *********************** PARTIAL_OUTPUT ************************
  // test 1: all allow sharing (only out)
  if(nr_out.allowsSharing() && nq_out.allowsSharing()) {
    // test 2: node descrs are subsets (only out)
    if(nr_out.isSubset(nq_out)) {
      // test 3: same out tag, but ...
      if(nr_edge.getOutputJackTag() == nq_edge.getOutputJackTag()) {
        // test 4: possible connection format ? (FIXME: use output property instead of single format!)
        if(nr_edge.getFormat() != 0) {
          if(nq_edge.getFormat() != 0) {
            const Format* fr = nr_edge.getFormat();
            const Format* fq = nq_edge.getFormat();
            if(fr->isFormatIncluded(fq)) { 
              return EdgeDescription::PARTIAL_OUTPUT;
            }
          }
          else { // no format for query -> format is identical
            return EdgeDescription::PARTIAL_OUTPUT;
          }
        }
        else { // no format for running edge
          if(nq_edge.getFormat() == 0) { // no format for query edge -> formats are identical (FIXME: work-around)
            return EdgeDescription::PARTIAL_OUTPUT;
          }
        }
      }
    }
  }

  return EdgeDescription::NONE;
}

string depth = "";
void ServerRegistry::shareSubGraphs(list<GraphDescription* >& graph_list,
                    GraphDescription& o_graph,
                    GraphDescription& r_graph,
                    GraphDescription& q_graph)
{
  list<int> q_graph_sources;
  list<int> r_graph_sources;

  // get all sources of query graph
  const SequenceValue< NodeDescription >& q_nds = q_graph.getNodeDescriptionList();
  for(size_t i = 0; i < q_nds.size(); ++i) {
    const list<int>& in_list  = q_graph.getInEdgeList(i);
    const list<int>& out_list = q_graph.getOutEdgeList(i);
    // node with no incoming edge but some outgoing edge -> source (FIXME: or node with unconnected input)
    if(in_list.empty() && !out_list.empty()) {
      q_graph_sources.push_back(i);
      // DEBUG_STREAM("sshar: "<< depth << "found q source : " << i << endl);
    }
  }

  // get all sources of running graph
  const SequenceValue< NodeDescription >& r_nds = r_graph.getNodeDescriptionList();
  for(size_t i = 0; i < r_nds.size(); ++i) {
    const list<int>& in_list  = r_graph.getInEdgeList(i);
    const list<int>& out_list = r_graph.getOutEdgeList(i);
    // node with no incoming edge but some outgoing edge -> source (FIXME: or node with unconnected input)
    if(in_list.empty() && !out_list.empty()) {
      r_graph_sources.push_back(i);
      // DEBUG_STREAM("sshar: "<< depth << "found r source : " << i << endl);
    }
  }

  DEBUG_STREAM("sshar: "<< depth << "q sources size : " << q_graph_sources.size() << endl);
  DEBUG_STREAM("sshar: "<< depth << "r sources size : " << r_graph_sources.size() << endl);

 // TEST 
  // set break condition to false
  bool break_because_of_overlap = false;

  // for all sources in query graph
  for(list<int>::iterator q_iter = q_graph_sources.begin(); q_iter != q_graph_sources.end(); ++q_iter) {
    if(break_because_of_overlap == true) break;
    // for all sources in running graph
    for(list<int>::iterator r_iter = r_graph_sources.begin(); r_iter != r_graph_sources.end(); ++r_iter) {
      if(break_because_of_overlap == true) break;
      // for all outgoing edges of current query node
      const list<int>& q_out_list = q_graph.getOutEdgeList(*q_iter);
      for(list<int>::const_iterator q_edge_iter = q_out_list.begin(); q_edge_iter != q_out_list.end(); ++q_edge_iter) {
	if(break_because_of_overlap == true) break;
        // for all outgoing edges of current running node
        const list<int>& r_out_list = r_graph.getOutEdgeList(*r_iter);
        for(list<int>::const_iterator r_edge_iter = r_out_list.begin(); r_edge_iter != r_out_list.end(); ++r_edge_iter) {
	  if(break_because_of_overlap == true) break;
          // get edges
          EdgeDescription& q_edge = q_graph.getEdge(*q_edge_iter);
          EdgeDescription& r_edge = r_graph.getEdge(*r_edge_iter);

          // init status variables
          EdgeDescription::OverlapStatus complete_status = EdgeDescription::NONE;
          EdgeDescription::OverlapStatus copy_status     = EdgeDescription::NONE;
          EdgeDescription::OverlapStatus output_status   = EdgeDescription::NONE;

          // for copy overlap and complete overlap, running and query edge both need to be connected at both ends
          if(r_edge.isOutConnected() && r_edge.isInConnected() && q_edge.isOutConnected() && q_edge.isInConnected()) {

            // partial copy overlap ******************************************************************
            copy_status =
              overlapPartialCopy(r_graph.getNodeDescription(r_edge.getOutputNodeID()),
                                 r_edge,
                                 q_graph.getNodeDescription(q_edge.getOutputNodeID()),
                                 q_edge);

            if(copy_status == EdgeDescription::PARTIAL_COPY) { // need to add additional output stream to r
              // create new solution from current o_graph and add to list
              GraphDescription* new_o_graph = new GraphDescription(o_graph);
              graph_list.push_back(new_o_graph);

              // need to know the node id from running graph lateron -> simply copy node from r graph
              NodeDescription out_node(r_graph.getNodeDescription(r_edge.getOutputNodeID()));
              size_t edge_num = new_o_graph->addOutEdge(out_node, r_edge.getOutputJackTag(), r_edge.getFormat());
              
              // set edge parameter according to overlap type and set corresponding id's from q
              EdgeDescription& edge   = new_o_graph->getEdge(edge_num);
              edge.setOverlapType(EdgeDescription::PARTIAL_COPY);
              edge.setLeftOutputNodeID(q_edge.getOutputNodeID());
              edge.setLeftEdgeID(q_edge.getEdgeID());
              edge.setRightEdgeID(r_edge.getEdgeID());
              
              // create new r_graph and remove current edge (since some overlap exists)
              GraphDescription new_r_graph(r_graph);
              DEBUG_STREAM("sshar: "<< depth << "remove edge in r : " << *r_edge_iter << endl);
              new_r_graph.removeEdge(*r_edge_iter);
              // create new q_graph and remove current edge (since some overlap exists)
              GraphDescription new_q_graph(q_graph);
              DEBUG_STREAM("sshar: "<< depth << "remove edge in q : " << *q_edge_iter << endl);

	      // if right node of edge in query still contains more than single incoming edges ( = current edge )
	      //      -> only remove current edge in query, keep other edges 
	      // else -> remove complete subgraph
	      if(q_graph.getInEdgeList(q_edge.getInputNodeID()).size() > 1) {
		// remove only current edge, i.e. right node is mux
		// node with additional input
		new_q_graph.removeEdge(*q_edge_iter);
	      }
	      else {
		// remove complete subgraph (since it is a copy
		// overlap and complete subgraph of query then needs
		// to be created new)
		new_q_graph.removeSubgraph(*q_edge_iter);
	      }

              depth += "  ";
              // recursively continue with rest of r and q and new overlap graph
              shareSubGraphs(graph_list, *new_o_graph, new_r_graph, new_q_graph);
              depth = depth.substr(0, depth.size() - 2);

              // since copy overlap exists -> check for complete overlap, too *********************************
	      if(!r_edge.isInHelper()) {
		complete_status =
		  overlapComplete(r_graph.getNodeDescription(r_edge.getOutputNodeID()),
				  r_edge,
				  r_graph.getNodeDescription(r_edge.getInputNodeID()),
				  q_graph.getNodeDescription(q_edge.getOutputNodeID()),
				  q_edge,
				  q_graph.getNodeDescription(q_edge.getInputNodeID()));
		
		// if query edge was tested successfully for some overlap -> create new graph description
		if(complete_status == EdgeDescription::COMPLETE) {
		  // create new solution from current o_graph and add to list
		  GraphDescription* new_o_graph = new GraphDescription(o_graph);
		  // TEST: ONLY add o_graph to list IF node q_edge.in_id (right node) has no more outputs (= sink)
		  //       (otherwise graph will be valued with -1 because of missing ongoing edge)
		  //if(q_graph.getOutEdgeList(q_edge.getInputNodeID()).size() == 0) {
		  //graph_list.push_back(new_o_graph);
		  //}
		  graph_list.push_back(new_o_graph);
		  
		  // need to know the node id from running graph lateron -> simply copy nodes from running graph
		  NodeDescription out_node(r_graph.getNodeDescription(r_edge.getOutputNodeID()));
		  NodeDescription in_node (r_graph.getNodeDescription(r_edge.getInputNodeID()));
		  size_t edge_num = new_o_graph->addEdge(out_node, in_node);
		  
		  // set edge parameter according to overlap type and set corresponding id's from q
		  EdgeDescription& edge   = new_o_graph->getEdge(edge_num);
		  edge.setOverlapType(EdgeDescription::COMPLETE);
		  edge.setLeftOutputNodeID(q_edge.getOutputNodeID());
		  edge.setLeftInputNodeID(q_edge.getInputNodeID());
		  edge.setLeftEdgeID(q_edge.getEdgeID());
		  edge.setRightEdgeID(r_edge.getEdgeID());
		  
		  // create new r_graph and remove current edge (since some overlap exists)
		  GraphDescription new_r_graph(r_graph);
		  DEBUG_STREAM("sshar: "<< depth << "remove edge in r : " << *r_edge_iter << endl);
		  new_r_graph.removeEdge(*r_edge_iter);
		  // create new q_graph and remove current edge (since some overlap exists)
		  GraphDescription new_q_graph(q_graph);
		  DEBUG_STREAM("sshar: "<< depth << "remove edge in q : " << *q_edge_iter << endl);
		  new_q_graph.removeEdge(*q_edge_iter);
		  
		  depth += "  ";

		  // if (no outgoing edge is available for source in r 
		  //     and still some outgoing edge is available for source in q)
		  // -> need to add unconnected out edge to source in r because copy overlap is possible
		  if((new_r_graph.getOutEdgeList(*r_iter).empty()) && (!new_q_graph.getOutEdgeList(*q_iter).empty())) {
		    new_r_graph.addOutHelpEdge(new_r_graph.getNodeDescription(*r_iter),
					       r_edge.getOutputJackTag(),
					       r_edge.getFormat());
		  }
		  
		  // recursively continue with rest of r and q and new overlap graph
		  shareSubGraphs(graph_list, *new_o_graph, new_r_graph, new_q_graph);
		  depth = depth.substr(0, depth.size() - 2);
		}
	      }
	    }
          } // end: connected?

          // for output overlap, running edge needs to be connected at out-end and unconnected at in-end
          // query edge needs to be connected at out-end and in-end
          if(r_edge.isOutConnected() && !r_edge.isInConnected() && q_edge.isOutConnected() && q_edge.isInConnected()) {
            // in any case, check for partial output overlap ********************************************
            output_status =
              overlapPartialOutput(r_graph.getNodeDescription(r_edge.getOutputNodeID()),
                                   r_edge,
                                   q_graph.getNodeDescription(q_edge.getOutputNodeID()),
                                   q_edge);

            if(output_status == EdgeDescription::PARTIAL_OUTPUT) { // need to connect unconnected output of r
              // create new solution from current o_graph and add to list
              GraphDescription* new_o_graph = new GraphDescription(o_graph);
              graph_list.push_back(new_o_graph);

              // need to know the node id from running graph lateron -> simply copy node from r graph, but tag and format from q !!!
              NodeDescription out_node(r_graph.getNodeDescription(r_edge.getOutputNodeID()));
              size_t edge_num = new_o_graph->addOutEdge(out_node, q_edge.getOutputJackTag(), q_edge.getFormat());

              // set edge parameter according to overlap type and set corresponding id's from q
              EdgeDescription& edge   = new_o_graph->getEdge(edge_num);
              edge.setOverlapType(EdgeDescription::PARTIAL_OUTPUT);
              edge.setLeftOutputNodeID(q_edge.getOutputNodeID());
              edge.setLeftEdgeID(q_edge.getEdgeID());
              edge.setRightEdgeID(r_edge.getEdgeID());

              // create new r_graph and remove current edge (since some overlap exists)
              GraphDescription new_r_graph(r_graph);
              DEBUG_STREAM("sshar: "<< depth << "remove edge in r : " << *r_edge_iter << endl);
              new_r_graph.removeEdge(*r_edge_iter);
              // create new q_graph and remove current edge (since some overlap exists)
              GraphDescription new_q_graph(q_graph);
              DEBUG_STREAM("sshar: "<< depth << "remove edge in q : " << *q_edge_iter << endl);

	      // if right node of edge in query still contains more than single incoming edges ( = current edge )
	      //      -> only remove current edge in query, keep other edges 
	      // else -> remove complete subgraph
	      if(q_graph.getInEdgeList(q_edge.getInputNodeID()).size() > 1) {
		// remove only current edge, i.e. right node is mux
		// node with additional input
		new_q_graph.removeEdge(*q_edge_iter);
	      }
	      else {
		// remove complete subgraph (since it is a copy
		// overlap and complete subgraph of query then needs
		// to be created new)
		new_q_graph.removeSubgraph(*q_edge_iter);
	      }

              depth += "  ";
              // recursively continue with rest of r and q and new overlap graph
              shareSubGraphs(graph_list, *new_o_graph, new_r_graph, new_q_graph);
              depth = depth.substr(0, depth.size() - 2);
            }
          } // end: connected?

          DEBUG_STREAM("sshar: "<< depth << "overlap: ");
          if(r_edge.isOutConnected()) {
            DEBUG_STREAM("sshar: "<< r_graph.getNodeDescription(r_edge.getOutputNodeID()).getNodeName() << " -> ");
          }
          else {
            DEBUG_STREAM("sshar: "<< "UNCONNECTED_OUT_ID -> ");
          }
          if(r_edge.isInConnected()) {
            DEBUG_STREAM("sshar: "<< r_graph.getNodeDescription(r_edge.getInputNodeID()).getNodeName()  << " and ");
          }
          else {
            DEBUG_STREAM("sshar: "<< "UNCONNECTED_IN_ID and ");
          }
          if(q_edge.isOutConnected()) {
            DEBUG_STREAM("sshar: "<< q_graph.getNodeDescription(q_edge.getOutputNodeID()).getNodeName() << " -> ");
          }
          else {
            DEBUG_STREAM("sshar: "<< "UNCONNECTED_OUT_ID -> ");
          }
          if(q_edge.isInConnected()) {
            DEBUG_STREAM("sshar: "<< q_graph.getNodeDescription(q_edge.getInputNodeID()).getNodeName()  << " : ");
          }
          else {
            DEBUG_STREAM("sshar: "<< "UNCONNECTED_IN_ID : ");
          }

          DEBUG_STREAM("sshar: "<< "complete: " << (complete_status != EdgeDescription::NONE ? "yes" : "no" ) << ", copy: " << (copy_status != EdgeDescription::NONE ? "yes" : "no" ) << ", output: " << (output_status != EdgeDescription::NONE ? "yes" : "no" ) << endl);

	  // TEST
	  if(copy_status != EdgeDescription::NONE || complete_status != EdgeDescription::NONE || output_status != EdgeDescription::NONE) {
	    // set both iterators for query (source nodes and edges) to end of lists
	    // since all other possibilities will be considered in next step of recursion
	    break_because_of_overlap = true;
	  }

        } // end: for all outgoing edges of current running node
      } // end: for all outgoing edges of current query node
    } // end: for all sources in running graph
  } // end: for all sources in query graph
}

int ServerRegistry::valueGraph(const GraphDescription& o, const GraphDescription& q)
{
  int value = 0;

  // check all edges
  for(size_t i = 0; i < o.getNumEdges(); ++i) {
    const EdgeDescription& e = o.getEdge(i);

    // edge is complete overlap
    if(e.getInputNodeID() != EdgeDescription::UNCONNECTED_IN_ID) {
      // check if all outputs of in_node in q ( = right node of edge ) are also present in in_node of o
      // (a complete overlap is only valid, if all edges can go on from right node of edge!)
      if(q.getOutEdgeList(e.getLeftInputNodeID()).size() == o.getOutEdgeList(e.getInputNodeID()).size()) {
        value += 2;
      }
      else { // some output is missing -> dead end in overlap, graph o has no value
        return -1;
      }
    }
    else { // edge is either partial copy or output overlap
      value += 1;
    }

    // check, whether all outgoing edges of current out_node in q (= left node) are somehow overlapped in o (important for demux nodes!)
    // FIXME: should be ">" instead of "<" ???
    if(q.getOutEdgeList(e.getLeftOutputNodeID()).size() < o.getOutEdgeList(e.getOutputNodeID()).size()) {
      // some outgoing edge in current out_node is not overlapped -> graph has no value
      return -1;
    }

    // check, whether all ingoing edges of current out_node in q (= left node) are somehow overlapped in o (important for mux nodes!)
    // FIXME: should be ">" instead of "<" ???
    if(q.getInEdgeList(e.getLeftOutputNodeID()).size() < o.getInEdgeList(e.getOutputNodeID()).size()) {
      // some ingoing edge in current out_node is not overlapped -> graph has no value
      return -1;
    }

  }
  return value;
}


pair<GraphDescription*, int> ServerRegistry::shareGraphs(GraphDescription& r_graph, GraphDescription& q_graph)
{
  Clock clock;
  Time t1, t2;
  t1 = clock.getTime();

  DEBUG_STREAM("sshar: "<< "> shareGraphs" << endl);

  list<GraphDescription* > graph_list;
  GraphDescription o_graph;

  // start recursive computation
  shareSubGraphs(graph_list, o_graph, const_cast<GraphDescription&>(r_graph), q_graph);

  // evaluate overlaps
  int value = 0;
  list<GraphDescription* >::iterator best_graph = graph_list.end();
  for(list<GraphDescription* >::iterator i = graph_list.begin(); i != graph_list.end(); ++i) {
    DEBUG_STREAM("sshar: "<< "--- overlap graph ----------------------" << endl);
    (*i)->showEdges();
    int tmp = valueGraph(**i, q_graph);
    if(tmp > value) {
      best_graph = i;
      value = tmp;
    }
    DEBUG_STREAM("sshar: "<< "==> value " << tmp << endl);
  }

  if(best_graph != graph_list.end()) {
    DEBUG_STREAM("sshar: "<< endl << "BEST OVERLAP with value " << value << endl);
    (*best_graph)->showEdges();

    t2 = clock.getTime();
    DEBUG_STREAM("sshar: "<< "< shareGraphs, total time: " << (t2 - t1) << endl);
    return pair<GraphDescription*, int> ( *best_graph, value);
  } else {
    /* Otherwise return 0, because the two GD can not be shared */
    t2 = clock.getTime();
    DEBUG_STREAM("sshar: "<< "< shareGraphs, total time: " << (t2 - t1) << endl);
    return pair<GraphDescription*, int> ( 0, 0);
  }

}


void ServerRegistry::searchSharedGraphs(GraphDescription& query_graph, GraphDescription& overlap_graph)
{
  Clock clock;
  Time t1, t2;
  t1 = clock.getTime();

  /* Stores the GraphDescription with the 'best' value */
  pair<GraphDescription*, int> best_gd_match(0,0);

  DEBUG_STREAM("sshar: "<< "> searchSharedGraphs on " << SProcessUserInfo::getInstance().getHostname() << endl);

  for (SessionIt_t it = __session_list.begin(); it != __session_list.end(); ++it) {
    /* Get the current GraphDescription. Because the ISession returns a copy, we store the return
     * value in a _var-type for correct memory management.
     */
    GraphDescription_var running_graph(it->second.i_session->getGraphDescription());

    /* Test, if query graph already has set a particular session (-> handover)
     * Then, only try to share query with the graph description of the given session 
     */
    bool perform_test = false;
    if(query_graph.getISession()) {
      if(query_graph.getISession()->getParentObject() == it->second.i_session->getParentObject()) {
	perform_test = true;
      }
    }
    else { /* no particular session -> try to share graphs */
      perform_test = true;
    }

    if(perform_test) {
      /* Try to share the query graph with the current running graph */
      pair<GraphDescription*, int> gd = shareGraphs(*running_graph, query_graph);

      /* If their exist an overlap graph and it has a better value than our currently best one */
      if ( gd.first && gd.second > best_gd_match.second ) {
	best_gd_match = gd;
	best_gd_match.first->setISession(*it->second.i_session);
      }
    }
  }

  /* If we find an overlap graph, assign it to the overlap graph */
  if ( best_gd_match.first) {
    DEBUG_STREAM("sshar: "<< best_gd_match.first->getISession() << endl);
    overlap_graph = *best_gd_match.first;
    DEBUG_STREAM("sshar: "<< overlap_graph.getISession() << endl);
    DEBUG_STREAM("sshar: "<< endl << "TOTAL: BEST OVERLAP with value " << best_gd_match.second << endl);
    best_gd_match.first->showEdges();
  }

  t2 = clock.getTime();
  DEBUG_STREAM("sshar: "<< "< searchSharedGraphs, total time: " << (t2 - t1) << endl);
}

void ServerRegistry::requestGraph(GraphDescription& gd, const UserInfo& user_info){
  if(!securityChecks(user_info, false)) return;

  vector<NodeDescription*> gds = gd.getNodeDescriptionList();
  /* Begin> Request the nodes */
  for(size_t i = 0; i < gds.size(); ++i) {
    /* Begin> if we should not request this node, continue */
    if (! gd.getRequestFlag(*gds[i]))
      continue;
    
    /* Ignore already allocated Nodes */
    if(gd.getINode(*gds[i])) 
      continue;

    list<Response> resp = initRequest(*gds[i]);
    if(resp.empty()) {
      /* release the already requested nodes */
      releaseGraph(gd, user_info);
      throw RegistryException("Node " + gds[i]->getNodeName() + " is not available");
    }

    /* Try to request one of these nodes */
    INode* inode = requestNode(resp, user_info, *gds[i]);

    gd.setINode(*gds[i], *inode);
  }

  
}

void ServerRegistry::setSharingType(GraphDescription& gd, const vector<NodeDescription::SharingType>& sharing_types, SharingRule rule) {
  if ( rule == ServerRegistry::LEFT_SIDE) {
    for (unsigned int i = 0; i < sharing_types.size(); ++i) {
      switch (sharing_types[i] )  {
        case NodeDescription::EXCLUSIVE_OR_SHARED :
          gd.getNodeDescriptionList()[i]->setSharingType(NodeDescription::EXCLUSIVE);
          break;
        case NMM::NodeDescription::SHARED_OR_EXCLUSIVE :
          gd.getNodeDescriptionList()[i]->setSharingType(NodeDescription::SHARED);
          break;
        case NMM::NodeDescription::EXCLUSIVE_THEN_SHARED_OR_SHARED :
          gd.getNodeDescriptionList()[i]->setSharingType(NodeDescription::EXCLUSIVE_THEN_SHARED);
          break;
        case NMM::NodeDescription::SHARED_OR_EXCLUSIVE_THEN_SHARED :
          gd.getNodeDescriptionList()[i]->setSharingType(NodeDescription::SHARED);
          break;
        default:
          break;
      }
    }
  } else {
    for (unsigned int i = 0; i < sharing_types.size(); ++i) {
      switch (sharing_types[i]) {
        case NodeDescription::EXCLUSIVE_OR_SHARED :
          gd.getNodeDescriptionList()[i]->setSharingType(NodeDescription::SHARED);
          break;
        case NodeDescription::SHARED_OR_EXCLUSIVE :
          gd.getNodeDescriptionList()[i]->setSharingType(NodeDescription::EXCLUSIVE);
          break;
        case NodeDescription::EXCLUSIVE_THEN_SHARED_OR_SHARED :
          gd.getNodeDescriptionList()[i]->setSharingType(NodeDescription::SHARED);
          break;
        case NodeDescription::SHARED_OR_EXCLUSIVE_THEN_SHARED :
          gd.getNodeDescriptionList()[i]->setSharingType(NodeDescription::EXCLUSIVE_THEN_SHARED);
          break;
        default:
          break;
      }
    }
  }
}

ISession* ServerRegistry::requestSharedGraph(GraphDescription& gd, const UserInfo& user_info)
  throw (const NMM::RegistryException&)
{
  securityChecks(user_info);

  vector<NodeDescription*>& gds = gd.getNodeDescriptionList();
  vector<NodeDescription::SharingType> sharing_types;
  sharing_types.resize(gds.size());

  /* Store the old sharing types */
  for (unsigned int i = 0; i < gds.size(); ++i) {
    if (gds[i])
      sharing_types[i] = gds[i]->getSharingType();
    else //set a valid value
      sharing_types[i] = NodeDescription::EXCLUSIVE;
  }

  SharingRule rule = LEFT_SIDE;
  while ( 1 ) {
    setSharingType(gd, sharing_types, rule);
    GraphDescription overlap;
    try {
     /* Begin> Check if at least one source should be shared. */
      bool is_source_shared = false;
      for (unsigned int i = 0; i < gds.size(); ++i) {
        if (gds[i]) {
          if ( gd.getInEdgeList(gds[i]->getInternalID()).empty() ) {
            if( gds[i]->getSharingType() == NodeDescription::SHARED ) {
              is_source_shared = true;
              break; //and we can stop
            }
          }
        }
      }
      /* End< */

      if ( !is_source_shared ) {
        cerr << "Try to request an exclusive graph" << endl;
        /* Handle new graphs */
        requestGraph(gd, user_info);
        /* Begin>  reset the sharing types of node that must not be allocated */
        for (unsigned int i = 0; i < gds.size(); ++i) {
          if (gds[i] && (!gd.getRequestFlag(*gd.getNodeDescriptionList()[i])) ) {
            gds[i]->setSharingType(sharing_types[i]);
          }
        }
        /* End< */

        /* Init a new Session */
        return initSession(gd, user_info, 0);
      } else {
        cerr << "Try to request a shared graph" << endl;
        /* Handle shareable graphs */
        searchSharedGraphs(gd, overlap);

        /* If we find no Session that can be shared, throw an Exception */
        if (!overlap.getISession()) {
          throw RegistryException("No Session for sharing found" );
        }

        /* Get the NodeDescription from the session because it stores the INodes*/
        GraphDescription_var running_graph(overlap.getISession()->getGraphDescription());


        /* Store the nodes that can be shared in given gd */
        mapSharedGraph(gd, overlap, *running_graph, user_info, sharing_types);

        /* Begin>  reset the sharing types of node that are not shared and must not be allocated */
        for (unsigned int i = 0; i < gds.size(); ++i) {
          if (!gd.getINode(*gd.getNodeDescriptionList()[i]) && (!gd.getRequestFlag(*gd.getNodeDescriptionList()[i])) ) {
            gds[i]->setSharingType(sharing_types[i]);
          }
        }
        /* End< */

        /* Init a new session for the shared graph */
        return initSession(gd, user_info, overlap.getISession());
      }
    } catch (const RegistryException& e) {
      if  ( rule == LEFT_SIDE ) {
        rule = RIGHT_SIDE;
        DEBUG_STREAM("sshar: "<< e << endl);
      } else {
        throw e;
      }
    }
  }
}


void ServerRegistry::mapSharedGraph(GraphDescription& query, GraphDescription& overlap, GraphDescription& running, const UserInfo& user_info, const vector<NodeDescription::SharingType>& sharing_types)
  throw (const RegistryException& )
{
  securityChecks(user_info);

  DEBUG_STREAM("sshar: "<< endl << "--------------> requestGraph for overlap " << endl);
  overlap.showEdges();


  // check all edges of overlap:
  // get all INodes from shared nodes in overlap gd and set INode in query gd
  // (use the node description of query since the id's have to match lateron!)
  // also, set the edge type in query graph to sharing type
  for(size_t i = 0; i < overlap.getNumEdges(); ++i) {
    const EdgeDescription& e = overlap.getEdge(i);

    // take corresponding left node ( = out id ) of query and overlap
    NodeDescription& q_left_node = query.getNodeDescription(e.getLeftOutputNodeID());
    NodeDescription& o_left_node = overlap.getNodeDescription(e.getOutputNodeID());
    // this node is shared -> set INode of overlap for query
    INode* r_left_inode = running.getINode(o_left_node);
    query.setINode(q_left_node, *r_left_inode);

    // set sharing type of edge in query
    EdgeDescription::OverlapStatus overlap_type = e.getOverlapType();

    if(overlap_type == EdgeDescription::COMPLETE) { // -> also set things up for right node of edge ( = in id )
      NodeDescription& q_right_node = query.getNodeDescription(e.getLeftInputNodeID());
      NodeDescription& o_right_node = overlap.getNodeDescription(e.getInputNodeID());
      INode* r_right_inode = running.getINode(o_right_node);
      
      /* copy the values from the running NodeDescription*/
      updateNodeDescription(q_right_node, o_right_node);
      updateNodeDescription(q_left_node, o_left_node);

      /* Store the INode in the graph */
      query.setINode(q_right_node, *r_right_inode);
    }
    else if(overlap_type == EdgeDescription::PARTIAL_COPY) {
      updateNodeDescription(q_left_node, o_left_node);
    }
    else if(overlap_type == EdgeDescription::PARTIAL_OUTPUT) {
      updateNodeDescription(q_left_node, o_left_node);
    }
    else {
      ERROR_STREAM("sshar: "<< "BUG!" << endl);
    }

    // TODO: get corresponding (!) edge in query and set overlap type
    EdgeDescription& q_e = const_cast<EdgeDescription& >(query.getEdge(e.getLeftEdgeID()));
    q_e.setOverlapType( overlap_type );
    DEBUG_STREAM("sshar: "<< "set overlap type of edge with q_id " << e.getLeftEdgeID() << " with id " << q_e.getEdgeID() << " of query to " << overlap_type << ", r_id = " << e.getRightEdgeID() << endl);

    // set the instream channel in query graph
    if ( overlap_type == EdgeDescription::COMPLETE && running.getEdge(e.getRightEdgeID()).getInstreamChannel() ) {
      q_e.setInstreamChannel( running.getEdge(e.getRightEdgeID()).getInstreamChannel()->copy() );
    }
  }

  // request all remaining node descr from registry (that is without an INode set)
  SequenceValue< NodeDescription >& nds = query.getNodeDescriptionList();
  /* Begin> Request the nodes */
  for(size_t i = 0; i < nds.size(); ++i) {
    // INode is already set
    if(query.getINode(nds[i]) && query.getRequestFlag(*nds[i])) {
      cerr << "Not requesting node " << nds[i]->getNodeName() << " since it is shared" << endl;
    }

    /* If the node should be allocated in the server registry but is not shared */
    if(!query.getINode(nds[i]) && query.getRequestFlag(*nds[i])) {
      /* Begin> If the Sharing type is != SHARED it is allowed to allocate the node */
      if ( sharing_types[i] != NodeDescription::SHARED ) {
        cerr << "Requesting node " << nds[i]->getNodeName() << " of graph description\n";
        list<Response> resp = initRequest(*nds[i]);
        if(resp.empty()) {
          throw RegistryException("Node " + nds[i]->getNodeName() + " is not available");
        }
        const INode* inode = requestNode(resp, user_info, *nds[i]);
        query.setINode(*nds[i], *inode);

        /* Begin> Store the sharing type in the GraphDescription which can be either EXCLUSIVE
         * or EXCLUSIVE_THEN_SHARED */
        if ( sharing_types[i] == NodeDescription::SHARED_OR_EXCLUSIVE
          || sharing_types[i] == NodeDescription::EXCLUSIVE_OR_SHARED
          || sharing_types[i] == NodeDescription::EXCLUSIVE
           )
        {
          nds[i]->setSharingType(NodeDescription::EXCLUSIVE);
        } else {
          nds[i]->setSharingType(NodeDescription::EXCLUSIVE_THEN_SHARED);
        }
        /* End< */
      } else {
        cerr << string("Can not request node ") + nds[i]->getNodeName() + string(" can not be shared") << endl;
        /* A node with sharing type SHARED, can not be shared */
        throw RegistryException(string("Can not request node ") + nds[i]->getNodeName() + string(" can not be shared"));
      }
    }
  }
  DEBUG_STREAM("sshar: "<< "--------------> FINISHED" << endl);
}


void ServerRegistry::releaseGraph(GraphDescription& , const UserInfo& ){

}

void ServerRegistry::releaseSession(const ISession& session, const UserInfo& info) {
  if(!securityChecks(info, false)) return;

  SessionIt_t it = __session_list.find( session.getParentObject() );
  if (it != __session_list.end()) {
    __session_list.erase(it);
  }
  else {
    ERROR_STREAM (" Could not delete session object. No proxy object for given ISession registered " << endl);
  }
}

INode* ServerRegistry::requestNode(const list<Response>& request_list, const UserInfo& user_info, NodeDescription& nd)
  throw (const RegistryException& )
{
  securityChecks(user_info);

  /* Try to request one of the Nodes in the response-list */
  for (list<Response>::const_iterator it = request_list.begin(); it != request_list.end(); ++it) {
    for(list<ObjectID>::const_iterator it_2 = it->second.begin(); it_2 != it->second.end(); ++it_2)
      try {
        /* Try to request the node */
        return requestNode(it->first, *it_2, user_info, nd);
      } catch (const Exception& e) {
        WARNING_STREAM(e.getComment() << endl);
      }
  }
  throw RegistryException("No such node available");
}

void ServerRegistry::writeToStream(OValueStream& ov) {
  writeToStream(ov, getChildList(getRegistryID()));  
}

void ServerRegistry::writeToStream(OValueStream& ov, list<RegistryID> child_list) {
  for( list<RegistryID>::const_iterator it = child_list.begin(); it != child_list.end(); ++it){
    writeToStream(ov, getChildList(*it));
    RegistryMemento rm = getMemento(*it);
    ov << rm;
  }  
}

void ServerRegistry::readFromStream(IValueStream& iv) {
  while(!iv.isEmpty()) {
    RegistryMemento rm;
    iv >> rm;
    __memento_list.push_back( rm );       
  }
  UserInfo user_info;
  for( list<RegistryMemento>::iterator it = __memento_list.begin(); it != __memento_list.end(); ++it){
    it->getRegistryID().setUserInfo(user_info);
  }

  setMementoObject(getChildList(getRegistryID()));
 
}

list<NodeDescription> ServerRegistry::getNodeDescriptionList(const RegistryID& reg_id) {
/* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->getNodeDescriptionList(reg_id);
    else
      /* The Registry is unknown and we throw an Exception */
      throw Exception("Unknown RegistryID");
  /* End< */

  return list<NodeDescription>();  
}

RegistryMemento ServerRegistry::getMemento(const RegistryID& reg_id) {
/* Begin> Check the Registry and forward the method call if reg_id != getOwnRegistryID() */
  if (reg_id != getOwnRegistryID())
    if (findRegistry(reg_id))
      return findRegistry(reg_id)->getMemento();
    else
      /* The Registry is unknown and we throw an Exception */
      throw Exception("Unknown RegistryID");
  /* End< */

  return getMemento();
}

RegistryMemento ServerRegistry::getMemento() {
  RegistryMemento rm;
  rm.setRegistryID( getRegistryID() );

  return rm;
}

void ServerRegistry::announceRegistry(const RegistryID& reg_id) {
  setMementoObject( reg_id );
}

void ServerRegistry::setMementoObject(const RegistryID& reg_id) {
    for (list<RegistryMemento>::iterator it = __memento_list.begin(); it != __memento_list.end(); ++it)
	{
	    if ( reg_id.getRegistryName() == it->getRegistryID().getRegistryName() ) {
		if (findRegistry(reg_id)) {
		  findRegistry(reg_id)->setMemento(*it);
		  __memento_list.erase(it);
		  return;
		}
		else {
		    ERROR_STREAM ("ServerRegistry: Internal Error. Could not find Registry " << reg_id.getRegistryName() << endl);
		}
	    }
	}
    DEBUG_STREAM2("ServerRegistry: No Memento for Registry " << reg_id.getRegistryName() << " available" <<  endl);
}

void ServerRegistry::setMementoObject(list<RegistryID> child_list) {
  for( list<RegistryID>::const_iterator it = child_list.begin(); it != child_list.end(); ++it){
    setMementoObject(*it);
    setMementoObject(getChildList(*it));
  }
}


nmm_bool ServerRegistry::isSubRegistry(const RegistryID& reg_id) const {
  for (list<RegistryMemento>::const_iterator it = __memento_list.begin(); it != __memento_list.end(); ++it) {
      if(it->getRegistryID() == reg_id) {
	  return true;
      }
  }
  return Registry::isSubRegistry(reg_id);
}
	 
void ServerRegistry::sendIServerRegistry() {
  if (__broadcaster) {  
    IPluginManager_var i_plugin_mgr ( SPluginManager::getInstance().getCheckedInterface<IPluginManager>());
    Event e("ClientRegistry::addServerRegistry", 2);
    e.setValue(0, new TInValue<string>( SProcessUserInfo::getInstance().getHostname() ));
    e.setValue(1, new TInValue<nmm_int32>(__port));
    __broadcaster->sendEvent(e);
  } else {
    WARNING_STREAM ( "ServerRegistry is not able to broadcast its information " << endl);
  }
}

void ServerRegistry::setUDPBroadcaster(UDPBroadcaster* broadcaster, int port) {
  __port = port;
  __broadcaster = broadcaster;
  if ( __broadcaster ) {
    __broadcaster->setEventReceiver(this);
  }
}

/******************************************************/


}//end namespace


#include "nmm/nmm_suffix.hpp"
