/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "Session.hpp"
#include "IServerRegistry.hpp"
#include "nmm/comm/ProxyObject.hpp"
#include "nmm/multimedia/INode.hpp"
#include "nmm/multimedia/connect.hpp"
#include "nmm/multimedia/IReferenceCount.hpp"

#include <iostream>

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  Session::Session(const GraphDescription& gd)
  {
    setGraphDescription(gd);
  }

  Session::~Session() {
    if (__gd.get()) {
      decreaseCount(*__gd);
      releaseNodes();
    }
  }

  GraphDescription* Session::getGraphDescription() const
  {
    if (__gd.get())
      return __gd->copy();

    return 0;
  }
  
// ****  ICompositeNode* Session::getCompositeNode() const
// ****  {
// ****    if (__comp_node.get())
// ****      return dynamic_cast<ICompositeNode*>(__comp_node->copy());
// ****
// ****    return 0;
// ****  }
  
  ISourceSynchronizer* Session::getSourceSynchronizer() const
  {
    if (__source_synchronizer.get())
      return dynamic_cast<ISourceSynchronizer*>(__source_synchronizer->copy());

    return 0;
  }
  
  ISinkSynchronizer* Session::getSinkSynchronizer() const
  {
    if (__sink_synchronizer.get())
      return dynamic_cast<ISinkSynchronizer*>(__sink_synchronizer->copy());
    
    return 0;
  }

  void Session::setSourceSynchronizer(const ISourceSynchronizer& synchronizer) 
  {
    __source_synchronizer.reset(dynamic_cast<ISourceSynchronizer* >(synchronizer.copy())); 
  }

  void Session::setSinkSynchronizer(const ISinkSynchronizer& synchronizer) 
  {  
    __sink_synchronizer.reset(dynamic_cast<ISinkSynchronizer* >(synchronizer.copy())); 
  }

  void Session::setGraphDescription(const GraphDescription& gd)
  {
    /* First increase the ref count of the new nodes */
    increaseCount(gd);

    /* Then decrease the ref count of the old nodes */
    if (__gd.get()) {
      decreaseCount(*__gd);
    }

    /* Store the new Graphdescription */
    __gd.reset(gd.copy());

    /* Store the Session inside the graphdescription */
    __gd->setISession(*getOwnInterface<ISession>());
  }
  
// ****  void Session::setCompositeNode(const ICompositeNode& comp_node)
// ****  {
// ****    __comp_node.reset(dynamic_cast<ICompositeNode* >(comp_node.copy()));
// ****  }

  void Session::increaseCount(const GraphDescription& gd) {
    for (unsigned int i = 0; i < gd.getINodeList().size(); ++i) {
      if ( gd.getINodeList()[i] && gd.getReleaseFlag(*gd.getNodeDescriptionList()[i])) {
        IReferenceCount_var ref ( gd.getINodeList()[i]->getParentObject()->getCheckedInterface<IReferenceCount>());
        ref->increaseCount();
      }
    }
  }

  void Session::decreaseCount(const GraphDescription& gd) {
    for (unsigned int i = 0; i < gd.getINodeList().size(); ++i) {
      if ( gd.getINodeList()[i] && gd.getReleaseFlag(*gd.getNodeDescriptionList()[i])) {
        IReferenceCount_var ref ( gd.getINodeList()[i]->getParentObject()->getCheckedInterface<IReferenceCount>());
        ref->decreaseCount();
      }
    }
  }

  void Session::releaseNodes() {
    /* Begin> Fast disconnect from running graph. Disconnect all edges whose in-node has refcount > 0 and out-node refcount = 0 */
    for (unsigned int i = 0; i< __gd->getNumEdges(); ++i) {
      EdgeDescription& edge = __gd->getEdge(i);

      /* Check if the edge has an unconnected in- or output */
      if ( edge.getOutputNodeID() == EdgeDescription::UNCONNECTED_OUT_ID || edge.getInputNodeID() == EdgeDescription::UNCONNECTED_IN_ID )
      {
        /* This is an unconnected edge and we must continue! */
        continue;
      }

      NodeDescription& current_out_nd = __gd->getNodeDescription(edge.getOutputNodeID());
      NodeDescription& current_in_nd = __gd->getNodeDescription(edge.getInputNodeID());
      GLOBAL_DEBUG_STREAM ("Check connection from  " << current_out_nd.getNodeName() << " to " << current_in_nd.getNodeName() << " for fast disconnect " << endl);

      if ( __gd->getINode( current_out_nd ) &&  __gd->getINode( current_in_nd ) ) {
        IReferenceCount_var out_ref ( __gd->getINode(current_out_nd)->getParentObject()->getCheckedInterface<IReferenceCount>());
        IReferenceCount_var in_ref ( __gd->getINode(current_in_nd)->getParentObject()->getCheckedInterface<IReferenceCount>());
        if ( out_ref->getCount() > 0) {
          /* Begin> If ref count of the ingoing edge is 0 */
          if ( in_ref->getCount() == 0) {
            /* and the input jack of node 'current_in_node' was connected using this GraphDescription, disconnect the input of this node */
            if ( __gd->getINode( current_in_nd )->isInputConnected( edge.getInputJackTag() ) && 
                 edge.getInstreamChannel() != 0 ) 
            {
              GLOBAL_DEBUG_STREAM ("Disconnect " << current_out_nd.getNodeName() << " from " << current_in_nd.getNodeName() << endl);
              
              /* Stop both nodes */
              __gd->getINode( current_out_nd )->reachActivated();
              __gd->getINode( current_in_nd )->reachActivated();
              /* Flush the in-node (but not its predecssor, because it must send these buffers to other nodes) */
              __gd->getINode( current_in_nd )->flush();
              __gd->getINode( current_in_nd )->reachInitialized();
              /* Try to disconnect this node */
              try {
                disconnect_input(__gd->getINode( current_in_nd ), edge.getInputJackTag());
              } catch (const Exception& e) {
                GLOBAL_ERROR_STREAM( e << endl );
              }
              /* start the successor node again */
              __gd->getINode( current_out_nd )->reachStarted();
              /* Remove the edge from the disconnected node */
              const_cast<list<int>&>(__gd->getInEdgeList(current_in_nd.getInternalID())).remove(edge.getEdgeID());
             } 
           }
          /* End< */
        }
      }
    }
    /* End< Fast disconnect */

    /* Begin> Stop all nodes that are not stopped and have a ref count == 0 */
    for (unsigned int i = 0; i < __gd->getINodeList().size(); ++i) {
      if ( __gd->getINodeList()[i] ) {
        IReferenceCount_var ref ( __gd->getINodeList()[i]->getParentObject()->getCheckedInterface<IReferenceCount>());
        if ( ref->getCount() == 0 && __gd->getINodeList()[i]->isStarted()) {
          /* Stop and flush the remaining nodes */
          __gd->getINodeList()[i]->stop();
          __gd->getINodeList()[i]->flush();
        }
      }
    }
    /* End< */

    /* Begin> Disconnect the remaining graph of nodes that must be released */
    GLOBAL_DEBUG_STREAM("Begin > Disconnecting Graph " << endl);
    bool is_new_sink= true;
    while ( is_new_sink ) {
      is_new_sink = false;
      for (unsigned int i = 0; i < __gd->getNodeDescriptionList().size(); ++i) {
        if (__gd->getNodeDescriptionList()[i]) {
          if (__gd->getINode(*__gd->getNodeDescriptionList()[i])) {
            IReferenceCount_var ref ( __gd->getINodeList()[i]->getParentObject()->getCheckedInterface<IReferenceCount>());
            if ( ref->getCount() == 0) {
              /* Begin> If ref count == 0 we must disconnect this node */
              int current_id = __gd->getNodeDescriptionList()[i]->getInternalID();

              /* Begin> Remove all unconnected out id's */
              const list<int>& out_id_list = __gd->getOutEdgeList(current_id);
              list<int> unconnected_outputs;
              for (list<int>::const_iterator it = out_id_list.begin(); it != out_id_list.end(); ++it) {
                if (__gd->getEdge(*it).getInputNodeID() == EdgeDescription::UNCONNECTED_IN_ID) {
                  GLOBAL_DEBUG_STREAM("Found unconnected out id of Node " << __gd->getNodeDescription(i).getNodeName()
				      << "from: " << __gd->getEdge(*it).getOutputJackTag() << endl );
		  unconnected_outputs.push_back(*it);
                }
              }
              for (list<int>::const_iterator it = unconnected_outputs.begin(); it != unconnected_outputs.end(); ++it) {
                const_cast<list<int>&> (out_id_list).remove(*it);
              }
              /* End< */

              /* If OutEdgeList is empty we have disconnected all predecossor nodes and if InEdgeList is not empty we must disconnect these connections */
              if ( __gd->getOutEdgeList( current_id ).empty() && (!__gd->getInEdgeList( current_id ).empty()) ) {
                /* Set flag to true because we must iterate again to check if there are some nodes that must be disconnectd */
                is_new_sink = true;
                /* Get the list of all incoming edges */
                const list <int>& prev_id_list = __gd->getInEdgeList( current_id );

                /* Change node's state to initialized */
                __gd->getINodeList()[i]->reachState(INode::INITIALIZED);
                /* Begin> Disconnect all specified connections */
                for (list<int>::const_iterator it = prev_id_list.begin(); it != prev_id_list.end(); ++it ) {
                  /* If no instream channel is provided by the GraphDescription, the nodes have not been connected
                   * using this GrpahDescription. In this case we must not disconnect the node */
                  if ( __gd->getEdge( *it ).getInstreamChannel() ) {
                    /* disconnect the specified connection in the server registry */
                    GLOBAL_DEBUG_STREAM("Disconnect input: " << __gd->getNodeDescription(i).getNodeName() << __gd->getEdge( *it ).getInputJackTag()<<endl);
                  
                    try {
                      disconnect_input(__gd->getINodeList()[i], __gd->getEdge( *it ).getInputJackTag());
                    } catch (const Exception& e) {
                      GLOBAL_WARNING_STREAM(e << endl);
                    }
                    int prev_id = __gd->getEdge( *it ).getOutputNodeID();
                    /* remove edge_id from the previous node-lists */
                    const_cast<list<int>&>(__gd->getOutEdgeList(prev_id)).remove( __gd->getEdge( *it ).getEdgeID() );
                  } 
                }
                const_cast<list<int>&>(prev_id_list).clear() ;
                /* End< */
              }

              /* Begin> release nodes with ref count 0 and releasing flag true */
              if ( __gd->getReleaseFlag(*__gd->getNodeDescriptionList()[i])
                   && __gd->getInEdgeList( current_id ).empty()
                   &&  __gd->getOutEdgeList( current_id ).empty() ) {
                Interface_var my_interface (__gd->getINodeList()[i]->getParentObject()->getRegistry());
		if (my_interface.get() ) {
                  IServerRegistry_var reg (my_interface->getParentObject()->getInterface<IServerRegistry>());  
                  if ( reg.get() ) {
                    __gd->getINodeList()[i]->reachConstructed();
                    GLOBAL_DEBUG_STREAM("deleting: "<<__gd->getNodeDescription(i).getNodeName()<<endl);
                    reg->releaseSharedNode(*__gd->getINodeList()[i]);
                  } else {
                    GLOBAL_DEBUG_STREAM( " OBJECT IS NOT MANAGED BY A SERVERREGISTRY " << endl );
                  }
		}
                delete __gd->getNodeDescriptionList()[i];
                __gd->getNodeDescriptionList()[i] = 0;
                delete __gd->getINodeList()[i];
                __gd->getINodeList()[i] = 0;
              }
              /* End< */
            }
          }
        }
      }
    }
    GLOBAL_DEBUG_STREAM("End < Disconnecting Graph " << endl );
    /* End< */
  }

  void Session::realizeGraph() {
    if (__gd.get()) {
      __gd.get()->realizeGraph();
    } else {
      throw Exception(string("realizeGraph failed because Session has no GraphDescription" ));
    }
  }

  void Session::activateGraph() {
    if (__gd.get()) {
      __gd->activateGraph();
    } else {
      throw Exception(string("activateGraph failed" ));
    }
  }

  GraphDescription* Session::getMasterGraph() {
    if (__master_graph.get())
      return __master_graph->copy();

    return 0;
  }

  void Session::startGraph() {
    if (__gd.get()) {
      __gd->startGraph();
    } else {
      throw Exception(string("activateGraph failed" ));
    }
  }

  void Session::stopGraph() {
    if (__gd.get()) {
      __gd->stopGraph();
    } else {
      throw Exception(string("stopGraph failed" ));
    }
  }

  void Session::setMasterGraph(const GraphDescription& master_graph){
    __master_graph.reset(master_graph.copy());
  }

}


#include "nmm/nmm_suffix.hpp"
