/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "StrategyDescription.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  StrategyDescription::StrategyDescription(const TypeInfo& ti)
    : SerializedValue(ti),
      __strategy_type(),
      __roles(0),
      __is_exclusive(false),
      __i_value_stream(),
      __i_value_stream_stack(0),
      __o_value_stream(),
      __o_value_stream_stack(0)
  {
    registerAttributes();
  }
  
  StrategyDescription::StrategyDescription(const StrategyDescription& other)
    : SerializedValue(other),
      __strategy_type(other.__strategy_type.getRValue()),
      __roles(other.__roles.getRValue()),
      __is_exclusive(other.__is_exclusive.getRValue()),
      __i_value_stream(other.__i_value_stream.getRValue()),
      __i_value_stream_stack(0),
      __o_value_stream(other.__o_value_stream.getRValue()),
      __o_value_stream_stack(0)
  {
    /* copy IValueStream and OValueStream objects */
    if(other.__i_value_stream_stack.getRValue()) {
      __i_value_stream_stack.setValue(other.__i_value_stream_stack.getRValue()->copy());
    }
    if(other.__o_value_stream_stack.getRValue()) {
      __o_value_stream_stack.setValue(other.__o_value_stream_stack.getRValue()->copy());
    }
    
    registerAttributes();
  }
  
  StrategyDescription& StrategyDescription::operator=(const StrategyDescription& other) {
    SerializedValue::operator=(other);
    
    if(this != &other) {
      __strategy_type = other.__strategy_type;
      __roles = other.__roles;
      __is_exclusive = other.__is_exclusive;
      __i_value_stream.setValue(other.__i_value_stream.getRValue());
      __o_value_stream.setValue(other.__o_value_stream.getRValue());
      
      /* copy IValueStream and OValueStream objects */
      if(other.__i_value_stream_stack.getRValue()) {
        __i_value_stream_stack.setValue(other.__i_value_stream_stack.getRValue()->copy());
      } else {
        __i_value_stream_stack.setValue(0);
      }
      if(other.__o_value_stream_stack.getRValue()) {
        __o_value_stream_stack.setValue(other.__o_value_stream_stack.getRValue()->copy());
      } else {
        __o_value_stream_stack.setValue(0);
      }
    }
    
    return *this;
  }
  
  StrategyDescription::~StrategyDescription() {
    /* delete IValueStream and OValueStream objects */
    if(__i_value_stream_stack.getRValue()) {
      delete __i_value_stream_stack.getRValue();
      __i_value_stream_stack.setValue(0);
    }
    if(__o_value_stream_stack.getRValue()) {
      delete __o_value_stream_stack.getRValue();
      __o_value_stream_stack.setValue(0);
    }
  }
  
  void StrategyDescription::registerAttributes() {
    insertArgument(&__strategy_type);
    insertArgument(&__roles);
    insertArgument(&__is_exclusive);
    insertArgument(&__i_value_stream);
    insertArgument(&__i_value_stream_stack);
    insertArgument(&__o_value_stream);
    insertArgument(&__o_value_stream_stack);
  }
  
  void StrategyDescription::setIValueStream(const string& i_value_stream) {
    __i_value_stream.setValue(i_value_stream);
    
    /* delete old IValueStream object */
    if(__i_value_stream_stack.getRValue()) {
      delete __i_value_stream_stack.getRValue();
      __i_value_stream_stack.setValue(0);
    }
  }
  
  void StrategyDescription::setOValueStream(const string& o_value_stream) {
    __o_value_stream.setValue(o_value_stream);
    
    /* delete old OValueStream object */
    if(__o_value_stream_stack.getRValue()) {
      delete __o_value_stream_stack.getRValue();
      __o_value_stream_stack.setValue(0);
    }
  }
  
  void StrategyDescription::setIValueStreamStack(const IValueStream* i_value_stream) {
    /* replace old IValueStream object with new one */
    if(__i_value_stream_stack.getRValue()) {
      delete __i_value_stream_stack.getRValue();
    }
    if(i_value_stream) {
      __i_value_stream_stack.setValue(i_value_stream->copy());
    } else {
      __i_value_stream_stack.setValue(0);
    }
    
    /* set input value stream type to empty */
    __i_value_stream.setValue(string());
  }
  
  void StrategyDescription::setOValueStreamStack(const OValueStream* o_value_stream) {
    /* replace old OValueStream object with new one */
    if(__o_value_stream_stack.getRValue()) {
      delete __o_value_stream_stack.getRValue();
    }
    if(o_value_stream) {
      __o_value_stream_stack.setValue(o_value_stream->copy());
    } else {
      __o_value_stream_stack.setValue(0);
    }
    
    /* set output value stream type to empty */
    __o_value_stream.setValue(string());
  }
}

#include "nmm/nmm_suffix.hpp"
