/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "DirReader.hpp"
#include "NamedObject.hpp"

#include <unistd.h>    // for getuid
#include <sys/stat.h>  // for mkdir
#include <sys/types.h> // for mkdir
#include <errno.h>
#include <dirent.h>
#include <algorithm>   // for sort
#include <iostream>


#include "nmm/nmm_prefix.hpp"

DirReader::DirReader() {
  setRootDir( "/" );
  setNiceDir( false );
}

DirReader::~DirReader(){

}

void DirReader::setRootDir( string dir ) {
  root_dir = dir;

  removeLastSlash(dir);

  cur_dir = root_dir;
}

bool DirReader::makeDir(const string dir) {
  if (mkdir(dir.c_str(), 00777) != 0) {
    switch (errno) {
    case EEXIST: // dir already exists (not necessarily as a directory).  This includes the case where pathname is a symbolic  link, dangling or not
      {
	// stat if writeable directory
        struct stat statbuf;
        lstat( dir.c_str(), &statbuf );
        if (getuid() == statbuf.st_uid) {
          // current user owns this dir, has he write perms?
           if ((S_IWUSR & statbuf.st_mode) == S_IWUSR) {
            return true;
          }
        }

        bool isingroup = false;
        gid_t groups[256];
        int groupcount = getgroups(256, groups);
        for (int i=0; i < groupcount; i++) {
          if (statbuf.st_gid == groups[i]) {
            isingroup = true;
            break;
          }
        }
        if (isingroup) {
          // process is in the group of dir, has group write prems?
          if ((S_IWGRP & statbuf.st_mode) == S_IWGRP) {
            return true;
          }

        }

        // have others write perms in dir?
        if ((S_IWOTH & statbuf.st_mode) == S_IWOTH) {
          return true;
        }

        return false;
      }

      break;
    case ENOENT: // A directory component in dir does not exist or is a dangling symbolic link
      {
	// recursive makeDir parent
	string parent_dir = dir.substr(0, dir.rfind("/", dir.size()-1));
	if (false == makeDir(parent_dir)) {
	  return false;
	}
	mkdir(dir.c_str(), 00777); // no need to check here as parent_dir could been created and so dir is creatable or exists already (by recursion)
      }
      break;
    default: // all other cases mean that it is not possible to write to or create dir
      return false;
      break;
    }
  }
  return true;
}

void DirReader::setNiceDir( bool nd ) {
  nice_dir = nd;

  if ( nice_dir ) {
    backdir = "Go Up";
  } else {
    backdir = "..";
  }
}

string DirReader::getRootDir() {
  return( (root_dir == "/") ? root_dir : root_dir+"/" );
  //return( root_dir+"/" );
}

string DirReader::getCurDir() {
  return( (cur_dir == "/") ? cur_dir : cur_dir+"/" );
  //return( cur_dir+"/" );
}

string DirReader::getCurDirWithoutRoot() {
  unsigned int end = cur_dir.length();
  unsigned int begin = getRootDir().length();
  return( cur_dir.substr( begin-1, end ) );
}

bool DirReader::myIsDir( string complete_path ) {
  struct stat statbuf;
  stat( complete_path.c_str(), &statbuf );
  return( S_ISDIR(statbuf.st_mode) );
}

bool DirReader::myIsReg( string complete_path ) {
  struct stat statbuf;
  stat( complete_path.c_str(), &statbuf );
  return( S_ISREG(statbuf.st_mode) );
}

bool DirReader::readDir( string completePath ) {
  unsigned int num_dir_items = 0;
  unsigned int num_reg_items = 0;
  int num_entries   = 16;

  DIR*           indir;
  struct dirent* indirEntry;
  string         name;

  // first delete all entries
  cur_dir_entries.clear();

  dirs.resize( num_entries );
  regFiles.resize( num_entries );

  dirs.clear();
  regFiles.clear();

  // First read the Names of all Sub-Directories ...
  if ( (indir = opendir( completePath.c_str() )) != 0 ) {
    // change directory
    //if ( chdir( completePath.c_str() ) == -1 ) return false;
    cur_dir = completePath;

    dirs.resize( num_entries );
    if ( !nice_dir ) {
      dirs[num_dir_items].isdir  = true;
      dirs[num_dir_items].path   = getCurDir();
      dirs[num_dir_items++].name = ".";
    }

    if ( cur_dir != root_dir ) {
      dirs[num_dir_items].isdir  = true;
      dirs[num_dir_items].path   = getCurDir();
      dirs[num_dir_items++].name = backdir;
    }

    // catch all directory entries
    while( (indirEntry = readdir(indir)) !=0 ) {
       name = indirEntry->d_name;
       
       // check if we realy have a directory entry and no "." dir
       if( myIsDir(completePath + "/" + name) && name != "." ) {

         // do not add hidden directories
         if ( name[0] != '.' ) {
           
           // if not enough space in "array" resize it
           if ( num_dir_items == dirs.size() ) {
             dirs.resize( dirs.size() + num_entries );
           }
           dirs[num_dir_items].isdir  = true;
           dirs[num_dir_items].path   = getCurDir();
           dirs[num_dir_items++].name = name;
         }
       }
    }
    closedir(indir);
  } else {
    GLOBAL_ERROR_STREAM("Error while opening directory " << completePath << endl);
    return false;
  }

  // adjust array size to exact item numbers
  dirs.resize(num_dir_items);

  // sort directory entries
  //sort( dirs.begin(), dirs.end(), cmpdirents() );

  // Now read all regular filenames
  if ( (indir = opendir( completePath.c_str() )) != 0 ) {
    // change directory
    //if ( chdir( completePath.c_str() ) == -1 ) return false;
    cur_dir = completePath;
    
    // catch all regular filename
    while((indirEntry = readdir(indir))!=0) {
      
      string name = indirEntry->d_name;
      
      // check if we really have a regular file
      if( myIsReg(completePath + "/" + name) ) {
        
        // do not add hidden files
        if ( indirEntry->d_name[0] != '.' ) {
          
          // if not enough space in "array" resize it
          if ( num_reg_items == regFiles.size() ) {
	    regFiles.resize(regFiles.size() + num_entries);
	  }
	  regFiles[num_reg_items].isdir  = false;
	  regFiles[num_reg_items].path   = getCurDir();
	  regFiles[num_reg_items++].name = indirEntry->d_name;
        }
      }
    }
    closedir(indir);
  }
  else {
    GLOBAL_ERROR_STREAM("Error while opening directory" << completePath << endl);
    return false;
  }

  // resize to fit exactly (because sort would bring
  // empty entries to be shown at top)
  regFiles.resize(num_reg_items);

  // sort list
  if ( num_reg_items > 0 )
    sort( regFiles.begin(), regFiles.end(), cmpregents() );

  // copy dir entries in front of regular files
  cur_dir_entries.insert( cur_dir_entries.begin(), regFiles.begin(), regFiles.end() );
  cur_dir_entries.insert( cur_dir_entries.begin(), dirs.begin(), dirs.end() );

  return true;
}

void DirReader::removeLastSlash(string& dir) {
  // remove the last "/", but only if dir != "/".
  if (dir != "/") {
    unsigned int pos = dir.find_last_of('/');
    if ( (pos+1) == dir.size() ) dir = dir.substr( 0, pos );
  }
}

bool DirReader::changeDir( string ndir ) {
  unsigned int pos;

  removeLastSlash(ndir);

  if ( ndir == "." ) return( readDir( cur_dir ) );

  if ( ndir == backdir ) {
    // ".." selected in rootpath shouldn't do anything.
    if ( cur_dir == root_dir ) return true;

    // go dir up
    pos = cur_dir.find_last_of('/');
    // if destination dir is "/" set it by hand.
    if (0 == pos) ndir = "/";
    else ndir = cur_dir.substr( 0, pos );

    return( readDir( ndir ) );
  }

  pos  = ndir.find_first_of('/');
  if ( pos == 0 ) {
    ndir = root_dir + ndir;
  } else {
    ndir = getCurDir() + ndir;
  }

  return( readDir( ndir ) );
}


string DirReader::getSuffix( string name ) {
  string::size_type pos = name.find_last_of('.');
  if ( pos != string::npos ) {
    string temp;
    temp = name.substr( pos+1, name.length() );
    return( temp );
  } else {
    return( "" );
  }
}

string DirReader::getNiceFileName( string name, int len ) {
  string::size_type pos = name.find_last_of('.');
  if ( pos != string::npos ) {
   // cut ending
   name = name.substr( 0, pos );
  }

  pos = name.find_last_of('/');
  if ( pos != string::npos ) {
    // cut previous directories
    name = name.substr( pos+1, name.length() );
  }

  // should file name be cut?
  if ( len > 10  ) {
    int nl = name.length();
    if ( nl > len ) {
      int cut_left  = len/2-2;
      int cut_right = nl-(len/2-1);
      name = name.substr( 0, cut_left )+"..."+name.substr( cut_right, nl );
    }
  }

  return( name );
}

string DirReader::getFileName( string name, int len ) {
  string::size_type pos = name.find_last_of('/');
  if ( pos != string::npos ) {
    // cut previous directories
    name = name.substr( pos+1, name.length() );
  }

  // should file name be cut?
  if ( len > 10  ) {
    int nl = name.length();
    if ( nl > len ) {
      int cut_left  = len/2-2;
      int cut_right = nl-(len/2-1);
      name = name.substr( 0, cut_left )+"..."+name.substr( cut_right, nl );
    }
  }

  return( name );
}

string DirReader::getNiceDirName( string name, int len ) {
  if ( len > 10  ) {
    name = getNiceFileName( name, len-3 );
    if ( cur_dir != root_dir ) name = "../"+name;
  }

  return( name );
}

vector<DirReader::directory_t> DirReader::getDirEntries() {
  return( cur_dir_entries );
}

vector<DirReader::directory_t> DirReader::getDirsOnlyEntries() {
  return( dirs );
}

vector<DirReader::directory_t> DirReader::getRegFiles() {
  return( regFiles );
}

void DirReader::importDirectory(  vector<directory_t> mdirs, vector<directory_t> reg_files ) {
  dirs     = mdirs;
  regFiles = reg_files;
  vector<directory_t> temp;

  /* put backdir in front of directory entries */
  temp.resize( 1 );
  temp[0].isdir  = true;
  temp[0].path   = getCurDir();
  temp[0].name   = backdir;
  dirs.insert( dirs.begin(), temp.begin(), temp.end() );

  cur_dir_entries.clear();
  cur_dir_entries.insert( cur_dir_entries.begin(), regFiles.begin(), regFiles.end() );
  cur_dir_entries.insert( cur_dir_entries.begin(), dirs.begin(), dirs.end() );
}

#include "nmm/nmm_suffix.hpp"
