/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "SecurityCheck.hpp"

#include "string_utils.hpp"
#include "directory_utils.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
 
  SecurityCheck::SecurityCheck()
    : NamedObject("SecurityCheck")
  {
    string nmmrcfile(getUserHomeDirectory() + "/" + getNMMConfigFilename());
    nmmconfig.parseFile(nmmrcfile);
  }

  bool SecurityCheck::checkFilenameForReading(const string& dir_plus_filename) {
    if (dir_plus_filename.empty()) {
      return true;
    }
    
    string allowedpaths;
    nmmconfig.getValue("allowedreadpaths", allowedpaths);
    if(isFilenameWithinDirectories(dir_plus_filename, allowedpaths)) {
      return true;
    } else {
      ERROR_STREAM("Given filename " << dir_plus_filename << " is not in allowed paths for reading " << allowedpaths << endl);
      return false;
    } 
  }
  
  bool SecurityCheck::checkFilenameForWriting(const string& dir_plus_filename) {
    if (dir_plus_filename.empty()) {
      return true;
    }
    
    string allowedpaths;
    nmmconfig.getValue("allowedwritepaths", allowedpaths);
    if(isFilenameWithinDirectories(dir_plus_filename, allowedpaths)) {
      return true;
    } else {
      ERROR_STREAM("Given filename " << dir_plus_filename << " is not in allowed paths for writing " << allowedpaths << endl);
      return false;
    } 
  }

  const string SecurityCheck::getPasswd() {
    string passwd;
    nmmconfig.getValue("passwd", passwd);
    return passwd;
  }

  bool SecurityCheck::checkPasswd(const string& passwd) {
    string _passwd;
    nmmconfig.getValue("passwd", _passwd);
    if(passwd == _passwd) {
      return true;
    }
    else {
      ERROR_STREAM("Given password does not match" << endl);
      return false;
    }
  }

  void SecurityCheck::parseFile(const string& filename) {
    nmmconfig.parseFile(filename);
  }

  bool SecurityCheck::isFilenameWithinDirectories(const string& dir_plus_filename, const string& directories) {
    // no restrictions -> filename is okay
    if(directories.size() == 0) {
      return true;
    }

    // get directory only from directory plus filename
    string directory = getDirectoryPart(dir_plus_filename);
    if(directory == std::string()) {
      // for a simple file name, directory is the current working directory
      directory = getNormalizedDirectoryName(getCurrentWorkingDirectory());
    }

    // normalize directory name for comparison
    directory = getNormalizedDirectoryName(directory);

    // iterate over all given directories
    size_t curr_pos_in_directories = 0;
    while(true) {
      // test if there is more then one directory in directories (separated with a ';')
      size_t pos_curr_dir_end = directories.find(";", curr_pos_in_directories);
      // check only the current directory
      if(pos_curr_dir_end != string::npos) {
	string curr_directory = getNormalizedDirectoryName(directories.substr(curr_pos_in_directories, pos_curr_dir_end - curr_pos_in_directories));
	if(directory.find(curr_directory) == 0) {
	  return true;
	}
	else { // otherwise, move to next entry and check if we are done
	  curr_pos_in_directories = pos_curr_dir_end + 1;
	  if(curr_pos_in_directories >= directories.size()) {
	    return false;
	  }
	}
      }
      else { // only one directory in directories OR last directory
	string curr_directory = getNormalizedDirectoryName(directories.substr(curr_pos_in_directories, directories.size() - curr_pos_in_directories));
	if(directory.find(curr_directory) == 0) {
	  return true;
	}
	else {
	  return false;
	}
      }
    }
  }
}

DEFINE_SINGLETON(NMM::SecurityCheck);

#include "nmm/nmm_suffix.hpp"
