/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
 
 
#include "directory_utils.hpp"

#include <sys/stat.h>
#include <dirent.h>
#include <errno.h>

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
 /* FileLocation stuff */
  std::string getNMMEnvVar(std::string envvarname) {
    /* get the envvarname environment variable */
    char* envvar = getenv(envvarname.c_str());
    if (NULL != envvar) {
      /* Variable is set, use value */
      return std::string(envvar);
    } else {
      /* Variable is not set */
      return std::string("");
    }
  }

  std::string getNMMPrefixDir() {
    /* this return the define */
    return std::string(INSTALL_PREFIX);
  }

  std::string getNMMPkgdataDir() {
    /* this return the define */
    return std::string(INSTALL_PKGDATADIR);
  }

  std::string getNMMLibDir() {
    /* this return the define */
    return std::string(INSTALL_LIBDIR);
  }

  std::string ensureTrailingSlash(const std::string val) {
    if ('/' != val[val.length()-1]) {
      return val + std::string("/");
    }
    return val;
  }

  std::string getNMMDir() {
    std::string retval = getNMMEnvVar("NMM_DEV_DIR");
    if ("" == retval) {
      retval = getNMMEnvVar("NMM_DIR");
      if ("" == retval) {
	retval = getNMMPrefixDir();
      }
    }
    
    return ensureTrailingSlash(retval);
  }

  std::string getNMMResourceDir() {
    std::string retval = getNMMEnvVar("NMM_DEV_DIR");
    if ("" != retval) {
      return ensureTrailingSlash(retval) + std::string("resources/");
    }


    retval = getNMMEnvVar("NMM_DIR");
    if ("" != retval) {
      return ensureTrailingSlash(retval) + std::string("share/nmm/");
    }

    return ensureTrailingSlash(getNMMPkgdataDir());
  }


  std::string getNMMResource(const std::string res) {
    return getNMMResourceDir() + res;
  }

  std::string getNMMPluginDir() {
    std::string retval = getNMMEnvVar("NMM_DEVLIB_DIR");
    if ("" != retval) {
      return ensureTrailingSlash(retval);
    }

    retval = getNMMEnvVar("NMM_DEV_DIR");
    if ("" != retval) {
      return ensureTrailingSlash(retval) + std::string("dev-lib/");
    }


    retval = getNMMEnvVar("NMM_DIR");
    if ("" != retval) {
      return ensureTrailingSlash(retval) + std::string("lib/");
    }

    return ensureTrailingSlash(getNMMLibDir()) + std::string("nmm/dev-lib");
  }

  std::string getNMMProjectPluginDir() {
    std::string retval = getNMMEnvVar("NMM_PROJECT_DIR");
    if ("" != retval) {
      return ensureTrailingSlash(retval);
    }

    return ensureTrailingSlash(getNMMPluginDir());
  }

  std::string getNMMLocaleDir() {
    std::string retval = getNMMEnvVar("NMM_DEV_DIR");
    if ("" != retval) {
      // if NMM_DEV_DIR is set use $NMM_DEV_DIR/resources/locale/
      return ensureTrailingSlash(retval) + std::string("resources/locale/");
    }


    retval = getNMMEnvVar("NMM_DIR");
    if ("" != retval) {
      // if NMM_DIR is set use $NMM_DIR/share/nmm/locale/
      // warning: if you change the installation dir in configure in
      // (defined there in localedir variable), you have to adapt the
      // std::string below (and the comment above :-) )
      return ensureTrailingSlash(retval) + std::string("share/nmm/locale/");
    }

    // there is no function, that simply returns the define!!!
    return ensureTrailingSlash(std::string(INSTALL_LOCALEDIR));
  }

  std::string getNMMConfigFilename()
  {
    return ".nmmrc";
  }

  std::string getAbsolutePath(const std::string& filename)
  {
    std::string tmp = filename;
    if(tmp.substr(0,2) == "./") {
      char* working_dir = new char[PATH_MAX];
      getcwd(working_dir, PATH_MAX);
      tmp = working_dir + tmp.substr(1, tmp.size() - 1);
      delete working_dir;
    }
    else if((tmp.substr(0,1) != "/") && (tmp.substr(1,2) != ":\\")) {
      char* working_dir = new char[PATH_MAX];
      getcwd(working_dir, PATH_MAX);
      tmp = working_dir + std::string("/") + tmp;
      delete working_dir;
    }
    return tmp;
  }
  
  std::string getUserHomeDirectory() {
    char* username = ::getenv("LOGNAME");
    char* home = getenv("HOME");
    std::string home_directory;
    if(home) {
      home_directory = home;
    } else {
      if(username && std::string(username) == "root") {
        home_directory = "/root";
      } else {
        home_directory = std::string("/home/") + username;
      }    
    }
    
    return home_directory;
  }
  
  std::list<std::string> getAllFilesInDirectory(const std::string& folder)
    throw(const Exception&)
  {
    struct dirent **namelist;
    struct stat buffer;
    
    std::list<std::string> file_list;
    
    int count = scandir(folder.c_str(), &namelist, 0, alphasort);
    std::string tmp;
    if (count < 0) {
      /* if the directory can not be accessed throw an exception */
      throw Exception("Scandir for directory " + folder + " failed ");
    } else {
      for (int i = 0; i< count; i++) {
        tmp = folder + "/" + namelist[i]->d_name;
        if (stat(tmp.c_str(), &buffer) == 0) {
          if (S_ISREG(buffer.st_mode)) {
            file_list.push_back(namelist[i]->d_name);
          }
        }
      }
    }

    free(namelist);
    return file_list;
  }

  std::string getCurrentWorkingDirectory() throw() {
    char* ret = 0;
    size_t size = PATH_MAX;
    std::string wd;
    do {
      char* buf = new char[size];
      ret = getcwd(buf, size);
      if(ret) {
        wd = buf;
      } else {
        size += 128;
      }
      delete[] buf;
    } while(ret == 0);
    return wd;
  }

  std::string getAbsolutePath(std::string path) throw() {
    if(path.substr(0,2) == "./") {
      path = getCurrentWorkingDirectory() + path.substr(1, path.size() - 1);
    } else if((path.substr(0,1) != "/") && (path.substr(1,2) != ":\\")) {
      path = getCurrentWorkingDirectory() + std::string("/") + path;
    }
    return path;
  }

  std::string getAbsoluteNativePath(std::string path) throw() {
    if(path.substr(0,2) == "./") {
      path = getCurrentWorkingDirectory() + path.substr(1, path.size() - 1);
    } else if(path.substr(0,1) != "/") {
      path = getCurrentWorkingDirectory() + std::string("/") + path;
    }
    return path;
  }

  void createDirectory(const std::string& directory) throw(const Exception&) {
    if (opendir(directory.c_str()) == NULL) {
      if (errno == ENOENT) {
        if ( mkdir(directory.c_str(),0777) != 0) {
          throw Exception(std::string("Could not create directory ") + directory);
        }
      }
    }
    /* End< */
  }

  std::string getNormalizedFileName(const std::string& filename)
  {
    return filename;
  }

  std::string getNormalizedDirectoryName(const std::string& dirname)
  {
    return ensureTrailingSlash(getNormalizedFileName(dirname));
  }

  std::string getDirectoryPart(const std::string& filename)
  {
    size_t pos_last_slash = filename.rfind("/", filename.size() - 1);

    // if a slash was found, take sub-string only
    if(pos_last_slash != std::string::npos) {
      return filename.substr(0, pos_last_slash);
    }
    else {
      return std::string();
    }
  }

}


#include "nmm/nmm_suffix.hpp"
