/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "imageconverter_utils.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  unsigned int yuv2rgb( int Y, int Cb, int Cr ) {
    int Ey, Epb, Epr;
    int Eg, Eb, Er;
    unsigned int result;

    Ey  = (Y-16);
    Epb = (Cb-128);
    Epr = (Cr-128);

    Eg = (298*Ey - 100*Epb - 208*Epr)/256;
    Eb = (298*Ey + 516*Epb)/256;
    Er = (298*Ey + 408*Epr)/256;

    if(Eg > 255)
      Eg = 255;
    if(Eg < 0)
      Eg = 0;

    if(Eb > 255)
      Eb = 255;
    if(Eb < 0)
      Eb = 0;

    if(Er > 255)
      Er = 255;
    if(Er < 0)
      Er = 0;

    result = (Eb << 16) | (Eg << 8) | Er;

    return result;
  }

  void convertYV12ToRGB24(unsigned char* rgb_image, unsigned char* yv12_image,
			unsigned int width, unsigned int height, bool use_alpha) {
    unsigned int framesize, x, y;

    unsigned char* y_plane;
    unsigned char* u_plane;
    unsigned char* v_plane;
    unsigned char* a_plane;
    unsigned int rgb;
    unsigned width_h = width>>1;
    unsigned int offset;
    unsigned char* yc;
    unsigned char* uc;
    unsigned char* vc;
    unsigned char* ac;
    unsigned char* dest = rgb_image;
    
    framesize = width*height;
    y_plane = yv12_image;
    u_plane = yv12_image+framesize;
    v_plane = u_plane+(framesize>>2);
    if ( use_alpha ) {
      a_plane = v_plane+(framesize>>2);
    }

    yc = y_plane;
    ac = a_plane;

    for( y=0; y<height; y++ ) {
      offset = width_h*(y>>1);
      uc = u_plane + offset;
      vc = v_plane + offset;
      for( x=0; x<width; x++ ) {

        rgb = yuv2rgb( *yc, *uc, *vc );
	if (use_alpha) {
          
	  unsigned char alpha = *ac;
	  *dest++ = ((rgb & 0xff) * alpha) / 255;
	  *dest++ = (((rgb>>8) & 0xff) * alpha) / 255;
	  *dest++ = (((rgb>>16) & 0xff) * alpha) / 255;
	  ac++;
        } else {
	  *dest++ = rgb & 0xff;
	  *dest++ = (rgb>>8) & 0xff;
	  *dest++ = (rgb>>16) & 0xff;
	}

	yc++;
	if (x%2) { uc++; vc++; }
      }
    }
  }

  void convertYV12ToRGB32(unsigned char* rgb_image, unsigned char* yv12_image,
			unsigned int width, unsigned int height, bool use_alpha) {
    unsigned int framesize, x, y;

    unsigned char* y_plane;
    unsigned char* u_plane;
    unsigned char* v_plane;
    unsigned char* a_plane;
    unsigned int rgb;
    unsigned width_h = width>>1;
    unsigned int offset;
    unsigned char* yc;
    unsigned char* uc;
    unsigned char* vc;
    unsigned char* ac;
    unsigned char* dest = rgb_image;
    
    framesize = width*height;
    y_plane = yv12_image;
    u_plane = yv12_image+framesize;
    v_plane = u_plane+(framesize>>2);
    if ( use_alpha ) {
      a_plane = v_plane+(framesize>>2);
    }

    yc = y_plane;
    ac = a_plane;

    for( y=0; y<height; y++ ) {
      offset = width_h*(y>>1);
      uc = u_plane + offset;
      vc = v_plane + offset;
      for( x=0; x<width; x++ ) {

        rgb = yuv2rgb( *yc, *uc, *vc );
	if (use_alpha) {
	  unsigned char alpha = *ac;
	  *dest++ = ((rgb & 0xff) * alpha) / 255;
	  *dest++ = (((rgb>>8) & 0xff) * alpha) / 255;
	  *dest++ = (((rgb>>16) & 0xff) * alpha) / 255;
          *dest++;
	  ac++;
        }else{
	  *dest++ = rgb & 0xff;
	  *dest++ = (rgb>>8) & 0xff;
	  *dest++ = (rgb>>16) & 0xff;
          *dest++;
	}

	yc++;
	if (x%2) { uc++; vc++; }
      }
    }
  }

  void convertRGBX32toRGB24InPlace(unsigned char* image,
				   unsigned int width,
				   unsigned int height ) {
    unsigned char* sp = image;
    unsigned char* dp = image;

    for (unsigned int i = 0; i < height; i++) {
      for (unsigned int j = 0; j < width; j++) {
	*dp++ = *sp++;
	*dp++ = *sp++;
	*dp++ = *sp++;
	sp++;
      }
    }
  }

}

#include "nmm/nmm_suffix.hpp"
