/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include "INetAddr.hpp"
#include "nmm/Exception.hpp"

#include "nmm_netutils.hpp"

#include <sstream>
#include <string>
#include <vector>
#include <errno.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <net/if.h>
#include <iostream>
#include <ifaddrs.h>

#include "nmm/nmm_prefix.hpp"
using namespace std;
namespace NMM {
  
  INetAddr::INetAddr(unsigned long addr, unsigned int port) {
    memset (&_address, 0, sizeof _address);
    _address.sin_family      = AF_INET;

    setPort(port);
    setAddress(addr);
    setTransportProtocol("tcp");
  }

  INetAddr::INetAddr(const char* addr, unsigned int port) {
    memset (&_address, 0, sizeof _address);
    _address.sin_family      = AF_INET;

    setPort(port);
    setAddress(addr);
    setTransportProtocol("tcp");
  }

  INetAddr::INetAddr() {
    memset (&_address, 0, sizeof _address);
    _address.sin_family      = AF_INET;

    setPort(0);
    setAddress(static_cast<unsigned long>(0)); //static_cast avoids ambigious overload with setAddress(const char*)
    setTransportProtocol("tcp");
  }

  INetAddr::INetAddr(const INetAddr& addr) {
    *this = addr;
  }

  INetAddr::~INetAddr(){}

  void INetAddr::setAddress(const char* addr) {
    struct hostent* host     = gethostbyname(addr);
    if (host) {
      memcpy(&_address.sin_addr.s_addr, host->h_addr_list[0], sizeof(unsigned long));
      _ip_address = getIpAddress();
    }
  }

  void INetAddr::setAddress(unsigned long addr) {
    /* convert the given addr to the network byte order */
    _address.sin_addr.s_addr = htonl(addr);
    _ip_address = addr;
  }
  
  void INetAddr::setAddress(struct sockaddr_in* addr) {
    memcpy(&_address, addr, sizeof(_address));
    _ip_address = ntohl(_address.sin_addr.s_addr);
    _port = ntohs(_address.sin_port);
  }
  
  std::string INetAddr::toString() const {
    std::ostringstream s;
    s << std::string(inet_ntoa(_address.sin_addr));
    s << std::string(":");
    s << _port;
    return s.str();
  }
  
  void INetAddr::updateAttributes() {
    setAddress(_ip_address);
    setPort(_port);
    setTransportProtocol(_trans_prot.c_str());
  }


  static unsigned long broadcast_address = 0;
  
  unsigned long INetAddr::getReceivingBroadcastAddress() {
  
    // if broadcast_address is != 0 we already read this address from the network device
    if (broadcast_address != 0) {
	  return broadcast_address;
	}

    /* If not, first get the hostname */
    std::vector<char> tmp(255, '0');
    memset(&tmp[0], 0x0, 255);

    if (gethostname(&tmp[0], 255) == -1) {
      switch(errno) {
        case EINVAL        : throw Exception("length of hostname is  negative  or,  for  sethostname,  length is larger than the maximum allowed size, or length is smaller than the actual size.");

        case EPERM         : throw Exception("For sethostname, the caller was not the superuser.");

        case EFAULT        : throw Exception("name is an invalid address.");

        case ENAMETOOLONG  : throw Exception("length is larger than the maximum allowed size, or length is smaller than the actual size.");

        default            : throw Exception("Unkown error");
      }
    }
    std::string __hostname = std::string(&tmp[0]);

    /* Now get the IP-Adress for this hostname */
    INetAddr receiving_addr;
    receiving_addr.setAddress(__hostname.c_str());

    struct ifaddrs *ifa = NULL;
    //get all interfaces 
    getifaddrs(&ifa);
 
    // find the interface that provides our address
    for (struct ifaddrs *i = ifa; i ; i = i->ifa_next) {
      // Check if the devices uses our ip address
      if ( ntohl(reinterpret_cast<sockaddr_in*>(i->ifa_addr)->sin_addr.s_addr) == receiving_addr.getIpAddress()) {
        // If so, check if there is a valid network address available
        if ( i->ifa_netmask ) { 
          in_addr addr;
          //Calculate the broadcast address 
          addr.s_addr = (reinterpret_cast<sockaddr_in*>(i->ifa_addr)->sin_addr.s_addr | ~(reinterpret_cast<sockaddr_in*>(i->ifa_netmask)->sin_addr.s_addr));

          broadcast_address = ntohl(addr.s_addr); 
          break;
        }                     
      }
    }

    freeifaddrs(ifa);
    
    if (broadcast_address == 0) {
      throw Exception("Could not get broadcast address");
    }

    return broadcast_address;
  }
  
  unsigned long INetAddr::getSendingBroadcastAddress() {
    return  getReceivingBroadcastAddress();
  }
}

#include "nmm/nmm_suffix.hpp"
