/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include "NetReactor.hpp"
#include "nmm/utils/thread/ThreadPool.hpp"

#include <cerrno>

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  NetReactor::NetReactor()
    : NamedObject("NetReactor"), __handler_mutex(ThreadMutex::RECURSIVE), __handler_cond_mutex(ThreadMutex::RECURSIVE),
      __timeout(10), __pipe_socket(2), __internal_char('a'), __internal_handler(*this),
      __reactor_task(this, &NetReactor::run ), __leader_thread(0)
  {
    if(pipe(&__pipe_socket[0]) != 0) {
      ERROR_STREAM("Could not create the internal pipe " << std::endl);
    }

    /*Begin > create a new poll-struct for the __socket_array*/
    __socket_array.push_back(pollfd());
    __socket_array.back().fd      = __pipe_socket[0];
    __socket_array.back().events  = NetReactor::NR_POLLERR | NetReactor::NR_POLLIN |NetReactor::NR_POLLRDNORM | NetReactor::NR_POLLRDBAND | NetReactor::NR_POLLPRI;
    __socket_array.back().revents = 0;
    /*End <*/

    /*Begin> register the NetHandler for the new socket*/
    __handler.insert(pair<SocketHandle, NetHandler*>(__pipe_socket[0], &__internal_handler));
    /*End<*/

    /* activate the thread */
    if (__is_running)
      requestLeaderThread();
  }

  NetReactor::~NetReactor()  {
    /* Begin> Set __is_running to false and stop the thread */
    __is_running = false;
    internalLock();
    internalUnlock();
    /*End< */
  
    /* Begin> Close the internal pipe */
    close (__pipe_socket[0]);
    close (__pipe_socket[1]);
    /* End< */
  }

  void NetReactor::run() {
    int retval = 0;
    while(__is_running) {
      __handler_mutex.lock();
      /* Begin> Call poll on registered sockets. During this time no socket can be removed */
      retval = poll(&__socket_array[0], __socket_array.size(), __timeout);
      /* End< */

      /* Begin> If changes on a socket occuerd call update*/
      if (retval > 0 ) {
	if (__is_running) {
	  if ( !checkSockets() )  { 
	    __handler_mutex.unlock();
	  } else {
	    /* Another thread is the leader and we must call break */
	    break;         
	  }
	} else {
	  __handler_mutex.unlock();
	}
      }

      if (retval < 0) {
	ERROR_STREAM( getError(errno) << endl );
	__handler_mutex.unlock();
      }

      if (retval == 0) {
	__handler_mutex.unlock();
      }
      /* End< */
    }
  }

  bool NetReactor::isSocketRegisterd(SocketHandle socket) {
    for (unsigned int i = 0; i < __socket_array.size(); ++i)
      if (__socket_array[i].fd == socket)
	return true;

    return false;
  }

  void NetReactor::changeEvents(SocketHandle socket, short events) {
    for (unsigned int i = 0; i<__socket_array.size(); ++i)
      if (__socket_array[i].fd == socket) {
	__socket_array[i].events = events;
      }
  }

  void NetReactor::registerSocket(SocketHandle socket,NetHandler* handler, short events) {
    /*Begin> Test if NetHandler is valid*/
    if (!handler) {
      ERROR_STREAM("Invalid NetHandler (0-ptr). Can not register a socket without a NetHandler" << endl);
      return;
    }
    /*End<*/

    /* lock the NetReactor */
    internalLock();

    /* Register the socket */
    internalRegisterSocket(socket, handler, events);

    /* unlock the NetReactor */  
    internalUnlock();
  }

  void NetReactor::internalRegisterSocket(SocketHandle socket,NetHandler* handler, short events) {
    /*Begin> Test if NetHandler is valid*/
    if (!handler) {
      ERROR_STREAM("Invalid NetHandler (0-ptr). Can not register a socket without a NetHandler" << endl);
      return;
    }
    /*End<*/

    /* Begin> Check if socket was already registered */
    for (unsigned int i = 0; i < __socket_array.size(); ++i)
      if (__socket_array[i].fd == socket )  {
	return;
      }

    /*Begin > add the error flags to the events, so they can be checked against revents in struct pollfd */
    events |= POLLERR;
    events |= POLLHUP;
    events |= POLLNVAL;
    /*End< */

    /*Begin > create a new poll-struct for the __socket_array*/
    __socket_array.push_back(pollfd());
    __socket_array.back().fd      = socket;
    __socket_array.back().events  = events;
    __socket_array.back().revents = 0;
    /*End <*/

    /*Begin> register the NetHandler for the new socket*/
    map<SocketHandle, NetHandler*>::iterator it = __handler.find( socket );
    if ( it == __handler.end() ) {
      __handler.insert(pair<SocketHandle, NetHandler*>(socket, handler));
    } else {
      if ( it->second != handler ) {
	__handler.erase(it);
      }
      __handler.insert(pair<SocketHandle, NetHandler*>(socket, handler));       
    }
    /*End<*/
  }


  void NetReactor::internalLock() {
    /* Write the byte in the internal pipe */
    __handler_cond_mutex.lock();
    write(__pipe_socket[1], &__internal_char, 1);
    __handler_mutex.lock();
  }

  void NetReactor::internalUnlock() {
    __handler_mutex.unlock();
    __handler_cond_mutex.unlock();
  }

  bool NetReactor::checkSockets() {
    for ( unsigned int i = 0; i< __socket_array.size(); ++i ) {
      /* Begin> check if an registered event is occured */
      if ( (__socket_array[i].revents & __socket_array[i].events) != 0) {
	if (i == 0) {
	  /* The first socket is the pipe which is used to wake-up the poll method call
	   * This socket must not be removed from our internal array
	   */
	  __internal_handler.receiveEvent(__socket_array[0].fd, __socket_array[0].events, __socket_array[0].revents );
	  __socket_array[i].revents = 0; //reset the revents
	} else {
	  map<SocketHandle, NetHandler*>::iterator it = __handler.find(__socket_array[i].fd);
	  if ( it != __handler.end()) {
	    if (it->second->requireThread()) {
	      /* Copy all required information, because we must unlock the mutex */
	      int revents    = __socket_array[i].revents;
	      int events     = __socket_array[i].events;
	      SocketHandle socket = __socket_array[i].fd;
	      NetHandler* handler = it->second;                    

	      /* Unregister the socket to avoid a second handling for the same socket */          
	      internalUnregisterSocket(socket);
	      /* request a new leader, because we need this thread */
	      requestLeaderThread();
	      /* Unlock the mutex, so that the remaining sockets can be handled*/
	      __handler_mutex.unlock();

	      /* Inform the handler about the events */
	      handler->receiveEvent(socket, events, revents);

	      /* return false, because this thread can not be reused by the run method */
	      return true;
	    } else {
	      /* Infrom the handler about occured events */
	      NetHandler::NetResult result = it->second->receiveEvent(__socket_array[i].fd, __socket_array[i].events, __socket_array[i].revents);
	      /* reset the events */
	      __socket_array[i].revents = 0;
	      /* Remove the socket if the Hanlder returns remove */
	      /* We can call this internal-method because the mutex is still locked */
	      if (result == NetHandler::REMOVE ) {
		internalUnregisterSocket(__socket_array[i].fd);
	      }
	    }
	  } else {
	    ERROR_STREAM("No handler registered for a socket " << endl);
	    ERROR_STREAM("Unregister this socket " << endl);
	    /* ERROR-HANDLING!!! > If no handler for this socket available, remove it! This must not happen!!*/
	    vector<pollfd>::iterator it = __socket_array.begin();
	    advance(it, i);
	    __socket_array.erase(it);
	    /* ERROR-Handling< */
	  }
	}
	/* End< */
      }
    }
    /* We can reuse the Thread, and return false */
    return false;
  }

  void NetReactor::requestLeaderThread() {
    __leader_thread = SThreadPool::getInstance().requestThread();
    __leader_thread->setTask(&__reactor_task);
    __leader_thread->iterateTask();
  }


  void NetReactor::unregisterSocket(SocketHandle socket) {
    /* Lock the NetReactor */
    internalLock();

    /* Unregister the socket */
    internalUnregisterSocket(socket);
  
    /* Unlock the NetReactor */
    internalUnlock();
  }

  void NetReactor::internalUnregisterSocket(SocketHandle socket) {
    /* Begin> remove the socket from socket array */
    bool socket_reg = false;
    if (__socket_array.size() > 0) {
      for (unsigned int i = 0; i< __socket_array.size(); ++i) {
	if (__socket_array[i].fd == socket) {
	  vector<pollfd>::iterator it = __socket_array.begin();
	  advance(it, i);        
	  __socket_array.erase(it);
	  socket_reg = true;
	  break;
	}
      }
    }
    /* End< */

    /* Begin> remove the NetHandler */
    __handler.erase(socket);
    /* End< */
  }

  void NetReactor::setTimeout(int timeout) {
    /* A value <= 0 the value is not used to avoid inifinte waiting */
    if (timeout > 0)
      __timeout = timeout;
  }

  const string& NetReactor::getError(int __errno) {
    switch (__errno) {
    case EBADF  : return EBADF_error;
    case ENOMEM : return ENOMEM_error;
    case EFAULT : return EFAULT_error;
    case EINTR  : return EINTR_error;
    default     : return unknown_error ;
    }
  }

  NetReactor::InternalNetHandler::InternalNetHandler(NetReactor& reactor)
    : __receive_signal(NetReactor::NR_POLLIN |NetReactor::NR_POLLRDNORM | NetReactor::NR_POLLRDBAND | NetReactor::NR_POLLPRI),
      __reactor(reactor)
  {

  }

  NetHandler::NetResult NetReactor::InternalNetHandler::receiveEvent(SocketHandle socket, short event, short revents)
  {
    if( (event & revents & __receive_signal) == 0) {
      GLOBAL_ERROR_STREAM2("Illegal network event" << std::endl);
      return VOID;
    }

    read(socket, &__internal_char, 1);

    __reactor.__handler_mutex.unlock();

    /* Wait until the register mehtod changes the fd-array*/
    __reactor.__handler_cond_mutex.lock();

    /* Unlock the mutex again */
    __reactor.__handler_mutex.lock();

    __reactor.__handler_cond_mutex.unlock();
    return VOID;
  }

  bool NetReactor::__is_running = true;
  const string NetReactor::EBADF_error  = "An invalid file descriptor was given in one of the sets.";
  const string NetReactor::ENOMEM_error = "There was no space to allocate file descriptor tables.";
  const string NetReactor::EFAULT_error = "The array given as argument was not contained in the calling program's address space.";
  const string NetReactor::EINTR_error  = "A signal occurred before any requested event.";
  const string NetReactor::unknown_error = "Unknown error occurd";
}

DEFINE_SINGLETON(NMM::NetReactor);

#include "nmm/nmm_suffix.hpp"
