/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include "SockAcceptor.hpp"
#include "INetAddr.hpp"

#include "nmm/utils/NamedObject.hpp"

#include <cerrno>

#include "nmm_netutils.hpp"

#include <fcntl.h> // for ftntl-controls

#include <sys/socket.h>

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  /****************** class SockAcceptor ******************/

  SockAcceptor::SockAcceptor()
    : __socket(-1)
  {
  }
  
  SockAcceptor::~SockAcceptor() {
    if (__socket != -1) {
      ::close(__socket);
      __socket = -1;
    }
  }
  
  bool SockAcceptor::isAcceptable() {
    if (__trans_prot == "tcp") {
      fd_set fds;
      struct timeval tv;
      
      FD_ZERO(&fds);
      FD_SET(__socket, &fds);
      tv.tv_sec = 0;
      tv.tv_usec = 1000;
      
      if (0 < select(__socket, &fds, &fds, NULL, &tv)) {
        return true;
      }
      
      return false;
    } else if (__trans_prot == "udp") {
      /* UDP accept always succeeds if the SockAcceptor is opened */
      return (__socket >= 0);
    } else {
      return false;
    }
  }
  
  void SockAcceptor::setNonBlocking() {
    if (__trans_prot == "tcp") {
      fcntl(__socket, F_SETFL, FNDELAY);
    }
  }
  
  SocketHandle SockAcceptor::getSocket() {
    return __socket;
  }
  
  void SockAcceptor::accept(Sock& s) throw (const ConnectionException&){
    if (__socket == -1) {
      GLOBAL_ERROR_STREAM("SockAcceptor::accept called, but SockAcceptor was not opened!" << endl);
      throw ConnectionException("SockAcceptor not opened", 0, ConnectionException::accept);
    }
    
    if(__trans_prot == "tcp") {
      /* accept a new connection and store the socket in the given stream */
      sockaddr_in peer_addr;
      socklen_t peer_addr_len = sizeof(peer_addr);
      
      int _tmp = ::accept(__socket, reinterpret_cast<sockaddr*>(&peer_addr), &peer_addr_len);
      if (_tmp == -1) {
        /* If acception failes, throw an exception */
        createAcceptException(errno);
      }
      
      /* Set peer address */
      INetAddr inet_addr;
      inet_addr.setTransportProtocol(__trans_prot);
      inet_addr.setAddress(&peer_addr);
      s.setPeerAddress(inet_addr);
      
      /* Set socket */
      s.setSocket(_tmp, true);
    } else if(__trans_prot == "udp") {
      /* Set socket */
      s.setSocket(__socket, true);
      
      /* Set socket to -1, SockAcceptor must be opened before accept can be called again */
      __socket = -1;
    }
    else {
      throw ConnectionException("Invalid protocol", 0, ConnectionException::accept);
    }
  }

  void SockAcceptor::open(const INetAddr& addr, const char* proto) throw (const ConnectionException&) {
    struct protoent* protocol = getprotobyname(proto);
    if (protocol == 0) {
      throw ConnectionException("Unknown protocol", 0, ConnectionException::open);
    }
    
    if (__socket != -1) {
      ::close(__socket);
      __socket = -1;
    }
    
    __trans_prot = proto;
    
    /* Begin> Handle TCP-Connection */
    if(strcmp(proto, "tcp") == 0) {
      /*Create a new socket for the given address */
      __socket = ::socket(PF_INET, SOCK_STREAM, protocol->p_proto);
      if (__socket == -1) {
        /*Create an exception*/
        createSocketException(errno);
      }
      
      /* set option for reusing (only for testing/development, should be removed sometime) */
      int on = 1;
      if ( setsockopt ( __socket, SOL_SOCKET, SO_REUSEADDR, ( const char* ) &on, sizeof ( on ) ) == -1 ) {
	GLOBAL_ERROR_STREAM2("SO_REUSEADDR failed in SockAcceptor" << endl);
      }
      else {
	GLOBAL_DEBUG_STREAM2("SO_REUSEADDR set for SockAcceptor" << endl);
      }
      
      /* Try to bind the socket to the given address */
      if (::bind(__socket, addr.getAddress(), addr.getSize()) == -1)
        /*Create a BindingException*/
        createBindException(errno);

      /* Try to start listening on the new socket */
      if (::listen(__socket, SOMAXCONN) == -1)
        /*Create a ListeningException */
        createListenException(errno);
    }
    /* End< */
    else if(strcmp(proto, "udp") == 0) {
      bool isMulticastAddress = IN_MULTICAST(addr.getIpAddress());
      __socket = ::socket(PF_INET, SOCK_DGRAM, protocol->p_proto);

      if (__socket == -1)
        createSocketException(errno);

      if (isMulticastAddress) {
	int on = 1;
	if ( setsockopt ( __socket, SOL_SOCKET, SO_REUSEADDR, ( const char* ) &on, sizeof ( on ) ) == -1 ) {
	  GLOBAL_ERROR_STREAM2("SO_REUSEADDR failed in SockAcceptor" << endl);
	}
	else {
	  GLOBAL_DEBUG_STREAM2("SO_REUSEADDR set for SockAcceptor" << endl);
	}
      }

      if (::bind(__socket, addr.getAddress(), addr.getSize()) == -1)
        createBindException(errno);

      if (isMulticastAddress) {
	struct ip_mreq mreq;
	mreq.imr_multiaddr.s_addr = htonl(addr.getIpAddress());
	mreq.imr_interface.s_addr = htonl(INADDR_ANY);
	
	setsockopt(__socket, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq));
      } 
    }
    else {
      throw ConnectionException("Unknown protocol", 0, ConnectionException::open);
    }
  }

  void SockAcceptor::tryOpen(INetAddr& addr, u_int32_t from, u_int32_t to, const char* proto)
    throw (const ConnectionException&)
  {
    for (u_int32_t i = from; i <= to; ++i) {
      addr.setPort(i);
      try {
	this->open(addr, proto);
	return;
      } catch(const ConnectionException& ce) {
	if (i == to)
	  throw ce;
      }
    }
  }
  
  void SockAcceptor::close()
    throw (const ConnectionException&)
  {
    if (__socket != -1) {
      ::close(__socket);
    }
  }
  
  void SockAcceptor::createShutdownException(int my_errno) {
    switch(my_errno) {
    case EBADF    :
      throw ConnectionException("Value is not a valid descriptor.", my_errno, ConnectionException::shutdown);

    case ENOTSOCK :
      throw ConnectionException("Value is a file, not a socket.", my_errno, ConnectionException::shutdown);

    case ENOTCONN :
      throw ConnectionException("The specified socket is not connected.", my_errno, ConnectionException::shutdown);

    default:
      throw ConnectionException("Unknown Error", my_errno, ConnectionException::shutdown);
    }
  }

  void SockAcceptor::createBindException(int my_errno) {
    switch(my_errno) {
    case EBADF     :
      throw ConnectionException("sockfd is not a valid descriptor.", my_errno, ConnectionException::bind);
      
    case EINVAL    :
      throw ConnectionException("The socket is already bound to an address.", my_errno, ConnectionException::bind);

    case EACCES    :
      throw ConnectionException("The address is protected, and the user is not the super-user.", my_errno, ConnectionException::bind);

    case ENOTSOCK  :
      throw ConnectionException("Argument is a descriptor for a file, not a socket.", my_errno, ConnectionException::bind);

    case EROFS  :
      throw ConnectionException("The socket inode would reside on a read-only file system.", my_errno, ConnectionException::bind);

    case  EFAULT :
      throw ConnectionException("my_addr points outside the user's accessible address space.", my_errno, ConnectionException::bind);

    case   ENAMETOOLONG:
      throw ConnectionException("my_addr is too long.", my_errno, ConnectionException::bind);

    case   ENOENT:
      throw ConnectionException("The file does not exist.", my_errno, ConnectionException::bind);

    case   ENOMEM:
      throw ConnectionException("Insufficient kernel memory was available.", my_errno, ConnectionException::bind);

    case   ENOTDIR:
      throw ConnectionException("A component of the path prefix is not a directory.", my_errno, ConnectionException::bind);
    case   EADDRINUSE:
      throw ConnectionException("Another socket is already listening on the same port.", my_errno, ConnectionException::bind);

    case   ELOOP:
      throw ConnectionException("  Too many symbolic links were encountered in resolving my_addr.", my_errno, ConnectionException::bind);
    default        :
      throw ConnectionException("Unknown Error occured", my_errno, ConnectionException::bind);
    }
  }

  void SockAcceptor::createListenException(int my_errno) {
    switch(my_errno) {
    case EBADF      :
      throw ConnectionException("The argument s is not a valid descriptor.", my_errno, ConnectionException::listen);

    case ENOTSOCK   :
      throw ConnectionException("The argument s is not a socket.", my_errno, ConnectionException::listen);

    case EOPNOTSUPP :
      throw ConnectionException("The socket is not of a type that supports the listen operation.", my_errno, ConnectionException::listen);

    default :
      throw ConnectionException("Unknown error.", my_errno, ConnectionException::listen);
    }
  }

  void SockAcceptor::createAcceptException(int my_errno) {
    switch(my_errno) {
    case EAGAIN       :
      throw ConnectionException("The socket is marked non-blocking and no connections are present to be accepted.", my_errno, ConnectionException::accept);

    case EBADF       :
      throw ConnectionException("The descriptor is invalid.", my_errno, ConnectionException::accept);

    case ENOTSOCK    :
      throw ConnectionException("The descriptor references a file, not a socket.", my_errno, ConnectionException::accept);

    case EOPNOTSUPP  :
      throw ConnectionException("The referenced socket is not of type SOCK_STREAM.", my_errno, ConnectionException::accept);

    case EFAULT      :
      throw ConnectionException("The addr parameter is not in a writable part of the user address space.", my_errno, ConnectionException::accept);

    case EPERM       :
      throw ConnectionException("Firewall rules forbid connection.", my_errno, ConnectionException::accept);

    case ENOBUFS     :
      throw ConnectionException("Not enough free memory.", my_errno, ConnectionException::accept);

    case ENOMEM      :
      throw ConnectionException("Not enough free memory.", my_errno, ConnectionException::accept);

    default          :
      throw ConnectionException("Unknwon error.", my_errno, ConnectionException::accept);
    }
  }

  void SockAcceptor::createSocketException(int my_errno) {
    switch(my_errno) {
    case EPROTONOSUPPORT :
      throw ConnectionException("The protocol type or the specified protocol is not supported within this domain.", my_errno, ConnectionException::socket);

    case ENFILE :
      throw ConnectionException("Not enough kernel memory to allocate a new socket structure.", my_errno, ConnectionException::socket);

    case EMFILE :
      throw ConnectionException("Process file table overflow.", my_errno, ConnectionException::socket);

    case EACCES :
      throw ConnectionException("Permission to create a socket of the specified type and/or protocol is denied.", my_errno, ConnectionException::socket);

    case ENOBUFS:
      throw ConnectionException("Insufficient memory is available. The socket cannot be created until sufficient resources are freed.", my_errno, ConnectionException::socket);

    case ENOMEM :
      throw ConnectionException("Insufficient memory is available. The socket cannot be created until sufficient resources are freed.", my_errno, ConnectionException::socket);

    case EINVAL :
      throw ConnectionException("Unknown protocol, or protocol family not available.", my_errno, ConnectionException::socket);
    }
  }

}

#include "nmm/nmm_suffix.hpp"
