/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "SockConnector.hpp"
#include "SockStream.hpp"
#include "SockDGram.hpp"
#include "nmm/utils/NamedObject.hpp"

#include <cerrno>

#include "nmm_netutils.hpp"

#include <netinet/tcp.h> // for TCP_NODELAY

#include <sys/socket.h>
#include <sys/fcntl.h>

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  /****************** class SockConnector ******************/
  SockConnector::SockConnector()
    : __connect_timeout(2, 0)
  {}

  SockConnector::~SockConnector() {}

  Sock* SockConnector::connect(const INetAddr& addr, const char* proto) throw(const ConnectionException&)
  {
    /* create a socket object */
    Sock* new_sock = 0;
    if(addr.getTransportProtocol() == "tcp") {
      /* create a stream socket object */
      new_sock = new SockStream();
    }
    else if(addr.getTransportProtocol() == "udp") {
      /* create a datagram socket object */
      new_sock = new SockDGram();
    }
    else {
      throw ConnectionException("Unknown transport protocol", 0, ConnectionException::connect);
    }

    /* establish connection */
    try {
      connect(*new_sock, addr, proto);
    }
    catch(const ConnectionException& e) {
      /* destroy the new socket object and rethrow exception */
      delete new_sock;
      throw;
    }

    return new_sock;
  }

  void SockConnector::connect(Sock& sock, const INetAddr& addr, const char* proto) throw(const ConnectionException&)
  {
    /* establish connection */
    SocketHandle new_socket = internalConnect(sock.getSocket(), addr, proto);

    /* If internalConnect created a new socket, pass it to the Sock object.
       Otherwise keep using the old socket handle because it may already be bound to a port! */
    if(new_socket != sock.getSocket()) {
      sock.setSocket(new_socket, true);
    }

    /* Set peer address */
    sock.setPeerAddress(addr);
  }

  SocketHandle SockConnector::internalConnect(SocketHandle socket, const INetAddr& addr, const char* proto) throw(const ConnectionException&)
  {
    /* request a protoent-struct for the given protocol */
    struct protoent* protocol = getprotobyname(proto);
    if (protocol == 0) {
      /* create a new socket */
      throw ConnectionException("Unknown Protocol", 0, ConnectionException::connect);
    }

    /* Initialize the socket */
    SocketHandle new_socket = socket;

    if(addr.getTransportProtocol() == "tcp") {
      /* create a new socket */
      new_socket = ::socket(PF_INET, SOCK_STREAM, protocol->p_proto);

      /* test the new socket and create an exception if an error occurs */
      if (new_socket == -1) {
        /* If an error occurs create an exception */
        createSocketException(errno);
      }

      /* set socket to non-blocking mode to achieve connect with a specific timeout */
      int flags = fcntl(new_socket, F_GETFL, 0);
      if(fcntl(new_socket, F_SETFL, flags | O_NONBLOCK) < 0) {
        /* fcntl failed */
        GLOBAL_ERROR_STREAM2("Failed to set socket to non-blocking mode for connect()" << std::endl);
        ::close(new_socket);
        createConnectionException(errno);
      }

      /* test the new socket and create an exception if an error occurs */
      if(::connect(new_socket, addr.getAddress(), addr.getSize()) == -1) {
        if(errno == EINPROGRESS) {
          /* create a file descriptor set containing the new socket */
          fd_set fds;
          FD_ZERO (&fds);
          FD_SET (new_socket, &fds);

          /* create the timeout value */
          struct timeval tval;
          tval.tv_sec = __connect_timeout.sec;
          tval.tv_usec = __connect_timeout.nsec / 1000;

          /* wait for socket to be writable (i.e. connect() finished) */
          int n = select(new_socket+1, NULL, &fds, NULL, &tval);
          if(n < 0) {
            /* failed to wait for connection */
            ::close(new_socket);
            createConnectionException(errno);
          }
          else if(n == 0) {
            /* connect() timed out */
            GLOBAL_ERROR_STREAM2("connect() timed out" << std::endl);
            ::close(new_socket);
            createConnectionException(ETIMEDOUT);
          }

          /* check for error on the socket */
          int error;
          socklen_t error_size = sizeof(error);
          if(getsockopt(new_socket, SOL_SOCKET, SO_ERROR, &error, &error_size) < 0) {
            /* failed to get error status */
            GLOBAL_ERROR_STREAM2("Failed to get error status of connect()" << std::endl);
            ::close(new_socket);
            createConnectionException(errno);
          }
          if(error != 0) {
            /* connect() has failed with the error code returned by getsockopt() */
            GLOBAL_ERROR_STREAM2("Connection failed" << std::endl);
            ::close(new_socket);
            createConnectionException(error);
          }
        }
        else {
          /* If an error occurs create an exception */
          GLOBAL_ERROR_STREAM2("Failed to initiate non-blocking connection" << std::endl);
          ::close(new_socket);
          createConnectionException(errno);
        }
      }

      /* set the socket back to blocking mode */
      if(fcntl (new_socket, F_SETFL, flags) < 0) {
        /* fcntl failed */
        GLOBAL_ERROR_STREAM2("Failed to set socket back to blocking mode after connect()" << std::endl);
        ::close(new_socket);
        createConnectionException(errno);
      }

      /* Begin> Set TCP_NODELAY */
      int value = true;
      int error = setsockopt (new_socket, IPPROTO_TCP, TCP_NODELAY, &value, sizeof (value));
      if (error == -1) {
        GLOBAL_ERROR_STREAM2("Could not disable tcp delay" << endl);
      }
      /* End< */
    }
    else if(addr.getTransportProtocol() == "udp") {
      int on = 1;
      bool isMulticastAddress = IN_MULTICAST(addr.getIpAddress());

      bool isBroadCastAddress = (addr.getIpAddress() == INetAddr::getSendingBroadcastAddress());

      if(isMulticastAddress || isBroadCastAddress) {
        /* we must create a new socket */
        new_socket = ::socket(AF_INET, SOCK_DGRAM, protocol->p_proto);

        if (new_socket == -1)
          createSocketException(errno);

        if (isMulticastAddress) {
          if ( setsockopt (new_socket, SOL_SOCKET, SO_REUSEADDR, ( const char* ) &on, sizeof ( on ) ) == -1 ) {
            GLOBAL_ERROR_STREAM2("SO_REUSEADDR failed in SockAcceptor" << endl);
          }
          else {
            GLOBAL_DEBUG_STREAM2("SO_REUSEADDR set for SockAcceptor" << endl);
          }
        }

        if (isBroadCastAddress) {
          if (setsockopt(new_socket, SOL_SOCKET, SO_BROADCAST, (const char* ) &on, sizeof(on)) == -1) {
            GLOBAL_ERROR_STREAM2("SO_BROADCAST failed in SockAcceptor" << endl);
          }
          else {
            GLOBAL_DEBUG_STREAM2("SO_BROADCAST set for SockAcceptor"<<endl);
          }
        }

      //Don't connect a broadcast socket. This mught work under Linux but not under MacOSX. However, it does not 
      //make sense to connect a broadcast socket
      if (!isBroadCastAddress) 
        if (::connect(new_socket, addr.getAddress(), addr.getSize()) == -1)
          createConnectionException(errno);

        if (isMulticastAddress) {
          struct ip_mreq mreq;
          mreq.imr_multiaddr.s_addr = htonl(addr.getIpAddress());
          mreq.imr_interface.s_addr = htonl(INADDR_ANY);

          setsockopt(new_socket, IPPROTO_IP, IP_ADD_MEMBERSHIP, &mreq, sizeof(mreq));
        }
      }
    }
    else {
      throw ConnectionException("Unknown transport protocol", 0, ConnectionException::connect);
    }

    /* return the old or new socket handle */
    return new_socket;
  }

  void SockConnector::createSocketException(int my_errno) {
    switch(my_errno) {
    case EPROTONOSUPPORT :
      throw ConnectionException("The protocol type or the specified protocol is not supported within this domain.", my_errno, ConnectionException::socket);

    case ENFILE :
      throw ConnectionException("Not enough kernel memory to allocate a new socket structure.", my_errno, ConnectionException::socket);

    case EMFILE :
      throw ConnectionException("Process file table overflow.", my_errno, ConnectionException::socket);

    case EACCES :
      throw ConnectionException("Permission to create a socket of the specified type and/or protocol is denied.", my_errno, ConnectionException::socket);

    case ENOBUFS:
      throw ConnectionException("Insufficient memory is available. The socket cannot be created until sufficient resources are freed.", my_errno, ConnectionException::socket);

    case ENOMEM :
      throw ConnectionException("Insufficient memory is available. The socket cannot be created until sufficient resources are freed.", my_errno, ConnectionException::socket);

    case EINVAL :
      throw ConnectionException("Unknown protocol, or protocol family not available.", my_errno, ConnectionException::socket);
    }
  }

  void SockConnector::createShutdownException(int my_errno) {
    switch(my_errno) {
    case EBADF    :
      throw ConnectionException("Value is not a valid descriptor.", my_errno, ConnectionException::shutdown);

    case ENOTSOCK :
      throw ConnectionException("Value is a file, not a socket.", my_errno, ConnectionException::shutdown);

    case ENOTCONN :
      throw ConnectionException("The specified socket is not connected.", my_errno, ConnectionException::shutdown);

    default:
      throw ConnectionException("Unknown Error", my_errno, ConnectionException::shutdown);
    }
  }



  void SockConnector::createConnectionException(int my_errno) {
    switch(my_errno) {
    case EBADF   :
      throw ConnectionException("Bad descriptor.", my_errno, ConnectionException::connect );

    case EFAULT  :
      throw ConnectionException("The socket structure address is outside the user's address space.", my_errno, ConnectionException::connect );

    case ENOTSOCK:
      throw ConnectionException("The descriptor is not associated with a socket.", my_errno, ConnectionException::connect );

    case EISCONN :
      throw ConnectionException("The socket is already connected.", my_errno, ConnectionException::connect );

    case ECONNREFUSED :
      throw ConnectionException("Connection refused at server.", my_errno, ConnectionException::connect );

    case ETIMEDOUT    :
      throw ConnectionException("Timeout while attempting connection. The server may be too busy  to  accept  new connections.", my_errno, ConnectionException::connect );

    case ENETUNREACH  :
      throw ConnectionException("Network is unreachable.", my_errno, ConnectionException::connect );

    case EADDRINUSE   :
      throw ConnectionException("Address is already in use.", my_errno, ConnectionException::connect );

    case EINPROGRESS  :
      throw ConnectionException("The  socket  is non-blocking and the connection cannot be completed immediately.", my_errno, ConnectionException::connect );

    case EALREADY     :
      throw ConnectionException("The socket is non-blocking and a previous connection attempt has  not  yet  been completed.", my_errno, ConnectionException::connect );

    case EAGAIN       :
      throw ConnectionException("No  more  free  local  ports  or  insufficient entries in the routing cache.", my_errno, ConnectionException::connect );

    case EAFNOSUPPORT :
      throw ConnectionException("The  passed  address  didn't  have  the  correct address family in its sa_family field.", my_errno, ConnectionException::connect );

    case EACCES :
      throw ConnectionException("The user tried to connect to a  broadcast  address  without  having  the  socket \
               broadcast flag enabled or the connection request failed because of a local fire-\
               wall rule.", my_errno, ConnectionException::connect );

    case EPERM  :
      throw ConnectionException("The user tried to connect to a  broadcast  address  without  having  the  socket \
               broadcast flag enabled or the connection request failed because of a local fire-\
               wall rule.", my_errno, ConnectionException::connect );
    default :
      throw  ConnectionException("Unknown error", my_errno, ConnectionException::connect );
    }
  }

}

#include "nmm/nmm_suffix.hpp"
