/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include "SockStream.hpp"
#include "nmm/utils/NamedObject.hpp"

#include <cerrno>

#include "nmm_netutils.hpp"

#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>    
#include <netinet/tcp.h>

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  /****************** class SockStream ******************/
  
  SockStream::SockStream(bool owner)
    :  __socket(-1),
       __own_socket(owner)
  {
    // XXX: fw: not necessary to create a socket here !! (SockConnector will do this)
    struct protoent* protocol = getprotobyname("tcp");
    
    __socket = ::socket(PF_INET, SOCK_STREAM, protocol->p_proto);
    if ( __socket == -1 ) {
      switch (errno){
        case EACCES : GLOBAL_ERROR_STREAM( " SockStream(): Permission to create a socket of the specified type and/or protocol is denied. " << endl); break;

        case EAFNOSUPPORT: GLOBAL_ERROR_STREAM( " SockStream(): The implementation does not support the specified address family." << endl); break;

        case EINVAL : GLOBAL_ERROR_STREAM( " SockStream(): Unknown protocol, or protocol family not available." << endl); break;

        case EMFILE : GLOBAL_ERROR_STREAM( " SockStream(): Process file table overflow." << endl); break;

        case ENFILE : GLOBAL_ERROR_STREAM( " SockStream(): The system limit on the total number of open files has been reached." << endl); break;

        case ENOBUFS : GLOBAL_ERROR_STREAM( " SockStream(): Insufficient memory is available.  The socket cannot be created until sufficient resources are freed." << endl); break;
	
	case ENOMEM : GLOBAL_ERROR_STREAM( " SockStream(): Insufficient memory is available.  The socket cannot be created until sufficient resources are freed." << endl); break;

        case EPROTONOSUPPORT : GLOBAL_ERROR_STREAM( " SockStream(): The protocol type or the specified protocol is not supported within this domain." << endl); break;   
      }
    }
    #ifdef BUILD_MACOSX
    if ( __socket > 0 ) {
    	  int on = 1;
      int retval = setsockopt(__socket, SOL_SOCKET, SO_NOSIGPIPE, &on, sizeof(on));
      if ( retval != 0 ) {
        cerr <<	"setsockopt(__socket, SOL_SOCKET, SO_NOSIGPIPE, &on, sizeof(on)); faile " << endl; 
      }
    }
    #endif
  }
  
  SockStream::SockStream(SocketHandle socket, bool owner)
    : __socket(socket),
      __own_socket(owner)
  {}
  
  SockStream::~SockStream() {
    /* if this object is the owner, close the socket */
    if (__own_socket && __socket >= 0) {
      close();
    }
  }
  
  void SockStream::setSocket(SocketHandle socket, bool owner) {
    /* If this object is the owner, close the old socket */
    if (__own_socket && __socket >= 0) {
      close();
    } 
    /* set the new socket */
    __socket = socket;

    #ifdef BUILD_MACOSX
    if ( __socket > 0 ) {
    	  int on = 1;
      int retval = setsockopt(__socket, SOL_SOCKET, SO_NOSIGPIPE, &on, sizeof(on));
      if ( retval != 0 ) {
        cerr <<	"setsockopt(__socket, SOL_SOCKET, SO_NOSIGPIPE, &on, sizeof(on)); failed " << endl; 
      }
    }
    #endif 
    
    /* set the owner flag */
    __own_socket = owner;
  }
  
  size_t SockStream::recv(void* buffer,size_t len)
    throw (const TransmissionException&)
  {
    ssize_t size = ::recv(__socket, buffer, len, /*MSG_NOSIGNAL |*/ /*MSG_WAITALL*/ 0);
    if (size < 0) {
      createReadTransmissionException(errno);
    }
    return size;
  }
  
  size_t SockStream::send(const void* buffer,size_t len)
    throw (const TransmissionException&)
  {
    ssize_t size = ::send(__socket, buffer, len, MSG_NOSIGNAL | MSG_WAITALL);
    if (size < 0) {
      createWriteTransmissionException(errno);
    }
    return size;
  }
  
  void SockStream::close() {
    /* close the connection with c-shutdown method */
    if ( __socket ) {  
       ::shutdown(__socket, SHUT_RDWR);
       
      if ( ::close(__socket ) == -1) {
        GLOBAL_ERROR_STREAM( "SockeStream::close() : Closing socket " << __socket << " failed" << endl);    
      } 
      __socket = -1;
    }
  }
  
  void SockStream::readStream(char* data, size_t length, size_t max_packet_size /* = 0 */ ) 
    throw (const TransmissionException&, const ConnectionClosed&)
  {
    if (max_packet_size == 0) {
      max_packet_size = length;
    }
    
    int received = 0;
    
    /* While not all data are received call recv() */
    while (length > 0)  {
      /* receive the */
      if ( (received = recv(data, length)) == 0) {
        /* the connection was closed */
        throw ConnectionClosed(EPIPE);
      }
      
      length -= received;
      data   += received;
    }
  }
  
  void SockStream::writeStream(const char* data, size_t length, size_t max_packet_size)
    throw(const TransmissionException&, const ConnectionClosed&)
  {
    if (max_packet_size == 0) {
      max_packet_size = length;
    }
    
    int transmit = 0;
    
    /* Wile not all data are written call send */
    while (length > 0)  {    
      if (length < max_packet_size)
        // calls SockStream::send and disables all signals
        transmit = send(data, length);
      else
        // calls SockStream::send and disables all signals
        transmit = send(data, max_packet_size);
      
      /* If return value is 0 the connection is closed */
      if (transmit == 0) {
        /* Throw a ConnectionClosed exception */
        throw ConnectionClosed(EPIPE);
      }
      
      /* go to the memory section which was not written */
      data   += transmit;
      /* remove the written bytes */
      length -=  transmit;
    }
  }
  
  INetAddr SockStream::getLocalAddress() const {
    struct sockaddr_in addr;
    socklen_t len_addr = sizeof(addr);
    
    if (getsockname(__socket, reinterpret_cast<struct sockaddr*>(&addr), &len_addr) == 0) {
      INetAddr address;
      address.setAddress(&addr);
      return address;
    } else {
      GLOBAL_ERROR_STREAM("Failed to get stream socket address" << endl);
      return INetAddr();
    }
  }
  
  const INetAddr& SockStream::getPeerAddress() const {
    return __peer_addr;
  }
  
  void SockStream::setPeerAddress(const INetAddr& addr) {
    __peer_addr = addr;
  }
  
  void SockStream::enableNoDelay(const bool& value) {
    int v1 = value;
    if (setsockopt (getSocket(), SOL_TCP, TCP_NODELAY, &v1, sizeof (v1))== -1)
      createSetSockOptException(errno);
  }

  bool SockStream::isDelayEnabled() const {
    int response = false;
#ifdef __CYGWIN__
    int size = sizeof(response);
#endif
#ifndef __CYGWIN__
#ifdef BUILD_PPC
    socklen_t  size  = sizeof(response);
#else
    unsigned int size = sizeof(response);
#endif
#endif
    if (getsockopt(getSocket(), SOL_TCP, TCP_NODELAY, reinterpret_cast<char*>(&response), &size) == -1)  {
      createSetSockOptException(errno);
    }
    return response;
  }

  void SockStream::setMaxSegment(const int& value){
    if ( setsockopt(getSocket(), SOL_TCP, TCP_MAXSEG, reinterpret_cast<const char*>(&value), sizeof(value)) == -1) {
      createSetSockOptException(errno);
    }
  }

  int  SockStream::getMaxSegment() const {
    int retval = 0;
#ifdef __CYGWIN__
    int size = sizeof(retval);
#endif
#ifndef __CYGWIN__
#ifdef BUILD_PPC
    socklen_t size = sizeof(retval);
#else
    unsigned int size = sizeof(retval);
#endif
#endif
    if (getsockopt(getSocket(), SOL_TCP, TCP_MAXSEG, reinterpret_cast<char*>(&retval), &size) == -1)  {
      createSetSockOptException(errno);
    }
    return size;
  }

#ifndef __CYGWIN__
  void SockStream::enablePartialFrames(const bool& value){
#ifndef BUILD_MACOSX
    if ( setsockopt(getSocket(), SOL_TCP,  TCP_CORK, reinterpret_cast<const char*>(&value), sizeof(value)) == -1) {
      createSetSockOptException(errno);
    }
#endif    
  }
#endif

#ifdef __CYGWIN__
  void SockStream::enablePartialFrames(const bool& ){}
#endif

  bool SockStream::isPartialFramesEnabled(){
    int retval = 0;
#ifndef __CYGWIN__
#ifndef BUILD_MACOSX
    unsigned int size = sizeof(retval);
    if (getsockopt(getSocket(), SOL_TCP, TCP_CORK, reinterpret_cast<char*>(&retval), &size) == -1)  {
      createSetSockOptException(errno);
    }
#endif
#endif
    return retval;
  }

  int  SockStream::getUnreadQueueCount() const{
    int value = 0;
    if ( ioctl(getSocket(), FIONREAD, &value) == -1) {
      createSetSockOptException(errno);
    }
    return value;
  }

  bool SockStream::allUrgentDataSend()   const{
    int value = 0;
    if ( ioctl(getSocket(), SIOCATMARK, &value) == -1) {
      createSetSockOptException(errno);
    }
    return value;
  }

  int  SockStream::getUnsendQueueCount() const{
    int value = 0;
#ifndef __CYGWIN__
    if ( ioctl(getSocket(), TIOCOUTQ, &value) == -1) {
      createSetSockOptException(errno);
    }
#endif
    return value;
  }

  void SockStream::enableKeepAlive(const bool& value) {
    if (setsockopt (getSocket(), SOL_SOCKET, SO_KEEPALIVE, &value, sizeof (value))== -1)
      createSetSockOptException(errno);
  }

  bool SockStream::getKeepAlive() const{
    int retval = 0;
#ifdef __CYGWIN__
    int size = sizeof(retval);
#endif
#ifndef __CYGWIN__
#ifdef BUILD_PPC
    socklen_t size = sizeof(retval);
#else
    unsigned int size = sizeof(retval);
#endif
#endif
    if (getsockopt(getSocket(), SOL_SOCKET, SO_KEEPALIVE, reinterpret_cast<char*>(&retval), &size) == -1)  {
      createSetSockOptException(errno);
    }
    return retval;
  }
  
}

#include "nmm/nmm_suffix.hpp"
