/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "Sock.hpp"

#include "nmm/utils/NamedObject.hpp"

#include <cerrno>       // for errnos

#include "nmm_netutils.hpp"
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <netinet/in.h>

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  Sock::Sock() throw() : __current_qos_tag(Sock::DEFAULT) {}
  
  void Sock::setRecvBufferSize(const int& value) {
    if(setsockopt(getSocket(), SOL_SOCKET, SO_RCVBUF,
                  &value, sizeof (value)) == -1) {
      createSetSockOptException(errno);
    }
  }  
  
  int Sock::getRecvBufferSize() const {
    int retval = 0;
    socklen_t size = sizeof(retval);

    if(getsockopt(getSocket(), SOL_SOCKET, SO_RCVBUF,
                  reinterpret_cast<char*>(&retval), &size) == -1) {
      createSetSockOptException(errno);
    }
    
    return retval;
  }
  
  void Sock::setSendBufferSize(const int& value) {
    if(setsockopt(getSocket(), SOL_SOCKET, SO_SNDBUF,
                  &value, sizeof (value))== -1) {
      createSetSockOptException(errno);
    }
  }
  
  int Sock::getSendBufferSize() const {
    int retval = 0;
    socklen_t size = sizeof(retval);

    if (getsockopt(getSocket(), SOL_SOCKET, SO_SNDBUF,
                   reinterpret_cast<char*>(&retval), &size) == -1)  {
      createSetSockOptException(errno);
    }
    
    return retval;
  }
  
  int Sock::getUsedSendBuffer() const {
    
    return 0;
  }


  void Sock::setPriority(const int& ) {}
  
  int Sock::getPriority() const {
    int retval = 0;
    socklen_t size = sizeof(retval);

    if(getsockopt(getSocket(), SOL_SOCKET, SO_RCVBUF,
                  reinterpret_cast<char*>(&retval), &size) == -1) {
      createSetSockOptException(errno);
    }
    
    return retval;
  }
  

  void Sock::setQoSTag(const QOS_TAG& qos_tag) {
    if (__current_qos_tag == qos_tag ) {
      return;
    }
    int value = 0;
    switch ( qos_tag ) {
      case Sock::DEFAULT: value = 0; break;
      case Sock::STREAMING_DATA : value = 128; break;  //should be 160 but needs root access
      case Sock::INBAND_DATA :    value = 128; break;
      case Sock::RMI_DATA :       value = 96;  break;
    }

    if ( setsockopt( getSocket(), IPPROTO_IP, IP_TOS, reinterpret_cast<char*>(&value), sizeof(value) ) == -1)
    {
      createSetSockOptException(errno);
    } 

    __current_qos_tag = qos_tag;
  }

  Sock::QOS_TAG Sock::getQoSTag() {
    int retval = 0;
    socklen_t size = sizeof(retval) ;
    if (getsockopt(getSocket(), IPPROTO_IP, IP_TOS,
                   reinterpret_cast<char*>(&retval), &size) == -1)  {
      createSetSockOptException(errno);
    }
    switch ( retval ) {
      case 0   : return Sock::DEFAULT; break;
      case 160 : return Sock::STREAMING_DATA; break;
      case 128 : return Sock::INBAND_DATA; break;
      case 96  : return Sock::RMI_DATA;  break;
      default  : return Sock::DEFAULT; break;
    }
   }


  void Sock::createSetSockOptException(int my_errno) throw() {
    switch(my_errno) {
    case EPIPE:
      GLOBAL_ERROR_STREAM1("The other end closed the socket unexpectedly or a read is executed on a shut down socket." << endl);
      break;
    case ETIMEDOUT:
      GLOBAL_ERROR_STREAM1("The other end didn't acknowledge retransmitted data after some time." << endl);
      break;  
    case EAFNOSUPPORT:
      GLOBAL_ERROR_STREAM1("Passed socket address type in sin_family was not AF_INET." << endl);
      break;
    default:
      GLOBAL_ERROR_STREAM1("Unknown error" << endl);
    }
  }
  
  void Sock::createWriteTransmissionException(int my_errno)
    throw(const TransmissionException&)
  {
    switch(my_errno) {
    case EBADF     : throw TransmissionException("fd is not a valid file descriptor or is not open for writing.", my_errno);

    case EINVAL    : throw TransmissionException("fd is attached to an object which is unsuitable for writing.", my_errno);

    case EFAULT    : throw TransmissionException("buf is outside your accessible address space.", my_errno);

    case EFBIG     : throw TransmissionException("An attempt was made to write a file that exceeds the implementation-defined maximum file size or the process' file size limit, or to  write  at  a position past than the maximum allowed offset.", my_errno);

    case EPIPE     : throw TransmissionException("fd  is connected to a pipe or socket whose reading end is closed.", my_errno);

    case EAGAIN    : throw TransmissionException("Non-blocking I/O has been selected using O_NONBLOCK and there was no room in the\
              pipe or socket connected to fd to write the data immediately.", my_errno);

    case EINTR     : throw TransmissionException("The call was interrupted by a signal before any data was written.", my_errno);

    case ENOSPC    : throw TransmissionException("The device containing the file referred to by fd has no room for the data.", my_errno);

    case EIO       : throw TransmissionException("A low-level I/O error occurred while modifying the inode.", my_errno);

    case ECONNRESET: throw TransmissionException("Connection reset by peer", my_errno);

    default        : throw TransmissionException("Unknown error", my_errno);
    }
  }

  void Sock::createReadTransmissionException(int my_errno)
    throw(const TransmissionException&)
  {
    switch (my_errno) {
    case EINTR  :
      throw TransmissionException("The call was interrupted by a signal before any data was read.", my_errno);

    case EAGAIN :
      throw TransmissionException("Non-blocking  I/O has been selected using O_NONBLOCK and no data was immediately\
               available for reading.", my_errno);

    case EIO    :
      throw TransmissionException("I/O error.", my_errno);

    case EISDIR :
      throw TransmissionException("fd refers to a directory.", my_errno);

    case EBADF  :
      throw TransmissionException("fd is not a valid file descriptor or is not open for reading.", my_errno);

    case EINVAL :
      throw TransmissionException("fd is attached to an object which is unsuitable for reading.", my_errno);

    case EFAULT :
      throw TransmissionException("buf is outside your accessible address space.", my_errno);

    default:
      throw TransmissionException("Unknown error", my_errno);
    }
  }

}

#include "nmm/nmm_suffix.hpp"
