/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "ThreadedNetHandler.hpp"
#include "NetReactor.hpp"

#include "nmm/utils/thread/MutexGuard.hpp"
#include "nmm/utils/thread/ThreadPool.hpp"

#include <errno.h>
#include <iostream>
#include <cassert>

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  ThreadedNetHandler::ThreadedNetHandler()
    : __net_reactor(SNetReactor::getInstance()),
      __notification(true),
      __thread_count(0), __event_mutex(), __event_condition(__event_mutex),
      __socket(-1), __events(0), __revents(0), __task(this, &ThreadedNetHandler::doCheckForEvents),
      __pipe_socket(2)
  {
    /* Create the internal pipe */
    if(pipe(&__pipe_socket[0]) != 0) {
      GLOBAL_ERROR_STREAM("Could not create the internal pipe " << std::endl);
    }
  }

  ThreadedNetHandler::~ThreadedNetHandler() {
    /* Close the internal pipe */
    close( __pipe_socket[0] );
    close( __pipe_socket[1] );
  }

  void ThreadedNetHandler::disableNotification() {
    /* Clear the notification flag */
    __notification = false;

    /* Interrupt the poll */
    write(__pipe_socket[1], &__internal_char, 1);

    /* Wait for all threads to finish */
    MutexGuard guard(__event_mutex);
    while(__thread_count > 0) {
      /* Wait for notification that the last thread has finished */
      __event_condition.wait();
    }
  }

  bool ThreadedNetHandler::requireThread() const {
    return true;
  }

  NetHandler::NetResult ThreadedNetHandler::receiveEvent(SocketHandle socket, short events, short revents) {
    /* Lock the event mutex */
    MutexGuard event_guard(__event_mutex);

    /* Store the arguments in this handler object */
    __socket = socket;
    __events = events;
    __revents = revents;

    /* Increment the thread counter */
    ++__thread_count;

    /* Start handling events. Initial state is HANDLE_EVENTS */
    return run(HANDLE_EVENTS);
  }

  void ThreadedNetHandler::requestLeaderThread() {
    /* Lock the event mutex to disallow concurrent calls */
    MutexGuard event_guard(__event_mutex);

    /* Increment the thread counter */
    ++__thread_count;

    /* Request a new worker thread */
    WorkerThread* worker = SThreadPool::getInstance().requestThread();

    /* Create a new task for invoking the handler. */
    worker->setTask(&__task);

    /* Run the worker thread */
    worker->iterateTask();
  }

  bool ThreadedNetHandler::isDataAvailable() {
    return false;
  }

  void ThreadedNetHandler::doCheckForEvents() {
    /* Lock the event mutex */
    MutexGuard event_guard(__event_mutex);

    /* Start event handling. Initial state is CHECK_FOR_EVENTS. */
    /* Discard the result of the run method*/
    run(CHECK_FOR_EVENTS);
  }

  NetHandler::NetResult ThreadedNetHandler::run(InitialState state) {
    /* -- The event mutex is locked by the caller -- */

    /* Every new thread is initially the leader */
    bool is_leader = true;

    // XXX: fw: TODO: Remove this, only for debugging
    static unsigned int __thread_id = 0;
    if(__thread_count == 1) __thread_id = 0;
    
    __thread_id++;

    /* Main loop */
    bool finished = false;
    while(!finished) {
      switch(state) {
        /* -- State CHECK_FOR_EVENTS: Check for new socket events -- */
        case CHECK_FOR_EVENTS:
        {
          /* Check if data is still available */
          if(!isDataAvailable()) {
            /* Unlock the mutex */
            __event_mutex.unlock();

            /* Recheck for socket events */
            pollfd tmp_struct[2];
            tmp_struct[0].fd = __socket;
            tmp_struct[0].events = __events;
            tmp_struct[1].fd = __pipe_socket[0];
            tmp_struct[1].events = NetReactor::NR_POLLERR | NetReactor::NR_POLLIN |NetReactor::NR_POLLRDNORM | NetReactor::NR_POLLRDBAND | NetReactor::NR_POLLPRI;
            int retval = ::poll(&tmp_struct[0], 2, 100);

            /* Lock the mutex */
            __event_mutex.lock();

            /* Update event mask */
            __revents = tmp_struct[0].revents;

            /* If no event occurred, the socket is registered again and this thread becomes idle */
            if(retval == 0) {
              state = IDLE;
              break;
            }

            /* If an error occurred, the socket is registered again and this thread becmoes idle */
            if(retval < 0) {
              GLOBAL_ERROR_STREAM1("ThreadedNetHandler: Socket error while rechecking for events" << endl);
              state = IDLE;
              break;
            }
          }

          /* Go to state HANDLE_EVENTS */
          state = HANDLE_EVENTS;
          break;
        }

        /* -- State CHECK_FOR_EVENTS: Handle socket events (IO, PROCESSING) -- */
        case HANDLE_EVENTS:
        {
          /* If notification is disabled, remove the event handler */
          if(!__notification ) {
            state = IDLE;
            break;
          }

          /* Unlock the mutex */
          __event_mutex.unlock();

          /* Handle events */
          NetResult result = handleEvents(__socket, __events & __revents);

          /* Lock the mutex */
          __event_mutex.lock();

          /* Check if this thread is still the leader */
          if(result != RECHECK) {
            /* This thread is no longer the leader and can go idle because it has nothing to do. */
            is_leader = false;
            state = IDLE;
            break;
          } else {
            /* Go to state CHECK_FOR_EVENTS */
            state = CHECK_FOR_EVENTS;
            break;
          }
        }

        /* -- State IDLE: This thread waits for work -- */
        case IDLE:
        {
          /* Decrement the event count */
          assert(__thread_count > 0);
          --__thread_count;

          /* Check if this was the last thread that finished */
          if(__thread_count == 0) {
            /* Notify all threads waiting in disableNotification */
            __event_condition.notifyAll();
          }

          finished = true;
          break;
        }
      }
    } while(!finished);

    /* If this thread is not the leader, take no special action */
    if(!is_leader) {
      return VOID;
    }

    /* If notification is disabled, return REMOVE */
    if(!__notification) {
      return REMOVE;
    }

    /* Re-register this socket at the reactor */
    getNetReactor().registerSocket(__socket, this, __events);
    return VOID;
  }

  char ThreadedNetHandler::__internal_char = 'a';

}

#include "nmm/nmm_suffix.hpp"
