/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "PNGUtils.hpp"
#include "nmm/utils/imageconverter_utils.hpp"
#include <vector>

#include "nmm/nmm_prefix.hpp"

namespace NMM {

  PNGUtils::PNGUtils() {
    width = 0;
    height = 0;
    rgb_buffer = NULL;
    setFormat( RGB24 );
  }

  PNGUtils::~PNGUtils(){
    destroyRGBBuffer();
  }

  void PNGUtils::destroyRGBBuffer() {
    if (rgb_buffer) {
      delete[] rgb_buffer;
      rgb_buffer = NULL;
    }
  }

  void PNGUtils::setFormat( colorformat cf ) {
    format = cf;
  }

  void PNGUtils::setDimension( unsigned int w, unsigned int h ) {
    if (w == width && h == height) return;
    width = w;
    height = h;
    destroyRGBBuffer();
    rgb_buffer = new unsigned char[width*height*3];    
  }

  void PNGUtils::writeImage( const string& filename, unsigned char* image ) {
    if ( format == YV12 ) {
      convertYV12ToRGB24( rgb_buffer, image, width, height, false );
      image = rgb_buffer;
    } else if ( format == YV12A ) {
      convertYV12ToRGB24( rgb_buffer, image, width, height, true );
      image = rgb_buffer;
    } else if ( format == RGBX32 ) {
      // If we have RGBX32, which means that the last 8 bits are padding bits,
      // convert the data
      convertRGBX32toRGB24InPlace(image, width, height);
    }

    writePNG(filename, image, width, height );
  }

  void PNGUtils::writePNG(const string& file_name, unsigned char* image, unsigned int width, unsigned int height) {
    FILE *fp;
    png_structp png_ptr;
    png_infop info_ptr;


    /* open the file */
    fp = fopen(file_name.c_str(), "wb");
    if ( !fp ) return;


    /* allocate the necessary structures */

    png_ptr = new png_struct;

    if ( !png_ptr ) {
      fclose(fp);
      return;
    }

    info_ptr = new png_info;

    if (!info_ptr) {
      fclose(fp);
      delete png_ptr;
      return;
    }

    /* set error handling */
    if ( setjmp(png_ptr->jmpbuf) ) {
#ifndef WIN32
		png_write_destroy(png_ptr);
#endif
      fclose(fp);
      delete png_ptr;
      delete info_ptr;

      /* If we get here, we had a problem reading the file */
      return;
    }

    /* initialize the structures */
    png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if ( !png_ptr ) return;

    info_ptr = png_create_info_struct(png_ptr);
    if ( !info_ptr ) {
      png_destroy_write_struct(&png_ptr, (png_infopp)NULL);
      return;
    }

    /* set up the output control if you are using standard C streams */
    png_init_io( png_ptr, fp );

    int color_type = PNG_COLOR_TYPE_RGB;
    int bytes_per_pixel = 3;

    png_set_bgr(png_ptr);

    png_set_IHDR(png_ptr, info_ptr, width, height,
                8, color_type, PNG_INTERLACE_NONE,
                PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);

    /* write the file information */
    png_write_info(png_ptr, info_ptr);

    /* the easiest way to write the image */
    vector<png_bytep> row_pointers(height);
    png_byte* pimage = (png_byte *) image;
    for( unsigned int i=0; i<height; i++ ) {
      row_pointers[i] = pimage;
      pimage += bytes_per_pixel*width;
    }
    png_write_image(png_ptr, &row_pointers[0]);


    /* write the rest of the file */
    png_write_end(png_ptr, info_ptr);

    /* clean up after the write, and free any memory allocated */

    //png_write_destroy(png_ptr);

    /* if you malloced the palette, free it here */
#ifndef WIN32
	if (info_ptr->palette) free(info_ptr->palette);
#endif
    /* free the structures */
    delete png_ptr;
    delete info_ptr;

    /* close the file */
    fclose(fp);
  }

} // NMM


#include "nmm/nmm_suffix.hpp"
