/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

 /**
  * This Class gives infos about the CPU of the host it is running on
  * On creation it starts a thread that periodically measures CPU utilisation
  * to calculate the amount of free CPU time
  */

#include "CPUInfo.hpp"
#include "nmm/utils/thread/ThreadCondition.hpp"
#include "nmm/utils/thread/MutexGuard.hpp"
#include <sstream>
#include <fstream>
#include <unistd.h>

#include "nmm/nmm_prefix.hpp"

namespace NMM {

// static member declaration
string CPUInfo::cpu_type;
string CPUInfo::CPUFILENAME;
string CPUInfo::STATFILENAME;
unsigned int CPUInfo::cpu_num;
double CPUInfo::cpu_speed;
vector<int> CPUInfo::cpu_free;
vector<CPUInfo::CPUUsage> CPUInfo::cpu_usage;
unsigned int CPUInfo::update_intervall;
bool CPUInfo::is_running;
bool CPUInfo::is_started;
bool CPUInfo::first_read = true;
ifstream CPUInfo::stat_file;

ThreadMutex CPUInfo::cv;
ThreadCondition* CPUInfo::mcv;

CPUInfo::CPUInfo() : NamedObject("CPUInfo") {
  if (! is_running) {
    mcv = new ThreadCondition(cv);
    is_running = true;
  }
  first_read = true;
}

void CPUInfo::start() {
  if (! is_started) {
    update_intervall = 2;

    string tmp;
    STATFILENAME = string("/proc/stat");
    CPUFILENAME  = string("/proc/cpuinfo");

    // get infos about the cpu

    ifstream cpuinfo( CPUFILENAME.c_str() ); // open the file in /proc

    // read the available data
    do {
      cpuinfo >> tmp;
    } while ( (tmp != string("processor")) && (! cpuinfo.eof()) );
    if (! (tmp != string("processor")))
      cpu_num = 1; // first processor found
    else {
      ERROR_STREAM("NO CPU FOUND !?" << endl);
      return; //exit(1);
    }

    do {
      cpuinfo >> tmp;
    } while (tmp != string("name"));
    cpuinfo >> tmp; // skip ':'

    stringstream tmp_stream("");
    do {
      cpuinfo >> tmp;
      if (tmp != string("stepping"))
        tmp_stream << tmp.c_str() ;
    } while (tmp != string("stepping"));

    cpu_type = tmp_stream.str();

    do {
      cpuinfo >> tmp;
    } while (tmp != string("bogomips"));
    cpuinfo >> tmp; // skip ':'
    cpuinfo >> cpu_speed;

    do {
      do {
        cpuinfo >> tmp;
      } while ( (tmp != string("processor")) && (! cpuinfo.eof()) );
      if (! (tmp != string("processor")))
        cpu_num++;
    } while(! cpuinfo.eof());

    cpuinfo.close();

    // initialize the counters that measure cpu_usage
    cpu_usage.clear();
    cpu_usage.resize(cpu_num);

    cpu_free.clear();
    cpu_free.resize(cpu_num);

    DEBUG_STREAM("reading /proc/stat" << endl);

    // read the initial cpu utilisation values
//    read_cpu_info();

    // set the cpu_free values to maximum because the first values read
    // do not allow any conclusion about free cpu time
    for (uint i=0;i< cpu_num; i++) {
      cpu_free[i] = (100 * update_intervall);
    }

    is_started = true;
    cpu_free_monitor_thread.start(run_static, NULL);
  }
}

CPUInfo::~CPUInfo() {
  if (is_started) {
    stop();
  }
}

void CPUInfo::stop() {
    cv.lock();
    is_started = false;
    mcv->notify();
    cv.unlock();
    // stop the running thread
    cpu_free_monitor_thread.join();
}

void CPUInfo::read_cpu_info() {
  string tmp;
  CPUUsage tmp_cpu_usage;

  if (first_read)
    stat_file.open(STATFILENAME.c_str());

  if (! stat_file) {
    GLOBAL_ERROR_STREAM("unable to open '/proc/stat' " << endl);
    return;
  }
  
  stat_file.sync();
  stat_file.seekg(0,ios::beg); // read always from beginning of the file

  for (unsigned int i=0; i< cpu_num; i++) {
    stringstream tmp_stream("");
    tmp_stream << "cpu" << i ; // measure cpu 0,1,...

    do { // read the data from proc
      /* Read a single line from stat_file ...*/
      string tmp_string;
      getline(stat_file, tmp_string);

      /* and parse it */
      stringstream read_stream;
      read_stream << tmp_string; 
      read_stream >>tmp >>tmp_cpu_usage.cpu_user >> tmp_cpu_usage.cpu_nice >> tmp_cpu_usage.cpu_system >> tmp_cpu_usage.cpu_free;
    } while (tmp != tmp_stream.str() && stat_file);

     /* Clear the output stream if something went wrong */
     if (!stat_file) {
       stat_file.clear();
     }
    
    // actual value = new value - old value
    if (! first_read) {
      cpu_free[i] = tmp_cpu_usage.cpu_free -  cpu_usage[i].cpu_free;
    }
    // store new value
    cpu_usage[i] = tmp_cpu_usage;
  }

  first_read = false;
}

const unsigned int CPUInfo::getFreeCPU(unsigned int cpu) {
  if (cpu >= cpu_num) {
    return 0;
  }
  unsigned int free_ms = cpu_free[cpu] / update_intervall;
  return free_ms;
}

const unsigned int CPUInfo::getTotalFreeCPU() {
  unsigned int free_ms = 0;

  for (uint i=0; i< cpu_num; i++) { // simply add free time of all cpu's
    free_ms += getFreeCPU(i);
  }
  return free_ms;
}

const unsigned int CPUInfo::getCPUNum() {
    return cpu_num;
}

const string CPUInfo::getCPUType() {
    return cpu_type;
}

const unsigned int CPUInfo::getCPUSpeed() {
    return (unsigned int) cpu_speed;
}

void CPUInfo::setUpdateIntervall(unsigned int seconds) {
  update_intervall = seconds;
}

void* CPUInfo::run_static(void*) {
  MutexGuard mg(cv);
  while (is_started) {
    read_cpu_info();
    mcv->relativeTimedWait(update_intervall);
  } 
  return 0;
}

}

#include "nmm/nmm_suffix.hpp"
