/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */


#include "VideoScaler.hpp"
#include "scaler.h"
#include "nmm/utils/NamedObject.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  VideoScaler::VideoScaler() {
    in_width = in_height = 0;
    out_width = out_height = 0;
    filtering = false;
    format = FORMAT_YV12;
    num_planes = 0;
    planar = false;
  }
  
  void VideoScaler::setInputSize(int width, int height) {
    in_width = width;
    in_height = height;
    
    GLOBAL_DEBUG_STREAM("VideoScaler: set input size to " << in_width << "x" << in_height << endl);
    
    initPlanes();
  }
  
  void VideoScaler::setOutputSize(int width, int height) {
    out_width = width;
    out_height = height;
    
    GLOBAL_DEBUG_STREAM("VideoScaler: set output size to " << out_width << "x" << out_height << endl);
    
    initPlanes();
  }
  
  void VideoScaler::setFormat(int format) {
    this->format = format;
    initPlanes();
  }
  
  void VideoScaler::initPlanes() {
    initPlanes(in_width, in_height, in_planes);
    initPlanes(out_width, out_height, out_planes);
  }
  
  void VideoScaler::initPlanes(int w, int h, ImagePlane* planes) {
    if (format == FORMAT_RGB) {
      planar = false;
      num_planes = 3;
      
      for (int i=0; i<num_planes; ++i) {
        planes[i].offset = i;
        planes[i].x_pixels = w;
        planes[i].y_pixels = h;
        planes[i].x_pitch = num_planes;
        planes[i].y_pitch = h * num_planes;
      }
    } else if (format == FORMAT_RGBA) {
      planar = false;
      num_planes = 4;
      
      for (int i=0; i<num_planes; ++i) {
        planes[i].offset = i;
        planes[i].x_pixels = w;
        planes[i].y_pixels = h;
        planes[i].x_pitch = num_planes;
        planes[i].y_pitch = h * num_planes;
      }
    } else if (format == FORMAT_YV12) {
      planar = true;
      num_planes = 3;
      
      planes[0].offset = 0;
      planes[0].x_pixels = w;
      planes[0].y_pixels = h;
      planes[0].x_pitch = 1;
      planes[0].y_pitch = w;
      
      planes[1].offset = w * h;
      planes[1].x_pixels = w >> 1;
      planes[1].y_pixels = h >> 1;
      planes[1].x_pitch = 1;
      planes[1].y_pitch = w >> 1;
      
      planes[2].offset = planes[1].offset + ((w >> 1) * (h >> 1));
      planes[2].x_pixels = w >> 1;
      planes[2].y_pixels = h >> 1;
      planes[2].x_pitch = 1;
      planes[2].y_pitch = w >> 1;
    } else if (format == FORMAT_YV12A) {
      planar = true;
      num_planes = 4;
      
      planes[0].offset = 0;
      planes[0].x_pixels = w;
      planes[0].y_pixels = h;
      planes[0].x_pitch = 1;
      planes[0].y_pitch = w;
      
      planes[1].offset = w * h;
      planes[1].x_pixels = w >> 1;
      planes[1].y_pixels = h >> 1;
      planes[1].x_pitch = 1;
      planes[1].y_pitch = w >> 1;
      
      planes[2].offset = planes[1].offset + ((w >> 1) * (h >> 1));
      planes[2].x_pixels = w >> 1;
      planes[2].y_pixels = h >> 1;
      planes[2].x_pitch = 1;
      planes[2].y_pitch = w >> 1;
      
      planes[3].offset = planes[2].offset + ((w >> 1) * (h >> 1));
      planes[3].x_pixels = w;
      planes[3].y_pixels = h;
      planes[3].x_pitch = 1;
      planes[3].y_pitch = w;
    } else {
      num_planes = 0;
    }
  }
  
  void VideoScaler::processData(char* out_data, const char* in_data) {
    
    int in_delta_x = (in_width << 16) / out_width;
    int in_delta_y = (in_height << 16) / out_height;
    
    // XXX: not using x_pitch (=> only planar formats currently supported)
    
    for (int i=0; i<num_planes; ++i) {
      if (filtering) {
        processPlaneFiltering(out_data + out_planes[i].offset,
                              out_planes[i].x_pixels,
                              out_planes[i].y_pixels,
                              out_planes[i].y_pitch,
                              in_data + in_planes[i].offset,
                              in_planes[i].y_pitch,
                              in_delta_x,
                              in_delta_y);
      } else {
        processPlane(out_data + out_planes[i].offset,
                     out_planes[i].x_pixels,
                     out_planes[i].y_pixels,
                     out_planes[i].y_pitch,
                     in_data + in_planes[i].offset,
                     in_planes[i].y_pitch,
                     in_delta_x,
                     in_delta_y);
      }
    }
  }
  
  void VideoScaler::processPlane(char* out_data, int out_width, int out_height, int out_pitch,
				 const char* in_data, int in_pitch,
				 int in_delta_x, int in_delta_y) {
    int iny = 0;
    for (int y=0; y<out_height; ++y) {
      scaleImageRow(out_data, out_width, in_data + (iny >> 16) * in_pitch, in_pitch, in_delta_x, (iny & 0xFFFF));
      out_data += out_pitch;
      iny += in_delta_y;
    }
  }

  void VideoScaler::processPlaneFiltering(char* out_data, int out_width, int out_height, int out_pitch,
					  const char* in_data, int in_pitch,
					  int in_delta_x, int in_delta_y) {
    int iny = 0;
    for (int y=0; y<out_height; ++y) {
      scaleImageRowFiltering(out_data, out_width, in_data + (iny >> 16) * in_pitch, in_pitch, in_delta_x, (iny & 0xFFFF));
      out_data += out_pitch;
      iny += in_delta_y;
    }
  }
}


#include "nmm/nmm_suffix.hpp"
