/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */



#include "Thread.hpp"
#include "nmm/utils/NamedObject.hpp"
#include <iostream>

#include <signal.h>
#include <unistd.h>
#include <sys/time.h>
#include <sys/types.h>

#include "nmm/nmm_prefix.hpp"

namespace NMM {

Thread::Thread(){
  fifo_scheduling_enabled = rr_scheduling_enabled = false;
}

Thread::~Thread() {
}

void Thread::start(  ThrFunction function, void *arg, pthread_attr_t* attr ) throw (const runtime_error&) {
#ifdef PROFILING
  if ( gprof_pthread_create(&thread, attr, function, arg) != 0) {
    GLOBAL_DEBUG_STREAM2("pthread_create failed!" << std::endl);
    throw (runtime_error("not enough system resources to create a process for the new thread"));
  }
  else {
    GLOBAL_DEBUG_STREAM2("gprof_pthread_create successful!" << std::endl);
  }
#else
  if ((fifo_scheduling_enabled) || (rr_scheduling_enabled)) {
    createDefaultAttribute();
    attr = &custom_sched_attr;
  }
  if ( pthread_create(&thread, attr, function, arg) != 0) {
    GLOBAL_ERROR_STREAM2("pthread_create failed!" << std::endl);
    throw (runtime_error("not enough system resources to create a process for the new thread"));
  }
  //getSchedPriotity();
#endif
}


Result Thread::join() {
  if (pthread_join(thread, 0) == 0)
    return SUCCESS;
  return FAILURE;
}

void Thread::kill()
{
  pthread_kill(thread, 9);
}

void Thread::enableCustomScheduling(Scheduling_Policy policy) {
  switch (policy) {
    case FIFO:
      fifo_scheduling_enabled = true;
      rr_scheduling_enabled   = false;
      break;
    case ROUND_ROBIN:
      fifo_scheduling_enabled = false;
      rr_scheduling_enabled   = true;
      break;
    case OTHER:
      fifo_scheduling_enabled = false;
      rr_scheduling_enabled   = false;
      break;
  }
}

void Thread::createDefaultAttribute() {
  GLOBAL_DEBUG_STREAM2("creating default pthread attribute" << std::endl);

  pthread_attr_init(&custom_sched_attr);
  pthread_attr_setinheritsched(&custom_sched_attr, PTHREAD_EXPLICIT_SCHED);
  if (rr_scheduling_enabled) {
      pthread_attr_setschedpolicy(&custom_sched_attr, SCHED_RR);
      rr_param.sched_priority = sched_get_priority_max(SCHED_RR);
      pthread_attr_setschedparam(&custom_sched_attr, &rr_param);
      GLOBAL_DEBUG_STREAM2("default pthread attribute (EXPLICIT ROUND ROBIN SCHEDULING) created" << std::endl);
  }
  if (fifo_scheduling_enabled) {
      pthread_attr_setschedpolicy(&custom_sched_attr, SCHED_FIFO);
      fifo_param.sched_priority = sched_get_priority_max(SCHED_FIFO);
      pthread_attr_setschedparam(&custom_sched_attr, &fifo_param);
      GLOBAL_DEBUG_STREAM2("default pthread attribute (EXPLICIT FIFO SCHEDULING) created" << std::endl);
  }
}

void Thread::setMinPriority() {
  // get the lowest possible priority
  int sched_priority;
  if (rr_scheduling_enabled) {
      sched_priority = sched_get_priority_min(SCHED_RR);
      rr_param.sched_priority = sched_priority;
      pthread_setschedparam(thread, SCHED_RR, &rr_param);
      getSchedPriotity();
//      GLOBAL_DEBUG_STREAM2("setting pthread priority to minimum ("<<  sched_priority << ")" << std::endl);
  }
  if (fifo_scheduling_enabled) {
      sched_priority = sched_get_priority_min(SCHED_FIFO);
      fifo_param.sched_priority = sched_priority;
      pthread_setschedparam(thread, SCHED_FIFO, &fifo_param);
      getSchedPriotity();
//      GLOBAL_DEBUG_STREAM2("setting pthread priority to minimum ("<<  sched_priority << ")" << std::endl);
  }
}
void Thread::setMidPriority() {
  // compute middle priority
  int sched_priority;
  if (rr_scheduling_enabled) {
      sched_priority = (sched_get_priority_min(SCHED_RR) + sched_get_priority_max(SCHED_RR)) / 2;
      rr_param.sched_priority = sched_priority;
      pthread_setschedparam(thread, SCHED_RR, &rr_param);
      getSchedPriotity();
//      GLOBAL_DEBUG_STREAM2("setting pthread middle priority ("<<  sched_priority << ")" << std::endl);
  }
  if (fifo_scheduling_enabled) {
      sched_priority = (sched_get_priority_min(SCHED_FIFO) + sched_get_priority_max(SCHED_FIFO)) / 2;
      fifo_param.sched_priority = sched_priority;
      pthread_setschedparam(thread, SCHED_FIFO, &fifo_param);
      getSchedPriotity();
//      GLOBAL_DEBUG_STREAM2("setting pthread middle priority ("<<  sched_priority << ")" << std::endl);
  }
}
void Thread::setMaxPriority() {
  // get the highest possible priority
  int sched_priority;
  if (rr_scheduling_enabled) {
      sched_priority = sched_get_priority_max(SCHED_RR);
      rr_param.sched_priority = sched_priority;
      pthread_setschedparam(thread, SCHED_RR, &rr_param);
      getSchedPriotity();
//      GLOBAL_DEBUG_STREAM2("setting pthread priority to maximum ("<<  sched_priority << ")" << std::endl);
  }
  if (fifo_scheduling_enabled) {
      sched_priority = sched_get_priority_max(SCHED_FIFO);
      fifo_param.sched_priority = sched_priority;
      pthread_setschedparam(thread, SCHED_FIFO, &fifo_param);
      getSchedPriotity();
//      GLOBAL_DEBUG_STREAM2("setting pthread priority to maximum ("<<  sched_priority << ")" << std::endl);
  }
}

int Thread::getSchedPriotity() {
  int returned_policy;
  struct sched_param parameters;
  int return_value = pthread_getschedparam(thread, &returned_policy, &parameters);
  if (return_value != 0) {
    GLOBAL_ERROR_STREAM2("pthread getsched failed!" << std::endl);
  }
  switch (returned_policy) {
    case SCHED_RR:
      GLOBAL_DEBUG_STREAM2( "pthread priority : " << parameters.sched_priority << " (round-robin scheduling)" << std::endl);
      break;
    case SCHED_FIFO:
      GLOBAL_DEBUG_STREAM2( "pthread priority : " << parameters.sched_priority << " (FIFO scheduling)" << std::endl);
      break;
    case SCHED_OTHER:
      GLOBAL_DEBUG_STREAM2( "pthread priority : " << parameters.sched_priority << " (other scheduling)" << std::endl);
      break;
  }
  return parameters.sched_priority;
}

void Thread::suspendThread(long millis) {
  struct ::timeval tv;
  if (millis > 0) {
    tv.tv_sec = millis / 1000;
    tv.tv_usec = (millis % 1000) * 1000;
    select(0, 0, 0, 0, &tv);
  }
}

}

#include "nmm/nmm_suffix.hpp"
