/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "Interval.hpp"
#include "Time.hpp"
#include "Rational.hpp"

#include <sstream>
#include <cmath>

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  Interval::Interval() {
    nsec = 0;
    sec  = 0;
  };

  Interval::Interval(const nmm_int64& nanoseconds) {
    nsec = nanoseconds % nano_per_sec;
    sec  = nanoseconds / nano_per_sec;

    if(nsec < 0) {
      nsec += nano_per_sec;
      sec--;
    }
  };

  Interval::Interval(const nmm_int64& seconds, const nmm_int64& nanoseconds)
  {
    nsec = nanoseconds;
    sec  = seconds;

    while(nsec < 0) {
      nsec += nano_per_sec;
      sec--;
    }
    while(nsec > nano_per_sec) {
      nsec -= nano_per_sec;
      sec++;
    }
  };

  Interval::Interval(const struct timeval& time) {
    nsec = time.tv_usec * 1000L;
    sec  = time.tv_sec;

    while(nsec < 0) {
      nsec += nano_per_sec;
      sec--;
    }
    while(nsec > nano_per_sec) {
      nsec -= nano_per_sec;
      sec++;
    }
  };

  Interval::Interval(const Rational& time)
  {
    sec = time.getNumerator() / time.getDenominator();
    double nanosec = static_cast<double>(time.getNumerator()) / time.getDenominator();
    nanosec -= sec;          // so the absolute value of nanosec is less than 1.
    nanosec *= nano_per_sec;

    nsec = static_cast<nmm_int64>(nanosec);

    if(nsec < 0) {
      nsec += nano_per_sec;
      sec--;
    }
  };

  Interval& Interval::operator+=(const Interval& interval) {
    nsec += interval.nsec;
    sec  += interval.sec;
    if(nsec >= nano_per_sec) {
      sec++;
      nsec -= nano_per_sec;
    }
    return *this;
  };

  Interval& Interval::operator-=(const Interval& interval) {
    nsec -= interval.nsec;
    sec  -= interval.sec;
    if(nsec < 0) {
      sec--;
      nsec += nano_per_sec;
    }
    return *this;
  };

  Interval& Interval::operator*=(const float& a)
  {
    double p = a * (double(sec) + double(nsec) / nano_per_sec);
    if(p >= 0) {
      sec = static_cast<int64_t>(p);
    } else {
      sec = static_cast<int64_t>(p) - 1;
    }
    nsec = static_cast<int64_t>((p - sec) * nano_per_sec);

    return *this;
  };

  Interval& Interval::operator/=(const int& a)
    throw (const IntervalDivisionByZeroException&)
  {
    if(a == 0) throw IntervalDivisionByZeroException();

    int64_t seconds     = sec  / static_cast<int64_t>(a);
    int64_t nanoseconds = nsec / static_cast<int64_t>(a);
    nanoseconds += ((sec % static_cast<int64_t>(a)) * nano_per_sec) / static_cast<int64_t>(a);
    
    while(nanoseconds < 0) {
      nanoseconds += nano_per_sec;
      seconds--;
    }
    while(nanoseconds > nano_per_sec) {
      nanoseconds -= nano_per_sec;
      seconds++;
    }

    sec  = seconds;
    nsec = nanoseconds;

    return *this;
  };

  Interval operator-(const Interval& interval) {
    Interval result;
    result -= interval;
    
    return result;
  };

  Interval operator+(const Interval& a, const Interval& b) {
    Interval result = a;
    result += b;

    return result;
  };

  Interval operator-(const Interval& a, const Interval& b) {
    Interval result = a;
    result -= b;

    return result;
  };

  Interval operator- (const Time& a, const Time& b) {
    Interval result;
    result.nsec = a.nsec - b.nsec;
    result.sec  = a.sec  - b.sec;
    if(result.nsec < 0) {
      result.nsec += nano_per_sec;
      result.sec--;
    }
    return result;
  };

  void suspendThread(const Interval& interval) {
    if(interval > Interval(0)) {
      struct timeval tv;
      tv.tv_usec = interval.nsec / 1000;
      tv.tv_sec  = interval.sec;
      select(0, 0, 0, 0, &tv);
    }
  };


  std::ostream& operator<<(std::ostream& s, const Interval& interval)
  {
    if (interval.sec < 0) {
      if (interval.sec == -1) {
	s << "-";
      }
      return s << interval.sec + 1 << " sec, " << nano_per_sec - interval.nsec << " nanosec";
    }
    return s << interval.sec << " sec, " << interval.nsec << " nanosec";
  };
  
  std::string Interval::getPrettyString() const {
    std::ostringstream strstr;
    strstr.width(4);
    strstr.fill('0');
    strstr << sec << ".";
    strstr.width(9);
    strstr.fill('0');
    strstr << nsec;
    return strstr.str();  
  }
}

#include "nmm/nmm_suffix.hpp"
