/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "LocalTime.hpp"
#include "Time.hpp"

#include <iostream>
#include <time.h>
#include <locale.h>
#include "timeutils.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  LocalTime::LocalTime()
    : __second(0),
      __minute(0),
      __hour(0),
      __day_of_month(0),
      __day_of_week(0),
      __day_of_year(0),
      __month(0),
      __gmt_offset(0),
      __year(0),
      __is_dst(0),
      __time_zone()
  {
    memset(&__tm_struct, 0, sizeof(struct ::tm));
  }
  
  LocalTime::LocalTime(const Time& time, bool local /*=true*/)
    : __second(0),
      __minute(0),
      __hour(0),
      __day_of_month(0),
      __day_of_week(0),
      __day_of_year(0),
      __month(0),
      __gmt_offset(0),
      __year(0),
      __is_dst(0),
      __time_zone()
  {
    reset(time, local);
  }
  
  LocalTime::LocalTime(const struct ::tm& tm_struct)
    : __second(0),
      __minute(0),
      __hour(0),
      __day_of_month(0),
      __day_of_week(0),
      __day_of_year(0),
      __month(0),
      __gmt_offset(0),
      __year(0),
      __is_dst(0),
      __time_zone()
  {
    reset(tm_struct);
  }
  
  LocalTime::LocalTime(const LocalTime& other)
    : __second(other.__second),
      __minute(other.__minute),
      __hour(other.__hour),
      __day_of_month(other.__day_of_month),
      __day_of_week(other.__day_of_week),
      __day_of_year(other.__day_of_year),
      __month(other.__month),
      __gmt_offset(other.__gmt_offset),
      __year(other.__year),
      __is_dst(other.__is_dst),
      __time_zone(other.__time_zone)
  {
    updateAttributes();
  }
  
  void LocalTime::reset(const struct ::tm& tm_struct) {
    __second = tm_struct.tm_sec;
    __minute = tm_struct.tm_min;
    __hour = tm_struct.tm_hour;
    __day_of_month = tm_struct.tm_mday;
    __day_of_week = tm_struct.tm_wday;
    __day_of_year = tm_struct.tm_yday;
    __month = tm_struct.tm_mon;
    __year = tm_struct.tm_year;
    
    // XXX: fw: if tm_isdst is negative, the information is not available!
    __is_dst = tm_struct.tm_isdst == 1;
    
    // XXX: fw: the following fields are GNU extensions!
#ifndef __CYGWIN__
    __gmt_offset = tm_struct.tm_gmtoff;
    __time_zone = std::string(tm_struct.tm_zone);
#endif    
    memcpy(&__tm_struct, &tm_struct, sizeof(struct ::tm));
  }
  
  void LocalTime::reset(const Time& time, bool local /*=true*/) {
    struct ::tm tm_value;
    
    // NOTE: fw: there seems to be no conversion function, but the manual says that the VALUE of time_t
    // NOTE: fw: is always the number of seconds since the epoch (like timeval.tv_sec???).
    // NOTE: fw: only the TYPE of time_t is not known, but the manual also says it is either a floating point
    // NOTE: fw: or integer type, so this conversion should be possible.
    time_t ct = static_cast<time_t>(time.sec);
    
    if (local) {
      ::localtime_r(&ct, &tm_value);
    } else {
      ::gmtime_r(&ct, &tm_value);
    }
    reset(tm_value);
    
  }
  
  std::string LocalTime::toString(const std::string& format) const {
    char buffer[4096];
    size_t size = sizeof(buffer);
    ::strftime(buffer, size, format.c_str(), &__tm_struct);
    return std::string(buffer);
  }
  
  void LocalTime::updateAttributes() {
    __tm_struct.tm_sec = __second;
    __tm_struct.tm_min = __minute;
    __tm_struct.tm_hour = __hour;
    __tm_struct.tm_mday = __day_of_month;
    __tm_struct.tm_mon = __month;
    __tm_struct.tm_year = __year;
    __tm_struct.tm_wday = __day_of_week;
    __tm_struct.tm_isdst = __is_dst;
    
    // XXX: fw: the following are GNU extensions
#ifndef __CYGWIN__
    __tm_struct.tm_gmtoff = __gmt_offset;
    
    // XXX: fw: does this really work?
    __tm_struct.tm_zone = const_cast<char*>(__time_zone.c_str());
#endif
  }
  
  void LocalTime::toGMT(){
    if(getGMTOffset() != 0){
      struct tm tmp(__tm_struct);
      time_t t = mktime(&tmp);  
#ifndef __CYGWIN__
      if((int)getGMTOffset() != tmp.tm_gmtoff)
        t += tmp.tm_gmtoff;
#endif
      gmtime_r(&t, &tmp);
      reset(tmp);
    }
  }
  
  void LocalTime::toLocalTimeZone(){
    struct tm tmp(__tm_struct);
    time_t t = mktime(&tmp);
 
    //determine current GMTOffset
    time_t cur = time(NULL);
    struct tm *curtm = localtime(&cur); 
       
 #ifndef __CYGWIN__   
  if((int)getGMTOffset() != tmp.tm_gmtoff){
      if (tmp.tm_gmtoff != curtm->tm_gmtoff){
        t += tmp.tm_gmtoff - curtm->tm_gmtoff;
      }else{
        t += tmp.tm_gmtoff;
      }
    }
#endif
    localtime_r(&t, &tmp);
    reset(tmp);
  }
 
  time_t LocalTime::getTime_t() const{
    struct tm tmp(__tm_struct);
    time_t t = mktime(&tmp);
    
    //determine current GMTOffset
    time_t cur = time(NULL);
    struct tm *curtm = localtime(&cur); 
   #ifndef __CYGWIN__
    //calculate time with respect to GMTOffset
    //because this is not done automatically
    if((int)getGMTOffset() != tmp.tm_gmtoff){
      if (tmp.tm_gmtoff != curtm->tm_gmtoff){
        t += tmp.tm_gmtoff - curtm->tm_gmtoff;
      }else{
        t += tmp.tm_gmtoff;
      }
    }
   #endif
    return t;
  }
  
  bool LocalTime::operator<(const LocalTime &lt) const {
    //need to copy struct tm because mktime updates internal variables of struct
    //struct is converted ti time_t for easier comparison
    struct tm st1 = __tm_struct;
    time_t t1 = mktime(&st1);
    struct tm st2 = lt.__tm_struct;
    time_t t2 = mktime(&st2);
    return t1 < t2;
  }
  
  bool LocalTime::operator==(const LocalTime &lt) const {
    //need to copy struct tm because mktime updates internal variables of struct
    //struct is converted ti time_t for easier comparison
    struct tm st1 = __tm_struct;
    time_t t1 = mktime(&st1);
    struct tm st2 = lt.__tm_struct;
    time_t t2 = mktime(&st2);
    return t1 == t2;
  }
 
  LocalTime LocalTime::operator-(const LocalTime &lt) {
    //need to copy struct tm because mktime updates internal variables of struct
    //struct is converted ti time_t for easier calculation
    struct tm st1 = __tm_struct;
    time_t t1 = mktime(&st1);
    struct tm st2 = lt.__tm_struct;
    time_t t2 = mktime(&st2);
 
    time_t ret = t1 - t2;
    st2 = *(localtime(&ret));
    return st2;
  }
  
  LocalTime LocalTime::operator+(const LocalTime &lt) {
    //need to copy struct tm because mktime updates internal variables of struct
    //struct is converted ti time_t for easier calculation
    struct tm st1 = __tm_struct;
    time_t t1 = mktime(&st1);
    struct tm st2 = lt.__tm_struct;
    time_t t2 = mktime(&st2);
     
    time_t ret = t1 + t2;
    st2 = *(localtime(&ret));
    return st2;
 
  }
  
  LocalTime LocalTime::operator-(const int &sub) {
    //need to copy struct tm because mktime updates internal variables of struct
    //struct is converted ti time_t for easier calculation
    struct tm st1 = __tm_struct;
    time_t t1 = mktime(&st1);
 
    time_t ret = t1 - sub;
    st1 = *(localtime(&ret));
    return st1;
  }
  
  LocalTime LocalTime::operator+(const int &add) {
    //need to copy struct tm because mktime updates internal variables of struct
    //struct is converted ti time_t for easier calculation
    struct ::tm st1(__tm_struct);
    time_t t1 = mktime(&st1);
 
    time_t ret = t1 + add;
    localtime_r(&ret, &st1);
    return st1;
  }

  void LocalTime::setSecond(uint8_t s) {
    __second = s;
    updateAttributes();
    mktime(&__tm_struct);
    reset(__tm_struct);
  }

  void LocalTime::setMinute(uint8_t m) {
    __minute = m;
    updateAttributes();
    mktime(&__tm_struct);
    reset(__tm_struct);
  }

  void LocalTime::setHour(uint8_t h) {
    __hour = h;
    updateAttributes();
    mktime(&__tm_struct);
    reset(__tm_struct);
  }

  void LocalTime::setDayOfMonth(uint8_t d) {
    __day_of_month = d;
    updateAttributes();
    mktime(&__tm_struct);
    reset(__tm_struct);
  }
  
  void LocalTime::setMonth(uint8_t m) {
    __month = m;
    updateAttributes();
    reset(__tm_struct);
  }
  
  void LocalTime::setYear(unsigned int y) {
    __year = y;
    updateAttributes();
    reset(__tm_struct);
  }
  
  void LocalTime::setGMTOffset(unsigned gmt_offset) {
    __gmt_offset = gmt_offset;
    updateAttributes();
    reset(__tm_struct);
  }
  
  void LocalTime::setDaylightSavingsTime(bool is_dst) {
    __is_dst = is_dst;
    updateAttributes();
    reset(__tm_struct);
  }
  
}

#include "nmm/nmm_suffix.hpp"
