/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */
#include "Rational.hpp"

#include <iostream>
using std::cerr;
using std::endl;

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  RationalDivisionByZeroException::RationalDivisionByZeroException() {
    cerr << "Rational - Division by zero." << endl;
  }
  
  nmm_int64 Rational::greatestCommonDivisor(nmm_int64 a, nmm_int64 b) 
  {
    if(a == 0 || b == 0) return 0;  // gcd(0, x) = gcd(x, 0) = 0

    if(a < 0) {      // make sure that the two integers are non-negative
      a = -a;
    }

    if(b < 0) {
      b = -b;
    }

    nmm_int64 t = 1;       // and perform Euklid's algorithm
    while(t != 0) {
      t = a % b;
      a = b;
      b = t;
    }

    return a;
  };

  nmm_int64 Rational::leastCommonMultiple(nmm_int64 a, nmm_int64 b)
  {
    if(a == 0 || b == 0) return 0;

    nmm_int64 lcm = a / greatestCommonDivisor(a, b) * b;
    if(lcm < 0) {
      return -lcm;
    }
    return lcm;
  };

  void Rational::reduce() 
  {
    if(numerator == 0) {
      denominator = 1;
      return;
    }

    nmm_int64 gcd = greatestCommonDivisor(numerator, denominator);
    numerator   /= gcd;
    denominator /= gcd;
  };

  void Rational::invert()
    throw (const RationalDivisionByZeroException&)
  {
    if (numerator == 0) throw RationalDivisionByZeroException();

    nmm_int64 a = numerator;
    numerator = denominator;
    denominator = a;
  };

  Rational Rational::getReciprocal()
    throw (const RationalDivisionByZeroException&)
  {
    if (numerator == 0) throw RationalDivisionByZeroException();

    return Rational(denominator, numerator);
  };

  Rational::Rational(const nmm_int64& numerator, const nmm_int64& denominator)
    throw (const RationalDivisionByZeroException&) 
  {
    if(denominator == 0) {
      this->numerator = 0;    // although an exception is thrown,
      this->denominator = 1;  // set the fraction to the value zero.
      throw RationalDivisionByZeroException();
    } else {
      if(denominator < 0) {                // the numerator should carry the sign 
	this->denominator = - denominator; // of the fraction, so the denominator
	this->numerator   = - numerator;   // is made positive
      } else {
	this->numerator   = numerator;
	this->denominator = denominator;
      }
    }
  };

  Rational& Rational::operator- ()
  {
    numerator = -numerator;

    return *this;
  };

  Rational& Rational::operator+=(const Rational& b)
  {
    numerator = numerator * b.denominator + b.numerator * denominator;
    denominator *= b.denominator;
    reduce();

    return *this;
  };

  Rational& Rational::operator-=(const Rational& b)
  {
    numerator = numerator * b.denominator - b.numerator * denominator;
    denominator *= b.denominator;
    reduce();

    return *this;
  };

  Rational& Rational::operator*=(const Rational& b)
  {
    numerator   *= b.numerator;
    denominator *= b.denominator;
    reduce();

    return *this;
  };

  Rational& Rational::operator/=(const Rational& b)
    throw (const RationalDivisionByZeroException&)
  {
    if(b.numerator == 0) throw RationalDivisionByZeroException();

    numerator   *= b.denominator;
    denominator *= b.numerator;
    if(denominator < 0) {
      denominator = -denominator;
      numerator   = -numerator;
    }
    reduce();

    return *this;
  };

  Rational operator+ (const Rational& a, const Rational& b)
  {
    Rational sum = a;
    return sum += b;
  };

  Rational operator- (const Rational& a, const Rational& b)
  {
    Rational diff = a;
    return diff -= b;
  };

  Rational operator* (const Rational& a, const Rational& b)
  {
    Rational prod = a;
    return prod *= b;
  };

  Rational operator/ (const Rational& a, const Rational& b)
  {
    Rational quot = a;
    return quot /= b;
  };

  std::ostream& operator<< (std::ostream& s, const Rational& a)
  {
    return s << a.numerator << " / " << a.denominator;
  }
}

#include "nmm/nmm_suffix.hpp"
