/* Network-Integrated Multimedia Middleware (NMM)
 * http://www.motama.com/nmm.html
 *  
 * Copyright (C) 2006-2007
 * Motama GmbH, Saarbruecken, Germany
 * http://www.motama.com
 *  
 * This file is part of NMM. 
 * 
 * NMM is free software; you can redistribute it and/or modify it    
 * under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; only version 2 of the License.      
 * 
 * NMM is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY 
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public 
 * License for more details.  
 * 
 * You should have received a copy of the GNU General Public License 
 * along with NMM; if not, write to the Free Software Foundation,    
 * Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.
 *
 */

#include "Time.hpp"
#include "Rational.hpp"
#include "Interval.hpp"

#include <sstream>
#include <iostream>

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  Time::Time() 
  {
    nsec = 0;
    sec  = 0;
  };
  
  Time::Time(const nmm_int64& nanoseconds) 
  {
    nsec = nanoseconds % nano_per_sec;
    sec  = nanoseconds / nano_per_sec;
    
    if(nsec < 0) {
      nsec += nano_per_sec;
      sec--;
    }
  };

  Time::Time(const nmm_int64& seconds, const nmm_int64& nanoseconds)
  {
    nsec = nanoseconds;
    sec  = seconds;
    
    while(nsec < 0) {
      nsec += nano_per_sec;
      sec--;
    }
    while(nsec > nano_per_sec) {
      nsec -= nano_per_sec;
      sec++;
    }
  };
  
  Time::Time(const struct timeval& time) 
  {
    nsec = time.tv_usec * 1000L;
    sec  = time.tv_sec;
    
    while(nsec < 0) {
      nsec += nano_per_sec;
      sec--;
    }
    while(nsec > nano_per_sec) {
      nsec -= nano_per_sec;
      sec++;
    }
  };

  Time::Time(const Rational& time)
  {
    sec = time.getNumerator() / time.getDenominator();
    double nanosec = static_cast<double>(time.getNumerator()) / time.getDenominator();
    nanosec -= sec;          // so the absolute value of nanosec is less than 1.
    nanosec *= nano_per_sec;

    nsec = static_cast<nmm_int64>(nanosec);

    if(nsec < 0) {
      nsec += nano_per_sec;
      sec--;
    }
  };

  Time& Time::operator+=(const Interval& interval) 
  {
    nsec += interval.nsec;
    sec  += interval.sec;
    if(nsec >= nano_per_sec) {
      sec++;
      nsec -= nano_per_sec;
    }

    return *this;
  };

  Time& Time::operator-=(const Interval& interval) 
  {
    nsec -= interval.nsec;
    sec  -= interval.sec;
    if(nsec < 0) {
      sec--;
      nsec += nano_per_sec;
    }

    return *this;
  };

  Time operator+(const Time& a, const Interval& b) {
    Time result = a;
    result += b;

    return result;
  };

  Time operator-(const Time& a, const Interval& b) {
    Time result = a;
    result -= b;
    
    return result;
  };

  std::string Time::getPrettyString() const {
    std::ostringstream strstr;
    strstr.width(4);
    strstr.fill('0');
    strstr << sec << ".";
    strstr.width(9);
    strstr.fill('0');
    strstr << nsec;
    return strstr.str();
  }


  std::ostream& operator<<(std::ostream& s, const Time& time)
  {
    if (time.sec < 0) {
      if (time.sec == -1) {
	s << "-";
      }
      return s << time.sec + 1 << " sec, " << nano_per_sec - time.nsec << " nanosec";
    }
    return s << time.sec << " sec, " << time.nsec << " nanosec";
  };

  
  IntervalDivisionByZeroException::IntervalDivisionByZeroException() {
    std::cerr << "Interval - Division by zero exception." << std::endl;
  };
  
}

#include "nmm/nmm_suffix.hpp"
