/* libxml++
 * Copyright (C) 2000  Ari Johnson
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 */

/* Modified by Motama GmbH, Copyright (C) 2006-2007 Motama GmbH */


#include "XMLNode.hpp"

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  XMLNode::XMLNode(const std::string &n)
    : _name(n), _content(std::string()) {
    if(_name.empty())
      _initialized = false;
    else
      _initialized = true;
  }

  XMLNode::XMLNode(const std::string &n, const std::string &c)
    : _name(n), _content(c) {
    _initialized = true;
  }
  
  XMLNode::XMLNode(const XMLNode &node) {
    *this = node;
  }

  const XMLNode& XMLNode::operator=(const XMLNode& node) {
    erase(); //remove all old nodes and properties

    _name = node.name();
    set_content(node.content());

    XMLNodeConstIterator curnode;
    XMLPropertyConstIterator curprop;

    for(curprop = node.properties().begin(); curprop != node.properties().end(); ++curprop)
      add_property((*curprop)->name(), (*curprop)->value());

    for(curnode = node.children().begin(); curnode != node.children().end(); ++curnode)
      add_child(new XMLNode(**curnode));

    return *this;
  }
  
  bool XMLNode::operator==(const XMLNode& ) {
    return false;
  }

  XMLNode::~XMLNode() {
    erase();
  }

  void XMLNode::erase() {
    XMLNodeIterator curchild;
    XMLPropertyIterator curprop;

    for(curchild = _children.begin(); curchild != _children.end(); ++curchild)
      delete *curchild;

    for(curprop = _proplist.begin(); curprop != _proplist.end(); ++curprop)
      delete *curprop;
  }

  const std::string & XMLNode::set_content(const std::string &c) {
    _content = c;

    return _content;
  }

  XMLNodeList XMLNode::children(const std::string &n) const {
    XMLNodeList retval;

    if(n.length() == 0)
      return _children;

    for(XMLNodeConstIterator cur = _children.begin(); cur != _children.end(); ++cur) {
      /*Begin> If name matches, insert the child */
      if((*cur)->name() == n)
        retval.push_back(*cur);
      /*End< */

      /*Begin> Else search the rest of the children for a given name*/
      else {
        XMLNodeList tmp_list = (*cur)->children(n);
        retval.insert(retval.end(), tmp_list.begin(), tmp_list.end());
      }
      /*End< */
    }

    return retval;
  }

  XMLNode *XMLNode::add_child(const std::string &n) {
    XMLNode *tmp;

    tmp = new XMLNode(n);

    return add_child(tmp);
  }

  XMLNode *XMLNode::add_child(XMLNode *n) {
    if(!n)
      return NULL;

    _children.insert(_children.end(), n);
    return n;
  }

  XMLNode *XMLNode::add_child(const std::string &n, int pos) {
    XMLNode *tmp;
  
    tmp = new XMLNode(n);

    if(!tmp)
      return NULL;
    XMLNodeIterator i = _children.begin();
    advance(i, pos);

    _children.insert(i, tmp);
    return tmp;
  }

  XMLNode *XMLNode::add_content(const std::string &c) {
    XMLNode *tmp;

    tmp = new XMLNode(std::string(), c);

    return add_child(tmp);
  }

  std::string XMLNode::getData() {
    XMLNodeList pcdata = children("text");
    XMLNodeIterator i = pcdata.begin();
    if ( i != pcdata.end() ) {
      XMLNode* i_data = *i;
      return( i_data->content() );
    }
    return ("");
  }

  void XMLNode::remove_child(XMLNode *n) {
    if(n)
      _children.remove(n);
  }

  XMLProperty *XMLNode::property(const std::string &n) {
    if(_prophash.find(n) == _prophash.end())
      return NULL;

    return _prophash[n];
  }

  XMLProperty *XMLNode::add_property(const std::string &n, const std::string &v) {
    XMLProperty *tmp;

    if(_prophash.find(n) != _prophash.end())
      return NULL;

    tmp = new XMLProperty(n, v);

    if(!tmp)
      return NULL;

    _prophash[tmp->name()] = tmp;
    _proplist.insert(_proplist.end(), tmp);

    return tmp;
  }

  void XMLNode::remove_property(const std::string &n) {
    if(_prophash.find(n) != _prophash.end()) {
      _proplist.remove(_prophash[n]);                   //memory leak!! Property is not deleted;
      _prophash.erase(n);
    }
  }
  
}

#include "nmm/nmm_suffix.hpp"
