/* libxml++
 * Copyright (C) 2000  Ari Johnson
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301  USA
 */

/* Modified by Motama GmbH, Copyright (C) 2006-2007 Motama GmbH */


#include "XMLTree.hpp"
#include "XMLNode.hpp"

#include <libxml/parser.h>
#include <libxml/encoding.h>

#include "nmm/nmm_prefix.hpp"

namespace NMM {
  
  static XMLNode *readnode(xmlNodePtr);
  static void writenode(xmlDocPtr, XMLNode *, xmlNodePtr, int = 0);

  XMLTree::XMLTree(const XMLTree& tree) {
    *this = tree;
  }
  
  const XMLTree& XMLTree::operator=(const XMLTree& tree) {
    delete _root;
    _root = 0;
    if (tree.root())
      _root = new XMLNode(*tree.root());

    _filename     = tree.filename();
    _compression  = tree.compression();
    _initialized  = true;

    return *this;
  }

  bool XMLTree::operator==(const XMLTree& tree) {
    if (tree.root()) {
      if (root()) {
        return (*root() == *tree.root());
      } else {
        return false;
      }
    } else {
      if(!root()) //both have no root node so they are equal
        return true;
    }
    return false;
  }

  XMLTree::~XMLTree() {
    delete _root;
  }

  XMLNode* XMLTree::set_root(XMLNode* n) {
    delete _root;
    return _root = n;
  }

  int XMLTree::set_compression(int c) {
    if(c > 9)
      c = 9;

    if(c < 0)
      c = 0;

    _compression = c;

    return _compression;
  }

  bool XMLTree::read(void) {
    xmlDocPtr doc;

    if(_root) {
      delete _root;
      _root = NULL;
    }

    xmlKeepBlanksDefault(0);

    doc = xmlParseFile(_filename.c_str());
    if(!doc) {
      _initialized = false;
      return false;
    }

    _root = readnode(xmlDocGetRootElement(doc));
    xmlFreeDoc(doc);
    _initialized = true;

    return true;
  }

  bool XMLTree::read_buffer(const std::string &buffer) {
    xmlDocPtr doc;

    _filename = "";

    if(_root) {
      delete _root;
      _root = NULL;
    }

    doc = xmlParseMemory((char *) buffer.c_str(), buffer.length());
    if(!doc) {
      _initialized = false;
      return false;
    }

    _root = readnode(xmlDocGetRootElement(doc));
    xmlFreeDoc(doc);
    _initialized = true;

    return true;
  }

  bool XMLTree::write(void) const {
    xmlDocPtr doc;
    XMLNodeList children;
    int result;

    xmlKeepBlanksDefault(0);
    doc = xmlNewDoc((xmlChar *) "1.0");
    xmlSetDocCompressMode(doc, _compression);
    writenode(doc, _root, doc->children, 1);
    result = xmlSaveFormatFile(_filename.c_str(), doc, 1);
    xmlFreeDoc(doc);

    if(result == -1)
      return false;

    return true;
  }

  std::string XMLTree::write_buffer(void) const {
    std::string retval;
    char *ptr;
    int len;
    xmlDocPtr doc;
    XMLNodeList children;

    xmlKeepBlanksDefault(0);
    doc = xmlNewDoc((xmlChar *) "1.0");
    xmlSetDocCompressMode(doc, _compression);
    writenode(doc, _root, doc->children, 1);
    xmlDocDumpMemory(doc, (xmlChar **) &ptr, &len);
    xmlFreeDoc(doc);

    retval = ptr;

    xmlFree(ptr);

    return retval;
  };

  XMLNodeList XMLTree::children(const std::string& /*name*/) const {
    if (root())
      return root()->children();

    return XMLNodeList(); //returns an empty list if no root node available
  }
  
  
  static XMLNode *readnode(xmlNodePtr node) {
    std::string name, content;
    xmlNodePtr child;
    XMLNode *tmp;
    xmlAttrPtr attr;

    if(node->name)
      name = (char *) node->name;
    else
      name = std::string();

    tmp = new XMLNode(name);

    for(attr = node->properties; attr; attr = attr->next) {
      name = (char *) attr->name;
      content = "";
      if(attr->children)
        content = (char *) attr->children->content;
      tmp->add_property(name, content);
    }

    if(node->content)
      tmp->set_content((char *) node->content);
    else
      tmp->set_content(std::string());

    for(child = node->children; child; child = child->next)
      tmp->add_child(readnode(child));

    return tmp;
  }


  static void writenode(xmlDocPtr doc, XMLNode *n, xmlNodePtr p, int root ) {
    XMLPropertyList props;
    XMLPropertyIterator curprop;
    XMLNodeList children;
    XMLNodeIterator curchild;
    xmlNodePtr node;

    if(root)
      node = doc->children = xmlNewDocNode(doc, NULL,
                                           (xmlChar *) n->name().c_str(),
                                           (xmlChar *) n->content().c_str());
    else
      node = xmlNewChild(p, NULL, (xmlChar *) n->name().c_str(),
                         (xmlChar *) n->content().c_str());

    if(n->is_content()) {
      node->type = XML_TEXT_NODE;
      xmlNodeSetContentLen(node, (const xmlChar *) n->content().c_str(),
                           n->content().length());
    }

    props = n->properties();
    for(curprop = props.begin(); curprop != props.end(); curprop++)
      xmlSetProp(node, (xmlChar *) (*curprop)->name().c_str(),
                 (xmlChar *) (*curprop)->value().c_str());

    children = n->children();
    for(curchild = children.begin(); curchild != children.end(); curchild++)
      writenode(doc, *curchild, node);
  }
  
}

#include "nmm/nmm_suffix.hpp"
